 td > p > blockquote{
    margin-left: 0 !important;
    line-height: 0 !important;
    margin-bottom: 0 !important;
} 

.section >  p  > blockquote{
    margin-top: 10px;
    margin-left: 0 !important;
    line-height: 0 !important;
    margin-bottom: 0 !important;
} 

a {
	cursor: pointer;
}

.description a {
    cursor: text;
}
.info {
  opacity: 0.45;  
  padding-left: 4px;
  /* position: absolute; */
  /* margin-top: 3px; */
  margin-top: -1px;
  margin-left: -2px; 
  width: auto;
}
.hljs-string {
    color: black;
}

textarea {
	height: 250px;
}

pre a:visited {color:#2980B9;}

.xml-element-name {
    color: darkred !important;
}

.xml-simpletype-name {
    color: green !important;
}

.xml-attribute-name {
    color: darkblue !important;
}

div.tab {overflow: hidden;background-color: inherit;}

/* Style the tab content */
.tabcontent {
  display: none;
  }
  
/*pre*/
.tabcontent pre {
  margin-top: 0;
  margin-bottom: 0;
  overflow: auto;
  -ms-word-wrap: normal;
  word-wrap: normal;
  overflow-wrap: normal;
  white-space: pre;
  max-height: 340px;
  }

/* Style the buttons inside the tab */
div.tab button {
  float: left;
  border: none;
  outline: none;
  cursor: pointer;
  padding: 8px 14px;
  transition: 0.3s;
  font-size: 90%;
  border-top-left-radius: 6px;
  border-top-right-radius: 6px;
  }

/* Change background color of buttons on hover */
div.tab button:hover {
  background-color: #ddd;
  }

/*button active*/
div.tab button.active {
  background-color: #ccc;
  }

/* Style the tab content */
.stabcontent {
  border: 1px solid #ccc;
  padding: 6px;
  display: none;
  }

/* pre content */
.stabcontent pre {
  overflow: auto;
  -ms-word-wrap: normal;
  word-wrap: normal;
  overflow-wrap: normal;
  white-space: pre;
  max-height: 450px;
  }

pre {
  	font-size: small;
   /* max-height: 400px; */
    overflow-x: hidden;
    white-space: pre;
    padding: 10px;
   /* background: #f3f6f6; */
    border: 1px solid #f3ecec;
    background: rgb(249, 249, 249);
    border-radius: 10px;
}

.stabcontent {
    border: 1px solid #f3ecec;
    padding: 15px;
}

.wy-table-responsive table td, .wy-table-responsive table th {
    white-space: inherit;
}

div.wy-side-nav-search 
{
    background-color: #F08B1D;
}

.tooltiptext-under td, .tooltiptext-under tr:nth-child(2n-1) td, .tooltiptext-under table.docutils:not(.field-list) tr:nth-child(2n-1) td
{
    background-color: transparent !IMPORTANT;
}

.tooltiptext-under table.docutils thead th {
    background-color: gray;   
}

table.docutils {

}

td p {
   font-size: 14.4px;
   line-height: 1em !important;
}

.section>p{
    margin : 0px;
}

footer {
    display: none;
}

h2 {
    visibility : hidden;
    height: 0px;
    width: 0px;
    margin: 0px;
}

.glossary-desc {
    display: none;
}

h2, h3, h4 {
    color: #F08B1D;
}

h2, h5 {
    cursor : pointer;
}

h3 {
    margin-top : 15px;
}

.show-more {
    color: #2980B9;
    cursor: pointer;
}

.docutils {
    width: 100%;
}

.wy-nav-content-wrap {
    background-color: rgb(252, 252, 252);
}

.wy-nav-content {
    max-width: 1280px;
}

.wy-nav-top {
    background: #F08B1D
}

.tooltiptext-under {
    visibility: hidden;
    background-color: #333333;
    color: #fff;
    text-align: left;
    border-radius: 6px;
    padding: 5px 10px 5px 10px;
    margin: 25px 0px 0px 10px;
    margin-left: -50px;
    position: absolute;
    z-index: 1;
    width: calc(100% - 320px);
    left: 362px; 
}


.tooltiptext-after {
    visibility: hidden;
    content: "";
    position: absolute;
    margin-top: 15px;
    margin-left: -25px;
    border-width: 5px;
    border-style: solid;
    border-color: transparent transparent #333333 transparent;
}

td a .tooltiptext-under {
    visibility: hidden;
    background-color: #333333;
    color: #fff;
    text-align: left;
    border-radius: 6px;
    padding: 5px 10px 5px 10px;
    margin: 20px 0px 0px 10px;
    margin-left: -50px;
    position: absolute;
    z-index: 1;
    width: calc(100% - 320px);
    left: 362px; 
}


td a .tooltiptext-after {
    visibility: hidden;
    content: "";
    position: absolute;
    margin-top: 10px;
    margin-left: 3px;
    border-width: 5px;
    border-style: solid;
    border-color: transparent transparent #333333 transparent;
}

.tooltip-parent:hover .tooltiptext-under,  .tooltip-parent:hover .tooltiptext-after, .tooltip-parent:hover .tooltiptext-after, .tooltip-parent:hover .tooltiptext-under{
    visibility: visible;
}

#wheel {
  z-index: 1;
  display: none;
	width: 100%;
	height: 100%;
	position: absolute; 
	background-image: url('../img/wheel.gif');
	background-repeat: no-repeat;
	background-attachment: fixed;
	background-position: center;
}



#search-wheel {    
    margin-top: 10vh;
    margin-left: 0vh;
}