var schemaString = `<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema" xmlns:lx="lixi.org.au/schema/appinfo_elements" xmlns:li="lixi.org.au/schema/appinfo_instructions" elementFormDefault="qualified" attributeFormDefault="unqualified" version="1.0">
  <xs:annotation>
    <xs:documentation>IMPORTANT: Copyright in this schema is owned by LIXI Limited and use of the schema is controlled by the LIXI End User Licence Agreement (EULA) which can be downloaded from http://www.lixi.org.au. The EULA does not permit you to extend the schema through the addition of new elements or attributes, or modify the content model or any data values in a manner that would be inconsistent with the Standard. However you are permitted to restrict the schema such that it specifies the structures for capturing only the information you wish to receive. This statement in no way limits or modifies the terms of the EULA, and you should ensure that you are familiar with your rights and obligations under the EULA before using the schema.</xs:documentation>
    <xs:documentation>This schema is the LIXI Master Schema WIP for next release. This field gets autopopulated by the process that creates the Transaction Schema from the LIXI Master Schema.</xs:documentation>
    <xs:appinfo>
      <lx:schemadetail type="Master" version="2.2.28" annotation="Full" lifecyclestage="Autopopulated by distribution process" timestamp="Autopopulated by distribution process"/>
      <li:subschema transactiontype="CAL" version="2.6.28"/>
      <li:subschema transactiontype="DAS" version="2.2.28"/>
      <li:subschema transactiontype="CNZ" version="2.1.1"/>
      <li:subschema transactiontype="CDA" version="2.0.26"/>
      <li:subschema transactiontype="SVC" version="2.0.17"/>
      <li:subschema transactiontype="LMI" version="0.0.13"/>
      <li:subschema transactiontype="VAL" version="0.0.13"/>
      <li:subschema transactiontype="ACC" version="0.0.10"/>
    </xs:appinfo>
  </xs:annotation>
  <xs:element name="Package">
    <xs:annotation>
      <xs:documentation>The root element that contains a message for exchange between two end-points involved in the workflow related to the processing of an application for a new financial product or change management of an existing financial product. The message contains instructions to the recipient requesting some workflow activity along with data related to: the applicant and the financial product involved; the publisher; one or more recipients; schema version information; and any related attachments.</xs:documentation>
      <xs:appinfo>
        <lx:path>Package</lx:path>
        <lx:label>Package</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:complexType>
      <xs:sequence>
        <xs:element name="Attachment" minOccurs="0" maxOccurs="unbounded">
          <xs:annotation>
            <xs:documentation>Contains information about a document that is attached to the application, either by being embedded or by reference to a URL, or both.</xs:documentation>
            <xs:appinfo>
              <lx:path>Package.Attachment</lx:path>
              <lx:label>Attachment</lx:label>
              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
            </xs:appinfo>
          </xs:annotation>
          <xs:complexType>
            <xs:sequence>
              <xs:element name="InlineAttachment" type="base64BinaryType" minOccurs="0" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>For each attachment, the document itself can be embedded in the XML in a text encoded format.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Attachment.InlineAttachment</lx:path>
                    <lx:label>Inline Attachment</lx:label>
                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
              </xs:element>
            </xs:sequence>
            <xs:attribute name="Filename" type="stringType" use="optional">
              <xs:annotation>
                <xs:documentation>For each attachment, the name for the attached file, including filename extension may be provided</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.Attachment.Filename</lx:path>
                  <lx:label>Filename</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
            <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
              <xs:annotation>
                <xs:documentation>System assigned unique identifier</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.Attachment.UniqueID</lx:path>
                  <lx:label>Unique ID</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
            <xs:attribute name="URI" type="stringType" use="optional">
              <xs:annotation>
                <xs:documentation>For each attachment, the URI that identifies where the attached file might be accessed from. While it is anticipated that it will generally be of the form "http://" or "ftp://", there is no requirement for this to be the case.</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.Attachment.URI</lx:path>
                  <lx:label>URI</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
          </xs:complexType>
        </xs:element>
        <xs:element name="Content" minOccurs="0" maxOccurs="1">
          <xs:annotation>
            <xs:documentation>Wrapper element for the actual data, the payload of the message</xs:documentation>
            <xs:appinfo>
              <lx:path>Package.Content</lx:path>
              <lx:label>Content</lx:label>
              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
            </xs:appinfo>
          </xs:annotation>
          <xs:complexType>
            <xs:sequence>
              <xs:element name="Application" minOccurs="0" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>The set of borrower- and broker-provided data for a single loan application.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Content.Application</lx:path>
                    <lx:label>Application</lx:label>
                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:sequence>
                    <xs:element name="AccountVariation" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>A request to vary an existing loan account. More than one request can be made on a single account, such as increase limit and add security. Not all requests are compatible, for instance increase limit and decrease limit.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.AccountVariation</lx:path>
                          <lx:label>Account Variation</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="AddBorrower" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to add a borrower to an existing loan account</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.AddBorrower</lx:path>
                                <lx:label>Add Borrower</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Borrower" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific uniquely identifiable borrower to add to an existing loan account</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.AddBorrower.x_Borrower</lx:path>
                                    <lx:label>Borrower</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
				                              <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="AddGuarantee" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to add a guarantee to an existing loan account.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.AddGuarantee</lx:path>
                                <lx:label>Add Guarantee</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_LendingGuarantee" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the Lending Guarantee.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.AddGuarantee.x_LendingGuarantee</lx:path>
                                    <lx:label>Lending Guarantee</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.LendingGuarantee</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="AddGuaranteeSecurity" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to add a security to an existing lending guarantee.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.AddGuaranteeSecurity</lx:path>
                                <lx:label>Add Guarantee Security</lx:label>
                                <li:transactions>CAL,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ActionDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date that the security is to be added.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.AddGuaranteeSecurity.ActionDate</lx:path>
                                    <lx:label>Action Date</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Security" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific uniquely identifiable security to add to an existing lending guarantee.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.AddGuaranteeSecurity.x_Security</lx:path>
                                    <lx:label>Security</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.GeneralSecurityAgreement</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="AddGuarantor" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to add a guarantor to an existing loan account, generally in conjunction with a limit increase and an additional security</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.AddGuarantor</lx:path>
                                <lx:label>Add Guarantor</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Guarantor" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific uniquely identifiable guarantor to add to an existing loan account</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.AddGuarantor.x_Guarantor</lx:path>
                                    <lx:label>Guarantor</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="AddSecurity" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to add a security to an existing loan account, generally in conjunction with a limit increase</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.AddSecurity</lx:path>
                                <lx:label>Add Security</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ActionDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date that the security is to be added.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.AddSecurity.ActionDate</lx:path>
                                    <lx:label>Action Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Security" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific uniquely identifiable security to add to an existing loan account</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.AddSecurity.x_Security</lx:path>
                                    <lx:label>Security</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.GeneralSecurityAgreement</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="BalanceIncrease" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to increase the balance on an existing loan account.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.BalanceIncrease</lx:path>
                                <lx:label>Balance Increase</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ActionDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date on which the balance is to be increased.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.BalanceIncrease.ActionDate</lx:path>
                                    <lx:label>Action Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncreaseAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount by which the balance is to be increased.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.BalanceIncrease.IncreaseAmount</lx:path>
                                    <lx:label>Increase Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NewBalance" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount to which the balance is to be increased.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.BalanceIncrease.NewBalance</lx:path>
                                    <lx:label>New Balance</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ChangeLoanTerm" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to change the term of the loan, either the fixed term, the total term, or both.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.ChangeLoanTerm</lx:path>
                                <lx:label>Change Loan Term</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ActionDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Date from which the loan term is to be changed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ChangeLoanTerm.ActionDate</lx:path>
                                    <lx:label>Action Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NewFixedLoanTerm" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of months for the new fixed rate term, if the fixed rate term can be changed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ChangeLoanTerm.NewFixedLoanTerm</lx:path>
                                    <lx:label>New Fixed Loan Term</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NewTotalLoanTerm" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of months for the total term of the loan, from the action date</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ChangeLoanTerm.NewTotalLoanTerm</lx:path>
                                    <lx:label>New Total Loan Term</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ChangeRepaymentAmount" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Request to change the amount of the periodic repayment</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.ChangeRepaymentAmount</lx:path>
                                <lx:label>Change Repayment Amount</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="RepaymentAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The requested periodic repayment amount. It will be up to the lender to determine whether the repayment funds are all applied against the principal or whether some goes into redraw.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ChangeRepaymentAmount.RepaymentAmount</lx:path>
                                    <lx:label>Repayment Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ChangeRepaymentFrequency" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Request to change the repayment frequency (e.g. monthly, fortnightly) for the existing account</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.ChangeRepaymentFrequency</lx:path>
                                <lx:label>Change Repayment Frequency</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="RepaymentFrequency" type="frequencyFullList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The requested repayment frequency (e.g. monthly, fortnightly)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ChangeRepaymentFrequency.RepaymentFrequency</lx:path>
                                    <lx:label>Repayment Frequency</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ChangeRepaymentType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Request to change the repayment type, e.g. interest only, principal and interest</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.ChangeRepaymentType</lx:path>
                                <lx:label>Change Repayment Type</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="RepaymentType" type="paymentTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The requested repayment type, e.g. interest only, principal and interest</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ChangeRepaymentType.RepaymentType</lx:path>
                                    <lx:label>Repayment Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="CloseAccount" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to close the existing loan account</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.CloseAccount</lx:path>
                                <lx:label>Close Account</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DestroyCard" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>If a card is attached to the account, should the card account also be terminated and the card destroyed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.CloseAccount.DestroyCard</lx:path>
                                    <lx:label>Destroy Card</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PayoutAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An estimate of the amount required to be transferred to clear the account balance and exit fees</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.CloseAccount.PayoutAmount</lx:path>
                                    <lx:label>Payout Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="InterestRateDiscount" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request for a discount on the existing loan account interest rate</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.InterestRateDiscount</lx:path>
                                <lx:label>Interest Rate Discount</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DiscountRate" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The discount being requested, as a delta from the base rate for the product</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.InterestRateDiscount.DiscountRate</lx:path>
                                    <lx:label>Discount Rate</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="LendingGuaranteeLimitDecrease" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to reduce the limit on an existing lending guarantee.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.LendingGuaranteeLimitDecrease</lx:path>
                                <lx:label>Lending Guarantee Limit Decrease</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DecreaseAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount that the lending guarantee limit is to be reduced by.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.LendingGuaranteeLimitDecrease.DecreaseAmount</lx:path>
                                    <lx:label>Decrease Amount</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NewLimit" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount that the lending guarantee limit is to be reduced to.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.LendingGuaranteeLimitDecrease.NewLimit</lx:path>
                                    <lx:label>New Limit</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="LendingGuaranteeLimitIncrease" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to increase the limit on an existing lending guarantee.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.LendingGuaranteeLimitIncrease</lx:path>
                                <lx:label>Lending Guarantee Limit Increase</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="IncreaseAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount that the lending guarantee limit is to be increased by.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.LendingGuaranteeLimitIncrease.IncreaseAmount</lx:path>
                                    <lx:label>Increase Amount</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NewLimit" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount that the lending guarantee limit is to be increased to.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.LendingGuaranteeLimitIncrease.NewLimit</lx:path>
                                    <lx:label>New Limit</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="LimitIncrease" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to increase the limit on an existing loan account</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.LimitIncrease</lx:path>
                                <lx:label>Limit Increase</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="NewLimitRequestedInForeignCurrency" type="amountInForeignCurrencyType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The amount that the credit limit is to be increased to, as requested in a foreign currency.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.AccountVariation.LimitIncrease.NewLimitRequestedInForeignCurrency</lx:path>
                                      <lx:label>New Limit Requested In Foreign Currency</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="ActionDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date on which the limit is to be increased</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.LimitIncrease.ActionDate</lx:path>
                                    <lx:label>Action Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncreaseAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount that the credit limit is to be increased by</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.LimitIncrease.IncreaseAmount</lx:path>
                                    <lx:label>Increase Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsNewLimitRequestedInForeignCurrency" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A flag to indicate that the new credit limit amount was requested in a foreign currency.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.LimitIncrease.IsNewLimitRequestedInForeignCurrency</lx:path>
                                    <lx:label>Is New Limit Requested In Foreign Currency</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NewLimit" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount that the credit limit is to be increased to</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.LimitIncrease.NewLimit</lx:path>
                                    <lx:label>New Limit</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ReduceBalance" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to reduce the balance on an existing loan account.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.ReduceBalance</lx:path>
                                <lx:label>Reduce Balance</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ActionDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date on which the balance is to be reduced.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ReduceBalance.ActionDate</lx:path>
                                    <lx:label>Action Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NewBalance" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount to which the balance is to be reduced.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ReduceBalance.NewBalance</lx:path>
                                    <lx:label>New Balance</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ReductionAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount by which the balance is to be decreased.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ReduceBalance.ReductionAmount</lx:path>
                                    <lx:label>Reduction Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ReduceLimit" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to reduce the limit of an existing loan account</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.ReduceLimit</lx:path>
                                <lx:label>Reduce Limit</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="NewLimitRequestedInForeignCurrency" type="amountInForeignCurrencyType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The amount that the credit limit is to be reduced to, as requested in a foreign currency.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.AccountVariation.ReduceLimit.NewLimitRequestedInForeignCurrency</lx:path>
                                      <lx:label>New Limit Requested In Foreign Currency</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="IsNewLimitRequestedInForeignCurrency" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A flag to indicate that the new credit limit amount was requested in a foreign currency.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ReduceLimit.IsNewLimitRequestedInForeignCurrency</lx:path>
                                    <lx:label>Is New Limit Requested In Foreign Currency</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LimitReduction" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount to reduce the existing account limit by. For example to reduce a $50,000 line of credit to $40,000, the Limit Reduction would be $10,000</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ReduceLimit.LimitReduction</lx:path>
                                    <lx:label>Limit Reduction</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NewLimit" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The new (reduced) limit for the loan account</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ReduceLimit.NewLimit</lx:path>
                                    <lx:label>New Limit</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ReleaseGuarantee" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to release a guarantee from an existing loan account.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.ReleaseGuarantee</lx:path>
                                <lx:label>Release Guarantee</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_LendingGuarantee" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the Lending Guarantee.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ReleaseGuarantee.x_LendingGuarantee</lx:path>
                                    <lx:label>Lending Guarantee</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.LendingGuarantee</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ReleaseGuaranteeSecurity" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to release a security from an existing lending guarantee.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.ReleaseGuaranteeSecurity</lx:path>
                                <lx:label>Release Guarantee Security</lx:label>
                                <li:transactions>CAL,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ActionDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date that the security is to be released.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ReleaseGuaranteeSecurity.ActionDate</lx:path>
                                    <lx:label>Action Date</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Reason" type="releaseSecurityReasonList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A reason for the request to release the security</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ReleaseGuaranteeSecurity.Reason</lx:path>
                                    <lx:label>Reason</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Security" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific uniquely identifiable security to remove from an existing lending guarantee.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ReleaseGuaranteeSecurity.x_Security</lx:path>
                                    <lx:label>Security</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.GeneralSecurityAgreement</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ReleaseGuarantor" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to release a guarantor from an existing loan account</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.ReleaseGuarantor</lx:path>
                                <lx:label>Release Guarantor</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Guarantor" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific uniquely identifiable guarantor to remove from an existing loan account</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ReleaseGuarantor.x_Guarantor</lx:path>
                                    <lx:label>Guarantor</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ReleaseSecurity" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to release a security from an existing loan account</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.ReleaseSecurity</lx:path>
                                <lx:label>Release Security</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ActionDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date that the security is to be released.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ReleaseSecurity.ActionDate</lx:path>
                                    <lx:label>Action Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Reason" type="releaseSecurityReasonList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A reason for the request to release the security</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ReleaseSecurity.Reason</lx:path>
                                    <lx:label>Reason</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Security" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific uniquely identifiable security to remove from an existing loan account</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.ReleaseSecurity.x_Security</lx:path>
                                    <lx:label>Security</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.GeneralSecurityAgreement</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="RemoveBorrower" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to remove a borrower from an existing loan account</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.RemoveBorrower</lx:path>
                                <lx:label>Remove Borrower</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Borrower" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific uniquely identifiable borrower to remove from an existing loan account</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.RemoveBorrower.x_Borrower</lx:path>
                                    <lx:label>Borrower</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="SplitAccount" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Changing a loan from single loan product to two or more products, generally within the same approved limit</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.SplitAccount</lx:path>
                                <lx:label>Split Account</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Account" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>The set of new account details for each split, including (as required) information such as loan product, loan term, loan amount, etc.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.AccountVariation.SplitAccount.Account</lx:path>
                                      <lx:label>Account</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="x_AccountDetails" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The set of new account details for each split, including (as required) information such as loan product, loan term, loan amount, etc.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.AccountVariation.SplitAccount.Account.x_AccountDetails</lx:path>
                                          <lx:label>Account Details</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="ActionDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date that the account is to be split</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.SplitAccount.ActionDate</lx:path>
                                    <lx:label>Action Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="SwitchProduct" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to switch an existing loan account to a new product</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.SwitchProduct</lx:path>
                                <lx:label>Switch Product</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_ProductDetails" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The set of loan product details for the new product, including (as required) information such as loan product, repayment type, loan term etc.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.SwitchProduct.x_ProductDetails</lx:path>
                                    <lx:label>Product Details</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="UpdatePartyDetails" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to change details, such as contact details, for a party on the account</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.AccountVariation.UpdatePartyDetails</lx:path>
                                <lx:label>Update Party Details</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_LoanParty" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific uniquely identifiable party whose details are being changed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.AccountVariation.UpdatePartyDetails.x_LoanParty</lx:path>
                                    <lx:label>Loan Party</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this account variation within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.AccountVariation.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Type" type="variationTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate the type of variation being specified</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.AccountVariation.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.AccountVariation.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="VariationDescription" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A financial institution's specific description that further describes the type of account variation.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.AccountVariation.VariationDescription</lx:path>
                              <lx:label>Variation Description</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_AccountToVary" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The existing loan account (i.e. liability) that is being varied.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.AccountVariation.x_AccountToVary</lx:path>
                              <lx:label>Account To Vary</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.Liability</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_LendingGuaranteeToVary" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The existing guarantee that is being varied.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.AccountVariation.x_LendingGuaranteeToVary</lx:path>
                              <lx:label>Guarantee To Vary</lx:label>
                              <li:transactions>CAL,DAS</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.LendingGuarantee</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_PostVariationAccount" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A cross reference to the loan account after it has been varied (a Loan Details element).</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.AccountVariation.x_PostVariationAccount</lx:path>
                              <lx:label>Post Variation Account</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.LoanDetails</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Address" type="addressType" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Any real or postal address, used by reference</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.Address</lx:path>
                          <lx:label>Address</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                    </xs:element>
                    <xs:element name="BusinessChannel" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Captures information about the organisation that the loan application is being submitted to. Typically, the business channel is the lender. For an application being submitted to a mortgage manager, the mortgage manager is the business channel.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.BusinessChannel</lx:path>
                          <lx:label>Business Channel</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains contact details for the business channel</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.BusinessChannel.Contact</lx:path>
                                <lx:label>Contact</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="ContactPerson" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Contact person for the business channel.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.Contact.ContactPerson</lx:path>
                                      <lx:label>Contact Person</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Email" type="emailPattern" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The email address of the contact person.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.BusinessChannel.Contact.ContactPerson.Email</lx:path>
                                          <lx:label>Email</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="FirstName" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A person's first name.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.BusinessChannel.Contact.ContactPerson.FirstName</lx:path>
                                          <lx:label>First Name</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NameTitle" type="nameTitleList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The name title this person uses when they are formally addressed should be captured.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.BusinessChannel.Contact.ContactPerson.NameTitle</lx:path>
                                          <lx:label>Name Title</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Role" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The role or job position of the contact person in the business channel.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.BusinessChannel.Contact.ContactPerson.Role</lx:path>
                                          <lx:label>Role</lx:label>
                                          <li:transactions>CAL,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Surname" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A person's surname.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.BusinessChannel.Contact.ContactPerson.Surname</lx:path>
                                          <lx:label>Surname</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="OfficeFax" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Contains a contact fax number for the business channel.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.Contact.OfficeFax</lx:path>
                                      <lx:label>Office Fax</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Contains a contact phone number for the business channel.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.Contact.OfficePhone</lx:path>
                                      <lx:label>Office Phone</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Email" type="emailPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures an email address for the distribution channel to which the loan application is being submitted.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.Contact.Email</lx:path>
                                    <lx:label>Email</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="WebAddress" type="urlPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An internet address for web content associated with the business channel.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.Contact.WebAddress</lx:path>
                                    <lx:label>Web Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Address" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Address should capture the lender's address .</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.Contact.x_Address</lx:path>
                                    <lx:label>Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="CreditAssessor" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The credit assessor that is associated with this application. The credit assessor works for the lender. They are responsible for evaluating whether the loan application complies with the lender's credit compliance rules. A loan may be assessed by more than one credit assessor.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.BusinessChannel.CreditAssessor</lx:path>
                                <lx:label>Credit Assessor</lx:label>
                                <li:transactions>ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The credit assessor's contact details.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.CreditAssessor.Contact</lx:path>
                                      <lx:label>Contact</lx:label>
                                      <li:transactions>ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="EmailAddress" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>A contact email address.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.CreditAssessor.Contact.EmailAddress</lx:path>
                                            <lx:label>Email Address</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Email" type="emailPattern" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The email address.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.BusinessChannel.CreditAssessor.Contact.EmailAddress.Email</lx:path>
                                                <lx:label>Email</lx:label>
                                                <li:transactions>ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="Mobile" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>A contact mobile phone number.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.CreditAssessor.Contact.Mobile</lx:path>
                                            <lx:label>Mobile</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>A contact office phone number.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.CreditAssessor.Contact.OfficePhone</lx:path>
                                            <lx:label>Office Phone</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The name of the credit assessor.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.CreditAssessor.PersonName</lx:path>
                                      <lx:label>Person Name</lx:label>
                                      <li:transactions>ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="PositionTitle" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The job position title of the credit assessor.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.CreditAssessor.PositionTitle</lx:path>
                                    <lx:label>Position Title</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.CreditAssessor.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="CreditManager" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The credit manager that is associated with this application. The credit manager is a senior credit assessor. They are ultimately responsible for approving the loan based on the lender's credit compliance rules, and the assessments given by the credit assessors.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.BusinessChannel.CreditManager</lx:path>
                                <lx:label>Credit Manager</lx:label>
                                <li:transactions>ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The credit manager's contact details.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.CreditManager.Contact</lx:path>
                                      <lx:label>Contact</lx:label>
                                      <li:transactions>ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="EmailAddress" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>A contact email address.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.CreditManager.Contact.EmailAddress</lx:path>
                                            <lx:label>Email Address</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Email" type="emailPattern" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The email address.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.BusinessChannel.CreditManager.Contact.EmailAddress.Email</lx:path>
                                                <lx:label>Email</lx:label>
                                                <li:transactions>ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="Mobile" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>A contact mobile phone number.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.CreditManager.Contact.Mobile</lx:path>
                                            <lx:label>Mobile</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>A contact office phone number.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.CreditManager.Contact.OfficePhone</lx:path>
                                            <lx:label>Office Phone</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The name of the credit manager.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.CreditManager.PersonName</lx:path>
                                      <lx:label>Person Name</lx:label>
                                      <li:transactions>ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="PositionTitle" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The job position title of the credit manager.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.CreditManager.PositionTitle</lx:path>
                                    <lx:label>Position Title</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.CreditManager.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PanelSolicitor" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Lender's approved panel solicitor</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.BusinessChannel.PanelSolicitor</lx:path>
                                <lx:label>Panel Solicitor</lx:label>
                                <li:transactions>CAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Solicitor" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Solicitor or the company who is approved to act on behalf of the lender. Panel Solicitor associated with the lender for a given loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.PanelSolicitor.x_Solicitor</lx:path>
                                    <lx:label>Solicitor</lx:label>
                                    <li:transactions>CAL,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="RelationshipManager" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The relationship manager that is associated with this application. The relationship manager works for the lender and is responsible for negotiating policy with brokers. The relationship manager negotiates a given loan application with the broker and then passes the loan application to the credit assessment team.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.BusinessChannel.RelationshipManager</lx:path>
                                <lx:label>Relationship Manager</lx:label>
                                <li:transactions>ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The relationship manager contact details.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.RelationshipManager.Contact</lx:path>
                                      <lx:label>Contact</lx:label>
                                      <li:transactions>ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="EmailAddress" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>A contact email address.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.RelationshipManager.Contact.EmailAddress</lx:path>
                                            <lx:label>Email Address</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Email" type="emailPattern" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The email address.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.BusinessChannel.RelationshipManager.Contact.EmailAddress.Email</lx:path>
                                                <lx:label>Email</lx:label>
                                                <li:transactions>ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="Mobile" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>A contact mobile phone number.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.RelationshipManager.Contact.Mobile</lx:path>
                                            <lx:label>Mobile</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>A contact office phone number.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.RelationshipManager.Contact.OfficePhone</lx:path>
                                            <lx:label>Office Phone</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The name of the relationship manager.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.RelationshipManager.PersonName</lx:path>
                                      <lx:label>Person Name</lx:label>
                                      <li:transactions>ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="PositionTitle" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The job position title of the relationship manager.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.RelationshipManager.PositionTitle</lx:path>
                                    <lx:label>Position Title</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.RelationshipManager.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="SalesConsultant" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The sales consultant that is associated with this application.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.BusinessChannel.SalesConsultant</lx:path>
                                <lx:label>Sales Consultant</lx:label>
                                <li:transactions>ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The sales consultant contact details.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.SalesConsultant.Contact</lx:path>
                                      <lx:label>Contact</lx:label>
                                      <li:transactions>ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="EmailAddress" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>A contact email address.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.SalesConsultant.Contact.EmailAddress</lx:path>
                                            <lx:label>Email Address</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Email" type="emailPattern" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The email address.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.BusinessChannel.SalesConsultant.Contact.EmailAddress.Email</lx:path>
                                                <lx:label>Email</lx:label>
                                                <li:transactions>ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="Mobile" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>A contact mobile phone number.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.SalesConsultant.Contact.Mobile</lx:path>
                                            <lx:label>Mobile</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>A contact office phone number.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.SalesConsultant.Contact.OfficePhone</lx:path>
                                            <lx:label>Office Phone</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The name of the sales consultant.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.SalesConsultant.PersonName</lx:path>
                                      <lx:label>Person Name</lx:label>
                                      <li:transactions>ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="PositionTitle" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The job position title of the sales consultant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.SalesConsultant.PositionTitle</lx:path>
                                    <lx:label>Position Title</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.SalesConsultant.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="SettlementSupportManager" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The settlement support manager that is associated with this application. The settlement support manager is a senior settlement support officer. They are responsible for authorising that the loan application can be submitted to the servicer.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportManager</lx:path>
                                <lx:label>settlement support manager</lx:label>
                                <li:transactions>ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The settlement support manager contact details.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportManager.Contact</lx:path>
                                      <lx:label>Contact</lx:label>
                                      <li:transactions>ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="EmailAddress" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>A contact email address.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportManager.Contact.EmailAddress</lx:path>
                                            <lx:label>Email Address</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Email" type="emailPattern" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The email address.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportManager.Contact.EmailAddress.Email</lx:path>
                                                <lx:label>Email</lx:label>
                                                <li:transactions>ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="Mobile" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>A contact mobile phone number.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportManager.Contact.Mobile</lx:path>
                                            <lx:label>Mobile</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>A contact office phone number.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportManager.Contact.OfficePhone</lx:path>
                                            <lx:label>Office Phone</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The name of the settlement support manager.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportManager.PersonName</lx:path>
                                      <lx:label>Person Name</lx:label>
                                      <li:transactions>ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="PositionTitle" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The job position title of the settlement support manager.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportManager.PositionTitle</lx:path>
                                    <lx:label>Position Title</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportManager.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="SettlementSupportOfficer" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The settlement support officer that is associated with this application. The settlement support officer works for the lender. They receive a loan application that has been approved by the lender's credit assessment team. They work to ensure that the loan application has all the required paperwork complete and is ready to submit to the servicer.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportOfficer</lx:path>
                                <lx:label>settlement support officer</lx:label>
                                <li:transactions>ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The settlement support officer contact details.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportOfficer.Contact</lx:path>
                                      <lx:label>Contact</lx:label>
                                      <li:transactions>ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="EmailAddress" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>A contact email address.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportOfficer.Contact.EmailAddress</lx:path>
                                            <lx:label>Email Address</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Email" type="emailPattern" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The email address.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportOfficer.Contact.EmailAddress.Email</lx:path>
                                                <lx:label>Email</lx:label>
                                                <li:transactions>ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="Mobile" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>A contact mobile phone number.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportOfficer.Contact.Mobile</lx:path>
                                            <lx:label>Mobile</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>A contact office phone number.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportOfficer.Contact.OfficePhone</lx:path>
                                            <lx:label>Office Phone</lx:label>
                                            <li:transactions>ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The name of the settlement support officer.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportOfficer.PersonName</lx:path>
                                      <lx:label>Person Name</lx:label>
                                      <li:transactions>ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="PositionTitle" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The job position title of the settlement support officer.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportOfficer.PositionTitle</lx:path>
                                    <lx:label>Position Title</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.BusinessChannel.SettlementSupportOfficer.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="ABN" type="abnPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The lender ABN.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.BusinessChannel.ABN</lx:path>
                              <lx:label>ABN</lx:label>
                              <li:transactions>CAL,CDA,DAS,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ACN" type="acnPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The lender ACN.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.BusinessChannel.ACN</lx:path>
                              <lx:label>ACN</lx:label>
                              <li:transactions>CAL,CDA,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CompanyName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The company name of the lender or mortgage manager.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.BusinessChannel.CompanyName</lx:path>
                              <lx:label>Company Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LenderID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures a unique identifier that has been assigned by the lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.BusinessChannel.LenderID</lx:path>
                              <lx:label>Lender ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LicenceNumber" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The licence number.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.BusinessChannel.LicenceNumber</lx:path>
                              <lx:label>Licence Number</lx:label>
                              <li:transactions>CAL,CDA,DAS,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LicenceType" type="licenceTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of credit or other licence that the lender operates under.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.BusinessChannel.LicenceType</lx:path>
                              <lx:label>Licence Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NZBN" type="nzbnPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The lenders NZBN.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.BusinessChannel.NZBN</lx:path>
                              <lx:label>NZBN</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OtherIdentifier" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures an alternative identifier by which the business channel is known, typically issued by the originating organisation rather than the lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.BusinessChannel.OtherIdentifier</lx:path>
                              <lx:label>Other Identifier</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Type" type="businessChannelTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Type of organisation that is the recipient of the application, typically a type of lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.BusinessChannel.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.BusinessChannel.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="CompanyApplicant" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Contains applicant details for business enterprises (including company, sole trader, partnership or government department) being either a borrower or a guarantor, and may be acting for a Trust.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.CompanyApplicant</lx:path>
                          <lx:label>Company Applicant</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="AuthorisedSignatory" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains a reference to an applicant or related party that is an authorised signatory of the company.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.AuthorisedSignatory</lx:path>
                                <lx:label>Authorised Signatory</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="PositionTitle" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The position or job title of the person that is a signatory</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.AuthorisedSignatory.PositionTitle</lx:path>
                                    <lx:label>Position Title</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Signatory" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Contains a reference to an applicant or related party that is an authorised signatory of the company.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.AuthorisedSignatory.x_Signatory</lx:path>
                                    <lx:label>Signatory</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="BeneficialOwner" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>A beneficial owner of the company. All beneficial owners of the company should be captured (even persons captured elsewhere as directors, shareholders, beneficiaries etc.)</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.BeneficialOwner</lx:path>
                                <lx:label>Beneficial Owner</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_BeneficialOwner" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the beneficial owner.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.BeneficialOwner.x_BeneficialOwner</lx:path>
                                    <lx:label>Beneficial Owner</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Business" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about the business of the company</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.Business</lx:path>
                                <lx:label>Business</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="ConcentrationRisk" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The company is at risk because of customer concentration, product concentration, industry concentration, economic-sector concentration, supplier concentration, or finance concentration.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.Business.ConcentrationRisk</lx:path>
                                      <lx:label>Concentration Risk</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="ConcentrationRiskDetails" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Details of the concentration risk.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.Business.ConcentrationRisk.ConcentrationRiskDetails</lx:path>
                                          <lx:label>Concentration Risk Details</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CustomerOrSupplierConcentration" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The company is at risk because it does more than 25 percent of business with one customer or supplier.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.Business.ConcentrationRisk.CustomerOrSupplierConcentration</lx:path>
                                          <lx:label>Customer or Supplier Concentration</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="Diversification" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The diversification of the company with respect to it's suppliers and customers.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.Business.Diversification</lx:path>
                                      <lx:label>Diversification</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Details" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The nature of the diversification of the diversified company.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.Business.Diversification.Details</lx:path>
                                          <lx:label>Details</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="DiversifiedDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date when the business diversified.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.Business.Diversification.DiversifiedDate</lx:path>
                                          <lx:label>Diversified Date</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IsDiversified" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates that the company is a diversified company.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.Business.Diversification.IsDiversified</lx:path>
                                          <lx:label>Is Diversified</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="ImportExport" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Information regarding a company's involvement in import and/or export.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.Business.ImportExport</lx:path>
                                      <lx:label>Import Export</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Details" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures a free text description the nature of the involvement in import and/or export.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.Business.ImportExport.Details</lx:path>
                                          <lx:label>Details</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IsInvolved" type="yesNoIntentList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates if the company is involved in import and/or export, or intends to become involved in import and/or export.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.Business.ImportExport.IsInvolved</lx:path>
                                          <lx:label>Is Involved</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="PropertyInvestment" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Information regarding a company's involvement in property investment.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.Business.PropertyInvestment</lx:path>
                                      <lx:label>Property Investment</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Details" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures a free text description the nature of the involvement in property investment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.Business.PropertyInvestment.Details</lx:path>
                                          <lx:label>Details</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IsInvolved" type="yesNoIntentList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates if the company is involved in property investment, or intends to become involved in property investment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.Business.PropertyInvestment.IsInvolved</lx:path>
                                          <lx:label>Is Involved</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="AustralianBIC" type="aubicPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A Business Industry Code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.AustralianBIC</lx:path>
                                    <lx:label>Australian BIC</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CustomIndustryCode" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An industry code, from a customised list supported by the lender where the lender requires a code that is not within ANZSIC, BIC or GICS.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.CustomIndustryCode</lx:path>
                                    <lx:label>Custom Industry Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FranchiseDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Details of the franchise.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.FranchiseDetails</lx:path>
                                    <lx:label>Franchise Details</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GICSCode" type="gicsCodePattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A Global Industry Classification Standard (GICS) code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.GICSCode</lx:path>
                                    <lx:label>GICS Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Industry" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures a free text description of the Industry Code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.Industry</lx:path>
                                    <lx:label>Industry</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IndustryCode" type="anzsicIndustryCodeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The four digit industry code, from the ANZSIC list of codes. If the nominal code is two or three digits, add '0's to the end to make the four digit code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.IndustryCode</lx:path>
                                    <lx:label>Industry Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsAFranchise" type="yesNoIntentList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether this business is a franchise or has a future intention to become a franchise.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.IsAFranchise</lx:path>
                                    <lx:label>Is A Franchise</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MainBusinessActivity" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The main activity of the business</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.MainBusinessActivity</lx:path>
                                    <lx:label>Main Business Activity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NewZealandBICCode" type="nzbicPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A Business Industry Classification Code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.NewZealandBICCode</lx:path>
                                    <lx:label>New Zealand BIC Code</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfEmployees" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of employees the company has in the business.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.NumberOfEmployees</lx:path>
                                    <lx:label>Number Of Employees</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfLocations" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of locations where the business operates should be captured.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.NumberOfLocations</lx:path>
                                    <lx:label>Number Of Locations</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OwnPremises" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the company owns the premises from which it operates the business.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.OwnPremises</lx:path>
                                    <lx:label>Own Premises</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PrincipalTradingAddressLeaseAgreementDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The details of the Lease Agreement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.PrincipalTradingAddressLeaseAgreementDetails</lx:path>
                                    <lx:label>Principal Trading Address Lease Agreement Details</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PrincipalTradingAddressLeased" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the premises from which the company operates its business is leased.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.PrincipalTradingAddressLeased</lx:path>
                                    <lx:label>Principal Trading Address Leased</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date that the company started operating the business described.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Business.StartDate</lx:path>
                                    <lx:label>Business Start Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains contact details for the company applicant.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.Contact</lx:path>
                                <lx:label>Contact</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="ContactPerson" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>A contact person at the company</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.Contact.ContactPerson</lx:path>
                                      <lx:label>Contact Person</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="FirstName" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A person's first name.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.Contact.ContactPerson.FirstName</lx:path>
                                          <lx:label>First Name</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NameTitle" type="nameTitleList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The name title this person uses when they are formally addressed should be captured.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.Contact.ContactPerson.NameTitle</lx:path>
                                          <lx:label>Name Title</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Role" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The role or job position of the contact person</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.Contact.ContactPerson.Role</lx:path>
                                          <lx:label>Role</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Surname" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A person's surname.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.Contact.ContactPerson.Surname</lx:path>
                                          <lx:label>Surname</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_ContactPerson" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A contact person at the company</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.Contact.ContactPerson.x_ContactPerson</lx:path>
                                          <lx:label>Contact Person</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="OfficeFax" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Fax number for the company</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.Contact.OfficeFax</lx:path>
                                      <lx:label>Office Fax</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="OfficeMobile" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Mobile phone number for the company.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.Contact.OfficeMobile</lx:path>
                                      <lx:label>Office Mobile</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Office number for the company</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.Contact.OfficePhone</lx:path>
                                      <lx:label>Office Phone</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="PreviousRegisteredAddressDuration" type="durationType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The duration the company applicant was registered at this address.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.Contact.PreviousRegisteredAddressDuration</lx:path>
                                      <lx:label>Previous Registered Address Duration</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="PrincipalTradingAddressDuration" type="durationType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The duration of the company applicant trading at this address.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.Contact.PrincipalTradingAddressDuration</lx:path>
                                      <lx:label>Principal Trading Address Duration</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="RegisteredAddressDuration" type="durationType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The duration the company applicant has been registered at this address.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.Contact.RegisteredAddressDuration</lx:path>
                                      <lx:label>Registered Address Duration</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Email" type="emailPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An email address for the company</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Contact.Email</lx:path>
                                    <lx:label>Email</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PreferredContact" type="preferredContactCompanyList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the preferred method for contacting this applicant</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Contact.PreferredContact</lx:path>
                                    <lx:label>Preferred Contact</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PreviousRegisteredAddressEndDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date the company applicant ended registration at this address.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Contact.PreviousRegisteredAddressEndDate</lx:path>
                                    <lx:label>Previous Registered Address End Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PreviousRegisteredAddressStartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date the company applicant registered this address.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Contact.PreviousRegisteredAddressStartDate</lx:path>
                                    <lx:label>Previous Registered Address Start Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PrincipalTradingAddressStartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date the company applicant started trading at this address.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Contact.PrincipalTradingAddressStartDate</lx:path>
                                    <lx:label>Principal Trading Address Start Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegisteredAddressStartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date the company applicant registered at this address.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Contact.RegisteredAddressStartDate</lx:path>
                                    <lx:label>Registered Address Start Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="WebAddress" type="urlPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A URL for the company's public accessible web site.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Contact.WebAddress</lx:path>
                                    <lx:label>Web Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_MailingAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Mailing address of the company</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Contact.x_MailingAddress</lx:path>
                                    <lx:label>Mailing Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_PreviousRegisteredAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Previous registered address of the company</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Contact.x_PreviousRegisteredAddress</lx:path>
                                    <lx:label>Previous Registered Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_PrincipalTradingAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The official address of the company applicant where business is carried out.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Contact.x_PrincipalTradingAddress</lx:path>
                                    <lx:label>Principal Trading Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_RegisteredAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Registered address of the company</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Contact.x_RegisteredAddress</lx:path>
                                    <lx:label>Registered Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="CreditHistory" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains details of an applicant's credit history as provided both by credit reference bureaus and/or by the applicant themselves.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.CreditHistory</lx:path>
                                <lx:label>Credit History</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A comment about the applicant's credit history, possibly in mitigation or explanation of credit issues.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.CreditHistory.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EndDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date when the credit issue ended.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.CreditHistory.EndDate</lx:path>
                                    <lx:label>End Date</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsCurrent" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether the credit issue is current. Selecting "No" means the issue described is a past credit issue.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.CreditHistory.IsCurrent</lx:path>
                                    <lx:label>Is Current</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Issue" type="creditHistoryIssueList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Describes the type of Credit Issues the applicant has experienced.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.CreditHistory.Issue</lx:path>
                                    <lx:label>Issue</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date when the credit issue started.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.CreditHistory.StartDate</lx:path>
                                    <lx:label>Start Date</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unique Identifier for this Credit History Element</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.CreditHistory.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="DeclaredIncome" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for the amount of income declared by the applicant. Used for Low Doc applications when full financial years for income are not provided.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.DeclaredIncome</lx:path>
                                <lx:label>Declared Income</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="IncomeAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount of income declared by the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.DeclaredIncome.IncomeAmount</lx:path>
                                    <lx:label>Income Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NetIncomeAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The net amount of income declared by the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.DeclaredIncome.NetIncomeAmount</lx:path>
                                    <lx:label>Net Income Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Director" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains a reference to an applicant or related party that is a director of the company.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.Director</lx:path>
                                <lx:label>Director</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Director" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Contains a reference to an applicant or related party that is a director of the company.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Director.x_Director</lx:path>
                                    <lx:label>Director</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ExistingCustomer" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Indicates whether the applicant is an existing customer of the lender being applied to.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.ExistingCustomer</lx:path>
                                <lx:label>Existing Customer</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="AccountNumber" type="financialAccountType" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Account details for an account held that identifies the party as an existing customer</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.ExistingCustomer.AccountNumber</lx:path>
                                      <lx:label>Account Number</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="CustomerNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A unique number assigned by the lender used to identify the customer</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.ExistingCustomer.CustomerNumber</lx:path>
                                    <lx:label>Customer Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CustomerSince" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The approximate year that the applicant became a customer</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.ExistingCustomer.CustomerSince</lx:path>
                                    <lx:label>Customer Since</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CustomerSinceMonth" type="mmMonthList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The month (if known) that the company applicant became a customer (only if the year is also provided in 'CustomerSince').</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.ExistingCustomer.CustomerSinceMonth</lx:path>
                                    <lx:label>Customer Since Month</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="FinancialAnalysis" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for the detailed financial analysis of the company</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.FinancialAnalysis</lx:path>
                                <lx:label>Financial Analysis</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="CompanyFinancials" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Container for the reference to the financial data</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.FinancialAnalysis.CompanyFinancials</lx:path>
                                      <lx:label>Company Financials</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="x_CompanyFinancials" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Reference to the element containing the financial analysis for each financial year, to be provided when Complete Financial Analysis = "Yes"</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.FinancialAnalysis.CompanyFinancials.x_CompanyFinancials</lx:path>
                                          <lx:label>Company Financials</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyFinancials</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="AnnualPaymentOnCommitments" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Total Principle and Interest payments on borrowings due to be paid in the next 12 months</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.FinancialAnalysis.AnnualPaymentOnCommitments</lx:path>
                                    <lx:label>Annual Payment On Commitments</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CompleteFinancialAnalysis" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether a complete financial analysis will be provided electronically, or only a summary view</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.FinancialAnalysis.CompleteFinancialAnalysis</lx:path>
                                    <lx:label>Complete Financial Analysis</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ForeignTaxAssociation" type="foreignTaxAssociationType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for Foreign Tax Associations.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.ForeignTaxAssociation</lx:path>
                                <lx:label>Foreign Tax Association</lx:label>
                                <li:transactions>CAL,CDA,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="IncomePrevious" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the share of the gross (pre-tax) income or income-related amounts in the previous financial year (immediately preceding the most recent full financial year), attributable to this applicant (for year = current year -2).</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious</lx:path>
                                <lx:label>Income Previous</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Addback" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Container element for Addback detail</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback</lx:path>
                                      <lx:label>Addback</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="OtherAddback" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>Other addback that cannot otherwise be captured under a specific addback category.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.OtherAddback</lx:path>
                                            <lx:label>Other Add Back</lx:label>
                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Amount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Amount that can be an addback to the income.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.OtherAddback.Amount</lx:path>
                                                <lx:label>Amount</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Decription of the other addback.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.OtherAddback.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Allowances" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Allowances amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.Allowances</lx:path>
                                          <lx:label>Allowances</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AmortisationOfGoodwill" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Amortisation Of Goodwill amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.AmortisationOfGoodwill</lx:path>
                                          <lx:label>Amortisation Of Goodwill</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Bonus" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Bonus amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.Bonus</lx:path>
                                          <lx:label>Bonus</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CarExpense" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Car expense amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.CarExpense</lx:path>
                                          <lx:label>Car Expense</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CarryForwardLosses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Carry Forward Losses amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.CarryForwardLosses</lx:path>
                                          <lx:label>Carry Forward Losses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Depreciation" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Depreciation amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.Depreciation</lx:path>
                                          <lx:label>Depreciation</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="HomeOffice" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Home Office amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.HomeOffice</lx:path>
                                          <lx:label>Home Office</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Interest" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Interest amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.Interest</lx:path>
                                          <lx:label>Interest</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Lease" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Lease amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.Lease</lx:path>
                                          <lx:label>Lease</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonCashBenefits" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-cash benefits amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.NonCashBenefits</lx:path>
                                          <lx:label>Non Cash Benefits</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonRecurringExpenses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-recurring expense amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.NonRecurringExpenses</lx:path>
                                          <lx:label>Non Recurring Expenses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PaymentToDirector" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The payment to director that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.PaymentToDirector</lx:path>
                                          <lx:label>Payment To Director</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Salary" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Salary amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.Salary</lx:path>
                                          <lx:label>Salary</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SuperannuationExcess" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Superannuation Excess amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.Addback.SuperannuationExcess</lx:path>
                                          <lx:label>Superannuation Excess</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="CompanyProfitAfterTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the company profit after tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.CompanyProfitAfterTax</lx:path>
                                    <lx:label>Company Profit After Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CompanyProfitBeforeTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the company net profit/profit before tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.CompanyProfitBeforeTax</lx:path>
                                    <lx:label>Company Profit Before Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EndDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The end date for the stated income, generally the end of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.EndDate</lx:path>
                                    <lx:label>End Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncomeGreaterThanPreviousYear" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>When comparing past Tax Office Assessments, indicate whether the past return has improved on the previous year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.IncomeGreaterThanPreviousYear</lx:path>
                                    <lx:label>Income Greater Than Previous Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The start date for the stated income, generally the start of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.StartDate</lx:path>
                                    <lx:label>Start Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TaxOfficeAssessments" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Are Tax Office Assessments for at least the past two years available as supporting evidence for this application</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.TaxOfficeAssessments</lx:path>
                                    <lx:label>Tax Office Assessments</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Accountant" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to the company's accountant for the full previous financial year. This should be populated when the accountant for the specified period is different to that declared as the current accountant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomePrevious.x_Accountant</lx:path>
                                    <lx:label>Accountant</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="IncomePrior" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the share of the gross (pre-tax) income or income-related amounts in the prior financial year (two years prior to the most recent full financial year), attributable to this applicant (for year = current year -3).</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior</lx:path>
                                <lx:label>Income Prior</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Addback" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Container element for Addback detail</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback</lx:path>
                                      <lx:label>Addback</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="OtherAddback" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>Other addback that cannot otherwise be captured under a specific addback category.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.OtherAddback</lx:path>
                                            <lx:label>Other Add Back</lx:label>
                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Amount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Amount that can be an addback to the income.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.OtherAddback.Amount</lx:path>
                                                <lx:label>Amount</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Decription of the other addback.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.OtherAddback.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Allowances" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Allowances amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.Allowances</lx:path>
                                          <lx:label>Allowances</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AmortisationOfGoodwill" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Amortisation Of Goodwill amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.AmortisationOfGoodwill</lx:path>
                                          <lx:label>Amortisation Of Goodwill</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Bonus" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Bonus amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.Bonus</lx:path>
                                          <lx:label>Bonus</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CarExpense" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Car expense amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.CarExpense</lx:path>
                                          <lx:label>Car Expense</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CarryForwardLosses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Carry Forward Losses amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.CarryForwardLosses</lx:path>
                                          <lx:label>Carry Forward Losses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Depreciation" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Depreciation amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.Depreciation</lx:path>
                                          <lx:label>Depreciation</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="HomeOffice" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Home Office amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.HomeOffice</lx:path>
                                          <lx:label>Home Office</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Interest" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Interest amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.Interest</lx:path>
                                          <lx:label>Interest</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Lease" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Lease amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.Lease</lx:path>
                                          <lx:label>Lease</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonCashBenefits" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-cash benefits amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.NonCashBenefits</lx:path>
                                          <lx:label>Non Cash Benefits</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonRecurringExpenses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-recurring expense amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.NonRecurringExpenses</lx:path>
                                          <lx:label>Non Recurring Expenses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PaymentToDirector" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The payment to director that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.PaymentToDirector</lx:path>
                                          <lx:label>Payment To Director</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Salary" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Salary amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.Salary</lx:path>
                                          <lx:label>Salary</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SuperannuationExcess" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Superannuation Excess amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.Addback.SuperannuationExcess</lx:path>
                                          <lx:label>Superannuation Excess</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="CompanyProfitAfterTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the company profit after tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.CompanyProfitAfterTax</lx:path>
                                    <lx:label>Company Profit After Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CompanyProfitBeforeTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the company net profit/profit before tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.CompanyProfitBeforeTax</lx:path>
                                    <lx:label>Company Profit Before Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EndDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The end date for the stated income, generally the end of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.EndDate</lx:path>
                                    <lx:label>End Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncomeGreaterThanPreviousYear" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>When comparing past Tax Office Assessments, indicate whether the past return has improved on the Prior year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.IncomeGreaterThanPreviousYear</lx:path>
                                    <lx:label>Income Greater Than Previous Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The start date for the stated income, generally the start of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.StartDate</lx:path>
                                    <lx:label>Start Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TaxOfficeAssessments" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Are Tax Office Assessments for at least the past two years available as supporting evidence for this application</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.TaxOfficeAssessments</lx:path>
                                    <lx:label>Tax Office Assessments</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Accountant" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to the company's accountant for the full prior financial year. This should be populated when the accountant for the specified period is different to that declared as the current accountant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomePrior.x_Accountant</lx:path>
                                    <lx:label>Accountant</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="IncomeRecent" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the share of the gross (pre-tax) income or income-related amounts in the most recent full financial year, attributable to this applicant (year = current year -1).</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent</lx:path>
                                <lx:label>Income Recent</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Addback" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Container element for Addback detail</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback</lx:path>
                                      <lx:label>Addback</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="OtherAddback" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>Other addback that cannot otherwise be captured under a specific addback category.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.OtherAddback</lx:path>
                                            <lx:label>Other Add Back</lx:label>
                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Amount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Amount that can be an addback to the income.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.OtherAddback.Amount</lx:path>
                                                <lx:label>Amount</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Decription of the other addback.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.OtherAddback.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Allowances" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Allowances amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.Allowances</lx:path>
                                          <lx:label>Allowances</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AmortisationOfGoodwill" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Amortisation Of Goodwill amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.AmortisationOfGoodwill</lx:path>
                                          <lx:label>Amortisation Of Goodwill</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Bonus" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Bonus amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.Bonus</lx:path>
                                          <lx:label>Bonus</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CarExpense" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Car expense amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.CarExpense</lx:path>
                                          <lx:label>Car Expense</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CarryForwardLosses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Carry Forward Losses amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.CarryForwardLosses</lx:path>
                                          <lx:label>Carry Forward Losses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Depreciation" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Depreciation amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.Depreciation</lx:path>
                                          <lx:label>Depreciation</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="HomeOffice" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Home Office amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.HomeOffice</lx:path>
                                          <lx:label>Home Office</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Interest" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Interest amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.Interest</lx:path>
                                          <lx:label>Interest</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Lease" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Lease amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.Lease</lx:path>
                                          <lx:label>Lease</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonCashBenefits" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-cash benefits amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.NonCashBenefits</lx:path>
                                          <lx:label>Non Cash Benefits</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonRecurringExpenses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-recurring expense amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.NonRecurringExpenses</lx:path>
                                          <lx:label>Non Recurring Expenses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PaymentToDirector" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The payment to director that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.PaymentToDirector</lx:path>
                                          <lx:label>Payment To Director</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Salary" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Salary amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.Salary</lx:path>
                                          <lx:label>Salary</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SuperannuationExcess" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Superannuation Excess amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.Addback.SuperannuationExcess</lx:path>
                                          <lx:label>Superannuation Excess</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="CompanyProfitAfterTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the company profit after tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.CompanyProfitAfterTax</lx:path>
                                    <lx:label>Company Profit After Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CompanyProfitBeforeTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the company net profit/profit before tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.CompanyProfitBeforeTax</lx:path>
                                    <lx:label>Company Profit Before Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EndDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The end date for the stated income, generally the end of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.EndDate</lx:path>
                                    <lx:label>End Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncomeGreaterThanPreviousYear" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>When comparing past Tax Office Assessments, indicate whether the past return has improved on the previous year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.IncomeGreaterThanPreviousYear</lx:path>
                                    <lx:label>Income Greater Than Previous Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProofCode" type="proofCodeCompanyList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Code to indicate the type of evidence of the income</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.ProofCode</lx:path>
                                    <lx:label>Proof Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProofSighted" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Has the Income evidence been sighted b the loan writer or assessor</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.ProofSighted</lx:path>
                                    <lx:label>Proof Sighted</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The start date for the stated income, generally the start of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.StartDate</lx:path>
                                    <lx:label>Start Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TaxOfficeAssessments" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Are Tax Office Assessments for at least the past two years available as supporting evidence for this application where the applicant is self-employed.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.TaxOfficeAssessments</lx:path>
                                    <lx:label>Tax Office Assessments</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Accountant" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to the company's accountant for the full most recent financial year. This should be populated when the accountant for the specified period is different to that declared as the current accountant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomeRecent.x_Accountant</lx:path>
                                    <lx:label>Accountant</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="IncomeYearToDate" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the current year to date share of the gross (pre-tax) income or income-related amounts, attributable to this applicant.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate</lx:path>
                                <lx:label>Income Year To Date</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Addback" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Container element for Addback detail</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback</lx:path>
                                      <lx:label>Addback</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="OtherAddback" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>Other addback that cannot otherwise be captured under a specific addback category.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.OtherAddback</lx:path>
                                            <lx:label>Other Add Back</lx:label>
                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Amount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Amount that can be an addback to the income.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.OtherAddback.Amount</lx:path>
                                                <lx:label>Amount</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Decription of the other addback.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.OtherAddback.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Allowances" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Allowances amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.Allowances</lx:path>
                                          <lx:label>Allowances</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AmortisationOfGoodwill" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Amortisation Of Goodwill amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.AmortisationOfGoodwill</lx:path>
                                          <lx:label>Amortisation Of Goodwill</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Bonus" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Bonus amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.Bonus</lx:path>
                                          <lx:label>Bonus</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CarExpense" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Car expense amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.CarExpense</lx:path>
                                          <lx:label>Car Expense</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CarryForwardLosses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Carry Forward Losses amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.CarryForwardLosses</lx:path>
                                          <lx:label>Carry Forward Losses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Depreciation" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Depreciation amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.Depreciation</lx:path>
                                          <lx:label>Depreciation</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="HomeOffice" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Home Office amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.HomeOffice</lx:path>
                                          <lx:label>Home Office</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Interest" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Interest amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.Interest</lx:path>
                                          <lx:label>Interest</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Lease" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Lease amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.Lease</lx:path>
                                          <lx:label>Lease</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonCashBenefits" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-cash benefits amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.NonCashBenefits</lx:path>
                                          <lx:label>Non Cash Benefits</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonRecurringExpenses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-recurring expense amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.NonRecurringExpenses</lx:path>
                                          <lx:label>Non Recurring Expenses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PaymentToDirector" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The payment to director that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.PaymentToDirector</lx:path>
                                          <lx:label>Payment To Director</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Salary" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Salary amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.Salary</lx:path>
                                          <lx:label>Salary</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SuperannuationExcess" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Superannuation Excess amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.Addback.SuperannuationExcess</lx:path>
                                          <lx:label>Superannuation Excess</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="CompanyProfitAfterTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the company profit after tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.CompanyProfitAfterTax</lx:path>
                                    <lx:label>Company Profit After Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CompanyProfitBeforeTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the company net profit/profit before tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.CompanyProfitBeforeTax</lx:path>
                                    <lx:label>Company Profit Before Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EndDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The end date for the stated income, generally the end of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.EndDate</lx:path>
                                    <lx:label>End Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The start date for the stated income, generally the start of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.IncomeYearToDate.StartDate</lx:path>
                                    <lx:label>Start Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Partner" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Container element for details of the partners that constitute this partnership.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.Partner</lx:path>
                                <lx:label>Partner</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="PartnerType" type="partnerTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The Partner Type describes whether the partner is a general or limited partner.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Partner.PartnerType</lx:path>
                                    <lx:label>Partner Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PercentOwned" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The ownership percentage for the specified partner.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Partner.PercentOwned</lx:path>
                                    <lx:label>Percent Owned</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Partner" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Contains a reference to an applicant or related party that is a partner of the business. These can be a person, a business enterprise or a trust.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Partner.x_Partner</lx:path>
                                    <lx:label>Partner</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="RelatedLegalEntities" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the related legal entities.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.RelatedLegalEntities</lx:path>
                                <lx:label>Related Legal Entities</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="CurrentCustomers" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Free text to capture information about the related legal entites that are already customers of the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.RelatedLegalEntities.CurrentCustomers</lx:path>
                                    <lx:label>Current Customers</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EntityCount" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of related legal entities.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.RelatedLegalEntities.EntityCount</lx:path>
                                    <lx:label>Entity Count</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ResponsibleLending" type="responsibleLendingType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Captures information about any significant changes in circumstances that will adversely affect the applicants' ability to make contracted loan repayments; the nature of the change, and how they will collectively continue to make repayments.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.ResponsibleLending</lx:path>
                                <lx:label>Responsible Lending</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="Shareholder" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains a reference to an applicant or related party that is a shareholder of the company.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.Shareholder</lx:path>
                                <lx:label>Shareholder</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="PercentOwned" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The ownership percentage for the specified shareholder</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Shareholder.PercentOwned</lx:path>
                                    <lx:label>Percent Owned</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Shareholder" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Contains a reference to an applicant or related party that is a shareholder of the company.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.Shareholder.x_Shareholder</lx:path>
                                    <lx:label>Shareholder</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="SourceOfWealth" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The source of the applicant's entire body of wealth.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyApplicant.SourceOfWealth</lx:path>
                                <lx:label>Source Of Wealth</lx:label>
                                <li:transactions>CAL,CDA,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Detail" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The details of the source of wealth.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.SourceOfWealth.Detail</lx:path>
                                    <lx:label>Detail</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Percentage" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The percentage of the wealth from this source.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.SourceOfWealth.Percentage</lx:path>
                                    <lx:label>Percentage</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="sourceOfFundsTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of source of wealth.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyApplicant.SourceOfWealth.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="ABN" type="abnPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the ABN of the company applicant</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.ABN</lx:path>
                              <lx:label>ABN</lx:label>
                              <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ABNVerified" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate that the business number has been verified by a search</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.ABNVerified</lx:path>
                              <lx:label>ABN Verified</lx:label>
                              <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ACN" type="acnPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the ACN of the company applicant</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.ACN</lx:path>
                              <lx:label>ACN</lx:label>
                              <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="AllowDirectMarketing" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether this Party will allow their details to be used for direct marketing purposes.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.AllowDirectMarketing</lx:path>
                              <lx:label>Allow Direct Marketing</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="AllowThirdPartyDisclosure" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether this Party will allow their details to be disclosed to third parties affiliated with this lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.AllowThirdPartyDisclosure</lx:path>
                              <lx:label>Allow Third Party Disclosure</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ApplicantType" type="applicantTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Whether the loan party is a borrower or guarantor</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.ApplicantType</lx:path>
                              <lx:label>Applicant Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ARBN" type="arbnPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The Australian Registered Body Number (ARBN) of the company applicant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.ARBN</lx:path>
                              <lx:label>ARBN</lx:label>
                              <li:transactions>CAL,CDA,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ARSN" type="arsnPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The Australian Registered Scheme Number (ARSN) of the company applicant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.ARSN</lx:path>
                              <lx:label>ARSN</lx:label>
                              <li:transactions>CAL,CDA,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BusinessName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The business name or trading name of the company should be captured.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.BusinessName</lx:path>
                              <lx:label>Business Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BusinessNameSameAsCompanyName" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates if the Business Name is the same as the Company Name.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.BusinessNameSameAsCompanyName</lx:path>
                              <lx:label>Business Name Same As Company Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BusinessStructure" type="businessStructureExcludingTrustList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Legal structure of this business (only populated where this applicant is a Business Enterprise). Trust is not a valid entry as trust entity applicants need to be entered as a Trust Applicant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.BusinessStructure</lx:path>
                              <lx:label>Business Structure</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CompanyName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The legal name of the company should be captured.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.CompanyName</lx:path>
                              <lx:label>Company Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CompanyNumber" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the registered Company Number of the company applicant</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.CompanyNumber</lx:path>
                              <lx:label>Company Number</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CreditStatus" type="creditStatusList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Summarises whether there are any issues indicated the credit history of the applicant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.CreditStatus</lx:path>
                              <lx:label>Credit Status</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CustomerTypeCode" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A lender assigned code for the customer type.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.CustomerTypeCode</lx:path>
                              <lx:label>Customer Type Code</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CustomerTypeDescription" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A lender assigned description of the customer type.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.CustomerTypeDescription</lx:path>
                              <lx:label>Customer Type Description</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DateRegistered" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates the date when the company was registered.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.DateRegistered</lx:path>
                              <lx:label>Date Registered</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ExchangeListedOn" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Contains the name of the stock exchange that a public company is listed on. Is only relevant where company is of a publicly listed type.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.ExchangeListedOn</lx:path>
                              <lx:label>Exchange Listed On</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ExemptStatus" type="exemptStatusList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the company status is exempt or non-exempt.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.ExemptStatus</lx:path>
                              <lx:label>Exempt Status</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="GovernmentOrganisation" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The applicant is a government organisation.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.GovernmentOrganisation</lx:path>
                              <lx:label>Government Organisation</lx:label>
                              <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="GSTRegisteredDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>If registered for GST, the date from which the company is registered</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.GSTRegisteredDate</lx:path>
                              <lx:label>GST Registered Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IRDNumber" type="irdNumberPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The IRD Number of this Company Applicant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.IRDNumber</lx:path>
                              <lx:label>IRD Number</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsExistingCustomer" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the applicant is an existing customer of the lender being applied to.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.IsExistingCustomer</lx:path>
                              <lx:label>Is Existing Customer</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsHoldingCompany" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicate if the company is a holding company</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.IsHoldingCompany</lx:path>
                              <lx:label>Is Holding Company</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LicenceNumber" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The licence number of the company should be captured.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.LicenceNumber</lx:path>
                              <lx:label>Licence Number</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LocalAgentOfForeignCompany" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The applicant is a local agent of a foreign company.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.LocalAgentOfForeignCompany</lx:path>
                              <lx:label>Local Agent Of Foreign Company</lx:label>
                              <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NumberOfCreditHistoryIssues" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The number of credit history issues associated with this applicant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.NumberOfCreditHistoryIssues</lx:path>
                              <lx:label>Number Of Credit History Issues</lx:label>
                              <li:transactions>ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NumberOfDirectors" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The number of directors that company has on its Board of Directors.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.NumberOfDirectors</lx:path>
                              <lx:label>Number Of Directors</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NumberOfPartners" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The number of partners that make up the partnership.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.NumberOfPartners</lx:path>
                              <lx:label>Number Of Partners</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NumberOfShareholders" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The approximate number of shareholders of the company</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.NumberOfShareholders</lx:path>
                              <lx:label>Number Of Shareholders</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NZBN" type="nzbnPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the NZBN of the company applicant</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.NZBN</lx:path>
                              <lx:label>NZBN</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NZBNVerified" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate that the business number has been verified by a search</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.NZBNVerified</lx:path>
                              <lx:label>NZBN Verified</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OECDCRSStatus" type="oecdCRSStatusList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the company is considered Active or Passive for OECD Common Reporting Standard purposes.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.OECDCRSStatus</lx:path>
                              <lx:label>OECD CRS Status</lx:label>
                              <li:transactions>CAL,CDA,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PrimaryApplicant" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the lender should consider the party to be the primary applicant when processing the loan application. The significance of the primary applicant to the application varies from lender to lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.PrimaryApplicant</lx:path>
                              <lx:label>Primary Applicant</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PublicAuthority" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether a company is a public authority like a council or government department.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.PublicAuthority</lx:path>
                              <lx:label>Public Authority</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ReasonForObtainingLoan" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A field to indicate what is the reason for requesting or obtaining the loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.ReasonForObtainingLoan</lx:path>
                              <lx:label>Reason For Obtaining Loan</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RegisteredForGST" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate whether the company is registered for GST</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.RegisteredForGST</lx:path>
                              <lx:label>Registered For GST</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RegisteredForLTC" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The company is registered to use Look Through Company (LTC) rules for an income year.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.RegisteredForLTC</lx:path>
                              <lx:label>Registered For LTC</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RegisteredInCountry" type="countryCodeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates the country where the company is registered.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.RegisteredInCountry</lx:path>
                              <lx:label>Registered In Country</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RegisteredInState" type="auStateList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates the Australian state where the company is registered.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.RegisteredInState</lx:path>
                              <lx:label>Registered In State</lx:label>
                              <li:transactions>CAL,CDA,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this company applicant within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="TradingMoreThanTwoYears" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicate if the business has been trading more than 2 years</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.TradingMoreThanTwoYears</lx:path>
                              <lx:label>Trading More Than Two Years</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="TypeOfIncorporation" type="typeOfIncorporationList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The company structure, in the form of the type of incorporation.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.TypeOfIncorporation</lx:path>
                              <lx:label>Type Of Incorporation</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Accountant" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Link to the company's current accountant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.x_Accountant</lx:path>
                              <lx:label>Accountant</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_SoleTrader" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Link to the natural person that is the sole trader legally liable for this business enterprise.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.x_SoleTrader</lx:path>
                              <lx:label>Sole Trader</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Solicitor" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Link to the applicant's solicitor.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.x_Solicitor</lx:path>
                              <lx:label>Solicitor</lx:label>
                              <li:transactions>CAL,DAS,ACC</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_TradeReference" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Link to the related company that is providing a trade reference.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyApplicant.x_TradeReference</lx:path>
                              <lx:label>Trade Reference</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="CompanyFinancials" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Contain element for company or self-employed financial report data</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.CompanyFinancials</lx:path>
                          <lx:label>Company Financials</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="BalanceSheet" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Balance sheet financials</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet</lx:path>
                                <lx:label>Balance Sheet</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Assets" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Asset items from the financial report</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Assets</lx:path>
                                      <lx:label>Assets</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Cash" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Cash at bank</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Assets.Cash</lx:path>
                                          <lx:label>Cash</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CashAndCashEquivalents" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>As listed in financial report</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Assets.CashAndCashEquivalents</lx:path>
                                          <lx:label>Cash And Cash Equivalents</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Intangibles" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Intangible assets, or goodwill</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Assets.Intangibles</lx:path>
                                          <lx:label>Intangibles</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Inventory" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Value of inventory, being materials, work in progress, and stock on hand</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Assets.Inventory</lx:path>
                                          <lx:label>Inventory</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TangibleAssets" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>(Total Assets - Intangibles)</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Assets.TangibleAssets</lx:path>
                                          <lx:label>Tangible Assets</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TotalAssets" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Total assets</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Assets.TotalAssets</lx:path>
                                          <lx:label>Total Assets</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TotalCurrentAssets" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Total current assets</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Assets.TotalCurrentAssets</lx:path>
                                          <lx:label>Total Current Assets</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TradeDebtors" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Value of the trade debtors (receivables)</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Assets.TradeDebtors</lx:path>
                                          <lx:label>Trade Debtors</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="Liabilities" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Liability items from the financial report</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Liabilities</lx:path>
                                      <lx:label>Liabilities</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="DebtAtNominalValue" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Total Borrowings</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Liabilities.DebtAtNominalValue</lx:path>
                                          <lx:label>Debt At Nominal Value</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TotalCurrentLiabilities" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Total current liabilities</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Liabilities.TotalCurrentLiabilities</lx:path>
                                          <lx:label>Total Current Liabilities</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TotalLiabilities" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Total liabilities</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Liabilities.TotalLiabilities</lx:path>
                                          <lx:label>Total Liabilities</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TradeCreditors" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Value of the trade creditors (accounts payable)</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Liabilities.TradeCreditors</lx:path>
                                          <lx:label>Trade Creditors</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Equity" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Equity on this balance sheet.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.Equity</lx:path>
                                    <lx:label>Equity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NetAssets" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>(Total Assets - Total Liabilities)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.NetAssets</lx:path>
                                    <lx:label>Net Assets</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaidUpCapital" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Paid Up Capital</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.PaidUpCapital</lx:path>
                                    <lx:label>Paid Up Capital</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RetainedEarnings" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount of retained earnings</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.RetainedEarnings</lx:path>
                                    <lx:label>Retained Earnings</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ShareholderFunds" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Shareholder Funds on this balance sheet.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.BalanceSheet.ShareholderFunds</lx:path>
                                    <lx:label>Shareholder Funds</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="CurrentMarketData" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Equity data from the financial report</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyFinancials.CurrentMarketData</lx:path>
                                <lx:label>Current Market Data</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ValueOfCommonEquity" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Number of Common Shares x current Market Price</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.CurrentMarketData.ValueOfCommonEquity</lx:path>
                                    <lx:label>Value Of Common Equity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ValueOfMinorityInterests" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Market Value of any shares held by a minority shareholder (if any)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.CurrentMarketData.ValueOfMinorityInterests</lx:path>
                                    <lx:label>Value Of Minority Interests</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ValueOfPreferredEquity" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Number of Preference Shares x current Market Price</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.CurrentMarketData.ValueOfPreferredEquity</lx:path>
                                    <lx:label>Value Of Preferred Equity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ProfitAndLoss" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Profit and Loss (income and expense) financials</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CompanyFinancials.ProfitAndLoss</lx:path>
                                <lx:label>Profit And Loss</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AmortisationExpense" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Amortisation as recorded in the financial report</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.ProfitAndLoss.AmortisationExpense</lx:path>
                                    <lx:label>Amortisation Expense</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CostOfGoodsSold" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Cost of goods sold on this profit and loss statement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.ProfitAndLoss.CostOfGoodsSold</lx:path>
                                    <lx:label>Cost of Goods Sold</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DepreciationExpense" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Depreciation as recorded in the financial report</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.ProfitAndLoss.DepreciationExpense</lx:path>
                                    <lx:label>Depreciation Expense</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EBIT" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Earnings before interest and tax</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.ProfitAndLoss.EBIT</lx:path>
                                    <lx:label>EBIT</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EBITDA" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Earnings before interest, tax, depreciation and amortisation</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.ProfitAndLoss.EBITDA</lx:path>
                                    <lx:label>EBITDA</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GrossProfit" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Sales less cost of goods sold (COGS)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.ProfitAndLoss.GrossProfit</lx:path>
                                    <lx:label>Gross Profit</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncomeTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Income tax paid or refunded for the financial year. A positive number indicates tax paid, a negative number indicates tax refunded.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.ProfitAndLoss.IncomeTax</lx:path>
                                    <lx:label>Income Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InterestExpense" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Total interest expense as listed in the financial report, including leasing expenses</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.ProfitAndLoss.InterestExpense</lx:path>
                                    <lx:label>Interest Expense</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NetProfitBeforeTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Net profit before tax</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.ProfitAndLoss.NetProfitBeforeTax</lx:path>
                                    <lx:label>Net Profit Before Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentsToDirectors" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Gross payments paid to the directors of the company</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.ProfitAndLoss.PaymentsToDirectors</lx:path>
                                    <lx:label>Payments To Directors</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Sales" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Total sales revenue</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.ProfitAndLoss.Sales</lx:path>
                                    <lx:label>Sales</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalExpenses" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Total Expenses on this profit and loss statement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CompanyFinancials.ProfitAndLoss.TotalExpenses</lx:path>
                                    <lx:label>TotalExpenses</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="EndDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The end date of the financials, generally a financial year end</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyFinancials.EndDate</lx:path>
                              <lx:label>End Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this company financials within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyFinancials.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="StartDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The start date of the financials, generally a financial year start</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyFinancials.StartDate</lx:path>
                              <lx:label>Start Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CompanyFinancials.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="ContributionFunds" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Details the amounts and sources of funds that will be contributed by the applicant to complete the contemplated transaction, generally a property purchase but may also be a refinance or other purchase. There may be 0 or more contributions.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.ContributionFunds</lx:path>
                          <lx:label>Contribution Funds</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="Amount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The amount of the particular contribution.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ContributionFunds.Amount</lx:path>
                              <lx:label>Amount</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Description" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A free form description of the contribution used when the selected Contribution Fund Type = "Other"</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ContributionFunds.Description</lx:path>
                              <lx:label>Description</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Loan" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the contribution is borrowed from a third party.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ContributionFunds.Loan</lx:path>
                              <lx:label>Loan</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this contribution funds within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ContributionFunds.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Type" type="contributionFundsTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of the particular contribution.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ContributionFunds.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ContributionFunds.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_AssociatedLoanAccount" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A cross reference to the loan account (Loan Details element) to which the contribution funds belong.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ContributionFunds.x_AssociatedLoanAccount</lx:path>
                              <lx:label>Associated Loan Account</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.LoanDetails</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_FundsHeldInAccount" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Reference to capture the bank account details for funds held with a financial institution.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ContributionFunds.x_FundsHeldInAccount</lx:path>
                              <lx:label>Funds Held In Account</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.Liability</lx:target>
                                <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="CustomerTransactionAnalysis" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Analysis of customer bank account transaction data.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.CustomerTransactionAnalysis</lx:path>
                          <lx:label>Customer Transaction Analysis</lx:label>
                          <li:transactions>CAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="CategorySet" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>A set of customer bank account transaction data classified into aggregated categories.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CustomerTransactionAnalysis.CategorySet</lx:path>
                                <lx:label>Category Set</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="AggregatedTransactions" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>A category of customer bank account transaction data.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.CustomerTransactionAnalysis.CategorySet.AggregatedTransactions</lx:path>
                                      <lx:label>Aggregated Transactions</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Amount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The dollar amount in this category of aggregated transaction data.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CustomerTransactionAnalysis.CategorySet.AggregatedTransactions.Amount</lx:path>
                                          <lx:label>Amount</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AmountOnMonthlyBasis" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The dollar amount on a monthly basis in this category of aggregated transaction data.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CustomerTransactionAnalysis.CategorySet.AggregatedTransactions.AmountOnMonthlyBasis</lx:path>
                                          <lx:label>Amount On Monthly Basis</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Category" type="aggregatedTransactionsCategoryTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The name of the category of aggregated transaction data.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CustomerTransactionAnalysis.CategorySet.AggregatedTransactions.Category</lx:path>
                                          <lx:label>Category</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OtherCategory" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A description of the category of aggregated transaction data, to be used when the category is 'Other'.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CustomerTransactionAnalysis.CategorySet.AggregatedTransactions.OtherCategory</lx:path>
                                          <lx:label>Other Category</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OtherCategoryCode" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A code indicating the category of this aggregate transaction, to be used when the category is 'Other'.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CustomerTransactionAnalysis.CategorySet.AggregatedTransactions.OtherCategoryCode</lx:path>
                                          <lx:label>Other Category Code</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Type" type="transactionTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The type of transaction: Debit or Credit.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CustomerTransactionAnalysis.CategorySet.AggregatedTransactions.Type</lx:path>
                                          <lx:label>Type</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.CustomerTransactionAnalysis.CategorySet.AggregatedTransactions.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="CategorySetName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the set of categorised data.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CustomerTransactionAnalysis.CategorySet.CategorySetName</lx:path>
                                    <lx:label>CategorySetName</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CustomerTransactionAnalysis.CategorySet.UniqueID</lx:path>
                                    <lx:label>UniqueID</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Result" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The results of the customer transaction analysis.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.CustomerTransactionAnalysis.Result</lx:path>
                                <lx:label>Result</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ManualCheckDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the manual check that the analysis has determined to be required.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CustomerTransactionAnalysis.Result.ManualCheckDetails</lx:path>
                                    <lx:label>Manual Check Details</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ReferForManualCheck" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The analysis has determined a manual check by the lender is required.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.CustomerTransactionAnalysis.Result.ReferForManualCheck</lx:path>
                                    <lx:label>Refer For Manual Check</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="AnalysisID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>An identifier for the customer transaction analysis provided by the data aggregator.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CustomerTransactionAnalysis.AnalysisID</lx:path>
                              <lx:label>Analysis ID</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DataAggregator" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The provider of the aggregated results of the customer transaction analysis.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CustomerTransactionAnalysis.DataAggregator</lx:path>
                              <lx:label>Data Aggregator</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="EndDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The end date of the period of the transaction analysis.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CustomerTransactionAnalysis.EndDate</lx:path>
                              <lx:label>End Date</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this customer transaction analysis within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CustomerTransactionAnalysis.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="StartDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The start date of the period of the transaction analysis.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CustomerTransactionAnalysis.StartDate</lx:path>
                              <lx:label>Start Date</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.CustomerTransactionAnalysis.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Declarations" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Captures declarations such as Broker Declarations that apply to the whole application.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.Declarations</lx:path>
                          <lx:label>Declarations</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="BrokerDeclarations" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for Broker Declarations.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Declarations.BrokerDeclarations</lx:path>
                                <lx:label>Broker Declarations</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ApplicantIsTheSamePersonShownInThePhotograph" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that the applicant is the same as the person shown in the photograph in the documents provided with this application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.ApplicantIsTheSamePersonShownInThePhotograph</lx:path>
                                    <lx:label>Applicant Is The Same Person Shown In The Photograph</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AssetIncomeLiabilityAndExpenseObtained" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that the applicant's assets, income, liability and expense information provided within the application are those obtained from the applicant(s) during the preliminary assessment.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.AssetIncomeLiabilityAndExpenseObtained</lx:path>
                                    <lx:label>Asset Income Liability And Expense Obtained</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AwareThatWilfulBreachOfIdentificationIsPunishable" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that they are aware that any wilful breach of identification obligations is made punishable under the Anti-Money Laundering/ Counter-Terrorism Financing Act 2006.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.AwareThatWilfulBreachOfIdentificationIsPunishable</lx:path>
                                    <lx:label>Aware That Wilful Breach Of Identification Is Punishable</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="BeenShareholderOrOfficerOfACompanyInDifficulties" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that they have asked whether applicant/s has been a shareholder or an officer of any company in difficulties (including bankruptcy).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.BeenShareholderOrOfficerOfACompanyInDifficulties</lx:path>
                                    <lx:label>Been Shareholder Or Officer Of A Company In Difficulties</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CompletedOriginalDocumentsAndSightedInThePresenceOfApplicant" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that they have completed the original documents detailed in this application and sighted it in the presence of all applicants.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.CompletedOriginalDocumentsAndSightedInThePresenceOfApplicant</lx:path>
                                    <lx:label>Completed Original Documents And Sighted In The Presence Of Applicant</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DateAndTimeOfDeclaration" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Date and time of the broker's declaration.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.DateAndTimeOfDeclaration</lx:path>
                                    <lx:label>Date And Time Of Declaration</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DetailsOfConflictsOfInterestProvided" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that they have provided details of any conflicts of interest relating to this application(s).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.DetailsOfConflictsOfInterestProvided</lx:path>
                                    <lx:label>Details Of Conflicts Of Interest Provided</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="HaveNoConcernsAboutApplicantIdentity" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that they have no concern about applicant's identity and believe them to be who they say they are despite not being successfully electronically verified.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.HaveNoConcernsAboutApplicantIdentity</lx:path>
                                    <lx:label>Have No Concerns About Applicant Identity</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InterestOnlyMeetsRequirements" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that the interest only period aligns with the applicant's requirements.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.InterestOnlyMeetsRequirements</lx:path>
                                    <lx:label>Declaration from the broker that the interest only period aligns with the applicant's requirements.</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InterestOnlyMeetsRequirementsDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The text of the declaration from the broker stating that the interest only period aligns with the applicant's requirements.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.InterestOnlyMeetsRequirementsDescription</lx:path>
                                    <lx:label>The text of the declaration from the broker stating that the interest only period aligns with the applicant's requirements.</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InterestOnlyRisksExplained" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that the applicant has had the additional risks and costs of an interest only term explained to them. The applicant has had the following explained: interest only loan repayments will not pay off any principal during the interest only term; the repayments required to pay out the loan will increase after the interest only period ends to cover both interest and principal reductions; and the applicant may pay more over the life of their loan than if there was no interest only term.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.InterestOnlyRisksExplained</lx:path>
                                    <lx:label>Declaration from the broker that the applicant has had the additional risks and costs of an interest only term explained to them.</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InterestOnlyRisksExplainedDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The text of the declaration from the broker explaining the applicant has had the additional risks and costs of an interest only term explained to them.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.InterestOnlyRisksExplainedDescription</lx:path>
                                    <lx:label>The text of the declaration from the broker explaining that the applicant has had the additional risks and costs of an interest only term explained to them.</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LenderInformedAboutAdditionalRequirementAndInformation" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that they have informed the lender of any additional information, requirements and objectives of which they have been informed or have identified.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.LenderInformedAboutAdditionalRequirementAndInformation</lx:path>
                                    <lx:label>Lender Informed About Additional Requirement And Information</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LMIApplicableAndExplainedToTheApplicant" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that they have explained Lenders Mortgage Insurance (LMI) to the applicant(s) where is applicable.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.LMIApplicableAndExplainedToTheApplicant</lx:path>
                                    <lx:label>LMI Applicable And Explained To The Applicant</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProductMeetsApplicantRequirements" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that they made reasonable inquiries and can state that the product meets the applicant(s) requirements and objectives as outlined by the applicant(s) in this application form and consider the loan as not unsuitable.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.ProductMeetsApplicantRequirements</lx:path>
                                    <lx:label>Product Meets Applicant Requirements</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RefinancingAndDebtConsolidationMeetsObjectives" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that they have considered the costs, risks and benefits for the applicant(s) of the refinancing/debt consolidation, including: fees payable as a result of the refinancing/debt consolidation (both in ending the old loan and in establishing the new loan); a comparison of the interest rate of the current loan and the proposed new loan; a comparison of the loan term and repayment amount of the current loan and the proposed new loan and whether the applicant(s) would get improved cash flow as a result of the refinancing; whether any debt consolidation involved would achieve administrative convenience for the applicant(s); whether the applicant can afford to make the change to the product and meet the new payments without substantial hardship; and whether the applicant will end up paying more interest if debt is refinanced/ consolidated. I am satisfied that the refinancing/debt consolidation meets the applicant(s) requirements and objectives.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.RefinancingAndDebtConsolidationMeetsObjectives</lx:path>
                                    <lx:label>Refinancing And Debt Consolidation Meets Objectives</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SubmittedSimilarApplicationToOtherLenders" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that they have asked whether applicant/s has submitted similar application to other lenders.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.SubmittedSimilarApplicationToOtherLenders</lx:path>
                                    <lx:label>Submitted Similar Application To Other Lenders</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UnsatisfiedJudgmentAgainstAnyApplicantOrCompany" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that they have asked whether there is any unsatisfied judgement entered in any court against any applicant, or any company of which any applicant is or was a shareholder or officer.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.UnsatisfiedJudgmentAgainstAnyApplicantOrCompany</lx:path>
                                    <lx:label>Unsatisfied Judgment Against Any Applicant Or Company</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="VerifyThatTheDataCapturedIsCorrect" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Declaration from the broker that they are verifying that the data captured is correct.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Declarations.BrokerDeclarations.VerifyThatTheDataCapturedIsCorrect</lx:path>
                                    <lx:label>Verify That The Data Captured Is Correct</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="DepositAccountDetails" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Carries the details of a deposit account being opened.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.DepositAccountDetails</lx:path>
                          <lx:label>Deposit Account Details</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="AccountOwners" type="percentOwnedType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Identifies the parties that are prospective owners of this account.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.DepositAccountDetails.AccountOwners</lx:path>
                                <lx:label>Account Owners</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="Commission" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information relating to the commission to be paid to the business partner that has sourced the deposit account application for the lender. This partner is typically a third party.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.DepositAccountDetails.Commission</lx:path>
                                <lx:label>Commission</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="CommissionAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount of upfront commission to be paid to the sales channel in the event that deposit account application is successful.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.Commission.CommissionAmount</lx:path>
                                    <lx:label>Commission Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CommissionPaid" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether a commission is payable to the sales channel in respect of the deposit account application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.Commission.CommissionPaid</lx:path>
                                    <lx:label>Commission Paid</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CommissionStructure" type="commissionStructureList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the commission plan involves payment of an upfront commission only, an upfront commission and a trail commission, a trail commission only, or another type of structure.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.Commission.CommissionStructure</lx:path>
                                    <lx:label>Commission Structure</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherCommissionStructureDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the commission structure if an accurate description is not specified in the enumerated values. Must be provided when Commission Structure = "Other"</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.Commission.OtherCommissionStructureDescription</lx:path>
                                    <lx:label>Other Commission Structure Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PromotionCode" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A code that specifies a particular promotion or special commission offer that applies to the deposit account application, specified by the submitting sales channel in order to tag the deal with the promotion code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.Commission.PromotionCode</lx:path>
                                    <lx:label>Promotion Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ThirdPartyReferee" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the sales channel is a third party referrer of business for the purpose of commission disclosure requirements.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.Commission.ThirdPartyReferee</lx:path>
                                    <lx:label>Third Party Referee</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Trail" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the trail commission percentage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.Commission.Trail</lx:path>
                                    <lx:label>Trail</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="DocumentationInstructions" type="documentationInstructionsType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information specifying where the lender is to send the account documents once the deposit account has been approved. This nomination over-rides an equivalent one at the Application level.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.DepositAccountDetails.DocumentationInstructions</lx:path>
                                <lx:label>Documentation Instructions</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="FeaturesSelected" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains information about the product features requested by the applicant or applicants.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.DepositAccountDetails.FeaturesSelected</lx:path>
                                <lx:label>Features Selected</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="ExtraFeature" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Information about a product feature that is requested on the deposit account.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.DepositAccountDetails.FeaturesSelected.ExtraFeature</lx:path>
                                      <lx:label>Extra Feature</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Description" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Specific information, value or enumeration related to the specific feature (name).</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.DepositAccountDetails.FeaturesSelected.ExtraFeature.Description</lx:path>
                                          <lx:label>Description</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Name" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The name of the specific feature being requested. Free text field although deposit taking institutions may provide enumerations.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.DepositAccountDetails.FeaturesSelected.ExtraFeature.Name</lx:path>
                                          <lx:label>Name</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Unique Identifier for this Extra Feature Element</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.DepositAccountDetails.FeaturesSelected.ExtraFeature.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="ChequeBook" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that a cheque book be issued for the deposit account.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.FeaturesSelected.ChequeBook</lx:path>
                                    <lx:label>Cheque Book</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CreditCard" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant has requested a credit card.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.FeaturesSelected.CreditCard</lx:path>
                                    <lx:label>Credit Card</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DebitCard" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant has requested a debit card.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.FeaturesSelected.DebitCard</lx:path>
                                    <lx:label>Debit Card</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DepositBook" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant has requested a deposit book</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.FeaturesSelected.DepositBook</lx:path>
                                    <lx:label>Deposit Book</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EFTPOSCard" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant has requested an EFTPOS card.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.FeaturesSelected.EFTPOSCard</lx:path>
                                    <lx:label>EFTPOS Card</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StatementEmailDelivery" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant has requested statements be delivered by email</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.FeaturesSelected.StatementEmailDelivery</lx:path>
                                    <lx:label>Statement Email Delivery</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StatementPaperDelivery" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant has requested to receive paper statements.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.FeaturesSelected.StatementPaperDelivery</lx:path>
                                    <lx:label>Statement Paper Delivery</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Applicant" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant that has selected these features (when this attribute is not populated, the selected features apply to the entire account).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.FeaturesSelected.x_Applicant</lx:path>
                                    <lx:label>Applicant</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Package" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about a package option that may be available for the individual deposit account.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.DepositAccountDetails.Package</lx:path>
                                <lx:label>Package</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Category" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures a deposit taking institution's defined category to which the requested package belongs.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.Package.Category</lx:path>
                                    <lx:label>Category</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Code" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A unique code representing the product package being requested by the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.Package.Code</lx:path>
                                    <lx:label>Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MemberID" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Membership Identifier of the applicant issued by the Organisation the applicant claims to belong to and therefore qualify for the package associated with that Organisation. May be provided when Organisation is specified.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.Package.MemberID</lx:path>
                                    <lx:label>Member ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the name of a package that the applicant is applying for. A package commonly gives the applicants certain benefits, such as an interest rate bonus. There may be rules that determine whether the applicants qualify for the package.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.Package.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OptionCode" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A unique code identifying an option that applies to a package.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.Package.OptionCode</lx:path>
                                    <lx:label>Option Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Organisation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A code or name that identifies the organisation under which the package is offered.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.Package.Organisation</lx:path>
                                    <lx:label>Organisation</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="SourceOfInitialDeposit" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The source of the funds to be used for the initial deposit.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.DepositAccountDetails.SourceOfInitialDeposit</lx:path>
                                <lx:label>Source Of Initial Deposit</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The details of the source of funds if the type is 'Other'.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.SourceOfInitialDeposit.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Percentage" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The percentage of the funds from this source.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.SourceOfInitialDeposit.Percentage</lx:path>
                                    <lx:label>Percentage</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="sourceOfFundsTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A general description of the source of funds for the initial deposit.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.DepositAccountDetails.SourceOfInitialDeposit.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="InitialDepositAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The dollar amount that will be deposited in the initial deposit.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DepositAccountDetails.InitialDepositAmount</lx:path>
                              <lx:label>Initial Deposit Amount</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OriginatorReferenceID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Unique identifier assigned to the deposit account detail by the originator system.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DepositAccountDetails.OriginatorReferenceID</lx:path>
                              <lx:label>Originator Reference ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProductCode" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures an identifying code for the deposit account product applied for. This code is typically issued by the Organisation and enables the Organisation's systems to uniquely identify the product.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DepositAccountDetails.ProductCode</lx:path>
                              <lx:label>Product Code</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProductName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the name of the deposit account product applied for.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DepositAccountDetails.ProductName</lx:path>
                              <lx:label>Product Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProposedAnnualInterestRate" type="percentType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The interest rate that applies to the deposit account. Typically expressed as an annual interest rate.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DepositAccountDetails.ProposedAnnualInterestRate</lx:path>
                              <lx:label>Proposed Annual Interest Rate</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this deposit account details within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DepositAccountDetails.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="StatementCycle" type="statementCycleList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates preferred frequency for statements</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DepositAccountDetails.StatementCycle</lx:path>
                              <lx:label>Statement Cycle</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Type" type="depositAccountTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Specifies the type of the deposit account such as transactional or term.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DepositAccountDetails.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DepositAccountDetails.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="DetailedComment" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Complex element containing comments related to any section of the application. By default, absent any reference to a Context, comments relate to the Application overall.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.DetailedComment</lx:path>
                          <lx:label>Detailed Comment</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Comment" type="stringType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A free text comment</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.DetailedComment.Comment</lx:path>
                                <lx:label>Comment</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="ContextDescription" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Text description of the item the comment refers to</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DetailedComment.ContextDescription</lx:path>
                              <lx:label>Context Description</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CreatedDate" type="dateTimeType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date that the comment was added to the file</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DetailedComment.CreatedDate</lx:path>
                              <lx:label>Created Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this detailed comment within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DetailedComment.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DetailedComment.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Author" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Reference to the author of the comment</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DetailedComment.x_Author</lx:path>
                              <lx:label>Author</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                <lx:target>Package.Content.RealEstateValuation.Instructor</lx:target>
                                <lx:target>Package.Content.RealEstateValuation.PrimaryValuer</lx:target>
                                <lx:target>Package.Content.Application.SalesChannel.LoanWriter</lx:target>
                                <lx:target>Package.Content.Application.BusinessChannel.Contact</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Context" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Pointer to the item that the comment refers to</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.DetailedComment.x_Context</lx:path>
                              <lx:label>Context</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="GeneralSecurityAgreement" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>General security agreement is used in commercial transactions and is a legal right granted by the grantor to a lender over the all existing and future assets of the grantor.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.GeneralSecurityAgreement</lx:path>
                          <lx:label>General Security Agreement</lx:label>
                          <li:transactions>ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Duration" type="durationType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Duration of general security agreement.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.GeneralSecurityAgreement.Duration</lx:path>
                                <lx:label>Duration</lx:label>
                                <li:transactions>ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="Grantor" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The grantor is the party (individual or organisation) who owns or has an interest in the personal property to which the security interest is attached.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.GeneralSecurityAgreement.Grantor</lx:path>
                                <lx:label>Grantor</lx:label>
                                <li:transactions>ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Grantor" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The cross reference to the party (individual or organisation) who owns or has an interest in the personal property to which the security interest is attached.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.GeneralSecurityAgreement.Grantor.x_Grantor</lx:path>
                                    <lx:label>Grantor</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="Type" type="generalSecurityAgreementTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of general security agreement.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.GeneralSecurityAgreement.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier for this general security agreement.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.GeneralSecurityAgreement.UniqueID</lx:path>
                              <lx:label>UniqueID</lx:label>
                              <li:transactions>ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Household" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>A household to which one or more person applicants belong, used in relation to the calculation of household living expenses</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.Household</lx:path>
                          <lx:label>Household</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Dependant" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>A person who depends on the applicant(s) for financial or other support</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Household.Dependant</lx:path>
                                <lx:label>Dependant</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="FinancialProvider" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>A party that provides financial support to the dependant.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Household.Dependant.FinancialProvider</lx:path>
                                      <lx:label>Financial Provider</lx:label>
                                      <li:transactions>CAL,CDA,SVC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="x_Party" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A cross reference to the party that provides financial support to the dependant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.Dependant.FinancialProvider.x_Party</lx:path>
                                          <lx:label>Party</lx:label>
                                          <li:transactions>CAL,CDA,SVC</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Age" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The age of the dependant at the time of submitting the application</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Household.Dependant.Age</lx:path>
                                    <lx:label>Age</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DateOfBirth" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The dependant's date of birth</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Household.Dependant.DateOfBirth</lx:path>
                                    <lx:label>Date Of Birth</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the dependant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Household.Dependant.Name</lx:path>
                                    <lx:label>Dependant Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unique Identifier for this Dependant Element</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Household.Dependant.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="EducationExpenses" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Education expenses contains information about the number and type of students attending school or otherwise participating in study. May be used as an input into the calculation of deemed education expenses in the lender's of living expense allowance.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Household.EducationExpenses</lx:path>
                                <lx:label>Education Expenses</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="NumberOfAdultsInFullTimeStudy" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the number of adults participating in full-time study.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Household.EducationExpenses.NumberOfAdultsInFullTimeStudy</lx:path>
                                    <lx:label>Number Of Adults In Full Time Study</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfAdultsInPartTimeStudy" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the number of adults participating in part-time study.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Household.EducationExpenses.NumberOfAdultsInPartTimeStudy</lx:path>
                                    <lx:label>Number Of Adults In Part Time Study</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfChildrenInPrivateSchool" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the number of dependants (children) attending private school.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Household.EducationExpenses.NumberOfChildrenInPrivateSchool</lx:path>
                                    <lx:label>Number Of Children In Private School</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfChildrenInPublicSchool" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the number of dependants (children) attending public school.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Household.EducationExpenses.NumberOfChildrenInPublicSchool</lx:path>
                                    <lx:label>Number Of Children In Public School</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ExpenseDetails" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The set of expenses declared for the household, including Living Expenses and Other Commitments.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Household.ExpenseDetails</lx:path>
                                <lx:label>Living Expenses</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="LivingExpense" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Contains the details of a single line item that forms part of the household's declared living expenses.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Household.ExpenseDetails.LivingExpense</lx:path>
                                      <lx:label>Living Expense Line Item</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="PercentResponsible" type="percentOwnedType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Where the party responsible for the expense needs to be established at an individual level, this element identifies the party that is responsible for the split and corresponding percentages. Each expense may be 'owned' by one or more parties, typically the applicants, so there may be more than one owner referenced. An expense may also be partially owned by a non-applicant, such as an applicant's spouse, in which case the third party owner may be referenced, or less than 100% ownership of the expense may be assigned.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.Household.ExpenseDetails.LivingExpense.PercentResponsible</lx:path>
                                            <lx:label>Percent Responsible</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Amount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The dollar amount of this particular living expense for the household as declared by the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.LivingExpense.Amount</lx:path>
                                          <lx:label>Amount</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Category" type="livingExpenseCategoryList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The category that this declared living expense falls under.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.LivingExpense.Category</lx:path>
                                          <lx:label>Living Expense Category</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Description" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A plain text description that provides additional detail supporting the Living Expense for the household as declared by the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.LivingExpense.Description</lx:path>
                                          <lx:label>Description</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="EndDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Where this living expense has an end date, this attribute contains that anticipated end date.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.LivingExpense.EndDate</lx:path>
                                          <lx:label>End Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Frequency" type="frequencyFullList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The frequency that corresponds to the declared amount for this particular Living Expense.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.LivingExpense.Frequency</lx:path>
                                          <lx:label>Frequency</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StartDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Where this living expense starts in the future, this attribute contains the anticipated future start date.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.LivingExpense.StartDate</lx:path>
                                          <lx:label>Start Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Unique Identifier for this living expense.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.LivingExpense.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="OtherCommitment" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Contains the details of a single line item that forms part of the household's commitments other than living expenses.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Household.ExpenseDetails.OtherCommitment</lx:path>
                                      <lx:label>Other Commitment Line Item</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="PercentResponsible" type="percentOwnedType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Where the party responsible for the commitment needs to be established at an individual level, this element identifies the party that is responsible for the split and corresponding percentages. Each commitment may be 'owned' by one or more parties, typically the applicants, so there may be more than one owner referenced. A commitment may also be partially owned by a non-applicant, such as an applicant's spouse, in which case the third party owner may be referenced, or less than 100% ownership of the commitment may be assigned.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.Household.ExpenseDetails.OtherCommitment.PercentResponsible</lx:path>
                                            <lx:label>Percent Responsible</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Amount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The dollar amount of this particular commitment for the household as declared by the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.OtherCommitment.Amount</lx:path>
                                          <lx:label>Amount</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Category" type="otherCommitmentCategoryList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The category that this declared other commitment falls under.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.OtherCommitment.Category</lx:path>
                                          <lx:label>Category</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Description" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A plain text description that provides additional detail supporting the commitment for the household as declared by the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.OtherCommitment.Description</lx:path>
                                          <lx:label>Description</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="EndDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Where this commitment has an end date, this attribute contains that anticipated end date.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.OtherCommitment.EndDate</lx:path>
                                          <lx:label>End Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Frequency" type="frequencyFullList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The frequency that corresponds to the declared amount for this particular commitment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.OtherCommitment.Frequency</lx:path>
                                          <lx:label>Frequency</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IsContinuing" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates that this commitment is continuing.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.OtherCommitment.IsContinuing</lx:path>
                                          <lx:label>Is Continuing</lx:label>
                                          <li:transactions>CDA,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StartDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Where this commitment starts in the future, this attribute contains the anticipated future start date.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.OtherCommitment.StartDate</lx:path>
                                          <lx:label>Business Start Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Unique Identifier for this commitment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.OtherCommitment.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="TotalSystemCalculatedLivingExpenses" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The total system calculated living expenses. This may be calculated using the Household Expenditure Measure (HEM) or calculated using a different method.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Household.ExpenseDetails.TotalSystemCalculatedLivingExpenses</lx:path>
                                      <lx:label>Total System Calculated Living Expenses</lx:label>
                                      <li:transactions>CAL,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Amount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The dollar amount of the total system calculated living expenses.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.TotalSystemCalculatedLivingExpenses.Amount</lx:path>
                                          <lx:label>Amount</lx:label>
                                          <li:transactions>CAL,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The frequency that corresponds to the amount for the total system calculated living expenses.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.TotalSystemCalculatedLivingExpenses.Frequency</lx:path>
                                          <lx:label>Frequency</lx:label>
                                          <li:transactions>CAL,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="TotalUserStatedLivingExpenses" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The total living expenses declared by the application parties.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Household.ExpenseDetails.TotalUserStatedLivingExpenses</lx:path>
                                      <lx:label>Total User Stated Living Expenses</lx:label>
                                      <li:transactions>CAL,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Amount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The dollar amount of the total user stated living expenses.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.TotalUserStatedLivingExpenses.Amount</lx:path>
                                          <lx:label>Amount</lx:label>
                                          <li:transactions>CAL,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The frequency that corresponds to the amount for total user stated living expenses.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Household.ExpenseDetails.TotalUserStatedLivingExpenses.Frequency</lx:path>
                                          <lx:label>Frequency</lx:label>
                                          <li:transactions>CAL,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="BrokerVerifiedExpense" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The broker has verified the expense amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Household.ExpenseDetails.BrokerVerifiedExpense</lx:path>
                                    <lx:label>Broker Verified Expense</lx:label>
                                    <li:transactions>CAL,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="BrokerVerifiedExpenseDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Details given of the expenses and sources used to verify the expenses.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Household.ExpenseDetails.BrokerVerifiedExpenseDetails</lx:path>
                                    <lx:label>Broker Verified Expense Details</lx:label>
                                    <li:transactions>CAL,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StatedLessThanCalculatedLivingExpensesDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Details given when stated living expenses are less than calculated living expenses.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Household.ExpenseDetails.StatedLessThanCalculatedLivingExpensesDetails</lx:path>
                                    <lx:label>Stated Less Than Calculated Living Expenses Details</lx:label>
                                    <li:transactions>CAL,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="Name" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A descriptive name for the household</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Household.Name</lx:path>
                              <lx:label>Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NumberOfAdults" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The number of adults in the household</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Household.NumberOfAdults</lx:path>
                              <lx:label>Number Of Adults</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NumberOfDependants" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The number of dependants in this applicant household.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Household.NumberOfDependants</lx:path>
                              <lx:label>Number Of Dependants</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this household within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Household.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Unique identification for the household</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Household.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Insurance" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Central insurance element capturing insurance details.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.Insurance</lx:path>
                          <lx:label>Insurance</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="AssociatedLoanAccount" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for the loan accounts associated with the insurance.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Insurance.AssociatedLoanAccount</lx:path>
                                <lx:label>Associated Loan Account</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_AssociatedLoanAccount" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The loan account that the insurance applies to</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Insurance.AssociatedLoanAccount.x_AssociatedLoanAccount</lx:path>
                                    <lx:label>Associated Loan Account</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="CommissionPayable" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for the commission payable to the party writing the insurance policy.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Insurance.CommissionPayable</lx:path>
                                <lx:label>Commission Payable</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Amount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount of commission to be paid to the party writing the insurance</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Insurance.CommissionPayable.Amount</lx:path>
                                    <lx:label>Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Percentage" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the percentage of commission payable</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Insurance.CommissionPayable.Percentage</lx:path>
                                    <lx:label>Percentage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Payer" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A reference to the Lender who pays the commission</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Insurance.CommissionPayable.x_Payer</lx:path>
                                    <lx:label>Payer</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.SalesChannel.Company</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="InsuredParty" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for the insured party reference</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Insurance.InsuredParty</lx:path>
                                <lx:label>Insured Party</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_InsuredParty" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Contains a reference to an applicant or related party that is the insured party</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Insurance.InsuredParty.x_InsuredParty</lx:path>
                                    <lx:label>Insured Party</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Premium" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element to capture insurance premium</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Insurance.Premium</lx:path>
                                <lx:label>Premium</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Amount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the insurance premium amount</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Insurance.Premium.Amount</lx:path>
                                    <lx:label>Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the insurance premium payment frequency</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Insurance.Premium.Frequency</lx:path>
                                    <lx:label>Frequency</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Term" type="durationType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The term of insurance cover. Applicable for limited term cover.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Insurance.Term</lx:path>
                                <lx:label>Term</lx:label>
                                <li:transactions>LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="ApprovedDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date that the policy was approved.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.ApprovedDate</lx:path>
                              <lx:label>Approved Date</lx:label>
                              <li:transactions>LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CustomerCode" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The customer code for this lenders mortgage insurance policy.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.CustomerCode</lx:path>
                              <lx:label>Customer Code</lx:label>
                              <li:transactions>LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DeclinedByOther" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>This application for lenders mortgage insurance has been declined by another insurer.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.DeclinedByOther</lx:path>
                              <lx:label>Declined By Other</lx:label>
                              <li:transactions>LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Description" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures a plain text description of the insurance when the selected Insurance Type = "Other"</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.Description</lx:path>
                              <lx:label>Description</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="EffectiveDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the date that the policy is effective from</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.EffectiveDate</lx:path>
                              <lx:label>Effective Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="EquityAcceptable" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The conditions regarding the source of equity have been met.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.EquityAcceptable</lx:path>
                              <lx:label>Equity Acceptable</lx:label>
                              <li:transactions>LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ExpiryDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the date that the policy expires</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.ExpiryDate</lx:path>
                              <lx:label>Expiry Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="InsuranceType" type="insuranceTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Defines the type of insurance held</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.InsuranceType</lx:path>
                              <lx:label>Insurance Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="InsuredAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Defines the amount insured</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.InsuredAmount</lx:path>
                              <lx:label>Insured Amount</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Insurer" type="insurerList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the name of insurer</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.Insurer</lx:path>
                              <lx:label>Insurer</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OtherInsurerName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>For an Insurer not listed, the name of the Insurer</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.OtherInsurerName</lx:path>
                              <lx:label>Other Insurer Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PercentCoverRequired" type="percentType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The amount that is covered as a percentage of the total loan amount.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.PercentCoverRequired</lx:path>
                              <lx:label>Percent Cover Required</lx:label>
                              <li:transactions>LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PolicyNumber" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the policy number</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.PolicyNumber</lx:path>
                              <lx:label>Policy Number</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PremiumInsured" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The premium amount will be covered by the insurance.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.PremiumInsured</lx:path>
                              <lx:label>Premium Insured</lx:label>
                              <li:transactions>LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProductCode" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The product code for this lenders mortgage insurance policy.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.ProductCode</lx:path>
                              <lx:label>Product Code</lx:label>
                              <li:transactions>LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProductName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The name of the insurance product being offered by the provider.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.ProductName</lx:path>
                              <lx:label>Product Name</lx:label>
                              <li:transactions>LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this insurance within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UnderwritingType" type="underwritingTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The underwriting type for this lenders mortgage insurance policy.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.UnderwritingType</lx:path>
                              <lx:label>Underwriting Type</lx:label>
                              <li:transactions>LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Insurance.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="LendingGuarantee" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>A loan guarantee offered by one or more guarantors against one or more loan accounts, with or without security.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.LendingGuarantee</lx:path>
                          <lx:label>Lending Guarantee</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="AssociatedLoanAccount" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The loan account(s) being secured with this guarantee</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LendingGuarantee.AssociatedLoanAccount</lx:path>
                                <lx:label>Associated Loan Account</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_AssociatedLoanAccount" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The loan account(s) being secured with this guarantee</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.AssociatedLoanAccount.x_AssociatedLoanAccount</lx:path>
                                    <lx:label>Associated Loan Account</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Guarantor" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for the loan party providing the guarantee</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LendingGuarantee.Guarantor</lx:path>
                                <lx:label>Guarantor</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Interview" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Information about the interview with the guarantor. Used to capture metadata about the interview itself rather than any content.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.Interview</lx:path>
                                      <lx:label>Interview</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Date" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date that the guarantor interview takes place.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.Interview.Date</lx:path>
                                          <lx:label>Interview Date</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IsFaceToFace" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The interview was conducted face to face with the guarantor.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.Interview.IsFaceToFace</lx:path>
                                          <lx:label>Face to Face</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IsOnlyPersonPresent" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The guarantor was the only person present at the interview.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.Interview.IsOnlyPersonPresent</lx:path>
                                          <lx:label>Only Person Present</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="LoanWriterConfirmations" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Confirmation that the loan-writer has completed specific tasks with respect to the guarantor.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.LoanWriterConfirmations</lx:path>
                                      <lx:label>Loan Writer Confirmations</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="AdvisedDocumentsImminentIfSuccessful" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation that the loan-writer has advised the Guarantor that if this assessment is successful, they will receive their Guarantee documents shortly.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.LoanWriterConfirmations.AdvisedDocumentsImminentIfSuccessful</lx:path>
                                          <lx:label>Advised Documents Imminent If Successful</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AdvisedIdentificationProcessRequired" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation that the loan-writer has advised the Guarantor that they must complete the applicable Customer Identification Process.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.LoanWriterConfirmations.AdvisedIdentificationProcessRequired</lx:path>
                                          <lx:label>Advised Identification Process Required</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AdvisedInformationUsedInAssessment" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation that the loan-writer has advised the Guarantor that the information collected during the interview will now be as part of the final assessment of the loan application.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.LoanWriterConfirmations.AdvisedInformationUsedInAssessment</lx:path>
                                          <lx:label>Advised Information Used In Assessment</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AdvisedOfWithdrawalInstructions" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation that the loan-writer has advised the Guarantor that if they wish to withdraw from this transaction before the funds are drawn, then they need to advise the lender as soon as possible.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.LoanWriterConfirmations.AdvisedOfWithdrawalInstructions</lx:path>
                                          <lx:label>Advised Of Withdrawal Instructions</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CompletedStatementOfPosition" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation that the loan-writer has completed a Statement of Financial Position for the Guarantor.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.LoanWriterConfirmations.CompletedStatementOfPosition</lx:path>
                                          <lx:label>Completed Statement Of Position</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ObtainedAgreementForJointStatementOfPosition" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation that the loan-writer has obtained agreement from both the borrower and the guarantor for the completion of a joint Statement of Position and that and each party has consented to the disclosure of his/her financial information to the other party.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.LoanWriterConfirmations.ObtainedAgreementForJointStatementOfPosition</lx:path>
                                          <lx:label>Obtained Agreement For Joint Statement Of Position</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ObtainedElectronicCommunicationConsent" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation that the loan writer has obtained the completed and signed guarantor document electronic communication consent form.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.LoanWriterConfirmations.ObtainedElectronicCommunicationConsent</lx:path>
                                          <lx:label>Obtained Electronic Communication Consent</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ObtainedSignedDeclaration" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation that the loan-writer has Obtained the Guarantor's signed Declaration.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.LoanWriterConfirmations.ObtainedSignedDeclaration</lx:path>
                                          <lx:label>Obtained Signed Declaration</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="VerifiedGuarantorIncome" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation that the loan-writer has verified the Guarantor's income according to the lender's policy.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.LoanWriterConfirmations.VerifiedGuarantorIncome</lx:path>
                                          <lx:label>Verified Guarantor Income</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="SecurityGuarantee" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Information about guarantor that relates specifically to a security guarantee.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.SecurityGuarantee</lx:path>
                                      <lx:label>Security Guarantee</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="ConfirmCouldMeetLoanRepayments" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation from the guarantor that if they were required to repay the guaranteed amount by borrowing funds, that they would be in a position to meet the loan repayments from their income in addition to reasonable living expenses.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.SecurityGuarantee.ConfirmCouldMeetLoanRepayments</lx:path>
                                          <lx:label>Confirmation that Guarantor Could Meet Loan Repayments</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ConfirmWouldSellFamilyHome" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation from the guarantor that if they were required to repay the guaranteed amount, they would need to sell the family home to meet the commitment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.SecurityGuarantee.ConfirmWouldSellFamilyHome</lx:path>
                                          <lx:label>Confirmation that Guarantor Would Need to Sell the Family Home</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UnderstandsSecurityGuarantee" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation from the guarantor that they understand that should the borrower be unable to meet the commitments of the loan or fails to meet those commitments for any reason, that they may be required to pay the guaranteed amount.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.SecurityGuarantee.UnderstandsSecurityGuarantee</lx:path>
                                          <lx:label>Confirmation that Guarantor Understands the Security Guarantee</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="ServicingGuarantee" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Information about guarantor that relates specifically to a servicing guarantee.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.ServicingGuarantee</lx:path>
                                      <lx:label>Servicing Guarantee</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="ConfirmIsBenefitToFamily" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation from the loan writer that (where the guarantor is a parent or step-parent) by making funds available to the borrower to meet loan repayments there is some benefit for the guarantor or the family (for example by enabling the borrower to leave the family home and/or assisting in making the borrower more financially independent).</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.ServicingGuarantee.ConfirmIsBenefitToFamily</lx:path>
                                          <lx:label>Confirm Is Benefit To Family</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ConfirmRelationshipToBorrower" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation from the loan writer that the relationship between the borrower (primary applicant) and the Guarantor captured at the Person level was specified correctly.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.ServicingGuarantee.ConfirmRelationshipToBorrower</lx:path>
                                          <lx:label>Confirm Relationship To Borrower</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ConfirmUnderstandsBorrowerReliance" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation from the guarantor that they understand that the borrower is relying on payments from you to meet repayments for the proposed loan, and that the borrower's other independent income will not be sufficient to service the loan repayments.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.ServicingGuarantee.ConfirmUnderstandsBorrowerReliance</lx:path>
                                          <lx:label>Confirm Understands the Borrowers Reliance</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ConfirmUnderstandsRepaymentsRequired" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Confirmation that the guarantor understands that throughout the term of the proposed loan they are expected to make repayments for the borrower, and that they are willing to do so.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.ServicingGuarantee.ConfirmUnderstandsRepaymentsRequired</lx:path>
                                          <lx:label>Confirm Understands that Repayments are Required</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="AccessType" type="guarantorAccessList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the level of access for each Guarantor.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.AccessType</lx:path>
                                    <lx:label>Access Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AdvisedToSeekIndependentAdvice" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The loan writer confirms whether:the guarantor has been advised to seek independent legal and/or financial advice.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.AdvisedToSeekIndependentAdvice</lx:path>
                                    <lx:label>Advised To Seek Independent Advice</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConfirmNotUnderPressure" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The guarantor confirms that they are not under any pressure to provide this guarantee.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.ConfirmNotUnderPressure</lx:path>
                                    <lx:label>Confirm Not Under Pressure</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DemonstratesReadingEnglish" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The guarantor can demonstrate that they read and speak English.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.DemonstratesReadingEnglish</lx:path>
                                    <lx:label>Understands English</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DocumentationDeliveryMethod" type="documentationDeliveryMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The guarantor's preferred method for the guarantee documents to be issued.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.DocumentationDeliveryMethod</lx:path>
                                    <lx:label>Documentation Delivery Method</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IndependentAdviceTypeRequired" type="independentAdviceTypeRequiredList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The loan writer confirms the nature of the independent advice that the guarantor requires.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.IndependentAdviceTypeRequired</lx:path>
                                    <lx:label>Independent Advice Type Required</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsUnderPowerOfAttorney" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The guarantee is to be executed under power of attorney on behalf of an individual.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.IsUnderPowerOfAttorney</lx:path>
                                    <lx:label>Is Under Power Of Attorney</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PreferredCorrespondence" type="preferredCorrespondenceList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The preferred method for this guarantor to receive correspondence relating to this loan such as welcome letters, loan statements, arrears notifications, changes of interest rate notifications, and other notifications.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.PreferredCorrespondence</lx:path>
                                    <lx:label>Preferred Correspondence</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SuspectDifficultyUnderstandingEnglish" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The loan writer suspects that the guarantor does not understand the effect of the Guarantee because of language or literacy difficulties.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.SuspectDifficultyUnderstandingEnglish</lx:path>
                                    <lx:label>Suspect Difficulty Understanding English</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SuspectDifficultyUnderstandingObligations" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The loan writer suspects that the guarantor is having difficulty understanding their obligations as Guarantor (for example if they are under the influence of alcohol, have impaired mental capacity, or have poor English).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.SuspectDifficultyUnderstandingObligations</lx:path>
                                    <lx:label>Suspect Difficulty Understanding Obligations</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SuspectMisrepresentation" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The loan writer suspects that the guarantor may be subject to misrepresentation as to why the Guarantee is needed.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.SuspectMisrepresentation</lx:path>
                                    <lx:label>Suspect Misrepresentation</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SuspectUnderBorrowerInfluence" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The loan writer suspects that, due to the nature of the relationship between the Guarantor and the Borrower, the guarantor may be under the Borrower's influence to provide the Guarantee.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.SuspectUnderBorrowerInfluence</lx:path>
                                    <lx:label>Suspect Under Borrower Influence</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SuspectUnderPressure" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The loan writer suspects that the guarantor is under pressure to give this Guarantee.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.SuspectUnderPressure</lx:path>
                                    <lx:label>Suspect Under Pressure</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SuspectUnfairConduct" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The loan writer suspects that the guarantor is subject to unreasonable or unfair conduct by someone involved in this transaction.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.SuspectUnfairConduct</lx:path>
                                    <lx:label>Suspect Unfair Conduct</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UnderstandsGuaranteeSigningPeriod" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The guarantor confirms that they understand that under the Banking Code of Practice they have at least three days to read and review the guarantee and other documents that are required to provide before returning the signed guarantee to the lender. If independent legal advice is obtained or another exception in the Banking Code of Practice applies, documents can be returned earlier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.UnderstandsGuaranteeSigningPeriod</lx:path>
                                    <lx:label>Understands Guarantee Signing Period</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UnderstandsGuarantorObligations" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The guarantor confirms that they understand the guarantee and their obligations as guarantor.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.UnderstandsGuarantorObligations</lx:path>
                                    <lx:label>Understands Guarantor Obligations</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Guarantor" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The loan party providing the guarantee</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Guarantor.x_Guarantor</lx:path>
                                    <lx:label>Guarantor</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Security" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for the security offered</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LendingGuarantee.Security</lx:path>
                                <lx:label>Security</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="SecurityAgreement" type="securityAgreementType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Security agreement details for this security.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LendingGuarantee.Security.SecurityAgreement</lx:path>
                                      <lx:label>Security Agreement</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Priority" type="securityPriorityList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The priority of this lender's mortgage on the security property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Security.Priority</lx:path>
                                    <lx:label>Priority</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Security.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Security" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Security being offered, if any</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LendingGuarantee.Security.x_Security</lx:path>
                                    <lx:label>Security</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.GeneralSecurityAgreement</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="CrossGuarantee" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The lending guarantee is a cross guarantee.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LendingGuarantee.CrossGuarantee</lx:path>
                              <lx:label>Cross Guarantee</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IncomeLimitAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A dollar limit on the income amount that can be used as a servicing guarantee a for loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LendingGuarantee.IncomeLimitAmount</lx:path>
                              <lx:label>Income Limit Amount</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IncomeLimitFrequency" type="frequencyShortList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The frequency of the income amount that can be used as a servicing guarantee a for loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LendingGuarantee.IncomeLimitFrequency</lx:path>
                              <lx:label>Income Limit Frequency</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Limit" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The dollar limit on the portion of the guarantor security property that can be used as a lending guarantee for the loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LendingGuarantee.Limit</lx:path>
                              <lx:label>Limit</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Limited" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The lending guarantee is a limited guarantee. The guarantor is not liable for the full amount of the loans covered by the lending guarantee.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LendingGuarantee.Limited</lx:path>
                              <lx:label>Limited</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LimitedToFacilityAmount" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The guarantee is limited to the outstanding facility amount.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LendingGuarantee.LimitedToFacilityAmount</lx:path>
                              <lx:label>Limited To Facility Amount</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this lending guarantee within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LendingGuarantee.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Type" type="lendingGuaranteeTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of guarantee that will be offered by the Guarantors in the application (security, income for serviceability or both).</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LendingGuarantee.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LendingGuarantee.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Liability" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>A liability item is a debt owed to a creditor or lender - such as a loan, line of credit, or tax bill.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.Liability</lx:path>
                          <lx:label>Liability</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="AccountNumber" type="financialAccountType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Account details of the liability</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.AccountNumber</lx:path>
                                <lx:label>Account Number</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="Arrears" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for capturing Arrears detail</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.Arrears</lx:path>
                                <lx:label>Arrears</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Amount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the amount in arrears</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Arrears.Amount</lx:path>
                                    <lx:label>Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfMissedPayments" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the number of missed payments</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Arrears.NumberOfMissedPayments</lx:path>
                                    <lx:label>Number Of Missed Payments</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ContinuingRepayment" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Details of the continuing repayment expense associated with this liability where it differs from the current repayment expense.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.ContinuingRepayment</lx:path>
                                <lx:label>Continuing Repayment</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="LoanPaymentScheduleType" type="loanPaymentScheduleTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of loan payment schedule.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.ContinuingRepayment.LoanPaymentScheduleType</lx:path>
                                    <lx:label>Loan Payment Schedule Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MinimumAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The minimum amount that must be paid.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.ContinuingRepayment.MinimumAmount</lx:path>
                                    <lx:label>Minimum Amount</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentTiming" type="paymentTimingList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the timing of payments with respect to their payment period.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.ContinuingRepayment.PaymentTiming</lx:path>
                                    <lx:label>Payment Timing</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentType" type="paymentTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of repayment.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.ContinuingRepayment.PaymentType</lx:path>
                                    <lx:label>Payment Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RemainingRepayments" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The total number of loan repayments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.ContinuingRepayment.RemainingRepayments</lx:path>
                                    <lx:label>Total Repayments</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RepaymentAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The repayment expense amount associated with repaying this liability.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.ContinuingRepayment.RepaymentAmount</lx:path>
                                    <lx:label>Repayment Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RepaymentFrequency" type="repaymentFrequencyList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The payment frequency of the repayment associated with this liability.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.ContinuingRepayment.RepaymentFrequency</lx:path>
                                    <lx:label>Repayment Frequency</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TaxDeductible" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the expenses associated with this liability are tax deductible.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.ContinuingRepayment.TaxDeductible</lx:path>
                                    <lx:label>Tax Deductible</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.ContinuingRepayment.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="DiscountMargin" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The interest rate discount margin on the liability.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.DiscountMargin</lx:path>
                                <lx:label>Discount Margin</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DiscountedAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The interest rate discount margin as a dollar amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.DiscountMargin.DiscountedAmount</lx:path>
                                    <lx:label>Discounted Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DiscountRate" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The interest rate discount margin as a percentage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.DiscountMargin.DiscountRate</lx:path>
                                    <lx:label>Discount Rate</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DiscountReason" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An explanation of why the applicant qualifies for or is entitled to the discount claimed.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.DiscountMargin.DiscountReason</lx:path>
                                    <lx:label>Discount Reason</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Duration" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The duration over which the discount is to be applied.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.DiscountMargin.Duration</lx:path>
                                    <lx:label>Duration</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DurationUnits" type="durationUnitsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The units of time of the specified duration.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.DiscountMargin.DurationUnits</lx:path>
                                    <lx:label>Duration Units</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The order of this discount margin within the sequence.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.DiscountMargin.SequenceNumber</lx:path>
                                    <lx:label>Sequence Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="DocumentationInstructions" type="documentationInstructionsType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Information specifying where correspondence is sent regarding this liability.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.DocumentationInstructions</lx:path>
                                <lx:label>Documentation Instructions</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="DSH" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Indicates whether this loan is a Defence Service Housing loan under the program between the Commonwealth Government and Westpac Banking Corporation.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.DSH</lx:path>
                                <lx:label>DSH</lx:label>
                                <li:transactions>CAL,CDA,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AssignmentCertificate" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Under the DSH loan program, a certificate of assignment assigns the subsidy that is paid by the Commonwealth to a credit provider if, in accordance with the certificate, the credit provider makes an advance to another person (the assignee). If there is an assignment certificate, the details of the certificate are captured here.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.DSH.AssignmentCertificate</lx:path>
                                    <lx:label>Assignment Certificate</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EntitlementCertificate" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Under the Defence Services Housing loan program, a certificate of entitlement certifies that a subsidy is payable by the Commonwealth to the credit provider if, in accordance with the agreement, the credit provider makes that advance to the applicant or the assignee of the applicant. If there is an entitlement certificate, details of the certificate must be provided.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.DSH.EntitlementCertificate</lx:path>
                                    <lx:label>Entitlement Certificate</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FileNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Under the Defence Services Housing loan program, the Commonwealth's file number for the qualifying applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.DSH.FileNumber</lx:path>
                                    <lx:label>File Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Subsidised" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the DSH loan being applied for is subsidised.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.DSH.Subsidised</lx:path>
                                    <lx:label>Subsidised</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Supplement" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the DSH loan being applied for is a supplement to an existing DSH loan, whereby additional available DSH loan funds are advanced to the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.DSH.Supplement</lx:path>
                                    <lx:label>Supplement</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="FeaturesSelected" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about the product features of the liability.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.FeaturesSelected</lx:path>
                                <lx:label>Features Selected</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="ExtraFeature" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Information about a product feature of the liability.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Liability.FeaturesSelected.ExtraFeature</lx:path>
                                      <lx:label>Extra Feature</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Description" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Specific information, value or enumeration related to the specific feature (name).</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.FeaturesSelected.ExtraFeature.Description</lx:path>
                                          <lx:label>Description</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Name" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The name of the specific feature of the liability. Free text field although lenders may provide enumerations</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.FeaturesSelected.ExtraFeature.Name</lx:path>
                                          <lx:label>Name</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Unique Identifier for this Extra Feature Element</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.FeaturesSelected.ExtraFeature.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="OffsetAccount" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Identifies the BSB, Account Number and Account Name for the offset account.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Liability.FeaturesSelected.OffsetAccount</lx:path>
                                      <lx:label>Offset Account</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="AccountNumber" type="financialAccountType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Account number details such as BSB, Account Number and Account Name for the offset account</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.Liability.FeaturesSelected.OffsetAccount.AccountNumber</lx:path>
                                            <lx:label>Account Number</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="IsExisting" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates whether the Offset Account detailed is an existing Offset Account or not</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.FeaturesSelected.OffsetAccount.IsExisting</lx:path>
                                          <lx:label>Is Existing</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Unique Identifier for this Offset Account Element</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.FeaturesSelected.OffsetAccount.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Account" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A cross reference to an account (a Non Real Estate Asset element) to be used as the offset account.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.FeaturesSelected.OffsetAccount.x_Account</lx:path>
                                          <lx:label>Account</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="ChequeBook" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that a cheque book be issued for the loan account.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.FeaturesSelected.ChequeBook</lx:path>
                                    <lx:label>Cheque Book</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CreditCard" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant has requested a credit card.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.FeaturesSelected.CreditCard</lx:path>
                                    <lx:label>Credit Card</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DebitCard" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant has requested a debit card.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.FeaturesSelected.DebitCard</lx:path>
                                    <lx:label>Debit Card</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DepositBook" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant has requested a deposit book</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.FeaturesSelected.DepositBook</lx:path>
                                    <lx:label>Deposit Book</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="HolidayLeave" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that the loan product allows them to temporarily pause making repayments while they are not earning a regular income or for some other acceptable reason.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.FeaturesSelected.HolidayLeave</lx:path>
                                    <lx:label>Holiday Leave</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Offset" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that a 100% offset account facility be established for the loan account.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.FeaturesSelected.Offset</lx:path>
                                    <lx:label>Offset</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OffsetPercentage" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the percentage of the balance in the offset account that should be used to reduce the outstanding balance on the loan account when calculating interest payable. Typically 100%, but some lenders offer partial offset accounts.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.FeaturesSelected.OffsetPercentage</lx:path>
                                    <lx:label>Offset Percentage</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ParentalLeave" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that the loan product allows them to reduce their repayments while they are not earning a regular income because they are taking maternity/parental leave.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.FeaturesSelected.ParentalLeave</lx:path>
                                    <lx:label>Parental Leave</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PartialOffset" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that an offset account be established for the loan account with less than 100% offset of the balance in the offset account against the outstanding loan balance.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.FeaturesSelected.PartialOffset</lx:path>
                                    <lx:label>Partial Offset</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Portability" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that the loan facility be portable between security properties.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.FeaturesSelected.Portability</lx:path>
                                    <lx:label>Portability</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProgressiveDraw" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that the loan funds be drawn down progressively, typically against progress of stages under a construction contract.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.FeaturesSelected.ProgressiveDraw</lx:path>
                                    <lx:label>Progressive Draw</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RateLock" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that the interest rate for a fixed rate product be 'locked' until settlement</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.FeaturesSelected.RateLock</lx:path>
                                    <lx:label>Rate Lock</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Redraw" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that the loan product permit the redraw of loan principal that has been repaid in advance of the amortisation schedule for the loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.FeaturesSelected.Redraw</lx:path>
                                    <lx:label>Redraw</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SplitLoan" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the applicant requests that the loan amount be split between two or more separate accounts or split facilities.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.FeaturesSelected.SplitLoan</lx:path>
                                    <lx:label>Split Loan</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="LendingPurpose" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The purpose for which the applicant intends to use the loan funds.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.LendingPurpose</lx:path>
                                <lx:label>Lending Purpose</lx:label>
                                <li:transactions>CAL,CDA,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ABSLendingPurpose" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The free text description of the ABS Lending Purpose Code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.LendingPurpose.ABSLendingPurpose</lx:path>
                                    <lx:label>ABS Lending Purpose</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ABSLendingPurposeCode" type="absLendingPurposeCodeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The ABS code for this purpose for which the loan funds will be used by the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.LendingPurpose.ABSLendingPurposeCode</lx:path>
                                    <lx:label>ABS Lending Purpose Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the purpose if none of the purposes specified in the lender's enumerated list of purposes adequately described the true purpose.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.LendingPurpose.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LenderCode" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The lender-specified code for this purpose for which the loan funds will be used by the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.LendingPurpose.LenderCode</lx:path>
                                    <lx:label>Lender Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="OriginalAmountRequestedInForeignCurrency" type="amountInForeignCurrencyType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The original loan amount as requested in a foreign currency, being the credit limit for a line of credit, or the fully drawn principal amount for an amortising loan.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.OriginalAmountRequestedInForeignCurrency</lx:path>
                                <lx:label>Original Amount Requested In Foreign Currency</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="OriginalTerm" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the original term of the credit product and any introductory terms, such as an interest only term or a fixed rate term that applies for an initial period at the start of the product.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.OriginalTerm</lx:path>
                                <lx:label>OriginalTerm</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="DistinctLoanPeriod" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>A distinct loan period, where each period may have a different payment type, interest type, rate composition.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Liability.OriginalTerm.DistinctLoanPeriod</lx:path>
                                      <lx:label>Distinct Loan Period</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Duration" type="durationType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The duration of the distinct loan period.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.Liability.OriginalTerm.DistinctLoanPeriod.Duration</lx:path>
                                            <lx:label>Duration</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="Repayment" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>A repayment that is made during this period.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.Liability.OriginalTerm.DistinctLoanPeriod.Repayment</lx:path>
                                            <lx:label>Repayment</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="x_ContinuingRepayment" type="referenceType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>A cross reference to the continuing repayment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.Liability.OriginalTerm.DistinctLoanPeriod.Repayment.x_ContinuingRepayment</lx:path>
                                                <lx:label>Repayment</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                <lx:references>
                                                  <lx:target>Package.Content.Application.Liability.ContinuingRepayment</lx:target>
                                                </lx:references>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="x_Repayment" type="referenceType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>A cross reference to the repayment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.Liability.OriginalTerm.DistinctLoanPeriod.Repayment.x_Repayment</lx:path>
                                                <lx:label>Repayment</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                <lx:references>
                                                  <lx:target>Package.Content.Application.Liability.Repayment</lx:target>
                                                </lx:references>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="EndDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The end date of the distinct loan period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.OriginalTerm.DistinctLoanPeriod.EndDate</lx:path>
                                          <lx:label>End Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="InterestType" type="interestTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The type of interest that is to be paid during the distinct loan period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.OriginalTerm.DistinctLoanPeriod.InterestType</lx:path>
                                          <lx:label>Interest Type</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PaymentType" type="paymentTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The type of repayment that is to be made during the distinct loan period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.OriginalTerm.DistinctLoanPeriod.PaymentType</lx:path>
                                          <lx:label>Payment Type</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The order of this distinct loan period within the sequence.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.OriginalTerm.DistinctLoanPeriod.SequenceNumber</lx:path>
                                          <lx:label>Sequence Number</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StartDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The start date of the distinct loan period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.OriginalTerm.DistinctLoanPeriod.StartDate</lx:path>
                                          <lx:label>Start Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.OriginalTerm.DistinctLoanPeriod.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_RateComposition" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A cross reference to the rate composition for this distinct loan period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.OriginalTerm.DistinctLoanPeriod.x_RateComposition</lx:path>
                                          <lx:label>Rate Composition</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.Liability.RateComposition</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="InterestOnlyEndDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date when the applicant will stop/stopped making Interest Only repayments on this liability.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.OriginalTerm.InterestOnlyEndDate</lx:path>
                                    <lx:label>Interest Only End Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InterestType" type="interestTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the original type of interest of the liability.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.OriginalTerm.InterestType</lx:path>
                                    <lx:label>Interest Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InterestTypeDuration" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The original duration of the fixed interest rate period, as a whole number multiple of the Units.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.OriginalTerm.InterestTypeDuration</lx:path>
                                    <lx:label>Interest Type Duration</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InterestTypeUnits" type="loanTermUnitsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The units of time that are used to define the duration of the fixed interest rate period.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.OriginalTerm.InterestTypeUnits</lx:path>
                                    <lx:label>Interest Type Units</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentType" type="paymentTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the original type of repayment of the liability.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.OriginalTerm.PaymentType</lx:path>
                                    <lx:label>Payment Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentTypeDuration" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The original duration of the interest only period, as a whole number multiple of the Units.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.OriginalTerm.PaymentTypeDuration</lx:path>
                                    <lx:label>Payment Type Duration</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentTypeUnits" type="loanTermUnitsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The units of time that are used to define the duration of the interest only period.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.OriginalTerm.PaymentTypeUnits</lx:path>
                                    <lx:label>Payment Type Units</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RolloverPeriodDuration" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The duration of the rollover period for a commercial bill.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.OriginalTerm.RolloverPeriodDuration</lx:path>
                                    <lx:label>RolloverPeriodDuration</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RolloverPeriodUnits" type="loanTermUnitsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The units of time that are used to define the rollover period for a commercial bill.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.OriginalTerm.RolloverPeriodUnits</lx:path>
                                    <lx:label>RolloverPeriodUnits</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalTermDuration" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the original duration of the total term, as a multiple of the units.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.OriginalTerm.TotalTermDuration</lx:path>
                                    <lx:label>Total Term Duration</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalTermUnits" type="loanTermUnitsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The units of time that are used to define the duration of the total term.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.OriginalTerm.TotalTermUnits</lx:path>
                                    <lx:label>Total Term Units</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Package" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about a package option that previously was part of the lender's offer.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.Package</lx:path>
                                <lx:label>Package</lx:label>
                                <li:transactions>CAL,CDA,CNZ,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Category" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Category in which the package was offered by the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Package.Category</lx:path>
                                    <lx:label>Category</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Code" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A unique code representing the product package that was offered by the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Package.Code</lx:path>
                                    <lx:label>Code</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MemberID" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Membership identifier of the applicant issued by the Organisation the applicant claims to belong to.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Package.MemberID</lx:path>
                                    <lx:label>Member ID</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Name of a package that the applicant previously applied for.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Package.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OptionCode" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A unique code identifying an option that previously applied to a package.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Package.OptionCode</lx:path>
                                    <lx:label>Option Code</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Organisation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A code or name that identifies the organisation under which the package was offered.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Package.Organisation</lx:path>
                                    <lx:label>Organisation</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PercentOwned" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Identifies the party/s that are responsible for the liability and the amount they are responsible for, either as a borrower or guarantor.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.PercentOwned</lx:path>
                                <lx:label>Percent Owned</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Owner" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>The details of the ownership of this liability, either as borrower or guarantor.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Liability.PercentOwned.Owner</lx:path>
                                      <lx:label>Owner</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="OwnerType" type="applicantTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Describes whether the linked party is a borrower or guarantor on this liability.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.PercentOwned.Owner.OwnerType</lx:path>
                                          <lx:label>Owner Type</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Percent" type="percentType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the percentage of the liability that the linked party is responsible for.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.PercentOwned.Owner.Percent</lx:path>
                                          <lx:label>Percent</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Party" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A reference to a party that is an responsible for the liability.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.PercentOwned.Owner.x_Party</lx:path>
                                          <lx:label>Party</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Proportions" type="proportionsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether the ownership proportions are equal across all owners, specified individually for each one, or not specified.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.PercentOwned.Proportions</lx:path>
                                    <lx:label>Proportions</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="RateComposition" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The structure of the interest rate to the borrower, based on the delivery rate with applied margins</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.RateComposition</lx:path>
                                <lx:label>Rate Composition</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="BaseRate" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The base rate on which other rates are calculated.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Liability.RateComposition.BaseRate</lx:path>
                                      <lx:label>Base Rate</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Code" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The base rate code.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.RateComposition.BaseRate.Code</lx:path>
                                          <lx:label>Code</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Name" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The base rate name.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.RateComposition.BaseRate.Name</lx:path>
                                          <lx:label>Name</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Rate" type="percentType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The base rate as a percentage.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.RateComposition.BaseRate.Rate</lx:path>
                                          <lx:label>Rate</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="CustomerRiskMargin" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The customer risk margin added (as a percentage) based on the risk the customer presents to the bank.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.RateComposition.CustomerRiskMargin</lx:path>
                                    <lx:label>Customer Risk Margin</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DisclosedToCustomer" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The details of the rate are disclosed to the customer.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.RateComposition.DisclosedToCustomer</lx:path>
                                    <lx:label>Disclosed To Customer</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IntroductoryMargin" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The introductory margin added (as a percentage) during the introductory period or honeymoon period.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.RateComposition.IntroductoryMargin</lx:path>
                                    <lx:label>Introductory Margin</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NetCustomerRiskMargin" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The customer risk margin minus the discount margin (as a percentage).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.RateComposition.NetCustomerRiskMargin</lx:path>
                                    <lx:label>Net Customer Risk Margin</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentTypeMargin" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The payment type margin added (as a percentage) based on the payment type.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.RateComposition.PaymentTypeMargin</lx:path>
                                    <lx:label>Payment Type Margin</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PricingConstruct" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The pricing construct used to calculate the Total Interest Rate. For example, Base Rate + Term Premium + Product Margin + Customer Risk Margin.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.RateComposition.PricingConstruct</lx:path>
                                    <lx:label>Pricing Construct</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProductMargin" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The product margin added (as a percentage) based on the specific product type.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.RateComposition.ProductMargin</lx:path>
                                    <lx:label>Product Margin</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TermPremium" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The term premium added (as a percentage) based on the loan term.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.RateComposition.TermPremium</lx:path>
                                    <lx:label>Term Premium</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalInterestRate" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The total interest rate for this rate composition scheme.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.RateComposition.TotalInterestRate</lx:path>
                                    <lx:label>Total Interest Rate</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.RateComposition.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="RemainingTerm" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for capturing the remaining term of liabilities not being refinanced.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.RemainingTerm</lx:path>
                                <lx:label>Remaining Term</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Duration" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The duration of the remaining term.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.RemainingTerm.Duration</lx:path>
                                    <lx:label>Duration</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Units" type="loanTermUnitsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The units used to define the duration of the remaining term.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.RemainingTerm.Units</lx:path>
                                    <lx:label>Units of Remaining Term Duration</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Repayment" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Details of the repayment expense associated with this loan</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.Repayment</lx:path>
                                <lx:label>Repayment</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="InterestPayment" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This payment covers the interest amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Repayment.InterestPayment</lx:path>
                                    <lx:label>Interest Payment</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LoanPaymentScheduleType" type="loanPaymentScheduleTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of loan payment schedule.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Repayment.LoanPaymentScheduleType</lx:path>
                                    <lx:label>Loan Payment Schedule Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MinimumAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The minimum amount that must be paid.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Repayment.MinimumAmount</lx:path>
                                    <lx:label>Minimum Amount</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentTiming" type="paymentTimingList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the timing of payments with respect to their payment period.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Repayment.PaymentTiming</lx:path>
                                    <lx:label>Payment Timing</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentType" type="paymentTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the type of repayment that is to be made during the specified term</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Repayment.PaymentType</lx:path>
                                    <lx:label>Payment Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PrincipalPayment" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This payment reduces the principal amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Repayment.PrincipalPayment</lx:path>
                                    <lx:label>Principal Payment</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Regular" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether payments are regular (same amount on regular intervals) or structured (irregular intervals and/or amounts).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Repayment.Regular</lx:path>
                                    <lx:label>Regular</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RemainingRepayments" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The total number of loan repayments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Repayment.RemainingRepayments</lx:path>
                                    <lx:label>Total Repayments</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RepaymentAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The repayment expense amount associated with repaying this loan</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Repayment.RepaymentAmount</lx:path>
                                    <lx:label>Repayment Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RepaymentFrequency" type="repaymentFrequencyList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The payment frequency of the repayment associated with this loan</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Repayment.RepaymentFrequency</lx:path>
                                    <lx:label>Repayment Frequency</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TaxDeductible" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the expenses associated with a liability are tax deductible (yes) or not tax deductible (no)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Repayment.TaxDeductible</lx:path>
                                    <lx:label>Tax Deductible</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Repayment.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Security" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Details to identify the security property for the loan</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.Security</lx:path>
                                <lx:label>Security</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="MortgagorDetails" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Details of the mortgagor(s) for the security for the liability</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Liability.Security.MortgagorDetails</lx:path>
                                      <lx:label>Mortgagor Details</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="ChangeNameFormRequired" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Flag to indicate whether a change of name form is required for the mortgage document</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.Security.MortgagorDetails.ChangeNameFormRequired</lx:path>
                                          <lx:label>Change Name Form Required</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Name" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The name of the mortgagor matching the existing name on the record</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.Security.MortgagorDetails.Name</lx:path>
                                          <lx:label>Name</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ReasonForDifferentName" type="reasonForDifferentNameList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The reason the name of the mortgagor does not match the name on the certificate of title. Only applicable in Victoria.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.Security.MortgagorDetails.ReasonForDifferentName</lx:path>
                                          <lx:label>Reason For Different Name</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Mortgagor" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Reference link to a mortgagor for the security. If a change of name is required, the new name will be constructed from the granular name fields of the linked Party</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Liability.Security.MortgagorDetails.x_Mortgagor</lx:path>
                                          <lx:label>Mortgagor</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                            <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="SecurityAgreement" type="securityAgreementType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Security agreement details for this security.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Liability.Security.SecurityAgreement</lx:path>
                                      <lx:label>Security Agreement</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="MortgageNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The identifying number of the mortgage that links the security to the loan</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Security.MortgageNumber</lx:path>
                                    <lx:label>Mortgage Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Priority" type="securityPriorityList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The priority of this lender's mortgage on the security property</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Security.Priority</lx:path>
                                    <lx:label>Priority</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PriorityAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The maximum dollar amount that the mortgagee is entitled to recover under the mortgage, to repay loaned funds including principal, interest and fees.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Security.PriorityAmount</lx:path>
                                    <lx:label>Priority Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RequiresOriginalMortgageDocument" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the original mortgage document will be required for the mortgage to be discharged.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Security.RequiresOriginalMortgageDocument</lx:path>
                                    <lx:label>Requires Original Mortgage Document</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SignedDocumentsRequired" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Signed documents will be required for this security.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Security.SignedDocumentsRequired</lx:path>
                                    <lx:label>Signed Documents Required</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Security.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Security" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Contains a reference link to the security (real estate or non real estate asset)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Security.x_Security</lx:path>
                                    <lx:label>Security</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.GeneralSecurityAgreement</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Software" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The software application used by the liability originator.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.Software</lx:path>
                                <lx:label>Software</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the software provider, for example the company name.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Software.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the software application used by the liability originator.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.Software.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="TermsAndConditions" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The terms and conditions of a loan contract or loan application.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Liability.TermsAndConditions</lx:path>
                                <lx:label>Terms And Conditions</lx:label>
                                <li:transactions>CAL,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="TermsCode" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The code used to identify the terms and conditions in the lender's system.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.TermsAndConditions.TermsCode</lx:path>
                                    <lx:label>TermsCode</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TermsName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name used by the lender for this set of terms and conditions.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Liability.TermsAndConditions.TermsName</lx:path>
                                    <lx:label>TermsName</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="AccelerationPercentage" type="percentType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The amount that is provided to the borrower as a percentage of the total amount that is provided from the invoices for an invoice financing loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.AccelerationPercentage</lx:path>
                              <lx:label>Acceleration Percentage</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="AnnualInterestRate" type="percentType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the annual interest rate that applies to the liability</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.AnnualInterestRate</lx:path>
                              <lx:label>Annual Interest Rate</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="AvailableForRedrawAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Advance or additional repayments that reduce the balance on a loan that are available to be redrawn at a later date.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.AvailableForRedrawAmount</lx:path>
                              <lx:label>Available For Redraw Amount</lx:label>
                              <li:transactions>CAL,CDA,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BalloonRepaymentAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The nominated balloon repayment amount.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.BalloonRepaymentAmount</lx:path>
                              <lx:label>Balloon Repayment Amount</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BalloonRepaymentDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The expected date of the balloon repayment.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.BalloonRepaymentDate</lx:path>
                              <lx:label>Balloon Repayment Date</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ClearingFromOtherSource" type="clearingFromOtherSourceList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the existing liability will be cleared, in whole or in part, using funds from a third party institution.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.ClearingFromOtherSource</lx:path>
                              <lx:label>Clearing From Other Source</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ClearingFromOtherSourceAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the exact amount that will be refinanced for this liability using funds from a source that is not the new loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.ClearingFromOtherSourceAmount</lx:path>
                              <lx:label>Clearing from Other Source Amount</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ClearingFromThisLoan" type="clearingFromThisLoanList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the existing loan will be cleared, in whole or in part, using funds from the loan being applied for.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.ClearingFromThisLoan</lx:path>
                              <lx:label>Clearing From This Loan</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ClearingFromThisLoanAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the exact amount that will be refinanced for this liability using funds sourced from a new loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.ClearingFromThisLoanAmount</lx:path>
                              <lx:label>Clearing from This Loan Amount</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ClosingOnSettlement" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>This revolving liability will be closed on settlement.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.ClosingOnSettlement</lx:path>
                              <lx:label>Closing On Settlement</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CreditCardType" type="creditCardTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates the type of credit card</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.CreditCardType</lx:path>
                              <lx:label>Credit Card Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CreditLimit" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the limit for a credit card or a line of credit, or the scheduled balance for an amortising loan. This is typically the balance that the lender uses to assess the liability as the borrower is likely able to draw funds up to the limit, or to redraw principal repaid ahead of the scheduled balance.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.CreditLimit</lx:path>
                              <lx:label>Credit Limit</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CreditRiskGrade" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A credit risk grade can be used to specify how likely the liability is to experience events such as payment defaults based on its history. A higher risk grade generally indicates higher risk.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.CreditRiskGrade</lx:path>
                              <lx:label>Credit Risk Grade</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Description" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures a free text description of the liability when the selected Liability Type = "Other".</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.Description</lx:path>
                              <lx:label>Description</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="HasArrears" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The applicant or financial system has declared a current arrears credit issue on this liability.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.HasArrears</lx:path>
                              <lx:label>Has Arrears</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="HasPreviousArrears" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Used to indicate if the applicant or financial system has declared a previous arrears Credit Issue against this liability.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.HasPreviousArrears</lx:path>
                              <lx:label>Has Previous Arrears</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="HasUndrawnFunds" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the liability has any undrawn funds.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.HasUndrawnFunds</lx:path>
                              <lx:label>Has Undrawn Funds</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="InAdvanceAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Advance or additional repayments that reduce the balance on a loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.InAdvanceAmount</lx:path>
                              <lx:label>In Advance Amount</lx:label>
                              <li:transactions>CAL,CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="InterestCalculationFrequency" type="interestChargeFrequencyList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The frequency at which interest is calculated on the loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.InterestCalculationFrequency</lx:path>
                              <lx:label>Interest Calculation Frequency</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="InterestChargeFrequency" type="interestChargeFrequencyList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The frequency at which interest is charged on the loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.InterestChargeFrequency</lx:path>
                              <lx:label>Interest Charge Frequency</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsOriginalAmountRequestedInForeignCurrency" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>A flag to indicate that the original amount was requested in a foreign currency.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.IsOriginalAmountRequestedInForeignCurrency</lx:path>
                              <lx:label>Is New Limit Requested In Foreign Currency</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LenderAssessmentReason" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A free text description of the reason the liability record requires assessment by the lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.LenderAssessmentReason</lx:path>
                              <lx:label>Lender Assessment Reason</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LenderAssessmentRequired" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>A flag to indicate the liability record requires assessment by the lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.LenderAssessmentRequired</lx:path>
                              <lx:label>Lender Assessment Required</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LimitExceededCurrently" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the liability's limit is currently exceeded.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.LimitExceededCurrently</lx:path>
                              <lx:label>Limit Exceeded Currently</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LimitExceededPreviously" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the liability's limit previously been exceeded.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.LimitExceededPreviously</lx:path>
                              <lx:label>Limit Exceeded Previously</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LVR" type="percentType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Liability's loan to value ratio.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.LVR</lx:path>
                              <lx:label>LVR</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="MaturityDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The maturity date of the loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.MaturityDate</lx:path>
                              <lx:label>Maturity Date</lx:label>
                              <li:transactions>CAL,CDA,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="MinimumRepaymentRate" type="percentType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The percentage rate used to calculate the minimum repayment for the liability.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.MinimumRepaymentRate</lx:path>
                              <lx:label>Minimum Repayment Rate</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NCCPStatus" type="nccpStatusList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The status of the liability as regulated or unregulated under the NCCP.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.NCCPStatus</lx:path>
                              <lx:label>NCCP Status</lx:label>
                              <li:transactions>CAL,CDA,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NegativelyGeared" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures if the Liability is used to a negatively geared tax liability.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.NegativelyGeared</lx:path>
                              <lx:label>Negatively Geared</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NegativelyGearedPercentage" type="percentType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the percentage of the liability used for negative gearing.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.NegativelyGearedPercentage</lx:path>
                              <lx:label>Negatively Geared Percentage</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NewLimit" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the new limit of the liability when the applicant proposes that they will reduce the limit prior to or at settlement of the loan application.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.NewLimit</lx:path>
                              <lx:label>New Limit</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NonCapitalisedInterest" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The dollar amount of the interest that has been accrued for the loan but has not yet been charged.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.NonCapitalisedInterest</lx:path>
                              <lx:label>Non Capitalised Interest</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OriginalAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the original loan amount, being the credit limit for a line of credit, or the fully drawn principal amount for an amortising loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.OriginalAmount</lx:path>
                              <lx:label>Original Amount</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OriginalLoanPurpose" type="primaryPurposeLoanPurposeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the original loan purpose</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.OriginalLoanPurpose</lx:path>
                              <lx:label>Original Loan Purpose</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OriginationDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The origination date of the liability.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.OriginationDate</lx:path>
                              <lx:label>Origination Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OriginatorReferenceID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Unique identifier assigned to the liability by the originator system.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.OriginatorReferenceID</lx:path>
                              <lx:label>Originator Reference ID</lx:label>
                              <li:transactions>CAL,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OutstandingBalance" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the outstanding, current or unpaid balance of the liability.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.OutstandingBalance</lx:path>
                              <lx:label>Outstanding Balance</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Overdrawn" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the liability is currently overdrawn.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.Overdrawn</lx:path>
                              <lx:label>Overdrawn</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProductCode" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures an identifying code for the existing liability. This code is typically issued by the lender and enables the lender's systems to uniquely identify the product.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.ProductCode</lx:path>
                              <lx:label>Product Code</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProductName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The product name of the existing liability.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.ProductName</lx:path>
                              <lx:label>Product Name</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RefinanceAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the exact amount that will be refinanced for existing liabilities.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.RefinanceAmount</lx:path>
                              <lx:label>Refinance Amount</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RefinanceCosts" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Borrowers must state the estimated cost associated with closing a credit facility. For each liability being refinanced the applicants will provide the Estimated Exit and Break Costs payable to the outgoing finance institution.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.RefinanceCosts</lx:path>
                              <lx:label>Refinance Costs</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RepaidDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>For a loan that has recently been repaid, the date that it was paid out</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.RepaidDate</lx:path>
                              <lx:label>Repaid Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RevertInterestRate" type="percentType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Estimate of the interest rate at the end of the fixed rate period or the end of the interest only period (whichever is longer). The revert interest rate will be an estimate based on the current variable rate.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.RevertInterestRate</lx:path>
                              <lx:label>Revert Interest Rate</lx:label>
                              <li:transactions>CAL,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Secured" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The liability is secured.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.Secured</lx:path>
                              <lx:label>Secured</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this liability within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SMSFLoan" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to note whether the liability belongs to a self-managed super fund.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.SMSFLoan</lx:path>
                              <lx:label>SMSF Loan</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Suspended" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that interest accrual and payments on the liability are currently suspended.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.Suspended</lx:path>
                              <lx:label>Suspended</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Type" type="liabilityTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Specifies the type of liability that is being described. Types include credit card limits, HECS debt, existing loans, etc.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UndrawnAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The maximum amount available to be drawn under a Credit agreement. Captures difference between the future credit limit less the Current Outstanding Balance.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.UndrawnAmount</lx:path>
                              <lx:label>Undrawn Amount</lx:label>
                              <li:transactions>CAL,CDA,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Verified" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>A flag to indicate that this liability data has been verified.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.Verified</lx:path>
                              <lx:label>Verified</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="WrittenOff" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>This liability has been written off.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.WrittenOff</lx:path>
                              <lx:label>Written Off</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_CustomerTransactionAnalysis" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A cross reference to the customer transaction analysis element from which the liability data was verified.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Liability.x_CustomerTransactionAnalysis</lx:path>
                              <lx:label>Customer Transaction Analysis</lx:label>
                              <li:transactions>CAL</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.CustomerTransactionAnalysis</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="LoanDetails" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Information about an individual credit facility that is being applied for including but not limited to: term loans; revolving credit facilities; overdrafts; credit cards; equipment finance; commercial bills; margin loans; personal loans; and reverse mortgages.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.LoanDetails</lx:path>
                          <lx:label>Loan Details</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="AccountNumber" type="financialAccountType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Account details of this loan account.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.AccountNumber</lx:path>
                                <lx:label>Account Number</lx:label>
                                <li:transactions>CDA</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="AmountRequestedInForeignCurrency" type="amountInForeignCurrencyType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The loan amount as requested in a foreign currency by the applicant for this loan account, exclusive of to-be-capitalised fees and/or LMI.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.AmountRequestedInForeignCurrency</lx:path>
                                <lx:label>Amount Requested In Foreign Currency</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="Borrowers" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Identifies the parties that are prospective borrowers on this loan account.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.Borrowers</lx:path>
                                <lx:label>Borrowers</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Owner" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Container that associates a reference to an owner or other responsible party with the percentage of the item that they own or are responsible for.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.Borrowers.Owner</lx:path>
                                      <lx:label>Owner</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="AccessType" type="accessTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Specifies the level of access for this borrower.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Borrowers.Owner.AccessType</lx:path>
                                          <lx:label>Access Type</lx:label>
                                          <li:transactions>ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Percent" type="percentType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the percentage of the item that referenced party owns or is responsible for.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Borrowers.Owner.Percent</lx:path>
                                          <lx:label>Percent</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PreferredCorrespondence" type="preferredCorrespondenceList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The preferred method for this borrower to receive correspondence relating to this loan such as welcome letters, loan statements, arrears notifications, changes of interest rate notifications, and other notifications.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Borrowers.Owner.PreferredCorrespondence</lx:path>
                                          <lx:label>Preferred Correspondence</lx:label>
                                          <li:transactions>ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PrimaryBorrower" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The lender should consider this borrower to be the primary borrower when processing the loan. The significance of the primary borrower to the loan varies from lender to lender.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Borrowers.Owner.PrimaryBorrower</lx:path>
                                          <lx:label>Primary Borrower</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Party" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Contains a reference to a party that is an owner of the item, or has a similar relationship such as a borrower or beneficiary.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Borrowers.Owner.x_Party</lx:path>
                                          <lx:label>Party</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Proportions" type="proportionsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether the ownership proportions are equal across all owners, specified individually for each one, or not specified.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Borrowers.Proportions</lx:path>
                                    <lx:label>Proportions</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="BulkReduction" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A bulk reduction is a scheduled future lump sum principal repayment. For example, the applicant may be yet to sell their home so take out bridging finance in order to fund the purchase of their new home pending the sale of their existing home. When the existing home is sold, they will make a lump sum payment in reduction of the loan. This payment is the bulk reduction. Another example would be where the applicant is expecting a superannuation payout.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.BulkReduction</lx:path>
                                <lx:label>Bulk Reduction</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Amount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount of the future bulk reduction repayment.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.BulkReduction.Amount</lx:path>
                                    <lx:label>Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EstimatedDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date on which it is anticipated that the bulk reduction repayment will be made.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.BulkReduction.EstimatedDate</lx:path>
                                    <lx:label>Estimated Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Commission" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information relating to the commission to be paid to the business partner that has sourced the loan application for the lender. This partner is typically a third party such as a mortgage aggregator or brokerage.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.Commission</lx:path>
                                <lx:label>Commission</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="CommissionAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount of upfront commission to be paid to the sales channel in the event that loan application settles.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Commission.CommissionAmount</lx:path>
                                    <lx:label>Commission Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CommissionPaid" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether a commission is payable to the sales channel in respect of the loan application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Commission.CommissionPaid</lx:path>
                                    <lx:label>Commission Paid</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CommissionStructure" type="commissionStructureList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the commission plan involves payment of an upfront commission only, an upfront commission and a trail commission, a trail commission only, or another type of structure.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Commission.CommissionStructure</lx:path>
                                    <lx:label>Commission Structure</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GSTAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The portion of the commission amount that is paid in goods and services tax (GST).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Commission.GSTAmount</lx:path>
                                    <lx:label>GST Amount</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ITCAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The portion of the commission amount that is claimed as input tax credit (ITC).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Commission.ITCAmount</lx:path>
                                    <lx:label>ITC Amount</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherCommissionStructureDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the commission structure if an accurate description is not specified in the enumerated values. Must be provided when Commission Structure = "Other"</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Commission.OtherCommissionStructureDescription</lx:path>
                                    <lx:label>Other Commission Structure Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PromotionCode" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A code that specifies a particular promotion or special commission offer that applies to the loan application, specified by the submitting sales channel in order to tag the deal with the promotion code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Commission.PromotionCode</lx:path>
                                    <lx:label>Promotion Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ThirdPartyReferee" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the sales channel is a third party referrer of business for the purpose of commission disclosure requirements.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Commission.ThirdPartyReferee</lx:path>
                                    <lx:label>Third Party Referee</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Trail" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the trail commission percentage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Commission.Trail</lx:path>
                                    <lx:label>Trail</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UpfrontPayment" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the upfront commission percentage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Commission.UpfrontPayment</lx:path>
                                    <lx:label>Upfront Payment</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="DiscountMargin" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains information about a discount in relation to the loan facility that the applicant claims they are entitled to. The discount claimed may be in the form of either a rate reduction margin or a dollar amount. The discount claimed is different from, and therefore may be in addition to, a general special offer that applies to all applicants, or discount that is available under a product package.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.DiscountMargin</lx:path>
                                <lx:label>Discount Margin</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DiscountedAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the dollar amount of the discount the applicant claims they qualify for or are entitled to. The dollar amount is to be deducted from the calculated periodic repayment.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.DiscountMargin.DiscountedAmount</lx:path>
                                    <lx:label>Discounted Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DiscountRate" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the percentage rate discount the applicant claims they qualify for or are entitled to. If the discount is for 0.6% off the variable rate which is 6.5%, the value to be entered in this field should be '0.6'. If the borrower gets 0.5 in addition to 0.2, then the net Discount Rate will be 0.7</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.DiscountMargin.DiscountRate</lx:path>
                                    <lx:label>Discount Rate</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DiscountReason" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures an explanation of why the applicant qualifies for or is entitled to the discount claimed.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.DiscountMargin.DiscountReason</lx:path>
                                    <lx:label>Discount Reason</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Duration" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The duration over which the discount is to be applied</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.DiscountMargin.Duration</lx:path>
                                    <lx:label>Duration</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DurationUnits" type="durationUnitsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The units of time of the specified duration</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.DiscountMargin.DurationUnits</lx:path>
                                    <lx:label>Duration Units</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The order of this discount margin within the sequence.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.DiscountMargin.SequenceNumber</lx:path>
                                    <lx:label>Sequence Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="DocumentationInstructions" type="documentationInstructionsType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information specifying where the lender is to send the loan documents once the loan has been approved. This nomination over-rides an equivalent one at the Application level.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.DocumentationInstructions</lx:path>
                                <lx:label>Documentation Instructions</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="DSH" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Indicates whether this loan is a Defence Service Housing loan under the program between the Commonwealth Government and Westpac Banking Corporation.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.DSH</lx:path>
                                <lx:label>DSH</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AssignmentCertificate" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Under the DSH loan program, a certificate of assignment assigns the subsidy that is paid by the Commonwealth to a credit provider if, in accordance with the certificate, the credit provider makes an advance to another person (the assignee). If there is an assignment certificate, the details of the certificate are captured here.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.DSH.AssignmentCertificate</lx:path>
                                    <lx:label>Assignment Certificate</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EntitlementCertificate" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Under the Defence Services Housing loan program, a certificate of entitlement certifies that a subsidy is payable by the Commonwealth to the credit provider if, in accordance with the agreement, the credit provider makes that advance to the applicant or the assignee of the applicant. If there is an entitlement certificate, details of the certificate must be provided.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.DSH.EntitlementCertificate</lx:path>
                                    <lx:label>Entitlement Certificate</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FileNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Under the Defence Services Housing loan program, the Commonwealth's file number for the qualifying applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.DSH.FileNumber</lx:path>
                                    <lx:label>File Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Subsidised" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the DSH loan being applied for is subsidised</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.DSH.Subsidised</lx:path>
                                    <lx:label>Subsidised</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Supplement" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the DSH loan being applied for is a supplement to an existing DSH loan, whereby additional available DSH loan funds are advanced to the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.DSH.Supplement</lx:path>
                                    <lx:label>Supplement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="EquityRelease" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about the loan amount requested that applies only when the loan product being applied for is a reverse mortgage, or equity release.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.EquityRelease</lx:path>
                                <lx:label>Equity Release</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="AccommodationBond" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Captures information about the bond that the applicant intends to deposit with an aged-care facility when they taking up residence.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.EquityRelease.AccommodationBond</lx:path>
                                      <lx:label>Accommodation Bond</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="ExpectedAdmissionDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>For an accommodation bond, the expected date that the aged person will take up residence in the aged care facility must be provided.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.EquityRelease.AccommodationBond.ExpectedAdmissionDate</lx:path>
                                          <lx:label>Expected Admission Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ServiceProviderName" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>For an accommodation bond, the name of the aged care facility that will receive the accommodation bond must be provided - the "aged care facility name";</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.EquityRelease.AccommodationBond.ServiceProviderName</lx:path>
                                          <lx:label>Service Provider Name</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ServiceProviderNumber" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>For an accommodation bond, the service provider number of the aged care facility, probably the unique RACS ID (Residential Aged Care Service) allocated by the Department of Health and Ageing must be provided</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.EquityRelease.AccommodationBond.ServiceProviderNumber</lx:path>
                                          <lx:label>Service Provider Number</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="Amount" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>For a reverse mortgage, contains information about the way that the applicant wants the loan funds to be advanced. The loan amount requested may be advanced in combinations of the following: 1. a lump sum paid at settlement; 2. instalment amounts paid periodically; and/or 3) a cash reserve amount that may be drawn down at will.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.EquityRelease.Amount</lx:path>
                                      <lx:label>Amount</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="CalculateAsPercentage" type="percentType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>For a reverse mortgage, captures the percentage of the value of the security property that the applicant requests to borrow. This is provided as an as an alternative to specifying an absolute loan amount. The actual loan amount will be calculated once the valuation of the security property has been completed.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.EquityRelease.Amount.CalculateAsPercentage</lx:path>
                                          <lx:label>Calculate As Percentage</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CashReserve" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the amount of the total loan amount requested that is to be advanced to the applicant as a cash reserve. Also known as a line of credit.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.EquityRelease.Amount.CashReserve</lx:path>
                                          <lx:label>Cash Reserve</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="InstalmentsAmount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the amount of the total loan amount requested that is to be advanced to the applicant in instalments. This amount is the sum of the instalment payments.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.EquityRelease.Amount.InstalmentsAmount</lx:path>
                                          <lx:label>Instalments Amount</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="LumpSum" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the amount of the total loan amount requested that is to be advanced to the applicant as a lump sum at settlement.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.EquityRelease.Amount.LumpSum</lx:path>
                                          <lx:label>Lump Sum</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="VaryOnValuation" type="varyOnValuationList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The amount of the loan advanced may vary if the amount the security property is valued at differs from the applicant's estimate of its value. The applicant may be able to specify what component of the payment should be varied in this situation. The applicant may specify that one of the lump sum amount OR the cash reserve amount OR the periodic payment be varied by any shortfall or surplus when the valuation amount differs from the applicant's estimate.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.EquityRelease.Amount.VaryOnValuation</lx:path>
                                          <lx:label>Vary On Valuation</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="Instalments" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Contains information about instalment payments, including the amount, number and regularity, and start date for the payments.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.EquityRelease.Instalments</lx:path>
                                      <lx:label>Instalments</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="AmountPerInstalment" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>For instalment payments with a specified amount, the amount of the monthly (or other period ) payments must be provided.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.EquityRelease.Instalments.AmountPerInstalment</lx:path>
                                          <lx:label>Amount Per Instalment</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Frequency" type="instalmentsFrequencyList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The frequency of instalment payments</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.EquityRelease.Instalments.Frequency</lx:path>
                                          <lx:label>Frequency</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Indexed" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant may elect to have their instalment payments indexed in order to adjust for inflation. The applicant must be able to specify if their instalment payments are to be indexed, and if so the rate of indexation to be applied.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.EquityRelease.Instalments.Indexed</lx:path>
                                          <lx:label>Indexed</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IndexRate" type="percentType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>If instalment payments are to be indexed, specifies the percentage rate of indexation to be applied.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.EquityRelease.Instalments.IndexRate</lx:path>
                                          <lx:label>Index Rate</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NumberOfInstalments" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>For instalment payments without a specified amount, the number of instalment payments to be made must be provided</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.EquityRelease.Instalments.NumberOfInstalments</lx:path>
                                          <lx:label>Number Of Instalments</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="ProtectedEquity" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An applicant for a reverse mortgage may specify that a percentage of the equity in the security property be protected against the increasing balance of the loan, to preserve it for the borrower's future needs. The lender cannot use proceeds of sale of the property that relate to the protected equity portion of the loan to repay the outstanding balance of the loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.EquityRelease.ProtectedEquity</lx:path>
                                    <lx:label>Protected Equity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProtectedEquityPercentage" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>If a percentage of the applicant's equity in the security property is to be protected, specifies the percentage of equity protected.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.EquityRelease.ProtectedEquityPercentage</lx:path>
                                    <lx:label>Protected Equity Percentage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="FeaturesSelected" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about the product features requested by the applicant.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected</lx:path>
                                <lx:label>Features Selected</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="DepositAccount" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>The features selected for this deposit account associated with the loan account.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.DepositAccount</lx:path>
                                      <lx:label>Deposit Account</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Holding" type="accountHoldingList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates whether the applicant has requested a deposit account held by a sole account holder or held by joint account holders.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.DepositAccount.Holding</lx:path>
                                          <lx:label>Holding</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="ExtraFeature" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Information about a product feature that is requested on the loan account.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.ExtraFeature</lx:path>
                                      <lx:label>Extra Feature</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Description" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Specific information, value or enumeration related to the specific feature (name).</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.ExtraFeature.Description</lx:path>
                                          <lx:label>Description</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Name" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The name of the specific feature being requested. Free text field although lenders may provide enumerations</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.ExtraFeature.Name</lx:path>
                                          <lx:label>Name</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Unique Identifier for this Extra Feature Element</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.ExtraFeature.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="OffsetAccount" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Identifies the BSB, Account Number and Account Name for the offset account.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.OffsetAccount</lx:path>
                                      <lx:label>Offset Account</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="AccountNumber" type="financialAccountType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Account number details such as BSB, Account Number and Account Name for the offset account</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.OffsetAccount.AccountNumber</lx:path>
                                            <lx:label>Account Number</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Holding" type="accountHoldingList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The offest account is held by a sole account holder or is held by joint account holders.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.OffsetAccount.Holding</lx:path>
                                          <lx:label>Holding</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IsExisting" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates whether the Offset Account detailed is an existing Offset Account or not</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.OffsetAccount.IsExisting</lx:path>
                                          <lx:label>Is Existing</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Unique Identifier for this Offset Account Element</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.OffsetAccount.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Account" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A cross reference to an account (a Non Real Estate Asset element) to be used as the offset account.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.OffsetAccount.x_Account</lx:path>
                                          <lx:label>Account</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="ChequeBook" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that a cheque book be issued for the loan account.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.ChequeBook</lx:path>
                                    <lx:label>Cheque Book</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CreditCard" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant has requested a credit card.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.CreditCard</lx:path>
                                    <lx:label>Credit Card</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DebitCard" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant has requested a debit card.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.DebitCard</lx:path>
                                    <lx:label>Debit Card</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DepositAccountRequested" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant has requested one or more deposit accounts be established for the loan account.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.DepositAccountRequested</lx:path>
                                    <lx:label>Deposit Account Requested</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DepositBook" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant has requested a deposit book</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.DepositBook</lx:path>
                                    <lx:label>Deposit Book</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="HolidayLeave" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that the loan product allows them to temporarily pause making repayments while they are not earning a regular income or for some other acceptable reason.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.HolidayLeave</lx:path>
                                    <lx:label>Holiday Leave</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Offset" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that a 100% offset account facility be established for the loan account.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.Offset</lx:path>
                                    <lx:label>Offset</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OffsetPercentage" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the percentage of the balance in the offset account that should be used to reduce the outstanding balance on the loan account when calculating interest payable. Typically 100%, but some lenders offer partial offset accounts.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.OffsetPercentage</lx:path>
                                    <lx:label>Offset Percentage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ParentalLeave" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that the loan product allows them to reduce their repayments while they are not earning a regular income because they are taking maternity/parental leave.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.ParentalLeave</lx:path>
                                    <lx:label>Parental Leave</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PartialOffset" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that an offset account be established for the loan account with less than 100% offset of the balance in the offset account against the outstanding loan balance.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.PartialOffset</lx:path>
                                    <lx:label>Partial Offset</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Portability" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that the loan facility be portable between security properties.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.Portability</lx:path>
                                    <lx:label>Portability</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProgressiveDraw" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that the loan funds be drawn down progressively, typically against progress of stages under a construction contract.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.ProgressiveDraw</lx:path>
                                    <lx:label>Progressive Draw</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RateLock" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that the interest rate for a fixed rate product be 'locked' until settlement</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.RateLock</lx:path>
                                    <lx:label>Rate Lock</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Redraw" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant requests that the loan product permit the redraw of loan principal that has been repaid in advance of the amortisation schedule for the loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.Redraw</lx:path>
                                    <lx:label>Redraw</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SplitLoan" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the applicant requests that the loan amount be split between two or more separate accounts or split facilities.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.FeaturesSelected.SplitLoan</lx:path>
                                    <lx:label>Split Loan</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="FundsDisbursement" type="fundsDisbursementType" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Specification of how funds are to be distributed.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.FundsDisbursement</lx:path>
                                <lx:label>Funds Disbursement</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="Guarantor" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>A reference to the party that is applying to be a guarantor on this loan account.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.Guarantor</lx:path>
                                <lx:label>Guarantor</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Guarantor" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A reference to the party that is applying to be a guarantor on this loan account.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Guarantor.x_Guarantor</lx:path>
                                    <lx:label>Guarantor</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="InterestPayable" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Interest that will be paid over the term of the loan.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.InterestPayable</lx:path>
                                <lx:label>Interest Payable</lx:label>
                                <li:transactions>DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="TotalAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The total interest amount that will be paid over the term of the loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.InterestPayable.TotalAmount</lx:path>
                                    <lx:label>Total Amount</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="LendingPurpose" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Details of the purpose for which the applicant intends to use the loan funds.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.LendingPurpose</lx:path>
                                <lx:label>Lending Purpose</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="PercentBenefit" type="percentOwnedType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Identifies the parties that will benefit from this lending purpose.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.LendingPurpose.PercentBenefit</lx:path>
                                      <lx:label>Percent Benefit</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="ABSLendingPurpose" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the free text description of the ABS Lending Purpose Code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LendingPurpose.ABSLendingPurpose</lx:path>
                                    <lx:label>ABS Lending Purpose</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ABSLendingPurposeCode" type="absLendingPurposeCodeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the ABS code for this purpose for which the loan funds will be used by the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LendingPurpose.ABSLendingPurposeCode</lx:path>
                                    <lx:label>ABS Lending Purpose Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures a description of the purpose if none of the purposes specified in the lender's enumerated list of purposes adequately described the true purpose.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LendingPurpose.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludesRefinancing" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Whether all or part of this credit facility will be used to refinance in full or in part one or many existing credit facilities.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LendingPurpose.IncludesRefinancing</lx:path>
                                    <lx:label>Includes Refinancing</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LenderCode" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the lender-specified code for this purpose for which the loan funds will be used by the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LendingPurpose.LenderCode</lx:path>
                                    <lx:label>Lender Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PayoutQuoteObtained" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Confirmation that a payout quote has been obtained with respect to all facilities to be refinanced using this facility.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LendingPurpose.PayoutQuoteObtained</lx:path>
                                    <lx:label>Payout Quote Obtained</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PurposeAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount of the loan funds that will be used for the specified purpose.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LendingPurpose.PurposeAmount</lx:path>
                                    <lx:label>Purpose Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PurposeOfFunds" type="purposeOfFundsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The purpose for which the loan funds will be used.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LendingPurpose.PurposeOfFunds</lx:path>
                                    <lx:label>Purpose of Funds</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RefinancingReason" type="principalRefinancingReasonList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Borrowers should inform the lender of the reason for refinancing liabilities</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LendingPurpose.RefinancingReason</lx:path>
                                    <lx:label>Refinancing Reason</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SecurityForMarginLoan" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicate if shares or an investment purchased using this loan are to be used as security for a margin loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LendingPurpose.SecurityForMarginLoan</lx:path>
                                    <lx:label>Security For Margin Loan</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unique Identifier for this Lending Purpose Element</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LendingPurpose.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="LoanPurpose" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about the main purpose for which the applicant intends to use the loan funds.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.LoanPurpose</lx:path>
                                <lx:label>Loan Purpose</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="NCCPStatus" type="nccpStatusList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the loan is regulated by the NCCP. This determination is made by the lender based on information provided by the applicant about the purpose for which the loan funds will be used, so this field is unlikely to be requested from the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LoanPurpose.NCCPStatus</lx:path>
                                    <lx:label>NCCP Status</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Occupancy" type="occupancyList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant's intention as to who will occupy the dwelling after the purchase completes.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LoanPurpose.Occupancy</lx:path>
                                    <lx:label>Occupancy</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OwnerBuilderApplication" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant intends to use the loan funds to fund the building of an owner/builder constructed property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LoanPurpose.OwnerBuilderApplication</lx:path>
                                    <lx:label>Owner Builder Application</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PrimaryPurpose" type="primaryPurposeLoanPurposeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the primary purpose for which the loan funds will be used by the applicant. Categories reflect the categories used in classifying a loan as regulated or unregulated under the NCCP.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LoanPurpose.PrimaryPurpose</lx:path>
                                    <lx:label>Primary Purpose</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RBALendingPurpose" type="rbaLendingPurposeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the RBA specific lending purpose for which the loan funds will be used by the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.LoanPurpose.RBALendingPurpose</lx:path>
                                    <lx:label>RBA Lending Purpose</lx:label>
                                    <li:transactions>CAL,CDA,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Package" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about a package option that may be available for the individual loan account.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.Package</lx:path>
                                <lx:label>Package</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Category" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures a lender defined category to which the requested package belongs.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Package.Category</lx:path>
                                    <lx:label>Category</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Code" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A unique code representing the product package being requested by the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Package.Code</lx:path>
                                    <lx:label>Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MemberID" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Membership Identifier of the applicant issued by the Organisation the applicant claims to belong to and therefore qualify for the package associated with that Organisation. May be provided when Organisation is specified.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Package.MemberID</lx:path>
                                    <lx:label>Member ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the name of a package that the applicant is applying for. A package commonly gives the borrowers certain benefits, such as an interest rate and fee discounts, for a fee. There may be rules that determine whether the applicants qualify for the package.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Package.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OptionCode" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A unique code identifying an option that applies to a package.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Package.OptionCode</lx:path>
                                    <lx:label>Option Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Organisation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A code or name that identifies the organisation under which the package is offered.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Package.Organisation</lx:path>
                                    <lx:label>Organisation</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PortfolioFund" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The portfolio fund to which this loan is assigned.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.PortfolioFund</lx:path>
                                <lx:label>Portfolio Fund</lx:label>
                                <li:transactions>ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="PortfolioFundID" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An identifier that can be used to uniquely identify the portfolio fund.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.PortfolioFund.PortfolioFundID</lx:path>
                                    <lx:label>Portfolio Fund ID</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PortfolioFundName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the portfolio fund.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.PortfolioFund.PortfolioFundName</lx:path>
                                    <lx:label>Portfolio Fund Name</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PreferredCorrespondent" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>A party who will be receiving correspondence on behalf of the applicant.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.PreferredCorrespondent</lx:path>
                                <lx:label>Preferred Correspondent</lx:label>
                                <li:transactions>ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AccessType" type="accessTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the level of access for this correspondent.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.PreferredCorrespondent.AccessType</lx:path>
                                    <lx:label>Access Type</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PreferredCorrespondence" type="preferredCorrespondenceList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The preferred method for this correspondent to receive correspondence relating to this loan such as welcome letters, loan statements, arrears notifications, changes of interest rate notifications, and other notifications.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.PreferredCorrespondent.PreferredCorrespondence</lx:path>
                                    <lx:label>Preferred Correspondence</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Correspondent" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the party who will be receiving correspondence on behalf of the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.PreferredCorrespondent.x_Correspondent</lx:path>
                                    <lx:label>Correspondent</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ProposedRepayment" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information that specifies how the borrower(s) intend to make regular or structured repayments.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment</lx:path>
                                <lx:label>Proposed Repayment</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Authoriser" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>A party that has authorised the proposed repayment.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.Authoriser</lx:path>
                                      <lx:label>Authoriser</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Authorised" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The authoriser has authorised the proposed repayment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.Authoriser.Authorised</lx:path>
                                          <lx:label>Authorised</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AuthorityVerified" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The authority of the authorising party has been verified.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.Authoriser.AuthorityVerified</lx:path>
                                          <lx:label>Authority Verified</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Party" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A cross reference to the authorising party.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.Authoriser.x_Party</lx:path>
                                          <lx:label>x_Party</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                            <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="CreditCard" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Details of the credit card to be used for repayments</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.CreditCard</lx:path>
                                      <lx:label>Credit Card</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="CardHolderName" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Cardholder name on the credit card to be used for repayments</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.CreditCard.CardHolderName</lx:path>
                                          <lx:label>Card Holder Name</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ExpiryMonth" type="mmMonthList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Month of expiry date of the credit card to be used for repayments</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.CreditCard.ExpiryMonth</lx:path>
                                          <lx:label>Expiry Month</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ExpiryYear" type="yyYearList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Year of expiry date of the credit card to be used for repayments</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.CreditCard.ExpiryYear</lx:path>
                                          <lx:label>Expiry Year</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Number" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Credit card number to be used for repayments</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.CreditCard.Number</lx:path>
                                          <lx:label>Number</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Type" type="creditCardTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Type of credit card to be used for repayments</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.CreditCard.Type</lx:path>
                                          <lx:label>Type</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="FromAccount" type="financialAccountType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Captures BSB, Account Number and Account Name for the account from which the repayments will be made.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.FromAccount</lx:path>
                                      <lx:label>From Account</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="RegularRepayment" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Container element for information associated with regular repayments</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment</lx:path>
                                      <lx:label>Regular Repayment</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Amount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Preferred or nominated total regular repayment amount (incorporating any GST or Stamp Duty in the case of Equipment Finance Transactions)</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.Amount</lx:path>
                                          <lx:label>Amount</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="DayOfMonth" type="dayOfMonthPattern" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The day of the month that the regular repayment will be made.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.DayOfMonth</lx:path>
                                          <lx:label>Day Of Month</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="DayOfWeek" type="dayOfWeekList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The day of the week that the regular repayment will be made.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.DayOfWeek</lx:path>
                                          <lx:label>Day Of Week</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="EndOfPeriod" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The payment will be made at the end of the end of the period e.g. end of month (EOM), end of quarter (EOQ), or end of year (EOY).</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.EndOfPeriod</lx:path>
                                          <lx:label>End Of Period</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="FirstRepaymentDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date that the first repayment of this regular repayment will occur.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.FirstRepaymentDate</lx:path>
                                          <lx:label>First Repayment Date</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Frequency" type="frequencyFullList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the applicants' preferred period for making loan repayments for the loan account. Captured when the loan product selected offers a range of repayment periods, enabling the applicant to nominate a period which best suits their needs in terms of income and the desire to reduce interest over the life of the loan by accelerating the reduction of the principal balance.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.Frequency</lx:path>
                                          <lx:label>Frequency</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="FrequencyInterval" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The time interval at which the repayment is repeated. Expressed as a multiple of the specified frequency.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.FrequencyInterval</lx:path>
                                          <lx:label>Frequency Interval</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="GSTAmount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The dollar amount of the GST incorporated into the Regular Payment (typically used in Equipment Finance Transactions).</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.GSTAmount</lx:path>
                                          <lx:label>GST Amount</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="InterestPayment" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>This payment covers the interest amount.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.InterestPayment</lx:path>
                                          <lx:label>Interest Payment</lx:label>
                                          <li:transactions>CAL,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="LastRepaymentDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date that the last repayment of this regular repayment will occur.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.LastRepaymentDate</lx:path>
                                          <lx:label>Last Repayment Date</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="LoanPaymentScheduleType" type="loanPaymentScheduleTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The type of loan payment schedule.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.LoanPaymentScheduleType</lx:path>
                                          <lx:label>Loan Payment Schedule Type</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="MinimumAmount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The minimum amount that must be paid.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.MinimumAmount</lx:path>
                                          <lx:label>Minimum Amount</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PrincipalPayment" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>This payment reduces the principal amount.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.PrincipalPayment</lx:path>
                                          <lx:label>Principal Payment</lx:label>
                                          <li:transactions>CAL,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StampDutyAmount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The dollar amount of the Stamp Duty incorporated into the Regular Payment (typically used in Equipment Finance Transactions).</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.StampDutyAmount</lx:path>
                                          <lx:label>Stamp Duty Amount</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TotalRepayments" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the total number of loan repayments.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.TotalRepayments</lx:path>
                                          <lx:label>Total Repayments</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Week" type="weekList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The week within the period that the regular repayment will be made.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment.Week</lx:path>
                                          <lx:label>Week</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="StructuredPayments" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Container element used when structured repayments rather than regular repayments are proposed or advised</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.StructuredPayments</lx:path>
                                      <lx:label>Structured Payments</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Payment" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>Details of itemised proposed payments. A given loan may have differing payment regimes at different periods in its total term (e.g. a fixed rate IO loan may have an initial Fixed Rate IO period, followed by a Variable Rate Interest Only period, followed by a Variable Rate Principal and Interest period.)</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.StructuredPayments.Payment</lx:path>
                                            <lx:label>Payment</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Amount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the total amount of the payment in the payment period, being the calculated repayment amount plus any service fee, GST or Stamp Duty.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.StructuredPayments.Payment.Amount</lx:path>
                                                <lx:label>Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Date" type="dateType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The date of the specified payment</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.StructuredPayments.Payment.Date</lx:path>
                                                <lx:label>Date</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GSTAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The dollar amount of the GST incorporated into this Structured Payment (typically used in Equipment Finance Transactions).</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.StructuredPayments.Payment.GSTAmount</lx:path>
                                                <lx:label>GST Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GSTOnTermChargesIncluded" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The GST on term charges is included in this payment. The GST amount for this payment includes both the GST charged on this payment amount, and the GST on term charges for the full term of the loan.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.StructuredPayments.Payment.GSTOnTermChargesIncluded</lx:path>
                                                <lx:label>GSTOnTermChargesIncluded</lx:label>
                                                <li:transactions>CAL,CDA,DAS</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GSTRebateAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>An input tax credit (ITC) amount that the applicant can claim on an eligible purchase.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.StructuredPayments.Payment.GSTRebateAmount</lx:path>
                                                <lx:label>GST Rebate Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GSTRebateAmountReplacingPayment" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The payment will only consist of the GST rebate amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.StructuredPayments.Payment.GSTRebateAmountReplacingPayment</lx:path>
                                                <lx:label>GST Rebate Amount Replacing Payment</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the sequence number of the payment or payment period.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.StructuredPayments.Payment.SequenceNumber</lx:path>
                                                <lx:label>Sequence Number</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="StampDutyAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The dollar amount of the Stamp Duty incorporated into this Structured Payment (typically used in Equipment Finance Transactions).</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.StructuredPayments.Payment.StampDutyAmount</lx:path>
                                                <lx:label>Stamp Duty Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Unique Identifier for this Payment Element</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.StructuredPayments.Payment.UniqueID</lx:path>
                                                <lx:label>Unique ID</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="BaseFrequency" type="frequencyFullList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Although structured payments may have arbitrary payment dates, the base frequency defines the fundamental unit of time used for determining payment dates and interest payment calculations.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.StructuredPayments.BaseFrequency</lx:path>
                                          <lx:label>Base Frequency</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="AnniversaryDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The preferred start (anniversary) date for regular repayments</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.AnniversaryDate</lx:path>
                                    <lx:label>Anniversary Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the repayment method that will be used to repay the loan when Method = "Other"</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Method" type="proposedRepaymentMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the payment method that will be used to effect the transfer of repayments on the loan to the lender. Examples include salary sacrifice, direct debit and Bpay.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.Method</lx:path>
                                    <lx:label>Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentTiming" type="paymentTimingList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the timing of payments with respect to their payment period.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.PaymentTiming</lx:path>
                                    <lx:label>Payment Timing</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PerDiemPaymentAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An optional partial-period payment to bring payment periods for several loans onto a common anniversary date (this is a total amount, so it includes GST and Stamp Duty).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.PerDiemPaymentAmount</lx:path>
                                    <lx:label>Per Diem Payment Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PerDiemPaymentDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The proposed date of the per diem payment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.PerDiemPaymentDate</lx:path>
                                    <lx:label>Per Diem Payment Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PerDiemPaymentGSTAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The dollar amount of the GST portion of within the Per Diem Payment Amount</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.PerDiemPaymentGSTAmount</lx:path>
                                    <lx:label>Per Diem Payment GST Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PerDiemPaymentStampDutyAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The dollar amount of the Stamp Duty portion of the Per Diem Payment Amount</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.PerDiemPaymentStampDutyAmount</lx:path>
                                    <lx:label>Per Diem Payment StampDuty Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Regular" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether payments are regular (same amount on regular intervals) or structured (irregular intervals and/or amounts)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.Regular</lx:path>
                                    <lx:label>Regular</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Account" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the account (a Non Real Estate Asset element) from which the repayments will be made.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.ProposedRepayment.x_Account</lx:path>
                                    <lx:label>Account</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="RateComposition" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The structure of the interest rate to the borrower, based on the delivery rate with applied margins</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.RateComposition</lx:path>
                                <lx:label>Rate Composition</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="BaseRate" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The base rate on which other rates are calculated.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.RateComposition.BaseRate</lx:path>
                                      <lx:label>Base Rate</lx:label>
                                      <li:transactions>CAL,DAS,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Code" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The base rate code.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.RateComposition.BaseRate.Code</lx:path>
                                          <lx:label>Code</lx:label>
                                          <li:transactions>CAL,DAS,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Name" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The base rate name.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.RateComposition.BaseRate.Name</lx:path>
                                          <lx:label>Name</lx:label>
                                          <li:transactions>CAL,DAS,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Rate" type="percentType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The base rate as a percentage.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.RateComposition.BaseRate.Rate</lx:path>
                                          <lx:label>Rate</lx:label>
                                          <li:transactions>CAL,DAS,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="CustomerRiskMargin" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The customer risk margin added (as a percentage) based on the risk the customer presents to the bank.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.CustomerRiskMargin</lx:path>
                                    <lx:label>Customer Risk Margin</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DisclosedToCustomer" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The details of the rate are disclosed to the customer.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.DisclosedToCustomer</lx:path>
                                    <lx:label>Disclosed To Customer</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IntroducerMargin" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the introducer margin percentage</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.IntroducerMargin</lx:path>
                                    <lx:label>Introducer Margin</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IntroductoryMargin" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The introductory margin added (as a percentage) during the introductory period or honeymoon period.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.IntroductoryMargin</lx:path>
                                    <lx:label>Introductory Margin</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NetCustomerRiskMargin" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The customer risk margin minus the discount margin (as a percentage).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.NetCustomerRiskMargin</lx:path>
                                    <lx:label>Net Customer Risk Margin</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OriginatorMargin" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the originator margin percentage</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.OriginatorMargin</lx:path>
                                    <lx:label>Originator Margin</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentTypeMargin" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The payment type margin added (as a percentage) based on the payment type.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.PaymentTypeMargin</lx:path>
                                    <lx:label>Payment Type Margin</lx:label>
                                    <li:transactions>CAL,DAS,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PricingConstruct" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The pricing construct used to calculate the rate i.e. Base Rate + Term Premium.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.PricingConstruct</lx:path>
                                    <lx:label>Pricing Construct</lx:label>
                                    <li:transactions>CAL,DAS,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProductMargin" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The product margin added (as a percentage) based on the specific product type.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.ProductMargin</lx:path>
                                    <lx:label>Product Margin</lx:label>
                                    <li:transactions>CAL,DAS,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SubsidyBaseAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Upfront fixed amount</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.SubsidyBaseAmount</lx:path>
                                    <lx:label>Subsidy Base Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SubsidyBasePercent" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Upfront commission percent</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.SubsidyBasePercent</lx:path>
                                    <lx:label>Subsidy Base Percent</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SubsidyRateSacrifice" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Rate added to the borrower rate</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.SubsidyRateSacrifice</lx:path>
                                    <lx:label>Subsidy Rate Sacrifice</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TermPremium" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The term premium added (as a percentage) based on the loan term.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.TermPremium</lx:path>
                                    <lx:label>Term Premium</lx:label>
                                    <li:transactions>CAL,DAS,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalInterestRate" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The total interest rate for this rate composition scheme.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.TotalInterestRate</lx:path>
                                    <lx:label>Total Interest Rate</lx:label>
                                    <li:transactions>CAL,DAS,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.RateComposition.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Security" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains a reference to a security property that is submitted with the loan application, and information about the relationship between the subject loan account and that security property.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.Security</lx:path>
                                <lx:label>Security</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="SecurityAgreement" type="securityAgreementType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Security agreement details for this security.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.Security.SecurityAgreement</lx:path>
                                      <lx:label>Security Agreement</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="AllocatedAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The portion of the loan amount that is allocated to this security property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Security.AllocatedAmount</lx:path>
                                    <lx:label>Allocated Amount</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Priority" type="securityPriorityList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the priority of the new mortgage that is offered as security for the loans being applied for. When the mortgage is already held by the lender being applied to, this is the priority of that existing mortgage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Security.Priority</lx:path>
                                    <lx:label>Priority</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Security.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Security" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Contains a reference to a security (real estate or non real estate asset) that is submitted with the loan application, and information about the relationship between the subject loan account and that security.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Security.x_Security</lx:path>
                                    <lx:label>Security</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.GeneralSecurityAgreement</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Software" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The software application used by the loan originator.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.Software</lx:path>
                                <lx:label>Software</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the software provider, for example the company name.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Software.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the software application used by the loan originator.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Software.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="StatementInstructions" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Information specifying where loan statements should be sent regarding this loan.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.StatementInstructions</lx:path>
                                <lx:label>Statement Instructions</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="NameOnStatement" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>An applicant whose name should be included on the loan statement.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.StatementInstructions.NameOnStatement</lx:path>
                                      <lx:label>NameOnStatement</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="x_Applicant" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A cross reference to the applicant whose name should be included on the loan statement.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.StatementInstructions.NameOnStatement.x_Applicant</lx:path>
                                          <lx:label>Applicant</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="SupplementaryCardholder" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains a link to a supplementary (non liable) cardholder.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.SupplementaryCardholder</lx:path>
                                <lx:label>Supplementary Cardholder</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_SupplementaryCardholder" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to the loan party, related party or household dependant that will be a supplementary (non liable) cardholder.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.SupplementaryCardholder.x_SupplementaryCardholder</lx:path>
                                    <lx:label>Supplementary Cardholder</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Term" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the total term of the credit product and any introductory terms, such as an interest only term (including interest only reasons) or a fixed rate term that applies for an initial period at the start of the product.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.Term</lx:path>
                                <lx:label>Term</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="DistinctLoanPeriod" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>A distinct loan period, where each period may have a different payment type, interest type, rate composition.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod</lx:path>
                                      <lx:label>Distinct Loan Period</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Duration" type="durationType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The duration of the distinct loan period.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod.Duration</lx:path>
                                            <lx:label>Duration</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="Repayment" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>A repayment that is made during this period.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod.Repayment</lx:path>
                                            <lx:label>Repayment</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="x_Repayment" type="referenceType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>A cross reference to the repayment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod.Repayment.x_Repayment</lx:path>
                                                <lx:label>Repayment</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                <lx:references>
                                                  <lx:target>Package.Content.Application.LoanDetails.ProposedRepayment.StructuredPayments.Payment</lx:target>
                                                  <lx:target>Package.Content.Application.LoanDetails.ProposedRepayment.RegularRepayment</lx:target>
                                                </lx:references>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="EndDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The end date of the distinct loan period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod.EndDate</lx:path>
                                          <lx:label>End Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="InterestType" type="interestTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The type of interest that is to be paid during the distinct loan period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod.InterestType</lx:path>
                                          <lx:label>Interest Type</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PaymentType" type="paymentTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The type of repayment that is to be made during the distinct loan period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod.PaymentType</lx:path>
                                          <lx:label>Payment Type</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The order of this distinct loan period within the sequence.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod.SequenceNumber</lx:path>
                                          <lx:label>Sequence Number</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StartDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The start date of the distinct loan period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod.StartDate</lx:path>
                                          <lx:label>Start Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TotalFeesAmount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The anticipated total dollar amount of all fees paid during this distinct loan period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod.TotalFeesAmount</lx:path>
                                          <lx:label>Total Fees Amount</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TotalInterestAmount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The anticipated total dollar amount of interest paid during this distinct loan period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod.TotalInterestAmount</lx:path>
                                          <lx:label>Total Interest Amount</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TotalRepayments" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The anticipated total number of repayments during this distinct loan period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod.TotalRepayments</lx:path>
                                          <lx:label>Total Repayments</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TotalRepaymentsAmount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The anticipated total dollar amount of all repayments including interest paid during this distinct loan period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod.TotalRepaymentsAmount</lx:path>
                                          <lx:label>Total Repayments Amount</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_RateComposition" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A cross reference to the rate composition for this distinct loan period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Term.DistinctLoanPeriod.x_RateComposition</lx:path>
                                          <lx:label>Rate Composition</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.LoanDetails.RateComposition</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="InterestOnlyReason" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Wrapper element for the applicant/s reasons for applying for an interest-only term within this this loan.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.LoanDetails.Term.InterestOnlyReason</lx:path>
                                      <lx:label>Interest Only Reason</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Description" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A free-text description provided by the applicant/s that provides further detail regarding this reason for applying for an interest-only term within this loan.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Term.InterestOnlyReason.Description</lx:path>
                                          <lx:label>Description</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Reason" type="interestOnlyReasonList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The reason that the applicant is applying for an interest-only term within this loan.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Term.InterestOnlyReason.Reason</lx:path>
                                          <lx:label>Reason</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Unique Identifier for this interest-only reason.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.LoanDetails.Term.InterestOnlyReason.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="GSTOnTermCharges" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The GST component of the term charges amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.GSTOnTermCharges</lx:path>
                                    <lx:label>GST On Term Charges</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GSTOnTermChargesCapitalised" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The GST component of the term charges will be added to the total amount financed for the loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.GSTOnTermChargesCapitalised</lx:path>
                                    <lx:label>GST On Term Charges Capitalised</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InterestType" type="interestTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the type of interest that is to be paid during the specified term.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.InterestType</lx:path>
                                    <lx:label>Interest Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InterestTypeDuration" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The duration of the fixed interest rate period, as a whole number multiple of the Units.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.InterestTypeDuration</lx:path>
                                    <lx:label>Interest Type Duration</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InterestTypeUnits" type="loanTermUnitsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The units of time that are used to define the duration of the fixed interest rate period.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.InterestTypeUnits</lx:path>
                                    <lx:label>Interest Type Units</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentType" type="paymentTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the type of repayment that is to be made during the specified term</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.PaymentType</lx:path>
                                    <lx:label>Payment Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentTypeDuration" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The duration of the interest only period, as a whole number multiple of the Units.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.PaymentTypeDuration</lx:path>
                                    <lx:label>Payment Type Duration</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentTypeUnits" type="loanTermUnitsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The units of time that are used to define the duration of the interest only period.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.PaymentTypeUnits</lx:path>
                                    <lx:label>Payment Type Units</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RolloverPeriodDuration" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The duration of the rollover period for a commercial bill.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.RolloverPeriodDuration</lx:path>
                                    <lx:label>RolloverPeriodDuration</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RolloverPeriodUnits" type="loanTermUnitsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The units of time that are used to define the rollover period for a commercial bill.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.RolloverPeriodUnits</lx:path>
                                    <lx:label>RolloverPeriodUnits</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TermChargesAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The dollar amount of the term charges over the full term of the loan. Term charges are the total interest amount charged, plus any undisclosed capitalised fees, plus brokerage fees.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.TermChargesAmount</lx:path>
                                    <lx:label>Term Charges Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalFeesAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The anticipated total dollar amount of all fees paid over the full term of the loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.TotalFeesAmount</lx:path>
                                    <lx:label>Total Fees Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalInterestAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The anticipated total dollar amount of interest paid over the full term of the loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.TotalInterestAmount</lx:path>
                                    <lx:label>Total Interest Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalRepayments" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The anticipated total number of repayments over the full term of the loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.TotalRepayments</lx:path>
                                    <lx:label>Total Repayments</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalRepaymentsAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The anticipated total dollar amount of all repayments including interest paid over the full term of the loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.TotalRepaymentsAmount</lx:path>
                                    <lx:label>Total Repayments Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalTermDuration" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The Duration of the specified loan term, as a multiple of the Units</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.TotalTermDuration</lx:path>
                                    <lx:label>Total Term Duration</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalTermType" type="totalTermTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of loan term should indicate that the loan term is either the total length of the term or the amortised term as required by the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.TotalTermType</lx:path>
                                    <lx:label>Total Term Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalTermUnits" type="loanTermUnitsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the period described by the number specified for term.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.Term.TotalTermUnits</lx:path>
                                    <lx:label>Total Term Units</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="TermsAndConditions" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The terms and conditions of a loan contract or loan application.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.LoanDetails.TermsAndConditions</lx:path>
                                <lx:label>Terms And Conditions</lx:label>
                                <li:transactions>CAL,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="TermsCode" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The code used to identify the terms and conditions in the lender's system.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.TermsAndConditions.TermsCode</lx:path>
                                    <lx:label>TermsCode</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TermsName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name used by the lender for this set of terms and conditions.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.TermsAndConditions.TermsName</lx:path>
                                    <lx:label>TermsName</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TermsVersion" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The version used by the lender for this set of terms and conditions.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.LoanDetails.TermsAndConditions.TermsVersion</lx:path>
                                    <lx:label>Terms Version</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="AccelerationPercentage" type="percentType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The amount that is provided to the borrower as a percentage of the total amount that is provided from the invoices for an invoice financing loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.AccelerationPercentage</lx:path>
                              <lx:label>Acceleration Percentage</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="AmountRequested" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the loan amount as requested by the applicant for this loan account, exclusive of to-be-capitalised fees and/or LMI.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.AmountRequested</lx:path>
                              <lx:label>Amount Requested</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="AmountRequestedInclusive" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The loan amount as requested by the applicant for this loan account, inclusive of to-be-capitalised fees and LMI.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.AmountRequestedInclusive</lx:path>
                              <lx:label>Amount Requested Inclusive</lx:label>
                              <li:transactions>CAL,CDA,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="AmountToBeFinanced" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the amount to be financed for this loan account, inclusive of any capitalised fees and/or LMI. This is the limit for limit based products.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.AmountToBeFinanced</lx:path>
                              <lx:label>Amount to Be Financed</lx:label>
                              <li:transactions>CDA,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BalloonRepaymentAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The proposed nominated balloon repayment amount</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.BalloonRepaymentAmount</lx:path>
                              <lx:label>Balloon Repayment Amount</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BalloonRepaymentDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The expected date of the proposed balloon repayment</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.BalloonRepaymentDate</lx:path>
                              <lx:label>Balloon Repayment Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BuyNowPayLater" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The loan is a buy now pay later facility according to the lender's policy.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.BuyNowPayLater</lx:path>
                              <lx:label>Buy Now Pay Later</lx:label>
                              <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ConformsWithUnderwritingStandards" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Mortgage conforms to the underwriting standards of the mortgage insurer.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.ConformsWithUnderwritingStandards</lx:path>
                              <lx:label>Conforms With Underwriting Standards</lx:label>
                              <li:transactions>ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ConsiderLowerLimitIfNotEligible" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that the applicant would like a lower limit to be considered if not eligible for the requested amount.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.ConsiderLowerLimitIfNotEligible</lx:path>
                              <lx:label>Consider Lower Limit If Not Eligible</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DocType" type="applicationDocTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the level of income verification documentation for each loan provided to the lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.DocType</lx:path>
                              <lx:label>Doc Type</lx:label>
                              <li:transactions>ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="EstimatedSettlementDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the date that the applicant estimates that the application will settle.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.EstimatedSettlementDate</lx:path>
                              <lx:label>Estimated Settlement Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Funder" type="funderList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the name of the loan funder.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.Funder</lx:path>
                              <lx:label>Funder</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Grade" type="gradeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Grade of the loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.Grade</lx:path>
                              <lx:label>Grade</lx:label>
                              <li:transactions>ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="InterestCalculationFrequency" type="interestChargeFrequencyList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The frequency at which interest will be calculated on the loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.InterestCalculationFrequency</lx:path>
                              <lx:label>Interest Calculation Frequency</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="InterestChargeFrequency" type="interestChargeFrequencyList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The frequency at which interest will be charged on the loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.InterestChargeFrequency</lx:path>
                              <lx:label>Interest Charge Frequency</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsAmountRequestedInForeignCurrency" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>A flag to indicate that the amount is requested in a foreign currency.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.IsAmountRequestedInForeignCurrency</lx:path>
                              <lx:label>Is Amount Requested In Foreign Currency</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LenderLoanApprovalDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date on which the loan was approved by the lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.LenderLoanApprovalDate</lx:path>
                              <lx:label>Lender Loan Approval Date</lx:label>
                              <li:transactions>LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LoanType" type="loanTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of credit facility.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.LoanType</lx:path>
                              <lx:label>Credit Facility Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LVR" type="percentType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Loan to value ratio.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.LVR</lx:path>
                              <lx:label>LVR</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="MainProduct" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Where a lender requires that the applicant nominate one loan account/product as being the primary or main account within a package or portfolio or accounts, indicates whether this is the main account.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.MainProduct</lx:path>
                              <lx:label>Main Product</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="MaturityDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The maturity date of the loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.MaturityDate</lx:path>
                              <lx:label>Maturity Date</lx:label>
                              <li:transactions>CAL,CDA,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="MinimumRepaymentRate" type="percentType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The percentage rate used to calculate the minimum repayment for the loan account.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.MinimumRepaymentRate</lx:path>
                              <lx:label>Minimum Repayment Rate</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NegativelyGeared" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the interest and maintenance expense will be greater than the rental income</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.NegativelyGeared</lx:path>
                              <lx:label>Negatively Geared</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NominateBalloonRepayment" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate whether a balloon repayment is proposed (yes) or not proposed (no)</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.NominateBalloonRepayment</lx:path>
                              <lx:label>Nominate Balloon Repayment</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OriginatorReferenceID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Unique identifier assigned to the loan detail by the originator system</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.OriginatorReferenceID</lx:path>
                              <lx:label>Originator Reference ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProductCode" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures an identifying code for the loan product applied for. This code is typically issued by the lender and enables the lender's systems to uniquely identify the product.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.ProductCode</lx:path>
                              <lx:label>Product Code</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProductName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the name of the loan product applied for.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.ProductName</lx:path>
                              <lx:label>Product Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProposedAnnualInterestRate" type="percentType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The interest rate that applies to the loan account. Typically expressed as an annual interest rate.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.ProposedAnnualInterestRate</lx:path>
                              <lx:label>Proposed Annual Interest Rate</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RevertInterestRate" type="percentType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Estimate of the interest rate at the end of the fixed rate period or the end of the interest only period (whichever is longer). The revert interest rate will be an estimate based on the current variable rate.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.RevertInterestRate</lx:path>
                              <lx:label>Revert Interest Rate</lx:label>
                              <li:transactions>CAL,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Secured" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Whether the credit product is to be secured or not.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.Secured</lx:path>
                              <lx:label>Secured</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this loan details within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SMSFLoan" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to note whether the loan is for a self-managed super fund.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.SMSFLoan</lx:path>
                              <lx:label>SMSF Loan</lx:label>
                              <li:transactions>ACC,CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SpecialConcessionCode" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures a lender issued code that identifies a program of the lender that provides certain specified concessions for the loan applied for.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.SpecialConcessionCode</lx:path>
                              <lx:label>Special Concession Code</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="StatementCycle" type="statementCycleList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates preferred frequency for statements</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.StatementCycle</lx:path>
                              <lx:label>Statement Cycle</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="TaxDeductible" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate if the loan repayments are tax deductible</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.TaxDeductible</lx:path>
                              <lx:label>Tax Deductible</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="TemplateID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Allow user to preconfigured the Loan attributes at the message receiving end</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.TemplateID</lx:path>
                              <lx:label>Template ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_MasterAgreement" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>For a loan application under a master agreement, the link to the agreement details</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.LoanDetails.x_MasterAgreement</lx:path>
                              <lx:label>Master Agreement</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.MasterAgreement</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="MasterAgreement" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>A Master Agreement finance facility is a contractual document under which supplementary agreements for loans can be executed, and existing loans can be brought in.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.MasterAgreement</lx:path>
                          <lx:label>Master Agreement</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="AccountToIncorporate" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>For a new agreement that will incorporate existing loan accounts, container for the existing account(s)</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.MasterAgreement.AccountToIncorporate</lx:path>
                                <lx:label>Account To Incorporate</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_ExistingAccount" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For a new agreement that will incorporate existing loan accounts, pointer to the existing account(s)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.MasterAgreement.AccountToIncorporate.x_ExistingAccount</lx:path>
                                    <lx:label>Existing Account</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Liability</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_NewAccount" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to a new account that is being incorporated into the master agreement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.MasterAgreement.AccountToIncorporate.x_NewAccount</lx:path>
                                    <lx:label>New Account</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PercentOwned" type="percentOwnedType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Identifies the parties that are owners on this master agreement.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.MasterAgreement.PercentOwned</lx:path>
                                <lx:label>Percent Owned</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="Security" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The security offered under the master agreement finance facility.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.MasterAgreement.Security</lx:path>
                                <lx:label>Security</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="PercentOwned" type="percentOwnedType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>This item is now deprecated and will be deleted in a future release. Identifies the parties that are owners on this master agreement.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.MasterAgreement.Security.PercentOwned</lx:path>
                                      <lx:label>Percent Owned</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="SecurityAgreement" type="securityAgreementType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Security agreement details for this security.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.MasterAgreement.Security.SecurityAgreement</lx:path>
                                      <lx:label>Security Agreement</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Priority" type="securityPriorityList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The priority of this lender's mortgage on the security property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.MasterAgreement.Security.Priority</lx:path>
                                    <lx:label>Priority</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.MasterAgreement.Security.UniqueID</lx:path>
                                    <lx:label>UniqueID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Security" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the security being offered.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.MasterAgreement.Security.x_Security</lx:path>
                                    <lx:label>Security</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.GeneralSecurityAgreement</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="ApprovedLimit" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Total approved loan amount.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.MasterAgreement.ApprovedLimit</lx:path>
                              <lx:label>Approved Limit</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DateOfExecution" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>For an existing agreement, the date of the agreement</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.MasterAgreement.DateOfExecution</lx:path>
                              <lx:label>Date Of Execution</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Description" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A name or description of the facility</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.MasterAgreement.Description</lx:path>
                              <lx:label>Description</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="EndDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>For an existing agreement, the end date of the agreement</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.MasterAgreement.EndDate</lx:path>
                              <lx:label>End Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LenderAgreementNumber" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>For an existing agreement, the lender's identifier of the agreement</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.MasterAgreement.LenderAgreementNumber</lx:path>
                              <lx:label>Lender Agreement Number</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="MasterFacilityLimit" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>For a master facility, the requested total umbrella limit</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.MasterAgreement.MasterFacilityLimit</lx:path>
                              <lx:label>Master Facility Limit</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProductCode" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>An identifying code for the master agreement finance facility product. This code is typically issued by the lender and enables the lender's systems to uniquely identify the product.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.MasterAgreement.ProductCode</lx:path>
                              <lx:label>Product Code</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProductName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The name of the master agreement finance facility product.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.MasterAgreement.ProductName</lx:path>
                              <lx:label>Product Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Purpose" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Purpose of the master agreement.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.MasterAgreement.Purpose</lx:path>
                              <lx:label>Purpose</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this master agreement within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.MasterAgreement.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Status" type="masterAgreementStatusList" use="optional">
                          <xs:annotation>
                            <xs:documentation>A flag to indicate whether the master agreement exists or is being requested</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.MasterAgreement.Status</lx:path>
                              <lx:label>Status</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Type" type="masterAgreementTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of agreement.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.MasterAgreement.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.MasterAgreement.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="NonRealEstateAsset" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Contains details about an asset item that is not real estate, such as a term deposit account.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.NonRealEstateAsset</lx:path>
                          <lx:label>Non Real Estate Asset</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="AgriculturalAsset" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Agricultural equipment, crops and livestock as assets</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset</lx:path>
                                <lx:label>Agricultural Asset</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AdditionalIDType" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Type of additional ID, such as Microdot, UPC, GTIN, Asset Number</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.AdditionalIDType</lx:path>
                                    <lx:label>Additional ID Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AdditionalIDValue" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For an additional ID, the value of the ID</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.AdditionalIDValue</lx:path>
                                    <lx:label>Additional ID Value</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Age" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age in years. 0=New (alternatively, use year of manufacture)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.Age</lx:path>
                                    <lx:label>Age</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Condition" type="conditionList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Asset condition</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.Condition</lx:path>
                                    <lx:label>Condition</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConditionDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comments on the general condition and appearance of the asset e.g. good working order.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.ConditionDescription</lx:path>
                                    <lx:label>Condition Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CropArea" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The crop growing area.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.CropArea</lx:path>
                                    <lx:label>Crop Area</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CropAreaUnits" type="landAreaUnitsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measure for the crop growing area.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.CropAreaUnits</lx:path>
                                    <lx:label>Crop Area Units</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EffectiveLife" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Effective life of the asset in years, as determined by the ATO</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.EffectiveLife</lx:path>
                                    <lx:label>Effective Life</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EngineHoursTotal" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Total engine hours</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.EngineHoursTotal</lx:path>
                                    <lx:label>Engine Hours Total</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EngineID" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Identification number of the engine</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.EngineID</lx:path>
                                    <lx:label>Engine ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExcessUsageCalculationMethod" type="excessUsageCalculationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Method to calculate excess usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.ExcessUsageCalculationMethod</lx:path>
                                    <lx:label>Excess Usage Calculation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExpectedHarvestYear" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The calendar year in which the harvest is expected.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.ExpectedHarvestYear</lx:path>
                                    <lx:label>Expected Harvest Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GoodToBeUsedLocation" type="goodToBeUsedLocationList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>On Business Premises, On-site and returned to Business Premises each day, On-site until job is complete</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.GoodToBeUsedLocation</lx:path>
                                    <lx:label>Good To Be Used Location</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludedUsage" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Allowed usage based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.IncludedUsage</lx:path>
                                    <lx:label>Included Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Make" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The make of the agricultural asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.Make</lx:path>
                                    <lx:label>Make</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Model" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The model of the agricultural asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.Model</lx:path>
                                    <lx:label>Model</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherInformation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Other information which may impact the value of the asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.OtherInformation</lx:path>
                                    <lx:label>Other Information</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Quantity" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Quantity of assets. Typically "1", however for assets that are not individually registrable, indicates the number of items for a bulk grouping of assets being purchased and financed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.Quantity</lx:path>
                                    <lx:label>Quantity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegisteredInState" type="auStateList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the Australian state where the asset is registered.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.RegisteredInState</lx:path>
                                    <lx:label>Registered In State</lx:label>
                                    <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegistrationExpiryDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Date of expiry of registration.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.RegistrationExpiryDate</lx:path>
                                    <lx:label>Registration Expiry Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegistrationNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Registration identifier for the asset e.g. license plate number.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.RegistrationNumber</lx:path>
                                    <lx:label>Registration Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SerialNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The uniquely identifiable manufacturer serial number associated with the asset such as VIN</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.SerialNumber</lx:path>
                                    <lx:label>Serial Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ServicingHistory" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comments on any service history that will impact on the value of the asset.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.ServicingHistory</lx:path>
                                    <lx:label>Servicing History</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="agriculturalAssetTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific type of agricultural asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageAtStart" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age of the equipment based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.UsageAtStart</lx:path>
                                    <lx:label>Usage At Start</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageUnitOfMeasurement" type="usageUnitOfMeasurementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measurement to quantify the usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.UsageUnitOfMeasurement</lx:path>
                                    <lx:label>Usage Unit Of Measurement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_GoodToBeUsedAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Location of Goods (address)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.x_GoodToBeUsedAddress</lx:path>
                                    <lx:label>Good To Be Used Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Year" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Year of Manufacture</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.AgriculturalAsset.Year</lx:path>
                                    <lx:label>Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Aircraft" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Aircraft including fixed and rotary winged aircraft, light and ultralight craft</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft</lx:path>
                                <lx:label>Aircraft</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AdditionalIDType" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Type of additional ID, such as Microdot, UPC, GTIN, Asset Number</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.AdditionalIDType</lx:path>
                                    <lx:label>Additional ID Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AdditionalIDValue" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For an additional ID, the value of the ID</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.AdditionalIDValue</lx:path>
                                    <lx:label>Additional ID Value</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Age" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age in years. 0=New (alternatively, use year of manufacture)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.Age</lx:path>
                                    <lx:label>Age</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AirFrameHours" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Total hours on airframe as per log book.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.AirFrameHours</lx:path>
                                    <lx:label>Air Frame Hours</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Condition" type="conditionList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Asset condition</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.Condition</lx:path>
                                    <lx:label>Condition</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConditionDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comments on the general condition and appearance of the asset e.g. good working order.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.ConditionDescription</lx:path>
                                    <lx:label>Condition Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the aircraft</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EffectiveLife" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Effective life of the asset in years, as determined by the ATO</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.EffectiveLife</lx:path>
                                    <lx:label>Effective Life</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EngineHoursTotal" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Total engine hours</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.EngineHoursTotal</lx:path>
                                    <lx:label>Engine Hours Total</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EngineHoursTTR" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Time to run until next service, in hours.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.EngineHoursTTR</lx:path>
                                    <lx:label>Engine Hours TTR</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EngineID" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Identification number of the engine</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.EngineID</lx:path>
                                    <lx:label>Engine ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExcessUsageCalculationMethod" type="excessUsageCalculationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Method to calculate excess usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.ExcessUsageCalculationMethod</lx:path>
                                    <lx:label>Excess Usage Calculation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludedUsage" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Allowed usage based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.IncludedUsage</lx:path>
                                    <lx:label>Included Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Make" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The make of the aircraft</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.Make</lx:path>
                                    <lx:label>Make</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Model" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The model of the aircraft</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.Model</lx:path>
                                    <lx:label>Model</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherInformation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Other information which may impact the value of the asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.OtherInformation</lx:path>
                                    <lx:label>Other Information</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Quantity" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Quantity of assets. Typically "1", however for assets that are not individually registrable, indicates the number of items for a bulk grouping of assets being purchased and financed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.Quantity</lx:path>
                                    <lx:label>Quantity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegistrationExpiryDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Date of expiry of registration.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.RegistrationExpiryDate</lx:path>
                                    <lx:label>Registration Expiry Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegistrationNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Registration identifier for the asset e.g. license plate number.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.RegistrationNumber</lx:path>
                                    <lx:label>Registration Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SerialNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The uniquely identifiable manufacturer serial number associated with the asset such as VIN</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.SerialNumber</lx:path>
                                    <lx:label>Serial Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ServicingHistory" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comments on any service history that will impact on the value of the asset.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.ServicingHistory</lx:path>
                                    <lx:label>Servicing History</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="aircraftTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific type of aircraft</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageAtStart" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age of the equipment based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.UsageAtStart</lx:path>
                                    <lx:label>Usage At Start</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageUnitOfMeasurement" type="usageUnitOfMeasurementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measurement to quantify the usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.UsageUnitOfMeasurement</lx:path>
                                    <lx:label>Usage Unit Of Measurement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_GoodToBeUsedAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Location of Goods (address)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.x_GoodToBeUsedAddress</lx:path>
                                    <lx:label>Good To Be Used Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Year" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Year of Manufacture</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Aircraft.Year</lx:path>
                                    <lx:label>Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Business" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for capturing Business details</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.Business</lx:path>
                                <lx:label>Business</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="FinancialAnalysis" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Container element for the detailed financial analysis of the business</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.NonRealEstateAsset.Business.FinancialAnalysis</lx:path>
                                      <lx:label>Financial Analysis</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="CompanyFinancials" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>Container for the reference to the financial data</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.NonRealEstateAsset.Business.FinancialAnalysis.CompanyFinancials</lx:path>
                                            <lx:label>Company Financials</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="x_CompanyFinancials" type="referenceType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Reference to the element containing the financial analysis for each financial year, to be provided when Complete Financial Analysis = "Yes"</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.NonRealEstateAsset.Business.FinancialAnalysis.CompanyFinancials.x_CompanyFinancials</lx:path>
                                                <lx:label>Company Financials</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                <lx:references>
                                                  <lx:target>Package.Content.Application.CompanyFinancials</lx:target>
                                                </lx:references>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="AnnualPaymentOnCommitments" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Total Principle and Interest payments on borrowings due to be paid in the next 12 months</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.Business.FinancialAnalysis.AnnualPaymentOnCommitments</lx:path>
                                          <lx:label>Annual Payment On Commitments</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CompleteFinancialAnalysis" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Flag to indicate whether a complete financial analysis will be provided electronically, or only a summary view</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.Business.FinancialAnalysis.CompleteFinancialAnalysis</lx:path>
                                          <lx:label>Complete Financial Analysis</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="BusinessName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the business</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Business.BusinessName</lx:path>
                                    <lx:label>Business Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MainBusinessActivity" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Description of the businesses main activity</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Business.MainBusinessActivity</lx:path>
                                    <lx:label>Main Business Activity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="CleaningEquipment" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for Cleaning Equipment asset</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment</lx:path>
                                <lx:label>Cleaning Equipment</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AdditionalIDType" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Type of additional ID, such as Microdot, UPC, GTIN, Asset Number</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.AdditionalIDType</lx:path>
                                    <lx:label>Additional ID Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AdditionalIDValue" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For an additional ID, the value of the ID</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.AdditionalIDValue</lx:path>
                                    <lx:label>Additional ID Value</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Age" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age in years. 0=New (alternatively, use year of manufacture)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.Age</lx:path>
                                    <lx:label>Age</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Condition" type="conditionList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Asset condition</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.Condition</lx:path>
                                    <lx:label>Condition</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConditionDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comments on the general condition and appearance of the asset e.g. good working order.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.ConditionDescription</lx:path>
                                    <lx:label>Condition Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the Cleaning Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EffectiveLife" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Effective life of the asset in years, as determined by the ATO</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.EffectiveLife</lx:path>
                                    <lx:label>Effective Life</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExcessUsageCalculationMethod" type="excessUsageCalculationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Method to calculate excess usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.ExcessUsageCalculationMethod</lx:path>
                                    <lx:label>Excess Usage Calculation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GoodToBeUsedLocation" type="goodToBeUsedLocationList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>On Business Premises, On-site and returned to Business Premises each day, On-site until job is complete</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.GoodToBeUsedLocation</lx:path>
                                    <lx:label>Good To Be Used Location</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludedUsage" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Allowed usage based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.IncludedUsage</lx:path>
                                    <lx:label>Included Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Make" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The make of the Cleaning Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.Make</lx:path>
                                    <lx:label>Make</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Model" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The model of the Cleaning Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.Model</lx:path>
                                    <lx:label>Model</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherInformation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Other information which may impact the value of the asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.OtherInformation</lx:path>
                                    <lx:label>Other Information</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Quantity" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Quantity of assets. Typically "1", however for assets that are not individually registrable, indicates the number of items for a bulk grouping of assets being purchased and financed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.Quantity</lx:path>
                                    <lx:label>Quantity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SerialNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The uniquely identifiable manufacturer serial number associated with the asset such as VIN</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.SerialNumber</lx:path>
                                    <lx:label>Serial Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="cleaningEquipmentTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific type of Cleaning Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageAtStart" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age of the equipment based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.UsageAtStart</lx:path>
                                    <lx:label>Usage At Start</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageUnitOfMeasurement" type="usageUnitOfMeasurementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measurement to quantify the usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.UsageUnitOfMeasurement</lx:path>
                                    <lx:label>Usage Unit Of Measurement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_GoodToBeUsedAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Location of Goods (address)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.x_GoodToBeUsedAddress</lx:path>
                                    <lx:label>Good To Be Used Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Year" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Year of Manufacture</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.CleaningEquipment.Year</lx:path>
                                    <lx:label>Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ContractDetails" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the purchase/sale price as specified in the Contract of Sale</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails</lx:path>
                                <lx:label>Contract Details</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ContractDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date of the Contract of Sale.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.ContractDate</lx:path>
                                    <lx:label>Contract Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ContractPriceAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the total purchase/sale price as specified in the Contract of Sale, inclusive of any GST component</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.ContractPriceAmount</lx:path>
                                    <lx:label>Contract Price Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DepositAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the total amount that is or was payable as a deposit for the security property being purchased.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.DepositAmount</lx:path>
                                    <lx:label>Deposit Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DepositPaid" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the amount of the purchase deposit that has been paid by the applicants as at the dateType the application is submitted.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.DepositPaid</lx:path>
                                    <lx:label>Deposit Paid</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EstimatedSettlementDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The purchase date of the asset, or the proposed settlement date for an asset being purchased or sold</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.EstimatedSettlementDate</lx:path>
                                    <lx:label>Estimated Settlement Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GSTAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>the GST component of the total contract price amount</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.GSTAmount</lx:path>
                                    <lx:label>GST Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GSTInclusive" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Tick the box is GST is included in purchase price</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.GSTInclusive</lx:path>
                                    <lx:label>GST Inclusive</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GSTOverwritten" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Tick the box if GST is to be overwritten using a modified figure from the calculated one</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.GSTOverwritten</lx:path>
                                    <lx:label>GST Overwritten</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InputTaxCredit" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This is the amount of income tax (GST) credits that are claimed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.InputTaxCredit</lx:path>
                                    <lx:label>Input Tax Credit</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InsurancePremium" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This is the insurance premium amount included on the invoice. For Motor vehicles this relates to CHP insurance only</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.InsurancePremium</lx:path>
                                    <lx:label>Insurance Premium</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LuxuryCarTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>To be completed for Finance Lease only. This applies to cars with a GST inclusive value over the LCT threshold</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.LuxuryCarTax</lx:path>
                                    <lx:label>Luxury Car Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NetTradeIn" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Net Trade In Value.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.NetTradeIn</lx:path>
                                    <lx:label>Net Trade In</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NonClaimableGST" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Non-claimable GST Amount</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.NonClaimableGST</lx:path>
                                    <lx:label>Non Claimable GST</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OnRoadCosts" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This is the on road costs.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.OnRoadCosts</lx:path>
                                    <lx:label>On Road Costs</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherCost" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Other costs being financed such as extended warranty or gap insurance</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.OtherCost</lx:path>
                                    <lx:label>Other Cost</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegistrationCost" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Cost of registration, amount included in the price.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.RegistrationCost</lx:path>
                                    <lx:label>Registration Cost</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StampDuty" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Stamp duty amount, included on the invoice</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.StampDuty</lx:path>
                                    <lx:label>Stamp Duty</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SupplierType" type="supplierTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Type of supplier, in the context of the purchase of the asset from the supplier</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.SupplierType</lx:path>
                                    <lx:label>Supplier Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalCost" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The net of the total purchase price including GST, luxury car tax, stamp duty, registration etc minus deposit, trade-in.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.TotalCost</lx:path>
                                    <lx:label>Total Cost</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TradeInAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Trade-in amount</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.TradeInAmount</lx:path>
                                    <lx:label>Trade In Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TradeInExistingFinance" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Amount of existing finance owing on the trade in, if applicable.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.TradeInExistingFinance</lx:path>
                                    <lx:label>Trade In Existing Finance</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Vendor" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to the Vendor of a purchased Asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ContractDetails.x_Vendor</lx:path>
                                    <lx:label>Vendor</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="EarthMovingMiningAndConstruction" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for Earth Moving Mining and Construction asset</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction</lx:path>
                                <lx:label>Earth Moving Mining And Construction</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AdditionalIDType" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Type of additional ID, such as Microdot, UPC, GTIN, Asset Number</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.AdditionalIDType</lx:path>
                                    <lx:label>Additional ID Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AdditionalIDValue" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For an additional ID, the value of the ID</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.AdditionalIDValue</lx:path>
                                    <lx:label>Additional ID Value</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Age" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age in years. 0=New (alternatively, use year of manufacture)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.Age</lx:path>
                                    <lx:label>Age</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Condition" type="conditionList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Asset condition</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.Condition</lx:path>
                                    <lx:label>Condition</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConditionDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comments on the general condition and appearance of the asset e.g. good working order.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.ConditionDescription</lx:path>
                                    <lx:label>Condition Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the Earth Moving Mining And Construction equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EffectiveLife" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Effective life of the asset in years, as determined by the ATO</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.EffectiveLife</lx:path>
                                    <lx:label>Effective Life</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExcessUsageCalculationMethod" type="excessUsageCalculationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Method to calculate excess usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.ExcessUsageCalculationMethod</lx:path>
                                    <lx:label>Excess Usage Calculation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GoodToBeUsedLocation" type="goodToBeUsedLocationList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>On Business Premises, On-site and returned to Business Premises each day, On-site until job is complete</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.GoodToBeUsedLocation</lx:path>
                                    <lx:label>Good To Be Used Location</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludedUsage" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Allowed usage based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.IncludedUsage</lx:path>
                                    <lx:label>Included Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Make" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The make of the Earth Moving Mining And Construction Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.Make</lx:path>
                                    <lx:label>Make</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Model" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The model of the Earth Moving Mining And Construction Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.Model</lx:path>
                                    <lx:label>Model</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherInformation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Other information which may impact the value of the asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.OtherInformation</lx:path>
                                    <lx:label>Other Information</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Quantity" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Quantity of assets. Typically "1", however for assets that are not individually registrable, indicates the number of items for a bulk grouping of assets being purchased and financed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.Quantity</lx:path>
                                    <lx:label>Quantity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SerialNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The uniquely identifiable manufacturer serial number associated with the asset such as VIN</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.SerialNumber</lx:path>
                                    <lx:label>Serial Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Subtype" type="earthMovingMiningAndConstructionSubtypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific subtype of Earth Moving Mining And Construction.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.Subtype</lx:path>
                                    <lx:label>Subtype</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="earthMovingMiningAndConstructionTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific type of Earth Moving Mining And Construction Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageAtStart" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age of the equipment based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.UsageAtStart</lx:path>
                                    <lx:label>Usage At Start</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageUnitOfMeasurement" type="usageUnitOfMeasurementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measurement to quantify the usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.UsageUnitOfMeasurement</lx:path>
                                    <lx:label>Usage Unit Of Measurement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_GoodToBeUsedAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Location of Goods (address)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.x_GoodToBeUsedAddress</lx:path>
                                    <lx:label>Good To Be Used Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Year" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Year of Manufacture</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EarthMovingMiningAndConstruction.Year</lx:path>
                                    <lx:label>Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Encumbrance" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>An encumbrance is a burden, impairment or third party right over property, that often makes it less valuable or easy for the owner to deal with. There may be multiple encumbrances over a single item of property.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.Encumbrance</lx:path>
                                <lx:label>Encumbrance</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="InFavourOf" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>The party to whom the encumbrance is in favour.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.NonRealEstateAsset.Encumbrance.InFavourOf</lx:path>
                                      <lx:label>In Favour Of</lx:label>
                                      <li:transactions>CAL,DAS</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Capacity" type="encumbranceInFavourOfCapacityList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The capacity of the party to whom the encumbrance is in favour.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.Encumbrance.InFavourOf.Capacity</lx:path>
                                          <lx:label>Capacity</lx:label>
                                          <li:transactions>CAL,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Name" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The name of the party to whom the encumbrance is in favour.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.Encumbrance.InFavourOf.Name</lx:path>
                                          <lx:label>Name</lx:label>
                                          <li:transactions>CAL,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_InFavourOf" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A cross reference to the party to whom the encumbrance is in favour.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.Encumbrance.InFavourOf.x_InFavourOf</lx:path>
                                          <lx:label>In Favour Of</lx:label>
                                          <li:transactions>CAL,DAS</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                            <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the encumbrance that is being specified as applying to the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Encumbrance.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EncumbranceType" type="nonRealEstateAssetEncumbranceList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of encumbrance that is being specified as applying to the real property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Encumbrance.EncumbranceType</lx:path>
                                    <lx:label>Encumbrance Type</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Priority" type="nonRealEstateSecurityPriorityList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The priority of the encumbrance claim against the asset.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Encumbrance.Priority</lx:path>
                                    <lx:label>Priority</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegisteredNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number issued by the relevant Office to identify the encumbrance, if registered.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Encumbrance.RegisteredNumber</lx:path>
                                    <lx:label>Registered Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegistrationDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date the encumbrance was registered with the Titles Office.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Encumbrance.RegistrationDate</lx:path>
                                    <lx:label>Registration Date</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unique Identifier for this Encumbrance Element</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Encumbrance.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="EstimatedValue" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The estimated value of the asset</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.EstimatedValue</lx:path>
                                <lx:label>Estimated Value</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="BalloonRVAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The residual value or final payment for the asset at the end of the term, as a $ amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EstimatedValue.BalloonRVAmount</lx:path>
                                    <lx:label>Balloon RV Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="BalloonRVInputPattern" type="balloonRVInputPatternList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether the residual value for the asset is a percentage or a fixed dollar amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EstimatedValue.BalloonRVInputPattern</lx:path>
                                    <lx:label>Balloon RV Input Pattern</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="BalloonRVPercent" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The residual value or final payment for the asset at the end of the term, as a % amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EstimatedValue.BalloonRVPercent</lx:path>
                                    <lx:label>Balloon RV Percent</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EstimateBasis" type="estimateBasisNonRealEstateAssetList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the estimate was based on the valuation of the applicant or a certified valuer.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EstimatedValue.EstimateBasis</lx:path>
                                    <lx:label>Estimate Basis</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EstimatedCGTLiability" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The estimate of the owners liability to pay capital gains tax on disposal of the asset. Relevant to the lender in determining the net proceeds likely to be realised by the owner if they sell the asset for the estimated value.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EstimatedValue.EstimatedCGTLiability</lx:path>
                                    <lx:label>Estimated CGT Liability</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MinimumResidualValue" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The minimum residual value of the asset, typically based on ATO guidelines. Can be used as part of the depreciation schedule.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EstimatedValue.MinimumResidualValue</lx:path>
                                    <lx:label>Minimum Residual Value</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TaxDepreciationMethod" type="taxDepreciationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether the depreciation schedule is straight line or diminishing value</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EstimatedValue.TaxDepreciationMethod</lx:path>
                                    <lx:label>Tax Depreciation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TaxDepreciationRate" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The depreciation rate of the asset, if known</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EstimatedValue.TaxDepreciationRate</lx:path>
                                    <lx:label>Tax Depreciation Rate</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Value" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The estimated value of the non real estate asset.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.EstimatedValue.Value</lx:path>
                                    <lx:label>Value</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="FinancialAsset" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Captures the details of a financial account asset or security - e.g. Cash, cash equivalents, and accounts at financial institutions</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.FinancialAsset</lx:path>
                                <lx:label>Financial Asset</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="AccountNumber" type="financialAccountType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Account details for the financial account asset</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.NonRealEstateAsset.FinancialAsset.AccountNumber</lx:path>
                                      <lx:label>Account Number</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="Shares" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Container element for financial assets that are shares</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.NonRealEstateAsset.FinancialAsset.Shares</lx:path>
                                      <lx:label>Shares</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="ClassOfShare" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The rights and restrictions attached to shares in a class distinguish it from other classes.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.FinancialAsset.Shares.ClassOfShare</lx:path>
                                          <lx:label>Class Of Share</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="HolderIdentificationNumber" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Holder Identification Number (HIN) uniquely identifies an individual as the holder of shares on the CHESS subregister.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.FinancialAsset.Shares.HolderIdentificationNumber</lx:path>
                                          <lx:label>Holder Identification Number</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IssuingCompany" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The company which has issued the shares.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.FinancialAsset.Shares.IssuingCompany</lx:path>
                                          <lx:label>Issuing Company</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="QuantityOfShares" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Number of shares owned by the individual.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.FinancialAsset.Shares.QuantityOfShares</lx:path>
                                          <lx:label>Quantity Of Shares</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Type" type="shareTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>When the Financial Asset Type = "Shares and Debentures", indicates the type of shares.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.FinancialAsset.Shares.Type</lx:path>
                                          <lx:label>Type</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the account</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.FinancialAsset.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TransferOwnershipToSMSF" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates if any amount is to be transferred to the SMSF.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.FinancialAsset.TransferOwnershipToSMSF</lx:path>
                                    <lx:label>Transfer Ownership To SMSF</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="financialAssetTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of financial asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.FinancialAsset.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="FundsDisbursement" type="fundsDisbursementType" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Specification of how funds are to be distributed.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.FundsDisbursement</lx:path>
                                <lx:label>Funds Disbursement</lx:label>
                                <li:transactions>CAL,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="HospitalityAndLeisure" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for Hospitality and Leisure asset</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure</lx:path>
                                <lx:label>Hospitality And Leisure</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AdditionalIDType" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Type of additional ID, such as Microdot, UPC, GTIN, Asset Number</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.AdditionalIDType</lx:path>
                                    <lx:label>Additional ID Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AdditionalIDValue" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For an additional ID, the value of the ID</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.AdditionalIDValue</lx:path>
                                    <lx:label>Additional ID Value</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Age" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age in years. 0=New (alternatively, use year of manufacture)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.Age</lx:path>
                                    <lx:label>Age</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Condition" type="conditionList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Asset condition</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.Condition</lx:path>
                                    <lx:label>Condition</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConditionDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comments on the general condition and appearance of the asset e.g. good working order.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.ConditionDescription</lx:path>
                                    <lx:label>Condition Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the Hospitality And Leisure Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EffectiveLife" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Effective life of the asset in years, as determined by the ATO</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.EffectiveLife</lx:path>
                                    <lx:label>Effective Life</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExcessUsageCalculationMethod" type="excessUsageCalculationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Method to calculate excess usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.ExcessUsageCalculationMethod</lx:path>
                                    <lx:label>Excess Usage Calculation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GoodToBeUsedLocation" type="goodToBeUsedLocationList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>On Business Premises, On-site and returned to Business Premises each day, On-site until job is complete</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.GoodToBeUsedLocation</lx:path>
                                    <lx:label>Good To Be Used Location</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludedUsage" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Allowed usage based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.IncludedUsage</lx:path>
                                    <lx:label>Included Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Make" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The make of the Hospitality And Leisure Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.Make</lx:path>
                                    <lx:label>Make</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Model" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The model of the Hospitality And Leisure Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.Model</lx:path>
                                    <lx:label>Model</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherInformation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Other information which may impact the value of the asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.OtherInformation</lx:path>
                                    <lx:label>Other Information</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Quantity" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Quantity of assets. Typically "1", however for assets that are not individually registrable, indicates the number of items for a bulk grouping of assets being purchased and financed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.Quantity</lx:path>
                                    <lx:label>Quantity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SerialNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The uniquely identifiable manufacturer serial number associated with the asset such as VIN</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.SerialNumber</lx:path>
                                    <lx:label>Serial Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="hospitalityAndLeisureTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific type of Hospitality And Leisure Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageAtStart" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age of the equipment based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.UsageAtStart</lx:path>
                                    <lx:label>Usage At Start</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageUnitOfMeasurement" type="usageUnitOfMeasurementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measurement to quantify the usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.UsageUnitOfMeasurement</lx:path>
                                    <lx:label>Usage Unit Of Measurement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_GoodToBeUsedAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Location of Goods (address)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.x_GoodToBeUsedAddress</lx:path>
                                    <lx:label>Good To Be Used Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Year" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Year of Manufacture</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.HospitalityAndLeisure.Year</lx:path>
                                    <lx:label>Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Insurance" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for insurance information.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.Insurance</lx:path>
                                <lx:label>Insurance</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Insurance" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to Insurance details</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Insurance.x_Insurance</lx:path>
                                    <lx:label>Insurance</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Insurance</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ITAndAVEquipment" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for IT and AV Equipment asset</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment</lx:path>
                                <lx:label>IT And AV Equipment</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AdditionalIDType" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Type of additional ID, such as Microdot, UPC, GTIN, Asset Number</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.AdditionalIDType</lx:path>
                                    <lx:label>Additional ID Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AdditionalIDValue" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For an additional ID, the value of the ID</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.AdditionalIDValue</lx:path>
                                    <lx:label>Additional ID Value</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Age" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age in years. 0=New (alternatively, use year of manufacture)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.Age</lx:path>
                                    <lx:label>Age</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Condition" type="conditionList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Asset condition</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.Condition</lx:path>
                                    <lx:label>Condition</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConditionDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comments on the general condition and appearance of the asset e.g. good working order.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.ConditionDescription</lx:path>
                                    <lx:label>Condition Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the IT and AV Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EffectiveLife" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Effective life of the asset in years, as determined by the ATO</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.EffectiveLife</lx:path>
                                    <lx:label>Effective Life</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExcessUsageCalculationMethod" type="excessUsageCalculationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Method to calculate excess usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.ExcessUsageCalculationMethod</lx:path>
                                    <lx:label>Excess Usage Calculation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GoodToBeUsedLocation" type="goodToBeUsedLocationList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>On Business Premises, On-site and returned to Business Premises each day, On-site until job is complete</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.GoodToBeUsedLocation</lx:path>
                                    <lx:label>Good To Be Used Location</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludedUsage" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Allowed usage based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.IncludedUsage</lx:path>
                                    <lx:label>Included Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Make" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The make of the IT and AV Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.Make</lx:path>
                                    <lx:label>Make</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Model" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The model of the IT and AV Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.Model</lx:path>
                                    <lx:label>Model</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherInformation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Other information which may impact the value of the asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.OtherInformation</lx:path>
                                    <lx:label>Other Information</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Quantity" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Quantity of assets. Typically "1", however for assets that are not individually registrable, indicates the number of items for a bulk grouping of assets being purchased and financed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.Quantity</lx:path>
                                    <lx:label>Quantity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SerialNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The uniquely identifiable manufacturer serial number associated with the asset such as VIN</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.SerialNumber</lx:path>
                                    <lx:label>Serial Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="itAndAVEquipmentTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific type of IT and AV Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageAtStart" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age of the equipment based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.UsageAtStart</lx:path>
                                    <lx:label>Usage At Start</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageUnitOfMeasurement" type="usageUnitOfMeasurementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measurement to quantify the usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.UsageUnitOfMeasurement</lx:path>
                                    <lx:label>Usage Unit Of Measurement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_GoodToBeUsedAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Location of Goods (address)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.x_GoodToBeUsedAddress</lx:path>
                                    <lx:label>Good To Be Used Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Year" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Year of Manufacture</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ITAndAVEquipment.Year</lx:path>
                                    <lx:label>Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="MaterialsHandlingAndLifting" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for Materials Handling and Lifting asset</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting</lx:path>
                                <lx:label>Materials Handling And Lifting</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AdditionalIDType" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Type of additional ID, such as Microdot, UPC, GTIN, Asset Number</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.AdditionalIDType</lx:path>
                                    <lx:label>Additional ID Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AdditionalIDValue" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For an additional ID, the value of the ID</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.AdditionalIDValue</lx:path>
                                    <lx:label>Additional ID Value</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Age" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age in years. 0=New (alternatively, use year of manufacture)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.Age</lx:path>
                                    <lx:label>Age</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Condition" type="conditionList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Asset condition</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.Condition</lx:path>
                                    <lx:label>Condition</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConditionDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comments on the general condition and appearance of the asset e.g. good working order.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.ConditionDescription</lx:path>
                                    <lx:label>Condition Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the Materials Handling and Lifting Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EffectiveLife" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Effective life of the asset in years, as determined by the ATO</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.EffectiveLife</lx:path>
                                    <lx:label>Effective Life</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExcessUsageCalculationMethod" type="excessUsageCalculationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Method to calculate excess usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.ExcessUsageCalculationMethod</lx:path>
                                    <lx:label>Excess Usage Calculation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GoodToBeUsedLocation" type="goodToBeUsedLocationList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>On Business Premises, On-site and returned to Business Premises each day, On-site until job is complete</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.GoodToBeUsedLocation</lx:path>
                                    <lx:label>Good To Be Used Location</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludedUsage" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Allowed usage based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.IncludedUsage</lx:path>
                                    <lx:label>Included Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Make" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The make of the Materials Handling and Lifting Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.Make</lx:path>
                                    <lx:label>Make</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Model" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The model of the Materials Handling and Lifting Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.Model</lx:path>
                                    <lx:label>Model</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherInformation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Other information which may impact the value of the asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.OtherInformation</lx:path>
                                    <lx:label>Other Information</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Quantity" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Quantity of assets. Typically "1", however for assets that are not individually registrable, indicates the number of items for a bulk grouping of assets being purchased and financed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.Quantity</lx:path>
                                    <lx:label>Quantity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SerialNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The uniquely identifiable manufacturer serial number associated with the asset such as VIN</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.SerialNumber</lx:path>
                                    <lx:label>Serial Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Subtype" type="materialsHandlingAndLiftingSubtypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific subtype of Materials Handling and Lifting Equipment.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.Subtype</lx:path>
                                    <lx:label>Subtype</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="materialsHandlingAndLiftingTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific type of Materials Handling and Lifting Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageAtStart" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age of the equipment based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.UsageAtStart</lx:path>
                                    <lx:label>Usage At Start</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageUnitOfMeasurement" type="usageUnitOfMeasurementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measurement to quantify the usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.UsageUnitOfMeasurement</lx:path>
                                    <lx:label>Usage Unit Of Measurement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_GoodToBeUsedAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Location of Goods (address)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.x_GoodToBeUsedAddress</lx:path>
                                    <lx:label>Good To Be Used Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Year" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Year of Manufacture</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MaterialsHandlingAndLifting.Year</lx:path>
                                    <lx:label>Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="MedicalEquipment" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for Medical Equipment asset</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.MedicalEquipment</lx:path>
                                <lx:label>Medical Equipment</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Condition" type="conditionList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The condition of the medical equipment.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MedicalEquipment.Condition</lx:path>
                                    <lx:label>Condition</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the Medical Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MedicalEquipment.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExcessUsageCalculationMethod" type="excessUsageCalculationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Method to calculate excess usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MedicalEquipment.ExcessUsageCalculationMethod</lx:path>
                                    <lx:label>Excess Usage Calculation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludedUsage" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Allowed usage based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MedicalEquipment.IncludedUsage</lx:path>
                                    <lx:label>Included Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Make" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The make of the medical equipment.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MedicalEquipment.Make</lx:path>
                                    <lx:label>Make</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Model" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The model of the medical equipment.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MedicalEquipment.Model</lx:path>
                                    <lx:label>Model</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="medicalEquipmentTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific type of Medical Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MedicalEquipment.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageAtStart" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age of the equipment based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MedicalEquipment.UsageAtStart</lx:path>
                                    <lx:label>Usage At Start</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageUnitOfMeasurement" type="usageUnitOfMeasurementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measurement to quantify the usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MedicalEquipment.UsageUnitOfMeasurement</lx:path>
                                    <lx:label>Usage Unit Of Measurement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="MobileComputing" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for Mobile Computing asset</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.MobileComputing</lx:path>
                                <lx:label>Mobile Computing</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Condition" type="conditionList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The condition of the mobile computing device.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MobileComputing.Condition</lx:path>
                                    <lx:label>Condition</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the Mobile Computing device</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MobileComputing.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExcessUsageCalculationMethod" type="excessUsageCalculationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Method to calculate excess usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MobileComputing.ExcessUsageCalculationMethod</lx:path>
                                    <lx:label>Excess Usage Calculation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludedUsage" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Allowed usage based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MobileComputing.IncludedUsage</lx:path>
                                    <lx:label>Included Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Make" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The make of the mobile computing device.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MobileComputing.Make</lx:path>
                                    <lx:label>Make</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Model" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The model of the mobile computing device.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MobileComputing.Model</lx:path>
                                    <lx:label>Model</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="mobileComputingTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific type of Mobile Computing device</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MobileComputing.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageAtStart" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age of the equipment based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MobileComputing.UsageAtStart</lx:path>
                                    <lx:label>Usage At Start</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageUnitOfMeasurement" type="usageUnitOfMeasurementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measurement to quantify the usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MobileComputing.UsageUnitOfMeasurement</lx:path>
                                    <lx:label>Usage Unit Of Measurement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="MotorVehicle" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about an asset that is a motor vehicle.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle</lx:path>
                                <lx:label>Motor Vehicle</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AdditionalIDType" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Type of additional ID, such as Microdot, UPC, GTIN, Asset Number</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.AdditionalIDType</lx:path>
                                    <lx:label>Additional ID Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AdditionalIDValue" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For an additional ID, the value of the ID</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.AdditionalIDValue</lx:path>
                                    <lx:label>Additional ID Value</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Age" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age in years. 0=New (alternatively, use year of manufacture)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Age</lx:path>
                                    <lx:label>Age</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Badge" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Sub-model designation of a motor vehicle, e.g. Sportivo</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Badge</lx:path>
                                    <lx:label>Badge</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Body" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Vehicle body type, e.g. Sedan</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Body</lx:path>
                                    <lx:label>Body</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Colour" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Vehicle colour</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Colour</lx:path>
                                    <lx:label>Colour</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Condition" type="conditionList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Asset condition</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Condition</lx:path>
                                    <lx:label>Condition</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConditionDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comments on the general condition and appearance of the asset e.g. good working order.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.ConditionDescription</lx:path>
                                    <lx:label>Condition Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Cylinders" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Number of cylinders in the engine</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Cylinders</lx:path>
                                    <lx:label>Cylinders</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the motor vehicle</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Doors" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Number of doors</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Doors</lx:path>
                                    <lx:label>Doors</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EffectiveLife" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Effective life of the asset in years, as determined by the ATO</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.EffectiveLife</lx:path>
                                    <lx:label>Effective Life</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EngineCapacity" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The displacement of the engine, generally expressed in CCs</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.EngineCapacity</lx:path>
                                    <lx:label>Engine Capacity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EngineHoursTotal" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Total engine hours</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.EngineHoursTotal</lx:path>
                                    <lx:label>Engine Hours Total</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EngineID" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Identification number of the engine</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.EngineID</lx:path>
                                    <lx:label>Engine ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExcessUsageCalculationMethod" type="excessUsageCalculationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Method to calculate excess usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.ExcessUsageCalculationMethod</lx:path>
                                    <lx:label>Excess Usage Calculation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FuelType" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of fuel used to power the vehicle</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.FuelType</lx:path>
                                    <lx:label>Fuel Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludedUsage" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Allowed usage based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.IncludedUsage</lx:path>
                                    <lx:label>Included Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Kilometres" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The odometer reading for a vehicle, in kms</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Kilometres</lx:path>
                                    <lx:label>Kilometres</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Make" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the make of the vehicle, such as Holden</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Make</lx:path>
                                    <lx:label>Make</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Model" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the model of the vehicle</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Model</lx:path>
                                    <lx:label>Model</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Options" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Description of optional items included in the asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Options</lx:path>
                                    <lx:label>Options</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherInformation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Other information which may impact the value of the asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.OtherInformation</lx:path>
                                    <lx:label>Other Information</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Quantity" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Quantity of assets. Typically "1", however for assets that are not individually registrable, indicates the number of items for a bulk grouping of assets being purchased and financed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Quantity</lx:path>
                                    <lx:label>Quantity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegisteredInState" type="auStateList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the Australian state where the vehicle is registered.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.RegisteredInState</lx:path>
                                    <lx:label>Registered In State</lx:label>
                                    <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegistrationExpiryDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Date of expiry of registration.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.RegistrationExpiryDate</lx:path>
                                    <lx:label>Registration Expiry Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegistrationNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Registration identifier for the asset e.g. license plate number.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.RegistrationNumber</lx:path>
                                    <lx:label>Registration Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SerialNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The uniquely identifiable manufacturer serial number associated with the asset such as VIN.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.SerialNumber</lx:path>
                                    <lx:label>Serial Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Series" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For a vehicle model such as a car that is manufactured with a series indicator, a description of the series. E.g. Series II</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Series</lx:path>
                                    <lx:label>Series</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Subtype" type="vehicleSubtypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific subtype of Motor Vehicle.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Subtype</lx:path>
                                    <lx:label>Subtype</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Transmission" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For a motor vehicle, the type of transmission</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Transmission</lx:path>
                                    <lx:label>Transmission</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="vehicleTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Vehicle type describes the type of vehicle such as a bike, large vehicle, 4WD, medium vehicle, small vehicle, or small/medium vehicle.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageAtStart" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age of the equipment based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.UsageAtStart</lx:path>
                                    <lx:label>Usage At Start</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageUnitOfMeasurement" type="usageUnitOfMeasurementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measurement to quantify the usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.UsageUnitOfMeasurement</lx:path>
                                    <lx:label>Usage Unit Of Measurement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Variant" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For a motor vehicle model that is produced in variants, a description of the variant</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Variant</lx:path>
                                    <lx:label>Variant</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_GoodToBeUsedAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Location of Goods (address)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.x_GoodToBeUsedAddress</lx:path>
                                    <lx:label>Good To Be Used Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Year" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the year in which the vehicle was first registered or the model year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.MotorVehicle.Year</lx:path>
                                    <lx:label>Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="OfficeEquipment" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for Office Equipment asset</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment</lx:path>
                                <lx:label>Office Equipment</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AdditionalIDType" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Type of additional ID, such as Microdot, UPC, GTIN, Asset Number</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.AdditionalIDType</lx:path>
                                    <lx:label>Additional ID Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AdditionalIDValue" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For an additional ID, the value of the ID</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.AdditionalIDValue</lx:path>
                                    <lx:label>Additional ID Value</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Age" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age in years. 0=New (alternatively, use year of manufacture)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.Age</lx:path>
                                    <lx:label>Age</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Condition" type="conditionList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Asset condition</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.Condition</lx:path>
                                    <lx:label>Condition</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConditionDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comments on the general condition and appearance of the asset e.g. good working order.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.ConditionDescription</lx:path>
                                    <lx:label>Condition Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the Office Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EffectiveLife" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Effective life of the asset in years, as determined by the ATO</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.EffectiveLife</lx:path>
                                    <lx:label>Effective Life</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExcessUsageCalculationMethod" type="excessUsageCalculationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Method to calculate excess usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.ExcessUsageCalculationMethod</lx:path>
                                    <lx:label>Excess Usage Calculation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GoodToBeUsedLocation" type="goodToBeUsedLocationList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>On Business Premises, On-site and returned to Business Premises each day, On-site until job is complete</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.GoodToBeUsedLocation</lx:path>
                                    <lx:label>Good To Be Used Location</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludedUsage" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Allowed usage based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.IncludedUsage</lx:path>
                                    <lx:label>Included Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Make" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The make of the Office Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.Make</lx:path>
                                    <lx:label>Make</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Model" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The model of the Office Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.Model</lx:path>
                                    <lx:label>Model</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherInformation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Other information which may impact the value of the asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.OtherInformation</lx:path>
                                    <lx:label>Other Information</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Quantity" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Quantity of assets. Typically "1", however for assets that are not individually registrable, indicates the number of items for a bulk grouping of assets being purchased and financed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.Quantity</lx:path>
                                    <lx:label>Quantity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SerialNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The uniquely identifiable manufacturer serial number associated with the asset such as VIN</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.SerialNumber</lx:path>
                                    <lx:label>Serial Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="officeEquipmentTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific type of Office Equipment</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageAtStart" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age of the equipment based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.UsageAtStart</lx:path>
                                    <lx:label>Usage At Start</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageUnitOfMeasurement" type="usageUnitOfMeasurementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measurement to quantify the usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.UsageUnitOfMeasurement</lx:path>
                                    <lx:label>Usage Unit Of Measurement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_GoodToBeUsedAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Location of Goods (address)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.x_GoodToBeUsedAddress</lx:path>
                                    <lx:label>Good To Be Used Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Year" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Year of Manufacture</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OfficeEquipment.Year</lx:path>
                                    <lx:label>Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="OtherAsset" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for other assets.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.OtherAsset</lx:path>
                                <lx:label>Other Asset</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Description of the asset, especially if Other Asset type is "Other"</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OtherAsset.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="otherAssetTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Container element for other non real estate assets</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.OtherAsset.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PercentOwned" type="percentOwnedType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>An asset may be owned by one or more parties, typically the applicants, so there may be more than one owner referenced. An asset may also be partially owned by a non-applicant, such as an applicant's spouse, in which case the third party owner may be referenced, or less than 100% ownership of the asset may be assigned. If the asset is to be purchased, then this indicates the expected ownership arrangement post-settlement</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.PercentOwned</lx:path>
                                <lx:label>Percent Owned</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="PlantEquipmentAndIndustrial" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for Plant Equipment and Industrial asset</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial</lx:path>
                                <lx:label>Plant Equipment And Industrial</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AdditionalIDType" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Type of additional ID, such as Microdot, UPC, GTIN, Asset Number</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.AdditionalIDType</lx:path>
                                    <lx:label>Additional ID Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AdditionalIDValue" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For an additional ID, the value of the ID</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.AdditionalIDValue</lx:path>
                                    <lx:label>Additional ID Value</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Age" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age in years. 0=New (alternatively, use year of manufacture)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.Age</lx:path>
                                    <lx:label>Age</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Condition" type="conditionList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Asset condition</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.Condition</lx:path>
                                    <lx:label>Condition</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConditionDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comments on the general condition and appearance of the asset e.g. good working order.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.ConditionDescription</lx:path>
                                    <lx:label>Condition Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the Plant Equipment and Industrial</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EffectiveLife" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Effective life of the asset in years, as determined by the ATO</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.EffectiveLife</lx:path>
                                    <lx:label>Effective Life</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EngineHoursTotal" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Total engine hours</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.EngineHoursTotal</lx:path>
                                    <lx:label>Engine Hours Total</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExcessUsageCalculationMethod" type="excessUsageCalculationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Method to calculate excess usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.ExcessUsageCalculationMethod</lx:path>
                                    <lx:label>Excess Usage Calculation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GoodToBeUsedLocation" type="goodToBeUsedLocationList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>On Business Premises, On-site and returned to Business Premises each day, On-site until job is complete</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.GoodToBeUsedLocation</lx:path>
                                    <lx:label>Good To Be Used Location</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludedUsage" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Allowed usage based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.IncludedUsage</lx:path>
                                    <lx:label>Included Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Make" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The make of the Plant Equipment and Industrial</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.Make</lx:path>
                                    <lx:label>Make</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Model" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The model of the Plant Equipment and Industrial</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.Model</lx:path>
                                    <lx:label>Model</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherInformation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Other information which may impact the value of the asset</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.OtherInformation</lx:path>
                                    <lx:label>Other Information</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Quantity" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Quantity of assets. Typically "1", however for assets that are not individually registrable, indicates the number of items for a bulk grouping of assets being purchased and financed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.Quantity</lx:path>
                                    <lx:label>Quantity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SerialNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The uniquely identifiable manufacturer serial number associated with the asset such as VIN</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.SerialNumber</lx:path>
                                    <lx:label>Serial Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ServicingHistory" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comments on any service history that will impact on the value of the asset.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.ServicingHistory</lx:path>
                                    <lx:label>Servicing History</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="plantEquipmentAndIndustrialTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific type of Plant Equipment and Industrial</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageAtStart" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age of the equipment based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.UsageAtStart</lx:path>
                                    <lx:label>Usage At Start</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageUnitOfMeasurement" type="usageUnitOfMeasurementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measurement to quantify the usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.UsageUnitOfMeasurement</lx:path>
                                    <lx:label>Usage Unit Of Measurement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_GoodToBeUsedAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Location of Goods (address)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.x_GoodToBeUsedAddress</lx:path>
                                    <lx:label>Good To Be Used Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Year" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Year of Manufacture</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PlantEquipmentAndIndustrial.Year</lx:path>
                                    <lx:label>Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PPSR" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about the asset in relation to its registration with the personal property security register</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR</lx:path>
                                <lx:label>PPSR</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Registration" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>A registration of a security interests on the PPSR.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.Registration</lx:path>
                                      <lx:label>Registration</lx:label>
                                      <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="RegistrationEvent" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>A transactions relating to a PPSR registration. Registration event is either Creation, Ammendment, or Discharge.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.Registration.RegistrationEvent</lx:path>
                                            <lx:label>Registration Event</lx:label>
                                            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="ChangeDateTime" type="dateTimeType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The date time that the PPSR registration event occurred. A registration change date time is provided only on ammendment events.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.Registration.RegistrationEvent.ChangeDateTime</lx:path>
                                                <lx:label>Change Date Time</lx:label>
                                                <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ChangeNumber" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>A unique sequential number that is allocated for certain transactions relating to a PPSR registration. A change number is provided on creation, ammendment, and discharge events.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.Registration.RegistrationEvent.ChangeNumber</lx:path>
                                                <lx:label>Change Number</lx:label>
                                                <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="PPSRTransactionID" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>A unique identifier of the transaction on the PPSR.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.Registration.RegistrationEvent.PPSRTransactionID</lx:path>
                                                <lx:label>PPSR Transaction ID</lx:label>
                                                <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Type" type="registrationEventTypeList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The type of PPSR registration event.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.Registration.RegistrationEvent.Type</lx:path>
                                                <lx:label>Type</lx:label>
                                                <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="GivingOfNoticeIdentifier" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>An optional identifier that a secured party can specify for the PPSR registration.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.Registration.GivingOfNoticeIdentifier</lx:path>
                                          <lx:label>Giving Of Notice Identifier</lx:label>
                                          <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PurchaseMoneySecurityInterest" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The PPSR registration will be a purchase money security interest (PMSI).</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.Registration.PurchaseMoneySecurityInterest</lx:path>
                                          <lx:label>Purchase Money Security Interest</lx:label>
                                          <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RegistrationEndDateTime" type="dateTimeType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date and time that the the PPSR registration will expire. A registration end date time is provided on creation, ammendment, and discharge events.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.Registration.RegistrationEndDateTime</lx:path>
                                          <lx:label>Registration End Date Time</lx:label>
                                          <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RegistrationNumber" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>An identifier assigned to a registration of collateral on the PPSR.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.Registration.RegistrationNumber</lx:path>
                                          <lx:label>Registration Number</lx:label>
                                          <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RegistrationStartDateTime" type="dateTimeType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date and time the registration was first registered on the PPSR. A registration start date time is provided on the creation event.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.Registration.RegistrationStartDateTime</lx:path>
                                          <lx:label>Registration Start Date Time</lx:label>
                                          <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RegistrationToken" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A token that authorises a user to amend or discharge a registration on the PPSR.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.Registration.RegistrationToken</lx:path>
                                          <lx:label>Registration Token</lx:label>
                                          <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SecuredPartyGroupNumber" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The secured party group number used to identify the secured party group (SPG).</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.Registration.SecuredPartyGroupNumber</lx:path>
                                          <lx:label>Secured Party Group Number</lx:label>
                                          <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="CollateralClass" type="collateralClassList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The collateral class type. A general description of the property covered by the the PPSR registration.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.CollateralClass</lx:path>
                                    <lx:label>Collateral Class</lx:label>
                                    <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CollateralType" type="collateralTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Type of collateral as defined by the PPS, Consumer means personal property held by an individual and is not used to carry on an enterprise that has an ABN. Commercial is personal property other than consumer.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.CollateralType</lx:path>
                                    <lx:label>Collateral Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DescriptionOfProceeds" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This item is now deprecated and will be deleted in a future release. PPSR Description of proceeds to be claimed.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.DescriptionOfProceeds</lx:path>
                                    <lx:label>Description Of Proceeds</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PMSI" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This item is now deprecated and will be deleted in a future release. Flag to indicate whether the PPSR registration will be a Purchase Money Security Interest. Select Yes where the lender is providing finance for the acquisition of the asset being offered as security.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.PMSI</lx:path>
                                    <lx:label>PMSI</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PPSRegistrable" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether the asset is registrable on the PPSR</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.PPSRegistrable</lx:path>
                                    <lx:label>PPS Registrable</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProceedsToBeClaimed" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This item is now deprecated and will be deleted in a future release. PPSR flag to indicate if proceeds are to be claimed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.ProceedsToBeClaimed</lx:path>
                                    <lx:label>Proceeds To Be Claimed</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SubjectToControl" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This item is now deprecated and will be deleted in a future release. PPSR flag to indicate whether assets are subject to control</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.SubjectToControl</lx:path>
                                    <lx:label>Subject To Control</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SubordinateRegistration" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This item is now deprecated and will be deleted in a future release. Flag to indicate whether PPSR registration subordinate to another security interest</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.SubordinateRegistration</lx:path>
                                    <lx:label>Subordinate Registration</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SubordinateToNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This item is now deprecated and will be deleted in a future release. PPSR Registration number that this registration is subordinate to</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.PPSR.SubordinateToNumber</lx:path>
                                    <lx:label>Subordinate To Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ToolsOfTrade" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for Tools of Trade asset</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.ToolsOfTrade</lx:path>
                                <lx:label>Tools Of Trade</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Condition" type="conditionList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The condition of the tools of trade.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ToolsOfTrade.Condition</lx:path>
                                    <lx:label>Condition</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the Tools of Trade</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ToolsOfTrade.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExcessUsageCalculationMethod" type="excessUsageCalculationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Method to calculate excess usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ToolsOfTrade.ExcessUsageCalculationMethod</lx:path>
                                    <lx:label>Excess Usage Calculation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludedUsage" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Allowed usage based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ToolsOfTrade.IncludedUsage</lx:path>
                                    <lx:label>Included Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Make" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The make of the tools of trade.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ToolsOfTrade.Make</lx:path>
                                    <lx:label>Make</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Model" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The model of the tools of trade.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ToolsOfTrade.Model</lx:path>
                                    <lx:label>Model</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="toolsOfTradeTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The specific type of Tools of Trade</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ToolsOfTrade.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageAtStart" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age of the equipment based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ToolsOfTrade.UsageAtStart</lx:path>
                                    <lx:label>Usage At Start</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageUnitOfMeasurement" type="usageUnitOfMeasurementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measurement to quantify the usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.ToolsOfTrade.UsageUnitOfMeasurement</lx:path>
                                    <lx:label>Usage Unit Of Measurement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Watercraft" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A boat or other vessel that travels on water.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.NonRealEstateAsset.Watercraft</lx:path>
                                <lx:label>Watercraft</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="OutboardMotor" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>An engine that has a propeller and a manufacturer's number, designed to be attached to a boat or vessel and intended for use to propel a boat or vessel.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.NonRealEstateAsset.Watercraft.OutboardMotor</lx:path>
                                      <lx:label>Outboard Motor</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="OutboardMotorManufacturerNumber" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The identifying numbers/letters that are stamped or attached to an outboard motor by its manufacturer.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.NonRealEstateAsset.Watercraft.OutboardMotor.OutboardMotorManufacturerNumber</lx:path>
                                          <lx:label>Outboard Motor Manufacturer Number</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="ExcessUsageCalculationMethod" type="excessUsageCalculationMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Method to calculate excess usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Watercraft.ExcessUsageCalculationMethod</lx:path>
                                    <lx:label>Excess Usage Calculation Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="HullIdentificationNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A Hull Identification Number (HIN) consist of a unique letters and numbers permanently marked on the hull of the vessel to positively identify a vessel. Boat manufacturers have been required to mark all new vessels with a HIN meeting the required international standard (ISO 10087:1995(E)) since February 2001.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Watercraft.HullIdentificationNumber</lx:path>
                                    <lx:label>Hull Identification Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncludedUsage" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Allowed usage based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Watercraft.IncludedUsage</lx:path>
                                    <lx:label>Included Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageAtStart" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Age of the equipment based on the applicable unit of measurement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Watercraft.UsageAtStart</lx:path>
                                    <lx:label>Usage At Start</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UsageUnitOfMeasurement" type="usageUnitOfMeasurementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of measurement to quantify the usage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Watercraft.UsageUnitOfMeasurement</lx:path>
                                    <lx:label>Usage Unit Of Measurement</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="WatercraftOfficialNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A number (e.g. 345678) allocated to the ship or watercraft by the Shipping Registration Office to registered ships via Maritime Safety Authority (MSA).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.NonRealEstateAsset.Watercraft.WatercraftOfficialNumber</lx:path>
                                    <lx:label>Watercraft Official Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="AmountToBeReduced" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The dollar amount by which the balance of this asset will be reduced prior to or at settlement.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.AmountToBeReduced</lx:path>
                              <lx:label>Amount To Be Reduced</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ContractOfSale" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that the Borrower has entered into a Contract Of Sale</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.ContractOfSale</lx:path>
                              <lx:label>Contract Of Sale</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Encumbered" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether there is an encumbrance on the Non-Real Estate Asset.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.Encumbered</lx:path>
                              <lx:label>Encumbered</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="FinancialTransactionType" type="financialTransactionTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Defines the type of Equipment Finance transaction.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.FinancialTransactionType</lx:path>
                              <lx:label>Financial Transaction Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LenderAssessmentReason" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A free text description of the reason the asset record requires assessment by the lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.LenderAssessmentReason</lx:path>
                              <lx:label>Lender Assessment Reason</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LenderAssessmentRequired" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>A flag to indicate the asset record requires assessment by the lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.LenderAssessmentRequired</lx:path>
                              <lx:label>Lender Assessment Required</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PrimarySecurity" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the non-real estate asset is being offered as the primary security for the loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.PrimarySecurity</lx:path>
                              <lx:label>Primary Security</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this non-real estate asset within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ToBeReduced" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The balance of this asset will be reduced prior to or at settlement. For example it may be used to pay off existing loans, as contributions for a new loan, or used for a planned purchase.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.ToBeReduced</lx:path>
                              <lx:label>To Be Reduced</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ToBeSold" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the asset will be sold prior to or at settlement</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.ToBeSold</lx:path>
                              <lx:label>To Be Sold</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ToBeUsedAsSecurity" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the asset is to be used as security for the new or refinanced loan(s)</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.ToBeUsedAsSecurity</lx:path>
                              <lx:label>To Be Used As Security</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Transaction" type="assetTransactionList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates the current status of the applicants' ownership of the property, or the transaction that is proposed by them in relation to the property.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.Transaction</lx:path>
                              <lx:label>Transaction</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Type" type="nonRealEstateAssetTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates what the type of non real estate asset being described.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Unique identification assigned to the Asset</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Verified" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>A flag to indicate that this asset data has been verified.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.Verified</lx:path>
                              <lx:label>Verified</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_CustomerTransactionAnalysis" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A cross reference to the customer transaction analysis element from which the asset data was verified.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.x_CustomerTransactionAnalysis</lx:path>
                              <lx:label>Customer Transaction Analysis</lx:label>
                              <li:transactions>CAL</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.CustomerTransactionAnalysis</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_VendorTaxInvoice" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Link to Vendor Tax Invoice details.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.NonRealEstateAsset.x_VendorTaxInvoice</lx:path>
                              <lx:label>Vendor Tax Invoice</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.VendorTaxInvoice</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="OtherExpense" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. An expense item is an outflow of money to another person or group to pay for an item or service, or for a category of costs.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.OtherExpense</lx:path>
                          <lx:label>Other Expense</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Arrears" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Container element for capturing Arrears detail</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.OtherExpense.Arrears</lx:path>
                                <lx:label>Arrears</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Amount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Captures the amount in arrears</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.OtherExpense.Arrears.Amount</lx:path>
                                    <lx:label>Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfMissedPayments" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Captures the number of missed payments</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.OtherExpense.Arrears.NumberOfMissedPayments</lx:path>
                                    <lx:label>Number Of Missed Payments</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ToBePaidOut" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Captures whether the amount in arrears is to be paid out from the Loan</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.OtherExpense.Arrears.ToBePaidOut</lx:path>
                                    <lx:label>To Be Paid Out</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PercentOwned" type="percentOwnedType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Identifies the 'owners' of the expense item and the amount they are responsible for. An expense may be owned by one or more parties, typically the applicants, so there may be more than one owner referenced. An expense may also be partially owned by a non-applicant, such as an applicant's spouse, in which case the third party owner may be referenced, or less than 100% ownership of the expense may be assigned.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.OtherExpense.PercentOwned</lx:path>
                                <lx:label>Percent Owned</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="Amount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. The amount of the expense paid or incurred in the period specified for the expense item.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherExpense.Amount</lx:path>
                              <lx:label>Amount</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Category" type="otherExpenseCategoryList" use="optional">
                          <xs:annotation>
                            <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Specifies the category of expense that is being described</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherExpense.Category</lx:path>
                              <lx:label>Category</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Description" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Captures a plain text description of the expense when the selected Expense Type = "Other"</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherExpense.Description</lx:path>
                              <lx:label>Description</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="EndDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Captures the end date of the period for which the expense being described is paid</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherExpense.EndDate</lx:path>
                              <lx:label>End Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                          <xs:annotation>
                            <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. The frequency or payment period for the expense paid or incurred in the period specified for the expense item.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherExpense.Frequency</lx:path>
                              <lx:label>Frequency</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="HasArrears" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Used to indicate if the applicant has declared an arrears Credit Issue</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherExpense.HasArrears</lx:path>
                              <lx:label>Has Arrears</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this other expense within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherExpense.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="StartDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Captures the start date of the period for which the expense being described is paid</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherExpense.StartDate</lx:path>
                              <lx:label>Start Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Type" type="otherExpenseTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Specifies the type of expense that is being described.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherExpense.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherExpense.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="OtherIncome" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Captures information about income earned by the applicants that is not captured under the Employment element for Person Applicants or Income elements for Company and Trust Applicants</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.OtherIncome</lx:path>
                          <lx:label>Other Income</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="PercentOwned" type="percentOwnedType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>An income item may be owned by one or more parties, typically the applicants, so there may be more than one owner referenced. An income item may also be partially owned by a non-applicant, such as an applicant's spouse, in which case the third party owner may be referenced, or less than 100% ownership of the income may be assigned.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.OtherIncome.PercentOwned</lx:path>
                                <lx:label>Percent Owned</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="Amount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The gross amount of income received in the period specified for the income item.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.Amount</lx:path>
                              <lx:label>Gross Amount</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BenefitsDescription" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures a free text description of the benefits, especially if the benefit type is "other"</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.BenefitsDescription</lx:path>
                              <lx:label>Benefits Description</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Country" type="countryCodeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the country in which the income is earned or sourced.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.Country</lx:path>
                              <lx:label>Country</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Description" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures a free text description of the income when the selected Income Type = "Other"</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.Description</lx:path>
                              <lx:label>Description</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="EndDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the end date of the period for which the income being described is earned or received.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.EndDate</lx:path>
                              <lx:label>End Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The frequency at which the applicant is receiving other income gross amount.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.Frequency</lx:path>
                              <lx:label>Gross Amount Frequency</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="GovernmentBenefitsType" type="governmentBenefitsTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Where the applicant has indicated government benefits as an income source, the type of benefits should be captured. Such as family, unemployed or widow's allowance.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.GovernmentBenefitsType</lx:path>
                              <lx:label>Government Benefits Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="GSTAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The dollar amount of the GST incorporated into the regular income.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.GSTAmount</lx:path>
                              <lx:label>GST Amount</lx:label>
                              <li:transactions>CAL,CNZ,CDA,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IncomeStatusOnOrBeforeSettlement" type="incomeStatusOnOrBeforeSettlementList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The income status on or before settlement.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.IncomeStatusOnOrBeforeSettlement</lx:path>
                              <lx:label>Income Status On Or Before Settlement</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IncomeVerification" type="incomeVerificationList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Specifies how the applicant intends to substantiate the income item.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.IncomeVerification</lx:path>
                              <lx:label>Income Verification</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsTaxable" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the income is subject to tax, and not one of the following categories: exempt income; non-assessable non-exempt income; or other non taxable amounts.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.IsTaxable</lx:path>
                              <lx:label>Is Taxable</lx:label>
                              <li:transactions>CAL,CNZ,CDA,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NetAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The net amount of income received in the period specified for the income item.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.NetAmount</lx:path>
                              <lx:label>Net Amount</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NetAmountFrequency" type="frequencyShortList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The frequency at which the applicant is receiving other income net amount.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.NetAmountFrequency</lx:path>
                              <lx:label>Net Amount Frequency</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NumberOfRepeats" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The number of times the income is earned or received in the period specified.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.NumberOfRepeats</lx:path>
                              <lx:label>NumberOfRepeats</lx:label>
                              <li:transactions>CAL,CNZ,CDA,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PreviousYearAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The income amount received in the previous year.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.PreviousYearAmount</lx:path>
                              <lx:label>Previous Year Amount</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PrimaryForeignCurrency" type="currencyCodeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The original currency that the income was earned in. For example US Dollars, British Pounds.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.PrimaryForeignCurrency</lx:path>
                              <lx:label>Primary Foreign Currency</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProofCode" type="proofCodeOtherList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Code to indicate the type of evidence of the income</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.ProofCode</lx:path>
                              <lx:label>Proof Code</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProofSighted" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the Income evidence been sighted by the loan writer or assessor</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.ProofSighted</lx:path>
                              <lx:label>Proof Sighted</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this other income within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="StartDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the start date of the period for which the income being described is earned or received.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.StartDate</lx:path>
                              <lx:label>Start Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Taxed" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Income taxed indicates whether the income has already been subjected to tax and therefore whether a net amount or gross amount has been supplied.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.Taxed</lx:path>
                              <lx:label>Taxed</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Type" type="otherIncomeTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Specifies what type of income is described by the income item.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Asset" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A cross reference to the asset (a Real Estate Asset element or Non Real Estate Asset element) from which this income is generated.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.OtherIncome.x_Asset</lx:path>
                              <lx:label>Asset</lx:label>
                              <li:transactions>CAL,CNZ,CDA,SVC,LMI</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Overview" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Container element for the user input information applying to the entire application.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.Overview</lx:path>
                          <lx:label>Overview</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="BranchDomicile" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The branch that the applicants request that the requested loan accounts should be domiciled.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Overview.BranchDomicile</lx:path>
                                <lx:label>Branch Domicile</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Contains contact details for the branch.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Overview.BranchDomicile.Contact</lx:path>
                                      <lx:label>Contact</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="OfficeFax" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>A fax number for the branch.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.Overview.BranchDomicile.Contact.OfficeFax</lx:path>
                                            <lx:label>Office Fax</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>An office phone number for the branch.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.Overview.BranchDomicile.Contact.OfficePhone</lx:path>
                                            <lx:label>Office Phone</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="BSB" type="bsbPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the BSB of the branch the applicants request that the requested loan accounts should be domiciled.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.BranchDomicile.BSB</lx:path>
                                    <lx:label>BSB</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InternalName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The internal name given to a branch location by a lender or creditor may be captured where one exists.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.BranchDomicile.InternalName</lx:path>
                                    <lx:label>Branch Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InternalNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An internal identification number given to a branch location by a lender or creditor may be captured where one exists.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.BranchDomicile.InternalNumber</lx:path>
                                    <lx:label>Internal Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="BranchSign" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The branch that the applicants request that the documents for the loan be made available for signing prior to settlement.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Overview.BranchSign</lx:path>
                                <lx:label>Branch Sign</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Contains contact details for the branch.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Overview.BranchSign.Contact</lx:path>
                                      <lx:label>Contact</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="OfficeFax" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>A fax number for the branch.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.Overview.BranchSign.Contact.OfficeFax</lx:path>
                                            <lx:label>Office Fax</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>An office phone number for the branch.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.Overview.BranchSign.Contact.OfficePhone</lx:path>
                                            <lx:label>Office Phone</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="BSB" type="bsbPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the BSB of the branch the applicants request that the documents for the loan be made available for signing prior to settlement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.BranchSign.BSB</lx:path>
                                    <lx:label>BSB</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InternalName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The internal name given to a branch location by a lender or creditor may be captured where one exists.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.BranchSign.InternalName</lx:path>
                                    <lx:label>Branch Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InternalNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An internal identification number given to a branch location by a lender or creditor may be captured where one exists.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.BranchSign.InternalNumber</lx:path>
                                    <lx:label>Internal Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="BranchStamp" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The financial institution's branch stamp.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Overview.BranchStamp</lx:path>
                                <lx:label>Branch Stamp</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Stamped" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The application has been stamped with the financial institution's branch stamp.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.BranchStamp.Stamped</lx:path>
                                    <lx:label>Stamped</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StampRequired" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The application requires the financial institution's branch stamp.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.BranchStamp.StampRequired</lx:path>
                                    <lx:label>Stamp Required</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="BridgingFinance" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details on the bridging finance request.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Overview.BridgingFinance</lx:path>
                                <lx:label>Bridging Finance</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="BridgingTerm" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The estimated period of time required for the bridging finance, in months.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.BridgingFinance.BridgingTerm</lx:path>
                                    <lx:label>Bridging Term Months</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CapitalisedInterestAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount of capitalised interest on the bridged finance accrued during the bridging period.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.BridgingFinance.CapitalisedInterestAmount</lx:path>
                                    <lx:label>Capitalised Interest Amount</lx:label>
                                    <li:transactions>CAL,CDA,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CapitaliseInterest" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether interest on the bridged finance is to be capitalised.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.BridgingFinance.CapitaliseInterest</lx:path>
                                    <lx:label>Capitalise Interest</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EndDebt" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The end amount of the bridging loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.BridgingFinance.EndDebt</lx:path>
                                    <lx:label>End Debt</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EstimatedSellingCosts" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The estimated cost of selling the to-be-sold property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.BridgingFinance.EstimatedSellingCosts</lx:path>
                                    <lx:label>Estimated Selling Costs</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PeakDebt" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The peak amount of the bridging loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.BridgingFinance.PeakDebt</lx:path>
                                    <lx:label>Peak Debt</lx:label>
                                    <li:transactions>CAL,CDA,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ExistingCustomer" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>An applicant that is an existing customer of the organisation to which the application is being made.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Overview.ExistingCustomer</lx:path>
                                <lx:label>Existing Customer</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="CustomerNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A unique number assigned by the lender used to identify the customer.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.ExistingCustomer.CustomerNumber</lx:path>
                                    <lx:label>CustomerNumber</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.ExistingCustomer.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="MortgageInsurance" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains overview information that applies to mortgage insurance.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Overview.MortgageInsurance</lx:path>
                                <lx:label>Mortgage Insurance</lx:label>
                                <li:transactions>ACC,CDA</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Capitalised" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the mortgage insurance premium will be included in the amount borrowed.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.MortgageInsurance.Capitalised</lx:path>
                                    <lx:label>Capitalised</lx:label>
                                    <li:transactions>ACC,CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Required" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether mortgage insurance is required to cover the application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.MortgageInsurance.Required</lx:path>
                                    <lx:label>Required</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PricingDetails" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Pricing details</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Overview.PricingDetails</lx:path>
                                <lx:label>Pricing Details</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Description</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.PricingDetails.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DiscountReason" type="discountReasonList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The reason the applicant believe they should get discount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.PricingDetails.DiscountReason</lx:path>
                                    <lx:label>Discount Reason</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FinancialInstitution" type="financialInstitutionList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Financial institution</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.PricingDetails.FinancialInstitution</lx:path>
                                    <lx:label>Financial Institution</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FinancialInstitutionDiscount" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Discount in percentage offered to the applicant by the other financial institution.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.PricingDetails.FinancialInstitutionDiscount</lx:path>
                                    <lx:label>Financial Institution Discount</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherFIName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For a financial institution not listed, the text name of the FI.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Overview.PricingDetails.OtherFIName</lx:path>
                                    <lx:label>Other FI Name</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Signature" type="signatureType" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>This item is now deprecated and will be deleted in a future release. A signature for this application.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Overview.Signature</lx:path>
                                <lx:label>Signature</lx:label>
                                <li:transactions>CAL,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="ApplicationType" type="applicationTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Gives a broad indication of what the main application type is for, particularly if the lender has different loan processing arrangements.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.ApplicationType</lx:path>
                              <lx:label>Application Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BrokerApplicationReferenceNumber" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A unique identifier assigned to the application by the organisation that is submitting the application to the lender. Used by the loan writer to identify the application in their business systems.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.BrokerApplicationReferenceNumber</lx:path>
                              <lx:label>Broker Application Reference Number</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BrokerApplicationSequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A number to identify the specific submission for an application in the broker's business system. In combination with the Broker Application Reference Number the Broker Application Sequence Number helps uniquely identifies both initial submission and resubmissions of the same application.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.BrokerApplicationSequenceNumber</lx:path>
                              <lx:label>Broker Application Sequence Number</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CombinationLoan" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that the loan application is for two or more loan accounts with different products.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.CombinationLoan</lx:path>
                              <lx:label>Combination Loan</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DocType" type="applicationDocTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the level of income verification documentation that the applicants will provide to the lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.DocType</lx:path>
                              <lx:label>Doc Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ExpectedSettlementDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the date on which the application is expected to be settled</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.ExpectedSettlementDate</lx:path>
                              <lx:label>Expected Settlement Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="FastRefinance" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether it is a fast refinance application.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.FastRefinance</lx:path>
                              <lx:label>Fast Refinance</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="HasSpecialCircumstances" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>This application has special circumstances that need to be considered.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.HasSpecialCircumstances</lx:path>
                              <lx:label>Has Special Circumstances</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsBridgingFinance" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to note whether the loan application includes funds for bridging between a purchase and a subsequent sale.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.IsBridgingFinance</lx:path>
                              <lx:label>Is Bridging Finance</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LenderApplicationReferenceNumber" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A unique identifier assigned to the application by the lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.LenderApplicationReferenceNumber</lx:path>
                              <lx:label>Lender Application Reference Number</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LenderPreapprovalReferenceNumber" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A unique identifier assigned to a Preapproval application by the lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.LenderPreapprovalReferenceNumber</lx:path>
                              <lx:label>Lender Preapproval Reference Number</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LinkedCommercialApplication" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>This application has a linked commercial application.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.LinkedCommercialApplication</lx:path>
                              <lx:label>Linked Commercial Application</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LodgementReferenceNumber" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A unique identifier assigned to the application by the point of sale system provider. Used by the point of sale system provider to identify the application in their application capture system.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.LodgementReferenceNumber</lx:path>
                              <lx:label>Lodgement Reference Number</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LodgementSequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A number to identify the specific submission for an application that has been resubmitted.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.LodgementSequenceNumber</lx:path>
                              <lx:label>Lodgement Sequence Number</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PrivateBanking" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>This application is managed by a private banking service.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.PrivateBanking</lx:path>
                              <lx:label>Private Banking</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ProPack" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate if the application is propack.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.ProPack</lx:path>
                              <lx:label>ProPack</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SMSFLoan" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to note whether the loan application is for a self managed superfund.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.SMSFLoan</lx:path>
                              <lx:label>SMSF Loan</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Urgent" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>This application is designated as urgent.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.Urgent</lx:path>
                              <lx:label>Urgent</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_MainContactPoint" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A defined main contact point for the application. Can be an applicant, a broker, a solicitor, etc.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.x_MainContactPoint</lx:path>
                              <lx:label>Main Contact Point</lx:label>
                              <li:transactions>CAL</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_NeedsAnalysis" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Links to the Needs Analysis that was performed prior to creating this application.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Overview.x_NeedsAnalysis</lx:path>
                              <lx:label>Needs Analysis</lx:label>
                              <li:transactions>CDA</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.NeedsAnalysis</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="PersonApplicant" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Contains applicant details where the applicant is a person, being either a borrower or a guarantor, and may be acting for a Trust.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.PersonApplicant</lx:path>
                          <lx:label>Person Applicant</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for the applicant contact details</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.Contact</lx:path>
                                <lx:label>Contact</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="CurrentAddress" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The current address details for the applicant</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.Contact.CurrentAddress</lx:path>
                                      <lx:label>Current Address</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Duration" type="durationType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Container element for capturing the duration at the current address.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Contact.CurrentAddress.Duration</lx:path>
                                            <lx:label>Duration</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="HousingStatus" type="housingStatusList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates the housing status of this applicant e.g. Whether they are boarding, renting, own their own home, living with parents, other, etc.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.CurrentAddress.HousingStatus</lx:path>
                                          <lx:label>Housing Status</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StartDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date that the applicant moved to the residential address</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.CurrentAddress.StartDate</lx:path>
                                          <lx:label>Start Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Landlord" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Reference to the detail of the landlord of the applicant's current address.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.CurrentAddress.x_Landlord</lx:path>
                                          <lx:label>Landlord</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_MailingAddress" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Reference to the applicant's current address for mailing documents</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.CurrentAddress.x_MailingAddress</lx:path>
                                          <lx:label>Mailing Address</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.Address</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_ResidentialAddress" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Reference to the applicant's current residential address</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.CurrentAddress.x_ResidentialAddress</lx:path>
                                          <lx:label>Residential Address</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.Address</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="EmailAddress" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Wrapper element for email address information.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.Contact.EmailAddress</lx:path>
                                      <lx:label>Email Address</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Email" type="emailPattern" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>An email address for the person</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.EmailAddress.Email</lx:path>
                                          <lx:label>Email</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="EmailType" type="contactEmailTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Specify whether the applicant email address is personal or from work</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.EmailAddress.EmailType</lx:path>
                                          <lx:label>Email Type</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PrimaryEmail" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>This email address is the primary email address of the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.EmailAddress.PrimaryEmail</lx:path>
                                          <lx:label>Primary Email</lx:label>
                                          <li:transactions>ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Unique Identifier for this Email Address Element</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.EmailAddress.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="FaxNumber" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Fax number for the Applicant.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.Contact.FaxNumber</lx:path>
                                      <lx:label>Fax Number</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="HomePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Home phone captures the phone number used predominantly while this person is at home.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.Contact.HomePhone</lx:path>
                                      <lx:label>Home Phone</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="Mobile" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Applicant mobile number.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.Contact.Mobile</lx:path>
                                      <lx:label>Mobile</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="PostSettlementAddress" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The address details for the applicant following settlement</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.Contact.PostSettlementAddress</lx:path>
                                      <lx:label>Post Settlement Address</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="HousingStatus" type="housingStatusList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates the housing status of this applicant e.g. Whether they are boarding, renting, own their own home, living with parents, other, etc.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.PostSettlementAddress.HousingStatus</lx:path>
                                          <lx:label>Housing Status</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StartDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date on which the post settlement residential address will be effective.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.PostSettlementAddress.StartDate</lx:path>
                                          <lx:label>Start Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_MailingAddress" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Reference to the applicant's address for mailing documents after settlement</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.PostSettlementAddress.x_MailingAddress</lx:path>
                                          <lx:label>Mailing Address</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.Address</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_ResidentialAddress" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Reference to the applicant's residential address after settlement</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.PostSettlementAddress.x_ResidentialAddress</lx:path>
                                          <lx:label>Residential Address</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.Address</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="PreviousAddress" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Container element for an address previous to the current address</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.Contact.PreviousAddress</lx:path>
                                      <lx:label>Previous Address</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Duration" type="durationType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Container element for capturing the duration at the previous address.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Contact.PreviousAddress.Duration</lx:path>
                                            <lx:label>Duration</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="EndDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date that the applicant left this address</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.PreviousAddress.EndDate</lx:path>
                                          <lx:label>End Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="HousingStatus" type="housingStatusList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates the housing status of this applicant e.g. Whether they are boarding, renting, own their own home, living with parents, other, etc.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.PreviousAddress.HousingStatus</lx:path>
                                          <lx:label>Housing Status</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StartDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date that the applicant moved to this address</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.PreviousAddress.StartDate</lx:path>
                                          <lx:label>Start Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_ResidentialAddress" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Reference to the applicant's previous residential address</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.PreviousAddress.x_ResidentialAddress</lx:path>
                                          <lx:label>Residential Address</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.Address</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="PriorAddress" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Container element for an address prior to the 'previous' address</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.Contact.PriorAddress</lx:path>
                                      <lx:label>Prior Address</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Duration" type="durationType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Container element for capturing the duration at the prior address.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Contact.PriorAddress.Duration</lx:path>
                                            <lx:label>Duration</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="EndDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date that the applicant left this address</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.PriorAddress.EndDate</lx:path>
                                          <lx:label>End Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="HousingStatus" type="housingStatusList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates the housing status of this applicant e.g. Whether they are boarding, renting, own their own home, living with parents, other, etc.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.PriorAddress.HousingStatus</lx:path>
                                          <lx:label>Housing Status</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StartDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date that the applicant moved to this address</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.PriorAddress.StartDate</lx:path>
                                          <lx:label>Start Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_ResidentialAddress" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Reference to the applicant's prior residential address</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Contact.PriorAddress.x_ResidentialAddress</lx:path>
                                          <lx:label>Residential Address</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.Address</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="WorkPhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Where an applicant has a work number.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.Contact.WorkPhone</lx:path>
                                      <lx:label>Work Phone</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="PreferredContact" type="preferredContactPersonList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the preferred method for contacting this applicant</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Contact.PreferredContact</lx:path>
                                    <lx:label>Preferred Contact</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="CreditHistory" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains details of an applicant's credit history as provided both by credit reference bureaus and/or by the applicant themselves.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.CreditHistory</lx:path>
                                <lx:label>Credit History</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="BorrowerProvidedHistory" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The borrower themself has provided this credit history issue.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.CreditHistory.BorrowerProvidedHistory</lx:path>
                                    <lx:label>Borrower Provided History</lx:label>
                                    <li:transactions>LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A comment about the applicant's credit history, possibly in mitigation or explanation of credit issues.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.CreditHistory.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EndDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date when the credit issue ended.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.CreditHistory.EndDate</lx:path>
                                    <lx:label>End Date</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsCurrent" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether the credit issue is current. Selecting "No" means the issue described is a past credit issue.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.CreditHistory.IsCurrent</lx:path>
                                    <lx:label>Is Current</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Issue" type="creditHistoryIssueList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Describes the type of Credit Issues the applicant has experienced</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.CreditHistory.Issue</lx:path>
                                    <lx:label>Issue</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date when the credit issue started.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.CreditHistory.StartDate</lx:path>
                                    <lx:label>Start Date</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unique Identifier for this Credit History Element</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.CreditHistory.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="DocumentationInstructions" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information specifying where the lender is to send the loan documents once the loan has been approved. A nomination of authority at the Loan level will over-ride this nomination.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.DocumentationInstructions</lx:path>
                                <lx:label>Documentation Instructions</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Method" type="documentationInstructionsMethodList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the documentation is to be sent by email or by post.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.DocumentationInstructions.Method</lx:path>
                                    <lx:label>Method</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SendDocumentsTo" type="sendDocumentsToPersonList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicator for the party selected to receive loan documentation</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.DocumentationInstructions.SendDocumentsTo</lx:path>
                                    <lx:label>Send Documents To</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_NominatedAuthority" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The nominated authority for receiving the loan documents</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.DocumentationInstructions.x_NominatedAuthority</lx:path>
                                    <lx:label>Nominated Authority</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Employment" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains information about the persons current or previous employment.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.Employment</lx:path>
                                <lx:label>Employment</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="ForeignEmployed" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Indicates that the employment is foreign employment.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed</lx:path>
                                      <lx:label>ForeignEmployed</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Duration" type="durationType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Container element for capturing the duration of employment.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Duration</lx:path>
                                            <lx:label>Duration</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="Income" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Details of the income income-related amounts.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income</lx:path>
                                            <lx:label>Income</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="BonusAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross bonus income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.BonusAmount</lx:path>
                                                <lx:label>Gross Bonus Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="BonusFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross bonus income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.BonusFrequency</lx:path>
                                                <lx:label>Gross Bonus Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CarAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross car allowance amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.CarAllowanceAmount</lx:path>
                                                <lx:label>Gross Car Allowance Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CarAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross car allowance income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.CarAllowanceFrequency</lx:path>
                                                <lx:label>Gross Car Allowance Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CommissionAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross commission income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.CommissionAmount</lx:path>
                                                <lx:label>Gross Commission Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CommissionFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross commission income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.CommissionFrequency</lx:path>
                                                <lx:label>Gross Commission Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GrossRegularOvertimeAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross regular overtime income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.GrossRegularOvertimeAmount</lx:path>
                                                <lx:label>Gross Regular Overtime Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GrossRegularOvertimeFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross regular overtime income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.GrossRegularOvertimeFrequency</lx:path>
                                                <lx:label>Gross Regular Overtime Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GrossSalaryAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross salary income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.GrossSalaryAmount</lx:path>
                                                <lx:label>Gross Salary Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GrossSalaryFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross salary income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.GrossSalaryFrequency</lx:path>
                                                <lx:label>Gross Salary Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetBonusAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net bonus income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.NetBonusAmount</lx:path>
                                                <lx:label>Net Bonus Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetBonusFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving bonus income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.NetBonusFrequency</lx:path>
                                                <lx:label>Net Bonus Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetCarAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net car allowance amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.NetCarAllowanceAmount</lx:path>
                                                <lx:label>Net Car Allowance Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetCarAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving car allowance income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.NetCarAllowanceFrequency</lx:path>
                                                <lx:label>Net Car Allowance Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetCommissionAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net commission income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.NetCommissionAmount</lx:path>
                                                <lx:label>Net Commission Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetCommissionFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving commission income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.NetCommissionFrequency</lx:path>
                                                <lx:label>Net Commission Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetRegularOvertimeAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net regular overtime income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.NetRegularOvertimeAmount</lx:path>
                                                <lx:label>Net Regular Overtime Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetRegularOvertimeFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving regular overtime income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.NetRegularOvertimeFrequency</lx:path>
                                                <lx:label>Net Regular Overtime Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetSalaryAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net salary income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.NetSalaryAmount</lx:path>
                                                <lx:label>Net Salary Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetSalaryFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving salary income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.NetSalaryFrequency</lx:path>
                                                <lx:label>Net Salary Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetWorkAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net work allowance amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.NetWorkAllowanceAmount</lx:path>
                                                <lx:label>Net Work Allowance Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetWorkAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving work allowance income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.NetWorkAllowanceFrequency</lx:path>
                                                <lx:label>Net Work Allowance Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetWorkersCompensationAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net workers compensation income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.NetWorkersCompensationAmount</lx:path>
                                                <lx:label>Net Workers Compensation Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetWorkersCompensationFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving workers compensation income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.NetWorkersCompensationFrequency</lx:path>
                                                <lx:label>Net Workers Compensation Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProofSighted" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Indicates that evidence of this income has been sighted by the loan writer or assessor.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.ProofSighted</lx:path>
                                                <lx:label>Proof Sighted</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="WorkAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross work allowance amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.WorkAllowanceAmount</lx:path>
                                                <lx:label>Gross Work Allowance Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="WorkAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross work allowance income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.WorkAllowanceFrequency</lx:path>
                                                <lx:label>Gross Work Allowance Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="WorkersCompensationAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross workers compensation income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.WorkersCompensationAmount</lx:path>
                                                <lx:label>Gross Workers Compensation Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="WorkersCompensationFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross workers compensation income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Income.WorkersCompensationFrequency</lx:path>
                                                <lx:label>Gross Workers Compensation Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="AverageHoursPerWeek" type="decimalType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>When the applicant is employed on a part time or casual basis the average hours per week need to be collected for serviceability calculations.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.AverageHoursPerWeek</lx:path>
                                          <lx:label>Average Hours Per Week</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Basis" type="foreignEmployedBasisList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should indicate whether the employment type is fulltime, part-time, contract, temporary, casual, commission only or seasonal.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Basis</lx:path>
                                          <lx:label>Basis</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CompanyCar" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Flag to indicate whether the person has a company supplied (and paid) car</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.CompanyCar</lx:path>
                                          <lx:label>Company Car</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CustomIndustryCode" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>An industry code, from a customised list supported by the lender where the lender requires a code that is not within ANZSIC, BIC or GICS.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.CustomIndustryCode</lx:path>
                                          <lx:label>Custom Industry Code</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CustomOccupationCode" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A custom occupation code. A free text field for any other occupation code used by an entity.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.CustomOccupationCode</lx:path>
                                          <lx:label>Custom Occupation Code</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="EmployerType" type="employerTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should indicate whether the employer type belongs to the Public (government) or Private sectors.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.EmployerType</lx:path>
                                          <lx:label>Employer Type</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="EndDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The end date indicates when the applicant finalised their employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.EndDate</lx:path>
                                          <lx:label>End Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="GICSCode" type="gicsCodePattern" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A Global Industry Classification Standard (GICS) code.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.GICSCode</lx:path>
                                          <lx:label>GICS Code</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IncomeStatusOnOrBeforeSettlement" type="incomeStatusOnOrBeforeSettlementList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The income status on or before settlement.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.IncomeStatusOnOrBeforeSettlement</lx:path>
                                          <lx:label>Income Status On Or Before Settlement</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Industry" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures a free text description of the Industry Code.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Industry</lx:path>
                                          <lx:label>Industry</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="MainBusinessActivity" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should indicate what type of business they are working in</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.MainBusinessActivity</lx:path>
                                          <lx:label>Main Business Activity</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Occupation" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Text description of the person's role or occupation for this employment</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Occupation</lx:path>
                                          <lx:label>Occupation</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OnProbation" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should indicate whether they are on probation with their current employer.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.OnProbation</lx:path>
                                          <lx:label>On Probation</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PositionTitle" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should submit their position title relating to this instance of employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.PositionTitle</lx:path>
                                          <lx:label>Position Title</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ProbationDateEnds" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the date on which the applicant's probation period ends.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.ProbationDateEnds</lx:path>
                                          <lx:label>Probation Date Ends</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ProbationDateStarts" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date of commencement of the probation period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.ProbationDateStarts</lx:path>
                                          <lx:label>Probation Start Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The order of this foreign employed within the sequence.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.SequenceNumber</lx:path>
                                          <lx:label>Sequence Number</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StartDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The start date indicates when the applicant commenced their employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.StartDate</lx:path>
                                          <lx:label>Start Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Status" type="employmentStatusList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The current or previous status of the employment</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.Status</lx:path>
                                          <lx:label>Status</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Employer" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The employer details for this instance of ForeignEmployed employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.ForeignEmployed.x_Employer</lx:path>
                                          <lx:label>Employer</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="NotEmployed" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Where the applicant is unemployed, the details should be submitted, if relevant to this instance of employment.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed</lx:path>
                                      <lx:label>Not Employed</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Duration" type="durationType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Container element for capturing the duration of employment.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Duration</lx:path>
                                            <lx:label>Duration</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="Income" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Income details for this period without employment.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income</lx:path>
                                            <lx:label>Income</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="GovernmentBenefitsAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the government benefits income gross amount received by the applicant, excluding Jobseeker Support (New Zealand) and Newstart Allowance (Australia) which are captured separately.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.GovernmentBenefitsAmount</lx:path>
                                                <lx:label>Gross Government Benefits Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GovernmentBenefitsFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the frequency at which the applicant is receiving government benefits income gross amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.GovernmentBenefitsFrequency</lx:path>
                                                <lx:label>Gross Government Benefits Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="JobseekerSupportAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the gross Jobseeker Support income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.JobseekerSupportAmount</lx:path>
                                                <lx:label>Gross Jobseeker Support Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="JobseekerSupportFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the frequency at which the applicant is receiving Jobseeker Support income gross amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.JobseekerSupportFrequency</lx:path>
                                                <lx:label>Gross Jobseeker Support Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetGovernmentBenefitsAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the net government benefits income amount received by the applicant, excluding Jobseeker Support (New Zealand) and Newstart Allowance (Australia) which are captured separately.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.NetGovernmentBenefitsAmount</lx:path>
                                                <lx:label>Net Government Benefits Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetGovernmentBenefitsFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving government benefits income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.NetGovernmentBenefitsFrequency</lx:path>
                                                <lx:label>Net Government Benefits Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetJobseekerSupportAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the net Jobseeker Support income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.NetJobseekerSupportAmount</lx:path>
                                                <lx:label>Net Jobseeker Support Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetJobseekerSupportFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving government Jobseeker Support income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.NetJobseekerSupportFrequency</lx:path>
                                                <lx:label>Net Jobseeker Support Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetNewstartAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the Net Newstart Allowance income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.NetNewstartAllowanceAmount</lx:path>
                                                <lx:label>Net Newstart Allowance Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetNewstartAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving Newstart Allowance income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.NetNewstartAllowanceFrequency</lx:path>
                                                <lx:label>Net Newstart Allowance Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetOtherIncomeAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures a net income amount received by the applicant, that is not captured elsewhere.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.NetOtherIncomeAmount</lx:path>
                                                <lx:label>Net Other Income Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetOtherIncomeFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving other income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.NetOtherIncomeFrequency</lx:path>
                                                <lx:label>Net Other Income Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetPrivatePensionAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the net pension income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.NetPrivatePensionAmount</lx:path>
                                                <lx:label>Net Private Pension Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetPrivatePensionFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving private pension income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.NetPrivatePensionFrequency</lx:path>
                                                <lx:label>Net Private Pension Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetSuperannuationAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the net superannuation amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.NetSuperannuationAmount</lx:path>
                                                <lx:label>Net Superannuation Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetSuperannuationFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving superannuation income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.NetSuperannuationFrequency</lx:path>
                                                <lx:label>Net Superannuation Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NewstartAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the Newstart Allowance income gross amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.NewstartAllowanceAmount</lx:path>
                                                <lx:label>Gross Newstart Allowance Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NewstartAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the frequency at which the applicant is receiving Newstart Allowance income gross amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.NewstartAllowanceFrequency</lx:path>
                                                <lx:label>Gross Newstart Allowance Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="OtherIncomeAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures a gross income amount received by the applicant, that is not captured elsewhere.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.OtherIncomeAmount</lx:path>
                                                <lx:label>Gross Other Income Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="OtherIncomeDescription" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>A description of the other income.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.OtherIncomeDescription</lx:path>
                                                <lx:label>Other Income Description</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="OtherIncomeFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the frequency at which the applicant is receiving other income gross amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.OtherIncomeFrequency</lx:path>
                                                <lx:label>Gross Other Income Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="PrivatePensionAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the gross pension income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.PrivatePensionAmount</lx:path>
                                                <lx:label>Gross Private Pension Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="PrivatePensionFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the frequency at which the applicant is receiving private pension income gross amount</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.PrivatePensionFrequency</lx:path>
                                                <lx:label>Gross Private Pension Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProofCode" type="proofCodeNotEmployedList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Code to indicate the type of evidence of the income</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.ProofCode</lx:path>
                                                <lx:label>Proof Code</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProofSighted" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Has the Income evidence been sighted by the loan writer or assessor</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.ProofSighted</lx:path>
                                                <lx:label>Proof Sighted</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="SMSF" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Flag to indicate whether the superannuation income received by the applicant is from a Self Managed Super Fund.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.SMSF</lx:path>
                                                <lx:label>SMSF</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="SuperannuationAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the gross superannuation amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.SuperannuationAmount</lx:path>
                                                <lx:label>Gross Superannuation Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="SuperannuationFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the frequency at which the applicant is receiving superannuation income gross amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Income.SuperannuationFrequency</lx:path>
                                                <lx:label>Gross Superannuation Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="EndDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The end date indicates when the applicant finished being not employed.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.EndDate</lx:path>
                                          <lx:label>End Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="HomeDuties" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicants should indicate whether they perform home duties.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.HomeDuties</lx:path>
                                          <lx:label>Home Duties</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IncomeStatusOnOrBeforeSettlement" type="incomeStatusOnOrBeforeSettlementList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The income status on or before settlement.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.IncomeStatusOnOrBeforeSettlement</lx:path>
                                          <lx:label>Income Status On Or Before Settlement</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OnBenefits" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicants should indicate whether they are receiving unemployment benefits.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.OnBenefits</lx:path>
                                          <lx:label>On Benefits</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Retired" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicants should indicate whether they are retired</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Retired</lx:path>
                                          <lx:label>Retired</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The order of this not employed within the sequence.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.SequenceNumber</lx:path>
                                          <lx:label>Sequence Number</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StartDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The start date indicates when the applicant commenced being not employed.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.StartDate</lx:path>
                                          <lx:label>Start Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Status" type="employmentStatusList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The current or previous status of the employment</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Status</lx:path>
                                          <lx:label>Status</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Student" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicants should indicate whether they are currently a student.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.Student</lx:path>
                                          <lx:label>Student</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.NotEmployed.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="PAYE" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Wrapper to contain details regarding PAYE employment.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE</lx:path>
                                      <lx:label>PAYE</lx:label>
                                      <li:transactions>CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Duration" type="durationType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Container element for capturing the duration of employment.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Duration</lx:path>
                                            <lx:label>Duration</lx:label>
                                            <li:transactions>CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="Income" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Details of the income or income-related amounts.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income</lx:path>
                                            <lx:label>Income</lx:label>
                                            <li:transactions>CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="BonusAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross bonus income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.BonusAmount</lx:path>
                                                <lx:label>Gross Bonus Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="BonusFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross bonus income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.BonusFrequency</lx:path>
                                                <lx:label>Gross Bonus Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CarAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross car allowance amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.CarAllowanceAmount</lx:path>
                                                <lx:label>Gross Car Allowance Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CarAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross car allowance income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.CarAllowanceFrequency</lx:path>
                                                <lx:label>Gross Car Allowance Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CommissionAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross commission income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.CommissionAmount</lx:path>
                                                <lx:label>Gross Commission Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CommissionFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross commission income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.CommissionFrequency</lx:path>
                                                <lx:label>Gross Commission Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GrossRegularOvertimeAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross regular overtime income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.GrossRegularOvertimeAmount</lx:path>
                                                <lx:label>Gross Regular Overtime Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GrossRegularOvertimeFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross regular overtime income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.GrossRegularOvertimeFrequency</lx:path>
                                                <lx:label>Gross Regular Overtime Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GrossSalaryAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross salary income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.GrossSalaryAmount</lx:path>
                                                <lx:label>Gross Salary Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GrossSalaryFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross salary income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.GrossSalaryFrequency</lx:path>
                                                <lx:label>Gross Salary Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetBonusAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net bonus income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.NetBonusAmount</lx:path>
                                                <lx:label>Net Bonus Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetBonusFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving bonus income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.NetBonusFrequency</lx:path>
                                                <lx:label>Net Bonus Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetCarAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net car allowance amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.NetCarAllowanceAmount</lx:path>
                                                <lx:label>Net Car Allowance Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetCarAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving car allowance income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.NetCarAllowanceFrequency</lx:path>
                                                <lx:label>Net Car Allowance Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetCommissionAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net commission income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.NetCommissionAmount</lx:path>
                                                <lx:label>Net Commission Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetCommissionFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving commission income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.NetCommissionFrequency</lx:path>
                                                <lx:label>Net Commission Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetRegularOvertimeAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net regular overtime income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.NetRegularOvertimeAmount</lx:path>
                                                <lx:label>Net Regular Overtime Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetRegularOvertimeFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving regular overtime income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.NetRegularOvertimeFrequency</lx:path>
                                                <lx:label>Net Regular Overtime Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetSalaryAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net salary income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.NetSalaryAmount</lx:path>
                                                <lx:label>Net Salary Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetSalaryFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving salary income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.NetSalaryFrequency</lx:path>
                                                <lx:label>Net Salary Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetWorkAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net work allowance amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.NetWorkAllowanceAmount</lx:path>
                                                <lx:label>Net Work Allowance Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetWorkAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving work allowance income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.NetWorkAllowanceFrequency</lx:path>
                                                <lx:label>Net Work Allowance Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetWorkersCompensationAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net workers compensation income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.NetWorkersCompensationAmount</lx:path>
                                                <lx:label>Net Workers Compensation Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetWorkersCompensationFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving workers compensation income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.NetWorkersCompensationFrequency</lx:path>
                                                <lx:label>Net Workers Compensation Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProofCode" type="proofCodePAYEList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Code to indicate the type of evidence of the income.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.ProofCode</lx:path>
                                                <lx:label>Proof Code</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProofSighted" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Indicates that evidence of this income has been sighted by the loan writer or assessor.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.ProofSighted</lx:path>
                                                <lx:label>Proof Sighted</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="WorkAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross work allowance amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.WorkAllowanceAmount</lx:path>
                                                <lx:label>Gross Work Allowance Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="WorkAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross work allowance income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.WorkAllowanceFrequency</lx:path>
                                                <lx:label>Gross Work Allowance Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="WorkersCompensationAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross workers compensation income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.WorkersCompensationAmount</lx:path>
                                                <lx:label>Gross Workers Compensation Amount</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="WorkersCompensationFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross workers compensation income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Income.WorkersCompensationFrequency</lx:path>
                                                <lx:label>Gross Workers Compensation Frequency</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="ANZSCOOccupationCode" type="anzscoOccupationCodeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The ANZSCO occupation code specifying the person's role or occupation in this employment. This attribute is restricted to the list of ANZSCO occupation codes published by the ABS on 26/03/2013 that superseded the ASCO Occupation Codes, although the superseded codes still remain in common use.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.ANZSCOOccupationCode</lx:path>
                                          <lx:label>ANZSCO Occupation Code</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AverageHoursPerWeek" type="decimalType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>When the applicant is employed on a part time or casual basis the average hours per week need to be collected for serviceability calculations.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.AverageHoursPerWeek</lx:path>
                                          <lx:label>Average Hours Per Week</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Basis" type="payeBasisList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should indicate whether the employment type is fulltime, part-time, contract, temporary, casual, commission only or seasonal.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Basis</lx:path>
                                          <lx:label>Basis</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CompanyCar" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Flag to indicate whether the person has a company supplied (and paid) car</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.CompanyCar</lx:path>
                                          <lx:label>Company Car</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CustomIndustryCode" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>An industry code, from a customised list supported by the lender where the lender requires a code that is not within ANZSIC, BIC or GICS.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.CustomIndustryCode</lx:path>
                                          <lx:label>Custom Industry Code</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="EmployerType" type="employerTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should indicate whether the employer type belongs to the Public (government) or Private sectors.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.EmployerType</lx:path>
                                          <lx:label>Employer Type</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="EndDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The end date indicates when the applicant finalised their employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.EndDate</lx:path>
                                          <lx:label>End Date</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="GICSCode" type="gicsCodePattern" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A Global Industry Classification Standard (GICS) code.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.GICSCode</lx:path>
                                          <lx:label>GICS Code</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Industry" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures a free text description of the Industry Code.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Industry</lx:path>
                                          <lx:label>Industry</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IndustryCode" type="anzsicIndustryCodeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The four digit industry code, from the ANZSIC list of codes. If the nominal code is two or three digits, add '0's to the end to make the four digit code.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.IndustryCode</lx:path>
                                          <lx:label>Industry Code</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="MainBusinessActivity" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should indicate what type of business they are working in</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.MainBusinessActivity</lx:path>
                                          <lx:label>Main Business Activity</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NewZealandBICCode" type="nzbicPattern" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A Business Industry Classification Code.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.NewZealandBICCode</lx:path>
                                          <lx:label>New Zealand BIC Code</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Occupation" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Text description of the person's role or occupation for this employment</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Occupation</lx:path>
                                          <lx:label>Occupation</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OccupationCode" type="absOccupationCodeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>An ABS occupation code specifying the person's role or occupation in this employment. This attribute remains but is likely to be deprecated in the future, with the ANZSCOOccupationCode attribute being restricted to the more recent ANZSCO Occupation Codes published by the ABS on 26/03/2013.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.OccupationCode</lx:path>
                                          <lx:label>Occupation Code</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OnProbation" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should indicate whether they are on probation with their current employer.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.OnProbation</lx:path>
                                          <lx:label>On Probation</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PositionTitle" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should submit their position title relating to this instance of employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.PositionTitle</lx:path>
                                          <lx:label>Position Title</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ProbationDateEnds" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the date on which the applicant's probation period ends.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.ProbationDateEnds</lx:path>
                                          <lx:label>Probation Date Ends</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The order of this PAYE within the sequence.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.SequenceNumber</lx:path>
                                          <lx:label>Sequence Number</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StartDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The start date indicates when the applicant commenced their employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.StartDate</lx:path>
                                          <lx:label>Start Date</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Status" type="employmentStatusList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The current or previous status of the employment</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.Status</lx:path>
                                          <lx:label>Status</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Employer" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should capture the employer details for this instance of PAYE employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYE.x_Employer</lx:path>
                                          <lx:label>Employer</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="PAYG" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Indicates that the employment is PAYG employment.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG</lx:path>
                                      <lx:label>PAYG</lx:label>
                                      <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Duration" type="durationType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Container element for capturing the duration of employment.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Duration</lx:path>
                                            <lx:label>Duration</lx:label>
                                            <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="Income" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Details of the income income-related amounts.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income</lx:path>
                                            <lx:label>Income</lx:label>
                                            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="BonusAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross bonus income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.BonusAmount</lx:path>
                                                <lx:label>Gross Bonus Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="BonusFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross bonus income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.BonusFrequency</lx:path>
                                                <lx:label>Gross Bonus Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CarAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross car allowance amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.CarAllowanceAmount</lx:path>
                                                <lx:label>Gross Car Allowance Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CarAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross car allowance income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.CarAllowanceFrequency</lx:path>
                                                <lx:label>Gross Car Allowance Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CommissionAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross commission income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.CommissionAmount</lx:path>
                                                <lx:label>Gross Commission Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CommissionFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross commission income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.CommissionFrequency</lx:path>
                                                <lx:label>Gross Commission Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GrossRegularOvertimeAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross regular overtime income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.GrossRegularOvertimeAmount</lx:path>
                                                <lx:label>Gross Regular Overtime Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GrossRegularOvertimeFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross regular overtime income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.GrossRegularOvertimeFrequency</lx:path>
                                                <lx:label>Gross Regular Overtime Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GrossSalaryAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross salary income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.GrossSalaryAmount</lx:path>
                                                <lx:label>Gross Salary Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GrossSalaryFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross salary income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.GrossSalaryFrequency</lx:path>
                                                <lx:label>Gross Salary Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetBonusAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net bonus income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.NetBonusAmount</lx:path>
                                                <lx:label>Net Bonus Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetBonusFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving bonus income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.NetBonusFrequency</lx:path>
                                                <lx:label>Net Bonus Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetCarAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net car allowance amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.NetCarAllowanceAmount</lx:path>
                                                <lx:label>Net Car Allowance Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetCarAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving car allowance income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.NetCarAllowanceFrequency</lx:path>
                                                <lx:label>Net Car Allowance Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetCommissionAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net commission income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.NetCommissionAmount</lx:path>
                                                <lx:label>Net Commission Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetCommissionFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving commission income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.NetCommissionFrequency</lx:path>
                                                <lx:label>Net Commission Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetRegularOvertimeAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net regular overtime income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.NetRegularOvertimeAmount</lx:path>
                                                <lx:label>Net Regular Overtime Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetRegularOvertimeFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving regular overtime income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.NetRegularOvertimeFrequency</lx:path>
                                                <lx:label>Net Regular Overtime Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetSalaryAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net salary income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.NetSalaryAmount</lx:path>
                                                <lx:label>Net Salary Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetSalaryFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving salary income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.NetSalaryFrequency</lx:path>
                                                <lx:label>Net Salary Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetWorkAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net work allowance amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.NetWorkAllowanceAmount</lx:path>
                                                <lx:label>Net Work Allowance Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetWorkAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving work allowance income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.NetWorkAllowanceFrequency</lx:path>
                                                <lx:label>Net Work Allowance Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetWorkersCompensationAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net workers compensation income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.NetWorkersCompensationAmount</lx:path>
                                                <lx:label>Net Workers Compensation Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetWorkersCompensationFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving workers compensation income net amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.NetWorkersCompensationFrequency</lx:path>
                                                <lx:label>Net Workers Compensation Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="OtherFringeBenefitAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The other fringe benefit amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.OtherFringeBenefitAmount</lx:path>
                                                <lx:label>Other Fringe Benefit Amount</lx:label>
                                                <li:transactions>CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="OtherFringeBenefitFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving other fringe benefit amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.OtherFringeBenefitFrequency</lx:path>
                                                <lx:label>Other Fringe Benefit Frequency</lx:label>
                                                <li:transactions>CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="PermanentAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The permanent allowance received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.PermanentAllowanceAmount</lx:path>
                                                <lx:label>Permanent Allowance Amount</lx:label>
                                                <li:transactions>CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="PermanentAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency at which the applicant is receiving permanent allowance.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.PermanentAllowanceFrequency</lx:path>
                                                <lx:label>Permanent Allowance Frequency</lx:label>
                                                <li:transactions>CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProofCode" type="proofCodePAYGList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Code to indicate the type of evidence of the income.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.ProofCode</lx:path>
                                                <lx:label>Proof Code</lx:label>
                                                <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProofSighted" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Indicates that evidence of this income has been sighted by the loan writer or assessor.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.ProofSighted</lx:path>
                                                <lx:label>Proof Sighted</lx:label>
                                                <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="WorkAllowanceAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross work allowance amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.WorkAllowanceAmount</lx:path>
                                                <lx:label>Gross Work Allowance Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="WorkAllowanceFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross work allowance income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.WorkAllowanceFrequency</lx:path>
                                                <lx:label>Gross Work Allowance Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="WorkersCompensationAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The gross workers compensation income amount received by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.WorkersCompensationAmount</lx:path>
                                                <lx:label>Gross Workers Compensation Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="WorkersCompensationFrequency" type="frequencyShortList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The frequency of the gross workers compensation income payment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Income.WorkersCompensationFrequency</lx:path>
                                                <lx:label>Gross Workers Compensation Frequency</lx:label>
                                                <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="ANZSCOOccupationCode" type="anzscoOccupationCodeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The ANZSCO occupation code specifying the person's role or occupation in this employment. This attribute is restricted to the list of ANZSCO occupation codes published by the ABS on 26/03/2013 that superseded the ASCO Occupation Codes, although the superseded codes still remain in common use.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.ANZSCOOccupationCode</lx:path>
                                          <lx:label>ANZSCO Occupation Code</lx:label>
                                          <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AustralianBIC" type="aubicPattern" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A Business Industry Code.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.AustralianBIC</lx:path>
                                          <lx:label>Australian BIC</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AverageHoursPerWeek" type="decimalType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>When the applicant is employed on a part time or casual basis the average hours per week need to be collected for serviceability calculations.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.AverageHoursPerWeek</lx:path>
                                          <lx:label>Average Hours Per Week</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Basis" type="paygBasisList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should indicate whether the employment type is fulltime, part-time, contract, temporary, casual, commission only or seasonal.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Basis</lx:path>
                                          <lx:label>Basis</lx:label>
                                          <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CompanyCar" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Flag to indicate whether the person has a company supplied (and paid) car</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.CompanyCar</lx:path>
                                          <lx:label>Company Car</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CustomIndustryCode" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>An industry code, from a customised list supported by the lender where the lender requires a code that is not within ANZSIC, BIC or GICS.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.CustomIndustryCode</lx:path>
                                          <lx:label>Custom Industry Code</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="EmployerType" type="employerTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should indicate whether the employer type belongs to the Public (government) or Private sectors.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.EmployerType</lx:path>
                                          <lx:label>Employer Type</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="EndDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The end date indicates when the applicant finalised their employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.EndDate</lx:path>
                                          <lx:label>End Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="GICSCode" type="gicsCodePattern" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A Global Industry Classification Standard (GICS) code.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.GICSCode</lx:path>
                                          <lx:label>GICS Code</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IncomeStatusOnOrBeforeSettlement" type="incomeStatusOnOrBeforeSettlementList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The income status on or before settlement.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.IncomeStatusOnOrBeforeSettlement</lx:path>
                                          <lx:label>Income Status On Or Before Settlement</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Industry" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures a free text description of the Industry Code.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Industry</lx:path>
                                          <lx:label>Industry</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IndustryCode" type="anzsicIndustryCodeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The four digit industry code, from the ANZSIC list of codes. If the nominal code is two or three digits, add '0's to the end to make the four digit code.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.IndustryCode</lx:path>
                                          <lx:label>Industry Code</lx:label>
                                          <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="MainBusinessActivity" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should indicate what type of business they are working in</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.MainBusinessActivity</lx:path>
                                          <lx:label>Main Business Activity</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Occupation" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Text description of the person's role or occupation for this employment</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Occupation</lx:path>
                                          <lx:label>Occupation</lx:label>
                                          <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OccupationCode" type="absOccupationCodeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>An ABS occupation code specifying the person's role or occupation in this employment. This attribute remains but is likely to be deprecated in the future, with the ANZSCOOccupationCode attribute being restricted to the more recent ANZSCO Occupation Codes published by the ABS on 26/03/2013.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.OccupationCode</lx:path>
                                          <lx:label>Occupation Code</lx:label>
                                          <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OnProbation" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should indicate whether they are on probation with their current employer.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.OnProbation</lx:path>
                                          <lx:label>On Probation</lx:label>
                                          <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PositionTitle" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should submit their position title relating to this instance of employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.PositionTitle</lx:path>
                                          <lx:label>Position Title</lx:label>
                                          <li:transactions>CAL,CDA,DAS,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ProbationDateEnds" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the date on which the applicant's probation period ends.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.ProbationDateEnds</lx:path>
                                          <lx:label>Probation Date Ends</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ProbationDateStarts" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date of commencement of the probation period.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.ProbationDateStarts</lx:path>
                                          <lx:label>Probation Start Date</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The order of this PAYG within the sequence.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.SequenceNumber</lx:path>
                                          <lx:label>Sequence Number</lx:label>
                                          <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StartDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The start date indicates when the applicant commenced their employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.StartDate</lx:path>
                                          <lx:label>Start Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Status" type="employmentStatusList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The current or previous status of the employment</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.Status</lx:path>
                                          <lx:label>Status</lx:label>
                                          <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Employer" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The employer details for this instance of PAYG employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.PAYG.x_Employer</lx:path>
                                          <lx:label>Employer</lx:label>
                                          <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="SelfEmployed" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Self-employed refers to people who 'work for themselves' in either wholly owned or part-owned organisations, including when operating as a (paid) Director. The organisation may or may not employ other people.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed</lx:path>
                                      <lx:label>Self Employed</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Business" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Container element for self employed business details</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Business</lx:path>
                                            <lx:label>Business</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="AustralianBIC" type="aubicPattern" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>A Business Industry Code.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Business.AustralianBIC</lx:path>
                                                <lx:label>Australian BIC</lx:label>
                                                <li:transactions>CAL,CDA,DAS</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CustomIndustryCode" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>An industry code, from a customised list supported by the lender where the lender requires a code that is not within ANZSIC, BIC or GICS.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Business.CustomIndustryCode</lx:path>
                                                <lx:label>Custom Industry Code</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="GICSCode" type="gicsCodePattern" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>A Global Industry Classification Standard (GICS) code.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Business.GICSCode</lx:path>
                                                <lx:label>GICS Code</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Industry" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures a free text description of the Industry Code.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Business.Industry</lx:path>
                                                <lx:label>Industry</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="IndustryCode" type="anzsicIndustryCodeList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the ANZSIC industry code</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Business.IndustryCode</lx:path>
                                                <lx:label>Industry Code</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="IsAFranchise" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures whether the business is a franchise</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Business.IsAFranchise</lx:path>
                                                <lx:label>Is A Franchise</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="MainBusinessActivity" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The applicant should indicate what type of business they are working in</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Business.MainBusinessActivity</lx:path>
                                                <lx:label>Main Business Activity</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NewZealandBICCode" type="nzbicPattern" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>A Business Industry Classification Code.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Business.NewZealandBICCode</lx:path>
                                                <lx:label>New Zealand BIC Code</lx:label>
                                                <li:transactions>CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NumberOfEmployees" type="integerType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The number of employees in the business.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Business.NumberOfEmployees</lx:path>
                                                <lx:label>Number Of Employees</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NumberOfLocations" type="integerType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The number of locations where the business operates should be captured.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Business.NumberOfLocations</lx:path>
                                                <lx:label>Number Of Locations</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="OwnPremises" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Indicates whether the sole trader or partners own the premises from which the business is operated</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Business.OwnPremises</lx:path>
                                                <lx:label>Own Premises</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="StartDate" type="dateType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The start date for the self employed business.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Business.StartDate</lx:path>
                                                <lx:label>Business Start Date</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="BusinessIncomePrevious" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Details of the share of the income or income-related amounts in the previous financial year (immediately preceding the most recent full financial year), attributable to this applicant (for year = current year -2).</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious</lx:path>
                                            <lx:label>Business Income Previous</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:sequence>
                                            <xs:element name="Addback" minOccurs="0" maxOccurs="1">
                                              <xs:annotation>
                                                <xs:documentation>Container element for Previous Business Income Addback</xs:documentation>
                                                <xs:appinfo>
                                                  <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback</lx:path>
                                                  <lx:label>Addback</lx:label>
                                                  <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                </xs:appinfo>
                                              </xs:annotation>
                                              <xs:complexType>
                                                <xs:sequence>
                                                  <xs:element name="OtherAddback" minOccurs="0" maxOccurs="unbounded">
                                                    <xs:annotation>
                                                      <xs:documentation>Other addback that cannot otherwise be captured under a specific addback category.</xs:documentation>
                                                      <xs:appinfo>
                                                        <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.OtherAddback</lx:path>
                                                        <lx:label>Other Add Back</lx:label>
                                                        <li:transactions>CAL,CDA,SVC</li:transactions>
                                                      </xs:appinfo>
                                                    </xs:annotation>
                                                    <xs:complexType>
                                                      <xs:attribute name="Amount" type="currencyType" use="optional">
                                                        <xs:annotation>
                                                          <xs:documentation>Amount that can be an addback to the income.</xs:documentation>
                                                          <xs:appinfo>
                                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.OtherAddback.Amount</lx:path>
                                                            <lx:label>Amount</lx:label>
                                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                                          </xs:appinfo>
                                                        </xs:annotation>
                                                      </xs:attribute>
                                                      <xs:attribute name="Description" type="stringType" use="optional">
                                                        <xs:annotation>
                                                          <xs:documentation>Decription of the other addback.</xs:documentation>
                                                          <xs:appinfo>
                                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.OtherAddback.Description</lx:path>
                                                            <lx:label>Description</lx:label>
                                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                                          </xs:appinfo>
                                                        </xs:annotation>
                                                      </xs:attribute>
                                                    </xs:complexType>
                                                  </xs:element>
                                                </xs:sequence>
                                                <xs:attribute name="Allowances" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Allowances amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.Allowances</lx:path>
                                                      <lx:label>Allowances</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="AmortisationOfGoodwill" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Amortisation Of Goodwill amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.AmortisationOfGoodwill</lx:path>
                                                      <lx:label>Amortisation Of Goodwill</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Bonus" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Bonus amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.Bonus</lx:path>
                                                      <lx:label>Bonus</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="CarExpense" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Car expense amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.CarExpense</lx:path>
                                                      <lx:label>Car Expense</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="CarryForwardLosses" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Carry Forward Losses amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.CarryForwardLosses</lx:path>
                                                      <lx:label>Carry Forward Losses</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Depreciation" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Depreciation amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.Depreciation</lx:path>
                                                      <lx:label>Depreciation</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="HomeOffice" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Home Office amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.HomeOffice</lx:path>
                                                      <lx:label>Home Office</lx:label>
                                                      <li:transactions>CNZ</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Interest" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Interest amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.Interest</lx:path>
                                                      <lx:label>Interest</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Lease" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Lease amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.Lease</lx:path>
                                                      <lx:label>Lease</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="NonCashBenefits" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Non-cash benefits amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.NonCashBenefits</lx:path>
                                                      <lx:label>Non Cash Benefits</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="NonRecurringExpenses" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Non-recurring expense amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.NonRecurringExpenses</lx:path>
                                                      <lx:label>Non Recurring Expenses</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Salary" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Salary amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.Salary</lx:path>
                                                      <lx:label>Salary</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="SuperannuationExcess" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Excess superannuation amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.Addback.SuperannuationExcess</lx:path>
                                                      <lx:label>Superannuation Excess</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                              </xs:complexType>
                                            </xs:element>
                                            <xs:element name="ForeignSourcedIncome" minOccurs="0" maxOccurs="1">
                                              <xs:annotation>
                                                <xs:documentation>Container element for any foreign sourced income</xs:documentation>
                                                <xs:appinfo>
                                                  <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.ForeignSourcedIncome</lx:path>
                                                  <lx:label>Foreign Sourced Income</lx:label>
                                                  <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                </xs:appinfo>
                                              </xs:annotation>
                                              <xs:complexType>
                                                <xs:attribute name="AUDAmount" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>The amount of foreign sourced income in Australian Dollars</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.ForeignSourcedIncome.AUDAmount</lx:path>
                                                      <lx:label>AUD Amount</lx:label>
                                                      <li:transactions>CAL,CDA,DAS</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="NZDAmount" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>The amount of foreign sourced income in New Zealand Dollars</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.ForeignSourcedIncome.NZDAmount</lx:path>
                                                      <lx:label>NZD Amount</lx:label>
                                                      <li:transactions>CNZ</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="PrimaryForeignCurrency" type="currencyCodeList" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>The original currency that the income was earned in. For example US Dollars, British Pounds.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.ForeignSourcedIncome.PrimaryForeignCurrency</lx:path>
                                                      <lx:label>Primary Foreign Currency</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                              </xs:complexType>
                                            </xs:element>
                                          </xs:sequence>
                                          <xs:attribute name="EndDate" type="dateType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The end date for the stated income, generally the end of a financial year</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.EndDate</lx:path>
                                                <lx:label>End Date</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="HasForeignSourcedIncome" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Has the applicant received any income from foreign sources</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.HasForeignSourcedIncome</lx:path>
                                                <lx:label>Has Foreign Sourced Income</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="IncomeGreaterThanPreviousYear" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>When comparing past Tax Office Assessments, indicate whether the past return has improved on the previous year.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.IncomeGreaterThanPreviousYear</lx:path>
                                                <lx:label>Income Greater Than Previous Year</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProfitAfterTax" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the profit after tax for the specified financial year.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.ProfitAfterTax</lx:path>
                                                <lx:label>Profit After Tax</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProfitBeforeTax" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the net profit/profit before tax for the specified financial year.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.ProfitBeforeTax</lx:path>
                                                <lx:label>Profit Before Tax</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="StartDate" type="dateType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The start date for the stated income, generally the start of a financial year</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.StartDate</lx:path>
                                                <lx:label>Start Date</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="TaxOfficeAssessments" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Are Tax Office Assessments for at least the past two years available as supporting evidence for this application where the applicant is self-employed.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.TaxOfficeAssessments</lx:path>
                                                <lx:label>Tax Office Assessments</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="x_Accountant" type="referenceType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Link to the person's accountant for the full previous financial year. This should be populated when the accountant for the specified period is different to that declared as the current accountant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrevious.x_Accountant</lx:path>
                                                <lx:label>Accountant</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                <lx:references>
                                                  <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                                </lx:references>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="BusinessIncomePrior" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Details of the share of the income or income-related amounts in the prior financial year (two years prior to the most recent full financial year), attributable to this applicant (for year = current year -3).</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior</lx:path>
                                            <lx:label>Business Income Prior</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:sequence>
                                            <xs:element name="Addback" minOccurs="0" maxOccurs="1">
                                              <xs:annotation>
                                                <xs:documentation>Container element for Prior Business Income Addback</xs:documentation>
                                                <xs:appinfo>
                                                  <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback</lx:path>
                                                  <lx:label>Addback</lx:label>
                                                  <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                </xs:appinfo>
                                              </xs:annotation>
                                              <xs:complexType>
                                                <xs:sequence>
                                                  <xs:element name="OtherAddback" minOccurs="0" maxOccurs="unbounded">
                                                    <xs:annotation>
                                                      <xs:documentation>Other addback that cannot otherwise be captured under a specific addback category.</xs:documentation>
                                                      <xs:appinfo>
                                                        <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.OtherAddback</lx:path>
                                                        <lx:label>Other Add Back</lx:label>
                                                        <li:transactions>CAL,CDA,SVC</li:transactions>
                                                      </xs:appinfo>
                                                    </xs:annotation>
                                                    <xs:complexType>
                                                      <xs:attribute name="Amount" type="currencyType" use="optional">
                                                        <xs:annotation>
                                                          <xs:documentation>Amount that can be an addback to the income.</xs:documentation>
                                                          <xs:appinfo>
                                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.OtherAddback.Amount</lx:path>
                                                            <lx:label>Amount</lx:label>
                                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                                          </xs:appinfo>
                                                        </xs:annotation>
                                                      </xs:attribute>
                                                      <xs:attribute name="Description" type="stringType" use="optional">
                                                        <xs:annotation>
                                                          <xs:documentation>Decription of the other addback.</xs:documentation>
                                                          <xs:appinfo>
                                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.OtherAddback.Description</lx:path>
                                                            <lx:label>Description</lx:label>
                                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                                          </xs:appinfo>
                                                        </xs:annotation>
                                                      </xs:attribute>
                                                    </xs:complexType>
                                                  </xs:element>
                                                </xs:sequence>
                                                <xs:attribute name="Allowances" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Allowances amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.Allowances</lx:path>
                                                      <lx:label>Allowances</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="AmortisationOfGoodwill" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Amortisation Of Goodwill amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.AmortisationOfGoodwill</lx:path>
                                                      <lx:label>Amortisation Of Goodwill</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Bonus" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Bonus amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.Bonus</lx:path>
                                                      <lx:label>Bonus</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="CarExpense" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Car expense amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.CarExpense</lx:path>
                                                      <lx:label>Car Expense</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="CarryForwardLosses" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Carry Forward Losses amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.CarryForwardLosses</lx:path>
                                                      <lx:label>Carry Forward Losses</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Depreciation" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Depreciation amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.Depreciation</lx:path>
                                                      <lx:label>Depreciation</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="HomeOffice" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Home Office amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.HomeOffice</lx:path>
                                                      <lx:label>Home Office</lx:label>
                                                      <li:transactions>CNZ</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Interest" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Interest amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.Interest</lx:path>
                                                      <lx:label>Interest</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Lease" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Lease amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.Lease</lx:path>
                                                      <lx:label>Lease</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="NonCashBenefits" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Non-cash benefits amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.NonCashBenefits</lx:path>
                                                      <lx:label>Non Cash Benefits</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="NonRecurringExpenses" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Non-recurring expense amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.NonRecurringExpenses</lx:path>
                                                      <lx:label>Non Recurring Expenses</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Salary" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Salary amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.Salary</lx:path>
                                                      <lx:label>Salary</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="SuperannuationExcess" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Excess superannuation amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.Addback.SuperannuationExcess</lx:path>
                                                      <lx:label>Superannuation Excess</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                              </xs:complexType>
                                            </xs:element>
                                            <xs:element name="ForeignSourcedIncome" minOccurs="0" maxOccurs="1">
                                              <xs:annotation>
                                                <xs:documentation>Container element for any foreign sourced income</xs:documentation>
                                                <xs:appinfo>
                                                  <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.ForeignSourcedIncome</lx:path>
                                                  <lx:label>Foreign Sourced Income</lx:label>
                                                  <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                </xs:appinfo>
                                              </xs:annotation>
                                              <xs:complexType>
                                                <xs:attribute name="AUDAmount" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>The amount of foreign sourced income in Australian Dollars</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.ForeignSourcedIncome.AUDAmount</lx:path>
                                                      <lx:label>AUD Amount</lx:label>
                                                      <li:transactions>CAL,CDA,DAS</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="NZDAmount" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>The amount of foreign sourced income in New Zealand Dollars</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.ForeignSourcedIncome.NZDAmount</lx:path>
                                                      <lx:label>NZD Amount</lx:label>
                                                      <li:transactions>CNZ</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="PrimaryForeignCurrency" type="currencyCodeList" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>The original currency that the income was earned in. For example US Dollars, British Pounds</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.ForeignSourcedIncome.PrimaryForeignCurrency</lx:path>
                                                      <lx:label>Primary Foreign Currency</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                              </xs:complexType>
                                            </xs:element>
                                          </xs:sequence>
                                          <xs:attribute name="EndDate" type="dateType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The end date for the stated income, generally the end of a financial year.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.EndDate</lx:path>
                                                <lx:label>End Date</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="HasForeignSourcedIncome" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Has the applicant received any income from foreign sources</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.HasForeignSourcedIncome</lx:path>
                                                <lx:label>Has Foreign Sourced Income</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="IncomeGreaterThanPreviousYear" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>When comparing past Tax Office Assessments, indicate whether the past return has improved on the Prior year.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.IncomeGreaterThanPreviousYear</lx:path>
                                                <lx:label>Income Greater Than Previous Year</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProfitAfterTax" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the profit after tax for the specified financial year.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.ProfitAfterTax</lx:path>
                                                <lx:label>Profit After Tax</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProfitBeforeTax" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the net profit/profit before tax for the specified financial year.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.ProfitBeforeTax</lx:path>
                                                <lx:label>Profit Before Tax</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="StartDate" type="dateType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The start date for the stated income, generally the start of a financial year.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.StartDate</lx:path>
                                                <lx:label>Start Date</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="TaxOfficeAssessments" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Are Tax Office Assessments for at least the past two years available as supporting evidence for this application where the applicant is self-employed.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.TaxOfficeAssessments</lx:path>
                                                <lx:label>Tax Office Assessments</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="x_Accountant" type="referenceType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Link to the person's accountant for the full prior financial year. This should be populated when the accountant for the specified period is different to that declared as the current accountant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomePrior.x_Accountant</lx:path>
                                                <lx:label>Accountant</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                <lx:references>
                                                  <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                                </lx:references>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="BusinessIncomeRecent" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Details of the share of the income or income-related amounts in the most recent full financial year, attributable to this applicant (year = current year -1).</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent</lx:path>
                                            <lx:label>Business Income Recent</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:sequence>
                                            <xs:element name="Addback" minOccurs="0" maxOccurs="1">
                                              <xs:annotation>
                                                <xs:documentation>Container element for Recent Business Income Addback</xs:documentation>
                                                <xs:appinfo>
                                                  <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback</lx:path>
                                                  <lx:label>Addback</lx:label>
                                                  <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                </xs:appinfo>
                                              </xs:annotation>
                                              <xs:complexType>
                                                <xs:sequence>
                                                  <xs:element name="OtherAddback" minOccurs="0" maxOccurs="unbounded">
                                                    <xs:annotation>
                                                      <xs:documentation>Other addback that cannot otherwise be captured under a specific addback category.</xs:documentation>
                                                      <xs:appinfo>
                                                        <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.OtherAddback</lx:path>
                                                        <lx:label>Other Add Back</lx:label>
                                                        <li:transactions>CAL,CDA,SVC</li:transactions>
                                                      </xs:appinfo>
                                                    </xs:annotation>
                                                    <xs:complexType>
                                                      <xs:attribute name="Amount" type="currencyType" use="optional">
                                                        <xs:annotation>
                                                          <xs:documentation>Amount that can be an addback to the income.</xs:documentation>
                                                          <xs:appinfo>
                                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.OtherAddback.Amount</lx:path>
                                                            <lx:label>Amount</lx:label>
                                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                                          </xs:appinfo>
                                                        </xs:annotation>
                                                      </xs:attribute>
                                                      <xs:attribute name="Description" type="stringType" use="optional">
                                                        <xs:annotation>
                                                          <xs:documentation>Decription of the other addback.</xs:documentation>
                                                          <xs:appinfo>
                                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.OtherAddback.Description</lx:path>
                                                            <lx:label>Description</lx:label>
                                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                                          </xs:appinfo>
                                                        </xs:annotation>
                                                      </xs:attribute>
                                                    </xs:complexType>
                                                  </xs:element>
                                                </xs:sequence>
                                                <xs:attribute name="Allowances" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Allowances amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.Allowances</lx:path>
                                                      <lx:label>Allowances</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="AmortisationOfGoodwill" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Amortisation Of Goodwill amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.AmortisationOfGoodwill</lx:path>
                                                      <lx:label>Amortisation Of Goodwill</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Bonus" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Bonus amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.Bonus</lx:path>
                                                      <lx:label>Bonus</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="CarExpense" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Car expense amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.CarExpense</lx:path>
                                                      <lx:label>Car Expense</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="CarryForwardLosses" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Carry Forward Losses amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.CarryForwardLosses</lx:path>
                                                      <lx:label>Carry Forward Losses</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Depreciation" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Depreciation amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.Depreciation</lx:path>
                                                      <lx:label>Depreciation</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="HomeOffice" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Home Office amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.HomeOffice</lx:path>
                                                      <lx:label>Home Office</lx:label>
                                                      <li:transactions>CNZ</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Interest" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Interest amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.Interest</lx:path>
                                                      <lx:label>Interest</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Lease" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Lease amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.Lease</lx:path>
                                                      <lx:label>Lease</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="NonCashBenefits" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Non-cash benefits amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.NonCashBenefits</lx:path>
                                                      <lx:label>Non Cash Benefits</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="NonRecurringExpenses" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Non-recurring expense amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.NonRecurringExpenses</lx:path>
                                                      <lx:label>Non Recurring Expenses</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Salary" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Salary amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.Salary</lx:path>
                                                      <lx:label>Salary</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="SuperannuationExcess" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Excess superannuation amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.Addback.SuperannuationExcess</lx:path>
                                                      <lx:label>Superannuation Excess</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                              </xs:complexType>
                                            </xs:element>
                                            <xs:element name="ForeignSourcedIncome" minOccurs="0" maxOccurs="1">
                                              <xs:annotation>
                                                <xs:documentation>Container element for any foreign sourced income</xs:documentation>
                                                <xs:appinfo>
                                                  <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.ForeignSourcedIncome</lx:path>
                                                  <lx:label>Foreign Sourced Income</lx:label>
                                                  <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                </xs:appinfo>
                                              </xs:annotation>
                                              <xs:complexType>
                                                <xs:attribute name="AUDAmount" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>The amount of foreign sourced income in Australian Dollars</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.ForeignSourcedIncome.AUDAmount</lx:path>
                                                      <lx:label>AUD Amount</lx:label>
                                                      <li:transactions>CAL,CDA,DAS</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="NZDAmount" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>The amount of foreign sourced income in New Zealand Dollars</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.ForeignSourcedIncome.NZDAmount</lx:path>
                                                      <lx:label>NZD Amount</lx:label>
                                                      <li:transactions>CNZ</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="PrimaryForeignCurrency" type="currencyCodeList" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>The original currency that the income was earned in. For example US Dollars, British Pounds</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.ForeignSourcedIncome.PrimaryForeignCurrency</lx:path>
                                                      <lx:label>Primary Foreign Currency</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                              </xs:complexType>
                                            </xs:element>
                                          </xs:sequence>
                                          <xs:attribute name="EndDate" type="dateType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The end date for the stated income, generally the end of a financial year</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.EndDate</lx:path>
                                                <lx:label>End Date</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="HasForeignSourcedIncome" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Has the applicant received any income from foreign sources</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.HasForeignSourcedIncome</lx:path>
                                                <lx:label>Has Foreign Sourced Income</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="IncomeGreaterThanPreviousYear" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>When comparing past Tax Office Assessments, indicate whether the past return has improved on the previous year.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.IncomeGreaterThanPreviousYear</lx:path>
                                                <lx:label>Income Greater Than Previous Year</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProfitAfterTax" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the profit after tax for the specified financial year.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.ProfitAfterTax</lx:path>
                                                <lx:label>Profit After Tax</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProfitBeforeTax" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the net profit/profit before tax for the specified financial year.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.ProfitBeforeTax</lx:path>
                                                <lx:label>Profit Before Tax</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProofCode" type="proofCodeSelfEmployedList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Code to indicate the type of evidence of the income</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.ProofCode</lx:path>
                                                <lx:label>Proof Code</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProofSighted" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Has the Income evidence been sighted by the loan writer or assessor</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.ProofSighted</lx:path>
                                                <lx:label>Proof Sighted</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="StartDate" type="dateType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The start date for the stated income, generally the start of a financial year</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.StartDate</lx:path>
                                                <lx:label>Start Date</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="TaxOfficeAssessments" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Are Tax Office Assessments for at least the past two years available as supporting evidence for this application where the applicant is self-employed.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.TaxOfficeAssessments</lx:path>
                                                <lx:label>Tax Office Assessments</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="x_Accountant" type="referenceType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Link to the person's accountant for the full most recent financial year. This should be populated when the accountant for the specified period is different to that declared as the current accountant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeRecent.x_Accountant</lx:path>
                                                <lx:label>Accountant</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                <lx:references>
                                                  <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                                </lx:references>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="BusinessIncomeYearToDate" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Details of the current year to date share of the gross (pre-tax) income or income-related amounts, attributable to this applicant.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate</lx:path>
                                            <lx:label>Business Income Year To Date</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:sequence>
                                            <xs:element name="Addback" minOccurs="0" maxOccurs="1">
                                              <xs:annotation>
                                                <xs:documentation>Container element for Year To Date Business Income Addback</xs:documentation>
                                                <xs:appinfo>
                                                  <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback</lx:path>
                                                  <lx:label>Addback</lx:label>
                                                  <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                </xs:appinfo>
                                              </xs:annotation>
                                              <xs:complexType>
                                                <xs:sequence>
                                                  <xs:element name="OtherAddback" minOccurs="0" maxOccurs="unbounded">
                                                    <xs:annotation>
                                                      <xs:documentation>Other addback that cannot otherwise be captured under a specific addback category.</xs:documentation>
                                                      <xs:appinfo>
                                                        <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.OtherAddback</lx:path>
                                                        <lx:label>Other Add Back</lx:label>
                                                        <li:transactions>CAL,CDA,SVC</li:transactions>
                                                      </xs:appinfo>
                                                    </xs:annotation>
                                                    <xs:complexType>
                                                      <xs:attribute name="Amount" type="currencyType" use="optional">
                                                        <xs:annotation>
                                                          <xs:documentation>Amount that can be an addback to the income.</xs:documentation>
                                                          <xs:appinfo>
                                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.OtherAddback.Amount</lx:path>
                                                            <lx:label>Amount</lx:label>
                                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                                          </xs:appinfo>
                                                        </xs:annotation>
                                                      </xs:attribute>
                                                      <xs:attribute name="Description" type="stringType" use="optional">
                                                        <xs:annotation>
                                                          <xs:documentation>Decription of the other addback.</xs:documentation>
                                                          <xs:appinfo>
                                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.OtherAddback.Description</lx:path>
                                                            <lx:label>Description</lx:label>
                                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                                          </xs:appinfo>
                                                        </xs:annotation>
                                                      </xs:attribute>
                                                    </xs:complexType>
                                                  </xs:element>
                                                </xs:sequence>
                                                <xs:attribute name="Allowances" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Allowances amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.Allowances</lx:path>
                                                      <lx:label>Allowances</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="AmortisationOfGoodwill" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Amortisation Of Goodwill amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.AmortisationOfGoodwill</lx:path>
                                                      <lx:label>Amortisation Of Goodwill</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Bonus" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Bonus amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.Bonus</lx:path>
                                                      <lx:label>Bonus</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="CarExpense" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Car expense amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.CarExpense</lx:path>
                                                      <lx:label>Car Expense</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="CarryForwardLosses" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Carry Forward Losses amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.CarryForwardLosses</lx:path>
                                                      <lx:label>Carry Forward Losses</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Depreciation" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Depreciation amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.Depreciation</lx:path>
                                                      <lx:label>Depreciation</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="HomeOffice" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Home Office amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.HomeOffice</lx:path>
                                                      <lx:label>Home Office</lx:label>
                                                      <li:transactions>CNZ</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Interest" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Interest amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.Interest</lx:path>
                                                      <lx:label>Interest</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Lease" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Lease amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.Lease</lx:path>
                                                      <lx:label>Lease</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="NonCashBenefits" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Non-cash benefits amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.NonCashBenefits</lx:path>
                                                      <lx:label>Non Cash Benefits</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="NonRecurringExpenses" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Non-recurring expense amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.NonRecurringExpenses</lx:path>
                                                      <lx:label>Non Recurring Expenses</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="Salary" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Salary amount that can be an addback to the income.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.Salary</lx:path>
                                                      <lx:label>Salary</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="SuperannuationExcess" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Excess superannuation amount that can be an addback to the income</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.Addback.SuperannuationExcess</lx:path>
                                                      <lx:label>Superannuation Excess</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                              </xs:complexType>
                                            </xs:element>
                                            <xs:element name="ForeignSourcedIncome" minOccurs="0" maxOccurs="1">
                                              <xs:annotation>
                                                <xs:documentation>Container element for any foreign sourced income</xs:documentation>
                                                <xs:appinfo>
                                                  <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.ForeignSourcedIncome</lx:path>
                                                  <lx:label>Foreign Sourced Income</lx:label>
                                                  <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                </xs:appinfo>
                                              </xs:annotation>
                                              <xs:complexType>
                                                <xs:attribute name="AUDAmount" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>The amount of foreign sourced income in Australian Dollars</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.ForeignSourcedIncome.AUDAmount</lx:path>
                                                      <lx:label>AUD Amount</lx:label>
                                                      <li:transactions>CAL,CDA,DAS</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="NZDAmount" type="currencyType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>The amount of foreign sourced income in New Zealand Dollars</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.ForeignSourcedIncome.NZDAmount</lx:path>
                                                      <lx:label>NZD Amount</lx:label>
                                                      <li:transactions>CNZ</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                                <xs:attribute name="PrimaryForeignCurrency" type="currencyCodeList" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>The original currency that the income was earned in. For example US Dollars, British Pounds.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.ForeignSourcedIncome.PrimaryForeignCurrency</lx:path>
                                                      <lx:label>Primary Foreign Currency</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                              </xs:complexType>
                                            </xs:element>
                                          </xs:sequence>
                                          <xs:attribute name="EndDate" type="dateType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The end date for the stated income, generally the end of a financial year</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.EndDate</lx:path>
                                                <lx:label>End Date</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="HasForeignSourcedIncome" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Has the applicant received any income from foreign sources</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.HasForeignSourcedIncome</lx:path>
                                                <lx:label>Has Foreign Sourced Income</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProfitAfterTax" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the profit after tax for the specified financial year.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.ProfitAfterTax</lx:path>
                                                <lx:label>Profit After Tax</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ProfitBeforeTax" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Captures the net profit/profit before tax for the specified financial year.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.ProfitBeforeTax</lx:path>
                                                <lx:label>Profit Before Tax</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="StartDate" type="dateType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The start date for the stated income, generally the start of a financial year</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.BusinessIncomeYearToDate.StartDate</lx:path>
                                                <lx:label>Start Date</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="DeclaredIncome" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Wrapper element for the amount of income declared by the applicant. Used for Low Doc applications when full financial years for income are not provided.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.DeclaredIncome</lx:path>
                                            <lx:label>Declared Income</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="IncomeAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The amount of income declared by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.DeclaredIncome.IncomeAmount</lx:path>
                                                <lx:label>Income Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NetIncomeAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The net amount of income declared by the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.DeclaredIncome.NetIncomeAmount</lx:path>
                                                <lx:label>Net Income Amount</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="Duration" type="durationType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Container element for capturing the duration of employment.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Duration</lx:path>
                                            <lx:label>Duration</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="FinancialAnalysis" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The detailed financial analysis of a self employed applicant's company</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.FinancialAnalysis</lx:path>
                                            <lx:label>Financial Analysis</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:sequence>
                                            <xs:element name="CompanyFinancials" minOccurs="0" maxOccurs="unbounded">
                                              <xs:annotation>
                                                <xs:documentation>Container for the reference to the financial data</xs:documentation>
                                                <xs:appinfo>
                                                  <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.FinancialAnalysis.CompanyFinancials</lx:path>
                                                  <lx:label>Company Financials</lx:label>
                                                  <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                </xs:appinfo>
                                              </xs:annotation>
                                              <xs:complexType>
                                                <xs:attribute name="x_CompanyFinancials" type="referenceType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Reference to the element containing the financial analysis for each financial year, to be provided when Complete Financial Analysis = "Yes"</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.FinancialAnalysis.CompanyFinancials.x_CompanyFinancials</lx:path>
                                                      <lx:label>Company Financials</lx:label>
                                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                                      <lx:references>
                                                        <lx:target>Package.Content.Application.CompanyFinancials</lx:target>
                                                      </lx:references>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                              </xs:complexType>
                                            </xs:element>
                                          </xs:sequence>
                                          <xs:attribute name="AnnualPaymentOnCommitments" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Total Principle and Interest payments on borrowings due to be paid in the next 12 months</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.FinancialAnalysis.AnnualPaymentOnCommitments</lx:path>
                                                <lx:label>Annual Payment On Commitments</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CompleteFinancialAnalysis" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Flag to indicate whether a complete financial analysis will be provided electronically, or only a summary view</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.FinancialAnalysis.CompleteFinancialAnalysis</lx:path>
                                                <lx:label>Complete Financial Analysis</lx:label>
                                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="ANZSCOOccupationCode" type="anzscoOccupationCodeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The ANZSCO occupation code specifying the person's role or occupation in this employment. This attribute is restricted to the list of ANZSCO occupation codes published by the ABS on 26/03/2013 that superseded the ASCO Occupation Codes, although the superseded codes still remain in common use.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.ANZSCOOccupationCode</lx:path>
                                          <lx:label>ANZSCO Occupation Code</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AverageHoursPerWeek" type="decimalType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>When the applicant is employed on a part time or casual basis the average hours per week need to be collected for serviceability calculations.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.AverageHoursPerWeek</lx:path>
                                          <lx:label>Average Hours Per Week</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Basis" type="selfEmployedBasisList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should indicate whether the employment type is fulltime, part-time or casual.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Basis</lx:path>
                                          <lx:label>Basis</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="EndDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The end date indicates when the applicant finalised their employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.EndDate</lx:path>
                                          <lx:label>End Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IncomeStatusOnOrBeforeSettlement" type="incomeStatusOnOrBeforeSettlementList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The income status on or before settlement.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.IncomeStatusOnOrBeforeSettlement</lx:path>
                                          <lx:label>Income Status On Or Before Settlement</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Occupation" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Text description of the person's role or occupation for this employment</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Occupation</lx:path>
                                          <lx:label>Occupation</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OccupationCode" type="absOccupationCodeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>An ABS occupation code specifying the person's role or occupation in this employment. This attribute remains but is likely to be deprecated in the future, with the ANZSCOOccupationCode attribute being restricted to the more recent ANZSCO Occupation Codes published by the ABS on 26/03/2013.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.OccupationCode</lx:path>
                                          <lx:label>Occupation Code</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The order of this self employed within the sequence.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.SequenceNumber</lx:path>
                                          <lx:label>Sequence Number</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StartDate" type="dateType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The start date indicates when the applicant commenced their employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.StartDate</lx:path>
                                          <lx:label>Start Date</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Status" type="employmentStatusList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The current or previous status of the employment</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.Status</lx:path>
                                          <lx:label>Status</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Accountant" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Link to the Applicant's accountant</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.x_Accountant</lx:path>
                                          <lx:label>Accountant</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Employer" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant should capture their company details for this instance of self-employment.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.Employment.SelfEmployed.x_Employer</lx:path>
                                          <lx:label>Employer</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The order of this employment within the sequence.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Employment.SequenceNumber</lx:path>
                                    <lx:label>Sequence Number</lx:label>
                                    <li:transactions>CAL,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ExistingCustomer" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Indicates whether the applicant is an existing customer of the lender being applied to.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.ExistingCustomer</lx:path>
                                <lx:label>Existing Customer</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="AccountNumber" type="financialAccountType" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Account details for an account held that identifies the party as an existing customer</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.ExistingCustomer.AccountNumber</lx:path>
                                      <lx:label>Account Number</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="CustomerNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A unique number assigned by the lender used to identify the customer</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ExistingCustomer.CustomerNumber</lx:path>
                                    <lx:label>Customer Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CustomerSince" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The approximate year that the applicant became a customer</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ExistingCustomer.CustomerSince</lx:path>
                                    <lx:label>Customer Since</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CustomerSinceMonth" type="mmMonthList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The month (if known) that the person applicant became a customer (only if the year is also provided in 'CustomerSince').</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ExistingCustomer.CustomerSinceMonth</lx:path>
                                    <lx:label>Customer Since Month</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="FinancialSituationCheck" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Financial situation check to verify the applicant's financial situation information is accurate.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.FinancialSituationCheck</lx:path>
                                <lx:label>Financial Situation Check</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DataSource" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The data provider for the applicant's financial situation check request.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.FinancialSituationCheck.DataSource</lx:path>
                                    <lx:label>Data Source</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ReferenceNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The reference number for the applicant's financial situation check request.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.FinancialSituationCheck.ReferenceNumber</lx:path>
                                    <lx:label>Reference Number</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ForeignTaxAssociation" type="foreignTaxAssociationType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for Foreign Tax Associations.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.ForeignTaxAssociation</lx:path>
                                <lx:label>Foreign Tax Association</lx:label>
                                <li:transactions>CAL,CDA,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="IdentityCheck" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Identity check to verify the applicant's identity information is accurate.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.IdentityCheck</lx:path>
                                <lx:label>Identity Check</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DataSource" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The data provider for the applicant's identity check request.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.IdentityCheck.DataSource</lx:path>
                                    <lx:label>Data Source</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Date" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date when the identity check request was sent to the provider.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.IdentityCheck.Date</lx:path>
                                    <lx:label>Date</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PoliticallyExposedPerson" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The person applicant's name matches a politically exposed person.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.IdentityCheck.PoliticallyExposedPerson</lx:path>
                                    <lx:label>Politically Exposed Person</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ReferenceNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The reference number for the applicant's identity check request.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.IdentityCheck.ReferenceNumber</lx:path>
                                    <lx:label>Reference Number</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Result" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The overall assessment result of the identity verification for the person applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.IdentityCheck.Result</lx:path>
                                    <lx:label>Result</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SanctionedPerson" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The person applicant's name matches a legitimate sanctioned person.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.IdentityCheck.SanctionedPerson</lx:path>
                                    <lx:label>Sanctioned Person</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="IndependentAdvice" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Independent party who applicant sought advice from.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.IndependentAdvice</lx:path>
                                <lx:label>Independent Advice</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Date" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date on which the independent advice was provided to the applicant by this adviser.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.IndependentAdvice.Date</lx:path>
                                    <lx:label>Date</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="independentAdviceTypeRequiredList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of independent advice that has been provided to the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.IndependentAdvice.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Address" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the address at which the advice was received.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.IndependentAdvice.x_Address</lx:path>
                                    <lx:label>Address</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Advisor" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the adviser that provided advice to the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.IndependentAdvice.x_Advisor</lx:path>
                                    <lx:label>Advisor</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Insurance" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for insurance information.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.Insurance</lx:path>
                                <lx:label>Insurance</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Insurance" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to Insurance details</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Insurance.x_Insurance</lx:path>
                                    <lx:label>Insurance</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Insurance</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="MaritalStatusDetails" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container for details related to the applicants marital status</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.MaritalStatusDetails</lx:path>
                                <lx:label>Marital Status Details</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="MaritalStatusChangeDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the date a person's marital status last changed</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.MaritalStatusDetails.MaritalStatusChangeDate</lx:path>
                                    <lx:label>Marital Status Change Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Spouse" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to the spouse of this applicant, generally either a co-applicant or a Related Party.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.MaritalStatusDetails.x_Spouse</lx:path>
                                    <lx:label>Spouse</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="NextOfKin" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about one of the applicant's next of kin.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.NextOfKin</lx:path>
                                <lx:label>Next Of Kin</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="KinRelationship" type="kinRelationshipList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the relationship the applicant has with the next of kin.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.NextOfKin.KinRelationship</lx:path>
                                    <lx:label>Kin Relationship</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Person" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to the applicant's next of kin</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.NextOfKin.x_Person</lx:path>
                                    <lx:label>Person</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="NominatedBorrower" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Joint Borrowers who reside at the same address may choose to nominate one of them as the person who will receive notices and other documents relating to the loan, so that they do not have to each receive their own copy of any notices and other documents</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.NominatedBorrower</lx:path>
                                <lx:label>Nominated Borrower</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Nominee" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to nominated borrower</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.NominatedBorrower.x_Nominee</lx:path>
                                    <lx:label>Nominee</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for the name of the person applicant.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.PersonName</lx:path>
                                <lx:label>Person Name</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="PowerOfAttorney" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for details of a power of attorney granted by the party</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.PowerOfAttorney</lx:path>
                                <lx:label>Power Of Attorney</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="DealingNumber" type="dealingNumberType" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Wrapper element for details regarding the Power of Attorney (POA) dealing including the registered number, book number or state in which the POA is in place.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.PowerOfAttorney.DealingNumber</lx:path>
                                      <lx:label>Dealing Number</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Type" type="powerOfAttorneyTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the type of power granted to the POA Holder.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.PowerOfAttorney.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_POAHolder" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Contains a link to the holder of the Power of Attorney to capture their name.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.PowerOfAttorney.x_POAHolder</lx:path>
                                    <lx:label>POA Holder</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PreviousName" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for the previous name of the applicant.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.PreviousName</lx:path>
                                <lx:label>Previous Name</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="FirstName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A person's first name.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.PreviousName.FirstName</lx:path>
                                    <lx:label>First Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MiddleNames" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A person's other name. Also referred to as their 'middle name'. For example - where someone's name is Frederick Joseph Gilbert - 'Joseph' would be considered their other name.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.PreviousName.MiddleNames</lx:path>
                                    <lx:label>Middle Names</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NameChangeReason" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The reason an applicant changed their previous name should be captured</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.PreviousName.NameChangeReason</lx:path>
                                    <lx:label>Name Change Reason</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NameTitle" type="nameTitleList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name title this person uses when they are formally addressed should be captured.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.PreviousName.NameTitle</lx:path>
                                    <lx:label>Name Title</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Surname" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A person's surname.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.PreviousName.Surname</lx:path>
                                    <lx:label>Surname</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Privacy" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Indicates whether the Party has agreed to allow disclosure and contains details of identity and credit checks.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.Privacy</lx:path>
                                <lx:label>Privacy</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AllowApplicationStatusUpdates" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant wants to receive updates on the status of the application directly from the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Privacy.AllowApplicationStatusUpdates</lx:path>
                                    <lx:label>Allow Application Status Updates</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AllowCreditBureauIdentityCheck" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant will allow a credit bureau identity check.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Privacy.AllowCreditBureauIdentityCheck</lx:path>
                                    <lx:label>Allow Credit Bureau Identity Check</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AllowCreditCheck" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether this Party will allow a credit check using their personal details as a reference.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Privacy.AllowCreditCheck</lx:path>
                                    <lx:label>Allow Credit Check</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AllowDirectMarketing" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether this Party will allow their personal details to be used for direct marketing purposes.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Privacy.AllowDirectMarketing</lx:path>
                                    <lx:label>Allow Direct Marketing</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AllowElectronicIdentityCheck" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant will allow an electronic check of their identity.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Privacy.AllowElectronicIdentityCheck</lx:path>
                                    <lx:label>Allow Electronic Identity Check</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AllowTelemarketing" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Each party may indicate if they are willing to be contacted for telemarketing</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Privacy.AllowTelemarketing</lx:path>
                                    <lx:label>Allow Telemarketing</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AllowThirdPartyDisclosure" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether this Party will allow their personal details to be disclosed to third parties affiliated with this lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Privacy.AllowThirdPartyDisclosure</lx:path>
                                    <lx:label>Allow Third Party Disclosure</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CreditAuthoritySigned" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates if the applicant has signed a Credit Authority form to request a credit report</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Privacy.CreditAuthoritySigned</lx:path>
                                    <lx:label>Credit Authority Signed</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DateSigned" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date on which the applicant signed the privacy consent/declaration that supports the responses provided to the privacy questions.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Privacy.DateSigned</lx:path>
                                    <lx:label>Date Signed</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PrivacyActConsentSigned" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates if the applicant has signed a Privacy Act consent form</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Privacy.PrivacyActConsentSigned</lx:path>
                                    <lx:label>Privacy Act Consent Signed</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ProofOfIdentity" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains details of a document used to verify identity.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity</lx:path>
                                <lx:label>Proof Of Identity</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="MedicareCard" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>A container for details specific to a Medicare Card</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.MedicareCard</lx:path>
                                      <lx:label>Medicare Card</lx:label>
                                      <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Colour" type="medicareCardColourList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The colour of the medicare card background.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.MedicareCard.Colour</lx:path>
                                          <lx:label>Colour</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IndividualReferenceNumber" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The reference number assigned to the individual using this medicare card as a form of identity.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.MedicareCard.IndividualReferenceNumber</lx:path>
                                          <lx:label>Individual Reference Number</lx:label>
                                          <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="AustralianStateOfIssue" type="auStateList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For Australian documents, the state of issue</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.AustralianStateOfIssue</lx:path>
                                    <lx:label>Australian State Of Issue</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CertifiedCopy" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether a certified copy of the proof of identity type has been provided.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.CertifiedCopy</lx:path>
                                    <lx:label>Certified Copy Sighted</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CountryOfIssue" type="countryCodeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The country from which the ID was issued</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.CountryOfIssue</lx:path>
                                    <lx:label>Country Of Issue</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DateDocumentVerified" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the date when the document was verified.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.DateDocumentVerified</lx:path>
                                    <lx:label>Date Document Verified</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DateOfIssue" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date of issue or start date of the ID</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.DateOfIssue</lx:path>
                                    <lx:label>Date Of Issue</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DOBVerified" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate that the applicants date of birth is verified by the document</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.DOBVerified</lx:path>
                                    <lx:label>DOB Verified</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DocumentCategory" type="documentCategoryList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indication of whether the document is photographic, non photographic or secondary</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.DocumentCategory</lx:path>
                                    <lx:label>Document Category</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DocumentNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The identifying number of the document</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.DocumentNumber</lx:path>
                                    <lx:label>Document Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DocumentType" type="documentTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of supporting document provided to verify the Party's identity.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.DocumentType</lx:path>
                                    <lx:label>Document Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExpiryDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The end date or expiry date of the ID</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.ExpiryDate</lx:path>
                                    <lx:label>Expiry Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsPreviousNameIdentification" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate that the document identifies the party by their previous name</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.IsPreviousNameIdentification</lx:path>
                                    <lx:label>Is Previous Name Identification</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IssuingOrganisation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The organisation that issued the ID</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.IssuingOrganisation</lx:path>
                                    <lx:label>Issuing Organisation</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MiddleNameOnDocument" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The middle name of the individual as it appears on the form of identification. This data is sometimes required for bureau matching and for this reason may be captured separately to the full name that appears on the form of identification.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.MiddleNameOnDocument</lx:path>
                                    <lx:label>Middle Name On Document</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NameOnDocument" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the person for which the document is providing evidence of their identity</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.NameOnDocument</lx:path>
                                    <lx:label>Name On Document</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NameVerified" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate that the applicants name is verified by the document</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.NameVerified</lx:path>
                                    <lx:label>Name Verified</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Original" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether the document is an original</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.Original</lx:path>
                                    <lx:label>Original Sighted</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures a brief description of the proof of identity document if Document Type = 'Other'</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.OtherDescription</lx:path>
                                    <lx:label>Other Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PhotographVerified" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the photograph of the applicant on the provided proof of identity document has been verified.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.PhotographVerified</lx:path>
                                    <lx:label>Photograph Verified</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PlaceOfIssue" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Holds the city where the document was issued. For overseas documents, the place or state of issue.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.PlaceOfIssue</lx:path>
                                    <lx:label>Place Of Issue</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ResidentialAddressVerified" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate that the applicants residential is verified by the document</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.ResidentialAddressVerified</lx:path>
                                    <lx:label>Residential Address Verified</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The order of this proof of identity within the sequence.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.SequenceNumber</lx:path>
                                    <lx:label>Sequence Number</lx:label>
                                    <li:transactions>CAL,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SignatureVerified" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the signature of this Party has been verified by an approved representative of the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.SignatureVerified</lx:path>
                                    <lx:label>Signature Verified</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_LocationDocumentVerified" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the location where the document was verified.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.ProofOfIdentity.x_LocationDocumentVerified</lx:path>
                                    <lx:label>Location Document Verified</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ResponsibleLending" type="responsibleLendingType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Captures information about any significant changes in circumstances that will adversely affect the applicant's ability to make contracted loan repayments; the nature of the change, and how they will continue to make repayments.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.ResponsibleLending</lx:path>
                                <lx:label>Responsible Lending</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="SourceOfFunds" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The source of the applicant's funds to be used in this transaction.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.SourceOfFunds</lx:path>
                                <lx:label>Source Of Funds</lx:label>
                                <li:transactions>CAL,CDA,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Detail" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The details of the source of funds.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.SourceOfFunds.Detail</lx:path>
                                    <lx:label>Detail</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Percentage" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The percentage of the funds from this source.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.SourceOfFunds.Percentage</lx:path>
                                    <lx:label>Percentage</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="sourceOfFundsTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of source of funds.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.SourceOfFunds.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="SourceOfWealth" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The source of the applicant's entire body of wealth.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.SourceOfWealth</lx:path>
                                <lx:label>Source Of Wealth</lx:label>
                                <li:transactions>CAL,CDA,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Detail" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The details of the source of wealth.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.SourceOfWealth.Detail</lx:path>
                                    <lx:label>Detail</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Percentage" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The percentage of the wealth from this source.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.SourceOfWealth.Percentage</lx:path>
                                    <lx:label>Percentage</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="sourceOfFundsTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of source of wealth.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.SourceOfWealth.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="TaxDeclarationDetails" type="taxDeclarationDetailsType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Carries details relating to the local tax declarations that have been or are being made to the financial institution.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.TaxDeclarationDetails</lx:path>
                                <lx:label>Tax Declaration Details</lx:label>
                                <li:transactions>CAL,CDA,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="Will" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for details on the party's will</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.PersonApplicant.Will</lx:path>
                                <lx:label>Will</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Executor" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures information about the executor of the will.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Will.x_Executor</lx:path>
                                    <lx:label>Executor</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_WillHeldBy" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies details of the party that currently has custody of the will. This may be the same person as the Executor, or different to the Executor. When different to the Executor then the Executor should also be identified.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.PersonApplicant.Will.x_WillHeldBy</lx:path>
                                    <lx:label>Will Held By</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="ApplicantType" type="applicantTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Whether the loan party is an borrower or guarantor</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.ApplicantType</lx:path>
                              <lx:label>Applicant Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Citizenship" type="countryCodeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The country that the person is a citizen of.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.Citizenship</lx:path>
                              <lx:label>Citizenship</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CompanyDirector" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the person is a director of a company.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.CompanyDirector</lx:path>
                              <lx:label>Company Director</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CountryOfBirth" type="countryCodeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The country in which this applicant was born.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.CountryOfBirth</lx:path>
                              <lx:label>Country of Birth</lx:label>
                              <li:transactions>CAL,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CreditStatus" type="creditStatusList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Summarises whether there are any issues indicated the credit history of the applicant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.CreditStatus</lx:path>
                              <lx:label>Credit Status</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CustomerTypeCode" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A lender assigned code for the customer type.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.CustomerTypeCode</lx:path>
                              <lx:label>Customer Type Code</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CustomerTypeDescription" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A lender assigned description of the customer type.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.CustomerTypeDescription</lx:path>
                              <lx:label>Customer Type Description</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DateOfBirth" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the date of birth for the applicant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.DateOfBirth</lx:path>
                              <lx:label>Date Of Birth</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DiscussedWithBeneficiaries" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>For a reverse mortgage, each applicant should specify if they have discussed the loan application with their beneficiaries. Beneficiaries will be materially affected by an equity release loan so it is important to confirm that they have been consulted.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.DiscussedWithBeneficiaries</lx:path>
                              <lx:label>Discussed With Beneficiaries</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="EligibleForFHOG" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The person should indicate whether they are eligible for a First Home Owners Grants as per the state governments requirements.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.EligibleForFHOG</lx:path>
                              <lx:label>Eligible For FHOG</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="FirstHomeBuyer" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The person should indicate whether they are a first home buyer in accordance with the State Revenue Office's definition of a first home buyer.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.FirstHomeBuyer</lx:path>
                              <lx:label>First Home Buyer</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="FirstPropertyBuyer" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The borrower is a first property buyer.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.FirstPropertyBuyer</lx:path>
                              <lx:label>First Property Buyer</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Gender" type="genderList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates the gender of the applicant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.Gender</lx:path>
                              <lx:label>Gender</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="HasAppliedForAustralianCitizenship" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the person has applied for Australian Citizenship.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.HasAppliedForAustralianCitizenship</lx:path>
                              <lx:label>Has Applied For Australian Citizenship</lx:label>
                              <li:transactions>CAL,CDA,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="HasAppliedForPermanentResidencyVisa" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the person has applied for a Permanent Residency Visa.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.HasAppliedForPermanentResidencyVisa</lx:path>
                              <lx:label>Has Applied For Permanent Residency Visa</lx:label>
                              <li:transactions>CAL,CDA,DAS,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="HasAWill" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that the party's property is being dealt with under a will that is being executed by an Executor.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.HasAWill</lx:path>
                              <lx:label>Has A Will</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="HasPreviousName" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the person has a previous name</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.HasPreviousName</lx:path>
                              <lx:label>Has Previous Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Immigrant" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the person came to this country to take up permanent residence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.Immigrant</lx:path>
                              <lx:label>Immigrant</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ImmigrationDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date the person arrived in this country to take up permanent residency.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.ImmigrationDate</lx:path>
                              <lx:label>Immigration Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IndependentFinancialAdvice" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The person should indicate whether they have received independent financial advice during the loan process.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.IndependentFinancialAdvice</lx:path>
                              <lx:label>Independent Financial Advice</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IndependentLegalAdvice" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The person should indicate whether they have received independent legal advice during the loan process.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.IndependentLegalAdvice</lx:path>
                              <lx:label>Independent Legal Advice</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IRDNumber" type="irdNumberPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The IRD Number of this Applicant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.IRDNumber</lx:path>
                              <lx:label>IRD Number</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsExistingCustomer" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the applicant is an existing customer of the lender being applied to.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.IsExistingCustomer</lx:path>
                              <lx:label>Is Existing Customer</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsLenderStaff" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the applicant is classified as a staff member of the lender being applied to. This may indicate that the applicant is entitled to a staff offer on the loan, or that specific risk controls should be applied in processing the loan application.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.IsLenderStaff</lx:path>
                              <lx:label>Is Lender Staff</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="JointNomination" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate whether the borrower is nominating another borrower to receive notices jointly.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.JointNomination</lx:path>
                              <lx:label>Joint Nomination</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="JointStatementOfPosition" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate whether the applicant will file a joint statement of financial position with their spouse</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.JointStatementOfPosition</lx:path>
                              <lx:label>Joint Statement Of Position</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LocalAgentOfForeignCompany" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The applicant is a local agent of a foreign company.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.LocalAgentOfForeignCompany</lx:path>
                              <lx:label>Local Agent Of Foreign Company</lx:label>
                              <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="MaritalStatus" type="maritalStatusList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The marital status of the applicant should be indicated</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.MaritalStatus</lx:path>
                              <lx:label>Marital Status</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="MonthsInCurrentProfession" type="decimalType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The months a person has worked in their current profession should be captured.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.MonthsInCurrentProfession</lx:path>
                              <lx:label>Months In Current Profession</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="MothersMaidenName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The mother's maiden name should be captured where this is used by the lender for identification purposes.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.MothersMaidenName</lx:path>
                              <lx:label>Mothers Maiden Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NumberOfCreditHistoryIssues" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The number of credit history issues associated with this applicant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.NumberOfCreditHistoryIssues</lx:path>
                              <lx:label>Number Of Credit History Issues</lx:label>
                              <li:transactions>ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OnBehalfOfUnincorporatedAssociation" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The applicant is applying for the loan on behalf of an unincorporated association.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.OnBehalfOfUnincorporatedAssociation</lx:path>
                              <lx:label>On Behalf Of Unincorporated Association</lx:label>
                              <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PermanentResidencyDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The start date of permanent legal residence. When permanent residence was granted, or the person arrived in Australia, whichever is the later.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.PermanentResidencyDate</lx:path>
                              <lx:label>Permanent Residency Date</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PlaceOfBirth" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The city, town, village or isolated dwelling in which this applicant was born.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.PlaceOfBirth</lx:path>
                              <lx:label>Place of Birth</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="POAGranted" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to Indicate that the party has granted a power of attorney and that the agent will execute the loan documentation</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.POAGranted</lx:path>
                              <lx:label>POA Granted</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PrimaryApplicant" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the lender should consider the party to be the primary applicant when processing the loan application. The significance of the primary applicant to the application varies from lender to lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.PrimaryApplicant</lx:path>
                              <lx:label>Primary Applicant</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PrincipalForeignResidence" type="countryCodeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The country(ies) that the person is a resident of</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.PrincipalForeignResidence</lx:path>
                              <lx:label>Principal Foreign Residence</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RelationshipToPrimaryApplicant" type="kinRelationshipList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The relationship between this person and the primary applicant (specifically for guarantor).</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.RelationshipToPrimaryApplicant</lx:path>
                              <lx:label>Relationship To Primary Applicant</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ResidencyStatus" type="residencyStatusList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates the applicant's residency status as either a permanent resident, temporary resident or non-resident.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.ResidencyStatus</lx:path>
                              <lx:label>Residency Status</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this person applicant within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="TemporaryVisaExpiryDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date on which the applicant's temporary visa expires.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.TemporaryVisaExpiryDate</lx:path>
                              <lx:label>Temporary Visa Expiry Date</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UnderDuress" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>For a reverse mortgage, each applicant may confirm (yes/no), that their application was not made under duress (from children, siblings etc.)</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.UnderDuress</lx:path>
                              <lx:label>Under Duress</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UnderstandApplication" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>For a reverse mortgage, each applicant may confirm (yes/no) that they understand what they are applying for.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.UnderstandApplication</lx:path>
                              <lx:label>Understand Application</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="VisaSubclassCode" type="visaSubclassCodePattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The three digit code for the visa subclass held by this applicant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.VisaSubclassCode</lx:path>
                              <lx:label>Visa Subclass Code</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Accountant" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Link to the person's current accountant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.x_Accountant</lx:path>
                              <lx:label>Accountant</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Household" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Links to a Household element</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.x_Household</lx:path>
                              <lx:label>Household</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.Household</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_PersonalReference" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Link to a related party who is providing a personal reference.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.x_PersonalReference</lx:path>
                              <lx:label>Personal Reference</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Solicitor" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Link to the applicant's solicitor</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.x_Solicitor</lx:path>
                              <lx:label>Solicitor</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="YearsInCurrentProfession" type="decimalType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The years a person has worked in their current profession should be captured. Note: this is different to the number of years with a current or previous employer.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.PersonApplicant.YearsInCurrentProfession</lx:path>
                              <lx:label>Years In Current Profession</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="ProductPackage" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Information about packages of products that are wrapped up, often including discounts or extra features offered through the package.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.ProductPackage</lx:path>
                          <lx:label>Product Package</lx:label>
                          <li:transactions>CAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="CreditCard" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Information regarding the credit card that is to be incorporated into the product package.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.ProductPackage.CreditCard</lx:path>
                                <lx:label>Credit Card</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Features" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Details of features requested on the new or existing credit card.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.ProductPackage.CreditCard.Features</lx:path>
                                      <lx:label>Features</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="SupplementaryCardholder" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>Details of the supplementary cardholder to add to the existing credit card.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.ProductPackage.CreditCard.Features.SupplementaryCardholder</lx:path>
                                            <lx:label>Supplementary Cardholder</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="x_SupplementaryCardholder" type="referenceType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The supplementary cardholder to add.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.ProductPackage.CreditCard.Features.SupplementaryCardholder.x_SupplementaryCardholder</lx:path>
                                                <lx:label>Supplementary Cardholder</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                                <lx:references>
                                                  <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                                  <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                                </lx:references>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="CardUpgrade" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The customer has agreed to a card upgrade if they are eligible.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.ProductPackage.CreditCard.Features.CardUpgrade</lx:path>
                                          <lx:label>Card Upgrade</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Limit" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The limit that has been requested on the credit card by the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.ProductPackage.CreditCard.Features.Limit</lx:path>
                                          <lx:label>Limit</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="LoyaltyProgramNumber" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The customer's loyalty program number.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.ProductPackage.CreditCard.Features.LoyaltyProgramNumber</lx:path>
                                          <lx:label>Loyalty Program Number</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ProductCategory" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The product category of the credit card.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.ProductPackage.CreditCard.Features.ProductCategory</lx:path>
                                          <lx:label>Product Category</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ProductName" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The product name that describes the credit card.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.ProductPackage.CreditCard.Features.ProductName</lx:path>
                                          <lx:label>Product Name</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_PrimaryCardholder" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The primary cardholder of the new credit card.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.ProductPackage.CreditCard.Features.x_PrimaryCardholder</lx:path>
                                          <lx:label>Primary Cardholder</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="InclusionType" type="accountStatusList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The status of the credit card that is to be incorporated into this product package.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.ProductPackage.CreditCard.InclusionType</lx:path>
                                    <lx:label>Inclusion Type</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Number" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The credit card number of the existing card to add to the package.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.ProductPackage.CreditCard.Number</lx:path>
                                    <lx:label>Number</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Identifier for this element that is unique within this message.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.ProductPackage.CreditCard.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_CreditCard" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The existing credit card that is to be added to the package.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.ProductPackage.CreditCard.x_CreditCard</lx:path>
                                    <lx:label>Credit Card</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Liability</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="DepositAccount" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Information regarding the deposit account that is to be incorporated into the product package.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.ProductPackage.DepositAccount</lx:path>
                                <lx:label>Deposit Account</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Overdraft" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Details of overdraft features requested on the new or existing account.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.ProductPackage.DepositAccount.Overdraft</lx:path>
                                      <lx:label>Overdraft</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="ConsiderLowerOverdraftLimitIfNotEligible" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Whether the applicant would like a lower limit to be considered if not eligable for the requested limit.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.ProductPackage.DepositAccount.Overdraft.ConsiderLowerOverdraftLimitIfNotEligible</lx:path>
                                          <lx:label>Consider Lower Overdraft Limit If Not Eligible</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IsOverdraftUsedForTemporaryExpenses" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Whether the overdraft is to be used for temporary expenses only.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.ProductPackage.DepositAccount.Overdraft.IsOverdraftUsedForTemporaryExpenses</lx:path>
                                          <lx:label>Is Overdraft Used For Temporary Expenses</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Limit" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The overdraft limit that has been requested on the accountby the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.ProductPackage.DepositAccount.Overdraft.Limit</lx:path>
                                          <lx:label>Limit</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OptionCode" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Lender assigned code for describing the overdraft option selected.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.ProductPackage.DepositAccount.Overdraft.OptionCode</lx:path>
                                          <lx:label>Option Code</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OverdraftRequested" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Whether an overdraft has been requested on the deposit account.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.ProductPackage.DepositAccount.Overdraft.OverdraftRequested</lx:path>
                                          <lx:label>Overdraft Requested</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="AlreadyExists" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A flag to indicate whether the deposit account already exists.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.ProductPackage.DepositAccount.AlreadyExists</lx:path>
                                    <lx:label>Already Exists</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Identifier for this element that is unique within this message.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.ProductPackage.DepositAccount.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_DepositAccount" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The existing account that is to be added to the package.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.ProductPackage.DepositAccount.x_DepositAccount</lx:path>
                                    <lx:label>Deposit Account</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Liability" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Information regarding any existing liabilities that are to be incorporated into the product package.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.ProductPackage.Liability</lx:path>
                                <lx:label>Liability</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Identifier for this element that is unique within this message.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.ProductPackage.Liability.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Liability" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The existing liability that is to be added to the package.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.ProductPackage.Liability.x_Liability</lx:path>
                                    <lx:label>Liability</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Liability</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="LoanDetails" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Information regarding the new loans that are to be incorporated into the product package.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.ProductPackage.LoanDetails</lx:path>
                                <lx:label>Loan Details</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Identifier for this element that is unique within this message.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.ProductPackage.LoanDetails.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_LoanDetails" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The new loan that is to be added to the package.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.ProductPackage.LoanDetails.x_LoanDetails</lx:path>
                                    <lx:label>Loan Details</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="ExistingPackage" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicator of whether this package already exists.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ProductPackage.ExistingPackage</lx:path>
                              <lx:label>Existing Package</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this product package within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ProductPackage.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Identifier for this product package that is unique within this message.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ProductPackage.UniqueID</lx:path>
                              <lx:label>Preferred Contact</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="ProductSet" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>A set of products.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.ProductSet</lx:path>
                          <lx:label>Product Set</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contact details for the product set.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.ProductSet.Contact</lx:path>
                                <lx:label>Contact</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_MailingAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Mailing address of the product set.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.ProductSet.Contact.x_MailingAddress</lx:path>
                                    <lx:label>Mailing Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Product" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>A product that belongs to this set.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.ProductSet.Product</lx:path>
                                <lx:label>Product</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_FinancialProduct" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to a financial product that belongs to this set.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.ProductSet.Product.x_FinancialProduct</lx:path>
                                    <lx:label>Financial Product</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
				      <lx:target>Package.Content.Application.DepositAccountDetails</lx:target>
				      <lx:target>Package.Content.Application.Liability</lx:target>
				      <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                      <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="LodgementReferenceNumber" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A unique identifier assigned to the product set by the point of sale system provider.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ProductSet.LodgementReferenceNumber</lx:path>
                              <lx:label>Lodgement Reference Number</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this product set within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ProductSet.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ProductSet.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_PrimaryApplicant" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A cross reference to the primary applicant for this set (a Person Applicant element, Company Applicant element, or Trust Applicant element).</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.ProductSet.x_PrimaryApplicant</lx:path>
                              <lx:label>Primary Applicant</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              <lx:references>
			        <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="RealEstateAsset" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Contains information about an asset which is real property.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.RealEstateAsset</lx:path>
                          <lx:label>Real Estate Asset</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Amenity" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains information about an amenity and its relative position to the real estate asset.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.Amenity</lx:path>
                                <lx:label>Amenity</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AmenityType" type="amenityTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Type of amenity that is being specified.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Amenity.AmenityType</lx:path>
                                    <lx:label>Amenity Type</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the amenity.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Amenity.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DistanceFromProperty" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Distance of the amenity from the property, in kilometres.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Amenity.DistanceFromProperty</lx:path>
                                    <lx:label>Distance From Property</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="AncillaryFeature" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Any feature on the property apart from the main dwelling.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.AncillaryFeature</lx:path>
                                <lx:label>Ancillary Feature</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Condition" type="ratingList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Condition rating of the ancillary feature.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.AncillaryFeature.Condition</lx:path>
                                    <lx:label>Condition</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the ancillary feature attached to the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.AncillaryFeature.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="ancillaryFeatureTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of ancillary feature attached to the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.AncillaryFeature.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Commercial" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about real estate that is commercial.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.Commercial</lx:path>
                                <lx:label>Commercial</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Type" type="commercialTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the type of commercial property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Commercial.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ConstructionDetails" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information relating to the construction of a dwelling on a security property.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails</lx:path>
                                <lx:label>Construction Details</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="ConstructionStage" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Payments are made to the builder in stages under a standard building contract. Certain information may be captured about each construction stage. This container will occur for each stage under the contract.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.ConstructionStage</lx:path>
                                      <lx:label>Construction Stage</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="StageAmount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The amount that is to be paid under the contract at completion of the specified construction stage.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.ConstructionStage.StageAmount</lx:path>
                                          <lx:label>Stage Amount</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="StageDescription" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A description of the construction stage. Common stages include laying of slab, framing, lockup, completion.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.ConstructionStage.StageDescription</lx:path>
                                          <lx:label>Stage Description</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Unique Identifier for this Construction Stage Element</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.ConstructionStage.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="OwnerBuilderExperience" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Contains information about the experience of an owner builder.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.OwnerBuilderExperience</lx:path>
                                      <lx:label>Owner Builder Experience</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="CompletedSimilarProjectInLast2Years" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates whether the owner builder has completed a similar project within the last 2 years</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.OwnerBuilderExperience.CompletedSimilarProjectInLast2Years</lx:path>
                                          <lx:label>Completed Similar Project In Last 2 Years</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="MinimalExperience" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates whether the applicant, as an owner / builder, has minimal experience working in the building industry and this is their first time as an owner / builder</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.OwnerBuilderExperience.MinimalExperience</lx:path>
                                          <lx:label>Minimal Experience</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OtherExperience" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures a free text description of the owner builder's other building experience.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.OwnerBuilderExperience.OtherExperience</lx:path>
                                          <lx:label>Other Experience</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="WorksInBuildingIndustry" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates whether the applicant, as an owner / builder, has experience working in the building industry</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.OwnerBuilderExperience.WorksInBuildingIndustry</lx:path>
                                          <lx:label>Works In Building Industry</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="BuilderType" type="builderTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the builder is the owner or a licensed builder.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.BuilderType</lx:path>
                                    <lx:label>Builder Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="BuildPriceAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The contracted or estimated build price for the dwelling</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.BuildPriceAmount</lx:path>
                                    <lx:label>Build Price Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CheckCostComment" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free text field for the valuer to provide a comment if the check cost is significantly different to previously agreed upon construction cost. Part of valuer additional comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.CheckCostComment</lx:path>
                                    <lx:label>Check Cost Comment</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConstructionInsuranceHeld" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the licensed builder or owner builder holds a policy of insurance in relation to the construction works that are proposed to be completed. Work in the event of the death, disappearance or insolvency of the contractor.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.ConstructionInsuranceHeld</lx:path>
                                    <lx:label>Construction Insurance Held</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CopyBuildersLicenceHeld" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the owner of the property being constructed is in possession of a copy of the builders registration or building licence.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.CopyBuildersLicenceHeld</lx:path>
                                    <lx:label>Copy Builders Licence Held</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CopyHomeOwnersWarrantyCertificateHeld" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether a copy of the home owners warranty certificate is held</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.CopyHomeOwnersWarrantyCertificateHeld</lx:path>
                                    <lx:label>Copy Home Owners Warranty Certificate Held</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CouncilApprovalHeld" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the owner of the property being constructed is in possession of building plans for the dwelling to be constructed that have been formally approved by the council responsible for planning in the area in which the property is located.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.CouncilApprovalHeld</lx:path>
                                    <lx:label>Council Approval Held</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DetailedCostingsHeld" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the owner of the property is in possession of a detailed schedule listing the costings that make up the contract price.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.DetailedCostingsHeld</lx:path>
                                    <lx:label>Detailed Costings Held</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EstimatedCompletionDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date on which construction is expected to be completed.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.EstimatedCompletionDate</lx:path>
                                    <lx:label>Estimated Completion Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExpertAdviceRecommended" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the valuer recommends further advice from an expert to verify the check cost and/or cost of out of contract items.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.ExpertAdviceRecommended</lx:path>
                                    <lx:label>Expert Advice Recommended</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FixedPriceContract" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that contract to construct the dwelling is for a fixed price amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.FixedPriceContract</lx:path>
                                    <lx:label>Fixed Price Contract</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InitialAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount of the loan funds that are to be initially advanced by the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.InitialAmount</lx:path>
                                    <lx:label>Initial Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LandValue" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>In the case of a purchase involving a house and land package, the amount of the contract price that is attributed to the purchase of the land.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.LandValue</lx:path>
                                    <lx:label>Land Value</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LandValueBasis" type="landValueEstimateBasisList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the estimate was based on the Purchase Price or Estimated Market Value</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.LandValueBasis</lx:path>
                                    <lx:label>Land Value Basis</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OutOfContractItemsComment" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free text string for the valuer to provide a comment on out of contract items required during construction of the property. Part of valuer additional comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.OutOfContractItemsComment</lx:path>
                                    <lx:label>Out Of Contract Items Comment</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OutOfContractItemsCost" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The total cost of out of contract items required during construction of the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.OutOfContractItemsCost</lx:path>
                                    <lx:label>Out Of Contract Items Cost</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OutOfContractItemsRequired" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that out of contract items were used during the construction of the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.OutOfContractItemsRequired</lx:path>
                                    <lx:label>Out Of Contract Items</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SatisfactoryTOCValuationHeld" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Satisfactory 'tentative on completion' valuation held</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.SatisfactoryTOCValuationHeld</lx:path>
                                    <lx:label>Satisfactory TOC Valuation Held</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SignedContract" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the owner of the property is in possession of a signed contract for the proposed type of construction</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.SignedContract</lx:path>
                                    <lx:label>Signed Contract</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TenPercentOfLendingAsCash" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate if the customer has 10% of the loan amount as cash on hand to cover any discrepancies</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.TenPercentOfLendingAsCash</lx:path>
                                    <lx:label>Ten Percent Of Lending As Cash</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The total cost of the contract. Will be the total construction amount in the case of a building contract, or the full contract price for a house and land package</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.TotalAmount</lx:path>
                                    <lx:label>Total Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalCheckCostAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An estimated total cost of constructing the property according to the valuer, referred to as check cost or check amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.TotalCheckCostAmount</lx:path>
                                    <lx:label>Total Check Cost Amount</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="constructionTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the type of construction</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Builder" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Reference link to the Builder Contact details and used when Builder Type is not "Owner Builder"</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ConstructionDetails.x_Builder</lx:path>
                                    <lx:label>Builder</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ContractDetails" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the purchase/sale price as specified in the Contract of Sale</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails</lx:path>
                                <lx:label>Contract Details</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ArmsLengthTransaction" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether the transacting parties are arms length parties</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.ArmsLengthTransaction</lx:path>
                                    <lx:label>Arms Length Transaction</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ContractDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date of the Contract of Sale.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.ContractDate</lx:path>
                                    <lx:label>Contract Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ContractPriceAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the total purchase/sale price as specified in the Contract of Sale.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.ContractPriceAmount</lx:path>
                                    <lx:label>Contract Price Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ContractSighted" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the valuer has sighted the full contract of sale.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.ContractSighted</lx:path>
                                    <lx:label>Contract Sighted</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ContractSightedComment" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free text string for the valuer to provide a comment if a full contract of sale was not sighted. Part of valuer additional comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.ContractSightedComment</lx:path>
                                    <lx:label>Contract Sighted Comment</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DepositAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the total amount that is or was payable as a deposit for the security property being purchased.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.DepositAmount</lx:path>
                                    <lx:label>Deposit Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DepositPaid" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the amount of the purchase deposit that has been paid by the applicants as at the date the application is submitted.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.DepositPaid</lx:path>
                                    <lx:label>Deposit Paid</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DepositPercentageRequested" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the percentage of the purchase price that the applicant will be required to provide as a deposit to the vendor if they enter into a contract to purchase a security property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.DepositPercentageRequested</lx:path>
                                    <lx:label>Deposit Percentage Requested</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EarlyReleaseOfDepositAuthorityRequested" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The vendor has requested a Section 27 Confirmation Letter from the bank that provides the information required for a Section 27 Deposit Release Statement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.EarlyReleaseOfDepositAuthorityRequested</lx:path>
                                    <lx:label>Early Release Of Deposit Authority Requested</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EstimatedSettlementDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The purchase date of the property, or the proposed settlement date for a property being purchased or sold</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.EstimatedSettlementDate</lx:path>
                                    <lx:label>Estimated Settlement Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FinanceApprovalDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>When a Contract of Sale is signed in SA and QLD, Finance Date is the date by which the loan must get formal approval.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.FinanceApprovalDate</lx:path>
                                    <lx:label>Finance Approval Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LicencedRealEstateAgentContract" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the Contract of Sale uses a form that has been developed for use by real estate agents, as opposed to lawyers and conveyancers. Relevant in States where this practice has not been superseded by a standard form contract of sale.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.LicencedRealEstateAgentContract</lx:path>
                                    <lx:label>Licenced Real Estate Agent Contract</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TransferOfLandAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The value of the property, as used for the calculation of stamp duty</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.TransferOfLandAmount</lx:path>
                                    <lx:label>Transfer Of Land Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UnderContract" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the real estate asset is under contract.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.UnderContract</lx:path>
                                    <lx:label>Under Contract</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UnderOffer" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the real estate asset is under offer.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.UnderOffer</lx:path>
                                    <lx:label>Under Offer</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Vendor" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Reference link to the Settlement Vendor Contact details</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.ContractDetails.x_Vendor</lx:path>
                                    <lx:label>Vendor</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Encumbrance" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>An encumbrance is a burden, impairment or third party right over property, that often makes it less valuable or easy for the owner to deal with. There may be multiple encumbrances over a single piece of real property.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.Encumbrance</lx:path>
                                <lx:label>Encumbrance</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="InFavourOf" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>The party to whom the encumbrance is in favour.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RealEstateAsset.Encumbrance.InFavourOf</lx:path>
                                      <lx:label>In Favour Of</lx:label>
                                      <li:transactions>CAL,DAS,VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Capacity" type="encumbranceInFavourOfCapacityList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The capacity of the party to whom the encumbrance is in favour.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.Encumbrance.InFavourOf.Capacity</lx:path>
                                          <lx:label>Capacity</lx:label>
                                          <li:transactions>CAL,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Name" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The name of the party to whom the encumbrance is in favour.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.Encumbrance.InFavourOf.Name</lx:path>
                                          <lx:label>Name</lx:label>
                                          <li:transactions>CAL,DAS,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_InFavourOf" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A cross reference to the party to whom the encumbrance is in favour.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.Encumbrance.InFavourOf.x_InFavourOf</lx:path>
                                          <lx:label>In Favour Of</lx:label>
                                          <li:transactions>CAL,DAS,VAL</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                            <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the encumbrance that is being specified as applying to the real property. An encumbrance may be registered with the Titles Office, in which case the registered encumbrance number may be provided.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Encumbrance.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EncumbranceAffectsValue" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether the specified encumbrance affects property value.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Encumbrance.EncumbranceAffectsValue</lx:path>
                                    <lx:label>Encumbrance Affects Value</lx:label>
                                    <li:transactions>DAS,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EncumbranceAffectsValueComment" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free text field for the valuer to provide a comment if the encumbrance affects property value. Part of valuer additional comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Encumbrance.EncumbranceAffectsValueComment</lx:path>
                                    <lx:label>Encumbrance Affects Value Comment</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EncumbranceType" type="realEstateAssetEncumbranceList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of encumbrance that is being specified as applying to the real property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Encumbrance.EncumbranceType</lx:path>
                                    <lx:label>Encumbrance Type</lx:label>
                                    <li:transactions>CAL,DAS,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Priority" type="securityPriorityList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The priority of the encumbrance claim against the asset.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Encumbrance.Priority</lx:path>
                                    <lx:label>Priority</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegisteredNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number issued by the Titles Office to identify the encumbrance, if registered. May be provided if the encumbrance has been registered with the Titles Office.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Encumbrance.RegisteredNumber</lx:path>
                                    <lx:label>Registered Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegistrationDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date the encumbrance was registered with the Titles Office.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Encumbrance.RegistrationDate</lx:path>
                                    <lx:label>Registration Date</lx:label>
                                    <li:transactions>CAL,DAS,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unique Identifier for this Encumbrance Element</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Encumbrance.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="EstimatedValue" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains details about the estimated value of the property</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.EstimatedValue</lx:path>
                                <lx:label>Estimated Value</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="EstimateBasis" type="estimateBasisRealEstateAssetList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the estimate was based on the valuation of the applicant or a certified valuer.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.EstimatedValue.EstimateBasis</lx:path>
                                    <lx:label>Estimate Basis</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EstimatedCGTLiability" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The estimate of the owners liability to pay capital gains tax on disposal of the asset. Relevant to the lender in determining the net proceeds likely to be realised by the owner if they sell the asset for the estimated value.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.EstimatedValue.EstimatedCGTLiability</lx:path>
                                    <lx:label>Estimated CGT Liability</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EstimateMethodology" type="estimateMethodologyList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The methodology used for estimating the value.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.EstimatedValue.EstimateMethodology</lx:path>
                                    <lx:label>Estimate Methodology</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Value" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The estimated value of the real estate.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.EstimatedValue.Value</lx:path>
                                    <lx:label>Value</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ValuedDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date the valuation was carried out.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.EstimatedValue.ValuedDate</lx:path>
                                    <lx:label>Valued Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Valuer" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Reference to the valuation firm that provided a certified or electronic valuation</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.EstimatedValue.x_Valuer</lx:path>
                                    <lx:label>Valuer</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="FundsDisbursement" type="fundsDisbursementType" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Specification of how funds are to be distributed.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.FundsDisbursement</lx:path>
                                <lx:label>Funds Disbursement</lx:label>
                                <li:transactions>CAL,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="FutureRentalIncome" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Information about rental income forecast to be earned from the property after construction or renovations.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.FutureRentalIncome</lx:path>
                                <lx:label>Future Rental Income</lx:label>
                                <li:transactions>CAL,CDA,SVC,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="PropertyPart" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>A part of the property from which rental income is earned.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RealEstateAsset.FutureRentalIncome.PropertyPart</lx:path>
                                      <lx:label>Property Part</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Type" type="propertyPartTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The type of property part.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.FutureRentalIncome.PropertyPart.Type</lx:path>
                                          <lx:label>Type</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="GrossRentalAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The gross amount of rent that is forecast after the rental property has been constructed or renovated.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.FutureRentalIncome.GrossRentalAmount</lx:path>
                                    <lx:label>Gross Rental Amount</lx:label>
                                    <li:transactions>CAL,CDA,SVC,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GrossRentalFrequency" type="frequencyFullList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The frequency of forecast gross rental payments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.FutureRentalIncome.GrossRentalFrequency</lx:path>
                                    <lx:label>Gross Rental Frequency</lx:label>
                                    <li:transactions>CAL,CDA,SVC,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NetRentalAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The net amount of rent that is forecast after the rental property has been constructed or renovated.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.FutureRentalIncome.NetRentalAmount</lx:path>
                                    <lx:label>Net Rental Amount</lx:label>
                                    <li:transactions>CAL,CDA,SVC,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NetRentalFrequency" type="frequencyFullList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The frequency of forecast net rental payments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.FutureRentalIncome.NetRentalFrequency</lx:path>
                                    <lx:label>Net Rental Frequency</lx:label>
                                    <li:transactions>CAL,CDA,SVC,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date at which the future rental income is forecast to commence.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.FutureRentalIncome.StartDate</lx:path>
                                    <lx:label>Start Date</lx:label>
                                    <li:transactions>CAL,CDA,SVC,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Owner" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the party that owns the future rental income.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.FutureRentalIncome.x_Owner</lx:path>
                                    <lx:label>x_Owner</lx:label>
                                    <li:transactions>CAL,CDA,SVC,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="HeritageListing" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of any Heritage Listings that apply to the real estate asset.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.HeritageListing</lx:path>
                                <lx:label>Heritage Listing</lx:label>
                                <li:transactions>CAL,CDA,DAS,SVC,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="HeritageIssuesComment" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free text field for the valuer to provide a comment if the heritage listing of a property adversely or beneficially affects its value. Part of valuer additional comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.HeritageListing.HeritageIssuesComment</lx:path>
                                    <lx:label>Heritage Issues Comment</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="HeritageListed" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate that the property is heritage listed.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.HeritageListing.HeritageListed</lx:path>
                                    <lx:label>Heritage Listed</lx:label>
                                    <li:transactions>CAL,CDA,DAS,SVC,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Industrial" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The industrial type indicates whether the property is used for non-specialised industrial purposes or light industrial purposes.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.Industrial</lx:path>
                                <lx:label>Industrial</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Type" type="industrialTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The industrial type indicates whether the property is used for non-specialised industrial purposes or light industrial purposes.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Industrial.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Insurance" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for insurance information.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.Insurance</lx:path>
                                <lx:label>Insurance</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Insurance" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to Insurance details</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Insurance.x_Insurance</lx:path>
                                    <lx:label>Insurance</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Insurance</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PercentOwned" type="percentOwnedType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>An asset may be owned by one or more parties, typically the applicants, so there may be more than one owner referenced. An asset may also be partially owned by a non-applicant, such as an applicant's spouse, in which case the third party owner may be referenced, or less than 100% ownership of the asset may be assigned. If the asset is to be purchased, then this indicates the expected ownership arrangement post-settlement. If the percentage ownership is not shown, then ownership is assumed to be in equal parts.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.PercentOwned</lx:path>
                                <lx:label>Percent Owned</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="PropertyExpense" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Expense associated with an investment property.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.PropertyExpense</lx:path>
                                <lx:label>Property Expense</lx:label>
                                <li:transactions>ACC,CAL,CDA,DAS,CNZ,LMI,SVC,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Amount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Dollar amount of investment property expense.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyExpense.Amount</lx:path>
                                    <lx:label>Amount</lx:label>
                                    <li:transactions>ACC,CAL,CDA,DAS,CNZ,LMI,SVC,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Frequency" type="frequencyFullList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The frequency of the expense occurrence.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyExpense.Frequency</lx:path>
                                    <lx:label>Frequency</lx:label>
                                    <li:transactions>ACC,CAL,CDA,DAS,CNZ,LMI,SVC,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PropertyFeatures" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Container for information about the property's features.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures</lx:path>
                                <lx:label>Property Features</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="EssentialRepairs" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Any essential repairs needed on the dwelling.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.EssentialRepairs</lx:path>
                                      <lx:label>Essential Repairs</lx:label>
                                      <li:transactions>VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Cost" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Cost of the essential repairs on the dwelling.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.EssentialRepairs.Cost</lx:path>
                                          <lx:label>Cost</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Description" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A description of the essential repairs needed on the dwelling. Part of valuer additional comments</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.EssentialRepairs.Description</lx:path>
                                          <lx:label>Description</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ExpertAdviceRecommended" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates that further advice from an expert is recommended by the valuer.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.EssentialRepairs.ExpertAdviceRecommended</lx:path>
                                          <lx:label>Expert Advice Recommended</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="Fixture" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>A fixture that is attached to the property.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.Fixture</lx:path>
                                      <lx:label>Fixture</lx:label>
                                      <li:transactions>VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Description" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A description of the fixture attached to the property.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.Fixture.Description</lx:path>
                                          <lx:label>Description</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Type" type="fixtureTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The type of fixture attached to the property.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.Fixture.Type</lx:path>
                                          <lx:label>Type</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.Fixture.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="ParkingFacility" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>A parking facility on the property.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.ParkingFacility</lx:path>
                                      <lx:label>Parking Facility</lx:label>
                                      <li:transactions>VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Description" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A description of the parking facility. To be used when the type is 'Other'.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.ParkingFacility.Description</lx:path>
                                          <lx:label>Description</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="FloorArea" type="decimalType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The floor area of the parking facility in square meters.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.ParkingFacility.FloorArea</lx:path>
                                          <lx:label>Floor Area</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NumberOfCarSpaces" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The number of car spaces in the parking facility.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.ParkingFacility.NumberOfCarSpaces</lx:path>
                                          <lx:label>Number Of Car Spaces</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Type" type="parkingFacilityTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The type of parking facility.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.ParkingFacility.Type</lx:path>
                                          <lx:label>Type</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.ParkingFacility.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="Service" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>A service connected to the property.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.Service</lx:path>
                                      <lx:label>Service</lx:label>
                                      <li:transactions>VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Description" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A description of the service. To be used when type is 'Other'.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.Service.Description</lx:path>
                                          <lx:label>Description</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Type" type="serviceTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The type of service connected to the property.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.Service.Type</lx:path>
                                          <lx:label>Type</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.Service.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="AllPropertyFeatures" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A general description of the features of the property</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.AllPropertyFeatures</lx:path>
                                    <lx:label>All Property Features</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ArchitecturalStyle" type="architecturalStyleList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The architectural style of the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.ArchitecturalStyle</lx:path>
                                    <lx:label>Architectural Style</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ArchitecturalStyleDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free-text description of the architectural style of the property. Part of valuer additional comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.ArchitecturalStyleDescription</lx:path>
                                    <lx:label>Architectural Style Description</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="BuiltYear" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The year the property was built.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.BuiltYear</lx:path>
                                    <lx:label>Built Year</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="BuiltYearType" type="builtDateTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Describes if the year the dwelling was built is an estimate or an exact date.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.BuiltYearType</lx:path>
                                    <lx:label>Built Year Type</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ExternalCondition" type="ratingList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The external condition of the real estate asset.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.ExternalCondition</lx:path>
                                    <lx:label>External Condition</lx:label>
                                    <li:transactions>LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FloorArea" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The floor area of the building on the property, in sqm. Where is more than one dwelling or structure on the property, then the lender's policy will determine how the applicant is to calculate the floor area. For a residential property, generally speaking this is the living area.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.FloorArea</lx:path>
                                    <lx:label>Floor Area</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FloorMaterial" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the material of which the floors are made.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.FloorMaterial</lx:path>
                                    <lx:label>Floor Material</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FloorMaterialType" type="floorMaterialList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The main material of which the floors are made.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.FloorMaterialType</lx:path>
                                    <lx:label>Floor Material Type</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InteriorLayout" type="ratingList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The quality of the interior layout of the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.InteriorLayout</lx:path>
                                    <lx:label>Interior Layout</lx:label>
                                    <li:transactions>LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InteriorWallLiningMaterial" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the material of which the interior wall lining is made.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.InteriorWallLiningMaterial</lx:path>
                                    <lx:label>Interior Wall Lining Material</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InteriorWallLiningMaterialType" type="interiorWallLiningMaterialList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The main material of which the interior wall lining is made.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.InteriorWallLiningMaterialType</lx:path>
                                    <lx:label>Interior Wall Lining MaterialType</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="InternalCondition" type="ratingList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The internal condition of the real estate asset.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.InternalCondition</lx:path>
                                    <lx:label>Internal Condition</lx:label>
                                    <li:transactions>LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LandArea" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The land area of the property. The units of measurement must also be specified with this figure.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.LandArea</lx:path>
                                    <lx:label>Land Area</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LandAreaUnits" type="landAreaUnitsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The units of measurement for the stated land area</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.LandAreaUnits</lx:path>
                                    <lx:label>Land Area Units</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ModifiedYear" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Year the dwelling was last modified.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.ModifiedYear</lx:path>
                                    <lx:label>Modified Year</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ModifiedYearType" type="builtDateTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Describes if the year the dwelling was modified is an estimate or an exact date.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.ModifiedYearType</lx:path>
                                    <lx:label>Modified Year Type</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfBathrooms" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the number of bathrooms in the dwelling.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfBathrooms</lx:path>
                                    <lx:label>Number Of Bathrooms</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfBedrooms" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the number of bedrooms in the dwelling.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfBedrooms</lx:path>
                                    <lx:label>Number Of Bedrooms</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfCarSpaces" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the number of car spaces in the dwelling.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfCarSpaces</lx:path>
                                    <lx:label>Number Of Car Spaces</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfDiningRooms" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of dining rooms in the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfDiningRooms</lx:path>
                                    <lx:label>Number of Dining Rooms</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfDwellings" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the number of dwellings on the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfDwellings</lx:path>
                                    <lx:label>Number Of Dwellings</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfEnsuites" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of ensuites in the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfEnsuites</lx:path>
                                    <lx:label>Number of Ensuites</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfKitchens" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of kitchens in the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfKitchens</lx:path>
                                    <lx:label>Number of Kitchens</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfLaundries" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of laundries in the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfLaundries</lx:path>
                                    <lx:label>Number of Laundries</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfLivingRooms" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of living rooms in the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfLivingRooms</lx:path>
                                    <lx:label>Number of Living Rooms</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfOtherRooms" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of other rooms in the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfOtherRooms</lx:path>
                                    <lx:label>Number of Other Rooms</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfReceptions" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of receptions in the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfReceptions</lx:path>
                                    <lx:label>Number of Receptions</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfSaunas" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of saunas in the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfSaunas</lx:path>
                                    <lx:label>Number of Saunas</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfStorerooms" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of storerooms in the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfStorerooms</lx:path>
                                    <lx:label>Number of Storerooms</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfStudies" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of studies in the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfStudies</lx:path>
                                    <lx:label>Number of Studies</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfTheatreRooms" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of theatre rooms in the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfTheatreRooms</lx:path>
                                    <lx:label>Number of Theatre Rooms</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfWalkInLinens" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of walk in linens in the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfWalkInLinens</lx:path>
                                    <lx:label>Number of Walk in Linens</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfWalkInRobes" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of walk in robes in the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfWalkInRobes</lx:path>
                                    <lx:label>Number of Walk in Robes</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfWCs" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of WCs in the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.NumberOfWCs</lx:path>
                                    <lx:label>Number of WCs</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OutdoorArea" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The outdoor area of the building on the property, in sqm. Includes areas of the building that are not counted in the floor area, such as the garage or veranda.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.OutdoorArea</lx:path>
                                    <lx:label>Outdoor Area</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PoolType" type="poolTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the type of pool on the property</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.PoolType</lx:path>
                                    <lx:label>Pool Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PropertyDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the property</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.PropertyDescription</lx:path>
                                    <lx:label>Property Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PropertyID" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A unique identifier assigned to the property by the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.PropertyID</lx:path>
                                    <lx:label>Property ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PropertyImage" type="urlPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the location of the property's images</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.PropertyImage</lx:path>
                                    <lx:label>Property Image</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RoadElevationGrade" type="roadElevationGradeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The property's elevation grade from the road.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.RoadElevationGrade</lx:path>
                                    <lx:label>Road Elevation Grade</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RoofMaterial" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the material that the roof is constructed from.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.RoofMaterial</lx:path>
                                    <lx:label>Roof Material</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RoofMaterialType" type="roofMaterialList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The main material of which the roof of the property is made.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.RoofMaterialType</lx:path>
                                    <lx:label>Roof Material Type</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StreetAppeal" type="ratingList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the street appeal of the property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.StreetAppeal</lx:path>
                                    <lx:label>StreetAppeal</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="WallMaterial" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the type of walls of the property</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.WallMaterial</lx:path>
                                    <lx:label>Wall Material</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="WallMaterialType" type="wallMaterialList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The main material of which the walls of the property are made.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.WallMaterialType</lx:path>
                                    <lx:label>Wall Material Type</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="WindowMaterial" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the material of which the windows are made.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.WindowMaterial</lx:path>
                                    <lx:label>Window Material</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="WindowMaterialType" type="windowMaterialList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The main material of which the windows are made.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyFeatures.WindowMaterialType</lx:path>
                                    <lx:label>Window Material Type</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PropertySearchValidation" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The results returned by a third party property search service. These services can return reports for clients, value estimates, and other property data for the specified property address. Lenders may base their decision whether to order or skip a full valuation report using the data returned by the search service.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.PropertySearchValidation</lx:path>
                                <lx:label>Property Search Validation</lx:label>
                                <li:transactions>CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="OwnerNames" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The names of the property owners as returned by the third-party search service.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertySearchValidation.OwnerNames</lx:path>
                                    <lx:label>Owner Names</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Reason" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A short description that includes the reason why the Result value is Warning or Refer.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertySearchValidation.Reason</lx:path>
                                    <lx:label>Reason</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Result" type="propertySearchValidationResultList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The property search validation result.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertySearchValidation.Result</lx:path>
                                    <lx:label>Result</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="URL" type="urlPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The link to the report generated by the third-party search service for the current property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertySearchValidation.URL</lx:path>
                                    <lx:label>Property Type Name</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PropertyType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Captures the property type, like House, Unit, Villa etc.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.PropertyType</lx:path>
                                <lx:label>Property Type</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="CategoryTypeName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the property category type</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyType.CategoryTypeName</lx:path>
                                    <lx:label>Category Type Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PropertyTypeCode" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Contains information about the property type</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyType.PropertyTypeCode</lx:path>
                                    <lx:label>Property Type Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PropertyTypeName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the property type, like House, Unit, Villa etc.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.PropertyType.PropertyTypeName</lx:path>
                                    <lx:label>Property Type Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="RentalIncome" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Information about rental income earned from the property.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome</lx:path>
                                <lx:label>Rental Income</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="PropertyPart" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>A part of the property from which rental income is earned.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.PropertyPart</lx:path>
                                      <lx:label>Property Part</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Type" type="propertyPartTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The type of property part.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.PropertyPart.Type</lx:path>
                                          <lx:label>Type</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Basis" type="rentalIncomeBasisList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The basis on which the rental income amount was determined.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.Basis</lx:path>
                                    <lx:label>Basis</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Comment" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free text field for the valuer to provide a comment on the rental income of the property, especially with the addition of furniture. Part of valuer additional comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.Comment</lx:path>
                                    <lx:label>Comment</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConfidencePercentage" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The level of confidence that the valuation provider has in the estimated rental income amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.ConfidencePercentage</lx:path>
                                    <lx:label>Confidence Percentage</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EvidenceOfTenancy" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the applicant can provide the lender with evidence that the property is tenanted, typically in the form of a lease.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.EvidenceOfTenancy</lx:path>
                                    <lx:label>Evidence Of Tenancy</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Frequency" type="frequencyFullList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The frequency of rental payments gross amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.Frequency</lx:path>
                                    <lx:label>Gross Amount Frequency</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Furnished" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates if the rental property is furnished.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.Furnished</lx:path>
                                    <lx:label>Furnished</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncomeStatusOnOrBeforeSettlement" type="incomeStatusOnOrBeforeSettlementList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The income status on or before settlement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.IncomeStatusOnOrBeforeSettlement</lx:path>
                                    <lx:label>Income Status On Or Before Settlement</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NetRentalAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The net amount of rent that is paid under the lease. The frequency (weekly, monthly etc.) for the rental amount specified should also be explicitly specified.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.NetRentalAmount</lx:path>
                                    <lx:label>Net Rental Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NetRentalAmountFrequency" type="frequencyShortList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The frequency of rental payments net amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.NetRentalAmountFrequency</lx:path>
                                    <lx:label>Net Rental Amount Frequency</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RentalAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The gross amount of rent that is paid under the lease. The frequency (weekly, monthly etc.) for the rental amount specified should also be explicitly specified.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.RentalAmount</lx:path>
                                    <lx:label>Gross Rental Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ShortTermRentalAccommodation" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The rental property is considered to be short-term rental accommodation.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.ShortTermRentalAccommodation</lx:path>
                                    <lx:label>Short Term Rental Accommodation</lx:label>
                                    <li:transactions>CAL,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ShortTermRentalDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free text field to further describe the rental income accommodation arrangement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.ShortTermRentalDescription</lx:path>
                                    <lx:label>Short Term Rental Description</lx:label>
                                    <li:transactions>CAL,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ValuedDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date on which the rental income amount was determined.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.ValuedDate</lx:path>
                                    <lx:label>Valued Date</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Owner" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the party that owns the rental income.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RentalIncome.x_Owner</lx:path>
                                    <lx:label>x_Owner</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Representation" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details regarding the specification of an agent (solicitor or conveyancer) to act for the borrowers or vendors at settlement</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.Representation</lx:path>
                                <lx:label>Representation</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Conveyancer" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Identifies who will be representing the applicant(s) at settlement. Points to the Related Party acting as representative.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Representation.x_Conveyancer</lx:path>
                                    <lx:label>Conveyancer</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_VendorConveyancer" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Identifies who will be representing the vendor(s) at settlement. Points to the Related Party acting as representative.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Representation.x_VendorConveyancer</lx:path>
                                    <lx:label>Vendor Conveyancer</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Residential" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about real estate that is residential.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.Residential</lx:path>
                                <lx:label>Residential</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Type" type="residentialTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies what type of dwelling is, or is to be, located on the residential property. For an established or new dwelling, specifies what type of dwelling is currently located on the residential property. For an off-the-plan purchase or construction loan, specifies what type of dwelling will be constructed. If vacant property is notionally Residential but without an intention to build in the immediate future, then the type would remain Vacant Land</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Residential.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="WillOwn25PercentOfComplex" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the owners of the property own or will own following purchase of the subject property, 25% or more by number of the units in the complex in which the subject property is located.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Residential.WillOwn25PercentOfComplex</lx:path>
                                    <lx:label>Will Own 25 Percent Of Complex</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="WillOwn3UnitsInComplex" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the owners of the property own or will own following purchase of the subject property, 3 or more units in the complex in which the subject property is located.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Residential.WillOwn3UnitsInComplex</lx:path>
                                    <lx:label>Will Own 3 Units In Complex</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="RestrictionOnUseOfLand" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A restriction on the use of land by deed of agreement between the parties.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.RestrictionOnUseOfLand</lx:path>
                                <lx:label>Restriction On Use Of Land</lx:label>
                                <li:transactions>CAL,CDA,DAS,SVC,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the covenant or restriction on use of land.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RestrictionOnUseOfLand.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegistrationNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A registration number to identify the covenant or restriction on use of land.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RestrictionOnUseOfLand.RegistrationNumber</lx:path>
                                    <lx:label>Registration Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RestrictionExists" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether a covenant, or restriction on use of land, exists.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.RestrictionOnUseOfLand.RestrictionExists</lx:path>
                                    <lx:label>Restriction Exists</lx:label>
                                    <li:transactions>CAL,CDA,DAS,SVC,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Rural" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about real estate that is rural.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.Rural</lx:path>
                                <lx:label>Rural</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="NumberOfBuildings" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of buildings on the rural property should be indicated.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Rural.NumberOfBuildings</lx:path>
                                    <lx:label>Number Of Buildings</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="ruralTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The size of the rural property should be indicated.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Rural.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Usage" type="ruralUsageList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The purpose of the rural property and how it is used should be indicated.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Rural.Usage</lx:path>
                                    <lx:label>Usage</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="SalesRecord" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>A record of a previous sale of the property.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.SalesRecord</lx:path>
                                <lx:label>Sales Record</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="SaleComment" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free text field for the valuer to provide a comment about the current sale if the sale is not considered to be in line with the market or is known to be affected by special circumstances or incentives. Part of valuer additional comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SalesRecord.SaleComment</lx:path>
                                    <lx:label>Sale Comment</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SaleDate" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date on which the property was previously sold.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SalesRecord.SaleDate</lx:path>
                                    <lx:label>Sale Date</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SalePrice" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The price for which the property was previously sold.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SalesRecord.SalePrice</lx:path>
                                    <lx:label>Sale Price</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="SimilarSalesEvidence" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Sales evidence on a property similar to the real property.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.SimilarSalesEvidence</lx:path>
                                <lx:label>Similar Sales Evidence</lx:label>
                                <li:transactions>LMI,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ConditionComparision" type="comparisonList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comparision of the condition of the similar property with the real property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SimilarSalesEvidence.ConditionComparision</lx:path>
                                    <lx:label>Condition Comparision</lx:label>
                                    <li:transactions>LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LandComparision" type="comparisonList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comparision of the land of the similar property with the real property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SimilarSalesEvidence.LandComparision</lx:path>
                                    <lx:label>Land Comparision</lx:label>
                                    <li:transactions>LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LastSaleDate" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date the similar property was previously sold on.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SimilarSalesEvidence.LastSaleDate</lx:path>
                                    <lx:label>LastSaleDate</lx:label>
                                    <li:transactions>LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LocationComparision" type="comparisonList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Comparision of the location of the similar property with the real property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SimilarSalesEvidence.LocationComparision</lx:path>
                                    <lx:label>Location Comparision</lx:label>
                                    <li:transactions>LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OverallComparision" type="comparisonList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Overall comparision of the similar property with the real property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SimilarSalesEvidence.OverallComparision</lx:path>
                                    <lx:label>Overall Comparision</lx:label>
                                    <li:transactions>LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SaleComment" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free text field for the valuer to provide a comment about the sale. Part of valuer additional comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SimilarSalesEvidence.SaleComment</lx:path>
                                    <lx:label>Sale Comment</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SalePrice" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The sale price of the similar property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SimilarSalesEvidence.SalePrice</lx:path>
                                    <lx:label>SalePrice</lx:label>
                                    <li:transactions>LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Address" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The address of the similar property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SimilarSalesEvidence.x_Address</lx:path>
                                    <lx:label>Address</lx:label>
                                    <li:transactions>LMI,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="SiteDimensions" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The dimensions of the real estate asset.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.SiteDimensions</lx:path>
                                <lx:label>Site Dimensions</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Depth" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The depth of the real estate asset in metres.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SiteDimensions.Depth</lx:path>
                                    <lx:label>Depth</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free-text description of the dimensions of a real estate asset.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SiteDimensions.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Frontage" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The frontage of the real estate asset in metres.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SiteDimensions.Frontage</lx:path>
                                    <lx:label>Frontage</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IrregularShape" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicate that the real estate asset is of an irregular shape.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.SiteDimensions.IrregularShape</lx:path>
                                    <lx:label>Irregular Shape</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Title" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains information relating to the title details for the real property.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.Title</lx:path>
                                <lx:label>Title</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="RegisteredProprietor" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>A registered proprietor on the title.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RealEstateAsset.Title.RegisteredProprietor</lx:path>
                                      <lx:label>Registered Proprietor</lx:label>
                                      <li:transactions>CAL,CDA,DAS</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="ChangeNameFormRequired" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A change of name form is required for the title.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.Title.RegisteredProprietor.ChangeNameFormRequired</lx:path>
                                          <lx:label>Change Name Form Required</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Name" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The name of the registered proprietor as it appears on the title.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.Title.RegisteredProprietor.Name</lx:path>
                                          <lx:label>Name</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_RegisteredProprietor" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A cross reference to the registered proprietor (a Person Applicant element or Company Applicant element).</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.Title.RegisteredProprietor.x_RegisteredProprietor</lx:path>
                                          <lx:label>Registered Proprietor</lx:label>
                                          <li:transactions>CAL,CDA,DAS</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                            <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="UnitDetails" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Contains information relating to entitlement details for the unit.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RealEstateAsset.Title.UnitDetails</lx:path>
                                      <lx:label>Unit Details</lx:label>
                                      <li:transactions>VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="SubjectUnitEntitlement" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Share in the total unit entitlement for this unit.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.Title.UnitDetails.SubjectUnitEntitlement</lx:path>
                                          <lx:label>Subject Unit Entitlement</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TotalUnitEntitlement" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Total entitlement for the parcel of land this unit is part of.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.Title.UnitDetails.TotalUnitEntitlement</lx:path>
                                          <lx:label>Total Unit Entitlement</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TotalUnitsInDevelopment" type="integerType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Total number of units in the strata plan for the land.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RealEstateAsset.Title.UnitDetails.TotalUnitsInDevelopment</lx:path>
                                          <lx:label>Total Units In Development</lx:label>
                                          <li:transactions>VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="AutoConsol" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Auto Consol number of the land title. Used in NSW.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.AutoConsol</lx:path>
                                    <lx:label>Auto Consol</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Block" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Block number of the land title. Used in ACT.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.Block</lx:path>
                                    <lx:label>Block</lx:label>
                                    <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="County" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Name of the County on the land title.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.County</lx:path>
                                    <lx:label>County</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="District" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>District or Division in which the land title is located. Used in ACT.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.District</lx:path>
                                    <lx:label>District</lx:label>
                                    <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DuplicateTitleIssued" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether a paper duplicate Certificate of Title has been issued.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.DuplicateTitleIssued</lx:path>
                                    <lx:label>Duplicate Title Issued</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Extent" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Extent of the land title. Used in WA.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.Extent</lx:path>
                                    <lx:label>Extent</lx:label>
                                    <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Folio" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Folio of the land title.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.Folio</lx:path>
                                    <lx:label>Folio</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Location" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Location name and/or number of the land title. Used in NT, WA.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.Location</lx:path>
                                    <lx:label>Location</lx:label>
                                    <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Lot" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Lot identifier of the land title.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.Lot</lx:path>
                                    <lx:label>Lot</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LotPlan" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Contains the Lot Plan detail</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.LotPlan</lx:path>
                                    <lx:label>Lot Plan</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Additional details in relation to the land title. Should be used when the complete title information cannot be captured in the specific title related data fields.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.OtherDetails</lx:path>
                                    <lx:label>Other Details</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Parcel" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Parcel number of the land title. Used in SA, NT.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.Parcel</lx:path>
                                    <lx:label>Parcel</lx:label>
                                    <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Parish" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Should be provided when Title Type = Torrens Title. Name of the Parish on the land title.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.Parish</lx:path>
                                    <lx:label>Parish</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Plan" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A Document ID that identifies a Registry Plan within a Land Registry.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.Plan</lx:path>
                                    <lx:label>Plan</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PlanType" type="planTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Plan Type for a torrens title</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.PlanType</lx:path>
                                    <lx:label>Plan Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RealPropertyDescriptor" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Contains the Real Property Descriptor</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.RealPropertyDescriptor</lx:path>
                                    <lx:label>Real Property Descriptor</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Register" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Register number for the land title. Used in NT.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.Register</lx:path>
                                    <lx:label>Register</lx:label>
                                    <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Section" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Section of the land title. Used in ACT.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.Section</lx:path>
                                    <lx:label>Section</lx:label>
                                    <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TenureType" type="tenureTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the type of title scheme that the owners derive their tenure or estate from.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.TenureType</lx:path>
                                    <lx:label>Tenure Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TitleReference" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An Identifier that identifies a Land Title, used where a text string for the title reference is suitable in the case that lot/plan or volume/folio identification is not suitable.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.TitleReference</lx:path>
                                    <lx:label>Title Reference</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TitleSighted" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A flag to indicate that the title search has been sighted by the valuer.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.TitleSighted</lx:path>
                                    <lx:label>Title Sighted</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TitleSystem" type="titleSystemList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the type of title regime that the property is governed by.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.TitleSystem</lx:path>
                                    <lx:label>Title System</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Unit" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unit of the land title. Used in ACT, NT.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.Unit</lx:path>
                                    <lx:label>Unit</lx:label>
                                    <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UnregisteredLand" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Should be provided when Title Type = Unregistered. A description of the land that is the subject of the unregistered interest in the subject property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.UnregisteredLand</lx:path>
                                    <lx:label>Unregistered Land</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UnregisteredPlan" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Should be provided when Title Type = Unregistered. A description of the plan that is the subject of the unregistered interest in the subject property.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.UnregisteredPlan</lx:path>
                                    <lx:label>Unregistered Plan</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Volume" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Should be provided when Title Type = "Torrens Title".</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Title.Volume</lx:path>
                                    <lx:label>Volume</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="VisitContact" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains details of the party who the lender's valuer should contact to arrange an inspection of the property.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.VisitContact</lx:path>
                                <lx:label>Visit Contact</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="BoatAccessOnly" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The property may be accessible by boat only, a fact that is relevant to the valuer. If so, this should be specified</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.VisitContact.BoatAccessOnly</lx:path>
                                    <lx:label>Boat Access Only</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures information or instructions which are of importance to the lender or valuer in arranging access to the property to conduct a valuation inspection.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.VisitContact.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FourWDAccessOnly" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The property may be accessible by 4WD only, a fact that is relevant to the valuer. If so, this should be specified</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.VisitContact.FourWDAccessOnly</lx:path>
                                    <lx:label>Four WD Access Only</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="visitContactTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the relationship of the valuation contact to the property to be inspected.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.VisitContact.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_ContactParty" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The party that should be contacted to arrange site visit</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.VisitContact.x_ContactParty</lx:path>
                                    <lx:label>Contact Party</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Zoning" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Captures information about the zoning of the land under the relevant planning scheme.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RealEstateAsset.Zoning</lx:path>
                                <lx:label>Zoning</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Type" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the zoning of the property under the applicable planning scheme/regulations.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Zoning.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ZoningEffect" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free text field for the valuer to provide a comment if the zoning of a property affects its value. Part of valuer additional comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RealEstateAsset.Zoning.ZoningEffect</lx:path>
                                    <lx:label>Zoning Effect</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="ApprovalInPrinciple" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that the applicant has not yet identified a specific property for purchase but that this property represents a general description of the type of property they are looking to purchase, so an approval in principle is sought from the lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.ApprovalInPrinciple</lx:path>
                              <lx:label>Approval In Principle</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Construction" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the applicant is going to perform construction work on the security property. Lender may use this as indicator that the loan being sought is or should be a construction loan.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.Construction</lx:path>
                              <lx:label>Construction</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ContractOfSale" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that the Borrower has entered into a Contract Of Sale</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.ContractOfSale</lx:path>
                              <lx:label>Contract Of Sale</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DataSource" type="dataSourceList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The source of data used in searching information about a real estate property. Information provided may relate to property address, size, value and other features.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.DataSource</lx:path>
                              <lx:label>Data Source</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Encumbered" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether there is an encumbrance on the Real Estate Asset.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.Encumbered</lx:path>
                              <lx:label>Encumbered</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ExpectedSellingPeriodComment" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The reason why the selling period of the property is expected to be greater than 6 months.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.ExpectedSellingPeriodComment</lx:path>
                              <lx:label>Expected Selling Period Comment</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ExpectedSellingPeriodGreaterThan6Months" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that the selling period of the property is expected to be greater than 6 months.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.ExpectedSellingPeriodGreaterThan6Months</lx:path>
                              <lx:label>Expected Selling Period Greater Than 6 Months</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Holding" type="holdingList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Specified the capacity in which the owners hold, or will hold in the case of a proposed purchase, their interest in the property.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.Holding</lx:path>
                              <lx:label>Holding</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="InnerCity" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The location of the real estate asset is considered to be inner city according to the lender's policy.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.InnerCity</lx:path>
                              <lx:label>Inner City</lx:label>
                              <li:transactions>ACC,CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="InvestmentPropertyLetter" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate the existence of a letter of intent saying that a security is being bought for investment and not for resale</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.InvestmentPropertyLetter</lx:path>
                              <lx:label>Investment Property Letter</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LegalRepresentation" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate whether an agent is being specified</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.LegalRepresentation</lx:path>
                              <lx:label>Legal Representation</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="MultipleDwellings" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>There is more than one dwelling on this real estate asset. The characteristics of each dwelling may be captured using a separate 'Property Features' element.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.MultipleDwellings</lx:path>
                              <lx:label>Multiple Dwellings</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Neighbourhood" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A description of the neighbourhood.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.Neighbourhood</lx:path>
                              <lx:label>Neighbourhood</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NRASConsortium" type="nrasConsortiumList" use="optional">
                          <xs:annotation>
                            <xs:documentation>List of approved National Rental Affordability Scheme consortium.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.NRASConsortium</lx:path>
                              <lx:label>NRAS Consortium</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NRASProperty" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate whether the property is an approved dwelling under the National Rental Affordability Scheme</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.NRASProperty</lx:path>
                              <lx:label>NRAS Property</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Occupancy" type="occupancyList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates who occupies the dwelling on the property. Where the property is being purchased, this is the purchaser's intention as to who will occupy the dwelling after the purchase completes.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.Occupancy</lx:path>
                              <lx:label>Occupancy</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OnMarketTransaction" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates if the property is an On Market Transaction</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.OnMarketTransaction</lx:path>
                              <lx:label>On Market Transaction</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PopulationSize" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The population size of the city, town, village, or an isolated dwelling this real estate asset is located in or associated with.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.PopulationSize</lx:path>
                              <lx:label>Population Size</lx:label>
                              <li:transactions>ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PrimaryLandUse" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the primary land use of the property, for example, Single Unit Dwelling</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.PrimaryLandUse</lx:path>
                              <lx:label>Primary Land Use</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PrimaryPurpose" type="primaryPurposeRealEstateAssetList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates the primary or predominant purpose for which the owners hold the property.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.PrimaryPurpose</lx:path>
                              <lx:label>Primary Purpose</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PrimarySecurity" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the property will be the primary security for the loan(s). Primary security is to be contrasted with collateral securities, which equally secure the borrowings but which are stamped collateral to the primary security.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.PrimarySecurity</lx:path>
                              <lx:label>Primary Security</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PrimaryUsage" type="primaryUsageList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Specifies what the primary usage of the property is, as reflected as a choice of elements (Industrial | Commercial | Rural | Residential)</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.PrimaryUsage</lx:path>
                              <lx:label>Primary Usage</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PropertyID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The property ID from the data provider</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.PropertyID</lx:path>
                              <lx:label>Property ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PurchaseUnderAOCWrap" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the property is subject to assignment of or a wrap contract.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.PurchaseUnderAOCWrap</lx:path>
                              <lx:label>Purchase Under AOC Wrap</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Remoteness" type="remotenessList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The remoteness of the real estate asset in relation to the main centres of population.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.Remoteness</lx:path>
                              <lx:label>Remoteness</lx:label>
                              <li:transactions>ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this realestate asset within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SiteArea" type="decimalType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Total Area of the Real Estate Asset.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.SiteArea</lx:path>
                              <lx:label>Site Area</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SiteAreaUnits" type="landAreaUnitsList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Units the total area of the Real Estate Asset is measured in.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.SiteAreaUnits</lx:path>
                              <lx:label>Site Area Units</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SiteDescription" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A description of the real estate asset.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.SiteDescription</lx:path>
                              <lx:label>Site Description</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Status" type="realEstateAssetStatusList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates the state of any current dwelling on the property</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.Status</lx:path>
                              <lx:label>Status</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ToBeSold" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the asset will be sold prior to or at settlement</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.ToBeSold</lx:path>
                              <lx:label>To Be Sold</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ToBeUsedAsSecurity" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the asset is to be used as security for the new or refinanced loan(s)</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.ToBeUsedAsSecurity</lx:path>
                              <lx:label>To Be Used As Security</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Transaction" type="assetTransactionList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates the current status of the applicants' ownership of the property, or the transaction that is proposed by them in relation to the property.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.Transaction</lx:path>
                              <lx:label>Transaction</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ValuationProgram" type="valuationProgramList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Expected or intended valuation type if such a recommendation is being made, e.g. may be used by a mortgage manager to inform the funder, or by a broker who is authorised to order valuations for the lender</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.ValuationProgram</lx:path>
                              <lx:label>Valuation Program</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ValuationRequired" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that the credit originator or system has determined that a valuation is required for the property.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.ValuationRequired</lx:path>
                              <lx:label>Valuation Required</lx:label>
                              <li:transactions>CAL,CDA,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Verified" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>A flag to indicate that this asset data has been verified.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.Verified</lx:path>
                              <lx:label>Verified</lx:label>
                              <li:transactions>CAL,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Address" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Address should capture the location details of real estate asset.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.x_Address</lx:path>
                              <lx:label>Address</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.Address</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_PropertyAgent" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>link to the agent managing the property on behalf of the owner.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RealEstateAsset.x_PropertyAgent</lx:path>
                              <lx:label>Property Agent</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="RelatedCompany" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>A business enterprise (including company, sole trader, partnership or government department) that is not an applicant but has a role in some aspect of the application.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.RelatedCompany</lx:path>
                          <lx:label>Related Company</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="BeneficialOwner" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>A beneficial owner of the related company.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RelatedCompany.BeneficialOwner</lx:path>
                                <lx:label>Beneficial Owner</lx:label>
                                <li:transactions>CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_BeneficialOwner" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the beneficial owner.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RelatedCompany.BeneficialOwner.x_BeneficialOwner</lx:path>
                                    <lx:label>Beneficial Owner</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for the related company contact details</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RelatedCompany.Contact</lx:path>
                                <lx:label>Contact</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="ContactPerson" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>A primary contact person at the company</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RelatedCompany.Contact.ContactPerson</lx:path>
                                      <lx:label>Contact Person</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="FirstName" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A person's first name.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RelatedCompany.Contact.ContactPerson.FirstName</lx:path>
                                          <lx:label>First Name</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NameTitle" type="nameTitleList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The name title this person uses when they are formally addressed should be captured.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RelatedCompany.Contact.ContactPerson.NameTitle</lx:path>
                                          <lx:label>Name Title</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Role" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The role or job position of the contact person</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RelatedCompany.Contact.ContactPerson.Role</lx:path>
                                          <lx:label>Role</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Surname" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A person's surname.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RelatedCompany.Contact.ContactPerson.Surname</lx:path>
                                          <lx:label>Surname</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_ContactPerson" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A reference to the primary contact person</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.RelatedCompany.Contact.ContactPerson.x_ContactPerson</lx:path>
                                          <lx:label>Contact Person</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="OfficeFax" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>A fax number for the related company</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RelatedCompany.Contact.OfficeFax</lx:path>
                                      <lx:label>Office Fax</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="OfficeMobile" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Mobile phone number for the related company.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RelatedCompany.Contact.OfficeMobile</lx:path>
                                      <lx:label>Office Mobile</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>An office phone number for the related company</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RelatedCompany.Contact.OfficePhone</lx:path>
                                      <lx:label>Office Phone</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Email" type="emailPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An email address for the related company</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RelatedCompany.Contact.Email</lx:path>
                                    <lx:label>Email</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="WebAddress" type="urlPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A URL for the company's public accessible web site.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RelatedCompany.Contact.WebAddress</lx:path>
                                    <lx:label>Web Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Address" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Address details for the party related to the application at the time of the application submission, and where this information is required by the lender</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RelatedCompany.Contact.x_Address</lx:path>
                                    <lx:label>Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Director" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains a reference to an applicant or related party that is a director of the company.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RelatedCompany.Director</lx:path>
                                <lx:label>Director</lx:label>
                                <li:transactions>CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Director" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Contains a reference to an applicant or related party that is a director of the company.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RelatedCompany.Director.x_Director</lx:path>
                                    <lx:label>Director</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Partner" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Container element for details of the partners that constitute this partnership.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RelatedCompany.Partner</lx:path>
                                <lx:label>Partner</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="PartnerType" type="partnerTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The Partner Type describes whether the partner is a general or limited partner.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RelatedCompany.Partner.PartnerType</lx:path>
                                    <lx:label>Partner Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PercentOwned" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The ownership percentage for the specified partner.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RelatedCompany.Partner.PercentOwned</lx:path>
                                    <lx:label>Percent Owned</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Partner" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Contains a reference to an applicant or related party that is a partner of the business. These can be a person, a business enterprise or a trust.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RelatedCompany.Partner.x_Partner</lx:path>
                                    <lx:label>Partner</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Shareholder" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains a reference to an applicant or related party that is a shareholder of the company.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RelatedCompany.Shareholder</lx:path>
                                <lx:label>Shareholder</lx:label>
                                <li:transactions>CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="PercentOwned" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The ownership percentage for the specified shareholder.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RelatedCompany.Shareholder.PercentOwned</lx:path>
                                    <lx:label>Percent Owned</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Shareholder" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Contains a reference to an applicant or related party that is a shareholder of the company.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RelatedCompany.Shareholder.x_Shareholder</lx:path>
                                    <lx:label>Shareholder</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="ABN" type="abnPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The ABN of the related company</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.ABN</lx:path>
                              <lx:label>ABN</lx:label>
                              <li:transactions>CAL,CDA,DAS,LMI,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ABNVerified" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate that the business number has been verified by a search</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.ABNVerified</lx:path>
                              <lx:label>ABN Verified</lx:label>
                              <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ACN" type="acnPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The ACN of the related company</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.ACN</lx:path>
                              <lx:label>ACN</lx:label>
                              <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BusinessName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The business name or trading name of the company</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.BusinessName</lx:path>
                              <lx:label>Business Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BusinessNameSameAsCompanyName" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates if the Business Name is the same as the Company Name.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.BusinessNameSameAsCompanyName</lx:path>
                              <lx:label>Business Name Same As Company Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BusinessStructure" type="businessStructureFullList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Legal structure of this business (only populated where this applicant is a Business Enterprise).</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.BusinessStructure</lx:path>
                              <lx:label>Business Structure</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CompanyName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The registered name of the company</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.CompanyName</lx:path>
                              <lx:label>Company Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CompanyNumber" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The registered Company Number of the related company.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.CompanyNumber</lx:path>
                              <lx:label>Company Number</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DateRegistered" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates the date when the company was registered.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.DateRegistered</lx:path>
                              <lx:label>Date Registered</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="GSTRegisteredDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>If registered for GST, the date from which the company is registered</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.GSTRegisteredDate</lx:path>
                              <lx:label>GST Registered Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NumberOfDirectors" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The number of directors of the related company.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.NumberOfDirectors</lx:path>
                              <lx:label>Number Of Directors</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NumberOfPartners" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The number of partners that make up the partnership.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.NumberOfPartners</lx:path>
                              <lx:label>Number Of Partners</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NumberOfShareholders" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The number of shareholders of the related company.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.NumberOfShareholders</lx:path>
                              <lx:label>Number Of Shareholders</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NZBN" type="nzbnPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The NZBN of the related company</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.NZBN</lx:path>
                              <lx:label>NZBN</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NZBNVerified" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate that the business number has been verified by a search</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.NZBNVerified</lx:path>
                              <lx:label>NZBN Verified</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RegisteredForGST" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate whether the company is registered for GST</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.RegisteredForGST</lx:path>
                              <lx:label>Registered For GST</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RegisteredInCountry" type="countryCodeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates the country where the company is registered.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.RegisteredInCountry</lx:path>
                              <lx:label>Registered In Country</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this related company within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="TPBNumber" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Tax practitioners board (TPB) registration number.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.TPBNumber</lx:path>
                              <lx:label>TPB Number</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedCompany.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="RelatedPerson" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>A person that is not an applicant but has a role in some aspect of the application.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.RelatedPerson</lx:path>
                          <lx:label>Related Person</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for the related person contact details</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RelatedPerson.Contact</lx:path>
                                <lx:label>Contact</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="HomePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>A home phone number for the related party.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RelatedPerson.Contact.HomePhone</lx:path>
                                      <lx:label>Home Phone</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="Mobile" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>A mobile number for the related party</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RelatedPerson.Contact.Mobile</lx:path>
                                      <lx:label>Mobile</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="WorkPhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>A work phone number for the related party.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.RelatedPerson.Contact.WorkPhone</lx:path>
                                      <lx:label>Work Phone</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Email" type="emailPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An email address for the related person</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RelatedPerson.Contact.Email</lx:path>
                                    <lx:label>Email</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="EmailType" type="contactEmailTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specify whether the email address is home or work</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RelatedPerson.Contact.EmailType</lx:path>
                                    <lx:label>Email Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PreferredContact" type="preferredContactPersonList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the preferred method for contacting this person</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RelatedPerson.Contact.PreferredContact</lx:path>
                                    <lx:label>Preferred Contact</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Address" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Address details for the party related to the application at the time of the application submission, and where this information is required by the lender</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.RelatedPerson.Contact.x_Address</lx:path>
                                    <lx:label>Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ForeignTaxAssociation" type="foreignTaxAssociationType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for Foreign Tax Associations.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RelatedPerson.ForeignTaxAssociation</lx:path>
                                <lx:label>Foreign Tax Association</lx:label>
                                <li:transactions>CAL,CDA,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for the name of the related person.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RelatedPerson.PersonName</lx:path>
                                <lx:label>Person Name</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="TaxDeclarationDetails" type="taxDeclarationDetailsType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Carries details relating to the local tax declarations that have been or are being made to the financial institution.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.RelatedPerson.TaxDeclarationDetails</lx:path>
                                <lx:label>Tax Declaration Details</lx:label>
                                <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="DateOfBirth" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The related person's date of birth.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedPerson.DateOfBirth</lx:path>
                              <lx:label>Date Of Birth</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this related person within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedPerson.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.RelatedPerson.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="SalesChannel" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Captures information about the sales channel through which the product is sold to the applicant.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.SalesChannel</lx:path>
                          <lx:label>Sales Channel</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Aggregator" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The company that the loan writer aggregates under</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.SalesChannel.Aggregator</lx:path>
                                <lx:label>Aggregator</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Container element for Aggregator Contact information</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.SalesChannel.Aggregator.Contact</lx:path>
                                      <lx:label>Contact</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="OfficeFax" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Contains a contact fax number for the aggregator.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.SalesChannel.Aggregator.Contact.OfficeFax</lx:path>
                                            <lx:label>Office Fax</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Contains a contact phone number for the aggregator.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.SalesChannel.Aggregator.Contact.OfficePhone</lx:path>
                                            <lx:label>Office Phone</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Email" type="emailPattern" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures an email address for the sales aggregator submitting the loan application.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.SalesChannel.Aggregator.Contact.Email</lx:path>
                                          <lx:label>Email</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="WebAddress" type="urlPattern" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>An internet address for web content associated with the aggregator</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.SalesChannel.Aggregator.Contact.WebAddress</lx:path>
                                          <lx:label>Web Address</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Address" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Address details for the aggregator at the time of the application submission.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.SalesChannel.Aggregator.Contact.x_Address</lx:path>
                                          <lx:label>Address</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.Address</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="ABN" type="abnPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The ABN of the aggregator</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Aggregator.ABN</lx:path>
                                    <lx:label>ABN</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AccreditationNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures a number such as the accreditation number for the aggregator that has been assigned by the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Aggregator.AccreditationNumber</lx:path>
                                    <lx:label>Accreditation Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CompanyName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the legal name of the aggregator</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Aggregator.CompanyName</lx:path>
                                    <lx:label>Company Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LicenceNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the credit licence number of the aggregator.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Aggregator.LicenceNumber</lx:path>
                                    <lx:label>Licence Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LicenceType" type="licenceTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the type of credit licence under which the aggregator operates.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Aggregator.LicenceType</lx:path>
                                    <lx:label>Licence Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NZBN" type="nzbnPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The NZBN of the aggregator</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Aggregator.NZBN</lx:path>
                                    <lx:label>NZBN</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherIdentifier" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures an alternative identifier by which the aggregator is known, typically issued by the originating organisation rather than the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Aggregator.OtherIdentifier</lx:path>
                                    <lx:label>Other Identifier</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Aggregator.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Commission" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information relating to the commission to be paid to the business partner that has sourced the loan application for the lender. This partner is typically a third party such as a mortgage aggregator or brokerage.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.SalesChannel.Commission</lx:path>
                                <lx:label>Commission</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="CommissionAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount of upfront commission to be paid to the sales channel in the event that loan application settles.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Commission.CommissionAmount</lx:path>
                                    <lx:label>Commission Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CommissionPaid" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether a commission is payable to the sales channel in respect of the loan application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Commission.CommissionPaid</lx:path>
                                    <lx:label>Commission Paid</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CommissionStructure" type="commissionStructureList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the commission plan involves payment of an upfront commission only, an upfront commission and a trail commission, a trail commission only, or another type of structure.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Commission.CommissionStructure</lx:path>
                                    <lx:label>Commission Structure</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherCommissionStructureDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the commission structure if an accurate description is not specified in the enumerated values. Must be provided when Commission Structure = "Other"</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Commission.OtherCommissionStructureDescription</lx:path>
                                    <lx:label>Other Commission Structure Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PromotionCode" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A code that specifies a particular promotion or special commission offer that applies to the loan application, specified by the submitting sales channel in order to tag the deal with the promotion code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Commission.PromotionCode</lx:path>
                                    <lx:label>Promotion Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ThirdPartyReferee" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the sales channel is a third party referrer of business for the purpose of commission disclosure requirements.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Commission.ThirdPartyReferee</lx:path>
                                    <lx:label>Third Party Referee</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Trail" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the trail commission percentage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Commission.Trail</lx:path>
                                    <lx:label>Trail</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UpfrontPayment" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the upfront commission percentage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Commission.UpfrontPayment</lx:path>
                                    <lx:label>Upfront Payment</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Company" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The company or bank branch that the loan writer works at</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.SalesChannel.Company</lx:path>
                                <lx:label>Company</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Container element for contact details of the Sales Channel Company</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.SalesChannel.Company.Contact</lx:path>
                                      <lx:label>Contact</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="OfficeFax" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Contains a contact fax number for the company.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.SalesChannel.Company.Contact.OfficeFax</lx:path>
                                            <lx:label>Office Fax</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Contains a contact phone number for the company.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.SalesChannel.Company.Contact.OfficePhone</lx:path>
                                            <lx:label>Office Phone</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Email" type="emailPattern" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures an email address for the sales channel company submitting the loan application.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.SalesChannel.Company.Contact.Email</lx:path>
                                          <lx:label>Email</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="WebAddress" type="urlPattern" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>An internet address for web content associated with the company</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.SalesChannel.Company.Contact.WebAddress</lx:path>
                                          <lx:label>Web Address</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Address" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Address details for the sales channel company at the time of the application submission.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.SalesChannel.Company.Contact.x_Address</lx:path>
                                          <lx:label>Address</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.Address</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="ABN" type="abnPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The ABN of the broker company or bank branch.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Company.ABN</lx:path>
                                    <lx:label>ABN</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AccreditationNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures a number such as the accreditation number for the company that has been assigned by the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Company.AccreditationNumber</lx:path>
                                    <lx:label>Accreditation Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="BSB" type="bsbPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>For a loan originating at a bank branch, the Bank/State/Branch number of the branch. This information may be collected where an application is originated by lending staff e.g. a Home Finance Manager.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Company.BSB</lx:path>
                                    <lx:label>BSB</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="BusinessName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the business.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Company.BusinessName</lx:path>
                                    <lx:label>Business Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="BusinessNameSameAsCompanyName" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates if the Business Name is the same as the Company Name.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Company.BusinessNameSameAsCompanyName</lx:path>
                                    <lx:label>Business Name Same As Company Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CompanyName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the broker company or bank branch.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Company.CompanyName</lx:path>
                                    <lx:label>Company Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LicenceNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the credit licence number of the company.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Company.LicenceNumber</lx:path>
                                    <lx:label>Licence Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LicenceType" type="licenceTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the type of credit licence under which the company operates.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Company.LicenceType</lx:path>
                                    <lx:label>Licence Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NZBN" type="nzbnPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The NZBN of the broker company or bank branch.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Company.NZBN</lx:path>
                                    <lx:label>NZBN</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherIdentifier" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures an alternative identifier by which the company is known, typically issued by the originating organisation rather than the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Company.OtherIdentifier</lx:path>
                                    <lx:label>Other Identifier</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Company.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Introducer" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of an enterprise or individual who introduced one or more applicants to the sales channel.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.SalesChannel.Introducer</lx:path>
                                <lx:label>Introducer</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="CompanyName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the enterprise that made the introduction.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Introducer.CompanyName</lx:path>
                                    <lx:label>Company Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ContactName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the individual who made the introduction or the name of an individual contact within the enterprise that made the introduction.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Introducer.ContactName</lx:path>
                                    <lx:label>Contact Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IntroducerID" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An identifier assigned to the introducer by the lender. Allows the lender to uniquely identify the introducer.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.Introducer.IntroducerID</lx:path>
                                    <lx:label>Introducer ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="LoanWriter" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for the loan writer details</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.SalesChannel.LoanWriter</lx:path>
                                <lx:label>Loan Writer</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Contains contact details for the Sales Channel Loan Writer</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.SalesChannel.LoanWriter.Contact</lx:path>
                                      <lx:label>Contact</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Fax" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Contains a fax number for the loan writer</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.SalesChannel.LoanWriter.Contact.Fax</lx:path>
                                            <lx:label>Fax</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="Mobile" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Contains a mobile number for the loan writer</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.SalesChannel.LoanWriter.Contact.Mobile</lx:path>
                                            <lx:label>Mobile</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Contains a contact phone number for the loan writer.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.SalesChannel.LoanWriter.Contact.OfficePhone</lx:path>
                                            <lx:label>Office Phone</lx:label>
                                            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Email" type="emailPattern" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>An email address should be captured for the broker or lender staff submitting the home loan application.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.SalesChannel.LoanWriter.Contact.Email</lx:path>
                                          <lx:label>Email</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="WebAddress" type="urlPattern" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>An internet address for web content associated with the loan writer.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.SalesChannel.LoanWriter.Contact.WebAddress</lx:path>
                                          <lx:label>Web Address</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Address" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Address details for the loan writer at the time of the application submission.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.SalesChannel.LoanWriter.Contact.x_Address</lx:path>
                                          <lx:label>Address</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.Address</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="AccreditationNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures a number such as the accreditation number for the loan writer that has been assigned by the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.LoanWriter.AccreditationNumber</lx:path>
                                    <lx:label>Accreditation Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Department" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the department within the sales channel that the loan writer belongs to.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.LoanWriter.Department</lx:path>
                                    <lx:label>Department</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FirstName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A person's first name.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.LoanWriter.FirstName</lx:path>
                                    <lx:label>First Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LicenceNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the credit licence number of loan writer.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.LoanWriter.LicenceNumber</lx:path>
                                    <lx:label>Licence Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LicenceType" type="licenceTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the type of credit licence under which the loan writer operates.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.LoanWriter.LicenceType</lx:path>
                                    <lx:label>Licence Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ManagerName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the name of the manager of the individual loan writer. Typically used when the sales channel is internal to the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.LoanWriter.ManagerName</lx:path>
                                    <lx:label>Manager Name</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NameTitle" type="nameTitleList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name title this person uses when they are formally addressed should be captured.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.LoanWriter.NameTitle</lx:path>
                                    <lx:label>Name Title</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherIdentifier" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures an alternative identifier by which the loan writer is known, typically issued by the originating organisation rather than the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.LoanWriter.OtherIdentifier</lx:path>
                                    <lx:label>Other Identifier</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PersonRole" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures a description of the role of the individual loan writer. Typically used when the sales channel is internal to the lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.LoanWriter.PersonRole</lx:path>
                                    <lx:label>Person Role</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Surname" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A person's surname.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.LoanWriter.Surname</lx:path>
                                    <lx:label>Surname</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A unique ID to reference the loan writer.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SalesChannel.LoanWriter.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="Type" type="salesChannelTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of sales channel that is submitting the application.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.SalesChannel.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="SessionIdentifier" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Contains information about a session the application is being processed within.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.SessionIdentifier</lx:path>
                          <lx:label>SessionIdentifier</lx:label>
                          <li:transactions>CDA,SVC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="SessionID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The identifier of the session as provided by the session handling system</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.SessionIdentifier.SessionID</lx:path>
                              <lx:label>SessionID</lx:label>
                              <li:transactions>CDA,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="System" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The system the session identifier was created by.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.SessionIdentifier.System</lx:path>
                              <lx:label>System</lx:label>
                              <li:transactions>CDA,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Settlement" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Details for the settlement of one or more securities.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.Settlement</lx:path>
                          <lx:label>Settlement</lx:label>
                          <li:transactions>CAL,DAS</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Security" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Information about the security that is to be settled.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Settlement.Security</lx:path>
                                <lx:label>Security</lx:label>
                                <li:transactions>CAL,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Security" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to a security.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Settlement.Security.x_Security</lx:path>
                                    <lx:label>Security</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.GeneralSecurityAgreement</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="PayoutFigure" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The amount to be transferred at settlement, including the amount for multiple loans/securities and could include any fees, shortfalls or surplus amounts that are to be transferred. The payout figure would be the final amount for the solicitor to transfer.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Settlement.PayoutFigure</lx:path>
                              <lx:label>Payout Figure</lx:label>
                              <li:transactions>CAL,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this settlement within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Settlement.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SettlementBookingDateTime" type="dateTimeType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date and time of this settlement booking.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Settlement.SettlementBookingDateTime</lx:path>
                              <lx:label>Settlement Booking Date Time</lx:label>
                              <li:transactions>CAL,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SettlementReferenceNumber" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A unique identifier assigned to the settlement by the lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Settlement.SettlementReferenceNumber</lx:path>
                              <lx:label>Settlement Reference Number</lx:label>
                              <li:transactions>CAL,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_SettlementBookingAddress" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A cross reference to the address at which this settlement is booked to occur.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Settlement.x_SettlementBookingAddress</lx:path>
                              <lx:label>Settlement Booking Address</lx:label>
                              <li:transactions>CAL,DAS</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.Address</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="SplitLoan" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>The collection of loan accounts or liabilities that comprise this split loan.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.SplitLoan</lx:path>
                          <lx:label>Split Loan</lx:label>
                          <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="SplitLoanComponent" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>A loan account or liability that is a component of the split loan.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.SplitLoan.SplitLoanComponent</lx:path>
                                <lx:label>Split Loan Component</lx:label>
                                <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Liability" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to a Liability element.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SplitLoan.SplitLoanComponent.x_Liability</lx:path>
                                    <lx:label>Liability</lx:label>
                                    <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Liability</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_LoanDetails" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to a Loan Details element.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.SplitLoan.SplitLoanComponent.x_LoanDetails</lx:path>
                                    <lx:label>Loan Details</lx:label>
                                    <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this split loan within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.SplitLoan.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.SplitLoan.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Summary" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Wrapper element for Application Summary data.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.Summary</lx:path>
                          <lx:label>Summary</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Applicants" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains summary information that applies to all applicants.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Summary.Applicants</lx:path>
                                <lx:label>Applicants</lx:label>
                                <li:transactions>CDA,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ApplicationRoleType" type="applicantsApplicationRoleTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the role of the applicants in the application such as joint or single.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Applicants.ApplicationRoleType</lx:path>
                                    <lx:label>Are All Applicants Individuals</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AreAllApplicantsIndividuals" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether all applicants are individuals.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Applicants.AreAllApplicantsIndividuals</lx:path>
                                    <lx:label>Are All Applicants Individuals</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AreAllApplicantsOfLegalAgeForApplication" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether all individual applicants have attained at least the number of years required by law for them to be making the application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Applicants.AreAllApplicantsOfLegalAgeForApplication</lx:path>
                                    <lx:label>Are All Applicants Of Legal Age For Application</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="AreAllApplicantsPermanentResidents" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether all individual applicants have an Australian permanent residency status or better (e.g. citizen).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Applicants.AreAllApplicantsPermanentResidents</lx:path>
                                    <lx:label>Are All Applicants Permanent Residents</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="HasBankruptApplicant" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether any applicant is currently bankrupt.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Applicants.HasBankruptApplicant</lx:path>
                                    <lx:label>Has Bankrupt Applicant</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="HasLiabilitiesInArrears" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the application contains one or more liabilities that are are currently in arrears.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Applicants.HasLiabilitiesInArrears</lx:path>
                                    <lx:label>Has Liabilities In Arrears</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsJointApplicantExistingNonIBCustomer" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the solitary joint applicant is an existing customer of the funding institution but does not have access to their internet banking platform.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Applicants.IsJointApplicantExistingNonIBCustomer</lx:path>
                                    <lx:label>Is Joint Applicant Existing Non IB Customer</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsPrimaryApplicantExistingCustomer" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the primary applicant is an existing customer of the funding institution.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Applicants.IsPrimaryApplicantExistingCustomer</lx:path>
                                    <lx:label>Is Primary Applicant Existing Customer</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsPrimaryApplicantExistingNonIBCustomer" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the primary applicant is an existing customer of the funding institution but does not have access to their internet banking platform.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Applicants.IsPrimaryApplicantExistingNonIBCustomer</lx:path>
                                    <lx:label>Is Primary Applicant Existing Non IB Customer</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsSecondApplicantPresentConfirmed" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the second applicant is confirmed to be present at the time the application is filled in.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Applicants.IsSecondApplicantPresentConfirmed</lx:path>
                                    <lx:label>Is Second Applicant Present Confirmed</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsSecondaryApplicantExistingCustomer" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the sole secondary applicant is an existing customer of the funding institution.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Applicants.IsSecondaryApplicantExistingCustomer</lx:path>
                                    <lx:label>Is Secondary Applicant Existing Customer</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfApplicants" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the number of applicants on the application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Applicants.NumberOfApplicants</lx:path>
                                    <lx:label>Number Of Applicants</lx:label>
                                    <li:transactions>CDA,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="DocumentationInstructions" type="documentationInstructionsType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information specifying where the lender is to send the loan documents once the loan has been approved. A nomination of authority at the Loan level will over-ride this nomination.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Summary.DocumentationInstructions</lx:path>
                                <lx:label>Documentation Instructions</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="Fee" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains information about a fee that is payable in connection with the loan application.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Summary.Fee</lx:path>
                                <lx:label>Fee</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="AccountNumber" type="financialAccountType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The details of the account from which the fees should be paid.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Summary.Fee.AccountNumber</lx:path>
                                      <lx:label>Account Number</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="ApplicableDuration" type="durationType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The length of time that the fee is applicable.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Summary.Fee.ApplicableDuration</lx:path>
                                      <lx:label>Applicable Duration</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="CreditCard" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Details of the credit card to be used to pay fees</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Summary.Fee.CreditCard</lx:path>
                                      <lx:label>Credit Card</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="CardHolderName" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Cardholder name on the credit card to be used to pay fees</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.Fee.CreditCard.CardHolderName</lx:path>
                                          <lx:label>Card Holder Name</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ExpiryMonth" type="mmMonthList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Month of expiry date of the credit card to be used to pay fees</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.Fee.CreditCard.ExpiryMonth</lx:path>
                                          <lx:label>Expiry Month</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ExpiryYear" type="yyYearList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Year of expiry date of the credit card to be used to pay fees</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.Fee.CreditCard.ExpiryYear</lx:path>
                                          <lx:label>Expiry Year</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Number" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Credit card number to be used to pay fees</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.Fee.CreditCard.Number</lx:path>
                                          <lx:label>Number</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Type" type="creditCardTypeList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Type of credit card to be used to pay fees</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.Fee.CreditCard.Type</lx:path>
                                          <lx:label>Type</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="PayableToAccount" type="financialAccountType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The account into which the fees should be paid.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Summary.Fee.PayableToAccount</lx:path>
                                      <lx:label>Payable To Account</lx:label>
                                      <li:transactions>CAL,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="Percentage" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The fee as a percentage.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Summary.Fee.Percentage</lx:path>
                                      <lx:label>Percentage</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="CalculationBasis" type="calculationBasisList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates which amount is used as the base value to calculate the fee, when the fee is given as a percentage rate.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.Fee.Percentage.CalculationBasis</lx:path>
                                          <lx:label>Calculation Basis</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Rate" type="percentType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The percentage rate of the fee.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.Fee.Percentage.Rate</lx:path>
                                          <lx:label>Rate</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Amount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The estimated or approximate amount of the fee.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.Amount</lx:path>
                                    <lx:label>Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CapitaliseFee" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the fees are to be capitalised against the loan amount</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.CapitaliseFee</lx:path>
                                    <lx:label>Capitalise Fee</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Category" type="feeCategoryList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the general category of fees such as bank fees</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.Category</lx:path>
                                    <lx:label>Category</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of the fee type if an accurate description is not specified in the enumerated values. Must be provided when Fee Type = "Other"</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FeeCode" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An identifying code for this fee. This code is typically issued by the lender and enables the lender's systems to identify the fee.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.FeeCode</lx:path>
                                    <lx:label>Fee Code</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Frequency" type="frequencyFullList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The frequency of the fee payments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.Frequency</lx:path>
                                    <lx:label>Frequency</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GSTAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The portion of the fee amount that is paid in goods and services tax (GST).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.GSTAmount</lx:path>
                                    <lx:label>GST Amount</lx:label>
                                    <li:transactions>ACC,CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsDisclosed" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>This fee is disclosed to the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.IsDisclosed</lx:path>
                                    <lx:label>IsDisclosed</lx:label>
                                    <li:transactions>CAL,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsEstimated" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the fee amount is estimated.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.IsEstimated</lx:path>
                                    <lx:label>Is Estimated</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ITCAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The portion of the fee amount that is claimed as input tax credit (ITC).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.ITCAmount</lx:path>
                                    <lx:label>ITC Amount</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfRepeats" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of repeats of the fee payments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.NumberOfRepeats</lx:path>
                                    <lx:label>NumberOfRepeats</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Paid" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the fee has already been paid</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.Paid</lx:path>
                                    <lx:label>Paid</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PayableTo" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the name of the party that the fee is payable to.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.PayableTo</lx:path>
                                    <lx:label>Payable To</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PayFeesFrom" type="payFeesFromList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The borrower-selected method of paying the fee</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.PayFeesFrom</lx:path>
                                    <lx:label>Pay Fees From</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PaymentTiming" type="feePaymentTimingList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The timing of the fee payment.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.PaymentTiming</lx:path>
                                    <lx:label>Payment Timing</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The start date of the fee payments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.StartDate</lx:path>
                                    <lx:label>StartDate</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="feeTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the type of fee being described</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unique Identifier for this Fee Element</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Account" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to an account (either a Financial Asset element, a Liability element, or a Loan Details element) from which the fee should be paid.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.x_Account</lx:path>
                                    <lx:label>Account</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Liability</lx:target>
                                      <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                      <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_FinancialProduct" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to the unique identifier (UniqueID) for the financial product that relates to the fee being incurred.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.x_FinancialProduct</lx:path>
                                    <lx:label>Financial Product</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_ProductSet" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the product set (a Product Set element) to which the fee is assigned.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.x_ProductSet</lx:path>
                                    <lx:label>Product Set</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.ProductSet</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Security" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to the unique identifier (UniqueID) for the security that relates to the fee being incurred.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Fee.x_Security</lx:path>
                                    <lx:label>Security</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                      <lx:target>Package.Content.Application.GeneralSecurityAgreement</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="LoanToValuationRatio" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains the various calculated loan to valuation ratio (LVR) values relative to the lending scenario described by the application.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Summary.LoanToValuationRatio</lx:path>
                                <lx:label>Loan To Valuation Ratio</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="ContributingValuation" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Contains links to the valuation that was used as the basis for this ratio calculation.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Summary.LoanToValuationRatio.ContributingValuation</lx:path>
                                      <lx:label>Contributing Valuation</lx:label>
                                      <li:transactions>CAL,CDA,CNZ,VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="x_Valuation" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Link to a valuation that was used in the calculation of this Loan To Valuation Ratio</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.LoanToValuationRatio.ContributingValuation.x_Valuation</lx:path>
                                          <lx:label>Valuation</lx:label>
                                          <li:transactions>CAL,CDA,CNZ,VAL</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                            <lx:target>Package.Content.RealEstateValuation</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="ApplicationLVR" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The basic LVR for the entire application. The calculation of this value can be specific so it is left to the parties making use of this LVR to agree on the calculation to be used.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.LoanToValuationRatio.ApplicationLVR</lx:path>
                                    <lx:label>Application LVR</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PeakDebtLVR" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The peak debt LVR calculated for bridging loan applications. The calculation of this value can be specific so it is left to the parties making use of this LVR to agree on the calculation to be used.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.LoanToValuationRatio.PeakDebtLVR</lx:path>
                                    <lx:label>Peak Debt LVR</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.LoanToValuationRatio.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Purpose" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains summary information that applies to the application purpose.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Summary.Purpose</lx:path>
                                <lx:label>Purpose</lx:label>
                                <li:transactions>CDA</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="IsGenuineSavingsType" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether any deposit contribution source of funds is from genuine savings accumulated over time.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Purpose.IsGenuineSavingsType</lx:path>
                                    <lx:label>Is Genuine Savings Type</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsInvestment" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the application capture or origination system has determined that the application is primarily intended for investment purposes.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Purpose.IsInvestment</lx:path>
                                    <lx:label>Is Investment</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="WillBalanceTransferReduceCosts" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Describes whether the balance transfer would reduce the borrower's costs.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Purpose.WillBalanceTransferReduceCosts</lx:path>
                                    <lx:label>Will Balance Transfer Reduce Costs</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="WillBalanceTransferReduceCostsDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Describes the reason(s) identified as to why the balance transfer would reduce the borrower's costs.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Purpose.WillBalanceTransferReduceCostsDetails</lx:path>
                                    <lx:label>Will Balance Transfer Reduce Costs Details</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ResponsibleLending" type="responsibleLendingType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Captures information about any significant changes in circumstances that will adversely affect the applicants' ability to make contracted loan repayments; the nature of the change, and how they will collectively continue to make repayments.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Summary.ResponsibleLending</lx:path>
                                <lx:label>Responsible Lending</lx:label>
                                <li:transactions>CDA</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="RolledUpAmounts" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Summary information relating to the exposure of loan parties and/or the institution providing funds or credit to loan parties.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Summary.RolledUpAmounts</lx:path>
                                <lx:label>Rolled Up Amounts</lx:label>
                                <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="NumberOfLoans" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the number of loan products captured in the application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.RolledUpAmounts.NumberOfLoans</lx:path>
                                    <lx:label>Number Of Loans</lx:label>
                                    <li:transactions>CDA,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The total amount requested/owed by the applicant(s) to the financial institution funding the application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.RolledUpAmounts.TotalAmount</lx:path>
                                    <lx:label>Total Amount</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalCreditLimitAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Amount that represents the total liabilities held by applicants and guarantors including approved loans that are not yet drawn.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.RolledUpAmounts.TotalCreditLimitAmount</lx:path>
                                    <lx:label>Total Credit Limit Amount</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalInstitutionLiabilityAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Amount that represents the total debts incurred by the applicant(s) from the same financial institution from whence the new loans will be funded.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.RolledUpAmounts.TotalInstitutionLiabilityAmount</lx:path>
                                    <lx:label>Total Institution Liability Amount</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalNewFunds" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The total amount of new funds requested in the application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.RolledUpAmounts.TotalNewFunds</lx:path>
                                    <lx:label>Total New Funds</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TotalSecurityAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Amount that represents the total assets value to be used as security held by applicants.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.RolledUpAmounts.TotalSecurityAmount</lx:path>
                                    <lx:label>Total Security Amount</lx:label>
                                    <li:transactions>CAL,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Securities" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains summary information that applies to assets that secure the funds requested by the application.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Summary.Securities</lx:path>
                                <lx:label>Securities</lx:label>
                                <li:transactions>CDA,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="IsLandSizeOverTheLimit" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the land size of any property asset to be used to secure the application funds is over the limit specified by the funding financial institution.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Securities.IsLandSizeOverTheLimit</lx:path>
                                    <lx:label>Is Land Size Over The Limit</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IsPropertyAreaLessThanLimit" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the living area of a property asset to be used to secure the application funds is under the limit specified by the funding financial institution.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Securities.IsPropertyAreaLessThanLimit</lx:path>
                                    <lx:label>Is Property Area Less Than Limit</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfSecurities" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the number of assets to be used to secure the application funds.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.Securities.NumberOfSecurities</lx:path>
                                    <lx:label>Number Of Securities</lx:label>
                                    <li:transactions>CDA,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ServiceabilityResults" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Contains summary information regarding the serviceability results calculated for this group of loan parties.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Summary.ServiceabilityResults</lx:path>
                                <lx:label>Serviceability Results</lx:label>
                                <li:transactions>CAL,CDA,CNZ,LMI,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Applicant" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Loan applicant/s (borrower/s or guarantor/s) that are included in this group.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Summary.ServiceabilityResults.Applicant</lx:path>
                                      <lx:label>Applicant</lx:label>
                                      <li:transactions>CAL,CNZ,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="x_Party" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Reference to the loan party that is included in this group.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.ServiceabilityResults.Applicant.x_Party</lx:path>
                                          <lx:label>Party</lx:label>
                                          <li:transactions>CAL,CNZ,LMI</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="NetDisposableIncome" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Contains the net disposable income calculated for the application.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Summary.ServiceabilityResults.NetDisposableIncome</lx:path>
                                      <lx:label>Net Disposable Income</lx:label>
                                      <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Amount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the amount of the net disposable income calculated for the application.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.ServiceabilityResults.NetDisposableIncome.Amount</lx:path>
                                          <lx:label>Amount</lx:label>
                                          <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the period over which the amount of the net disposable income is considered disposable.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.ServiceabilityResults.NetDisposableIncome.Frequency</lx:path>
                                          <lx:label>Frequency</lx:label>
                                          <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="TotalGrossIncome" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The total income of the application parties before accounting for taxes and deductions.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalGrossIncome</lx:path>
                                      <lx:label>Total Gross Income</lx:label>
                                      <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Amount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The total income amount of the application parties before accounting for taxes and deductions.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalGrossIncome.Amount</lx:path>
                                          <lx:label>Amount</lx:label>
                                          <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the period over which the total gross income amount is calculated.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalGrossIncome.Frequency</lx:path>
                                          <lx:label>Frequency</lx:label>
                                          <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="TotalNetIncome" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The total income of the application parties after all deductions such as taxes have been taken.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalNetIncome</lx:path>
                                      <lx:label>Total Net Income</lx:label>
                                      <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Amount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The total income amount of the application parties after all deductions such as taxes have been taken.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalNetIncome.Amount</lx:path>
                                          <lx:label>Amount</lx:label>
                                          <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the period over which the total net income amount is calculated.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalNetIncome.Frequency</lx:path>
                                          <lx:label>Frequency</lx:label>
                                          <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="TotalSystemCalculatedExpenses" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The total system calculated expenses.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalSystemCalculatedExpenses</lx:path>
                                      <lx:label>Total System Calculated Living Expenses</lx:label>
                                      <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Amount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The total system calculated expense amount.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalSystemCalculatedExpenses.Amount</lx:path>
                                          <lx:label>Amount</lx:label>
                                          <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The frequency of the total system calculated expenses.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalSystemCalculatedExpenses.Frequency</lx:path>
                                          <lx:label>Frequency</lx:label>
                                          <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="TotalSystemCalculatedLivingExpenses" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The estimated living expenses calculated by or on behalf of the application capture or origination system.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalSystemCalculatedLivingExpenses</lx:path>
                                      <lx:label>Total System Calculated Living Expenses</lx:label>
                                      <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Amount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The estimated living expenses amount calculated by or on behalf of the application capture or origination system.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalSystemCalculatedLivingExpenses.Amount</lx:path>
                                          <lx:label>Amount</lx:label>
                                          <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the period over which the living expenses were calculated.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalSystemCalculatedLivingExpenses.Frequency</lx:path>
                                          <lx:label>Frequency</lx:label>
                                          <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="TotalUserStatedLivingExpenses" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The total of living expenses declared by the application parties.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalUserStatedLivingExpenses</lx:path>
                                      <lx:label>Total User Stated Living Expenses</lx:label>
                                      <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Amount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The total amount of living expenses declared by the application parties.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalUserStatedLivingExpenses.Amount</lx:path>
                                          <lx:label>Amount</lx:label>
                                          <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Captures the period over which the living expenses were captured.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.Summary.ServiceabilityResults.TotalUserStatedLivingExpenses.Frequency</lx:path>
                                          <lx:label>Frequency</lx:label>
                                          <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="CalculationIdentifier" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An identifier issued by the serviceability calculator that can uniquely identify a serviceability calculation.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.ServiceabilityResults.CalculationIdentifier</lx:path>
                                    <lx:label>Calculation Identifier</lx:label>
                                    <li:transactions>CDA,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DSR" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the debt servicing ratio that applies to the parties to the application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.ServiceabilityResults.DSR</lx:path>
                                    <lx:label>DSR</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DTI" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Debt to income ratio (DTI). The total credit limit of all debts to the gross annual income ratio that applies to the parties to the application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.ServiceabilityResults.DTI</lx:path>
                                    <lx:label>DTI</lx:label>
                                    <li:transactions>CAL,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GroupIdentifier" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An identifer that can be used to identify the group.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.ServiceabilityResults.GroupIdentifier</lx:path>
                                    <lx:label>Group Identifier</lx:label>
                                    <li:transactions>CAL,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GroupName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>If these serviceability results are for a group, this defines a name assigned to the group.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.ServiceabilityResults.GroupName</lx:path>
                                    <lx:label>Group Name</lx:label>
                                    <li:transactions>CAL,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LTI" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Loan to income ratio (LTI). The total loan repayment amounts to total actual gross income ratio that applies to the parties to the application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.ServiceabilityResults.LTI</lx:path>
                                    <lx:label>LTI</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MaximumLoanAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Maximum recommended amount of the loan, combining the individual commitments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.ServiceabilityResults.MaximumLoanAmount</lx:path>
                                    <lx:label>Maximum Loan Amount</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MaximumRepaymentAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The dollar amount of the maximum recommended sum of gross repayments for this borrower or borrowers.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.ServiceabilityResults.MaximumRepaymentAmount</lx:path>
                                    <lx:label>Maximum Repayment Amount</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MaximumRepaymentFrequency" type="repaymentFrequencyList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The frequency of the maximum recommended sum of gross repayments for this borrower or borrowers.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.ServiceabilityResults.MaximumRepaymentFrequency</lx:path>
                                    <lx:label>Maximum Repayment Frequency</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MRIM" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Monthly mortgage repayment to income measure (MRIM). The monthly mortgage repayment to sensitised monthly income ratio that applies to the parties to the application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.ServiceabilityResults.MRIM</lx:path>
                                    <lx:label>MRIM</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NSR" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the net surplus ratio that applies to the parties to the application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.ServiceabilityResults.NSR</lx:path>
                                    <lx:label>NSR</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,LMI,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Serviceable" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures whether the parties to the loan are able to service the funds they will be liable for should the application be successful.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.ServiceabilityResults.Serviceable</lx:path>
                                    <lx:label>Serviceable</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.Summary.ServiceabilityResults.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Signature" type="signatureType" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>A signature for this application.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.Summary.Signature</lx:path>
                                <lx:label>Signature</lx:label>
                                <li:transactions>CAL,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="AllPartiesAgreeToElectronicSignature" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>All parties have agreed to use electronic signatures.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Summary.AllPartiesAgreeToElectronicSignature</lx:path>
                              <lx:label>All Parties Agree To Electronic Signature</lx:label>
                              <li:transactions>CAL,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="FeesDisclosureDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date that the fees were disclosed to the borrower may be specified</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.Summary.FeesDisclosureDate</lx:path>
                              <lx:label>Fees Disclosure Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="TrustApplicant" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Contains information about an applicant who is applying in their capacity as trustee of a trust. A trustee may be either a person or a company.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.TrustApplicant</lx:path>
                          <lx:label>Trust Applicant</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="BeneficialOwner" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>A beneficial owner of the trust. All beneficial owners of the trust should be captured (even persons captured elsewhere as directors, shareholders, beneficiaries etc.)</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.BeneficialOwner</lx:path>
                                <lx:label>Beneficial Owner</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_BeneficialOwner" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the beneficial owner.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.BeneficialOwner.x_BeneficialOwner</lx:path>
                                    <lx:label>Beneficial Owner</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Beneficiary" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Wrapper for the trust beneficiaries</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.Beneficiary</lx:path>
                                <lx:label>Beneficiary</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Beneficiary" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to a beneficiary of the Trust. A Trust Beneficiary is a person, company, trust or other legal entity who may be entitled in some way to receive distributions of the assets or income of the Trust.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Beneficiary.x_Beneficiary</lx:path>
                                    <lx:label>Beneficiary</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Business" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about the business of the trust</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.Business</lx:path>
                                <lx:label>Business</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AustralianBIC" type="aubicPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A Business Industry Code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Business.AustralianBIC</lx:path>
                                    <lx:label>Australian BIC</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CustomIndustryCode" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An industry code, from a customised list supported by the lender where the lender requires a code that is not within ANZSIC, BIC or GICS.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Business.CustomIndustryCode</lx:path>
                                    <lx:label>Custom Industry Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GICSCode" type="gicsCodePattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A Global Industry Classification Standard (GICS) code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Business.GICSCode</lx:path>
                                    <lx:label>GICS Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Industry" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures a free text description of the Industry Code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Business.Industry</lx:path>
                                    <lx:label>Industry</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IndustryCode" type="anzsicIndustryCodeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The four digit industry code, from the ANZSIC list of codes. If the nominal code is two or three digits, add '0's to the end to make the four digit code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Business.IndustryCode</lx:path>
                                    <lx:label>Industry Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="MainBusinessActivity" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The main activity of the business</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Business.MainBusinessActivity</lx:path>
                                    <lx:label>Main Business Activity</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NewZealandBICCode" type="nzbicPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A Business Industry Classification Code.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Business.NewZealandBICCode</lx:path>
                                    <lx:label>New Zealand BIC Code</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Contact" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains contact details for the trust applicant.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.Contact</lx:path>
                                <lx:label>Contact</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="ContactPerson" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Contains information for the contact person for the trust</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.TrustApplicant.Contact.ContactPerson</lx:path>
                                      <lx:label>Contact Person</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="FirstName" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A person's first name.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.Contact.ContactPerson.FirstName</lx:path>
                                          <lx:label>First Name</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NameTitle" type="nameTitleList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The name title this person uses when they are formally addressed should be captured.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.Contact.ContactPerson.NameTitle</lx:path>
                                          <lx:label>Name Title</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Role" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The role or job position of the contact person</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.Contact.ContactPerson.Role</lx:path>
                                          <lx:label>Role</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Surname" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A person's surname.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.Contact.ContactPerson.Surname</lx:path>
                                          <lx:label>Surname</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_ContactPerson" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A contact person for the trust</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.Contact.ContactPerson.x_ContactPerson</lx:path>
                                          <lx:label>Contact Person</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="OfficeFax" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Fax number for the trust</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.TrustApplicant.Contact.OfficeFax</lx:path>
                                      <lx:label>Office Fax</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="OfficeMobile" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Mobile phone number for the trust.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.TrustApplicant.Contact.OfficeMobile</lx:path>
                                      <lx:label>Office Mobile</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Office number for the trust</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.TrustApplicant.Contact.OfficePhone</lx:path>
                                      <lx:label>Office Phone</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="PrincipalTradingAddressDuration" type="durationType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The duration of the trust applicant trading at this address.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.TrustApplicant.Contact.PrincipalTradingAddressDuration</lx:path>
                                      <lx:label>Principal Trading Address Duration</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                                <xs:element name="RegisteredAddressDuration" type="durationType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The duration of registration of the trust applicant at this address.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.TrustApplicant.Contact.RegisteredAddressDuration</lx:path>
                                      <lx:label>Registered Address Duration</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="Email" type="emailPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An email address for the trust</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Contact.Email</lx:path>
                                    <lx:label>Email</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PreferredContact" type="preferredContactCompanyList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Specifies the preferred method for contacting this applicant</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Contact.PreferredContact</lx:path>
                                    <lx:label>Preferred Contact</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PrincipalTradingAddressStartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date the trust applicant began trading at this address.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Contact.PrincipalTradingAddressStartDate</lx:path>
                                    <lx:label>Principal Trading Address Start Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RegisteredAddressStartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date the trust applicant registered at this address.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Contact.RegisteredAddressStartDate</lx:path>
                                    <lx:label>Registered Address Start Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="WebAddress" type="urlPattern" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A URL for the trust's public accessible web site.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Contact.WebAddress</lx:path>
                                    <lx:label>Web Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_MailingAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Mailing address of the trust</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Contact.x_MailingAddress</lx:path>
                                    <lx:label>Mailing Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_PrincipalTradingAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The official address of the trust where business is carried out.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Contact.x_PrincipalTradingAddress</lx:path>
                                    <lx:label>Principal Trading Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_RegisteredAddress" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Registered address of the trust</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Contact.x_RegisteredAddress</lx:path>
                                    <lx:label>Registered Address</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.Address</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="DealingNumber" type="dealingNumberType" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for Dealing Number details.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.DealingNumber</lx:path>
                                <lx:label>Dealing Number</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="DeclaredIncome" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for the amount of income declared by the applicant. Used for Low Doc applications when full financial years for income are not provided.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.DeclaredIncome</lx:path>
                                <lx:label>Declared Income</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="IncomeAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount of income declared by the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.DeclaredIncome.IncomeAmount</lx:path>
                                    <lx:label>Income Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NetIncomeAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The net amount of income declared by the applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.DeclaredIncome.NetIncomeAmount</lx:path>
                                    <lx:label>Net Income Amount</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ExistingCustomer" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains details of the existing customer details for this trust applicant where it is an existing customer of the lender being applied to.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.ExistingCustomer</lx:path>
                                <lx:label>Existing Customer</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="AccountNumber" type="financialAccountType" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Account details for an account held that identifies the party as an existing customer</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.TrustApplicant.ExistingCustomer.AccountNumber</lx:path>
                                      <lx:label>Account Number</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="CustomerNumber" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A unique number assigned by the lender used to identify the customer</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.ExistingCustomer.CustomerNumber</lx:path>
                                    <lx:label>Customer Number</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CustomerSince" type="gYearType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The approximate year that the applicant became a customer</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.ExistingCustomer.CustomerSince</lx:path>
                                    <lx:label>Customer Since</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CustomerSinceMonth" type="mmMonthList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The month (if known) that the trust applicant became a customer (only if the year is also provided in 'CustomerSince').</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.ExistingCustomer.CustomerSinceMonth</lx:path>
                                    <lx:label>Customer Since Month</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="FinancialAnalysis" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Container element for the detailed financial analysis of the company</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.FinancialAnalysis</lx:path>
                                <lx:label>Financial Analysis</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="CompanyFinancials" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Container for the reference to the financial data</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.TrustApplicant.FinancialAnalysis.CompanyFinancials</lx:path>
                                      <lx:label>Company Financials</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="x_CompanyFinancials" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Reference to the element containing the financial analysis for each financial year, to be provided when Complete Financial Analysis = "Yes"</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.FinancialAnalysis.CompanyFinancials.x_CompanyFinancials</lx:path>
                                          <lx:label>Company Financials</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.CompanyFinancials</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="AnnualPaymentOnCommitments" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Total Principle and Interest payments on borrowings due to be paid in the next 12 months</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.FinancialAnalysis.AnnualPaymentOnCommitments</lx:path>
                                    <lx:label>Annual Payment On Commitments</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CompleteFinancialAnalysis" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag to indicate whether a complete financial analysis will be provided electronically, or only a summary view</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.FinancialAnalysis.CompleteFinancialAnalysis</lx:path>
                                    <lx:label>Complete Financial Analysis</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ForeignTaxAssociation" type="foreignTaxAssociationType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for Foreign Tax Associations.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.ForeignTaxAssociation</lx:path>
                                <lx:label>Foreign Tax Association</lx:label>
                                <li:transactions>CAL,CDA,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="IncomePrevious" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the share of the gross (pre-tax) income or income-related amounts in the previous financial year (immediately preceding the most recent full financial year), attributable to this applicant (for year = current year -2).</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious</lx:path>
                                <lx:label>Income Previous</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Addback" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Container element for Addback detail</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback</lx:path>
                                      <lx:label>Addback</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="OtherAddback" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>Other addback that cannot otherwise be captured under a specific addback category.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.OtherAddback</lx:path>
                                            <lx:label>Other Add Back</lx:label>
                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Amount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Amount that can be an addback to the income.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.OtherAddback.Amount</lx:path>
                                                <lx:label>Amount</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Decription of the other addback.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.OtherAddback.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Allowances" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Allowances amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.Allowances</lx:path>
                                          <lx:label>Allowances</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AmortisationOfGoodwill" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Amortisation Of Goodwill amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.AmortisationOfGoodwill</lx:path>
                                          <lx:label>Amortisation Of Goodwill</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Bonus" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Bonus amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.Bonus</lx:path>
                                          <lx:label>Bonus</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CarryForwardLosses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Carry Forward Losses amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.CarryForwardLosses</lx:path>
                                          <lx:label>Carry Forward Losses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Depreciation" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Depreciation as recorded in the financial report</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.Depreciation</lx:path>
                                          <lx:label>Depreciation</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="HomeOffice" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Home Office amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.HomeOffice</lx:path>
                                          <lx:label>Home Office</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Interest" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Total interest expense as listed in the financial report, including leasing expenses</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.Interest</lx:path>
                                          <lx:label>Interest</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Lease" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Lease amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.Lease</lx:path>
                                          <lx:label>Lease</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonCashBenefits" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-cash benefits amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.NonCashBenefits</lx:path>
                                          <lx:label>Non Cash Benefits</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonRecurringExpenses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-recurring expense amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.NonRecurringExpenses</lx:path>
                                          <lx:label>Non Recurring Expenses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PaymentToTrustee" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The payment to trustee that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.PaymentToTrustee</lx:path>
                                          <lx:label>Payment To Trustee</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Salary" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Salary amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.Salary</lx:path>
                                          <lx:label>Salary</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SuperannuationExcess" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Super excess amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.Addback.SuperannuationExcess</lx:path>
                                          <lx:label>Superannuation Excess</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="EndDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The end date for the stated income, generally the end of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.EndDate</lx:path>
                                    <lx:label>End Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncomeGreaterThanPreviousYear" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>When comparing past Tax Office Assessments, indicate whether the past return has improved on the previous year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.IncomeGreaterThanPreviousYear</lx:path>
                                    <lx:label>Income Greater Than Previous Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProfitAfterTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the profit after tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.ProfitAfterTax</lx:path>
                                    <lx:label>Profit After Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProfitBeforeTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the net profit/profit before tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.ProfitBeforeTax</lx:path>
                                    <lx:label>Profit Before Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The start date for the stated income, generally the start of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.StartDate</lx:path>
                                    <lx:label>Start Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TaxOfficeAssessments" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Are Tax Office Assessments for at least the past two years available as supporting evidence for this application</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.TaxOfficeAssessments</lx:path>
                                    <lx:label>Tax Office Assessments</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Accountant" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to the trust's accountant for the full previous financial year. This should be populated when the accountant for the specified period is different to that declared as the current accountant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomePrevious.x_Accountant</lx:path>
                                    <lx:label>Accountant</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="IncomePrior" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the share of the gross (pre-tax) income or income-related amounts in the prior financial year (two years prior to the most recent full financial year), attributable to this applicant (for year = current year -3).</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.IncomePrior</lx:path>
                                <lx:label>Income Prior</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Addback" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Container element for Addback detail</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback</lx:path>
                                      <lx:label>Addback</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="OtherAddback" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>Other addback that cannot otherwise be captured under a specific addback category.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.OtherAddback</lx:path>
                                            <lx:label>Other Add Back</lx:label>
                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Amount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Amount that can be an addback to the income.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.OtherAddback.Amount</lx:path>
                                                <lx:label>Amount</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Decription of the other addback.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.OtherAddback.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Allowances" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Allowances amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.Allowances</lx:path>
                                          <lx:label>Allowances</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AmortisationOfGoodwill" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Amortisation Of Goodwill amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.AmortisationOfGoodwill</lx:path>
                                          <lx:label>Amortisation Of Goodwill</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Bonus" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Bonus amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.Bonus</lx:path>
                                          <lx:label>Bonus</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CarryForwardLosses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Carry Forward Losses amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.CarryForwardLosses</lx:path>
                                          <lx:label>Carry Forward Losses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Depreciation" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Depreciation as recorded in the financial report</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.Depreciation</lx:path>
                                          <lx:label>Depreciation</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="HomeOffice" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Home Office amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.HomeOffice</lx:path>
                                          <lx:label>Home Office</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Interest" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The Trust's total interest expense in the prior tax year as listed in the financial report, including leasing expenses</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.Interest</lx:path>
                                          <lx:label>Interest</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Lease" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Lease amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.Lease</lx:path>
                                          <lx:label>Lease</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonCashBenefits" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-cash benefits amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.NonCashBenefits</lx:path>
                                          <lx:label>Non Cash Benefits</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonRecurringExpenses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-recurring expense amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.NonRecurringExpenses</lx:path>
                                          <lx:label>Non Recurring Expenses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PaymentToTrustee" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The payment to trustee that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.PaymentToTrustee</lx:path>
                                          <lx:label>Payment To Trustee</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Salary" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Salary amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.Salary</lx:path>
                                          <lx:label>Salary</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SuperannuationExcess" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Super excess amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.Addback.SuperannuationExcess</lx:path>
                                          <lx:label>Superannuation Excess</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="EndDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The end date for the stated income, generally the end of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.EndDate</lx:path>
                                    <lx:label>End Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncomeGreaterThanPreviousYear" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>When comparing past Tax Office Assessments, indicate whether the past return has improved on the prior year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.IncomeGreaterThanPreviousYear</lx:path>
                                    <lx:label>Income Greater Than Previous Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProfitAfterTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the profit after tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.ProfitAfterTax</lx:path>
                                    <lx:label>Profit After Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProfitBeforeTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the net profit/profit before tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.ProfitBeforeTax</lx:path>
                                    <lx:label>Profit Before Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The start date for the stated income, generally the start of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.StartDate</lx:path>
                                    <lx:label>Start Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TaxOfficeAssessments" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Are Tax Office Assessments for at least the past two years available as supporting evidence for this application</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.TaxOfficeAssessments</lx:path>
                                    <lx:label>Tax Office Assessments</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Accountant" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to the trust's accountant for the full prior financial year. This should be populated when the accountant for the specified period is different to that declared as the current accountant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomePrior.x_Accountant</lx:path>
                                    <lx:label>Accountant</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="IncomeRecent" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the share of the gross (pre-tax) income or income-related amounts in the most recent full financial year, attributable to this applicant (year = current year -1).</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent</lx:path>
                                <lx:label>Income Recent</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Addback" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Container element for Addback detail</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback</lx:path>
                                      <lx:label>Addback</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="OtherAddback" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>Other addback that cannot otherwise be captured under a specific addback category.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.OtherAddback</lx:path>
                                            <lx:label>Other Add Back</lx:label>
                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Amount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Amount that can be an addback to the income.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.OtherAddback.Amount</lx:path>
                                                <lx:label>Amount</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Decription of the other addback.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.OtherAddback.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Allowances" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Allowances amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.Allowances</lx:path>
                                          <lx:label>Allowances</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AmortisationOfGoodwill" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Amortisation Of Goodwill amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.AmortisationOfGoodwill</lx:path>
                                          <lx:label>Amortisation Of Goodwill</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Bonus" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Bonus amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.Bonus</lx:path>
                                          <lx:label>Bonus</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CarryForwardLosses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Carry Forward Losses amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.CarryForwardLosses</lx:path>
                                          <lx:label>Carry Forward Losses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Depreciation" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Depreciation as recorded in the financial report</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.Depreciation</lx:path>
                                          <lx:label>Depreciation</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="HomeOffice" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Home Office amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.HomeOffice</lx:path>
                                          <lx:label>Home Office</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Interest" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Total interest expense as listed in the financial report, including leasing expenses</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.Interest</lx:path>
                                          <lx:label>Interest</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Lease" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Lease amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.Lease</lx:path>
                                          <lx:label>Lease</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonCashBenefits" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-cash benefits amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.NonCashBenefits</lx:path>
                                          <lx:label>Non Cash Benefits</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonRecurringExpenses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-recurring expense amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.NonRecurringExpenses</lx:path>
                                          <lx:label>Non Recurring Expenses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PaymentToTrustee" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The payment to trustee that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.PaymentToTrustee</lx:path>
                                          <lx:label>Payment To Trustee</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Salary" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Salary amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.Salary</lx:path>
                                          <lx:label>Salary</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SuperannuationExcess" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Super excess amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.Addback.SuperannuationExcess</lx:path>
                                          <lx:label>Superannuation Excess</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="EndDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The end date for the stated income, generally the end of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.EndDate</lx:path>
                                    <lx:label>End Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncomeGreaterThanPreviousYear" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>When comparing past Tax Office Assessments, indicate whether the past return has improved on the previous year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.IncomeGreaterThanPreviousYear</lx:path>
                                    <lx:label>Income Greater Than Previous Year</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProfitAfterTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the profit after tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.ProfitAfterTax</lx:path>
                                    <lx:label>Profit After Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProfitBeforeTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the net profit/profit before tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.ProfitBeforeTax</lx:path>
                                    <lx:label>Profit Before Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProofCode" type="proofCodeTrustList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Code to indicate the type of evidence of the income</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.ProofCode</lx:path>
                                    <lx:label>Proof Code</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProofSighted" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the Income evidence been sighted by the loan writer or assessor</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.ProofSighted</lx:path>
                                    <lx:label>Proof Sighted</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The start date for the stated income, generally the start of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.StartDate</lx:path>
                                    <lx:label>Start Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="TaxOfficeAssessments" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Are Tax Office Assessments for at least the past two years available as supporting evidence for this application</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.TaxOfficeAssessments</lx:path>
                                    <lx:label>Tax Office Assessments</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Accountant" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to the trust's accountant for the full most recent financial year. This should be populated when the accountant for the specified period is different to that declared as the current accountant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomeRecent.x_Accountant</lx:path>
                                    <lx:label>Accountant</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="IncomeYearToDate" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the current year to date share of the gross (pre-tax) income or income-related amounts, attributable to this applicant.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate</lx:path>
                                <lx:label>Income Year To Date</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Addback" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Container element for Addback detail</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback</lx:path>
                                      <lx:label>Addback</lx:label>
                                      <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="OtherAddback" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>Other addback that cannot otherwise be captured under a specific addback category.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.OtherAddback</lx:path>
                                            <lx:label>Other Add Back</lx:label>
                                            <li:transactions>CAL,CDA,SVC</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Amount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Amount that can be an addback to the income.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.OtherAddback.Amount</lx:path>
                                                <lx:label>Amount</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Decription of the other addback.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.OtherAddback.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL,CDA,SVC</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Allowances" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Allowances amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.Allowances</lx:path>
                                          <lx:label>Allowances</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AmortisationOfGoodwill" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Amortisation Of Goodwill amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.AmortisationOfGoodwill</lx:path>
                                          <lx:label>Amortisation Of Goodwill</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Bonus" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Bonus amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.Bonus</lx:path>
                                          <lx:label>Bonus</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CarryForwardLosses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Carry Forward Losses amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.CarryForwardLosses</lx:path>
                                          <lx:label>Carry Forward Losses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Depreciation" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Depreciation as recorded in the financial report</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.Depreciation</lx:path>
                                          <lx:label>Depreciation</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="HomeOffice" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Home Office amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.HomeOffice</lx:path>
                                          <lx:label>Home Office</lx:label>
                                          <li:transactions>CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Interest" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Total interest expense as listed in the financial report, including leasing expenses</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.Interest</lx:path>
                                          <lx:label>Interest</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Lease" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Lease amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.Lease</lx:path>
                                          <lx:label>Lease</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonCashBenefits" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-cash benefits amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.NonCashBenefits</lx:path>
                                          <lx:label>Non Cash Benefits</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="NonRecurringExpenses" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Non-recurring expense amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.NonRecurringExpenses</lx:path>
                                          <lx:label>Non Recurring Expenses</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PaymentToTrustee" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The payment to trustee that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.PaymentToTrustee</lx:path>
                                          <lx:label>Payment To Trustee</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Salary" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Salary amount that can be an addback to the income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.Salary</lx:path>
                                          <lx:label>Salary</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SuperannuationExcess" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Super excess amount that can be an addback to the income</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.Addback.SuperannuationExcess</lx:path>
                                          <lx:label>Superannuation Excess</lx:label>
                                          <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="EndDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The end date for the stated income, generally the end of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.EndDate</lx:path>
                                    <lx:label>End Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProfitAfterTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the profit after tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.ProfitAfterTax</lx:path>
                                    <lx:label>Profit After Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ProfitBeforeTax" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the net profit/profit before tax for the specified financial year.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.ProfitBeforeTax</lx:path>
                                    <lx:label>Profit Before Tax</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="StartDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The start date for the stated income, generally the start of a financial year</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.IncomeYearToDate.StartDate</lx:path>
                                    <lx:label>Start Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Settlor" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Details of a settlor of the trust.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.Settlor</lx:path>
                                <lx:label>Settlor</lx:label>
                                <li:transactions>CAL,CDA,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Settlor" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to a person who is a settlor of the trust.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Settlor.x_Settlor</lx:path>
                                    <lx:label>Settlor</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="SourceOfWealth" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The source of the applicant's entire body of wealth.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.SourceOfWealth</lx:path>
                                <lx:label>Source Of Wealth</lx:label>
                                <li:transactions>CAL,CDA,DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Detail" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The details of the source of wealth.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.SourceOfWealth.Detail</lx:path>
                                    <lx:label>Detail</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Percentage" type="percentType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The percentage of the wealth from this source.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.SourceOfWealth.Percentage</lx:path>
                                    <lx:label>Percentage</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Type" type="sourceOfFundsTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of source of wealth.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.SourceOfWealth.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CAL,CDA,DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="TrustDeedVariation" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Container element for the date and description of a trust variation</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.TrustDeedVariation</lx:path>
                                <lx:label>Trust Deed Variation</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Date" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date of the variation to the trust</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.TrustDeedVariation.Date</lx:path>
                                    <lx:label>Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The description or type of variation that was effected</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.TrustDeedVariation.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Unique Identifier for this Trust Deed Variation Element</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.TrustDeedVariation.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Trustee" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Wrapper for the trustees</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Application.TrustApplicant.Trustee</lx:path>
                                <lx:label>Trustee</lx:label>
                                <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="AppointmentDate" type="dateType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date that the party was appointed as a trustee</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Trustee.AppointmentDate</lx:path>
                                    <lx:label>Appointment Date</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IndependentTrustee" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>An Independent Trustee is a person or corporate entity (company) who derives no benefit from the trust assets. The FMA issues licences to provide independent trustee services.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Trustee.IndependentTrustee</lx:path>
                                    <lx:label>Independent Trustee</lx:label>
                                    <li:transactions>CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Trustee" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Links to the applicant or related party that is a trustee of the trust.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Application.TrustApplicant.Trustee.x_Trustee</lx:path>
                                    <lx:label>Trustee</lx:label>
                                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="ABN" type="abnPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the ABN of the trust applicant</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.ABN</lx:path>
                              <lx:label>ABN</lx:label>
                              <li:transactions>CAL,CDA,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ApplicantType" type="applicantTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Whether the trust is a borrower or a guarantor</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.ApplicantType</lx:path>
                              <lx:label>Applicant Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,LMI,SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BusinessName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The business name or trading name of the Trust</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.BusinessName</lx:path>
                              <lx:label>Business Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BusinessNameSameAsTrustName" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates if the Business Name is the same as the Trust Name.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.BusinessNameSameAsTrustName</lx:path>
                              <lx:label>Business Name Same As Trust Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CountryEstablished" type="countryCodeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The country where the Trust was established</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.CountryEstablished</lx:path>
                              <lx:label>Country Established</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CustomerTypeCode" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A lender assigned code for the customer type.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.CustomerTypeCode</lx:path>
                              <lx:label>Customer Type Code</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CustomerTypeDescription" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A lender assigned description of the customer type.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.CustomerTypeDescription</lx:path>
                              <lx:label>Customer Type Description</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DocumentationType" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of documentation under which the Trust is constituted e.g. Deed of Trust</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.DocumentationType</lx:path>
                              <lx:label>Documentation Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Established" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag that indicates whether the trust has already been established.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.Established</lx:path>
                              <lx:label>Established</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="EstablishmentDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date that the trust was established</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.EstablishmentDate</lx:path>
                              <lx:label>Establishment Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="GovernmentOrganisation" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The applicant is a government organisation.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.GovernmentOrganisation</lx:path>
                              <lx:label>Government Organisation</lx:label>
                              <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IRDNumber" type="irdNumberPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The IRD Number of this Trust Applicant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.IRDNumber</lx:path>
                              <lx:label>IRD Number</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsExistingCustomer" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the trust applicant is an existing customer of the lender being applied to.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.IsExistingCustomer</lx:path>
                              <lx:label>Is Existing Customer</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NumberOfBeneficiaries" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The number of beneficiaries of the Trust</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.NumberOfBeneficiaries</lx:path>
                              <lx:label>Number Of Beneficiaries</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NumberOfTrustees" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The number of trustees of the Trust</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.NumberOfTrustees</lx:path>
                              <lx:label>Number Of Trustees</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NZBN" type="nzbnPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures the NZBN of the trust applicant</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.NZBN</lx:path>
                              <lx:label>NZBN</lx:label>
                              <li:transactions>CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OECDCRSStatus" type="oecdCRSStatusList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the trust is considered Active or Passive for OECD Common Reporting Standard purposes.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.OECDCRSStatus</lx:path>
                              <lx:label>OECD CRS Status</lx:label>
                              <li:transactions>CAL,CDA,DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PrimaryApplicant" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the lender should consider the party to be the primary applicant when processing the loan application. The significance of the primary applicant to the application varies from lender to lender.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.PrimaryApplicant</lx:path>
                              <lx:label>Primary Applicant</lx:label>
                              <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this trust applicant within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SettlorName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>This item is now deprecated and will be deleted in a release in 2019. The name of the person or company who originally provided the property on behalf of the Trust Beneficiaries, thereby establishing the Trust.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.SettlorName</lx:path>
                              <lx:label>Settlor Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="TrustName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The legal name of the Trust.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.TrustName</lx:path>
                              <lx:label>Trust Name</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="TrustPurpose" type="trustPurposeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The purpose for which the trust has been established</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.TrustPurpose</lx:path>
                              <lx:label>Trust Purpose</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="TrustStructure" type="trustStructureList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The structure of the trust as with particular regard to the discretion of the trustee</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.TrustStructure</lx:path>
                              <lx:label>Trust Structure</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="VestingDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>For a fixed trust, the date on which the property of the trust will vest in the beneficiaries.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.VestingDate</lx:path>
                              <lx:label>Vesting Date</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Accountant" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Link to the trust's current accountant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Application.TrustApplicant.x_Accountant</lx:path>
                              <lx:label>Accountant</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="VendorTaxInvoice" type="vendorTaxInvoiceType" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Contains the details of a vendor tax invoice for assets purchased or being purchased as part of the transaction</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Application.VendorTaxInvoice</lx:path>
                          <lx:label>Vendor Tax Invoice</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                    </xs:element>
                  </xs:sequence>
                  <xs:attribute name="ProductionData" type="yesNoList" use="optional">
                    <xs:annotation>
                      <xs:documentation>Indicates whether this file contains production or test data.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.Application.ProductionData</lx:path>
                        <lx:label>Production Data</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                    <xs:annotation>
                      <xs:documentation>System assigned unique identifier.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.Application.UniqueID</lx:path>
                        <lx:label>Unique ID</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                </xs:complexType>
              </xs:element>
              <xs:element name="CreditAssessment" minOccurs="0" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>Captures details relating to the credit assessment of an application, such as policy exceptions, reports, scores and an overall recommendation.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Content.CreditAssessment</lx:path>
                    <lx:label>Credit Assessment</lx:label>
                    <li:transactions>ACC,CDA,LMI</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:sequence>
                    <xs:element name="DecisionScore" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>A score that is typically provided by a decision engine to provide assistance in assessing the credit worthiness of an actor in the application such as an applicant or security.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.CreditAssessment.DecisionScore</lx:path>
                          <lx:label>Decision Score</lx:label>
                          <li:transactions>ACC,CDA,LMI</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="Determined" type="creditAssessmentDecisionScoreDeterminedList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates whether the decision score is Pre-Bureau or Post-Bureau or based on Behavioural Score.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.DecisionScore.Determined</lx:path>
                              <lx:label>Determined</lx:label>
                              <li:transactions>CDA,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Recommendation" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The recommendation driven by the score as defined by lending policy.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.DecisionScore.Recommendation</lx:path>
                              <lx:label>Recommendation</lx:label>
                              <li:transactions>ACC,CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RiskGrade" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The risk grading assigned to the decision score as per lending policy. The lower the number the higher the risk associated with the decision score.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.DecisionScore.RiskGrade</lx:path>
                              <lx:label>RiskGrade</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Score" type="decimalType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The decision score numeric value.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.DecisionScore.Score</lx:path>
                              <lx:label>Score</lx:label>
                              <li:transactions>ACC,CDA,LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Party" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Link to the party the decision score was determined for.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.DecisionScore.x_Party</lx:path>
                              <lx:label>Party</lx:label>
                              <li:transactions>ACC,CDA,LMI</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Security" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Link to the security the decision score was determined for.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.DecisionScore.x_Security</lx:path>
                              <lx:label>Security</lx:label>
                              <li:transactions>CDA,LMI</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                <lx:target>Package.Content.Application.GeneralSecurityAgreement</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="PolicyException" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>A credit policy that has not had its requirements met via the application data provided that needs to be addressed by some means before full approval can be granted.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.CreditAssessment.PolicyException</lx:path>
                          <lx:label>Policy Exception</lx:label>
                          <li:transactions>CDA</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Override" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the policy exception override.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.CreditAssessment.PolicyException.Override</lx:path>
                                <lx:label>Override</lx:label>
                                <li:transactions>CDA</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Category" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The category of the reason provided for overriding a policy exception.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.CreditAssessment.PolicyException.Override.Category</lx:path>
                                    <lx:label>Category</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Reason" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The reason provided for overriding a policy exception.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.CreditAssessment.PolicyException.Override.Reason</lx:path>
                                    <lx:label>Reason</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Timestamp" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date and time the override was performed.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.CreditAssessment.PolicyException.Override.Timestamp</lx:path>
                                    <lx:label>Timestamp</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UserID" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The system generated identifier of the user who overrode the policy exception.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.CreditAssessment.PolicyException.Override.UserID</lx:path>
                                    <lx:label>UserID</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="Category" type="creditAssessmentPolicyExceptionCategoryList" use="optional">
                          <xs:annotation>
                            <xs:documentation>How the policy exception can be broadly categorised as such as fraud, score, serviceability etc.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.PolicyException.Category</lx:path>
                              <lx:label>Category</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Code" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A code that can be used to identify the policy exception type uniquely from other types.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.PolicyException.Code</lx:path>
                              <lx:label>Code</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CreditDecisionEngineIdentifier" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>An identifier issued by the credit decision engine that can uniquely identify the policy exception.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.PolicyException.CreditDecisionEngineIdentifier</lx:path>
                              <lx:label>Credit Decision Engine Identifier</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DelegatedAuthorityLevel" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The level of delegated authority required to override the policy exception.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.PolicyException.DelegatedAuthorityLevel</lx:path>
                              <lx:label>Delegated Authority Level</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Description" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A text description of the policy exception. This can include the definition of the policy and why it would be raised.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.PolicyException.Description</lx:path>
                              <lx:label>Description</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Outcome" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The outcome derived from the policy exception. For example, the outcome might be a referral or a decline based on the credit policy and the importance placed on the exception.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.PolicyException.Outcome</lx:path>
                              <lx:label>Outcome</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Overridden" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Whether the policy exception has been overridden to allow the application to progress toward full or unconditional approval.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.PolicyException.Overridden</lx:path>
                              <lx:label>Overridden</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier for the policy exception.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.PolicyException.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Party" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Link to the party that triggered the policy exception.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.PolicyException.x_Party</lx:path>
                              <lx:label>Party</lx:label>
                              <li:transactions>CDA</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Security" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Link to the security that triggered the policy exception.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.PolicyException.x_Security</lx:path>
                              <lx:label>Security</lx:label>
                              <li:transactions>CDA</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                <lx:target>Package.Content.Application.GeneralSecurityAgreement</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Recommendation" minOccurs="1" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>The overall recommendation for the application such as full or unconditional approval, referral for manual assessment, decline etc.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.CreditAssessment.Recommendation</lx:path>
                          <lx:label>Recommendation</lx:label>
                          <li:transactions>CDA</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="MaximumCreditLimit" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The maximum credit amount that can be offered to the applicants as determined by credit rules most likely contained in a credit decision engine.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.Recommendation.MaximumCreditLimit</lx:path>
                              <lx:label>Maximum Credit Limit</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OverallRecommendation" type="creditAssessmentRecommendationList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The overall recommended decision determined upon assessment of the application for credit worthiness.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.Recommendation.OverallRecommendation</lx:path>
                              <lx:label>Overall Recommendation</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Reason" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A brief description as to how the overall recommendation was determined.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.Recommendation.Reason</lx:path>
                              <lx:label>Reason</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Report" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>A bureau, identity, credit or other type of report that may be used in assessing the credit worthiness of an application or its actors.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.CreditAssessment.Report</lx:path>
                          <lx:label>Report</lx:label>
                          <li:transactions>ACC,CDA</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Content" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The raw content of the report. Only one Content element should exist for each Content.Type.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.CreditAssessment.Report.Content</lx:path>
                                <lx:label>Content</lx:label>
                                <li:transactions>CDA</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Type" type="reportContentTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of the content such as string, XML or HTML.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.CreditAssessment.Report.Content.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Attachment" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Link to the actual report content in encoded form.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.CreditAssessment.Report.Content.x_Attachment</lx:path>
                                    <lx:label>Attachment</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Attachment</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="RelatedSearch" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Details of any related searches that may have been performed that are related to the report returned.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.CreditAssessment.Report.RelatedSearch</lx:path>
                                <lx:label>Related Search</lx:label>
                                <li:transactions>CDA</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Wrapper element for the name of the person that the related search is for.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.CreditAssessment.Report.RelatedSearch.PersonName</lx:path>
                                      <lx:label>Person Name</lx:label>
                                      <li:transactions>CDA</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="ReferenceType" type="relatedSearchReferenceTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of relationship to the parent report that triggered a related search.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.CreditAssessment.Report.RelatedSearch.ReferenceType</lx:path>
                                    <lx:label>Reference Type</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="FormatVersion" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned version relating to the report format.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.Report.FormatVersion</lx:path>
                              <lx:label>Format Version</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Name" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The name of the report such as identity check for J Doe, credit report etc.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.Report.Name</lx:path>
                              <lx:label>Name</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RequestDateTime" type="dateTimeType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date and time the report was requested.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.Report.RequestDateTime</lx:path>
                              <lx:label>Request Date Time</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Source" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The system the report was sourced from e.g. 'Bureau', 'PPSR', 'Veda IDMatrix' etc.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.Report.Source</lx:path>
                              <lx:label>Source</lx:label>
                              <li:transactions>ACC,CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Type" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of the report such as bureau or identity. This is not presently an enumerated type as there are too many unknown report types that may be valid.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.Report.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier for the report.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.Report.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>ACC,CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Party" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The party the report pertains to.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.Report.x_Party</lx:path>
                              <lx:label>Party</lx:label>
                              <li:transactions>ACC,CDA</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                                <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Security" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The security the report pertains to.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.CreditAssessment.Report.x_Security</lx:path>
                              <lx:label>Security</lx:label>
                              <li:transactions>CDA</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                <lx:target>Package.Content.Application.GeneralSecurityAgreement</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                  </xs:sequence>
                  <xs:attribute name="x_Application" type="referenceType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The application the credit assessment has been conducted on.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.CreditAssessment.x_Application</lx:path>
                        <lx:label>Application</lx:label>
                        <li:transactions>CDA</li:transactions>
                        <lx:references>
                          <lx:target>Package.Content.Application</lx:target>
                        </lx:references>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                </xs:complexType>
              </xs:element>
              <xs:element name="History" minOccurs="0" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>A placeholder to group the various types of history that are associated with an application for financial item(s).</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Content.History</lx:path>
                    <lx:label>History</lx:label>
                    <li:transactions>CDA</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:sequence>
                    <xs:element name="Status" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Details relating to each change in status of the application.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.History.Status</lx:path>
                          <lx:label>Status</lx:label>
                          <li:transactions>CDA</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="DateTime" type="dateTimeType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date and time at which the application was granted this status.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.History.Status.DateTime</lx:path>
                              <lx:label>Date Time</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="InitiatedByName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The name of the system or user that initiated changing the application to this status.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.History.Status.InitiatedByName</lx:path>
                              <lx:label>Initiated By Name</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Name" type="statusNameApplicationInstructionsList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The name of this application status.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.History.Status.Name</lx:path>
                              <lx:label>Name</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.History.Status.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Submission" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Details relating to each submission of the application.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.History.Submission</lx:path>
                          <lx:label>Submission</lx:label>
                          <li:transactions>CDA</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Submitter" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains details of the person or system that submitted the application.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.History.Submission.Submitter</lx:path>
                                <lx:label>Office Fax</lx:label>
                                <li:transactions>CDA</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Department" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the department within the submitting organisation that the submitter belongs to.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.History.Submission.Submitter.Department</lx:path>
                                    <lx:label>Department</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the name of the submitter, whether the submitter is a person or a system.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.History.Submission.Submitter.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UserID" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The system generated identifier of the user who submitted the application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.History.Submission.Submitter.UserID</lx:path>
                                    <lx:label>User ID</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="DateTime" type="dateTimeType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date and time at which the submission is specified as having occurred.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.History.Submission.DateTime</lx:path>
                              <lx:label>Date Time</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Type" type="submissionTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indication of the type of submission that has been made, such as for a credit decision or from a broker.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.History.Submission.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.History.Submission.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                  </xs:sequence>
                  <xs:attribute name="x_Application" type="referenceType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The application to which the history applies.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.History.x_Application</lx:path>
                        <lx:label>Application</lx:label>
                        <li:transactions>CDA</li:transactions>
                        <lx:references>
                          <lx:target>Package.Content.Application</lx:target>
                        </lx:references>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                </xs:complexType>
              </xs:element>
              <xs:element name="NeedsAnalysis" minOccurs="0" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>An overview of the customers needs in requesting products from the financial institution.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Content.NeedsAnalysis</lx:path>
                    <lx:label>Needs Analysis</lx:label>
                    <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:sequence>
                    <xs:element name="BenefitToApplicants" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Information about the benefit derived by the applicants.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.NeedsAnalysis.BenefitToApplicants</lx:path>
                          <lx:label>Benefit To Applicants</lx:label>
                          <li:transactions>CAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="AllApplicantsBenefit" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that all applicants and co-applicants clearly benefit from the producrts within this application.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.BenefitToApplicants.AllApplicantsBenefit</lx:path>
                              <lx:label>All Applicants Benefit</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="BenefitEnquiries" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Describes the enquiries that have been made to ascertain the level of benefit to each applicant.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.BenefitToApplicants.BenefitEnquiries</lx:path>
                              <lx:label>Benefit Enquiries</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="CoApplicantInterview" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Interview that must be completed with each co-applicant.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview</lx:path>
                          <lx:label>Co Applicant Interview</lx:label>
                          <li:transactions>CAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Applicant" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The results of the interview with a co-applicant.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Applicant</lx:path>
                                <lx:label>Applicant</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="BenefitEnquiries" type="benefitEnquiriesList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>How the co-applicant substantially benefits from the products within this application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Applicant.BenefitEnquiries</lx:path>
                                    <lx:label>Benefit Enquiries</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="BenefitEnquiriesDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A description of how the co-applicant substantially benefits from the products within this application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Applicant.BenefitEnquiriesDescription</lx:path>
                                    <lx:label>Benefit Enquiries Description</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CoApplicantBenefit" type="coApplicantBenefitList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates whether the co-applicant substantially benefits from the products within this application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Applicant.CoApplicantBenefit</lx:path>
                                    <lx:label>Co Applicant Benefit</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CoApplicantReason" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The reason the borrower wants to be a co-applicant.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Applicant.CoApplicantReason</lx:path>
                                    <lx:label>Co Applicant Reason</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the co-applicant, to be used to identify the co-applicant when the application has not yet been created.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Applicant.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Number" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant number (e.g. Applicant 1, Applicant 2) to be used to identify the applicant when the application has not yet been created.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Applicant.Number</lx:path>
                                    <lx:label>Number</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SuspectDifficultyUnderstandingObligations" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the interviewer suspects the co-applicant is having difficulty understanding their obligations.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Applicant.SuspectDifficultyUnderstandingObligations</lx:path>
                                    <lx:label>Suspect Difficulty Understanding Obligations</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SuspectExperiencingFinancialAbuse" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the interviewer suspects the co-applicant is experiencing financial abuse.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Applicant.SuspectExperiencingFinancialAbuse</lx:path>
                                    <lx:label>Suspect Experiencing Financial Abuse</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SuspectUnderPressure" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the interviewer suspects the co-applicant is under pressure to enter the loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Applicant.SuspectUnderPressure</lx:path>
                                    <lx:label>Suspect Under Pressure</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SuspectUnfairConduct" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the interviewer suspects the co-applicant is subject to unreasonable or unfair conduct by someone involved in the loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Applicant.SuspectUnfairConduct</lx:path>
                                    <lx:label>Suspect Unfair Conduct</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UnderstandsCoApplicantObligations" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates that the co-applicant understands their obligations as co-applicant, and the difference between co-applicant and guarantor.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Applicant.UnderstandsCoApplicantObligations</lx:path>
                                    <lx:label>Understands Co Applicant Obligations</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Applicant" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the applicant (a Person Applicant element).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Applicant.x_Applicant</lx:path>
                                    <lx:label>Applicant</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Witness" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>A person that was a witness to the co-applicant interview.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Witness</lx:path>
                                <lx:label>Witness</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ANZSCOOccupationCode" type="anzscoOccupationCodeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Occupation code (ANZSCO) of the person who was a witness to the co-applicant interview.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Witness.ANZSCOOccupationCode</lx:path>
                                    <lx:label>ANZSCO Occupation Code</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Occupation" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Occupation description of the person who was a witness to the co-applicant interview.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Witness.Occupation</lx:path>
                                    <lx:label>Occupation</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Witness" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the person that was a witness to the co-applicant interview.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.CoApplicantInterview.Witness.x_Witness</lx:path>
                                    <lx:label>Witness</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="FutureCircumstances" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Planned or anticipated changes to the applicant/s future financial circumstances.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.NeedsAnalysis.FutureCircumstances</lx:path>
                          <lx:label>FutureCircumstances</lx:label>
                          <li:transactions>CAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Applicant" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>An applicant to which these future financial circumstances apply.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.NeedsAnalysis.FutureCircumstances.Applicant</lx:path>
                                <lx:label>Applicant</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Name" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the applicant, to be used to identify the applicant when the application has not yet been created.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.FutureCircumstances.Applicant.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Number" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant number (e.g. Applicant 1, Applicant 2) to be used to identify the applicant when the application has not yet been created.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.FutureCircumstances.Applicant.Number</lx:path>
                                    <lx:label>Number</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Applicant" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the applicant (a Person Applicant element).</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.FutureCircumstances.Applicant.x_Applicant</lx:path>
                                    <lx:label>Applicant</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="ResponsibleLending" type="responsibleLendingType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Planned or anticipated changes in circumstances that will adversely affect the applicants' ability to make future loan repayments; the nature of the change, and how they will collectively continue to make repayments.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.NeedsAnalysis.FutureCircumstances.ResponsibleLending</lx:path>
                                <lx:label>Responsible Lending</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                        </xs:sequence>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Interview" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Information about the needs analysis interview.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.NeedsAnalysis.Interview</lx:path>
                          <lx:label>Interview</lx:label>
                          <li:transactions>CAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Attendee" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Interview Attendee.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.NeedsAnalysis.Interview.Attendee</lx:path>
                                <lx:label>Interview</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Party" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The party that attended the interview.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.Interview.Attendee.x_Party</lx:path>
                                    <lx:label>Party</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.RelatedPerson</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="AllApplicantsPresent" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that all applicants and co-applicants were present for the needs analysis interview.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Interview.AllApplicantsPresent</lx:path>
                              <lx:label>All Applicants Present</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="AllApplicantsUnderstandEnglish" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that all the attendees of the needs analysis interview were clearly able to understand English.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Interview.AllApplicantsUnderstandEnglish</lx:path>
                              <lx:label>All Applicants Understand English</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Date" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date that the needs analysis interview takes place.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Interview.Date</lx:path>
                              <lx:label>Interview Date</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="InterpreterRecommended" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that the services of an interpreter were recommended.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Interview.InterpreterRecommended</lx:path>
                              <lx:label>Interpreter Recommended</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Location" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The location at which the needs analysis interview takes place.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Interview.x_Location</lx:path>
                              <lx:label>Location</lx:label>
                              <li:transactions>CAL</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.Address</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="LoanTermSought" type="durationType" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>The duration of the term of credit sought for this loan.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.NeedsAnalysis.LoanTermSought</lx:path>
                          <lx:label>Loan Term Sought</lx:label>
                          <li:transactions>CAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                    </xs:element>
                    <xs:element name="Preferences" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>The preferences of the applicant with respect to the requirements, type of facilities, or features of the products they are seeking.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.NeedsAnalysis.Preferences</lx:path>
                          <lx:label>Preferences</lx:label>
                          <li:transactions>CAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="FundsAccessTypeDetails" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the applicant's funds access type preferences.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails</lx:path>
                                <lx:label>Funds Access Type Details</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="OffsetAccount" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The applicant's preferences for offset account funds access type.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.OffsetAccount</lx:path>
                                      <lx:label>Offset Account</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Reason" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The reason offset account funds access type is important to the applicant.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.OffsetAccount.Reason</lx:path>
                                            <lx:label>Reason</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="AllowsAccessToFunds" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Offset account funds access type is important to the applicant to allow access to funds.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.OffsetAccount.Reason.AllowsAccessToFunds</lx:path>
                                                <lx:label>Allows Access to Funds</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="AllowsPayingOffLoanSooner" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Offset account funds access type is important to the applicant to allow paying off loan sooner.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.OffsetAccount.Reason.AllowsPayingOffLoanSooner</lx:path>
                                                <lx:label>Allows Paying off Loan Sooner</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Details of the other reason offset account funds access type is important to the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.OffsetAccount.Reason.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ForTaxPurposes" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Offset account funds access type is important to the applicant for tax purposes (based on financial or tax advice).</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.OffsetAccount.Reason.ForTaxPurposes</lx:path>
                                                <lx:label>For Tax Purposes</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Other" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The reason offset account funds access type is important to the applicant cannot be captured in the specific reason data fields.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.OffsetAccount.Reason.Other</lx:path>
                                                <lx:label>Other</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Importance" type="importanceList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The importance of this feature to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.OffsetAccount.Importance</lx:path>
                                          <lx:label>Importance</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RisksExplained" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The risks of this feature have been explained to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.OffsetAccount.RisksExplained</lx:path>
                                          <lx:label>Risks Explained</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="Redraw" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The applicant's preferences for redraw funds access type.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.Redraw</lx:path>
                                      <lx:label>Redraw</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Reason" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The reason redraw funds access type is important to the applicant.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.Redraw.Reason</lx:path>
                                            <lx:label>Reason</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Details of the other reason redraw funds access type is important to the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.Redraw.Reason.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="FlexibilityToAccessPrepaidFundsIfNeeded" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Redraw funds access type is important to the applicant for flexibility to access prepaid funds if needed.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.Redraw.Reason.FlexibilityToAccessPrepaidFundsIfNeeded</lx:path>
                                                <lx:label>Flexibility to Access Prepaid Funds If Needed</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Other" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The reason redraw funds access type is important to the applicant cannot be captured in the specific reason data fields.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.Redraw.Reason.Other</lx:path>
                                                <lx:label>Other</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Importance" type="importanceList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The importance of this feature to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.Redraw.Importance</lx:path>
                                          <lx:label>Importance</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RisksExplained" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The risks of this feature have been explained to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessTypeDetails.Redraw.RisksExplained</lx:path>
                                          <lx:label>RisksExplained</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="InterestRateTypeDetails" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the applicant's interest rate type preference.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails</lx:path>
                                <lx:label>Interest Rate Type Details</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="FixedAndVariableRate" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The applicant's preferences for a fixed and variable rate loan (i.e split loan).</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedAndVariableRate</lx:path>
                                      <lx:label>Fixed and Variable Rate</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="FixedPeriodDuration" type="durationType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The preferred duration of the fixed rate period.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedAndVariableRate.FixedPeriodDuration</lx:path>
                                            <lx:label>Fixed Period Duration</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="Reason" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The reason fixed and variable rate is important to the applicant.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedAndVariableRate.Reason</lx:path>
                                            <lx:label>Reason</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Details of the other reason fixed and variable rate is important to the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedAndVariableRate.Reason.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Flexibility" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Fixed and variable rate is important to the applicant to retain a degree of flexibility in relation to increased repayments, redraws and/or early repayment of part of the loan.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedAndVariableRate.Reason.Flexibility</lx:path>
                                                <lx:label>Flexibility</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="LimitingRateIncreaseRisk" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Fixed and variable rate is important to the applicant to limit risk of increasing variable interest rate to obtain some benefit from potential future decreases in the interest rate.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedAndVariableRate.Reason.LimitingRateIncreaseRisk</lx:path>
                                                <lx:label>Limiting Rate Increase Risk</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="MakeBudgetingEasier" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Fixed and variable rate is important to the applicant to make budgeting easier than if the entire loan were variable.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedAndVariableRate.Reason.MakeBudgetingEasier</lx:path>
                                                <lx:label>Make Budgeting Easier</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Other" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The reason fixed and variable rate is important to the applicant cannot be captured in the specific reason data fields.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedAndVariableRate.Reason.Other</lx:path>
                                                <lx:label>Other</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Importance" type="importanceList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The importance of this feature to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedAndVariableRate.Importance</lx:path>
                                          <lx:label>Importance</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RisksExplained" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The risks of this feature have been explained to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedAndVariableRate.RisksExplained</lx:path>
                                          <lx:label>Risks Explained</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="FixedRate" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The applicant's preferences for a fixed rate loan.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedRate</lx:path>
                                      <lx:label>Fixed Rate</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="FixedPeriodDuration" type="durationType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The preferred duration of the fixed rate period.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedRate.FixedPeriodDuration</lx:path>
                                            <lx:label>Fixed Period Duration</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="Reason" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The reason fixed rate is important to the applicant.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedRate.Reason</lx:path>
                                            <lx:label>Reason</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="AvoidingRateIncreaseRisk" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Fixed rate is important to the applicant to avoid risk of increasing variable interest rate during fixed interest period.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedRate.Reason.AvoidingRateIncreaseRisk</lx:path>
                                                <lx:label>Avoiding Rate Increase Risk</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Details of the other reason fixed rate is important to the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedRate.Reason.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="MakeBudgetingEasier" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Fixed rate is important to the applicant to make budgeting easier.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedRate.Reason.MakeBudgetingEasier</lx:path>
                                                <lx:label>Make Budgeting Easier</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Other" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The reason fixed rate is important to the applicant cannot be captured in the specific reason data fields.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedRate.Reason.Other</lx:path>
                                                <lx:label>Other</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Importance" type="importanceList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The importance of this feature to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedRate.Importance</lx:path>
                                          <lx:label>Importance</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RisksExplained" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The risks of this feature have been explained to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.FixedRate.RisksExplained</lx:path>
                                          <lx:label>Risks Explained</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="VariableRate" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The applicant's preferences for a variable rate loan.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.VariableRate</lx:path>
                                      <lx:label>Variable Rate</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Reason" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The reason variable rate is important to the applicant.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.VariableRate.Reason</lx:path>
                                            <lx:label>Variable Rate Reason</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Details of the other reason variable rate is important to the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.VariableRate.Reason.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Flexibility" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Variable rate is important to the applicant to retain flexibility with respect to repayment, redraw and/or early repayment of loan.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.VariableRate.Reason.Flexibility</lx:path>
                                                <lx:label>Flexibility</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Other" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The reason variable rate is important to the applicant cannot be captured in the specific reason data fields.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.VariableRate.Reason.Other</lx:path>
                                                <lx:label>Other</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="PotentialRateDecreases" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Variable rate is important to the applicant to take advantage of potential future decreases in the interest rate.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.VariableRate.Reason.PotentialRateDecreases</lx:path>
                                                <lx:label>Potential Rate Decreases</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Importance" type="importanceList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The importance of this feature to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.VariableRate.Importance</lx:path>
                                          <lx:label>Importance</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RisksExplained" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The risks of this feature have been explained to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateTypeDetails.VariableRate.RisksExplained</lx:path>
                                          <lx:label>Risks Explained</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="RepaymentTypeDetails" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Details of the applicant's repayment type preferences.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails</lx:path>
                                <lx:label>Repayment Type Details</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="InterestInAdvance" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The applicant's preferences for interest in advance repayment type.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestInAdvance</lx:path>
                                      <lx:label>Interest in Advance</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Reason" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The reason interest in advance repayment type is important to the applicant.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestInAdvance.Reason</lx:path>
                                            <lx:label>Reason</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Details of the other reason interest in advance is important to the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestInAdvance.Reason.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="DiscountsOnInterestRate" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Interest in advance is important to the applicant for discounts on interest rate.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestInAdvance.Reason.DiscountsOnInterestRate</lx:path>
                                                <lx:label>Discounts on Interest Rate</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ForTaxPurposes" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Interest in advance is important to the applicant for tax purposes.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestInAdvance.Reason.ForTaxPurposes</lx:path>
                                                <lx:label>For Tax Purposes</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Other" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The reason interest in advance is important to the applicant cannot be captured in the specific reason data fields.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestInAdvance.Reason.Other</lx:path>
                                                <lx:label>Other</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="ToAssistWithCashFlowAndBudgeting" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Interest in advance is important to the applicant to assist with cash flow and budgeting.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestInAdvance.Reason.ToAssistWithCashFlowAndBudgeting</lx:path>
                                                <lx:label>To Assist with Cash Flow and Budgeting</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Importance" type="importanceList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The importance of this feature to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestInAdvance.Importance</lx:path>
                                          <lx:label>Importance</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RisksExplained" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The risks of this feature have been explained to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestInAdvance.RisksExplained</lx:path>
                                          <lx:label>Risks Explained</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="InterestOnly" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The applicant's preferences for interest only repayment type.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly</lx:path>
                                      <lx:label>Interest Only</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="InterestOnlyDuration" type="durationType" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The preferred duration of the interest only period.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.InterestOnlyDuration</lx:path>
                                            <lx:label>Interest Only Duration</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                      </xs:element>
                                      <xs:element name="Reason" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The reason interest only repayment type is important to the applicant.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Reason</lx:path>
                                            <lx:label>Reason</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="AccommodateAnticipatedNonRecurringExpenseItem" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Interest only is important to the applicant to accommodate an anticipated nonrecurring expense item (e.g. education, renovation/construction, furniture).</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Reason.AccommodateAnticipatedNonRecurringExpenseItem</lx:path>
                                                <lx:label>Accommodate Anticipated Non Recurring Expense Item</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="AccommodateTemporaryReductionInIncome" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Interest only is important to the applicant to accommodate a temporary reduction in income (e.g. parental leave, changing circumstances).</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Reason.AccommodateTemporaryReductionInIncome</lx:path>
                                                <lx:label>Accommodate Temporary Reduction in Income</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="CreateFundsForInvestmentPurposes" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Interest only is important to the applicant to create funds for investment purposes.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Reason.CreateFundsForInvestmentPurposes</lx:path>
                                                <lx:label>Create Funds for Investment Purposes</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Details of the other reason interest only is important to the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Reason.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="MaximiseCashFlow" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Interest only is important to the applicant to maximise cash flow.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Reason.MaximiseCashFlow</lx:path>
                                                <lx:label>Maximise Cash Flow</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="MinimiseCashFlowImpactDuringConstruction" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Interest only is important to the applicant to minimise cash flow impact during construction.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Reason.MinimiseCashFlowImpactDuringConstruction</lx:path>
                                                <lx:label>Minimise Cash Flow Impact During Construction</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="MinimiseCashFlowImpactDuringTheBridgingPeriod" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Interest only is important to the applicant to minimise cash flow Impact during the bridging period.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Reason.MinimiseCashFlowImpactDuringTheBridgingPeriod</lx:path>
                                                <lx:label>Minimise Cash Flow Impact During the Bridging Period</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Other" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The reason interest only is important to the applicant cannot be captured in the specific reason data fields.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Reason.Other</lx:path>
                                                <lx:label>Other</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="PlanToConvertToInvestmentProperty" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Interest only is important to the applicant because they plan to convert the property to an investment property.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Reason.PlanToConvertToInvestmentProperty</lx:path>
                                                <lx:label>Plan to Convert to Investment Property</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="PrincipalReductionsInAnOffsetFacility" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Interest only is important to the applicant for principal reductions in an offset facility.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Reason.PrincipalReductionsInAnOffsetFacility</lx:path>
                                                <lx:label>Principal Reductions in an Offset Facility</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Recommendation" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The reason interest only is important to the applicant is recommendation provided by an independent financial advisor or accountant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Reason.Recommendation</lx:path>
                                                <lx:label>Recommendation</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="TaxationOrAccountingReasons" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The reason interest only is important to the applicant is taxation or accounting reasons (based on financial or tax advice), including: release funds for investment purposes (e.g. shares, investment property, super contributions); priority is paying off nondeductable debts (this loan is for investment purposes); and plan to convert to an investment property in future.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Reason.TaxationOrAccountingReasons</lx:path>
                                                <lx:label>Taxation or Accounting Reasons</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="VariableAndUnpredictableIncome" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The reason interest only is important to the applicant is variable and unpredictable income.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Reason.VariableAndUnpredictableIncome</lx:path>
                                                <lx:label>Variable and Unpredictable Income</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Importance" type="importanceList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The importance of this feature to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.Importance</lx:path>
                                          <lx:label>Importance</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="InterestOnlyDurationReason" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The reason why the proposed length of the interest only term is appropriate for the customer's needs.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.InterestOnlyDurationReason</lx:path>
                                          <lx:label>Interest Only Duration Reason</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RepaymentFrequency" type="frequencyShortList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The preferred repayment frequency.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.RepaymentFrequency</lx:path>
                                          <lx:label>Repayment Frequency</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RisksExplained" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The risks of this feature have been explained to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.InterestOnly.RisksExplained</lx:path>
                                          <lx:label>Risks Explained</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="LineOfCredit" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The applicant's preferences for line of credit repayment type.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit</lx:path>
                                      <lx:label>Line of Credit</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Reason" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The reason line of credit repayment type is important to the applicant.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.Reason</lx:path>
                                            <lx:label>Reason</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="AllowsAccessToFundsAtAnyTime" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Line of credit repayment type is important to the applicant to allows access to funds at any time.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.Reason.AllowsAccessToFundsAtAnyTime</lx:path>
                                                <lx:label>Allows Access to Funds at Any Time</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="AnticipatedVariableCashFlows" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Line of credit repayment type is important to the applicant due to anticipated variable cash flows.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.Reason.AnticipatedVariableCashFlows</lx:path>
                                                <lx:label>Anticipated Variable Cash Flows</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="AssistsWithInvestmentPurchasesAndTaxPlanning" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Line of credit repayment type is important to the applicant to assist with investment purchases and tax planning.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.Reason.AssistsWithInvestmentPurchasesAndTaxPlanning</lx:path>
                                                <lx:label>Assists with Investment Purchases and Tax Planning</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Details of the other reason interest only is important to the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.Reason.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="FlexibilityToManageRepayments" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Line of credit repayment type is important to the applicant for flexibility to manage repayments.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.Reason.FlexibilityToManageRepayments</lx:path>
                                                <lx:label>Flexibility to Manage Repayments</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="NoNeedToApplyForFurtherLendingInFuture" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Line of credit repayment type is important to the applicant because there is no need to apply for further lending in future.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.Reason.NoNeedToApplyForFurtherLendingInFuture</lx:path>
                                                <lx:label>No Need to Apply for Further Lending in Future</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="OngoingNeedForFunds" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Line of credit repayment type is important to the applicant for an ongoing need for funds - planning to make a few purchases over a period of time.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.Reason.OngoingNeedForFunds</lx:path>
                                                <lx:label>Ongoing Need for Funds</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Other" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The reason line of credit repayment type is important to the applicant cannot be captured in the specific reason data fields.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.Reason.Other</lx:path>
                                                <lx:label>Other</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="RepaymentPlan" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The applicant's plan for paying off line of credit at the end of or during the term.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.RepaymentPlan</lx:path>
                                            <lx:label>Repayment Plan</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="CoApplicantIncome" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The applicant's plan for paying off line of credit is co-applicant's income.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.RepaymentPlan.CoApplicantIncome</lx:path>
                                                <lx:label>Co Applicant Income</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Details of the other reason line of credit is important to the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.RepaymentPlan.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Downsizing" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The applicant's plan for paying off line of credit is downsizing.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.RepaymentPlan.Downsizing</lx:path>
                                                <lx:label>Downsizing</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="IncomeFromOtherInvestments" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The applicant's plan for paying off line of credit is income from other investments.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.RepaymentPlan.IncomeFromOtherInvestments</lx:path>
                                                <lx:label>Income from Other Investments</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Other" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The applicant's plan for paying off line of credit cannot be captured in the specific data fields.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.RepaymentPlan.Other</lx:path>
                                                <lx:label>Other</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="RepaymentOfLoanPriorToEndOfTerm" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The applicant's plan for paying off line of credit is repayment of loan prior to end of term.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.RepaymentPlan.RepaymentOfLoanPriorToEndOfTerm</lx:path>
                                                <lx:label>Repayment of Loan Prior to End of Term</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="SaleOfAssets" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The applicant's plan for paying off line of credit is sale of assets.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.RepaymentPlan.SaleOfAssets</lx:path>
                                                <lx:label>Sale of Assets</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Savings" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The applicant's plan for paying off line of credit is savings.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.RepaymentPlan.Savings</lx:path>
                                                <lx:label>Savings</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Importance" type="importanceList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The importance of this feature to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.Importance</lx:path>
                                          <lx:label>Importance</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RisksExplained" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The risks of this feature have been explained to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.LineOfCredit.RisksExplained</lx:path>
                                          <lx:label>Risks Explained</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="PrincipalAndInterest" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The applicant's preferences for principal and interest repayment type.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.PrincipalAndInterest</lx:path>
                                      <lx:label>Principal and Interest</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Reason" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>The reason principal and interest repayment type is important to the applicant.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.PrincipalAndInterest.Reason</lx:path>
                                            <lx:label>Reason</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="BuildUpEquityFromTheStart" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Principal and interest repayment type is important to the applicant to build up equity from the start.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.PrincipalAndInterest.Reason.BuildUpEquityFromTheStart</lx:path>
                                                <lx:label>Build up Equity from the Start</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Details of the other reason principal and interest repayment type is important to the applicant.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.PrincipalAndInterest.Reason.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="HigherLendingLimit" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Principal and interest repayment type is important to the applicant for a higher lending limit.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.PrincipalAndInterest.Reason.HigherLendingLimit</lx:path>
                                                <lx:label>Higher Lending Limit</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="LowerDepositRequired" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Principal and interest repayment type is important to the applicant for a lower deposit required.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.PrincipalAndInterest.Reason.LowerDepositRequired</lx:path>
                                                <lx:label>Lower Deposit Required</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="MinimiseInterestPaidOverLifeOfLoan" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Principal and interest repayment type is important to the applicant to minimise interest paid over life of loan.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.PrincipalAndInterest.Reason.MinimiseInterestPaidOverLifeOfLoan</lx:path>
                                                <lx:label>Minimise Interest Paid over Life of Loan</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Other" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The reason principal and interest repayment type is important to the applicant cannot be captured in the specific reason data fields.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.PrincipalAndInterest.Reason.Other</lx:path>
                                                <lx:label>Other</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="Importance" type="importanceList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The importance of this feature to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.PrincipalAndInterest.Importance</lx:path>
                                          <lx:label>Importance</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RepaymentFrequency" type="frequencyShortList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The preferred repayment frequency.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.PrincipalAndInterest.RepaymentFrequency</lx:path>
                                          <lx:label>Repayment Frequency</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RisksExplained" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The risks of this feature have been explained to the applicant.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentTypeDetails.PrincipalAndInterest.RisksExplained</lx:path>
                                          <lx:label>Risks Explained</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="ConflictDescription" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A description of the conflict in the loan feature preferences.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Preferences.ConflictDescription</lx:path>
                              <lx:label>Conflict Description</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ConflictExists" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>A flag to indicate that a conflict exists in the loan feature preferences.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Preferences.ConflictExists</lx:path>
                              <lx:label>Conflict Exists</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="FundsAccessType" type="fundsAccessTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The applicant/s stated preference for type of facilities and/or features that provide access to funds.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Preferences.FundsAccessType</lx:path>
                              <lx:label>Funds Access Type</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="InterestRateType" type="interestRateTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The applicant/s stated preference for type of interest rate.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Preferences.InterestRateType</lx:path>
                              <lx:label>Interest Rate Type</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="OtherPreferences" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A free text field to capture any other applicant/s stated preference otherwise not specifically captured.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Preferences.OtherPreferences</lx:path>
                              <lx:label>Other Preferences</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PreferredLenders" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>A flag to indicate that the applicant has preferred lenders.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Preferences.PreferredLenders</lx:path>
                              <lx:label>Preferred Lenders</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PreferredLendersDetails" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Details of the applicant's preferred lenders.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Preferences.PreferredLendersDetails</lx:path>
                              <lx:label>Preferred Lenders Details</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PrioritiesAndReasons" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The priorities and reasons for the priorities in loan feature preferences.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Preferences.PrioritiesAndReasons</lx:path>
                              <lx:label>Priorities and Reasons</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RepaymentType" type="paymentTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The applicant/s stated preference for type of repayment.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Preferences.RepaymentType</lx:path>
                              <lx:label>Repayment Type</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Summary" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A summary of the applicant/s preferences with respect to requirements or features of the products they are seeking.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.Preferences.Summary</lx:path>
                              <lx:label>Summary</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="PurposeOfApplication" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>The purpose associated with a monetary amount that the customer has requested.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.NeedsAnalysis.PurposeOfApplication</lx:path>
                          <lx:label>Purpose Of Funds</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="Amount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The monetary amount the customer has requested for the stated purpose.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.PurposeOfApplication.Amount</lx:path>
                              <lx:label>Amount</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Description" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The description of the purpose the associated monetary amount is intended to be used for.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.PurposeOfApplication.Description</lx:path>
                              <lx:label>Description</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Purpose" type="purposeOfApplicationList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The purpose the associated monetary amount is intended to be used for.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.PurposeOfApplication.Purpose</lx:path>
                              <lx:label>Purpose</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="RefinancingAndConsolidation" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Details of refinancing or consolidating debts with the proceeds of the proposed loan.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation</lx:path>
                          <lx:label>Refinancing and Consolidation</lx:label>
                          <li:transactions>CAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="RefinancingReason" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The reason for wanting to refinance the existing loan.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason</lx:path>
                                <lx:label>Refinancing Reason</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="CloseToEndOfCurrentLoanTerm" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The reason for refinancing the existing loan is that the end of current loan term is close.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.CloseToEndOfCurrentLoanTerm</lx:path>
                                    <lx:label>Close To End Of Current Loan Term</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CloseToEndOfCurrentLoanTermDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Details of end of current loan term.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.CloseToEndOfCurrentLoanTermDetails</lx:path>
                                    <lx:label>Close To End Of Current Loan Term Details</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DebtConsolidation" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The reason for refinancing the existing loan is to reduce number of lenders or consolidate debts.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.DebtConsolidation</lx:path>
                                    <lx:label>Debt Consolidation</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DebtConsolidationDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Details of reduced number of lenders or consolidated debts.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.DebtConsolidationDetails</lx:path>
                                    <lx:label>Debt Consolidation Details</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DissatisfactionWithCurrentLender" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The reason for refinancing the existing loan is dissatisfaction with the applicant's current lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.DissatisfactionWithCurrentLender</lx:path>
                                    <lx:label>Dissatisfaction With Current Lender</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DissatisfactionWithCurrentLenderDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Details of dissatisfaction with the applicant's current lender.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.DissatisfactionWithCurrentLenderDetails</lx:path>
                                    <lx:label>Dissatisfaction With Current Lender Details</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GreaterFlexibility" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The reason for refinancing the existing loan is greater flexibility, or extra or specific loan features under the new loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.GreaterFlexibility</lx:path>
                                    <lx:label>Greater Flexibility</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="GreaterFlexibilityDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Details of greater flexibility, or extra or specific loan features.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.GreaterFlexibilityDetails</lx:path>
                                    <lx:label>Greater Flexibility Details</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncreaseTotalLoanAmount" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The reason for refinancing the existing loan is to increase the total loan amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.IncreaseTotalLoanAmount</lx:path>
                                    <lx:label>Increase Total Loan Amount</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncreaseTotalLoanAmountDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Details of the increase to the total loan amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.IncreaseTotalLoanAmountDetails</lx:path>
                                    <lx:label>Increase Total Loan Amount Details</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LowerInterestRate" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The reason for refinancing the existing loan is to lower the interest rate.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.LowerInterestRate</lx:path>
                                    <lx:label>Lower Interest Rate</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LowerInterestRateDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Details of current and anticipated interest rates under the existing loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.LowerInterestRateDetails</lx:path>
                                    <lx:label>Lower Interest Rate Details</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Other" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The reason for refinancing the existing loan cannot be captured in the specific refinancing reason data fields.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.Other</lx:path>
                                    <lx:label>Other</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="OtherDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Details of the other reason for refinancing the existing loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.OtherDetails</lx:path>
                                    <lx:label>Other Details</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ReducedRepayments" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The reason for refinancing the existing loan is to reduce repayments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.ReducedRepayments</lx:path>
                                    <lx:label>Reduced Repayments</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ReducedRepaymentsDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Details of current and anticipated repayments under the existing loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.RefinancingReason.ReducedRepaymentsDetails</lx:path>
                                    <lx:label>Reduced Repayments Details</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="DetailsOfCreditCardPlan" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Details of plans to close or reduce limits of existing credit cards.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.DetailsOfCreditCardPlan</lx:path>
                              <lx:label>Details of Credit Card Plan</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ExplainedIncreaseInterestRisk" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The broker has discussed with the applicant(s) that they may end up paying more interest, particularly if the applicant(s) incurs further credit card, personal loan or car loan debt after consolidating existing debts.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.ExplainedIncreaseInterestRisk</lx:path>
                              <lx:label>Explained Increase Interest Risk</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ExplainedIncreaseLoanTermRisk" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The broker has confirmed the applicant(s) understands that an increase in the term of the loan may cause the applicant(s) to pay more interest and mean the applicant(s)'s equity in the property builds up at a slower rate.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.ExplainedIncreaseLoanTermRisk</lx:path>
                              <lx:label>Explained Increase Loan Term Risk</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PlanToCloseOrReduceCreditCard" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Existing credit card limits will be reduced or cancelled.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.NeedsAnalysis.RefinancingAndConsolidation.PlanToCloseOrReduceCreditCard</lx:path>
                              <lx:label>Plan to Close or Reduce Credit Card</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="RetirementDetails" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Retirement details of the applicant.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails</lx:path>
                          <lx:label>Retirement Details</lx:label>
                          <li:transactions>CAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Applicant" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>An applicant to which these retirement details apply.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant</lx:path>
                                <lx:label>Applicant</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="RepaymentOptions" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>The repayment options with which the applicant intends to meet the loan repayments if they plan to retire during the proposed loan term.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions</lx:path>
                                      <lx:label>Repayment Options</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="CoApplicantIncomeDetails" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Co applicant income details.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.CoApplicantIncomeDetails</lx:path>
                                            <lx:label>Co Applicant Income Details</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:sequence>
                                            <xs:element name="CoApplicant" minOccurs="0" maxOccurs="unbounded">
                                              <xs:annotation>
                                                <xs:documentation>Details of the co applicant that would pay the loan monthly expense in the event that the applicant is not able to do.</xs:documentation>
                                                <xs:appinfo>
                                                  <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.CoApplicantIncomeDetails.CoApplicant</lx:path>
                                                  <lx:label>Co Applicant</lx:label>
                                                  <li:transactions>CAL</li:transactions>
                                                </xs:appinfo>
                                              </xs:annotation>
                                              <xs:complexType>
                                                <xs:attribute name="x_Applicant" type="referenceType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>A cross reference to the co applicant that can pay some or all of the loan monthly expense in the event that the applicant is not able to do.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.CoApplicantIncomeDetails.CoApplicant.x_Applicant</lx:path>
                                                      <lx:label>Applicant</lx:label>
                                                      <li:transactions>CAL</li:transactions>
                                                      <lx:references>
                                                        <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                                        <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                                        <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                                      </lx:references>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                              </xs:complexType>
                                            </xs:element>
                                          </xs:sequence>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Co applicant income details description.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.CoApplicantIncomeDetails.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="DownsizingHomeDetails" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Details of the applicant's plan to downsize their home.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.DownsizingHomeDetails</lx:path>
                                            <lx:label>Downsizing Home Details</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:sequence>
                                            <xs:element name="AcceptableAsset" minOccurs="0" maxOccurs="unbounded">
                                              <xs:annotation>
                                                <xs:documentation>An asset accepted by the lender to cover the equity shortfall.</xs:documentation>
                                                <xs:appinfo>
                                                  <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.DownsizingHomeDetails.AcceptableAsset</lx:path>
                                                  <lx:label>Acceptable Asset</lx:label>
                                                  <li:transactions>CAL</li:transactions>
                                                </xs:appinfo>
                                              </xs:annotation>
                                              <xs:complexType>
                                                <xs:attribute name="x_Asset" type="referenceType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>A cross reference to the real or non-real estate assets that is an acceptable asset by the lender to cover equity shortfall.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.DownsizingHomeDetails.AcceptableAsset.x_Asset</lx:path>
                                                      <lx:label>Asset</lx:label>
                                                      <li:transactions>CAL</li:transactions>
                                                      <lx:references>
                                                        <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                                        <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                                      </lx:references>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                              </xs:complexType>
                                            </xs:element>
                                            <xs:element name="PropertyToSell" minOccurs="0" maxOccurs="unbounded">
                                              <xs:annotation>
                                                <xs:documentation>A property that will be sold by the applicant in order to downsize.</xs:documentation>
                                                <xs:appinfo>
                                                  <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.DownsizingHomeDetails.PropertyToSell</lx:path>
                                                  <lx:label>Property To Sell</lx:label>
                                                  <li:transactions>CAL</li:transactions>
                                                </xs:appinfo>
                                              </xs:annotation>
                                              <xs:complexType>
                                                <xs:attribute name="x_Property" type="referenceType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>A cross reference to the property that will be sold.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.DownsizingHomeDetails.PropertyToSell.x_Property</lx:path>
                                                      <lx:label>Property</lx:label>
                                                      <li:transactions>CAL</li:transactions>
                                                      <lx:references>
                                                        <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                                      </lx:references>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                              </xs:complexType>
                                            </xs:element>
                                          </xs:sequence>
                                          <xs:attribute name="AcceptableAssetTotalValue" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The total estimated value of assets that could be used to cover the equity shortfall.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.DownsizingHomeDetails.AcceptableAssetTotalValue</lx:path>
                                                <lx:label>Acceptable Asset Total Value</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="AssetsToEquityShortfallRatio" type="decimalType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The total value of acceptable assets divided by the shortfall.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.DownsizingHomeDetails.AssetsToEquityShortfallRatio</lx:path>
                                                <lx:label>Assets To Equity Shortfall Ratio</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="AvailableEquity" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The total equity available to the applicant. The total outstanding balance of all liabilities owed by the applicant subtracted from estimated value of all properties being sold.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.DownsizingHomeDetails.AvailableEquity</lx:path>
                                                <lx:label>Available Equity</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>A general description the applicant's plan to downsize their home.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.DownsizingHomeDetails.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="DownsizeToPropertyValue" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The estimated value of the property or properties being purchased by the applicant in order to downsize.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.DownsizingHomeDetails.DownsizeToPropertyValue</lx:path>
                                                <lx:label>Downsize To Property Value</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="EquityShortfallAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The dollar amount that the equity falls short of covering the cost of the new property.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.DownsizingHomeDetails.EquityShortfallAmount</lx:path>
                                                <lx:label>Equity Shortfall Amount</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="PropertyToSellTotalValue" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The total value of all properties that will be sold by the applicant in order to downsize.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.DownsizingHomeDetails.PropertyToSellTotalValue</lx:path>
                                                <lx:label>Property To Sell Total Value</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="LenderAcceptableAssetsDetails" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Details of financial assets compared to total liabilities.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.LenderAcceptableAssetsDetails</lx:path>
                                            <lx:label>Lender Acceptable Assets Details</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:sequence>
                                            <xs:element name="Asset" minOccurs="0" maxOccurs="unbounded">
                                              <xs:annotation>
                                                <xs:documentation>Real estate and non real estate assets.</xs:documentation>
                                                <xs:appinfo>
                                                  <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.LenderAcceptableAssetsDetails.Asset</lx:path>
                                                  <lx:label>Asset</lx:label>
                                                  <li:transactions>CAL</li:transactions>
                                                </xs:appinfo>
                                              </xs:annotation>
                                              <xs:complexType>
                                                <xs:attribute name="x_Asset" type="referenceType" use="optional">
                                                  <xs:annotation>
                                                    <xs:documentation>Asset that will be used to payoff existing or new liabilities.</xs:documentation>
                                                    <xs:appinfo>
                                                      <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.LenderAcceptableAssetsDetails.Asset.x_Asset</lx:path>
                                                      <lx:label>Asset</lx:label>
                                                      <li:transactions>CAL</li:transactions>
                                                      <lx:references>
                                                        <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                                        <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                                      </lx:references>
                                                    </xs:appinfo>
                                                  </xs:annotation>
                                                </xs:attribute>
                                              </xs:complexType>
                                            </xs:element>
                                          </xs:sequence>
                                          <xs:attribute name="Description" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Additional explanation or comment.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.LenderAcceptableAssetsDetails.Description</lx:path>
                                                <lx:label>Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="TotalLenderAcceptableAssetsAmount" type="currencyType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Total dollar value of all assets.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.LenderAcceptableAssetsDetails.TotalLenderAcceptableAssetsAmount</lx:path>
                                                <lx:label>Total Lender Acceptable Assets Amount</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="TotalLenderAcceptableAssetsAmountToTotalLiabilityRatio" type="decimalType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Ratio Of Assets To Total Liability.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.LenderAcceptableAssetsDetails.TotalLenderAcceptableAssetsAmountToTotalLiabilityRatio</lx:path>
                                                <lx:label>Total Lender Acceptable Assets Amount To Total Liability Ratio</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="RecurringIncomeFromSuperannuationDetails" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Details of the recurring income from superannuation.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.RecurringIncomeFromSuperannuationDetails</lx:path>
                                            <lx:label>Recurring Income From Superannuation Details</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="FinancialPlannerDocumentation" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Financial planner documentation is present.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.RecurringIncomeFromSuperannuationDetails.FinancialPlannerDocumentation</lx:path>
                                                <lx:label>Financial Planner Documentation</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="FinancialPlannerDocumentationDescription" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Extra details of the financial planner documentation.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.RecurringIncomeFromSuperannuationDetails.FinancialPlannerDocumentationDescription</lx:path>
                                                <lx:label>Financial Planner Documentation Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="SuperStatementShowingProjectedRecurringIncome" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>A super statement showing projected recurring income documentation is present.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.RecurringIncomeFromSuperannuationDetails.SuperStatementShowingProjectedRecurringIncome</lx:path>
                                                <lx:label>Super Statement Showing Projected Recurring Income</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="SuperStatementShowingProjectedRecurringIncomeDescription" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Extra details of the super statement showing projected recurring income documentation.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.RecurringIncomeFromSuperannuationDetails.SuperStatementShowingProjectedRecurringIncomeDescription</lx:path>
                                                <lx:label>Super Statement Showing Projected Recurring Income Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="SaleOfAssetsDetails" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Details of the sale of assets.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SaleOfAssetsDetails</lx:path>
                                            <lx:label>Sale Of Assets Details</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="LenderHeldSecurity" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The asset is a security held by the lender.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SaleOfAssetsDetails.LenderHeldSecurity</lx:path>
                                                <lx:label>Lender Held Security</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="LenderHeldSecurityDescription" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Extra details regarding the security asset held by the lender.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SaleOfAssetsDetails.LenderHeldSecurityDescription</lx:path>
                                                <lx:label>Lender Held Security Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="OtherAssets" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The sale of assets refers to sale of other assets.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SaleOfAssetsDetails.OtherAssets</lx:path>
                                                <lx:label>Other Assets</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="OtherAssetsDescription" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Extra details of the sales of other assets.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SaleOfAssetsDetails.OtherAssetsDescription</lx:path>
                                                <lx:label>Other Assets Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="OtherLenderHeldSecurity" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The asset is a security held by the other lender.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SaleOfAssetsDetails.OtherLenderHeldSecurity</lx:path>
                                                <lx:label>Other Lender Held Security</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="OtherLenderHeldSecurityDescription" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Extra details of the security asset held by the other lender.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SaleOfAssetsDetails.OtherLenderHeldSecurityDescription</lx:path>
                                                <lx:label>Other Lender Held Security Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="Shares" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Specifies if the sale of assets refers to sale of shares.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SaleOfAssetsDetails.Shares</lx:path>
                                                <lx:label>Shares</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="SharesDescription" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Extra details of the sales of shares.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SaleOfAssetsDetails.SharesDescription</lx:path>
                                                <lx:label>Shares Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="SavingsDetails" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Information regarding the savings.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SavingsDetails</lx:path>
                                            <lx:label>Savings Details</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="LenderHeldSavingsAccount" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The savings are held in one or more accounts with this lender.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SavingsDetails.LenderHeldSavingsAccount</lx:path>
                                                <lx:label>Lender Held Accounts</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="LenderHeldSavingsAccountDescription" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Information regarding the accounts held by the lender.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SavingsDetails.LenderHeldSavingsAccountDescription</lx:path>
                                                <lx:label>Lender Held Account Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="OtherLenderHeldSavingsAccount" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>The savings are held in one or more accounts with another financial institution.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SavingsDetails.OtherLenderHeldSavingsAccount</lx:path>
                                                <lx:label>Other Financial Institution Held Accounts</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="OtherLenderHeldSavingsAccountDescription" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Information regarding the accounts held by another financial institution.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SavingsDetails.OtherLenderHeldSavingsAccountDescription</lx:path>
                                                <lx:label>Other Financial Institution Held Accounts Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                      <xs:element name="SuperannuationLumpSumFollowingRetirementDetails" minOccurs="0" maxOccurs="1">
                                        <xs:annotation>
                                          <xs:documentation>Details of the superannuation lump sum following retirement.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SuperannuationLumpSumFollowingRetirementDetails</lx:path>
                                            <lx:label>Superannuation Lump Sum Following Retirement Details</lx:label>
                                            <li:transactions>CAL</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="FinancialPlannerDocumentation" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Financial planner documentation is present.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SuperannuationLumpSumFollowingRetirementDetails.FinancialPlannerDocumentation</lx:path>
                                                <lx:label>Financial Planner Documentation</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="FinancialPlannerDocumentationDescription" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Extra details of the financial planner documentation.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SuperannuationLumpSumFollowingRetirementDetails.FinancialPlannerDocumentationDescription</lx:path>
                                                <lx:label>Financial Planner Documentation Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="SuperStatementShowingProjectedSuperannuationAmount" type="yesNoList" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>A super statement showing the projected Superannuation amount is present.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SuperannuationLumpSumFollowingRetirementDetails.SuperStatementShowingProjectedSuperannuationAmount</lx:path>
                                                <lx:label>Super Statement Showing Projected Superannuation Amount</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                          <xs:attribute name="SuperStatementShowingProjectedSuperannuationAmountDescription" type="stringType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>Extra details of the super statement showing the projected Superannuation amount.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SuperannuationLumpSumFollowingRetirementDetails.SuperStatementShowingProjectedSuperannuationAmountDescription</lx:path>
                                                <lx:label>Super Statement Showing Projected Superannuation Amount Description</lx:label>
                                                <li:transactions>CAL</li:transactions>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="AllRepaymentStrategiesConsidered" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant has considered all exit strategy options.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.AllRepaymentStrategiesConsidered</lx:path>
                                          <lx:label>All Repayment Strategies Considered</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="AllRepaymentStrategiesConsideredDescription" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A description to support the applicant's choice of exit strategy.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.AllRepaymentStrategiesConsideredDescription</lx:path>
                                          <lx:label>All Repayment Strategies Considered Description</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CoApplicantIncome" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant intends to meet the loan repayments with their co-applicant's income.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.CoApplicantIncome</lx:path>
                                          <lx:label>Co Applicant Income</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Description" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Details of the other repayment option.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.Description</lx:path>
                                          <lx:label>Description</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="DownsizingHome" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant intends to meet the loan repayments by downsizing their home.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.DownsizingHome</lx:path>
                                          <lx:label>Downsizing Home</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="IncomeFromOtherInvestments" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant intends to meet the loan repayments with income from other investments.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.IncomeFromOtherInvestments</lx:path>
                                          <lx:label>Income From Other Investments</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="LenderAcceptableAssets" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Financial assets compared to total liabilities.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.LenderAcceptableAssets</lx:path>
                                          <lx:label>Lender Acceptable Assets</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Other" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicants plan to meet the loan repayments cannot be captured in the specific repayment options data fields.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.Other</lx:path>
                                          <lx:label>Other</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RecurringIncomeFromSuperannuation" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant intends to meet the loan repayments with recurring income from superannuation.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.RecurringIncomeFromSuperannuation</lx:path>
                                          <lx:label>Recurring Income From Superannuation</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RepaymentOfLoanPriorToRetirement" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant intends to repay the loan prior to retirement.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.RepaymentOfLoanPriorToRetirement</lx:path>
                                          <lx:label>Repayment Of Loan Prior To Retirement</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SaleOfAssets" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant intends to meet the loan repayments by sale of assets.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SaleOfAssets</lx:path>
                                          <lx:label>Sale Of Assets</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Savings" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant intends to meet the loan repayments with their savings.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.Savings</lx:path>
                                          <lx:label>Savings</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SuperannuationLumpSumFollowingRetirement" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant intends to meet the loan repayments with a superannuation lump sum following retirement.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.SuperannuationLumpSumFollowingRetirement</lx:path>
                                          <lx:label>Superannuation Lump Sum Following Retirement</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="TotalLiability" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Total dollar value of all existing an new liabilities.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.TotalLiability</lx:path>
                                          <lx:label>Total Liability</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="WorkPastStatutoryRetirementAge" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The applicant plan to work past the statutory retirement age.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.RepaymentOptions.WorkPastStatutoryRetirementAge</lx:path>
                                          <lx:label>WorkPastStatutoryRetirementAge</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="AgeAtEndOfLoan" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant's age at the end of the loan.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.AgeAtEndOfLoan</lx:path>
                                    <lx:label>Age At End Of Loan</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ImminentlyRetiring" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant will be retiring imminently. Refers to an applicant who is over a certain age or X number of years from retirement. Each lender can have different age/retirement years limits based on internal policy.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.ImminentlyRetiring</lx:path>
                                    <lx:label>Imminently Retiring</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IntendedRetirementAge" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The age at which the applicant is planning to retire.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.IntendedRetirementAge</lx:path>
                                    <lx:label>Intended Retirement Age</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IntendedRetirementAgeDetails" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Details of the intended retirement age e.g. the reason why the intended retirement age is greater than the normal retirement age.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.IntendedRetirementAgeDetails</lx:path>
                                    <lx:label>IntendedRetirementAgeDetails</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the applicant, to be used to identify the applicant when the application has not yet been created.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Number" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant number (e.g. Applicant 1, Applicant 2) to be used to identify the applicant when the application has not yet been created.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.Number</lx:path>
                                    <lx:label>Number</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UnderstandsImpactOnRetirementPlans" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The broker has confirmed the applicant understands that if the term of the loan extends past their planned retirement age it may impact their retirement plans.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.UnderstandsImpactOnRetirementPlans</lx:path>
                                    <lx:label>Understands Impact On Retirement Plans</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Applicant" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the applicant (a Person Applicant element)</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.Applicant.x_Applicant</lx:path>
                                    <lx:label>Applicant</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="RepaymentOptions" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The repayment options with which the applicant intends to meet the loan repayments if they plan to retire during the proposed loan term.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions</lx:path>
                                <lx:label>Repayment Options</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="RecurringIncomeFromSuperannuationDetails" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Details of the recurring income from superannuation.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.RecurringIncomeFromSuperannuationDetails</lx:path>
                                      <lx:label>Recurring Income From Superannuation Details</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="FinancialPlannerDocumentation" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Financial planner documentation is present.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.RecurringIncomeFromSuperannuationDetails.FinancialPlannerDocumentation</lx:path>
                                          <lx:label>Financial Planner Documentation</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="FinancialPlannerDocumentationDescription" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Extra details of the financial planner documentation.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.RecurringIncomeFromSuperannuationDetails.FinancialPlannerDocumentationDescription</lx:path>
                                          <lx:label>Financial Planner Documentation Description</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SuperStatementShowingProjectedRecurringIncome" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A super statement showing projected recurring income documentation is present.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.RecurringIncomeFromSuperannuationDetails.SuperStatementShowingProjectedRecurringIncome</lx:path>
                                          <lx:label>Super Statement Showing Projected Recurring Income</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SuperStatementShowingProjectedRecurringIncomeDescription" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Extra details of the super statement showing projected recurring income documentation.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.RecurringIncomeFromSuperannuationDetails.SuperStatementShowingProjectedRecurringIncomeDescription</lx:path>
                                          <lx:label>Super Statement Showing Projected Recurring Income Description</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="SaleOfAssetsDetails" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Details of the sale of assets.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SaleOfAssetsDetails</lx:path>
                                      <lx:label>Sale Of Assets Details</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="LenderHeldSecurity" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The asset is a security held by the lender.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SaleOfAssetsDetails.LenderHeldSecurity</lx:path>
                                          <lx:label>Lender Held Security</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="LenderHeldSecurityDescription" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Extra details regarding the security asset held by the lender.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SaleOfAssetsDetails.LenderHeldSecurityDescription</lx:path>
                                          <lx:label>Lender Held Security Description</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OtherAssets" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The sale of assets refers to sale of other assets.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SaleOfAssetsDetails.OtherAssets</lx:path>
                                          <lx:label>Other Assets</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OtherAssetsDescription" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Extra details of the sales of other assets.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SaleOfAssetsDetails.OtherAssetsDescription</lx:path>
                                          <lx:label>Other Assets Description</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OtherLenderHeldProperty" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>This items has been marked for deprecation. Please use OtherLenderHeldSecurity instead.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SaleOfAssetsDetails.OtherLenderHeldProperty</lx:path>
                                          <lx:label>Other Lender Held Property</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OtherLenderHeldPropertyDescription" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>This items has been marked for deprecation. Please use OtherLenderHeldSecurityDescription instead.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SaleOfAssetsDetails.OtherLenderHeldPropertyDescription</lx:path>
                                          <lx:label>Other Lender Held Property Description</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OtherLenderHeldSecurity" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The asset is a security held by the other lender.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SaleOfAssetsDetails.OtherLenderHeldSecurity</lx:path>
                                          <lx:label>Other Lender Held Security</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OtherLenderHeldSecurityDescription" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Extra details of the security asset held by the other lender.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SaleOfAssetsDetails.OtherLenderHeldSecurityDescription</lx:path>
                                          <lx:label>Other Lender Held Security Description</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Shares" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Specifies if the sale of assets refers to sale of shares.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SaleOfAssetsDetails.Shares</lx:path>
                                          <lx:label>Shares</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SharesDescription" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Extra details of the sales of shares.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SaleOfAssetsDetails.SharesDescription</lx:path>
                                          <lx:label>Shares Description</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="SavingsDetails" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Information regarding the savings.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SavingsDetails</lx:path>
                                      <lx:label>Savings Details</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="LenderHeldSavingsAccount" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The savings are held in one or more accounts with this lender.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SavingsDetails.LenderHeldSavingsAccount</lx:path>
                                          <lx:label>Lender Held Accounts</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="LenderHeldSavingsAccountDescription" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Information regarding the accounts held by the lender.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SavingsDetails.LenderHeldSavingsAccountDescription</lx:path>
                                          <lx:label>Lender Held Account Description</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OtherLenderHeldSavingsAccount" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The savings are held in one or more accounts with another financial institution.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SavingsDetails.OtherLenderHeldSavingsAccount</lx:path>
                                          <lx:label>Other Financial Institution Held Accounts</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="OtherLenderHeldSavingsAccountDescription" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Information regarding the accounts held by another financial institution.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SavingsDetails.OtherLenderHeldSavingsAccountDescription</lx:path>
                                          <lx:label>Other Financial Institution Held Accounts Description</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="SuperannuationLumpSumFollowingRetirementDetails" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Details of the superannuation lump sum following retirement.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SuperannuationLumpSumFollowingRetirementDetails</lx:path>
                                      <lx:label>Superannuation Lump Sum Following Retirement Details</lx:label>
                                      <li:transactions>CAL</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="FinancialPlannerDocumentation" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Financial planner documentation is present.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SuperannuationLumpSumFollowingRetirementDetails.FinancialPlannerDocumentation</lx:path>
                                          <lx:label>Financial Planner Documentation</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="FinancialPlannerDocumentationDescription" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Extra details of the financial planner documentation.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SuperannuationLumpSumFollowingRetirementDetails.FinancialPlannerDocumentationDescription</lx:path>
                                          <lx:label>Financial Planner Documentation Description</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SuperStatementShowingProjectedSuperannuationAmount" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A super statement showing the projected Superannuation amount is present.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SuperannuationLumpSumFollowingRetirementDetails.SuperStatementShowingProjectedSuperannuationAmount</lx:path>
                                          <lx:label>Super Statement Showing Projected Superannuation Amount</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="SuperStatementShowingProjectedSuperannuationAmountDescription" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Extra details of the super statement showing the projected Superannuation amount.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SuperannuationLumpSumFollowingRetirementDetails.SuperStatementShowingProjectedSuperannuationAmountDescription</lx:path>
                                          <lx:label>Super Statement Showing Projected Superannuation Amount Description</lx:label>
                                          <li:transactions>CAL</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="CoApplicantIncome" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant intends to meet the loan repayments with their co-applicant's income.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.CoApplicantIncome</lx:path>
                                    <lx:label>Co Applicant Income</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Details of the other repayment option.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DownsizingHome" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant intends to meet the loan repayments by downsizing their home.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.DownsizingHome</lx:path>
                                    <lx:label>Downsizing Home</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="IncomeFromOtherInvestments" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant intends to meet the loan repayments with income from other investments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.IncomeFromOtherInvestments</lx:path>
                                    <lx:label>Income From Other Investments</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Other" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicants plan to meet the loan repayments cannot be captured in the specific repayment options data fields.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.Other</lx:path>
                                    <lx:label>Other</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RecurringIncomeFromSuperannuation" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant intends to meet the loan repayments with recurring income from superannuation.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.RecurringIncomeFromSuperannuation</lx:path>
                                    <lx:label>Recurring Income From Superannuation</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RepaymentOfLoanPriorToRetirement" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant intends to repay the loan prior to retirement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.RepaymentOfLoanPriorToRetirement</lx:path>
                                    <lx:label>Repayment Of Loan Prior To Retirement</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SaleOfAssets" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant intends to meet the loan repayments by sale of assets.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SaleOfAssets</lx:path>
                                    <lx:label>Sale Of Assets</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Savings" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant intends to meet the loan repayments with their savings.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.Savings</lx:path>
                                    <lx:label>Savings</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="SuperannuationLumpSumFollowingRetirement" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant intends to meet the loan repayments with a superannuation lump sum following retirement.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.SuperannuationLumpSumFollowingRetirement</lx:path>
                                    <lx:label>Superannuation Lump Sum Following Retirement</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="WorkPastStatutoryRetirementAge" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The applicant plan to work past the statutory retirement age.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.NeedsAnalysis.RetirementDetails.RepaymentOptions.WorkPastStatutoryRetirementAge</lx:path>
                                    <lx:label>WorkPastStatutoryRetirementAge</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                      </xs:complexType>
                    </xs:element>
                  </xs:sequence>
                  <xs:attribute name="LoanAmountSought" type="currencyType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The total amount of credit sought for this loan.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.NeedsAnalysis.LoanAmountSought</lx:path>
                        <lx:label>Loan Amount Sought</lx:label>
                        <li:transactions>CAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="LoanTermSoughtDescription" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>A description of the discussion between the broker and applicant regarding the duration of the loan term sought and the reason this duration was selected.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.NeedsAnalysis.LoanTermSoughtDescription</lx:path>
                        <lx:label>Loan Term Sought Description</lx:label>
                        <li:transactions>CAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="PrimaryApplicationPurpose" type="primaryPurposeLoanPurposeList" use="optional">
                    <xs:annotation>
                      <xs:documentation>The primary purpose for this loan.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.NeedsAnalysis.PrimaryApplicationPurpose</lx:path>
                        <lx:label>Primary Application Purpose</lx:label>
                        <li:transactions>CAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ProductSelection" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>Product selection having regard to the outcome of the requirements and objectives enquiries.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.NeedsAnalysis.ProductSelection</lx:path>
                        <lx:label>Product Selection</lx:label>
                        <li:transactions>CAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="PurposeSummary" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>A summary of the applicant/s stated purpose for the application.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.NeedsAnalysis.PurposeSummary</lx:path>
                        <lx:label>Description</lx:label>
                        <li:transactions>CAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                    <xs:annotation>
                      <xs:documentation>Unique Identifier for this NeedsAnalysis element.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.NeedsAnalysis.UniqueID</lx:path>
                        <lx:label>Unique ID</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                </xs:complexType>
              </xs:element>
              <xs:element name="RealEstateValuation" minOccurs="0" maxOccurs="unbounded">
                <xs:annotation>
                  <xs:documentation>A valuation for a real estate asset.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Content.RealEstateValuation</lx:path>
                    <lx:label>Real Estate Valuation</lx:label>
                    <li:transactions>CDA,LMI,VAL</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:sequence>
                    <xs:element name="Instructor" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>The party that has instructed the valuation i.e. the lender or broker that ordered the valuation.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.RealEstateValuation.Instructor</lx:path>
                          <lx:label>Instructor</lx:label>
                          <li:transactions>VAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Mobile" type="phoneType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The mobile number for the valuation instructor.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.Instructor.Mobile</lx:path>
                                <lx:label>Mobile</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The office phone number for the valuation instructor.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.Instructor.OfficePhone</lx:path>
                                <lx:label>Office Phone</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The valuation instructor contact person's name.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.Instructor.PersonName</lx:path>
                                <lx:label>Person Name</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="CompanyName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The company name of the valuation instructor.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.Instructor.CompanyName</lx:path>
                              <lx:label>Company Name</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Email" type="emailPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The email address for the valuation instructor.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.Instructor.Email</lx:path>
                              <lx:label>Email</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="InstructorID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>An identifier for the valuation instructor.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.Instructor.InstructorID</lx:path>
                              <lx:label>Instructor ID</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Address" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The address of the valuation instructor.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.Instructor.x_Address</lx:path>
                              <lx:label>Address</lx:label>
                              <li:transactions>VAL</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.Address</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="PlatformProvider" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>The valuation platform provider for the real estate valuation.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.RealEstateValuation.PlatformProvider</lx:path>
                          <lx:label>Platform Provider</lx:label>
                          <li:transactions>VAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Mobile" type="phoneType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The mobile number for the valuation platform provider.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.PlatformProvider.Mobile</lx:path>
                                <lx:label>Mobile</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The office phone number for the valuation platform provider.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.PlatformProvider.OfficePhone</lx:path>
                                <lx:label>Office Phone</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The valuation platform provider contact person's name.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.PlatformProvider.PersonName</lx:path>
                                <lx:label>Person Name</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="CompanyName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The company name of the valuation platform provider.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.PlatformProvider.CompanyName</lx:path>
                              <lx:label>Company Name</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Email" type="emailPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The email address for the valuation platform provider.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.PlatformProvider.Email</lx:path>
                              <lx:label>Email</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PlatformProviderID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>An identifier for the valuation platform provider.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.PlatformProvider.PlatformProviderID</lx:path>
                              <lx:label>Platform Provider ID</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Address" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The address of the valuation platform provider.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.PlatformProvider.x_Address</lx:path>
                              <lx:label>Address</lx:label>
                              <li:transactions>VAL</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.Address</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="PreviousValuation" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>A previous valuation of the same real estate asset that is valued by this valuation.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.RealEstateValuation.PreviousValuation</lx:path>
                          <lx:label>Previous Valuation</lx:label>
                          <li:transactions>VAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="TotalValuedAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The total value of the real estate asset as reported by the previous valuation.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.PreviousValuation.TotalValuedAmount</lx:path>
                              <lx:label>Total Valued Amount</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ValuationType" type="realEstateValuationTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The classification of the previous valuation, such as full registered, desktop, AVM, kerbside.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.PreviousValuation.ValuationType</lx:path>
                              <lx:label>Valuation Type</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ValuedDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date the previous valuation was carried out.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.PreviousValuation.ValuedDate</lx:path>
                              <lx:label>Valued Date</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ValuerAssignedID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A unique ID assigned to the previous valuation.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.PreviousValuation.ValuerAssignedID</lx:path>
                              <lx:label>Valuer Assigned ID</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="PrimaryValuer" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Details of the primary valuer responsible for conducting the real estate valuation.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.RealEstateValuation.PrimaryValuer</lx:path>
                          <lx:label>Primary Valuer</lx:label>
                          <li:transactions>VAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Capacity" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The capacity in which the primary valuer has reviewed the valuation.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.PrimaryValuer.Capacity</lx:path>
                                <lx:label>Capacity</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="CompanyName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The parent entity on whose behalf the primary valuer has carried out the valuation.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.RealEstateValuation.PrimaryValuer.Capacity.CompanyName</lx:path>
                                    <lx:label>Company Name</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Description of capacity in which the primary valuer has carried out the valuation.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.RealEstateValuation.PrimaryValuer.Capacity.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The primary valuer's name.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.PrimaryValuer.PersonName</lx:path>
                                <lx:label>Person Name</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="APICertificationType" type="apiCertificationTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of Australian Property Institute (API) certification the primary valuer holds.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.PrimaryValuer.APICertificationType</lx:path>
                              <lx:label>API Certification Type</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ValuerID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A unique identifier for the valuer.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.PrimaryValuer.ValuerID</lx:path>
                              <lx:label>Valuer ID</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Qualification" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>A specific qualification (warning, caution, assumption, condition, or limitation) that should be considered when interpreting the results of the valuation report.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.RealEstateValuation.Qualification</lx:path>
                          <lx:label>Qualification</lx:label>
                          <li:transactions>VAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="Comment" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A free-text field for the valuer to provide a qualification.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.Qualification.Comment</lx:path>
                              <lx:label>Comment</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Heading" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The heading line of the qualification.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.Qualification.Heading</lx:path>
                              <lx:label>Heading</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The order of this qualification within the sequence.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.Qualification.SequenceNumber</lx:path>
                              <lx:label>Sequence Number</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="ReplacementCostEstimate" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>The replacement cost estimate for the property.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.RealEstateValuation.ReplacementCostEstimate</lx:path>
                          <lx:label>Replacement Cost Estimate</lx:label>
                          <li:transactions>VAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="BodyCorporateResponsibility" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that the replacement cost is the responsibility of the body corporate.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ReplacementCostEstimate.BodyCorporateResponsibility</lx:path>
                              <lx:label>Body Corporate Responsibility</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CostEstimate" type="decimalType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The dollar amount of the replacement cost estimate for the property.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ReplacementCostEstimate.CostEstimate</lx:path>
                              <lx:label>Cost Estimate</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ExpertAdviceRecommended" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that further advice from an expert is recommended by the valuer.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ReplacementCostEstimate.ExpertAdviceRecommended</lx:path>
                              <lx:label>Expert Advice Recommended</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="RiskAnalysis" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>The risk analysis captured for the valuation.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.RealEstateValuation.RiskAnalysis</lx:path>
                          <lx:label>Risk Analysis</lx:label>
                          <li:transactions>CDA,LMI,VAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="RiskRating" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>A risk rating based on the API PropertyPro risk rating model for real estate.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.RiskAnalysis.RiskRating</lx:path>
                                <lx:label>Risk Rating</lx:label>
                                <li:transactions>CDA,LMI,VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Comment" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free text field for the valuer to provide a comment if the potential risk affects the property value. Part of valuer additional comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.RealEstateValuation.RiskAnalysis.RiskRating.Comment</lx:path>
                                    <lx:label>Comment</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free text description of the type of the potential risk associated with selling real estate. To be used when rating type is 'Other'.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.RealEstateValuation.RiskAnalysis.RiskRating.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Rating" type="valuationRiskRatingList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The rating of a potential risk associated with selling real estate.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.RealEstateValuation.RiskAnalysis.RiskRating.Rating</lx:path>
                                    <lx:label>Rating</lx:label>
                                    <li:transactions>CDA,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="RatingType" type="valuationRiskRatingTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The type of a potential risk associated with selling real estate.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.RealEstateValuation.RiskAnalysis.RiskRating.RatingType</lx:path>
                                    <lx:label>Rating Type</lx:label>
                                    <li:transactions>CDA,LMI,VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="SupervisingValuer" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Details of the supervising valuer responsible for reviewing the real estate valuation, sometimes called a counter signatory.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.RealEstateValuation.SupervisingValuer</lx:path>
                          <lx:label>Supervising Valuer</lx:label>
                          <li:transactions>VAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Capacity" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The capacity in which the supervising valuer has reviewed the valuation.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.SupervisingValuer.Capacity</lx:path>
                                <lx:label>Capacity</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="CompanyName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The parent entity on whose behalf the supervising valuer has reviewed the valuation.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.RealEstateValuation.SupervisingValuer.Capacity.CompanyName</lx:path>
                                    <lx:label>Company Name</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Description" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Description of capacity in which the supervising valuer has reviewed the valuation.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.RealEstateValuation.SupervisingValuer.Capacity.Description</lx:path>
                                    <lx:label>Description</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The supervising valuer's name.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.SupervisingValuer.PersonName</lx:path>
                                <lx:label>Person Name</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="APICertificationType" type="apiCertificationTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of Australian Property Institute (API) certification the supervising valuer holds.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.SupervisingValuer.APICertificationType</lx:path>
                              <lx:label>API Certification Type</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ValuerID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A unique identifier for the valuer.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.SupervisingValuer.ValuerID</lx:path>
                              <lx:label>Valuer ID</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="SupportingDocument" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Supporting documents that have been provided.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.RealEstateValuation.SupportingDocument</lx:path>
                          <lx:label>Additional Document</lx:label>
                          <li:transactions>VAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="DocumentDescription" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A description of the supporting document.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.SupportingDocument.DocumentDescription</lx:path>
                              <lx:label>Document Description</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DocumentReturnMandatory" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates if it is necessary to return the document.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.SupportingDocument.DocumentReturnMandatory</lx:path>
                              <lx:label>Document Return Mandatory</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DocumentType" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of the supporting document.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.SupportingDocument.DocumentType</lx:path>
                              <lx:label>Document Type</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.SupportingDocument.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Attachment" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A cross reference to the attached document.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.SupportingDocument.x_Attachment</lx:path>
                              <lx:label>x_Attachment</lx:label>
                              <li:transactions>VAL</li:transactions>
                              <lx:references>
                                <lx:target>Package.Attachment</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="ValuationFirm" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>The valuation firm for the real estate valuation.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.RealEstateValuation.ValuationFirm</lx:path>
                          <lx:label>Valuation Firm</lx:label>
                          <li:transactions>VAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Mobile" type="phoneType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The mobile number for the valuation firm.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.ValuationFirm.Mobile</lx:path>
                                <lx:label>Mobile</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="OfficePhone" type="phoneType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The office phone number for the valuation firm.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.ValuationFirm.OfficePhone</lx:path>
                                <lx:label>Office Phone</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                          <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The valuation firm contact person's name.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.ValuationFirm.PersonName</lx:path>
                                <lx:label>Person Name</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="CompanyName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The company name of the valuation firm.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ValuationFirm.CompanyName</lx:path>
                              <lx:label>Company Name</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Email" type="emailPattern" use="optional">
                          <xs:annotation>
                            <xs:documentation>The email address for the valuation firm.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ValuationFirm.Email</lx:path>
                              <lx:label>Email</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ValuationFirmID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>An identifier for the valuation firm.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ValuationFirm.ValuationFirmID</lx:path>
                              <lx:label>Valuation Firm ID</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="x_Address" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The address of the valuation firm.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ValuationFirm.x_Address</lx:path>
                              <lx:label>Address</lx:label>
                              <li:transactions>VAL</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.Address</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="ValuationOrder" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>The details of the valuation order.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.RealEstateValuation.ValuationOrder</lx:path>
                          <lx:label>Valuation Order</lx:label>
                          <li:transactions>VAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="AppointmentDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date the valuation is planned to be carried out, recorded in case the actual valuation is carried out on another date.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ValuationOrder.AppointmentDate</lx:path>
                              <lx:label>Appointment Date</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="CreatedDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date the valuation was created in the requestor system.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ValuationOrder.CreatedDate</lx:path>
                              <lx:label>Created Date</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DueDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date by which the requestor expects the valuation.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ValuationOrder.DueDate</lx:path>
                              <lx:label>Due Date</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ExpectedValuationDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The expected date the valuation will be be submitted to the requestor.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ValuationOrder.ExpectedValuationDate</lx:path>
                              <lx:label>Expected Valuation Date</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="PreferredValuationDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date the valuation is preferred to be carried out by the requestor.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ValuationOrder.PreferredValuationDate</lx:path>
                              <lx:label>Preferred Valuation Date</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="ReminderDate" type="dateType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date of reminder to complete the valuation.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ValuationOrder.ReminderDate</lx:path>
                              <lx:label>Reminder Date</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RequestorAssignedID" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>A unique ID assigned to the valuation by requestor.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ValuationOrder.RequestorAssignedID</lx:path>
                              <lx:label>Requestor Assigned ID</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="ValuationVariation" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>A request to vary an exisiting valuation order.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.RealEstateValuation.ValuationVariation</lx:path>
                          <lx:label>Valuation Variation</lx:label>
                          <li:transactions>VAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="ChangeFeeAmount" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>A request to vary the amount of the valuation fee to be paid to the valuer.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.RealEstateValuation.ValuationVariation.ChangeFeeAmount</lx:path>
                                <lx:label>Change Fee Amount</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="FeePrevious" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The previous fee amount</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.RealEstateValuation.ValuationVariation.ChangeFeeAmount.FeePrevious</lx:path>
                                    <lx:label>Fee Previous</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="FeeRequested" type="decimalType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The requested fee amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.RealEstateValuation.ValuationVariation.ChangeFeeAmount.FeeRequested</lx:path>
                                    <lx:label>Fee Requested</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Status" type="realEstateValuationFeeStatusList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The status of the valuation fee variation request.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.RealEstateValuation.ValuationVariation.ChangeFeeAmount.Status</lx:path>
                                    <lx:label>Status</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="Type" type="valuationVariationTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of valuation variation.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.RealEstateValuation.ValuationVariation.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                  </xs:sequence>
                  <xs:attribute name="AVMReferenceID" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>A unique reference ID assigned to the valuation by the automated valuation model (AVM) system.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.AVMReferenceID</lx:path>
                        <lx:label>AVMReferenceID</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ConfidenceLevel" type="decimalType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The level of confidence that the valuation provider has in their valuation.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.ConfidenceLevel</lx:path>
                        <lx:label>Confidence Level</lx:label>
                        <li:transactions>LMI,VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="DocumentsToSightComment" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>A free text field for the valuer to provide a comment for documents yet to be sighted or not sighted. Part of valuer additional comments.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.DocumentsToSightComment</lx:path>
                        <lx:label>Documents To Sight Comment</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ExpiryDate" type="dateType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The expiry date of the valuation.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.ExpiryDate</lx:path>
                        <lx:label>Expiry Date</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ImprovementsValue" type="currencyType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The total value of the improvements to the real estate asset.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.ImprovementsValue</lx:path>
                        <lx:label>Improvements Value</lx:label>
                        <li:transactions>LMI,VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="InstructorComment" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>A free text field for the instructor to provide a comment for the valuer regarding the valuation.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.InstructorComment</lx:path>
                        <lx:label>Instructor Comment</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="InstructorToSight" type="yesNoList" use="optional">
                    <xs:annotation>
                      <xs:documentation>Indicates that the valuation report must be sighted and approved by the instructor before it can be finalised.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.InstructorToSight</lx:path>
                        <lx:label>Instructor To Sight</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="IssueDate" type="dateType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The date the valuation report was sent to the instructor.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.IssueDate</lx:path>
                        <lx:label>Issue Date</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="LandValue" type="currencyType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The land value component of the real estate asset value.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.LandValue</lx:path>
                        <lx:label>Land Value</lx:label>
                        <li:transactions>LMI,VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="Marketability" type="ratingList" use="optional">
                    <xs:annotation>
                      <xs:documentation>The general marketability of the real property.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.Marketability</lx:path>
                        <lx:label>Marketability</lx:label>
                        <li:transactions>LMI,VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="OtherValue" type="currencyType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The total value of items beside land/insurance/improvements in the the real property.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.OtherValue</lx:path>
                        <lx:label>Other Value</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="RecommendSecurity" type="yesNoList" use="optional">
                    <xs:annotation>
                      <xs:documentation>Indicates if security as collateral is recommended by the valuer.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.RecommendSecurity</lx:path>
                        <lx:label>Recommend Security</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="RequestedPriority" type="realEstateRequestedPriorityList" use="optional">
                    <xs:annotation>
                      <xs:documentation>The priority for the requested valuation.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.RequestedPriority</lx:path>
                        <lx:label>Requested Priority</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="SaleInLineWithLocalMarket" type="yesNoList" use="optional">
                    <xs:annotation>
                      <xs:documentation>Indicates that the current sale price is in line with the current local market.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.SaleInLineWithLocalMarket</lx:path>
                        <lx:label>Sale In Line With Local Market</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="SaleInLineWithLocalMarketComment" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>A free text field for the valuer to provide a comment if the sale is not in line with the current local market. Part of valuer additional comments.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.SaleInLineWithLocalMarketComment</lx:path>
                        <lx:label>Sale In Line With Local Market Comment</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="TotalValuedAmount" type="currencyType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The total value of the real estate asset.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.TotalValuedAmount</lx:path>
                        <lx:label>Total Valued Amount</lx:label>
                        <li:transactions>CDA,LMI,VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="TwoTieredMarketing" type="yesNoList" use="optional">
                    <xs:annotation>
                      <xs:documentation>Indicates if the real estate agent is engaged in two-tiered marketing.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.TwoTieredMarketing</lx:path>
                        <lx:label>Two Tiered Marketing</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                    <xs:annotation>
                      <xs:documentation>Unique Identifier for this Valuation element.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.UniqueID</lx:path>
                        <lx:label>Unique ID</lx:label>
                        <li:transactions>CDA,LMI,VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ValuationMethodology" type="estimateMethodologyList" use="optional">
                    <xs:annotation>
                      <xs:documentation>The methodology used to estimate the value of the real estate asset.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.ValuationMethodology</lx:path>
                        <lx:label>Valuation Methodology</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ValuationReason" type="realEstateValuationReasonList" use="optional">
                    <xs:annotation>
                      <xs:documentation>The reason for conducting the valuation.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.ValuationReason</lx:path>
                        <lx:label>Valuation Reason</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ValuationReasonDescription" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>A description on the reason for conducting the valuation.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.ValuationReasonDescription</lx:path>
                        <lx:label>Valuation Reason Description</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ValuationType" type="realEstateValuationTypeList" use="optional">
                    <xs:annotation>
                      <xs:documentation>The classification of the valuation, such as full registered, desktop, AVM, kerbside.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.ValuationType</lx:path>
                        <lx:label>Valuation Type</lx:label>
                        <li:transactions>CDA,LMI,VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ValuedDate" type="dateType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The date the valuation was carried out.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.ValuedDate</lx:path>
                        <lx:label>ValuedDate</lx:label>
                        <li:transactions>CDA,LMI,VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ValueRangeMaximum" type="currencyType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The maximum total value of the real estate asset.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.ValueRangeMaximum</lx:path>
                        <lx:label>Value Range Maximum</lx:label>
                        <li:transactions>LMI,VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ValueRangeMinimum" type="currencyType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The minimum total value of the real estate asset.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.ValueRangeMinimum</lx:path>
                        <lx:label>Value Range Minimum</lx:label>
                        <li:transactions>LMI,VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ValuerAssignedID" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>A unique ID assigned to valuation by Valuer.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.ValuerAssignedID</lx:path>
                        <lx:label>Valuer Assigned ID</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ValuerComment" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>A free text field for the valuer to provide a comment about the valuation, or any amendments, on completion of the order. Part of valuer additional comments.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.ValuerComment</lx:path>
                        <lx:label>Valuer Comment</lx:label>
                        <li:transactions>VAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="x_Asset" type="referenceType" use="optional">
                    <xs:annotation>
                      <xs:documentation>Link to the valued asset.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.x_Asset</lx:path>
                        <lx:label>Asset</lx:label>
                        <li:transactions>CDA,LMI,VAL</li:transactions>
                        <lx:references>
                          <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                        </lx:references>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="x_Valuer" type="referenceType" use="optional">
                    <xs:annotation>
                      <xs:documentation>Link to an individual or a valuation firm that provided the valuation.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.RealEstateValuation.x_Valuer</lx:path>
                        <lx:label>Valuer</lx:label>
                        <li:transactions>CDA</li:transactions>
                        <lx:references>
                          <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                        </lx:references>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                </xs:complexType>
              </xs:element>
              <xs:element name="Serviceability" minOccurs="0" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>Details relating to the serviceability calculations for an application.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Content.Serviceability</lx:path>
                    <lx:label>Serviceability</lx:label>
                    <li:transactions>SVC</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:sequence>
                    <xs:element name="Report" minOccurs="0" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Report of the serviceability calculation that may be used in assessing serviceability of an application or its actors.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Serviceability.Report</lx:path>
                          <lx:label>Report</lx:label>
                          <li:transactions>SVC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Content" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>The content of the serviceability report.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Serviceability.Report.Content</lx:path>
                                <lx:label>Content</lx:label>
                                <li:transactions>SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Type" type="reportContentTypeList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The content type of the attached report.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Report.Content.Type</lx:path>
                                    <lx:label>Type</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_Attachment" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A cross reference to the the attached report.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Report.Content.x_Attachment</lx:path>
                                    <lx:label>Attachment</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Attachment</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="FormatVersion" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned version relating to the report format.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Serviceability.Report.FormatVersion</lx:path>
                              <lx:label>Format Version</lx:label>
                              <li:transactions>SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="RequestDateTime" type="dateTimeType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The date and time the report was requested.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Serviceability.Report.RequestDateTime</lx:path>
                              <lx:label>Request Date Time</lx:label>
                              <li:transactions>SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Source" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The system the report was sourced from.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Serviceability.Report.Source</lx:path>
                              <lx:label>Source</lx:label>
                              <li:transactions>SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Results" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Calculated serviceability results.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.Serviceability.Results</lx:path>
                          <lx:label>Serviceability Results</lx:label>
                          <li:transactions>SVC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Applicant" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Loan applicant/s (borrower/s or guarantor/s) that are included in this group.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Serviceability.Results.Applicant</lx:path>
                                <lx:label>Applicant</lx:label>
                                <li:transactions>SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="x_Party" type="referenceType" use="required">
                                <xs:annotation>
                                  <xs:documentation>Reference to the loan party that is included in this group.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.Applicant.x_Party</lx:path>
                                    <lx:label>Party</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                      <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                      <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="LoanDetail" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Serviceability results that relate to a single loan.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Serviceability.Results.LoanDetail</lx:path>
                                <lx:label>Net Disposable Income</lx:label>
                                <li:transactions>SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Sensitivity" minOccurs="0" maxOccurs="1">
                                  <xs:annotation>
                                    <xs:documentation>Details of the buffer applied to the interest rate and the resultant repayment amounts.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Content.Serviceability.Results.LoanDetail.Sensitivity</lx:path>
                                      <lx:label>Total Gross Income</lx:label>
                                      <li:transactions>SVC</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="InterestRate" type="percentType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The sensitised interest rate used for a serviceability calculation that includes an additional buffer applied to the contractual interest rate.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Serviceability.Results.LoanDetail.Sensitivity.InterestRate</lx:path>
                                          <lx:label>Interest Rate</lx:label>
                                          <li:transactions>SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RepaymentAmount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The repayment derived using the corresponding sensitised interest rate.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Serviceability.Results.LoanDetail.Sensitivity.RepaymentAmount</lx:path>
                                          <lx:label>Repayment Amount</lx:label>
                                          <li:transactions>SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="RepaymentFrequency" type="frequencyShortList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The frequency corresponding to the sensitivity repayment amount.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Content.Serviceability.Results.LoanDetail.Sensitivity.RepaymentFrequency</lx:path>
                                          <lx:label>Repayment Frequency</lx:label>
                                          <li:transactions>SVC</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="MaximumLoanAmount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Maximum recommended amount of this commitment.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.LoanDetail.MaximumLoanAmount</lx:path>
                                    <lx:label>Maximum Loan Amount</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_LoanDetail" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A reference to the LoanDetail record to which this applies.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.LoanDetail.x_LoanDetail</lx:path>
                                    <lx:label>Loan Detail</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="NetDisposableIncome" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The net disposable income calculated for the application.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Serviceability.Results.NetDisposableIncome</lx:path>
                                <lx:label>Net Disposable Income</lx:label>
                                <li:transactions>SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Amount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The amount of the net disposable income calculated for the application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.NetDisposableIncome.Amount</lx:path>
                                    <lx:label>Amount</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Captures the period over which the amount of the net disposable income is considered disposable.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.NetDisposableIncome.Frequency</lx:path>
                                    <lx:label>Frequency</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="TotalGrossIncome" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The total income of the application parties before accounting for taxes and deductions.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Serviceability.Results.TotalGrossIncome</lx:path>
                                <lx:label>Total Gross Income</lx:label>
                                <li:transactions>SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Amount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The total income amount of the application parties before accounting for taxes and deductions.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.TotalGrossIncome.Amount</lx:path>
                                    <lx:label>Amount</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The period over which the total gross income amount is calculated.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.TotalGrossIncome.Frequency</lx:path>
                                    <lx:label>Frequency</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="TotalNetIncome" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The total income of the application parties after all deductions such as taxes have been taken.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Serviceability.Results.TotalNetIncome</lx:path>
                                <lx:label>Total Net Income</lx:label>
                                <li:transactions>SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Amount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The total income amount of the application parties after all deductions such as taxes have been taken.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.TotalNetIncome.Amount</lx:path>
                                    <lx:label>Amount</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The period over which the total net income amount is calculated.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.TotalNetIncome.Frequency</lx:path>
                                    <lx:label>Frequency</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="TotalSystemCalculatedExpenses" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The total system calculated expenses.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Serviceability.Results.TotalSystemCalculatedExpenses</lx:path>
                                <lx:label>Total System Calculated Living Expenses</lx:label>
                                <li:transactions>SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Amount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The total system calculated expense amount.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.TotalSystemCalculatedExpenses.Amount</lx:path>
                                    <lx:label>Amount</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The frequency of the total system calculated expenses.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.TotalSystemCalculatedExpenses.Frequency</lx:path>
                                    <lx:label>Frequency</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="TotalSystemCalculatedLivingExpenses" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The estimated living expenses calculated by or on behalf of the application capture or origination system.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Serviceability.Results.TotalSystemCalculatedLivingExpenses</lx:path>
                                <lx:label>Total System Calculated Living Expenses</lx:label>
                                <li:transactions>SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Amount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The estimated living expenses amount calculated by or on behalf of the application capture or origination system.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.TotalSystemCalculatedLivingExpenses.Amount</lx:path>
                                    <lx:label>Amount</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The period over which the living expenses were calculated.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.TotalSystemCalculatedLivingExpenses.Frequency</lx:path>
                                    <lx:label>Frequency</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="TotalUserStatedLivingExpenses" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The total of living expenses declared by the application parties.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Content.Serviceability.Results.TotalUserStatedLivingExpenses</lx:path>
                                <lx:label>Total User Stated Living Expenses</lx:label>
                                <li:transactions>SVC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="Amount" type="currencyType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The total amount of living expenses declared by the application parties.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.TotalUserStatedLivingExpenses.Amount</lx:path>
                                    <lx:label>Amount</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Frequency" type="frequencyShortList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The period over which the living expenses were captured.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Content.Serviceability.Results.TotalUserStatedLivingExpenses.Frequency</lx:path>
                                    <lx:label>Frequency</lx:label>
                                    <li:transactions>SVC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="DSR" type="decimalType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The debt servicing ratio (DSR) that applies to the applicants.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Serviceability.Results.DSR</lx:path>
                              <lx:label>DSR</lx:label>
                              <li:transactions>SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="DTI" type="decimalType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Debt to income ratio (DTI). The ratio of the total credit limit of all debts to the gross annual income that applies to the parties to the application.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Serviceability.Results.DTI</lx:path>
                              <lx:label>DTI</lx:label>
                              <li:transactions>SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="GroupIdentifier" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>An identifer that can be used to identify the group.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Serviceability.Results.GroupIdentifier</lx:path>
                              <lx:label>Group Identifier</lx:label>
                              <li:transactions>SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="GroupName" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>If these serviceability results are for a group, this defines a name assigned to the group.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Serviceability.Results.GroupName</lx:path>
                              <lx:label>Group Name</lx:label>
                              <li:transactions>SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="LTI" type="decimalType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Loan to income ratio (LTI). The total loan repayment amounts to total actual gross income ratio that applies to the parties to the application.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Serviceability.Results.LTI</lx:path>
                              <lx:label>LTI</lx:label>
                              <li:transactions>SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="MaximumLoanAmount" type="currencyType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Maximum recommended amount of the loan, combining the individual commitments.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Serviceability.Results.MaximumLoanAmount</lx:path>
                              <lx:label>Maximum Loan Amount</lx:label>
                              <li:transactions>SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="MRIM" type="decimalType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Monthly mortgage repayment to income measure (MRIM). The monthly mortgage repayment to sensitised monthly income ratio that applies to the parties to the application.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Serviceability.Results.MRIM</lx:path>
                              <lx:label>MRIM</lx:label>
                              <li:transactions>SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="NSR" type="decimalType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The net surplus ratio (NSR) that applies to the applicants.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Serviceability.Results.NSR</lx:path>
                              <lx:label>NSR</lx:label>
                              <li:transactions>SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Serviceable" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Captures whether the party/s to the loan are able to service the funds they will be liable for should the application be successful.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Serviceability.Results.Serviceable</lx:path>
                              <lx:label>Serviceable</lx:label>
                              <li:transactions>SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Unique identifier for this serviceability results.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.Serviceability.Results.UniqueID</lx:path>
                              <lx:label>UniqueID</lx:label>
                              <li:transactions>SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                  </xs:sequence>
                  <xs:attribute name="CalculationIdentifier" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>An identifier issued by the serviceability calculator that can uniquely identify a serviceability calculation.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.Serviceability.CalculationIdentifier</lx:path>
                        <lx:label>Calculator Identifier</lx:label>
                        <li:transactions>SVC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                </xs:complexType>
              </xs:element>
              <xs:element name="StatementOfPosition" minOccurs="0" maxOccurs="unbounded">
                <xs:annotation>
                  <xs:documentation>Details of a Statement of Financial Position.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Content.StatementOfPosition</lx:path>
                    <lx:label>Statement Of Position</lx:label>
                    <li:transactions>CAL</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:sequence>
                    <xs:element name="Applicant" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>The applicant covered by this Statement of Position.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Content.StatementOfPosition.Applicant</lx:path>
                          <lx:label>Applicant</lx:label>
                          <li:transactions>CAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="x_Applicant" type="referenceType" use="optional">
                          <xs:annotation>
                            <xs:documentation>Link to the applicant covered by this Statement of Position.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Content.StatementOfPosition.Applicant.x_Applicant</lx:path>
                              <lx:label>Applicant</lx:label>
                              <li:transactions>CAL</li:transactions>
                              <lx:references>
                                <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                              </lx:references>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                  </xs:sequence>
                  <xs:attribute name="Date" type="dateType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The date of the Statement of Position.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.StatementOfPosition.Date</lx:path>
                        <lx:label>Date</lx:label>
                        <li:transactions>CAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="IsSigned" type="yesNoList" use="optional">
                    <xs:annotation>
                      <xs:documentation>Indicates that the Statement of Position has been fully signed.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.StatementOfPosition.IsSigned</lx:path>
                        <lx:label>Is Signed</lx:label>
                        <li:transactions>CAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="SequenceNumber" type="integerType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The order of this statement of position within the sequence.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.StatementOfPosition.SequenceNumber</lx:path>
                        <lx:label>Sequence Number</lx:label>
                        <li:transactions>CAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                    <xs:annotation>
                      <xs:documentation>Unique Identifier for this Statement of Position.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Content.StatementOfPosition.UniqueID</lx:path>
                        <lx:label>Unique ID</lx:label>
                        <li:transactions>CAL</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                </xs:complexType>
              </xs:element>
            </xs:sequence>
          </xs:complexType>
        </xs:element>
        <xs:element name="Instructions" minOccurs="1" maxOccurs="1">
          <xs:annotation>
            <xs:documentation>Determines the scope of the message. The message can be about the Application, Valuation, Mortgage Insurance, Loan Documents or an Error.</xs:documentation>
            <xs:appinfo>
              <lx:path>Package.Instructions</lx:path>
              <lx:label>Instructions</lx:label>
              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
            </xs:appinfo>
          </xs:annotation>
          <xs:complexType>
            <xs:choice>
              <xs:element name="AccountCreationInstructions" minOccurs="1" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>Loan account creation submission information or loan account creation status updates.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Instructions.AccountCreationInstructions</lx:path>
                    <lx:label>Account Creation Instructions</lx:label>
                    <li:transactions>ACC</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:choice>
                    <xs:element name="Submit" minOccurs="1" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Loan account creation submission information.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Instructions.AccountCreationInstructions.Submit</lx:path>
                          <lx:label>Submit</lx:label>
                          <li:transactions>ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="IsAccountVariation" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate whether the application being submitted is an account variation.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.AccountCreationInstructions.Submit.IsAccountVariation</lx:path>
                              <lx:label>Is Account Variation</lx:label>
                              <li:transactions>ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsResubmission" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate whether the application being submitted is a resubmission.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.AccountCreationInstructions.Submit.IsResubmission</lx:path>
                              <lx:label>Is Resubmission</lx:label>
                              <li:transactions>ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Update" minOccurs="1" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Application status or events information.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Instructions.AccountCreationInstructions.Update</lx:path>
                          <lx:label>Update</lx:label>
                          <li:transactions>ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Event" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>An event that occurs during the loan account creation process.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Instructions.AccountCreationInstructions.Update.Event</lx:path>
                                <lx:label>Event</lx:label>
                                <li:transactions>ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DateTime" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date and time at which the event is specified as having occurred.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.AccountCreationInstructions.Update.Event.DateTime</lx:path>
                                    <lx:label>Date Time</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free-text field to provide event specific details.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.AccountCreationInstructions.Update.Event.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="eventNameAccountCreationInstructionsList" use="required">
                                <xs:annotation>
                                  <xs:documentation>The standardised name for the event that has occurred during loan account creation processing.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.AccountCreationInstructions.Update.Event.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.AccountCreationInstructions.Update.Event.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Status" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The status that the application has achieved in the application workflow.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Instructions.AccountCreationInstructions.Update.Status</lx:path>
                                <lx:label>Status</lx:label>
                                <li:transactions>ACC</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DateTime" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date and time at which the application achieved the status.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.AccountCreationInstructions.Update.Status.DateTime</lx:path>
                                    <lx:label>Date Time</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free-text field for providing details about the status, such as comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.AccountCreationInstructions.Update.Status.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="statusNameAccountCreationInstructionsList" use="required">
                                <xs:annotation>
                                  <xs:documentation>The status of the application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.AccountCreationInstructions.Update.Status.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>ACC</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                      </xs:complexType>
                    </xs:element>
                  </xs:choice>
                </xs:complexType>
              </xs:element>
              <xs:element name="ApplicationInstructions" minOccurs="1" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>Wrapper element for submission information or application status updates.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Instructions.ApplicationInstructions</lx:path>
                    <lx:label>Application Instructions</lx:label>
                    <li:transactions>CAL,DAS,CNZ</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:choice>
                    <xs:element name="Submit" minOccurs="1" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Wrapper element for submission information.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Instructions.ApplicationInstructions.Submit</lx:path>
                          <lx:label>Submit</lx:label>
                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Condition" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Information provided in response to a request regarding conditions that must be met before the application can be approved.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Instructions.ApplicationInstructions.Submit.Condition</lx:path>
                                <lx:label>Condition</lx:label>
                                <li:transactions>CAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="ConditionResponseDescription" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A comment associated with the condition response status.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ApplicationInstructions.Submit.Condition.ConditionResponseDescription</lx:path>
                                    <lx:label>Condition Response Description</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConditionResponseStatus" type="conditionResponseStatusList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the degree to which the uploading party deems this response will meet the loan condition.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ApplicationInstructions.Submit.Condition.ConditionResponseStatus</lx:path>
                                    <lx:label>Condition Response Status</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LoanConditionText" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Detailed description of the condition.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ApplicationInstructions.Submit.Condition.LoanConditionText</lx:path>
                                    <lx:label>Loan Condition Text</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="x_SupportingDocument" type="referenceType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The unique ID of an attached document that satisfies this condition.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ApplicationInstructions.Submit.Condition.x_SupportingDocument</lx:path>
                                    <lx:label>Supporting Document</lx:label>
                                    <li:transactions>CAL</li:transactions>
                                    <lx:references>
                                      <lx:target>Package.Attachment</lx:target>
                                    </lx:references>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="AssessmentType" type="assessmentTypeApplicationInstructionsList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates the type of assessment.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.ApplicationInstructions.Submit.AssessmentType</lx:path>
                              <lx:label>Assessment Type</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsAccountVariation" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate whether the application being submitted is an account variation.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.ApplicationInstructions.Submit.IsAccountVariation</lx:path>
                              <lx:label>Is Account Variation</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsPricingRequest" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Client can submit pricing request to see how much they can borrow and at what interest rate.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.ApplicationInstructions.Submit.IsPricingRequest</lx:path>
                              <lx:label>Is Pricing Request</lx:label>
                              <li:transactions>CAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsResubmission" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate whether the application being submitted is a resubmission.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.ApplicationInstructions.Submit.IsResubmission</lx:path>
                              <lx:label>Is Resubmission</lx:label>
                              <li:transactions>CAL,DAS,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsSubmissionDocuments" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicate whether the package contains only submission documents for an already submitted application.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.ApplicationInstructions.Submit.IsSubmissionDocuments</lx:path>
                              <lx:label>Is Submission Documents</lx:label>
                              <li:transactions>CAL,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsSupportingDocuments" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate whether the submission only contains supporting documents for an already submitted application.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.ApplicationInstructions.Submit.IsSupportingDocuments</lx:path>
                              <lx:label>Is Supporting Documents</lx:label>
                              <li:transactions>CAL,CNZ</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Update" minOccurs="1" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Wrapper element for application Status or Events information.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Instructions.ApplicationInstructions.Update</lx:path>
                          <lx:label>Update</lx:label>
                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Event" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for application events.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Instructions.ApplicationInstructions.Update.Event</lx:path>
                                <lx:label>Event</lx:label>
                                <li:transactions>CAL,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DateTime" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date and time at which the event is specified as having occurred.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ApplicationInstructions.Update.Event.DateTime</lx:path>
                                    <lx:label>Date Time</lx:label>
                                    <li:transactions>CAL,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free-text field for the Broker/Lender to provide event specific details.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ApplicationInstructions.Update.Event.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>CAL,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="eventNameApplicationInstructionsList" use="required">
                                <xs:annotation>
                                  <xs:documentation>The standardised name for the event that has occurred during application processing.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ApplicationInstructions.Update.Event.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>CAL,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ApplicationInstructions.Update.Event.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CAL,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Status" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Contains information about the status that the application has achieved in the application workflow.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Instructions.ApplicationInstructions.Update.Status</lx:path>
                                <lx:label>Status</lx:label>
                                <li:transactions>CAL,DAS,CNZ</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:choice>
                                <xs:element name="Condition" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Wrapper element for information regarding conditions that must be met before the application can be approved. To be used when the Status Name is "Conditionally Approved".</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Condition</lx:path>
                                      <lx:label>Condition</lx:label>
                                      <li:transactions>CAL,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Product" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>Where a specific financial product (or multiple products, or product set) is conditional upon this requirement being met, the link to the product, products/s or product set can be captured here.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Condition.Product</lx:path>
                                            <lx:label>Product</lx:label>
                                            <li:transactions>CAL,DAS,CNZ</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="x_FinancialProduct" type="referenceType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>A cross reference to the financial product (or multiple products, or product set) that is conditional upon this requirement being met.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Condition.Product.x_FinancialProduct</lx:path>
                                                <lx:label>Financial Product</lx:label>
                                                <li:transactions>CAL,DAS,CNZ</li:transactions>
                                                <lx:references>
                                                  <lx:target>Package.Content.Application.Liability</lx:target>
                                                  <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                                  <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                                </lx:references>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="ConditionOwner" type="conditionOwnerApplicationInstructionsList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates the party responsible for satisfying the condition.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Condition.ConditionOwner</lx:path>
                                          <lx:label>Condition Owner</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ConditionStatus" type="conditionStatusApplicationInstructionsList" use="required">
                                      <xs:annotation>
                                        <xs:documentation>Indicates the current status of the condition.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Condition.ConditionStatus</lx:path>
                                          <lx:label>Condition Status</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="ConditionType" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The type of condition.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Condition.ConditionType</lx:path>
                                          <lx:label>Condition Type</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CreatedDateTime" type="dateTimeType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date and time at which the condition was issued.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Condition.CreatedDateTime</lx:path>
                                          <lx:label>Created Date Time</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="DocRequirement" type="yesNoList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Flag that indicates whether this condition is satisfied by the provision of a document.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Condition.DocRequirement</lx:path>
                                          <lx:label>Doc Requirement</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="LoanConditionText" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Detailed description of the condition.E.g. "Privacy Consent form signed by each individual applicant"</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Condition.LoanConditionText</lx:path>
                                          <lx:label>Loan Condition Text</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="PreconditionToStage" type="preConditionToStageApplicationInstructionsList" use="required">
                                      <xs:annotation>
                                        <xs:documentation>The condition must be Satisfied or Waived before the application can enter this stage.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Condition.PreconditionToStage</lx:path>
                                          <lx:label>Precondition To Stage</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Condition.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UpdatedDateTime" type="dateTimeType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date and time at which the condition was last updated.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Condition.UpdatedDateTime</lx:path>
                                          <lx:label>Updated Date Time</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_SupportingDocument" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The unique ID of an attached document that satisfies this condition.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Condition.x_SupportingDocument</lx:path>
                                          <lx:label>Supporting Document</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Attachment</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="Declined" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Wrapper element for information when the application is declined. To be used when the Status Name is "Declined".</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Declined</lx:path>
                                      <lx:label>Declined</lx:label>
                                      <li:transactions>CAL,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="CreditReportingBody" type="creditReportingBodyList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The credit reporting body used to provide the credit report/s, where Reason is "Credit History".</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Declined.CreditReportingBody</lx:path>
                                          <lx:label>Credit Reporting Body</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Details" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A free-text field for providing details about the reason why the application has been declined. E.g. when Reason = "Security" the Details would capture the details such as "The property is located in a high density area and it can't be accepted as security"</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Declined.Details</lx:path>
                                          <lx:label>Details</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="Reason" type="declinedReasonApplicationInstructionsList" use="required">
                                      <xs:annotation>
                                        <xs:documentation>The reason why the application has been declined.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Declined.Reason</lx:path>
                                          <lx:label>Reason</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Declined.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                                <xs:element name="PreApproved" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Wrapper element for information when the application is pre-approved. To be used when the Status Name is "PreApproved".</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.PreApproved</lx:path>
                                      <lx:label>Pre Approved</lx:label>
                                      <li:transactions>CAL,DAS,CNZ</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="EstablishmentAndGovernmentFees" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The amount of establishment and government fees to be paid should the applicant proceed to approval and draw down with the pre approved loan.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.PreApproved.EstablishmentAndGovernmentFees</lx:path>
                                          <lx:label>Establishment And Government Fees</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="EstimatedBorrowingPower" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The maximum loan amount estimated by the lender to be supported by the pre approval application.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.PreApproved.EstimatedBorrowingPower</lx:path>
                                          <lx:label>Estimated Borrowing Power</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="EstimatedRepaymentAmount" type="currencyType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The repayment amount estimated by the lender when pre approving the application.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.PreApproved.EstimatedRepaymentAmount</lx:path>
                                          <lx:label>Estimated Repayment Amount</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.PreApproved.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>CAL,DAS,CNZ</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:choice>
                              <xs:attribute name="DateTime" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date and time at which the application achieved the status.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.DateTime</lx:path>
                                    <lx:label>Date Time</lx:label>
                                    <li:transactions>CAL,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free-text field for providing details about the Status, such as comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>CAL,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="statusNameApplicationInstructionsList" use="required">
                                <xs:annotation>
                                  <xs:documentation>The status of the loan application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ApplicationInstructions.Update.Status.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>CAL,DAS,CNZ</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                      </xs:complexType>
                    </xs:element>
                  </xs:choice>
                </xs:complexType>
              </xs:element>
              <xs:element name="CreditDecisionInstructions" minOccurs="1" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>Instructions regarding the submission and responses related to the credit decision process, including events, errors, and status updates.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Instructions.CreditDecisionInstructions</lx:path>
                    <lx:label>Credit Decision Instructions</lx:label>
                    <li:transactions>CDA</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:choice>
                    <xs:element name="Submit" minOccurs="1" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Submission information related to the credit decision process.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Instructions.CreditDecisionInstructions.Submit</lx:path>
                          <lx:label>Submit</lx:label>
                          <li:transactions>CDA</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="AssessmentType" type="creditAssessmentTypeInstructionList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of credit assessment being requested by the credit decision process.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.CreditDecisionInstructions.Submit.AssessmentType</lx:path>
                              <lx:label>Assessment Type</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsResubmission" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Whether the credit assessment request is a resubmission.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.CreditDecisionInstructions.Submit.IsResubmission</lx:path>
                              <lx:label>Is Resubmission</lx:label>
                              <li:transactions>CDA</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Update" minOccurs="1" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Updates related to the credit decision process.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Instructions.CreditDecisionInstructions.Update</lx:path>
                          <lx:label>Update</lx:label>
                          <li:transactions>CDA</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Event" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Events related to the credit decision process.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Instructions.CreditDecisionInstructions.Update.Event</lx:path>
                                <lx:label>Event</lx:label>
                                <li:transactions>CDA</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DateTime" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date and time at which the event is specified as having occurred.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.CreditDecisionInstructions.Update.Event.DateTime</lx:path>
                                    <lx:label>Date Time</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free-text field for the submitter to provide event specific details.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.CreditDecisionInstructions.Update.Event.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="eventNameCreditDecisionInstructionsList" use="required">
                                <xs:annotation>
                                  <xs:documentation>The standardised name for the event that has occurred during processing of the credit decision assessment. Note that these include events from the credit application process, not just the credit decision process.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.CreditDecisionInstructions.Update.Event.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.CreditDecisionInstructions.Update.Event.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Status" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>Workflow status of the credit decision process.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Instructions.CreditDecisionInstructions.Update.Status</lx:path>
                                <lx:label>Status</lx:label>
                                <li:transactions>CDA</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DateTime" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date and time at which the credit decision process achieved the current status.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.CreditDecisionInstructions.Update.Status.DateTime</lx:path>
                                    <lx:label>Date Time</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Further details of the status of the credit decision process, such as comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.CreditDecisionInstructions.Update.Status.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="statusNameCreditDecisionAssessmentList" use="required">
                                <xs:annotation>
                                  <xs:documentation>The name of the status reached in the credit decision workflow.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.CreditDecisionInstructions.Update.Status.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>CDA</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                      </xs:complexType>
                    </xs:element>
                  </xs:choice>
                </xs:complexType>
              </xs:element>
              <xs:element name="DocumentsAndSettlementInstructions" minOccurs="1" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>Wrapper element for Documents and Settlement information or status updates.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions</lx:path>
                    <lx:label>Documents And Settlement Instructions</lx:label>
                    <li:transactions>DAS</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:choice>
                    <xs:element name="Submit" minOccurs="1" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Wrapper element for Documents and Settlement submission information.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit</lx:path>
                          <lx:label>Submit</lx:label>
                          <li:transactions>DAS</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="AdditionalDocument" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Request for additional documentation to be produced.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.AdditionalDocument</lx:path>
                                <lx:label>Additional Document</lx:label>
                                <li:transactions>DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Context" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Wrapper element for the context of the additional document to be produced.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.AdditionalDocument.Context</lx:path>
                                      <lx:label>Context</lx:label>
                                      <li:transactions>DAS</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="Type" type="contexTypeAdditionalDocumentList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The context type of the additional document.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.AdditionalDocument.Context.Type</lx:path>
                                          <lx:label>Type</lx:label>
                                          <li:transactions>DAS</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_Context" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Pointer to the item that the context type refers to.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.AdditionalDocument.Context.x_Context</lx:path>
                                          <lx:label>Context</lx:label>
                                          <li:transactions>DAS</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.Address</lx:target>
                                            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                                            <lx:target>Package.Content.Application.Liability</lx:target>
                                            <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                            <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                                            <lx:target>Package.Content.Application.RealEstateAsset</lx:target>
                                            <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="DocumentName" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The name of the additional document to be produced.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.AdditionalDocument.DocumentName</lx:path>
                                    <lx:label>Document Name</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="NumberOfCopies" type="integerType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The number of copies of the additional document to be produced.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.AdditionalDocument.NumberOfCopies</lx:path>
                                    <lx:label>Number Of Copies</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Condition" minOccurs="1" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for information regarding conditions that must be met as part of the Documentation process.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.Condition</lx:path>
                                <lx:label>Condition</lx:label>
                                <li:transactions>DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Product" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Where a specific financial product (or multiple products, or product set) is conditional upon this requirement being met, the link to the product, products/s or product set can be captured here.</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.Condition.Product</lx:path>
                                      <lx:label>Product</lx:label>
                                      <li:transactions>DAS</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:attribute name="x_FinancialProduct" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>A cross reference to the financial product (or multiple products, or product set) that is conditional upon this requirement being met.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.Condition.Product.x_FinancialProduct</lx:path>
                                          <lx:label>Financial Product</lx:label>
                                          <li:transactions>DAS</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Content.Application.Liability</lx:target>
                                            <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                            <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="ConditionOwner" type="conditionOwnerLoanDocumentsInstructionsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Indicates the party responsible for satisfying the condition.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.Condition.ConditionOwner</lx:path>
                                    <lx:label>Condition Owner</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="ConditionStatus" type="conditionStatusLoanDocumentsInstructionsList" use="required">
                                <xs:annotation>
                                  <xs:documentation>Indicates the current status of the condition.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.Condition.ConditionStatus</lx:path>
                                    <lx:label>Condition Status</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="CreatedDateTime" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date and time at which the condition was issued.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.Condition.CreatedDateTime</lx:path>
                                    <lx:label>Created Date Time</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="DocRequirement" type="yesNoList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Flag that indicates whether this condition is satisfied by the provision of a document.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.Condition.DocRequirement</lx:path>
                                    <lx:label>Doc Requirement</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="LoanConditionText" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>Detailed description of the condition.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.Condition.LoanConditionText</lx:path>
                                    <lx:label>Loan Condition Text</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="PreconditionToStage" type="preConditionToStageLoanDocumentsInstructionsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The condition must be Satisfied or Waived before the application can enter this stage.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.Condition.PreconditionToStage</lx:path>
                                    <lx:label>Precondition To Stage</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned identifier uniquely identifying the Condition element.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.Condition.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UpdatedDateTime" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date and time at which the condition was last updated.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.Condition.UpdatedDateTime</lx:path>
                                    <lx:label>Updated Date Time</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                        <xs:attribute name="IsResubmission" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Flag to indicate whether the Document or Settlement information being submitted is a resubmission.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Submit.IsResubmission</lx:path>
                              <lx:label>Is Resubmission</lx:label>
                              <li:transactions>DAS</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Update" minOccurs="1" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Wrapper element for Documentation or Settlement Events information.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Update</lx:path>
                          <lx:label>Update</lx:label>
                          <li:transactions>DAS</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Event" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for Documentation or Settlement events.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Update.Event</lx:path>
                                <lx:label>Event</lx:label>
                                <li:transactions>DAS</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DateTime" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date and time at which the event is specified as having occurred.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Update.Event.DateTime</lx:path>
                                    <lx:label>Date Time</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free-text field for the Broker/Lender/Settlement Agent to provide event specific details.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Update.Event.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="eventNameLoanDocumentsInstructionsList" use="required">
                                <xs:annotation>
                                  <xs:documentation>The standardised name for the event that has occurred during Documentation or Settlement processing.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Update.Event.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier for this Update Event.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.DocumentsAndSettlementInstructions.Update.Event.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>DAS</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                      </xs:complexType>
                    </xs:element>
                  </xs:choice>
                </xs:complexType>
              </xs:element>
              <xs:element name="ErrorInstructions" minOccurs="1" maxOccurs="unbounded">
                <xs:annotation>
                  <xs:documentation>Wrapper element for error information.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Instructions.ErrorInstructions</lx:path>
                    <lx:label>Error Instructions</lx:label>
                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:sequence>
                    <xs:element name="Annotation" minOccurs="0" maxOccurs="unbounded">
                      <xs:annotation>
                        <xs:documentation>Contains the text of the error message. May occur more than once, with message text for each different audience e.g.. layperson, technician.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Instructions.ErrorInstructions.Annotation</lx:path>
                          <lx:label>Annotation</lx:label>
                          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="Details" type="stringType" use="optional">
                          <xs:annotation>
                            <xs:documentation>The contents of the error message.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.ErrorInstructions.Annotation.Details</lx:path>
                              <lx:label>Details</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="Type" type="annotationTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>An indication of the intended audience for the message annotation. Can be used by the receiving system to determine the audience for the message e.g. technical versus end user error messages.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.ErrorInstructions.Annotation.Type</lx:path>
                              <lx:label>Type</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                          <xs:annotation>
                            <xs:documentation>System assigned unique identifier.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.ErrorInstructions.Annotation.UniqueID</lx:path>
                              <lx:label>Unique ID</lx:label>
                              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                  </xs:sequence>
                  <xs:attribute name="ErrorSource" type="errorSourceList" use="optional">
                    <xs:annotation>
                      <xs:documentation>Where Error Type is "Data Error", the technology that was used to generate the error.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Instructions.ErrorInstructions.ErrorSource</lx:path>
                        <lx:label>Error Source</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ErrorSourceVersion" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>Where Error Type is "Data Error", the version of the technology identified in Error Source that was used to generate the error.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Instructions.ErrorInstructions.ErrorSourceVersion</lx:path>
                        <lx:label>Error Source Version</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="Type" type="errorInstructionsTypeList" use="required">
                    <xs:annotation>
                      <xs:documentation>The standardised name for the type of error.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Instructions.ErrorInstructions.Type</lx:path>
                        <lx:label>Type</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                    <xs:annotation>
                      <xs:documentation>System assigned unique identifier.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Instructions.ErrorInstructions.UniqueID</lx:path>
                        <lx:label>Unique ID</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="XPath" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The position in the XML application instance that this message relates to, in the form of an XPath expression.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Instructions.ErrorInstructions.XPath</lx:path>
                        <lx:label>XPath</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                </xs:complexType>
              </xs:element>
              <xs:element name="LendersMortgageInsuranceInstructions" minOccurs="1" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>Instructions regarding the submission and responses related to the lenders mortgage insurance process.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions</lx:path>
                    <lx:label>Lenders Mortgage Insurance Instructions</lx:label>
                    <li:transactions>LMI</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:choice>
                    <xs:element name="Submit" minOccurs="1" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Submission information related to the lenders mortgage insurance process.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Submit</lx:path>
                          <lx:label>Submit</lx:label>
                          <li:transactions>LMI</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="SubmissionType" type="lendersMortgageInsuranceInstructionList" use="optional">
                          <xs:annotation>
                            <xs:documentation>The type of lenders mortgage insurance submission.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Submit.SubmissionType</lx:path>
                              <lx:label>Submission Type</lx:label>
                              <li:transactions>LMI</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Update" minOccurs="1" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Lenders mortgage insurance status or events information.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update</lx:path>
                          <lx:label>Update</lx:label>
                          <li:transactions>LMI</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Event" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>Wrapper element for lenders mortgage insurance events.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Event</lx:path>
                                <lx:label>Event</lx:label>
                                <li:transactions>LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DateTime" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date and time at which the event is specified as having occurred.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Event.DateTime</lx:path>
                                    <lx:label>Date Time</lx:label>
                                    <li:transactions>LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free-text field to provide event specific details.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Event.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="eventNameLendersMortgageInsuranceInstructionsList" use="required">
                                <xs:annotation>
                                  <xs:documentation>The standardised name for the event that has occurred during the LMI processing.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Event.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Event.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Status" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The status that the lenders mortgage insurance application has achieved in the workflow.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Status</lx:path>
                                <lx:label>Status</lx:label>
                                <li:transactions>LMI</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:sequence>
                                <xs:element name="Condition" minOccurs="0" maxOccurs="unbounded">
                                  <xs:annotation>
                                    <xs:documentation>Wrapper element for information regarding conditions that must be met before the lenders mortgage insurance can be approved. To be used when the Status Name is "Conditionally Approved".</xs:documentation>
                                    <xs:appinfo>
                                      <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Status.Condition</lx:path>
                                      <lx:label>Condition</lx:label>
                                      <li:transactions>LMI</li:transactions>
                                    </xs:appinfo>
                                  </xs:annotation>
                                  <xs:complexType>
                                    <xs:sequence>
                                      <xs:element name="Product" minOccurs="0" maxOccurs="unbounded">
                                        <xs:annotation>
                                          <xs:documentation>Where a specific financial product (or multiple products, or product set) is conditional upon this requirement being met, the link to the product, products/s or product set can be captured here.</xs:documentation>
                                          <xs:appinfo>
                                            <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Status.Condition.Product</lx:path>
                                            <lx:label>Product</lx:label>
                                            <li:transactions>LMI</li:transactions>
                                          </xs:appinfo>
                                        </xs:annotation>
                                        <xs:complexType>
                                          <xs:attribute name="x_FinancialProduct" type="referenceType" use="optional">
                                            <xs:annotation>
                                              <xs:documentation>A cross reference to the financial product (or multiple products, or product set) that is conditional upon this requirement being met.</xs:documentation>
                                              <xs:appinfo>
                                                <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Status.Condition.Product.x_FinancialProduct</lx:path>
                                                <lx:label>Financial Product</lx:label>
                                                <li:transactions>LMI</li:transactions>
                                                <lx:references>
                                                  <lx:target>Package.Content.Application.Liability</lx:target>
                                                  <lx:target>Package.Content.Application.LoanDetails</lx:target>
                                                  <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
                                                </lx:references>
                                              </xs:appinfo>
                                            </xs:annotation>
                                          </xs:attribute>
                                        </xs:complexType>
                                      </xs:element>
                                    </xs:sequence>
                                    <xs:attribute name="ConditionStatus" type="conditionStatusLendersMortgageInsuranceInstructionsList" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Indicates the current status of the condition.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Status.Condition.ConditionStatus</lx:path>
                                          <lx:label>ConditionStatus</lx:label>
                                          <li:transactions>LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="CreatedDateTime" type="dateTimeType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date and time at which the condition was issued.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Status.Condition.CreatedDateTime</lx:path>
                                          <lx:label>Created Date Time</lx:label>
                                          <li:transactions>LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="LoanConditionText" type="stringType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>Detailed description of the condition.E.g. "Privacy Consent form signed by each individual applicant"</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Status.Condition.LoanConditionText</lx:path>
                                          <lx:label>Loan Condition Text</lx:label>
                                          <li:transactions>LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>System assigned unique identifier.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Status.Condition.UniqueID</lx:path>
                                          <lx:label>Unique ID</lx:label>
                                          <li:transactions>LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="UpdatedDateTime" type="dateTimeType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The date and time at which the condition was last updated.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Status.Condition.UpdatedDateTime</lx:path>
                                          <lx:label>Updated Date Time</lx:label>
                                          <li:transactions>LMI</li:transactions>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                    <xs:attribute name="x_SupportingDocument" type="referenceType" use="optional">
                                      <xs:annotation>
                                        <xs:documentation>The unique ID of an attached document that satisfies this condition.</xs:documentation>
                                        <xs:appinfo>
                                          <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Status.Condition.x_SupportingDocument</lx:path>
                                          <lx:label>Supporting Document</lx:label>
                                          <li:transactions>LMI</li:transactions>
                                          <lx:references>
                                            <lx:target>Package.Attachment</lx:target>
                                          </lx:references>
                                        </xs:appinfo>
                                      </xs:annotation>
                                    </xs:attribute>
                                  </xs:complexType>
                                </xs:element>
                              </xs:sequence>
                              <xs:attribute name="DateTime" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date and time at which the application achieved the status.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Status.DateTime</lx:path>
                                    <lx:label>DateTime</lx:label>
                                    <li:transactions>LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free-text field for providing details about the status, such as comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Status.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="statusNameLendersMortgageInsuranceInstructionsList" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The status of the lenders mortgage insurance application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.LendersMortgageInsuranceInstructions.Update.Status.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>LMI</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                      </xs:complexType>
                    </xs:element>
                  </xs:choice>
                </xs:complexType>
              </xs:element>
              <xs:element name="ServiceabilityInstructions" minOccurs="1" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>Instructions regarding the submission and responses related to the serviceability calculation.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Instructions.ServiceabilityInstructions</lx:path>
                    <lx:label>Serviceability Instructions</lx:label>
                    <li:transactions>SVC</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:sequence>
                    <xs:element name="Submit" minOccurs="1" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Submission information related to the serviceability calculation.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Instructions.ServiceabilityInstructions.Submit</lx:path>
                          <lx:label>Submit</lx:label>
                          <li:transactions>SVC</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="ServiceabilityAssessmentType" type="serviceabilityAssessmentTypeList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates the type of serviceability assessment.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.ServiceabilityInstructions.Submit.ServiceabilityAssessmentType</lx:path>
                              <lx:label>Serviceability Assessment Type</lx:label>
                              <li:transactions>SVC</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                  </xs:sequence>
                </xs:complexType>
              </xs:element>
              <xs:element name="ValuationInstructions" minOccurs="1" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>Valuation information or status updates.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Instructions.ValuationInstructions</lx:path>
                    <lx:label>Valuation Instructions</lx:label>
                    <li:transactions>VAL</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:choice>
                    <xs:element name="Submit" minOccurs="1" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Valuation submission information.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Instructions.ValuationInstructions.Submit</lx:path>
                          <lx:label>Submit</lx:label>
                          <li:transactions>VAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:attribute name="IsInquiry" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that the request being submitted is an inquiry for a quote.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.ValuationInstructions.Submit.IsInquiry</lx:path>
                              <lx:label>IsInquiry</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                        <xs:attribute name="IsVariation" type="yesNoList" use="optional">
                          <xs:annotation>
                            <xs:documentation>Indicates that the request being submitted is a fee variation.</xs:documentation>
                            <xs:appinfo>
                              <lx:path>Package.Instructions.ValuationInstructions.Submit.IsVariation</lx:path>
                              <lx:label>Is Valuation Variation</lx:label>
                              <li:transactions>VAL</li:transactions>
                            </xs:appinfo>
                          </xs:annotation>
                        </xs:attribute>
                      </xs:complexType>
                    </xs:element>
                    <xs:element name="Update" minOccurs="1" maxOccurs="1">
                      <xs:annotation>
                        <xs:documentation>Valuation status or events information.</xs:documentation>
                        <xs:appinfo>
                          <lx:path>Package.Instructions.ValuationInstructions.Update</lx:path>
                          <lx:label>Update</lx:label>
                          <li:transactions>VAL</li:transactions>
                        </xs:appinfo>
                      </xs:annotation>
                      <xs:complexType>
                        <xs:sequence>
                          <xs:element name="Event" minOccurs="0" maxOccurs="unbounded">
                            <xs:annotation>
                              <xs:documentation>A valuation event.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Instructions.ValuationInstructions.Update.Event</lx:path>
                                <lx:label>Event</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DateTime" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date and time at which the event is specified as having occurred.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ValuationInstructions.Update.Event.DateTime</lx:path>
                                    <lx:label>Date Time</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free-text field for the broker/lender/valuer to provide event specific details.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ValuationInstructions.Update.Event.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="eventNameValuationInstructionsList" use="required">
                                <xs:annotation>
                                  <xs:documentation>The standardised name for the event that has occurred during valuation processing.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ValuationInstructions.Update.Event.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Subject" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The subject line of the event details.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ValuationInstructions.Update.Event.Subject</lx:path>
                                    <lx:label>Subject</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>System assigned unique identifier for this Update Event.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ValuationInstructions.Update.Event.UniqueID</lx:path>
                                    <lx:label>Unique ID</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                          <xs:element name="Status" minOccurs="0" maxOccurs="1">
                            <xs:annotation>
                              <xs:documentation>The status that the application has achieved in the application workflow.</xs:documentation>
                              <xs:appinfo>
                                <lx:path>Package.Instructions.ValuationInstructions.Update.Status</lx:path>
                                <lx:label>Status</lx:label>
                                <li:transactions>VAL</li:transactions>
                              </xs:appinfo>
                            </xs:annotation>
                            <xs:complexType>
                              <xs:attribute name="DateTime" type="dateTimeType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>The date and time at which the application achieved the status.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ValuationInstructions.Update.Status.DateTime</lx:path>
                                    <lx:label>Date Time</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Details" type="stringType" use="optional">
                                <xs:annotation>
                                  <xs:documentation>A free-text field for providing details about the Status, such as comments.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ValuationInstructions.Update.Status.Details</lx:path>
                                    <lx:label>Details</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                              <xs:attribute name="Name" type="statusNameValuationInstructionsList" use="required">
                                <xs:annotation>
                                  <xs:documentation>The status of the loan application.</xs:documentation>
                                  <xs:appinfo>
                                    <lx:path>Package.Instructions.ValuationInstructions.Update.Status.Name</lx:path>
                                    <lx:label>Name</lx:label>
                                    <li:transactions>VAL</li:transactions>
                                  </xs:appinfo>
                                </xs:annotation>
                              </xs:attribute>
                            </xs:complexType>
                          </xs:element>
                        </xs:sequence>
                      </xs:complexType>
                    </xs:element>
                  </xs:choice>
                </xs:complexType>
              </xs:element>
            </xs:choice>
          </xs:complexType>
        </xs:element>
        <xs:element name="Publisher" minOccurs="1" maxOccurs="1">
          <xs:annotation>
            <xs:documentation>The publisher of the specific XML file, being the party responsible for the content of the file as well as the work instruction, and to whom the response message would be delivered. For a mortgage broker sending an application to a lender, this would likely be the aggregator if they are providing the software, or perhaps the broker company or broker otherwise. For a lender requesting LMI, the lender is the publisher. For the LMI response, the LMI provider is the publisher.</xs:documentation>
            <xs:appinfo>
              <lx:path>Package.Publisher</lx:path>
              <lx:label>Publisher</lx:label>
              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
            </xs:appinfo>
          </xs:annotation>
          <xs:complexType>
            <xs:sequence>
              <xs:element name="PhoneNumber" type="phoneType" minOccurs="0" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>A phone number for the publisher</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Publisher.PhoneNumber</lx:path>
                    <lx:label>Phone Number</lx:label>
                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
              </xs:element>
              <xs:element name="RelatedSoftware" minOccurs="0" maxOccurs="unbounded">
                <xs:annotation>
                  <xs:documentation>Details of software that takes part in the creation of this message instance but is not the main software that is regarded as the message publisher.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Publisher.RelatedSoftware</lx:path>
                    <lx:label>Related Software</lx:label>
                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:attribute name="Description" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>A free-text description of the related software provider, for example the company name.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Publisher.RelatedSoftware.Description</lx:path>
                        <lx:label>Description</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="Environment" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The environment of the related software used by the message publisher.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Publisher.RelatedSoftware.Environment</lx:path>
                        <lx:label>Environment</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="LIXICode" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The LIXI-specified code of the licensee that provides the related software.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Publisher.RelatedSoftware.LIXICode</lx:path>
                        <lx:label>LIXI Code</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="Name" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The name of the related software.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Publisher.RelatedSoftware.Name</lx:path>
                        <lx:label>Name</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="TechnicalEmail" type="emailPattern" use="optional">
                    <xs:annotation>
                      <xs:documentation>Captures the email address to be used by the recipient to advise the sender's technical support contact of any technical problems with the related software.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Publisher.RelatedSoftware.TechnicalEmail</lx:path>
                        <lx:label>Technical Email</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="Version" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The version of the software application.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Publisher.RelatedSoftware.Version</lx:path>
                        <lx:label>Version</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                </xs:complexType>
              </xs:element>
              <xs:element name="Software" minOccurs="0" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>Wrapper element for the publisher software details</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Publisher.Software</lx:path>
                    <lx:label>Software</lx:label>
                    <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:attribute name="Description" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>A free-text Description of the software provider, for example the company name.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Publisher.Software.Description</lx:path>
                        <lx:label>Description</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="Environment" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The environment of the software used by the message publisher.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Publisher.Software.Environment</lx:path>
                        <lx:label>Environment</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="LIXICode" type="stringType" use="required">
                    <xs:annotation>
                      <xs:documentation>The LIXI-specified code of the licensee that provides the Software.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Publisher.Software.LIXICode</lx:path>
                        <lx:label>LIXI Code</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="Name" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The name of the software application used by the message publisher.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Publisher.Software.Name</lx:path>
                        <lx:label>Name</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="TechnicalEmail" type="emailPattern" use="optional">
                    <xs:annotation>
                      <xs:documentation>Captures the email address to be used by the recipient to advise the sender's technical support contact of any technical problems with the application.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Publisher.Software.TechnicalEmail</lx:path>
                        <lx:label>Technical Email</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="Version" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The version of the software application used by the message publisher.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Publisher.Software.Version</lx:path>
                        <lx:label>Version</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                </xs:complexType>
              </xs:element>
            </xs:sequence>
            <xs:attribute name="CompanyName" type="stringType" use="optional">
              <xs:annotation>
                <xs:documentation>The company name of the publisher</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.Publisher.CompanyName</lx:path>
                  <lx:label>Company Name</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
            <xs:attribute name="ContactName" type="stringType" use="optional">
              <xs:annotation>
                <xs:documentation>Contact person from the publisher</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.Publisher.ContactName</lx:path>
                  <lx:label>Contact Name</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
            <xs:attribute name="Email" type="emailPattern" use="optional">
              <xs:annotation>
                <xs:documentation>An email address of the publisher contact person</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.Publisher.Email</lx:path>
                  <lx:label>Email</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
            <xs:attribute name="LIXICode" type="stringType" use="required">
              <xs:annotation>
                <xs:documentation>The LIXI-specified code of the sender/publisher of the Package.</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.Publisher.LIXICode</lx:path>
                  <lx:label>LIXI Code</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
            <xs:attribute name="PublishedDateTime" type="dateTimeType" use="optional">
              <xs:annotation>
                <xs:documentation>The date and time that the XML file was generated.</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.Publisher.PublishedDateTime</lx:path>
                  <lx:label>Published Date Time</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
          </xs:complexType>
        </xs:element>
        <xs:element name="Recipient" minOccurs="1" maxOccurs="unbounded">
          <xs:annotation>
            <xs:documentation>The intended recipients (company or company-specified routing) for the Package from the Publisher.</xs:documentation>
            <xs:appinfo>
              <lx:path>Package.Recipient</lx:path>
              <lx:label>Recipient</lx:label>
              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
            </xs:appinfo>
          </xs:annotation>
          <xs:complexType>
            <xs:sequence>
              <xs:element name="Software" minOccurs="0" maxOccurs="1">
                <xs:annotation>
                  <xs:documentation>The software used by the message recipient.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.Recipient.Software</lx:path>
                    <lx:label>Software</lx:label>
                    <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:attribute name="Environment" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The environment of the software used by the message recipient.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.Recipient.Software.Environment</lx:path>
                        <lx:label>Environment</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                </xs:complexType>
              </xs:element>
            </xs:sequence>
            <xs:attribute name="Description" type="stringType" use="optional">
              <xs:annotation>
                <xs:documentation>A free-text description of the recipient, for example the company name.</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.Recipient.Description</lx:path>
                  <lx:label>Description</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
            <xs:attribute name="LIXICode" type="stringType" use="required">
              <xs:annotation>
                <xs:documentation>The LIXI-specified code for the licensee that is the intended recipient for the Package.</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.Recipient.LIXICode</lx:path>
                  <lx:label>LIXI Code</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
            <xs:attribute name="RoutingCode" type="stringType" use="optional">
              <xs:annotation>
                <xs:documentation>A recipient-specified code that enables the recipient to route the Package to a specific destination with the organisation.</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.Recipient.RoutingCode</lx:path>
                  <lx:label>Routing Code</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
          </xs:complexType>
        </xs:element>
        <xs:element name="SchemaVersion" minOccurs="1" maxOccurs="1">
          <xs:annotation>
            <xs:documentation>Contains information about the package schema version</xs:documentation>
            <xs:appinfo>
              <lx:path>Package.SchemaVersion</lx:path>
              <lx:label>Schema Version</lx:label>
              <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
            </xs:appinfo>
          </xs:annotation>
          <xs:complexType>
            <xs:attribute name="GuidebookCode" type="stringType" use="optional">
              <xs:annotation>
                <xs:documentation>A LIXI-provided unique identifier for the 'type' of guidebook (e.g. full doc residential home loan) for the specific guidebook owner.</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.SchemaVersion.GuidebookCode</lx:path>
                  <lx:label>Guidebook Code</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
            <xs:attribute name="GuidebookName" type="stringType" use="optional">
              <xs:annotation>
                <xs:documentation>If a message was created using an Electronic Guidebook, this attribute contains the human readable name of that Electronic Guidebook.</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.SchemaVersion.GuidebookName</lx:path>
                  <lx:label>Guidebook Name</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
            <xs:attribute name="GuidebookVersion" type="stringType" use="optional">
              <xs:annotation>
                <xs:documentation>Captures the message recipient (lender, MM, funder, insurer etc) guidebook version number that the application XML was written to conform with, for the specific type of guidebook (guidebook code)</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.SchemaVersion.GuidebookVersion</lx:path>
                  <lx:label>Guidebook Version</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
            <xs:attribute name="LIXICustomVersion" type="stringType" use="optional">
              <xs:annotation>
                <xs:documentation>Each LIXI Schema that is customised by restriction is allocated a Custom Version Identifier. If the XML instance is created from a LIXI Transaction Schema the LIXICustomVersion is optional and should remain empty. If the XML instance is created from a schema that has been Customised by Restriction, this attribute is mandatory and holds the composite identifier constructed from: the LIXI Code of the author organisation; the Customisation Code (jointly defined by LIXI and the customisation author); and the datetimestamp of when the Customised Schema was created.</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.SchemaVersion.LIXICustomVersion</lx:path>
                  <lx:label>LIXI Version</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
            <xs:attribute name="LIXITransactionType" type="stringType" use="required" fixed="Autopopulated by distribution process">
              <xs:annotation>
                <xs:documentation>Identifies the transaction type of the schema, for example CAL (for a Credit Application) or DAS (for Documents Preparation and Settlement).</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.SchemaVersion.LIXITransactionType</lx:path>
                  <lx:label>LIXI Transaction Type</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
            <xs:attribute name="LIXIVersion" type="stringType" use="required" fixed="2.2.28">
              <xs:annotation>
                <xs:documentation>Captures the LIXI version number that the application XML was written to conform with.</xs:documentation>
                <xs:appinfo>
                  <lx:path>Package.SchemaVersion.LIXIVersion</lx:path>
                  <lx:label>LIXI Version</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:attribute>
          </xs:complexType>
        </xs:element>
        <xs:element name="TransformMetadata" minOccurs="0" maxOccurs="1">
          <xs:annotation>
            <xs:documentation>Wrapper element for storing metadata required for enabling round trip transforms from a different data model into a message conforming to this schema and back again without loss of fidelity. Note that no data that carries meaning in this LIXI 2.x schema may be carried in this element.</xs:documentation>
            <xs:appinfo>
              <lx:path>Package.TransformMetadata</lx:path>
              <lx:label>Transform Metadata</lx:label>
              <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
            </xs:appinfo>
          </xs:annotation>
          <xs:complexType>
            <xs:sequence>
              <xs:element name="Identifier" minOccurs="0" maxOccurs="unbounded">
                <xs:annotation>
                  <xs:documentation>Identifier metadata required for enabling round trip transformations from a different data model into a LIXI 2.x message and back again.</xs:documentation>
                  <xs:appinfo>
                    <lx:path>Package.TransformMetadata.Identifier</lx:path>
                    <lx:label>Identifier</lx:label>
                    <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                  </xs:appinfo>
                </xs:annotation>
                <xs:complexType>
                  <xs:attribute name="AttributeName" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The name of the attribute in this schema to which this Identifier relates.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.TransformMetadata.Identifier.AttributeName</lx:path>
                        <lx:label>Attribute Name</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="ExternalID" type="stringType" use="optional">
                    <xs:annotation>
                      <xs:documentation>The value of the identifier in the external data model.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.TransformMetadata.Identifier.ExternalID</lx:path>
                        <lx:label>External Identifier</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
                    <xs:annotation>
                      <xs:documentation>System assigned unique identifier</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.TransformMetadata.Identifier.UniqueID</lx:path>
                        <lx:label>Unique ID</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                  <xs:attribute name="x_Context" type="referenceType" use="optional">
                    <xs:annotation>
                      <xs:documentation>A link to a UniqueID field identifying an element within this LIXI 2.x message instance to which this external data model Identifier belongs.</xs:documentation>
                      <xs:appinfo>
                        <lx:path>Package.TransformMetadata.Identifier.x_Context</lx:path>
                        <lx:label>Identifier</lx:label>
                        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
                      </xs:appinfo>
                    </xs:annotation>
                  </xs:attribute>
                </xs:complexType>
              </xs:element>
            </xs:sequence>
          </xs:complexType>
        </xs:element>
      </xs:sequence>
      <xs:attribute name="ProductionData" type="yesNoList" use="optional">
        <xs:annotation>
          <xs:documentation>Indicates whether this file contains production or test data.</xs:documentation>
          <xs:appinfo>
            <lx:path>Package.ProductionData</lx:path>
            <lx:label>Production Data</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:attribute>
      <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
        <xs:annotation>
          <xs:documentation>System assigned unique identifier</xs:documentation>
          <xs:appinfo>
            <lx:path>Package.UniqueID</lx:path>
            <lx:label>Unique ID</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:attribute>
    </xs:complexType>
  </xs:element>
  <xs:complexType name="addressType">
    <xs:annotation>
      <xs:documentation>addressType</xs:documentation>
      <xs:appinfo>
        <lx:path>addressType</lx:path>
        <lx:label>addressType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="DeliveryPoint" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>Details for assisting in more accurately locating and making deliveries to the address.</xs:documentation>
          <xs:appinfo>
            <lx:path>addressType.DeliveryPoint</lx:path>
            <lx:label>Delivery Point</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:attribute name="Identifier" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>Uniquely identifies every delivery point in the Postal Address File (PAF) or equivalent in other countries.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.DeliveryPoint.Identifier</lx:path>
                <lx:label>Identifier</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="IdentifierBarcode" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The barcode encoded version of the Delivery Point Identifier.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.DeliveryPoint.IdentifierBarcode</lx:path>
                <lx:label>Identifier Barcode</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
      <xs:element name="DXBox" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>Wrapper element for the DX Box address</xs:documentation>
          <xs:appinfo>
            <lx:path>addressType.DXBox</lx:path>
            <lx:label>DX Box</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:attribute name="Exchange" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>An exchange box in an authorised DX system</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.DXBox.Exchange</lx:path>
                <lx:label>Exchange</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="Number" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>An identification number, used for tracking the documents from the moment they're collected to the time they arrive at their destination.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.DXBox.Number</lx:path>
                <lx:label>Number</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="Provider" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The name of the company providing DX mailing services</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.DXBox.Provider</lx:path>
                <lx:label>Provider</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
      <xs:element name="NonStandard" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>Wrapper element for the non standard address</xs:documentation>
          <xs:appinfo>
            <lx:path>addressType.NonStandard</lx:path>
            <lx:label>Non Standard</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:attribute name="Line1" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>Line 1 of the address.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.NonStandard.Line1</lx:path>
                <lx:label>Line 1</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="Line2" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>Line 2 of the address.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.NonStandard.Line2</lx:path>
                <lx:label>Line 2</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
      <xs:element name="POBox" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>Wrapper element for the standard PO Box address</xs:documentation>
          <xs:appinfo>
            <lx:path>addressType.POBox</lx:path>
            <lx:label>PO Box</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:attribute name="Exchange" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>An exchange box in an authorised PO system</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.POBox.Exchange</lx:path>
                <lx:label>Exchange</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="Number" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>An identification number, forming part of an address, for the channel of postal delivery.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.POBox.Number</lx:path>
                <lx:label>Number</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="Type" type="poBoxTypeList" use="optional">
            <xs:annotation>
              <xs:documentation>The specification of the identification of a postal delivery service such as General Post Office Box, Community Mail Bag, etc. to clearly distinguish it from another when applicable.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.POBox.Type</lx:path>
                <lx:label>Type</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
      <xs:element name="Standard" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>Wrapper element for the standard address</xs:documentation>
          <xs:appinfo>
            <lx:path>addressType.Standard</lx:path>
            <lx:label>Standard</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:attribute name="BuildingName" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The full name used to identify the physical building or property. Usually this information is not abbreviated. Should include any reference to a wing or other components of a building complex, if applicable.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.Standard.BuildingName</lx:path>
                <lx:label>Building Name</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="Level" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>Descriptors used to identify the floor or level of a multi-storey building or complex.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.Standard.Level</lx:path>
                <lx:label>Level</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="LevelType" type="levelTypeList" use="optional">
            <xs:annotation>
              <xs:documentation>The building level type of a standard address.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.Standard.LevelType</lx:path>
                <lx:label>Level Type</lx:label>
                <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="LotSection" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The lot or section number of the property. Usually only used until a street number is assigned.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.Standard.LotSection</lx:path>
                <lx:label>Lot Section</lx:label>
                <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="StreetName" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The full street name used to identify the street location of the property.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.Standard.StreetName</lx:path>
                <lx:label>Street Name</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="StreetNumber" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>To record the numeric or numeric/alpha reference number of a house or property.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.Standard.StreetNumber</lx:path>
                <lx:label>Street Number</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="StreetSuffix" type="streetSuffixList" use="optional">
            <xs:annotation>
              <xs:documentation>A suffix used to further describe the part of the address that offers directional information. For addresses in New Zealand, this attribute is used for the field referred to as street direction.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.Standard.StreetSuffix</lx:path>
                <lx:label>Street Suffix</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="StreetType" type="streetTypeList" use="optional">
            <xs:annotation>
              <xs:documentation>Used to identify the thoroughfare type</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.Standard.StreetType</lx:path>
                <lx:label>Street Type</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="ToStreetNumber" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The upper bound of a range given for the street number of a house or property.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.Standard.ToStreetNumber</lx:path>
                <lx:label>To Street Number</lx:label>
                <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="ToUnitNumber" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The upper bound of a range given for the unit number of a house or property.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.Standard.ToUnitNumber</lx:path>
                <lx:label>To Unit Number</lx:label>
                <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="Unit" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The specification of the type of a separately identifiable portion within a building complex or marina with its associated number or identifier to clearly distinguish it from another. Can either be depicted by numerals or alpha characters, or a mixture of both.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.Standard.Unit</lx:path>
                <lx:label>Unit</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="UnitType" type="unitTypeList" use="optional">
            <xs:annotation>
              <xs:documentation>The type of unit.</xs:documentation>
              <xs:appinfo>
                <lx:path>addressType.Standard.UnitType</lx:path>
                <lx:label>UnitType</lx:label>
                <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
    </xs:sequence>
    <xs:attribute name="AustralianPostCode" type="auPostCodePattern" use="optional">
      <xs:annotation>
        <xs:documentation>A four digit numeric descriptor for a postal delivery area, aligned with placename, suburb or locality and in some circumstances a unique Postal Delivery Type.</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.AustralianPostCode</lx:path>
          <lx:label>Australian Post Code</lx:label>
          <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="AustralianState" type="auStateList" use="optional">
      <xs:annotation>
        <xs:documentation>The defined State or Territory in Australia (in abbreviated format) that the specific placename/address is located.</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.AustralianState</lx:path>
          <lx:label>Australian State</lx:label>
          <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="City" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>The name of the town, city or island the delivery point is located in or associated with. Australian addresses do not have a City as per Australia Post guidelines. According to New Zealand Post Address Standards, the Town/City is a mandatory element of all addresses, i.e. urban, rural and delivery service addresses.</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.City</lx:path>
          <lx:label>Town/City</lx:label>
          <li:transactions>CAL,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="Country" type="countryCodeList" use="optional">
      <xs:annotation>
        <xs:documentation>The destination country of the address</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.Country</lx:path>
          <lx:label>Country</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="DataSource" type="dataSourceList" use="optional">
      <xs:annotation>
        <xs:documentation>The source of data used in searching information about an address.</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.DataSource</lx:path>
          <lx:label>Data Source</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="GNAF_ID" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>Geocoded National Address File. A G-NAF ID is a unique combination of address detail used to identify each property in Australia</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.GNAF_ID</lx:path>
          <lx:label>GNAF_ID</lx:label>
          <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="Latitude" type="decimalType" use="optional">
      <xs:annotation>
        <xs:documentation>Latitude is a geographic coordinate that specifies the north-south position of a point on the Earth's surface</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.Latitude</lx:path>
          <lx:label>Latitude</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="LGAName" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>A local government area (LGA) is an administrative division of a country that a local government is responsible for. The size of an LGA varies by country but it is generally a subdivision of a state, province, division, or territory.</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.LGAName</lx:path>
          <lx:label>LGA Name</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="Longitude" type="decimalType" use="optional">
      <xs:annotation>
        <xs:documentation>Longitude is a geographic coordinate that specifies the east-west position of a point on the Earth's surface</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.Longitude</lx:path>
          <lx:label>Longitude</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="NewZealandPostCode" type="nzPostCodePattern" use="optional">
      <xs:annotation>
        <xs:documentation>A four digit numeric descriptor for a postal delivery area, aligned with placename, suburb or locality and in some circumstances a unique Postal Delivery Type.</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.NewZealandPostCode</lx:path>
          <lx:label>New Zealand Post Code</lx:label>
          <li:transactions>CNZ</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="OverseasPostCode" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>A series of letters and/or digits for a postal delivery area overseas.</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.OverseasPostCode</lx:path>
          <lx:label>Overseas Post Code</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="OverseasState" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>The State, Province or Territory overseas that the specific placename/address is located.</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.OverseasState</lx:path>
          <lx:label>Overseas State</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="SequenceNumber" type="integerType" use="optional">
      <xs:annotation>
        <xs:documentation>The order of this address within the sequence.</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.SequenceNumber</lx:path>
          <lx:label>Sequence Number</lx:label>
          <li:transactions>CAL,DAS,CNZ</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="Suburb" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>The full name of the placename or Post Office of delivery containing the specific address, which may include a Delivery Centre or a Business Centre.</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.Suburb</lx:path>
          <lx:label>Suburb</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="TBAAddress" type="yesNoList" use="optional">
      <xs:annotation>
        <xs:documentation>For Pre-Approval loans, the address of the property is not required.</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.TBAAddress</lx:path>
          <lx:label>TBA Address</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="Type" type="addressTypeList" use="optional">
      <xs:annotation>
        <xs:documentation>Determines if the address is Standard, Non Standard, a DX Box or a PO BOX</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.Type</lx:path>
          <lx:label>Type</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
      <xs:annotation>
        <xs:documentation>Unique identification assigned to this address.</xs:documentation>
        <xs:appinfo>
          <lx:path>addressType.UniqueID</lx:path>
          <lx:label>Unique ID</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="amountInForeignCurrencyType">
    <xs:annotation>
      <xs:documentation>amountInForeignCurrencyType</xs:documentation>
      <xs:appinfo>
        <lx:path>amountInForeignCurrencyType</lx:path>
        <lx:label>amountInForeignCurrencyType</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:attribute name="Amount" type="currencyType" use="optional">
      <xs:annotation>
        <xs:documentation>The amount in units of the foreign currency.</xs:documentation>
        <xs:appinfo>
          <lx:path>amountInForeignCurrencyType.Amount</lx:path>
          <lx:label>Amount</lx:label>
          <li:transactions>CAL</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="CurrencyCode" type="currencyCodeList" use="optional">
      <xs:annotation>
        <xs:documentation>The currency unit used to express the amount.</xs:documentation>
        <xs:appinfo>
          <lx:path>amountInForeignCurrencyType.CurrencyCode</lx:path>
          <lx:label>Currency Code</lx:label>
          <li:transactions>CAL</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="ExchangeDateTime" type="dateTimeType" use="optional">
      <xs:annotation>
        <xs:documentation>The date time at which the Australian dollar (AUD) value of the amount in a foreign currency was calculated.</xs:documentation>
        <xs:appinfo>
          <lx:path>amountInForeignCurrencyType.ExchangeDateTime</lx:path>
          <lx:label>Exchange Date Time</lx:label>
          <li:transactions>CAL</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="ExchangeRate" type="percentType" use="optional">
      <xs:annotation>
        <xs:documentation>The exchange rate that was used to determine the Australian dollar (AUD) value of the amount in a foreign currency.</xs:documentation>
        <xs:appinfo>
          <lx:path>amountInForeignCurrencyType.ExchangeRate</lx:path>
          <lx:label>Exchange Rate</lx:label>
          <li:transactions>CAL</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="dealingNumberType">
    <xs:annotation>
      <xs:documentation>dealingNumberType</xs:documentation>
      <xs:appinfo>
        <lx:path>dealingNumberType</lx:path>
        <lx:label>dealingNumberType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:attribute name="AustralianState" type="auStateList" use="optional">
      <xs:annotation>
        <xs:documentation>The defined State or Territory in Australia (in abbreviated format) that the specific Dealing is registered in</xs:documentation>
        <xs:appinfo>
          <lx:path>dealingNumberType.AustralianState</lx:path>
          <lx:label>Australian State</lx:label>
          <li:transactions>CAL,CDA,DAS</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="BookNumber" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>Indicates the book number associated with the Dealing</xs:documentation>
        <xs:appinfo>
          <lx:path>dealingNumberType.BookNumber</lx:path>
          <lx:label>Book Number</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="RegisteredNumber" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>Indicates the registered number associated with the Dealing</xs:documentation>
        <xs:appinfo>
          <lx:path>dealingNumberType.RegisteredNumber</lx:path>
          <lx:label>Registered Number</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
      <xs:annotation>
        <xs:documentation>Unique Identifier for this Dealing Number Element</xs:documentation>
        <xs:appinfo>
          <lx:path>dealingNumberType.UniqueID</lx:path>
          <lx:label>Unique ID</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="documentationInstructionsType">
    <xs:annotation>
      <xs:documentation>Instructions on sending documentation to a nominated authority.</xs:documentation>
      <xs:appinfo>
        <lx:path>documentationInstructionsType</lx:path>
        <lx:label>Documentation Instructions Type</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:attribute name="Method" type="documentationInstructionsMethodList" use="optional">
      <xs:annotation>
        <xs:documentation>The method by which the documentation will be sent to the nominated authority.</xs:documentation>
        <xs:appinfo>
          <lx:path>documentationInstructionsType.Method</lx:path>
          <lx:label>Method</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="SendDocumentsTo" type="sendDocumentsToList" use="optional">
      <xs:annotation>
        <xs:documentation>The type of nominated authority to receive the documentation.</xs:documentation>
        <xs:appinfo>
          <lx:path>documentationInstructionsType.SendDocumentsTo</lx:path>
          <lx:label>Send Documents To</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="x_DeliverTo" type="referenceType" use="optional">
      <xs:annotation>
        <xs:documentation>A cross reference to the location for delivery (an Address element).</xs:documentation>
        <xs:appinfo>
          <lx:path>documentationInstructionsType.x_DeliverTo</lx:path>
          <lx:label>Deliver To</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          <lx:references>
            <lx:target>Package.Content.Application.Address</lx:target>
          </lx:references>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="x_NominatedAuthority" type="referenceType" use="optional">
      <xs:annotation>
        <xs:documentation>The nominated authority to receive the documentation.</xs:documentation>
        <xs:appinfo>
          <lx:path>documentationInstructionsType.x_NominatedAuthority</lx:path>
          <lx:label>Nominated Authority</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          <lx:references>
            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
            <lx:target>Package.Content.Application.RelatedCompany</lx:target>
            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
            <lx:target>Package.Content.Application.TrustApplicant</lx:target>
          </lx:references>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="durationType">
    <xs:annotation>
      <xs:documentation>A complex type used for defining durations in terms of units and length of time (as a multiple of the units).</xs:documentation>
      <xs:appinfo>
        <lx:path>durationType</lx:path>
        <lx:label>Duration Type</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:attribute name="Length" type="integerType" use="optional">
      <xs:annotation>
        <xs:documentation>The length of the duration as a multiple of the units.</xs:documentation>
        <xs:appinfo>
          <lx:path>durationType.Length</lx:path>
          <lx:label>Length</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="Units" type="durationUnitsList" use="optional">
      <xs:annotation>
        <xs:documentation>The units used to define the duration.</xs:documentation>
        <xs:appinfo>
          <lx:path>durationType.Units</lx:path>
          <lx:label>Units</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="financialAccountType">
    <xs:annotation>
      <xs:documentation>financialAccountType</xs:documentation>
      <xs:appinfo>
        <lx:path>financialAccountType</lx:path>
        <lx:label>financialAccountType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="BranchDomicile" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>The branch where the account should be domiciled.</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAccountType.BranchDomicile</lx:path>
            <lx:label>BranchDomicile</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:attribute name="InternalName" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The internal name given to a branch location by a lender or creditor may be captured where one exists.</xs:documentation>
              <xs:appinfo>
                <lx:path>financialAccountType.BranchDomicile.InternalName</lx:path>
                <lx:label>Internal Name</lx:label>
                <li:transactions>CAL,CDA,DAS,SVC,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="InternalNumber" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>An internal identification number given to a branch location by a lender or creditor may be captured where one exists.</xs:documentation>
              <xs:appinfo>
                <lx:path>financialAccountType.BranchDomicile.InternalNumber</lx:path>
                <lx:label>Internal Number</lx:label>
                <li:transactions>CAL,CDA,DAS,SVC,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
      <xs:element name="Status" minOccurs="0" maxOccurs="unbounded">
        <xs:annotation>
          <xs:documentation>The status the account is in according to the origination system for the given dates. The status may have been obtained some time in the past from a source system such as a core banking system.</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAccountType.Status</lx:path>
            <lx:label>Status</lx:label>
            <li:transactions>CDA,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:attribute name="EstablishedDate" type="dateTimeType" use="optional">
            <xs:annotation>
              <xs:documentation>The date that the source system established the status for the account.</xs:documentation>
              <xs:appinfo>
                <lx:path>financialAccountType.Status.EstablishedDate</lx:path>
                <lx:label>Established Date</lx:label>
                <li:transactions>CDA,SVC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="LastRetrievedDate" type="dateTimeType" use="optional">
            <xs:annotation>
              <xs:documentation>The date that the status was last retrieved from the source system.</xs:documentation>
              <xs:appinfo>
                <lx:path>financialAccountType.Status.LastRetrievedDate</lx:path>
                <lx:label>Last Retrieved Date</lx:label>
                <li:transactions>CDA,SVC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="Name" type="financialAccountStatusList" use="optional">
            <xs:annotation>
              <xs:documentation>The status of the account, such as active, in arrears or closed.</xs:documentation>
              <xs:appinfo>
                <lx:path>financialAccountType.Status.Name</lx:path>
                <lx:label>Name</lx:label>
                <li:transactions>CDA,SVC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
    </xs:sequence>
    <xs:attribute name="AccountName" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>The free form name assigned to the account.</xs:documentation>
        <xs:appinfo>
          <lx:path>financialAccountType.AccountName</lx:path>
          <lx:label>Account Name</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="AccountNumber" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>The unique number assigned to this account. (typically by the financial institution)</xs:documentation>
        <xs:appinfo>
          <lx:path>financialAccountType.AccountNumber</lx:path>
          <lx:label>Account Number</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="AccountTypeName" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>The lender specific product name that describes the account.</xs:documentation>
        <xs:appinfo>
          <lx:path>financialAccountType.AccountTypeName</lx:path>
          <lx:label>Account Type Name</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="BSB" type="bsbPattern" use="optional">
      <xs:annotation>
        <xs:documentation>The BSB number identifying where this account is held.</xs:documentation>
        <xs:appinfo>
          <lx:path>financialAccountType.BSB</lx:path>
          <lx:label>BSB</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="FinancialInstitution" type="financialInstitutionList" use="optional">
      <xs:annotation>
        <xs:documentation>The financial institution where the account is held.</xs:documentation>
        <xs:appinfo>
          <lx:path>financialAccountType.FinancialInstitution</lx:path>
          <lx:label>Financial Institution</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="OtherFIName" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>For a financial institution not listed, the text name of the FI.</xs:documentation>
        <xs:appinfo>
          <lx:path>financialAccountType.OtherFIName</lx:path>
          <lx:label>Other FI Name</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="foreignTaxAssociationType">
    <xs:annotation>
      <xs:documentation>foreignTaxAssociationType</xs:documentation>
      <xs:appinfo>
        <lx:path>foreignTaxAssociationType</lx:path>
        <lx:label>foreignTaxAssociationType</lx:label>
        <li:transactions>CAL,CDA,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="Detail" minOccurs="0" maxOccurs="unbounded">
        <xs:annotation>
          <xs:documentation>Carries foreign tax association information for sharing amongst OECD nations under the common reporting standard.</xs:documentation>
          <xs:appinfo>
            <lx:path>foreignTaxAssociationType.Detail</lx:path>
            <lx:label>Detail</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:attribute name="Country" type="countryCodeList" use="optional">
            <xs:annotation>
              <xs:documentation>Captures the country of the foreign tax association.</xs:documentation>
              <xs:appinfo>
                <lx:path>foreignTaxAssociationType.Detail.Country</lx:path>
                <lx:label>Country</lx:label>
                <li:transactions>CAL,CDA,DAS</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="Status" type="foreignTaxAssociationStatusList" use="optional">
            <xs:annotation>
              <xs:documentation>The status of the person's knowledge of this foreign tax association.</xs:documentation>
              <xs:appinfo>
                <lx:path>foreignTaxAssociationType.Detail.Status</lx:path>
                <lx:label>Status</lx:label>
                <li:transactions>CAL,CDA,DAS</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="TaxIdentificationNumber" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>Captures the tax identification number of the foreign tax association in the format specified by the country the association is with.</xs:documentation>
              <xs:appinfo>
                <lx:path>foreignTaxAssociationType.Detail.TaxIdentificationNumber</lx:path>
                <lx:label>Tax Identification Number</lx:label>
                <li:transactions>CAL,CDA,DAS</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="TINNotProvidedReason" type="tinNotProvidedReasonList" use="optional">
            <xs:annotation>
              <xs:documentation>If the Tax Identification Number or equivalent is not provided, this captures the reason the TIN was not provided.</xs:documentation>
              <xs:appinfo>
                <lx:path>foreignTaxAssociationType.Detail.TINNotProvidedReason</lx:path>
                <lx:label>TIN Not Provided Reason</lx:label>
                <li:transactions>CAL,CDA,DAS</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="TINNotProvidedReasonDetail" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>If the entity is required to provide a TIN, can be issued a TIN, but has been unable to obtain a TIN, further details should be provided here.</xs:documentation>
              <xs:appinfo>
                <lx:path>foreignTaxAssociationType.Detail.TINNotProvidedReasonDetail</lx:path>
                <lx:label>TIN Not Provided Reason Detail</lx:label>
                <li:transactions>CAL,CDA,DAS</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
    </xs:sequence>
    <xs:attribute name="SelfCertificationDeclaration" type="yesNoList" use="optional">
      <xs:annotation>
        <xs:documentation>Indicates that the person providing the details of the foreign tax association on behalf of the entity has completed the declaration of self certification.</xs:documentation>
        <xs:appinfo>
          <lx:path>foreignTaxAssociationType.SelfCertificationDeclaration</lx:path>
          <lx:label>SelfCertificationDeclaration</lx:label>
          <li:transactions>CAL,CDA,DAS</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="Status" type="foreignTaxAssociationStatusList" use="optional">
      <xs:annotation>
        <xs:documentation>Captures the status of the person's knowledge of any foreign tax associations they may have.</xs:documentation>
        <xs:appinfo>
          <lx:path>foreignTaxAssociationType.Status</lx:path>
          <lx:label>Status</lx:label>
          <li:transactions>CAL,CDA,DAS</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="fundsDisbursementType">
    <xs:annotation>
      <xs:documentation>Specification of how funds are to be distributed.</xs:documentation>
      <xs:appinfo>
        <lx:path>fundsDisbursementType</lx:path>
        <lx:label>fundsDisbursementType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="AccountNumber" type="financialAccountType" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>The account number to receive the specified funds at settlement.</xs:documentation>
          <xs:appinfo>
            <lx:path>fundsDisbursementType.AccountNumber</lx:path>
            <lx:label>Account Number</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:element>
      <xs:element name="BillerDetails" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>The BPAY biller details.</xs:documentation>
          <xs:appinfo>
            <lx:path>fundsDisbursementType.BillerDetails</lx:path>
            <lx:label>Biller Details</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:attribute name="BillerCode" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The BPAY biller code.</xs:documentation>
              <xs:appinfo>
                <lx:path>fundsDisbursementType.BillerDetails.BillerCode</lx:path>
                <lx:label>Biller Code</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="BillerName" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The BPAY biller name.</xs:documentation>
              <xs:appinfo>
                <lx:path>fundsDisbursementType.BillerDetails.BillerName</lx:path>
                <lx:label>Biller Name</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="CRN" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The BPAY customer reference number (CRN).</xs:documentation>
              <xs:appinfo>
                <lx:path>fundsDisbursementType.BillerDetails.CRN</lx:path>
                <lx:label>CRN</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
      <xs:element name="PersonName" type="personNameType" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>The name of the person receiving the disbursed funds.</xs:documentation>
          <xs:appinfo>
            <lx:path>fundsDisbursementType.PersonName</lx:path>
            <lx:label>Person Name</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:element>
    </xs:sequence>
    <xs:attribute name="Amount" type="currencyType" use="optional">
      <xs:annotation>
        <xs:documentation>Indicates the amount of funds to be disbursed at settlement.</xs:documentation>
        <xs:appinfo>
          <lx:path>fundsDisbursementType.Amount</lx:path>
          <lx:label>Amount</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="CompanyName" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>The registered name of the company receiving the disbursed funds.</xs:documentation>
        <xs:appinfo>
          <lx:path>fundsDisbursementType.CompanyName</lx:path>
          <lx:label>Company Name</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="Method" type="surplusFundsDisbursementMethodList" use="optional">
      <xs:annotation>
        <xs:documentation>Specification of the method of distributing the funds.</xs:documentation>
        <xs:appinfo>
          <lx:path>fundsDisbursementType.Method</lx:path>
          <lx:label>Method</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="PurposeOfFunds" type="purposeOfFundsList" use="optional">
      <xs:annotation>
        <xs:documentation>The purpose for which the disbursed funds will be used.</xs:documentation>
        <xs:appinfo>
          <lx:path>fundsDisbursementType.PurposeOfFunds</lx:path>
          <lx:label>Purpose Of Funds</lx:label>
          <li:transactions>ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="PurposeOfFundsDescription" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>A description of the purpose if none of the purposes specified in the list of purposes adequately described the true purpose.</xs:documentation>
        <xs:appinfo>
          <lx:path>fundsDisbursementType.PurposeOfFundsDescription</lx:path>
          <lx:label>Purpose Of Funds Description</lx:label>
          <li:transactions>ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
      <xs:annotation>
        <xs:documentation>Unique Identifier for this Funds Disbursement Element.</xs:documentation>
        <xs:appinfo>
          <lx:path>fundsDisbursementType.UniqueID</lx:path>
          <lx:label>Unique ID</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="x_Account" type="referenceType" use="optional">
      <xs:annotation>
        <xs:documentation>The loan account or transaction account to receive the funds at settlement.</xs:documentation>
        <xs:appinfo>
          <lx:path>fundsDisbursementType.x_Account</lx:path>
          <lx:label>Account</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          <lx:references>
            <lx:target>Package.Content.Application.DepositAccountDetails</lx:target>
            <lx:target>Package.Content.Application.Liability</lx:target>
            <lx:target>Package.Content.Application.LoanDetails</lx:target>
            <lx:target>Package.Content.Application.NonRealEstateAsset</lx:target>
          </lx:references>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="x_Address" type="referenceType" use="optional">
      <xs:annotation>
        <xs:documentation>A cross reference to the address of the party receiving the disbursed funds.</xs:documentation>
        <xs:appinfo>
          <lx:path>fundsDisbursementType.x_Address</lx:path>
          <lx:label>Address</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          <lx:references>
            <lx:target>Package.Content.Application.Address</lx:target>
          </lx:references>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="percentOwnedType">
    <xs:annotation>
      <xs:documentation>percentOwnedType</xs:documentation>
      <xs:appinfo>
        <lx:path>percentOwnedType</lx:path>
        <lx:label>percentOwnedType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="Owner" minOccurs="0" maxOccurs="unbounded">
        <xs:annotation>
          <xs:documentation>Container that associates a reference to an owner or other responsible party with the percentage of the item that they own or are responsible for.</xs:documentation>
          <xs:appinfo>
            <lx:path>percentOwnedType.Owner</lx:path>
            <lx:label>Owner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:attribute name="Percent" type="percentType" use="optional">
            <xs:annotation>
              <xs:documentation>Captures the percentage of the item that referenced party owns or is responsible for.</xs:documentation>
              <xs:appinfo>
                <lx:path>percentOwnedType.Owner.Percent</lx:path>
                <lx:label>Percent</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="x_Party" type="referenceType" use="optional">
            <xs:annotation>
              <xs:documentation>Contains a reference to a party that is an owner of the item, or has a similar relationship such as a borrower or beneficiary.</xs:documentation>
              <xs:appinfo>
                <lx:path>percentOwnedType.Owner.x_Party</lx:path>
                <lx:label>Party</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
                <lx:references>
                  <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                  <lx:target>Package.Content.Application.PersonApplicant</lx:target>
                  <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                </lx:references>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
    </xs:sequence>
    <xs:attribute name="Proportions" type="proportionsList" use="optional">
      <xs:annotation>
        <xs:documentation>Flag to indicate whether the ownership proportions are equal across all owners, specified individually for each one, or not specified.</xs:documentation>
        <xs:appinfo>
          <lx:path>percentOwnedType.Proportions</lx:path>
          <lx:label>Proportions</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="personNameType">
    <xs:annotation>
      <xs:documentation>personNameType</xs:documentation>
      <xs:appinfo>
        <lx:path>personNameType</lx:path>
        <lx:label>personNameType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:attribute name="FirstName" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>A person's first name.</xs:documentation>
        <xs:appinfo>
          <lx:path>personNameType.FirstName</lx:path>
          <lx:label>First Name</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="KnownAs" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>The familiar name, nickname or preferred name by which the person wishes to be known.</xs:documentation>
        <xs:appinfo>
          <lx:path>personNameType.KnownAs</lx:path>
          <lx:label>Known As</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="MiddleNames" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>A person's other name. Also referred to as their 'middle name'. For example - where someone's name is Frederick Joseph Gilbert - 'Joseph' would be considered their other name.</xs:documentation>
        <xs:appinfo>
          <lx:path>personNameType.MiddleNames</lx:path>
          <lx:label>Middle Names</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="NameTitle" type="nameTitleList" use="optional">
      <xs:annotation>
        <xs:documentation>The name title this person uses when they are formally addressed should be captured.</xs:documentation>
        <xs:appinfo>
          <lx:path>personNameType.NameTitle</lx:path>
          <lx:label>Name Title</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="OtherNameTitle" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>The alternate or other titles this person uses when they are formally addressed. This can include multiple titles and/or appropriate abbreviations.</xs:documentation>
        <xs:appinfo>
          <lx:path>personNameType.OtherNameTitle</lx:path>
          <lx:label>Other Name Title</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="Surname" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>A person's surname.</xs:documentation>
        <xs:appinfo>
          <lx:path>personNameType.Surname</lx:path>
          <lx:label>Surname</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="phoneType">
    <xs:annotation>
      <xs:documentation>phoneType</xs:documentation>
      <xs:appinfo>
        <lx:path>phoneType</lx:path>
        <lx:label>phoneType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:attribute name="AustralianDialingCode" type="auDialingCodeList" use="optional">
      <xs:annotation>
        <xs:documentation>The Dialing Code prefix for an Australian STD landline phone number. For example, if the applicant lives in Sydney, the Area Code is 02. If the applicant lives in Melbourne, the Area Code is 03.</xs:documentation>
        <xs:appinfo>
          <lx:path>phoneType.AustralianDialingCode</lx:path>
          <lx:label>Australian Dialing Code</lx:label>
          <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="CountryCode" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>Captures the international dialing code for the country in which the fixed phone number is located, e.g. "61" for Australia. Maximum length is 3 digits.</xs:documentation>
        <xs:appinfo>
          <lx:path>phoneType.CountryCode</lx:path>
          <lx:label>Country Code</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="NewZealandDialingCode" type="nzDialingCodeList" use="optional">
      <xs:annotation>
        <xs:documentation>The Area Code prefix for a New Zealand telephone number.</xs:documentation>
        <xs:appinfo>
          <lx:path>phoneType.NewZealandDialingCode</lx:path>
          <lx:label>New Zealand Area Code</lx:label>
          <li:transactions>CNZ</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="Number" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>The local-dial number. For local rate, free call, satellite or similar special non-overseas numbers where the area or dialing code is not otherwise available, the full phone number. Required, allowed once only.</xs:documentation>
        <xs:appinfo>
          <lx:path>phoneType.Number</lx:path>
          <lx:label>Number</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="OverseasDialingCode" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>The Dialing Code prefix for an overseas phone number</xs:documentation>
        <xs:appinfo>
          <lx:path>phoneType.OverseasDialingCode</lx:path>
          <lx:label>Overseas Dialing Code</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="responsibleLendingType">
    <xs:annotation>
      <xs:documentation>Captures information about any significant changes in circumstances that will adversely affect the applicant's or applicants' ability to make contracted loan repayments; the nature of the change, and how they will continue to make repayments.</xs:documentation>
      <xs:appinfo>
        <lx:path>responsibleLendingType</lx:path>
        <lx:label>Responsible Lending</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="Mitigant" minOccurs="0" maxOccurs="unbounded">
        <xs:annotation>
          <xs:documentation>In relation to the significant changes, it must be possible to provide one or more mitigants for the significant changes</xs:documentation>
          <xs:appinfo>
            <lx:path>responsibleLendingType.Mitigant</lx:path>
            <lx:label>Mitigant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:sequence>
            <xs:element name="SaleOfAssets" minOccurs="0" maxOccurs="1">
              <xs:annotation>
                <xs:documentation>Information regarding the sale of assets.</xs:documentation>
                <xs:appinfo>
                  <lx:path>responsibleLendingType.Mitigant.SaleOfAssets</lx:path>
                  <lx:label>Sale Of Assets</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                </xs:appinfo>
              </xs:annotation>
              <xs:complexType>
                <xs:attribute name="LenderHeldSecurity" type="yesNoList" use="optional">
                  <xs:annotation>
                    <xs:documentation>The asset is a security held by the lender.</xs:documentation>
                    <xs:appinfo>
                      <lx:path>responsibleLendingType.Mitigant.SaleOfAssets.LenderHeldSecurity</lx:path>
                      <lx:label>Lender Held Security</lx:label>
                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                    </xs:appinfo>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="LenderHeldSecurityDescription" type="stringType" use="optional">
                  <xs:annotation>
                    <xs:documentation>Details regarding the security asset held by the lender.</xs:documentation>
                    <xs:appinfo>
                      <lx:path>responsibleLendingType.Mitigant.SaleOfAssets.LenderHeldSecurityDescription</lx:path>
                      <lx:label>Lender Held Security Description</lx:label>
                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                    </xs:appinfo>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="OtherAssets" type="yesNoList" use="optional">
                  <xs:annotation>
                    <xs:documentation>The sale of assets refers to sale of other assets.</xs:documentation>
                    <xs:appinfo>
                      <lx:path>responsibleLendingType.Mitigant.SaleOfAssets.OtherAssets</lx:path>
                      <lx:label>Other Assets</lx:label>
                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                    </xs:appinfo>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="OtherAssetsDescription" type="stringType" use="optional">
                  <xs:annotation>
                    <xs:documentation>Details regarding the sales of other assets.</xs:documentation>
                    <xs:appinfo>
                      <lx:path>responsibleLendingType.Mitigant.SaleOfAssets.OtherAssetsDescription</lx:path>
                      <lx:label>Other Assets Description</lx:label>
                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                    </xs:appinfo>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="OtherLenderHeldSecurity" type="yesNoList" use="optional">
                  <xs:annotation>
                    <xs:documentation>The asset is a security held by another lender.</xs:documentation>
                    <xs:appinfo>
                      <lx:path>responsibleLendingType.Mitigant.SaleOfAssets.OtherLenderHeldSecurity</lx:path>
                      <lx:label>Other Lender Held Security</lx:label>
                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                    </xs:appinfo>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="OtherLenderHeldSecurityDescription" type="stringType" use="optional">
                  <xs:annotation>
                    <xs:documentation>Details regarding the security asset held by the other lender.</xs:documentation>
                    <xs:appinfo>
                      <lx:path>responsibleLendingType.Mitigant.SaleOfAssets.OtherLenderHeldSecurityDescription</lx:path>
                      <lx:label>Other Lender Held Security Description</lx:label>
                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                    </xs:appinfo>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="Shares" type="yesNoList" use="optional">
                  <xs:annotation>
                    <xs:documentation>The sale of assets refers to sale of shares.</xs:documentation>
                    <xs:appinfo>
                      <lx:path>responsibleLendingType.Mitigant.SaleOfAssets.Shares</lx:path>
                      <lx:label>Shares</lx:label>
                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                    </xs:appinfo>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="SharesDescription" type="stringType" use="optional">
                  <xs:annotation>
                    <xs:documentation>Details regarding the sales of shares.</xs:documentation>
                    <xs:appinfo>
                      <lx:path>responsibleLendingType.Mitigant.SaleOfAssets.SharesDescription</lx:path>
                      <lx:label>Shares Description</lx:label>
                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                    </xs:appinfo>
                  </xs:annotation>
                </xs:attribute>
              </xs:complexType>
            </xs:element>
            <xs:element name="SavingsOrSuperannuation" minOccurs="0" maxOccurs="1">
              <xs:annotation>
                <xs:documentation>Information regarding the savings or superannuation.</xs:documentation>
                <xs:appinfo>
                  <lx:path>responsibleLendingType.Mitigant.SavingsOrSuperannuation</lx:path>
                  <lx:label>Savings Or Superannuation</lx:label>
                  <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                </xs:appinfo>
              </xs:annotation>
              <xs:complexType>
                <xs:attribute name="LenderHeldSavingsAccount" type="yesNoList" use="optional">
                  <xs:annotation>
                    <xs:documentation>The savings are held in the lender account.</xs:documentation>
                    <xs:appinfo>
                      <lx:path>responsibleLendingType.Mitigant.SavingsOrSuperannuation.LenderHeldSavingsAccount</lx:path>
                      <lx:label>Lender Held Savings Account</lx:label>
                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                    </xs:appinfo>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="LenderHeldSavingsAccountDescription" type="stringType" use="optional">
                  <xs:annotation>
                    <xs:documentation>Information regarding the savings account held by the lender.</xs:documentation>
                    <xs:appinfo>
                      <lx:path>responsibleLendingType.Mitigant.SavingsOrSuperannuation.LenderHeldSavingsAccountDescription</lx:path>
                      <lx:label>Lender Held Savings Account Description</lx:label>
                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                    </xs:appinfo>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="OtherLenderHeldSavingsAccount" type="yesNoList" use="optional">
                  <xs:annotation>
                    <xs:documentation>The savings are held in another financial institution account.</xs:documentation>
                    <xs:appinfo>
                      <lx:path>responsibleLendingType.Mitigant.SavingsOrSuperannuation.OtherLenderHeldSavingsAccount</lx:path>
                      <lx:label>Other Lender Held Savings Account</lx:label>
                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                    </xs:appinfo>
                  </xs:annotation>
                </xs:attribute>
                <xs:attribute name="OtherLenderHeldSavingsAccountDescription" type="stringType" use="optional">
                  <xs:annotation>
                    <xs:documentation>Information regarding the savings account held by another lender.</xs:documentation>
                    <xs:appinfo>
                      <lx:path>responsibleLendingType.Mitigant.SavingsOrSuperannuation.OtherLenderHeldSavingsAccountDescription</lx:path>
                      <lx:label>Other Lender Held Savings Account Description</lx:label>
                      <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
                    </xs:appinfo>
                  </xs:annotation>
                </xs:attribute>
              </xs:complexType>
            </xs:element>
          </xs:sequence>
          <xs:attribute name="Description" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>Captures a free text description of the mitigant factor.</xs:documentation>
              <xs:appinfo>
                <lx:path>responsibleLendingType.Mitigant.Description</lx:path>
                <lx:label>Description</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="Factor" type="mitigantFactorList" use="optional">
            <xs:annotation>
              <xs:documentation>In relation to the significant changes, it must be possible to provide one or more mitigants for the significant changes</xs:documentation>
              <xs:appinfo>
                <lx:path>responsibleLendingType.Mitigant.Factor</lx:path>
                <lx:label>Factor</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
            <xs:annotation>
              <xs:documentation>Unique Identifier for this Mitigant Element</xs:documentation>
              <xs:appinfo>
                <lx:path>responsibleLendingType.Mitigant.UniqueID</lx:path>
                <lx:label>Unique ID</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
      <xs:element name="SignificantChange" minOccurs="0" maxOccurs="unbounded">
        <xs:annotation>
          <xs:documentation>The kind of change the applicant is expecting.</xs:documentation>
          <xs:appinfo>
            <lx:path>responsibleLendingType.SignificantChange</lx:path>
            <lx:label>Significant Change</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:attribute name="Change" type="significantChangeList" use="optional">
            <xs:annotation>
              <xs:documentation>The kind of change the applicant is expecting.</xs:documentation>
              <xs:appinfo>
                <lx:path>responsibleLendingType.SignificantChange.Change</lx:path>
                <lx:label>Change</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="Description" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>Captures a free text description of the change.</xs:documentation>
              <xs:appinfo>
                <lx:path>responsibleLendingType.SignificantChange.Description</lx:path>
                <lx:label>Description</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="EndDate" type="dateType" use="optional">
            <xs:annotation>
              <xs:documentation>The anticipated end date of the change in circumstances</xs:documentation>
              <xs:appinfo>
                <lx:path>responsibleLendingType.SignificantChange.EndDate</lx:path>
                <lx:label>End Date</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="MonthlyFinancialImpact" type="currencyType" use="optional">
            <xs:annotation>
              <xs:documentation>Financial impact per month of the change in circumstances.</xs:documentation>
              <xs:appinfo>
                <lx:path>responsibleLendingType.SignificantChange.MonthlyFinancialImpact</lx:path>
                <lx:label>Monthly Financial Impact</lx:label>
                <li:transactions>CAL</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="StartDate" type="dateType" use="optional">
            <xs:annotation>
              <xs:documentation>The anticipated start date of the change in circumstances</xs:documentation>
              <xs:appinfo>
                <lx:path>responsibleLendingType.SignificantChange.StartDate</lx:path>
                <lx:label>Start Date</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
            <xs:annotation>
              <xs:documentation>Unique Identifier for this Significant Change Element</xs:documentation>
              <xs:appinfo>
                <lx:path>responsibleLendingType.SignificantChange.UniqueID</lx:path>
                <lx:label>Unique ID</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
    </xs:sequence>
    <xs:attribute name="AnticipatedChanges" type="yesNoList" use="optional">
      <xs:annotation>
        <xs:documentation>Each borrower to inform the lender if they are aware of any significant changes in circumstances that will adversely affect their ability to make contracted loan repayments</xs:documentation>
        <xs:appinfo>
          <lx:path>responsibleLendingType.AnticipatedChanges</lx:path>
          <lx:label>Anticipated Changes In Circumstances</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="securityAgreementType">
    <xs:annotation>
      <xs:documentation>Security agreement details.</xs:documentation>
      <xs:appinfo>
        <lx:path>securityAgreementType</lx:path>
        <lx:label>Security Agreement Type</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="TermsAndConditions" minOccurs="0" maxOccurs="unbounded">
        <xs:annotation>
          <xs:documentation>The terms and conditions of an asset used as a security.</xs:documentation>
          <xs:appinfo>
            <lx:path>securityAgreementType.TermsAndConditions</lx:path>
            <lx:label>Terms And Conditions</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:attribute name="TermsCode" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The code used to identify the terms and conditions in the lender's system.</xs:documentation>
              <xs:appinfo>
                <lx:path>securityAgreementType.TermsAndConditions.TermsCode</lx:path>
                <lx:label>Terms Code</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="TermsName" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The name used by the lender for this set of terms and conditions.</xs:documentation>
              <xs:appinfo>
                <lx:path>securityAgreementType.TermsAndConditions.TermsName</lx:path>
                <lx:label>Terms Name</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="TermsVersion" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>The version used by the lender for this set of terms and conditions.</xs:documentation>
              <xs:appinfo>
                <lx:path>securityAgreementType.TermsAndConditions.TermsVersion</lx:path>
                <lx:label>Terms Version</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
    </xs:sequence>
    <xs:attribute name="Held" type="yesNoList" use="optional">
      <xs:annotation>
        <xs:documentation>A security agreement is held for this security.</xs:documentation>
        <xs:appinfo>
          <lx:path>securityAgreementType.Held</lx:path>
          <lx:label>Held</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="ToBeDischarged" type="yesNoList" use="optional">
      <xs:annotation>
        <xs:documentation>The security agreement that is held for this security is to be discharged.</xs:documentation>
        <xs:appinfo>
          <lx:path>securityAgreementType.ToBeDischarged</lx:path>
          <lx:label>To Be Discharged</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="Type" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>The type of security agreement held over the security.</xs:documentation>
        <xs:appinfo>
          <lx:path>securityAgreementType.Type</lx:path>
          <lx:label>Type</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="signatureType">
    <xs:annotation>
      <xs:documentation>A signature for a document.</xs:documentation>
      <xs:appinfo>
        <lx:path>signatureType</lx:path>
        <lx:label>signatureType</lx:label>
        <li:transactions>CAL,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="Capacity" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>The capacity in which the signatory has signed this document.</xs:documentation>
          <xs:appinfo>
            <lx:path>signatureType.Capacity</lx:path>
            <lx:label>Capacity</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:attribute name="Type" type="capacityTypeList" use="optional">
            <xs:annotation>
              <xs:documentation>The type of capacity in which the signatory has signed this document.</xs:documentation>
              <xs:appinfo>
                <lx:path>signatureType.Capacity.Type</lx:path>
                <lx:label>Type</lx:label>
                <li:transactions>CAL,DAS</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="x_ParentEntity" type="referenceType" use="optional">
            <xs:annotation>
              <xs:documentation>A cross reference to the parent entity (a Company Applicant or Trust Applicant element) on whose behalf the signatory has signed this document.</xs:documentation>
              <xs:appinfo>
                <lx:path>signatureType.Capacity.x_ParentEntity</lx:path>
                <lx:label>Parent Entity</lx:label>
                <li:transactions>CAL,DAS</li:transactions>
                <lx:references>
                  <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
                  <lx:target>Package.Content.Application.RelatedCompany</lx:target>
                  <lx:target>Package.Content.Application.TrustApplicant</lx:target>
                </lx:references>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="x_ProductSet" type="referenceType" use="optional">
            <xs:annotation>
              <xs:documentation>A cross reference to the product set (a Product Set element) to which this signature applies.</xs:documentation>
              <xs:appinfo>
                <lx:path>signatureType.Capacity.x_ProductSet</lx:path>
                <lx:label>Product Set</lx:label>
                <li:transactions>CAL,DAS</li:transactions>
                <lx:references>
                  <lx:target>Package.Content.Application.ProductSet</lx:target>
                </lx:references>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
      <xs:element name="Contact" minOccurs="0" maxOccurs="1">
        <xs:annotation>
          <xs:documentation>Contact details for the signatory.</xs:documentation>
          <xs:appinfo>
            <lx:path>signatureType.Contact</lx:path>
            <lx:label>Contact</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:sequence>
            <xs:element name="Mobile" type="phoneType" minOccurs="0" maxOccurs="1">
              <xs:annotation>
                <xs:documentation>Mobile phone number for the signatory.</xs:documentation>
                <xs:appinfo>
                  <lx:path>signatureType.Contact.Mobile</lx:path>
                  <lx:label>Mobile</lx:label>
                  <li:transactions>CAL,DAS</li:transactions>
                </xs:appinfo>
              </xs:annotation>
            </xs:element>
          </xs:sequence>
          <xs:attribute name="Email" type="emailPattern" use="optional">
            <xs:annotation>
              <xs:documentation>Email address for the signatory.</xs:documentation>
              <xs:appinfo>
                <lx:path>signatureType.Contact.Email</lx:path>
                <lx:label>Email</lx:label>
                <li:transactions>CAL,DAS</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
    </xs:sequence>
    <xs:attribute name="Date" type="dateType" use="optional">
      <xs:annotation>
        <xs:documentation>The date on which the application was signed.</xs:documentation>
        <xs:appinfo>
          <lx:path>signatureType.Date</lx:path>
          <lx:label>Date</lx:label>
          <li:transactions>CAL,DAS</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="ElectronicSignature" type="yesNoList" use="optional">
      <xs:annotation>
        <xs:documentation>The signatory has signed the document with an electronic signature.</xs:documentation>
        <xs:appinfo>
          <lx:path>signatureType.ElectronicSignature</lx:path>
          <lx:label>Electronic Signature</lx:label>
          <li:transactions>CAL,DAS</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="Rank" type="integerType" use="optional">
      <xs:annotation>
        <xs:documentation>The rank of the signatory for this document, which can be used to determine the access (read-only or read-write for example) that the signatory has in providing input.</xs:documentation>
        <xs:appinfo>
          <lx:path>signatureType.Rank</lx:path>
          <lx:label>Rank</lx:label>
          <li:transactions>CAL,DAS</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
      <xs:annotation>
        <xs:documentation>System assigned unique identifier.</xs:documentation>
        <xs:appinfo>
          <lx:path>signatureType.UniqueID</lx:path>
          <lx:label>Unique ID</lx:label>
          <li:transactions>CAL,DAS</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="x_Signatory" type="referenceType" use="optional">
      <xs:annotation>
        <xs:documentation>A cross reference to the party (a Person Applicant element or Related Person element) that is acting as signatory.</xs:documentation>
        <xs:appinfo>
          <lx:path>signatureType.x_Signatory</lx:path>
          <lx:label>Signatory</lx:label>
          <li:transactions>CAL,DAS</li:transactions>
          <lx:references>
            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
          </lx:references>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="taxDeclarationDetailsType">
    <xs:annotation>
      <xs:documentation>Carries details relating to the local tax declarations that have been or are being made to the financial institution.</xs:documentation>
      <xs:appinfo>
        <lx:path>taxDeclarationDetailsType</lx:path>
        <lx:label>Tax Declaration Details Type</lx:label>
        <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:attribute name="AlreadyProvided" type="yesNoList" use="optional">
      <xs:annotation>
        <xs:documentation>Captures whether local tax details have already been provided to the financial institution.</xs:documentation>
        <xs:appinfo>
          <lx:path>taxDeclarationDetailsType.AlreadyProvided</lx:path>
          <lx:label>Already Provided</lx:label>
          <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="ExemptionCategory" type="taxDeclarationExemptionCategoryList" use="optional">
      <xs:annotation>
        <xs:documentation>The category the exemption to providing a local tax identifier falls into.</xs:documentation>
        <xs:appinfo>
          <lx:path>taxDeclarationDetailsType.ExemptionCategory</lx:path>
          <lx:label>Exemption Category</lx:label>
          <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="Undeclared" type="yesNoList" use="optional">
      <xs:annotation>
        <xs:documentation>Whether the party has not declared the tax identifier to the institution that has requested it.</xs:documentation>
        <xs:appinfo>
          <lx:path>taxDeclarationDetailsType.Undeclared</lx:path>
          <lx:label>Undeclared</lx:label>
          <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:complexType name="vendorTaxInvoiceType">
    <xs:annotation>
      <xs:documentation>vendorTaxInvoiceType</xs:documentation>
      <xs:appinfo>
        <lx:path>vendorTaxInvoiceType</lx:path>
        <lx:label>vendorTaxInvoiceType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="Asset" minOccurs="0" maxOccurs="unbounded">
        <xs:annotation>
          <xs:documentation>Defines an asset involved in the transaction.</xs:documentation>
          <xs:appinfo>
            <lx:path>vendorTaxInvoiceType.Asset</lx:path>
            <lx:label>Asset</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
        <xs:complexType>
          <xs:attribute name="AssetDescription" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>A free text description of the Asset.</xs:documentation>
              <xs:appinfo>
                <lx:path>vendorTaxInvoiceType.Asset.AssetDescription</lx:path>
                <lx:label>Asset Description</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="DeliveryDate" type="dateType" use="optional">
            <xs:annotation>
              <xs:documentation>Delivery date of the asset.</xs:documentation>
              <xs:appinfo>
                <lx:path>vendorTaxInvoiceType.Asset.DeliveryDate</lx:path>
                <lx:label>Delivery Date</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="DepositPaid" type="currencyType" use="optional">
            <xs:annotation>
              <xs:documentation>The dollar amount of deposit paid for the asset.</xs:documentation>
              <xs:appinfo>
                <lx:path>vendorTaxInvoiceType.Asset.DepositPaid</lx:path>
                <lx:label>Deposit Paid</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="FinalPosition" type="currencyType" use="optional">
            <xs:annotation>
              <xs:documentation>The final position on the asset in dollars.</xs:documentation>
              <xs:appinfo>
                <lx:path>vendorTaxInvoiceType.Asset.FinalPosition</lx:path>
                <lx:label>Final Position</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="GoodsDelivery" type="goodsDeliveryList" use="optional">
            <xs:annotation>
              <xs:documentation>Flag to indicate whether the goods have been delivered, and if not, when.</xs:documentation>
              <xs:appinfo>
                <lx:path>vendorTaxInvoiceType.Asset.GoodsDelivery</lx:path>
                <lx:label>Goods Delivery</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="GSTComponent" type="currencyType" use="optional">
            <xs:annotation>
              <xs:documentation>Dollar amount of the GST component.</xs:documentation>
              <xs:appinfo>
                <lx:path>vendorTaxInvoiceType.Asset.GSTComponent</lx:path>
                <lx:label>GST Component</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="NonTaxableAmount" type="currencyType" use="optional">
            <xs:annotation>
              <xs:documentation>Other costs that do not attract GST, such as fee component, luxury car tax, GST-free components etc. Cost of Goods Sold Is needed as some on-road costs do not attract GST, including fee component, tax components, luxury tax, etc.</xs:documentation>
              <xs:appinfo>
                <lx:path>vendorTaxInvoiceType.Asset.NonTaxableAmount</lx:path>
                <lx:label>Non Taxable Amount</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="OnCosts" type="currencyType" use="optional">
            <xs:annotation>
              <xs:documentation>Non taxable on-costs. For a motor vehicle on road costs include the total of registration, CTP, stamp duty and other similar costs.</xs:documentation>
              <xs:appinfo>
                <lx:path>vendorTaxInvoiceType.Asset.OnCosts</lx:path>
                <lx:label>On Costs</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="Payout" type="currencyType" use="optional">
            <xs:annotation>
              <xs:documentation>Dollar amount of the payout.</xs:documentation>
              <xs:appinfo>
                <lx:path>vendorTaxInvoiceType.Asset.Payout</lx:path>
                <lx:label>Payout</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="Refund" type="currencyType" use="optional">
            <xs:annotation>
              <xs:documentation>Dollar amount of the refund.</xs:documentation>
              <xs:appinfo>
                <lx:path>vendorTaxInvoiceType.Asset.Refund</lx:path>
                <lx:label>Refund</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="TaxableAmountOfAsset" type="currencyType" use="optional">
            <xs:annotation>
              <xs:documentation>Base purchase price, ex tax. May include freight or delivery.</xs:documentation>
              <xs:appinfo>
                <lx:path>vendorTaxInvoiceType.Asset.TaxableAmountOfAsset</lx:path>
                <lx:label>Taxable Amount Of Asset</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="TotalCostOfAsset" type="currencyType" use="optional">
            <xs:annotation>
              <xs:documentation>Dollar amount of the total cost of the asset.</xs:documentation>
              <xs:appinfo>
                <lx:path>vendorTaxInvoiceType.Asset.TotalCostOfAsset</lx:path>
                <lx:label>Total Cost Of Asset</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="TradeIn" type="currencyType" use="optional">
            <xs:annotation>
              <xs:documentation>Dollar amount of the trade-in.</xs:documentation>
              <xs:appinfo>
                <lx:path>vendorTaxInvoiceType.Asset.TradeIn</lx:path>
                <lx:label>Trade In</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
          <xs:attribute name="UniqueAssetIdentifier" type="stringType" use="optional">
            <xs:annotation>
              <xs:documentation>A unique Identifier for this asset.</xs:documentation>
              <xs:appinfo>
                <lx:path>vendorTaxInvoiceType.Asset.UniqueAssetIdentifier</lx:path>
                <lx:label>Unique Asset Identifier</lx:label>
                <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
              </xs:appinfo>
            </xs:annotation>
          </xs:attribute>
        </xs:complexType>
      </xs:element>
    </xs:sequence>
    <xs:attribute name="InvoiceNumber" type="stringType" use="optional">
      <xs:annotation>
        <xs:documentation>The invoice number as supplied by the vendor.</xs:documentation>
        <xs:appinfo>
          <lx:path>vendorTaxInvoiceType.InvoiceNumber</lx:path>
          <lx:label>Invoice Number</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="SequenceNumber" type="integerType" use="optional">
      <xs:annotation>
        <xs:documentation>The order of this vendor tax invoice within the sequence.</xs:documentation>
        <xs:appinfo>
          <lx:path>vendorTaxInvoiceType.SequenceNumber</lx:path>
          <lx:label>Sequence Number</lx:label>
          <li:transactions>CAL,DAS,CNZ</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="TaxInvoiceDate" type="dateType" use="optional">
      <xs:annotation>
        <xs:documentation>The date of the tax invoice.</xs:documentation>
        <xs:appinfo>
          <lx:path>vendorTaxInvoiceType.TaxInvoiceDate</lx:path>
          <lx:label>Tax Invoice Date</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="UniqueID" type="uniqueIDType" use="optional">
      <xs:annotation>
        <xs:documentation>Identifier for the Invoice.</xs:documentation>
        <xs:appinfo>
          <lx:path>vendorTaxInvoiceType.UniqueID</lx:path>
          <lx:label>Unique ID</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="x_DeliverTo" type="referenceType" use="optional">
      <xs:annotation>
        <xs:documentation>Party accepting delivery, and the location for delivery. For a commercial loan this is most commonly the customer, however it may be at a different address than the main applicant address.</xs:documentation>
        <xs:appinfo>
          <lx:path>vendorTaxInvoiceType.x_DeliverTo</lx:path>
          <lx:label>Deliver To</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          <lx:references>
            <lx:target>Package.Content.Application.Address</lx:target>
          </lx:references>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="x_Purchaser" type="referenceType" use="optional">
      <xs:annotation>
        <xs:documentation>This is the party receiving legal title to the goods, hence who is responsible for paying the invoice. Can be either the lender or the end customer depending on the finance structure.</xs:documentation>
        <xs:appinfo>
          <lx:path>vendorTaxInvoiceType.x_Purchaser</lx:path>
          <lx:label>Purchaser</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          <lx:references>
            <lx:target>Package.Content.Application.CompanyApplicant</lx:target>
            <lx:target>Package.Content.Application.PersonApplicant</lx:target>
            <lx:target>Package.Content.Application.RelatedCompany</lx:target>
            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
            <lx:target>Package.Content.Application.TrustApplicant</lx:target>
          </lx:references>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
    <xs:attribute name="x_Vendor" type="referenceType" use="optional">
      <xs:annotation>
        <xs:documentation>The party from whom the applicant is acquiring the specified asset(s) and hence to whom the loan funds (purchase/lease etc.) are paid.</xs:documentation>
        <xs:appinfo>
          <lx:path>vendorTaxInvoiceType.x_Vendor</lx:path>
          <lx:label>Vendor</lx:label>
          <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          <lx:references>
            <lx:target>Package.Content.Application.RelatedCompany</lx:target>
            <lx:target>Package.Content.Application.RelatedPerson</lx:target>
          </lx:references>
        </xs:appinfo>
      </xs:annotation>
    </xs:attribute>
  </xs:complexType>
  <xs:simpleType name="absLendingPurposeCodeList">
    <xs:annotation>
      <xs:documentation>absLendingPurposeCodeList</xs:documentation>
      <xs:appinfo>
        <lx:path>absLendingPurposeCodeList</lx:path>
        <lx:label>absLendingPurposeCodeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="ABS-121">
        <xs:annotation>
          <xs:documentation>Construction of dwelling - House</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-121</lx:path>
            <lx:label>Construction of dwelling - House</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-123">
        <xs:annotation>
          <xs:documentation>Construction of dwelling - Other dwelling</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-123</lx:path>
            <lx:label>Construction of dwelling - Other dwelling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-125">
        <xs:annotation>
          <xs:documentation>Purchase of newly erected dwelling - House</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-125</lx:path>
            <lx:label>Purchase of newly erected dwelling - House</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-127">
        <xs:annotation>
          <xs:documentation>Purchase of newly erected dwelling - Other dwelling</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-127</lx:path>
            <lx:label>Purchase of newly erected dwelling - Other dwelling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-129">
        <xs:annotation>
          <xs:documentation>Purchase of established dwelling - House</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-129</lx:path>
            <lx:label>Purchase of established dwelling - House</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-131">
        <xs:annotation>
          <xs:documentation>Purchase of established dwelling - Other dwelling</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-131</lx:path>
            <lx:label>Purchase of established dwelling - Other dwelling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-133">
        <xs:annotation>
          <xs:documentation>Refinance existing home loans - Loans held with other financial institutions</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-133</lx:path>
            <lx:label>Refinance existing home loans - Loans held with other financial institutions</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-143">
        <xs:annotation>
          <xs:documentation>Structural and non-structural alterations and/or additions to dwellings (excluding swimming pools, maintenance, repairs and other home improvements)</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-143</lx:path>
            <lx:label>Structural and non-structural alterations and/or additions to dwellings (excluding swimming pools, maintenance, repairs and other home improvements)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-221">
        <xs:annotation>
          <xs:documentation>Purchase of motor cars and station wagons - New</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-221</lx:path>
            <lx:label>Purchase of motor cars and station wagons - New</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-223">
        <xs:annotation>
          <xs:documentation>Purchase of motor cars and station wagons - Used</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-223</lx:path>
            <lx:label>Purchase of motor cars and station wagons - Used</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-225">
        <xs:annotation>
          <xs:documentation>Purchase of other motor vehicles</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-225</lx:path>
            <lx:label>Purchase of other motor vehicles</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-227">
        <xs:annotation>
          <xs:documentation>Purchase of motor cycles, scooters</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-227</lx:path>
            <lx:label>Purchase of motor cycles, scooters</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-229">
        <xs:annotation>
          <xs:documentation>Purchase of boats, caravans and trailers</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-229</lx:path>
            <lx:label>Purchase of boats, caravans and trailers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-231">
        <xs:annotation>
          <xs:documentation>Purchase of individual residential blocks of land</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-231</lx:path>
            <lx:label>Purchase of individual residential blocks of land</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-233">
        <xs:annotation>
          <xs:documentation>Unsecured loans for alterations and/or additions to dwellings</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-233</lx:path>
            <lx:label>Unsecured loans for alterations and/or additions to dwellings</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-235">
        <xs:annotation>
          <xs:documentation>Unsecured loans to assist in the purchase or construction of housing for owner occupation</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-235</lx:path>
            <lx:label>Unsecured loans to assist in the purchase or construction of housing for owner occupation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-237">
        <xs:annotation>
          <xs:documentation>Purchase of household and personal goods</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-237</lx:path>
            <lx:label>Purchase of household and personal goods</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-239">
        <xs:annotation>
          <xs:documentation>Debt consolidation (excluding refinance of personal loans)</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-239</lx:path>
            <lx:label>Debt consolidation (excluding refinance of personal loans)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-241">
        <xs:annotation>
          <xs:documentation>Travel and holidays</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-241</lx:path>
            <lx:label>Travel and holidays</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-243">
        <xs:annotation>
          <xs:documentation>Refinance of personal loans (excluding investment loans)</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-243</lx:path>
            <lx:label>Refinance of personal loans (excluding investment loans)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-245">
        <xs:annotation>
          <xs:documentation>Personal investment in dwellings for rent/resale</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-245</lx:path>
            <lx:label>Personal investment in dwellings for rent/resale</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-247">
        <xs:annotation>
          <xs:documentation>Other personal investment (including shares and other investment assets)</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-247</lx:path>
            <lx:label>Other personal investment (including shares and other investment assets)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-248">
        <xs:annotation>
          <xs:documentation>Refinancing of investment loans (including purchase of dwellings for rent/resale, purchase of shares and other investment assets)</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-248</lx:path>
            <lx:label>Refinancing of investment loans (including purchase of dwellings for rent/resale, purchase of shares and other investment assets)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-249">
        <xs:annotation>
          <xs:documentation>Other purposes (including swimming pools and home improvements - other than alterations and additions, motor accessories and any other purpose)</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-249</lx:path>
            <lx:label>Other purposes (including swimming pools and home improvements - other than alterations and additions, motor accessories and any other purpose)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-320">
        <xs:annotation>
          <xs:documentation>Construction of dwellings for rent/resale</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-320</lx:path>
            <lx:label>Construction of dwellings for rent/resale</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-321">
        <xs:annotation>
          <xs:documentation>Construction of non-residential buildings - Shop</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-321</lx:path>
            <lx:label>Construction of non-residential buildings - Shop</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-322">
        <xs:annotation>
          <xs:documentation>Construction of non-residential buildings - Offices</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-322</lx:path>
            <lx:label>Construction of non-residential buildings - Offices</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-323">
        <xs:annotation>
          <xs:documentation>Construction of non-residential buildings - Industrial buildings</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-323</lx:path>
            <lx:label>Construction of non-residential buildings - Industrial buildings</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-324">
        <xs:annotation>
          <xs:documentation>Construction of non-residential buildings - Other</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-324</lx:path>
            <lx:label>Construction of non-residential buildings - Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-326">
        <xs:annotation>
          <xs:documentation>Construction of non-building structures</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-326</lx:path>
            <lx:label>Construction of non-building structures</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-327">
        <xs:annotation>
          <xs:documentation>Structural and non-structural alterations and/or additions to buildings (excluding maintenance, repairs and other improvements that don't include building work)</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-327</lx:path>
            <lx:label>Structural and non-structural alterations and/or additions to buildings (excluding maintenance, repairs and other improvements that don't include building work)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-329">
        <xs:annotation>
          <xs:documentation>Purchase of land and buildings - Dwellings for rent/resale</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-329</lx:path>
            <lx:label>Purchase of land and buildings - Dwellings for rent/resale</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-330">
        <xs:annotation>
          <xs:documentation>Purchase of land and buildings - Shops</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-330</lx:path>
            <lx:label>Purchase of land and buildings - Shops</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-331">
        <xs:annotation>
          <xs:documentation>Purchase of land and buildings - Offices</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-331</lx:path>
            <lx:label>Purchase of land and buildings - Offices</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-332">
        <xs:annotation>
          <xs:documentation>Purchase of land and buildings - Industrial buildings (including factories, automotive repair facilities, etc.)</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-332</lx:path>
            <lx:label>Purchase of land and buildings - Industrial buildings (including factories, automotive repair facilities, etc.)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-333">
        <xs:annotation>
          <xs:documentation>Purchase of land and buildings - Other non-residential buildings (including hotels, buildings for educational, religious, health, entertainment and recreational purposes)</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-333</lx:path>
            <lx:label>Purchase of land and buildings - Other non-residential buildings (including hotels, buildings for educational, religious, health, entertainment and recreational purposes)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-335">
        <xs:annotation>
          <xs:documentation>Purchase of land and buildings - Rural property</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-335</lx:path>
            <lx:label>Purchase of land and buildings - Rural property</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-336">
        <xs:annotation>
          <xs:documentation>Purchase of land and buildings - Land for residential block development</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-336</lx:path>
            <lx:label>Purchase of land and buildings - Land for residential block development</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-337">
        <xs:annotation>
          <xs:documentation>Purchase of land and buildings - Other land purchases</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-337</lx:path>
            <lx:label>Purchase of land and buildings - Other land purchases</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-340">
        <xs:annotation>
          <xs:documentation>Wholesale Finance (including purchase of goods by wholesalers or retailers, traders' stocks held under bailment or floor plan schemes or working capital)</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-340</lx:path>
            <lx:label>Wholesale Finance (including purchase of goods by wholesalers or retailers, traders' stocks held under bailment or floor plan schemes or working capital)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-341">
        <xs:annotation>
          <xs:documentation>Purchase of plant and equipment - Motor vehicles (excluding motorcycles, trailers, caravans, trains, boats and planes)</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-341</lx:path>
            <lx:label>Purchase of plant and equipment - Motor vehicles (excluding motorcycles, trailers, caravans, trains, boats and planes)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-342">
        <xs:annotation>
          <xs:documentation>Purchase of plant and equipment - Other transport equipment (including motorcycles, trailers, caravans, trains, boats and planes)</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-342</lx:path>
            <lx:label>Purchase of plant and equipment - Other transport equipment (including motorcycles, trailers, caravans, trains, boats and planes)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-343">
        <xs:annotation>
          <xs:documentation>Purchase of plant and equipment - Other plant equipment (including construction and earthmoving equipment, agricultural machinery and office equipment)</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-343</lx:path>
            <lx:label>Purchase of plant and equipment - Other plant equipment (including construction and earthmoving equipment, agricultural machinery and office equipment)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-348">
        <xs:annotation>
          <xs:documentation>Refinance of existing commercial loans</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-348</lx:path>
            <lx:label>Refinance of existing commercial loans</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABS-349">
        <xs:annotation>
          <xs:documentation>Other purpose (including factoring - whether secured by trade debts or to purchase trade debts, maintenance, repairs and other improvements to property not involving building work)</xs:documentation>
          <xs:appinfo>
            <lx:path>absLendingPurposeCodeList.ABS-349</lx:path>
            <lx:label>Other purpose (including factoring - whether secured by trade debts or to purchase trade debts, maintenance, repairs and other improvements to property not involving building work)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="absOccupationCodeList">
    <xs:annotation>
      <xs:documentation>absOccupationCodeList</xs:documentation>
      <xs:appinfo>
        <lx:path>absOccupationCodeList</lx:path>
        <lx:label>absOccupationCodeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="1">
        <xs:annotation>
          <xs:documentation>MANAGERS AND ADMINISTRATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1</lx:path>
            <lx:label>MANAGERS AND ADMINISTRATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="11">
        <xs:annotation>
          <xs:documentation>GENERALIST MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.11</lx:path>
            <lx:label>GENERALIST MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="111">
        <xs:annotation>
          <xs:documentation>GENERAL MANAGERS AND ADMINISTRATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.111</lx:path>
            <lx:label>GENERAL MANAGERS AND ADMINISTRATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1111">
        <xs:annotation>
          <xs:documentation>LEGISLATORS AND GOVERNMENT APPOINTED OFFICIALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1111</lx:path>
            <lx:label>LEGISLATORS AND GOVERNMENT APPOINTED OFFICIALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1111-11">
        <xs:annotation>
          <xs:documentation>Parliamentarian or Councillor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1111-11</lx:path>
            <lx:label>Parliamentarian or Councillor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1111-13">
        <xs:annotation>
          <xs:documentation>Judge</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1111-13</lx:path>
            <lx:label>Judge</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1111-15">
        <xs:annotation>
          <xs:documentation>Magistrate</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1111-15</lx:path>
            <lx:label>Magistrate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1111-17">
        <xs:annotation>
          <xs:documentation>Tribunal Member</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1111-17</lx:path>
            <lx:label>Tribunal Member</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1111-79">
        <xs:annotation>
          <xs:documentation>Legislators and Government Appointed Officials nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1111-79</lx:path>
            <lx:label>Legislators and Government Appointed Officials nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1112">
        <xs:annotation>
          <xs:documentation>GENERAL MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1112</lx:path>
            <lx:label>GENERAL MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1112-11">
        <xs:annotation>
          <xs:documentation>General Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1112-11</lx:path>
            <lx:label>General Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="119">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS GENERALIST MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.119</lx:path>
            <lx:label>MISCELLANEOUS GENERALIST MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1191">
        <xs:annotation>
          <xs:documentation>BUILDING AND CONSTRUCTION MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1191</lx:path>
            <lx:label>BUILDING AND CONSTRUCTION MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1191-11">
        <xs:annotation>
          <xs:documentation>Construction Project Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1191-11</lx:path>
            <lx:label>Construction Project Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1191-13">
        <xs:annotation>
          <xs:documentation>Project Builder</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1191-13</lx:path>
            <lx:label>Project Builder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1192">
        <xs:annotation>
          <xs:documentation>IMPORTERS, EXPORTERS AND WHOLESALERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1192</lx:path>
            <lx:label>IMPORTERS, EXPORTERS AND WHOLESALERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1192-11">
        <xs:annotation>
          <xs:documentation>Importer or Exporter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1192-11</lx:path>
            <lx:label>Importer or Exporter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1192-13">
        <xs:annotation>
          <xs:documentation>Wholesaler</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1192-13</lx:path>
            <lx:label>Wholesaler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1193">
        <xs:annotation>
          <xs:documentation>MANUFACTURERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1193</lx:path>
            <lx:label>MANUFACTURERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1193-11">
        <xs:annotation>
          <xs:documentation>Manufacturer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1193-11</lx:path>
            <lx:label>Manufacturer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="12">
        <xs:annotation>
          <xs:documentation>SPECIALIST MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.12</lx:path>
            <lx:label>SPECIALIST MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121">
        <xs:annotation>
          <xs:documentation>RESOURCE MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.121</lx:path>
            <lx:label>RESOURCE MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1211">
        <xs:annotation>
          <xs:documentation>FINANCE MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1211</lx:path>
            <lx:label>FINANCE MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1211-11">
        <xs:annotation>
          <xs:documentation>Finance Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1211-11</lx:path>
            <lx:label>Finance Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1212">
        <xs:annotation>
          <xs:documentation>COMPANY SECRETARIES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1212</lx:path>
            <lx:label>COMPANY SECRETARIES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1212-11">
        <xs:annotation>
          <xs:documentation>Company Secretary</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1212-11</lx:path>
            <lx:label>Company Secretary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1213">
        <xs:annotation>
          <xs:documentation>HUMAN RESOURCE MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1213</lx:path>
            <lx:label>HUMAN RESOURCE MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1213-11">
        <xs:annotation>
          <xs:documentation>Human Resource Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1213-11</lx:path>
            <lx:label>Human Resource Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="122">
        <xs:annotation>
          <xs:documentation>ENGINEERING, DISTRIBUTION AND PROCESS MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.122</lx:path>
            <lx:label>ENGINEERING, DISTRIBUTION AND PROCESS MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1221">
        <xs:annotation>
          <xs:documentation>ENGINEERING MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1221</lx:path>
            <lx:label>ENGINEERING MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1221-11">
        <xs:annotation>
          <xs:documentation>Engineering Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1221-11</lx:path>
            <lx:label>Engineering Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1222">
        <xs:annotation>
          <xs:documentation>PRODUCTION MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1222</lx:path>
            <lx:label>PRODUCTION MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1222-11">
        <xs:annotation>
          <xs:documentation>Production Manager (Manufacturing)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1222-11</lx:path>
            <lx:label>Production Manager (Manufacturing)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1222-13">
        <xs:annotation>
          <xs:documentation>Production Manager (Mining)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1222-13</lx:path>
            <lx:label>Production Manager (Mining)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1223">
        <xs:annotation>
          <xs:documentation>SUPPLY AND DISTRIBUTION MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1223</lx:path>
            <lx:label>SUPPLY AND DISTRIBUTION MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1223-11">
        <xs:annotation>
          <xs:documentation>Supply and Distribution Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1223-11</lx:path>
            <lx:label>Supply and Distribution Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1224">
        <xs:annotation>
          <xs:documentation>INFORMATION TECHNOLOGY MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1224</lx:path>
            <lx:label>INFORMATION TECHNOLOGY MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1224-11">
        <xs:annotation>
          <xs:documentation>Information Technology Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1224-11</lx:path>
            <lx:label>Information Technology Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="123">
        <xs:annotation>
          <xs:documentation>SALES AND MARKETING MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.123</lx:path>
            <lx:label>SALES AND MARKETING MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1231">
        <xs:annotation>
          <xs:documentation>SALES AND MARKETING MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1231</lx:path>
            <lx:label>SALES AND MARKETING MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1231-11">
        <xs:annotation>
          <xs:documentation>Sales and Marketing Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1231-11</lx:path>
            <lx:label>Sales and Marketing Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="129">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS SPECIALIST MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.129</lx:path>
            <lx:label>MISCELLANEOUS SPECIALIST MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1291">
        <xs:annotation>
          <xs:documentation>POLICY AND PLANNING MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1291</lx:path>
            <lx:label>POLICY AND PLANNING MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1291-11">
        <xs:annotation>
          <xs:documentation>Policy and Planning Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1291-11</lx:path>
            <lx:label>Policy and Planning Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1292">
        <xs:annotation>
          <xs:documentation>HEALTH SERVICES MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1292</lx:path>
            <lx:label>HEALTH SERVICES MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1292-11">
        <xs:annotation>
          <xs:documentation>Director of Nursing</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1292-11</lx:path>
            <lx:label>Director of Nursing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1292-13">
        <xs:annotation>
          <xs:documentation>Medical Administrator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1292-13</lx:path>
            <lx:label>Medical Administrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1293">
        <xs:annotation>
          <xs:documentation>EDUCATION MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1293</lx:path>
            <lx:label>EDUCATION MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1293-11">
        <xs:annotation>
          <xs:documentation>School Principal</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1293-11</lx:path>
            <lx:label>School Principal</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1293-13">
        <xs:annotation>
          <xs:documentation>Faculty Head</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1293-13</lx:path>
            <lx:label>Faculty Head</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1293-15">
        <xs:annotation>
          <xs:documentation>Regional Education Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1293-15</lx:path>
            <lx:label>Regional Education Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1293-79">
        <xs:annotation>
          <xs:documentation>Education Managers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1293-79</lx:path>
            <lx:label>Education Managers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1294">
        <xs:annotation>
          <xs:documentation>COMMISSIONED OFFICERS (MANAGEMENT)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1294</lx:path>
            <lx:label>COMMISSIONED OFFICERS (MANAGEMENT)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1294-11">
        <xs:annotation>
          <xs:documentation>Commissioned Defence Force Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1294-11</lx:path>
            <lx:label>Commissioned Defence Force Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1294-13">
        <xs:annotation>
          <xs:documentation>Commissioned Fire Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1294-13</lx:path>
            <lx:label>Commissioned Fire Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1294-15">
        <xs:annotation>
          <xs:documentation>Commissioned Police Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1294-15</lx:path>
            <lx:label>Commissioned Police Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1294-81">
        <xs:annotation>
          <xs:documentation>Trainee Commissioned Defence Force Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1294-81</lx:path>
            <lx:label>Trainee Commissioned Defence Force Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1295">
        <xs:annotation>
          <xs:documentation>CHILD CARE CO-ORDINATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1295</lx:path>
            <lx:label>CHILD CARE CO-ORDINATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1295-11">
        <xs:annotation>
          <xs:documentation>Child Care Co-ordinator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1295-11</lx:path>
            <lx:label>Child Care Co-ordinator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1296">
        <xs:annotation>
          <xs:documentation>MEDIA PRODUCERS AND ARTISTIC DIRECTORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1296</lx:path>
            <lx:label>MEDIA PRODUCERS AND ARTISTIC DIRECTORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1296-11">
        <xs:annotation>
          <xs:documentation>Media Producer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1296-11</lx:path>
            <lx:label>Media Producer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1296-13">
        <xs:annotation>
          <xs:documentation>Artistic Director</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1296-13</lx:path>
            <lx:label>Artistic Director</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1299">
        <xs:annotation>
          <xs:documentation>OTHER SPECIALIST MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1299</lx:path>
            <lx:label>OTHER SPECIALIST MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1299-11">
        <xs:annotation>
          <xs:documentation>Research and Development Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1299-11</lx:path>
            <lx:label>Research and Development Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1299-13">
        <xs:annotation>
          <xs:documentation>Laboratory Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1299-13</lx:path>
            <lx:label>Laboratory Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1299-15">
        <xs:annotation>
          <xs:documentation>Welfare Centre Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1299-15</lx:path>
            <lx:label>Welfare Centre Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1299-17">
        <xs:annotation>
          <xs:documentation>Environment, Parks and Land Care Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1299-17</lx:path>
            <lx:label>Environment, Parks and Land Care Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1299-19">
        <xs:annotation>
          <xs:documentation>Sports Administrator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1299-19</lx:path>
            <lx:label>Sports Administrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1299-79">
        <xs:annotation>
          <xs:documentation>Specialist Managers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1299-79</lx:path>
            <lx:label>Specialist Managers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="13">
        <xs:annotation>
          <xs:documentation>FARMERS AND FARM MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.13</lx:path>
            <lx:label>FARMERS AND FARM MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="131">
        <xs:annotation>
          <xs:documentation>FARMERS AND FARM MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.131</lx:path>
            <lx:label>FARMERS AND FARM MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1311">
        <xs:annotation>
          <xs:documentation>MIXED CROP AND LIVESTOCK FARMERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1311</lx:path>
            <lx:label>MIXED CROP AND LIVESTOCK FARMERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1311-11">
        <xs:annotation>
          <xs:documentation>Mixed Crop and Livestock Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1311-11</lx:path>
            <lx:label>Mixed Crop and Livestock Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1312">
        <xs:annotation>
          <xs:documentation>LIVESTOCK FARMERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1312</lx:path>
            <lx:label>LIVESTOCK FARMERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1312-11">
        <xs:annotation>
          <xs:documentation>Mixed Livestock Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1312-11</lx:path>
            <lx:label>Mixed Livestock Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1312-13">
        <xs:annotation>
          <xs:documentation>Beef Cattle Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1312-13</lx:path>
            <lx:label>Beef Cattle Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1312-15">
        <xs:annotation>
          <xs:documentation>Dairy Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1312-15</lx:path>
            <lx:label>Dairy Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1312-17">
        <xs:annotation>
          <xs:documentation>Sheep Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1312-17</lx:path>
            <lx:label>Sheep Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1312-19">
        <xs:annotation>
          <xs:documentation>Pig Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1312-19</lx:path>
            <lx:label>Pig Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1312-21">
        <xs:annotation>
          <xs:documentation>Poultry Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1312-21</lx:path>
            <lx:label>Poultry Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1312-23">
        <xs:annotation>
          <xs:documentation>Horse Breeder</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1312-23</lx:path>
            <lx:label>Horse Breeder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1312-25">
        <xs:annotation>
          <xs:documentation>Apiarist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1312-25</lx:path>
            <lx:label>Apiarist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1312-79">
        <xs:annotation>
          <xs:documentation>Livestock Farmers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1312-79</lx:path>
            <lx:label>Livestock Farmers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1313">
        <xs:annotation>
          <xs:documentation>CROP FARMERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1313</lx:path>
            <lx:label>CROP FARMERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1313-11">
        <xs:annotation>
          <xs:documentation>Grain, Oilseed and Pasture Grower</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1313-11</lx:path>
            <lx:label>Grain, Oilseed and Pasture Grower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1313-13">
        <xs:annotation>
          <xs:documentation>Sugar Cane Grower</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1313-13</lx:path>
            <lx:label>Sugar Cane Grower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1313-15">
        <xs:annotation>
          <xs:documentation>Tobacco Grower</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1313-15</lx:path>
            <lx:label>Tobacco Grower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1313-17">
        <xs:annotation>
          <xs:documentation>Fruit and Nut Grower</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1313-17</lx:path>
            <lx:label>Fruit and Nut Grower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1313-19">
        <xs:annotation>
          <xs:documentation>Vegetable Grower</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1313-19</lx:path>
            <lx:label>Vegetable Grower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1313-21">
        <xs:annotation>
          <xs:documentation>Flower Grower</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1313-21</lx:path>
            <lx:label>Flower Grower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1313-79">
        <xs:annotation>
          <xs:documentation>Crop Farmers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1313-79</lx:path>
            <lx:label>Crop Farmers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1314">
        <xs:annotation>
          <xs:documentation>AQUACULTURE FARMERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1314</lx:path>
            <lx:label>AQUACULTURE FARMERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1314-11">
        <xs:annotation>
          <xs:documentation>Aquaculture Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.1314-11</lx:path>
            <lx:label>Aquaculture Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2">
        <xs:annotation>
          <xs:documentation>PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2</lx:path>
            <lx:label>PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="21">
        <xs:annotation>
          <xs:documentation>SCIENCE, BUILDING AND ENGINEERING PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.21</lx:path>
            <lx:label>SCIENCE, BUILDING AND ENGINEERING PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211">
        <xs:annotation>
          <xs:documentation>NATURAL AND PHYSICAL SCIENCE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.211</lx:path>
            <lx:label>NATURAL AND PHYSICAL SCIENCE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2111">
        <xs:annotation>
          <xs:documentation>CHEMISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2111</lx:path>
            <lx:label>CHEMISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2111-11">
        <xs:annotation>
          <xs:documentation>Chemist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2111-11</lx:path>
            <lx:label>Chemist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2112">
        <xs:annotation>
          <xs:documentation>GEOLOGISTS AND GEOPHYSICISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2112</lx:path>
            <lx:label>GEOLOGISTS AND GEOPHYSICISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2112-11">
        <xs:annotation>
          <xs:documentation>Geologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2112-11</lx:path>
            <lx:label>Geologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2112-13">
        <xs:annotation>
          <xs:documentation>Geophysicist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2112-13</lx:path>
            <lx:label>Geophysicist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2113">
        <xs:annotation>
          <xs:documentation>LIFE SCIENTISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2113</lx:path>
            <lx:label>LIFE SCIENTISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2113-11">
        <xs:annotation>
          <xs:documentation>Anatomist or Physiologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2113-11</lx:path>
            <lx:label>Anatomist or Physiologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2113-13">
        <xs:annotation>
          <xs:documentation>Botanist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2113-13</lx:path>
            <lx:label>Botanist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2113-15">
        <xs:annotation>
          <xs:documentation>Zoologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2113-15</lx:path>
            <lx:label>Zoologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2113-17">
        <xs:annotation>
          <xs:documentation>Biochemist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2113-17</lx:path>
            <lx:label>Biochemist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2113-19">
        <xs:annotation>
          <xs:documentation>Marine Biologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2113-19</lx:path>
            <lx:label>Marine Biologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2113-79">
        <xs:annotation>
          <xs:documentation>Life Scientists nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2113-79</lx:path>
            <lx:label>Life Scientists nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2114">
        <xs:annotation>
          <xs:documentation>ENVIRONMENTAL AND AGRICULTURAL SCIENCE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2114</lx:path>
            <lx:label>ENVIRONMENTAL AND AGRICULTURAL SCIENCE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2114-11">
        <xs:annotation>
          <xs:documentation>Environmental Research Scientist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2114-11</lx:path>
            <lx:label>Environmental Research Scientist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2114-13">
        <xs:annotation>
          <xs:documentation>Forester</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2114-13</lx:path>
            <lx:label>Forester</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2114-15">
        <xs:annotation>
          <xs:documentation>Park Ranger</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2114-15</lx:path>
            <lx:label>Park Ranger</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2114-17">
        <xs:annotation>
          <xs:documentation>Soil Scientist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2114-17</lx:path>
            <lx:label>Soil Scientist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2114-19">
        <xs:annotation>
          <xs:documentation>Agricultural Scientist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2114-19</lx:path>
            <lx:label>Agricultural Scientist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2114-21">
        <xs:annotation>
          <xs:documentation>Agricultural Adviser</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2114-21</lx:path>
            <lx:label>Agricultural Adviser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2114-79">
        <xs:annotation>
          <xs:documentation>Environmental and Agricultural Science Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2114-79</lx:path>
            <lx:label>Environmental and Agricultural Science Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2115">
        <xs:annotation>
          <xs:documentation>MEDICAL SCIENTISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2115</lx:path>
            <lx:label>MEDICAL SCIENTISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2115-11">
        <xs:annotation>
          <xs:documentation>Medical Scientist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2115-11</lx:path>
            <lx:label>Medical Scientist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2119">
        <xs:annotation>
          <xs:documentation>OTHER NATURAL AND PHYSICAL SCIENCE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2119</lx:path>
            <lx:label>OTHER NATURAL AND PHYSICAL SCIENCE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2119-11">
        <xs:annotation>
          <xs:documentation>Physicist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2119-11</lx:path>
            <lx:label>Physicist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2119-13">
        <xs:annotation>
          <xs:documentation>Meteorologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2119-13</lx:path>
            <lx:label>Meteorologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2119-15">
        <xs:annotation>
          <xs:documentation>Extractive Metallurgist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2119-15</lx:path>
            <lx:label>Extractive Metallurgist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2119-17">
        <xs:annotation>
          <xs:documentation>Physical Metallurgist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2119-17</lx:path>
            <lx:label>Physical Metallurgist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2119-19">
        <xs:annotation>
          <xs:documentation>Materials Scientist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2119-19</lx:path>
            <lx:label>Materials Scientist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2119-79">
        <xs:annotation>
          <xs:documentation>Natural and Physical Science Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2119-79</lx:path>
            <lx:label>Natural and Physical Science Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212">
        <xs:annotation>
          <xs:documentation>BUILDING AND ENGINEERING PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.212</lx:path>
            <lx:label>BUILDING AND ENGINEERING PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2121">
        <xs:annotation>
          <xs:documentation>ARCHITECTS AND LANDSCAPE ARCHITECTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2121</lx:path>
            <lx:label>ARCHITECTS AND LANDSCAPE ARCHITECTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2121-11">
        <xs:annotation>
          <xs:documentation>Architect</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2121-11</lx:path>
            <lx:label>Architect</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2121-13">
        <xs:annotation>
          <xs:documentation>Landscape Architect</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2121-13</lx:path>
            <lx:label>Landscape Architect</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2122">
        <xs:annotation>
          <xs:documentation>QUANTITY SURVEYORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2122</lx:path>
            <lx:label>QUANTITY SURVEYORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2122-11">
        <xs:annotation>
          <xs:documentation>Quantity Surveyor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2122-11</lx:path>
            <lx:label>Quantity Surveyor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2123">
        <xs:annotation>
          <xs:documentation>CARTOGRAPHERS AND SURVEYORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2123</lx:path>
            <lx:label>CARTOGRAPHERS AND SURVEYORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2123-11">
        <xs:annotation>
          <xs:documentation>Cartographer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2123-11</lx:path>
            <lx:label>Cartographer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2123-13">
        <xs:annotation>
          <xs:documentation>Surveyor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2123-13</lx:path>
            <lx:label>Surveyor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2124">
        <xs:annotation>
          <xs:documentation>CIVIL ENGINEERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2124</lx:path>
            <lx:label>CIVIL ENGINEERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2124-11">
        <xs:annotation>
          <xs:documentation>Civil Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2124-11</lx:path>
            <lx:label>Civil Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2125">
        <xs:annotation>
          <xs:documentation>ELECTRICAL AND ELECTRONICS ENGINEERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2125</lx:path>
            <lx:label>ELECTRICAL AND ELECTRONICS ENGINEERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2125-11">
        <xs:annotation>
          <xs:documentation>Electrical Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2125-11</lx:path>
            <lx:label>Electrical Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2125-13">
        <xs:annotation>
          <xs:documentation>Electronics Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2125-13</lx:path>
            <lx:label>Electronics Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2126">
        <xs:annotation>
          <xs:documentation>MECHANICAL, PRODUCTION AND PLANT ENGINEERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2126</lx:path>
            <lx:label>MECHANICAL, PRODUCTION AND PLANT ENGINEERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2126-11">
        <xs:annotation>
          <xs:documentation>Mechanical Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2126-11</lx:path>
            <lx:label>Mechanical Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2126-13">
        <xs:annotation>
          <xs:documentation>Production or Plant Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2126-13</lx:path>
            <lx:label>Production or Plant Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2127">
        <xs:annotation>
          <xs:documentation>MINING AND MATERIALS ENGINEERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2127</lx:path>
            <lx:label>MINING AND MATERIALS ENGINEERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2127-11">
        <xs:annotation>
          <xs:documentation>Mining Engineer (excluding Petroleum)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2127-11</lx:path>
            <lx:label>Mining Engineer (excluding Petroleum)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2127-13">
        <xs:annotation>
          <xs:documentation>Petroleum Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2127-13</lx:path>
            <lx:label>Petroleum Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2127-15">
        <xs:annotation>
          <xs:documentation>Materials Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2127-15</lx:path>
            <lx:label>Materials Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2128">
        <xs:annotation>
          <xs:documentation>ENGINEERING TECHNOLOGISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2128</lx:path>
            <lx:label>ENGINEERING TECHNOLOGISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2128-11">
        <xs:annotation>
          <xs:documentation>Civil Engineering Technologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2128-11</lx:path>
            <lx:label>Civil Engineering Technologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2128-13">
        <xs:annotation>
          <xs:documentation>Mechanical Engineering Technologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2128-13</lx:path>
            <lx:label>Mechanical Engineering Technologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2128-15">
        <xs:annotation>
          <xs:documentation>Electrical or Electronics Engineering Technologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2128-15</lx:path>
            <lx:label>Electrical or Electronics Engineering Technologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2128-79">
        <xs:annotation>
          <xs:documentation>Engineering Technologists nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2128-79</lx:path>
            <lx:label>Engineering Technologists nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2129">
        <xs:annotation>
          <xs:documentation>OTHER BUILDING AND ENGINEERING PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2129</lx:path>
            <lx:label>OTHER BUILDING AND ENGINEERING PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2129-11">
        <xs:annotation>
          <xs:documentation>Aeronautical Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2129-11</lx:path>
            <lx:label>Aeronautical Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2129-13">
        <xs:annotation>
          <xs:documentation>Agricultural Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2129-13</lx:path>
            <lx:label>Agricultural Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2129-15">
        <xs:annotation>
          <xs:documentation>Biomedical Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2129-15</lx:path>
            <lx:label>Biomedical Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2129-17">
        <xs:annotation>
          <xs:documentation>Chemical Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2129-17</lx:path>
            <lx:label>Chemical Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2129-19">
        <xs:annotation>
          <xs:documentation>Industrial Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2129-19</lx:path>
            <lx:label>Industrial Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2129-21">
        <xs:annotation>
          <xs:documentation>Naval Architect</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2129-21</lx:path>
            <lx:label>Naval Architect</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2129-79">
        <xs:annotation>
          <xs:documentation>Building and Engineering Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2129-79</lx:path>
            <lx:label>Building and Engineering Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="22">
        <xs:annotation>
          <xs:documentation>BUSINESS AND INFORMATION PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.22</lx:path>
            <lx:label>BUSINESS AND INFORMATION PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="221">
        <xs:annotation>
          <xs:documentation>ACCOUNTANTS, AUDITORS AND CORPORATE TREASURERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.221</lx:path>
            <lx:label>ACCOUNTANTS, AUDITORS AND CORPORATE TREASURERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2211">
        <xs:annotation>
          <xs:documentation>ACCOUNTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2211</lx:path>
            <lx:label>ACCOUNTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2211-11">
        <xs:annotation>
          <xs:documentation>Accountant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2211-11</lx:path>
            <lx:label>Accountant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2212">
        <xs:annotation>
          <xs:documentation>AUDITORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2212</lx:path>
            <lx:label>AUDITORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2212-11">
        <xs:annotation>
          <xs:documentation>External Auditor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2212-11</lx:path>
            <lx:label>External Auditor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2212-13">
        <xs:annotation>
          <xs:documentation>Internal Auditor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2212-13</lx:path>
            <lx:label>Internal Auditor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2213">
        <xs:annotation>
          <xs:documentation>CORPORATE TREASURERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2213</lx:path>
            <lx:label>CORPORATE TREASURERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2213-11">
        <xs:annotation>
          <xs:documentation>Corporate Treasurer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2213-11</lx:path>
            <lx:label>Corporate Treasurer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222">
        <xs:annotation>
          <xs:documentation>SALES, MARKETING AND ADVERTISING PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.222</lx:path>
            <lx:label>SALES, MARKETING AND ADVERTISING PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2221">
        <xs:annotation>
          <xs:documentation>MARKETING AND ADVERTISING PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2221</lx:path>
            <lx:label>MARKETING AND ADVERTISING PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2221-11">
        <xs:annotation>
          <xs:documentation>Public Relations Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2221-11</lx:path>
            <lx:label>Public Relations Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2221-13">
        <xs:annotation>
          <xs:documentation>Marketing Specialist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2221-13</lx:path>
            <lx:label>Marketing Specialist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2221-15">
        <xs:annotation>
          <xs:documentation>Market Research Analyst</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2221-15</lx:path>
            <lx:label>Market Research Analyst</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2221-17">
        <xs:annotation>
          <xs:documentation>Advertising Specialist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2221-17</lx:path>
            <lx:label>Advertising Specialist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2222">
        <xs:annotation>
          <xs:documentation>TECHNICAL SALES REPRESENTATIVES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2222</lx:path>
            <lx:label>TECHNICAL SALES REPRESENTATIVES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2222-11">
        <xs:annotation>
          <xs:documentation>Sales Representative (Industrial Products)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2222-11</lx:path>
            <lx:label>Sales Representative (Industrial Products)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2222-13">
        <xs:annotation>
          <xs:documentation>Sales Representative (Information and Communication Products)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2222-13</lx:path>
            <lx:label>Sales Representative (Information and Communication Products)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2222-15">
        <xs:annotation>
          <xs:documentation>Sales Representative (Medical and Pharmaceutical Products)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2222-15</lx:path>
            <lx:label>Sales Representative (Medical and Pharmaceutical Products)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2222-79">
        <xs:annotation>
          <xs:documentation>Technical Sales Representatives nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2222-79</lx:path>
            <lx:label>Technical Sales Representatives nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="223">
        <xs:annotation>
          <xs:documentation>COMPUTING PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.223</lx:path>
            <lx:label>COMPUTING PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2231">
        <xs:annotation>
          <xs:documentation>COMPUTING PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2231</lx:path>
            <lx:label>COMPUTING PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2231-11">
        <xs:annotation>
          <xs:documentation>Systems Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2231-11</lx:path>
            <lx:label>Systems Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2231-13">
        <xs:annotation>
          <xs:documentation>Systems Designer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2231-13</lx:path>
            <lx:label>Systems Designer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2231-15">
        <xs:annotation>
          <xs:documentation>Software Designer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2231-15</lx:path>
            <lx:label>Software Designer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2231-17">
        <xs:annotation>
          <xs:documentation>Applications and Analyst Programmer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2231-17</lx:path>
            <lx:label>Applications and Analyst Programmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2231-19">
        <xs:annotation>
          <xs:documentation>Systems Programmer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2231-19</lx:path>
            <lx:label>Systems Programmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2231-21">
        <xs:annotation>
          <xs:documentation>Computer Systems Auditor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2231-21</lx:path>
            <lx:label>Computer Systems Auditor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2231-79">
        <xs:annotation>
          <xs:documentation>Computing Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2231-79</lx:path>
            <lx:label>Computing Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="229">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS BUSINESS AND INFORMATION PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.229</lx:path>
            <lx:label>MISCELLANEOUS BUSINESS AND INFORMATION PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2291">
        <xs:annotation>
          <xs:documentation>HUMAN RESOURCE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2291</lx:path>
            <lx:label>HUMAN RESOURCE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2291-11">
        <xs:annotation>
          <xs:documentation>Personnel Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2291-11</lx:path>
            <lx:label>Personnel Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2291-13">
        <xs:annotation>
          <xs:documentation>Personnel Consultant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2291-13</lx:path>
            <lx:label>Personnel Consultant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2291-15">
        <xs:annotation>
          <xs:documentation>Industrial Relations Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2291-15</lx:path>
            <lx:label>Industrial Relations Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2291-17">
        <xs:annotation>
          <xs:documentation>Training Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2291-17</lx:path>
            <lx:label>Training Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2292">
        <xs:annotation>
          <xs:documentation>LIBRARIANS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2292</lx:path>
            <lx:label>LIBRARIANS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2292-11">
        <xs:annotation>
          <xs:documentation>Librarian</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2292-11</lx:path>
            <lx:label>Librarian</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2293">
        <xs:annotation>
          <xs:documentation>MATHEMATICIANS, STATISTICIANS AND ACTUARIES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2293</lx:path>
            <lx:label>MATHEMATICIANS, STATISTICIANS AND ACTUARIES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2293-11">
        <xs:annotation>
          <xs:documentation>Mathematician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2293-11</lx:path>
            <lx:label>Mathematician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2293-13">
        <xs:annotation>
          <xs:documentation>Statistician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2293-13</lx:path>
            <lx:label>Statistician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2293-15">
        <xs:annotation>
          <xs:documentation>Actuary</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2293-15</lx:path>
            <lx:label>Actuary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2294">
        <xs:annotation>
          <xs:documentation>BUSINESS AND ORGANISATION ANALYSTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2294</lx:path>
            <lx:label>BUSINESS AND ORGANISATION ANALYSTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2294-11">
        <xs:annotation>
          <xs:documentation>Management Consultant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2294-11</lx:path>
            <lx:label>Management Consultant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2294-13">
        <xs:annotation>
          <xs:documentation>Organisation and Methods Analyst</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2294-13</lx:path>
            <lx:label>Organisation and Methods Analyst</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2294-15">
        <xs:annotation>
          <xs:documentation>Quality Assurance Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2294-15</lx:path>
            <lx:label>Quality Assurance Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2295">
        <xs:annotation>
          <xs:documentation>PROPERTY PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2295</lx:path>
            <lx:label>PROPERTY PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2295-11">
        <xs:annotation>
          <xs:documentation>Valuer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2295-11</lx:path>
            <lx:label>Valuer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2295-13">
        <xs:annotation>
          <xs:documentation>Land Economist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2295-13</lx:path>
            <lx:label>Land Economist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2299">
        <xs:annotation>
          <xs:documentation>OTHER BUSINESS AND INFORMATION PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2299</lx:path>
            <lx:label>OTHER BUSINESS AND INFORMATION PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2299-11">
        <xs:annotation>
          <xs:documentation>Health Information Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2299-11</lx:path>
            <lx:label>Health Information Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2299-13">
        <xs:annotation>
          <xs:documentation>Records Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2299-13</lx:path>
            <lx:label>Records Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2299-15">
        <xs:annotation>
          <xs:documentation>Archivist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2299-15</lx:path>
            <lx:label>Archivist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2299-17">
        <xs:annotation>
          <xs:documentation>Policy Analyst</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2299-17</lx:path>
            <lx:label>Policy Analyst</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2299-19">
        <xs:annotation>
          <xs:documentation>Intelligence Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2299-19</lx:path>
            <lx:label>Intelligence Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2299-79">
        <xs:annotation>
          <xs:documentation>Business and Information Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2299-79</lx:path>
            <lx:label>Business and Information Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="23">
        <xs:annotation>
          <xs:documentation>HEALTH PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.23</lx:path>
            <lx:label>HEALTH PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231">
        <xs:annotation>
          <xs:documentation>MEDICAL PRACTITIONERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.231</lx:path>
            <lx:label>MEDICAL PRACTITIONERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2311">
        <xs:annotation>
          <xs:documentation>GENERALIST MEDICAL PRACTITIONERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2311</lx:path>
            <lx:label>GENERALIST MEDICAL PRACTITIONERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2311-11">
        <xs:annotation>
          <xs:documentation>General Medical Practitioner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2311-11</lx:path>
            <lx:label>General Medical Practitioner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2311-81">
        <xs:annotation>
          <xs:documentation>Medical Practitioner in Training</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2311-81</lx:path>
            <lx:label>Medical Practitioner in Training</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2312">
        <xs:annotation>
          <xs:documentation>SPECIALIST MEDICAL PRACTITIONERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2312</lx:path>
            <lx:label>SPECIALIST MEDICAL PRACTITIONERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2312-11">
        <xs:annotation>
          <xs:documentation>Anaesthetist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2312-11</lx:path>
            <lx:label>Anaesthetist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2312-13">
        <xs:annotation>
          <xs:documentation>Dermatologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2312-13</lx:path>
            <lx:label>Dermatologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2312-15">
        <xs:annotation>
          <xs:documentation>Emergency Medicine Specialist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2312-15</lx:path>
            <lx:label>Emergency Medicine Specialist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2312-17">
        <xs:annotation>
          <xs:documentation>Obstetrician and Gynaecologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2312-17</lx:path>
            <lx:label>Obstetrician and Gynaecologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2312-19">
        <xs:annotation>
          <xs:documentation>Ophthalmologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2312-19</lx:path>
            <lx:label>Ophthalmologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2312-21">
        <xs:annotation>
          <xs:documentation>Paediatrician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2312-21</lx:path>
            <lx:label>Paediatrician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2312-23">
        <xs:annotation>
          <xs:documentation>Pathologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2312-23</lx:path>
            <lx:label>Pathologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2312-25">
        <xs:annotation>
          <xs:documentation>Specialist Physician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2312-25</lx:path>
            <lx:label>Specialist Physician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2312-27">
        <xs:annotation>
          <xs:documentation>Psychiatrist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2312-27</lx:path>
            <lx:label>Psychiatrist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2312-29">
        <xs:annotation>
          <xs:documentation>Radiologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2312-29</lx:path>
            <lx:label>Radiologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2312-31">
        <xs:annotation>
          <xs:documentation>Surgeon</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2312-31</lx:path>
            <lx:label>Surgeon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2312-79">
        <xs:annotation>
          <xs:documentation>Specialist Medical Practitioners nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2312-79</lx:path>
            <lx:label>Specialist Medical Practitioners nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232">
        <xs:annotation>
          <xs:documentation>NURSING PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.232</lx:path>
            <lx:label>NURSING PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2321">
        <xs:annotation>
          <xs:documentation>NURSE MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2321</lx:path>
            <lx:label>NURSE MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2321-11">
        <xs:annotation>
          <xs:documentation>Nurse Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2321-11</lx:path>
            <lx:label>Nurse Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2322">
        <xs:annotation>
          <xs:documentation>NURSE EDUCATORS AND RESEARCHERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2322</lx:path>
            <lx:label>NURSE EDUCATORS AND RESEARCHERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2322-11">
        <xs:annotation>
          <xs:documentation>Nurse Educator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2322-11</lx:path>
            <lx:label>Nurse Educator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2322-13">
        <xs:annotation>
          <xs:documentation>Nurse Researcher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2322-13</lx:path>
            <lx:label>Nurse Researcher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2323">
        <xs:annotation>
          <xs:documentation>REGISTERED NURSES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2323</lx:path>
            <lx:label>REGISTERED NURSES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2323-11">
        <xs:annotation>
          <xs:documentation>Registered Nurse</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2323-11</lx:path>
            <lx:label>Registered Nurse</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2324">
        <xs:annotation>
          <xs:documentation>REGISTERED MIDWIVES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2324</lx:path>
            <lx:label>REGISTERED MIDWIVES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2324-11">
        <xs:annotation>
          <xs:documentation>Registered Midwife</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2324-11</lx:path>
            <lx:label>Registered Midwife</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2325">
        <xs:annotation>
          <xs:documentation>REGISTERED MENTAL HEALTH NURSES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2325</lx:path>
            <lx:label>REGISTERED MENTAL HEALTH NURSES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2325-11">
        <xs:annotation>
          <xs:documentation>Registered Mental Health Nurse</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2325-11</lx:path>
            <lx:label>Registered Mental Health Nurse</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2326">
        <xs:annotation>
          <xs:documentation>REGISTERED DEVELOPMENTAL DISABILITY NURSES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2326</lx:path>
            <lx:label>REGISTERED DEVELOPMENTAL DISABILITY NURSES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2326-11">
        <xs:annotation>
          <xs:documentation>Registered Developmental Disability Nurse</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2326-11</lx:path>
            <lx:label>Registered Developmental Disability Nurse</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="238">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS HEALTH PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.238</lx:path>
            <lx:label>MISCELLANEOUS HEALTH PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2381">
        <xs:annotation>
          <xs:documentation>DENTAL PRACTITIONERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2381</lx:path>
            <lx:label>DENTAL PRACTITIONERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2381-11">
        <xs:annotation>
          <xs:documentation>Dentist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2381-11</lx:path>
            <lx:label>Dentist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2381-13">
        <xs:annotation>
          <xs:documentation>Dental Specialist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2381-13</lx:path>
            <lx:label>Dental Specialist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2382">
        <xs:annotation>
          <xs:documentation>PHARMACISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2382</lx:path>
            <lx:label>PHARMACISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2382-11">
        <xs:annotation>
          <xs:documentation>Hospital Pharmacist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2382-11</lx:path>
            <lx:label>Hospital Pharmacist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2382-13">
        <xs:annotation>
          <xs:documentation>Industrial Pharmacist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2382-13</lx:path>
            <lx:label>Industrial Pharmacist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2382-15">
        <xs:annotation>
          <xs:documentation>Retail Pharmacist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2382-15</lx:path>
            <lx:label>Retail Pharmacist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2383">
        <xs:annotation>
          <xs:documentation>OCCUPATIONAL THERAPISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2383</lx:path>
            <lx:label>OCCUPATIONAL THERAPISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2383-11">
        <xs:annotation>
          <xs:documentation>Occupational Therapist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2383-11</lx:path>
            <lx:label>Occupational Therapist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2384">
        <xs:annotation>
          <xs:documentation>OPTOMETRISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2384</lx:path>
            <lx:label>OPTOMETRISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2384-11">
        <xs:annotation>
          <xs:documentation>Optometrist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2384-11</lx:path>
            <lx:label>Optometrist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2385">
        <xs:annotation>
          <xs:documentation>PHYSIOTHERAPISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2385</lx:path>
            <lx:label>PHYSIOTHERAPISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2385-11">
        <xs:annotation>
          <xs:documentation>Physiotherapist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2385-11</lx:path>
            <lx:label>Physiotherapist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2386">
        <xs:annotation>
          <xs:documentation>SPEECH PATHOLOGISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2386</lx:path>
            <lx:label>SPEECH PATHOLOGISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2386-11">
        <xs:annotation>
          <xs:documentation>Speech Pathologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2386-11</lx:path>
            <lx:label>Speech Pathologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2387">
        <xs:annotation>
          <xs:documentation>CHIROPRACTORS AND OSTEOPATHS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2387</lx:path>
            <lx:label>CHIROPRACTORS AND OSTEOPATHS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2387-11">
        <xs:annotation>
          <xs:documentation>Chiropractor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2387-11</lx:path>
            <lx:label>Chiropractor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2387-13">
        <xs:annotation>
          <xs:documentation>Osteopath</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2387-13</lx:path>
            <lx:label>Osteopath</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2388">
        <xs:annotation>
          <xs:documentation>PODIATRISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2388</lx:path>
            <lx:label>PODIATRISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2388-11">
        <xs:annotation>
          <xs:documentation>Podiatrist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2388-11</lx:path>
            <lx:label>Podiatrist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2391">
        <xs:annotation>
          <xs:documentation>MEDICAL IMAGING PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2391</lx:path>
            <lx:label>MEDICAL IMAGING PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2391-11">
        <xs:annotation>
          <xs:documentation>Medical Diagnostic Radiographer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2391-11</lx:path>
            <lx:label>Medical Diagnostic Radiographer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2391-13">
        <xs:annotation>
          <xs:documentation>Radiation Therapist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2391-13</lx:path>
            <lx:label>Radiation Therapist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2391-15">
        <xs:annotation>
          <xs:documentation>Nuclear Medicine Technologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2391-15</lx:path>
            <lx:label>Nuclear Medicine Technologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2391-17">
        <xs:annotation>
          <xs:documentation>Sonographer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2391-17</lx:path>
            <lx:label>Sonographer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2392">
        <xs:annotation>
          <xs:documentation>VETERINARIANS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2392</lx:path>
            <lx:label>VETERINARIANS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2392-11">
        <xs:annotation>
          <xs:documentation>Veterinarian</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2392-11</lx:path>
            <lx:label>Veterinarian</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2393">
        <xs:annotation>
          <xs:documentation>DIETITIANS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2393</lx:path>
            <lx:label>DIETITIANS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2393-11">
        <xs:annotation>
          <xs:documentation>Dietitian</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2393-11</lx:path>
            <lx:label>Dietitian</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2394">
        <xs:annotation>
          <xs:documentation>NATURAL THERAPY PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2394</lx:path>
            <lx:label>NATURAL THERAPY PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2394-11">
        <xs:annotation>
          <xs:documentation>Naturopath</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2394-11</lx:path>
            <lx:label>Naturopath</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2394-13">
        <xs:annotation>
          <xs:documentation>Acupuncturist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2394-13</lx:path>
            <lx:label>Acupuncturist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2394-79">
        <xs:annotation>
          <xs:documentation>Natural Therapy Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2394-79</lx:path>
            <lx:label>Natural Therapy Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2399">
        <xs:annotation>
          <xs:documentation>OTHER HEALTH PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2399</lx:path>
            <lx:label>OTHER HEALTH PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2399-11">
        <xs:annotation>
          <xs:documentation>Audiologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2399-11</lx:path>
            <lx:label>Audiologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2399-13">
        <xs:annotation>
          <xs:documentation>Orthoptist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2399-13</lx:path>
            <lx:label>Orthoptist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2399-15">
        <xs:annotation>
          <xs:documentation>Orthotist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2399-15</lx:path>
            <lx:label>Orthotist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2399-79">
        <xs:annotation>
          <xs:documentation>Health Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2399-79</lx:path>
            <lx:label>Health Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="24">
        <xs:annotation>
          <xs:documentation>EDUCATION PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.24</lx:path>
            <lx:label>EDUCATION PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="241">
        <xs:annotation>
          <xs:documentation>SCHOOL TEACHERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.241</lx:path>
            <lx:label>SCHOOL TEACHERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2411">
        <xs:annotation>
          <xs:documentation>PRE-PRIMARY SCHOOL TEACHERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2411</lx:path>
            <lx:label>PRE-PRIMARY SCHOOL TEACHERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2411-11">
        <xs:annotation>
          <xs:documentation>Pre-Primary School Teacher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2411-11</lx:path>
            <lx:label>Pre-Primary School Teacher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2412">
        <xs:annotation>
          <xs:documentation>PRIMARY SCHOOL TEACHERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2412</lx:path>
            <lx:label>PRIMARY SCHOOL TEACHERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2412-11">
        <xs:annotation>
          <xs:documentation>Primary School Teacher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2412-11</lx:path>
            <lx:label>Primary School Teacher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2413">
        <xs:annotation>
          <xs:documentation>SECONDARY SCHOOL TEACHERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2413</lx:path>
            <lx:label>SECONDARY SCHOOL TEACHERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2413-11">
        <xs:annotation>
          <xs:documentation>Secondary School Teacher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2413-11</lx:path>
            <lx:label>Secondary School Teacher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2414">
        <xs:annotation>
          <xs:documentation>SPECIAL EDUCATION TEACHERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2414</lx:path>
            <lx:label>SPECIAL EDUCATION TEACHERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2414-11">
        <xs:annotation>
          <xs:documentation>Special Needs Teacher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2414-11</lx:path>
            <lx:label>Special Needs Teacher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2414-13">
        <xs:annotation>
          <xs:documentation>Teacher of the Hearing Impaired</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2414-13</lx:path>
            <lx:label>Teacher of the Hearing Impaired</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2414-15">
        <xs:annotation>
          <xs:documentation>Teacher of the Sight Impaired</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2414-15</lx:path>
            <lx:label>Teacher of the Sight Impaired</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2414-79">
        <xs:annotation>
          <xs:documentation>Special Education Teachers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2414-79</lx:path>
            <lx:label>Special Education Teachers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="242">
        <xs:annotation>
          <xs:documentation>UNIVERSITY AND VOCATIONAL EDUCATION TEACHERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.242</lx:path>
            <lx:label>UNIVERSITY AND VOCATIONAL EDUCATION TEACHERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2421">
        <xs:annotation>
          <xs:documentation>UNIVERSITY LECTURERS AND TUTORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2421</lx:path>
            <lx:label>UNIVERSITY LECTURERS AND TUTORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2421-11">
        <xs:annotation>
          <xs:documentation>University Lecturer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2421-11</lx:path>
            <lx:label>University Lecturer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2421-13">
        <xs:annotation>
          <xs:documentation>University Tutor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2421-13</lx:path>
            <lx:label>University Tutor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2422">
        <xs:annotation>
          <xs:documentation>VOCATIONAL EDUCATION TEACHERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2422</lx:path>
            <lx:label>VOCATIONAL EDUCATION TEACHERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2422-11">
        <xs:annotation>
          <xs:documentation>Vocational Education Teacher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2422-11</lx:path>
            <lx:label>Vocational Education Teacher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="249">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS EDUCATION PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.249</lx:path>
            <lx:label>MISCELLANEOUS EDUCATION PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2491">
        <xs:annotation>
          <xs:documentation>EXTRA-SYSTEMIC TEACHERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2491</lx:path>
            <lx:label>EXTRA-SYSTEMIC TEACHERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2491-11">
        <xs:annotation>
          <xs:documentation>Art Teacher (Private)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2491-11</lx:path>
            <lx:label>Art Teacher (Private)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2491-13">
        <xs:annotation>
          <xs:documentation>Music Teacher (Private)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2491-13</lx:path>
            <lx:label>Music Teacher (Private)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2491-15">
        <xs:annotation>
          <xs:documentation>Dance Teacher (Private)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2491-15</lx:path>
            <lx:label>Dance Teacher (Private)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2491-17">
        <xs:annotation>
          <xs:documentation>Drama Teacher (Private)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2491-17</lx:path>
            <lx:label>Drama Teacher (Private)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2491-79">
        <xs:annotation>
          <xs:documentation>Extra-Systemic Teachers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2491-79</lx:path>
            <lx:label>Extra-Systemic Teachers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2492">
        <xs:annotation>
          <xs:documentation>ENGLISH AS A SECOND LANGUAGE TEACHERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2492</lx:path>
            <lx:label>ENGLISH AS A SECOND LANGUAGE TEACHERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2492-11">
        <xs:annotation>
          <xs:documentation>English as a Second Language Teacher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2492-11</lx:path>
            <lx:label>English as a Second Language Teacher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2493">
        <xs:annotation>
          <xs:documentation>EDUCATION OFFICERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2493</lx:path>
            <lx:label>EDUCATION OFFICERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2493-11">
        <xs:annotation>
          <xs:documentation>Education Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2493-11</lx:path>
            <lx:label>Education Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="25">
        <xs:annotation>
          <xs:documentation>SOCIAL, ARTS AND MISCELLANEOUS PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.25</lx:path>
            <lx:label>SOCIAL, ARTS AND MISCELLANEOUS PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251">
        <xs:annotation>
          <xs:documentation>SOCIAL WELFARE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.251</lx:path>
            <lx:label>SOCIAL WELFARE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2511">
        <xs:annotation>
          <xs:documentation>SOCIAL WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2511</lx:path>
            <lx:label>SOCIAL WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2511-11">
        <xs:annotation>
          <xs:documentation>Social Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2511-11</lx:path>
            <lx:label>Social Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2512">
        <xs:annotation>
          <xs:documentation>WELFARE AND COMMUNITY WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2512</lx:path>
            <lx:label>WELFARE AND COMMUNITY WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2512-11">
        <xs:annotation>
          <xs:documentation>Welfare Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2512-11</lx:path>
            <lx:label>Welfare Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2512-13">
        <xs:annotation>
          <xs:documentation>Community Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2512-13</lx:path>
            <lx:label>Community Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2513">
        <xs:annotation>
          <xs:documentation>COUNSELLORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2513</lx:path>
            <lx:label>COUNSELLORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2513-11">
        <xs:annotation>
          <xs:documentation>Rehabilitation Counsellor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2513-11</lx:path>
            <lx:label>Rehabilitation Counsellor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2513-13">
        <xs:annotation>
          <xs:documentation>Drug and Alcohol Counsellor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2513-13</lx:path>
            <lx:label>Drug and Alcohol Counsellor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2513-15">
        <xs:annotation>
          <xs:documentation>Family Counsellor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2513-15</lx:path>
            <lx:label>Family Counsellor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2513-17">
        <xs:annotation>
          <xs:documentation>Careers Counsellor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2513-17</lx:path>
            <lx:label>Careers Counsellor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2513-19">
        <xs:annotation>
          <xs:documentation>Student Counsellor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2513-19</lx:path>
            <lx:label>Student Counsellor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2513-79">
        <xs:annotation>
          <xs:documentation>Counsellors nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2513-79</lx:path>
            <lx:label>Counsellors nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2514">
        <xs:annotation>
          <xs:documentation>PSYCHOLOGISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2514</lx:path>
            <lx:label>PSYCHOLOGISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2514-11">
        <xs:annotation>
          <xs:documentation>Clinical Psychologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2514-11</lx:path>
            <lx:label>Clinical Psychologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2514-13">
        <xs:annotation>
          <xs:documentation>Educational Psychologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2514-13</lx:path>
            <lx:label>Educational Psychologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2514-15">
        <xs:annotation>
          <xs:documentation>Organisational Psychologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2514-15</lx:path>
            <lx:label>Organisational Psychologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2514-79">
        <xs:annotation>
          <xs:documentation>Psychologists nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2514-79</lx:path>
            <lx:label>Psychologists nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2515">
        <xs:annotation>
          <xs:documentation>MINISTERS OF RELIGION</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2515</lx:path>
            <lx:label>MINISTERS OF RELIGION</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2515-11">
        <xs:annotation>
          <xs:documentation>Minister of Religion</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2515-11</lx:path>
            <lx:label>Minister of Religion</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS SOCIAL PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.252</lx:path>
            <lx:label>MISCELLANEOUS SOCIAL PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2521">
        <xs:annotation>
          <xs:documentation>LEGAL PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2521</lx:path>
            <lx:label>LEGAL PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2521-11">
        <xs:annotation>
          <xs:documentation>Barrister</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2521-11</lx:path>
            <lx:label>Barrister</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2521-13">
        <xs:annotation>
          <xs:documentation>Solicitor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2521-13</lx:path>
            <lx:label>Solicitor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2521-79">
        <xs:annotation>
          <xs:documentation>Legal Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2521-79</lx:path>
            <lx:label>Legal Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2522">
        <xs:annotation>
          <xs:documentation>ECONOMISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2522</lx:path>
            <lx:label>ECONOMISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2522-11">
        <xs:annotation>
          <xs:documentation>Economist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2522-11</lx:path>
            <lx:label>Economist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2523">
        <xs:annotation>
          <xs:documentation>URBAN AND REGIONAL PLANNERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2523</lx:path>
            <lx:label>URBAN AND REGIONAL PLANNERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2523-11">
        <xs:annotation>
          <xs:documentation>Urban and Regional Planner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2523-11</lx:path>
            <lx:label>Urban and Regional Planner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2529">
        <xs:annotation>
          <xs:documentation>OTHER SOCIAL PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2529</lx:path>
            <lx:label>OTHER SOCIAL PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2529-11">
        <xs:annotation>
          <xs:documentation>Historian</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2529-11</lx:path>
            <lx:label>Historian</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2529-13">
        <xs:annotation>
          <xs:documentation>Interpreter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2529-13</lx:path>
            <lx:label>Interpreter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2529-15">
        <xs:annotation>
          <xs:documentation>Translator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2529-15</lx:path>
            <lx:label>Translator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2529-79">
        <xs:annotation>
          <xs:documentation>Social Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2529-79</lx:path>
            <lx:label>Social Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253">
        <xs:annotation>
          <xs:documentation>ARTISTS AND RELATED PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.253</lx:path>
            <lx:label>ARTISTS AND RELATED PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2531">
        <xs:annotation>
          <xs:documentation>VISUAL ARTS AND CRAFTS PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2531</lx:path>
            <lx:label>VISUAL ARTS AND CRAFTS PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2531-11">
        <xs:annotation>
          <xs:documentation>Painter (Visual Arts)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2531-11</lx:path>
            <lx:label>Painter (Visual Arts)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2531-13">
        <xs:annotation>
          <xs:documentation>Sculptor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2531-13</lx:path>
            <lx:label>Sculptor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2531-15">
        <xs:annotation>
          <xs:documentation>Potter or Ceramic Artist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2531-15</lx:path>
            <lx:label>Potter or Ceramic Artist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2531-79">
        <xs:annotation>
          <xs:documentation>Visual Arts and Crafts Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2531-79</lx:path>
            <lx:label>Visual Arts and Crafts Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2532">
        <xs:annotation>
          <xs:documentation>PHOTOGRAPHERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2532</lx:path>
            <lx:label>PHOTOGRAPHERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2532-11">
        <xs:annotation>
          <xs:documentation>Photographer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2532-11</lx:path>
            <lx:label>Photographer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2533">
        <xs:annotation>
          <xs:documentation>DESIGNERS AND ILLUSTRATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2533</lx:path>
            <lx:label>DESIGNERS AND ILLUSTRATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2533-11">
        <xs:annotation>
          <xs:documentation>Fashion Designer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2533-11</lx:path>
            <lx:label>Fashion Designer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2533-13">
        <xs:annotation>
          <xs:documentation>Graphic Designer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2533-13</lx:path>
            <lx:label>Graphic Designer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2533-15">
        <xs:annotation>
          <xs:documentation>Industrial Designer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2533-15</lx:path>
            <lx:label>Industrial Designer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2533-17">
        <xs:annotation>
          <xs:documentation>Interior Designer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2533-17</lx:path>
            <lx:label>Interior Designer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2533-19">
        <xs:annotation>
          <xs:documentation>Illustrator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2533-19</lx:path>
            <lx:label>Illustrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2534">
        <xs:annotation>
          <xs:documentation>JOURNALISTS AND RELATED PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2534</lx:path>
            <lx:label>JOURNALISTS AND RELATED PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2534-11">
        <xs:annotation>
          <xs:documentation>Editor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2534-11</lx:path>
            <lx:label>Editor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2534-13">
        <xs:annotation>
          <xs:documentation>Print Journalist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2534-13</lx:path>
            <lx:label>Print Journalist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2534-15">
        <xs:annotation>
          <xs:documentation>Television Journalist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2534-15</lx:path>
            <lx:label>Television Journalist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2534-17">
        <xs:annotation>
          <xs:documentation>Radio Journalist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2534-17</lx:path>
            <lx:label>Radio Journalist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2534-19">
        <xs:annotation>
          <xs:documentation>Copywriter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2534-19</lx:path>
            <lx:label>Copywriter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2534-21">
        <xs:annotation>
          <xs:documentation>Technical Writer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2534-21</lx:path>
            <lx:label>Technical Writer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2534-79">
        <xs:annotation>
          <xs:documentation>Journalists and Related Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2534-79</lx:path>
            <lx:label>Journalists and Related Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2535">
        <xs:annotation>
          <xs:documentation>AUTHORS AND RELATED PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2535</lx:path>
            <lx:label>AUTHORS AND RELATED PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2535-11">
        <xs:annotation>
          <xs:documentation>Author</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2535-11</lx:path>
            <lx:label>Author</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2535-13">
        <xs:annotation>
          <xs:documentation>Book Editor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2535-13</lx:path>
            <lx:label>Book Editor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2535-15">
        <xs:annotation>
          <xs:documentation>Script Editor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2535-15</lx:path>
            <lx:label>Script Editor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2536">
        <xs:annotation>
          <xs:documentation>FILM, TELEVISION, RADIO AND STAGE DIRECTORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2536</lx:path>
            <lx:label>FILM, TELEVISION, RADIO AND STAGE DIRECTORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2536-11">
        <xs:annotation>
          <xs:documentation>Art Director (Film, Television or Stage)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2536-11</lx:path>
            <lx:label>Art Director (Film, Television or Stage)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2536-13">
        <xs:annotation>
          <xs:documentation>Director (Film, Television, Radio or Stage)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2536-13</lx:path>
            <lx:label>Director (Film, Television, Radio or Stage)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2536-15">
        <xs:annotation>
          <xs:documentation>Director of Photography</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2536-15</lx:path>
            <lx:label>Director of Photography</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2536-17">
        <xs:annotation>
          <xs:documentation>Film and Video Editor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2536-17</lx:path>
            <lx:label>Film and Video Editor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2536-19">
        <xs:annotation>
          <xs:documentation>Stage Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2536-19</lx:path>
            <lx:label>Stage Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2536-21">
        <xs:annotation>
          <xs:documentation>Program Director (Radio or Television)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2536-21</lx:path>
            <lx:label>Program Director (Radio or Television)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2536-23">
        <xs:annotation>
          <xs:documentation>Technical Director</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2536-23</lx:path>
            <lx:label>Technical Director</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2536-79">
        <xs:annotation>
          <xs:documentation>Film, Television, Radio and Stage Directors nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2536-79</lx:path>
            <lx:label>Film, Television, Radio and Stage Directors nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2537">
        <xs:annotation>
          <xs:documentation>MUSICIANS AND RELATED PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2537</lx:path>
            <lx:label>MUSICIANS AND RELATED PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2537-11">
        <xs:annotation>
          <xs:documentation>Music Director</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2537-11</lx:path>
            <lx:label>Music Director</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2537-13">
        <xs:annotation>
          <xs:documentation>Singer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2537-13</lx:path>
            <lx:label>Singer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2537-15">
        <xs:annotation>
          <xs:documentation>Instrumental Musician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2537-15</lx:path>
            <lx:label>Instrumental Musician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2537-17">
        <xs:annotation>
          <xs:documentation>Composer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2537-17</lx:path>
            <lx:label>Composer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2537-79">
        <xs:annotation>
          <xs:documentation>Musicians and Related Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2537-79</lx:path>
            <lx:label>Musicians and Related Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2538">
        <xs:annotation>
          <xs:documentation>ACTORS, DANCERS AND RELATED PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2538</lx:path>
            <lx:label>ACTORS, DANCERS AND RELATED PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2538-11">
        <xs:annotation>
          <xs:documentation>Actor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2538-11</lx:path>
            <lx:label>Actor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2538-13">
        <xs:annotation>
          <xs:documentation>Dancer or Choreographer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2538-13</lx:path>
            <lx:label>Dancer or Choreographer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2538-79">
        <xs:annotation>
          <xs:documentation>Actors, Dancers and Related Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2538-79</lx:path>
            <lx:label>Actors, Dancers and Related Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2539">
        <xs:annotation>
          <xs:documentation>MEDIA PRESENTERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2539</lx:path>
            <lx:label>MEDIA PRESENTERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2539-11">
        <xs:annotation>
          <xs:documentation>Radio Presenter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2539-11</lx:path>
            <lx:label>Radio Presenter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2539-13">
        <xs:annotation>
          <xs:documentation>Television Presenter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2539-13</lx:path>
            <lx:label>Television Presenter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.254</lx:path>
            <lx:label>MISCELLANEOUS PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2541">
        <xs:annotation>
          <xs:documentation>AIR TRANSPORT PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2541</lx:path>
            <lx:label>AIR TRANSPORT PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2541-11">
        <xs:annotation>
          <xs:documentation>Aircraft Pilot</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2541-11</lx:path>
            <lx:label>Aircraft Pilot</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2541-13">
        <xs:annotation>
          <xs:documentation>Air Traffic Controller</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2541-13</lx:path>
            <lx:label>Air Traffic Controller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2541-15">
        <xs:annotation>
          <xs:documentation>Flight Service Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2541-15</lx:path>
            <lx:label>Flight Service Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2541-17">
        <xs:annotation>
          <xs:documentation>Flight Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2541-17</lx:path>
            <lx:label>Flight Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2541-19">
        <xs:annotation>
          <xs:documentation>Flying Instructor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2541-19</lx:path>
            <lx:label>Flying Instructor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2541-79">
        <xs:annotation>
          <xs:documentation>Air Transport Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2541-79</lx:path>
            <lx:label>Air Transport Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2542">
        <xs:annotation>
          <xs:documentation>SEA TRANSPORT PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2542</lx:path>
            <lx:label>SEA TRANSPORT PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2542-11">
        <xs:annotation>
          <xs:documentation>Ship's Master</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2542-11</lx:path>
            <lx:label>Ship's Master</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2542-13">
        <xs:annotation>
          <xs:documentation>Master Fisher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2542-13</lx:path>
            <lx:label>Master Fisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2542-15">
        <xs:annotation>
          <xs:documentation>Ship's Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2542-15</lx:path>
            <lx:label>Ship's Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2542-17">
        <xs:annotation>
          <xs:documentation>Ship's Surveyor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2542-17</lx:path>
            <lx:label>Ship's Surveyor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2542-19">
        <xs:annotation>
          <xs:documentation>Ship's Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2542-19</lx:path>
            <lx:label>Ship's Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2542-79">
        <xs:annotation>
          <xs:documentation>Sea Transport Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2542-79</lx:path>
            <lx:label>Sea Transport Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2543">
        <xs:annotation>
          <xs:documentation>OCCUPATIONAL AND ENVIRONMENTAL HEALTH PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2543</lx:path>
            <lx:label>OCCUPATIONAL AND ENVIRONMENTAL HEALTH PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2543-11">
        <xs:annotation>
          <xs:documentation>Occupational Health and Safety Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2543-11</lx:path>
            <lx:label>Occupational Health and Safety Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2543-13">
        <xs:annotation>
          <xs:documentation>Environmental Health Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2543-13</lx:path>
            <lx:label>Environmental Health Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2549">
        <xs:annotation>
          <xs:documentation>OTHER PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2549</lx:path>
            <lx:label>OTHER PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2549-11">
        <xs:annotation>
          <xs:documentation>Conservator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2549-11</lx:path>
            <lx:label>Conservator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2549-13">
        <xs:annotation>
          <xs:documentation>Electorate Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2549-13</lx:path>
            <lx:label>Electorate Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2549-15">
        <xs:annotation>
          <xs:documentation>Patents Examiner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2549-15</lx:path>
            <lx:label>Patents Examiner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2549-17">
        <xs:annotation>
          <xs:documentation>Oenologist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2549-17</lx:path>
            <lx:label>Oenologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2549-19">
        <xs:annotation>
          <xs:documentation>Recreation Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2549-19</lx:path>
            <lx:label>Recreation Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2549-21">
        <xs:annotation>
          <xs:documentation>Museum or Gallery Curator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2549-21</lx:path>
            <lx:label>Museum or Gallery Curator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2549-79">
        <xs:annotation>
          <xs:documentation>Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.2549-79</lx:path>
            <lx:label>Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3">
        <xs:annotation>
          <xs:documentation>ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3</lx:path>
            <lx:label>ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="31">
        <xs:annotation>
          <xs:documentation>SCIENCE, ENGINEERING AND RELATED ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.31</lx:path>
            <lx:label>SCIENCE, ENGINEERING AND RELATED ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311">
        <xs:annotation>
          <xs:documentation>MEDICAL AND SCIENCE TECHNICAL OFFICERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.311</lx:path>
            <lx:label>MEDICAL AND SCIENCE TECHNICAL OFFICERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3111">
        <xs:annotation>
          <xs:documentation>MEDICAL TECHNICAL OFFICERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3111</lx:path>
            <lx:label>MEDICAL TECHNICAL OFFICERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3111-11">
        <xs:annotation>
          <xs:documentation>Medical Laboratory Technical Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3111-11</lx:path>
            <lx:label>Medical Laboratory Technical Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3111-79">
        <xs:annotation>
          <xs:documentation>Medical Technical Officers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3111-79</lx:path>
            <lx:label>Medical Technical Officers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3112">
        <xs:annotation>
          <xs:documentation>SCIENCE TECHNICAL OFFICERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3112</lx:path>
            <lx:label>SCIENCE TECHNICAL OFFICERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3112-11">
        <xs:annotation>
          <xs:documentation>Chemistry Technical Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3112-11</lx:path>
            <lx:label>Chemistry Technical Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3112-13">
        <xs:annotation>
          <xs:documentation>Earth Science Technical Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3112-13</lx:path>
            <lx:label>Earth Science Technical Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3112-15">
        <xs:annotation>
          <xs:documentation>Life Science Technical Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3112-15</lx:path>
            <lx:label>Life Science Technical Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3112-17">
        <xs:annotation>
          <xs:documentation>Agricultural Technical Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3112-17</lx:path>
            <lx:label>Agricultural Technical Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3112-79">
        <xs:annotation>
          <xs:documentation>Science Technical Officers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3112-79</lx:path>
            <lx:label>Science Technical Officers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312">
        <xs:annotation>
          <xs:documentation>BUILDING AND ENGINEERING ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.312</lx:path>
            <lx:label>BUILDING AND ENGINEERING ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3121">
        <xs:annotation>
          <xs:documentation>BUILDING, ARCHITECTURAL AND SURVEYING ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3121</lx:path>
            <lx:label>BUILDING, ARCHITECTURAL AND SURVEYING ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3121-11">
        <xs:annotation>
          <xs:documentation>Building Associate</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3121-11</lx:path>
            <lx:label>Building Associate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3121-13">
        <xs:annotation>
          <xs:documentation>Architectural Associate</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3121-13</lx:path>
            <lx:label>Architectural Associate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3121-15">
        <xs:annotation>
          <xs:documentation>Surveying and Cartographic Associate</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3121-15</lx:path>
            <lx:label>Surveying and Cartographic Associate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3121-17">
        <xs:annotation>
          <xs:documentation>Building Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3121-17</lx:path>
            <lx:label>Building Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3121-19">
        <xs:annotation>
          <xs:documentation>Plumbing Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3121-19</lx:path>
            <lx:label>Plumbing Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3121-21">
        <xs:annotation>
          <xs:documentation>Plumbing Engineering Associate</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3121-21</lx:path>
            <lx:label>Plumbing Engineering Associate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3122">
        <xs:annotation>
          <xs:documentation>CIVIL ENGINEERING ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3122</lx:path>
            <lx:label>CIVIL ENGINEERING ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3122-11">
        <xs:annotation>
          <xs:documentation>Civil Engineering Associate</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3122-11</lx:path>
            <lx:label>Civil Engineering Associate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3122-13">
        <xs:annotation>
          <xs:documentation>Civil Engineering Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3122-13</lx:path>
            <lx:label>Civil Engineering Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3123">
        <xs:annotation>
          <xs:documentation>ELECTRICAL ENGINEERING ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3123</lx:path>
            <lx:label>ELECTRICAL ENGINEERING ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3123-11">
        <xs:annotation>
          <xs:documentation>Electrical Engineering Associate</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3123-11</lx:path>
            <lx:label>Electrical Engineering Associate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3123-13">
        <xs:annotation>
          <xs:documentation>Electrical Engineering Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3123-13</lx:path>
            <lx:label>Electrical Engineering Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3124">
        <xs:annotation>
          <xs:documentation>ELECTRONIC ENGINEERING ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3124</lx:path>
            <lx:label>ELECTRONIC ENGINEERING ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3124-11">
        <xs:annotation>
          <xs:documentation>Electronic Engineering Associate</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3124-11</lx:path>
            <lx:label>Electronic Engineering Associate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3124-13">
        <xs:annotation>
          <xs:documentation>Electronic Engineering Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3124-13</lx:path>
            <lx:label>Electronic Engineering Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3125">
        <xs:annotation>
          <xs:documentation>MECHANICAL ENGINEERING ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3125</lx:path>
            <lx:label>MECHANICAL ENGINEERING ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3125-11">
        <xs:annotation>
          <xs:documentation>Mechanical Engineering Associate</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3125-11</lx:path>
            <lx:label>Mechanical Engineering Associate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3125-13">
        <xs:annotation>
          <xs:documentation>Mechanical Engineering Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3125-13</lx:path>
            <lx:label>Mechanical Engineering Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3129">
        <xs:annotation>
          <xs:documentation>OTHER BUILDING AND ENGINEERING ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3129</lx:path>
            <lx:label>OTHER BUILDING AND ENGINEERING ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3129-11">
        <xs:annotation>
          <xs:documentation>Biomedical Engineering Associate</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3129-11</lx:path>
            <lx:label>Biomedical Engineering Associate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3129-13">
        <xs:annotation>
          <xs:documentation>Metallurgical and Materials Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3129-13</lx:path>
            <lx:label>Metallurgical and Materials Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3129-15">
        <xs:annotation>
          <xs:documentation>Mine Deputy</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3129-15</lx:path>
            <lx:label>Mine Deputy</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3129-79">
        <xs:annotation>
          <xs:documentation>Building and Engineering Associate Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3129-79</lx:path>
            <lx:label>Building and Engineering Associate Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="32">
        <xs:annotation>
          <xs:documentation>BUSINESS AND ADMINISTRATION ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.32</lx:path>
            <lx:label>BUSINESS AND ADMINISTRATION ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="321">
        <xs:annotation>
          <xs:documentation>FINANCE ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.321</lx:path>
            <lx:label>FINANCE ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3211">
        <xs:annotation>
          <xs:documentation>BRANCH ACCOUNTANTS AND MANAGERS (FINANCIAL INSTITUTION)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3211</lx:path>
            <lx:label>BRANCH ACCOUNTANTS AND MANAGERS (FINANCIAL INSTITUTION)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3211-11">
        <xs:annotation>
          <xs:documentation>Branch Accountant (Financial Institution)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3211-11</lx:path>
            <lx:label>Branch Accountant (Financial Institution)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3211-13">
        <xs:annotation>
          <xs:documentation>Financial Institution Branch Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3211-13</lx:path>
            <lx:label>Financial Institution Branch Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3212">
        <xs:annotation>
          <xs:documentation>FINANCIAL DEALERS AND BROKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3212</lx:path>
            <lx:label>FINANCIAL DEALERS AND BROKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3212-11">
        <xs:annotation>
          <xs:documentation>Stockbroking Dealer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3212-11</lx:path>
            <lx:label>Stockbroking Dealer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3212-13">
        <xs:annotation>
          <xs:documentation>Futures Trader</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3212-13</lx:path>
            <lx:label>Futures Trader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3212-15">
        <xs:annotation>
          <xs:documentation>Financial Market Dealer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3212-15</lx:path>
            <lx:label>Financial Market Dealer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3212-17">
        <xs:annotation>
          <xs:documentation>Commodities Trader</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3212-17</lx:path>
            <lx:label>Commodities Trader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3212-19">
        <xs:annotation>
          <xs:documentation>Insurance Broker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3212-19</lx:path>
            <lx:label>Insurance Broker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3212-21">
        <xs:annotation>
          <xs:documentation>Bookmaker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3212-21</lx:path>
            <lx:label>Bookmaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3212-79">
        <xs:annotation>
          <xs:documentation>Financial Dealers and Brokers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3212-79</lx:path>
            <lx:label>Financial Dealers and Brokers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3213">
        <xs:annotation>
          <xs:documentation>FINANCIAL INVESTMENT ADVISERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3213</lx:path>
            <lx:label>FINANCIAL INVESTMENT ADVISERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3213-11">
        <xs:annotation>
          <xs:documentation>Financial Investment Adviser</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3213-11</lx:path>
            <lx:label>Financial Investment Adviser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="329">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS BUSINESS AND ADMINISTRATION ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.329</lx:path>
            <lx:label>MISCELLANEOUS BUSINESS AND ADMINISTRATION ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3291">
        <xs:annotation>
          <xs:documentation>OFFICE MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3291</lx:path>
            <lx:label>OFFICE MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3291-11">
        <xs:annotation>
          <xs:documentation>Office Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3291-11</lx:path>
            <lx:label>Office Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3292">
        <xs:annotation>
          <xs:documentation>PROJECT AND PROGRAM ADMINISTRATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3292</lx:path>
            <lx:label>PROJECT AND PROGRAM ADMINISTRATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3292-11">
        <xs:annotation>
          <xs:documentation>Project or Program Administrator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3292-11</lx:path>
            <lx:label>Project or Program Administrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3293">
        <xs:annotation>
          <xs:documentation>REAL ESTATE ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3293</lx:path>
            <lx:label>REAL ESTATE ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3293-11">
        <xs:annotation>
          <xs:documentation>Real Estate Agency Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3293-11</lx:path>
            <lx:label>Real Estate Agency Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3293-13">
        <xs:annotation>
          <xs:documentation>Property Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3293-13</lx:path>
            <lx:label>Property Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3293-15">
        <xs:annotation>
          <xs:documentation>Real Estate Salesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3293-15</lx:path>
            <lx:label>Real Estate Salesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3294">
        <xs:annotation>
          <xs:documentation>COMPUTING SUPPORT TECHNICIANS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3294</lx:path>
            <lx:label>COMPUTING SUPPORT TECHNICIANS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3294-11">
        <xs:annotation>
          <xs:documentation>Computing Support Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3294-11</lx:path>
            <lx:label>Computing Support Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="33">
        <xs:annotation>
          <xs:documentation>MANAGING SUPERVISORS (SALES AND SERVICE)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.33</lx:path>
            <lx:label>MANAGING SUPERVISORS (SALES AND SERVICE)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="331">
        <xs:annotation>
          <xs:documentation>SHOP MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.331</lx:path>
            <lx:label>SHOP MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3311">
        <xs:annotation>
          <xs:documentation>SHOP MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3311</lx:path>
            <lx:label>SHOP MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3311-11">
        <xs:annotation>
          <xs:documentation>Shop Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3311-11</lx:path>
            <lx:label>Shop Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="332">
        <xs:annotation>
          <xs:documentation>HOSPITALITY AND ACCOMMODATION MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.332</lx:path>
            <lx:label>HOSPITALITY AND ACCOMMODATION MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3321">
        <xs:annotation>
          <xs:documentation>RESTAURANT AND CATERING MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3321</lx:path>
            <lx:label>RESTAURANT AND CATERING MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3321-11">
        <xs:annotation>
          <xs:documentation>Restaurant and Catering Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3321-11</lx:path>
            <lx:label>Restaurant and Catering Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3322">
        <xs:annotation>
          <xs:documentation>CHEFS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3322</lx:path>
            <lx:label>CHEFS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3322-01">
        <xs:annotation>
          <xs:documentation>Head Chef</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3322-01</lx:path>
            <lx:label>Head Chef</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3322-11">
        <xs:annotation>
          <xs:documentation>Chef</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3322-11</lx:path>
            <lx:label>Chef</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3323">
        <xs:annotation>
          <xs:documentation>HOTEL AND MOTEL MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3323</lx:path>
            <lx:label>HOTEL AND MOTEL MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3323-11">
        <xs:annotation>
          <xs:documentation>Hotel or Motel Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3323-11</lx:path>
            <lx:label>Hotel or Motel Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3324">
        <xs:annotation>
          <xs:documentation>CLUB MANAGERS (LICENSED PREMISES)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3324</lx:path>
            <lx:label>CLUB MANAGERS (LICENSED PREMISES)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3324-11">
        <xs:annotation>
          <xs:documentation>Club Manager (Licensed Premises)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3324-11</lx:path>
            <lx:label>Club Manager (Licensed Premises)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3325">
        <xs:annotation>
          <xs:documentation>CARAVAN PARK AND CAMPING GROUND MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3325</lx:path>
            <lx:label>CARAVAN PARK AND CAMPING GROUND MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3325-11">
        <xs:annotation>
          <xs:documentation>Caravan Park and Camping Ground Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3325-11</lx:path>
            <lx:label>Caravan Park and Camping Ground Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3329">
        <xs:annotation>
          <xs:documentation>OTHER HOSPITALITY AND ACCOMMODATION MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3329</lx:path>
            <lx:label>OTHER HOSPITALITY AND ACCOMMODATION MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3329-11">
        <xs:annotation>
          <xs:documentation>Other Hospitality and Accommodation Managers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3329-11</lx:path>
            <lx:label>Other Hospitality and Accommodation Managers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="339">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS MANAGING SUPERVISORS (SALES AND SERVICE)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.339</lx:path>
            <lx:label>MISCELLANEOUS MANAGING SUPERVISORS (SALES AND SERVICE)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3391">
        <xs:annotation>
          <xs:documentation>SPORT AND RECREATION MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3391</lx:path>
            <lx:label>SPORT AND RECREATION MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3391-11">
        <xs:annotation>
          <xs:documentation>Fitness Centre Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3391-11</lx:path>
            <lx:label>Fitness Centre Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3391-13">
        <xs:annotation>
          <xs:documentation>Other Sports Centre Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3391-13</lx:path>
            <lx:label>Other Sports Centre Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3391-15">
        <xs:annotation>
          <xs:documentation>Amusement Centre Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3391-15</lx:path>
            <lx:label>Amusement Centre Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3391-79">
        <xs:annotation>
          <xs:documentation>Sport and Recreation Managers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3391-79</lx:path>
            <lx:label>Sport and Recreation Managers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3392">
        <xs:annotation>
          <xs:documentation>CUSTOMER SERVICE MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3392</lx:path>
            <lx:label>CUSTOMER SERVICE MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3392-11">
        <xs:annotation>
          <xs:documentation>Customer Service Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3392-11</lx:path>
            <lx:label>Customer Service Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3393">
        <xs:annotation>
          <xs:documentation>TRANSPORT COMPANY MANAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3393</lx:path>
            <lx:label>TRANSPORT COMPANY MANAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3393-11">
        <xs:annotation>
          <xs:documentation>Transport Company Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3393-11</lx:path>
            <lx:label>Transport Company Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3399">
        <xs:annotation>
          <xs:documentation>OTHER MANAGING SUPERVISORS (SALES AND SERVICE)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3399</lx:path>
            <lx:label>OTHER MANAGING SUPERVISORS (SALES AND SERVICE)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3399-11">
        <xs:annotation>
          <xs:documentation>Post Office Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3399-11</lx:path>
            <lx:label>Post Office Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3399-13">
        <xs:annotation>
          <xs:documentation>Railway Station Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3399-13</lx:path>
            <lx:label>Railway Station Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3399-15">
        <xs:annotation>
          <xs:documentation>Betting Agency Branch Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3399-15</lx:path>
            <lx:label>Betting Agency Branch Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3399-17">
        <xs:annotation>
          <xs:documentation>Hair and Beauty Salon Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3399-17</lx:path>
            <lx:label>Hair and Beauty Salon Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3399-19">
        <xs:annotation>
          <xs:documentation>Car Rental Agency Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3399-19</lx:path>
            <lx:label>Car Rental Agency Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3399-21">
        <xs:annotation>
          <xs:documentation>Fleet Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3399-21</lx:path>
            <lx:label>Fleet Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3399-23">
        <xs:annotation>
          <xs:documentation>Stock and Station Agent</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3399-23</lx:path>
            <lx:label>Stock and Station Agent</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3399-25">
        <xs:annotation>
          <xs:documentation>Travel Agency Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3399-25</lx:path>
            <lx:label>Travel Agency Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3399-27">
        <xs:annotation>
          <xs:documentation>Theatre or Cinema Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3399-27</lx:path>
            <lx:label>Theatre or Cinema Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3399-29">
        <xs:annotation>
          <xs:documentation>Funeral Director</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3399-29</lx:path>
            <lx:label>Funeral Director</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3399-79">
        <xs:annotation>
          <xs:documentation>Managing Supervisors (Sales and Service) nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3399-79</lx:path>
            <lx:label>Managing Supervisors (Sales and Service) nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="34">
        <xs:annotation>
          <xs:documentation>HEALTH AND WELFARE ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.34</lx:path>
            <lx:label>HEALTH AND WELFARE ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="341">
        <xs:annotation>
          <xs:documentation>ENROLLED NURSES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.341</lx:path>
            <lx:label>ENROLLED NURSES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3411">
        <xs:annotation>
          <xs:documentation>ENROLLED NURSES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3411</lx:path>
            <lx:label>ENROLLED NURSES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3411-11">
        <xs:annotation>
          <xs:documentation>Enrolled Nurse</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3411-11</lx:path>
            <lx:label>Enrolled Nurse</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342">
        <xs:annotation>
          <xs:documentation>WELFARE ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.342</lx:path>
            <lx:label>WELFARE ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3421">
        <xs:annotation>
          <xs:documentation>WELFARE ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3421</lx:path>
            <lx:label>WELFARE ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3421-11">
        <xs:annotation>
          <xs:documentation>Parole or Probation Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3421-11</lx:path>
            <lx:label>Parole or Probation Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3421-13">
        <xs:annotation>
          <xs:documentation>Youth Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3421-13</lx:path>
            <lx:label>Youth Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3421-15">
        <xs:annotation>
          <xs:documentation>Residential Care Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3421-15</lx:path>
            <lx:label>Residential Care Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3421-17">
        <xs:annotation>
          <xs:documentation>Disabilities Services Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3421-17</lx:path>
            <lx:label>Disabilities Services Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3421-19">
        <xs:annotation>
          <xs:documentation>Family Support Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3421-19</lx:path>
            <lx:label>Family Support Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="349">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS HEALTH AND WELFARE ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.349</lx:path>
            <lx:label>MISCELLANEOUS HEALTH AND WELFARE ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3491">
        <xs:annotation>
          <xs:documentation>AMBULANCE OFFICERS AND PARAMEDICS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3491</lx:path>
            <lx:label>AMBULANCE OFFICERS AND PARAMEDICS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3491-11">
        <xs:annotation>
          <xs:documentation>Ambulance Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3491-11</lx:path>
            <lx:label>Ambulance Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3491-13">
        <xs:annotation>
          <xs:documentation>Intensive Care Ambulance Paramedic</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3491-13</lx:path>
            <lx:label>Intensive Care Ambulance Paramedic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3492">
        <xs:annotation>
          <xs:documentation>DENTAL ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3492</lx:path>
            <lx:label>DENTAL ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3492-11">
        <xs:annotation>
          <xs:documentation>Dental Therapist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3492-11</lx:path>
            <lx:label>Dental Therapist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3492-13">
        <xs:annotation>
          <xs:documentation>Dental Hygienist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3492-13</lx:path>
            <lx:label>Dental Hygienist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3492-15">
        <xs:annotation>
          <xs:documentation>Dental Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3492-15</lx:path>
            <lx:label>Dental Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3493">
        <xs:annotation>
          <xs:documentation>ABORIGINAL AND TORRES STRAIT ISLANDER HEALTH WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3493</lx:path>
            <lx:label>ABORIGINAL AND TORRES STRAIT ISLANDER HEALTH WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3493-11">
        <xs:annotation>
          <xs:documentation>Aboriginal and Torres Strait Islander Health Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3493-11</lx:path>
            <lx:label>Aboriginal and Torres Strait Islander Health Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3494">
        <xs:annotation>
          <xs:documentation>MASSAGE THERAPISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3494</lx:path>
            <lx:label>MASSAGE THERAPISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3494-11">
        <xs:annotation>
          <xs:documentation>Massage Therapist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3494-11</lx:path>
            <lx:label>Massage Therapist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="39">
        <xs:annotation>
          <xs:documentation>OTHER ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.39</lx:path>
            <lx:label>OTHER ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="391">
        <xs:annotation>
          <xs:documentation>POLICE OFFICERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.391</lx:path>
            <lx:label>POLICE OFFICERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3911">
        <xs:annotation>
          <xs:documentation>POLICE OFFICERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3911</lx:path>
            <lx:label>POLICE OFFICERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3911-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Police Officers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3911-01</lx:path>
            <lx:label>Supervisor, Police Officers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3911-11">
        <xs:annotation>
          <xs:documentation>Police Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3911-11</lx:path>
            <lx:label>Police Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.399</lx:path>
            <lx:label>MISCELLANEOUS ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3991">
        <xs:annotation>
          <xs:documentation>PRIMARY PRODUCTS INSPECTORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3991</lx:path>
            <lx:label>PRIMARY PRODUCTS INSPECTORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3991-11">
        <xs:annotation>
          <xs:documentation>Primary Products Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3991-11</lx:path>
            <lx:label>Primary Products Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3992">
        <xs:annotation>
          <xs:documentation>SAFETY INSPECTORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3992</lx:path>
            <lx:label>SAFETY INSPECTORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3992-11">
        <xs:annotation>
          <xs:documentation>Safety Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3992-11</lx:path>
            <lx:label>Safety Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3993">
        <xs:annotation>
          <xs:documentation>SPORTSPERSONS, COACHES AND RELATED SUPPORT WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3993</lx:path>
            <lx:label>SPORTSPERSONS, COACHES AND RELATED SUPPORT WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3993-11">
        <xs:annotation>
          <xs:documentation>Jockey</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3993-11</lx:path>
            <lx:label>Jockey</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3993-13">
        <xs:annotation>
          <xs:documentation>Golfer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3993-13</lx:path>
            <lx:label>Golfer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3993-15">
        <xs:annotation>
          <xs:documentation>Footballer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3993-15</lx:path>
            <lx:label>Footballer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3993-17">
        <xs:annotation>
          <xs:documentation>Other Sportsperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3993-17</lx:path>
            <lx:label>Other Sportsperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3993-19">
        <xs:annotation>
          <xs:documentation>Gymnastics Coach</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3993-19</lx:path>
            <lx:label>Gymnastics Coach</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3993-21">
        <xs:annotation>
          <xs:documentation>Tennis Coach</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3993-21</lx:path>
            <lx:label>Tennis Coach</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3993-23">
        <xs:annotation>
          <xs:documentation>Swimming Coach</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3993-23</lx:path>
            <lx:label>Swimming Coach</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3993-25">
        <xs:annotation>
          <xs:documentation>Horseriding Coach</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3993-25</lx:path>
            <lx:label>Horseriding Coach</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3993-27">
        <xs:annotation>
          <xs:documentation>Other Sports Coach</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3993-27</lx:path>
            <lx:label>Other Sports Coach</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3993-29">
        <xs:annotation>
          <xs:documentation>Sports Development Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3993-29</lx:path>
            <lx:label>Sports Development Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3993-31">
        <xs:annotation>
          <xs:documentation>Sports Umpire</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3993-31</lx:path>
            <lx:label>Sports Umpire</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3993-33">
        <xs:annotation>
          <xs:documentation>Horse or Dog Racing Official</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3993-33</lx:path>
            <lx:label>Horse or Dog Racing Official</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3993-35">
        <xs:annotation>
          <xs:documentation>Other Sports Official</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3993-35</lx:path>
            <lx:label>Other Sports Official</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3994">
        <xs:annotation>
          <xs:documentation>SENIOR NON-COMMISSIONED DEFENCE FORCE OFFICERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3994</lx:path>
            <lx:label>SENIOR NON-COMMISSIONED DEFENCE FORCE OFFICERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3994-11">
        <xs:annotation>
          <xs:documentation>Senior Non-Commissioned Defence Force Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3994-11</lx:path>
            <lx:label>Senior Non-Commissioned Defence Force Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3995">
        <xs:annotation>
          <xs:documentation>SENIOR FIRE FIGHTERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3995</lx:path>
            <lx:label>SENIOR FIRE FIGHTERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3995-11">
        <xs:annotation>
          <xs:documentation>Senior Fire Fighter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3995-11</lx:path>
            <lx:label>Senior Fire Fighter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3996">
        <xs:annotation>
          <xs:documentation>RETAIL BUYERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3996</lx:path>
            <lx:label>RETAIL BUYERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3996-11">
        <xs:annotation>
          <xs:documentation>Retail Buyer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3996-11</lx:path>
            <lx:label>Retail Buyer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3997">
        <xs:annotation>
          <xs:documentation>LIBRARY TECHNICIANS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3997</lx:path>
            <lx:label>LIBRARY TECHNICIANS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3997-11">
        <xs:annotation>
          <xs:documentation>Library Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3997-11</lx:path>
            <lx:label>Library Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3999">
        <xs:annotation>
          <xs:documentation>OTHER MISCELLANEOUS ASSOCIATE PROFESSIONALS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3999</lx:path>
            <lx:label>OTHER MISCELLANEOUS ASSOCIATE PROFESSIONALS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3999-11">
        <xs:annotation>
          <xs:documentation>Interior Decorator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3999-11</lx:path>
            <lx:label>Interior Decorator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3999-13">
        <xs:annotation>
          <xs:documentation>Museum or Art Gallery Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3999-13</lx:path>
            <lx:label>Museum or Art Gallery Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3999-15">
        <xs:annotation>
          <xs:documentation>Radio Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3999-15</lx:path>
            <lx:label>Radio Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3999-17">
        <xs:annotation>
          <xs:documentation>Private Investigator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3999-17</lx:path>
            <lx:label>Private Investigator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3999-19">
        <xs:annotation>
          <xs:documentation>Security Adviser</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3999-19</lx:path>
            <lx:label>Security Adviser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3999-79">
        <xs:annotation>
          <xs:documentation>Associate Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.3999-79</lx:path>
            <lx:label>Associate Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4">
        <xs:annotation>
          <xs:documentation>TRADESPERSONS AND RELATED WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4</lx:path>
            <lx:label>TRADESPERSONS AND RELATED WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="41">
        <xs:annotation>
          <xs:documentation>MECHANICAL AND FABRICATION ENGINEERING TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.41</lx:path>
            <lx:label>MECHANICAL AND FABRICATION ENGINEERING TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411">
        <xs:annotation>
          <xs:documentation>MECHANICAL ENGINEERING TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.411</lx:path>
            <lx:label>MECHANICAL ENGINEERING TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4111">
        <xs:annotation>
          <xs:documentation>GENERAL MECHANICAL ENGINEERING TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4111</lx:path>
            <lx:label>GENERAL MECHANICAL ENGINEERING TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4111-01">
        <xs:annotation>
          <xs:documentation>Supervisor, General Mechanical Engineering Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4111-01</lx:path>
            <lx:label>Supervisor, General Mechanical Engineering Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4111-11">
        <xs:annotation>
          <xs:documentation>General Mechanical Engineering Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4111-11</lx:path>
            <lx:label>General Mechanical Engineering Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4111-81">
        <xs:annotation>
          <xs:documentation>Apprentice General Mechanical Engineering Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4111-81</lx:path>
            <lx:label>Apprentice General Mechanical Engineering Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4112">
        <xs:annotation>
          <xs:documentation>METAL FITTERS AND MACHINISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4112</lx:path>
            <lx:label>METAL FITTERS AND MACHINISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4112-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Metal Fitters and Machinists</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4112-01</lx:path>
            <lx:label>Supervisor, Metal Fitters and Machinists</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4112-11">
        <xs:annotation>
          <xs:documentation>Fitter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4112-11</lx:path>
            <lx:label>Fitter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4112-13">
        <xs:annotation>
          <xs:documentation>Metal Machinist (First Class)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4112-13</lx:path>
            <lx:label>Metal Machinist (First Class)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4112-15">
        <xs:annotation>
          <xs:documentation>Textile, Clothing or Footwear Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4112-15</lx:path>
            <lx:label>Textile, Clothing or Footwear Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4112-81">
        <xs:annotation>
          <xs:documentation>Apprentice Fitter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4112-81</lx:path>
            <lx:label>Apprentice Fitter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4112-83">
        <xs:annotation>
          <xs:documentation>Apprentice Metal Machinist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4112-83</lx:path>
            <lx:label>Apprentice Metal Machinist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4112-85">
        <xs:annotation>
          <xs:documentation>Apprentice Textile, Clothing or Footwear Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4112-85</lx:path>
            <lx:label>Apprentice Textile, Clothing or Footwear Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4113">
        <xs:annotation>
          <xs:documentation>TOOLMAKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4113</lx:path>
            <lx:label>TOOLMAKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4113-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Toolmakers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4113-01</lx:path>
            <lx:label>Supervisor, Toolmakers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4113-11">
        <xs:annotation>
          <xs:documentation>Toolmaker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4113-11</lx:path>
            <lx:label>Toolmaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4113-81">
        <xs:annotation>
          <xs:documentation>Apprentice Toolmaker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4113-81</lx:path>
            <lx:label>Apprentice Toolmaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4114">
        <xs:annotation>
          <xs:documentation>AIRCRAFT MAINTENANCE ENGINEERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4114</lx:path>
            <lx:label>AIRCRAFT MAINTENANCE ENGINEERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4114-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Aircraft Maintenance Engineers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4114-01</lx:path>
            <lx:label>Supervisor, Aircraft Maintenance Engineers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4114-11">
        <xs:annotation>
          <xs:documentation>Aircraft Maintenance Engineer (Mechanical)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4114-11</lx:path>
            <lx:label>Aircraft Maintenance Engineer (Mechanical)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4114-13">
        <xs:annotation>
          <xs:documentation>Aircraft Maintenance Engineer (Structures)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4114-13</lx:path>
            <lx:label>Aircraft Maintenance Engineer (Structures)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4114-15">
        <xs:annotation>
          <xs:documentation>Aircraft Maintenance Engineer (Avionics)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4114-15</lx:path>
            <lx:label>Aircraft Maintenance Engineer (Avionics)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4114-81">
        <xs:annotation>
          <xs:documentation>Apprentice Aircraft Maintenance Engineer (Mechanical)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4114-81</lx:path>
            <lx:label>Apprentice Aircraft Maintenance Engineer (Mechanical)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4114-83">
        <xs:annotation>
          <xs:documentation>Apprentice Aircraft Maintenance Engineer (Structures)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4114-83</lx:path>
            <lx:label>Apprentice Aircraft Maintenance Engineer (Structures)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4114-85">
        <xs:annotation>
          <xs:documentation>Apprentice Aircraft Maintenance Engineer (Avionics)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4114-85</lx:path>
            <lx:label>Apprentice Aircraft Maintenance Engineer (Avionics)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4115">
        <xs:annotation>
          <xs:documentation>PRECISION METAL TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4115</lx:path>
            <lx:label>PRECISION METAL TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4115-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Precision Metal Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4115-01</lx:path>
            <lx:label>Supervisor, Precision Metal Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4115-11">
        <xs:annotation>
          <xs:documentation>Precision Instrument Maker and Repairer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4115-11</lx:path>
            <lx:label>Precision Instrument Maker and Repairer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4115-13">
        <xs:annotation>
          <xs:documentation>Watch and Clock Maker and Repairer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4115-13</lx:path>
            <lx:label>Watch and Clock Maker and Repairer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4115-15">
        <xs:annotation>
          <xs:documentation>Locksmith</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4115-15</lx:path>
            <lx:label>Locksmith</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4115-17">
        <xs:annotation>
          <xs:documentation>Saw Maker and Repairer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4115-17</lx:path>
            <lx:label>Saw Maker and Repairer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4115-19">
        <xs:annotation>
          <xs:documentation>Gunsmith</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4115-19</lx:path>
            <lx:label>Gunsmith</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4115-21">
        <xs:annotation>
          <xs:documentation>Engraver</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4115-21</lx:path>
            <lx:label>Engraver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4115-81">
        <xs:annotation>
          <xs:documentation>Apprentice Precision Metal Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4115-81</lx:path>
            <lx:label>Apprentice Precision Metal Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="412">
        <xs:annotation>
          <xs:documentation>FABRICATION ENGINEERING TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.412</lx:path>
            <lx:label>FABRICATION ENGINEERING TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4121">
        <xs:annotation>
          <xs:documentation>GENERAL FABRICATION ENGINEERING TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4121</lx:path>
            <lx:label>GENERAL FABRICATION ENGINEERING TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4121-01">
        <xs:annotation>
          <xs:documentation>Supervisor, General Fabrication Engineering Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4121-01</lx:path>
            <lx:label>Supervisor, General Fabrication Engineering Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4121-11">
        <xs:annotation>
          <xs:documentation>General Fabrication Engineering Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4121-11</lx:path>
            <lx:label>General Fabrication Engineering Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4121-81">
        <xs:annotation>
          <xs:documentation>Apprentice General Fabrication Engineering Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4121-81</lx:path>
            <lx:label>Apprentice General Fabrication Engineering Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4122">
        <xs:annotation>
          <xs:documentation>STRUCTURAL STEEL AND WELDING TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4122</lx:path>
            <lx:label>STRUCTURAL STEEL AND WELDING TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4122-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Structural Steel and Welding Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4122-01</lx:path>
            <lx:label>Supervisor, Structural Steel and Welding Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4122-11">
        <xs:annotation>
          <xs:documentation>Metal Fabricator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4122-11</lx:path>
            <lx:label>Metal Fabricator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4122-13">
        <xs:annotation>
          <xs:documentation>Pressure Welder</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4122-13</lx:path>
            <lx:label>Pressure Welder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4122-15">
        <xs:annotation>
          <xs:documentation>Welder (First Class)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4122-15</lx:path>
            <lx:label>Welder (First Class)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4122-81">
        <xs:annotation>
          <xs:documentation>Apprentice Metal Fabricator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4122-81</lx:path>
            <lx:label>Apprentice Metal Fabricator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4122-83">
        <xs:annotation>
          <xs:documentation>Apprentice Welder</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4122-83</lx:path>
            <lx:label>Apprentice Welder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4123">
        <xs:annotation>
          <xs:documentation>FORGING TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4123</lx:path>
            <lx:label>FORGING TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4123-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Forging Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4123-01</lx:path>
            <lx:label>Supervisor, Forging Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4123-11">
        <xs:annotation>
          <xs:documentation>Blacksmith</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4123-11</lx:path>
            <lx:label>Blacksmith</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4123-13">
        <xs:annotation>
          <xs:documentation>Farrier</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4123-13</lx:path>
            <lx:label>Farrier</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4123-81">
        <xs:annotation>
          <xs:documentation>Apprentice Blacksmith</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4123-81</lx:path>
            <lx:label>Apprentice Blacksmith</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4123-83">
        <xs:annotation>
          <xs:documentation>Apprentice Farrier</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4123-83</lx:path>
            <lx:label>Apprentice Farrier</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4124">
        <xs:annotation>
          <xs:documentation>SHEETMETAL TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4124</lx:path>
            <lx:label>SHEETMETAL TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4124-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Sheetmetal Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4124-01</lx:path>
            <lx:label>Supervisor, Sheetmetal Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4124-11">
        <xs:annotation>
          <xs:documentation>Sheetmetal Worker (First Class)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4124-11</lx:path>
            <lx:label>Sheetmetal Worker (First Class)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4124-81">
        <xs:annotation>
          <xs:documentation>Apprentice Sheetmetal Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4124-81</lx:path>
            <lx:label>Apprentice Sheetmetal Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4125">
        <xs:annotation>
          <xs:documentation>METAL CASTING TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4125</lx:path>
            <lx:label>METAL CASTING TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4125-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Metal Casting Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4125-01</lx:path>
            <lx:label>Supervisor, Metal Casting Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4125-11">
        <xs:annotation>
          <xs:documentation>Metal Casting Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4125-11</lx:path>
            <lx:label>Metal Casting Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4125-81">
        <xs:annotation>
          <xs:documentation>Apprentice Metal Casting Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4125-81</lx:path>
            <lx:label>Apprentice Metal Casting Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4126">
        <xs:annotation>
          <xs:documentation>METAL FINISHING TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4126</lx:path>
            <lx:label>METAL FINISHING TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4126-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Metal Finishing Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4126-01</lx:path>
            <lx:label>Supervisor, Metal Finishing Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4126-11">
        <xs:annotation>
          <xs:documentation>Metal Polisher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4126-11</lx:path>
            <lx:label>Metal Polisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4126-13">
        <xs:annotation>
          <xs:documentation>Electroplater (First Class)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4126-13</lx:path>
            <lx:label>Electroplater (First Class)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4126-81">
        <xs:annotation>
          <xs:documentation>Apprentice Metal Polisher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4126-81</lx:path>
            <lx:label>Apprentice Metal Polisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4126-83">
        <xs:annotation>
          <xs:documentation>Apprentice Electroplater</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4126-83</lx:path>
            <lx:label>Apprentice Electroplater</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="42">
        <xs:annotation>
          <xs:documentation>AUTOMOTIVE TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.42</lx:path>
            <lx:label>AUTOMOTIVE TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="421">
        <xs:annotation>
          <xs:documentation>AUTOMOTIVE TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.421</lx:path>
            <lx:label>AUTOMOTIVE TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4211">
        <xs:annotation>
          <xs:documentation>MOTOR MECHANICS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4211</lx:path>
            <lx:label>MOTOR MECHANICS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4211-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Motor Mechanics</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4211-01</lx:path>
            <lx:label>Supervisor, Motor Mechanics</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4211-11">
        <xs:annotation>
          <xs:documentation>Motor Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4211-11</lx:path>
            <lx:label>Motor Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4211-81">
        <xs:annotation>
          <xs:documentation>Apprentice Motor Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4211-81</lx:path>
            <lx:label>Apprentice Motor Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4212">
        <xs:annotation>
          <xs:documentation>AUTOMOTIVE ELECTRICIANS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4212</lx:path>
            <lx:label>AUTOMOTIVE ELECTRICIANS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4212-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Automotive Electricians</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4212-01</lx:path>
            <lx:label>Supervisor, Automotive Electricians</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4212-11">
        <xs:annotation>
          <xs:documentation>Automotive Electrician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4212-11</lx:path>
            <lx:label>Automotive Electrician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4212-81">
        <xs:annotation>
          <xs:documentation>Apprentice Automotive Electrician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4212-81</lx:path>
            <lx:label>Apprentice Automotive Electrician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4213">
        <xs:annotation>
          <xs:documentation>PANEL BEATERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4213</lx:path>
            <lx:label>PANEL BEATERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4213-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Panel Beaters</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4213-01</lx:path>
            <lx:label>Supervisor, Panel Beaters</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4213-11">
        <xs:annotation>
          <xs:documentation>Panel Beater</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4213-11</lx:path>
            <lx:label>Panel Beater</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4213-81">
        <xs:annotation>
          <xs:documentation>Apprentice Panel Beater</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4213-81</lx:path>
            <lx:label>Apprentice Panel Beater</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4214">
        <xs:annotation>
          <xs:documentation>VEHICLE PAINTERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4214</lx:path>
            <lx:label>VEHICLE PAINTERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4214-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Vehicle Painters</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4214-01</lx:path>
            <lx:label>Supervisor, Vehicle Painters</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4214-11">
        <xs:annotation>
          <xs:documentation>Vehicle Painter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4214-11</lx:path>
            <lx:label>Vehicle Painter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4214-81">
        <xs:annotation>
          <xs:documentation>Apprentice Vehicle Painter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4214-81</lx:path>
            <lx:label>Apprentice Vehicle Painter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4215">
        <xs:annotation>
          <xs:documentation>VEHICLE BODY MAKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4215</lx:path>
            <lx:label>VEHICLE BODY MAKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4215-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Vehicle Body Makers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4215-01</lx:path>
            <lx:label>Supervisor, Vehicle Body Makers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4215-11">
        <xs:annotation>
          <xs:documentation>Vehicle Body Maker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4215-11</lx:path>
            <lx:label>Vehicle Body Maker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4215-81">
        <xs:annotation>
          <xs:documentation>Apprentice Vehicle Body Maker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4215-81</lx:path>
            <lx:label>Apprentice Vehicle Body Maker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4216">
        <xs:annotation>
          <xs:documentation>VEHICLE TRIMMERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4216</lx:path>
            <lx:label>VEHICLE TRIMMERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4216-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Vehicle Trimmers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4216-01</lx:path>
            <lx:label>Supervisor, Vehicle Trimmers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4216-11">
        <xs:annotation>
          <xs:documentation>Vehicle Trimmer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4216-11</lx:path>
            <lx:label>Vehicle Trimmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4216-81">
        <xs:annotation>
          <xs:documentation>Apprentice Vehicle Trimmer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4216-81</lx:path>
            <lx:label>Apprentice Vehicle Trimmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="43">
        <xs:annotation>
          <xs:documentation>ELECTRICAL AND ELECTRONICS TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.43</lx:path>
            <lx:label>ELECTRICAL AND ELECTRONICS TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="431">
        <xs:annotation>
          <xs:documentation>ELECTRICAL AND ELECTRONICS TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.431</lx:path>
            <lx:label>ELECTRICAL AND ELECTRONICS TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4311">
        <xs:annotation>
          <xs:documentation>ELECTRICIANS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4311</lx:path>
            <lx:label>ELECTRICIANS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4311-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Electricians</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4311-01</lx:path>
            <lx:label>Supervisor, Electricians</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4311-11">
        <xs:annotation>
          <xs:documentation>General Electrician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4311-11</lx:path>
            <lx:label>General Electrician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4311-13">
        <xs:annotation>
          <xs:documentation>Electrician (Special Class)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4311-13</lx:path>
            <lx:label>Electrician (Special Class)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4311-15">
        <xs:annotation>
          <xs:documentation>Lift Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4311-15</lx:path>
            <lx:label>Lift Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4311-81">
        <xs:annotation>
          <xs:documentation>Apprentice Electrician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4311-81</lx:path>
            <lx:label>Apprentice Electrician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4311-83">
        <xs:annotation>
          <xs:documentation>Apprentice Lift Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4311-83</lx:path>
            <lx:label>Apprentice Lift Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4312">
        <xs:annotation>
          <xs:documentation>REFRIGERATION AND AIRCONDITIONING MECHANICS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4312</lx:path>
            <lx:label>REFRIGERATION AND AIRCONDITIONING MECHANICS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4312-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Refrigeration and Airconditioning Mechanics</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4312-01</lx:path>
            <lx:label>Supervisor, Refrigeration and Airconditioning Mechanics</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4312-11">
        <xs:annotation>
          <xs:documentation>Refrigeration and Airconditioning Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4312-11</lx:path>
            <lx:label>Refrigeration and Airconditioning Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4312-81">
        <xs:annotation>
          <xs:documentation>Apprentice Refrigeration and Airconditioning Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4312-81</lx:path>
            <lx:label>Apprentice Refrigeration and Airconditioning Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4313">
        <xs:annotation>
          <xs:documentation>ELECTRICAL DISTRIBUTION TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4313</lx:path>
            <lx:label>ELECTRICAL DISTRIBUTION TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4313-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Electrical Distribution Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4313-01</lx:path>
            <lx:label>Supervisor, Electrical Distribution Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4313-11">
        <xs:annotation>
          <xs:documentation>Electrical Powerline Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4313-11</lx:path>
            <lx:label>Electrical Powerline Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4313-13">
        <xs:annotation>
          <xs:documentation>Cable Jointer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4313-13</lx:path>
            <lx:label>Cable Jointer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4313-81">
        <xs:annotation>
          <xs:documentation>Apprentice Electrical Powerline Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4313-81</lx:path>
            <lx:label>Apprentice Electrical Powerline Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4313-83">
        <xs:annotation>
          <xs:documentation>Apprentice Cable Jointer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4313-83</lx:path>
            <lx:label>Apprentice Cable Jointer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4314">
        <xs:annotation>
          <xs:documentation>ELECTRONIC INSTRUMENT TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4314</lx:path>
            <lx:label>ELECTRONIC INSTRUMENT TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4314-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Electronic Instrument Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4314-01</lx:path>
            <lx:label>Supervisor, Electronic Instrument Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4314-11">
        <xs:annotation>
          <xs:documentation>General Electronic Instrument Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4314-11</lx:path>
            <lx:label>General Electronic Instrument Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4314-13">
        <xs:annotation>
          <xs:documentation>Electronic Instrument Tradesperson (Special Class)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4314-13</lx:path>
            <lx:label>Electronic Instrument Tradesperson (Special Class)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4314-81">
        <xs:annotation>
          <xs:documentation>Apprentice Electronic Instrument Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4314-81</lx:path>
            <lx:label>Apprentice Electronic Instrument Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4315">
        <xs:annotation>
          <xs:documentation>ELECTRONIC AND OFFICE EQUIPMENT TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4315</lx:path>
            <lx:label>ELECTRONIC AND OFFICE EQUIPMENT TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4315-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Electronic and Office Equipment Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4315-01</lx:path>
            <lx:label>Supervisor, Electronic and Office Equipment Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4315-11">
        <xs:annotation>
          <xs:documentation>Electronic Equipment Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4315-11</lx:path>
            <lx:label>Electronic Equipment Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4315-13">
        <xs:annotation>
          <xs:documentation>Business Machine Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4315-13</lx:path>
            <lx:label>Business Machine Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4315-81">
        <xs:annotation>
          <xs:documentation>Apprentice Electronic Equipment Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4315-81</lx:path>
            <lx:label>Apprentice Electronic Equipment Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4315-83">
        <xs:annotation>
          <xs:documentation>Apprentice Business Machine Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4315-83</lx:path>
            <lx:label>Apprentice Business Machine Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4316">
        <xs:annotation>
          <xs:documentation>COMMUNICATIONS TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4316</lx:path>
            <lx:label>COMMUNICATIONS TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4316-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Communications Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4316-01</lx:path>
            <lx:label>Supervisor, Communications Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4316-11">
        <xs:annotation>
          <xs:documentation>General Communications Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4316-11</lx:path>
            <lx:label>General Communications Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4316-13">
        <xs:annotation>
          <xs:documentation>Communications Linesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4316-13</lx:path>
            <lx:label>Communications Linesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4316-81">
        <xs:annotation>
          <xs:documentation>Apprentice General Communications Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4316-81</lx:path>
            <lx:label>Apprentice General Communications Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4316-83">
        <xs:annotation>
          <xs:documentation>Apprentice Communications Linesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4316-83</lx:path>
            <lx:label>Apprentice Communications Linesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="44">
        <xs:annotation>
          <xs:documentation>CONSTRUCTION TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.44</lx:path>
            <lx:label>CONSTRUCTION TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="441">
        <xs:annotation>
          <xs:documentation>STRUCTURAL CONSTRUCTION TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.441</lx:path>
            <lx:label>STRUCTURAL CONSTRUCTION TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4411">
        <xs:annotation>
          <xs:documentation>CARPENTRY AND JOINERY TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4411</lx:path>
            <lx:label>CARPENTRY AND JOINERY TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4411-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Carpentry and Joinery Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4411-01</lx:path>
            <lx:label>Supervisor, Carpentry and Joinery Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4411-11">
        <xs:annotation>
          <xs:documentation>Carpenter and Joiner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4411-11</lx:path>
            <lx:label>Carpenter and Joiner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4411-13">
        <xs:annotation>
          <xs:documentation>Carpenter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4411-13</lx:path>
            <lx:label>Carpenter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4411-15">
        <xs:annotation>
          <xs:documentation>Joiner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4411-15</lx:path>
            <lx:label>Joiner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4411-81">
        <xs:annotation>
          <xs:documentation>Apprentice Carpenter and Joiner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4411-81</lx:path>
            <lx:label>Apprentice Carpenter and Joiner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4411-83">
        <xs:annotation>
          <xs:documentation>Apprentice Carpenter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4411-83</lx:path>
            <lx:label>Apprentice Carpenter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4411-85">
        <xs:annotation>
          <xs:documentation>Apprentice Joiner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4411-85</lx:path>
            <lx:label>Apprentice Joiner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4412">
        <xs:annotation>
          <xs:documentation>FIBROUS PLASTERERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4412</lx:path>
            <lx:label>FIBROUS PLASTERERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4412-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Fibrous Plasterers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4412-01</lx:path>
            <lx:label>Supervisor, Fibrous Plasterers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4412-11">
        <xs:annotation>
          <xs:documentation>Fibrous Plasterer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4412-11</lx:path>
            <lx:label>Fibrous Plasterer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4412-81">
        <xs:annotation>
          <xs:documentation>Apprentice Fibrous Plasterer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4412-81</lx:path>
            <lx:label>Apprentice Fibrous Plasterer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4413">
        <xs:annotation>
          <xs:documentation>ROOF SLATERS AND TILERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4413</lx:path>
            <lx:label>ROOF SLATERS AND TILERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4413-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Roof Slaters and Tilers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4413-01</lx:path>
            <lx:label>Supervisor, Roof Slaters and Tilers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4413-11">
        <xs:annotation>
          <xs:documentation>Roof Slater and Tiler</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4413-11</lx:path>
            <lx:label>Roof Slater and Tiler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4413-81">
        <xs:annotation>
          <xs:documentation>Apprentice Roof Slater and Tiler</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4413-81</lx:path>
            <lx:label>Apprentice Roof Slater and Tiler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4414">
        <xs:annotation>
          <xs:documentation>BRICKLAYERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4414</lx:path>
            <lx:label>BRICKLAYERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4414-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Bricklayers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4414-01</lx:path>
            <lx:label>Supervisor, Bricklayers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4414-11">
        <xs:annotation>
          <xs:documentation>Bricklayer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4414-11</lx:path>
            <lx:label>Bricklayer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4414-81">
        <xs:annotation>
          <xs:documentation>Apprentice Bricklayer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4414-81</lx:path>
            <lx:label>Apprentice Bricklayer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4415">
        <xs:annotation>
          <xs:documentation>SOLID PLASTERERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4415</lx:path>
            <lx:label>SOLID PLASTERERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4415-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Solid Plasterers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4415-01</lx:path>
            <lx:label>Supervisor, Solid Plasterers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4415-11">
        <xs:annotation>
          <xs:documentation>Solid Plasterer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4415-11</lx:path>
            <lx:label>Solid Plasterer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4415-81">
        <xs:annotation>
          <xs:documentation>Apprentice Solid Plasterer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4415-81</lx:path>
            <lx:label>Apprentice Solid Plasterer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4416">
        <xs:annotation>
          <xs:documentation>WALL AND FLOOR TILERS AND STONEMASONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4416</lx:path>
            <lx:label>WALL AND FLOOR TILERS AND STONEMASONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4416-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Wall and Floor Tilers and Stonemasons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4416-01</lx:path>
            <lx:label>Supervisor, Wall and Floor Tilers and Stonemasons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4416-11">
        <xs:annotation>
          <xs:documentation>Wall and Floor Tiler</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4416-11</lx:path>
            <lx:label>Wall and Floor Tiler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4416-13">
        <xs:annotation>
          <xs:documentation>Stonemason</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4416-13</lx:path>
            <lx:label>Stonemason</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4416-81">
        <xs:annotation>
          <xs:documentation>Apprentice Wall and Floor Tiler</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4416-81</lx:path>
            <lx:label>Apprentice Wall and Floor Tiler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4416-83">
        <xs:annotation>
          <xs:documentation>Apprentice Stonemason</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4416-83</lx:path>
            <lx:label>Apprentice Stonemason</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="442">
        <xs:annotation>
          <xs:documentation>FINAL FINISHES CONSTRUCTION TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.442</lx:path>
            <lx:label>FINAL FINISHES CONSTRUCTION TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4421">
        <xs:annotation>
          <xs:documentation>PAINTERS AND DECORATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4421</lx:path>
            <lx:label>PAINTERS AND DECORATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4421-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Painters and Decorators</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4421-01</lx:path>
            <lx:label>Supervisor, Painters and Decorators</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4421-11">
        <xs:annotation>
          <xs:documentation>Painter and Decorator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4421-11</lx:path>
            <lx:label>Painter and Decorator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4421-81">
        <xs:annotation>
          <xs:documentation>Apprentice Painter and Decorator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4421-81</lx:path>
            <lx:label>Apprentice Painter and Decorator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4422">
        <xs:annotation>
          <xs:documentation>SIGNWRITERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4422</lx:path>
            <lx:label>SIGNWRITERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4422-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Signwriters</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4422-01</lx:path>
            <lx:label>Supervisor, Signwriters</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4422-11">
        <xs:annotation>
          <xs:documentation>Signwriter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4422-11</lx:path>
            <lx:label>Signwriter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4422-81">
        <xs:annotation>
          <xs:documentation>Apprentice Signwriter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4422-81</lx:path>
            <lx:label>Apprentice Signwriter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4423">
        <xs:annotation>
          <xs:documentation>FLOOR FINISHERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4423</lx:path>
            <lx:label>FLOOR FINISHERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4423-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Floor Finishers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4423-01</lx:path>
            <lx:label>Supervisor, Floor Finishers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4423-11">
        <xs:annotation>
          <xs:documentation>Floor Finisher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4423-11</lx:path>
            <lx:label>Floor Finisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4423-81">
        <xs:annotation>
          <xs:documentation>Apprentice Floor Finisher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4423-81</lx:path>
            <lx:label>Apprentice Floor Finisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="443">
        <xs:annotation>
          <xs:documentation>PLUMBERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.443</lx:path>
            <lx:label>PLUMBERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4431">
        <xs:annotation>
          <xs:documentation>PLUMBERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4431</lx:path>
            <lx:label>PLUMBERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4431-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Plumbers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4431-01</lx:path>
            <lx:label>Supervisor, Plumbers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4431-11">
        <xs:annotation>
          <xs:documentation>General Plumber</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4431-11</lx:path>
            <lx:label>General Plumber</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4431-13">
        <xs:annotation>
          <xs:documentation>Gasfitter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4431-13</lx:path>
            <lx:label>Gasfitter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4431-15">
        <xs:annotation>
          <xs:documentation>Drainer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4431-15</lx:path>
            <lx:label>Drainer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4431-17">
        <xs:annotation>
          <xs:documentation>Roof Plumber</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4431-17</lx:path>
            <lx:label>Roof Plumber</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4431-19">
        <xs:annotation>
          <xs:documentation>Mechanical Services and Airconditioning Plumber</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4431-19</lx:path>
            <lx:label>Mechanical Services and Airconditioning Plumber</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4431-81">
        <xs:annotation>
          <xs:documentation>Apprentice General Plumber</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4431-81</lx:path>
            <lx:label>Apprentice General Plumber</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4431-83">
        <xs:annotation>
          <xs:documentation>Apprentice Gasfitter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4431-83</lx:path>
            <lx:label>Apprentice Gasfitter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4431-85">
        <xs:annotation>
          <xs:documentation>Apprentice Drainer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4431-85</lx:path>
            <lx:label>Apprentice Drainer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4431-87">
        <xs:annotation>
          <xs:documentation>Apprentice Roof Plumber</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4431-87</lx:path>
            <lx:label>Apprentice Roof Plumber</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4431-89">
        <xs:annotation>
          <xs:documentation>Apprentice Mechanical Services and Airconditioning Plumber</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4431-89</lx:path>
            <lx:label>Apprentice Mechanical Services and Airconditioning Plumber</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="45">
        <xs:annotation>
          <xs:documentation>FOOD TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.45</lx:path>
            <lx:label>FOOD TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451">
        <xs:annotation>
          <xs:documentation>FOOD TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.451</lx:path>
            <lx:label>FOOD TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4511">
        <xs:annotation>
          <xs:documentation>MEAT TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4511</lx:path>
            <lx:label>MEAT TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4511-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Meat Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4511-01</lx:path>
            <lx:label>Supervisor, Meat Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4511-11">
        <xs:annotation>
          <xs:documentation>Butcher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4511-11</lx:path>
            <lx:label>Butcher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4511-13">
        <xs:annotation>
          <xs:documentation>Smallgoods Maker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4511-13</lx:path>
            <lx:label>Smallgoods Maker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4511-15">
        <xs:annotation>
          <xs:documentation>Slaughterperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4511-15</lx:path>
            <lx:label>Slaughterperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4511-81">
        <xs:annotation>
          <xs:documentation>Apprentice Butcher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4511-81</lx:path>
            <lx:label>Apprentice Butcher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4511-83">
        <xs:annotation>
          <xs:documentation>Apprentice Smallgoods Maker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4511-83</lx:path>
            <lx:label>Apprentice Smallgoods Maker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4511-85">
        <xs:annotation>
          <xs:documentation>Apprentice Slaughterperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4511-85</lx:path>
            <lx:label>Apprentice Slaughterperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4512">
        <xs:annotation>
          <xs:documentation>BAKERS AND PASTRYCOOKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4512</lx:path>
            <lx:label>BAKERS AND PASTRYCOOKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4512-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Bakers and Pastrycooks</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4512-01</lx:path>
            <lx:label>Supervisor, Bakers and Pastrycooks</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4512-11">
        <xs:annotation>
          <xs:documentation>Baker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4512-11</lx:path>
            <lx:label>Baker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4512-13">
        <xs:annotation>
          <xs:documentation>Pastrycook</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4512-13</lx:path>
            <lx:label>Pastrycook</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4512-81">
        <xs:annotation>
          <xs:documentation>Apprentice Baker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4512-81</lx:path>
            <lx:label>Apprentice Baker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4512-83">
        <xs:annotation>
          <xs:documentation>Apprentice Pastrycook</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4512-83</lx:path>
            <lx:label>Apprentice Pastrycook</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4513">
        <xs:annotation>
          <xs:documentation>COOKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4513</lx:path>
            <lx:label>COOKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4513-11">
        <xs:annotation>
          <xs:documentation>Cook</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4513-11</lx:path>
            <lx:label>Cook</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4513-81">
        <xs:annotation>
          <xs:documentation>Apprentice Cook or Chef</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4513-81</lx:path>
            <lx:label>Apprentice Cook or Chef</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4519">
        <xs:annotation>
          <xs:documentation>OTHER FOOD TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4519</lx:path>
            <lx:label>OTHER FOOD TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4519-11">
        <xs:annotation>
          <xs:documentation>Miller</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4519-11</lx:path>
            <lx:label>Miller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4519-13">
        <xs:annotation>
          <xs:documentation>Buttermaker or Cheesemaker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4519-13</lx:path>
            <lx:label>Buttermaker or Cheesemaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4519-15">
        <xs:annotation>
          <xs:documentation>Confectioner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4519-15</lx:path>
            <lx:label>Confectioner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4519-81">
        <xs:annotation>
          <xs:documentation>Apprentice Miller</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4519-81</lx:path>
            <lx:label>Apprentice Miller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4519-83">
        <xs:annotation>
          <xs:documentation>Apprentice Buttermaker or Cheesemaker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4519-83</lx:path>
            <lx:label>Apprentice Buttermaker or Cheesemaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4519-85">
        <xs:annotation>
          <xs:documentation>Apprentice Confectioner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4519-85</lx:path>
            <lx:label>Apprentice Confectioner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="46">
        <xs:annotation>
          <xs:documentation>SKILLED AGRICULTURAL AND HORTICULTURAL WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.46</lx:path>
            <lx:label>SKILLED AGRICULTURAL AND HORTICULTURAL WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="461">
        <xs:annotation>
          <xs:documentation>SKILLED AGRICULTURAL WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.461</lx:path>
            <lx:label>SKILLED AGRICULTURAL WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4611">
        <xs:annotation>
          <xs:documentation>FARM OVERSEERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4611</lx:path>
            <lx:label>FARM OVERSEERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4611-11">
        <xs:annotation>
          <xs:documentation>Farm Overseer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4611-11</lx:path>
            <lx:label>Farm Overseer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4612">
        <xs:annotation>
          <xs:documentation>SHEARERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4612</lx:path>
            <lx:label>SHEARERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4612-11">
        <xs:annotation>
          <xs:documentation>Shearer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4612-11</lx:path>
            <lx:label>Shearer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4613">
        <xs:annotation>
          <xs:documentation>WOOL, HIDE AND SKIN CLASSERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4613</lx:path>
            <lx:label>WOOL, HIDE AND SKIN CLASSERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4613-11">
        <xs:annotation>
          <xs:documentation>Wool Classer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4613-11</lx:path>
            <lx:label>Wool Classer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4613-13">
        <xs:annotation>
          <xs:documentation>Hide and Skin Classer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4613-13</lx:path>
            <lx:label>Hide and Skin Classer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4614">
        <xs:annotation>
          <xs:documentation>ANIMAL TRAINERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4614</lx:path>
            <lx:label>ANIMAL TRAINERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4614-11">
        <xs:annotation>
          <xs:documentation>Horse Trainer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4614-11</lx:path>
            <lx:label>Horse Trainer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4614-79">
        <xs:annotation>
          <xs:documentation>Animal Trainers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4614-79</lx:path>
            <lx:label>Animal Trainers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="462">
        <xs:annotation>
          <xs:documentation>HORTICULTURAL TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.462</lx:path>
            <lx:label>HORTICULTURAL TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4621">
        <xs:annotation>
          <xs:documentation>NURSERYPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4621</lx:path>
            <lx:label>NURSERYPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4621-11">
        <xs:annotation>
          <xs:documentation>Nurseryperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4621-11</lx:path>
            <lx:label>Nurseryperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4621-81">
        <xs:annotation>
          <xs:documentation>Apprentice Nurseryperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4621-81</lx:path>
            <lx:label>Apprentice Nurseryperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4622">
        <xs:annotation>
          <xs:documentation>GREENKEEPERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4622</lx:path>
            <lx:label>GREENKEEPERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4622-11">
        <xs:annotation>
          <xs:documentation>Greenkeeper</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4622-11</lx:path>
            <lx:label>Greenkeeper</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4622-81">
        <xs:annotation>
          <xs:documentation>Apprentice Greenkeeper</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4622-81</lx:path>
            <lx:label>Apprentice Greenkeeper</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4623">
        <xs:annotation>
          <xs:documentation>GARDENERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4623</lx:path>
            <lx:label>GARDENERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4623-01">
        <xs:annotation>
          <xs:documentation>Head Gardener</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4623-01</lx:path>
            <lx:label>Head Gardener</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4623-11">
        <xs:annotation>
          <xs:documentation>General Gardener</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4623-11</lx:path>
            <lx:label>General Gardener</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4623-13">
        <xs:annotation>
          <xs:documentation>Landscape Gardener</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4623-13</lx:path>
            <lx:label>Landscape Gardener</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4623-15">
        <xs:annotation>
          <xs:documentation>Tree Surgeon</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4623-15</lx:path>
            <lx:label>Tree Surgeon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4623-81">
        <xs:annotation>
          <xs:documentation>Apprentice General Gardener</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4623-81</lx:path>
            <lx:label>Apprentice General Gardener</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4623-83">
        <xs:annotation>
          <xs:documentation>Apprentice Landscape Gardener</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4623-83</lx:path>
            <lx:label>Apprentice Landscape Gardener</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4623-85">
        <xs:annotation>
          <xs:documentation>Apprentice Tree Surgeon</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4623-85</lx:path>
            <lx:label>Apprentice Tree Surgeon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="49">
        <xs:annotation>
          <xs:documentation>OTHER TRADESPERSONS AND RELATED WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.49</lx:path>
            <lx:label>OTHER TRADESPERSONS AND RELATED WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="491">
        <xs:annotation>
          <xs:documentation>PRINTING TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.491</lx:path>
            <lx:label>PRINTING TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4911">
        <xs:annotation>
          <xs:documentation>GRAPHIC PRE-PRESS TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4911</lx:path>
            <lx:label>GRAPHIC PRE-PRESS TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4911-11">
        <xs:annotation>
          <xs:documentation>Graphic Pre-Press Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4911-11</lx:path>
            <lx:label>Graphic Pre-Press Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4911-81">
        <xs:annotation>
          <xs:documentation>Apprentice Graphic Pre-Press Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4911-81</lx:path>
            <lx:label>Apprentice Graphic Pre-Press Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4912">
        <xs:annotation>
          <xs:documentation>PRINTING MACHINISTS AND SMALL OFFSET PRINTERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4912</lx:path>
            <lx:label>PRINTING MACHINISTS AND SMALL OFFSET PRINTERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4912-11">
        <xs:annotation>
          <xs:documentation>Printing Machinist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4912-11</lx:path>
            <lx:label>Printing Machinist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4912-13">
        <xs:annotation>
          <xs:documentation>Small Offset Printer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4912-13</lx:path>
            <lx:label>Small Offset Printer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4912-81">
        <xs:annotation>
          <xs:documentation>Apprentice Printing Machinist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4912-81</lx:path>
            <lx:label>Apprentice Printing Machinist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4912-83">
        <xs:annotation>
          <xs:documentation>Apprentice Small Offset Printer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4912-83</lx:path>
            <lx:label>Apprentice Small Offset Printer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4913">
        <xs:annotation>
          <xs:documentation>BINDERS AND FINISHERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4913</lx:path>
            <lx:label>BINDERS AND FINISHERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4913-11">
        <xs:annotation>
          <xs:documentation>Binder and Finisher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4913-11</lx:path>
            <lx:label>Binder and Finisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4913-81">
        <xs:annotation>
          <xs:documentation>Apprentice Binder and Finisher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4913-81</lx:path>
            <lx:label>Apprentice Binder and Finisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4914">
        <xs:annotation>
          <xs:documentation>SCREEN PRINTERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4914</lx:path>
            <lx:label>SCREEN PRINTERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4914-11">
        <xs:annotation>
          <xs:documentation>Screen Printer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4914-11</lx:path>
            <lx:label>Screen Printer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4914-81">
        <xs:annotation>
          <xs:documentation>Apprentice Screen Printer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4914-81</lx:path>
            <lx:label>Apprentice Screen Printer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="492">
        <xs:annotation>
          <xs:documentation>WOOD TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.492</lx:path>
            <lx:label>WOOD TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4921">
        <xs:annotation>
          <xs:documentation>WOOD MACHINISTS AND TURNERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4921</lx:path>
            <lx:label>WOOD MACHINISTS AND TURNERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4921-11">
        <xs:annotation>
          <xs:documentation>Wood Machinist (A-Grade)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4921-11</lx:path>
            <lx:label>Wood Machinist (A-Grade)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4921-13">
        <xs:annotation>
          <xs:documentation>Wood Turner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4921-13</lx:path>
            <lx:label>Wood Turner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4921-81">
        <xs:annotation>
          <xs:documentation>Apprentice Wood Machinist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4921-81</lx:path>
            <lx:label>Apprentice Wood Machinist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4921-83">
        <xs:annotation>
          <xs:documentation>Apprentice Wood Turner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4921-83</lx:path>
            <lx:label>Apprentice Wood Turner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4922">
        <xs:annotation>
          <xs:documentation>CABINETMAKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4922</lx:path>
            <lx:label>CABINETMAKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4922-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Cabinetmakers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4922-01</lx:path>
            <lx:label>Supervisor, Cabinetmakers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4922-11">
        <xs:annotation>
          <xs:documentation>Cabinetmaker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4922-11</lx:path>
            <lx:label>Cabinetmaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4922-81">
        <xs:annotation>
          <xs:documentation>Apprentice Cabinetmaker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4922-81</lx:path>
            <lx:label>Apprentice Cabinetmaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4929">
        <xs:annotation>
          <xs:documentation>OTHER WOOD TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4929</lx:path>
            <lx:label>OTHER WOOD TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4929-11">
        <xs:annotation>
          <xs:documentation>Picture Framer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4929-11</lx:path>
            <lx:label>Picture Framer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4929-13">
        <xs:annotation>
          <xs:documentation>Furniture Finisher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4929-13</lx:path>
            <lx:label>Furniture Finisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4929-79">
        <xs:annotation>
          <xs:documentation>Wood Tradespersons nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4929-79</lx:path>
            <lx:label>Wood Tradespersons nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4929-81">
        <xs:annotation>
          <xs:documentation>Apprentice Picture Framer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4929-81</lx:path>
            <lx:label>Apprentice Picture Framer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4929-83">
        <xs:annotation>
          <xs:documentation>Apprentice Furniture Finisher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4929-83</lx:path>
            <lx:label>Apprentice Furniture Finisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4929-99">
        <xs:annotation>
          <xs:documentation>Apprentice Wood Tradespersons nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4929-99</lx:path>
            <lx:label>Apprentice Wood Tradespersons nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="493">
        <xs:annotation>
          <xs:documentation>HAIRDRESSERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.493</lx:path>
            <lx:label>HAIRDRESSERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4931">
        <xs:annotation>
          <xs:documentation>HAIRDRESSERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4931</lx:path>
            <lx:label>HAIRDRESSERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4931-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Hairdressers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4931-01</lx:path>
            <lx:label>Supervisor, Hairdressers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4931-11">
        <xs:annotation>
          <xs:documentation>Hairdresser</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4931-11</lx:path>
            <lx:label>Hairdresser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4931-81">
        <xs:annotation>
          <xs:documentation>Apprentice Hairdresser</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4931-81</lx:path>
            <lx:label>Apprentice Hairdresser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="494">
        <xs:annotation>
          <xs:documentation>TEXTILE, CLOTHING AND RELATED TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.494</lx:path>
            <lx:label>TEXTILE, CLOTHING AND RELATED TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4941">
        <xs:annotation>
          <xs:documentation>CLOTHING TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4941</lx:path>
            <lx:label>CLOTHING TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4941-11">
        <xs:annotation>
          <xs:documentation>General Clothing Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4941-11</lx:path>
            <lx:label>General Clothing Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4941-13">
        <xs:annotation>
          <xs:documentation>Tailor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4941-13</lx:path>
            <lx:label>Tailor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4941-15">
        <xs:annotation>
          <xs:documentation>Dressmaker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4941-15</lx:path>
            <lx:label>Dressmaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4941-17">
        <xs:annotation>
          <xs:documentation>Apparel Cutter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4941-17</lx:path>
            <lx:label>Apparel Cutter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4941-19">
        <xs:annotation>
          <xs:documentation>Patternmaker-Grader (Clothing)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4941-19</lx:path>
            <lx:label>Patternmaker-Grader (Clothing)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4941-79">
        <xs:annotation>
          <xs:documentation>Clothing Tradespersons nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4941-79</lx:path>
            <lx:label>Clothing Tradespersons nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4941-81">
        <xs:annotation>
          <xs:documentation>Apprentice Clothing Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4941-81</lx:path>
            <lx:label>Apprentice Clothing Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4942">
        <xs:annotation>
          <xs:documentation>UPHOLSTERERS AND BEDDING TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4942</lx:path>
            <lx:label>UPHOLSTERERS AND BEDDING TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4942-11">
        <xs:annotation>
          <xs:documentation>Furniture Upholsterer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4942-11</lx:path>
            <lx:label>Furniture Upholsterer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4942-79">
        <xs:annotation>
          <xs:documentation>Upholsterers and Bedding Tradespersons nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4942-79</lx:path>
            <lx:label>Upholsterers and Bedding Tradespersons nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4942-81">
        <xs:annotation>
          <xs:documentation>Apprentice Upholsterer or Bedding Tradespersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4942-81</lx:path>
            <lx:label>Apprentice Upholsterer or Bedding Tradespersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4943">
        <xs:annotation>
          <xs:documentation>FOOTWEAR TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4943</lx:path>
            <lx:label>FOOTWEAR TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4943-11">
        <xs:annotation>
          <xs:documentation>Shoemaker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4943-11</lx:path>
            <lx:label>Shoemaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4943-13">
        <xs:annotation>
          <xs:documentation>Medical Grade Shoemaker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4943-13</lx:path>
            <lx:label>Medical Grade Shoemaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4943-81">
        <xs:annotation>
          <xs:documentation>Apprentice Shoemaker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4943-81</lx:path>
            <lx:label>Apprentice Shoemaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4944">
        <xs:annotation>
          <xs:documentation>LEATHER GOODS, CANVAS GOODS AND SAIL MAKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4944</lx:path>
            <lx:label>LEATHER GOODS, CANVAS GOODS AND SAIL MAKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4944-11">
        <xs:annotation>
          <xs:documentation>Leather Goods Maker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4944-11</lx:path>
            <lx:label>Leather Goods Maker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4944-13">
        <xs:annotation>
          <xs:documentation>Canvas Goods Maker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4944-13</lx:path>
            <lx:label>Canvas Goods Maker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4944-15">
        <xs:annotation>
          <xs:documentation>Sail Maker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4944-15</lx:path>
            <lx:label>Sail Maker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4944-81">
        <xs:annotation>
          <xs:documentation>Apprentice Leather Goods, Canvas Goods or Sail Maker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4944-81</lx:path>
            <lx:label>Apprentice Leather Goods, Canvas Goods or Sail Maker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="498">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS TRADESPERSONS AND RELATED WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.498</lx:path>
            <lx:label>MISCELLANEOUS TRADESPERSONS AND RELATED WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4981">
        <xs:annotation>
          <xs:documentation>MARINE CONSTRUCTION TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4981</lx:path>
            <lx:label>MARINE CONSTRUCTION TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4981-11">
        <xs:annotation>
          <xs:documentation>Shipwright</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4981-11</lx:path>
            <lx:label>Shipwright</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4981-13">
        <xs:annotation>
          <xs:documentation>Boat Builder and Repairer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4981-13</lx:path>
            <lx:label>Boat Builder and Repairer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4981-81">
        <xs:annotation>
          <xs:documentation>Apprentice Shipwright</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4981-81</lx:path>
            <lx:label>Apprentice Shipwright</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4981-83">
        <xs:annotation>
          <xs:documentation>Apprentice Boat Builder and Repairer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4981-83</lx:path>
            <lx:label>Apprentice Boat Builder and Repairer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4982">
        <xs:annotation>
          <xs:documentation>GLASS TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4982</lx:path>
            <lx:label>GLASS TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4982-11">
        <xs:annotation>
          <xs:documentation>Flat Glass Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4982-11</lx:path>
            <lx:label>Flat Glass Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4982-13">
        <xs:annotation>
          <xs:documentation>Glass Blower</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4982-13</lx:path>
            <lx:label>Glass Blower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4982-81">
        <xs:annotation>
          <xs:documentation>Apprentice Flat Glass Tradesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4982-81</lx:path>
            <lx:label>Apprentice Flat Glass Tradesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4982-83">
        <xs:annotation>
          <xs:documentation>Apprentice Glass Blower</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4982-83</lx:path>
            <lx:label>Apprentice Glass Blower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4983">
        <xs:annotation>
          <xs:documentation>JEWELLERS AND RELATED TRADESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4983</lx:path>
            <lx:label>JEWELLERS AND RELATED TRADESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4983-11">
        <xs:annotation>
          <xs:documentation>Jeweller</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4983-11</lx:path>
            <lx:label>Jeweller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4983-13">
        <xs:annotation>
          <xs:documentation>Gem Cutter and Polisher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4983-13</lx:path>
            <lx:label>Gem Cutter and Polisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4983-81">
        <xs:annotation>
          <xs:documentation>Apprentice Jeweller</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4983-81</lx:path>
            <lx:label>Apprentice Jeweller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4983-83">
        <xs:annotation>
          <xs:documentation>Apprentice Gem Cutter and Polisher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4983-83</lx:path>
            <lx:label>Apprentice Gem Cutter and Polisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4984">
        <xs:annotation>
          <xs:documentation>FLORISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4984</lx:path>
            <lx:label>FLORISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4984-11">
        <xs:annotation>
          <xs:documentation>Florist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4984-11</lx:path>
            <lx:label>Florist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4985">
        <xs:annotation>
          <xs:documentation>FIRE FIGHTERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4985</lx:path>
            <lx:label>FIRE FIGHTERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4985-11">
        <xs:annotation>
          <xs:documentation>Fire Fighter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4985-11</lx:path>
            <lx:label>Fire Fighter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4986">
        <xs:annotation>
          <xs:documentation>DRILLERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4986</lx:path>
            <lx:label>DRILLERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4986-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Drillers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4986-01</lx:path>
            <lx:label>Supervisor, Drillers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4986-11">
        <xs:annotation>
          <xs:documentation>Driller</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4986-11</lx:path>
            <lx:label>Driller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4987">
        <xs:annotation>
          <xs:documentation>CHEMICAL, PETROLEUM AND GAS PLANT OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4987</lx:path>
            <lx:label>CHEMICAL, PETROLEUM AND GAS PLANT OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4987-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Chemical, Petroleum and Gas Plant Operators</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4987-01</lx:path>
            <lx:label>Supervisor, Chemical, Petroleum and Gas Plant Operators</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4987-11">
        <xs:annotation>
          <xs:documentation>Chemical Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4987-11</lx:path>
            <lx:label>Chemical Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4987-13">
        <xs:annotation>
          <xs:documentation>Petroleum and Gas Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4987-13</lx:path>
            <lx:label>Petroleum and Gas Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4988">
        <xs:annotation>
          <xs:documentation>POWER GENERATION PLANT OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4988</lx:path>
            <lx:label>POWER GENERATION PLANT OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4988-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Power Generation Plant Operators</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4988-01</lx:path>
            <lx:label>Supervisor, Power Generation Plant Operators</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4988-11">
        <xs:annotation>
          <xs:documentation>Power Generation Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4988-11</lx:path>
            <lx:label>Power Generation Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4991">
        <xs:annotation>
          <xs:documentation>DEFENCE FORCE MEMBERS NOT ELSEWHERE INCLUDED</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4991</lx:path>
            <lx:label>DEFENCE FORCE MEMBERS NOT ELSEWHERE INCLUDED</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4991-11">
        <xs:annotation>
          <xs:documentation>Defence Force Member Not Elsewhere Included</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4991-11</lx:path>
            <lx:label>Defence Force Member Not Elsewhere Included</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4991-81">
        <xs:annotation>
          <xs:documentation>Trainee Defence Force Member Not Elsewhere Included</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4991-81</lx:path>
            <lx:label>Trainee Defence Force Member Not Elsewhere Included</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4992">
        <xs:annotation>
          <xs:documentation>PERFORMING ARTS SUPPORT WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4992</lx:path>
            <lx:label>PERFORMING ARTS SUPPORT WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4992-11">
        <xs:annotation>
          <xs:documentation>Sound Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4992-11</lx:path>
            <lx:label>Sound Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4992-13">
        <xs:annotation>
          <xs:documentation>Camera Operator (Film, Television or Video)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4992-13</lx:path>
            <lx:label>Camera Operator (Film, Television or Video)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4992-15">
        <xs:annotation>
          <xs:documentation>Television Equipment Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4992-15</lx:path>
            <lx:label>Television Equipment Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4992-17">
        <xs:annotation>
          <xs:documentation>Broadcast Transmitter Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4992-17</lx:path>
            <lx:label>Broadcast Transmitter Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4992-19">
        <xs:annotation>
          <xs:documentation>Motion Picture Projectionist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4992-19</lx:path>
            <lx:label>Motion Picture Projectionist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4992-21">
        <xs:annotation>
          <xs:documentation>Light Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4992-21</lx:path>
            <lx:label>Light Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4992-23">
        <xs:annotation>
          <xs:documentation>Production Assistant (Film, Television or Radio)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4992-23</lx:path>
            <lx:label>Production Assistant (Film, Television or Radio)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4992-25">
        <xs:annotation>
          <xs:documentation>Production Assistant (Theatre)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4992-25</lx:path>
            <lx:label>Production Assistant (Theatre)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4992-27">
        <xs:annotation>
          <xs:documentation>Make Up Artist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4992-27</lx:path>
            <lx:label>Make Up Artist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4992-79">
        <xs:annotation>
          <xs:documentation>Performing Arts Support Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4992-79</lx:path>
            <lx:label>Performing Arts Support Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4999">
        <xs:annotation>
          <xs:documentation>OTHER MISCELLANEOUS TRADESPERSONS AND RELATED WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4999</lx:path>
            <lx:label>OTHER MISCELLANEOUS TRADESPERSONS AND RELATED WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4999-11">
        <xs:annotation>
          <xs:documentation>Optical Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4999-11</lx:path>
            <lx:label>Optical Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4999-13">
        <xs:annotation>
          <xs:documentation>Diver</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4999-13</lx:path>
            <lx:label>Diver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4999-15">
        <xs:annotation>
          <xs:documentation>Aircraft Safety Equipment Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4999-15</lx:path>
            <lx:label>Aircraft Safety Equipment Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4999-17">
        <xs:annotation>
          <xs:documentation>Piano Tuner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4999-17</lx:path>
            <lx:label>Piano Tuner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4999-79">
        <xs:annotation>
          <xs:documentation>Tradespersons and Related Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.4999-79</lx:path>
            <lx:label>Tradespersons and Related Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5">
        <xs:annotation>
          <xs:documentation>ADVANCED CLERICAL AND SERVICE WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5</lx:path>
            <lx:label>ADVANCED CLERICAL AND SERVICE WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="51">
        <xs:annotation>
          <xs:documentation>SECRETARIES AND PERSONAL ASSISTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.51</lx:path>
            <lx:label>SECRETARIES AND PERSONAL ASSISTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="511">
        <xs:annotation>
          <xs:documentation>SECRETARIES AND PERSONAL ASSISTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.511</lx:path>
            <lx:label>SECRETARIES AND PERSONAL ASSISTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5111">
        <xs:annotation>
          <xs:documentation>SECRETARIES AND PERSONAL ASSISTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5111</lx:path>
            <lx:label>SECRETARIES AND PERSONAL ASSISTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5111-11">
        <xs:annotation>
          <xs:documentation>Secretary</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5111-11</lx:path>
            <lx:label>Secretary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5111-13">
        <xs:annotation>
          <xs:documentation>Personal Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5111-13</lx:path>
            <lx:label>Personal Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="59">
        <xs:annotation>
          <xs:documentation>OTHER ADVANCED CLERICAL AND SERVICE WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.59</lx:path>
            <lx:label>OTHER ADVANCED CLERICAL AND SERVICE WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="591">
        <xs:annotation>
          <xs:documentation>ADVANCED NUMERICAL CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.591</lx:path>
            <lx:label>ADVANCED NUMERICAL CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5911">
        <xs:annotation>
          <xs:documentation>BOOKKEEPERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5911</lx:path>
            <lx:label>BOOKKEEPERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5911-11">
        <xs:annotation>
          <xs:documentation>Bookkeeper</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5911-11</lx:path>
            <lx:label>Bookkeeper</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5912">
        <xs:annotation>
          <xs:documentation>CREDIT AND LOANS OFFICERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5912</lx:path>
            <lx:label>CREDIT AND LOANS OFFICERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5912-11">
        <xs:annotation>
          <xs:documentation>Credit and Loans Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5912-11</lx:path>
            <lx:label>Credit and Loans Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS ADVANCED CLERICAL AND SERVICE WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.599</lx:path>
            <lx:label>MISCELLANEOUS ADVANCED CLERICAL AND SERVICE WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5991">
        <xs:annotation>
          <xs:documentation>ADVANCED LEGAL AND RELATED CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5991</lx:path>
            <lx:label>ADVANCED LEGAL AND RELATED CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5991-11">
        <xs:annotation>
          <xs:documentation>Clerk of Court</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5991-11</lx:path>
            <lx:label>Clerk of Court</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5991-13">
        <xs:annotation>
          <xs:documentation>Law Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5991-13</lx:path>
            <lx:label>Law Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5991-15">
        <xs:annotation>
          <xs:documentation>Trust Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5991-15</lx:path>
            <lx:label>Trust Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5992">
        <xs:annotation>
          <xs:documentation>COURT AND HANSARD REPORTERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5992</lx:path>
            <lx:label>COURT AND HANSARD REPORTERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5992-11">
        <xs:annotation>
          <xs:documentation>Court or Hansard Reporter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5992-11</lx:path>
            <lx:label>Court or Hansard Reporter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5993">
        <xs:annotation>
          <xs:documentation>INSURANCE AGENTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5993</lx:path>
            <lx:label>INSURANCE AGENTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5993-11">
        <xs:annotation>
          <xs:documentation>Insurance Agent</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5993-11</lx:path>
            <lx:label>Insurance Agent</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5994">
        <xs:annotation>
          <xs:documentation>INSURANCE RISK SURVEYORS, INVESTIGATORS AND LOSS ADJUSTERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5994</lx:path>
            <lx:label>INSURANCE RISK SURVEYORS, INVESTIGATORS AND LOSS ADJUSTERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5994-11">
        <xs:annotation>
          <xs:documentation>Insurance Risk Surveyor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5994-11</lx:path>
            <lx:label>Insurance Risk Surveyor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5994-13">
        <xs:annotation>
          <xs:documentation>Insurance Investigator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5994-13</lx:path>
            <lx:label>Insurance Investigator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5994-15">
        <xs:annotation>
          <xs:documentation>Insurance Loss Adjuster</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5994-15</lx:path>
            <lx:label>Insurance Loss Adjuster</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5995">
        <xs:annotation>
          <xs:documentation>DESKTOP PUBLISHING OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5995</lx:path>
            <lx:label>DESKTOP PUBLISHING OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5995-11">
        <xs:annotation>
          <xs:documentation>Desktop Publishing Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5995-11</lx:path>
            <lx:label>Desktop Publishing Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5996">
        <xs:annotation>
          <xs:documentation>TRAVEL ATTENDANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5996</lx:path>
            <lx:label>TRAVEL ATTENDANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5996-11">
        <xs:annotation>
          <xs:documentation>Flight Service Director</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5996-11</lx:path>
            <lx:label>Flight Service Director</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5996-13">
        <xs:annotation>
          <xs:documentation>Flight Attendant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5996-13</lx:path>
            <lx:label>Flight Attendant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5996-79">
        <xs:annotation>
          <xs:documentation>Travel Attendants nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5996-79</lx:path>
            <lx:label>Travel Attendants nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5999">
        <xs:annotation>
          <xs:documentation>OTHER MISCELLANEOUS ADVANCED CLERICAL AND SERVICE WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5999</lx:path>
            <lx:label>OTHER MISCELLANEOUS ADVANCED CLERICAL AND SERVICE WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5999-11">
        <xs:annotation>
          <xs:documentation>Radio Despatcher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5999-11</lx:path>
            <lx:label>Radio Despatcher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5999-13">
        <xs:annotation>
          <xs:documentation>Auctioneer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5999-13</lx:path>
            <lx:label>Auctioneer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5999-15">
        <xs:annotation>
          <xs:documentation>Visual Merchandiser</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5999-15</lx:path>
            <lx:label>Visual Merchandiser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5999-17">
        <xs:annotation>
          <xs:documentation>Photographer's Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.5999-17</lx:path>
            <lx:label>Photographer's Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6">
        <xs:annotation>
          <xs:documentation>INTERMEDIATE CLERICAL, SALES AND SERVICE WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6</lx:path>
            <lx:label>INTERMEDIATE CLERICAL, SALES AND SERVICE WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="61">
        <xs:annotation>
          <xs:documentation>INTERMEDIATE CLERICAL WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.61</lx:path>
            <lx:label>INTERMEDIATE CLERICAL WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="611">
        <xs:annotation>
          <xs:documentation>GENERAL CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.611</lx:path>
            <lx:label>GENERAL CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6111">
        <xs:annotation>
          <xs:documentation>GENERAL CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6111</lx:path>
            <lx:label>GENERAL CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6111-11">
        <xs:annotation>
          <xs:documentation>General Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6111-11</lx:path>
            <lx:label>General Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="612">
        <xs:annotation>
          <xs:documentation>KEYBOARD OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.612</lx:path>
            <lx:label>KEYBOARD OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6121">
        <xs:annotation>
          <xs:documentation>KEYBOARD OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6121</lx:path>
            <lx:label>KEYBOARD OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6121-11">
        <xs:annotation>
          <xs:documentation>Typist and Word Processing Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6121-11</lx:path>
            <lx:label>Typist and Word Processing Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6121-13">
        <xs:annotation>
          <xs:documentation>Data Entry Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6121-13</lx:path>
            <lx:label>Data Entry Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="613">
        <xs:annotation>
          <xs:documentation>RECEPTIONISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.613</lx:path>
            <lx:label>RECEPTIONISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6131">
        <xs:annotation>
          <xs:documentation>RECEPTIONISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6131</lx:path>
            <lx:label>RECEPTIONISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6131-11">
        <xs:annotation>
          <xs:documentation>Receptionist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6131-11</lx:path>
            <lx:label>Receptionist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="614">
        <xs:annotation>
          <xs:documentation>INTERMEDIATE NUMERICAL CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.614</lx:path>
            <lx:label>INTERMEDIATE NUMERICAL CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6141">
        <xs:annotation>
          <xs:documentation>ACCOUNTING CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6141</lx:path>
            <lx:label>ACCOUNTING CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6141-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Accounting Clerks</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6141-01</lx:path>
            <lx:label>Supervisor, Accounting Clerks</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6141-11">
        <xs:annotation>
          <xs:documentation>Accounts Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6141-11</lx:path>
            <lx:label>Accounts Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6141-13">
        <xs:annotation>
          <xs:documentation>Credit Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6141-13</lx:path>
            <lx:label>Credit Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6141-15">
        <xs:annotation>
          <xs:documentation>Cost Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6141-15</lx:path>
            <lx:label>Cost Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6142">
        <xs:annotation>
          <xs:documentation>PAYROLL CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6142</lx:path>
            <lx:label>PAYROLL CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6142-11">
        <xs:annotation>
          <xs:documentation>Payroll Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6142-11</lx:path>
            <lx:label>Payroll Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6143">
        <xs:annotation>
          <xs:documentation>BANK WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6143</lx:path>
            <lx:label>BANK WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6143-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Bank Workers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6143-01</lx:path>
            <lx:label>Supervisor, Bank Workers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6143-11">
        <xs:annotation>
          <xs:documentation>Bank Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6143-11</lx:path>
            <lx:label>Bank Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6144">
        <xs:annotation>
          <xs:documentation>INSURANCE CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6144</lx:path>
            <lx:label>INSURANCE CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6144-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Insurance Clerks</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6144-01</lx:path>
            <lx:label>Supervisor, Insurance Clerks</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6144-11">
        <xs:annotation>
          <xs:documentation>Insurance Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6144-11</lx:path>
            <lx:label>Insurance Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6145">
        <xs:annotation>
          <xs:documentation>MONEY MARKET AND STATISTICAL CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6145</lx:path>
            <lx:label>MONEY MARKET AND STATISTICAL CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6145-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Money Market and Statistical Clerks</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6145-01</lx:path>
            <lx:label>Supervisor, Money Market and Statistical Clerks</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6145-11">
        <xs:annotation>
          <xs:documentation>Money Market Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6145-11</lx:path>
            <lx:label>Money Market Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6145-13">
        <xs:annotation>
          <xs:documentation>Statistical Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6145-13</lx:path>
            <lx:label>Statistical Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="615">
        <xs:annotation>
          <xs:documentation>MATERIAL RECORDING AND DESPATCHING CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.615</lx:path>
            <lx:label>MATERIAL RECORDING AND DESPATCHING CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6151">
        <xs:annotation>
          <xs:documentation>PRODUCTION RECORDING CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6151</lx:path>
            <lx:label>PRODUCTION RECORDING CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6151-11">
        <xs:annotation>
          <xs:documentation>Production Recording Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6151-11</lx:path>
            <lx:label>Production Recording Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6152">
        <xs:annotation>
          <xs:documentation>TRANSPORT AND DESPATCHING CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6152</lx:path>
            <lx:label>TRANSPORT AND DESPATCHING CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6152-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Transport and Despatching Clerks</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6152-01</lx:path>
            <lx:label>Supervisor, Transport and Despatching Clerks</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6152-11">
        <xs:annotation>
          <xs:documentation>Receiving and Despatching Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6152-11</lx:path>
            <lx:label>Receiving and Despatching Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6152-13">
        <xs:annotation>
          <xs:documentation>Import-Export Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6152-13</lx:path>
            <lx:label>Import-Export Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6153">
        <xs:annotation>
          <xs:documentation>STOCK AND PURCHASING CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6153</lx:path>
            <lx:label>STOCK AND PURCHASING CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6153-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Stock and Purchasing Clerks</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6153-01</lx:path>
            <lx:label>Supervisor, Stock and Purchasing Clerks</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6153-11">
        <xs:annotation>
          <xs:documentation>Stock Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6153-11</lx:path>
            <lx:label>Stock Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6153-13">
        <xs:annotation>
          <xs:documentation>Purchasing Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6153-13</lx:path>
            <lx:label>Purchasing Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6153-15">
        <xs:annotation>
          <xs:documentation>Order Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6153-15</lx:path>
            <lx:label>Order Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6153-17">
        <xs:annotation>
          <xs:documentation>Sales Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6153-17</lx:path>
            <lx:label>Sales Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="619">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS INTERMEDIATE CLERICAL WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.619</lx:path>
            <lx:label>MISCELLANEOUS INTERMEDIATE CLERICAL WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6191">
        <xs:annotation>
          <xs:documentation>INQUIRY AND ADMISSIONS CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6191</lx:path>
            <lx:label>INQUIRY AND ADMISSIONS CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6191-11">
        <xs:annotation>
          <xs:documentation>Inquiry Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6191-11</lx:path>
            <lx:label>Inquiry Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6191-13">
        <xs:annotation>
          <xs:documentation>Admissions Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6191-13</lx:path>
            <lx:label>Admissions Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6192">
        <xs:annotation>
          <xs:documentation>LIBRARY ASSISTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6192</lx:path>
            <lx:label>LIBRARY ASSISTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6192-11">
        <xs:annotation>
          <xs:documentation>Library Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6192-11</lx:path>
            <lx:label>Library Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6193">
        <xs:annotation>
          <xs:documentation>PERSONNEL CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6193</lx:path>
            <lx:label>PERSONNEL CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6193-11">
        <xs:annotation>
          <xs:documentation>Personnel Records Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6193-11</lx:path>
            <lx:label>Personnel Records Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6193-13">
        <xs:annotation>
          <xs:documentation>Employment Office Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6193-13</lx:path>
            <lx:label>Employment Office Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6194">
        <xs:annotation>
          <xs:documentation>INTERMEDIATE INSPECTORS AND EXAMINERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6194</lx:path>
            <lx:label>INTERMEDIATE INSPECTORS AND EXAMINERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6194-11">
        <xs:annotation>
          <xs:documentation>Customs Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6194-11</lx:path>
            <lx:label>Customs Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6194-13">
        <xs:annotation>
          <xs:documentation>Taxation Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6194-13</lx:path>
            <lx:label>Taxation Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6194-15">
        <xs:annotation>
          <xs:documentation>Social Security Assessor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6194-15</lx:path>
            <lx:label>Social Security Assessor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6194-17">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Licence Examiner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6194-17</lx:path>
            <lx:label>Motor Vehicle Licence Examiner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6194-19">
        <xs:annotation>
          <xs:documentation>Transport Operations Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6194-19</lx:path>
            <lx:label>Transport Operations Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6194-21">
        <xs:annotation>
          <xs:documentation>Train Examiner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6194-21</lx:path>
            <lx:label>Train Examiner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6194-23">
        <xs:annotation>
          <xs:documentation>Water Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6194-23</lx:path>
            <lx:label>Water Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6194-79">
        <xs:annotation>
          <xs:documentation>Intermediate Inspectors and Examiners nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6194-79</lx:path>
            <lx:label>Intermediate Inspectors and Examiners nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6199">
        <xs:annotation>
          <xs:documentation>OTHER INTERMEDIATE CLERICAL WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6199</lx:path>
            <lx:label>OTHER INTERMEDIATE CLERICAL WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6199-11">
        <xs:annotation>
          <xs:documentation>Proof Reader</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6199-11</lx:path>
            <lx:label>Proof Reader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6199-13">
        <xs:annotation>
          <xs:documentation>Debt Collector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6199-13</lx:path>
            <lx:label>Debt Collector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6199-15">
        <xs:annotation>
          <xs:documentation>Mail Supervisor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6199-15</lx:path>
            <lx:label>Mail Supervisor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6199-17">
        <xs:annotation>
          <xs:documentation>Court Orderly</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6199-17</lx:path>
            <lx:label>Court Orderly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6199-19">
        <xs:annotation>
          <xs:documentation>Bailiff or Sheriff</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6199-19</lx:path>
            <lx:label>Bailiff or Sheriff</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6199-21">
        <xs:annotation>
          <xs:documentation>Coding Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6199-21</lx:path>
            <lx:label>Coding Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6199-79">
        <xs:annotation>
          <xs:documentation>Intermediate Clerical Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6199-79</lx:path>
            <lx:label>Intermediate Clerical Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="62">
        <xs:annotation>
          <xs:documentation>INTERMEDIATE SALES AND RELATED WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.62</lx:path>
            <lx:label>INTERMEDIATE SALES AND RELATED WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621">
        <xs:annotation>
          <xs:documentation>INTERMEDIATE SALES AND RELATED WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.621</lx:path>
            <lx:label>INTERMEDIATE SALES AND RELATED WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6211">
        <xs:annotation>
          <xs:documentation>SALES REPRESENTATIVES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6211</lx:path>
            <lx:label>SALES REPRESENTATIVES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6211-11">
        <xs:annotation>
          <xs:documentation>Sales Representative (Personal and Household Goods)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6211-11</lx:path>
            <lx:label>Sales Representative (Personal and Household Goods)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6211-13">
        <xs:annotation>
          <xs:documentation>Sales Representative (Business Services)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6211-13</lx:path>
            <lx:label>Sales Representative (Business Services)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6211-15">
        <xs:annotation>
          <xs:documentation>Sales Representative (Builder's and Plumber's Supplies)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6211-15</lx:path>
            <lx:label>Sales Representative (Builder's and Plumber's Supplies)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6211-17">
        <xs:annotation>
          <xs:documentation>Sales Representative (Motor Vehicle Parts and Accessories)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6211-17</lx:path>
            <lx:label>Sales Representative (Motor Vehicle Parts and Accessories)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6211-79">
        <xs:annotation>
          <xs:documentation>Sales Representatives nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6211-79</lx:path>
            <lx:label>Sales Representatives nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6212">
        <xs:annotation>
          <xs:documentation>MOTOR VEHICLE AND RELATED PRODUCTS SALESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6212</lx:path>
            <lx:label>MOTOR VEHICLE AND RELATED PRODUCTS SALESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6212-11">
        <xs:annotation>
          <xs:documentation>Motor Vehicle and Caravan Salesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6212-11</lx:path>
            <lx:label>Motor Vehicle and Caravan Salesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6212-13">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Parts Interpreter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6212-13</lx:path>
            <lx:label>Motor Vehicle Parts Interpreter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6213">
        <xs:annotation>
          <xs:documentation>RETAIL AND CHECKOUT SUPERVISORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6213</lx:path>
            <lx:label>RETAIL AND CHECKOUT SUPERVISORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6213-11">
        <xs:annotation>
          <xs:documentation>Retail Supervisor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6213-11</lx:path>
            <lx:label>Retail Supervisor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6213-13">
        <xs:annotation>
          <xs:documentation>Checkout Supervisor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6213-13</lx:path>
            <lx:label>Checkout Supervisor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="63">
        <xs:annotation>
          <xs:documentation>INTERMEDIATE SERVICE WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.63</lx:path>
            <lx:label>INTERMEDIATE SERVICE WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="631">
        <xs:annotation>
          <xs:documentation>CARERS AND AIDES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.631</lx:path>
            <lx:label>CARERS AND AIDES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6311">
        <xs:annotation>
          <xs:documentation>EDUCATION AIDES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6311</lx:path>
            <lx:label>EDUCATION AIDES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6311-11">
        <xs:annotation>
          <xs:documentation>Pre-School Aide</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6311-11</lx:path>
            <lx:label>Pre-School Aide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6311-13">
        <xs:annotation>
          <xs:documentation>Integration Aide</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6311-13</lx:path>
            <lx:label>Integration Aide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6311-15">
        <xs:annotation>
          <xs:documentation>Teachers' Aide</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6311-15</lx:path>
            <lx:label>Teachers' Aide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6311-17">
        <xs:annotation>
          <xs:documentation>Aboriginal and Torres Strait Islander Education Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6311-17</lx:path>
            <lx:label>Aboriginal and Torres Strait Islander Education Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6312">
        <xs:annotation>
          <xs:documentation>CHILDREN'S CARE WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6312</lx:path>
            <lx:label>CHILDREN'S CARE WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6312-11">
        <xs:annotation>
          <xs:documentation>Child Care Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6312-11</lx:path>
            <lx:label>Child Care Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6312-13">
        <xs:annotation>
          <xs:documentation>Family Day Care Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6312-13</lx:path>
            <lx:label>Family Day Care Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6312-15">
        <xs:annotation>
          <xs:documentation>Nanny</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6312-15</lx:path>
            <lx:label>Nanny</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6313">
        <xs:annotation>
          <xs:documentation>SPECIAL CARE WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6313</lx:path>
            <lx:label>SPECIAL CARE WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6313-11">
        <xs:annotation>
          <xs:documentation>Hostel Parent</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6313-11</lx:path>
            <lx:label>Hostel Parent</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6313-13">
        <xs:annotation>
          <xs:documentation>Child or Youth Residential Care Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6313-13</lx:path>
            <lx:label>Child or Youth Residential Care Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6313-15">
        <xs:annotation>
          <xs:documentation>Refuge Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6313-15</lx:path>
            <lx:label>Refuge Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6313-17">
        <xs:annotation>
          <xs:documentation>Aged or Disabled Person Carer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6313-17</lx:path>
            <lx:label>Aged or Disabled Person Carer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6313-19">
        <xs:annotation>
          <xs:documentation>Therapy Aide</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6313-19</lx:path>
            <lx:label>Therapy Aide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6314">
        <xs:annotation>
          <xs:documentation>PERSONAL CARE AND NURSING ASSISTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6314</lx:path>
            <lx:label>PERSONAL CARE AND NURSING ASSISTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6314-11">
        <xs:annotation>
          <xs:documentation>Personal Care Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6314-11</lx:path>
            <lx:label>Personal Care Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6314-13">
        <xs:annotation>
          <xs:documentation>Nursing Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6314-13</lx:path>
            <lx:label>Nursing Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="632">
        <xs:annotation>
          <xs:documentation>HOSPITALITY WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.632</lx:path>
            <lx:label>HOSPITALITY WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6321">
        <xs:annotation>
          <xs:documentation>HOTEL SERVICE SUPERVISORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6321</lx:path>
            <lx:label>HOTEL SERVICE SUPERVISORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6321-11">
        <xs:annotation>
          <xs:documentation>Hotel Service Supervisor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6321-11</lx:path>
            <lx:label>Hotel Service Supervisor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6322">
        <xs:annotation>
          <xs:documentation>BAR ATTENDANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6322</lx:path>
            <lx:label>BAR ATTENDANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6322-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Bar Attendants</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6322-01</lx:path>
            <lx:label>Supervisor, Bar Attendants</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6322-11">
        <xs:annotation>
          <xs:documentation>Bar Attendant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6322-11</lx:path>
            <lx:label>Bar Attendant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6323">
        <xs:annotation>
          <xs:documentation>WAITERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6323</lx:path>
            <lx:label>WAITERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6323-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Waiters</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6323-01</lx:path>
            <lx:label>Supervisor, Waiters</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6323-11">
        <xs:annotation>
          <xs:documentation>General Waiter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6323-11</lx:path>
            <lx:label>General Waiter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6323-13">
        <xs:annotation>
          <xs:documentation>Drink Waiter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6323-13</lx:path>
            <lx:label>Drink Waiter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6324">
        <xs:annotation>
          <xs:documentation>HOSPITALITY TRAINEES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6324</lx:path>
            <lx:label>HOSPITALITY TRAINEES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6324-11">
        <xs:annotation>
          <xs:documentation>Hospitality Trainee</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6324-11</lx:path>
            <lx:label>Hospitality Trainee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="639">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS INTERMEDIATE SERVICE WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.639</lx:path>
            <lx:label>MISCELLANEOUS INTERMEDIATE SERVICE WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6391">
        <xs:annotation>
          <xs:documentation>DENTAL ASSISTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6391</lx:path>
            <lx:label>DENTAL ASSISTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6391-11">
        <xs:annotation>
          <xs:documentation>Dental Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6391-11</lx:path>
            <lx:label>Dental Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6392">
        <xs:annotation>
          <xs:documentation>VETERINARY NURSES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6392</lx:path>
            <lx:label>VETERINARY NURSES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6392-11">
        <xs:annotation>
          <xs:documentation>Veterinary Nurse</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6392-11</lx:path>
            <lx:label>Veterinary Nurse</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6393">
        <xs:annotation>
          <xs:documentation>PRISON OFFICERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6393</lx:path>
            <lx:label>PRISON OFFICERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6393-11">
        <xs:annotation>
          <xs:documentation>Prison Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6393-11</lx:path>
            <lx:label>Prison Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6394">
        <xs:annotation>
          <xs:documentation>GAMING WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6394</lx:path>
            <lx:label>GAMING WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6394-11">
        <xs:annotation>
          <xs:documentation>Gaming Pit Boss</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6394-11</lx:path>
            <lx:label>Gaming Pit Boss</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6394-13">
        <xs:annotation>
          <xs:documentation>Gaming Table Supervisor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6394-13</lx:path>
            <lx:label>Gaming Table Supervisor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6394-15">
        <xs:annotation>
          <xs:documentation>Gaming Dealer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6394-15</lx:path>
            <lx:label>Gaming Dealer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6395">
        <xs:annotation>
          <xs:documentation>PERSONAL CARE CONSULTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6395</lx:path>
            <lx:label>PERSONAL CARE CONSULTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6395-11">
        <xs:annotation>
          <xs:documentation>Beauty Therapist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6395-11</lx:path>
            <lx:label>Beauty Therapist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6395-13">
        <xs:annotation>
          <xs:documentation>Natural Remedy Consultant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6395-13</lx:path>
            <lx:label>Natural Remedy Consultant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6395-15">
        <xs:annotation>
          <xs:documentation>Weight Loss Consultant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6395-15</lx:path>
            <lx:label>Weight Loss Consultant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6396">
        <xs:annotation>
          <xs:documentation>FITNESS INSTRUCTORS AND RELATED WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6396</lx:path>
            <lx:label>FITNESS INSTRUCTORS AND RELATED WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6396-11">
        <xs:annotation>
          <xs:documentation>Fitness Instructor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6396-11</lx:path>
            <lx:label>Fitness Instructor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6396-13">
        <xs:annotation>
          <xs:documentation>Outdoor Adventure Leader</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6396-13</lx:path>
            <lx:label>Outdoor Adventure Leader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6397">
        <xs:annotation>
          <xs:documentation>TRAVEL AND TOURISM AGENTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6397</lx:path>
            <lx:label>TRAVEL AND TOURISM AGENTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6397-11">
        <xs:annotation>
          <xs:documentation>Travel Agent</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6397-11</lx:path>
            <lx:label>Travel Agent</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6397-13">
        <xs:annotation>
          <xs:documentation>Tourist Information Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6397-13</lx:path>
            <lx:label>Tourist Information Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6397-15">
        <xs:annotation>
          <xs:documentation>Tour Guide</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6397-15</lx:path>
            <lx:label>Tour Guide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6399">
        <xs:annotation>
          <xs:documentation>OTHER INTERMEDIATE SERVICE WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6399</lx:path>
            <lx:label>OTHER INTERMEDIATE SERVICE WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6399-11">
        <xs:annotation>
          <xs:documentation>Museum or Gallery Attendant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6399-11</lx:path>
            <lx:label>Museum or Gallery Attendant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6399-13">
        <xs:annotation>
          <xs:documentation>Driving Instructor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6399-13</lx:path>
            <lx:label>Driving Instructor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6399-15">
        <xs:annotation>
          <xs:documentation>Animal Attendant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6399-15</lx:path>
            <lx:label>Animal Attendant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6399-17">
        <xs:annotation>
          <xs:documentation>Pest and Weed Controller</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6399-17</lx:path>
            <lx:label>Pest and Weed Controller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6399-79">
        <xs:annotation>
          <xs:documentation>Intermediate Service Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.6399-79</lx:path>
            <lx:label>Intermediate Service Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7">
        <xs:annotation>
          <xs:documentation>INTERMEDIATE PRODUCTION AND TRANSPORT WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7</lx:path>
            <lx:label>INTERMEDIATE PRODUCTION AND TRANSPORT WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="71">
        <xs:annotation>
          <xs:documentation>INTERMEDIATE PLANT OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.71</lx:path>
            <lx:label>INTERMEDIATE PLANT OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711">
        <xs:annotation>
          <xs:documentation>MOBILE PLANT OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.711</lx:path>
            <lx:label>MOBILE PLANT OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7111">
        <xs:annotation>
          <xs:documentation>MOBILE CONSTRUCTION PLANT OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7111</lx:path>
            <lx:label>MOBILE CONSTRUCTION PLANT OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7111-11">
        <xs:annotation>
          <xs:documentation>General Construction Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7111-11</lx:path>
            <lx:label>General Construction Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7111-13">
        <xs:annotation>
          <xs:documentation>Bulldozer Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7111-13</lx:path>
            <lx:label>Bulldozer Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7111-15">
        <xs:annotation>
          <xs:documentation>Backhoe Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7111-15</lx:path>
            <lx:label>Backhoe Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7111-17">
        <xs:annotation>
          <xs:documentation>Loader Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7111-17</lx:path>
            <lx:label>Loader Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7111-19">
        <xs:annotation>
          <xs:documentation>Grader Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7111-19</lx:path>
            <lx:label>Grader Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7111-21">
        <xs:annotation>
          <xs:documentation>Excavator Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7111-21</lx:path>
            <lx:label>Excavator Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7111-23">
        <xs:annotation>
          <xs:documentation>Paving Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7111-23</lx:path>
            <lx:label>Paving Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7111-25">
        <xs:annotation>
          <xs:documentation>Road Roller Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7111-25</lx:path>
            <lx:label>Road Roller Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7112">
        <xs:annotation>
          <xs:documentation>FORKLIFT DRIVERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7112</lx:path>
            <lx:label>FORKLIFT DRIVERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7112-11">
        <xs:annotation>
          <xs:documentation>Forklift Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7112-11</lx:path>
            <lx:label>Forklift Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7119">
        <xs:annotation>
          <xs:documentation>OTHER MOBILE PLANT OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7119</lx:path>
            <lx:label>OTHER MOBILE PLANT OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7119-11">
        <xs:annotation>
          <xs:documentation>Agricultural and Horticultural Mobile Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7119-11</lx:path>
            <lx:label>Agricultural and Horticultural Mobile Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7119-13">
        <xs:annotation>
          <xs:documentation>Railway Track Repair Mobile Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7119-13</lx:path>
            <lx:label>Railway Track Repair Mobile Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7119-15">
        <xs:annotation>
          <xs:documentation>Streetsweeper Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7119-15</lx:path>
            <lx:label>Streetsweeper Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7119-17">
        <xs:annotation>
          <xs:documentation>Logging Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7119-17</lx:path>
            <lx:label>Logging Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7119-79">
        <xs:annotation>
          <xs:documentation>Mobile Plant Operators nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7119-79</lx:path>
            <lx:label>Mobile Plant Operators nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712">
        <xs:annotation>
          <xs:documentation>INTERMEDIATE STATIONARY PLANT OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.712</lx:path>
            <lx:label>INTERMEDIATE STATIONARY PLANT OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7121">
        <xs:annotation>
          <xs:documentation>ENGINE AND BOILER OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7121</lx:path>
            <lx:label>ENGINE AND BOILER OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7121-11">
        <xs:annotation>
          <xs:documentation>Engine or Boiler Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7121-11</lx:path>
            <lx:label>Engine or Boiler Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7122">
        <xs:annotation>
          <xs:documentation>CRANE, HOIST AND LIFT OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7122</lx:path>
            <lx:label>CRANE, HOIST AND LIFT OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7122-11">
        <xs:annotation>
          <xs:documentation>Crane, Hoist or Lift Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7122-11</lx:path>
            <lx:label>Crane, Hoist or Lift Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7123">
        <xs:annotation>
          <xs:documentation>ENGINEERING PRODUCTION SYSTEMS WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7123</lx:path>
            <lx:label>ENGINEERING PRODUCTION SYSTEMS WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7123-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Engineering Production Systems Workers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7123-01</lx:path>
            <lx:label>Supervisor, Engineering Production Systems Workers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7123-11">
        <xs:annotation>
          <xs:documentation>Engineering Production Systems Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7123-11</lx:path>
            <lx:label>Engineering Production Systems Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7124">
        <xs:annotation>
          <xs:documentation>PULP AND PAPER MILL OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7124</lx:path>
            <lx:label>PULP AND PAPER MILL OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7124-11">
        <xs:annotation>
          <xs:documentation>Pulp Mill Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7124-11</lx:path>
            <lx:label>Pulp Mill Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7124-13">
        <xs:annotation>
          <xs:documentation>Paper Mill Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7124-13</lx:path>
            <lx:label>Paper Mill Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7129">
        <xs:annotation>
          <xs:documentation>OTHER INTERMEDIATE STATIONARY PLANT OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7129</lx:path>
            <lx:label>OTHER INTERMEDIATE STATIONARY PLANT OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7129-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Other Intermediate Stationary Plant Operators</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7129-01</lx:path>
            <lx:label>Supervisor, Other Intermediate Stationary Plant Operators</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7129-11">
        <xs:annotation>
          <xs:documentation>Railway Signal Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7129-11</lx:path>
            <lx:label>Railway Signal Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7129-13">
        <xs:annotation>
          <xs:documentation>Train Controller</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7129-13</lx:path>
            <lx:label>Train Controller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7129-15">
        <xs:annotation>
          <xs:documentation>Concrete Pump Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7129-15</lx:path>
            <lx:label>Concrete Pump Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7129-17">
        <xs:annotation>
          <xs:documentation>Cement Production Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7129-17</lx:path>
            <lx:label>Cement Production Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7129-19">
        <xs:annotation>
          <xs:documentation>Concrete Batching Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7129-19</lx:path>
            <lx:label>Concrete Batching Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7129-21">
        <xs:annotation>
          <xs:documentation>Water and Waste Water Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7129-21</lx:path>
            <lx:label>Water and Waste Water Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7129-23">
        <xs:annotation>
          <xs:documentation>Bulk Materials Handling Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7129-23</lx:path>
            <lx:label>Bulk Materials Handling Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7129-79">
        <xs:annotation>
          <xs:documentation>Intermediate Stationary Plant Operators nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7129-79</lx:path>
            <lx:label>Intermediate Stationary Plant Operators nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="72">
        <xs:annotation>
          <xs:documentation>INTERMEDIATE MACHINE OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.72</lx:path>
            <lx:label>INTERMEDIATE MACHINE OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721">
        <xs:annotation>
          <xs:documentation>INTERMEDIATE TEXTILE, CLOTHING AND RELATED MACHINE OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.721</lx:path>
            <lx:label>INTERMEDIATE TEXTILE, CLOTHING AND RELATED MACHINE OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7211">
        <xs:annotation>
          <xs:documentation>SEWING MACHINISTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7211</lx:path>
            <lx:label>SEWING MACHINISTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7211-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Sewing Machinists</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7211-01</lx:path>
            <lx:label>Supervisor, Sewing Machinists</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7211-11">
        <xs:annotation>
          <xs:documentation>Sewing Machinist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7211-11</lx:path>
            <lx:label>Sewing Machinist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7212">
        <xs:annotation>
          <xs:documentation>TEXTILE AND FOOTWEAR PRODUCTION MACHINE OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7212</lx:path>
            <lx:label>TEXTILE AND FOOTWEAR PRODUCTION MACHINE OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7212-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Textile and Footwear Production Machine Operators</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7212-01</lx:path>
            <lx:label>Supervisor, Textile and Footwear Production Machine Operators</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7212-11">
        <xs:annotation>
          <xs:documentation>Yarn Carding and Spinning Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7212-11</lx:path>
            <lx:label>Yarn Carding and Spinning Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7212-13">
        <xs:annotation>
          <xs:documentation>Weaving Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7212-13</lx:path>
            <lx:label>Weaving Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7212-15">
        <xs:annotation>
          <xs:documentation>Knitting Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7212-15</lx:path>
            <lx:label>Knitting Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7212-17">
        <xs:annotation>
          <xs:documentation>Textile Dyeing and Finishing Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7212-17</lx:path>
            <lx:label>Textile Dyeing and Finishing Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7212-19">
        <xs:annotation>
          <xs:documentation>Hide and Skin Processing Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7212-19</lx:path>
            <lx:label>Hide and Skin Processing Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7212-21">
        <xs:annotation>
          <xs:documentation>Footwear Production Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7212-21</lx:path>
            <lx:label>Footwear Production Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7212-79">
        <xs:annotation>
          <xs:documentation>Textile and Footwear Production Machine Operators nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7212-79</lx:path>
            <lx:label>Textile and Footwear Production Machine Operators nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="729">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS INTERMEDIATE MACHINE OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.729</lx:path>
            <lx:label>MISCELLANEOUS INTERMEDIATE MACHINE OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7291">
        <xs:annotation>
          <xs:documentation>PLASTICS PRODUCTION MACHINE OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7291</lx:path>
            <lx:label>PLASTICS PRODUCTION MACHINE OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7291-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Plastics Production Machine Operators</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7291-01</lx:path>
            <lx:label>Supervisor, Plastics Production Machine Operators</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7291-11">
        <xs:annotation>
          <xs:documentation>General Plastics Production Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7291-11</lx:path>
            <lx:label>General Plastics Production Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7291-13">
        <xs:annotation>
          <xs:documentation>Plastics Compounding and Reclamation Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7291-13</lx:path>
            <lx:label>Plastics Compounding and Reclamation Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7291-15">
        <xs:annotation>
          <xs:documentation>Reinforced Plastic and Composite Production Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7291-15</lx:path>
            <lx:label>Reinforced Plastic and Composite Production Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7291-17">
        <xs:annotation>
          <xs:documentation>Plastic Cablemaking Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7291-17</lx:path>
            <lx:label>Plastic Cablemaking Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7291-79">
        <xs:annotation>
          <xs:documentation>Plastics Production Machine Operators nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7291-79</lx:path>
            <lx:label>Plastics Production Machine Operators nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7292">
        <xs:annotation>
          <xs:documentation>RUBBER PRODUCTION MACHINE OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7292</lx:path>
            <lx:label>RUBBER PRODUCTION MACHINE OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7292-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Rubber Production Machine Operators</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7292-01</lx:path>
            <lx:label>Supervisor, Rubber Production Machine Operators</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7292-11">
        <xs:annotation>
          <xs:documentation>Tyre Production Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7292-11</lx:path>
            <lx:label>Tyre Production Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7292-79">
        <xs:annotation>
          <xs:documentation>Rubber Production Machine Operators nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7292-79</lx:path>
            <lx:label>Rubber Production Machine Operators nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7293">
        <xs:annotation>
          <xs:documentation>CHEMICAL PRODUCTION MACHINE OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7293</lx:path>
            <lx:label>CHEMICAL PRODUCTION MACHINE OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7293-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Chemical Production Machine Operators</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7293-01</lx:path>
            <lx:label>Supervisor, Chemical Production Machine Operators</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7293-11">
        <xs:annotation>
          <xs:documentation>Chemical Production Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7293-11</lx:path>
            <lx:label>Chemical Production Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7294">
        <xs:annotation>
          <xs:documentation>WOOD PROCESSING MACHINE OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7294</lx:path>
            <lx:label>WOOD PROCESSING MACHINE OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7294-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Wood Processing Machine Operators</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7294-01</lx:path>
            <lx:label>Supervisor, Wood Processing Machine Operators</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7294-11">
        <xs:annotation>
          <xs:documentation>Wood Processing Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7294-11</lx:path>
            <lx:label>Wood Processing Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7295">
        <xs:annotation>
          <xs:documentation>PAPER PRODUCTS MACHINE OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7295</lx:path>
            <lx:label>PAPER PRODUCTS MACHINE OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7295-11">
        <xs:annotation>
          <xs:documentation>Paper Products Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7295-11</lx:path>
            <lx:label>Paper Products Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7296">
        <xs:annotation>
          <xs:documentation>GLASS PRODUCTION MACHINE OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7296</lx:path>
            <lx:label>GLASS PRODUCTION MACHINE OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7296-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Glass Production Machine Operators</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7296-01</lx:path>
            <lx:label>Supervisor, Glass Production Machine Operators</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7296-11">
        <xs:annotation>
          <xs:documentation>Glass Production Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7296-11</lx:path>
            <lx:label>Glass Production Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7297">
        <xs:annotation>
          <xs:documentation>CLAY, STONE AND CONCRETE PROCESSING MACHINE OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7297</lx:path>
            <lx:label>CLAY, STONE AND CONCRETE PROCESSING MACHINE OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7297-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Clay, Stone and Concrete Processing Machine Operators</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7297-01</lx:path>
            <lx:label>Supervisor, Clay, Stone and Concrete Processing Machine Operators</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7297-11">
        <xs:annotation>
          <xs:documentation>Clay Products Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7297-11</lx:path>
            <lx:label>Clay Products Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7297-13">
        <xs:annotation>
          <xs:documentation>Stone Processing Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7297-13</lx:path>
            <lx:label>Stone Processing Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7297-15">
        <xs:annotation>
          <xs:documentation>Concrete Products Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7297-15</lx:path>
            <lx:label>Concrete Products Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7297-79">
        <xs:annotation>
          <xs:documentation>Clay, Stone and Concrete Processing Machine Operators nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7297-79</lx:path>
            <lx:label>Clay, Stone and Concrete Processing Machine Operators nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7298">
        <xs:annotation>
          <xs:documentation>PHOTOGRAPHIC DEVELOPERS AND PRINTERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7298</lx:path>
            <lx:label>PHOTOGRAPHIC DEVELOPERS AND PRINTERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7298-11">
        <xs:annotation>
          <xs:documentation>Photographic Developer and Printer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7298-11</lx:path>
            <lx:label>Photographic Developer and Printer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7299">
        <xs:annotation>
          <xs:documentation>OTHER INTERMEDIATE MACHINE OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7299</lx:path>
            <lx:label>OTHER INTERMEDIATE MACHINE OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7299-11">
        <xs:annotation>
          <xs:documentation>Industrial Spray Painter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7299-11</lx:path>
            <lx:label>Industrial Spray Painter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7299-79">
        <xs:annotation>
          <xs:documentation>Intermediate Machine Operators nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7299-79</lx:path>
            <lx:label>Intermediate Machine Operators nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="73">
        <xs:annotation>
          <xs:documentation>ROAD AND RAIL TRANSPORT DRIVERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.73</lx:path>
            <lx:label>ROAD AND RAIL TRANSPORT DRIVERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="731">
        <xs:annotation>
          <xs:documentation>ROAD AND RAIL TRANSPORT DRIVERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.731</lx:path>
            <lx:label>ROAD AND RAIL TRANSPORT DRIVERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7311">
        <xs:annotation>
          <xs:documentation>TRUCK DRIVERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7311</lx:path>
            <lx:label>TRUCK DRIVERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7311-11">
        <xs:annotation>
          <xs:documentation>Heavy Truck Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7311-11</lx:path>
            <lx:label>Heavy Truck Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7311-13">
        <xs:annotation>
          <xs:documentation>Furniture Removalist</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7311-13</lx:path>
            <lx:label>Furniture Removalist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7312">
        <xs:annotation>
          <xs:documentation>BUS AND TRAM DRIVERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7312</lx:path>
            <lx:label>BUS AND TRAM DRIVERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7312-11">
        <xs:annotation>
          <xs:documentation>Bus Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7312-11</lx:path>
            <lx:label>Bus Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7312-13">
        <xs:annotation>
          <xs:documentation>Tram Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7312-13</lx:path>
            <lx:label>Tram Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7313">
        <xs:annotation>
          <xs:documentation>AUTOMOBILE DRIVERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7313</lx:path>
            <lx:label>AUTOMOBILE DRIVERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7313-11">
        <xs:annotation>
          <xs:documentation>Automobile Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7313-11</lx:path>
            <lx:label>Automobile Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7314">
        <xs:annotation>
          <xs:documentation>DELIVERY DRIVERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7314</lx:path>
            <lx:label>DELIVERY DRIVERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7314-11">
        <xs:annotation>
          <xs:documentation>Delivery Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7314-11</lx:path>
            <lx:label>Delivery Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7315">
        <xs:annotation>
          <xs:documentation>TRAIN DRIVERS AND ASSISTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7315</lx:path>
            <lx:label>TRAIN DRIVERS AND ASSISTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7315-11">
        <xs:annotation>
          <xs:documentation>Train Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7315-11</lx:path>
            <lx:label>Train Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7315-13">
        <xs:annotation>
          <xs:documentation>Train Driver's Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7315-13</lx:path>
            <lx:label>Train Driver's Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="79">
        <xs:annotation>
          <xs:documentation>OTHER INTERMEDIATE PRODUCTION AND TRANSPORT WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.79</lx:path>
            <lx:label>OTHER INTERMEDIATE PRODUCTION AND TRANSPORT WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="791">
        <xs:annotation>
          <xs:documentation>INTERMEDIATE MINING AND CONSTRUCTION WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.791</lx:path>
            <lx:label>INTERMEDIATE MINING AND CONSTRUCTION WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7911">
        <xs:annotation>
          <xs:documentation>MINERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7911</lx:path>
            <lx:label>MINERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7911-11">
        <xs:annotation>
          <xs:documentation>Miner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7911-11</lx:path>
            <lx:label>Miner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7912">
        <xs:annotation>
          <xs:documentation>BLASTING WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7912</lx:path>
            <lx:label>BLASTING WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7912-11">
        <xs:annotation>
          <xs:documentation>Blasting Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7912-11</lx:path>
            <lx:label>Blasting Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7913">
        <xs:annotation>
          <xs:documentation>STRUCTURAL STEEL CONSTRUCTION WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7913</lx:path>
            <lx:label>STRUCTURAL STEEL CONSTRUCTION WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7913-11">
        <xs:annotation>
          <xs:documentation>Scaffolder</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7913-11</lx:path>
            <lx:label>Scaffolder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7913-13">
        <xs:annotation>
          <xs:documentation>Steel Fixer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7913-13</lx:path>
            <lx:label>Steel Fixer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7913-15">
        <xs:annotation>
          <xs:documentation>Structural Steel Erector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7913-15</lx:path>
            <lx:label>Structural Steel Erector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7913-17">
        <xs:annotation>
          <xs:documentation>Construction Rigger</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7913-17</lx:path>
            <lx:label>Construction Rigger</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7914">
        <xs:annotation>
          <xs:documentation>INSULATION AND HOME IMPROVEMENTS INSTALLERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7914</lx:path>
            <lx:label>INSULATION AND HOME IMPROVEMENTS INSTALLERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7914-11">
        <xs:annotation>
          <xs:documentation>Building Insulation Installer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7914-11</lx:path>
            <lx:label>Building Insulation Installer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7914-13">
        <xs:annotation>
          <xs:documentation>Home Improvements Installer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7914-13</lx:path>
            <lx:label>Home Improvements Installer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="799">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS INTERMEDIATE PRODUCTION AND TRANSPORT WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.799</lx:path>
            <lx:label>MISCELLANEOUS INTERMEDIATE PRODUCTION AND TRANSPORT WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7991">
        <xs:annotation>
          <xs:documentation>MOTOR VEHICLE PARTS AND ACCESSORIES FITTERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7991</lx:path>
            <lx:label>MOTOR VEHICLE PARTS AND ACCESSORIES FITTERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7991-11">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Parts and Accessories Fitter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7991-11</lx:path>
            <lx:label>Motor Vehicle Parts and Accessories Fitter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7992">
        <xs:annotation>
          <xs:documentation>PRODUCT QUALITY CONTROLLERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7992</lx:path>
            <lx:label>PRODUCT QUALITY CONTROLLERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7992-11">
        <xs:annotation>
          <xs:documentation>Product Examiner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7992-11</lx:path>
            <lx:label>Product Examiner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7992-13">
        <xs:annotation>
          <xs:documentation>Product Grader</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7992-13</lx:path>
            <lx:label>Product Grader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7992-15">
        <xs:annotation>
          <xs:documentation>Product Tester</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7992-15</lx:path>
            <lx:label>Product Tester</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7993">
        <xs:annotation>
          <xs:documentation>STOREPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7993</lx:path>
            <lx:label>STOREPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7993-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Storepersons</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7993-01</lx:path>
            <lx:label>Supervisor, Storepersons</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7993-11">
        <xs:annotation>
          <xs:documentation>Storeperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7993-11</lx:path>
            <lx:label>Storeperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7994">
        <xs:annotation>
          <xs:documentation>SEAFARERS AND FISHING HANDS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7994</lx:path>
            <lx:label>SEAFARERS AND FISHING HANDS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7994-11">
        <xs:annotation>
          <xs:documentation>Seafarer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7994-11</lx:path>
            <lx:label>Seafarer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7994-13">
        <xs:annotation>
          <xs:documentation>Fishing Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7994-13</lx:path>
            <lx:label>Fishing Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7995">
        <xs:annotation>
          <xs:documentation>FORESTRY AND LOGGING WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7995</lx:path>
            <lx:label>FORESTRY AND LOGGING WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7995-01">
        <xs:annotation>
          <xs:documentation>Supervisor, Forestry and Logging Workers</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7995-01</lx:path>
            <lx:label>Supervisor, Forestry and Logging Workers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7995-11">
        <xs:annotation>
          <xs:documentation>Tree Faller</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7995-11</lx:path>
            <lx:label>Tree Faller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7995-13">
        <xs:annotation>
          <xs:documentation>Forestry Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7995-13</lx:path>
            <lx:label>Forestry Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7995-15">
        <xs:annotation>
          <xs:documentation>Logging Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7995-15</lx:path>
            <lx:label>Logging Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7996">
        <xs:annotation>
          <xs:documentation>PRINTING HANDS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7996</lx:path>
            <lx:label>PRINTING HANDS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7996-11">
        <xs:annotation>
          <xs:documentation>Printing Table Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7996-11</lx:path>
            <lx:label>Printing Table Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7996-13">
        <xs:annotation>
          <xs:documentation>Printer's Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.7996-13</lx:path>
            <lx:label>Printer's Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8">
        <xs:annotation>
          <xs:documentation>ELEMENTARY CLERICAL, SALES AND SERVICE WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8</lx:path>
            <lx:label>ELEMENTARY CLERICAL, SALES AND SERVICE WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="81">
        <xs:annotation>
          <xs:documentation>ELEMENTARY CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.81</lx:path>
            <lx:label>ELEMENTARY CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811">
        <xs:annotation>
          <xs:documentation>ELEMENTARY CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.811</lx:path>
            <lx:label>ELEMENTARY CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8111">
        <xs:annotation>
          <xs:documentation>REGISTRY AND FILING CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8111</lx:path>
            <lx:label>REGISTRY AND FILING CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8111-11">
        <xs:annotation>
          <xs:documentation>Registry or Filing Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8111-11</lx:path>
            <lx:label>Registry or Filing Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8112">
        <xs:annotation>
          <xs:documentation>MAIL SORTING CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8112</lx:path>
            <lx:label>MAIL SORTING CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8112-11">
        <xs:annotation>
          <xs:documentation>Mail Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8112-11</lx:path>
            <lx:label>Mail Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8112-13">
        <xs:annotation>
          <xs:documentation>Postal Sorting Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8112-13</lx:path>
            <lx:label>Postal Sorting Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8113">
        <xs:annotation>
          <xs:documentation>SWITCHBOARD OPERATORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8113</lx:path>
            <lx:label>SWITCHBOARD OPERATORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8113-11">
        <xs:annotation>
          <xs:documentation>Switchboard Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8113-11</lx:path>
            <lx:label>Switchboard Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8114">
        <xs:annotation>
          <xs:documentation>MESSENGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8114</lx:path>
            <lx:label>MESSENGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8114-11">
        <xs:annotation>
          <xs:documentation>Courier</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8114-11</lx:path>
            <lx:label>Courier</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8114-13">
        <xs:annotation>
          <xs:documentation>Postal Delivery Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8114-13</lx:path>
            <lx:label>Postal Delivery Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8115">
        <xs:annotation>
          <xs:documentation>BETTING CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8115</lx:path>
            <lx:label>BETTING CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8115-11">
        <xs:annotation>
          <xs:documentation>Betting Agency Counter Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8115-11</lx:path>
            <lx:label>Betting Agency Counter Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8115-13">
        <xs:annotation>
          <xs:documentation>Telephone Betting Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8115-13</lx:path>
            <lx:label>Telephone Betting Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8115-15">
        <xs:annotation>
          <xs:documentation>Bookmaker's Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8115-15</lx:path>
            <lx:label>Bookmaker's Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8115-79">
        <xs:annotation>
          <xs:documentation>Betting Clerks nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8115-79</lx:path>
            <lx:label>Betting Clerks nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8116">
        <xs:annotation>
          <xs:documentation>OFFICE TRAINEES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8116</lx:path>
            <lx:label>OFFICE TRAINEES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8116-11">
        <xs:annotation>
          <xs:documentation>Office Trainee</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8116-11</lx:path>
            <lx:label>Office Trainee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8119">
        <xs:annotation>
          <xs:documentation>OTHER ELEMENTARY CLERKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8119</lx:path>
            <lx:label>OTHER ELEMENTARY CLERKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8119-11">
        <xs:annotation>
          <xs:documentation>Meter Reader</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8119-11</lx:path>
            <lx:label>Meter Reader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8119-13">
        <xs:annotation>
          <xs:documentation>Classified Advertising Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8119-13</lx:path>
            <lx:label>Classified Advertising Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8119-15">
        <xs:annotation>
          <xs:documentation>Interviewer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8119-15</lx:path>
            <lx:label>Interviewer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8119-17">
        <xs:annotation>
          <xs:documentation>Parking Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8119-17</lx:path>
            <lx:label>Parking Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8119-79">
        <xs:annotation>
          <xs:documentation>Elementary Clerks nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8119-79</lx:path>
            <lx:label>Elementary Clerks nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="82">
        <xs:annotation>
          <xs:documentation>ELEMENTARY SALES WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.82</lx:path>
            <lx:label>ELEMENTARY SALES WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821">
        <xs:annotation>
          <xs:documentation>SALES ASSISTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.821</lx:path>
            <lx:label>SALES ASSISTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8211">
        <xs:annotation>
          <xs:documentation>SALES ASSISTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8211</lx:path>
            <lx:label>SALES ASSISTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8211-11">
        <xs:annotation>
          <xs:documentation>Sales Assistant (Food and Drink Products)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8211-11</lx:path>
            <lx:label>Sales Assistant (Food and Drink Products)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8211-13">
        <xs:annotation>
          <xs:documentation>Sales Assistant (Fabric, Clothing and Footwear)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8211-13</lx:path>
            <lx:label>Sales Assistant (Fabric, Clothing and Footwear)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8211-15">
        <xs:annotation>
          <xs:documentation>Sales Assistant (Other Personal and Household Goods)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8211-15</lx:path>
            <lx:label>Sales Assistant (Other Personal and Household Goods)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8211-17">
        <xs:annotation>
          <xs:documentation>Sales Assistant (Postal Services)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8211-17</lx:path>
            <lx:label>Sales Assistant (Postal Services)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8211-79">
        <xs:annotation>
          <xs:documentation>Sales Assistants nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8211-79</lx:path>
            <lx:label>Sales Assistants nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="829">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS ELEMENTARY SALES WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.829</lx:path>
            <lx:label>MISCELLANEOUS ELEMENTARY SALES WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8291">
        <xs:annotation>
          <xs:documentation>CHECKOUT OPERATORS AND CASHIERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8291</lx:path>
            <lx:label>CHECKOUT OPERATORS AND CASHIERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8291-11">
        <xs:annotation>
          <xs:documentation>Checkout Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8291-11</lx:path>
            <lx:label>Checkout Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8291-13">
        <xs:annotation>
          <xs:documentation>Office Cashier</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8291-13</lx:path>
            <lx:label>Office Cashier</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8292">
        <xs:annotation>
          <xs:documentation>TICKET SALESPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8292</lx:path>
            <lx:label>TICKET SALESPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8292-11">
        <xs:annotation>
          <xs:documentation>Ticket Seller</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8292-11</lx:path>
            <lx:label>Ticket Seller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8292-13">
        <xs:annotation>
          <xs:documentation>Transport Conductor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8292-13</lx:path>
            <lx:label>Transport Conductor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8293">
        <xs:annotation>
          <xs:documentation>STREET VENDORS AND RELATED WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8293</lx:path>
            <lx:label>STREET VENDORS AND RELATED WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8293-11">
        <xs:annotation>
          <xs:documentation>Street Vendor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8293-11</lx:path>
            <lx:label>Street Vendor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8293-13">
        <xs:annotation>
          <xs:documentation>Cash Van Salesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8293-13</lx:path>
            <lx:label>Cash Van Salesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8293-15">
        <xs:annotation>
          <xs:documentation>Door-to-Door Salesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8293-15</lx:path>
            <lx:label>Door-to-Door Salesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8294">
        <xs:annotation>
          <xs:documentation>TELEMARKETERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8294</lx:path>
            <lx:label>TELEMARKETERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8294-11">
        <xs:annotation>
          <xs:documentation>Telemarketer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8294-11</lx:path>
            <lx:label>Telemarketer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8295">
        <xs:annotation>
          <xs:documentation>SALES DEMONSTRATORS AND MODELS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8295</lx:path>
            <lx:label>SALES DEMONSTRATORS AND MODELS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8295-11">
        <xs:annotation>
          <xs:documentation>Sales Demonstrator</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8295-11</lx:path>
            <lx:label>Sales Demonstrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8295-13">
        <xs:annotation>
          <xs:documentation>Model</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8295-13</lx:path>
            <lx:label>Model</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8296">
        <xs:annotation>
          <xs:documentation>SERVICE STATION ATTENDANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8296</lx:path>
            <lx:label>SERVICE STATION ATTENDANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8296-11">
        <xs:annotation>
          <xs:documentation>Service Station Attendant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8296-11</lx:path>
            <lx:label>Service Station Attendant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8297">
        <xs:annotation>
          <xs:documentation>SALES AND SERVICE TRAINEES</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8297</lx:path>
            <lx:label>SALES AND SERVICE TRAINEES</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8297-11">
        <xs:annotation>
          <xs:documentation>Sales and Service Trainee</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8297-11</lx:path>
            <lx:label>Sales and Service Trainee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8299">
        <xs:annotation>
          <xs:documentation>OTHER ELEMENTARY SALES WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8299</lx:path>
            <lx:label>OTHER ELEMENTARY SALES WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8299-11">
        <xs:annotation>
          <xs:documentation>Rental Salesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8299-11</lx:path>
            <lx:label>Rental Salesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8299-13">
        <xs:annotation>
          <xs:documentation>Materials Recycler</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8299-13</lx:path>
            <lx:label>Materials Recycler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8299-79">
        <xs:annotation>
          <xs:documentation>Elementary Sales Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8299-79</lx:path>
            <lx:label>Elementary Sales Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="83">
        <xs:annotation>
          <xs:documentation>ELEMENTARY SERVICE WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.83</lx:path>
            <lx:label>ELEMENTARY SERVICE WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831">
        <xs:annotation>
          <xs:documentation>ELEMENTARY SERVICE WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.831</lx:path>
            <lx:label>ELEMENTARY SERVICE WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8311">
        <xs:annotation>
          <xs:documentation>GUARDS AND SECURITY OFFICERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8311</lx:path>
            <lx:label>GUARDS AND SECURITY OFFICERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8311-11">
        <xs:annotation>
          <xs:documentation>Security Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8311-11</lx:path>
            <lx:label>Security Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8311-13">
        <xs:annotation>
          <xs:documentation>Armoured Car Escort</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8311-13</lx:path>
            <lx:label>Armoured Car Escort</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8311-79">
        <xs:annotation>
          <xs:documentation>Guards and Security Officers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8311-79</lx:path>
            <lx:label>Guards and Security Officers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8312">
        <xs:annotation>
          <xs:documentation>USHERS, PORTERS AND RELATED WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8312</lx:path>
            <lx:label>USHERS, PORTERS AND RELATED WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8312-11">
        <xs:annotation>
          <xs:documentation>Ticket Collector or Usher</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8312-11</lx:path>
            <lx:label>Ticket Collector or Usher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8312-13">
        <xs:annotation>
          <xs:documentation>Luggage Porter or Doorperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8312-13</lx:path>
            <lx:label>Luggage Porter or Doorperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8313">
        <xs:annotation>
          <xs:documentation>DOMESTIC HOUSEKEEPERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8313</lx:path>
            <lx:label>DOMESTIC HOUSEKEEPERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8313-11">
        <xs:annotation>
          <xs:documentation>Domestic Housekeeper</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8313-11</lx:path>
            <lx:label>Domestic Housekeeper</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8314">
        <xs:annotation>
          <xs:documentation>CARETAKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8314</lx:path>
            <lx:label>CARETAKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8314-11">
        <xs:annotation>
          <xs:documentation>Caretaker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8314-11</lx:path>
            <lx:label>Caretaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8315">
        <xs:annotation>
          <xs:documentation>LAUNDRY WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8315</lx:path>
            <lx:label>LAUNDRY WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8315-11">
        <xs:annotation>
          <xs:documentation>Laundry Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8315-11</lx:path>
            <lx:label>Laundry Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8319">
        <xs:annotation>
          <xs:documentation>OTHER ELEMENTARY SERVICE WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8319</lx:path>
            <lx:label>OTHER ELEMENTARY SERVICE WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8319-11">
        <xs:annotation>
          <xs:documentation>Railways Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8319-11</lx:path>
            <lx:label>Railways Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8319-13">
        <xs:annotation>
          <xs:documentation>Hair and Beauty Salon Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8319-13</lx:path>
            <lx:label>Hair and Beauty Salon Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8319-15">
        <xs:annotation>
          <xs:documentation>Sterilisation Technology Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8319-15</lx:path>
            <lx:label>Sterilisation Technology Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8319-17">
        <xs:annotation>
          <xs:documentation>Car Park Attendant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8319-17</lx:path>
            <lx:label>Car Park Attendant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8319-19">
        <xs:annotation>
          <xs:documentation>Crossing Supervisor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8319-19</lx:path>
            <lx:label>Crossing Supervisor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8319-21">
        <xs:annotation>
          <xs:documentation>Trolley Collector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8319-21</lx:path>
            <lx:label>Trolley Collector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8319-23">
        <xs:annotation>
          <xs:documentation>Leaflet and Newspaper Deliverer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8319-23</lx:path>
            <lx:label>Leaflet and Newspaper Deliverer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8319-25">
        <xs:annotation>
          <xs:documentation>Examination Supervisor</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8319-25</lx:path>
            <lx:label>Examination Supervisor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8319-27">
        <xs:annotation>
          <xs:documentation>Prostitute or Escort</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8319-27</lx:path>
            <lx:label>Prostitute or Escort</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8319-79">
        <xs:annotation>
          <xs:documentation>Elementary Service Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.8319-79</lx:path>
            <lx:label>Elementary Service Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9">
        <xs:annotation>
          <xs:documentation>LABOURERS AND RELATED WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9</lx:path>
            <lx:label>LABOURERS AND RELATED WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="91">
        <xs:annotation>
          <xs:documentation>CLEANERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.91</lx:path>
            <lx:label>CLEANERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="911">
        <xs:annotation>
          <xs:documentation>CLEANERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.911</lx:path>
            <lx:label>CLEANERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9111">
        <xs:annotation>
          <xs:documentation>CLEANERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9111</lx:path>
            <lx:label>CLEANERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9111-11">
        <xs:annotation>
          <xs:documentation>Commercial Cleaner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9111-11</lx:path>
            <lx:label>Commercial Cleaner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9111-13">
        <xs:annotation>
          <xs:documentation>Domestic Cleaner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9111-13</lx:path>
            <lx:label>Domestic Cleaner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9111-15">
        <xs:annotation>
          <xs:documentation>Carpet Cleaner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9111-15</lx:path>
            <lx:label>Carpet Cleaner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9111-17">
        <xs:annotation>
          <xs:documentation>Vehicle Cleaner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9111-17</lx:path>
            <lx:label>Vehicle Cleaner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9111-19">
        <xs:annotation>
          <xs:documentation>Window Cleaner</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9111-19</lx:path>
            <lx:label>Window Cleaner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9111-79">
        <xs:annotation>
          <xs:documentation>Cleaners nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9111-79</lx:path>
            <lx:label>Cleaners nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="92">
        <xs:annotation>
          <xs:documentation>FACTORY LABOURERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.92</lx:path>
            <lx:label>FACTORY LABOURERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="921">
        <xs:annotation>
          <xs:documentation>PROCESS WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.921</lx:path>
            <lx:label>PROCESS WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9211">
        <xs:annotation>
          <xs:documentation>ENGINEERING PRODUCTION PROCESS WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9211</lx:path>
            <lx:label>ENGINEERING PRODUCTION PROCESS WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9211-11">
        <xs:annotation>
          <xs:documentation>Engineering Production Process Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9211-11</lx:path>
            <lx:label>Engineering Production Process Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9212">
        <xs:annotation>
          <xs:documentation>PRODUCT ASSEMBLERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9212</lx:path>
            <lx:label>PRODUCT ASSEMBLERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9212-11">
        <xs:annotation>
          <xs:documentation>Product Assembler</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9212-11</lx:path>
            <lx:label>Product Assembler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9213">
        <xs:annotation>
          <xs:documentation>MEAT AND FISH PROCESS WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9213</lx:path>
            <lx:label>MEAT AND FISH PROCESS WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9213-11">
        <xs:annotation>
          <xs:documentation>Meatworks Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9213-11</lx:path>
            <lx:label>Meatworks Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9213-13">
        <xs:annotation>
          <xs:documentation>Meat Boner and Slicer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9213-13</lx:path>
            <lx:label>Meat Boner and Slicer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9213-15">
        <xs:annotation>
          <xs:documentation>Poultry Process Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9213-15</lx:path>
            <lx:label>Poultry Process Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9213-17">
        <xs:annotation>
          <xs:documentation>Fish Process Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9213-17</lx:path>
            <lx:label>Fish Process Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9214">
        <xs:annotation>
          <xs:documentation>OTHER FOOD FACTORY HANDS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9214</lx:path>
            <lx:label>OTHER FOOD FACTORY HANDS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9214-11">
        <xs:annotation>
          <xs:documentation>Food and Drink Processing Machine Attendant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9214-11</lx:path>
            <lx:label>Food and Drink Processing Machine Attendant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9214-13">
        <xs:annotation>
          <xs:documentation>Baking Factory Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9214-13</lx:path>
            <lx:label>Baking Factory Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9214-15">
        <xs:annotation>
          <xs:documentation>Dairy Factory Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9214-15</lx:path>
            <lx:label>Dairy Factory Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9214-17">
        <xs:annotation>
          <xs:documentation>Fruit and Vegetable Factory Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9214-17</lx:path>
            <lx:label>Fruit and Vegetable Factory Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9214-19">
        <xs:annotation>
          <xs:documentation>Food Products Millhand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9214-19</lx:path>
            <lx:label>Food Products Millhand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9214-79">
        <xs:annotation>
          <xs:documentation>Food Factory Hands nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9214-79</lx:path>
            <lx:label>Food Factory Hands nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9215">
        <xs:annotation>
          <xs:documentation>WOOD PRODUCTS FACTORY HANDS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9215</lx:path>
            <lx:label>WOOD PRODUCTS FACTORY HANDS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9215-11">
        <xs:annotation>
          <xs:documentation>Sawmill Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9215-11</lx:path>
            <lx:label>Sawmill Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9215-13">
        <xs:annotation>
          <xs:documentation>Timberyard Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9215-13</lx:path>
            <lx:label>Timberyard Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9215-15">
        <xs:annotation>
          <xs:documentation>Wood and Wood Products Factory Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9215-15</lx:path>
            <lx:label>Wood and Wood Products Factory Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9215-17">
        <xs:annotation>
          <xs:documentation>Pulp and Paper Mill General Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9215-17</lx:path>
            <lx:label>Pulp and Paper Mill General Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9219">
        <xs:annotation>
          <xs:documentation>OTHER PROCESS WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9219</lx:path>
            <lx:label>OTHER PROCESS WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9219-11">
        <xs:annotation>
          <xs:documentation>Cement and Concrete Plant Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9219-11</lx:path>
            <lx:label>Cement and Concrete Plant Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9219-13">
        <xs:annotation>
          <xs:documentation>Clay Processing Factory Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9219-13</lx:path>
            <lx:label>Clay Processing Factory Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9219-15">
        <xs:annotation>
          <xs:documentation>Chemical Plant Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9219-15</lx:path>
            <lx:label>Chemical Plant Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9219-17">
        <xs:annotation>
          <xs:documentation>Rubber and Plastics Factory Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9219-17</lx:path>
            <lx:label>Rubber and Plastics Factory Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9219-19">
        <xs:annotation>
          <xs:documentation>Hide and Skin Processing Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9219-19</lx:path>
            <lx:label>Hide and Skin Processing Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9219-21">
        <xs:annotation>
          <xs:documentation>Glass Processing Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9219-21</lx:path>
            <lx:label>Glass Processing Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9219-23">
        <xs:annotation>
          <xs:documentation>Fabric and Textile Factory Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9219-23</lx:path>
            <lx:label>Fabric and Textile Factory Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9219-25">
        <xs:annotation>
          <xs:documentation>Footwear Factory Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9219-25</lx:path>
            <lx:label>Footwear Factory Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9219-79">
        <xs:annotation>
          <xs:documentation>Process Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9219-79</lx:path>
            <lx:label>Process Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="922">
        <xs:annotation>
          <xs:documentation>PRODUCT PACKAGERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.922</lx:path>
            <lx:label>PRODUCT PACKAGERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9221">
        <xs:annotation>
          <xs:documentation>HAND PACKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9221</lx:path>
            <lx:label>HAND PACKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9221-11">
        <xs:annotation>
          <xs:documentation>Hand Packer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9221-11</lx:path>
            <lx:label>Hand Packer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9222">
        <xs:annotation>
          <xs:documentation>PACKAGERS AND CONTAINER FILLERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9222</lx:path>
            <lx:label>PACKAGERS AND CONTAINER FILLERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9222-11">
        <xs:annotation>
          <xs:documentation>Packager and Container Filler</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9222-11</lx:path>
            <lx:label>Packager and Container Filler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="99">
        <xs:annotation>
          <xs:documentation>OTHER LABOURERS AND RELATED WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.99</lx:path>
            <lx:label>OTHER LABOURERS AND RELATED WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="991">
        <xs:annotation>
          <xs:documentation>MINING, CONSTRUCTION AND RELATED LABOURERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.991</lx:path>
            <lx:label>MINING, CONSTRUCTION AND RELATED LABOURERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9911">
        <xs:annotation>
          <xs:documentation>MINING SUPPORT WORKERS AND DRILLER'S ASSISTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9911</lx:path>
            <lx:label>MINING SUPPORT WORKERS AND DRILLER'S ASSISTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9911-11">
        <xs:annotation>
          <xs:documentation>Mining Support Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9911-11</lx:path>
            <lx:label>Mining Support Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9911-13">
        <xs:annotation>
          <xs:documentation>Driller's Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9911-13</lx:path>
            <lx:label>Driller's Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9912">
        <xs:annotation>
          <xs:documentation>EARTHMOVING LABOURERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9912</lx:path>
            <lx:label>EARTHMOVING LABOURERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9912-11">
        <xs:annotation>
          <xs:documentation>Earthmoving Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9912-11</lx:path>
            <lx:label>Earthmoving Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9913">
        <xs:annotation>
          <xs:documentation>PAVING AND SURFACING LABOURERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9913</lx:path>
            <lx:label>PAVING AND SURFACING LABOURERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9913-11">
        <xs:annotation>
          <xs:documentation>Paving and Surfacing Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9913-11</lx:path>
            <lx:label>Paving and Surfacing Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9914">
        <xs:annotation>
          <xs:documentation>SURVEY HANDS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9914</lx:path>
            <lx:label>SURVEY HANDS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9914-11">
        <xs:annotation>
          <xs:documentation>Survey Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9914-11</lx:path>
            <lx:label>Survey Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9915">
        <xs:annotation>
          <xs:documentation>RAILWAY LABOURERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9915</lx:path>
            <lx:label>RAILWAY LABOURERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9915-11">
        <xs:annotation>
          <xs:documentation>Railway Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9915-11</lx:path>
            <lx:label>Railway Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9916">
        <xs:annotation>
          <xs:documentation>CONSTRUCTION AND PLUMBER'S ASSISTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9916</lx:path>
            <lx:label>CONSTRUCTION AND PLUMBER'S ASSISTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9916-11">
        <xs:annotation>
          <xs:documentation>Construction Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9916-11</lx:path>
            <lx:label>Construction Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9916-13">
        <xs:annotation>
          <xs:documentation>Plumber's Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9916-13</lx:path>
            <lx:label>Plumber's Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9917">
        <xs:annotation>
          <xs:documentation>CONCRETERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9917</lx:path>
            <lx:label>CONCRETERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9917-11">
        <xs:annotation>
          <xs:documentation>Concreter</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9917-11</lx:path>
            <lx:label>Concreter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9918">
        <xs:annotation>
          <xs:documentation>ELECTRICAL AND TELECOMMUNICATIONS TRADES ASSISTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9918</lx:path>
            <lx:label>ELECTRICAL AND TELECOMMUNICATIONS TRADES ASSISTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9918-11">
        <xs:annotation>
          <xs:documentation>Electrical or Telecommunications Trades Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9918-11</lx:path>
            <lx:label>Electrical or Telecommunications Trades Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9919">
        <xs:annotation>
          <xs:documentation>OTHER MINING, CONSTRUCTION AND RELATED LABOURERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9919</lx:path>
            <lx:label>OTHER MINING, CONSTRUCTION AND RELATED LABOURERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9919-11">
        <xs:annotation>
          <xs:documentation>Lagger</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9919-11</lx:path>
            <lx:label>Lagger</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9919-13">
        <xs:annotation>
          <xs:documentation>Crane Chaser</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9919-13</lx:path>
            <lx:label>Crane Chaser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9919-15">
        <xs:annotation>
          <xs:documentation>Fence Erector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9919-15</lx:path>
            <lx:label>Fence Erector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="992">
        <xs:annotation>
          <xs:documentation>AGRICULTURAL AND HORTICULTURAL LABOURERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.992</lx:path>
            <lx:label>AGRICULTURAL AND HORTICULTURAL LABOURERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9921">
        <xs:annotation>
          <xs:documentation>FARM HANDS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9921</lx:path>
            <lx:label>FARM HANDS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9921-11">
        <xs:annotation>
          <xs:documentation>General Farm Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9921-11</lx:path>
            <lx:label>General Farm Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9921-13">
        <xs:annotation>
          <xs:documentation>Fruit, Vegetable or Nut Farm Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9921-13</lx:path>
            <lx:label>Fruit, Vegetable or Nut Farm Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9921-15">
        <xs:annotation>
          <xs:documentation>Stud Hand or Stable Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9921-15</lx:path>
            <lx:label>Stud Hand or Stable Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9921-17">
        <xs:annotation>
          <xs:documentation>Shearing Shed Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9921-17</lx:path>
            <lx:label>Shearing Shed Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9921-79">
        <xs:annotation>
          <xs:documentation>Farm Hands nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9921-79</lx:path>
            <lx:label>Farm Hands nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9922">
        <xs:annotation>
          <xs:documentation>NURSERY AND GARDEN LABOURERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9922</lx:path>
            <lx:label>NURSERY AND GARDEN LABOURERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9922-11">
        <xs:annotation>
          <xs:documentation>Horticultural Nursery Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9922-11</lx:path>
            <lx:label>Horticultural Nursery Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9922-13">
        <xs:annotation>
          <xs:documentation>Garden Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9922-13</lx:path>
            <lx:label>Garden Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9929">
        <xs:annotation>
          <xs:documentation>OTHER AGRICULTURAL AND HORTICULTURAL LABOURERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9929</lx:path>
            <lx:label>OTHER AGRICULTURAL AND HORTICULTURAL LABOURERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9929-11">
        <xs:annotation>
          <xs:documentation>Shooter-Trapper</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9929-11</lx:path>
            <lx:label>Shooter-Trapper</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9929-13">
        <xs:annotation>
          <xs:documentation>Rural Trainee</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9929-13</lx:path>
            <lx:label>Rural Trainee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9929-79">
        <xs:annotation>
          <xs:documentation>Agricultural and Horticultural Labourers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9929-79</lx:path>
            <lx:label>Agricultural and Horticultural Labourers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="993">
        <xs:annotation>
          <xs:documentation>ELEMENTARY FOOD PREPARATION AND RELATED WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.993</lx:path>
            <lx:label>ELEMENTARY FOOD PREPARATION AND RELATED WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9931">
        <xs:annotation>
          <xs:documentation>KITCHENHANDS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9931</lx:path>
            <lx:label>KITCHENHANDS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9931-11">
        <xs:annotation>
          <xs:documentation>Kitchenhand</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9931-11</lx:path>
            <lx:label>Kitchenhand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9932">
        <xs:annotation>
          <xs:documentation>FAST FOOD COOKS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9932</lx:path>
            <lx:label>FAST FOOD COOKS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9932-11">
        <xs:annotation>
          <xs:documentation>Fast Food Cook</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9932-11</lx:path>
            <lx:label>Fast Food Cook</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9933">
        <xs:annotation>
          <xs:documentation>FOOD TRADES ASSISTANTS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9933</lx:path>
            <lx:label>FOOD TRADES ASSISTANTS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9933-11">
        <xs:annotation>
          <xs:documentation>Pastrycook's Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9933-11</lx:path>
            <lx:label>Pastrycook's Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9933-79">
        <xs:annotation>
          <xs:documentation>Food Trades Assistants nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9933-79</lx:path>
            <lx:label>Food Trades Assistants nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="999">
        <xs:annotation>
          <xs:documentation>MISCELLANEOUS LABOURERS AND RELATED WORKERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.999</lx:path>
            <lx:label>MISCELLANEOUS LABOURERS AND RELATED WORKERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9991">
        <xs:annotation>
          <xs:documentation>GARBAGE COLLECTORS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9991</lx:path>
            <lx:label>GARBAGE COLLECTORS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9991-11">
        <xs:annotation>
          <xs:documentation>Garbage Collector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9991-11</lx:path>
            <lx:label>Garbage Collector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9992">
        <xs:annotation>
          <xs:documentation>FREIGHT AND FURNITURE HANDLERS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9992</lx:path>
            <lx:label>FREIGHT AND FURNITURE HANDLERS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9992-11">
        <xs:annotation>
          <xs:documentation>Freight Handler (Road and Rail)</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9992-11</lx:path>
            <lx:label>Freight Handler (Road and Rail)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9992-13">
        <xs:annotation>
          <xs:documentation>Waterside Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9992-13</lx:path>
            <lx:label>Waterside Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9992-15">
        <xs:annotation>
          <xs:documentation>Truck Driver's Offsider</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9992-15</lx:path>
            <lx:label>Truck Driver's Offsider</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9993">
        <xs:annotation>
          <xs:documentation>HANDYPERSONS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9993</lx:path>
            <lx:label>HANDYPERSONS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9993-11">
        <xs:annotation>
          <xs:documentation>Handyperson</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9993-11</lx:path>
            <lx:label>Handyperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9999">
        <xs:annotation>
          <xs:documentation>OTHER MISCELLANEOUS</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9999</lx:path>
            <lx:label>OTHER MISCELLANEOUS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9999-11">
        <xs:annotation>
          <xs:documentation>Vending Machine Attendant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9999-11</lx:path>
            <lx:label>Vending Machine Attendant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9999-13">
        <xs:annotation>
          <xs:documentation>Sign Erector</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9999-13</lx:path>
            <lx:label>Sign Erector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9999-15">
        <xs:annotation>
          <xs:documentation>Mechanic's Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9999-15</lx:path>
            <lx:label>Mechanic's Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9999-17">
        <xs:annotation>
          <xs:documentation>Stormwater, Drainage and Sewerage System Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9999-17</lx:path>
            <lx:label>Stormwater, Drainage and Sewerage System Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9999-79">
        <xs:annotation>
          <xs:documentation>Labourers and Related Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>absOccupationCodeList.9999-79</lx:path>
            <lx:label>Labourers and Related Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="accessTypeList">
    <xs:annotation>
      <xs:documentation>accessTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>accessTypeList</lx:path>
        <lx:label>accessTypeList</lx:label>
        <li:transactions>ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Full Access">
        <xs:annotation>
          <xs:documentation>Full Access</xs:documentation>
          <xs:appinfo>
            <lx:path>accessTypeList.Full Access</lx:path>
            <lx:label>Full Access</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Information Only Access">
        <xs:annotation>
          <xs:documentation>Information Only Access</xs:documentation>
          <xs:appinfo>
            <lx:path>accessTypeList.Information Only Access</lx:path>
            <lx:label>Information Only Access</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="No Access">
        <xs:annotation>
          <xs:documentation>No Access</xs:documentation>
          <xs:appinfo>
            <lx:path>accessTypeList.No Access</lx:path>
            <lx:label>No Access</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="accountHoldingList">
    <xs:annotation>
      <xs:documentation>accountHoldingList</xs:documentation>
      <xs:appinfo>
        <lx:path>accountHoldingList</lx:path>
        <lx:label>accountHoldingList</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Joint">
        <xs:annotation>
          <xs:documentation>This account is jointly held by two or more individuals.</xs:documentation>
          <xs:appinfo>
            <lx:path>accountHoldingList.Joint</lx:path>
            <lx:label>Joint</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sole">
        <xs:annotation>
          <xs:documentation>This account is held by a sole individual.</xs:documentation>
          <xs:appinfo>
            <lx:path>accountHoldingList.Sole</lx:path>
            <lx:label>Sole</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="accountStatusList">
    <xs:annotation>
      <xs:documentation>List of status values to describe the existence of an account.</xs:documentation>
      <xs:appinfo>
        <lx:path>accountStatusList</lx:path>
        <lx:label>Account Status List</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Add Existing">
        <xs:annotation>
          <xs:documentation>Add Existing</xs:documentation>
          <xs:appinfo>
            <lx:path>accountStatusList.Add Existing</lx:path>
            <lx:label>Add Existing</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Create New">
        <xs:annotation>
          <xs:documentation>Create New</xs:documentation>
          <xs:appinfo>
            <lx:path>accountStatusList.Create New</lx:path>
            <lx:label>Create New</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Upgrade Existing">
        <xs:annotation>
          <xs:documentation>Upgrade Existing</xs:documentation>
          <xs:appinfo>
            <lx:path>accountStatusList.Upgrade Existing</lx:path>
            <lx:label>Upgrade Existing</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="addressTypeList">
    <xs:annotation>
      <xs:documentation>addressTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>addressTypeList</lx:path>
        <lx:label>addressTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="DX Box">
        <xs:annotation>
          <xs:documentation>DX Box</xs:documentation>
          <xs:appinfo>
            <lx:path>addressTypeList.DX Box</lx:path>
            <lx:label>DX Box</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Non Standard">
        <xs:annotation>
          <xs:documentation>Non Standard</xs:documentation>
          <xs:appinfo>
            <lx:path>addressTypeList.Non Standard</lx:path>
            <lx:label>Non Standard</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PO Box or Bag">
        <xs:annotation>
          <xs:documentation>PO Box or Bag</xs:documentation>
          <xs:appinfo>
            <lx:path>addressTypeList.PO Box or Bag</lx:path>
            <lx:label>PO Box or Bag</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Standard">
        <xs:annotation>
          <xs:documentation>Standard</xs:documentation>
          <xs:appinfo>
            <lx:path>addressTypeList.Standard</lx:path>
            <lx:label>Standard</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="aggregatedTransactionsCategoryTypeList">
    <xs:annotation>
      <xs:documentation>aggregatedTransactionsCategoryTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>aggregatedTransactionsCategoryTypeList</lx:path>
        <lx:label>aggregatedTransactionsCategoryTypeList</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Alimony or Maintenance Payments">
        <xs:annotation>
          <xs:documentation>Alimony or Maintenance Payments</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Alimony or Maintenance Payments</lx:path>
            <lx:label>Alimony or Maintenance Payments</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ambulance Insurance">
        <xs:annotation>
          <xs:documentation>Ambulance Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Ambulance Insurance</lx:path>
            <lx:label>Ambulance Insurance</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Animal Purchases">
        <xs:annotation>
          <xs:documentation>Animal Purchases</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Animal Purchases</lx:path>
            <lx:label>Animal Purchases</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Annuities Income">
        <xs:annotation>
          <xs:documentation>Annuities Income</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Annuities Income</lx:path>
            <lx:label>Annuities Income</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Capital Building Costs">
        <xs:annotation>
          <xs:documentation>Capital Building Costs</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Capital Building Costs</lx:path>
            <lx:label>Capital Building Costs</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Child Care">
        <xs:annotation>
          <xs:documentation>Child Care</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Child Care</lx:path>
            <lx:label>Child Care</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Child Support Payment Income">
        <xs:annotation>
          <xs:documentation>Child Support Payment Income</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Child Support Payment Income</lx:path>
            <lx:label>Child Support Payment Income</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Clothing and Accessories">
        <xs:annotation>
          <xs:documentation>Clothing and Accessories</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Clothing and Accessories</lx:path>
            <lx:label>Clothing and Accessories</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Contracted Recreational Memberships and Subscriptions">
        <xs:annotation>
          <xs:documentation>Contracted Recreational Memberships and Subscriptions</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Contracted Recreational Memberships and Subscriptions</lx:path>
            <lx:label>Contracted Recreational Memberships and Subscriptions</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Card Repayments">
        <xs:annotation>
          <xs:documentation>Credit Card Repayments</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Credit Card Repayments</lx:path>
            <lx:label>Credit Card Repayments</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dishonour Fees and Fines">
        <xs:annotation>
          <xs:documentation>Dishonour Fees and Fines</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Dishonour Fees and Fines</lx:path>
            <lx:label>Dishonour Fees and Fines</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dividends Income">
        <xs:annotation>
          <xs:documentation>Dividends Income</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Dividends Income</lx:path>
            <lx:label>Dividends Income</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Finance Repayments Excluding Mortgages and Credit Cards">
        <xs:annotation>
          <xs:documentation>Finance Repayments Excluding Mortgages and Credit Cards</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Finance Repayments Excluding Mortgages and Credit Cards</lx:path>
            <lx:label>Finance Repayments Excluding Mortgages and Credit Cards</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Financial Institution Charges, Fees and Related Duties and Taxes">
        <xs:annotation>
          <xs:documentation>Financial Institution Charges, Fees and Related Duties and Taxes</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Financial Institution Charges, Fees and Related Duties and Taxes</lx:path>
            <lx:label>Financial Institution Charges, Fees and Related Duties and Taxes</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Foreign Sourced Income">
        <xs:annotation>
          <xs:documentation>Foreign Sourced Income</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Foreign Sourced Income</lx:path>
            <lx:label>Foreign Sourced Income</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gambling">
        <xs:annotation>
          <xs:documentation>Gambling</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Gambling</lx:path>
            <lx:label>Gambling</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gifts and Donations">
        <xs:annotation>
          <xs:documentation>Gifts and Donations</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Gifts and Donations</lx:path>
            <lx:label>Gifts and Donations</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits">
        <xs:annotation>
          <xs:documentation>Government Benefits</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Government Benefits</lx:path>
            <lx:label>Government Benefits</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits - Abstudy">
        <xs:annotation>
          <xs:documentation>Government Benefits - Abstudy</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Government Benefits - Abstudy</lx:path>
            <lx:label>Government Benefits - Abstudy</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits - Age Pension">
        <xs:annotation>
          <xs:documentation>Government Benefits - Age Pension</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Government Benefits - Age Pension</lx:path>
            <lx:label>Government Benefits - Age Pension</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits - Austudy">
        <xs:annotation>
          <xs:documentation>Government Benefits - Austudy</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Government Benefits - Austudy</lx:path>
            <lx:label>Government Benefits - Austudy</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits - Carer Payment">
        <xs:annotation>
          <xs:documentation>Government Benefits - Carer Payment</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Government Benefits - Carer Payment</lx:path>
            <lx:label>Government Benefits - Carer Payment</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits - Crisis Payment">
        <xs:annotation>
          <xs:documentation>Government Benefits - Crisis Payment</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Government Benefits - Crisis Payment</lx:path>
            <lx:label>Government Benefits - Crisis Payment</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits - Disability Support Pension">
        <xs:annotation>
          <xs:documentation>Government Benefits - Disability Support Pension</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Government Benefits - Disability Support Pension</lx:path>
            <lx:label>Government Benefits - Disability Support Pension</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits - Family Allowance">
        <xs:annotation>
          <xs:documentation>Government Benefits - Family Allowance</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Government Benefits - Family Allowance</lx:path>
            <lx:label>Government Benefits - Family Allowance</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits - Mobility Allowance">
        <xs:annotation>
          <xs:documentation>Government Benefits - Mobility Allowance</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Government Benefits - Mobility Allowance</lx:path>
            <lx:label>Government Benefits - Mobility Allowance</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits - Newstart">
        <xs:annotation>
          <xs:documentation>Government Benefits - Newstart</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Government Benefits - Newstart</lx:path>
            <lx:label>Government Benefits - Newstart</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits - Other">
        <xs:annotation>
          <xs:documentation>Government Benefits - Other</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Government Benefits - Other</lx:path>
            <lx:label>Government Benefits - Other</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits - Parenting Payment">
        <xs:annotation>
          <xs:documentation>Government Benefits - Parenting Payment</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Government Benefits - Parenting Payment</lx:path>
            <lx:label>Government Benefits - Parenting Payment</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits - Widows Allowance">
        <xs:annotation>
          <xs:documentation>Government Benefits - Widows Allowance</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Government Benefits - Widows Allowance</lx:path>
            <lx:label>Government Benefits - Widows Allowance</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits - Youth Allowance">
        <xs:annotation>
          <xs:documentation>Government Benefits - Youth Allowance</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Government Benefits - Youth Allowance</lx:path>
            <lx:label>Government Benefits - Youth Allowance</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Groceries">
        <xs:annotation>
          <xs:documentation>Groceries</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Groceries</lx:path>
            <lx:label>Groceries</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HECS">
        <xs:annotation>
          <xs:documentation>HECS</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.HECS</lx:path>
            <lx:label>HECS</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Higher Education">
        <xs:annotation>
          <xs:documentation>Higher Education</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Higher Education</lx:path>
            <lx:label>Higher Education</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Holidays, Travel and Accommodation">
        <xs:annotation>
          <xs:documentation>Holidays, Travel and Accommodation</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Holidays, Travel and Accommodation</lx:path>
            <lx:label>Holidays, Travel and Accommodation</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hospital, Medical and Dental Insurance">
        <xs:annotation>
          <xs:documentation>Hospital, Medical and Dental Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Hospital, Medical and Dental Insurance</lx:path>
            <lx:label>Hospital, Medical and Dental Insurance</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="House and Contents Insurance">
        <xs:annotation>
          <xs:documentation>House and Contents Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.House and Contents Insurance</lx:path>
            <lx:label>House and Contents Insurance</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Household and Garden Furnishings and Equipment">
        <xs:annotation>
          <xs:documentation>Household and Garden Furnishings and Equipment</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Household and Garden Furnishings and Equipment</lx:path>
            <lx:label>Household and Garden Furnishings and Equipment</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Household Services">
        <xs:annotation>
          <xs:documentation>Household Services</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Household Services</lx:path>
            <lx:label>Household Services</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Income Tax">
        <xs:annotation>
          <xs:documentation>Income Tax</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Income Tax</lx:path>
            <lx:label>Income Tax</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Interest Income">
        <xs:annotation>
          <xs:documentation>Interest Income</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Interest Income</lx:path>
            <lx:label>Interest Income</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Investment Expenses">
        <xs:annotation>
          <xs:documentation>Investment Expenses</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Investment Expenses</lx:path>
            <lx:label>Investment Expenses</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Jewellery and Watches">
        <xs:annotation>
          <xs:documentation>Jewellery and Watches</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Jewellery and Watches</lx:path>
            <lx:label>Jewellery and Watches</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Life Insurance">
        <xs:annotation>
          <xs:documentation>Life Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Life Insurance</lx:path>
            <lx:label>Life Insurance</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Medical and Health">
        <xs:annotation>
          <xs:documentation>Medical and Health</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Medical and Health</lx:path>
            <lx:label>Medical and Health</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mortgage Repayments">
        <xs:annotation>
          <xs:documentation>Mortgage Repayments</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Mortgage Repayments</lx:path>
            <lx:label>Mortgage Repayments</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Net Salary">
        <xs:annotation>
          <xs:documentation>Net Salary</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Net Salary</lx:path>
            <lx:label>Net Salary</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Nursing Home Charges">
        <xs:annotation>
          <xs:documentation>Nursing Home Charges</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Nursing Home Charges</lx:path>
            <lx:label>Nursing Home Charges</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other Income">
        <xs:annotation>
          <xs:documentation>Other Income</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Other Income</lx:path>
            <lx:label>Other Income</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other Insurance">
        <xs:annotation>
          <xs:documentation>Other Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Other Insurance</lx:path>
            <lx:label>Other Insurance</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Personal Care">
        <xs:annotation>
          <xs:documentation>Personal Care</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Personal Care</lx:path>
            <lx:label>Personal Care</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Private Pension Income">
        <xs:annotation>
          <xs:documentation>Private Pension Income</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Private Pension Income</lx:path>
            <lx:label>Private Pension Income</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Private Schooling Expenses">
        <xs:annotation>
          <xs:documentation>Private Schooling Expenses</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Private Schooling Expenses</lx:path>
            <lx:label>Private Schooling Expenses</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Professional Services, Union and Association Fees">
        <xs:annotation>
          <xs:documentation>Professional Services, Union and Association Fees</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Professional Services, Union and Association Fees</lx:path>
            <lx:label>Professional Services, Union and Association Fees</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Property Maintenance">
        <xs:annotation>
          <xs:documentation>Property Maintenance</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Property Maintenance</lx:path>
            <lx:label>Property Maintenance</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Public or Government Primary and Secondary Education">
        <xs:annotation>
          <xs:documentation>Public or Government Primary and Secondary Education</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Public or Government Primary and Secondary Education</lx:path>
            <lx:label>Public or Government Primary and Secondary Education</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase of Property">
        <xs:annotation>
          <xs:documentation>Purchase of Property</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Purchase of Property</lx:path>
            <lx:label>Purchase of Property</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Recreation and Entertainment">
        <xs:annotation>
          <xs:documentation>Recreation and Entertainment</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Recreation and Entertainment</lx:path>
            <lx:label>Recreation and Entertainment</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rent Payments">
        <xs:annotation>
          <xs:documentation>Rent Payments</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Rent Payments</lx:path>
            <lx:label>Rent Payments</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rental Income">
        <xs:annotation>
          <xs:documentation>Rental Income</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Rental Income</lx:path>
            <lx:label>Rental Income</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sickness and Personal Accident Insurance">
        <xs:annotation>
          <xs:documentation>Sickness and Personal Accident Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Sickness and Personal Accident Insurance</lx:path>
            <lx:label>Sickness and Personal Accident Insurance</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Strata Fees">
        <xs:annotation>
          <xs:documentation>Strata Fees</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Strata Fees</lx:path>
            <lx:label>Strata Fees</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Superannuation Income">
        <xs:annotation>
          <xs:documentation>Superannuation Income</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Superannuation Income</lx:path>
            <lx:label>Superannuation Income</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Superannuation Payments">
        <xs:annotation>
          <xs:documentation>Superannuation Payments</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Superannuation Payments</lx:path>
            <lx:label>Superannuation Payments</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tax and Government Fees and Payments">
        <xs:annotation>
          <xs:documentation>Tax and Government Fees and Payments</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Tax and Government Fees and Payments</lx:path>
            <lx:label>Tax and Government Fees and Payments</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tax Refund">
        <xs:annotation>
          <xs:documentation>Tax Refund</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Tax Refund</lx:path>
            <lx:label>Tax Refund</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Telephone, Internet, Pay TV and Media Streaming Subscriptions">
        <xs:annotation>
          <xs:documentation>Telephone, Internet, Pay TV and Media Streaming Subscriptions</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Telephone, Internet, Pay TV and Media Streaming Subscriptions</lx:path>
            <lx:label>Telephone, Internet, Pay TV and Media Streaming Subscriptions</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Transport">
        <xs:annotation>
          <xs:documentation>Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Transport</lx:path>
            <lx:label>Transport</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Utilities">
        <xs:annotation>
          <xs:documentation>Utilities</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Utilities</lx:path>
            <lx:label>Utilities</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vehicle Insurance">
        <xs:annotation>
          <xs:documentation>Vehicle Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Vehicle Insurance</lx:path>
            <lx:label>Vehicle Insurance</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vehicle Purchase">
        <xs:annotation>
          <xs:documentation>Vehicle Purchase</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Vehicle Purchase</lx:path>
            <lx:label>Vehicle Purchase</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Workers Compensation Income">
        <xs:annotation>
          <xs:documentation>Workers Compensation Income</xs:documentation>
          <xs:appinfo>
            <lx:path>aggregatedTransactionsCategoryTypeList.Workers Compensation Income</lx:path>
            <lx:label>Workers Compensation Income</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="agriculturalAssetTypeList">
    <xs:annotation>
      <xs:documentation>agriculturalAssetTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>agriculturalAssetTypeList</lx:path>
        <lx:label>agriculturalAssetTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="All Terrain Vehicle">
        <xs:annotation>
          <xs:documentation>All Terrain Vehicle</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.All Terrain Vehicle</lx:path>
            <lx:label>All Terrain Vehicle</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Crops / Hay">
        <xs:annotation>
          <xs:documentation>Crops / Hay</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Crops / Hay</lx:path>
            <lx:label>Crops / Hay</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fencing">
        <xs:annotation>
          <xs:documentation>Fencing</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Fencing</lx:path>
            <lx:label>Fencing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fertiliser / Slurry">
        <xs:annotation>
          <xs:documentation>Fertiliser / Slurry</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Fertiliser / Slurry</lx:path>
            <lx:label>Fertiliser / Slurry</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GPS / Laser / Levelling">
        <xs:annotation>
          <xs:documentation>GPS / Laser / Levelling</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.GPS / Laser / Levelling</lx:path>
            <lx:label>GPS / Laser / Levelling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Handling / Storage">
        <xs:annotation>
          <xs:documentation>Handling / Storage</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Handling / Storage</lx:path>
            <lx:label>Handling / Storage</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Harvester / Header">
        <xs:annotation>
          <xs:documentation>Harvester / Header</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Harvester / Header</lx:path>
            <lx:label>Harvester / Header</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Irrigation">
        <xs:annotation>
          <xs:documentation>Irrigation</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Irrigation</lx:path>
            <lx:label>Irrigation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Livestock">
        <xs:annotation>
          <xs:documentation>Livestock</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Livestock</lx:path>
            <lx:label>Livestock</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Seeding / Planting">
        <xs:annotation>
          <xs:documentation>Seeding / Planting</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Seeding / Planting</lx:path>
            <lx:label>Seeding / Planting</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sprayer">
        <xs:annotation>
          <xs:documentation>Sprayer</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Sprayer</lx:path>
            <lx:label>Sprayer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Storage">
        <xs:annotation>
          <xs:documentation>Storage</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Storage</lx:path>
            <lx:label>Storage</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tillage Equipment">
        <xs:annotation>
          <xs:documentation>Tillage Equipment</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Tillage Equipment</lx:path>
            <lx:label>Tillage Equipment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tractor - Tracked">
        <xs:annotation>
          <xs:documentation>Tractor - Tracked</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Tractor - Tracked</lx:path>
            <lx:label>Tractor - Tracked</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tractor - Wheeled">
        <xs:annotation>
          <xs:documentation>Tractor - Wheeled</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Tractor - Wheeled</lx:path>
            <lx:label>Tractor - Wheeled</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tractor Front Attachments">
        <xs:annotation>
          <xs:documentation>Tractor Front Attachments</xs:documentation>
          <xs:appinfo>
            <lx:path>agriculturalAssetTypeList.Tractor Front Attachments</lx:path>
            <lx:label>Tractor Front Attachments</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="aircraftTypeList">
    <xs:annotation>
      <xs:documentation>aircraftTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>aircraftTypeList</lx:path>
        <lx:label>aircraftTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Engine">
        <xs:annotation>
          <xs:documentation>Engine</xs:documentation>
          <xs:appinfo>
            <lx:path>aircraftTypeList.Engine</lx:path>
            <lx:label>Engine</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fixed Wing - Cylinder">
        <xs:annotation>
          <xs:documentation>Fixed Wing - Cylinder</xs:documentation>
          <xs:appinfo>
            <lx:path>aircraftTypeList.Fixed Wing - Cylinder</lx:path>
            <lx:label>Fixed Wing - Cylinder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fixed Wing - Jet">
        <xs:annotation>
          <xs:documentation>Fixed Wing - Jet</xs:documentation>
          <xs:appinfo>
            <lx:path>aircraftTypeList.Fixed Wing - Jet</lx:path>
            <lx:label>Fixed Wing - Jet</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fixed Wing - Turbo Prop">
        <xs:annotation>
          <xs:documentation>Fixed Wing - Turbo Prop</xs:documentation>
          <xs:appinfo>
            <lx:path>aircraftTypeList.Fixed Wing - Turbo Prop</lx:path>
            <lx:label>Fixed Wing - Turbo Prop</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Helicopter">
        <xs:annotation>
          <xs:documentation>Helicopter</xs:documentation>
          <xs:appinfo>
            <lx:path>aircraftTypeList.Helicopter</lx:path>
            <lx:label>Helicopter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>aircraftTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ultra Light">
        <xs:annotation>
          <xs:documentation>Ultra Light</xs:documentation>
          <xs:appinfo>
            <lx:path>aircraftTypeList.Ultra Light</lx:path>
            <lx:label>Ultra Light</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="amenityTypeList">
    <xs:annotation>
      <xs:documentation>amenityTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>amenityTypeList</lx:path>
        <lx:label>amenityTypeList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Beach">
        <xs:annotation>
          <xs:documentation>Beach.</xs:documentation>
          <xs:appinfo>
            <lx:path>amenityTypeList.Beach</lx:path>
            <lx:label>Beach</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other.</xs:documentation>
          <xs:appinfo>
            <lx:path>amenityTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Post Office">
        <xs:annotation>
          <xs:documentation>Post Office.</xs:documentation>
          <xs:appinfo>
            <lx:path>amenityTypeList.Post Office</lx:path>
            <lx:label>Post Office</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Public Transport">
        <xs:annotation>
          <xs:documentation>Public Transport.</xs:documentation>
          <xs:appinfo>
            <lx:path>amenityTypeList.Public Transport</lx:path>
            <lx:label>Public Transport</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="School">
        <xs:annotation>
          <xs:documentation>School.</xs:documentation>
          <xs:appinfo>
            <lx:path>amenityTypeList.School</lx:path>
            <lx:label>School</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Town Center">
        <xs:annotation>
          <xs:documentation>Town Center.</xs:documentation>
          <xs:appinfo>
            <lx:path>amenityTypeList.Town Center</lx:path>
            <lx:label>Town Center</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="ancillaryFeatureTypeList">
    <xs:annotation>
      <xs:documentation>ancillaryFeatureTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>ancillaryFeatureTypeList</lx:path>
        <lx:label>ancillaryFeatureTypeList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Alfresco Area">
        <xs:annotation>
          <xs:documentation>Alfresco Area.</xs:documentation>
          <xs:appinfo>
            <lx:path>ancillaryFeatureTypeList.Alfresco Area</lx:path>
            <lx:label>Alfresco Area</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Barbecue">
        <xs:annotation>
          <xs:documentation>Barbecue.</xs:documentation>
          <xs:appinfo>
            <lx:path>ancillaryFeatureTypeList.Barbecue</lx:path>
            <lx:label>Barbecue</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Driveway">
        <xs:annotation>
          <xs:documentation>Driveway.</xs:documentation>
          <xs:appinfo>
            <lx:path>ancillaryFeatureTypeList.Driveway</lx:path>
            <lx:label>Driveway</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fencing">
        <xs:annotation>
          <xs:documentation>Fencing.</xs:documentation>
          <xs:appinfo>
            <lx:path>ancillaryFeatureTypeList.Fencing</lx:path>
            <lx:label>Fencing</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gazebo">
        <xs:annotation>
          <xs:documentation>Gazebo.</xs:documentation>
          <xs:appinfo>
            <lx:path>ancillaryFeatureTypeList.Gazebo</lx:path>
            <lx:label>Gazebo</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="In Ground Pool">
        <xs:annotation>
          <xs:documentation>In Ground Pool.</xs:documentation>
          <xs:appinfo>
            <lx:path>ancillaryFeatureTypeList.In Ground Pool</lx:path>
            <lx:label>In Ground Pool</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Landscaping">
        <xs:annotation>
          <xs:documentation>Landscaping.</xs:documentation>
          <xs:appinfo>
            <lx:path>ancillaryFeatureTypeList.Landscaping</lx:path>
            <lx:label>Landscaping</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other.</xs:documentation>
          <xs:appinfo>
            <lx:path>ancillaryFeatureTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Paving">
        <xs:annotation>
          <xs:documentation>Paving.</xs:documentation>
          <xs:appinfo>
            <lx:path>ancillaryFeatureTypeList.Paving</lx:path>
            <lx:label>Paving</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Shed">
        <xs:annotation>
          <xs:documentation>Shed.</xs:documentation>
          <xs:appinfo>
            <lx:path>ancillaryFeatureTypeList.Shed</lx:path>
            <lx:label>Shed</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tennis Court">
        <xs:annotation>
          <xs:documentation>Tennis Court.</xs:documentation>
          <xs:appinfo>
            <lx:path>ancillaryFeatureTypeList.Tennis Court</lx:path>
            <lx:label>Tennis Court</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="annotationTypeList">
    <xs:annotation>
      <xs:documentation>annotationTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>annotationTypeList</lx:path>
        <lx:label>annotationTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Code">
        <xs:annotation>
          <xs:documentation>Code</xs:documentation>
          <xs:appinfo>
            <lx:path>annotationTypeList.Code</lx:path>
            <lx:label>Code</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="End User Message">
        <xs:annotation>
          <xs:documentation>End User Message</xs:documentation>
          <xs:appinfo>
            <lx:path>annotationTypeList.End User Message</lx:path>
            <lx:label>End User Message</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Message">
        <xs:annotation>
          <xs:documentation>Message</xs:documentation>
          <xs:appinfo>
            <lx:path>annotationTypeList.Message</lx:path>
            <lx:label>Message</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Technical">
        <xs:annotation>
          <xs:documentation>Technical</xs:documentation>
          <xs:appinfo>
            <lx:path>annotationTypeList.Technical</lx:path>
            <lx:label>Technical</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="anzscoOccupationCodeList">
    <xs:annotation>
      <xs:documentation>The list of Occupation Codes defined by the Australian and New Zealand Standard Classification of Occupations, 2013, Version 1.2, Released at 11:30 AM (CANBERRA TIME) 26/06/2013. This list of classifications (ANZSCO Version 1.2) builds on a review conducted in 2009 (ANZSCO First Edition, Revision 1) following the classification's introduction in 2006.</xs:documentation>
      <xs:appinfo>
        <lx:path>anzscoOccupationCodeList</lx:path>
        <lx:label>anzscoOccupationCodeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="100000">
        <xs:annotation>
          <xs:documentation>Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.100000</lx:path>
            <lx:label>Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="111000">
        <xs:annotation>
          <xs:documentation>Chief Executives, General Managers and Legislators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.111000</lx:path>
            <lx:label>Chief Executives, General Managers and Legislators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="111111">
        <xs:annotation>
          <xs:documentation>Chief Executive or Managing Director</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.111111</lx:path>
            <lx:label>Chief Executive or Managing Director</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="111200">
        <xs:annotation>
          <xs:documentation>General Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.111200</lx:path>
            <lx:label>General Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="111211">
        <xs:annotation>
          <xs:documentation>Corporate General Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.111211</lx:path>
            <lx:label>Corporate General Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="111212">
        <xs:annotation>
          <xs:documentation>Defence Force Senior Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.111212</lx:path>
            <lx:label>Defence Force Senior Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="111300">
        <xs:annotation>
          <xs:documentation>Legislators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.111300</lx:path>
            <lx:label>Legislators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="111311">
        <xs:annotation>
          <xs:documentation>Local Government Legislator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.111311</lx:path>
            <lx:label>Local Government Legislator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="111312">
        <xs:annotation>
          <xs:documentation>Member of Parliament</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.111312</lx:path>
            <lx:label>Member of Parliament</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="111399">
        <xs:annotation>
          <xs:documentation>Legislators nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.111399</lx:path>
            <lx:label>Legislators nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121000">
        <xs:annotation>
          <xs:documentation>Farmers and Farm Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121000</lx:path>
            <lx:label>Farmers and Farm Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121111">
        <xs:annotation>
          <xs:documentation>Aquaculture Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121111</lx:path>
            <lx:label>Aquaculture Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121200">
        <xs:annotation>
          <xs:documentation>Crop Farmers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121200</lx:path>
            <lx:label>Crop Farmers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121211">
        <xs:annotation>
          <xs:documentation>Cotton Grower</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121211</lx:path>
            <lx:label>Cotton Grower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121212">
        <xs:annotation>
          <xs:documentation>Flower Grower</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121212</lx:path>
            <lx:label>Flower Grower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121213">
        <xs:annotation>
          <xs:documentation>Fruit or Nut Grower</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121213</lx:path>
            <lx:label>Fruit or Nut Grower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121214">
        <xs:annotation>
          <xs:documentation>Grain, Oilseed or Pasture Grower</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121214</lx:path>
            <lx:label>Grain, Oilseed or Pasture Grower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121215">
        <xs:annotation>
          <xs:documentation>Grape Grower</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121215</lx:path>
            <lx:label>Grape Grower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121216">
        <xs:annotation>
          <xs:documentation>Mixed Crop Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121216</lx:path>
            <lx:label>Mixed Crop Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121217">
        <xs:annotation>
          <xs:documentation>Sugar Cane Grower</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121217</lx:path>
            <lx:label>Sugar Cane Grower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121218">
        <xs:annotation>
          <xs:documentation>Turf Grower</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121218</lx:path>
            <lx:label>Turf Grower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121221">
        <xs:annotation>
          <xs:documentation>Vegetable Grower</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121221</lx:path>
            <lx:label>Vegetable Grower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121299">
        <xs:annotation>
          <xs:documentation>Crop Farmers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121299</lx:path>
            <lx:label>Crop Farmers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121300">
        <xs:annotation>
          <xs:documentation>Livestock Farmers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121300</lx:path>
            <lx:label>Livestock Farmers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121311">
        <xs:annotation>
          <xs:documentation>Apiarist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121311</lx:path>
            <lx:label>Apiarist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121312">
        <xs:annotation>
          <xs:documentation>Beef Cattle Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121312</lx:path>
            <lx:label>Beef Cattle Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121313">
        <xs:annotation>
          <xs:documentation>Dairy Cattle Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121313</lx:path>
            <lx:label>Dairy Cattle Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121314">
        <xs:annotation>
          <xs:documentation>Deer Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121314</lx:path>
            <lx:label>Deer Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121315">
        <xs:annotation>
          <xs:documentation>Goat Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121315</lx:path>
            <lx:label>Goat Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121316">
        <xs:annotation>
          <xs:documentation>Horse Breeder</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121316</lx:path>
            <lx:label>Horse Breeder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121317">
        <xs:annotation>
          <xs:documentation>Mixed Livestock Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121317</lx:path>
            <lx:label>Mixed Livestock Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121318">
        <xs:annotation>
          <xs:documentation>Pig Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121318</lx:path>
            <lx:label>Pig Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121321">
        <xs:annotation>
          <xs:documentation>Poultry Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121321</lx:path>
            <lx:label>Poultry Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121322">
        <xs:annotation>
          <xs:documentation>Sheep Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121322</lx:path>
            <lx:label>Sheep Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121399">
        <xs:annotation>
          <xs:documentation>Livestock Farmers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121399</lx:path>
            <lx:label>Livestock Farmers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="121411">
        <xs:annotation>
          <xs:documentation>Mixed Crop and Livestock Farmer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.121411</lx:path>
            <lx:label>Mixed Crop and Livestock Farmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="130000">
        <xs:annotation>
          <xs:documentation>Specialist Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.130000</lx:path>
            <lx:label>Specialist Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="131100">
        <xs:annotation>
          <xs:documentation>Advertising, Public Relations and Sales Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.131100</lx:path>
            <lx:label>Advertising, Public Relations and Sales Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="131112">
        <xs:annotation>
          <xs:documentation>Sales and Marketing Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.131112</lx:path>
            <lx:label>Sales and Marketing Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="131113">
        <xs:annotation>
          <xs:documentation>Advertising Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.131113</lx:path>
            <lx:label>Advertising Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="131114">
        <xs:annotation>
          <xs:documentation>Public Relations Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.131114</lx:path>
            <lx:label>Public Relations Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="132000">
        <xs:annotation>
          <xs:documentation>Business Administration Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.132000</lx:path>
            <lx:label>Business Administration Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="132111">
        <xs:annotation>
          <xs:documentation>Corporate Services Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.132111</lx:path>
            <lx:label>Corporate Services Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="132211">
        <xs:annotation>
          <xs:documentation>Finance Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.132211</lx:path>
            <lx:label>Finance Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="132311">
        <xs:annotation>
          <xs:documentation>Human Resource Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.132311</lx:path>
            <lx:label>Human Resource Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="132411">
        <xs:annotation>
          <xs:documentation>Policy and Planning Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.132411</lx:path>
            <lx:label>Policy and Planning Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="132511">
        <xs:annotation>
          <xs:documentation>Research and Development Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.132511</lx:path>
            <lx:label>Research and Development Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133000">
        <xs:annotation>
          <xs:documentation>Construction, Distribution and Production Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133000</lx:path>
            <lx:label>Construction, Distribution and Production Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133100">
        <xs:annotation>
          <xs:documentation>Construction Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133100</lx:path>
            <lx:label>Construction Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133111">
        <xs:annotation>
          <xs:documentation>Construction Project Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133111</lx:path>
            <lx:label>Construction Project Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133112">
        <xs:annotation>
          <xs:documentation>Project Builder</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133112</lx:path>
            <lx:label>Project Builder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133211">
        <xs:annotation>
          <xs:documentation>Engineering Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133211</lx:path>
            <lx:label>Engineering Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133300">
        <xs:annotation>
          <xs:documentation>Importers, Exporters and Wholesalers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133300</lx:path>
            <lx:label>Importers, Exporters and Wholesalers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133311">
        <xs:annotation>
          <xs:documentation>Importer or Exporter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133311</lx:path>
            <lx:label>Importer or Exporter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133312">
        <xs:annotation>
          <xs:documentation>Wholesaler</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133312</lx:path>
            <lx:label>Wholesaler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133411">
        <xs:annotation>
          <xs:documentation>Manufacturer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133411</lx:path>
            <lx:label>Manufacturer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133500">
        <xs:annotation>
          <xs:documentation>Production Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133500</lx:path>
            <lx:label>Production Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133511">
        <xs:annotation>
          <xs:documentation>Production Manager (Forestry)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133511</lx:path>
            <lx:label>Production Manager (Forestry)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133512">
        <xs:annotation>
          <xs:documentation>Production Manager (Manufacturing)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133512</lx:path>
            <lx:label>Production Manager (Manufacturing)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133513">
        <xs:annotation>
          <xs:documentation>Production Manager (Mining)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133513</lx:path>
            <lx:label>Production Manager (Mining)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133611">
        <xs:annotation>
          <xs:documentation>Supply and Distribution Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133611</lx:path>
            <lx:label>Supply and Distribution Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="133612">
        <xs:annotation>
          <xs:documentation>Procurement Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.133612</lx:path>
            <lx:label>Procurement Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="134000">
        <xs:annotation>
          <xs:documentation>Education, Health and Welfare Services Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.134000</lx:path>
            <lx:label>Education, Health and Welfare Services Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="134111">
        <xs:annotation>
          <xs:documentation>Child Care Centre Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.134111</lx:path>
            <lx:label>Child Care Centre Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="134200">
        <xs:annotation>
          <xs:documentation>Health and Welfare Services Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.134200</lx:path>
            <lx:label>Health and Welfare Services Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="134211">
        <xs:annotation>
          <xs:documentation>Medical Administrator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.134211</lx:path>
            <lx:label>Medical Administrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="134212">
        <xs:annotation>
          <xs:documentation>Nursing Clinical Director</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.134212</lx:path>
            <lx:label>Nursing Clinical Director</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="134213">
        <xs:annotation>
          <xs:documentation>Primary Health Organisation Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.134213</lx:path>
            <lx:label>Primary Health Organisation Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="134214">
        <xs:annotation>
          <xs:documentation>Welfare Centre Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.134214</lx:path>
            <lx:label>Welfare Centre Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="134299">
        <xs:annotation>
          <xs:documentation>Health and Welfare Services Managers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.134299</lx:path>
            <lx:label>Health and Welfare Services Managers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="134311">
        <xs:annotation>
          <xs:documentation>School Principal</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.134311</lx:path>
            <lx:label>School Principal</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="134400">
        <xs:annotation>
          <xs:documentation>Other Education Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.134400</lx:path>
            <lx:label>Other Education Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="134411">
        <xs:annotation>
          <xs:documentation>Faculty Head</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.134411</lx:path>
            <lx:label>Faculty Head</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="134412">
        <xs:annotation>
          <xs:documentation>Regional Education Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.134412</lx:path>
            <lx:label>Regional Education Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="134499">
        <xs:annotation>
          <xs:documentation>Education Managers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.134499</lx:path>
            <lx:label>Education Managers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="135100">
        <xs:annotation>
          <xs:documentation>ICT Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.135100</lx:path>
            <lx:label>ICT Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="135111">
        <xs:annotation>
          <xs:documentation>Chief Information Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.135111</lx:path>
            <lx:label>Chief Information Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="135112">
        <xs:annotation>
          <xs:documentation>ICT Project Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.135112</lx:path>
            <lx:label>ICT Project Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="135199">
        <xs:annotation>
          <xs:documentation>ICT Managers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.135199</lx:path>
            <lx:label>ICT Managers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="139000">
        <xs:annotation>
          <xs:documentation>Miscellaneous Specialist Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.139000</lx:path>
            <lx:label>Miscellaneous Specialist Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="139100">
        <xs:annotation>
          <xs:documentation>Commissioned Officers (Management) nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.139100</lx:path>
            <lx:label>Commissioned Officers (Management) nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="139111">
        <xs:annotation>
          <xs:documentation>Commissioned Defence Force Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.139111</lx:path>
            <lx:label>Commissioned Defence Force Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="139112">
        <xs:annotation>
          <xs:documentation>Commissioned Fire Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.139112</lx:path>
            <lx:label>Commissioned Fire Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="139113">
        <xs:annotation>
          <xs:documentation>Commissioned Police Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.139113</lx:path>
            <lx:label>Commissioned Police Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="139211">
        <xs:annotation>
          <xs:documentation>Senior Non-commissioned Defence Force Member</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.139211</lx:path>
            <lx:label>Senior Non-commissioned Defence Force Member</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="139900">
        <xs:annotation>
          <xs:documentation>Other Specialist Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.139900</lx:path>
            <lx:label>Other Specialist Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="139911">
        <xs:annotation>
          <xs:documentation>Arts Administrator or Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.139911</lx:path>
            <lx:label>Arts Administrator or Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="139912">
        <xs:annotation>
          <xs:documentation>Environmental Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.139912</lx:path>
            <lx:label>Environmental Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="139913">
        <xs:annotation>
          <xs:documentation>Laboratory Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.139913</lx:path>
            <lx:label>Laboratory Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="139914">
        <xs:annotation>
          <xs:documentation>Quality Assurance Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.139914</lx:path>
            <lx:label>Quality Assurance Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="139915">
        <xs:annotation>
          <xs:documentation>Sports Administrator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.139915</lx:path>
            <lx:label>Sports Administrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="139999">
        <xs:annotation>
          <xs:documentation>Specialist Managers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.139999</lx:path>
            <lx:label>Specialist Managers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="140000">
        <xs:annotation>
          <xs:documentation>Hospitality, Retail and Service Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.140000</lx:path>
            <lx:label>Hospitality, Retail and Service Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="141000">
        <xs:annotation>
          <xs:documentation>Accommodation and Hospitality Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.141000</lx:path>
            <lx:label>Accommodation and Hospitality Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="141111">
        <xs:annotation>
          <xs:documentation>Cafe or Restaurant Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.141111</lx:path>
            <lx:label>Cafe or Restaurant Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="141211">
        <xs:annotation>
          <xs:documentation>Caravan Park and Camping Ground Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.141211</lx:path>
            <lx:label>Caravan Park and Camping Ground Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="141311">
        <xs:annotation>
          <xs:documentation>Hotel or Motel Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.141311</lx:path>
            <lx:label>Hotel or Motel Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="141411">
        <xs:annotation>
          <xs:documentation>Licensed Club Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.141411</lx:path>
            <lx:label>Licensed Club Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="141900">
        <xs:annotation>
          <xs:documentation>Other Accommodation and Hospitality Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.141900</lx:path>
            <lx:label>Other Accommodation and Hospitality Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="141911">
        <xs:annotation>
          <xs:documentation>Bed and Breakfast Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.141911</lx:path>
            <lx:label>Bed and Breakfast Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="141912">
        <xs:annotation>
          <xs:documentation>Retirement Village Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.141912</lx:path>
            <lx:label>Retirement Village Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="141999">
        <xs:annotation>
          <xs:documentation>Accommodation and Hospitality Managers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.141999</lx:path>
            <lx:label>Accommodation and Hospitality Managers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="142100">
        <xs:annotation>
          <xs:documentation>Retail Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.142100</lx:path>
            <lx:label>Retail Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="142111">
        <xs:annotation>
          <xs:documentation>Retail Manager (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.142111</lx:path>
            <lx:label>Retail Manager (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="142112">
        <xs:annotation>
          <xs:documentation>Antique Dealer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.142112</lx:path>
            <lx:label>Antique Dealer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="142113">
        <xs:annotation>
          <xs:documentation>Betting Agency Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.142113</lx:path>
            <lx:label>Betting Agency Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="142114">
        <xs:annotation>
          <xs:documentation>Hair or Beauty Salon Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.142114</lx:path>
            <lx:label>Hair or Beauty Salon Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="142115">
        <xs:annotation>
          <xs:documentation>Post Office Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.142115</lx:path>
            <lx:label>Post Office Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="142116">
        <xs:annotation>
          <xs:documentation>Travel Agency Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.142116</lx:path>
            <lx:label>Travel Agency Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149000">
        <xs:annotation>
          <xs:documentation>Miscellaneous Hospitality, Retail and Service Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149000</lx:path>
            <lx:label>Miscellaneous Hospitality, Retail and Service Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149100">
        <xs:annotation>
          <xs:documentation>Amusement, Fitness and Sports Centre Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149100</lx:path>
            <lx:label>Amusement, Fitness and Sports Centre Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149111">
        <xs:annotation>
          <xs:documentation>Amusement Centre Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149111</lx:path>
            <lx:label>Amusement Centre Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149112">
        <xs:annotation>
          <xs:documentation>Fitness Centre Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149112</lx:path>
            <lx:label>Fitness Centre Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149113">
        <xs:annotation>
          <xs:documentation>Sports Centre Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149113</lx:path>
            <lx:label>Sports Centre Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149200">
        <xs:annotation>
          <xs:documentation>Call or Contact Centre and Customer Service Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149200</lx:path>
            <lx:label>Call or Contact Centre and Customer Service Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149211">
        <xs:annotation>
          <xs:documentation>Call or Contact Centre Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149211</lx:path>
            <lx:label>Call or Contact Centre Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149212">
        <xs:annotation>
          <xs:documentation>Customer Service Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149212</lx:path>
            <lx:label>Customer Service Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149311">
        <xs:annotation>
          <xs:documentation>Conference and Event Organiser</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149311</lx:path>
            <lx:label>Conference and Event Organiser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149400">
        <xs:annotation>
          <xs:documentation>Transport Services Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149400</lx:path>
            <lx:label>Transport Services Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149411">
        <xs:annotation>
          <xs:documentation>Fleet Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149411</lx:path>
            <lx:label>Fleet Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149412">
        <xs:annotation>
          <xs:documentation>Railway Station Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149412</lx:path>
            <lx:label>Railway Station Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149413">
        <xs:annotation>
          <xs:documentation>Transport Company Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149413</lx:path>
            <lx:label>Transport Company Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149900">
        <xs:annotation>
          <xs:documentation>Other Hospitality, Retail and Service Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149900</lx:path>
            <lx:label>Other Hospitality, Retail and Service Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149911">
        <xs:annotation>
          <xs:documentation>Boarding Kennel or Cattery Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149911</lx:path>
            <lx:label>Boarding Kennel or Cattery Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149912">
        <xs:annotation>
          <xs:documentation>Cinema or Theatre Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149912</lx:path>
            <lx:label>Cinema or Theatre Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149913">
        <xs:annotation>
          <xs:documentation>Facilities Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149913</lx:path>
            <lx:label>Facilities Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149914">
        <xs:annotation>
          <xs:documentation>Financial Institution Branch Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149914</lx:path>
            <lx:label>Financial Institution Branch Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149915">
        <xs:annotation>
          <xs:documentation>Equipment Hire Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149915</lx:path>
            <lx:label>Equipment Hire Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="149999">
        <xs:annotation>
          <xs:documentation>Hospitality, Retail and Service Managers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.149999</lx:path>
            <lx:label>Hospitality, Retail and Service Managers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="200000">
        <xs:annotation>
          <xs:documentation>Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.200000</lx:path>
            <lx:label>Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="210000">
        <xs:annotation>
          <xs:documentation>Arts and Media Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.210000</lx:path>
            <lx:label>Arts and Media Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211000">
        <xs:annotation>
          <xs:documentation>Arts Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211000</lx:path>
            <lx:label>Arts Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211100">
        <xs:annotation>
          <xs:documentation>Actors, Dancers and Other Entertainers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211100</lx:path>
            <lx:label>Actors, Dancers and Other Entertainers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211111">
        <xs:annotation>
          <xs:documentation>Actor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211111</lx:path>
            <lx:label>Actor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211112">
        <xs:annotation>
          <xs:documentation>Dancer or Choreographer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211112</lx:path>
            <lx:label>Dancer or Choreographer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211113">
        <xs:annotation>
          <xs:documentation>Entertainer or Variety Artist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211113</lx:path>
            <lx:label>Entertainer or Variety Artist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211199">
        <xs:annotation>
          <xs:documentation>Actors, Dancers and Other Entertainers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211199</lx:path>
            <lx:label>Actors, Dancers and Other Entertainers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211200">
        <xs:annotation>
          <xs:documentation>Music Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211200</lx:path>
            <lx:label>Music Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211211">
        <xs:annotation>
          <xs:documentation>Composer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211211</lx:path>
            <lx:label>Composer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211212">
        <xs:annotation>
          <xs:documentation>Music Director</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211212</lx:path>
            <lx:label>Music Director</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211213">
        <xs:annotation>
          <xs:documentation>Musician (Instrumental)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211213</lx:path>
            <lx:label>Musician (Instrumental)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211214">
        <xs:annotation>
          <xs:documentation>Singer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211214</lx:path>
            <lx:label>Singer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211299">
        <xs:annotation>
          <xs:documentation>Music Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211299</lx:path>
            <lx:label>Music Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211311">
        <xs:annotation>
          <xs:documentation>Photographer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211311</lx:path>
            <lx:label>Photographer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211400">
        <xs:annotation>
          <xs:documentation>Visual Arts and Crafts Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211400</lx:path>
            <lx:label>Visual Arts and Crafts Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211411">
        <xs:annotation>
          <xs:documentation>Painter (Visual Arts)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211411</lx:path>
            <lx:label>Painter (Visual Arts)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211412">
        <xs:annotation>
          <xs:documentation>Potter or Ceramic Artist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211412</lx:path>
            <lx:label>Potter or Ceramic Artist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211413">
        <xs:annotation>
          <xs:documentation>Sculptor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211413</lx:path>
            <lx:label>Sculptor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="211499">
        <xs:annotation>
          <xs:documentation>Visual Arts and Crafts Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.211499</lx:path>
            <lx:label>Visual Arts and Crafts Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212000">
        <xs:annotation>
          <xs:documentation>Media Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212000</lx:path>
            <lx:label>Media Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212100">
        <xs:annotation>
          <xs:documentation>Artistic Directors, and Media Producers and Presenters nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212100</lx:path>
            <lx:label>Artistic Directors, and Media Producers and Presenters nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212111">
        <xs:annotation>
          <xs:documentation>Artistic Director</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212111</lx:path>
            <lx:label>Artistic Director</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212112">
        <xs:annotation>
          <xs:documentation>Media Producer (excluding Video)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212112</lx:path>
            <lx:label>Media Producer (excluding Video)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212113">
        <xs:annotation>
          <xs:documentation>Radio Presenter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212113</lx:path>
            <lx:label>Radio Presenter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212114">
        <xs:annotation>
          <xs:documentation>Television Presenter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212114</lx:path>
            <lx:label>Television Presenter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212200">
        <xs:annotation>
          <xs:documentation>Authors, and Book and Script Editors nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212200</lx:path>
            <lx:label>Authors, and Book and Script Editors nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212211">
        <xs:annotation>
          <xs:documentation>Author</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212211</lx:path>
            <lx:label>Author</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212212">
        <xs:annotation>
          <xs:documentation>Book or Script Editor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212212</lx:path>
            <lx:label>Book or Script Editor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212300">
        <xs:annotation>
          <xs:documentation>Film, Television, Radio and Stage Directors nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212300</lx:path>
            <lx:label>Film, Television, Radio and Stage Directors nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212311">
        <xs:annotation>
          <xs:documentation>Art Director (Film, Television or Stage)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212311</lx:path>
            <lx:label>Art Director (Film, Television or Stage)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212312">
        <xs:annotation>
          <xs:documentation>Director (Film, Television, Radio or Stage)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212312</lx:path>
            <lx:label>Director (Film, Television, Radio or Stage)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212313">
        <xs:annotation>
          <xs:documentation>Director of Photography</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212313</lx:path>
            <lx:label>Director of Photography</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212314">
        <xs:annotation>
          <xs:documentation>Film and Video Editor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212314</lx:path>
            <lx:label>Film and Video Editor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212315">
        <xs:annotation>
          <xs:documentation>Program Director (Television or Radio)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212315</lx:path>
            <lx:label>Program Director (Television or Radio)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212316">
        <xs:annotation>
          <xs:documentation>Stage Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212316</lx:path>
            <lx:label>Stage Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212317">
        <xs:annotation>
          <xs:documentation>Technical Director</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212317</lx:path>
            <lx:label>Technical Director</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212318">
        <xs:annotation>
          <xs:documentation>Video Producer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212318</lx:path>
            <lx:label>Video Producer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212399">
        <xs:annotation>
          <xs:documentation>Film, Television, Radio and Stage Directors nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212399</lx:path>
            <lx:label>Film, Television, Radio and Stage Directors nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212400">
        <xs:annotation>
          <xs:documentation>Journalists and Other Writers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212400</lx:path>
            <lx:label>Journalists and Other Writers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212411">
        <xs:annotation>
          <xs:documentation>Copywriter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212411</lx:path>
            <lx:label>Copywriter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212412">
        <xs:annotation>
          <xs:documentation>Newspaper or Periodical Editor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212412</lx:path>
            <lx:label>Newspaper or Periodical Editor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212413">
        <xs:annotation>
          <xs:documentation>Print Journalist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212413</lx:path>
            <lx:label>Print Journalist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212414">
        <xs:annotation>
          <xs:documentation>Radio Journalist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212414</lx:path>
            <lx:label>Radio Journalist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212415">
        <xs:annotation>
          <xs:documentation>Technical Writer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212415</lx:path>
            <lx:label>Technical Writer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212416">
        <xs:annotation>
          <xs:documentation>Television Journalist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212416</lx:path>
            <lx:label>Television Journalist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="212499">
        <xs:annotation>
          <xs:documentation>Journalists and Other Writers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.212499</lx:path>
            <lx:label>Journalists and Other Writers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="220000">
        <xs:annotation>
          <xs:documentation>Business, Human Resource and Marketing Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.220000</lx:path>
            <lx:label>Business, Human Resource and Marketing Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="221000">
        <xs:annotation>
          <xs:documentation>Accountants, Auditors and Company Secretaries nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.221000</lx:path>
            <lx:label>Accountants, Auditors and Company Secretaries nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="221100">
        <xs:annotation>
          <xs:documentation>Accountants nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.221100</lx:path>
            <lx:label>Accountants nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="221111">
        <xs:annotation>
          <xs:documentation>Accountant (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.221111</lx:path>
            <lx:label>Accountant (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="221112">
        <xs:annotation>
          <xs:documentation>Management Accountant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.221112</lx:path>
            <lx:label>Management Accountant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="221113">
        <xs:annotation>
          <xs:documentation>Taxation Accountant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.221113</lx:path>
            <lx:label>Taxation Accountant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="221200">
        <xs:annotation>
          <xs:documentation>Auditors, Company Secretaries and Corporate Treasurers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.221200</lx:path>
            <lx:label>Auditors, Company Secretaries and Corporate Treasurers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="221211">
        <xs:annotation>
          <xs:documentation>Company Secretary</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.221211</lx:path>
            <lx:label>Company Secretary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="221212">
        <xs:annotation>
          <xs:documentation>Corporate Treasurer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.221212</lx:path>
            <lx:label>Corporate Treasurer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="221213">
        <xs:annotation>
          <xs:documentation>External Auditor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.221213</lx:path>
            <lx:label>External Auditor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="221214">
        <xs:annotation>
          <xs:documentation>Internal Auditor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.221214</lx:path>
            <lx:label>Internal Auditor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222000">
        <xs:annotation>
          <xs:documentation>Financial Brokers and Dealers, and Investment Advisers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.222000</lx:path>
            <lx:label>Financial Brokers and Dealers, and Investment Advisers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222100">
        <xs:annotation>
          <xs:documentation>Financial Brokers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.222100</lx:path>
            <lx:label>Financial Brokers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222111">
        <xs:annotation>
          <xs:documentation>Commodities Trader</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.222111</lx:path>
            <lx:label>Commodities Trader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222112">
        <xs:annotation>
          <xs:documentation>Finance Broker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.222112</lx:path>
            <lx:label>Finance Broker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222113">
        <xs:annotation>
          <xs:documentation>Insurance Broker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.222113</lx:path>
            <lx:label>Insurance Broker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222199">
        <xs:annotation>
          <xs:documentation>Financial Brokers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.222199</lx:path>
            <lx:label>Financial Brokers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222200">
        <xs:annotation>
          <xs:documentation>Financial Dealers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.222200</lx:path>
            <lx:label>Financial Dealers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222211">
        <xs:annotation>
          <xs:documentation>Financial Market Dealer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.222211</lx:path>
            <lx:label>Financial Market Dealer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222212">
        <xs:annotation>
          <xs:documentation>Futures Trader</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.222212</lx:path>
            <lx:label>Futures Trader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222213">
        <xs:annotation>
          <xs:documentation>Stockbroking Dealer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.222213</lx:path>
            <lx:label>Stockbroking Dealer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222299">
        <xs:annotation>
          <xs:documentation>Financial Dealers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.222299</lx:path>
            <lx:label>Financial Dealers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222300">
        <xs:annotation>
          <xs:documentation>Financial Investment Advisers and Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.222300</lx:path>
            <lx:label>Financial Investment Advisers and Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222311">
        <xs:annotation>
          <xs:documentation>Financial Investment Adviser</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.222311</lx:path>
            <lx:label>Financial Investment Adviser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="222312">
        <xs:annotation>
          <xs:documentation>Financial Investment Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.222312</lx:path>
            <lx:label>Financial Investment Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="223000">
        <xs:annotation>
          <xs:documentation>Human Resource and Training Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.223000</lx:path>
            <lx:label>Human Resource and Training Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="223100">
        <xs:annotation>
          <xs:documentation>Human Resource Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.223100</lx:path>
            <lx:label>Human Resource Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="223111">
        <xs:annotation>
          <xs:documentation>Human Resource Adviser</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.223111</lx:path>
            <lx:label>Human Resource Adviser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="223112">
        <xs:annotation>
          <xs:documentation>Recruitment Consultant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.223112</lx:path>
            <lx:label>Recruitment Consultant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="223113">
        <xs:annotation>
          <xs:documentation>Workplace Relations Adviser</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.223113</lx:path>
            <lx:label>Workplace Relations Adviser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="223211">
        <xs:annotation>
          <xs:documentation>ICT Trainer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.223211</lx:path>
            <lx:label>ICT Trainer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="223311">
        <xs:annotation>
          <xs:documentation>Training and Development Professional</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.223311</lx:path>
            <lx:label>Training and Development Professional</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224000">
        <xs:annotation>
          <xs:documentation>Information and Organisation Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224000</lx:path>
            <lx:label>Information and Organisation Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224100">
        <xs:annotation>
          <xs:documentation>Actuaries, Mathematicians and Statisticians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224100</lx:path>
            <lx:label>Actuaries, Mathematicians and Statisticians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224111">
        <xs:annotation>
          <xs:documentation>Actuary</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224111</lx:path>
            <lx:label>Actuary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224112">
        <xs:annotation>
          <xs:documentation>Mathematician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224112</lx:path>
            <lx:label>Mathematician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224113">
        <xs:annotation>
          <xs:documentation>Statistician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224113</lx:path>
            <lx:label>Statistician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224200">
        <xs:annotation>
          <xs:documentation>Archivists, Curators and Records Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224200</lx:path>
            <lx:label>Archivists, Curators and Records Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224211">
        <xs:annotation>
          <xs:documentation>Archivist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224211</lx:path>
            <lx:label>Archivist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224212">
        <xs:annotation>
          <xs:documentation>Gallery or Museum Curator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224212</lx:path>
            <lx:label>Gallery or Museum Curator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224213">
        <xs:annotation>
          <xs:documentation>Health Information Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224213</lx:path>
            <lx:label>Health Information Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224214">
        <xs:annotation>
          <xs:documentation>Records Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224214</lx:path>
            <lx:label>Records Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224311">
        <xs:annotation>
          <xs:documentation>Economist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224311</lx:path>
            <lx:label>Economist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224400">
        <xs:annotation>
          <xs:documentation>Intelligence and Policy Analysts nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224400</lx:path>
            <lx:label>Intelligence and Policy Analysts nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224411">
        <xs:annotation>
          <xs:documentation>Intelligence Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224411</lx:path>
            <lx:label>Intelligence Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224412">
        <xs:annotation>
          <xs:documentation>Policy Analyst</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224412</lx:path>
            <lx:label>Policy Analyst</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224500">
        <xs:annotation>
          <xs:documentation>Land Economists and Valuers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224500</lx:path>
            <lx:label>Land Economists and Valuers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224511">
        <xs:annotation>
          <xs:documentation>Land Economist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224511</lx:path>
            <lx:label>Land Economist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224512">
        <xs:annotation>
          <xs:documentation>Valuer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224512</lx:path>
            <lx:label>Valuer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224611">
        <xs:annotation>
          <xs:documentation>Librarian</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224611</lx:path>
            <lx:label>Librarian</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224700">
        <xs:annotation>
          <xs:documentation>Management and Organisation Analysts nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224700</lx:path>
            <lx:label>Management and Organisation Analysts nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224711">
        <xs:annotation>
          <xs:documentation>Management Consultant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224711</lx:path>
            <lx:label>Management Consultant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224712">
        <xs:annotation>
          <xs:documentation>Organisation and Methods Analyst</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224712</lx:path>
            <lx:label>Organisation and Methods Analyst</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224900">
        <xs:annotation>
          <xs:documentation>Other Information and Organisation Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224900</lx:path>
            <lx:label>Other Information and Organisation Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224911">
        <xs:annotation>
          <xs:documentation>Electorate Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224911</lx:path>
            <lx:label>Electorate Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224912">
        <xs:annotation>
          <xs:documentation>Liaison Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224912</lx:path>
            <lx:label>Liaison Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224913">
        <xs:annotation>
          <xs:documentation>Migration Agent</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224913</lx:path>
            <lx:label>Migration Agent</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224914">
        <xs:annotation>
          <xs:documentation>Patents Examiner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224914</lx:path>
            <lx:label>Patents Examiner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="224999">
        <xs:annotation>
          <xs:documentation>Information and Organisation Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.224999</lx:path>
            <lx:label>Information and Organisation Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="225000">
        <xs:annotation>
          <xs:documentation>Sales, Marketing and Public Relations Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.225000</lx:path>
            <lx:label>Sales, Marketing and Public Relations Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="225100">
        <xs:annotation>
          <xs:documentation>Advertising and Marketing Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.225100</lx:path>
            <lx:label>Advertising and Marketing Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="225111">
        <xs:annotation>
          <xs:documentation>Advertising Specialist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.225111</lx:path>
            <lx:label>Advertising Specialist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="225112">
        <xs:annotation>
          <xs:documentation>Market Research Analyst</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.225112</lx:path>
            <lx:label>Market Research Analyst</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="225113">
        <xs:annotation>
          <xs:documentation>Marketing Specialist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.225113</lx:path>
            <lx:label>Marketing Specialist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="225200">
        <xs:annotation>
          <xs:documentation>ICT Sales Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.225200</lx:path>
            <lx:label>ICT Sales Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="225211">
        <xs:annotation>
          <xs:documentation>ICT Account Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.225211</lx:path>
            <lx:label>ICT Account Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="225212">
        <xs:annotation>
          <xs:documentation>ICT Business Development Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.225212</lx:path>
            <lx:label>ICT Business Development Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="225213">
        <xs:annotation>
          <xs:documentation>ICT Sales Representative</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.225213</lx:path>
            <lx:label>ICT Sales Representative</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="225311">
        <xs:annotation>
          <xs:documentation>Public Relations Professional</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.225311</lx:path>
            <lx:label>Public Relations Professional</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="225400">
        <xs:annotation>
          <xs:documentation>Technical Sales Representatives nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.225400</lx:path>
            <lx:label>Technical Sales Representatives nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="225411">
        <xs:annotation>
          <xs:documentation>Sales Representative (Industrial Products)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.225411</lx:path>
            <lx:label>Sales Representative (Industrial Products)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="225412">
        <xs:annotation>
          <xs:documentation>Sales Representative (Medical and Pharmaceutical Products)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.225412</lx:path>
            <lx:label>Sales Representative (Medical and Pharmaceutical Products)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="225499">
        <xs:annotation>
          <xs:documentation>Technical Sales Representatives nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.225499</lx:path>
            <lx:label>Technical Sales Representatives nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="230000">
        <xs:annotation>
          <xs:documentation>Design, Engineering, Science and Transport Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.230000</lx:path>
            <lx:label>Design, Engineering, Science and Transport Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231000">
        <xs:annotation>
          <xs:documentation>Air and Marine Transport Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.231000</lx:path>
            <lx:label>Air and Marine Transport Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231100">
        <xs:annotation>
          <xs:documentation>Air Transport Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.231100</lx:path>
            <lx:label>Air Transport Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231111">
        <xs:annotation>
          <xs:documentation>Aeroplane Pilot</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.231111</lx:path>
            <lx:label>Aeroplane Pilot</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231112">
        <xs:annotation>
          <xs:documentation>Air Traffic Controller</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.231112</lx:path>
            <lx:label>Air Traffic Controller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231113">
        <xs:annotation>
          <xs:documentation>Flying Instructor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.231113</lx:path>
            <lx:label>Flying Instructor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231114">
        <xs:annotation>
          <xs:documentation>Helicopter Pilot</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.231114</lx:path>
            <lx:label>Helicopter Pilot</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231199">
        <xs:annotation>
          <xs:documentation>Air Transport Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.231199</lx:path>
            <lx:label>Air Transport Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231200">
        <xs:annotation>
          <xs:documentation>Marine Transport Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.231200</lx:path>
            <lx:label>Marine Transport Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231211">
        <xs:annotation>
          <xs:documentation>Master Fisher</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.231211</lx:path>
            <lx:label>Master Fisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231212">
        <xs:annotation>
          <xs:documentation>Ship's Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.231212</lx:path>
            <lx:label>Ship's Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231213">
        <xs:annotation>
          <xs:documentation>Ship's Master</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.231213</lx:path>
            <lx:label>Ship's Master</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231214">
        <xs:annotation>
          <xs:documentation>Ship's Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.231214</lx:path>
            <lx:label>Ship's Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231215">
        <xs:annotation>
          <xs:documentation>Marine Surveyor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.231215</lx:path>
            <lx:label>Marine Surveyor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="231299">
        <xs:annotation>
          <xs:documentation>Marine Transport Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.231299</lx:path>
            <lx:label>Marine Transport Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232000">
        <xs:annotation>
          <xs:documentation>Architects, Designers, Planners and Surveyors nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232000</lx:path>
            <lx:label>Architects, Designers, Planners and Surveyors nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232100">
        <xs:annotation>
          <xs:documentation>Architects and Landscape Architects nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232100</lx:path>
            <lx:label>Architects and Landscape Architects nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232111">
        <xs:annotation>
          <xs:documentation>Architect</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232111</lx:path>
            <lx:label>Architect</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232112">
        <xs:annotation>
          <xs:documentation>Landscape Architect</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232112</lx:path>
            <lx:label>Landscape Architect</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232200">
        <xs:annotation>
          <xs:documentation>Surveyors and Spatial Scientists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232200</lx:path>
            <lx:label>Surveyors and Spatial Scientists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232212">
        <xs:annotation>
          <xs:documentation>Surveyor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232212</lx:path>
            <lx:label>Surveyor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232213">
        <xs:annotation>
          <xs:documentation>Cartographer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232213</lx:path>
            <lx:label>Cartographer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232214">
        <xs:annotation>
          <xs:documentation>Other Spatial Scientist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232214</lx:path>
            <lx:label>Other Spatial Scientist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232300">
        <xs:annotation>
          <xs:documentation>Fashion, Industrial and Jewellery Designers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232300</lx:path>
            <lx:label>Fashion, Industrial and Jewellery Designers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232311">
        <xs:annotation>
          <xs:documentation>Fashion Designer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232311</lx:path>
            <lx:label>Fashion Designer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232312">
        <xs:annotation>
          <xs:documentation>Industrial Designer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232312</lx:path>
            <lx:label>Industrial Designer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232313">
        <xs:annotation>
          <xs:documentation>Jewellery Designer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232313</lx:path>
            <lx:label>Jewellery Designer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232400">
        <xs:annotation>
          <xs:documentation>Graphic and Web Designers, and Illustrators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232400</lx:path>
            <lx:label>Graphic and Web Designers, and Illustrators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232411">
        <xs:annotation>
          <xs:documentation>Graphic Designer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232411</lx:path>
            <lx:label>Graphic Designer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232412">
        <xs:annotation>
          <xs:documentation>Illustrator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232412</lx:path>
            <lx:label>Illustrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232413">
        <xs:annotation>
          <xs:documentation>Multimedia Designer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232413</lx:path>
            <lx:label>Multimedia Designer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232414">
        <xs:annotation>
          <xs:documentation>Web Designer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232414</lx:path>
            <lx:label>Web Designer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232511">
        <xs:annotation>
          <xs:documentation>Interior Designer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232511</lx:path>
            <lx:label>Interior Designer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="232611">
        <xs:annotation>
          <xs:documentation>Urban and Regional Planner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.232611</lx:path>
            <lx:label>Urban and Regional Planner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233000">
        <xs:annotation>
          <xs:documentation>Engineering Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233000</lx:path>
            <lx:label>Engineering Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233100">
        <xs:annotation>
          <xs:documentation>Chemical and Materials Engineers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233100</lx:path>
            <lx:label>Chemical and Materials Engineers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233111">
        <xs:annotation>
          <xs:documentation>Chemical Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233111</lx:path>
            <lx:label>Chemical Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233112">
        <xs:annotation>
          <xs:documentation>Materials Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233112</lx:path>
            <lx:label>Materials Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233200">
        <xs:annotation>
          <xs:documentation>Civil Engineering Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233200</lx:path>
            <lx:label>Civil Engineering Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233211">
        <xs:annotation>
          <xs:documentation>Civil Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233211</lx:path>
            <lx:label>Civil Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233212">
        <xs:annotation>
          <xs:documentation>Geotechnical Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233212</lx:path>
            <lx:label>Geotechnical Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233213">
        <xs:annotation>
          <xs:documentation>Quantity Surveyor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233213</lx:path>
            <lx:label>Quantity Surveyor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233214">
        <xs:annotation>
          <xs:documentation>Structural Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233214</lx:path>
            <lx:label>Structural Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233215">
        <xs:annotation>
          <xs:documentation>Transport Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233215</lx:path>
            <lx:label>Transport Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233311">
        <xs:annotation>
          <xs:documentation>Electrical Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233311</lx:path>
            <lx:label>Electrical Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233411">
        <xs:annotation>
          <xs:documentation>Electronics Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233411</lx:path>
            <lx:label>Electronics Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233500">
        <xs:annotation>
          <xs:documentation>Industrial, Mechanical and Production Engineers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233500</lx:path>
            <lx:label>Industrial, Mechanical and Production Engineers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233511">
        <xs:annotation>
          <xs:documentation>Industrial Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233511</lx:path>
            <lx:label>Industrial Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233512">
        <xs:annotation>
          <xs:documentation>Mechanical Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233512</lx:path>
            <lx:label>Mechanical Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233513">
        <xs:annotation>
          <xs:documentation>Production or Plant Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233513</lx:path>
            <lx:label>Production or Plant Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233600">
        <xs:annotation>
          <xs:documentation>Mining Engineers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233600</lx:path>
            <lx:label>Mining Engineers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233611">
        <xs:annotation>
          <xs:documentation>Mining Engineer (excluding Petroleum)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233611</lx:path>
            <lx:label>Mining Engineer (excluding Petroleum)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233612">
        <xs:annotation>
          <xs:documentation>Petroleum Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233612</lx:path>
            <lx:label>Petroleum Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233900">
        <xs:annotation>
          <xs:documentation>Other Engineering Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233900</lx:path>
            <lx:label>Other Engineering Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233911">
        <xs:annotation>
          <xs:documentation>Aeronautical Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233911</lx:path>
            <lx:label>Aeronautical Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233912">
        <xs:annotation>
          <xs:documentation>Agricultural Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233912</lx:path>
            <lx:label>Agricultural Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233913">
        <xs:annotation>
          <xs:documentation>Biomedical Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233913</lx:path>
            <lx:label>Biomedical Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233914">
        <xs:annotation>
          <xs:documentation>Engineering Technologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233914</lx:path>
            <lx:label>Engineering Technologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233915">
        <xs:annotation>
          <xs:documentation>Environmental Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233915</lx:path>
            <lx:label>Environmental Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233916">
        <xs:annotation>
          <xs:documentation>Naval Architect</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233916</lx:path>
            <lx:label>Naval Architect</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="233999">
        <xs:annotation>
          <xs:documentation>Engineering Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.233999</lx:path>
            <lx:label>Engineering Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234000">
        <xs:annotation>
          <xs:documentation>Natural and Physical Science Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234000</lx:path>
            <lx:label>Natural and Physical Science Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234100">
        <xs:annotation>
          <xs:documentation>Agricultural and Forestry Scientists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234100</lx:path>
            <lx:label>Agricultural and Forestry Scientists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234111">
        <xs:annotation>
          <xs:documentation>Agricultural Consultant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234111</lx:path>
            <lx:label>Agricultural Consultant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234112">
        <xs:annotation>
          <xs:documentation>Agricultural Scientist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234112</lx:path>
            <lx:label>Agricultural Scientist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234113">
        <xs:annotation>
          <xs:documentation>Forester</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234113</lx:path>
            <lx:label>Forester</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234200">
        <xs:annotation>
          <xs:documentation>Chemists, and Food and Wine Scientists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234200</lx:path>
            <lx:label>Chemists, and Food and Wine Scientists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234211">
        <xs:annotation>
          <xs:documentation>Chemist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234211</lx:path>
            <lx:label>Chemist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234212">
        <xs:annotation>
          <xs:documentation>Food Technologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234212</lx:path>
            <lx:label>Food Technologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234213">
        <xs:annotation>
          <xs:documentation>Wine Maker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234213</lx:path>
            <lx:label>Wine Maker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234300">
        <xs:annotation>
          <xs:documentation>Environmental Scientists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234300</lx:path>
            <lx:label>Environmental Scientists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234311">
        <xs:annotation>
          <xs:documentation>Conservation Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234311</lx:path>
            <lx:label>Conservation Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234312">
        <xs:annotation>
          <xs:documentation>Environmental Consultant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234312</lx:path>
            <lx:label>Environmental Consultant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234313">
        <xs:annotation>
          <xs:documentation>Environmental Research Scientist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234313</lx:path>
            <lx:label>Environmental Research Scientist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234314">
        <xs:annotation>
          <xs:documentation>Park Ranger</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234314</lx:path>
            <lx:label>Park Ranger</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234399">
        <xs:annotation>
          <xs:documentation>Environmental Scientists nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234399</lx:path>
            <lx:label>Environmental Scientists nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234400">
        <xs:annotation>
          <xs:documentation>Geologists, Geophysicists and Hydrogeologists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234400</lx:path>
            <lx:label>Geologists, Geophysicists and Hydrogeologists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234411">
        <xs:annotation>
          <xs:documentation>Geologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234411</lx:path>
            <lx:label>Geologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234412">
        <xs:annotation>
          <xs:documentation>Geophysicist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234412</lx:path>
            <lx:label>Geophysicist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234413">
        <xs:annotation>
          <xs:documentation>Hydrogeologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234413</lx:path>
            <lx:label>Hydrogeologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234500">
        <xs:annotation>
          <xs:documentation>Life Scientists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234500</lx:path>
            <lx:label>Life Scientists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234511">
        <xs:annotation>
          <xs:documentation>Life Scientist (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234511</lx:path>
            <lx:label>Life Scientist (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234513">
        <xs:annotation>
          <xs:documentation>Biochemist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234513</lx:path>
            <lx:label>Biochemist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234514">
        <xs:annotation>
          <xs:documentation>Biotechnologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234514</lx:path>
            <lx:label>Biotechnologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234515">
        <xs:annotation>
          <xs:documentation>Botanist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234515</lx:path>
            <lx:label>Botanist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234516">
        <xs:annotation>
          <xs:documentation>Marine Biologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234516</lx:path>
            <lx:label>Marine Biologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234517">
        <xs:annotation>
          <xs:documentation>Microbiologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234517</lx:path>
            <lx:label>Microbiologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234518">
        <xs:annotation>
          <xs:documentation>Zoologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234518</lx:path>
            <lx:label>Zoologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234599">
        <xs:annotation>
          <xs:documentation>Life Scientists nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234599</lx:path>
            <lx:label>Life Scientists nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234611">
        <xs:annotation>
          <xs:documentation>Medical Laboratory Scientist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234611</lx:path>
            <lx:label>Medical Laboratory Scientist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234711">
        <xs:annotation>
          <xs:documentation>Veterinarian</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234711</lx:path>
            <lx:label>Veterinarian</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234900">
        <xs:annotation>
          <xs:documentation>Other Natural and Physical Science Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234900</lx:path>
            <lx:label>Other Natural and Physical Science Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234911">
        <xs:annotation>
          <xs:documentation>Conservator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234911</lx:path>
            <lx:label>Conservator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234912">
        <xs:annotation>
          <xs:documentation>Metallurgist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234912</lx:path>
            <lx:label>Metallurgist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234913">
        <xs:annotation>
          <xs:documentation>Meteorologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234913</lx:path>
            <lx:label>Meteorologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234914">
        <xs:annotation>
          <xs:documentation>Physicist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234914</lx:path>
            <lx:label>Physicist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234915">
        <xs:annotation>
          <xs:documentation>Exercise Physiologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234915</lx:path>
            <lx:label>Exercise Physiologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="234999">
        <xs:annotation>
          <xs:documentation>Natural and Physical Science Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.234999</lx:path>
            <lx:label>Natural and Physical Science Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="240000">
        <xs:annotation>
          <xs:documentation>Education Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.240000</lx:path>
            <lx:label>Education Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="241000">
        <xs:annotation>
          <xs:documentation>School Teachers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.241000</lx:path>
            <lx:label>School Teachers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="241111">
        <xs:annotation>
          <xs:documentation>Early Childhood (Pre-primary School) Teacher</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.241111</lx:path>
            <lx:label>Early Childhood (Pre-primary School) Teacher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="241213">
        <xs:annotation>
          <xs:documentation>Primary School Teacher</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.241213</lx:path>
            <lx:label>Primary School Teacher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="241311">
        <xs:annotation>
          <xs:documentation>Middle School Teacher</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.241311</lx:path>
            <lx:label>Middle School Teacher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="241411">
        <xs:annotation>
          <xs:documentation>Secondary School Teacher</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.241411</lx:path>
            <lx:label>Secondary School Teacher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="241500">
        <xs:annotation>
          <xs:documentation>Special Education Teachers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.241500</lx:path>
            <lx:label>Special Education Teachers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="241511">
        <xs:annotation>
          <xs:documentation>Special Needs Teacher</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.241511</lx:path>
            <lx:label>Special Needs Teacher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="241512">
        <xs:annotation>
          <xs:documentation>Teacher of the Hearing Impaired</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.241512</lx:path>
            <lx:label>Teacher of the Hearing Impaired</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="241513">
        <xs:annotation>
          <xs:documentation>Teacher of the Sight Impaired</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.241513</lx:path>
            <lx:label>Teacher of the Sight Impaired</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="241599">
        <xs:annotation>
          <xs:documentation>Special Education Teachers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.241599</lx:path>
            <lx:label>Special Education Teachers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="242000">
        <xs:annotation>
          <xs:documentation>Tertiary Education Teachers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.242000</lx:path>
            <lx:label>Tertiary Education Teachers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="242100">
        <xs:annotation>
          <xs:documentation>University Lecturers and Tutors nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.242100</lx:path>
            <lx:label>University Lecturers and Tutors nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="242111">
        <xs:annotation>
          <xs:documentation>University Lecturer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.242111</lx:path>
            <lx:label>University Lecturer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="242112">
        <xs:annotation>
          <xs:documentation>University Tutor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.242112</lx:path>
            <lx:label>University Tutor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="242211">
        <xs:annotation>
          <xs:documentation>Vocational Education Teacher</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.242211</lx:path>
            <lx:label>Vocational Education Teacher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="249000">
        <xs:annotation>
          <xs:documentation>Miscellaneous Education Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.249000</lx:path>
            <lx:label>Miscellaneous Education Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="249100">
        <xs:annotation>
          <xs:documentation>Education Advisers and Reviewers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.249100</lx:path>
            <lx:label>Education Advisers and Reviewers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="249111">
        <xs:annotation>
          <xs:documentation>Education Adviser</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.249111</lx:path>
            <lx:label>Education Adviser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="249112">
        <xs:annotation>
          <xs:documentation>Education Reviewer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.249112</lx:path>
            <lx:label>Education Reviewer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="249200">
        <xs:annotation>
          <xs:documentation>Private Tutors and Teachers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.249200</lx:path>
            <lx:label>Private Tutors and Teachers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="249211">
        <xs:annotation>
          <xs:documentation>Art Teacher (Private Tuition)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.249211</lx:path>
            <lx:label>Art Teacher (Private Tuition)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="249212">
        <xs:annotation>
          <xs:documentation>Dance Teacher (Private Tuition)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.249212</lx:path>
            <lx:label>Dance Teacher (Private Tuition)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="249213">
        <xs:annotation>
          <xs:documentation>Drama Teacher (Private Tuition)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.249213</lx:path>
            <lx:label>Drama Teacher (Private Tuition)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="249214">
        <xs:annotation>
          <xs:documentation>Music Teacher (Private Tuition)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.249214</lx:path>
            <lx:label>Music Teacher (Private Tuition)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="249299">
        <xs:annotation>
          <xs:documentation>Private Tutors and Teachers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.249299</lx:path>
            <lx:label>Private Tutors and Teachers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="249311">
        <xs:annotation>
          <xs:documentation>Teacher of English to Speakers of Other Languages</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.249311</lx:path>
            <lx:label>Teacher of English to Speakers of Other Languages</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="250000">
        <xs:annotation>
          <xs:documentation>Health Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.250000</lx:path>
            <lx:label>Health Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251000">
        <xs:annotation>
          <xs:documentation>Health Diagnostic and Promotion Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251000</lx:path>
            <lx:label>Health Diagnostic and Promotion Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251100">
        <xs:annotation>
          <xs:documentation>Nutrition Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251100</lx:path>
            <lx:label>Nutrition Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251111">
        <xs:annotation>
          <xs:documentation>Dietitian</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251111</lx:path>
            <lx:label>Dietitian</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251112">
        <xs:annotation>
          <xs:documentation>Nutritionist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251112</lx:path>
            <lx:label>Nutritionist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251200">
        <xs:annotation>
          <xs:documentation>Medical Imaging Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251200</lx:path>
            <lx:label>Medical Imaging Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251211">
        <xs:annotation>
          <xs:documentation>Medical Diagnostic Radiographer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251211</lx:path>
            <lx:label>Medical Diagnostic Radiographer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251212">
        <xs:annotation>
          <xs:documentation>Medical Radiation Therapist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251212</lx:path>
            <lx:label>Medical Radiation Therapist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251213">
        <xs:annotation>
          <xs:documentation>Nuclear Medicine Technologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251213</lx:path>
            <lx:label>Nuclear Medicine Technologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251214">
        <xs:annotation>
          <xs:documentation>Sonographer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251214</lx:path>
            <lx:label>Sonographer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251300">
        <xs:annotation>
          <xs:documentation>Occupational and Environmental Health Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251300</lx:path>
            <lx:label>Occupational and Environmental Health Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251311">
        <xs:annotation>
          <xs:documentation>Environmental Health Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251311</lx:path>
            <lx:label>Environmental Health Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251312">
        <xs:annotation>
          <xs:documentation>Occupational Health and Safety Adviser</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251312</lx:path>
            <lx:label>Occupational Health and Safety Adviser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251400">
        <xs:annotation>
          <xs:documentation>Optometrists and Orthoptists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251400</lx:path>
            <lx:label>Optometrists and Orthoptists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251411">
        <xs:annotation>
          <xs:documentation>Optometrist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251411</lx:path>
            <lx:label>Optometrist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251412">
        <xs:annotation>
          <xs:documentation>Orthoptist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251412</lx:path>
            <lx:label>Orthoptist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251500">
        <xs:annotation>
          <xs:documentation>Pharmacists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251500</lx:path>
            <lx:label>Pharmacists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251511">
        <xs:annotation>
          <xs:documentation>Hospital Pharmacist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251511</lx:path>
            <lx:label>Hospital Pharmacist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251512">
        <xs:annotation>
          <xs:documentation>Industrial Pharmacist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251512</lx:path>
            <lx:label>Industrial Pharmacist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251513">
        <xs:annotation>
          <xs:documentation>Retail Pharmacist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251513</lx:path>
            <lx:label>Retail Pharmacist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251900">
        <xs:annotation>
          <xs:documentation>Other Health Diagnostic and Promotion Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251900</lx:path>
            <lx:label>Other Health Diagnostic and Promotion Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251911">
        <xs:annotation>
          <xs:documentation>Health Promotion Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251911</lx:path>
            <lx:label>Health Promotion Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251912">
        <xs:annotation>
          <xs:documentation>Orthotist or Prosthetist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251912</lx:path>
            <lx:label>Orthotist or Prosthetist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="251999">
        <xs:annotation>
          <xs:documentation>Health Diagnostic and Promotion Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.251999</lx:path>
            <lx:label>Health Diagnostic and Promotion Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252000">
        <xs:annotation>
          <xs:documentation>Health Therapy Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252000</lx:path>
            <lx:label>Health Therapy Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252100">
        <xs:annotation>
          <xs:documentation>Chiropractors and Osteopaths nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252100</lx:path>
            <lx:label>Chiropractors and Osteopaths nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252111">
        <xs:annotation>
          <xs:documentation>Chiropractor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252111</lx:path>
            <lx:label>Chiropractor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252112">
        <xs:annotation>
          <xs:documentation>Osteopath</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252112</lx:path>
            <lx:label>Osteopath</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252200">
        <xs:annotation>
          <xs:documentation>Complementary Health Therapists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252200</lx:path>
            <lx:label>Complementary Health Therapists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252211">
        <xs:annotation>
          <xs:documentation>Acupuncturist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252211</lx:path>
            <lx:label>Acupuncturist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252212">
        <xs:annotation>
          <xs:documentation>Homoeopath</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252212</lx:path>
            <lx:label>Homoeopath</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252213">
        <xs:annotation>
          <xs:documentation>Naturopath</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252213</lx:path>
            <lx:label>Naturopath</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252214">
        <xs:annotation>
          <xs:documentation>Traditional Chinese Medicine Practitioner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252214</lx:path>
            <lx:label>Traditional Chinese Medicine Practitioner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252299">
        <xs:annotation>
          <xs:documentation>Complementary Health Therapists nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252299</lx:path>
            <lx:label>Complementary Health Therapists nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252300">
        <xs:annotation>
          <xs:documentation>Dental Practitioners nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252300</lx:path>
            <lx:label>Dental Practitioners nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252311">
        <xs:annotation>
          <xs:documentation>Dental Specialist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252311</lx:path>
            <lx:label>Dental Specialist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252312">
        <xs:annotation>
          <xs:documentation>Dentist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252312</lx:path>
            <lx:label>Dentist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252411">
        <xs:annotation>
          <xs:documentation>Occupational Therapist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252411</lx:path>
            <lx:label>Occupational Therapist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252511">
        <xs:annotation>
          <xs:documentation>Physiotherapist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252511</lx:path>
            <lx:label>Physiotherapist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252611">
        <xs:annotation>
          <xs:documentation>Podiatrist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252611</lx:path>
            <lx:label>Podiatrist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252700">
        <xs:annotation>
          <xs:documentation>Audiologists and Speech Pathologists \ Therapists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252700</lx:path>
            <lx:label>Audiologists and Speech Pathologists \ Therapists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252711">
        <xs:annotation>
          <xs:documentation>Audiologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252711</lx:path>
            <lx:label>Audiologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="252712">
        <xs:annotation>
          <xs:documentation>Speech Pathologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.252712</lx:path>
            <lx:label>Speech Pathologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253000">
        <xs:annotation>
          <xs:documentation>Medical Practitioners nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253000</lx:path>
            <lx:label>Medical Practitioners nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253100">
        <xs:annotation>
          <xs:documentation>General Practitioners and Resident Medical Officers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253100</lx:path>
            <lx:label>General Practitioners and Resident Medical Officers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253111">
        <xs:annotation>
          <xs:documentation>General Practitioner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253111</lx:path>
            <lx:label>General Practitioner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253112">
        <xs:annotation>
          <xs:documentation>Resident Medical Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253112</lx:path>
            <lx:label>Resident Medical Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253211">
        <xs:annotation>
          <xs:documentation>Anaesthetist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253211</lx:path>
            <lx:label>Anaesthetist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253300">
        <xs:annotation>
          <xs:documentation>Specialist Physicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253300</lx:path>
            <lx:label>Specialist Physicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253311">
        <xs:annotation>
          <xs:documentation>Specialist Physician (General Medicine)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253311</lx:path>
            <lx:label>Specialist Physician (General Medicine)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253312">
        <xs:annotation>
          <xs:documentation>Cardiologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253312</lx:path>
            <lx:label>Cardiologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253313">
        <xs:annotation>
          <xs:documentation>Clinical Haematologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253313</lx:path>
            <lx:label>Clinical Haematologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253314">
        <xs:annotation>
          <xs:documentation>Medical Oncologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253314</lx:path>
            <lx:label>Medical Oncologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253315">
        <xs:annotation>
          <xs:documentation>Endocrinologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253315</lx:path>
            <lx:label>Endocrinologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253316">
        <xs:annotation>
          <xs:documentation>Gastroenterologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253316</lx:path>
            <lx:label>Gastroenterologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253317">
        <xs:annotation>
          <xs:documentation>Intensive Care Specialist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253317</lx:path>
            <lx:label>Intensive Care Specialist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253318">
        <xs:annotation>
          <xs:documentation>Neurologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253318</lx:path>
            <lx:label>Neurologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253321">
        <xs:annotation>
          <xs:documentation>Paediatrician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253321</lx:path>
            <lx:label>Paediatrician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253322">
        <xs:annotation>
          <xs:documentation>Renal Medicine Specialist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253322</lx:path>
            <lx:label>Renal Medicine Specialist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253323">
        <xs:annotation>
          <xs:documentation>Rheumatologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253323</lx:path>
            <lx:label>Rheumatologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253324">
        <xs:annotation>
          <xs:documentation>Thoracic Medicine Specialist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253324</lx:path>
            <lx:label>Thoracic Medicine Specialist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253399">
        <xs:annotation>
          <xs:documentation>Specialist Physicians nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253399</lx:path>
            <lx:label>Specialist Physicians nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253411">
        <xs:annotation>
          <xs:documentation>Psychiatrist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253411</lx:path>
            <lx:label>Psychiatrist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253500">
        <xs:annotation>
          <xs:documentation>Surgeons nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253500</lx:path>
            <lx:label>Surgeons nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253511">
        <xs:annotation>
          <xs:documentation>Surgeon (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253511</lx:path>
            <lx:label>Surgeon (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253512">
        <xs:annotation>
          <xs:documentation>Cardiothoracic Surgeon</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253512</lx:path>
            <lx:label>Cardiothoracic Surgeon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253513">
        <xs:annotation>
          <xs:documentation>Neurosurgeon</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253513</lx:path>
            <lx:label>Neurosurgeon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253514">
        <xs:annotation>
          <xs:documentation>Orthopaedic Surgeon</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253514</lx:path>
            <lx:label>Orthopaedic Surgeon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253515">
        <xs:annotation>
          <xs:documentation>Otorhinolaryngologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253515</lx:path>
            <lx:label>Otorhinolaryngologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253516">
        <xs:annotation>
          <xs:documentation>Paediatric Surgeon</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253516</lx:path>
            <lx:label>Paediatric Surgeon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253517">
        <xs:annotation>
          <xs:documentation>Plastic and Reconstructive Surgeon</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253517</lx:path>
            <lx:label>Plastic and Reconstructive Surgeon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253518">
        <xs:annotation>
          <xs:documentation>Urologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253518</lx:path>
            <lx:label>Urologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253521">
        <xs:annotation>
          <xs:documentation>Vascular Surgeon</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253521</lx:path>
            <lx:label>Vascular Surgeon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253900">
        <xs:annotation>
          <xs:documentation>Other Medical Practitioners nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253900</lx:path>
            <lx:label>Other Medical Practitioners nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253911">
        <xs:annotation>
          <xs:documentation>Dermatologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253911</lx:path>
            <lx:label>Dermatologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253912">
        <xs:annotation>
          <xs:documentation>Emergency Medicine Specialist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253912</lx:path>
            <lx:label>Emergency Medicine Specialist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253913">
        <xs:annotation>
          <xs:documentation>Obstetrician and Gynaecologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253913</lx:path>
            <lx:label>Obstetrician and Gynaecologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253914">
        <xs:annotation>
          <xs:documentation>Ophthalmologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253914</lx:path>
            <lx:label>Ophthalmologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253915">
        <xs:annotation>
          <xs:documentation>Pathologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253915</lx:path>
            <lx:label>Pathologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253917">
        <xs:annotation>
          <xs:documentation>Diagnostic and Interventional Radiologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253917</lx:path>
            <lx:label>Diagnostic and Interventional Radiologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253918">
        <xs:annotation>
          <xs:documentation>Radiation Oncologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253918</lx:path>
            <lx:label>Radiation Oncologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="253999">
        <xs:annotation>
          <xs:documentation>Medical Practitioners nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.253999</lx:path>
            <lx:label>Medical Practitioners nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254000">
        <xs:annotation>
          <xs:documentation>Midwifery and Nursing Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254000</lx:path>
            <lx:label>Midwifery and Nursing Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254111">
        <xs:annotation>
          <xs:documentation>Midwife</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254111</lx:path>
            <lx:label>Midwife</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254200">
        <xs:annotation>
          <xs:documentation>Nurse Educators and Researchers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254200</lx:path>
            <lx:label>Nurse Educators and Researchers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254211">
        <xs:annotation>
          <xs:documentation>Nurse Educator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254211</lx:path>
            <lx:label>Nurse Educator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254212">
        <xs:annotation>
          <xs:documentation>Nurse Researcher</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254212</lx:path>
            <lx:label>Nurse Researcher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254311">
        <xs:annotation>
          <xs:documentation>Nurse Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254311</lx:path>
            <lx:label>Nurse Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254400">
        <xs:annotation>
          <xs:documentation>Registered Nurses nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254400</lx:path>
            <lx:label>Registered Nurses nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254411">
        <xs:annotation>
          <xs:documentation>Nurse Practitioner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254411</lx:path>
            <lx:label>Nurse Practitioner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254412">
        <xs:annotation>
          <xs:documentation>Registered Nurse (Aged Care)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254412</lx:path>
            <lx:label>Registered Nurse (Aged Care)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254413">
        <xs:annotation>
          <xs:documentation>Registered Nurse (Child and Family Health)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254413</lx:path>
            <lx:label>Registered Nurse (Child and Family Health)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254414">
        <xs:annotation>
          <xs:documentation>Registered Nurse (Community Health)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254414</lx:path>
            <lx:label>Registered Nurse (Community Health)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254415">
        <xs:annotation>
          <xs:documentation>Registered Nurse (Critical Care and Emergency)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254415</lx:path>
            <lx:label>Registered Nurse (Critical Care and Emergency)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254416">
        <xs:annotation>
          <xs:documentation>Registered Nurse (Developmental Disability)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254416</lx:path>
            <lx:label>Registered Nurse (Developmental Disability)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254417">
        <xs:annotation>
          <xs:documentation>Registered Nurse (Disability and Rehabilitation)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254417</lx:path>
            <lx:label>Registered Nurse (Disability and Rehabilitation)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254418">
        <xs:annotation>
          <xs:documentation>Registered Nurse (Medical)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254418</lx:path>
            <lx:label>Registered Nurse (Medical)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254421">
        <xs:annotation>
          <xs:documentation>Registered Nurse (Medical Practice)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254421</lx:path>
            <lx:label>Registered Nurse (Medical Practice)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254422">
        <xs:annotation>
          <xs:documentation>Registered Nurse (Mental Health)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254422</lx:path>
            <lx:label>Registered Nurse (Mental Health)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254423">
        <xs:annotation>
          <xs:documentation>Registered Nurse (Perioperative)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254423</lx:path>
            <lx:label>Registered Nurse (Perioperative)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254424">
        <xs:annotation>
          <xs:documentation>Registered Nurse (Surgical)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254424</lx:path>
            <lx:label>Registered Nurse (Surgical)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254425">
        <xs:annotation>
          <xs:documentation>Registered Nurse (Paediatrics)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254425</lx:path>
            <lx:label>Registered Nurse (Paediatrics)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="254499">
        <xs:annotation>
          <xs:documentation>Registered Nurses nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.254499</lx:path>
            <lx:label>Registered Nurses nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="260000">
        <xs:annotation>
          <xs:documentation>ICT Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.260000</lx:path>
            <lx:label>ICT Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="261000">
        <xs:annotation>
          <xs:documentation>Business and Systems Analysts, and Programmers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.261000</lx:path>
            <lx:label>Business and Systems Analysts, and Programmers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="261100">
        <xs:annotation>
          <xs:documentation>ICT Business and Systems Analysts nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.261100</lx:path>
            <lx:label>ICT Business and Systems Analysts nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="261111">
        <xs:annotation>
          <xs:documentation>ICT Business Analyst</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.261111</lx:path>
            <lx:label>ICT Business Analyst</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="261112">
        <xs:annotation>
          <xs:documentation>Systems Analyst</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.261112</lx:path>
            <lx:label>Systems Analyst</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="261200">
        <xs:annotation>
          <xs:documentation>Multimedia Specialists and Web Developers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.261200</lx:path>
            <lx:label>Multimedia Specialists and Web Developers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="261211">
        <xs:annotation>
          <xs:documentation>Multimedia Specialist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.261211</lx:path>
            <lx:label>Multimedia Specialist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="261212">
        <xs:annotation>
          <xs:documentation>Web Developer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.261212</lx:path>
            <lx:label>Web Developer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="261300">
        <xs:annotation>
          <xs:documentation>Software and Applications Programmers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.261300</lx:path>
            <lx:label>Software and Applications Programmers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="261311">
        <xs:annotation>
          <xs:documentation>Analyst Programmer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.261311</lx:path>
            <lx:label>Analyst Programmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="261312">
        <xs:annotation>
          <xs:documentation>Developer Programmer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.261312</lx:path>
            <lx:label>Developer Programmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="261313">
        <xs:annotation>
          <xs:documentation>Software Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.261313</lx:path>
            <lx:label>Software Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="261314">
        <xs:annotation>
          <xs:documentation>Software Tester</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.261314</lx:path>
            <lx:label>Software Tester</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="261399">
        <xs:annotation>
          <xs:documentation>Software and Applications Programmers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.261399</lx:path>
            <lx:label>Software and Applications Programmers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="262100">
        <xs:annotation>
          <xs:documentation>Database and Systems Administrators, and ICT Security Specialists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.262100</lx:path>
            <lx:label>Database and Systems Administrators, and ICT Security Specialists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="262111">
        <xs:annotation>
          <xs:documentation>Database Administrator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.262111</lx:path>
            <lx:label>Database Administrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="262112">
        <xs:annotation>
          <xs:documentation>ICT Security Specialist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.262112</lx:path>
            <lx:label>ICT Security Specialist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="262113">
        <xs:annotation>
          <xs:documentation>Systems Administrator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.262113</lx:path>
            <lx:label>Systems Administrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="263000">
        <xs:annotation>
          <xs:documentation>ICT Network and Support Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.263000</lx:path>
            <lx:label>ICT Network and Support Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="263100">
        <xs:annotation>
          <xs:documentation>Computer Network Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.263100</lx:path>
            <lx:label>Computer Network Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="263111">
        <xs:annotation>
          <xs:documentation>Computer Network and Systems Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.263111</lx:path>
            <lx:label>Computer Network and Systems Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="263112">
        <xs:annotation>
          <xs:documentation>Network Administrator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.263112</lx:path>
            <lx:label>Network Administrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="263113">
        <xs:annotation>
          <xs:documentation>Network Analyst</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.263113</lx:path>
            <lx:label>Network Analyst</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="263200">
        <xs:annotation>
          <xs:documentation>ICT Support and Test Engineers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.263200</lx:path>
            <lx:label>ICT Support and Test Engineers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="263211">
        <xs:annotation>
          <xs:documentation>ICT Quality Assurance Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.263211</lx:path>
            <lx:label>ICT Quality Assurance Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="263212">
        <xs:annotation>
          <xs:documentation>ICT Support Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.263212</lx:path>
            <lx:label>ICT Support Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="263213">
        <xs:annotation>
          <xs:documentation>ICT Systems Test Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.263213</lx:path>
            <lx:label>ICT Systems Test Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="263299">
        <xs:annotation>
          <xs:documentation>ICT Support and Test Engineers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.263299</lx:path>
            <lx:label>ICT Support and Test Engineers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="263300">
        <xs:annotation>
          <xs:documentation>Telecommunications Engineering Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.263300</lx:path>
            <lx:label>Telecommunications Engineering Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="263311">
        <xs:annotation>
          <xs:documentation>Telecommunications Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.263311</lx:path>
            <lx:label>Telecommunications Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="263312">
        <xs:annotation>
          <xs:documentation>Telecommunications Network Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.263312</lx:path>
            <lx:label>Telecommunications Network Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="270000">
        <xs:annotation>
          <xs:documentation>Legal, Social and Welfare Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.270000</lx:path>
            <lx:label>Legal, Social and Welfare Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="271000">
        <xs:annotation>
          <xs:documentation>Legal Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.271000</lx:path>
            <lx:label>Legal Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="271111">
        <xs:annotation>
          <xs:documentation>Barrister</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.271111</lx:path>
            <lx:label>Barrister</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="271200">
        <xs:annotation>
          <xs:documentation>Judicial and Other Legal Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.271200</lx:path>
            <lx:label>Judicial and Other Legal Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="271211">
        <xs:annotation>
          <xs:documentation>Judge</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.271211</lx:path>
            <lx:label>Judge</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="271212">
        <xs:annotation>
          <xs:documentation>Magistrate</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.271212</lx:path>
            <lx:label>Magistrate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="271213">
        <xs:annotation>
          <xs:documentation>Tribunal Member</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.271213</lx:path>
            <lx:label>Tribunal Member</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="271214">
        <xs:annotation>
          <xs:documentation>Intellectual Property Lawyer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.271214</lx:path>
            <lx:label>Intellectual Property Lawyer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="271299">
        <xs:annotation>
          <xs:documentation>Judicial and Other Legal Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.271299</lx:path>
            <lx:label>Judicial and Other Legal Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="271311">
        <xs:annotation>
          <xs:documentation>Solicitor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.271311</lx:path>
            <lx:label>Solicitor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272000">
        <xs:annotation>
          <xs:documentation>Social and Welfare Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272000</lx:path>
            <lx:label>Social and Welfare Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272100">
        <xs:annotation>
          <xs:documentation>Counsellors nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272100</lx:path>
            <lx:label>Counsellors nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272111">
        <xs:annotation>
          <xs:documentation>Careers Counsellor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272111</lx:path>
            <lx:label>Careers Counsellor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272112">
        <xs:annotation>
          <xs:documentation>Drug and Alcohol Counsellor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272112</lx:path>
            <lx:label>Drug and Alcohol Counsellor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272113">
        <xs:annotation>
          <xs:documentation>Family and Marriage Counsellor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272113</lx:path>
            <lx:label>Family and Marriage Counsellor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272114">
        <xs:annotation>
          <xs:documentation>Rehabilitation Counsellor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272114</lx:path>
            <lx:label>Rehabilitation Counsellor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272115">
        <xs:annotation>
          <xs:documentation>Student Counsellor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272115</lx:path>
            <lx:label>Student Counsellor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272199">
        <xs:annotation>
          <xs:documentation>Counsellors nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272199</lx:path>
            <lx:label>Counsellors nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272211">
        <xs:annotation>
          <xs:documentation>Minister of Religion</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272211</lx:path>
            <lx:label>Minister of Religion</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272300">
        <xs:annotation>
          <xs:documentation>Psychologists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272300</lx:path>
            <lx:label>Psychologists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272311">
        <xs:annotation>
          <xs:documentation>Clinical Psychologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272311</lx:path>
            <lx:label>Clinical Psychologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272312">
        <xs:annotation>
          <xs:documentation>Educational Psychologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272312</lx:path>
            <lx:label>Educational Psychologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272313">
        <xs:annotation>
          <xs:documentation>Organisational Psychologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272313</lx:path>
            <lx:label>Organisational Psychologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272314">
        <xs:annotation>
          <xs:documentation>Psychotherapist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272314</lx:path>
            <lx:label>Psychotherapist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272399">
        <xs:annotation>
          <xs:documentation>Psychologists nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272399</lx:path>
            <lx:label>Psychologists nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272400">
        <xs:annotation>
          <xs:documentation>Social Professionals nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272400</lx:path>
            <lx:label>Social Professionals nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272411">
        <xs:annotation>
          <xs:documentation>Historian</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272411</lx:path>
            <lx:label>Historian</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272412">
        <xs:annotation>
          <xs:documentation>Interpreter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272412</lx:path>
            <lx:label>Interpreter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272413">
        <xs:annotation>
          <xs:documentation>Translator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272413</lx:path>
            <lx:label>Translator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272414">
        <xs:annotation>
          <xs:documentation>Archaeologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272414</lx:path>
            <lx:label>Archaeologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272499">
        <xs:annotation>
          <xs:documentation>Social Professionals nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272499</lx:path>
            <lx:label>Social Professionals nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272511">
        <xs:annotation>
          <xs:documentation>Social Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272511</lx:path>
            <lx:label>Social Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272600">
        <xs:annotation>
          <xs:documentation>Welfare, Recreation and Community Arts Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272600</lx:path>
            <lx:label>Welfare, Recreation and Community Arts Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272611">
        <xs:annotation>
          <xs:documentation>Community Arts Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272611</lx:path>
            <lx:label>Community Arts Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272612">
        <xs:annotation>
          <xs:documentation>Recreation Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272612</lx:path>
            <lx:label>Recreation Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="272613">
        <xs:annotation>
          <xs:documentation>Welfare Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.272613</lx:path>
            <lx:label>Welfare Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="300000">
        <xs:annotation>
          <xs:documentation>Technicians and Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.300000</lx:path>
            <lx:label>Technicians and Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="310000">
        <xs:annotation>
          <xs:documentation>Engineering, ICT and Science Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.310000</lx:path>
            <lx:label>Engineering, ICT and Science Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311000">
        <xs:annotation>
          <xs:documentation>Agricultural, Medical and Science Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311000</lx:path>
            <lx:label>Agricultural, Medical and Science Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311111">
        <xs:annotation>
          <xs:documentation>Agricultural Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311111</lx:path>
            <lx:label>Agricultural Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311200">
        <xs:annotation>
          <xs:documentation>Medical Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311200</lx:path>
            <lx:label>Medical Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311211">
        <xs:annotation>
          <xs:documentation>Anaesthetic Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311211</lx:path>
            <lx:label>Anaesthetic Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311212">
        <xs:annotation>
          <xs:documentation>Cardiac Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311212</lx:path>
            <lx:label>Cardiac Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311213">
        <xs:annotation>
          <xs:documentation>Medical Laboratory Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311213</lx:path>
            <lx:label>Medical Laboratory Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311214">
        <xs:annotation>
          <xs:documentation>Operating Theatre Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311214</lx:path>
            <lx:label>Operating Theatre Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311215">
        <xs:annotation>
          <xs:documentation>Pharmacy Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311215</lx:path>
            <lx:label>Pharmacy Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311216">
        <xs:annotation>
          <xs:documentation>Pathology Collector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311216</lx:path>
            <lx:label>Pathology Collector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311299">
        <xs:annotation>
          <xs:documentation>Medical Technicians nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311299</lx:path>
            <lx:label>Medical Technicians nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311300">
        <xs:annotation>
          <xs:documentation>Primary Products Inspectors nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311300</lx:path>
            <lx:label>Primary Products Inspectors nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311311">
        <xs:annotation>
          <xs:documentation>Fisheries Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311311</lx:path>
            <lx:label>Fisheries Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311312">
        <xs:annotation>
          <xs:documentation>Meat Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311312</lx:path>
            <lx:label>Meat Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311313">
        <xs:annotation>
          <xs:documentation>Quarantine Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311313</lx:path>
            <lx:label>Quarantine Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311399">
        <xs:annotation>
          <xs:documentation>Primary Products Inspectors nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311399</lx:path>
            <lx:label>Primary Products Inspectors nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311400">
        <xs:annotation>
          <xs:documentation>Science Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311400</lx:path>
            <lx:label>Science Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311411">
        <xs:annotation>
          <xs:documentation>Chemistry Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311411</lx:path>
            <lx:label>Chemistry Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311412">
        <xs:annotation>
          <xs:documentation>Earth Science Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311412</lx:path>
            <lx:label>Earth Science Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311413">
        <xs:annotation>
          <xs:documentation>Life Science Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311413</lx:path>
            <lx:label>Life Science Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311414">
        <xs:annotation>
          <xs:documentation>School Laboratory Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311414</lx:path>
            <lx:label>School Laboratory Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311415">
        <xs:annotation>
          <xs:documentation>Hydrographer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311415</lx:path>
            <lx:label>Hydrographer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="311499">
        <xs:annotation>
          <xs:documentation>Science Technicians nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.311499</lx:path>
            <lx:label>Science Technicians nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312000">
        <xs:annotation>
          <xs:documentation>Building and Engineering Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312000</lx:path>
            <lx:label>Building and Engineering Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312100">
        <xs:annotation>
          <xs:documentation>Architectural, Building and Surveying Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312100</lx:path>
            <lx:label>Architectural, Building and Surveying Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312111">
        <xs:annotation>
          <xs:documentation>Architectural Draftsperson</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312111</lx:path>
            <lx:label>Architectural Draftsperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312112">
        <xs:annotation>
          <xs:documentation>Building Associate</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312112</lx:path>
            <lx:label>Building Associate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312113">
        <xs:annotation>
          <xs:documentation>Building Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312113</lx:path>
            <lx:label>Building Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312114">
        <xs:annotation>
          <xs:documentation>Construction Estimator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312114</lx:path>
            <lx:label>Construction Estimator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312115">
        <xs:annotation>
          <xs:documentation>Plumbing Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312115</lx:path>
            <lx:label>Plumbing Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312116">
        <xs:annotation>
          <xs:documentation>Surveying or Spatial Science Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312116</lx:path>
            <lx:label>Surveying or Spatial Science Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312199">
        <xs:annotation>
          <xs:documentation>Architectural, Building and Surveying Technicians nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312199</lx:path>
            <lx:label>Architectural, Building and Surveying Technicians nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312200">
        <xs:annotation>
          <xs:documentation>Civil Engineering Draftspersons and Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312200</lx:path>
            <lx:label>Civil Engineering Draftspersons and Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312211">
        <xs:annotation>
          <xs:documentation>Civil Engineering Draftsperson</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312211</lx:path>
            <lx:label>Civil Engineering Draftsperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312212">
        <xs:annotation>
          <xs:documentation>Civil Engineering Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312212</lx:path>
            <lx:label>Civil Engineering Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312300">
        <xs:annotation>
          <xs:documentation>Electrical Engineering Draftspersons and Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312300</lx:path>
            <lx:label>Electrical Engineering Draftspersons and Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312311">
        <xs:annotation>
          <xs:documentation>Electrical Engineering Draftsperson</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312311</lx:path>
            <lx:label>Electrical Engineering Draftsperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312312">
        <xs:annotation>
          <xs:documentation>Electrical Engineering Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312312</lx:path>
            <lx:label>Electrical Engineering Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312400">
        <xs:annotation>
          <xs:documentation>Electronic Engineering Draftspersons and Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312400</lx:path>
            <lx:label>Electronic Engineering Draftspersons and Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312411">
        <xs:annotation>
          <xs:documentation>Electronic Engineering Draftsperson</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312411</lx:path>
            <lx:label>Electronic Engineering Draftsperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312412">
        <xs:annotation>
          <xs:documentation>Electronic Engineering Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312412</lx:path>
            <lx:label>Electronic Engineering Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312500">
        <xs:annotation>
          <xs:documentation>Mechanical Engineering Draftspersons and Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312500</lx:path>
            <lx:label>Mechanical Engineering Draftspersons and Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312511">
        <xs:annotation>
          <xs:documentation>Mechanical Engineering Draftsperson</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312511</lx:path>
            <lx:label>Mechanical Engineering Draftsperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312512">
        <xs:annotation>
          <xs:documentation>Mechanical Engineering Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312512</lx:path>
            <lx:label>Mechanical Engineering Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312611">
        <xs:annotation>
          <xs:documentation>Safety Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312611</lx:path>
            <lx:label>Safety Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312900">
        <xs:annotation>
          <xs:documentation>Other Building and Engineering Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312900</lx:path>
            <lx:label>Other Building and Engineering Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312911">
        <xs:annotation>
          <xs:documentation>Maintenance Planner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312911</lx:path>
            <lx:label>Maintenance Planner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312912">
        <xs:annotation>
          <xs:documentation>Metallurgical or Materials Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312912</lx:path>
            <lx:label>Metallurgical or Materials Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312913">
        <xs:annotation>
          <xs:documentation>Mine Deputy</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312913</lx:path>
            <lx:label>Mine Deputy</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="312999">
        <xs:annotation>
          <xs:documentation>Building and Engineering Technicians nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.312999</lx:path>
            <lx:label>Building and Engineering Technicians nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="313000">
        <xs:annotation>
          <xs:documentation>ICT and Telecommunications Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.313000</lx:path>
            <lx:label>ICT and Telecommunications Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="313100">
        <xs:annotation>
          <xs:documentation>ICT Support Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.313100</lx:path>
            <lx:label>ICT Support Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="313111">
        <xs:annotation>
          <xs:documentation>Hardware Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.313111</lx:path>
            <lx:label>Hardware Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="313112">
        <xs:annotation>
          <xs:documentation>ICT Customer Support Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.313112</lx:path>
            <lx:label>ICT Customer Support Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="313113">
        <xs:annotation>
          <xs:documentation>Web Administrator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.313113</lx:path>
            <lx:label>Web Administrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="313199">
        <xs:annotation>
          <xs:documentation>ICT Support Technicians nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.313199</lx:path>
            <lx:label>ICT Support Technicians nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="313200">
        <xs:annotation>
          <xs:documentation>Telecommunications Technical Specialists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.313200</lx:path>
            <lx:label>Telecommunications Technical Specialists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="313211">
        <xs:annotation>
          <xs:documentation>Radiocommunications Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.313211</lx:path>
            <lx:label>Radiocommunications Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="313212">
        <xs:annotation>
          <xs:documentation>Telecommunications Field Engineer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.313212</lx:path>
            <lx:label>Telecommunications Field Engineer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="313213">
        <xs:annotation>
          <xs:documentation>Telecommunications Network Planner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.313213</lx:path>
            <lx:label>Telecommunications Network Planner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="313214">
        <xs:annotation>
          <xs:documentation>Telecommunications Technical Officer or Technologist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.313214</lx:path>
            <lx:label>Telecommunications Technical Officer or Technologist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="320000">
        <xs:annotation>
          <xs:documentation>Automotive and Engineering Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.320000</lx:path>
            <lx:label>Automotive and Engineering Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="321000">
        <xs:annotation>
          <xs:documentation>Automotive Electricians and Mechanics nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.321000</lx:path>
            <lx:label>Automotive Electricians and Mechanics nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="321111">
        <xs:annotation>
          <xs:documentation>Automotive Electrician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.321111</lx:path>
            <lx:label>Automotive Electrician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="321200">
        <xs:annotation>
          <xs:documentation>Motor Mechanics nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.321200</lx:path>
            <lx:label>Motor Mechanics nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="321211">
        <xs:annotation>
          <xs:documentation>Motor Mechanic (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.321211</lx:path>
            <lx:label>Motor Mechanic (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="321212">
        <xs:annotation>
          <xs:documentation>Diesel Motor Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.321212</lx:path>
            <lx:label>Diesel Motor Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="321213">
        <xs:annotation>
          <xs:documentation>Motorcycle Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.321213</lx:path>
            <lx:label>Motorcycle Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="321214">
        <xs:annotation>
          <xs:documentation>Small Engine Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.321214</lx:path>
            <lx:label>Small Engine Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="322000">
        <xs:annotation>
          <xs:documentation>Fabrication Engineering Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.322000</lx:path>
            <lx:label>Fabrication Engineering Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="322100">
        <xs:annotation>
          <xs:documentation>Metal Casting, Forging and Finishing Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.322100</lx:path>
            <lx:label>Metal Casting, Forging and Finishing Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="322111">
        <xs:annotation>
          <xs:documentation>Blacksmith</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.322111</lx:path>
            <lx:label>Blacksmith</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="322112">
        <xs:annotation>
          <xs:documentation>Electroplater</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.322112</lx:path>
            <lx:label>Electroplater</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="322113">
        <xs:annotation>
          <xs:documentation>Farrier</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.322113</lx:path>
            <lx:label>Farrier</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="322114">
        <xs:annotation>
          <xs:documentation>Metal Casting Trades Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.322114</lx:path>
            <lx:label>Metal Casting Trades Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="322115">
        <xs:annotation>
          <xs:documentation>Metal Polisher</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.322115</lx:path>
            <lx:label>Metal Polisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="322211">
        <xs:annotation>
          <xs:documentation>Sheetmetal Trades Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.322211</lx:path>
            <lx:label>Sheetmetal Trades Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="322300">
        <xs:annotation>
          <xs:documentation>Structural Steel and Welding Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.322300</lx:path>
            <lx:label>Structural Steel and Welding Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="322311">
        <xs:annotation>
          <xs:documentation>Metal Fabricator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.322311</lx:path>
            <lx:label>Metal Fabricator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="322312">
        <xs:annotation>
          <xs:documentation>Pressure Welder</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.322312</lx:path>
            <lx:label>Pressure Welder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="322313">
        <xs:annotation>
          <xs:documentation>Welder (First Class)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.322313</lx:path>
            <lx:label>Welder (First Class)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323000">
        <xs:annotation>
          <xs:documentation>Mechanical Engineering Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323000</lx:path>
            <lx:label>Mechanical Engineering Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323100">
        <xs:annotation>
          <xs:documentation>Aircraft Maintenance Engineers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323100</lx:path>
            <lx:label>Aircraft Maintenance Engineers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323111">
        <xs:annotation>
          <xs:documentation>Aircraft Maintenance Engineer (Avionics)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323111</lx:path>
            <lx:label>Aircraft Maintenance Engineer (Avionics)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323112">
        <xs:annotation>
          <xs:documentation>Aircraft Maintenance Engineer (Mechanical)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323112</lx:path>
            <lx:label>Aircraft Maintenance Engineer (Mechanical)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323113">
        <xs:annotation>
          <xs:documentation>Aircraft Maintenance Engineer (Structures)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323113</lx:path>
            <lx:label>Aircraft Maintenance Engineer (Structures)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323200">
        <xs:annotation>
          <xs:documentation>Metal Fitters and Machinists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323200</lx:path>
            <lx:label>Metal Fitters and Machinists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323211">
        <xs:annotation>
          <xs:documentation>Fitter (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323211</lx:path>
            <lx:label>Fitter (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323212">
        <xs:annotation>
          <xs:documentation>Fitter and Turner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323212</lx:path>
            <lx:label>Fitter and Turner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323213">
        <xs:annotation>
          <xs:documentation>Fitter-Welder</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323213</lx:path>
            <lx:label>Fitter-Welder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323214">
        <xs:annotation>
          <xs:documentation>Metal Machinist (First Class)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323214</lx:path>
            <lx:label>Metal Machinist (First Class)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323215">
        <xs:annotation>
          <xs:documentation>Textile, Clothing and Footwear Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323215</lx:path>
            <lx:label>Textile, Clothing and Footwear Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323299">
        <xs:annotation>
          <xs:documentation>Metal Fitters and Machinists nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323299</lx:path>
            <lx:label>Metal Fitters and Machinists nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323300">
        <xs:annotation>
          <xs:documentation>Precision Metal Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323300</lx:path>
            <lx:label>Precision Metal Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323311">
        <xs:annotation>
          <xs:documentation>Engraver</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323311</lx:path>
            <lx:label>Engraver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323312">
        <xs:annotation>
          <xs:documentation>Gunsmith</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323312</lx:path>
            <lx:label>Gunsmith</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323313">
        <xs:annotation>
          <xs:documentation>Locksmith</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323313</lx:path>
            <lx:label>Locksmith</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323314">
        <xs:annotation>
          <xs:documentation>Precision Instrument Maker and Repairer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323314</lx:path>
            <lx:label>Precision Instrument Maker and Repairer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323315">
        <xs:annotation>
          <xs:documentation>Saw Doctor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323315</lx:path>
            <lx:label>Saw Doctor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323316">
        <xs:annotation>
          <xs:documentation>Watch and Clock Maker and Repairer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323316</lx:path>
            <lx:label>Watch and Clock Maker and Repairer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323400">
        <xs:annotation>
          <xs:documentation>Toolmakers and Engineering Patternmakers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323400</lx:path>
            <lx:label>Toolmakers and Engineering Patternmakers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323411">
        <xs:annotation>
          <xs:documentation>Engineering Patternmaker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323411</lx:path>
            <lx:label>Engineering Patternmaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="323412">
        <xs:annotation>
          <xs:documentation>Toolmaker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.323412</lx:path>
            <lx:label>Toolmaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="324000">
        <xs:annotation>
          <xs:documentation>Panelbeaters, and Vehicle Body Builders, Trimmers and Painters nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.324000</lx:path>
            <lx:label>Panelbeaters, and Vehicle Body Builders, Trimmers and Painters nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="324111">
        <xs:annotation>
          <xs:documentation>Panelbeater</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.324111</lx:path>
            <lx:label>Panelbeater</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="324200">
        <xs:annotation>
          <xs:documentation>Vehicle Body Builders and Trimmers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.324200</lx:path>
            <lx:label>Vehicle Body Builders and Trimmers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="324211">
        <xs:annotation>
          <xs:documentation>Vehicle Body Builder</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.324211</lx:path>
            <lx:label>Vehicle Body Builder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="324212">
        <xs:annotation>
          <xs:documentation>Vehicle Trimmer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.324212</lx:path>
            <lx:label>Vehicle Trimmer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="324311">
        <xs:annotation>
          <xs:documentation>Vehicle Painter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.324311</lx:path>
            <lx:label>Vehicle Painter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="330000">
        <xs:annotation>
          <xs:documentation>Construction Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.330000</lx:path>
            <lx:label>Construction Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="331000">
        <xs:annotation>
          <xs:documentation>Bricklayers, and Carpenters and Joiners nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.331000</lx:path>
            <lx:label>Bricklayers, and Carpenters and Joiners nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="331100">
        <xs:annotation>
          <xs:documentation>Bricklayers and Stonemasons nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.331100</lx:path>
            <lx:label>Bricklayers and Stonemasons nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="331111">
        <xs:annotation>
          <xs:documentation>Bricklayer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.331111</lx:path>
            <lx:label>Bricklayer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="331112">
        <xs:annotation>
          <xs:documentation>Stonemason</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.331112</lx:path>
            <lx:label>Stonemason</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="331211">
        <xs:annotation>
          <xs:documentation>Carpenter and Joiner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.331211</lx:path>
            <lx:label>Carpenter and Joiner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="331212">
        <xs:annotation>
          <xs:documentation>Carpenter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.331212</lx:path>
            <lx:label>Carpenter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="331213">
        <xs:annotation>
          <xs:documentation>Joiner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.331213</lx:path>
            <lx:label>Joiner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="332000">
        <xs:annotation>
          <xs:documentation>Floor Finishers and Painting Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.332000</lx:path>
            <lx:label>Floor Finishers and Painting Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="332111">
        <xs:annotation>
          <xs:documentation>Floor Finisher</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.332111</lx:path>
            <lx:label>Floor Finisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="332211">
        <xs:annotation>
          <xs:documentation>Painting Trades Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.332211</lx:path>
            <lx:label>Painting Trades Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="333000">
        <xs:annotation>
          <xs:documentation>Glaziers, Plasterers and Tilers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.333000</lx:path>
            <lx:label>Glaziers, Plasterers and Tilers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="333111">
        <xs:annotation>
          <xs:documentation>Glazier</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.333111</lx:path>
            <lx:label>Glazier</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="333200">
        <xs:annotation>
          <xs:documentation>Plasterers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.333200</lx:path>
            <lx:label>Plasterers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="333211">
        <xs:annotation>
          <xs:documentation>Fibrous Plasterer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.333211</lx:path>
            <lx:label>Fibrous Plasterer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="333212">
        <xs:annotation>
          <xs:documentation>Solid Plasterer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.333212</lx:path>
            <lx:label>Solid Plasterer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="333311">
        <xs:annotation>
          <xs:documentation>Roof Tiler</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.333311</lx:path>
            <lx:label>Roof Tiler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="333411">
        <xs:annotation>
          <xs:documentation>Wall and Floor Tiler</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.333411</lx:path>
            <lx:label>Wall and Floor Tiler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="334100">
        <xs:annotation>
          <xs:documentation>Plumbers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.334100</lx:path>
            <lx:label>Plumbers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="334111">
        <xs:annotation>
          <xs:documentation>Plumber (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.334111</lx:path>
            <lx:label>Plumber (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="334112">
        <xs:annotation>
          <xs:documentation>Airconditioning and Mechanical Services Plumber</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.334112</lx:path>
            <lx:label>Airconditioning and Mechanical Services Plumber</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="334113">
        <xs:annotation>
          <xs:documentation>Drainer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.334113</lx:path>
            <lx:label>Drainer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="334114">
        <xs:annotation>
          <xs:documentation>Gasfitter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.334114</lx:path>
            <lx:label>Gasfitter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="334115">
        <xs:annotation>
          <xs:documentation>Roof Plumber</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.334115</lx:path>
            <lx:label>Roof Plumber</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="340000">
        <xs:annotation>
          <xs:documentation>Electrotechnology and Telecommunications Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.340000</lx:path>
            <lx:label>Electrotechnology and Telecommunications Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="341100">
        <xs:annotation>
          <xs:documentation>Electricians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.341100</lx:path>
            <lx:label>Electricians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="341111">
        <xs:annotation>
          <xs:documentation>Electrician (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.341111</lx:path>
            <lx:label>Electrician (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="341112">
        <xs:annotation>
          <xs:documentation>Electrician (Special Class)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.341112</lx:path>
            <lx:label>Electrician (Special Class)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="341113">
        <xs:annotation>
          <xs:documentation>Lift Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.341113</lx:path>
            <lx:label>Lift Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342000">
        <xs:annotation>
          <xs:documentation>Electronics and Telecommunications Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342000</lx:path>
            <lx:label>Electronics and Telecommunications Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342111">
        <xs:annotation>
          <xs:documentation>Airconditioning and Refrigeration Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342111</lx:path>
            <lx:label>Airconditioning and Refrigeration Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342200">
        <xs:annotation>
          <xs:documentation>Electrical Distribution Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342200</lx:path>
            <lx:label>Electrical Distribution Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342211">
        <xs:annotation>
          <xs:documentation>Electrical Linesworker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342211</lx:path>
            <lx:label>Electrical Linesworker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342212">
        <xs:annotation>
          <xs:documentation>Technical Cable Jointer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342212</lx:path>
            <lx:label>Technical Cable Jointer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342300">
        <xs:annotation>
          <xs:documentation>Electronics Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342300</lx:path>
            <lx:label>Electronics Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342311">
        <xs:annotation>
          <xs:documentation>Business Machine Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342311</lx:path>
            <lx:label>Business Machine Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342312">
        <xs:annotation>
          <xs:documentation>Communications Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342312</lx:path>
            <lx:label>Communications Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342313">
        <xs:annotation>
          <xs:documentation>Electronic Equipment Trades Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342313</lx:path>
            <lx:label>Electronic Equipment Trades Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342314">
        <xs:annotation>
          <xs:documentation>Electronic Instrument Trades Worker (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342314</lx:path>
            <lx:label>Electronic Instrument Trades Worker (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342315">
        <xs:annotation>
          <xs:documentation>Electronic Instrument Trades Worker (Special Class)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342315</lx:path>
            <lx:label>Electronic Instrument Trades Worker (Special Class)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342400">
        <xs:annotation>
          <xs:documentation>Telecommunications Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342400</lx:path>
            <lx:label>Telecommunications Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342411">
        <xs:annotation>
          <xs:documentation>Cabler (Data and Telecommunications)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342411</lx:path>
            <lx:label>Cabler (Data and Telecommunications)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342412">
        <xs:annotation>
          <xs:documentation>Telecommunications Cable Jointer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342412</lx:path>
            <lx:label>Telecommunications Cable Jointer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342413">
        <xs:annotation>
          <xs:documentation>Telecommunications Linesworker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342413</lx:path>
            <lx:label>Telecommunications Linesworker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="342414">
        <xs:annotation>
          <xs:documentation>Telecommunications Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.342414</lx:path>
            <lx:label>Telecommunications Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="351000">
        <xs:annotation>
          <xs:documentation>Food Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.351000</lx:path>
            <lx:label>Food Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="351100">
        <xs:annotation>
          <xs:documentation>Bakers and Pastrycooks nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.351100</lx:path>
            <lx:label>Bakers and Pastrycooks nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="351111">
        <xs:annotation>
          <xs:documentation>Baker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.351111</lx:path>
            <lx:label>Baker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="351112">
        <xs:annotation>
          <xs:documentation>Pastrycook</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.351112</lx:path>
            <lx:label>Pastrycook</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="351211">
        <xs:annotation>
          <xs:documentation>Butcher or Smallgoods Maker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.351211</lx:path>
            <lx:label>Butcher or Smallgoods Maker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="351311">
        <xs:annotation>
          <xs:documentation>Chef</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.351311</lx:path>
            <lx:label>Chef</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="351411">
        <xs:annotation>
          <xs:documentation>Cook</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.351411</lx:path>
            <lx:label>Cook</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="360000">
        <xs:annotation>
          <xs:documentation>Skilled Animal and Horticultural Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.360000</lx:path>
            <lx:label>Skilled Animal and Horticultural Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="361000">
        <xs:annotation>
          <xs:documentation>Animal Attendants and Trainers, and Shearers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.361000</lx:path>
            <lx:label>Animal Attendants and Trainers, and Shearers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="361100">
        <xs:annotation>
          <xs:documentation>Animal Attendants and Trainers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.361100</lx:path>
            <lx:label>Animal Attendants and Trainers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="361111">
        <xs:annotation>
          <xs:documentation>Dog Handler or Trainer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.361111</lx:path>
            <lx:label>Dog Handler or Trainer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="361112">
        <xs:annotation>
          <xs:documentation>Horse Trainer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.361112</lx:path>
            <lx:label>Horse Trainer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="361113">
        <xs:annotation>
          <xs:documentation>Pet Groomer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.361113</lx:path>
            <lx:label>Pet Groomer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="361114">
        <xs:annotation>
          <xs:documentation>Zookeeper</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.361114</lx:path>
            <lx:label>Zookeeper</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="361115">
        <xs:annotation>
          <xs:documentation>Kennel Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.361115</lx:path>
            <lx:label>Kennel Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="361199">
        <xs:annotation>
          <xs:documentation>Animal Attendants and Trainers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.361199</lx:path>
            <lx:label>Animal Attendants and Trainers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="361211">
        <xs:annotation>
          <xs:documentation>Shearer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.361211</lx:path>
            <lx:label>Shearer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="361311">
        <xs:annotation>
          <xs:documentation>Veterinary Nurse</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.361311</lx:path>
            <lx:label>Veterinary Nurse</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="362000">
        <xs:annotation>
          <xs:documentation>Horticultural Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.362000</lx:path>
            <lx:label>Horticultural Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="362111">
        <xs:annotation>
          <xs:documentation>Florist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.362111</lx:path>
            <lx:label>Florist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="362200">
        <xs:annotation>
          <xs:documentation>Gardeners nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.362200</lx:path>
            <lx:label>Gardeners nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="362211">
        <xs:annotation>
          <xs:documentation>Gardener (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.362211</lx:path>
            <lx:label>Gardener (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="362212">
        <xs:annotation>
          <xs:documentation>Arborist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.362212</lx:path>
            <lx:label>Arborist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="362213">
        <xs:annotation>
          <xs:documentation>Landscape Gardener</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.362213</lx:path>
            <lx:label>Landscape Gardener</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="362311">
        <xs:annotation>
          <xs:documentation>Greenkeeper</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.362311</lx:path>
            <lx:label>Greenkeeper</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="362411">
        <xs:annotation>
          <xs:documentation>Nurseryperson</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.362411</lx:path>
            <lx:label>Nurseryperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="390000">
        <xs:annotation>
          <xs:documentation>Other Technicians and Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.390000</lx:path>
            <lx:label>Other Technicians and Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="391111">
        <xs:annotation>
          <xs:documentation>Hairdresser</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.391111</lx:path>
            <lx:label>Hairdresser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="392000">
        <xs:annotation>
          <xs:documentation>Printing Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.392000</lx:path>
            <lx:label>Printing Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="392100">
        <xs:annotation>
          <xs:documentation>Print Finishers and Screen Printers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.392100</lx:path>
            <lx:label>Print Finishers and Screen Printers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="392111">
        <xs:annotation>
          <xs:documentation>Print Finisher</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.392111</lx:path>
            <lx:label>Print Finisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="392112">
        <xs:annotation>
          <xs:documentation>Screen Printer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.392112</lx:path>
            <lx:label>Screen Printer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="392211">
        <xs:annotation>
          <xs:documentation>Graphic Pre-press Trades Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.392211</lx:path>
            <lx:label>Graphic Pre-press Trades Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="392300">
        <xs:annotation>
          <xs:documentation>Printers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.392300</lx:path>
            <lx:label>Printers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="392311">
        <xs:annotation>
          <xs:documentation>Printing Machinist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.392311</lx:path>
            <lx:label>Printing Machinist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="392312">
        <xs:annotation>
          <xs:documentation>Small Offset Printer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.392312</lx:path>
            <lx:label>Small Offset Printer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="393000">
        <xs:annotation>
          <xs:documentation>Textile, Clothing and Footwear Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.393000</lx:path>
            <lx:label>Textile, Clothing and Footwear Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="393100">
        <xs:annotation>
          <xs:documentation>Canvas and Leather Goods Makers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.393100</lx:path>
            <lx:label>Canvas and Leather Goods Makers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="393111">
        <xs:annotation>
          <xs:documentation>Canvas Goods Fabricator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.393111</lx:path>
            <lx:label>Canvas Goods Fabricator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="393112">
        <xs:annotation>
          <xs:documentation>Leather Goods Maker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.393112</lx:path>
            <lx:label>Leather Goods Maker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="393113">
        <xs:annotation>
          <xs:documentation>Sail Maker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.393113</lx:path>
            <lx:label>Sail Maker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="393114">
        <xs:annotation>
          <xs:documentation>Shoemaker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.393114</lx:path>
            <lx:label>Shoemaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="393200">
        <xs:annotation>
          <xs:documentation>Clothing Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.393200</lx:path>
            <lx:label>Clothing Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="393211">
        <xs:annotation>
          <xs:documentation>Apparel Cutter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.393211</lx:path>
            <lx:label>Apparel Cutter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="393212">
        <xs:annotation>
          <xs:documentation>Clothing Patternmaker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.393212</lx:path>
            <lx:label>Clothing Patternmaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="393213">
        <xs:annotation>
          <xs:documentation>Dressmaker or Tailor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.393213</lx:path>
            <lx:label>Dressmaker or Tailor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="393299">
        <xs:annotation>
          <xs:documentation>Clothing Trades Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.393299</lx:path>
            <lx:label>Clothing Trades Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="393311">
        <xs:annotation>
          <xs:documentation>Upholsterer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.393311</lx:path>
            <lx:label>Upholsterer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="394000">
        <xs:annotation>
          <xs:documentation>Wood Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.394000</lx:path>
            <lx:label>Wood Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="394111">
        <xs:annotation>
          <xs:documentation>Cabinetmaker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.394111</lx:path>
            <lx:label>Cabinetmaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="394200">
        <xs:annotation>
          <xs:documentation>Wood Machinists and Other Wood Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.394200</lx:path>
            <lx:label>Wood Machinists and Other Wood Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="394211">
        <xs:annotation>
          <xs:documentation>Furniture Finisher</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.394211</lx:path>
            <lx:label>Furniture Finisher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="394212">
        <xs:annotation>
          <xs:documentation>Picture Framer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.394212</lx:path>
            <lx:label>Picture Framer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="394213">
        <xs:annotation>
          <xs:documentation>Wood Machinist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.394213</lx:path>
            <lx:label>Wood Machinist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="394214">
        <xs:annotation>
          <xs:documentation>Wood Turner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.394214</lx:path>
            <lx:label>Wood Turner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="394299">
        <xs:annotation>
          <xs:documentation>Wood Machinists and Other Wood Trades Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.394299</lx:path>
            <lx:label>Wood Machinists and Other Wood Trades Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399000">
        <xs:annotation>
          <xs:documentation>Miscellaneous Technicians and Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399000</lx:path>
            <lx:label>Miscellaneous Technicians and Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399100">
        <xs:annotation>
          <xs:documentation>Boat Builders and Shipwrights nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399100</lx:path>
            <lx:label>Boat Builders and Shipwrights nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399111">
        <xs:annotation>
          <xs:documentation>Boat Builder and Repairer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399111</lx:path>
            <lx:label>Boat Builder and Repairer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399112">
        <xs:annotation>
          <xs:documentation>Shipwright</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399112</lx:path>
            <lx:label>Shipwright</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399200">
        <xs:annotation>
          <xs:documentation>Chemical, Gas, Petroleum and Power Generation Plant Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399200</lx:path>
            <lx:label>Chemical, Gas, Petroleum and Power Generation Plant Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399211">
        <xs:annotation>
          <xs:documentation>Chemical Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399211</lx:path>
            <lx:label>Chemical Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399212">
        <xs:annotation>
          <xs:documentation>Gas or Petroleum Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399212</lx:path>
            <lx:label>Gas or Petroleum Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399213">
        <xs:annotation>
          <xs:documentation>Power Generation Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399213</lx:path>
            <lx:label>Power Generation Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399300">
        <xs:annotation>
          <xs:documentation>Gallery, Library and Museum Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399300</lx:path>
            <lx:label>Gallery, Library and Museum Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399311">
        <xs:annotation>
          <xs:documentation>Gallery or Museum Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399311</lx:path>
            <lx:label>Gallery or Museum Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399312">
        <xs:annotation>
          <xs:documentation>Library Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399312</lx:path>
            <lx:label>Library Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399411">
        <xs:annotation>
          <xs:documentation>Jeweller</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399411</lx:path>
            <lx:label>Jeweller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399500">
        <xs:annotation>
          <xs:documentation>Performing Arts Technicians nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399500</lx:path>
            <lx:label>Performing Arts Technicians nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399511">
        <xs:annotation>
          <xs:documentation>Broadcast Transmitter Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399511</lx:path>
            <lx:label>Broadcast Transmitter Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399512">
        <xs:annotation>
          <xs:documentation>Camera Operator (Film, Television or Video)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399512</lx:path>
            <lx:label>Camera Operator (Film, Television or Video)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399513">
        <xs:annotation>
          <xs:documentation>Light Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399513</lx:path>
            <lx:label>Light Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399514">
        <xs:annotation>
          <xs:documentation>Make Up Artist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399514</lx:path>
            <lx:label>Make Up Artist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399515">
        <xs:annotation>
          <xs:documentation>Musical Instrument Maker or Repairer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399515</lx:path>
            <lx:label>Musical Instrument Maker or Repairer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399516">
        <xs:annotation>
          <xs:documentation>Sound Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399516</lx:path>
            <lx:label>Sound Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399517">
        <xs:annotation>
          <xs:documentation>Television Equipment Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399517</lx:path>
            <lx:label>Television Equipment Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399599">
        <xs:annotation>
          <xs:documentation>Performing Arts Technicians nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399599</lx:path>
            <lx:label>Performing Arts Technicians nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399611">
        <xs:annotation>
          <xs:documentation>Signwriter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399611</lx:path>
            <lx:label>Signwriter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399900">
        <xs:annotation>
          <xs:documentation>Other Miscellaneous Technicians and Trades Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399900</lx:path>
            <lx:label>Other Miscellaneous Technicians and Trades Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399911">
        <xs:annotation>
          <xs:documentation>Diver</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399911</lx:path>
            <lx:label>Diver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399912">
        <xs:annotation>
          <xs:documentation>Interior Decorator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399912</lx:path>
            <lx:label>Interior Decorator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399913">
        <xs:annotation>
          <xs:documentation>Optical Dispenser</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399913</lx:path>
            <lx:label>Optical Dispenser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399914">
        <xs:annotation>
          <xs:documentation>Optical Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399914</lx:path>
            <lx:label>Optical Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399915">
        <xs:annotation>
          <xs:documentation>Photographer's Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399915</lx:path>
            <lx:label>Photographer's Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399916">
        <xs:annotation>
          <xs:documentation>Plastics Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399916</lx:path>
            <lx:label>Plastics Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399917">
        <xs:annotation>
          <xs:documentation>Wool Classer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399917</lx:path>
            <lx:label>Wool Classer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399918">
        <xs:annotation>
          <xs:documentation>Fire Protection Equipment Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399918</lx:path>
            <lx:label>Fire Protection Equipment Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="399999">
        <xs:annotation>
          <xs:documentation>Technicians and Trades Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.399999</lx:path>
            <lx:label>Technicians and Trades Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="400000">
        <xs:annotation>
          <xs:documentation>Community and Personal Service Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.400000</lx:path>
            <lx:label>Community and Personal Service Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411000">
        <xs:annotation>
          <xs:documentation>Health and Welfare Support Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411000</lx:path>
            <lx:label>Health and Welfare Support Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411100">
        <xs:annotation>
          <xs:documentation>Ambulance Officers and Paramedics nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411100</lx:path>
            <lx:label>Ambulance Officers and Paramedics nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411111">
        <xs:annotation>
          <xs:documentation>Ambulance Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411111</lx:path>
            <lx:label>Ambulance Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411112">
        <xs:annotation>
          <xs:documentation>Intensive Care Ambulance Paramedic</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411112</lx:path>
            <lx:label>Intensive Care Ambulance Paramedic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411200">
        <xs:annotation>
          <xs:documentation>Dental Hygienists, Technicians and Therapists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411200</lx:path>
            <lx:label>Dental Hygienists, Technicians and Therapists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411211">
        <xs:annotation>
          <xs:documentation>Dental Hygienist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411211</lx:path>
            <lx:label>Dental Hygienist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411212">
        <xs:annotation>
          <xs:documentation>Dental Prosthetist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411212</lx:path>
            <lx:label>Dental Prosthetist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411213">
        <xs:annotation>
          <xs:documentation>Dental Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411213</lx:path>
            <lx:label>Dental Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411214">
        <xs:annotation>
          <xs:documentation>Dental Therapist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411214</lx:path>
            <lx:label>Dental Therapist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411311">
        <xs:annotation>
          <xs:documentation>Diversional Therapist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411311</lx:path>
            <lx:label>Diversional Therapist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411400">
        <xs:annotation>
          <xs:documentation>Enrolled and Mothercraft Nurses nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411400</lx:path>
            <lx:label>Enrolled and Mothercraft Nurses nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411411">
        <xs:annotation>
          <xs:documentation>Enrolled Nurse</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411411</lx:path>
            <lx:label>Enrolled Nurse</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411412">
        <xs:annotation>
          <xs:documentation>Mothercraft Nurse</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411412</lx:path>
            <lx:label>Mothercraft Nurse</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411511">
        <xs:annotation>
          <xs:documentation>Aboriginal and Torres Strait Islander Health Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411511</lx:path>
            <lx:label>Aboriginal and Torres Strait Islander Health Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411611">
        <xs:annotation>
          <xs:documentation>Massage Therapist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411611</lx:path>
            <lx:label>Massage Therapist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411700">
        <xs:annotation>
          <xs:documentation>Welfare Support Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411700</lx:path>
            <lx:label>Welfare Support Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411711">
        <xs:annotation>
          <xs:documentation>Community Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411711</lx:path>
            <lx:label>Community Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411712">
        <xs:annotation>
          <xs:documentation>Disabilities Services Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411712</lx:path>
            <lx:label>Disabilities Services Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411713">
        <xs:annotation>
          <xs:documentation>Family Support Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411713</lx:path>
            <lx:label>Family Support Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411714">
        <xs:annotation>
          <xs:documentation>Parole or Probation Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411714</lx:path>
            <lx:label>Parole or Probation Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411715">
        <xs:annotation>
          <xs:documentation>Residential Care Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411715</lx:path>
            <lx:label>Residential Care Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="411716">
        <xs:annotation>
          <xs:documentation>Youth Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.411716</lx:path>
            <lx:label>Youth Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="420000">
        <xs:annotation>
          <xs:documentation>Carers and Aides nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.420000</lx:path>
            <lx:label>Carers and Aides nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="421100">
        <xs:annotation>
          <xs:documentation>Child Carers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.421100</lx:path>
            <lx:label>Child Carers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="421111">
        <xs:annotation>
          <xs:documentation>Child Care Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.421111</lx:path>
            <lx:label>Child Care Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="421112">
        <xs:annotation>
          <xs:documentation>Family Day Care Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.421112</lx:path>
            <lx:label>Family Day Care Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="421113">
        <xs:annotation>
          <xs:documentation>Nanny</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.421113</lx:path>
            <lx:label>Nanny</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="421114">
        <xs:annotation>
          <xs:documentation>Out of School Hours Care Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.421114</lx:path>
            <lx:label>Out of School Hours Care Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="422100">
        <xs:annotation>
          <xs:documentation>Education Aides nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.422100</lx:path>
            <lx:label>Education Aides nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="422111">
        <xs:annotation>
          <xs:documentation>Aboriginal and Torres Strait Islander Education Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.422111</lx:path>
            <lx:label>Aboriginal and Torres Strait Islander Education Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="422112">
        <xs:annotation>
          <xs:documentation>Integration Aide</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.422112</lx:path>
            <lx:label>Integration Aide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="422115">
        <xs:annotation>
          <xs:documentation>Preschool Aide</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.422115</lx:path>
            <lx:label>Preschool Aide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="422116">
        <xs:annotation>
          <xs:documentation>Teachers' Aide</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.422116</lx:path>
            <lx:label>Teachers' Aide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="423000">
        <xs:annotation>
          <xs:documentation>Personal Carers and Assistants nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.423000</lx:path>
            <lx:label>Personal Carers and Assistants nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="423111">
        <xs:annotation>
          <xs:documentation>Aged or Disabled Carer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.423111</lx:path>
            <lx:label>Aged or Disabled Carer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="423211">
        <xs:annotation>
          <xs:documentation>Dental Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.423211</lx:path>
            <lx:label>Dental Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="423300">
        <xs:annotation>
          <xs:documentation>Nursing Support and Personal Care Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.423300</lx:path>
            <lx:label>Nursing Support and Personal Care Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="423311">
        <xs:annotation>
          <xs:documentation>Hospital Orderly</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.423311</lx:path>
            <lx:label>Hospital Orderly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="423312">
        <xs:annotation>
          <xs:documentation>Nursing Support Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.423312</lx:path>
            <lx:label>Nursing Support Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="423313">
        <xs:annotation>
          <xs:documentation>Personal Care Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.423313</lx:path>
            <lx:label>Personal Care Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="423314">
        <xs:annotation>
          <xs:documentation>Therapy Aide</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.423314</lx:path>
            <lx:label>Therapy Aide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="423400">
        <xs:annotation>
          <xs:documentation>Special Care Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.423400</lx:path>
            <lx:label>Special Care Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="423411">
        <xs:annotation>
          <xs:documentation>Child or Youth Residential Care Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.423411</lx:path>
            <lx:label>Child or Youth Residential Care Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="423412">
        <xs:annotation>
          <xs:documentation>Hostel Parent</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.423412</lx:path>
            <lx:label>Hostel Parent</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="423413">
        <xs:annotation>
          <xs:documentation>Refuge Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.423413</lx:path>
            <lx:label>Refuge Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="431000">
        <xs:annotation>
          <xs:documentation>Hospitality Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.431000</lx:path>
            <lx:label>Hospitality Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="431100">
        <xs:annotation>
          <xs:documentation>Bar Attendants and Baristas nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.431100</lx:path>
            <lx:label>Bar Attendants and Baristas nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="431111">
        <xs:annotation>
          <xs:documentation>Bar Attendant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.431111</lx:path>
            <lx:label>Bar Attendant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="431112">
        <xs:annotation>
          <xs:documentation>Barista</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.431112</lx:path>
            <lx:label>Barista</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="431211">
        <xs:annotation>
          <xs:documentation>Cafe Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.431211</lx:path>
            <lx:label>Cafe Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="431311">
        <xs:annotation>
          <xs:documentation>Gaming Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.431311</lx:path>
            <lx:label>Gaming Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="431411">
        <xs:annotation>
          <xs:documentation>Hotel Service Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.431411</lx:path>
            <lx:label>Hotel Service Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="431511">
        <xs:annotation>
          <xs:documentation>Waiter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.431511</lx:path>
            <lx:label>Waiter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="431900">
        <xs:annotation>
          <xs:documentation>Other Hospitality Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.431900</lx:path>
            <lx:label>Other Hospitality Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="431911">
        <xs:annotation>
          <xs:documentation>Bar Useful or Busser</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.431911</lx:path>
            <lx:label>Bar Useful or Busser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="431912">
        <xs:annotation>
          <xs:documentation>Doorperson or Luggage Porter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.431912</lx:path>
            <lx:label>Doorperson or Luggage Porter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="431999">
        <xs:annotation>
          <xs:documentation>Hospitality Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.431999</lx:path>
            <lx:label>Hospitality Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="440000">
        <xs:annotation>
          <xs:documentation>Protective Service Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.440000</lx:path>
            <lx:label>Protective Service Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="441000">
        <xs:annotation>
          <xs:documentation>Defence Force Members, Fire Fighters and Police nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.441000</lx:path>
            <lx:label>Defence Force Members, Fire Fighters and Police nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="441111">
        <xs:annotation>
          <xs:documentation>Defence Force Member - Other Ranks</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.441111</lx:path>
            <lx:label>Defence Force Member - Other Ranks</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="441200">
        <xs:annotation>
          <xs:documentation>Fire and Emergency Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.441200</lx:path>
            <lx:label>Fire and Emergency Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="441211">
        <xs:annotation>
          <xs:documentation>Emergency Service Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.441211</lx:path>
            <lx:label>Emergency Service Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="441212">
        <xs:annotation>
          <xs:documentation>Fire Fighter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.441212</lx:path>
            <lx:label>Fire Fighter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="441300">
        <xs:annotation>
          <xs:documentation>Police nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.441300</lx:path>
            <lx:label>Police nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="441311">
        <xs:annotation>
          <xs:documentation>Detective</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.441311</lx:path>
            <lx:label>Detective</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="441312">
        <xs:annotation>
          <xs:documentation>Police Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.441312</lx:path>
            <lx:label>Police Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="442000">
        <xs:annotation>
          <xs:documentation>Prison and Security Officers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.442000</lx:path>
            <lx:label>Prison and Security Officers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="442111">
        <xs:annotation>
          <xs:documentation>Prison Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.442111</lx:path>
            <lx:label>Prison Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="442200">
        <xs:annotation>
          <xs:documentation>Security Officers and Guards nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.442200</lx:path>
            <lx:label>Security Officers and Guards nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="442211">
        <xs:annotation>
          <xs:documentation>Alarm, Security or Surveillance Monitor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.442211</lx:path>
            <lx:label>Alarm, Security or Surveillance Monitor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="442212">
        <xs:annotation>
          <xs:documentation>Armoured Car Escort</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.442212</lx:path>
            <lx:label>Armoured Car Escort</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="442213">
        <xs:annotation>
          <xs:documentation>Crowd Controller</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.442213</lx:path>
            <lx:label>Crowd Controller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="442214">
        <xs:annotation>
          <xs:documentation>Private Investigator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.442214</lx:path>
            <lx:label>Private Investigator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="442215">
        <xs:annotation>
          <xs:documentation>Retail Loss Prevention Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.442215</lx:path>
            <lx:label>Retail Loss Prevention Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="442216">
        <xs:annotation>
          <xs:documentation>Security Consultant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.442216</lx:path>
            <lx:label>Security Consultant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="442217">
        <xs:annotation>
          <xs:documentation>Security Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.442217</lx:path>
            <lx:label>Security Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="442299">
        <xs:annotation>
          <xs:documentation>Security Officers and Guards nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.442299</lx:path>
            <lx:label>Security Officers and Guards nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="450000">
        <xs:annotation>
          <xs:documentation>Sports and Personal Service Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.450000</lx:path>
            <lx:label>Sports and Personal Service Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451000">
        <xs:annotation>
          <xs:documentation>Personal Service and Travel Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451000</lx:path>
            <lx:label>Personal Service and Travel Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451111">
        <xs:annotation>
          <xs:documentation>Beauty Therapist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451111</lx:path>
            <lx:label>Beauty Therapist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451211">
        <xs:annotation>
          <xs:documentation>Driving Instructor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451211</lx:path>
            <lx:label>Driving Instructor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451300">
        <xs:annotation>
          <xs:documentation>Funeral Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451300</lx:path>
            <lx:label>Funeral Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451311">
        <xs:annotation>
          <xs:documentation>Funeral Director</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451311</lx:path>
            <lx:label>Funeral Director</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451399">
        <xs:annotation>
          <xs:documentation>Funeral Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451399</lx:path>
            <lx:label>Funeral Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451400">
        <xs:annotation>
          <xs:documentation>Gallery, Museum and Tour Guides nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451400</lx:path>
            <lx:label>Gallery, Museum and Tour Guides nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451411">
        <xs:annotation>
          <xs:documentation>Gallery or Museum Guide</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451411</lx:path>
            <lx:label>Gallery or Museum Guide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451412">
        <xs:annotation>
          <xs:documentation>Tour Guide</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451412</lx:path>
            <lx:label>Tour Guide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451500">
        <xs:annotation>
          <xs:documentation>Personal Care Consultants nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451500</lx:path>
            <lx:label>Personal Care Consultants nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451511">
        <xs:annotation>
          <xs:documentation>Natural Remedy Consultant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451511</lx:path>
            <lx:label>Natural Remedy Consultant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451512">
        <xs:annotation>
          <xs:documentation>Weight Loss Consultant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451512</lx:path>
            <lx:label>Weight Loss Consultant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451513">
        <xs:annotation>
          <xs:documentation>Herbalist (Western)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451513</lx:path>
            <lx:label>Herbalist (Western)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451600">
        <xs:annotation>
          <xs:documentation>Tourism and Travel Advisers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451600</lx:path>
            <lx:label>Tourism and Travel Advisers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451611">
        <xs:annotation>
          <xs:documentation>Tourist Information Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451611</lx:path>
            <lx:label>Tourist Information Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451612">
        <xs:annotation>
          <xs:documentation>Travel Consultant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451612</lx:path>
            <lx:label>Travel Consultant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451700">
        <xs:annotation>
          <xs:documentation>Travel Attendants nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451700</lx:path>
            <lx:label>Travel Attendants nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451711">
        <xs:annotation>
          <xs:documentation>Flight Attendant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451711</lx:path>
            <lx:label>Flight Attendant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451799">
        <xs:annotation>
          <xs:documentation>Travel Attendants nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451799</lx:path>
            <lx:label>Travel Attendants nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451800">
        <xs:annotation>
          <xs:documentation>Other Personal Service Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451800</lx:path>
            <lx:label>Other Personal Service Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451811">
        <xs:annotation>
          <xs:documentation>Civil Celebrant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451811</lx:path>
            <lx:label>Civil Celebrant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451812">
        <xs:annotation>
          <xs:documentation>Hair or Beauty Salon Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451812</lx:path>
            <lx:label>Hair or Beauty Salon Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451813">
        <xs:annotation>
          <xs:documentation>Sex Worker or Escort</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451813</lx:path>
            <lx:label>Sex Worker or Escort</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451814">
        <xs:annotation>
          <xs:documentation>Body Artist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451814</lx:path>
            <lx:label>Body Artist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451815">
        <xs:annotation>
          <xs:documentation>First Aid Trainer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451815</lx:path>
            <lx:label>First Aid Trainer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451816">
        <xs:annotation>
          <xs:documentation>Religious Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451816</lx:path>
            <lx:label>Religious Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="451899">
        <xs:annotation>
          <xs:documentation>Personal Service Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.451899</lx:path>
            <lx:label>Personal Service Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452000">
        <xs:annotation>
          <xs:documentation>Sports and Fitness Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452000</lx:path>
            <lx:label>Sports and Fitness Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452111">
        <xs:annotation>
          <xs:documentation>Fitness Instructor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452111</lx:path>
            <lx:label>Fitness Instructor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452200">
        <xs:annotation>
          <xs:documentation>Outdoor Adventure Guides nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452200</lx:path>
            <lx:label>Outdoor Adventure Guides nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452211">
        <xs:annotation>
          <xs:documentation>Bungy Jump Master</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452211</lx:path>
            <lx:label>Bungy Jump Master</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452212">
        <xs:annotation>
          <xs:documentation>Fishing Guide</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452212</lx:path>
            <lx:label>Fishing Guide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452213">
        <xs:annotation>
          <xs:documentation>Hunting Guide</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452213</lx:path>
            <lx:label>Hunting Guide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452214">
        <xs:annotation>
          <xs:documentation>Mountain or Glacier Guide</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452214</lx:path>
            <lx:label>Mountain or Glacier Guide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452215">
        <xs:annotation>
          <xs:documentation>Outdoor Adventure Instructor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452215</lx:path>
            <lx:label>Outdoor Adventure Instructor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452216">
        <xs:annotation>
          <xs:documentation>Trekking Guide</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452216</lx:path>
            <lx:label>Trekking Guide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452217">
        <xs:annotation>
          <xs:documentation>Whitewater Rafting Guide</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452217</lx:path>
            <lx:label>Whitewater Rafting Guide</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452299">
        <xs:annotation>
          <xs:documentation>Outdoor Adventure Guides nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452299</lx:path>
            <lx:label>Outdoor Adventure Guides nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452300">
        <xs:annotation>
          <xs:documentation>Sports Coaches, Instructors and Officials nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452300</lx:path>
            <lx:label>Sports Coaches, Instructors and Officials nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452311">
        <xs:annotation>
          <xs:documentation>Diving Instructor (Open Water)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452311</lx:path>
            <lx:label>Diving Instructor (Open Water)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452312">
        <xs:annotation>
          <xs:documentation>Gymnastics Coach or Instructor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452312</lx:path>
            <lx:label>Gymnastics Coach or Instructor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452313">
        <xs:annotation>
          <xs:documentation>Horse Riding Coach or Instructor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452313</lx:path>
            <lx:label>Horse Riding Coach or Instructor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452314">
        <xs:annotation>
          <xs:documentation>Snowsport Instructor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452314</lx:path>
            <lx:label>Snowsport Instructor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452315">
        <xs:annotation>
          <xs:documentation>Swimming Coach or Instructor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452315</lx:path>
            <lx:label>Swimming Coach or Instructor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452316">
        <xs:annotation>
          <xs:documentation>Tennis Coach</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452316</lx:path>
            <lx:label>Tennis Coach</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452317">
        <xs:annotation>
          <xs:documentation>Other Sports Coach or Instructor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452317</lx:path>
            <lx:label>Other Sports Coach or Instructor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452318">
        <xs:annotation>
          <xs:documentation>Dog or Horse Racing Official</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452318</lx:path>
            <lx:label>Dog or Horse Racing Official</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452321">
        <xs:annotation>
          <xs:documentation>Sports Development Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452321</lx:path>
            <lx:label>Sports Development Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452322">
        <xs:annotation>
          <xs:documentation>Sports Umpire</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452322</lx:path>
            <lx:label>Sports Umpire</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452323">
        <xs:annotation>
          <xs:documentation>Other Sports Official</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452323</lx:path>
            <lx:label>Other Sports Official</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452400">
        <xs:annotation>
          <xs:documentation>Sportspersons nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452400</lx:path>
            <lx:label>Sportspersons nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452411">
        <xs:annotation>
          <xs:documentation>Footballer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452411</lx:path>
            <lx:label>Footballer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452412">
        <xs:annotation>
          <xs:documentation>Golfer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452412</lx:path>
            <lx:label>Golfer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452413">
        <xs:annotation>
          <xs:documentation>Jockey</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452413</lx:path>
            <lx:label>Jockey</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452414">
        <xs:annotation>
          <xs:documentation>Lifeguard</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452414</lx:path>
            <lx:label>Lifeguard</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="452499">
        <xs:annotation>
          <xs:documentation>Sportspersons nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.452499</lx:path>
            <lx:label>Sportspersons nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="500000">
        <xs:annotation>
          <xs:documentation>Clerical and Administrative Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.500000</lx:path>
            <lx:label>Clerical and Administrative Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="510000">
        <xs:annotation>
          <xs:documentation>Office Managers and Program Administrators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.510000</lx:path>
            <lx:label>Office Managers and Program Administrators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="511100">
        <xs:annotation>
          <xs:documentation>Contract, Program and Project Administrators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.511100</lx:path>
            <lx:label>Contract, Program and Project Administrators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="511111">
        <xs:annotation>
          <xs:documentation>Contract Administrator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.511111</lx:path>
            <lx:label>Contract Administrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="511112">
        <xs:annotation>
          <xs:documentation>Program or Project Administrator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.511112</lx:path>
            <lx:label>Program or Project Administrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="512000">
        <xs:annotation>
          <xs:documentation>Office and Practice Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.512000</lx:path>
            <lx:label>Office and Practice Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="512111">
        <xs:annotation>
          <xs:documentation>Office Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.512111</lx:path>
            <lx:label>Office Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="512200">
        <xs:annotation>
          <xs:documentation>Practice Managers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.512200</lx:path>
            <lx:label>Practice Managers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="512211">
        <xs:annotation>
          <xs:documentation>Health Practice Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.512211</lx:path>
            <lx:label>Health Practice Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="512299">
        <xs:annotation>
          <xs:documentation>Practice Managers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.512299</lx:path>
            <lx:label>Practice Managers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="521000">
        <xs:annotation>
          <xs:documentation>Personal Assistants and Secretaries nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.521000</lx:path>
            <lx:label>Personal Assistants and Secretaries nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="521111">
        <xs:annotation>
          <xs:documentation>Personal Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.521111</lx:path>
            <lx:label>Personal Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="521200">
        <xs:annotation>
          <xs:documentation>Secretaries nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.521200</lx:path>
            <lx:label>Secretaries nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="521211">
        <xs:annotation>
          <xs:documentation>Secretary (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.521211</lx:path>
            <lx:label>Secretary (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="521212">
        <xs:annotation>
          <xs:documentation>Legal Secretary</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.521212</lx:path>
            <lx:label>Legal Secretary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="530000">
        <xs:annotation>
          <xs:documentation>General Clerical Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.530000</lx:path>
            <lx:label>General Clerical Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="531111">
        <xs:annotation>
          <xs:documentation>General Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.531111</lx:path>
            <lx:label>General Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="532100">
        <xs:annotation>
          <xs:documentation>Keyboard Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.532100</lx:path>
            <lx:label>Keyboard Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="532111">
        <xs:annotation>
          <xs:documentation>Data Entry Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.532111</lx:path>
            <lx:label>Data Entry Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="532112">
        <xs:annotation>
          <xs:documentation>Machine Shorthand Reporter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.532112</lx:path>
            <lx:label>Machine Shorthand Reporter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="532113">
        <xs:annotation>
          <xs:documentation>Word Processing Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.532113</lx:path>
            <lx:label>Word Processing Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="540000">
        <xs:annotation>
          <xs:documentation>Inquiry Clerks and Receptionists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.540000</lx:path>
            <lx:label>Inquiry Clerks and Receptionists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="541000">
        <xs:annotation>
          <xs:documentation>Call or Contact Centre Information Clerks nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.541000</lx:path>
            <lx:label>Call or Contact Centre Information Clerks nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="541100">
        <xs:annotation>
          <xs:documentation>Call or Contact Centre Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.541100</lx:path>
            <lx:label>Call or Contact Centre Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="541111">
        <xs:annotation>
          <xs:documentation>Call or Contact Centre Team Leader</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.541111</lx:path>
            <lx:label>Call or Contact Centre Team Leader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="541112">
        <xs:annotation>
          <xs:documentation>Call or Contact Centre Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.541112</lx:path>
            <lx:label>Call or Contact Centre Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="541211">
        <xs:annotation>
          <xs:documentation>Information Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.541211</lx:path>
            <lx:label>Information Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="542100">
        <xs:annotation>
          <xs:documentation>Receptionists nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.542100</lx:path>
            <lx:label>Receptionists nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="542111">
        <xs:annotation>
          <xs:documentation>Receptionist (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.542111</lx:path>
            <lx:label>Receptionist (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="542112">
        <xs:annotation>
          <xs:documentation>Admissions Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.542112</lx:path>
            <lx:label>Admissions Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="542113">
        <xs:annotation>
          <xs:documentation>Hotel or Motel Receptionist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.542113</lx:path>
            <lx:label>Hotel or Motel Receptionist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="542114">
        <xs:annotation>
          <xs:documentation>Medical Receptionist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.542114</lx:path>
            <lx:label>Medical Receptionist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="550000">
        <xs:annotation>
          <xs:documentation>Numerical Clerks nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.550000</lx:path>
            <lx:label>Numerical Clerks nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="551000">
        <xs:annotation>
          <xs:documentation>Accounting Clerks and Bookkeepers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.551000</lx:path>
            <lx:label>Accounting Clerks and Bookkeepers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="551100">
        <xs:annotation>
          <xs:documentation>Accounting Clerks nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.551100</lx:path>
            <lx:label>Accounting Clerks nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="551111">
        <xs:annotation>
          <xs:documentation>Accounts Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.551111</lx:path>
            <lx:label>Accounts Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="551112">
        <xs:annotation>
          <xs:documentation>Cost Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.551112</lx:path>
            <lx:label>Cost Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="551211">
        <xs:annotation>
          <xs:documentation>Bookkeeper</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.551211</lx:path>
            <lx:label>Bookkeeper</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="551311">
        <xs:annotation>
          <xs:documentation>Payroll Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.551311</lx:path>
            <lx:label>Payroll Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="552000">
        <xs:annotation>
          <xs:documentation>Financial and Insurance Clerks nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.552000</lx:path>
            <lx:label>Financial and Insurance Clerks nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="552111">
        <xs:annotation>
          <xs:documentation>Bank Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.552111</lx:path>
            <lx:label>Bank Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="552211">
        <xs:annotation>
          <xs:documentation>Credit or Loans Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.552211</lx:path>
            <lx:label>Credit or Loans Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="552300">
        <xs:annotation>
          <xs:documentation>Insurance, Money Market and Statistical Clerks nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.552300</lx:path>
            <lx:label>Insurance, Money Market and Statistical Clerks nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="552311">
        <xs:annotation>
          <xs:documentation>Bookmaker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.552311</lx:path>
            <lx:label>Bookmaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="552312">
        <xs:annotation>
          <xs:documentation>Insurance Consultant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.552312</lx:path>
            <lx:label>Insurance Consultant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="552313">
        <xs:annotation>
          <xs:documentation>Money Market Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.552313</lx:path>
            <lx:label>Money Market Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="552314">
        <xs:annotation>
          <xs:documentation>Statistical Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.552314</lx:path>
            <lx:label>Statistical Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561000">
        <xs:annotation>
          <xs:documentation>Clerical and Office Support Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561000</lx:path>
            <lx:label>Clerical and Office Support Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561100">
        <xs:annotation>
          <xs:documentation>Betting Clerks nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561100</lx:path>
            <lx:label>Betting Clerks nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561111">
        <xs:annotation>
          <xs:documentation>Betting Agency Counter Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561111</lx:path>
            <lx:label>Betting Agency Counter Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561112">
        <xs:annotation>
          <xs:documentation>Bookmaker's Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561112</lx:path>
            <lx:label>Bookmaker's Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561113">
        <xs:annotation>
          <xs:documentation>Telephone Betting Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561113</lx:path>
            <lx:label>Telephone Betting Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561199">
        <xs:annotation>
          <xs:documentation>Betting Clerks nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561199</lx:path>
            <lx:label>Betting Clerks nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561200">
        <xs:annotation>
          <xs:documentation>Couriers and Postal Deliverers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561200</lx:path>
            <lx:label>Couriers and Postal Deliverers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561211">
        <xs:annotation>
          <xs:documentation>Courier</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561211</lx:path>
            <lx:label>Courier</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561212">
        <xs:annotation>
          <xs:documentation>Postal Delivery Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561212</lx:path>
            <lx:label>Postal Delivery Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561311">
        <xs:annotation>
          <xs:documentation>Filing or Registry Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561311</lx:path>
            <lx:label>Filing or Registry Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561400">
        <xs:annotation>
          <xs:documentation>Mail Sorters nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561400</lx:path>
            <lx:label>Mail Sorters nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561411">
        <xs:annotation>
          <xs:documentation>Mail Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561411</lx:path>
            <lx:label>Mail Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561412">
        <xs:annotation>
          <xs:documentation>Postal Sorting Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561412</lx:path>
            <lx:label>Postal Sorting Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561511">
        <xs:annotation>
          <xs:documentation>Survey Interviewer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561511</lx:path>
            <lx:label>Survey Interviewer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561611">
        <xs:annotation>
          <xs:documentation>Switchboard Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561611</lx:path>
            <lx:label>Switchboard Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561900">
        <xs:annotation>
          <xs:documentation>Other Clerical and Office Support Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561900</lx:path>
            <lx:label>Other Clerical and Office Support Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561911">
        <xs:annotation>
          <xs:documentation>Classified Advertising Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561911</lx:path>
            <lx:label>Classified Advertising Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561912">
        <xs:annotation>
          <xs:documentation>Meter Reader</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561912</lx:path>
            <lx:label>Meter Reader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561913">
        <xs:annotation>
          <xs:documentation>Parking Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561913</lx:path>
            <lx:label>Parking Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="561999">
        <xs:annotation>
          <xs:documentation>Clerical and Office Support Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.561999</lx:path>
            <lx:label>Clerical and Office Support Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="590000">
        <xs:annotation>
          <xs:documentation>Other Clerical and Administrative Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.590000</lx:path>
            <lx:label>Other Clerical and Administrative Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="591000">
        <xs:annotation>
          <xs:documentation>Logistics Clerks nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.591000</lx:path>
            <lx:label>Logistics Clerks nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="591100">
        <xs:annotation>
          <xs:documentation>Purchasing and Supply Logistics Clerks nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.591100</lx:path>
            <lx:label>Purchasing and Supply Logistics Clerks nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="591112">
        <xs:annotation>
          <xs:documentation>Production Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.591112</lx:path>
            <lx:label>Production Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="591113">
        <xs:annotation>
          <xs:documentation>Purchasing Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.591113</lx:path>
            <lx:label>Purchasing Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="591115">
        <xs:annotation>
          <xs:documentation>Stock Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.591115</lx:path>
            <lx:label>Stock Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="591116">
        <xs:annotation>
          <xs:documentation>Warehouse Administrator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.591116</lx:path>
            <lx:label>Warehouse Administrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="591117">
        <xs:annotation>
          <xs:documentation>Order Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.591117</lx:path>
            <lx:label>Order Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="591200">
        <xs:annotation>
          <xs:documentation>Transport and Despatch Clerks nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.591200</lx:path>
            <lx:label>Transport and Despatch Clerks nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="591211">
        <xs:annotation>
          <xs:documentation>Despatching and Receiving Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.591211</lx:path>
            <lx:label>Despatching and Receiving Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="591212">
        <xs:annotation>
          <xs:documentation>Import-Export Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.591212</lx:path>
            <lx:label>Import-Export Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599000">
        <xs:annotation>
          <xs:documentation>Miscellaneous Clerical and Administrative Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599000</lx:path>
            <lx:label>Miscellaneous Clerical and Administrative Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599100">
        <xs:annotation>
          <xs:documentation>Conveyancers and Legal Executives nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599100</lx:path>
            <lx:label>Conveyancers and Legal Executives nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599111">
        <xs:annotation>
          <xs:documentation>Conveyancer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599111</lx:path>
            <lx:label>Conveyancer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599112">
        <xs:annotation>
          <xs:documentation>Legal Executive</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599112</lx:path>
            <lx:label>Legal Executive</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599200">
        <xs:annotation>
          <xs:documentation>Court and Legal Clerks nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599200</lx:path>
            <lx:label>Court and Legal Clerks nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599211">
        <xs:annotation>
          <xs:documentation>Clerk of Court</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599211</lx:path>
            <lx:label>Clerk of Court</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599212">
        <xs:annotation>
          <xs:documentation>Court Bailiff or Sheriff</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599212</lx:path>
            <lx:label>Court Bailiff or Sheriff</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599213">
        <xs:annotation>
          <xs:documentation>Court Orderly</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599213</lx:path>
            <lx:label>Court Orderly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599214">
        <xs:annotation>
          <xs:documentation>Law Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599214</lx:path>
            <lx:label>Law Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599215">
        <xs:annotation>
          <xs:documentation>Trust Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599215</lx:path>
            <lx:label>Trust Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599311">
        <xs:annotation>
          <xs:documentation>Debt Collector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599311</lx:path>
            <lx:label>Debt Collector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599411">
        <xs:annotation>
          <xs:documentation>Human Resources Clerk</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599411</lx:path>
            <lx:label>Human Resources Clerk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599500">
        <xs:annotation>
          <xs:documentation>Inspectors and Regulatory Officers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599500</lx:path>
            <lx:label>Inspectors and Regulatory Officers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599511">
        <xs:annotation>
          <xs:documentation>Customs Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599511</lx:path>
            <lx:label>Customs Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599512">
        <xs:annotation>
          <xs:documentation>Immigration Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599512</lx:path>
            <lx:label>Immigration Officer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599513">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Licence Examiner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599513</lx:path>
            <lx:label>Motor Vehicle Licence Examiner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599514">
        <xs:annotation>
          <xs:documentation>Noxious Weeds and Pest Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599514</lx:path>
            <lx:label>Noxious Weeds and Pest Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599515">
        <xs:annotation>
          <xs:documentation>Social Security Assessor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599515</lx:path>
            <lx:label>Social Security Assessor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599516">
        <xs:annotation>
          <xs:documentation>Taxation Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599516</lx:path>
            <lx:label>Taxation Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599517">
        <xs:annotation>
          <xs:documentation>Train Examiner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599517</lx:path>
            <lx:label>Train Examiner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599518">
        <xs:annotation>
          <xs:documentation>Transport Operations Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599518</lx:path>
            <lx:label>Transport Operations Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599521">
        <xs:annotation>
          <xs:documentation>Water Inspector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599521</lx:path>
            <lx:label>Water Inspector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599599">
        <xs:annotation>
          <xs:documentation>Inspectors and Regulatory Officers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599599</lx:path>
            <lx:label>Inspectors and Regulatory Officers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599600">
        <xs:annotation>
          <xs:documentation>Insurance Investigators, Loss Adjusters and Risk Surveyors nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599600</lx:path>
            <lx:label>Insurance Investigators, Loss Adjusters and Risk Surveyors nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599611">
        <xs:annotation>
          <xs:documentation>Insurance Investigator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599611</lx:path>
            <lx:label>Insurance Investigator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599612">
        <xs:annotation>
          <xs:documentation>Insurance Loss Adjuster</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599612</lx:path>
            <lx:label>Insurance Loss Adjuster</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599613">
        <xs:annotation>
          <xs:documentation>Insurance Risk Surveyor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599613</lx:path>
            <lx:label>Insurance Risk Surveyor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599711">
        <xs:annotation>
          <xs:documentation>Library Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599711</lx:path>
            <lx:label>Library Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599900">
        <xs:annotation>
          <xs:documentation>Other Miscellaneous Clerical and Administrative Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599900</lx:path>
            <lx:label>Other Miscellaneous Clerical and Administrative Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599912">
        <xs:annotation>
          <xs:documentation>Production Assistant (Film, Television, Radio or Stage)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599912</lx:path>
            <lx:label>Production Assistant (Film, Television, Radio or Stage)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599913">
        <xs:annotation>
          <xs:documentation>Proof Reader</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599913</lx:path>
            <lx:label>Proof Reader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599914">
        <xs:annotation>
          <xs:documentation>Radio Despatcher</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599914</lx:path>
            <lx:label>Radio Despatcher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599915">
        <xs:annotation>
          <xs:documentation>Clinical Coder</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599915</lx:path>
            <lx:label>Clinical Coder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599916">
        <xs:annotation>
          <xs:documentation>Facilities Administrator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599916</lx:path>
            <lx:label>Facilities Administrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="599999">
        <xs:annotation>
          <xs:documentation>Clerical and Administrative Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.599999</lx:path>
            <lx:label>Clerical and Administrative Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="600000">
        <xs:annotation>
          <xs:documentation>Sales Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.600000</lx:path>
            <lx:label>Sales Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="610000">
        <xs:annotation>
          <xs:documentation>Sales Representatives and Agents nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.610000</lx:path>
            <lx:label>Sales Representatives and Agents nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="611000">
        <xs:annotation>
          <xs:documentation>Insurance Agents and Sales Representatives nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.611000</lx:path>
            <lx:label>Insurance Agents and Sales Representatives nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="611100">
        <xs:annotation>
          <xs:documentation>Auctioneers, and Stock and Station Agents nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.611100</lx:path>
            <lx:label>Auctioneers, and Stock and Station Agents nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="611111">
        <xs:annotation>
          <xs:documentation>Auctioneer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.611111</lx:path>
            <lx:label>Auctioneer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="611112">
        <xs:annotation>
          <xs:documentation>Stock and Station Agent</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.611112</lx:path>
            <lx:label>Stock and Station Agent</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="611211">
        <xs:annotation>
          <xs:documentation>Insurance Agent</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.611211</lx:path>
            <lx:label>Insurance Agent</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="611300">
        <xs:annotation>
          <xs:documentation>Sales Representatives nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.611300</lx:path>
            <lx:label>Sales Representatives nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="611311">
        <xs:annotation>
          <xs:documentation>Sales Representative (Building and Plumbing Supplies)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.611311</lx:path>
            <lx:label>Sales Representative (Building and Plumbing Supplies)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="611312">
        <xs:annotation>
          <xs:documentation>Sales Representative (Business Services)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.611312</lx:path>
            <lx:label>Sales Representative (Business Services)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="611313">
        <xs:annotation>
          <xs:documentation>Sales Representative (Motor Vehicle Parts and Accessories)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.611313</lx:path>
            <lx:label>Sales Representative (Motor Vehicle Parts and Accessories)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="611314">
        <xs:annotation>
          <xs:documentation>Sales Representative (Personal and Household Goods)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.611314</lx:path>
            <lx:label>Sales Representative (Personal and Household Goods)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="611399">
        <xs:annotation>
          <xs:documentation>Sales Representatives nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.611399</lx:path>
            <lx:label>Sales Representatives nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="612100">
        <xs:annotation>
          <xs:documentation>Real Estate Sales Agents nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.612100</lx:path>
            <lx:label>Real Estate Sales Agents nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="612111">
        <xs:annotation>
          <xs:documentation>Business Broker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.612111</lx:path>
            <lx:label>Business Broker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="612112">
        <xs:annotation>
          <xs:documentation>Property Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.612112</lx:path>
            <lx:label>Property Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="612113">
        <xs:annotation>
          <xs:documentation>Real Estate Agency Principal</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.612113</lx:path>
            <lx:label>Real Estate Agency Principal</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="612114">
        <xs:annotation>
          <xs:documentation>Real Estate Agent</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.612114</lx:path>
            <lx:label>Real Estate Agent</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="612115">
        <xs:annotation>
          <xs:documentation>Real Estate Representative</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.612115</lx:path>
            <lx:label>Real Estate Representative</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621000">
        <xs:annotation>
          <xs:documentation>Sales Assistants and Salespersons nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621000</lx:path>
            <lx:label>Sales Assistants and Salespersons nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621111">
        <xs:annotation>
          <xs:documentation>Sales Assistant (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621111</lx:path>
            <lx:label>Sales Assistant (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621211">
        <xs:annotation>
          <xs:documentation>ICT Sales Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621211</lx:path>
            <lx:label>ICT Sales Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621300">
        <xs:annotation>
          <xs:documentation>Motor Vehicle and Vehicle Parts Salespersons nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621300</lx:path>
            <lx:label>Motor Vehicle and Vehicle Parts Salespersons nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621311">
        <xs:annotation>
          <xs:documentation>Motor Vehicle or Caravan Salesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621311</lx:path>
            <lx:label>Motor Vehicle or Caravan Salesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621312">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Parts Interpreter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621312</lx:path>
            <lx:label>Motor Vehicle Parts Interpreter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621411">
        <xs:annotation>
          <xs:documentation>Pharmacy Sales Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621411</lx:path>
            <lx:label>Pharmacy Sales Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621511">
        <xs:annotation>
          <xs:documentation>Retail Supervisor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621511</lx:path>
            <lx:label>Retail Supervisor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621611">
        <xs:annotation>
          <xs:documentation>Service Station Attendant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621611</lx:path>
            <lx:label>Service Station Attendant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621700">
        <xs:annotation>
          <xs:documentation>Street Vendors and Related Salespersons nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621700</lx:path>
            <lx:label>Street Vendors and Related Salespersons nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621711">
        <xs:annotation>
          <xs:documentation>Cash Van Salesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621711</lx:path>
            <lx:label>Cash Van Salesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621712">
        <xs:annotation>
          <xs:documentation>Door-to-door Salesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621712</lx:path>
            <lx:label>Door-to-door Salesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621713">
        <xs:annotation>
          <xs:documentation>Street Vendor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621713</lx:path>
            <lx:label>Street Vendor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621900">
        <xs:annotation>
          <xs:documentation>Other Sales Assistants and Salespersons nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621900</lx:path>
            <lx:label>Other Sales Assistants and Salespersons nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621911">
        <xs:annotation>
          <xs:documentation>Materials Recycler</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621911</lx:path>
            <lx:label>Materials Recycler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621912">
        <xs:annotation>
          <xs:documentation>Rental Salesperson</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621912</lx:path>
            <lx:label>Rental Salesperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="621999">
        <xs:annotation>
          <xs:documentation>Sales Assistants and Salespersons nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.621999</lx:path>
            <lx:label>Sales Assistants and Salespersons nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="630000">
        <xs:annotation>
          <xs:documentation>Sales Support Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.630000</lx:path>
            <lx:label>Sales Support Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="631100">
        <xs:annotation>
          <xs:documentation>Checkout Operators and Office Cashiers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.631100</lx:path>
            <lx:label>Checkout Operators and Office Cashiers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="631111">
        <xs:annotation>
          <xs:documentation>Checkout Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.631111</lx:path>
            <lx:label>Checkout Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="631112">
        <xs:annotation>
          <xs:documentation>Office Cashier</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.631112</lx:path>
            <lx:label>Office Cashier</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="639000">
        <xs:annotation>
          <xs:documentation>Miscellaneous Sales Support Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.639000</lx:path>
            <lx:label>Miscellaneous Sales Support Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="639100">
        <xs:annotation>
          <xs:documentation>Models and Sales Demonstrators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.639100</lx:path>
            <lx:label>Models and Sales Demonstrators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="639111">
        <xs:annotation>
          <xs:documentation>Model</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.639111</lx:path>
            <lx:label>Model</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="639112">
        <xs:annotation>
          <xs:documentation>Sales Demonstrator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.639112</lx:path>
            <lx:label>Sales Demonstrator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="639200">
        <xs:annotation>
          <xs:documentation>Retail and Wool Buyers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.639200</lx:path>
            <lx:label>Retail and Wool Buyers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="639211">
        <xs:annotation>
          <xs:documentation>Retail Buyer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.639211</lx:path>
            <lx:label>Retail Buyer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="639212">
        <xs:annotation>
          <xs:documentation>Wool Buyer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.639212</lx:path>
            <lx:label>Wool Buyer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="639311">
        <xs:annotation>
          <xs:documentation>Telemarketer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.639311</lx:path>
            <lx:label>Telemarketer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="639400">
        <xs:annotation>
          <xs:documentation>Ticket Salespersons nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.639400</lx:path>
            <lx:label>Ticket Salespersons nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="639411">
        <xs:annotation>
          <xs:documentation>Ticket Seller</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.639411</lx:path>
            <lx:label>Ticket Seller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="639412">
        <xs:annotation>
          <xs:documentation>Transport Conductor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.639412</lx:path>
            <lx:label>Transport Conductor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="639511">
        <xs:annotation>
          <xs:documentation>Visual Merchandiser</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.639511</lx:path>
            <lx:label>Visual Merchandiser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="639911">
        <xs:annotation>
          <xs:documentation>Other Sales Support Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.639911</lx:path>
            <lx:label>Other Sales Support Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="700000">
        <xs:annotation>
          <xs:documentation>Machinery Operators and Drivers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.700000</lx:path>
            <lx:label>Machinery Operators and Drivers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="710000">
        <xs:annotation>
          <xs:documentation>Machine and Stationary Plant Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.710000</lx:path>
            <lx:label>Machine and Stationary Plant Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711000">
        <xs:annotation>
          <xs:documentation>Machine Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711000</lx:path>
            <lx:label>Machine Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711100">
        <xs:annotation>
          <xs:documentation>Clay, Concrete, Glass and Stone Processing Machine Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711100</lx:path>
            <lx:label>Clay, Concrete, Glass and Stone Processing Machine Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711111">
        <xs:annotation>
          <xs:documentation>Clay Products Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711111</lx:path>
            <lx:label>Clay Products Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711112">
        <xs:annotation>
          <xs:documentation>Concrete Products Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711112</lx:path>
            <lx:label>Concrete Products Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711113">
        <xs:annotation>
          <xs:documentation>Glass Production Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711113</lx:path>
            <lx:label>Glass Production Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711114">
        <xs:annotation>
          <xs:documentation>Stone Processing Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711114</lx:path>
            <lx:label>Stone Processing Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711199">
        <xs:annotation>
          <xs:documentation>Clay, Concrete, Glass and Stone Processing Machine Operators nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711199</lx:path>
            <lx:label>Clay, Concrete, Glass and Stone Processing Machine Operators nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711211">
        <xs:annotation>
          <xs:documentation>Industrial Spraypainter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711211</lx:path>
            <lx:label>Industrial Spraypainter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711300">
        <xs:annotation>
          <xs:documentation>Paper and Wood Processing Machine Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711300</lx:path>
            <lx:label>Paper and Wood Processing Machine Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711311">
        <xs:annotation>
          <xs:documentation>Paper Products Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711311</lx:path>
            <lx:label>Paper Products Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711313">
        <xs:annotation>
          <xs:documentation>Sawmilling Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711313</lx:path>
            <lx:label>Sawmilling Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711314">
        <xs:annotation>
          <xs:documentation>Other Wood Processing Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711314</lx:path>
            <lx:label>Other Wood Processing Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711411">
        <xs:annotation>
          <xs:documentation>Photographic Developer and Printer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711411</lx:path>
            <lx:label>Photographic Developer and Printer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711500">
        <xs:annotation>
          <xs:documentation>Plastics and Rubber Production Machine Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711500</lx:path>
            <lx:label>Plastics and Rubber Production Machine Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711511">
        <xs:annotation>
          <xs:documentation>Plastic Cablemaking Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711511</lx:path>
            <lx:label>Plastic Cablemaking Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711512">
        <xs:annotation>
          <xs:documentation>Plastic Compounding and Reclamation Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711512</lx:path>
            <lx:label>Plastic Compounding and Reclamation Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711513">
        <xs:annotation>
          <xs:documentation>Plastics Fabricator or Welder</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711513</lx:path>
            <lx:label>Plastics Fabricator or Welder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711514">
        <xs:annotation>
          <xs:documentation>Plastics Production Machine Operator (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711514</lx:path>
            <lx:label>Plastics Production Machine Operator (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711515">
        <xs:annotation>
          <xs:documentation>Reinforced Plastic and Composite Production Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711515</lx:path>
            <lx:label>Reinforced Plastic and Composite Production Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711516">
        <xs:annotation>
          <xs:documentation>Rubber Production Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711516</lx:path>
            <lx:label>Rubber Production Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711599">
        <xs:annotation>
          <xs:documentation>Plastics and Rubber Production Machine Operators nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711599</lx:path>
            <lx:label>Plastics and Rubber Production Machine Operators nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711611">
        <xs:annotation>
          <xs:documentation>Sewing Machinist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711611</lx:path>
            <lx:label>Sewing Machinist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711700">
        <xs:annotation>
          <xs:documentation>Textile and Footwear Production Machine Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711700</lx:path>
            <lx:label>Textile and Footwear Production Machine Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711711">
        <xs:annotation>
          <xs:documentation>Footwear Production Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711711</lx:path>
            <lx:label>Footwear Production Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711712">
        <xs:annotation>
          <xs:documentation>Hide and Skin Processing Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711712</lx:path>
            <lx:label>Hide and Skin Processing Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711713">
        <xs:annotation>
          <xs:documentation>Knitting Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711713</lx:path>
            <lx:label>Knitting Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711714">
        <xs:annotation>
          <xs:documentation>Textile Dyeing and Finishing Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711714</lx:path>
            <lx:label>Textile Dyeing and Finishing Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711715">
        <xs:annotation>
          <xs:documentation>Weaving Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711715</lx:path>
            <lx:label>Weaving Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711716">
        <xs:annotation>
          <xs:documentation>Yarn Carding and Spinning Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711716</lx:path>
            <lx:label>Yarn Carding and Spinning Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711799">
        <xs:annotation>
          <xs:documentation>Textile and Footwear Production Machine Operators nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711799</lx:path>
            <lx:label>Textile and Footwear Production Machine Operators nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711900">
        <xs:annotation>
          <xs:documentation>Other Machine Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711900</lx:path>
            <lx:label>Other Machine Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711911">
        <xs:annotation>
          <xs:documentation>Chemical Production Machine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711911</lx:path>
            <lx:label>Chemical Production Machine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711912">
        <xs:annotation>
          <xs:documentation>Motion Picture Projectionist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711912</lx:path>
            <lx:label>Motion Picture Projectionist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711913">
        <xs:annotation>
          <xs:documentation>Sand Blaster</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711913</lx:path>
            <lx:label>Sand Blaster</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711914">
        <xs:annotation>
          <xs:documentation>Sterilisation Technician</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711914</lx:path>
            <lx:label>Sterilisation Technician</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="711999">
        <xs:annotation>
          <xs:documentation>Machine Operators nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.711999</lx:path>
            <lx:label>Machine Operators nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712000">
        <xs:annotation>
          <xs:documentation>Stationary Plant Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712000</lx:path>
            <lx:label>Stationary Plant Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712111">
        <xs:annotation>
          <xs:documentation>Crane, Hoist or Lift Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712111</lx:path>
            <lx:label>Crane, Hoist or Lift Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712200">
        <xs:annotation>
          <xs:documentation>Drillers, Miners and Shot Firers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712200</lx:path>
            <lx:label>Drillers, Miners and Shot Firers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712211">
        <xs:annotation>
          <xs:documentation>Driller</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712211</lx:path>
            <lx:label>Driller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712212">
        <xs:annotation>
          <xs:documentation>Miner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712212</lx:path>
            <lx:label>Miner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712213">
        <xs:annotation>
          <xs:documentation>Shot Firer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712213</lx:path>
            <lx:label>Shot Firer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712311">
        <xs:annotation>
          <xs:documentation>Engineering Production Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712311</lx:path>
            <lx:label>Engineering Production Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712900">
        <xs:annotation>
          <xs:documentation>Other Stationary Plant Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712900</lx:path>
            <lx:label>Other Stationary Plant Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712911">
        <xs:annotation>
          <xs:documentation>Boiler or Engine Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712911</lx:path>
            <lx:label>Boiler or Engine Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712912">
        <xs:annotation>
          <xs:documentation>Bulk Materials Handling Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712912</lx:path>
            <lx:label>Bulk Materials Handling Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712913">
        <xs:annotation>
          <xs:documentation>Cement Production Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712913</lx:path>
            <lx:label>Cement Production Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712914">
        <xs:annotation>
          <xs:documentation>Concrete Batching Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712914</lx:path>
            <lx:label>Concrete Batching Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712915">
        <xs:annotation>
          <xs:documentation>Concrete Pump Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712915</lx:path>
            <lx:label>Concrete Pump Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712916">
        <xs:annotation>
          <xs:documentation>Paper and Pulp Mill Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712916</lx:path>
            <lx:label>Paper and Pulp Mill Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712917">
        <xs:annotation>
          <xs:documentation>Railway Signal Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712917</lx:path>
            <lx:label>Railway Signal Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712918">
        <xs:annotation>
          <xs:documentation>Train Controller</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712918</lx:path>
            <lx:label>Train Controller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712921">
        <xs:annotation>
          <xs:documentation>Waste Water or Water Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712921</lx:path>
            <lx:label>Waste Water or Water Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712922">
        <xs:annotation>
          <xs:documentation>Weighbridge Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712922</lx:path>
            <lx:label>Weighbridge Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="712999">
        <xs:annotation>
          <xs:documentation>Stationary Plant Operators nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.712999</lx:path>
            <lx:label>Stationary Plant Operators nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721000">
        <xs:annotation>
          <xs:documentation>Mobile Plant Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721000</lx:path>
            <lx:label>Mobile Plant Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721100">
        <xs:annotation>
          <xs:documentation>Agricultural, Forestry and Horticultural Plant Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721100</lx:path>
            <lx:label>Agricultural, Forestry and Horticultural Plant Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721111">
        <xs:annotation>
          <xs:documentation>Agricultural and Horticultural Mobile Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721111</lx:path>
            <lx:label>Agricultural and Horticultural Mobile Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721112">
        <xs:annotation>
          <xs:documentation>Logging Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721112</lx:path>
            <lx:label>Logging Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721200">
        <xs:annotation>
          <xs:documentation>Earthmoving Plant Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721200</lx:path>
            <lx:label>Earthmoving Plant Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721211">
        <xs:annotation>
          <xs:documentation>Earthmoving Plant Operator (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721211</lx:path>
            <lx:label>Earthmoving Plant Operator (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721212">
        <xs:annotation>
          <xs:documentation>Backhoe Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721212</lx:path>
            <lx:label>Backhoe Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721213">
        <xs:annotation>
          <xs:documentation>Bulldozer Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721213</lx:path>
            <lx:label>Bulldozer Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721214">
        <xs:annotation>
          <xs:documentation>Excavator Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721214</lx:path>
            <lx:label>Excavator Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721215">
        <xs:annotation>
          <xs:documentation>Grader Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721215</lx:path>
            <lx:label>Grader Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721216">
        <xs:annotation>
          <xs:documentation>Loader Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721216</lx:path>
            <lx:label>Loader Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721311">
        <xs:annotation>
          <xs:documentation>Forklift Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721311</lx:path>
            <lx:label>Forklift Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721900">
        <xs:annotation>
          <xs:documentation>Other Mobile Plant Operators nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721900</lx:path>
            <lx:label>Other Mobile Plant Operators nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721911">
        <xs:annotation>
          <xs:documentation>Aircraft Baggage Handler and Airline Ground Crew</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721911</lx:path>
            <lx:label>Aircraft Baggage Handler and Airline Ground Crew</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721912">
        <xs:annotation>
          <xs:documentation>Linemarker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721912</lx:path>
            <lx:label>Linemarker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721913">
        <xs:annotation>
          <xs:documentation>Paving Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721913</lx:path>
            <lx:label>Paving Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721914">
        <xs:annotation>
          <xs:documentation>Railway Track Plant Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721914</lx:path>
            <lx:label>Railway Track Plant Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721915">
        <xs:annotation>
          <xs:documentation>Road Roller Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721915</lx:path>
            <lx:label>Road Roller Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721916">
        <xs:annotation>
          <xs:documentation>Streetsweeper Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721916</lx:path>
            <lx:label>Streetsweeper Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="721999">
        <xs:annotation>
          <xs:documentation>Mobile Plant Operators nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.721999</lx:path>
            <lx:label>Mobile Plant Operators nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="730000">
        <xs:annotation>
          <xs:documentation>Road and Rail Drivers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.730000</lx:path>
            <lx:label>Road and Rail Drivers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="731000">
        <xs:annotation>
          <xs:documentation>Automobile, Bus and Rail Drivers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.731000</lx:path>
            <lx:label>Automobile, Bus and Rail Drivers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="731100">
        <xs:annotation>
          <xs:documentation>Automobile Drivers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.731100</lx:path>
            <lx:label>Automobile Drivers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="731111">
        <xs:annotation>
          <xs:documentation>Chauffeur</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.731111</lx:path>
            <lx:label>Chauffeur</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="731112">
        <xs:annotation>
          <xs:documentation>Taxi Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.731112</lx:path>
            <lx:label>Taxi Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="731199">
        <xs:annotation>
          <xs:documentation>Automobile Drivers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.731199</lx:path>
            <lx:label>Automobile Drivers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="731200">
        <xs:annotation>
          <xs:documentation>Bus and Coach Drivers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.731200</lx:path>
            <lx:label>Bus and Coach Drivers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="731211">
        <xs:annotation>
          <xs:documentation>Bus Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.731211</lx:path>
            <lx:label>Bus Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="731212">
        <xs:annotation>
          <xs:documentation>Charter and Tour Bus Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.731212</lx:path>
            <lx:label>Charter and Tour Bus Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="731213">
        <xs:annotation>
          <xs:documentation>Passenger Coach Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.731213</lx:path>
            <lx:label>Passenger Coach Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="731300">
        <xs:annotation>
          <xs:documentation>Train and Tram Drivers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.731300</lx:path>
            <lx:label>Train and Tram Drivers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="731311">
        <xs:annotation>
          <xs:documentation>Train Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.731311</lx:path>
            <lx:label>Train Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="731312">
        <xs:annotation>
          <xs:documentation>Tram Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.731312</lx:path>
            <lx:label>Tram Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="732111">
        <xs:annotation>
          <xs:documentation>Delivery Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.732111</lx:path>
            <lx:label>Delivery Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="733100">
        <xs:annotation>
          <xs:documentation>Truck Drivers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.733100</lx:path>
            <lx:label>Truck Drivers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="733111">
        <xs:annotation>
          <xs:documentation>Truck Driver (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.733111</lx:path>
            <lx:label>Truck Driver (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="733112">
        <xs:annotation>
          <xs:documentation>Aircraft Refueller</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.733112</lx:path>
            <lx:label>Aircraft Refueller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="733113">
        <xs:annotation>
          <xs:documentation>Furniture Removalist</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.733113</lx:path>
            <lx:label>Furniture Removalist</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="733114">
        <xs:annotation>
          <xs:documentation>Tanker Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.733114</lx:path>
            <lx:label>Tanker Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="733115">
        <xs:annotation>
          <xs:documentation>Tow Truck Driver</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.733115</lx:path>
            <lx:label>Tow Truck Driver</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="741111">
        <xs:annotation>
          <xs:documentation>Storeperson</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.741111</lx:path>
            <lx:label>Storeperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="800000">
        <xs:annotation>
          <xs:documentation>Labourers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.800000</lx:path>
            <lx:label>Labourers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811000">
        <xs:annotation>
          <xs:documentation>Cleaners and Laundry Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811000</lx:path>
            <lx:label>Cleaners and Laundry Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811111">
        <xs:annotation>
          <xs:documentation>Car Detailer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811111</lx:path>
            <lx:label>Car Detailer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811211">
        <xs:annotation>
          <xs:documentation>Commercial Cleaner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811211</lx:path>
            <lx:label>Commercial Cleaner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811311">
        <xs:annotation>
          <xs:documentation>Domestic Cleaner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811311</lx:path>
            <lx:label>Domestic Cleaner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811400">
        <xs:annotation>
          <xs:documentation>Housekeepers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811400</lx:path>
            <lx:label>Housekeepers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811411">
        <xs:annotation>
          <xs:documentation>Commercial Housekeeper</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811411</lx:path>
            <lx:label>Commercial Housekeeper</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811412">
        <xs:annotation>
          <xs:documentation>Domestic Housekeeper</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811412</lx:path>
            <lx:label>Domestic Housekeeper</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811500">
        <xs:annotation>
          <xs:documentation>Laundry Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811500</lx:path>
            <lx:label>Laundry Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811511">
        <xs:annotation>
          <xs:documentation>Laundry Worker (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811511</lx:path>
            <lx:label>Laundry Worker (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811512">
        <xs:annotation>
          <xs:documentation>Drycleaner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811512</lx:path>
            <lx:label>Drycleaner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811513">
        <xs:annotation>
          <xs:documentation>Ironer or Presser</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811513</lx:path>
            <lx:label>Ironer or Presser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811600">
        <xs:annotation>
          <xs:documentation>Other Cleaners nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811600</lx:path>
            <lx:label>Other Cleaners nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811611">
        <xs:annotation>
          <xs:documentation>Carpet Cleaner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811611</lx:path>
            <lx:label>Carpet Cleaner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811612">
        <xs:annotation>
          <xs:documentation>Window Cleaner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811612</lx:path>
            <lx:label>Window Cleaner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="811699">
        <xs:annotation>
          <xs:documentation>Cleaners nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.811699</lx:path>
            <lx:label>Cleaners nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821000">
        <xs:annotation>
          <xs:documentation>Construction and Mining Labourers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821000</lx:path>
            <lx:label>Construction and Mining Labourers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821100">
        <xs:annotation>
          <xs:documentation>Building and Plumbing Labourers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821100</lx:path>
            <lx:label>Building and Plumbing Labourers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821111">
        <xs:annotation>
          <xs:documentation>Builder's Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821111</lx:path>
            <lx:label>Builder's Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821112">
        <xs:annotation>
          <xs:documentation>Drainage, Sewerage and Stormwater Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821112</lx:path>
            <lx:label>Drainage, Sewerage and Stormwater Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821113">
        <xs:annotation>
          <xs:documentation>Earthmoving Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821113</lx:path>
            <lx:label>Earthmoving Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821114">
        <xs:annotation>
          <xs:documentation>Plumber's Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821114</lx:path>
            <lx:label>Plumber's Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821211">
        <xs:annotation>
          <xs:documentation>Concreter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821211</lx:path>
            <lx:label>Concreter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821311">
        <xs:annotation>
          <xs:documentation>Fencer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821311</lx:path>
            <lx:label>Fencer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821400">
        <xs:annotation>
          <xs:documentation>Insulation and Home Improvement Installers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821400</lx:path>
            <lx:label>Insulation and Home Improvement Installers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821411">
        <xs:annotation>
          <xs:documentation>Building Insulation Installer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821411</lx:path>
            <lx:label>Building Insulation Installer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821412">
        <xs:annotation>
          <xs:documentation>Home Improvement Installer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821412</lx:path>
            <lx:label>Home Improvement Installer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821511">
        <xs:annotation>
          <xs:documentation>Paving and Surfacing Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821511</lx:path>
            <lx:label>Paving and Surfacing Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821611">
        <xs:annotation>
          <xs:documentation>Railway Track Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821611</lx:path>
            <lx:label>Railway Track Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821700">
        <xs:annotation>
          <xs:documentation>Structural Steel Construction Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821700</lx:path>
            <lx:label>Structural Steel Construction Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821711">
        <xs:annotation>
          <xs:documentation>Construction Rigger</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821711</lx:path>
            <lx:label>Construction Rigger</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821712">
        <xs:annotation>
          <xs:documentation>Scaffolder</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821712</lx:path>
            <lx:label>Scaffolder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821713">
        <xs:annotation>
          <xs:documentation>Steel Fixer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821713</lx:path>
            <lx:label>Steel Fixer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821714">
        <xs:annotation>
          <xs:documentation>Structural Steel Erector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821714</lx:path>
            <lx:label>Structural Steel Erector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821900">
        <xs:annotation>
          <xs:documentation>Other Construction and Mining Labourers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821900</lx:path>
            <lx:label>Other Construction and Mining Labourers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821911">
        <xs:annotation>
          <xs:documentation>Crane Chaser</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821911</lx:path>
            <lx:label>Crane Chaser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821912">
        <xs:annotation>
          <xs:documentation>Driller's Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821912</lx:path>
            <lx:label>Driller's Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821913">
        <xs:annotation>
          <xs:documentation>Lagger</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821913</lx:path>
            <lx:label>Lagger</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821914">
        <xs:annotation>
          <xs:documentation>Mining Support Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821914</lx:path>
            <lx:label>Mining Support Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="821915">
        <xs:annotation>
          <xs:documentation>Surveyor's Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.821915</lx:path>
            <lx:label>Surveyor's Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="830000">
        <xs:annotation>
          <xs:documentation>Factory Process Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.830000</lx:path>
            <lx:label>Factory Process Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831000">
        <xs:annotation>
          <xs:documentation>Food Process Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831000</lx:path>
            <lx:label>Food Process Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831100">
        <xs:annotation>
          <xs:documentation>Food and Drink Factory Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831100</lx:path>
            <lx:label>Food and Drink Factory Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831111">
        <xs:annotation>
          <xs:documentation>Baking Factory Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831111</lx:path>
            <lx:label>Baking Factory Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831112">
        <xs:annotation>
          <xs:documentation>Brewery Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831112</lx:path>
            <lx:label>Brewery Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831113">
        <xs:annotation>
          <xs:documentation>Confectionery Maker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831113</lx:path>
            <lx:label>Confectionery Maker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831114">
        <xs:annotation>
          <xs:documentation>Dairy Products Maker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831114</lx:path>
            <lx:label>Dairy Products Maker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831115">
        <xs:annotation>
          <xs:documentation>Fruit and Vegetable Factory Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831115</lx:path>
            <lx:label>Fruit and Vegetable Factory Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831116">
        <xs:annotation>
          <xs:documentation>Grain Mill Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831116</lx:path>
            <lx:label>Grain Mill Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831117">
        <xs:annotation>
          <xs:documentation>Sugar Mill Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831117</lx:path>
            <lx:label>Sugar Mill Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831118">
        <xs:annotation>
          <xs:documentation>Winery Cellar Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831118</lx:path>
            <lx:label>Winery Cellar Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831199">
        <xs:annotation>
          <xs:documentation>Food and Drink Factory Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831199</lx:path>
            <lx:label>Food and Drink Factory Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831200">
        <xs:annotation>
          <xs:documentation>Meat Boners and Slicers, and Slaughterers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831200</lx:path>
            <lx:label>Meat Boners and Slicers, and Slaughterers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831211">
        <xs:annotation>
          <xs:documentation>Meat Boner and Slicer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831211</lx:path>
            <lx:label>Meat Boner and Slicer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831212">
        <xs:annotation>
          <xs:documentation>Slaughterer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831212</lx:path>
            <lx:label>Slaughterer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831300">
        <xs:annotation>
          <xs:documentation>Meat, Poultry and Seafood Process Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831300</lx:path>
            <lx:label>Meat, Poultry and Seafood Process Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831311">
        <xs:annotation>
          <xs:documentation>Meat Process Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831311</lx:path>
            <lx:label>Meat Process Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831312">
        <xs:annotation>
          <xs:documentation>Poultry Process Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831312</lx:path>
            <lx:label>Poultry Process Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="831313">
        <xs:annotation>
          <xs:documentation>Seafood Process Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.831313</lx:path>
            <lx:label>Seafood Process Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="832000">
        <xs:annotation>
          <xs:documentation>Packers and Product Assemblers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.832000</lx:path>
            <lx:label>Packers and Product Assemblers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="832100">
        <xs:annotation>
          <xs:documentation>Packers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.832100</lx:path>
            <lx:label>Packers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="832111">
        <xs:annotation>
          <xs:documentation>Chocolate Packer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.832111</lx:path>
            <lx:label>Chocolate Packer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="832112">
        <xs:annotation>
          <xs:documentation>Container Filler</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.832112</lx:path>
            <lx:label>Container Filler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="832113">
        <xs:annotation>
          <xs:documentation>Fruit and Vegetable Packer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.832113</lx:path>
            <lx:label>Fruit and Vegetable Packer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="832114">
        <xs:annotation>
          <xs:documentation>Meat Packer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.832114</lx:path>
            <lx:label>Meat Packer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="832115">
        <xs:annotation>
          <xs:documentation>Seafood Packer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.832115</lx:path>
            <lx:label>Seafood Packer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="832199">
        <xs:annotation>
          <xs:documentation>Packers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.832199</lx:path>
            <lx:label>Packers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="832211">
        <xs:annotation>
          <xs:documentation>Product Assembler</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.832211</lx:path>
            <lx:label>Product Assembler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839000">
        <xs:annotation>
          <xs:documentation>Miscellaneous Factory Process Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839000</lx:path>
            <lx:label>Miscellaneous Factory Process Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839111">
        <xs:annotation>
          <xs:documentation>Metal Engineering Process Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839111</lx:path>
            <lx:label>Metal Engineering Process Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839200">
        <xs:annotation>
          <xs:documentation>Plastics and Rubber Factory Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839200</lx:path>
            <lx:label>Plastics and Rubber Factory Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839211">
        <xs:annotation>
          <xs:documentation>Plastics Factory Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839211</lx:path>
            <lx:label>Plastics Factory Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839212">
        <xs:annotation>
          <xs:documentation>Rubber Factory Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839212</lx:path>
            <lx:label>Rubber Factory Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839300">
        <xs:annotation>
          <xs:documentation>Product Quality Controllers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839300</lx:path>
            <lx:label>Product Quality Controllers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839311">
        <xs:annotation>
          <xs:documentation>Product Examiner</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839311</lx:path>
            <lx:label>Product Examiner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839312">
        <xs:annotation>
          <xs:documentation>Product Grader</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839312</lx:path>
            <lx:label>Product Grader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839313">
        <xs:annotation>
          <xs:documentation>Product Tester</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839313</lx:path>
            <lx:label>Product Tester</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839400">
        <xs:annotation>
          <xs:documentation>Timber and Wood Process Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839400</lx:path>
            <lx:label>Timber and Wood Process Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839411">
        <xs:annotation>
          <xs:documentation>Paper and Pulp Mill Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839411</lx:path>
            <lx:label>Paper and Pulp Mill Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839412">
        <xs:annotation>
          <xs:documentation>Sawmill or Timber Yard Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839412</lx:path>
            <lx:label>Sawmill or Timber Yard Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839413">
        <xs:annotation>
          <xs:documentation>Wood and Wood Products Factory Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839413</lx:path>
            <lx:label>Wood and Wood Products Factory Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839900">
        <xs:annotation>
          <xs:documentation>Other Factory Process Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839900</lx:path>
            <lx:label>Other Factory Process Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839911">
        <xs:annotation>
          <xs:documentation>Cement and Concrete Plant Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839911</lx:path>
            <lx:label>Cement and Concrete Plant Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839912">
        <xs:annotation>
          <xs:documentation>Chemical Plant Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839912</lx:path>
            <lx:label>Chemical Plant Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839913">
        <xs:annotation>
          <xs:documentation>Clay Processing Factory Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839913</lx:path>
            <lx:label>Clay Processing Factory Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839914">
        <xs:annotation>
          <xs:documentation>Fabric and Textile Factory Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839914</lx:path>
            <lx:label>Fabric and Textile Factory Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839915">
        <xs:annotation>
          <xs:documentation>Footwear Factory Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839915</lx:path>
            <lx:label>Footwear Factory Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839916">
        <xs:annotation>
          <xs:documentation>Glass Processing Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839916</lx:path>
            <lx:label>Glass Processing Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839917">
        <xs:annotation>
          <xs:documentation>Hide and Skin Processing Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839917</lx:path>
            <lx:label>Hide and Skin Processing Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839918">
        <xs:annotation>
          <xs:documentation>Recycling Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839918</lx:path>
            <lx:label>Recycling Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="839999">
        <xs:annotation>
          <xs:documentation>Factory Process Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.839999</lx:path>
            <lx:label>Factory Process Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841000">
        <xs:annotation>
          <xs:documentation>Farm, Forestry and Garden Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841000</lx:path>
            <lx:label>Farm, Forestry and Garden Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841111">
        <xs:annotation>
          <xs:documentation>Aquaculture Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841111</lx:path>
            <lx:label>Aquaculture Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841200">
        <xs:annotation>
          <xs:documentation>Crop Farm Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841200</lx:path>
            <lx:label>Crop Farm Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841211">
        <xs:annotation>
          <xs:documentation>Fruit or Nut Farm Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841211</lx:path>
            <lx:label>Fruit or Nut Farm Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841212">
        <xs:annotation>
          <xs:documentation>Fruit or Nut Picker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841212</lx:path>
            <lx:label>Fruit or Nut Picker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841213">
        <xs:annotation>
          <xs:documentation>Grain, Oilseed or Pasture Farm Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841213</lx:path>
            <lx:label>Grain, Oilseed or Pasture Farm Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841214">
        <xs:annotation>
          <xs:documentation>Vegetable Farm Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841214</lx:path>
            <lx:label>Vegetable Farm Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841215">
        <xs:annotation>
          <xs:documentation>Vegetable Picker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841215</lx:path>
            <lx:label>Vegetable Picker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841216">
        <xs:annotation>
          <xs:documentation>Vineyard Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841216</lx:path>
            <lx:label>Vineyard Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841217">
        <xs:annotation>
          <xs:documentation>Mushroom Picker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841217</lx:path>
            <lx:label>Mushroom Picker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841299">
        <xs:annotation>
          <xs:documentation>Crop Farm Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841299</lx:path>
            <lx:label>Crop Farm Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841300">
        <xs:annotation>
          <xs:documentation>Forestry and Logging Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841300</lx:path>
            <lx:label>Forestry and Logging Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841311">
        <xs:annotation>
          <xs:documentation>Forestry Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841311</lx:path>
            <lx:label>Forestry Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841312">
        <xs:annotation>
          <xs:documentation>Logging Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841312</lx:path>
            <lx:label>Logging Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841313">
        <xs:annotation>
          <xs:documentation>Tree Faller</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841313</lx:path>
            <lx:label>Tree Faller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841400">
        <xs:annotation>
          <xs:documentation>Garden and Nursery Labourers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841400</lx:path>
            <lx:label>Garden and Nursery Labourers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841411">
        <xs:annotation>
          <xs:documentation>Garden Labourer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841411</lx:path>
            <lx:label>Garden Labourer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841412">
        <xs:annotation>
          <xs:documentation>Horticultural Nursery Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841412</lx:path>
            <lx:label>Horticultural Nursery Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841500">
        <xs:annotation>
          <xs:documentation>Livestock Farm Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841500</lx:path>
            <lx:label>Livestock Farm Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841511">
        <xs:annotation>
          <xs:documentation>Beef Cattle Farm Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841511</lx:path>
            <lx:label>Beef Cattle Farm Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841512">
        <xs:annotation>
          <xs:documentation>Dairy Cattle Farm Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841512</lx:path>
            <lx:label>Dairy Cattle Farm Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841513">
        <xs:annotation>
          <xs:documentation>Mixed Livestock Farm Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841513</lx:path>
            <lx:label>Mixed Livestock Farm Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841514">
        <xs:annotation>
          <xs:documentation>Poultry Farm Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841514</lx:path>
            <lx:label>Poultry Farm Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841515">
        <xs:annotation>
          <xs:documentation>Sheep Farm Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841515</lx:path>
            <lx:label>Sheep Farm Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841516">
        <xs:annotation>
          <xs:documentation>Stablehand</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841516</lx:path>
            <lx:label>Stablehand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841517">
        <xs:annotation>
          <xs:documentation>Wool Handler</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841517</lx:path>
            <lx:label>Wool Handler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841599">
        <xs:annotation>
          <xs:documentation>Livestock Farm Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841599</lx:path>
            <lx:label>Livestock Farm Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841611">
        <xs:annotation>
          <xs:documentation>Mixed Crop and Livestock Farm Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841611</lx:path>
            <lx:label>Mixed Crop and Livestock Farm Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841900">
        <xs:annotation>
          <xs:documentation>Other Farm, Forestry and Garden Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841900</lx:path>
            <lx:label>Other Farm, Forestry and Garden Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841911">
        <xs:annotation>
          <xs:documentation>Hunter-Trapper</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841911</lx:path>
            <lx:label>Hunter-Trapper</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841913">
        <xs:annotation>
          <xs:documentation>Pest Controller</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841913</lx:path>
            <lx:label>Pest Controller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="841999">
        <xs:annotation>
          <xs:documentation>Farm, Forestry and Garden Workers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.841999</lx:path>
            <lx:label>Farm, Forestry and Garden Workers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="851000">
        <xs:annotation>
          <xs:documentation>Food Preparation Assistants nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.851000</lx:path>
            <lx:label>Food Preparation Assistants nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="851111">
        <xs:annotation>
          <xs:documentation>Fast Food Cook</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.851111</lx:path>
            <lx:label>Fast Food Cook</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="851200">
        <xs:annotation>
          <xs:documentation>Food Trades Assistants nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.851200</lx:path>
            <lx:label>Food Trades Assistants nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="851211">
        <xs:annotation>
          <xs:documentation>Pastrycook's Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.851211</lx:path>
            <lx:label>Pastrycook's Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="851299">
        <xs:annotation>
          <xs:documentation>Food Trades Assistants nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.851299</lx:path>
            <lx:label>Food Trades Assistants nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="851311">
        <xs:annotation>
          <xs:documentation>Kitchenhand</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.851311</lx:path>
            <lx:label>Kitchenhand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="890000">
        <xs:annotation>
          <xs:documentation>Other Labourers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.890000</lx:path>
            <lx:label>Other Labourers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="891000">
        <xs:annotation>
          <xs:documentation>Freight Handlers and Shelf Fillers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.891000</lx:path>
            <lx:label>Freight Handlers and Shelf Fillers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="891100">
        <xs:annotation>
          <xs:documentation>Freight and Furniture Handlers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.891100</lx:path>
            <lx:label>Freight and Furniture Handlers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="891111">
        <xs:annotation>
          <xs:documentation>Freight Handler (Rail or Road)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.891111</lx:path>
            <lx:label>Freight Handler (Rail or Road)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="891112">
        <xs:annotation>
          <xs:documentation>Truck Driver's Offsider</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.891112</lx:path>
            <lx:label>Truck Driver's Offsider</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="891113">
        <xs:annotation>
          <xs:documentation>Waterside Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.891113</lx:path>
            <lx:label>Waterside Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="891211">
        <xs:annotation>
          <xs:documentation>Shelf Filler</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.891211</lx:path>
            <lx:label>Shelf Filler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899000">
        <xs:annotation>
          <xs:documentation>Miscellaneous Labourers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899000</lx:path>
            <lx:label>Miscellaneous Labourers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899111">
        <xs:annotation>
          <xs:documentation>Caretaker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899111</lx:path>
            <lx:label>Caretaker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899200">
        <xs:annotation>
          <xs:documentation>Deck and Fishing Hands nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899200</lx:path>
            <lx:label>Deck and Fishing Hands nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899211">
        <xs:annotation>
          <xs:documentation>Deck Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899211</lx:path>
            <lx:label>Deck Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899212">
        <xs:annotation>
          <xs:documentation>Fishing Hand</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899212</lx:path>
            <lx:label>Fishing Hand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899311">
        <xs:annotation>
          <xs:documentation>Handyperson</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899311</lx:path>
            <lx:label>Handyperson</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899400">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Parts and Accessories Fitters nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899400</lx:path>
            <lx:label>Motor Vehicle Parts and Accessories Fitters nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899411">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Parts and Accessories Fitter (General)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899411</lx:path>
            <lx:label>Motor Vehicle Parts and Accessories Fitter (General)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899412">
        <xs:annotation>
          <xs:documentation>Autoglazier</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899412</lx:path>
            <lx:label>Autoglazier</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899413">
        <xs:annotation>
          <xs:documentation>Exhaust and Muffler Repairer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899413</lx:path>
            <lx:label>Exhaust and Muffler Repairer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899414">
        <xs:annotation>
          <xs:documentation>Radiator Repairer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899414</lx:path>
            <lx:label>Radiator Repairer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899415">
        <xs:annotation>
          <xs:documentation>Tyre Fitter</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899415</lx:path>
            <lx:label>Tyre Fitter</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899500">
        <xs:annotation>
          <xs:documentation>Printing Assistants and Table Workers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899500</lx:path>
            <lx:label>Printing Assistants and Table Workers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899511">
        <xs:annotation>
          <xs:documentation>Printer's Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899511</lx:path>
            <lx:label>Printer's Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899512">
        <xs:annotation>
          <xs:documentation>Printing Table Worker</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899512</lx:path>
            <lx:label>Printing Table Worker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899611">
        <xs:annotation>
          <xs:documentation>Recycling or Rubbish Collector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899611</lx:path>
            <lx:label>Recycling or Rubbish Collector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899711">
        <xs:annotation>
          <xs:documentation>Vending Machine Attendant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899711</lx:path>
            <lx:label>Vending Machine Attendant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899900">
        <xs:annotation>
          <xs:documentation>Other Miscellaneous Labourers nfd</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899900</lx:path>
            <lx:label>Other Miscellaneous Labourers nfd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899911">
        <xs:annotation>
          <xs:documentation>Bicycle Mechanic</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899911</lx:path>
            <lx:label>Bicycle Mechanic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899912">
        <xs:annotation>
          <xs:documentation>Car Park Attendant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899912</lx:path>
            <lx:label>Car Park Attendant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899913">
        <xs:annotation>
          <xs:documentation>Crossing Supervisor</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899913</lx:path>
            <lx:label>Crossing Supervisor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899914">
        <xs:annotation>
          <xs:documentation>Electrical or Telecommunications Trades Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899914</lx:path>
            <lx:label>Electrical or Telecommunications Trades Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899915">
        <xs:annotation>
          <xs:documentation>Leaflet or Newspaper Deliverer</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899915</lx:path>
            <lx:label>Leaflet or Newspaper Deliverer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899916">
        <xs:annotation>
          <xs:documentation>Mechanic's Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899916</lx:path>
            <lx:label>Mechanic's Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899917">
        <xs:annotation>
          <xs:documentation>Railways Assistant</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899917</lx:path>
            <lx:label>Railways Assistant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899918">
        <xs:annotation>
          <xs:documentation>Sign Erector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899918</lx:path>
            <lx:label>Sign Erector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899921">
        <xs:annotation>
          <xs:documentation>Ticket Collector or Usher</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899921</lx:path>
            <lx:label>Ticket Collector or Usher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899922">
        <xs:annotation>
          <xs:documentation>Trolley Collector</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899922</lx:path>
            <lx:label>Trolley Collector</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899923">
        <xs:annotation>
          <xs:documentation>Road Traffic Controller</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899923</lx:path>
            <lx:label>Road Traffic Controller</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="899999">
        <xs:annotation>
          <xs:documentation>Labourers nec</xs:documentation>
          <xs:appinfo>
            <lx:path>anzscoOccupationCodeList.899999</lx:path>
            <lx:label>Labourers nec</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="anzsicIndustryCodeList">
    <xs:annotation>
      <xs:documentation>anzsicIndustryCodeList</xs:documentation>
      <xs:appinfo>
        <lx:path>anzsicIndustryCodeList</lx:path>
        <lx:label>anzsicIndustryCodeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="0100">
        <xs:annotation>
          <xs:documentation>Agriculture</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0100</lx:path>
            <lx:label>Agriculture</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0110">
        <xs:annotation>
          <xs:documentation>Nursery and Floriculture Production</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0110</lx:path>
            <lx:label>Nursery and Floriculture Production</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0111">
        <xs:annotation>
          <xs:documentation>Nursery Production (Under Cover)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0111</lx:path>
            <lx:label>Nursery Production (Under Cover)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0112">
        <xs:annotation>
          <xs:documentation>Nursery Production (Outdoors)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0112</lx:path>
            <lx:label>Nursery Production (Outdoors)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0113">
        <xs:annotation>
          <xs:documentation>Turf Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0113</lx:path>
            <lx:label>Turf Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0114">
        <xs:annotation>
          <xs:documentation>Floriculture Production (Under Cover)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0114</lx:path>
            <lx:label>Floriculture Production (Under Cover)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0115">
        <xs:annotation>
          <xs:documentation>Floriculture Production (Outdoors)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0115</lx:path>
            <lx:label>Floriculture Production (Outdoors)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0120">
        <xs:annotation>
          <xs:documentation>Mushroom and Vegetable Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0120</lx:path>
            <lx:label>Mushroom and Vegetable Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0121">
        <xs:annotation>
          <xs:documentation>Mushroom Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0121</lx:path>
            <lx:label>Mushroom Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0122">
        <xs:annotation>
          <xs:documentation>Vegetable Growing (Under Cover)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0122</lx:path>
            <lx:label>Vegetable Growing (Under Cover)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0123">
        <xs:annotation>
          <xs:documentation>Vegetable Growing (Outdoors)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0123</lx:path>
            <lx:label>Vegetable Growing (Outdoors)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0130">
        <xs:annotation>
          <xs:documentation>Fruit and Tree Nut Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0130</lx:path>
            <lx:label>Fruit and Tree Nut Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0131">
        <xs:annotation>
          <xs:documentation>Grape Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0131</lx:path>
            <lx:label>Grape Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0132">
        <xs:annotation>
          <xs:documentation>Kiwifruit Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0132</lx:path>
            <lx:label>Kiwifruit Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0133">
        <xs:annotation>
          <xs:documentation>Berry Fruit Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0133</lx:path>
            <lx:label>Berry Fruit Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0134">
        <xs:annotation>
          <xs:documentation>Apple and Pear Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0134</lx:path>
            <lx:label>Apple and Pear Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0135">
        <xs:annotation>
          <xs:documentation>Stone Fruit Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0135</lx:path>
            <lx:label>Stone Fruit Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0136">
        <xs:annotation>
          <xs:documentation>Citrus Fruit Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0136</lx:path>
            <lx:label>Citrus Fruit Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0137">
        <xs:annotation>
          <xs:documentation>Olive Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0137</lx:path>
            <lx:label>Olive Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0139">
        <xs:annotation>
          <xs:documentation>Other Fruit and Tree Nut Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0139</lx:path>
            <lx:label>Other Fruit and Tree Nut Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0140">
        <xs:annotation>
          <xs:documentation>Sheep, Beef Cattle and Grain Farming</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0140</lx:path>
            <lx:label>Sheep, Beef Cattle and Grain Farming</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0141">
        <xs:annotation>
          <xs:documentation>Sheep Farming (Specialised)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0141</lx:path>
            <lx:label>Sheep Farming (Specialised)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0142">
        <xs:annotation>
          <xs:documentation>Beef Cattle Farming (Specialised)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0142</lx:path>
            <lx:label>Beef Cattle Farming (Specialised)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0143">
        <xs:annotation>
          <xs:documentation>Beef Cattle Feedlots (Specialised)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0143</lx:path>
            <lx:label>Beef Cattle Feedlots (Specialised)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0144">
        <xs:annotation>
          <xs:documentation>Sheep-Beef Cattle Farming</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0144</lx:path>
            <lx:label>Sheep-Beef Cattle Farming</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0145">
        <xs:annotation>
          <xs:documentation>Grain-Sheep or Grain-Beef Cattle Farming</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0145</lx:path>
            <lx:label>Grain-Sheep or Grain-Beef Cattle Farming</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0146">
        <xs:annotation>
          <xs:documentation>Rice Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0146</lx:path>
            <lx:label>Rice Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0149">
        <xs:annotation>
          <xs:documentation>Other Grain Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0149</lx:path>
            <lx:label>Other Grain Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0150">
        <xs:annotation>
          <xs:documentation>Other Crop Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0150</lx:path>
            <lx:label>Other Crop Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0151">
        <xs:annotation>
          <xs:documentation>Sugar Cane Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0151</lx:path>
            <lx:label>Sugar Cane Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0152">
        <xs:annotation>
          <xs:documentation>Cotton Growing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0152</lx:path>
            <lx:label>Cotton Growing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0159">
        <xs:annotation>
          <xs:documentation>Other Crop Growing n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0159</lx:path>
            <lx:label>Other Crop Growing n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0160">
        <xs:annotation>
          <xs:documentation>Dairy Cattle Farming</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0160</lx:path>
            <lx:label>Dairy Cattle Farming</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0170">
        <xs:annotation>
          <xs:documentation>Poultry Farming</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0170</lx:path>
            <lx:label>Poultry Farming</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0171">
        <xs:annotation>
          <xs:documentation>Poultry Farming (Meat)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0171</lx:path>
            <lx:label>Poultry Farming (Meat)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0172">
        <xs:annotation>
          <xs:documentation>Poultry Farming (Eggs)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0172</lx:path>
            <lx:label>Poultry Farming (Eggs)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0180">
        <xs:annotation>
          <xs:documentation>Deer Farming</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0180</lx:path>
            <lx:label>Deer Farming</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0190">
        <xs:annotation>
          <xs:documentation>Other Livestock Farming</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0190</lx:path>
            <lx:label>Other Livestock Farming</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0191">
        <xs:annotation>
          <xs:documentation>Horse Farming</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0191</lx:path>
            <lx:label>Horse Farming</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0192">
        <xs:annotation>
          <xs:documentation>Pig Farming</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0192</lx:path>
            <lx:label>Pig Farming</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0193">
        <xs:annotation>
          <xs:documentation>Beekeeping</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0193</lx:path>
            <lx:label>Beekeeping</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0199">
        <xs:annotation>
          <xs:documentation>Other Livestock Farming n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0199</lx:path>
            <lx:label>Other Livestock Farming n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0200">
        <xs:annotation>
          <xs:documentation>Aquaculture</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0200</lx:path>
            <lx:label>Aquaculture</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0201">
        <xs:annotation>
          <xs:documentation>Offshore Longline and Rack Aquaculture</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0201</lx:path>
            <lx:label>Offshore Longline and Rack Aquaculture</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0202">
        <xs:annotation>
          <xs:documentation>Offshore Caged Aquaculture</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0202</lx:path>
            <lx:label>Offshore Caged Aquaculture</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0203">
        <xs:annotation>
          <xs:documentation>Onshore Aquaculture</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0203</lx:path>
            <lx:label>Onshore Aquaculture</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0300">
        <xs:annotation>
          <xs:documentation>Forestry and Logging</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0300</lx:path>
            <lx:label>Forestry and Logging</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0301">
        <xs:annotation>
          <xs:documentation>Forestry</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0301</lx:path>
            <lx:label>Forestry</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0302">
        <xs:annotation>
          <xs:documentation>Logging</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0302</lx:path>
            <lx:label>Logging</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0400">
        <xs:annotation>
          <xs:documentation>Fishing, Hunting and Trapping</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0400</lx:path>
            <lx:label>Fishing, Hunting and Trapping</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0410">
        <xs:annotation>
          <xs:documentation>Fishing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0410</lx:path>
            <lx:label>Fishing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0411">
        <xs:annotation>
          <xs:documentation>Rock Lobster and Crab Potting</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0411</lx:path>
            <lx:label>Rock Lobster and Crab Potting</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0412">
        <xs:annotation>
          <xs:documentation>Prawn Fishing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0412</lx:path>
            <lx:label>Prawn Fishing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0413">
        <xs:annotation>
          <xs:documentation>Line Fishing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0413</lx:path>
            <lx:label>Line Fishing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0414">
        <xs:annotation>
          <xs:documentation>Fish Trawling, Seining and Netting</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0414</lx:path>
            <lx:label>Fish Trawling, Seining and Netting</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0419">
        <xs:annotation>
          <xs:documentation>Other Fishing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0419</lx:path>
            <lx:label>Other Fishing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0420">
        <xs:annotation>
          <xs:documentation>Hunting and Trapping</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0420</lx:path>
            <lx:label>Hunting and Trapping</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0500">
        <xs:annotation>
          <xs:documentation>Agriculture, Forestry and Fishing Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0500</lx:path>
            <lx:label>Agriculture, Forestry and Fishing Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0510">
        <xs:annotation>
          <xs:documentation>Forestry Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0510</lx:path>
            <lx:label>Forestry Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0520">
        <xs:annotation>
          <xs:documentation>Agriculture and Fishing Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0520</lx:path>
            <lx:label>Agriculture and Fishing Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0521">
        <xs:annotation>
          <xs:documentation>Cotton Ginning</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0521</lx:path>
            <lx:label>Cotton Ginning</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0522">
        <xs:annotation>
          <xs:documentation>Shearing Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0522</lx:path>
            <lx:label>Shearing Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0529">
        <xs:annotation>
          <xs:documentation>Other Agriculture and Fishing Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0529</lx:path>
            <lx:label>Other Agriculture and Fishing Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0600">
        <xs:annotation>
          <xs:documentation>Coal Mining</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0600</lx:path>
            <lx:label>Coal Mining</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0700">
        <xs:annotation>
          <xs:documentation>Oil and Gas Extraction</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0700</lx:path>
            <lx:label>Oil and Gas Extraction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0800">
        <xs:annotation>
          <xs:documentation>Metal Ore Mining</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0800</lx:path>
            <lx:label>Metal Ore Mining</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0801">
        <xs:annotation>
          <xs:documentation>Iron Ore Mining</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0801</lx:path>
            <lx:label>Iron Ore Mining</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0802">
        <xs:annotation>
          <xs:documentation>Bauxite Mining</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0802</lx:path>
            <lx:label>Bauxite Mining</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0803">
        <xs:annotation>
          <xs:documentation>Copper Ore Mining</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0803</lx:path>
            <lx:label>Copper Ore Mining</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0804">
        <xs:annotation>
          <xs:documentation>Gold Ore Mining</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0804</lx:path>
            <lx:label>Gold Ore Mining</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0805">
        <xs:annotation>
          <xs:documentation>Mineral Sand Mining</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0805</lx:path>
            <lx:label>Mineral Sand Mining</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0806">
        <xs:annotation>
          <xs:documentation>Nickel Ore Mining</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0806</lx:path>
            <lx:label>Nickel Ore Mining</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0807">
        <xs:annotation>
          <xs:documentation>Silver-Lead-Zinc Ore Mining</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0807</lx:path>
            <lx:label>Silver-Lead-Zinc Ore Mining</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0809">
        <xs:annotation>
          <xs:documentation>Other Metal Ore Mining</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0809</lx:path>
            <lx:label>Other Metal Ore Mining</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0900">
        <xs:annotation>
          <xs:documentation>Non-Metallic Mineral Mining and Quarrying</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0900</lx:path>
            <lx:label>Non-Metallic Mineral Mining and Quarrying</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0910">
        <xs:annotation>
          <xs:documentation>Construction Material Mining</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0910</lx:path>
            <lx:label>Construction Material Mining</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0911">
        <xs:annotation>
          <xs:documentation>Gravel and Sand Quarrying</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0911</lx:path>
            <lx:label>Gravel and Sand Quarrying</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0919">
        <xs:annotation>
          <xs:documentation>Other Construction Material Mining</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0919</lx:path>
            <lx:label>Other Construction Material Mining</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0990">
        <xs:annotation>
          <xs:documentation>Other Non-Metallic Mineral Mining and Quarrying</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.0990</lx:path>
            <lx:label>Other Non-Metallic Mineral Mining and Quarrying</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1000">
        <xs:annotation>
          <xs:documentation>Exploration and Other Mining Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1000</lx:path>
            <lx:label>Exploration and Other Mining Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1010">
        <xs:annotation>
          <xs:documentation>Exploration</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1010</lx:path>
            <lx:label>Exploration</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1011">
        <xs:annotation>
          <xs:documentation>Petroleum Exploration</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1011</lx:path>
            <lx:label>Petroleum Exploration</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1012">
        <xs:annotation>
          <xs:documentation>Mineral Exploration</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1012</lx:path>
            <lx:label>Mineral Exploration</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1090">
        <xs:annotation>
          <xs:documentation>Other Mining Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1090</lx:path>
            <lx:label>Other Mining Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1100">
        <xs:annotation>
          <xs:documentation>Food Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1100</lx:path>
            <lx:label>Food Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1110">
        <xs:annotation>
          <xs:documentation>Meat and Meat Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1110</lx:path>
            <lx:label>Meat and Meat Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1111">
        <xs:annotation>
          <xs:documentation>Meat Processing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1111</lx:path>
            <lx:label>Meat Processing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1112">
        <xs:annotation>
          <xs:documentation>Poultry Processing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1112</lx:path>
            <lx:label>Poultry Processing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1113">
        <xs:annotation>
          <xs:documentation>Cured Meat and Smallgoods Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1113</lx:path>
            <lx:label>Cured Meat and Smallgoods Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1120">
        <xs:annotation>
          <xs:documentation>Seafood Processing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1120</lx:path>
            <lx:label>Seafood Processing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1130">
        <xs:annotation>
          <xs:documentation>Dairy Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1130</lx:path>
            <lx:label>Dairy Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1131">
        <xs:annotation>
          <xs:documentation>Milk and Cream Processing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1131</lx:path>
            <lx:label>Milk and Cream Processing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1132">
        <xs:annotation>
          <xs:documentation>Ice Cream Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1132</lx:path>
            <lx:label>Ice Cream Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1133">
        <xs:annotation>
          <xs:documentation>Cheese and Other Dairy Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1133</lx:path>
            <lx:label>Cheese and Other Dairy Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1140">
        <xs:annotation>
          <xs:documentation>Fruit and Vegetable Processing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1140</lx:path>
            <lx:label>Fruit and Vegetable Processing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1150">
        <xs:annotation>
          <xs:documentation>Oil and Fat Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1150</lx:path>
            <lx:label>Oil and Fat Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1160">
        <xs:annotation>
          <xs:documentation>Grain Mill and Cereal Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1160</lx:path>
            <lx:label>Grain Mill and Cereal Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1161">
        <xs:annotation>
          <xs:documentation>Grain Mill Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1161</lx:path>
            <lx:label>Grain Mill Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1162">
        <xs:annotation>
          <xs:documentation>Cereal, Pasta and Baking Mix Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1162</lx:path>
            <lx:label>Cereal, Pasta and Baking Mix Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1170">
        <xs:annotation>
          <xs:documentation>Bakery Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1170</lx:path>
            <lx:label>Bakery Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1171">
        <xs:annotation>
          <xs:documentation>Bread Manufacturing (Factory based)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1171</lx:path>
            <lx:label>Bread Manufacturing (Factory based)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1172">
        <xs:annotation>
          <xs:documentation>Cake and Pastry Manufacturing (Factory based)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1172</lx:path>
            <lx:label>Cake and Pastry Manufacturing (Factory based)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1173">
        <xs:annotation>
          <xs:documentation>Biscuit Manufacturing (Factory based)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1173</lx:path>
            <lx:label>Biscuit Manufacturing (Factory based)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1174">
        <xs:annotation>
          <xs:documentation>Bakery Product Manufacturing (Non-factory based)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1174</lx:path>
            <lx:label>Bakery Product Manufacturing (Non-factory based)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1180">
        <xs:annotation>
          <xs:documentation>Sugar and Confectionery Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1180</lx:path>
            <lx:label>Sugar and Confectionery Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1181">
        <xs:annotation>
          <xs:documentation>Sugar Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1181</lx:path>
            <lx:label>Sugar Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1182">
        <xs:annotation>
          <xs:documentation>Confectionery Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1182</lx:path>
            <lx:label>Confectionery Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1190">
        <xs:annotation>
          <xs:documentation>Other Food Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1190</lx:path>
            <lx:label>Other Food Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1191">
        <xs:annotation>
          <xs:documentation>Potato, Corn and Other Crisp Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1191</lx:path>
            <lx:label>Potato, Corn and Other Crisp Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1192">
        <xs:annotation>
          <xs:documentation>Prepared Animal and Bird Feed Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1192</lx:path>
            <lx:label>Prepared Animal and Bird Feed Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1199">
        <xs:annotation>
          <xs:documentation>Other Food Product Manufacturing n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1199</lx:path>
            <lx:label>Other Food Product Manufacturing n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1200">
        <xs:annotation>
          <xs:documentation>Beverage and Tobacco Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1200</lx:path>
            <lx:label>Beverage and Tobacco Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1210">
        <xs:annotation>
          <xs:documentation>Beverage Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1210</lx:path>
            <lx:label>Beverage Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1211">
        <xs:annotation>
          <xs:documentation>Soft Drink, Cordial and Syrup Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1211</lx:path>
            <lx:label>Soft Drink, Cordial and Syrup Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1212">
        <xs:annotation>
          <xs:documentation>Beer Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1212</lx:path>
            <lx:label>Beer Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1213">
        <xs:annotation>
          <xs:documentation>Spirit Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1213</lx:path>
            <lx:label>Spirit Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1214">
        <xs:annotation>
          <xs:documentation>Wine and Other Alcoholic Beverage Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1214</lx:path>
            <lx:label>Wine and Other Alcoholic Beverage Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1220">
        <xs:annotation>
          <xs:documentation>Cigarette and Tobacco Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1220</lx:path>
            <lx:label>Cigarette and Tobacco Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1300">
        <xs:annotation>
          <xs:documentation>Textile, Leather, Clothing and Footwear Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1300</lx:path>
            <lx:label>Textile, Leather, Clothing and Footwear Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1310">
        <xs:annotation>
          <xs:documentation>Textile Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1310</lx:path>
            <lx:label>Textile Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1311">
        <xs:annotation>
          <xs:documentation>Wool Scouring</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1311</lx:path>
            <lx:label>Wool Scouring</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1312">
        <xs:annotation>
          <xs:documentation>Natural Textile Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1312</lx:path>
            <lx:label>Natural Textile Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1313">
        <xs:annotation>
          <xs:documentation>Synthetic Textile Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1313</lx:path>
            <lx:label>Synthetic Textile Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1320">
        <xs:annotation>
          <xs:documentation>Leather Tanning, Fur Dressing and Leather Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1320</lx:path>
            <lx:label>Leather Tanning, Fur Dressing and Leather Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1330">
        <xs:annotation>
          <xs:documentation>Textile Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1330</lx:path>
            <lx:label>Textile Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1331">
        <xs:annotation>
          <xs:documentation>Textile Floor Covering Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1331</lx:path>
            <lx:label>Textile Floor Covering Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1332">
        <xs:annotation>
          <xs:documentation>Rope, Cordage and Twine Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1332</lx:path>
            <lx:label>Rope, Cordage and Twine Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1333">
        <xs:annotation>
          <xs:documentation>Cut and Sewn Textile Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1333</lx:path>
            <lx:label>Cut and Sewn Textile Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1334">
        <xs:annotation>
          <xs:documentation>Textile Finishing and Other Textile Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1334</lx:path>
            <lx:label>Textile Finishing and Other Textile Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1340">
        <xs:annotation>
          <xs:documentation>Knitted Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1340</lx:path>
            <lx:label>Knitted Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1350">
        <xs:annotation>
          <xs:documentation>Clothing and Footwear Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1350</lx:path>
            <lx:label>Clothing and Footwear Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1351">
        <xs:annotation>
          <xs:documentation>Clothing Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1351</lx:path>
            <lx:label>Clothing Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1352">
        <xs:annotation>
          <xs:documentation>Footwear Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1352</lx:path>
            <lx:label>Footwear Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1400">
        <xs:annotation>
          <xs:documentation>Wood Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1400</lx:path>
            <lx:label>Wood Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1410">
        <xs:annotation>
          <xs:documentation>Log Sawmilling and Timber Dressing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1410</lx:path>
            <lx:label>Log Sawmilling and Timber Dressing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1411">
        <xs:annotation>
          <xs:documentation>Log Sawmilling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1411</lx:path>
            <lx:label>Log Sawmilling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1412">
        <xs:annotation>
          <xs:documentation>Wood Chipping</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1412</lx:path>
            <lx:label>Wood Chipping</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1413">
        <xs:annotation>
          <xs:documentation>Timber Resawing and Dressing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1413</lx:path>
            <lx:label>Timber Resawing and Dressing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1490">
        <xs:annotation>
          <xs:documentation>Other Wood Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1490</lx:path>
            <lx:label>Other Wood Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1491">
        <xs:annotation>
          <xs:documentation>Prefabricated Wooden Building Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1491</lx:path>
            <lx:label>Prefabricated Wooden Building Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1492">
        <xs:annotation>
          <xs:documentation>Wooden Structural Fitting and Component Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1492</lx:path>
            <lx:label>Wooden Structural Fitting and Component Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1493">
        <xs:annotation>
          <xs:documentation>Veneer and Plywood Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1493</lx:path>
            <lx:label>Veneer and Plywood Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1494">
        <xs:annotation>
          <xs:documentation>Reconstituted Wood Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1494</lx:path>
            <lx:label>Reconstituted Wood Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1499">
        <xs:annotation>
          <xs:documentation>Other Wood Product Manufacturing n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1499</lx:path>
            <lx:label>Other Wood Product Manufacturing n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1500">
        <xs:annotation>
          <xs:documentation>Pulp, Paper and Converted Paper Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1500</lx:path>
            <lx:label>Pulp, Paper and Converted Paper Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1510">
        <xs:annotation>
          <xs:documentation>Pulp, Paper and Paperboard Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1510</lx:path>
            <lx:label>Pulp, Paper and Paperboard Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1520">
        <xs:annotation>
          <xs:documentation>Converted Paper Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1520</lx:path>
            <lx:label>Converted Paper Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1521">
        <xs:annotation>
          <xs:documentation>Corrugated Paperboard and Paperboard Container Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1521</lx:path>
            <lx:label>Corrugated Paperboard and Paperboard Container Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1522">
        <xs:annotation>
          <xs:documentation>Paper Bag Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1522</lx:path>
            <lx:label>Paper Bag Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1523">
        <xs:annotation>
          <xs:documentation>Paper Stationery Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1523</lx:path>
            <lx:label>Paper Stationery Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1524">
        <xs:annotation>
          <xs:documentation>Sanitary Paper Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1524</lx:path>
            <lx:label>Sanitary Paper Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1529">
        <xs:annotation>
          <xs:documentation>Other Converted Paper Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1529</lx:path>
            <lx:label>Other Converted Paper Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1600">
        <xs:annotation>
          <xs:documentation>Printing (including the Reproduction of Recorded Media)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1600</lx:path>
            <lx:label>Printing (including the Reproduction of Recorded Media)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1610">
        <xs:annotation>
          <xs:documentation>Printing and Printing Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1610</lx:path>
            <lx:label>Printing and Printing Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1611">
        <xs:annotation>
          <xs:documentation>Printing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1611</lx:path>
            <lx:label>Printing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1612">
        <xs:annotation>
          <xs:documentation>Printing Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1612</lx:path>
            <lx:label>Printing Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1620">
        <xs:annotation>
          <xs:documentation>Reproduction of Recorded Media</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1620</lx:path>
            <lx:label>Reproduction of Recorded Media</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1700">
        <xs:annotation>
          <xs:documentation>Petroleum and Coal Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1700</lx:path>
            <lx:label>Petroleum and Coal Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1701">
        <xs:annotation>
          <xs:documentation>Petroleum Refining and Petroleum Fuel Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1701</lx:path>
            <lx:label>Petroleum Refining and Petroleum Fuel Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1709">
        <xs:annotation>
          <xs:documentation>Other Petroleum and Coal Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1709</lx:path>
            <lx:label>Other Petroleum and Coal Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1800">
        <xs:annotation>
          <xs:documentation>Basic Chemical and Chemical Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1800</lx:path>
            <lx:label>Basic Chemical and Chemical Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1810">
        <xs:annotation>
          <xs:documentation>Basic Chemical Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1810</lx:path>
            <lx:label>Basic Chemical Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1811">
        <xs:annotation>
          <xs:documentation>Industrial Gas Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1811</lx:path>
            <lx:label>Industrial Gas Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1812">
        <xs:annotation>
          <xs:documentation>Basic Organic Chemical Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1812</lx:path>
            <lx:label>Basic Organic Chemical Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1813">
        <xs:annotation>
          <xs:documentation>Basic Inorganic Chemical Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1813</lx:path>
            <lx:label>Basic Inorganic Chemical Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1820">
        <xs:annotation>
          <xs:documentation>Basic Polymer Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1820</lx:path>
            <lx:label>Basic Polymer Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1821">
        <xs:annotation>
          <xs:documentation>Synthetic Resin and Synthetic Rubber Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1821</lx:path>
            <lx:label>Synthetic Resin and Synthetic Rubber Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1829">
        <xs:annotation>
          <xs:documentation>Other Basic Polymer Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1829</lx:path>
            <lx:label>Other Basic Polymer Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1830">
        <xs:annotation>
          <xs:documentation>Fertiliser and Pesticide Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1830</lx:path>
            <lx:label>Fertiliser and Pesticide Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1831">
        <xs:annotation>
          <xs:documentation>Fertiliser Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1831</lx:path>
            <lx:label>Fertiliser Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1832">
        <xs:annotation>
          <xs:documentation>Pesticide Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1832</lx:path>
            <lx:label>Pesticide Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1840">
        <xs:annotation>
          <xs:documentation>Pharmaceutical and Medicinal Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1840</lx:path>
            <lx:label>Pharmaceutical and Medicinal Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1841">
        <xs:annotation>
          <xs:documentation>Human Pharmaceutical and Medicinal Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1841</lx:path>
            <lx:label>Human Pharmaceutical and Medicinal Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1842">
        <xs:annotation>
          <xs:documentation>Veterinary Pharmaceutical and Medicinal Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1842</lx:path>
            <lx:label>Veterinary Pharmaceutical and Medicinal Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1850">
        <xs:annotation>
          <xs:documentation>Cleaning Compound and Toiletry Preparation Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1850</lx:path>
            <lx:label>Cleaning Compound and Toiletry Preparation Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1851">
        <xs:annotation>
          <xs:documentation>Cleaning Compound Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1851</lx:path>
            <lx:label>Cleaning Compound Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1852">
        <xs:annotation>
          <xs:documentation>Cosmetic and Toiletry Preparation Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1852</lx:path>
            <lx:label>Cosmetic and Toiletry Preparation Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1890">
        <xs:annotation>
          <xs:documentation>Other Basic Chemical Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1890</lx:path>
            <lx:label>Other Basic Chemical Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1891">
        <xs:annotation>
          <xs:documentation>Photographic Chemical Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1891</lx:path>
            <lx:label>Photographic Chemical Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1892">
        <xs:annotation>
          <xs:documentation>Explosive Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1892</lx:path>
            <lx:label>Explosive Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1899">
        <xs:annotation>
          <xs:documentation>Other Basic Chemical Product Manufacturing n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1899</lx:path>
            <lx:label>Other Basic Chemical Product Manufacturing n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1900">
        <xs:annotation>
          <xs:documentation>Polymer Product and Rubber Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1900</lx:path>
            <lx:label>Polymer Product and Rubber Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1910">
        <xs:annotation>
          <xs:documentation>Polymer Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1910</lx:path>
            <lx:label>Polymer Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1911">
        <xs:annotation>
          <xs:documentation>Polymer Film and Sheet Packaging Material Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1911</lx:path>
            <lx:label>Polymer Film and Sheet Packaging Material Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1912">
        <xs:annotation>
          <xs:documentation>Rigid and Semi-Rigid Polymer Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1912</lx:path>
            <lx:label>Rigid and Semi-Rigid Polymer Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1913">
        <xs:annotation>
          <xs:documentation>Polymer Foam Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1913</lx:path>
            <lx:label>Polymer Foam Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1914">
        <xs:annotation>
          <xs:documentation>Tyre Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1914</lx:path>
            <lx:label>Tyre Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1915">
        <xs:annotation>
          <xs:documentation>Adhesive Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1915</lx:path>
            <lx:label>Adhesive Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1916">
        <xs:annotation>
          <xs:documentation>Paint and Coatings Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1916</lx:path>
            <lx:label>Paint and Coatings Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1919">
        <xs:annotation>
          <xs:documentation>Other Polymer Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1919</lx:path>
            <lx:label>Other Polymer Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1920">
        <xs:annotation>
          <xs:documentation>Natural Rubber Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.1920</lx:path>
            <lx:label>Natural Rubber Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2000">
        <xs:annotation>
          <xs:documentation>Non-Metallic Mineral Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2000</lx:path>
            <lx:label>Non-Metallic Mineral Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2010">
        <xs:annotation>
          <xs:documentation>Glass and Glass Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2010</lx:path>
            <lx:label>Glass and Glass Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2020">
        <xs:annotation>
          <xs:documentation>Ceramic Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2020</lx:path>
            <lx:label>Ceramic Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2021">
        <xs:annotation>
          <xs:documentation>Clay Brick Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2021</lx:path>
            <lx:label>Clay Brick Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2029">
        <xs:annotation>
          <xs:documentation>Other Ceramic Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2029</lx:path>
            <lx:label>Other Ceramic Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2030">
        <xs:annotation>
          <xs:documentation>Cement, Lime, Plaster and Concrete Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2030</lx:path>
            <lx:label>Cement, Lime, Plaster and Concrete Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2031">
        <xs:annotation>
          <xs:documentation>Cement and Lime Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2031</lx:path>
            <lx:label>Cement and Lime Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2032">
        <xs:annotation>
          <xs:documentation>Plaster Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2032</lx:path>
            <lx:label>Plaster Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2033">
        <xs:annotation>
          <xs:documentation>Ready-Mixed Concrete Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2033</lx:path>
            <lx:label>Ready-Mixed Concrete Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2034">
        <xs:annotation>
          <xs:documentation>Concrete Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2034</lx:path>
            <lx:label>Concrete Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2090">
        <xs:annotation>
          <xs:documentation>Other Non-Metallic Mineral Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2090</lx:path>
            <lx:label>Other Non-Metallic Mineral Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2100">
        <xs:annotation>
          <xs:documentation>Primary Metal and Metal Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2100</lx:path>
            <lx:label>Primary Metal and Metal Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2110">
        <xs:annotation>
          <xs:documentation>Iron Smelting and Steel Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2110</lx:path>
            <lx:label>Iron Smelting and Steel Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2120">
        <xs:annotation>
          <xs:documentation>Basic Ferrous Metal Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2120</lx:path>
            <lx:label>Basic Ferrous Metal Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2121">
        <xs:annotation>
          <xs:documentation>Iron and Steel Casting</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2121</lx:path>
            <lx:label>Iron and Steel Casting</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2122">
        <xs:annotation>
          <xs:documentation>Steel Pipe and Tube Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2122</lx:path>
            <lx:label>Steel Pipe and Tube Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2130">
        <xs:annotation>
          <xs:documentation>Basic Non-Ferrous Metal Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2130</lx:path>
            <lx:label>Basic Non-Ferrous Metal Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2131">
        <xs:annotation>
          <xs:documentation>Alumina Production</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2131</lx:path>
            <lx:label>Alumina Production</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2132">
        <xs:annotation>
          <xs:documentation>Aluminium Smelting</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2132</lx:path>
            <lx:label>Aluminium Smelting</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2133">
        <xs:annotation>
          <xs:documentation>Copper, Silver, Lead and Zinc Smelting and Refining</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2133</lx:path>
            <lx:label>Copper, Silver, Lead and Zinc Smelting and Refining</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2139">
        <xs:annotation>
          <xs:documentation>Other Basic Non-Ferrous Metal Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2139</lx:path>
            <lx:label>Other Basic Non-Ferrous Metal Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2140">
        <xs:annotation>
          <xs:documentation>Basic Non-Ferrous Metal Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2140</lx:path>
            <lx:label>Basic Non-Ferrous Metal Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2141">
        <xs:annotation>
          <xs:documentation>Non-Ferrous Metal Casting</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2141</lx:path>
            <lx:label>Non-Ferrous Metal Casting</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2142">
        <xs:annotation>
          <xs:documentation>Aluminium Rolling, Drawing, Extruding</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2142</lx:path>
            <lx:label>Aluminium Rolling, Drawing, Extruding</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2149">
        <xs:annotation>
          <xs:documentation>Other Basic Non-Ferrous Metal Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2149</lx:path>
            <lx:label>Other Basic Non-Ferrous Metal Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2200">
        <xs:annotation>
          <xs:documentation>Fabricated Metal Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2200</lx:path>
            <lx:label>Fabricated Metal Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2210">
        <xs:annotation>
          <xs:documentation>Iron and Steel Forging</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2210</lx:path>
            <lx:label>Iron and Steel Forging</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2220">
        <xs:annotation>
          <xs:documentation>Structural Metal Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2220</lx:path>
            <lx:label>Structural Metal Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2221">
        <xs:annotation>
          <xs:documentation>Structural Steel Fabricating</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2221</lx:path>
            <lx:label>Structural Steel Fabricating</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2222">
        <xs:annotation>
          <xs:documentation>Prefabricated Metal Building Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2222</lx:path>
            <lx:label>Prefabricated Metal Building Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2223">
        <xs:annotation>
          <xs:documentation>Architectural Aluminium Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2223</lx:path>
            <lx:label>Architectural Aluminium Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2224">
        <xs:annotation>
          <xs:documentation>Metal Roof and Guttering Manufacturing (except Aluminium)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2224</lx:path>
            <lx:label>Metal Roof and Guttering Manufacturing (except Aluminium)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2229">
        <xs:annotation>
          <xs:documentation>Other Structural Metal Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2229</lx:path>
            <lx:label>Other Structural Metal Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2230">
        <xs:annotation>
          <xs:documentation>Metal Container Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2230</lx:path>
            <lx:label>Metal Container Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2231">
        <xs:annotation>
          <xs:documentation>Boiler, Tank and Other Heavy Gauge Metal Container Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2231</lx:path>
            <lx:label>Boiler, Tank and Other Heavy Gauge Metal Container Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2239">
        <xs:annotation>
          <xs:documentation>Other Metal Container Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2239</lx:path>
            <lx:label>Other Metal Container Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2240">
        <xs:annotation>
          <xs:documentation>Sheet Metal Product Manufacturing (except Metal Structural and Container</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2240</lx:path>
            <lx:label>Sheet Metal Product Manufacturing (except Metal Structural and Container</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2290">
        <xs:annotation>
          <xs:documentation>Other Fabricated Metal Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2290</lx:path>
            <lx:label>Other Fabricated Metal Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2291">
        <xs:annotation>
          <xs:documentation>Spring and Wire Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2291</lx:path>
            <lx:label>Spring and Wire Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2292">
        <xs:annotation>
          <xs:documentation>Nut, Bolt, Screw and Rivet Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2292</lx:path>
            <lx:label>Nut, Bolt, Screw and Rivet Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2293">
        <xs:annotation>
          <xs:documentation>Metal Coating and Finishing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2293</lx:path>
            <lx:label>Metal Coating and Finishing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2299">
        <xs:annotation>
          <xs:documentation>Other Fabricated Metal Product Manufacturing n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2299</lx:path>
            <lx:label>Other Fabricated Metal Product Manufacturing n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2300">
        <xs:annotation>
          <xs:documentation>Transport Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2300</lx:path>
            <lx:label>Transport Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2310">
        <xs:annotation>
          <xs:documentation>Motor Vehicle and Motor Vehicle Part Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2310</lx:path>
            <lx:label>Motor Vehicle and Motor Vehicle Part Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2311">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2311</lx:path>
            <lx:label>Motor Vehicle Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2312">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Body and Trailer Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2312</lx:path>
            <lx:label>Motor Vehicle Body and Trailer Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2313">
        <xs:annotation>
          <xs:documentation>Automotive Electrical Component Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2313</lx:path>
            <lx:label>Automotive Electrical Component Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2319">
        <xs:annotation>
          <xs:documentation>Other Motor Vehicle Parts Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2319</lx:path>
            <lx:label>Other Motor Vehicle Parts Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2390">
        <xs:annotation>
          <xs:documentation>Other Transport Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2390</lx:path>
            <lx:label>Other Transport Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2391">
        <xs:annotation>
          <xs:documentation>Shipbuilding and Repair Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2391</lx:path>
            <lx:label>Shipbuilding and Repair Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2392">
        <xs:annotation>
          <xs:documentation>Boatbuilding and Repair Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2392</lx:path>
            <lx:label>Boatbuilding and Repair Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2393">
        <xs:annotation>
          <xs:documentation>Railway Rolling Stock Manufacturing and Repair Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2393</lx:path>
            <lx:label>Railway Rolling Stock Manufacturing and Repair Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2394">
        <xs:annotation>
          <xs:documentation>Aircraft Manufacturing and Repair Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2394</lx:path>
            <lx:label>Aircraft Manufacturing and Repair Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2399">
        <xs:annotation>
          <xs:documentation>Other Transport Equipment Manufacturing n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2399</lx:path>
            <lx:label>Other Transport Equipment Manufacturing n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2400">
        <xs:annotation>
          <xs:documentation>Machinery and Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2400</lx:path>
            <lx:label>Machinery and Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2410">
        <xs:annotation>
          <xs:documentation>Professional and Scientific Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2410</lx:path>
            <lx:label>Professional and Scientific Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2411">
        <xs:annotation>
          <xs:documentation>Photographic, Optical and Ophthalmic Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2411</lx:path>
            <lx:label>Photographic, Optical and Ophthalmic Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2412">
        <xs:annotation>
          <xs:documentation>Medical and Surgical Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2412</lx:path>
            <lx:label>Medical and Surgical Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2419">
        <xs:annotation>
          <xs:documentation>Other Professional and Scientific Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2419</lx:path>
            <lx:label>Other Professional and Scientific Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2420">
        <xs:annotation>
          <xs:documentation>Computer and Electronic Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2420</lx:path>
            <lx:label>Computer and Electronic Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2421">
        <xs:annotation>
          <xs:documentation>Computer and Electronic Office Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2421</lx:path>
            <lx:label>Computer and Electronic Office Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2422">
        <xs:annotation>
          <xs:documentation>Communication Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2422</lx:path>
            <lx:label>Communication Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2429">
        <xs:annotation>
          <xs:documentation>Other Electronic Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2429</lx:path>
            <lx:label>Other Electronic Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2430">
        <xs:annotation>
          <xs:documentation>Electrical Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2430</lx:path>
            <lx:label>Electrical Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2431">
        <xs:annotation>
          <xs:documentation>Electric Cable and Wire Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2431</lx:path>
            <lx:label>Electric Cable and Wire Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2432">
        <xs:annotation>
          <xs:documentation>Electric Lighting Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2432</lx:path>
            <lx:label>Electric Lighting Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2439">
        <xs:annotation>
          <xs:documentation>Other Electrical Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2439</lx:path>
            <lx:label>Other Electrical Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2440">
        <xs:annotation>
          <xs:documentation>Domestic Appliance Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2440</lx:path>
            <lx:label>Domestic Appliance Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2441">
        <xs:annotation>
          <xs:documentation>Whiteware Appliance Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2441</lx:path>
            <lx:label>Whiteware Appliance Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2449">
        <xs:annotation>
          <xs:documentation>Other Domestic Appliance Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2449</lx:path>
            <lx:label>Other Domestic Appliance Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2450">
        <xs:annotation>
          <xs:documentation>Pump, Compressor, Heating and Ventilation Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2450</lx:path>
            <lx:label>Pump, Compressor, Heating and Ventilation Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2451">
        <xs:annotation>
          <xs:documentation>Pump and Compressor Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2451</lx:path>
            <lx:label>Pump and Compressor Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2452">
        <xs:annotation>
          <xs:documentation>Fixed Space Heating, Cooling and Ventilation Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2452</lx:path>
            <lx:label>Fixed Space Heating, Cooling and Ventilation Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2460">
        <xs:annotation>
          <xs:documentation>Specialised Machinery and Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2460</lx:path>
            <lx:label>Specialised Machinery and Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2461">
        <xs:annotation>
          <xs:documentation>Agricultural Machinery and Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2461</lx:path>
            <lx:label>Agricultural Machinery and Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2462">
        <xs:annotation>
          <xs:documentation>Mining and Construction Machinery Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2462</lx:path>
            <lx:label>Mining and Construction Machinery Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2463">
        <xs:annotation>
          <xs:documentation>Machine Tool and Parts Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2463</lx:path>
            <lx:label>Machine Tool and Parts Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2469">
        <xs:annotation>
          <xs:documentation>Other Specialised Machinery and Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2469</lx:path>
            <lx:label>Other Specialised Machinery and Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2490">
        <xs:annotation>
          <xs:documentation>Other Machinery and Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2490</lx:path>
            <lx:label>Other Machinery and Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2491">
        <xs:annotation>
          <xs:documentation>Lifting and Material Handling Equipment Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2491</lx:path>
            <lx:label>Lifting and Material Handling Equipment Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2499">
        <xs:annotation>
          <xs:documentation>Other Machinery and Equipment Manufacturing n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2499</lx:path>
            <lx:label>Other Machinery and Equipment Manufacturing n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2500">
        <xs:annotation>
          <xs:documentation>Furniture and Other Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2500</lx:path>
            <lx:label>Furniture and Other Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2510">
        <xs:annotation>
          <xs:documentation>Furniture Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2510</lx:path>
            <lx:label>Furniture Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2511">
        <xs:annotation>
          <xs:documentation>Wooden Furniture and Upholstered Seat Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2511</lx:path>
            <lx:label>Wooden Furniture and Upholstered Seat Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2512">
        <xs:annotation>
          <xs:documentation>Metal Furniture Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2512</lx:path>
            <lx:label>Metal Furniture Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2513">
        <xs:annotation>
          <xs:documentation>Mattress Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2513</lx:path>
            <lx:label>Mattress Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2519">
        <xs:annotation>
          <xs:documentation>Other Furniture Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2519</lx:path>
            <lx:label>Other Furniture Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2590">
        <xs:annotation>
          <xs:documentation>Other Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2590</lx:path>
            <lx:label>Other Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2591">
        <xs:annotation>
          <xs:documentation>Jewellery and Silverware Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2591</lx:path>
            <lx:label>Jewellery and Silverware Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2592">
        <xs:annotation>
          <xs:documentation>Toy, Sporting and Recreational Product Manufacturing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2592</lx:path>
            <lx:label>Toy, Sporting and Recreational Product Manufacturing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2599">
        <xs:annotation>
          <xs:documentation>Other Manufacturing n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2599</lx:path>
            <lx:label>Other Manufacturing n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2600">
        <xs:annotation>
          <xs:documentation>Electricity Supply</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2600</lx:path>
            <lx:label>Electricity Supply</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2610">
        <xs:annotation>
          <xs:documentation>Electricity Generation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2610</lx:path>
            <lx:label>Electricity Generation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2611">
        <xs:annotation>
          <xs:documentation>Fossil Fuel Electricity Generation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2611</lx:path>
            <lx:label>Fossil Fuel Electricity Generation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2612">
        <xs:annotation>
          <xs:documentation>Hydro-Electricity Generation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2612</lx:path>
            <lx:label>Hydro-Electricity Generation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2619">
        <xs:annotation>
          <xs:documentation>Other Electricity Generation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2619</lx:path>
            <lx:label>Other Electricity Generation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2620">
        <xs:annotation>
          <xs:documentation>Electricity Transmission</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2620</lx:path>
            <lx:label>Electricity Transmission</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2630">
        <xs:annotation>
          <xs:documentation>Electricity Distribution</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2630</lx:path>
            <lx:label>Electricity Distribution</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2640">
        <xs:annotation>
          <xs:documentation>On Selling Electricity and Electricity Market Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2640</lx:path>
            <lx:label>On Selling Electricity and Electricity Market Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2700">
        <xs:annotation>
          <xs:documentation>Gas Supply</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2700</lx:path>
            <lx:label>Gas Supply</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2800">
        <xs:annotation>
          <xs:documentation>Water Supply, Sewerage and Drainage Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2800</lx:path>
            <lx:label>Water Supply, Sewerage and Drainage Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2810">
        <xs:annotation>
          <xs:documentation>Water Supply, Sewerage and Drainage Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2810</lx:path>
            <lx:label>Water Supply, Sewerage and Drainage Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2811">
        <xs:annotation>
          <xs:documentation>Water Supply</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2811</lx:path>
            <lx:label>Water Supply</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2812">
        <xs:annotation>
          <xs:documentation>Sewerage and Drainage Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2812</lx:path>
            <lx:label>Sewerage and Drainage Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2900">
        <xs:annotation>
          <xs:documentation>Waste Collection, Treatment and Disposal Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2900</lx:path>
            <lx:label>Waste Collection, Treatment and Disposal Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2910">
        <xs:annotation>
          <xs:documentation>Waste Collection Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2910</lx:path>
            <lx:label>Waste Collection Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2911">
        <xs:annotation>
          <xs:documentation>Solid Waste Collection Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2911</lx:path>
            <lx:label>Solid Waste Collection Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2919">
        <xs:annotation>
          <xs:documentation>Other Waste Collection Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2919</lx:path>
            <lx:label>Other Waste Collection Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2920">
        <xs:annotation>
          <xs:documentation>Waste Treatment, Disposal and Remediation Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2920</lx:path>
            <lx:label>Waste Treatment, Disposal and Remediation Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2921">
        <xs:annotation>
          <xs:documentation>Waste Treatment and Disposal Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2921</lx:path>
            <lx:label>Waste Treatment and Disposal Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2922">
        <xs:annotation>
          <xs:documentation>Waste Remediation and Materials Recovery Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.2922</lx:path>
            <lx:label>Waste Remediation and Materials Recovery Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3000">
        <xs:annotation>
          <xs:documentation>Building Construction</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3000</lx:path>
            <lx:label>Building Construction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3010">
        <xs:annotation>
          <xs:documentation>Residential Building Construction</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3010</lx:path>
            <lx:label>Residential Building Construction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3011">
        <xs:annotation>
          <xs:documentation>House Construction</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3011</lx:path>
            <lx:label>House Construction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3019">
        <xs:annotation>
          <xs:documentation>Other Residential Building Construction</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3019</lx:path>
            <lx:label>Other Residential Building Construction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3020">
        <xs:annotation>
          <xs:documentation>Non-Residential Building Construction</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3020</lx:path>
            <lx:label>Non-Residential Building Construction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3100">
        <xs:annotation>
          <xs:documentation>Heavy and Civil Engineering Construction</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3100</lx:path>
            <lx:label>Heavy and Civil Engineering Construction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3101">
        <xs:annotation>
          <xs:documentation>Road and Bridge Construction</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3101</lx:path>
            <lx:label>Road and Bridge Construction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3109">
        <xs:annotation>
          <xs:documentation>Other Heavy and Civil Engineering Construction</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3109</lx:path>
            <lx:label>Other Heavy and Civil Engineering Construction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3200">
        <xs:annotation>
          <xs:documentation>Construction Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3200</lx:path>
            <lx:label>Construction Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3210">
        <xs:annotation>
          <xs:documentation>Land Development and Site Preparation Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3210</lx:path>
            <lx:label>Land Development and Site Preparation Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3211">
        <xs:annotation>
          <xs:documentation>Land Development and Subdivision</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3211</lx:path>
            <lx:label>Land Development and Subdivision</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3212">
        <xs:annotation>
          <xs:documentation>Site Preparation Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3212</lx:path>
            <lx:label>Site Preparation Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3220">
        <xs:annotation>
          <xs:documentation>Building Structure Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3220</lx:path>
            <lx:label>Building Structure Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3221">
        <xs:annotation>
          <xs:documentation>Concreting Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3221</lx:path>
            <lx:label>Concreting Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3222">
        <xs:annotation>
          <xs:documentation>Bricklaying Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3222</lx:path>
            <lx:label>Bricklaying Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3223">
        <xs:annotation>
          <xs:documentation>Roofing Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3223</lx:path>
            <lx:label>Roofing Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3224">
        <xs:annotation>
          <xs:documentation>Structural Steel Erection Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3224</lx:path>
            <lx:label>Structural Steel Erection Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3230">
        <xs:annotation>
          <xs:documentation>Building Installation Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3230</lx:path>
            <lx:label>Building Installation Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3231">
        <xs:annotation>
          <xs:documentation>Plumbing Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3231</lx:path>
            <lx:label>Plumbing Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3232">
        <xs:annotation>
          <xs:documentation>Electrical Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3232</lx:path>
            <lx:label>Electrical Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3233">
        <xs:annotation>
          <xs:documentation>Air Conditioning and Heating Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3233</lx:path>
            <lx:label>Air Conditioning and Heating Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3234">
        <xs:annotation>
          <xs:documentation>Fire and Security Alarm Installation Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3234</lx:path>
            <lx:label>Fire and Security Alarm Installation Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3239">
        <xs:annotation>
          <xs:documentation>Other Building Installation Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3239</lx:path>
            <lx:label>Other Building Installation Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3240">
        <xs:annotation>
          <xs:documentation>Building Completion Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3240</lx:path>
            <lx:label>Building Completion Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3241">
        <xs:annotation>
          <xs:documentation>Plastering and Ceiling Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3241</lx:path>
            <lx:label>Plastering and Ceiling Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3242">
        <xs:annotation>
          <xs:documentation>Carpentry Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3242</lx:path>
            <lx:label>Carpentry Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3243">
        <xs:annotation>
          <xs:documentation>Tiling and Carpeting Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3243</lx:path>
            <lx:label>Tiling and Carpeting Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3244">
        <xs:annotation>
          <xs:documentation>Painting and Decorating Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3244</lx:path>
            <lx:label>Painting and Decorating Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3245">
        <xs:annotation>
          <xs:documentation>Glazing Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3245</lx:path>
            <lx:label>Glazing Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3290">
        <xs:annotation>
          <xs:documentation>Other Construction Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3290</lx:path>
            <lx:label>Other Construction Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3291">
        <xs:annotation>
          <xs:documentation>Landscape Construction Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3291</lx:path>
            <lx:label>Landscape Construction Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3292">
        <xs:annotation>
          <xs:documentation>Hire of Construction Machinery with Operator</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3292</lx:path>
            <lx:label>Hire of Construction Machinery with Operator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3299">
        <xs:annotation>
          <xs:documentation>Other Construction Services n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3299</lx:path>
            <lx:label>Other Construction Services n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3300">
        <xs:annotation>
          <xs:documentation>Basic Material Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3300</lx:path>
            <lx:label>Basic Material Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3310">
        <xs:annotation>
          <xs:documentation>Agricultural Product Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3310</lx:path>
            <lx:label>Agricultural Product Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3311">
        <xs:annotation>
          <xs:documentation>Wool Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3311</lx:path>
            <lx:label>Wool Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3312">
        <xs:annotation>
          <xs:documentation>Cereal Grain Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3312</lx:path>
            <lx:label>Cereal Grain Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3319">
        <xs:annotation>
          <xs:documentation>Other Agricultural Product Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3319</lx:path>
            <lx:label>Other Agricultural Product Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3320">
        <xs:annotation>
          <xs:documentation>Mineral, Metal and Chemical Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3320</lx:path>
            <lx:label>Mineral, Metal and Chemical Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3321">
        <xs:annotation>
          <xs:documentation>Petroleum Product Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3321</lx:path>
            <lx:label>Petroleum Product Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3322">
        <xs:annotation>
          <xs:documentation>Metal and Mineral Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3322</lx:path>
            <lx:label>Metal and Mineral Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3323">
        <xs:annotation>
          <xs:documentation>Industrial and Agricultural Chemical Product Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3323</lx:path>
            <lx:label>Industrial and Agricultural Chemical Product Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3330">
        <xs:annotation>
          <xs:documentation>Timber and Hardware Goods Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3330</lx:path>
            <lx:label>Timber and Hardware Goods Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3331">
        <xs:annotation>
          <xs:documentation>Timber Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3331</lx:path>
            <lx:label>Timber Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3332">
        <xs:annotation>
          <xs:documentation>Plumbing Goods Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3332</lx:path>
            <lx:label>Plumbing Goods Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3339">
        <xs:annotation>
          <xs:documentation>Other Hardware Goods Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3339</lx:path>
            <lx:label>Other Hardware Goods Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3400">
        <xs:annotation>
          <xs:documentation>Machinery and Equipment Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3400</lx:path>
            <lx:label>Machinery and Equipment Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3410">
        <xs:annotation>
          <xs:documentation>Specialised Industrial Machinery and Equipment Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3410</lx:path>
            <lx:label>Specialised Industrial Machinery and Equipment Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3411">
        <xs:annotation>
          <xs:documentation>Agricultural and Construction Machinery Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3411</lx:path>
            <lx:label>Agricultural and Construction Machinery Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3419">
        <xs:annotation>
          <xs:documentation>Other Specialised Industrial Machinery and Equipment Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3419</lx:path>
            <lx:label>Other Specialised Industrial Machinery and Equipment Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3490">
        <xs:annotation>
          <xs:documentation>Other Machinery and Equipment Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3490</lx:path>
            <lx:label>Other Machinery and Equipment Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3491">
        <xs:annotation>
          <xs:documentation>Professional and Scientific Goods Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3491</lx:path>
            <lx:label>Professional and Scientific Goods Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3492">
        <xs:annotation>
          <xs:documentation>Computer and Computer Peripheral Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3492</lx:path>
            <lx:label>Computer and Computer Peripheral Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3493">
        <xs:annotation>
          <xs:documentation>Telecommunication Goods Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3493</lx:path>
            <lx:label>Telecommunication Goods Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3494">
        <xs:annotation>
          <xs:documentation>Other Electrical and Electronic Goods Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3494</lx:path>
            <lx:label>Other Electrical and Electronic Goods Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3499">
        <xs:annotation>
          <xs:documentation>Other Machinery and Equipment Wholesaling n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3499</lx:path>
            <lx:label>Other Machinery and Equipment Wholesaling n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3500">
        <xs:annotation>
          <xs:documentation>Motor Vehicle and Motor Vehicle Parts Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3500</lx:path>
            <lx:label>Motor Vehicle and Motor Vehicle Parts Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3501">
        <xs:annotation>
          <xs:documentation>Car Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3501</lx:path>
            <lx:label>Car Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3502">
        <xs:annotation>
          <xs:documentation>Commercial Vehicle Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3502</lx:path>
            <lx:label>Commercial Vehicle Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3503">
        <xs:annotation>
          <xs:documentation>Trailer and Other Motor Vehicle Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3503</lx:path>
            <lx:label>Trailer and Other Motor Vehicle Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3504">
        <xs:annotation>
          <xs:documentation>Motor Vehicle New Parts Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3504</lx:path>
            <lx:label>Motor Vehicle New Parts Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3505">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Dismantling and Used Parts Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3505</lx:path>
            <lx:label>Motor Vehicle Dismantling and Used Parts Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3600">
        <xs:annotation>
          <xs:documentation>Grocery, Liquor and Tobacco Product Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3600</lx:path>
            <lx:label>Grocery, Liquor and Tobacco Product Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3601">
        <xs:annotation>
          <xs:documentation>General Line Grocery Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3601</lx:path>
            <lx:label>General Line Grocery Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3602">
        <xs:annotation>
          <xs:documentation>Meat, Poultry and Smallgoods Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3602</lx:path>
            <lx:label>Meat, Poultry and Smallgoods Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3603">
        <xs:annotation>
          <xs:documentation>Dairy Produce Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3603</lx:path>
            <lx:label>Dairy Produce Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3604">
        <xs:annotation>
          <xs:documentation>Fish and Seafood Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3604</lx:path>
            <lx:label>Fish and Seafood Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3605">
        <xs:annotation>
          <xs:documentation>Fruit and Vegetable Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3605</lx:path>
            <lx:label>Fruit and Vegetable Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3606">
        <xs:annotation>
          <xs:documentation>Liquor and Tobacco Product Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3606</lx:path>
            <lx:label>Liquor and Tobacco Product Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3609">
        <xs:annotation>
          <xs:documentation>Other Grocery Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3609</lx:path>
            <lx:label>Other Grocery Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3700">
        <xs:annotation>
          <xs:documentation>Other Goods Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3700</lx:path>
            <lx:label>Other Goods Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3710">
        <xs:annotation>
          <xs:documentation>Textile, Clothing and Footwear Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3710</lx:path>
            <lx:label>Textile, Clothing and Footwear Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3711">
        <xs:annotation>
          <xs:documentation>Textile Product Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3711</lx:path>
            <lx:label>Textile Product Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3712">
        <xs:annotation>
          <xs:documentation>Clothing and Footwear Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3712</lx:path>
            <lx:label>Clothing and Footwear Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3720">
        <xs:annotation>
          <xs:documentation>Pharmaceutical and Toiletry Goods Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3720</lx:path>
            <lx:label>Pharmaceutical and Toiletry Goods Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3730">
        <xs:annotation>
          <xs:documentation>Furniture, Floor Covering and Other Goods Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3730</lx:path>
            <lx:label>Furniture, Floor Covering and Other Goods Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3731">
        <xs:annotation>
          <xs:documentation>Furniture and Floor Covering Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3731</lx:path>
            <lx:label>Furniture and Floor Covering Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3732">
        <xs:annotation>
          <xs:documentation>Jewellery and Watch Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3732</lx:path>
            <lx:label>Jewellery and Watch Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3733">
        <xs:annotation>
          <xs:documentation>Kitchen and Diningware Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3733</lx:path>
            <lx:label>Kitchen and Diningware Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3734">
        <xs:annotation>
          <xs:documentation>Toy and Sporting Goods Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3734</lx:path>
            <lx:label>Toy and Sporting Goods Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3735">
        <xs:annotation>
          <xs:documentation>Book and Magazine Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3735</lx:path>
            <lx:label>Book and Magazine Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3736">
        <xs:annotation>
          <xs:documentation>Paper Product Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3736</lx:path>
            <lx:label>Paper Product Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3739">
        <xs:annotation>
          <xs:documentation>Other Goods Wholesaling n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3739</lx:path>
            <lx:label>Other Goods Wholesaling n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3800">
        <xs:annotation>
          <xs:documentation>Commission-Based Wholesaling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3800</lx:path>
            <lx:label>Commission-Based Wholesaling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3900">
        <xs:annotation>
          <xs:documentation>Motor Vehicle and Motor Vehicle Parts Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3900</lx:path>
            <lx:label>Motor Vehicle and Motor Vehicle Parts Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3910">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3910</lx:path>
            <lx:label>Motor Vehicle Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3911">
        <xs:annotation>
          <xs:documentation>Car Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3911</lx:path>
            <lx:label>Car Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3912">
        <xs:annotation>
          <xs:documentation>Motor Cycle Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3912</lx:path>
            <lx:label>Motor Cycle Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3913">
        <xs:annotation>
          <xs:documentation>Trailer and Other Motor Vehicle Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3913</lx:path>
            <lx:label>Trailer and Other Motor Vehicle Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3920">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Parts and Tyre Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3920</lx:path>
            <lx:label>Motor Vehicle Parts and Tyre Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3921">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Parts Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3921</lx:path>
            <lx:label>Motor Vehicle Parts Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3922">
        <xs:annotation>
          <xs:documentation>Tyre Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.3922</lx:path>
            <lx:label>Tyre Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4000">
        <xs:annotation>
          <xs:documentation>Fuel Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4000</lx:path>
            <lx:label>Fuel Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4100">
        <xs:annotation>
          <xs:documentation>Food Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4100</lx:path>
            <lx:label>Food Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4110">
        <xs:annotation>
          <xs:documentation>Supermarket and Grocery Stores</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4110</lx:path>
            <lx:label>Supermarket and Grocery Stores</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4120">
        <xs:annotation>
          <xs:documentation>Specialised Food Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4120</lx:path>
            <lx:label>Specialised Food Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4121">
        <xs:annotation>
          <xs:documentation>Fresh Meat, Fish and Poultry Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4121</lx:path>
            <lx:label>Fresh Meat, Fish and Poultry Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4122">
        <xs:annotation>
          <xs:documentation>Fruit and Vegetable Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4122</lx:path>
            <lx:label>Fruit and Vegetable Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4123">
        <xs:annotation>
          <xs:documentation>Liquor Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4123</lx:path>
            <lx:label>Liquor Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4129">
        <xs:annotation>
          <xs:documentation>Other Specialised Food Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4129</lx:path>
            <lx:label>Other Specialised Food Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4200">
        <xs:annotation>
          <xs:documentation>Other Store-Based Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4200</lx:path>
            <lx:label>Other Store-Based Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4210">
        <xs:annotation>
          <xs:documentation>Furniture, Floor Coverings, Houseware and Textile Goods Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4210</lx:path>
            <lx:label>Furniture, Floor Coverings, Houseware and Textile Goods Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4211">
        <xs:annotation>
          <xs:documentation>Furniture Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4211</lx:path>
            <lx:label>Furniture Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4212">
        <xs:annotation>
          <xs:documentation>Floor Coverings Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4212</lx:path>
            <lx:label>Floor Coverings Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4213">
        <xs:annotation>
          <xs:documentation>Houseware Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4213</lx:path>
            <lx:label>Houseware Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4214">
        <xs:annotation>
          <xs:documentation>Manchester and Other Textile Goods Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4214</lx:path>
            <lx:label>Manchester and Other Textile Goods Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4220">
        <xs:annotation>
          <xs:documentation>Electrical and Electronic Goods Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4220</lx:path>
            <lx:label>Electrical and Electronic Goods Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4221">
        <xs:annotation>
          <xs:documentation>Electrical, Electronic and Gas Appliance Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4221</lx:path>
            <lx:label>Electrical, Electronic and Gas Appliance Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4222">
        <xs:annotation>
          <xs:documentation>Computer and Computer Peripheral Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4222</lx:path>
            <lx:label>Computer and Computer Peripheral Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4229">
        <xs:annotation>
          <xs:documentation>Other Electrical and Electronic Goods Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4229</lx:path>
            <lx:label>Other Electrical and Electronic Goods Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4230">
        <xs:annotation>
          <xs:documentation>Hardware, Building and Garden Supplies Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4230</lx:path>
            <lx:label>Hardware, Building and Garden Supplies Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4231">
        <xs:annotation>
          <xs:documentation>Hardware and Building Supplies Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4231</lx:path>
            <lx:label>Hardware and Building Supplies Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4232">
        <xs:annotation>
          <xs:documentation>Garden Supplies Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4232</lx:path>
            <lx:label>Garden Supplies Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4240">
        <xs:annotation>
          <xs:documentation>Recreational Goods Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4240</lx:path>
            <lx:label>Recreational Goods Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4241">
        <xs:annotation>
          <xs:documentation>Sport and Camping Equipment Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4241</lx:path>
            <lx:label>Sport and Camping Equipment Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4242">
        <xs:annotation>
          <xs:documentation>Entertainment Media Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4242</lx:path>
            <lx:label>Entertainment Media Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4243">
        <xs:annotation>
          <xs:documentation>Toy and Game Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4243</lx:path>
            <lx:label>Toy and Game Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4244">
        <xs:annotation>
          <xs:documentation>Newspaper and Book Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4244</lx:path>
            <lx:label>Newspaper and Book Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4245">
        <xs:annotation>
          <xs:documentation>Marine Equipment Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4245</lx:path>
            <lx:label>Marine Equipment Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4250">
        <xs:annotation>
          <xs:documentation>Clothing, Footwear and Personal Accessory Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4250</lx:path>
            <lx:label>Clothing, Footwear and Personal Accessory Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4251">
        <xs:annotation>
          <xs:documentation>Clothing Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4251</lx:path>
            <lx:label>Clothing Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4252">
        <xs:annotation>
          <xs:documentation>Footwear Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4252</lx:path>
            <lx:label>Footwear Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4253">
        <xs:annotation>
          <xs:documentation>Watch and Jewellery Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4253</lx:path>
            <lx:label>Watch and Jewellery Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4259">
        <xs:annotation>
          <xs:documentation>Other Personal Accessory Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4259</lx:path>
            <lx:label>Other Personal Accessory Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4260">
        <xs:annotation>
          <xs:documentation>Department Stores</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4260</lx:path>
            <lx:label>Department Stores</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4270">
        <xs:annotation>
          <xs:documentation>Pharmaceutical and Other Store-Based Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4270</lx:path>
            <lx:label>Pharmaceutical and Other Store-Based Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4271">
        <xs:annotation>
          <xs:documentation>Pharmaceutical, Cosmetic and Toiletry Goods Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4271</lx:path>
            <lx:label>Pharmaceutical, Cosmetic and Toiletry Goods Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4272">
        <xs:annotation>
          <xs:documentation>Stationery Goods Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4272</lx:path>
            <lx:label>Stationery Goods Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4273">
        <xs:annotation>
          <xs:documentation>Antique and Used Goods Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4273</lx:path>
            <lx:label>Antique and Used Goods Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4274">
        <xs:annotation>
          <xs:documentation>Flower Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4274</lx:path>
            <lx:label>Flower Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4279">
        <xs:annotation>
          <xs:documentation>Other Store-Based Retailing n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4279</lx:path>
            <lx:label>Other Store-Based Retailing n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4300">
        <xs:annotation>
          <xs:documentation>Non-Store Retailing and Retail Commission-Based Buying and/or Selling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4300</lx:path>
            <lx:label>Non-Store Retailing and Retail Commission-Based Buying and/or Selling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4310">
        <xs:annotation>
          <xs:documentation>Non-Store Retailing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4310</lx:path>
            <lx:label>Non-Store Retailing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4320">
        <xs:annotation>
          <xs:documentation>Retail Commission-Based Buying and/or Selling</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4320</lx:path>
            <lx:label>Retail Commission-Based Buying and/or Selling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4400">
        <xs:annotation>
          <xs:documentation>Accommodation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4400</lx:path>
            <lx:label>Accommodation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4500">
        <xs:annotation>
          <xs:documentation>Food and Beverage Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4500</lx:path>
            <lx:label>Food and Beverage Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4510">
        <xs:annotation>
          <xs:documentation>Cafes, Restaurants and Takeaway Food Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4510</lx:path>
            <lx:label>Cafes, Restaurants and Takeaway Food Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4511">
        <xs:annotation>
          <xs:documentation>Cafes and Restaurants</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4511</lx:path>
            <lx:label>Cafes and Restaurants</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4512">
        <xs:annotation>
          <xs:documentation>Takeaway Food Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4512</lx:path>
            <lx:label>Takeaway Food Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4513">
        <xs:annotation>
          <xs:documentation>Catering Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4513</lx:path>
            <lx:label>Catering Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4520">
        <xs:annotation>
          <xs:documentation>Pubs, Taverns and Bars</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4520</lx:path>
            <lx:label>Pubs, Taverns and Bars</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4530">
        <xs:annotation>
          <xs:documentation>Clubs (Hospitality)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4530</lx:path>
            <lx:label>Clubs (Hospitality)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4600">
        <xs:annotation>
          <xs:documentation>Road Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4600</lx:path>
            <lx:label>Road Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4610">
        <xs:annotation>
          <xs:documentation>Road Freight Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4610</lx:path>
            <lx:label>Road Freight Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4620">
        <xs:annotation>
          <xs:documentation>Road Passenger Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4620</lx:path>
            <lx:label>Road Passenger Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4621">
        <xs:annotation>
          <xs:documentation>Interurban and Rural Bus Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4621</lx:path>
            <lx:label>Interurban and Rural Bus Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4622">
        <xs:annotation>
          <xs:documentation>Urban Bus Transport (Including Tramway)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4622</lx:path>
            <lx:label>Urban Bus Transport (Including Tramway)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4623">
        <xs:annotation>
          <xs:documentation>Taxi and Other Road Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4623</lx:path>
            <lx:label>Taxi and Other Road Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4700">
        <xs:annotation>
          <xs:documentation>Rail Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4700</lx:path>
            <lx:label>Rail Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4710">
        <xs:annotation>
          <xs:documentation>Rail Freight Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4710</lx:path>
            <lx:label>Rail Freight Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4720">
        <xs:annotation>
          <xs:documentation>Rail Passenger Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4720</lx:path>
            <lx:label>Rail Passenger Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4800">
        <xs:annotation>
          <xs:documentation>Water Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4800</lx:path>
            <lx:label>Water Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4810">
        <xs:annotation>
          <xs:documentation>Water Freight Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4810</lx:path>
            <lx:label>Water Freight Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4820">
        <xs:annotation>
          <xs:documentation>Water Passenger Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4820</lx:path>
            <lx:label>Water Passenger Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4900">
        <xs:annotation>
          <xs:documentation>Air and Space Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.4900</lx:path>
            <lx:label>Air and Space Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5000">
        <xs:annotation>
          <xs:documentation>Other Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5000</lx:path>
            <lx:label>Other Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5010">
        <xs:annotation>
          <xs:documentation>Scenic and Sightseeing Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5010</lx:path>
            <lx:label>Scenic and Sightseeing Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5020">
        <xs:annotation>
          <xs:documentation>Pipeline and Other Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5020</lx:path>
            <lx:label>Pipeline and Other Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5021">
        <xs:annotation>
          <xs:documentation>Pipeline Transport</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5021</lx:path>
            <lx:label>Pipeline Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5029">
        <xs:annotation>
          <xs:documentation>Other Transport n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5029</lx:path>
            <lx:label>Other Transport n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5100">
        <xs:annotation>
          <xs:documentation>Postal and Courier Pick-up and Delivery Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5100</lx:path>
            <lx:label>Postal and Courier Pick-up and Delivery Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5101">
        <xs:annotation>
          <xs:documentation>Postal Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5101</lx:path>
            <lx:label>Postal Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5102">
        <xs:annotation>
          <xs:documentation>Courier Pick-up and Delivery Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5102</lx:path>
            <lx:label>Courier Pick-up and Delivery Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5200">
        <xs:annotation>
          <xs:documentation>Transport Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5200</lx:path>
            <lx:label>Transport Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5210">
        <xs:annotation>
          <xs:documentation>Water Transport Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5210</lx:path>
            <lx:label>Water Transport Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5211">
        <xs:annotation>
          <xs:documentation>Stevedoring Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5211</lx:path>
            <lx:label>Stevedoring Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5212">
        <xs:annotation>
          <xs:documentation>Port and Water Transport Terminal Operations</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5212</lx:path>
            <lx:label>Port and Water Transport Terminal Operations</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5219">
        <xs:annotation>
          <xs:documentation>Other Water Transport Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5219</lx:path>
            <lx:label>Other Water Transport Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5220">
        <xs:annotation>
          <xs:documentation>Airport Operations and Other Air Transport Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5220</lx:path>
            <lx:label>Airport Operations and Other Air Transport Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5290">
        <xs:annotation>
          <xs:documentation>Other Transport Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5290</lx:path>
            <lx:label>Other Transport Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5291">
        <xs:annotation>
          <xs:documentation>Customs Agency Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5291</lx:path>
            <lx:label>Customs Agency Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5292">
        <xs:annotation>
          <xs:documentation>Freight Forwarding Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5292</lx:path>
            <lx:label>Freight Forwarding Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5299">
        <xs:annotation>
          <xs:documentation>Other Transport Support Services n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5299</lx:path>
            <lx:label>Other Transport Support Services n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5300">
        <xs:annotation>
          <xs:documentation>Warehousing and Storage Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5300</lx:path>
            <lx:label>Warehousing and Storage Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5301">
        <xs:annotation>
          <xs:documentation>Grain Storage Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5301</lx:path>
            <lx:label>Grain Storage Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5309">
        <xs:annotation>
          <xs:documentation>Other Warehousing and Storage Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5309</lx:path>
            <lx:label>Other Warehousing and Storage Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5400">
        <xs:annotation>
          <xs:documentation>Publishing (except Internet and Music Publishing)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5400</lx:path>
            <lx:label>Publishing (except Internet and Music Publishing)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5410">
        <xs:annotation>
          <xs:documentation>Newspaper, Periodical, Book and Directory Publishing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5410</lx:path>
            <lx:label>Newspaper, Periodical, Book and Directory Publishing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5411">
        <xs:annotation>
          <xs:documentation>Newspaper Publishing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5411</lx:path>
            <lx:label>Newspaper Publishing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5412">
        <xs:annotation>
          <xs:documentation>Magazine and Other Periodical Publishing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5412</lx:path>
            <lx:label>Magazine and Other Periodical Publishing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5413">
        <xs:annotation>
          <xs:documentation>Book Publishing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5413</lx:path>
            <lx:label>Book Publishing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5414">
        <xs:annotation>
          <xs:documentation>Directory and Mailing List Publishing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5414</lx:path>
            <lx:label>Directory and Mailing List Publishing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5419">
        <xs:annotation>
          <xs:documentation>Other Publishing (except Software, Music and Internet)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5419</lx:path>
            <lx:label>Other Publishing (except Software, Music and Internet)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5420">
        <xs:annotation>
          <xs:documentation>Software Publishing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5420</lx:path>
            <lx:label>Software Publishing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5500">
        <xs:annotation>
          <xs:documentation>Motion Picture and Sound Recording Activities</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5500</lx:path>
            <lx:label>Motion Picture and Sound Recording Activities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5510">
        <xs:annotation>
          <xs:documentation>Motion Picture and Video Activities</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5510</lx:path>
            <lx:label>Motion Picture and Video Activities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5511">
        <xs:annotation>
          <xs:documentation>Motion Picture and Video Production</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5511</lx:path>
            <lx:label>Motion Picture and Video Production</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5512">
        <xs:annotation>
          <xs:documentation>Motion Picture and Video Distribution</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5512</lx:path>
            <lx:label>Motion Picture and Video Distribution</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5513">
        <xs:annotation>
          <xs:documentation>Motion Picture Exhibition</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5513</lx:path>
            <lx:label>Motion Picture Exhibition</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5514">
        <xs:annotation>
          <xs:documentation>Post-production Services and Other Motion Picture and Video Activities</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5514</lx:path>
            <lx:label>Post-production Services and Other Motion Picture and Video Activities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5520">
        <xs:annotation>
          <xs:documentation>Sound Recording and Music Publishing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5520</lx:path>
            <lx:label>Sound Recording and Music Publishing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5521">
        <xs:annotation>
          <xs:documentation>Music Publishing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5521</lx:path>
            <lx:label>Music Publishing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5522">
        <xs:annotation>
          <xs:documentation>Music and Other Sound Recording Activities</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5522</lx:path>
            <lx:label>Music and Other Sound Recording Activities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5600">
        <xs:annotation>
          <xs:documentation>Broadcasting (except Internet)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5600</lx:path>
            <lx:label>Broadcasting (except Internet)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5610">
        <xs:annotation>
          <xs:documentation>Radio Broadcasting</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5610</lx:path>
            <lx:label>Radio Broadcasting</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5620">
        <xs:annotation>
          <xs:documentation>Television Broadcasting</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5620</lx:path>
            <lx:label>Television Broadcasting</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5621">
        <xs:annotation>
          <xs:documentation>Free-to-Air Television Broadcasting</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5621</lx:path>
            <lx:label>Free-to-Air Television Broadcasting</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5622">
        <xs:annotation>
          <xs:documentation>Cable and Other Subscription Broadcasting</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5622</lx:path>
            <lx:label>Cable and Other Subscription Broadcasting</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5700">
        <xs:annotation>
          <xs:documentation>Internet Publishing and Broadcasting</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5700</lx:path>
            <lx:label>Internet Publishing and Broadcasting</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5800">
        <xs:annotation>
          <xs:documentation>Telecommunications Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5800</lx:path>
            <lx:label>Telecommunications Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5801">
        <xs:annotation>
          <xs:documentation>Wired Telecommunications Network Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5801</lx:path>
            <lx:label>Wired Telecommunications Network Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5802">
        <xs:annotation>
          <xs:documentation>Other Telecommunications Network Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5802</lx:path>
            <lx:label>Other Telecommunications Network Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5809">
        <xs:annotation>
          <xs:documentation>Other Telecommunications Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5809</lx:path>
            <lx:label>Other Telecommunications Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5900">
        <xs:annotation>
          <xs:documentation>Internet Service Providers, Web Search Portals and Data Processing Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5900</lx:path>
            <lx:label>Internet Service Providers, Web Search Portals and Data Processing Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5910">
        <xs:annotation>
          <xs:documentation>Internet Service Providers and Web Search Portals</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5910</lx:path>
            <lx:label>Internet Service Providers and Web Search Portals</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5920">
        <xs:annotation>
          <xs:documentation>Data Processing, Web Hosting and Electronic Information Storage Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5920</lx:path>
            <lx:label>Data Processing, Web Hosting and Electronic Information Storage Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5921">
        <xs:annotation>
          <xs:documentation>Data Processing and Web Hosting Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5921</lx:path>
            <lx:label>Data Processing and Web Hosting Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5922">
        <xs:annotation>
          <xs:documentation>Electronic Information Storage Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.5922</lx:path>
            <lx:label>Electronic Information Storage Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6000">
        <xs:annotation>
          <xs:documentation>Library and Other Information Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6000</lx:path>
            <lx:label>Library and Other Information Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6010">
        <xs:annotation>
          <xs:documentation>Libraries and Archives</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6010</lx:path>
            <lx:label>Libraries and Archives</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6020">
        <xs:annotation>
          <xs:documentation>Other Information Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6020</lx:path>
            <lx:label>Other Information Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6200">
        <xs:annotation>
          <xs:documentation>Finance</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6200</lx:path>
            <lx:label>Finance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6210">
        <xs:annotation>
          <xs:documentation>Central Banking</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6210</lx:path>
            <lx:label>Central Banking</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6220">
        <xs:annotation>
          <xs:documentation>Depository Financial Intermediation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6220</lx:path>
            <lx:label>Depository Financial Intermediation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6221">
        <xs:annotation>
          <xs:documentation>Banking</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6221</lx:path>
            <lx:label>Banking</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6222">
        <xs:annotation>
          <xs:documentation>Building Society Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6222</lx:path>
            <lx:label>Building Society Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6223">
        <xs:annotation>
          <xs:documentation>Credit Union Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6223</lx:path>
            <lx:label>Credit Union Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6229">
        <xs:annotation>
          <xs:documentation>Other Depository Financial Intermediation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6229</lx:path>
            <lx:label>Other Depository Financial Intermediation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6230">
        <xs:annotation>
          <xs:documentation>Non-Depository Financing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6230</lx:path>
            <lx:label>Non-Depository Financing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6240">
        <xs:annotation>
          <xs:documentation>Financial Asset Investing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6240</lx:path>
            <lx:label>Financial Asset Investing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6300">
        <xs:annotation>
          <xs:documentation>Insurance and Superannuation Funds</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6300</lx:path>
            <lx:label>Insurance and Superannuation Funds</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6310">
        <xs:annotation>
          <xs:documentation>Life Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6310</lx:path>
            <lx:label>Life Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6320">
        <xs:annotation>
          <xs:documentation>Health and General Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6320</lx:path>
            <lx:label>Health and General Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6321">
        <xs:annotation>
          <xs:documentation>Health Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6321</lx:path>
            <lx:label>Health Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6322">
        <xs:annotation>
          <xs:documentation>General Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6322</lx:path>
            <lx:label>General Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6330">
        <xs:annotation>
          <xs:documentation>Superannuation Funds</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6330</lx:path>
            <lx:label>Superannuation Funds</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6400">
        <xs:annotation>
          <xs:documentation>Auxiliary Finance and Insurance Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6400</lx:path>
            <lx:label>Auxiliary Finance and Insurance Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6410">
        <xs:annotation>
          <xs:documentation>Auxiliary Finance and Investment Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6410</lx:path>
            <lx:label>Auxiliary Finance and Investment Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6411">
        <xs:annotation>
          <xs:documentation>Financial Asset Broking Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6411</lx:path>
            <lx:label>Financial Asset Broking Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6419">
        <xs:annotation>
          <xs:documentation>Other Auxiliary Finance and Investment Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6419</lx:path>
            <lx:label>Other Auxiliary Finance and Investment Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6420">
        <xs:annotation>
          <xs:documentation>Auxiliary Insurance Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6420</lx:path>
            <lx:label>Auxiliary Insurance Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6600">
        <xs:annotation>
          <xs:documentation>Rental and Hiring Services (except Real Estate)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6600</lx:path>
            <lx:label>Rental and Hiring Services (except Real Estate)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6610">
        <xs:annotation>
          <xs:documentation>Motor Vehicle and Transport Equipment Rental and Hiring</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6610</lx:path>
            <lx:label>Motor Vehicle and Transport Equipment Rental and Hiring</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6611">
        <xs:annotation>
          <xs:documentation>Passenger Car Rental and Hiring</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6611</lx:path>
            <lx:label>Passenger Car Rental and Hiring</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6619">
        <xs:annotation>
          <xs:documentation>Other Motor Vehicle and Transport Equipment Rental and Hiring</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6619</lx:path>
            <lx:label>Other Motor Vehicle and Transport Equipment Rental and Hiring</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6620">
        <xs:annotation>
          <xs:documentation>Farm Animal and Bloodstock Leasing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6620</lx:path>
            <lx:label>Farm Animal and Bloodstock Leasing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6630">
        <xs:annotation>
          <xs:documentation>Other Goods and Equipment Rental and Hiring</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6630</lx:path>
            <lx:label>Other Goods and Equipment Rental and Hiring</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6631">
        <xs:annotation>
          <xs:documentation>Heavy Machinery and Scaffolding Rental and Hiring</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6631</lx:path>
            <lx:label>Heavy Machinery and Scaffolding Rental and Hiring</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6632">
        <xs:annotation>
          <xs:documentation>Video and Other Electronic Media Rental and Hiring</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6632</lx:path>
            <lx:label>Video and Other Electronic Media Rental and Hiring</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6639">
        <xs:annotation>
          <xs:documentation>Other Goods and Equipment Rental and Hiring n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6639</lx:path>
            <lx:label>Other Goods and Equipment Rental and Hiring n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6640">
        <xs:annotation>
          <xs:documentation>Non-Financial Intangible Assets (Except Copyrights) Leasing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6640</lx:path>
            <lx:label>Non-Financial Intangible Assets (Except Copyrights) Leasing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6700">
        <xs:annotation>
          <xs:documentation>Property Operators and Real Estate Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6700</lx:path>
            <lx:label>Property Operators and Real Estate Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6710">
        <xs:annotation>
          <xs:documentation>Property Operators</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6710</lx:path>
            <lx:label>Property Operators</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6711">
        <xs:annotation>
          <xs:documentation>Residential Property Operators</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6711</lx:path>
            <lx:label>Residential Property Operators</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6712">
        <xs:annotation>
          <xs:documentation>Non-Residential Property Operators</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6712</lx:path>
            <lx:label>Non-Residential Property Operators</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6720">
        <xs:annotation>
          <xs:documentation>Real Estate Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6720</lx:path>
            <lx:label>Real Estate Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6900">
        <xs:annotation>
          <xs:documentation>Professional, Scientific and Technical Services (Except Computer System Design and Related Services)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6900</lx:path>
            <lx:label>Professional, Scientific and Technical Services (Except Computer System Design and Related Services)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6910">
        <xs:annotation>
          <xs:documentation>Scientific Research Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6910</lx:path>
            <lx:label>Scientific Research Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6920">
        <xs:annotation>
          <xs:documentation>Architectural, Engineering and Technical Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6920</lx:path>
            <lx:label>Architectural, Engineering and Technical Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6921">
        <xs:annotation>
          <xs:documentation>Architectural Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6921</lx:path>
            <lx:label>Architectural Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6922">
        <xs:annotation>
          <xs:documentation>Surveying and Mapping Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6922</lx:path>
            <lx:label>Surveying and Mapping Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6923">
        <xs:annotation>
          <xs:documentation>Engineering Design and Engineering Consulting Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6923</lx:path>
            <lx:label>Engineering Design and Engineering Consulting Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6924">
        <xs:annotation>
          <xs:documentation>Other Specialised Design Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6924</lx:path>
            <lx:label>Other Specialised Design Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6925">
        <xs:annotation>
          <xs:documentation>Scientific Testing and Analysis Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6925</lx:path>
            <lx:label>Scientific Testing and Analysis Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6930">
        <xs:annotation>
          <xs:documentation>Legal and Accounting Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6930</lx:path>
            <lx:label>Legal and Accounting Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6931">
        <xs:annotation>
          <xs:documentation>Legal Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6931</lx:path>
            <lx:label>Legal Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6932">
        <xs:annotation>
          <xs:documentation>Accounting Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6932</lx:path>
            <lx:label>Accounting Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6940">
        <xs:annotation>
          <xs:documentation>Advertising Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6940</lx:path>
            <lx:label>Advertising Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6950">
        <xs:annotation>
          <xs:documentation>Market Research and Statistical Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6950</lx:path>
            <lx:label>Market Research and Statistical Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6960">
        <xs:annotation>
          <xs:documentation>Management and Related Consulting Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6960</lx:path>
            <lx:label>Management and Related Consulting Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6961">
        <xs:annotation>
          <xs:documentation>Corporate Head Office Management Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6961</lx:path>
            <lx:label>Corporate Head Office Management Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6962">
        <xs:annotation>
          <xs:documentation>Management Advice and Related Consulting Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6962</lx:path>
            <lx:label>Management Advice and Related Consulting Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6970">
        <xs:annotation>
          <xs:documentation>Veterinary Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6970</lx:path>
            <lx:label>Veterinary Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6990">
        <xs:annotation>
          <xs:documentation>Other Professional, Scientific and Technical Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6990</lx:path>
            <lx:label>Other Professional, Scientific and Technical Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6991">
        <xs:annotation>
          <xs:documentation>Professional Photographic Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6991</lx:path>
            <lx:label>Professional Photographic Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="6999">
        <xs:annotation>
          <xs:documentation>Other Professional, Scientific and Technical Services n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.6999</lx:path>
            <lx:label>Other Professional, Scientific and Technical Services n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7000">
        <xs:annotation>
          <xs:documentation>Computer System Design and Related Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7000</lx:path>
            <lx:label>Computer System Design and Related Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7200">
        <xs:annotation>
          <xs:documentation>Administrative Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7200</lx:path>
            <lx:label>Administrative Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7210">
        <xs:annotation>
          <xs:documentation>Employment Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7210</lx:path>
            <lx:label>Employment Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7211">
        <xs:annotation>
          <xs:documentation>Employment Placement and Recruitment Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7211</lx:path>
            <lx:label>Employment Placement and Recruitment Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7212">
        <xs:annotation>
          <xs:documentation>Labour Supply Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7212</lx:path>
            <lx:label>Labour Supply Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7220">
        <xs:annotation>
          <xs:documentation>Travel Agency and Tour Arrangement Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7220</lx:path>
            <lx:label>Travel Agency and Tour Arrangement Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7290">
        <xs:annotation>
          <xs:documentation>Other Administrative Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7290</lx:path>
            <lx:label>Other Administrative Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7291">
        <xs:annotation>
          <xs:documentation>Office Administrative Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7291</lx:path>
            <lx:label>Office Administrative Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7292">
        <xs:annotation>
          <xs:documentation>Document Preparation Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7292</lx:path>
            <lx:label>Document Preparation Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7293">
        <xs:annotation>
          <xs:documentation>Credit Reporting and Debt Collection Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7293</lx:path>
            <lx:label>Credit Reporting and Debt Collection Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7294">
        <xs:annotation>
          <xs:documentation>Call Centre Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7294</lx:path>
            <lx:label>Call Centre Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7299">
        <xs:annotation>
          <xs:documentation>Other Administrative Services n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7299</lx:path>
            <lx:label>Other Administrative Services n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7300">
        <xs:annotation>
          <xs:documentation>Building Cleaning, Pest Control and Other Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7300</lx:path>
            <lx:label>Building Cleaning, Pest Control and Other Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7310">
        <xs:annotation>
          <xs:documentation>Building Cleaning, Pest Control and Gardening Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7310</lx:path>
            <lx:label>Building Cleaning, Pest Control and Gardening Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7311">
        <xs:annotation>
          <xs:documentation>Building and Other Industrial Cleaning Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7311</lx:path>
            <lx:label>Building and Other Industrial Cleaning Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7312">
        <xs:annotation>
          <xs:documentation>Building Pest Control Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7312</lx:path>
            <lx:label>Building Pest Control Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7313">
        <xs:annotation>
          <xs:documentation>Gardening Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7313</lx:path>
            <lx:label>Gardening Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7320">
        <xs:annotation>
          <xs:documentation>Packaging Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7320</lx:path>
            <lx:label>Packaging Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7500">
        <xs:annotation>
          <xs:documentation>Public Administration</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7500</lx:path>
            <lx:label>Public Administration</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7510">
        <xs:annotation>
          <xs:documentation>Central Government Administration</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7510</lx:path>
            <lx:label>Central Government Administration</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7520">
        <xs:annotation>
          <xs:documentation>State Government Administration</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7520</lx:path>
            <lx:label>State Government Administration</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7530">
        <xs:annotation>
          <xs:documentation>Local Government Administration</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7530</lx:path>
            <lx:label>Local Government Administration</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7540">
        <xs:annotation>
          <xs:documentation>Justice</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7540</lx:path>
            <lx:label>Justice</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7550">
        <xs:annotation>
          <xs:documentation>Government Representation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7550</lx:path>
            <lx:label>Government Representation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7551">
        <xs:annotation>
          <xs:documentation>Domestic Government Representation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7551</lx:path>
            <lx:label>Domestic Government Representation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7552">
        <xs:annotation>
          <xs:documentation>Foreign Government Representation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7552</lx:path>
            <lx:label>Foreign Government Representation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7600">
        <xs:annotation>
          <xs:documentation>Defence</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7600</lx:path>
            <lx:label>Defence</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7700">
        <xs:annotation>
          <xs:documentation>Public Order, Safety and Regulatory Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7700</lx:path>
            <lx:label>Public Order, Safety and Regulatory Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7710">
        <xs:annotation>
          <xs:documentation>Public Order and Safety Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7710</lx:path>
            <lx:label>Public Order and Safety Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7711">
        <xs:annotation>
          <xs:documentation>Police Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7711</lx:path>
            <lx:label>Police Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7712">
        <xs:annotation>
          <xs:documentation>Investigation and Security Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7712</lx:path>
            <lx:label>Investigation and Security Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7713">
        <xs:annotation>
          <xs:documentation>Fire Protection and Other Emergency Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7713</lx:path>
            <lx:label>Fire Protection and Other Emergency Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7714">
        <xs:annotation>
          <xs:documentation>Correctional and Detention Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7714</lx:path>
            <lx:label>Correctional and Detention Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7719">
        <xs:annotation>
          <xs:documentation>Other Public Order and Safety Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7719</lx:path>
            <lx:label>Other Public Order and Safety Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="7720">
        <xs:annotation>
          <xs:documentation>Regulatory Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.7720</lx:path>
            <lx:label>Regulatory Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8000">
        <xs:annotation>
          <xs:documentation>Preschool and School Education</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8000</lx:path>
            <lx:label>Preschool and School Education</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8010">
        <xs:annotation>
          <xs:documentation>Preschool Education</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8010</lx:path>
            <lx:label>Preschool Education</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8020">
        <xs:annotation>
          <xs:documentation>School Education</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8020</lx:path>
            <lx:label>School Education</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8021">
        <xs:annotation>
          <xs:documentation>Primary Education</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8021</lx:path>
            <lx:label>Primary Education</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8022">
        <xs:annotation>
          <xs:documentation>Secondary Education</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8022</lx:path>
            <lx:label>Secondary Education</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8023">
        <xs:annotation>
          <xs:documentation>Combined Primary and Secondary Education</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8023</lx:path>
            <lx:label>Combined Primary and Secondary Education</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8024">
        <xs:annotation>
          <xs:documentation>Special School Education</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8024</lx:path>
            <lx:label>Special School Education</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8100">
        <xs:annotation>
          <xs:documentation>Tertiary Education</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8100</lx:path>
            <lx:label>Tertiary Education</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8101">
        <xs:annotation>
          <xs:documentation>Technical and Vocational Education and Training</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8101</lx:path>
            <lx:label>Technical and Vocational Education and Training</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8102">
        <xs:annotation>
          <xs:documentation>Higher Education</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8102</lx:path>
            <lx:label>Higher Education</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8200">
        <xs:annotation>
          <xs:documentation>Adult, Community and Other Education</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8200</lx:path>
            <lx:label>Adult, Community and Other Education</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8210">
        <xs:annotation>
          <xs:documentation>Adult, Community and Other Education</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8210</lx:path>
            <lx:label>Adult, Community and Other Education</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8211">
        <xs:annotation>
          <xs:documentation>Sports and Physical Recreation Instruction</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8211</lx:path>
            <lx:label>Sports and Physical Recreation Instruction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8212">
        <xs:annotation>
          <xs:documentation>Arts Education</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8212</lx:path>
            <lx:label>Arts Education</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8219">
        <xs:annotation>
          <xs:documentation>Adult, Community and Other Education n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8219</lx:path>
            <lx:label>Adult, Community and Other Education n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8220">
        <xs:annotation>
          <xs:documentation>Educational Support Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8220</lx:path>
            <lx:label>Educational Support Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8400">
        <xs:annotation>
          <xs:documentation>Hospitals</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8400</lx:path>
            <lx:label>Hospitals</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8401">
        <xs:annotation>
          <xs:documentation>Hospitals (Except Psychiatric Hospitals)</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8401</lx:path>
            <lx:label>Hospitals (Except Psychiatric Hospitals)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8402">
        <xs:annotation>
          <xs:documentation>Psychiatric Hospitals</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8402</lx:path>
            <lx:label>Psychiatric Hospitals</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8500">
        <xs:annotation>
          <xs:documentation>Medical and Other Health Care Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8500</lx:path>
            <lx:label>Medical and Other Health Care Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8510">
        <xs:annotation>
          <xs:documentation>Medical Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8510</lx:path>
            <lx:label>Medical Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8511">
        <xs:annotation>
          <xs:documentation>General Practice Medical Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8511</lx:path>
            <lx:label>General Practice Medical Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8512">
        <xs:annotation>
          <xs:documentation>Specialist Medical Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8512</lx:path>
            <lx:label>Specialist Medical Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8520">
        <xs:annotation>
          <xs:documentation>Pathology and Diagnostic Imaging Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8520</lx:path>
            <lx:label>Pathology and Diagnostic Imaging Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8530">
        <xs:annotation>
          <xs:documentation>Allied Health Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8530</lx:path>
            <lx:label>Allied Health Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8531">
        <xs:annotation>
          <xs:documentation>Dental Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8531</lx:path>
            <lx:label>Dental Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8532">
        <xs:annotation>
          <xs:documentation>Optometry and Optical Dispensing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8532</lx:path>
            <lx:label>Optometry and Optical Dispensing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8533">
        <xs:annotation>
          <xs:documentation>Physiotherapy Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8533</lx:path>
            <lx:label>Physiotherapy Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8534">
        <xs:annotation>
          <xs:documentation>Chiropractic and Osteopathic Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8534</lx:path>
            <lx:label>Chiropractic and Osteopathic Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8539">
        <xs:annotation>
          <xs:documentation>Other Allied Health Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8539</lx:path>
            <lx:label>Other Allied Health Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8590">
        <xs:annotation>
          <xs:documentation>Other Health Care Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8590</lx:path>
            <lx:label>Other Health Care Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8591">
        <xs:annotation>
          <xs:documentation>Ambulance Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8591</lx:path>
            <lx:label>Ambulance Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8599">
        <xs:annotation>
          <xs:documentation>Other Health Care Services n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8599</lx:path>
            <lx:label>Other Health Care Services n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8600">
        <xs:annotation>
          <xs:documentation>Residential Care Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8600</lx:path>
            <lx:label>Residential Care Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8601">
        <xs:annotation>
          <xs:documentation>Aged Care Residential Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8601</lx:path>
            <lx:label>Aged Care Residential Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8609">
        <xs:annotation>
          <xs:documentation>Other Residential Care Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8609</lx:path>
            <lx:label>Other Residential Care Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8700">
        <xs:annotation>
          <xs:documentation>Social Assistance Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8700</lx:path>
            <lx:label>Social Assistance Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8710">
        <xs:annotation>
          <xs:documentation>Child Care Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8710</lx:path>
            <lx:label>Child Care Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8790">
        <xs:annotation>
          <xs:documentation>Other Social Assistance Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8790</lx:path>
            <lx:label>Other Social Assistance Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8900">
        <xs:annotation>
          <xs:documentation>Heritage Activities</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8900</lx:path>
            <lx:label>Heritage Activities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8910">
        <xs:annotation>
          <xs:documentation>Museum Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8910</lx:path>
            <lx:label>Museum Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8920">
        <xs:annotation>
          <xs:documentation>Parks and Gardens Operations</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8920</lx:path>
            <lx:label>Parks and Gardens Operations</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8921">
        <xs:annotation>
          <xs:documentation>Zoological and Botanical Gardens Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8921</lx:path>
            <lx:label>Zoological and Botanical Gardens Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="8922">
        <xs:annotation>
          <xs:documentation>Nature Reserves and Conservation Parks Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.8922</lx:path>
            <lx:label>Nature Reserves and Conservation Parks Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9000">
        <xs:annotation>
          <xs:documentation>Creative and Performing Arts Activities</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9000</lx:path>
            <lx:label>Creative and Performing Arts Activities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9001">
        <xs:annotation>
          <xs:documentation>Performing Arts Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9001</lx:path>
            <lx:label>Performing Arts Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9002">
        <xs:annotation>
          <xs:documentation>Creative Artists, Musicians, Writers and Performers</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9002</lx:path>
            <lx:label>Creative Artists, Musicians, Writers and Performers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9003">
        <xs:annotation>
          <xs:documentation>Performing Arts Venue Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9003</lx:path>
            <lx:label>Performing Arts Venue Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9100">
        <xs:annotation>
          <xs:documentation>Sports and Recreation Activities</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9100</lx:path>
            <lx:label>Sports and Recreation Activities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9110">
        <xs:annotation>
          <xs:documentation>Sports and Physical Recreation Activities</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9110</lx:path>
            <lx:label>Sports and Physical Recreation Activities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9111">
        <xs:annotation>
          <xs:documentation>Health and Fitness Centres and Gymnasia Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9111</lx:path>
            <lx:label>Health and Fitness Centres and Gymnasia Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9112">
        <xs:annotation>
          <xs:documentation>Sports and Physical Recreation Clubs and Sports Professionals</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9112</lx:path>
            <lx:label>Sports and Physical Recreation Clubs and Sports Professionals</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9113">
        <xs:annotation>
          <xs:documentation>Sports and Physical Recreation Venues, Grounds and Facilities Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9113</lx:path>
            <lx:label>Sports and Physical Recreation Venues, Grounds and Facilities Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9114">
        <xs:annotation>
          <xs:documentation>Sports and Physical Recreation Administrative Service</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9114</lx:path>
            <lx:label>Sports and Physical Recreation Administrative Service</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9120">
        <xs:annotation>
          <xs:documentation>Horse and Dog Racing Activities</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9120</lx:path>
            <lx:label>Horse and Dog Racing Activities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9121">
        <xs:annotation>
          <xs:documentation>Horse and Dog Racing Administration and Track Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9121</lx:path>
            <lx:label>Horse and Dog Racing Administration and Track Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9129">
        <xs:annotation>
          <xs:documentation>Other Horse and Dog Racing Activities</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9129</lx:path>
            <lx:label>Other Horse and Dog Racing Activities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9130">
        <xs:annotation>
          <xs:documentation>Amusement and Other Recreation Activities</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9130</lx:path>
            <lx:label>Amusement and Other Recreation Activities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9131">
        <xs:annotation>
          <xs:documentation>Amusement Parks and Centres Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9131</lx:path>
            <lx:label>Amusement Parks and Centres Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9139">
        <xs:annotation>
          <xs:documentation>Amusement and Other Recreational Activities n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9139</lx:path>
            <lx:label>Amusement and Other Recreational Activities n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9200">
        <xs:annotation>
          <xs:documentation>Gambling Activities</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9200</lx:path>
            <lx:label>Gambling Activities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9201">
        <xs:annotation>
          <xs:documentation>Casino Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9201</lx:path>
            <lx:label>Casino Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9202">
        <xs:annotation>
          <xs:documentation>Lottery Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9202</lx:path>
            <lx:label>Lottery Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9209">
        <xs:annotation>
          <xs:documentation>Other Gambling Activities</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9209</lx:path>
            <lx:label>Other Gambling Activities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9400">
        <xs:annotation>
          <xs:documentation>Repair and Maintenance</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9400</lx:path>
            <lx:label>Repair and Maintenance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9410">
        <xs:annotation>
          <xs:documentation>Automotive Repair and Maintenance</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9410</lx:path>
            <lx:label>Automotive Repair and Maintenance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9411">
        <xs:annotation>
          <xs:documentation>Automotive Electrical Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9411</lx:path>
            <lx:label>Automotive Electrical Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9412">
        <xs:annotation>
          <xs:documentation>Automotive Body, Paint and Interior Repair</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9412</lx:path>
            <lx:label>Automotive Body, Paint and Interior Repair</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9419">
        <xs:annotation>
          <xs:documentation>Other Automotive Repair and Maintenance</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9419</lx:path>
            <lx:label>Other Automotive Repair and Maintenance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9420">
        <xs:annotation>
          <xs:documentation>Machinery and Equipment Repair and Maintenance</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9420</lx:path>
            <lx:label>Machinery and Equipment Repair and Maintenance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9421">
        <xs:annotation>
          <xs:documentation>Domestic Appliance Repair and Maintenance</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9421</lx:path>
            <lx:label>Domestic Appliance Repair and Maintenance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9422">
        <xs:annotation>
          <xs:documentation>Electronic (except Domestic Appliance) and Precision Equipment Repair</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9422</lx:path>
            <lx:label>Electronic (except Domestic Appliance) and Precision Equipment Repair</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9429">
        <xs:annotation>
          <xs:documentation>Other Machinery and Equipment Repair and Maintenance</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9429</lx:path>
            <lx:label>Other Machinery and Equipment Repair and Maintenance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9490">
        <xs:annotation>
          <xs:documentation>Other Repair and Maintenance</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9490</lx:path>
            <lx:label>Other Repair and Maintenance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9491">
        <xs:annotation>
          <xs:documentation>Clothing and Footwear Repair</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9491</lx:path>
            <lx:label>Clothing and Footwear Repair</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9499">
        <xs:annotation>
          <xs:documentation>Other Repair and Maintenance n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9499</lx:path>
            <lx:label>Other Repair and Maintenance n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9500">
        <xs:annotation>
          <xs:documentation>Personal and Other Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9500</lx:path>
            <lx:label>Personal and Other Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9510">
        <xs:annotation>
          <xs:documentation>Personal Care Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9510</lx:path>
            <lx:label>Personal Care Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9511">
        <xs:annotation>
          <xs:documentation>Hairdressing and Beauty Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9511</lx:path>
            <lx:label>Hairdressing and Beauty Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9512">
        <xs:annotation>
          <xs:documentation>Diet and Weight Reduction Centre Operation</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9512</lx:path>
            <lx:label>Diet and Weight Reduction Centre Operation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9520">
        <xs:annotation>
          <xs:documentation>Funeral, Crematorium and Cemetery Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9520</lx:path>
            <lx:label>Funeral, Crematorium and Cemetery Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9530">
        <xs:annotation>
          <xs:documentation>Other Personal Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9530</lx:path>
            <lx:label>Other Personal Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9531">
        <xs:annotation>
          <xs:documentation>Laundry and Dry-Cleaning Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9531</lx:path>
            <lx:label>Laundry and Dry-Cleaning Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9532">
        <xs:annotation>
          <xs:documentation>Photographic Film Processing</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9532</lx:path>
            <lx:label>Photographic Film Processing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9533">
        <xs:annotation>
          <xs:documentation>Parking Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9533</lx:path>
            <lx:label>Parking Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9534">
        <xs:annotation>
          <xs:documentation>Brothel Keeping and Prostitution Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9534</lx:path>
            <lx:label>Brothel Keeping and Prostitution Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9539">
        <xs:annotation>
          <xs:documentation>Other Personal Services n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9539</lx:path>
            <lx:label>Other Personal Services n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9540">
        <xs:annotation>
          <xs:documentation>Religious Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9540</lx:path>
            <lx:label>Religious Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9550">
        <xs:annotation>
          <xs:documentation>Civic, Professional and Other Interest Group Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9550</lx:path>
            <lx:label>Civic, Professional and Other Interest Group Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9551">
        <xs:annotation>
          <xs:documentation>Business and Professional Association Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9551</lx:path>
            <lx:label>Business and Professional Association Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9552">
        <xs:annotation>
          <xs:documentation>Labour Association Services</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9552</lx:path>
            <lx:label>Labour Association Services</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9559">
        <xs:annotation>
          <xs:documentation>Other Interest Group Services n.e.c.</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9559</lx:path>
            <lx:label>Other Interest Group Services n.e.c.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9600">
        <xs:annotation>
          <xs:documentation>Private Households Employing Staff and Undifferentiated Goods- and Service-Producing Activities of Households for Own Use</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9600</lx:path>
            <lx:label>Private Households Employing Staff and Undifferentiated Goods- and Service-Producing Activities of Households for Own Use</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9601">
        <xs:annotation>
          <xs:documentation>Private Households Employing Staff</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9601</lx:path>
            <lx:label>Private Households Employing Staff</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9602">
        <xs:annotation>
          <xs:documentation>Undifferentiated Goods-Producing Activities of Private Households for Own Use</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9602</lx:path>
            <lx:label>Undifferentiated Goods-Producing Activities of Private Households for Own Use</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="9603">
        <xs:annotation>
          <xs:documentation>Undifferentiated Service-Producing Activities of Private Households for Own Use</xs:documentation>
          <xs:appinfo>
            <lx:path>anzsicIndustryCodeList.9603</lx:path>
            <lx:label>Undifferentiated Service-Producing Activities of Private Households for Own Use</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="apiCertificationTypeList">
    <xs:annotation>
      <xs:documentation>apiCertificationTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>apiCertificationTypeList</lx:path>
        <lx:label>apiCertificationTypeList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="AAPI">
        <xs:annotation>
          <xs:documentation>An Associate of the API.</xs:documentation>
          <xs:appinfo>
            <lx:path>apiCertificationTypeList.AAPI</lx:path>
            <lx:label>AAPI</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FAPI">
        <xs:annotation>
          <xs:documentation>A Fellow of the API.</xs:documentation>
          <xs:appinfo>
            <lx:path>apiCertificationTypeList.FAPI</lx:path>
            <lx:label>FAPI</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LFAPI">
        <xs:annotation>
          <xs:documentation>A Life Fellow of the API.</xs:documentation>
          <xs:appinfo>
            <lx:path>apiCertificationTypeList.LFAPI</lx:path>
            <lx:label>LFAPI</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PMAPI">
        <xs:annotation>
          <xs:documentation>A Provisional Member of the API.</xs:documentation>
          <xs:appinfo>
            <lx:path>apiCertificationTypeList.PMAPI</lx:path>
            <lx:label>PMAPI</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="applicantsApplicationRoleTypeList">
    <xs:annotation>
      <xs:documentation>applicantsApplicationRoleTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>applicantsApplicationRoleTypeList</lx:path>
        <lx:label>applicantsApplicationRoleTypeList</lx:label>
        <li:transactions>CDA</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Joint">
        <xs:annotation>
          <xs:documentation>Applicants are jointly applying for the selected financial products.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicantsApplicationRoleTypeList.Joint</lx:path>
            <lx:label>Applicants are jointly applying for the selected financial products.</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Single">
        <xs:annotation>
          <xs:documentation>A single applicant is applying for the selected financial products.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicantsApplicationRoleTypeList.Single</lx:path>
            <lx:label>A single applicant is applying for the selected financial products.</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="applicantTypeList">
    <xs:annotation>
      <xs:documentation>applicantTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>applicantTypeList</lx:path>
        <lx:label>applicantTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Borrower">
        <xs:annotation>
          <xs:documentation>The party is applying in the capacity of a borrower on the loan.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicantTypeList.Borrower</lx:path>
            <lx:label>The party is applying in the capacity of a borrower on the loan.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Borrower and Guarantor">
        <xs:annotation>
          <xs:documentation>The party is applying in the capacity of both borrower and guarantor on the loan.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicantTypeList.Borrower and Guarantor</lx:path>
            <lx:label>Borrower and Guarantor</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Guarantor">
        <xs:annotation>
          <xs:documentation>The party is applying in the capacity of a guarantor on the loan.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicantTypeList.Guarantor</lx:path>
            <lx:label>The party is applying in the capacity of a guarantor on the loan.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="applicationDocTypeList">
    <xs:annotation>
      <xs:documentation>applicationDocTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>applicationDocTypeList</lx:path>
        <lx:label>applicationDocTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Alt Doc">
        <xs:annotation>
          <xs:documentation>Indicates that the applicants will provide alternative/non-traditional documentation in support of their financial position.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicationDocTypeList.Alt Doc</lx:path>
            <lx:label>Alt Doc</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Full Doc">
        <xs:annotation>
          <xs:documentation>Indicates that the applicants will provide documentary evidence in support of their financial position, including income.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicationDocTypeList.Full Doc</lx:path>
            <lx:label>Indicates that the applicants will provide documentary evidence in support of their financial position, including income.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lo Doc">
        <xs:annotation>
          <xs:documentation>Indicates that the applicants will provide limited documentation in support of their financial position, but will instead typically provide a declaration as to financial position in particular income.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicationDocTypeList.Lo Doc</lx:path>
            <lx:label>Indicates that the applicants will provide limited documentation in support of their financial position, but will instead typically provide a declaration as to financial position in particular income.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="No Doc">
        <xs:annotation>
          <xs:documentation>Indicates that the applicants will not provide documentary evidence in support of their financial position. Also known as an asset lend.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicationDocTypeList.No Doc</lx:path>
            <lx:label>Indicates that the applicants will not provide documentary evidence in support of their financial position. Also known as an asset lend.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="applicationTypeList">
    <xs:annotation>
      <xs:documentation>applicationTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>applicationTypeList</lx:path>
        <lx:label>applicationTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Deposit Account">
        <xs:annotation>
          <xs:documentation>The application is for a deposit account.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicationTypeList.Deposit Account</lx:path>
            <lx:label>Deposit Account</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Facility Drawdown">
        <xs:annotation>
          <xs:documentation>The application is for a loan within an existing approved umbrella master facility.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicationTypeList.Facility Drawdown</lx:path>
            <lx:label>The application is for a loan within an existing approved umbrella master facility.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lease">
        <xs:annotation>
          <xs:documentation>Application is for a finance or operating lease.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicationTypeList.Lease</lx:path>
            <lx:label>Application is for a finance or operating lease.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan">
        <xs:annotation>
          <xs:documentation>Application is for a standard loan facility.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicationTypeList.Loan</lx:path>
            <lx:label>Application is for a standard loan facility.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Variation">
        <xs:annotation>
          <xs:documentation>The application is to vary one or more existing loan accounts.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicationTypeList.Loan Variation</lx:path>
            <lx:label>The application is to vary one or more existing loan accounts.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Overdraft">
        <xs:annotation>
          <xs:documentation>Loan application is for an overdraft or similar non-amortising loan facility.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicationTypeList.Overdraft</lx:path>
            <lx:label>Loan application is for an overdraft or similar non-amortising loan facility.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Preapproval">
        <xs:annotation>
          <xs:documentation>Loan application is for an evaluation of a potential borrower by a lender that determines whether the borrower qualifies for a loan from the lender, or the maximum amount that the lender would be willing to lend.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicationTypeList.Preapproval</lx:path>
            <lx:label>Loan application is for preapproval of applicantion.</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Prequalification">
        <xs:annotation>
          <xs:documentation>Loan application is for prequalification of applicants, usually prior to identification of the security property to be purchased using the loan funds.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicationTypeList.Prequalification</lx:path>
            <lx:label>Loan application is for prequalification of applicants, usually prior to identification of the security property to be purchased using the loan funds.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unsecured">
        <xs:annotation>
          <xs:documentation>Loan application is for an unsecured loan facility.</xs:documentation>
          <xs:appinfo>
            <lx:path>applicationTypeList.Unsecured</lx:path>
            <lx:label>Loan application is for an unsecured loan facility.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="architecturalStyleList">
    <xs:annotation>
      <xs:documentation>A list of the architectural styles of real estate property.</xs:documentation>
      <xs:appinfo>
        <lx:path>architecturalStyleList</lx:path>
        <lx:label>architecturalStyleList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Art Deco">
        <xs:annotation>
          <xs:documentation>Art Deco</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Art Deco</lx:path>
            <lx:label>Art Deco</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Californian Bungalow">
        <xs:annotation>
          <xs:documentation>Californian Bungalow</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Californian Bungalow</lx:path>
            <lx:label>Californian Bungalow</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Colonial">
        <xs:annotation>
          <xs:documentation>Colonial</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Colonial</lx:path>
            <lx:label>Colonial</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Contemporary">
        <xs:annotation>
          <xs:documentation>Contemporary</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Contemporary</lx:path>
            <lx:label>Contemporary</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Early Victorian">
        <xs:annotation>
          <xs:documentation>Early Victorian</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Early Victorian</lx:path>
            <lx:label>Early Victorian</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Edwardian">
        <xs:annotation>
          <xs:documentation>Edwardian</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Edwardian</lx:path>
            <lx:label>Edwardian</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Federation">
        <xs:annotation>
          <xs:documentation>Federation</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Federation</lx:path>
            <lx:label>Federation</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Inter War">
        <xs:annotation>
          <xs:documentation>Inter War</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Inter War</lx:path>
            <lx:label>Inter War</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Late Victorian">
        <xs:annotation>
          <xs:documentation>Late Victorian</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Late Victorian</lx:path>
            <lx:label>Late Victorian</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mid Century Modern">
        <xs:annotation>
          <xs:documentation>Mid Century Modern</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Mid Century Modern</lx:path>
            <lx:label>Mid Century Modern</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Post War">
        <xs:annotation>
          <xs:documentation>Post War</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Post War</lx:path>
            <lx:label>Post War</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Prefabricated">
        <xs:annotation>
          <xs:documentation>Prefabricated</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Prefabricated</lx:path>
            <lx:label>Prefabricated</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Queen Anne">
        <xs:annotation>
          <xs:documentation>Queen Anne</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Queen Anne</lx:path>
            <lx:label>Queen Anne</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Queenslander">
        <xs:annotation>
          <xs:documentation>Queenslander</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Queenslander</lx:path>
            <lx:label>Queenslander</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ranch Style">
        <xs:annotation>
          <xs:documentation>Ranch Style</xs:documentation>
          <xs:appinfo>
            <lx:path>architecturalStyleList.Ranch Style</lx:path>
            <lx:label>Ranch Style</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="assessmentTypeApplicationInstructionsList">
    <xs:annotation>
      <xs:documentation>assessmentTypeApplicationInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>assessmentTypeApplicationInstructionsList</lx:path>
        <lx:label>assessmentTypeApplicationInstructionsList</lx:label>
        <li:transactions>CAL,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Full">
        <xs:annotation>
          <xs:documentation>Full</xs:documentation>
          <xs:appinfo>
            <lx:path>assessmentTypeApplicationInstructionsList.Full</lx:path>
            <lx:label>Full</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Indicative">
        <xs:annotation>
          <xs:documentation>Indicative</xs:documentation>
          <xs:appinfo>
            <lx:path>assessmentTypeApplicationInstructionsList.Indicative</lx:path>
            <lx:label>Indicative</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Quote">
        <xs:annotation>
          <xs:documentation>Quote</xs:documentation>
          <xs:appinfo>
            <lx:path>assessmentTypeApplicationInstructionsList.Quote</lx:path>
            <lx:label>Quote</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="assetTransactionList">
    <xs:annotation>
      <xs:documentation>assetTransactionList</xs:documentation>
      <xs:appinfo>
        <lx:path>assetTransactionList</lx:path>
        <lx:label>assetTransactionList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Owns">
        <xs:annotation>
          <xs:documentation>The property is already owned by the parties, without a mortgage.</xs:documentation>
          <xs:appinfo>
            <lx:path>assetTransactionList.Owns</lx:path>
            <lx:label>Owns</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Owns Existing Mortgage">
        <xs:annotation>
          <xs:documentation>The property is already owned by the parties, with a mortgage.</xs:documentation>
          <xs:appinfo>
            <lx:path>assetTransactionList.Owns Existing Mortgage</lx:path>
            <lx:label>Owns Existing Mortgage</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchasing">
        <xs:annotation>
          <xs:documentation>The property is being purchased by the parties specified as the owners of the property.</xs:documentation>
          <xs:appinfo>
            <lx:path>assetTransactionList.Purchasing</lx:path>
            <lx:label>The property is being purchased by the parties specified as the owners of the property.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sold">
        <xs:annotation>
          <xs:documentation>The property has recently been sold.</xs:documentation>
          <xs:appinfo>
            <lx:path>assetTransactionList.Sold</lx:path>
            <lx:label>The property has recently been sold.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="auDialingCodeList">
    <xs:annotation>
      <xs:documentation>auDialingCodeList</xs:documentation>
      <xs:appinfo>
        <lx:path>auDialingCodeList</lx:path>
        <lx:label>auDialingCodeList</lx:label>
        <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="02">
        <xs:annotation>
          <xs:documentation>02</xs:documentation>
          <xs:appinfo>
            <lx:path>auDialingCodeList.02</lx:path>
            <lx:label>02</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="03">
        <xs:annotation>
          <xs:documentation>03</xs:documentation>
          <xs:appinfo>
            <lx:path>auDialingCodeList.03</lx:path>
            <lx:label>03</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="04">
        <xs:annotation>
          <xs:documentation>04</xs:documentation>
          <xs:appinfo>
            <lx:path>auDialingCodeList.04</lx:path>
            <lx:label>04</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="05">
        <xs:annotation>
          <xs:documentation>05</xs:documentation>
          <xs:appinfo>
            <lx:path>auDialingCodeList.05</lx:path>
            <lx:label>05</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="07">
        <xs:annotation>
          <xs:documentation>07</xs:documentation>
          <xs:appinfo>
            <lx:path>auDialingCodeList.07</lx:path>
            <lx:label>07</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="08">
        <xs:annotation>
          <xs:documentation>08</xs:documentation>
          <xs:appinfo>
            <lx:path>auDialingCodeList.08</lx:path>
            <lx:label>08</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="13">
        <xs:annotation>
          <xs:documentation>13</xs:documentation>
          <xs:appinfo>
            <lx:path>auDialingCodeList.13</lx:path>
            <lx:label>13</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1300">
        <xs:annotation>
          <xs:documentation>1300</xs:documentation>
          <xs:appinfo>
            <lx:path>auDialingCodeList.1300</lx:path>
            <lx:label>1300</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="180">
        <xs:annotation>
          <xs:documentation>180</xs:documentation>
          <xs:appinfo>
            <lx:path>auDialingCodeList.180</lx:path>
            <lx:label>180</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1800">
        <xs:annotation>
          <xs:documentation>1800</xs:documentation>
          <xs:appinfo>
            <lx:path>auDialingCodeList.1800</lx:path>
            <lx:label>1800</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="auStateList">
    <xs:annotation>
      <xs:documentation>auStateList</xs:documentation>
      <xs:appinfo>
        <lx:path>auStateList</lx:path>
        <lx:label>auStateList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="ACT">
        <xs:annotation>
          <xs:documentation>ACT</xs:documentation>
          <xs:appinfo>
            <lx:path>auStateList.ACT</lx:path>
            <lx:label>ACT</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NSW">
        <xs:annotation>
          <xs:documentation>NSW</xs:documentation>
          <xs:appinfo>
            <lx:path>auStateList.NSW</lx:path>
            <lx:label>NSW</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NT">
        <xs:annotation>
          <xs:documentation>NT</xs:documentation>
          <xs:appinfo>
            <lx:path>auStateList.NT</lx:path>
            <lx:label>NT</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="QLD">
        <xs:annotation>
          <xs:documentation>QLD</xs:documentation>
          <xs:appinfo>
            <lx:path>auStateList.QLD</lx:path>
            <lx:label>QLD</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SA">
        <xs:annotation>
          <xs:documentation>SA</xs:documentation>
          <xs:appinfo>
            <lx:path>auStateList.SA</lx:path>
            <lx:label>SA</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TAS">
        <xs:annotation>
          <xs:documentation>TAS</xs:documentation>
          <xs:appinfo>
            <lx:path>auStateList.TAS</lx:path>
            <lx:label>TAS</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="VIC">
        <xs:annotation>
          <xs:documentation>VIC</xs:documentation>
          <xs:appinfo>
            <lx:path>auStateList.VIC</lx:path>
            <lx:label>VIC</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="WA">
        <xs:annotation>
          <xs:documentation>WA</xs:documentation>
          <xs:appinfo>
            <lx:path>auStateList.WA</lx:path>
            <lx:label>WA</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="balloonRVInputPatternList">
    <xs:annotation>
      <xs:documentation>balloonRVInputPatternList</xs:documentation>
      <xs:appinfo>
        <lx:path>balloonRVInputPatternList</lx:path>
        <lx:label>balloonRVInputPatternList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Amount">
        <xs:annotation>
          <xs:documentation>Amount</xs:documentation>
          <xs:appinfo>
            <lx:path>balloonRVInputPatternList.Amount</lx:path>
            <lx:label>Amount</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Percent">
        <xs:annotation>
          <xs:documentation>Percent</xs:documentation>
          <xs:appinfo>
            <lx:path>balloonRVInputPatternList.Percent</lx:path>
            <lx:label>Percent</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="benefitEnquiriesList">
    <xs:annotation>
      <xs:documentation>benefitEnquiriesList</xs:documentation>
      <xs:appinfo>
        <lx:path>benefitEnquiriesList</lx:path>
        <lx:label>benefitEnquiriesList</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other.</xs:documentation>
          <xs:appinfo>
            <lx:path>benefitEnquiriesList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Proportion in Assets Purchased">
        <xs:annotation>
          <xs:documentation>The co-applicant will acquire a reasonably proportionate legal or equitable interest in assets purchased with the loan funds.</xs:documentation>
          <xs:appinfo>
            <lx:path>benefitEnquiriesList.Proportion in Assets Purchased</lx:path>
            <lx:label>Proportion in Assets Purchased</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Repay Debts or Other Obligations">
        <xs:annotation>
          <xs:documentation>A reasonable portion of the loans funds will be used to repay debts or other obligations owed by the co-applicant.</xs:documentation>
          <xs:appinfo>
            <lx:path>benefitEnquiriesList.Repay Debts or Other Obligations</lx:path>
            <lx:label>Repay Debts or Other Obligations</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="builderTypeList">
    <xs:annotation>
      <xs:documentation>builderTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>builderTypeList</lx:path>
        <lx:label>builderTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Licensed Builder">
        <xs:annotation>
          <xs:documentation>Licensed Builder</xs:documentation>
          <xs:appinfo>
            <lx:path>builderTypeList.Licensed Builder</lx:path>
            <lx:label>Licensed Builder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Owner Builder">
        <xs:annotation>
          <xs:documentation>Owner Builder</xs:documentation>
          <xs:appinfo>
            <lx:path>builderTypeList.Owner Builder</lx:path>
            <lx:label>Owner Builder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unlicensed Builder">
        <xs:annotation>
          <xs:documentation>Unlicensed Builder</xs:documentation>
          <xs:appinfo>
            <lx:path>builderTypeList.Unlicensed Builder</lx:path>
            <lx:label>Unlicensed Builder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="builtDateTypeList">
    <xs:annotation>
      <xs:documentation>builtDateTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>builtDateTypeList</lx:path>
        <lx:label>builtDateTypeList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Built About">
        <xs:annotation>
          <xs:documentation>Built About</xs:documentation>
          <xs:appinfo>
            <lx:path>builtDateTypeList.Built About</lx:path>
            <lx:label>Built About</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Year Built">
        <xs:annotation>
          <xs:documentation>Year Built</xs:documentation>
          <xs:appinfo>
            <lx:path>builtDateTypeList.Year Built</lx:path>
            <lx:label>Year Built</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="businessChannelTypeList">
    <xs:annotation>
      <xs:documentation>businessChannelTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>businessChannelTypeList</lx:path>
        <lx:label>businessChannelTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Bank">
        <xs:annotation>
          <xs:documentation>Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>businessChannelTypeList.Bank</lx:path>
            <lx:label>Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Building Society">
        <xs:annotation>
          <xs:documentation>Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>businessChannelTypeList.Building Society</lx:path>
            <lx:label>Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Union">
        <xs:annotation>
          <xs:documentation>Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>businessChannelTypeList.Credit Union</lx:path>
            <lx:label>Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mortgage Manager">
        <xs:annotation>
          <xs:documentation>Mortgage Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>businessChannelTypeList.Mortgage Manager</lx:path>
            <lx:label>Mortgage Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mutual Bank">
        <xs:annotation>
          <xs:documentation>Mutual Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>businessChannelTypeList.Mutual Bank</lx:path>
            <lx:label>Mutual Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Non Bank Lender">
        <xs:annotation>
          <xs:documentation>Non Bank Lender</xs:documentation>
          <xs:appinfo>
            <lx:path>businessChannelTypeList.Non Bank Lender</lx:path>
            <lx:label>Non Bank Lender</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="businessStructureExcludingTrustList">
    <xs:annotation>
      <xs:documentation>A list of the types of business structure for an enterprise, excluding trusts.</xs:documentation>
      <xs:appinfo>
        <lx:path>businessStructureExcludingTrustList</lx:path>
        <lx:label>Business Structure Excluding Trusts</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Company">
        <xs:annotation>
          <xs:documentation>A company is a separate legal entity, unlike a sole trader or a partnership structure. This means the company has the same rights as a natural person and can incur debt, sue and be sued. The company's owners (the shareholders) can limit their personal liability and are generally not liable for company debts.</xs:documentation>
          <xs:appinfo>
            <lx:path>businessStructureExcludingTrustList.Company</lx:path>
            <lx:label>Company</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Partnership">
        <xs:annotation>
          <xs:documentation>A partnership is a type of business structure that involves a number of entities who carry on a business together.</xs:documentation>
          <xs:appinfo>
            <lx:path>businessStructureExcludingTrustList.Partnership</lx:path>
            <lx:label>Partnership</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sole Trader">
        <xs:annotation>
          <xs:documentation>A sole trader business structure is a person trading as the individual legally responsible for all aspects of the business.</xs:documentation>
          <xs:appinfo>
            <lx:path>businessStructureExcludingTrustList.Sole Trader</lx:path>
            <lx:label>Sole Trader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="businessStructureFullList">
    <xs:annotation>
      <xs:documentation>A list of the types of business structure for an enterprise.</xs:documentation>
      <xs:appinfo>
        <lx:path>businessStructureFullList</lx:path>
        <lx:label>Business Structure Full List</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Company">
        <xs:annotation>
          <xs:documentation>A company is a separate legal entity, unlike a sole trader or a partnership structure. This means the company has the same rights as a natural person and can incur debt, sue and be sued. The company's owners (the shareholders) can limit their personal liability and are generally not liable for company debts.</xs:documentation>
          <xs:appinfo>
            <lx:path>businessStructureFullList.Company</lx:path>
            <lx:label>Company</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Partnership">
        <xs:annotation>
          <xs:documentation>A partnership is a type of business structure that involves a number of entities who carry on a business together.</xs:documentation>
          <xs:appinfo>
            <lx:path>businessStructureFullList.Partnership</lx:path>
            <lx:label>Partnership</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sole Trader">
        <xs:annotation>
          <xs:documentation>A sole trader business structure is a person trading as the individual legally responsible for all aspects of the business.</xs:documentation>
          <xs:appinfo>
            <lx:path>businessStructureFullList.Sole Trader</lx:path>
            <lx:label>Sole Trader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trust">
        <xs:annotation>
          <xs:documentation>A Trust is an obligation imposed on a person - a trustee - to hold property or assets (such as business assets) for the benefit of others, known as beneficiaries.</xs:documentation>
          <xs:appinfo>
            <lx:path>businessStructureFullList.Trust</lx:path>
            <lx:label>Sole Trader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="calculationBasisList">
    <xs:annotation>
      <xs:documentation>A list of amounts that can be used as the base value to calculate a fee, when the fee is given as a percentage rate.</xs:documentation>
      <xs:appinfo>
        <lx:path>calculationBasisList</lx:path>
        <lx:label>calculationBasisList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Balance">
        <xs:annotation>
          <xs:documentation>The balance of the loan amount is to be used as the base value to calculate the fee amount.</xs:documentation>
          <xs:appinfo>
            <lx:path>calculationBasisList.Balance</lx:path>
            <lx:label>Balance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Limit">
        <xs:annotation>
          <xs:documentation>The limit of the loan amount is to be used as the base value to calculate the fee amount.</xs:documentation>
          <xs:appinfo>
            <lx:path>calculationBasisList.Limit</lx:path>
            <lx:label>Limit</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="capacityTypeList">
    <xs:annotation>
      <xs:documentation>The capacity in which a signatory has signed a document.</xs:documentation>
      <xs:appinfo>
        <lx:path>capacityTypeList</lx:path>
        <lx:label>capacityTypeList</lx:label>
        <li:transactions>CAL,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Banker">
        <xs:annotation>
          <xs:documentation>Banker</xs:documentation>
          <xs:appinfo>
            <lx:path>capacityTypeList.Banker</lx:path>
            <lx:label>Banker</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Borrower">
        <xs:annotation>
          <xs:documentation>Borrower</xs:documentation>
          <xs:appinfo>
            <lx:path>capacityTypeList.Borrower</lx:path>
            <lx:label>Borrower</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Director">
        <xs:annotation>
          <xs:documentation>Director</xs:documentation>
          <xs:appinfo>
            <lx:path>capacityTypeList.Director</lx:path>
            <lx:label>Director</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Guarantor">
        <xs:annotation>
          <xs:documentation>Guarantor</xs:documentation>
          <xs:appinfo>
            <lx:path>capacityTypeList.Guarantor</lx:path>
            <lx:label>Guarantor</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="System">
        <xs:annotation>
          <xs:documentation>System</xs:documentation>
          <xs:appinfo>
            <lx:path>capacityTypeList.System</lx:path>
            <lx:label>System</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trustee">
        <xs:annotation>
          <xs:documentation>Trustee</xs:documentation>
          <xs:appinfo>
            <lx:path>capacityTypeList.Trustee</lx:path>
            <lx:label>Trustee</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="cleaningEquipmentTypeList">
    <xs:annotation>
      <xs:documentation>cleaningEquipmentTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>cleaningEquipmentTypeList</lx:path>
        <lx:label>cleaningEquipmentTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Dry Cleaning">
        <xs:annotation>
          <xs:documentation>Dry Cleaning</xs:documentation>
          <xs:appinfo>
            <lx:path>cleaningEquipmentTypeList.Dry Cleaning</lx:path>
            <lx:label>Dry Cleaning</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>cleaningEquipmentTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vacuum Equipment">
        <xs:annotation>
          <xs:documentation>Vacuum Equipment</xs:documentation>
          <xs:appinfo>
            <lx:path>cleaningEquipmentTypeList.Vacuum Equipment</lx:path>
            <lx:label>Vacuum Equipment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Washers and Dryers">
        <xs:annotation>
          <xs:documentation>Washers and Dryers</xs:documentation>
          <xs:appinfo>
            <lx:path>cleaningEquipmentTypeList.Washers and Dryers</lx:path>
            <lx:label>Washers and Dryers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Waste Bins">
        <xs:annotation>
          <xs:documentation>Waste Bins</xs:documentation>
          <xs:appinfo>
            <lx:path>cleaningEquipmentTypeList.Waste Bins</lx:path>
            <lx:label>Waste Bins</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="clearingFromOtherSourceList">
    <xs:annotation>
      <xs:documentation>A list of options available for selecting the source of funds from a third party institution used to clear an existing liability (as opposed to funds from the loan under application).</xs:documentation>
      <xs:appinfo>
        <lx:path>clearingFromOtherSourceList</lx:path>
        <lx:label>Clearing From Other Source List</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="No">
        <xs:annotation>
          <xs:documentation>No part of the existing loan will be paid out using loan funds from a third party institution.</xs:documentation>
          <xs:appinfo>
            <lx:path>clearingFromOtherSourceList.No</lx:path>
            <lx:label>No: None of the existing loan will be paid out using loan funds from a third party institution.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Partial">
        <xs:annotation>
          <xs:documentation>The existing loan will be partially paid out using loan funds being applied for.</xs:documentation>
          <xs:appinfo>
            <lx:path>clearingFromOtherSourceList.Partial</lx:path>
            <lx:label>Partial: The existing loan will be partially paid out using funds from a third party institution.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Yes">
        <xs:annotation>
          <xs:documentation>The existing loan will be completely paid out using loan funds from a third party institution.</xs:documentation>
          <xs:appinfo>
            <lx:path>clearingFromOtherSourceList.Yes</lx:path>
            <lx:label>Yes: The existing loan will be completely paid out using loan funds from a third party institution.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="clearingFromThisLoanList">
    <xs:annotation>
      <xs:documentation>clearingFromThisLoanList</xs:documentation>
      <xs:appinfo>
        <lx:path>clearingFromThisLoanList</lx:path>
        <lx:label>clearingFromThisLoanList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="No">
        <xs:annotation>
          <xs:documentation>No part of the existing loan will be paid out using loan funds being applied for.</xs:documentation>
          <xs:appinfo>
            <lx:path>clearingFromThisLoanList.No</lx:path>
            <lx:label>No part of the existing loan will be paid out using loan funds being applied for.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Partial">
        <xs:annotation>
          <xs:documentation>The existing loan will be partially paid out using loan funds being applied for.</xs:documentation>
          <xs:appinfo>
            <lx:path>clearingFromThisLoanList.Partial</lx:path>
            <lx:label>The existing loan will be partially paid out using loan funds being applied for.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Yes">
        <xs:annotation>
          <xs:documentation>The existing loan will be completely paid out using loan funds being applied for.</xs:documentation>
          <xs:appinfo>
            <lx:path>clearingFromThisLoanList.Yes</lx:path>
            <lx:label>The existing loan will be completely paid out using loan funds being applied for.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="coApplicantBenefitList">
    <xs:annotation>
      <xs:documentation>coApplicantBenefitList</xs:documentation>
      <xs:appinfo>
        <lx:path>coApplicantBenefitList</lx:path>
        <lx:label>coApplicantBenefitList</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="No Substantial Benefit">
        <xs:annotation>
          <xs:documentation>The co-applicant has confirmed that they will not receive a substantial benefit from the loan.</xs:documentation>
          <xs:appinfo>
            <lx:path>coApplicantBenefitList.No Substantial Benefit</lx:path>
            <lx:label>No Substantial Benefit</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="No Substantial Benefit and Proceed">
        <xs:annotation>
          <xs:documentation>The co-applicant has confirmed that they will not receive a substantial benefit from the loan and have confirmed they still want to proceed as a co-applicant.</xs:documentation>
          <xs:appinfo>
            <lx:path>coApplicantBenefitList.No Substantial Benefit and Proceed</lx:path>
            <lx:label>No Substantial Benefit and Proceed</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Substantial Benefit">
        <xs:annotation>
          <xs:documentation>The co-applicant has confirmed that they will receive a substantial benefit from the loan.</xs:documentation>
          <xs:appinfo>
            <lx:path>coApplicantBenefitList.Substantial Benefit</lx:path>
            <lx:label>Substantial Benefit</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="collateralClassList">
    <xs:annotation>
      <xs:documentation>The collateral class types that can be selected as a general description of the property covered by a PPSR registration.</xs:documentation>
      <xs:appinfo>
        <lx:path>collateralClassList</lx:path>
        <lx:label>collateralClassList</lx:label>
        <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Account">
        <xs:annotation>
          <xs:documentation>Account</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Account</lx:path>
            <lx:label>Account</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Agriculture">
        <xs:annotation>
          <xs:documentation>Agriculture</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Agriculture</lx:path>
            <lx:label>Agriculture</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Aircraft">
        <xs:annotation>
          <xs:documentation>Aircraft</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Aircraft</lx:path>
            <lx:label>Aircraft</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Aircraft Engine">
        <xs:annotation>
          <xs:documentation>Aircraft Engine</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Aircraft Engine</lx:path>
            <lx:label>Aircraft Engine</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Airframe">
        <xs:annotation>
          <xs:documentation>Airframe</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Airframe</lx:path>
            <lx:label>Airframe</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="All Present and After Acquired Property With Exception">
        <xs:annotation>
          <xs:documentation>All Present and After Acquired Property With Exception</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.All Present and After Acquired Property With Exception</lx:path>
            <lx:label>All Present and After Acquired Property With Exception</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="All Present and After Acquired Property Without Exception">
        <xs:annotation>
          <xs:documentation>All Present and After Acquired Property Without Exception</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.All Present and After Acquired Property Without Exception</lx:path>
            <lx:label>All Present and After Acquired Property Without Exception</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Chattel Paper">
        <xs:annotation>
          <xs:documentation>Chattel Paper</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Chattel Paper</lx:path>
            <lx:label>Chattel Paper</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Circuit Layout">
        <xs:annotation>
          <xs:documentation>Circuit Layout</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Circuit Layout</lx:path>
            <lx:label>Circuit Layout</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Copyright">
        <xs:annotation>
          <xs:documentation>Copyright</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Copyright</lx:path>
            <lx:label>Copyright</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Crops">
        <xs:annotation>
          <xs:documentation>Crops</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Crops</lx:path>
            <lx:label>Crops</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Currency">
        <xs:annotation>
          <xs:documentation>Currency</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Currency</lx:path>
            <lx:label>Currency</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Design">
        <xs:annotation>
          <xs:documentation>Design</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Design</lx:path>
            <lx:label>Design</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Document of Title">
        <xs:annotation>
          <xs:documentation>Document of Title</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Document of Title</lx:path>
            <lx:label>Document of Title</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="General Intangible">
        <xs:annotation>
          <xs:documentation>General Intangible</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.General Intangible</lx:path>
            <lx:label>General Intangible</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Helicopter">
        <xs:annotation>
          <xs:documentation>Helicopter</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Helicopter</lx:path>
            <lx:label>Helicopter</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Intellectual Property">
        <xs:annotation>
          <xs:documentation>Intellectual Property</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Intellectual Property</lx:path>
            <lx:label>Intellectual Property</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Intermediate Security">
        <xs:annotation>
          <xs:documentation>Intermediate Security</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Intermediate Security</lx:path>
            <lx:label>Intermediate Security</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Investment Instrument">
        <xs:annotation>
          <xs:documentation>Investment Instrument</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Investment Instrument</lx:path>
            <lx:label>Investment Instrument</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Livestock">
        <xs:annotation>
          <xs:documentation>Livestock</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Livestock</lx:path>
            <lx:label>Livestock</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Motor Vehicle">
        <xs:annotation>
          <xs:documentation>Motor Vehicle</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Motor Vehicle</lx:path>
            <lx:label>Motor Vehicle</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Negotiable Instrument">
        <xs:annotation>
          <xs:documentation>Negotiable Instrument</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Negotiable Instrument</lx:path>
            <lx:label>Negotiable Instrument</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other Goods">
        <xs:annotation>
          <xs:documentation>Other Goods</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Other Goods</lx:path>
            <lx:label>Other Goods</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Patent">
        <xs:annotation>
          <xs:documentation>Patent</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Patent</lx:path>
            <lx:label>Patent</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Plant Breeders Right">
        <xs:annotation>
          <xs:documentation>Plant Breeders Right</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Plant Breeders Right</lx:path>
            <lx:label>Plant Breeders Right</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Small Aircraft">
        <xs:annotation>
          <xs:documentation>Small Aircraft</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Small Aircraft</lx:path>
            <lx:label>Small Aircraft</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trade Mark">
        <xs:annotation>
          <xs:documentation>Trade Mark</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Trade Mark</lx:path>
            <lx:label>Trade Mark</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unknown">
        <xs:annotation>
          <xs:documentation>Unknown</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Unknown</lx:path>
            <lx:label>Unknown</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unsupported">
        <xs:annotation>
          <xs:documentation>Unsupported</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Unsupported</lx:path>
            <lx:label>Unsupported</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Watercraft">
        <xs:annotation>
          <xs:documentation>Watercraft</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralClassList.Watercraft</lx:path>
            <lx:label>Watercraft</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="collateralTypeList">
    <xs:annotation>
      <xs:documentation>collateralTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>collateralTypeList</lx:path>
        <lx:label>collateralTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Commercial">
        <xs:annotation>
          <xs:documentation>Commercial</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralTypeList.Commercial</lx:path>
            <lx:label>Commercial</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Consumer">
        <xs:annotation>
          <xs:documentation>Consumer</xs:documentation>
          <xs:appinfo>
            <lx:path>collateralTypeList.Consumer</lx:path>
            <lx:label>Consumer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="commercialTypeList">
    <xs:annotation>
      <xs:documentation>commercialTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>commercialTypeList</lx:path>
        <lx:label>commercialTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Block of Units or Flats">
        <xs:annotation>
          <xs:documentation>A block of units or flats where the units/flats all come under a single title.</xs:documentation>
          <xs:appinfo>
            <lx:path>commercialTypeList.Block of Units or Flats</lx:path>
            <lx:label>Block of Units or Flats.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Factory">
        <xs:annotation>
          <xs:documentation>Factory</xs:documentation>
          <xs:appinfo>
            <lx:path>commercialTypeList.Factory</lx:path>
            <lx:label>Factory</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Multiple Dwelling Development">
        <xs:annotation>
          <xs:documentation>A multiple dwelling development is where the units may exist under individual titles.</xs:documentation>
          <xs:appinfo>
            <lx:path>commercialTypeList.Multiple Dwelling Development</lx:path>
            <lx:label>Multiple Dwelling Development</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Non Specialised Commercial">
        <xs:annotation>
          <xs:documentation>Non Specialised Commercial</xs:documentation>
          <xs:appinfo>
            <lx:path>commercialTypeList.Non Specialised Commercial</lx:path>
            <lx:label>Non Specialised Commercial</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Offices">
        <xs:annotation>
          <xs:documentation>Offices</xs:documentation>
          <xs:appinfo>
            <lx:path>commercialTypeList.Offices</lx:path>
            <lx:label>Offices</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Commercial premises of a type that is not described in any of the other enumerated values.</xs:documentation>
          <xs:appinfo>
            <lx:path>commercialTypeList.Other</lx:path>
            <lx:label>Commercial premises of a type that is not described in any of the other enumerated values.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Professional Chambers">
        <xs:annotation>
          <xs:documentation>Professional chambers, such as barristers chambers.</xs:documentation>
          <xs:appinfo>
            <lx:path>commercialTypeList.Professional Chambers</lx:path>
            <lx:label>Professional chambers, such as barristers chambers.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Residential Commercial">
        <xs:annotation>
          <xs:documentation>Mixed usage residential commercial, such as a shop front with an apartment residence over.</xs:documentation>
          <xs:appinfo>
            <lx:path>commercialTypeList.Residential Commercial</lx:path>
            <lx:label>Mixed usage residential commercial, such as a shop front with an apartment residence over.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Retirement Village">
        <xs:annotation>
          <xs:documentation>Retirement village accommodation.</xs:documentation>
          <xs:appinfo>
            <lx:path>commercialTypeList.Retirement Village</lx:path>
            <lx:label>Retirement village accommodation.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vacant Land">
        <xs:annotation>
          <xs:documentation>Land that does not have a structure erected on it.</xs:documentation>
          <xs:appinfo>
            <lx:path>commercialTypeList.Vacant Land</lx:path>
            <lx:label>Land that does not have a structure erected on it.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Warehouse">
        <xs:annotation>
          <xs:documentation>Warehouse</xs:documentation>
          <xs:appinfo>
            <lx:path>commercialTypeList.Warehouse</lx:path>
            <lx:label>Warehouse</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="commissionStructureList">
    <xs:annotation>
      <xs:documentation>commissionStructureList</xs:documentation>
      <xs:appinfo>
        <lx:path>commissionStructureList</lx:path>
        <lx:label>commissionStructureList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>commissionStructureList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trail Only">
        <xs:annotation>
          <xs:documentation>An ongoing trail commission will be paid post-settlement, no upfront commission will be paid.</xs:documentation>
          <xs:appinfo>
            <lx:path>commissionStructureList.Trail Only</lx:path>
            <lx:label>An ongoing trail commission will be paid post-settlement, no upfront commission will be paid.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Up Front Only">
        <xs:annotation>
          <xs:documentation>An up front commission only will be paid on settlement of the loan, no trail commission is payable.</xs:documentation>
          <xs:appinfo>
            <lx:path>commissionStructureList.Up Front Only</lx:path>
            <lx:label>An up front commission only will be paid on settlement of the loan, no trail commission is payable.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Up Front Plus Trail">
        <xs:annotation>
          <xs:documentation>An up front commission only will be paid on settlement of the loan, as well as an ongoing trail commission.</xs:documentation>
          <xs:appinfo>
            <lx:path>commissionStructureList.Up Front Plus Trail</lx:path>
            <lx:label>An up front commission only will be paid on settlement of the loan, as well as an ongoing trail commission.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="comparisonList">
    <xs:annotation>
      <xs:documentation>comparisonList</xs:documentation>
      <xs:appinfo>
        <lx:path>comparisonList</lx:path>
        <lx:label>comparisonList</lx:label>
        <li:transactions>LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Comparable">
        <xs:annotation>
          <xs:documentation>Comparable</xs:documentation>
          <xs:appinfo>
            <lx:path>comparisonList.Comparable</lx:path>
            <lx:label>Comparable</lx:label>
            <li:transactions>LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Inferior">
        <xs:annotation>
          <xs:documentation>Inferior</xs:documentation>
          <xs:appinfo>
            <lx:path>comparisonList.Inferior</lx:path>
            <lx:label>Inferior</lx:label>
            <li:transactions>LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Superior">
        <xs:annotation>
          <xs:documentation>Superior</xs:documentation>
          <xs:appinfo>
            <lx:path>comparisonList.Superior</lx:path>
            <lx:label>Superior</lx:label>
            <li:transactions>LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="conditionList">
    <xs:annotation>
      <xs:documentation>conditionList</xs:documentation>
      <xs:appinfo>
        <lx:path>conditionList</lx:path>
        <lx:label>conditionList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Demo">
        <xs:annotation>
          <xs:documentation>Demo</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionList.Demo</lx:path>
            <lx:label>Demo</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="New">
        <xs:annotation>
          <xs:documentation>New</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionList.New</lx:path>
            <lx:label>New</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Refurbished">
        <xs:annotation>
          <xs:documentation>Refurbished</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionList.Refurbished</lx:path>
            <lx:label>Refurbished</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Used">
        <xs:annotation>
          <xs:documentation>Used</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionList.Used</lx:path>
            <lx:label>Used</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="conditionOwnerApplicationInstructionsList">
    <xs:annotation>
      <xs:documentation>conditionOwnerApplicationInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>conditionOwnerApplicationInstructionsList</lx:path>
        <lx:label>conditionOwnerApplicationInstructionsList</lx:label>
        <li:transactions>CAL,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Applicant">
        <xs:annotation>
          <xs:documentation>The current responsible party for actioning the Condition is the Applicant</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionOwnerApplicationInstructionsList.Applicant</lx:path>
            <lx:label>The current responsible party for actioning the Condition is the Applicant</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Borrower's Solicitor">
        <xs:annotation>
          <xs:documentation>The current responsible party for actioning the Condition is the Borrower's Solicitor</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionOwnerApplicationInstructionsList.Borrower's Solicitor</lx:path>
            <lx:label>The current responsible party for actioning the Condition is the Borrower's Solicitor</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Branch">
        <xs:annotation>
          <xs:documentation>The current responsible party for actioning the Condition is the Branch</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionOwnerApplicationInstructionsList.Branch</lx:path>
            <lx:label>The current responsible party for actioning the Condition is the Branch</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Broker">
        <xs:annotation>
          <xs:documentation>The current responsible party for actioning the Condition is the Broker</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionOwnerApplicationInstructionsList.Broker</lx:path>
            <lx:label>The current responsible party for actioning the Condition is the Broker</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Officer">
        <xs:annotation>
          <xs:documentation>The current responsible party for actioning the Condition is the Loan Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionOwnerApplicationInstructionsList.Loan Officer</lx:path>
            <lx:label>The current responsible party for actioning the Condition is the Loan Officer</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Agent">
        <xs:annotation>
          <xs:documentation>The current responsible party for actioning the Condition is the Settlement Agent</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionOwnerApplicationInstructionsList.Settlement Agent</lx:path>
            <lx:label>The current responsible party for actioning the Condition is the Settlement Agent</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="conditionOwnerLoanDocumentsInstructionsList">
    <xs:annotation>
      <xs:documentation>conditionOwnerLoanDocumentsInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>conditionOwnerLoanDocumentsInstructionsList</lx:path>
        <lx:label>conditionOwnerLoanDocumentsInstructionsList</lx:label>
        <li:transactions>DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Loan Officer">
        <xs:annotation>
          <xs:documentation>The current responsible party for actioning the Condition is the Loan Officer</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionOwnerLoanDocumentsInstructionsList.Loan Officer</lx:path>
            <lx:label>The current responsible party for actioning the Condition is the Loan Officer</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Agent">
        <xs:annotation>
          <xs:documentation>The current responsible party for actioning the Condition is the Settlement Agent</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionOwnerLoanDocumentsInstructionsList.Settlement Agent</lx:path>
            <lx:label>The current responsible party for actioning the Condition is the Settlement Agent</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="conditionResponseStatusList">
    <xs:annotation>
      <xs:documentation>conditionResponseStatusList</xs:documentation>
      <xs:appinfo>
        <lx:path>conditionResponseStatusList</lx:path>
        <lx:label>conditionResponseStatusList</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Met">
        <xs:annotation>
          <xs:documentation>The submitting party attests that the condition has been met, and if supporting documentation is required it is attached.</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionResponseStatusList.Met</lx:path>
            <lx:label>Met</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Not Met">
        <xs:annotation>
          <xs:documentation>The submitting party attests that the condition has not been met. The condition response description may include a request that the condition be waived.</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionResponseStatusList.Not Met</lx:path>
            <lx:label>Not Met</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Partially Met">
        <xs:annotation>
          <xs:documentation>The submitting party attests that the condition has been met in part, and if supporting documentation is required it is attached.</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionResponseStatusList.Partially Met</lx:path>
            <lx:label>Partially Met</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="conditionStatusApplicationInstructionsList">
    <xs:annotation>
      <xs:documentation>conditionStatusApplicationInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>conditionStatusApplicationInstructionsList</lx:path>
        <lx:label>conditionStatusApplicationInstructionsList</lx:label>
        <li:transactions>CAL,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Not Satisfied">
        <xs:annotation>
          <xs:documentation>The Condition has not been satisfied</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionStatusApplicationInstructionsList.Not Satisfied</lx:path>
            <lx:label>The Condition has not been satisfied</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Open">
        <xs:annotation>
          <xs:documentation>The Condition status is Open and awaiting actioning</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionStatusApplicationInstructionsList.Open</lx:path>
            <lx:label>The Condition status is Open and awaiting actioning</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Referred to Assessor">
        <xs:annotation>
          <xs:documentation>The Condition has been referred for actioning</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionStatusApplicationInstructionsList.Referred to Assessor</lx:path>
            <lx:label>The Condition has been referred for actioning</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Satisfied">
        <xs:annotation>
          <xs:documentation>The Condition has been satisfied</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionStatusApplicationInstructionsList.Satisfied</lx:path>
            <lx:label>The Condition has been satisfied</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Waived">
        <xs:annotation>
          <xs:documentation>The Condition has been waived</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionStatusApplicationInstructionsList.Waived</lx:path>
            <lx:label>The Condition has been waived</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="conditionStatusLendersMortgageInsuranceInstructionsList">
    <xs:annotation>
      <xs:documentation>conditionStatusLendersMortgageInsuranceInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>conditionStatusLendersMortgageInsuranceInstructionsList</lx:path>
        <lx:label>conditionStatusLendersMortgageInsuranceInstructionsList</lx:label>
        <li:transactions>LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Not Satisfied">
        <xs:annotation>
          <xs:documentation>The Condition has not been satisfied</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionStatusLendersMortgageInsuranceInstructionsList.Not Satisfied</lx:path>
            <lx:label>The Condition has not been satisfied</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Open">
        <xs:annotation>
          <xs:documentation>The Condition status is Open and awaiting actioning</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionStatusLendersMortgageInsuranceInstructionsList.Open</lx:path>
            <lx:label>The Condition status is Open and awaiting actioning</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Satisfied">
        <xs:annotation>
          <xs:documentation>The Condition has been satisfied</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionStatusLendersMortgageInsuranceInstructionsList.Satisfied</lx:path>
            <lx:label>The Condition has been satisfied</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="conditionStatusLoanDocumentsInstructionsList">
    <xs:annotation>
      <xs:documentation>conditionStatusLoanDocumentsInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>conditionStatusLoanDocumentsInstructionsList</lx:path>
        <lx:label>conditionStatusLoanDocumentsInstructionsList</lx:label>
        <li:transactions>DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Not Satisfied">
        <xs:annotation>
          <xs:documentation>The Condition has not been satisfied</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionStatusLoanDocumentsInstructionsList.Not Satisfied</lx:path>
            <lx:label>The Condition has not been satisfied</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Open">
        <xs:annotation>
          <xs:documentation>The Condition status is Open and awaiting actioning</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionStatusLoanDocumentsInstructionsList.Open</lx:path>
            <lx:label>The Condition status is Open and awaiting actioning</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Referred">
        <xs:annotation>
          <xs:documentation>The Condition has been referred for actioning</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionStatusLoanDocumentsInstructionsList.Referred</lx:path>
            <lx:label>The Condition has been referred for actioning</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Satisfied">
        <xs:annotation>
          <xs:documentation>The Condition has been satisfied</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionStatusLoanDocumentsInstructionsList.Satisfied</lx:path>
            <lx:label>The Condition has been satisfied</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Waived">
        <xs:annotation>
          <xs:documentation>The Condition has been waived</xs:documentation>
          <xs:appinfo>
            <lx:path>conditionStatusLoanDocumentsInstructionsList.Waived</lx:path>
            <lx:label>The Condition has been waived</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="constructionTypeList">
    <xs:annotation>
      <xs:documentation>constructionTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>constructionTypeList</lx:path>
        <lx:label>constructionTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Full Construction">
        <xs:annotation>
          <xs:documentation>Full Construction</xs:documentation>
          <xs:appinfo>
            <lx:path>constructionTypeList.Full Construction</lx:path>
            <lx:label>Full Construction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Home Extension">
        <xs:annotation>
          <xs:documentation>Home Extension</xs:documentation>
          <xs:appinfo>
            <lx:path>constructionTypeList.Home Extension</lx:path>
            <lx:label>Home Extension</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="House Land Package">
        <xs:annotation>
          <xs:documentation>House Land Package</xs:documentation>
          <xs:appinfo>
            <lx:path>constructionTypeList.House Land Package</lx:path>
            <lx:label>House Land Package</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Minor Non Structural Repairs">
        <xs:annotation>
          <xs:documentation>Minor Non Structural Repairs</xs:documentation>
          <xs:appinfo>
            <lx:path>constructionTypeList.Minor Non Structural Repairs</lx:path>
            <lx:label>Minor Non Structural Repairs</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Renovation">
        <xs:annotation>
          <xs:documentation>Renovation</xs:documentation>
          <xs:appinfo>
            <lx:path>constructionTypeList.Renovation</lx:path>
            <lx:label>Renovation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Weather Tightness Repairs">
        <xs:annotation>
          <xs:documentation>Weather Tightness Repairs addresses a building's weathertightness and durability issues in order to make the dwelling compliant with the NZ Building Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>constructionTypeList.Weather Tightness Repairs</lx:path>
            <lx:label>Weather Tightness Repairs</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="contactEmailTypeList">
    <xs:annotation>
      <xs:documentation>contactEmailTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>contactEmailTypeList</lx:path>
        <lx:label>contactEmailTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Home">
        <xs:annotation>
          <xs:documentation>Email address is a home or personal address.</xs:documentation>
          <xs:appinfo>
            <lx:path>contactEmailTypeList.Home</lx:path>
            <lx:label>Email address is a home or personal address.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Work">
        <xs:annotation>
          <xs:documentation>Email address is a work or business address.</xs:documentation>
          <xs:appinfo>
            <lx:path>contactEmailTypeList.Work</lx:path>
            <lx:label>Email address is a work or business address.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="contexTypeAdditionalDocumentList">
    <xs:annotation>
      <xs:documentation>contexTypeAdditionalDocumentList</xs:documentation>
      <xs:appinfo>
        <lx:path>contexTypeAdditionalDocumentList</lx:path>
        <lx:label>contexTypeAdditionalDocumentList</lx:label>
        <li:transactions>DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Address">
        <xs:annotation>
          <xs:documentation>Address</xs:documentation>
          <xs:appinfo>
            <lx:path>contexTypeAdditionalDocumentList.Address</lx:path>
            <lx:label>Address</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Liability">
        <xs:annotation>
          <xs:documentation>Liability</xs:documentation>
          <xs:appinfo>
            <lx:path>contexTypeAdditionalDocumentList.Liability</lx:path>
            <lx:label>Liability</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Account">
        <xs:annotation>
          <xs:documentation>Loan Account</xs:documentation>
          <xs:appinfo>
            <lx:path>contexTypeAdditionalDocumentList.Loan Account</lx:path>
            <lx:label>Loan Account</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Party">
        <xs:annotation>
          <xs:documentation>Party</xs:documentation>
          <xs:appinfo>
            <lx:path>contexTypeAdditionalDocumentList.Party</lx:path>
            <lx:label>Party</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security">
        <xs:annotation>
          <xs:documentation>Security</xs:documentation>
          <xs:appinfo>
            <lx:path>contexTypeAdditionalDocumentList.Security</lx:path>
            <lx:label>Security</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="contributionFundsTypeList">
    <xs:annotation>
      <xs:documentation>contributionFundsTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>contributionFundsTypeList</lx:path>
        <lx:label>contributionFundsTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Cash">
        <xs:annotation>
          <xs:documentation>The source of funds is cash or equivalent, other than genuine savings.</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.Cash</lx:path>
            <lx:label>The source of funds is cash or equivalent, other than genuine savings.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Deposit Paid">
        <xs:annotation>
          <xs:documentation>The source of funds is money already paid as deposit.</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.Deposit Paid</lx:path>
            <lx:label>Deposit Paid</lx:label>
            <li:transactions>CAL,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Existing Equity">
        <xs:annotation>
          <xs:documentation>Existing Equity</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.Existing Equity</lx:path>
            <lx:label>Existing Equity</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="First Home Owner Grant">
        <xs:annotation>
          <xs:documentation>The source of funds is a First Home Owners Grant program to assist first home buyers to purchase an O/O home.</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.First Home Owner Grant</lx:path>
            <lx:label>The source of funds is a First Home Owners Grant program to assist first home buyers to purchase an O/O home.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Genuine Savings">
        <xs:annotation>
          <xs:documentation>The source of funds is savings that have been accumulated over time, with a savings institution or a liquidatable investment.</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.Genuine Savings</lx:path>
            <lx:label>The source of funds is savings that have been accumulated over time, with a savings institution or a liquidatable investment.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gift">
        <xs:annotation>
          <xs:documentation>The source of funds is a gift from a third party.</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.Gift</lx:path>
            <lx:label>The source of funds is a gift from a third party.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Inheritance">
        <xs:annotation>
          <xs:documentation>The funds will be provided from an inheritance that the borrower(s) have the right to use.</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.Inheritance</lx:path>
            <lx:label>The funds will be provided from an ineritance that the borrower(s) have the right to use.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KiwiSaver">
        <xs:annotation>
          <xs:documentation>The source of funds is a KiwiSaver.</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.KiwiSaver</lx:path>
            <lx:label>KiwiSaver</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Net Proceeds from Sale of Property">
        <xs:annotation>
          <xs:documentation>The net proceeds of sale of a property, after payment of all costs associated with the sale including agent's fees.</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.Net Proceeds from Sale of Property</lx:path>
            <lx:label>The net proceeds of sale of a property, after payment of all costs associated with the sale including agent's fees.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>The source of funds is not described by any of the other enumerated values.</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.Other</lx:path>
            <lx:label>The source of funds is not described by any of the other enumerated values.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Redraw">
        <xs:annotation>
          <xs:documentation>Redraw</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.Redraw</lx:path>
            <lx:label>Redraw</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sale of Non Real Estate Asset">
        <xs:annotation>
          <xs:documentation>Sale of Non Real Estate Asset</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.Sale of Non Real Estate Asset</lx:path>
            <lx:label>Sale of Non Real Estate Asset</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Secured VISA">
        <xs:annotation>
          <xs:documentation>Secured VISA</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.Secured VISA</lx:path>
            <lx:label>Secured VISA</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Superannuation">
        <xs:annotation>
          <xs:documentation>The funds will be provided from a superannuation fund.</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.Superannuation</lx:path>
            <lx:label>The funds will be provided from a superannuation fund.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Term Deposit">
        <xs:annotation>
          <xs:documentation>The funds will be provided from an existing term deposit.</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.Term Deposit</lx:path>
            <lx:label>The funds will be provided from an existing term deposit.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unsecured Borrowing">
        <xs:annotation>
          <xs:documentation>The funds will be borrowed from a third party.</xs:documentation>
          <xs:appinfo>
            <lx:path>contributionFundsTypeList.Unsecured Borrowing</lx:path>
            <lx:label>The funds will be borrowed from a third party.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="countryCodeList">
    <xs:annotation>
      <xs:documentation>countryCodeList</xs:documentation>
      <xs:appinfo>
        <lx:path>countryCodeList</lx:path>
        <lx:label>countryCodeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="AD">
        <xs:annotation>
          <xs:documentation>Andorra</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AD</lx:path>
            <lx:label>Andorra</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AE">
        <xs:annotation>
          <xs:documentation>United Arab Emirates</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AE</lx:path>
            <lx:label>United Arab Emirates</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AF">
        <xs:annotation>
          <xs:documentation>Afghanistan</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AF</lx:path>
            <lx:label>Afghanistan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AG">
        <xs:annotation>
          <xs:documentation>Antigua and Barbuda</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AG</lx:path>
            <lx:label>Antigua and Barbuda</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AI">
        <xs:annotation>
          <xs:documentation>Anguilla</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AI</lx:path>
            <lx:label>Anguilla</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AL">
        <xs:annotation>
          <xs:documentation>Albania</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AL</lx:path>
            <lx:label>Albania</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AM">
        <xs:annotation>
          <xs:documentation>Armenia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AM</lx:path>
            <lx:label>Armenia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AO">
        <xs:annotation>
          <xs:documentation>Angola</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AO</lx:path>
            <lx:label>Angola</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AQ">
        <xs:annotation>
          <xs:documentation>Antarctica</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AQ</lx:path>
            <lx:label>Antarctica</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AR">
        <xs:annotation>
          <xs:documentation>Argentina</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AR</lx:path>
            <lx:label>Argentina</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AS">
        <xs:annotation>
          <xs:documentation>American Samoa</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AS</lx:path>
            <lx:label>American Samoa</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AT">
        <xs:annotation>
          <xs:documentation>Austria</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AT</lx:path>
            <lx:label>Austria</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AU">
        <xs:annotation>
          <xs:documentation>Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AU</lx:path>
            <lx:label>Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AW">
        <xs:annotation>
          <xs:documentation>Aruba</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AW</lx:path>
            <lx:label>Aruba</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AX">
        <xs:annotation>
          <xs:documentation>Åland Islands</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AX</lx:path>
            <lx:label>Åland Islands</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AZ">
        <xs:annotation>
          <xs:documentation>Azerbaijan</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.AZ</lx:path>
            <lx:label>Azerbaijan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BA">
        <xs:annotation>
          <xs:documentation>Bosnia and Herzegovina</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BA</lx:path>
            <lx:label>Bosnia and Herzegovina</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BB">
        <xs:annotation>
          <xs:documentation>Barbados</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BB</lx:path>
            <lx:label>Barbados</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BD">
        <xs:annotation>
          <xs:documentation>Bangladesh</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BD</lx:path>
            <lx:label>Bangladesh</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BE">
        <xs:annotation>
          <xs:documentation>Belgium</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BE</lx:path>
            <lx:label>Belgium</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BF">
        <xs:annotation>
          <xs:documentation>Burkina Faso</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BF</lx:path>
            <lx:label>Burkina Faso</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BG">
        <xs:annotation>
          <xs:documentation>Bulgaria</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BG</lx:path>
            <lx:label>Bulgaria</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BH">
        <xs:annotation>
          <xs:documentation>Bahrain</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BH</lx:path>
            <lx:label>Bahrain</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BI">
        <xs:annotation>
          <xs:documentation>Burundi</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BI</lx:path>
            <lx:label>Burundi</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BJ">
        <xs:annotation>
          <xs:documentation>Benin</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BJ</lx:path>
            <lx:label>Benin</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BL">
        <xs:annotation>
          <xs:documentation>Saint Barthélemy</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BL</lx:path>
            <lx:label>Saint Barthélemy</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BM">
        <xs:annotation>
          <xs:documentation>Bermuda</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BM</lx:path>
            <lx:label>Bermuda</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BN">
        <xs:annotation>
          <xs:documentation>Brunei Darussalam</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BN</lx:path>
            <lx:label>Brunei Darussalam</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BO">
        <xs:annotation>
          <xs:documentation>Bolivia, Plurinational State of</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BO</lx:path>
            <lx:label>Bolivia, Plurinational State of</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BQ">
        <xs:annotation>
          <xs:documentation>Bonaire, Sint Eustatius and Saba</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BQ</lx:path>
            <lx:label>Bonaire, Sint Eustatius and Saba</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BR">
        <xs:annotation>
          <xs:documentation>Brazil</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BR</lx:path>
            <lx:label>Brazil</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BS">
        <xs:annotation>
          <xs:documentation>Bahamas</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BS</lx:path>
            <lx:label>Bahamas</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BT">
        <xs:annotation>
          <xs:documentation>Bhutan</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BT</lx:path>
            <lx:label>Bhutan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BV">
        <xs:annotation>
          <xs:documentation>Bouvet Island</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BV</lx:path>
            <lx:label>Bouvet Island</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BW">
        <xs:annotation>
          <xs:documentation>Botswana</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BW</lx:path>
            <lx:label>Botswana</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BY">
        <xs:annotation>
          <xs:documentation>Belarus</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BY</lx:path>
            <lx:label>Belarus</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BZ">
        <xs:annotation>
          <xs:documentation>Belize</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.BZ</lx:path>
            <lx:label>Belize</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CA">
        <xs:annotation>
          <xs:documentation>Canada</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CA</lx:path>
            <lx:label>Canada</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CC">
        <xs:annotation>
          <xs:documentation>Cocos (Keeling) Islands</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CC</lx:path>
            <lx:label>Cocos (Keeling) Islands</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CD">
        <xs:annotation>
          <xs:documentation>Congo, the Democratic Republic of the</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CD</lx:path>
            <lx:label>Congo, the Democratic Republic of the</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CF">
        <xs:annotation>
          <xs:documentation>Central African Republic</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CF</lx:path>
            <lx:label>Central African Republic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CG">
        <xs:annotation>
          <xs:documentation>Congo</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CG</lx:path>
            <lx:label>Congo</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CH">
        <xs:annotation>
          <xs:documentation>Switzerland</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CH</lx:path>
            <lx:label>Switzerland</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CI">
        <xs:annotation>
          <xs:documentation>Côte d'Ivoire</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CI</lx:path>
            <lx:label>Côte d'Ivoire</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CK">
        <xs:annotation>
          <xs:documentation>Cook Islands</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CK</lx:path>
            <lx:label>Cook Islands</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CL">
        <xs:annotation>
          <xs:documentation>Chile</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CL</lx:path>
            <lx:label>Chile</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CM">
        <xs:annotation>
          <xs:documentation>Cameroon</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CM</lx:path>
            <lx:label>Cameroon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CN">
        <xs:annotation>
          <xs:documentation>China</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CN</lx:path>
            <lx:label>China</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CO">
        <xs:annotation>
          <xs:documentation>Colombia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CO</lx:path>
            <lx:label>Colombia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CR">
        <xs:annotation>
          <xs:documentation>Costa Rica</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CR</lx:path>
            <lx:label>Costa Rica</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CU">
        <xs:annotation>
          <xs:documentation>Cuba</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CU</lx:path>
            <lx:label>Cuba</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CV">
        <xs:annotation>
          <xs:documentation>Cabo Verde</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CV</lx:path>
            <lx:label>Cabo Verde</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CW">
        <xs:annotation>
          <xs:documentation>Curaçao</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CW</lx:path>
            <lx:label>Curaçao</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CX">
        <xs:annotation>
          <xs:documentation>Christmas Island</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CX</lx:path>
            <lx:label>Christmas Island</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CY">
        <xs:annotation>
          <xs:documentation>Cyprus</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CY</lx:path>
            <lx:label>Cyprus</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CZ">
        <xs:annotation>
          <xs:documentation>Czech Republic</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.CZ</lx:path>
            <lx:label>Czech Republic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="DE">
        <xs:annotation>
          <xs:documentation>Germany</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.DE</lx:path>
            <lx:label>Germany</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="DJ">
        <xs:annotation>
          <xs:documentation>Djibouti</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.DJ</lx:path>
            <lx:label>Djibouti</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="DK">
        <xs:annotation>
          <xs:documentation>Denmark</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.DK</lx:path>
            <lx:label>Denmark</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="DM">
        <xs:annotation>
          <xs:documentation>Dominica</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.DM</lx:path>
            <lx:label>Dominica</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="DO">
        <xs:annotation>
          <xs:documentation>Dominican Republic</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.DO</lx:path>
            <lx:label>Dominican Republic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="DZ">
        <xs:annotation>
          <xs:documentation>Algeria</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.DZ</lx:path>
            <lx:label>Algeria</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="EC">
        <xs:annotation>
          <xs:documentation>Ecuador</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.EC</lx:path>
            <lx:label>Ecuador</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="EE">
        <xs:annotation>
          <xs:documentation>Estonia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.EE</lx:path>
            <lx:label>Estonia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="EG">
        <xs:annotation>
          <xs:documentation>Egypt</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.EG</lx:path>
            <lx:label>Egypt</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="EH">
        <xs:annotation>
          <xs:documentation>Western Sahara</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.EH</lx:path>
            <lx:label>Western Sahara</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ER">
        <xs:annotation>
          <xs:documentation>Eritrea</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.ER</lx:path>
            <lx:label>Eritrea</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ES">
        <xs:annotation>
          <xs:documentation>Spain</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.ES</lx:path>
            <lx:label>Spain</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ET">
        <xs:annotation>
          <xs:documentation>Ethiopia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.ET</lx:path>
            <lx:label>Ethiopia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FI">
        <xs:annotation>
          <xs:documentation>Finland</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.FI</lx:path>
            <lx:label>Finland</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FJ">
        <xs:annotation>
          <xs:documentation>Fiji</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.FJ</lx:path>
            <lx:label>Fiji</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FK">
        <xs:annotation>
          <xs:documentation>Falkland Islands (Malvinas)</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.FK</lx:path>
            <lx:label>Falkland Islands (Malvinas)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FM">
        <xs:annotation>
          <xs:documentation>Micronesia, Federated States of</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.FM</lx:path>
            <lx:label>Micronesia, Federated States of</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FO">
        <xs:annotation>
          <xs:documentation>Faroe Islands</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.FO</lx:path>
            <lx:label>Faroe Islands</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FR">
        <xs:annotation>
          <xs:documentation>France</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.FR</lx:path>
            <lx:label>France</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GA">
        <xs:annotation>
          <xs:documentation>Gabon</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GA</lx:path>
            <lx:label>Gabon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GB">
        <xs:annotation>
          <xs:documentation>United Kingdom</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GB</lx:path>
            <lx:label>United Kingdom</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GD">
        <xs:annotation>
          <xs:documentation>Grenada</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GD</lx:path>
            <lx:label>Grenada</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GE">
        <xs:annotation>
          <xs:documentation>Georgia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GE</lx:path>
            <lx:label>Georgia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GF">
        <xs:annotation>
          <xs:documentation>French Guiana</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GF</lx:path>
            <lx:label>French Guiana</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GG">
        <xs:annotation>
          <xs:documentation>Guernsey</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GG</lx:path>
            <lx:label>Guernsey</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GH">
        <xs:annotation>
          <xs:documentation>Ghana</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GH</lx:path>
            <lx:label>Ghana</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GI">
        <xs:annotation>
          <xs:documentation>Gibraltar</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GI</lx:path>
            <lx:label>Gibraltar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GL">
        <xs:annotation>
          <xs:documentation>Greenland</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GL</lx:path>
            <lx:label>Greenland</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GM">
        <xs:annotation>
          <xs:documentation>Gambia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GM</lx:path>
            <lx:label>Gambia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GN">
        <xs:annotation>
          <xs:documentation>Guinea</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GN</lx:path>
            <lx:label>Guinea</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GP">
        <xs:annotation>
          <xs:documentation>Guadeloupe</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GP</lx:path>
            <lx:label>Guadeloupe</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GQ">
        <xs:annotation>
          <xs:documentation>Equatorial Guinea</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GQ</lx:path>
            <lx:label>Equatorial Guinea</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GR">
        <xs:annotation>
          <xs:documentation>Greece</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GR</lx:path>
            <lx:label>Greece</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GS">
        <xs:annotation>
          <xs:documentation>South Georgia and the South Sandwich Islands</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GS</lx:path>
            <lx:label>South Georgia and the South Sandwich Islands</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GT">
        <xs:annotation>
          <xs:documentation>Guatemala</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GT</lx:path>
            <lx:label>Guatemala</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GU">
        <xs:annotation>
          <xs:documentation>Guam</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GU</lx:path>
            <lx:label>Guam</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GW">
        <xs:annotation>
          <xs:documentation>Guinea-Bissau</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GW</lx:path>
            <lx:label>Guinea-Bissau</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GY">
        <xs:annotation>
          <xs:documentation>Guyana</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.GY</lx:path>
            <lx:label>Guyana</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HK">
        <xs:annotation>
          <xs:documentation>Hong Kong</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.HK</lx:path>
            <lx:label>Hong Kong</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HM">
        <xs:annotation>
          <xs:documentation>Heard Island and McDonald Islands</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.HM</lx:path>
            <lx:label>Heard Island and McDonald Islands</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HN">
        <xs:annotation>
          <xs:documentation>Honduras</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.HN</lx:path>
            <lx:label>Honduras</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HR">
        <xs:annotation>
          <xs:documentation>Croatia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.HR</lx:path>
            <lx:label>Croatia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HT">
        <xs:annotation>
          <xs:documentation>Haiti</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.HT</lx:path>
            <lx:label>Haiti</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HU">
        <xs:annotation>
          <xs:documentation>Hungary</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.HU</lx:path>
            <lx:label>Hungary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ID">
        <xs:annotation>
          <xs:documentation>Indonesia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.ID</lx:path>
            <lx:label>Indonesia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IE">
        <xs:annotation>
          <xs:documentation>Ireland</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.IE</lx:path>
            <lx:label>Ireland</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IL">
        <xs:annotation>
          <xs:documentation>Israel</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.IL</lx:path>
            <lx:label>Israel</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IM">
        <xs:annotation>
          <xs:documentation>Isle of Man</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.IM</lx:path>
            <lx:label>Isle of Man</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IN">
        <xs:annotation>
          <xs:documentation>India</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.IN</lx:path>
            <lx:label>India</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IO">
        <xs:annotation>
          <xs:documentation>British Indian Ocean Territory</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.IO</lx:path>
            <lx:label>British Indian Ocean Territory</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IQ">
        <xs:annotation>
          <xs:documentation>Iraq</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.IQ</lx:path>
            <lx:label>Iraq</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IR">
        <xs:annotation>
          <xs:documentation>Iran, Islamic Republic of</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.IR</lx:path>
            <lx:label>Iran, Islamic Republic of</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IS">
        <xs:annotation>
          <xs:documentation>Iceland</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.IS</lx:path>
            <lx:label>Iceland</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IT">
        <xs:annotation>
          <xs:documentation>Italy</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.IT</lx:path>
            <lx:label>Italy</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="JE">
        <xs:annotation>
          <xs:documentation>Jersey</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.JE</lx:path>
            <lx:label>Jersey</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="JM">
        <xs:annotation>
          <xs:documentation>Jamaica</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.JM</lx:path>
            <lx:label>Jamaica</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="JO">
        <xs:annotation>
          <xs:documentation>Jordan</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.JO</lx:path>
            <lx:label>Jordan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="JP">
        <xs:annotation>
          <xs:documentation>Japan</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.JP</lx:path>
            <lx:label>Japan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KE">
        <xs:annotation>
          <xs:documentation>Kenya</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.KE</lx:path>
            <lx:label>Kenya</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KG">
        <xs:annotation>
          <xs:documentation>Kyrgyzstan</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.KG</lx:path>
            <lx:label>Kyrgyzstan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KH">
        <xs:annotation>
          <xs:documentation>Cambodia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.KH</lx:path>
            <lx:label>Cambodia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KI">
        <xs:annotation>
          <xs:documentation>Kiribati</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.KI</lx:path>
            <lx:label>Kiribati</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KM">
        <xs:annotation>
          <xs:documentation>Comoros</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.KM</lx:path>
            <lx:label>Comoros</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KN">
        <xs:annotation>
          <xs:documentation>Saint Kitts and Nevis</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.KN</lx:path>
            <lx:label>Saint Kitts and Nevis</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KP">
        <xs:annotation>
          <xs:documentation>Korea, Democratic People's Republic of</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.KP</lx:path>
            <lx:label>Korea, Democratic People's Republic of</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KR">
        <xs:annotation>
          <xs:documentation>Korea, Republic of</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.KR</lx:path>
            <lx:label>Korea, Republic of</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KW">
        <xs:annotation>
          <xs:documentation>Kuwait</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.KW</lx:path>
            <lx:label>Kuwait</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KY">
        <xs:annotation>
          <xs:documentation>Cayman Islands</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.KY</lx:path>
            <lx:label>Cayman Islands</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KZ">
        <xs:annotation>
          <xs:documentation>Kazakhstan</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.KZ</lx:path>
            <lx:label>Kazakhstan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LA">
        <xs:annotation>
          <xs:documentation>Lao People's Democratic Republic</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.LA</lx:path>
            <lx:label>Lao People's Democratic Republic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LB">
        <xs:annotation>
          <xs:documentation>Lebanon</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.LB</lx:path>
            <lx:label>Lebanon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LC">
        <xs:annotation>
          <xs:documentation>Saint Lucia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.LC</lx:path>
            <lx:label>Saint Lucia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LI">
        <xs:annotation>
          <xs:documentation>Liechtenstein</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.LI</lx:path>
            <lx:label>Liechtenstein</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LK">
        <xs:annotation>
          <xs:documentation>Sri Lanka</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.LK</lx:path>
            <lx:label>Sri Lanka</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LR">
        <xs:annotation>
          <xs:documentation>Liberia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.LR</lx:path>
            <lx:label>Liberia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LS">
        <xs:annotation>
          <xs:documentation>Lesotho</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.LS</lx:path>
            <lx:label>Lesotho</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LT">
        <xs:annotation>
          <xs:documentation>Lithuania</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.LT</lx:path>
            <lx:label>Lithuania</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LU">
        <xs:annotation>
          <xs:documentation>Luxembourg</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.LU</lx:path>
            <lx:label>Luxembourg</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LV">
        <xs:annotation>
          <xs:documentation>Latvia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.LV</lx:path>
            <lx:label>Latvia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LY">
        <xs:annotation>
          <xs:documentation>Libya</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.LY</lx:path>
            <lx:label>Libya</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MA">
        <xs:annotation>
          <xs:documentation>Morocco</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MA</lx:path>
            <lx:label>Morocco</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MC">
        <xs:annotation>
          <xs:documentation>Monaco</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MC</lx:path>
            <lx:label>Monaco</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MD">
        <xs:annotation>
          <xs:documentation>Moldova, Republic of</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MD</lx:path>
            <lx:label>Moldova, Republic of</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ME">
        <xs:annotation>
          <xs:documentation>Montenegro</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.ME</lx:path>
            <lx:label>Montenegro</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MF">
        <xs:annotation>
          <xs:documentation>Saint Martin (French part)</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MF</lx:path>
            <lx:label>Saint Martin (French part)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MG">
        <xs:annotation>
          <xs:documentation>Madagascar</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MG</lx:path>
            <lx:label>Madagascar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MH">
        <xs:annotation>
          <xs:documentation>Marshall Islands</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MH</lx:path>
            <lx:label>Marshall Islands</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MK">
        <xs:annotation>
          <xs:documentation>Macedonia, the former Yugoslav Republic of</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MK</lx:path>
            <lx:label>Macedonia, the former Yugoslav Republic of</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ML">
        <xs:annotation>
          <xs:documentation>Mali</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.ML</lx:path>
            <lx:label>Mali</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MM">
        <xs:annotation>
          <xs:documentation>Myanmar</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MM</lx:path>
            <lx:label>Myanmar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MN">
        <xs:annotation>
          <xs:documentation>Mongolia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MN</lx:path>
            <lx:label>Mongolia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MO">
        <xs:annotation>
          <xs:documentation>Macao</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MO</lx:path>
            <lx:label>Macao</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MP">
        <xs:annotation>
          <xs:documentation>Northern Mariana Islands</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MP</lx:path>
            <lx:label>Northern Mariana Islands</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MQ">
        <xs:annotation>
          <xs:documentation>Martinique</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MQ</lx:path>
            <lx:label>Martinique</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MR">
        <xs:annotation>
          <xs:documentation>Mauritania</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MR</lx:path>
            <lx:label>Mauritania</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MS">
        <xs:annotation>
          <xs:documentation>Montserrat</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MS</lx:path>
            <lx:label>Montserrat</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MT">
        <xs:annotation>
          <xs:documentation>Malta</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MT</lx:path>
            <lx:label>Malta</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MU">
        <xs:annotation>
          <xs:documentation>Mauritius</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MU</lx:path>
            <lx:label>Mauritius</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MV">
        <xs:annotation>
          <xs:documentation>Maldives</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MV</lx:path>
            <lx:label>Maldives</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MW">
        <xs:annotation>
          <xs:documentation>Malawi</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MW</lx:path>
            <lx:label>Malawi</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MX">
        <xs:annotation>
          <xs:documentation>Mexico</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MX</lx:path>
            <lx:label>Mexico</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MY">
        <xs:annotation>
          <xs:documentation>Malaysia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MY</lx:path>
            <lx:label>Malaysia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MZ">
        <xs:annotation>
          <xs:documentation>Mozambique</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.MZ</lx:path>
            <lx:label>Mozambique</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NA">
        <xs:annotation>
          <xs:documentation>Namibia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.NA</lx:path>
            <lx:label>Namibia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NC">
        <xs:annotation>
          <xs:documentation>New Caledonia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.NC</lx:path>
            <lx:label>New Caledonia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NE">
        <xs:annotation>
          <xs:documentation>Niger</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.NE</lx:path>
            <lx:label>Niger</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NF">
        <xs:annotation>
          <xs:documentation>Norfolk Island</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.NF</lx:path>
            <lx:label>Norfolk Island</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NG">
        <xs:annotation>
          <xs:documentation>Nigeria</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.NG</lx:path>
            <lx:label>Nigeria</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NI">
        <xs:annotation>
          <xs:documentation>Nicaragua</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.NI</lx:path>
            <lx:label>Nicaragua</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NL">
        <xs:annotation>
          <xs:documentation>Netherlands</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.NL</lx:path>
            <lx:label>Netherlands</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NO">
        <xs:annotation>
          <xs:documentation>Norway</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.NO</lx:path>
            <lx:label>Norway</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NP">
        <xs:annotation>
          <xs:documentation>Nepal</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.NP</lx:path>
            <lx:label>Nepal</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NR">
        <xs:annotation>
          <xs:documentation>Nauru</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.NR</lx:path>
            <lx:label>Nauru</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NU">
        <xs:annotation>
          <xs:documentation>Niue</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.NU</lx:path>
            <lx:label>Niue</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NZ">
        <xs:annotation>
          <xs:documentation>New Zealand</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.NZ</lx:path>
            <lx:label>New Zealand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="OM">
        <xs:annotation>
          <xs:documentation>Oman</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.OM</lx:path>
            <lx:label>Oman</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PA">
        <xs:annotation>
          <xs:documentation>Panama</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.PA</lx:path>
            <lx:label>Panama</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PE">
        <xs:annotation>
          <xs:documentation>Peru</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.PE</lx:path>
            <lx:label>Peru</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PF">
        <xs:annotation>
          <xs:documentation>French Polynesia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.PF</lx:path>
            <lx:label>French Polynesia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PG">
        <xs:annotation>
          <xs:documentation>Papua New Guinea</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.PG</lx:path>
            <lx:label>Papua New Guinea</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PH">
        <xs:annotation>
          <xs:documentation>Philippines</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.PH</lx:path>
            <lx:label>Philippines</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PK">
        <xs:annotation>
          <xs:documentation>Pakistan</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.PK</lx:path>
            <lx:label>Pakistan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PL">
        <xs:annotation>
          <xs:documentation>Poland</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.PL</lx:path>
            <lx:label>Poland</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PM">
        <xs:annotation>
          <xs:documentation>Saint Pierre and Miquelon</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.PM</lx:path>
            <lx:label>Saint Pierre and Miquelon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PN">
        <xs:annotation>
          <xs:documentation>Pitcairn</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.PN</lx:path>
            <lx:label>Pitcairn</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PR">
        <xs:annotation>
          <xs:documentation>Puerto Rico</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.PR</lx:path>
            <lx:label>Puerto Rico</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PS">
        <xs:annotation>
          <xs:documentation>Palestine, State of</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.PS</lx:path>
            <lx:label>Palestine, State of</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PT">
        <xs:annotation>
          <xs:documentation>Portugal</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.PT</lx:path>
            <lx:label>Portugal</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PW">
        <xs:annotation>
          <xs:documentation>Palau</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.PW</lx:path>
            <lx:label>Palau</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PY">
        <xs:annotation>
          <xs:documentation>Paraguay</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.PY</lx:path>
            <lx:label>Paraguay</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="QA">
        <xs:annotation>
          <xs:documentation>Qatar</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.QA</lx:path>
            <lx:label>Qatar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RE">
        <xs:annotation>
          <xs:documentation>Réunion</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.RE</lx:path>
            <lx:label>Réunion</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RO">
        <xs:annotation>
          <xs:documentation>Romania</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.RO</lx:path>
            <lx:label>Romania</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RS">
        <xs:annotation>
          <xs:documentation>Serbia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.RS</lx:path>
            <lx:label>Serbia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RU">
        <xs:annotation>
          <xs:documentation>Russian Federation</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.RU</lx:path>
            <lx:label>Russian Federation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RW">
        <xs:annotation>
          <xs:documentation>Rwanda</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.RW</lx:path>
            <lx:label>Rwanda</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SA">
        <xs:annotation>
          <xs:documentation>Saudi Arabia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SA</lx:path>
            <lx:label>Saudi Arabia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SB">
        <xs:annotation>
          <xs:documentation>Solomon Islands</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SB</lx:path>
            <lx:label>Solomon Islands</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SC">
        <xs:annotation>
          <xs:documentation>Seychelles</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SC</lx:path>
            <lx:label>Seychelles</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SD">
        <xs:annotation>
          <xs:documentation>Sudan</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SD</lx:path>
            <lx:label>Sudan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SE">
        <xs:annotation>
          <xs:documentation>Sweden</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SE</lx:path>
            <lx:label>Sweden</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SG">
        <xs:annotation>
          <xs:documentation>Singapore</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SG</lx:path>
            <lx:label>Singapore</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SH">
        <xs:annotation>
          <xs:documentation>Saint Helena, Ascension and Tristan da Cunha</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SH</lx:path>
            <lx:label>Saint Helena, Ascension and Tristan da Cunha</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SI">
        <xs:annotation>
          <xs:documentation>Slovenia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SI</lx:path>
            <lx:label>Slovenia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SJ">
        <xs:annotation>
          <xs:documentation>Svalbard and Jan Mayen</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SJ</lx:path>
            <lx:label>Svalbard and Jan Mayen</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SK">
        <xs:annotation>
          <xs:documentation>Slovakia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SK</lx:path>
            <lx:label>Slovakia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SL">
        <xs:annotation>
          <xs:documentation>Sierra Leone</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SL</lx:path>
            <lx:label>Sierra Leone</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SM">
        <xs:annotation>
          <xs:documentation>San Marino</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SM</lx:path>
            <lx:label>San Marino</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SN">
        <xs:annotation>
          <xs:documentation>Senegal</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SN</lx:path>
            <lx:label>Senegal</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SO">
        <xs:annotation>
          <xs:documentation>Somalia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SO</lx:path>
            <lx:label>Somalia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SR">
        <xs:annotation>
          <xs:documentation>Suriname</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SR</lx:path>
            <lx:label>Suriname</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SS">
        <xs:annotation>
          <xs:documentation>South Sudan</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SS</lx:path>
            <lx:label>South Sudan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ST">
        <xs:annotation>
          <xs:documentation>Sao Tome and Principe</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.ST</lx:path>
            <lx:label>Sao Tome and Principe</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SV">
        <xs:annotation>
          <xs:documentation>El Salvador</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SV</lx:path>
            <lx:label>El Salvador</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SX">
        <xs:annotation>
          <xs:documentation>Sint Maarten (Dutch part)</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SX</lx:path>
            <lx:label>Sint Maarten (Dutch part)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SY">
        <xs:annotation>
          <xs:documentation>Syrian Arab Republic</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SY</lx:path>
            <lx:label>Syrian Arab Republic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SZ">
        <xs:annotation>
          <xs:documentation>Swaziland</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.SZ</lx:path>
            <lx:label>Swaziland</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TC">
        <xs:annotation>
          <xs:documentation>Turks and Caicos Islands</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TC</lx:path>
            <lx:label>Turks and Caicos Islands</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TD">
        <xs:annotation>
          <xs:documentation>Chad</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TD</lx:path>
            <lx:label>Chad</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TF">
        <xs:annotation>
          <xs:documentation>French Southern Territories</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TF</lx:path>
            <lx:label>French Southern Territories</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TG">
        <xs:annotation>
          <xs:documentation>Togo</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TG</lx:path>
            <lx:label>Togo</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TH">
        <xs:annotation>
          <xs:documentation>Thailand</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TH</lx:path>
            <lx:label>Thailand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TJ">
        <xs:annotation>
          <xs:documentation>Tajikistan</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TJ</lx:path>
            <lx:label>Tajikistan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TK">
        <xs:annotation>
          <xs:documentation>Tokelau</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TK</lx:path>
            <lx:label>Tokelau</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TL">
        <xs:annotation>
          <xs:documentation>Timor-Leste</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TL</lx:path>
            <lx:label>Timor-Leste</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TM">
        <xs:annotation>
          <xs:documentation>Turkmenistan</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TM</lx:path>
            <lx:label>Turkmenistan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TN">
        <xs:annotation>
          <xs:documentation>Tunisia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TN</lx:path>
            <lx:label>Tunisia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TO">
        <xs:annotation>
          <xs:documentation>Tonga</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TO</lx:path>
            <lx:label>Tonga</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TR">
        <xs:annotation>
          <xs:documentation>Turkey</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TR</lx:path>
            <lx:label>Turkey</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TT">
        <xs:annotation>
          <xs:documentation>Trinidad and Tobago</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TT</lx:path>
            <lx:label>Trinidad and Tobago</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TV">
        <xs:annotation>
          <xs:documentation>Tuvalu</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TV</lx:path>
            <lx:label>Tuvalu</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TW">
        <xs:annotation>
          <xs:documentation>Taiwan, Province of China</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TW</lx:path>
            <lx:label>Taiwan, Province of China</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TZ">
        <xs:annotation>
          <xs:documentation>Tanzania, United Republic of</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.TZ</lx:path>
            <lx:label>Tanzania, United Republic of</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="UA">
        <xs:annotation>
          <xs:documentation>Ukraine</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.UA</lx:path>
            <lx:label>Ukraine</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="UG">
        <xs:annotation>
          <xs:documentation>Uganda</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.UG</lx:path>
            <lx:label>Uganda</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="UM">
        <xs:annotation>
          <xs:documentation>United States Minor Outlying Islands</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.UM</lx:path>
            <lx:label>United States Minor Outlying Islands</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="US">
        <xs:annotation>
          <xs:documentation>United States</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.US</lx:path>
            <lx:label>United States</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="UY">
        <xs:annotation>
          <xs:documentation>Uruguay</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.UY</lx:path>
            <lx:label>Uruguay</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="UZ">
        <xs:annotation>
          <xs:documentation>Uzbekistan</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.UZ</lx:path>
            <lx:label>Uzbekistan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="VA">
        <xs:annotation>
          <xs:documentation>Holy See (Vatican City State)</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.VA</lx:path>
            <lx:label>Holy See (Vatican City State)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="VC">
        <xs:annotation>
          <xs:documentation>Saint Vincent and the Grenadines</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.VC</lx:path>
            <lx:label>Saint Vincent and the Grenadines</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="VE">
        <xs:annotation>
          <xs:documentation>Venezuela, Bolivarian Republic of</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.VE</lx:path>
            <lx:label>Venezuela, Bolivarian Republic of</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="VG">
        <xs:annotation>
          <xs:documentation>Virgin Islands, British</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.VG</lx:path>
            <lx:label>Virgin Islands, British</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="VI">
        <xs:annotation>
          <xs:documentation>Virgin Islands, U.S.</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.VI</lx:path>
            <lx:label>Virgin Islands, U.S.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="VN">
        <xs:annotation>
          <xs:documentation>Viet Nam</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.VN</lx:path>
            <lx:label>Viet Nam</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="VU">
        <xs:annotation>
          <xs:documentation>Vanuatu</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.VU</lx:path>
            <lx:label>Vanuatu</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="WF">
        <xs:annotation>
          <xs:documentation>Wallis and Futuna</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.WF</lx:path>
            <lx:label>Wallis and Futuna</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="WS">
        <xs:annotation>
          <xs:documentation>Samoa</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.WS</lx:path>
            <lx:label>Samoa</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="YE">
        <xs:annotation>
          <xs:documentation>Yemen</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.YE</lx:path>
            <lx:label>Yemen</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="YT">
        <xs:annotation>
          <xs:documentation>Mayotte</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.YT</lx:path>
            <lx:label>Mayotte</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ZA">
        <xs:annotation>
          <xs:documentation>South Africa</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.ZA</lx:path>
            <lx:label>South Africa</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ZM">
        <xs:annotation>
          <xs:documentation>Zambia</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.ZM</lx:path>
            <lx:label>Zambia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ZW">
        <xs:annotation>
          <xs:documentation>Zimbabwe</xs:documentation>
          <xs:appinfo>
            <lx:path>countryCodeList.ZW</lx:path>
            <lx:label>Zimbabwe</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="creditAssessmentDecisionScoreDeterminedList">
    <xs:annotation>
      <xs:documentation>creditAssessmentDecisionScoreDeterminedList</xs:documentation>
      <xs:appinfo>
        <lx:path>creditAssessmentDecisionScoreDeterminedList</lx:path>
        <lx:label>creditAssessmentDecisionScoreDeterminedList</lx:label>
        <li:transactions>CDA,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Behavioural Score">
        <xs:annotation>
          <xs:documentation>The decision score was determined via behavioural score.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentDecisionScoreDeterminedList.Behavioural Score</lx:path>
            <lx:label>Behavioural Score</lx:label>
            <li:transactions>CDA,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Post Bureau">
        <xs:annotation>
          <xs:documentation>The decision score was determined after bureau interaction.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentDecisionScoreDeterminedList.Post Bureau</lx:path>
            <lx:label>Post Bureau</lx:label>
            <li:transactions>CDA,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pre Bureau">
        <xs:annotation>
          <xs:documentation>The decision score was determined prior to bureau interaction.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentDecisionScoreDeterminedList.Pre Bureau</lx:path>
            <lx:label>Pre Bureau</lx:label>
            <li:transactions>CDA,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Raw Bureau">
        <xs:annotation>
          <xs:documentation>The score was determined directly by the bureau.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentDecisionScoreDeterminedList.Raw Bureau</lx:path>
            <lx:label>Raw Bureau</lx:label>
            <li:transactions>CDA,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="creditAssessmentPolicyExceptionCategoryList">
    <xs:annotation>
      <xs:documentation>creditAssessmentPolicyExceptionCategoryList</xs:documentation>
      <xs:appinfo>
        <lx:path>creditAssessmentPolicyExceptionCategoryList</lx:path>
        <lx:label>creditAssessmentPolicyExceptionCategoryList</lx:label>
        <li:transactions>CDA</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Bankruptcy">
        <xs:annotation>
          <xs:documentation>The policy exception is related to bankruptcy.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentPolicyExceptionCategoryList.Bankruptcy</lx:path>
            <lx:label>Bankruptcy</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Court">
        <xs:annotation>
          <xs:documentation>The policy exception is related to court judgements.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentPolicyExceptionCategoryList.Court</lx:path>
            <lx:label>Court</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Rating">
        <xs:annotation>
          <xs:documentation>The policy exception is related to the applicant's credit rating.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentPolicyExceptionCategoryList.Credit Rating</lx:path>
            <lx:label>Credit Rating</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fraud">
        <xs:annotation>
          <xs:documentation>The policy exception is related to fraud.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentPolicyExceptionCategoryList.Fraud</lx:path>
            <lx:label>Fraud</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan to Valuation Ratio">
        <xs:annotation>
          <xs:documentation>The policy exception is related to loan to valuation ratio.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentPolicyExceptionCategoryList.Loan to Valuation Ratio</lx:path>
            <lx:label>Loan to Valuation Ratio</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Score">
        <xs:annotation>
          <xs:documentation>The policy exception is related to the score determined for the overall application or an actor in the application context such as a party or security.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentPolicyExceptionCategoryList.Score</lx:path>
            <lx:label>Score</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security">
        <xs:annotation>
          <xs:documentation>The policy exception is related to the application security.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentPolicyExceptionCategoryList.Security</lx:path>
            <lx:label>Security</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="creditAssessmentRecommendationList">
    <xs:annotation>
      <xs:documentation>creditAssessmentRecommendationList</xs:documentation>
      <xs:appinfo>
        <lx:path>creditAssessmentRecommendationList</lx:path>
        <lx:label>creditAssessmentRecommendationList</lx:label>
        <li:transactions>CDA</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Approve">
        <xs:annotation>
          <xs:documentation>The recommendation is to approve the application.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentRecommendationList.Approve</lx:path>
            <lx:label>Approve</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Decline">
        <xs:annotation>
          <xs:documentation>The recommendation is to decline the application.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentRecommendationList.Decline</lx:path>
            <lx:label>Decline</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Error">
        <xs:annotation>
          <xs:documentation>A recommendation could not be reached due to an error in the processing.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentRecommendationList.Error</lx:path>
            <lx:label>Error</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Refer">
        <xs:annotation>
          <xs:documentation>The recommendation is to refer the application for further assessment which is often handled manually.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentRecommendationList.Refer</lx:path>
            <lx:label>Refer</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Refer Fraud">
        <xs:annotation>
          <xs:documentation>The recommendation is to refer the application for further assessment due to possible fraudulent activities being detected which is often handled manually.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentRecommendationList.Refer Fraud</lx:path>
            <lx:label>Refer Fraud</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Withdraw">
        <xs:annotation>
          <xs:documentation>The recommendation is to withdraw the application.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentRecommendationList.Withdraw</lx:path>
            <lx:label>Withdraw</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="creditAssessmentTypeInstructionList">
    <xs:annotation>
      <xs:documentation>creditAssessmentTypeInstructionList</xs:documentation>
      <xs:appinfo>
        <lx:path>creditAssessmentTypeInstructionList</lx:path>
        <lx:label>creditAssessmentTypeInstructionList</lx:label>
        <li:transactions>CDA</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Post Bureau">
        <xs:annotation>
          <xs:documentation>The credit assessment should occur after bureau reports have been retrieved and their data can be used in the process.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentTypeInstructionList.Post Bureau</lx:path>
            <lx:label>Post Bureau</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pre Bureau">
        <xs:annotation>
          <xs:documentation>The credit assessment should occur before bureau reports have been retrieved. As bureau report retrieval can incur a fee, this type of assessment can help in filtering out cases a credit provider would not usually take on, before they are charged.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditAssessmentTypeInstructionList.Pre Bureau</lx:path>
            <lx:label>Pre Bureau</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="creditCardTypeList">
    <xs:annotation>
      <xs:documentation>creditCardTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>creditCardTypeList</lx:path>
        <lx:label>creditCardTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="American Express">
        <xs:annotation>
          <xs:documentation>American Express</xs:documentation>
          <xs:appinfo>
            <lx:path>creditCardTypeList.American Express</lx:path>
            <lx:label>American Express</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bank Card">
        <xs:annotation>
          <xs:documentation>Bank Card</xs:documentation>
          <xs:appinfo>
            <lx:path>creditCardTypeList.Bank Card</lx:path>
            <lx:label>Bank Card</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Diners Club">
        <xs:annotation>
          <xs:documentation>Diners Club</xs:documentation>
          <xs:appinfo>
            <lx:path>creditCardTypeList.Diners Club</lx:path>
            <lx:label>Diners Club</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MasterCard">
        <xs:annotation>
          <xs:documentation>MasterCard</xs:documentation>
          <xs:appinfo>
            <lx:path>creditCardTypeList.MasterCard</lx:path>
            <lx:label>MasterCard</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other Credit Card">
        <xs:annotation>
          <xs:documentation>Other Credit Card</xs:documentation>
          <xs:appinfo>
            <lx:path>creditCardTypeList.Other Credit Card</lx:path>
            <lx:label>Other Credit Card</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Visa">
        <xs:annotation>
          <xs:documentation>Visa</xs:documentation>
          <xs:appinfo>
            <lx:path>creditCardTypeList.Visa</lx:path>
            <lx:label>Visa</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="creditHistoryIssueList">
    <xs:annotation>
      <xs:documentation>creditHistoryIssueList</xs:documentation>
      <xs:appinfo>
        <lx:path>creditHistoryIssueList</lx:path>
        <lx:label>creditHistoryIssueList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Application for Hardship with Existing Lender">
        <xs:annotation>
          <xs:documentation>Application for Hardship with Existing Lender</xs:documentation>
          <xs:appinfo>
            <lx:path>creditHistoryIssueList.Application for Hardship with Existing Lender</lx:path>
            <lx:label>Application for Hardship with Existing Lender</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Arrears">
        <xs:annotation>
          <xs:documentation>Arrears</xs:documentation>
          <xs:appinfo>
            <lx:path>creditHistoryIssueList.Arrears</lx:path>
            <lx:label>Arrears</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bankruptcy">
        <xs:annotation>
          <xs:documentation>Bankruptcy</xs:documentation>
          <xs:appinfo>
            <lx:path>creditHistoryIssueList.Bankruptcy</lx:path>
            <lx:label>Bankruptcy</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Default">
        <xs:annotation>
          <xs:documentation>Default</xs:documentation>
          <xs:appinfo>
            <lx:path>creditHistoryIssueList.Default</lx:path>
            <lx:label>Default</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Financial Stress from Current Commitments">
        <xs:annotation>
          <xs:documentation>Financial Stress from Current Commitments</xs:documentation>
          <xs:appinfo>
            <lx:path>creditHistoryIssueList.Financial Stress from Current Commitments</lx:path>
            <lx:label>Financial Stress from Current Commitments</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Insolvency">
        <xs:annotation>
          <xs:documentation>Insolvency</xs:documentation>
          <xs:appinfo>
            <lx:path>creditHistoryIssueList.Insolvency</lx:path>
            <lx:label>Insolvency</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Judgement">
        <xs:annotation>
          <xs:documentation>Judgement</xs:documentation>
          <xs:appinfo>
            <lx:path>creditHistoryIssueList.Judgement</lx:path>
            <lx:label>Judgement</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Legal Proceeding">
        <xs:annotation>
          <xs:documentation>Legal Proceeding</xs:documentation>
          <xs:appinfo>
            <lx:path>creditHistoryIssueList.Legal Proceeding</lx:path>
            <lx:label>Legal Proceeding</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>creditHistoryIssueList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Scheme of Arrangement">
        <xs:annotation>
          <xs:documentation>Scheme of Arrangement</xs:documentation>
          <xs:appinfo>
            <lx:path>creditHistoryIssueList.Scheme of Arrangement</lx:path>
            <lx:label>Scheme of Arrangement</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Writ">
        <xs:annotation>
          <xs:documentation>Writ</xs:documentation>
          <xs:appinfo>
            <lx:path>creditHistoryIssueList.Writ</lx:path>
            <lx:label>Writ</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="creditReportingBodyList">
    <xs:annotation>
      <xs:documentation>creditReportingBodyList</xs:documentation>
      <xs:appinfo>
        <lx:path>creditReportingBodyList</lx:path>
        <lx:label>creditReportingBodyList</lx:label>
        <li:transactions>CAL,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="D and B">
        <xs:annotation>
          <xs:documentation>The credit reporting body who reported the default is Dun and Bradstreet</xs:documentation>
          <xs:appinfo>
            <lx:path>creditReportingBodyList.D and B</lx:path>
            <lx:label>The credit reporting body who reported the default is Dun and Bradstreet</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Experian">
        <xs:annotation>
          <xs:documentation>The credit reporting body who reported the default is Experian</xs:documentation>
          <xs:appinfo>
            <lx:path>creditReportingBodyList.Experian</lx:path>
            <lx:label>The credit reporting body who reported the default is Experian</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Veda">
        <xs:annotation>
          <xs:documentation>The credit reporting body who reported the default is Veda</xs:documentation>
          <xs:appinfo>
            <lx:path>creditReportingBodyList.Veda</lx:path>
            <lx:label>The credit reporting body who reported the default is Veda</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="creditStatusList">
    <xs:annotation>
      <xs:documentation>creditStatusList</xs:documentation>
      <xs:appinfo>
        <lx:path>creditStatusList</lx:path>
        <lx:label>creditStatusList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Clean">
        <xs:annotation>
          <xs:documentation>The party's credit history discloses no defaults, writ and summonses, bankruptcy/insolvency or other credit issues; and the applicant shall be able to indicate whether they have answered "No" to both questions on credit performance in the Applicant Declaration Authority and Acknowledgement section of the application form.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditStatusList.Clean</lx:path>
            <lx:label>The party's credit history discloses no defaults, writ and summonses, bankruptcy/insolvency or other credit issues; and the applicant shall be able to indicate whether they have answered "No" to both questions on credit performance in the Applicant Declaration Authority and Acknowledgement section of the application form.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unknown">
        <xs:annotation>
          <xs:documentation>Unknown</xs:documentation>
          <xs:appinfo>
            <lx:path>creditStatusList.Unknown</lx:path>
            <lx:label>Unknown</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Warning">
        <xs:annotation>
          <xs:documentation>The party's credit history discloses defaults, writ and summonses, bankruptcy/insolvency or other credit issues; or the applicant has answered "Yes" to one of the questions on credit performance in the Applicant Declaration Authority and Acknowledgement section of the application form.</xs:documentation>
          <xs:appinfo>
            <lx:path>creditStatusList.Warning</lx:path>
            <lx:label>The party's credit history discloses defaults, writ and summonses, bankruptcy/insolvency or other credit issues; or the applicant has answered "Yes" to one of the questions on credit performance in the Applicant Declaration Authority and Acknowledgement section of the application form.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="currencyCodeList">
    <xs:annotation>
      <xs:documentation>currencyCodeList</xs:documentation>
      <xs:appinfo>
        <lx:path>currencyCodeList</lx:path>
        <lx:label>currencyCodeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="AED">
        <xs:annotation>
          <xs:documentation>UAE Dirham</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.AED</lx:path>
            <lx:label>UAE Dirham</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AFN">
        <xs:annotation>
          <xs:documentation>Afghani</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.AFN</lx:path>
            <lx:label>Afghani</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ALL">
        <xs:annotation>
          <xs:documentation>Lek</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.ALL</lx:path>
            <lx:label>Lek</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AMD">
        <xs:annotation>
          <xs:documentation>Armenian Dram</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.AMD</lx:path>
            <lx:label>Armenian Dram</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ANG">
        <xs:annotation>
          <xs:documentation>Netherlands Antillean Guilder</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.ANG</lx:path>
            <lx:label>Netherlands Antillean Guilder</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AOA">
        <xs:annotation>
          <xs:documentation>Kwanza</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.AOA</lx:path>
            <lx:label>Kwanza</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ARS">
        <xs:annotation>
          <xs:documentation>Argentine Peso</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.ARS</lx:path>
            <lx:label>Argentine Peso</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AUD">
        <xs:annotation>
          <xs:documentation>Australian Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.AUD</lx:path>
            <lx:label>Australian Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AWG">
        <xs:annotation>
          <xs:documentation>Aruban Florin</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.AWG</lx:path>
            <lx:label>Aruban Florin</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AZN">
        <xs:annotation>
          <xs:documentation>Azerbaijanian Manat</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.AZN</lx:path>
            <lx:label>Azerbaijanian Manat</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BAM">
        <xs:annotation>
          <xs:documentation>Convertible Mark</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BAM</lx:path>
            <lx:label>Convertible Mark</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BBD">
        <xs:annotation>
          <xs:documentation>Barbados Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BBD</lx:path>
            <lx:label>Barbados Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BDT">
        <xs:annotation>
          <xs:documentation>Taka</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BDT</lx:path>
            <lx:label>Taka</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BGN">
        <xs:annotation>
          <xs:documentation>Bulgarian Lev</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BGN</lx:path>
            <lx:label>Bulgarian Lev</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BHD">
        <xs:annotation>
          <xs:documentation>Bahraini Dinar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BHD</lx:path>
            <lx:label>Bahraini Dinar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BIF">
        <xs:annotation>
          <xs:documentation>Burundi Franc</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BIF</lx:path>
            <lx:label>Burundi Franc</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BMD">
        <xs:annotation>
          <xs:documentation>Bermudian Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BMD</lx:path>
            <lx:label>Bermudian Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BND">
        <xs:annotation>
          <xs:documentation>Brunei Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BND</lx:path>
            <lx:label>Brunei Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BOB">
        <xs:annotation>
          <xs:documentation>Boliviano</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BOB</lx:path>
            <lx:label>Boliviano</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BOV">
        <xs:annotation>
          <xs:documentation>Mvdol</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BOV</lx:path>
            <lx:label>Mvdol</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BRL">
        <xs:annotation>
          <xs:documentation>Brazilian Real</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BRL</lx:path>
            <lx:label>Brazilian Real</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BSD">
        <xs:annotation>
          <xs:documentation>Bahamian Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BSD</lx:path>
            <lx:label>Bahamian Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BTN">
        <xs:annotation>
          <xs:documentation>Ngultrum</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BTN</lx:path>
            <lx:label>Ngultrum</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BWP">
        <xs:annotation>
          <xs:documentation>Pula</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BWP</lx:path>
            <lx:label>Pula</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BYR">
        <xs:annotation>
          <xs:documentation>Belarussian Ruble</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BYR</lx:path>
            <lx:label>Belarussian Ruble</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BZD">
        <xs:annotation>
          <xs:documentation>Belize Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.BZD</lx:path>
            <lx:label>Belize Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CAD">
        <xs:annotation>
          <xs:documentation>Canadian Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.CAD</lx:path>
            <lx:label>Canadian Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CDF">
        <xs:annotation>
          <xs:documentation>Congolese Franc</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.CDF</lx:path>
            <lx:label>Congolese Franc</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CHE">
        <xs:annotation>
          <xs:documentation>WIR Euro</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.CHE</lx:path>
            <lx:label>WIR Euro</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CHF">
        <xs:annotation>
          <xs:documentation>Swiss Franc</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.CHF</lx:path>
            <lx:label>Swiss Franc</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CHW">
        <xs:annotation>
          <xs:documentation>WIR Franc</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.CHW</lx:path>
            <lx:label>WIR Franc</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CLF">
        <xs:annotation>
          <xs:documentation>Unidad de Fomento</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.CLF</lx:path>
            <lx:label>Unidad de Fomento</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CLP">
        <xs:annotation>
          <xs:documentation>Chilean Peso</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.CLP</lx:path>
            <lx:label>Chilean Peso</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CNY">
        <xs:annotation>
          <xs:documentation>Yuan Renminbi</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.CNY</lx:path>
            <lx:label>Yuan Renminbi</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="COP">
        <xs:annotation>
          <xs:documentation>Colombian Peso</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.COP</lx:path>
            <lx:label>Colombian Peso</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="COU">
        <xs:annotation>
          <xs:documentation>Unidad de Valor Real</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.COU</lx:path>
            <lx:label>Unidad de Valor Real</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CRC">
        <xs:annotation>
          <xs:documentation>Costa Rican Colon</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.CRC</lx:path>
            <lx:label>Costa Rican Colon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CUC">
        <xs:annotation>
          <xs:documentation>Peso Convertible</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.CUC</lx:path>
            <lx:label>Peso Convertible</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CUP">
        <xs:annotation>
          <xs:documentation>Cuban Peso</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.CUP</lx:path>
            <lx:label>Cuban Peso</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CVE">
        <xs:annotation>
          <xs:documentation>Cape Verde Escudo</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.CVE</lx:path>
            <lx:label>Cape Verde Escudo</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CZK">
        <xs:annotation>
          <xs:documentation>Czech Koruna</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.CZK</lx:path>
            <lx:label>Czech Koruna</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="DJF">
        <xs:annotation>
          <xs:documentation>Djibouti Franc</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.DJF</lx:path>
            <lx:label>Djibouti Franc</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="DKK">
        <xs:annotation>
          <xs:documentation>Danish Krone</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.DKK</lx:path>
            <lx:label>Danish Krone</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="DOP">
        <xs:annotation>
          <xs:documentation>Dominican Peso</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.DOP</lx:path>
            <lx:label>Dominican Peso</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="DZD">
        <xs:annotation>
          <xs:documentation>Algerian Dinar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.DZD</lx:path>
            <lx:label>Algerian Dinar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="EGP">
        <xs:annotation>
          <xs:documentation>Egyptian Pound</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.EGP</lx:path>
            <lx:label>Egyptian Pound</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ERN">
        <xs:annotation>
          <xs:documentation>Nakfa</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.ERN</lx:path>
            <lx:label>Nakfa</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ETB">
        <xs:annotation>
          <xs:documentation>Ethiopian Birr</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.ETB</lx:path>
            <lx:label>Ethiopian Birr</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="EUR">
        <xs:annotation>
          <xs:documentation>Euro</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.EUR</lx:path>
            <lx:label>Euro</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FJD">
        <xs:annotation>
          <xs:documentation>Fiji Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.FJD</lx:path>
            <lx:label>Fiji Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FKP">
        <xs:annotation>
          <xs:documentation>Falkland Islands Pound</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.FKP</lx:path>
            <lx:label>Falkland Islands Pound</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GBP">
        <xs:annotation>
          <xs:documentation>Pound Sterling</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.GBP</lx:path>
            <lx:label>Pound Sterling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GEL">
        <xs:annotation>
          <xs:documentation>Lari</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.GEL</lx:path>
            <lx:label>Lari</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GHS">
        <xs:annotation>
          <xs:documentation>Ghana Cedi</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.GHS</lx:path>
            <lx:label>Ghana Cedi</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GIP">
        <xs:annotation>
          <xs:documentation>Gibraltar Pound</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.GIP</lx:path>
            <lx:label>Gibraltar Pound</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GMD">
        <xs:annotation>
          <xs:documentation>Dalasi</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.GMD</lx:path>
            <lx:label>Dalasi</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GNF">
        <xs:annotation>
          <xs:documentation>Guinea Franc</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.GNF</lx:path>
            <lx:label>Guinea Franc</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GTQ">
        <xs:annotation>
          <xs:documentation>Quetzal</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.GTQ</lx:path>
            <lx:label>Quetzal</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GYD">
        <xs:annotation>
          <xs:documentation>Guyana Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.GYD</lx:path>
            <lx:label>Guyana Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HKD">
        <xs:annotation>
          <xs:documentation>Hong Kong Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.HKD</lx:path>
            <lx:label>Hong Kong Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HNL">
        <xs:annotation>
          <xs:documentation>Lempira</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.HNL</lx:path>
            <lx:label>Lempira</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HRK">
        <xs:annotation>
          <xs:documentation>Croatian Kuna</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.HRK</lx:path>
            <lx:label>Croatian Kuna</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HTG">
        <xs:annotation>
          <xs:documentation>Gourde</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.HTG</lx:path>
            <lx:label>Gourde</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HUF">
        <xs:annotation>
          <xs:documentation>Forint</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.HUF</lx:path>
            <lx:label>Forint</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IDR">
        <xs:annotation>
          <xs:documentation>Rupiah</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.IDR</lx:path>
            <lx:label>Rupiah</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ILS">
        <xs:annotation>
          <xs:documentation>New Israeli Sheqel</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.ILS</lx:path>
            <lx:label>New Israeli Sheqel</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="INR">
        <xs:annotation>
          <xs:documentation>Indian Rupee</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.INR</lx:path>
            <lx:label>Indian Rupee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IQD">
        <xs:annotation>
          <xs:documentation>Iraqi Dinar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.IQD</lx:path>
            <lx:label>Iraqi Dinar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IRR">
        <xs:annotation>
          <xs:documentation>Iranian Rial</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.IRR</lx:path>
            <lx:label>Iranian Rial</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ISK">
        <xs:annotation>
          <xs:documentation>Iceland Krona</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.ISK</lx:path>
            <lx:label>Iceland Krona</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="JMD">
        <xs:annotation>
          <xs:documentation>Jamaican Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.JMD</lx:path>
            <lx:label>Jamaican Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="JOD">
        <xs:annotation>
          <xs:documentation>Jordanian Dinar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.JOD</lx:path>
            <lx:label>Jordanian Dinar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="JPY">
        <xs:annotation>
          <xs:documentation>Yen</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.JPY</lx:path>
            <lx:label>Yen</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KES">
        <xs:annotation>
          <xs:documentation>Kenyan Shilling</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.KES</lx:path>
            <lx:label>Kenyan Shilling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KGS">
        <xs:annotation>
          <xs:documentation>Som</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.KGS</lx:path>
            <lx:label>Som</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KHR">
        <xs:annotation>
          <xs:documentation>Riel</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.KHR</lx:path>
            <lx:label>Riel</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KMF">
        <xs:annotation>
          <xs:documentation>Comoro Franc</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.KMF</lx:path>
            <lx:label>Comoro Franc</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KPW">
        <xs:annotation>
          <xs:documentation>North Korean Won</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.KPW</lx:path>
            <lx:label>North Korean Won</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KRW">
        <xs:annotation>
          <xs:documentation>Won</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.KRW</lx:path>
            <lx:label>Won</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KWD">
        <xs:annotation>
          <xs:documentation>Kuwaiti Dinar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.KWD</lx:path>
            <lx:label>Kuwaiti Dinar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KYD">
        <xs:annotation>
          <xs:documentation>Cayman Islands Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.KYD</lx:path>
            <lx:label>Cayman Islands Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KZT">
        <xs:annotation>
          <xs:documentation>Tenge</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.KZT</lx:path>
            <lx:label>Tenge</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LAK">
        <xs:annotation>
          <xs:documentation>Kip</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.LAK</lx:path>
            <lx:label>Kip</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LBP">
        <xs:annotation>
          <xs:documentation>Lebanese Pound</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.LBP</lx:path>
            <lx:label>Lebanese Pound</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LKR">
        <xs:annotation>
          <xs:documentation>Sri Lanka Rupee</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.LKR</lx:path>
            <lx:label>Sri Lanka Rupee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LRD">
        <xs:annotation>
          <xs:documentation>Liberian Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.LRD</lx:path>
            <lx:label>Liberian Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LSL">
        <xs:annotation>
          <xs:documentation>Loti</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.LSL</lx:path>
            <lx:label>Loti</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LTL">
        <xs:annotation>
          <xs:documentation>Lithuanian Litas</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.LTL</lx:path>
            <lx:label>Lithuanian Litas</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LYD">
        <xs:annotation>
          <xs:documentation>Libyan Dinar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.LYD</lx:path>
            <lx:label>Libyan Dinar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MAD">
        <xs:annotation>
          <xs:documentation>Moroccan Dirham</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MAD</lx:path>
            <lx:label>Moroccan Dirham</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MDL">
        <xs:annotation>
          <xs:documentation>Moldovan Leu</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MDL</lx:path>
            <lx:label>Moldovan Leu</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MGA">
        <xs:annotation>
          <xs:documentation>Malagasy Ariary</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MGA</lx:path>
            <lx:label>Malagasy Ariary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MKD">
        <xs:annotation>
          <xs:documentation>Denar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MKD</lx:path>
            <lx:label>Denar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MMK">
        <xs:annotation>
          <xs:documentation>Kyat</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MMK</lx:path>
            <lx:label>Kyat</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MNT">
        <xs:annotation>
          <xs:documentation>Tugrik</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MNT</lx:path>
            <lx:label>Tugrik</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MOP">
        <xs:annotation>
          <xs:documentation>Pataca</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MOP</lx:path>
            <lx:label>Pataca</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MRO">
        <xs:annotation>
          <xs:documentation>Ouguiya</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MRO</lx:path>
            <lx:label>Ouguiya</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MUR">
        <xs:annotation>
          <xs:documentation>Mauritius Rupee</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MUR</lx:path>
            <lx:label>Mauritius Rupee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MVR">
        <xs:annotation>
          <xs:documentation>Rufiyaa</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MVR</lx:path>
            <lx:label>Rufiyaa</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MWK">
        <xs:annotation>
          <xs:documentation>Kwacha</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MWK</lx:path>
            <lx:label>Kwacha</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MXN">
        <xs:annotation>
          <xs:documentation>Mexican Peso</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MXN</lx:path>
            <lx:label>Mexican Peso</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MXV">
        <xs:annotation>
          <xs:documentation>Mexican Unidad de Inversion (UDI)</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MXV</lx:path>
            <lx:label>Mexican Unidad de Inversion (UDI)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MYR">
        <xs:annotation>
          <xs:documentation>Malaysian Ringgit</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MYR</lx:path>
            <lx:label>Malaysian Ringgit</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MZN">
        <xs:annotation>
          <xs:documentation>Mozambique Metical</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.MZN</lx:path>
            <lx:label>Mozambique Metical</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NAD">
        <xs:annotation>
          <xs:documentation>Namibia Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.NAD</lx:path>
            <lx:label>Namibia Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NGN">
        <xs:annotation>
          <xs:documentation>Naira</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.NGN</lx:path>
            <lx:label>Naira</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NIO">
        <xs:annotation>
          <xs:documentation>Cordoba Oro</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.NIO</lx:path>
            <lx:label>Cordoba Oro</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NOK">
        <xs:annotation>
          <xs:documentation>Norwegian Krone</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.NOK</lx:path>
            <lx:label>Norwegian Krone</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NPR">
        <xs:annotation>
          <xs:documentation>Nepalese Rupee</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.NPR</lx:path>
            <lx:label>Nepalese Rupee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NZD">
        <xs:annotation>
          <xs:documentation>New Zealand Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.NZD</lx:path>
            <lx:label>New Zealand Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="OMR">
        <xs:annotation>
          <xs:documentation>Rial Omani</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.OMR</lx:path>
            <lx:label>Rial Omani</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PAB">
        <xs:annotation>
          <xs:documentation>Balboa</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.PAB</lx:path>
            <lx:label>Balboa</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PEN">
        <xs:annotation>
          <xs:documentation>Nuevo Sol</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.PEN</lx:path>
            <lx:label>Nuevo Sol</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PGK">
        <xs:annotation>
          <xs:documentation>Kina</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.PGK</lx:path>
            <lx:label>Kina</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PHP">
        <xs:annotation>
          <xs:documentation>Philippine Peso</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.PHP</lx:path>
            <lx:label>Philippine Peso</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PKR">
        <xs:annotation>
          <xs:documentation>Pakistan Rupee</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.PKR</lx:path>
            <lx:label>Pakistan Rupee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PLN">
        <xs:annotation>
          <xs:documentation>Zloty</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.PLN</lx:path>
            <lx:label>Zloty</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PYG">
        <xs:annotation>
          <xs:documentation>Guarani</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.PYG</lx:path>
            <lx:label>Guarani</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="QAR">
        <xs:annotation>
          <xs:documentation>Qatari Rial</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.QAR</lx:path>
            <lx:label>Qatari Rial</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RON">
        <xs:annotation>
          <xs:documentation>New Romanian Leu</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.RON</lx:path>
            <lx:label>New Romanian Leu</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RSD">
        <xs:annotation>
          <xs:documentation>Serbian Dinar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.RSD</lx:path>
            <lx:label>Serbian Dinar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RUB">
        <xs:annotation>
          <xs:documentation>Russian Ruble</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.RUB</lx:path>
            <lx:label>Russian Ruble</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RWF">
        <xs:annotation>
          <xs:documentation>Rwanda Franc</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.RWF</lx:path>
            <lx:label>Rwanda Franc</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SAR">
        <xs:annotation>
          <xs:documentation>Saudi Riyal</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.SAR</lx:path>
            <lx:label>Saudi Riyal</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SBD">
        <xs:annotation>
          <xs:documentation>Solomon Islands Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.SBD</lx:path>
            <lx:label>Solomon Islands Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SCR">
        <xs:annotation>
          <xs:documentation>Seychelles Rupee</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.SCR</lx:path>
            <lx:label>Seychelles Rupee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SDG">
        <xs:annotation>
          <xs:documentation>Sudanese Pound</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.SDG</lx:path>
            <lx:label>Sudanese Pound</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SEK">
        <xs:annotation>
          <xs:documentation>Swedish Krona</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.SEK</lx:path>
            <lx:label>Swedish Krona</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SGD">
        <xs:annotation>
          <xs:documentation>Singapore Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.SGD</lx:path>
            <lx:label>Singapore Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SHP">
        <xs:annotation>
          <xs:documentation>Saint Helena Pound</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.SHP</lx:path>
            <lx:label>Saint Helena Pound</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SLL">
        <xs:annotation>
          <xs:documentation>Leone</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.SLL</lx:path>
            <lx:label>Leone</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SOS">
        <xs:annotation>
          <xs:documentation>Somali Shilling</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.SOS</lx:path>
            <lx:label>Somali Shilling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SRD">
        <xs:annotation>
          <xs:documentation>Surinam Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.SRD</lx:path>
            <lx:label>Surinam Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SSP">
        <xs:annotation>
          <xs:documentation>South Sudanese Pound</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.SSP</lx:path>
            <lx:label>South Sudanese Pound</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="STD">
        <xs:annotation>
          <xs:documentation>Dobra</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.STD</lx:path>
            <lx:label>Dobra</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SVC">
        <xs:annotation>
          <xs:documentation>El Salvador Colon</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.SVC</lx:path>
            <lx:label>El Salvador Colon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SYP">
        <xs:annotation>
          <xs:documentation>Syrian Pound</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.SYP</lx:path>
            <lx:label>Syrian Pound</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SZL">
        <xs:annotation>
          <xs:documentation>Lilangeni</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.SZL</lx:path>
            <lx:label>Lilangeni</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="THB">
        <xs:annotation>
          <xs:documentation>Baht</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.THB</lx:path>
            <lx:label>Baht</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TJS">
        <xs:annotation>
          <xs:documentation>Somoni</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.TJS</lx:path>
            <lx:label>Somoni</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TMT">
        <xs:annotation>
          <xs:documentation>Turkmenistan New Manat</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.TMT</lx:path>
            <lx:label>Turkmenistan New Manat</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TND">
        <xs:annotation>
          <xs:documentation>Tunisian Dinar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.TND</lx:path>
            <lx:label>Tunisian Dinar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TOP">
        <xs:annotation>
          <xs:documentation>Pa'anga</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.TOP</lx:path>
            <lx:label>Pa'anga</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TRY">
        <xs:annotation>
          <xs:documentation>Turkish Lira</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.TRY</lx:path>
            <lx:label>Turkish Lira</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TTD">
        <xs:annotation>
          <xs:documentation>Trinidad and Tobago Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.TTD</lx:path>
            <lx:label>Trinidad and Tobago Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TWD">
        <xs:annotation>
          <xs:documentation>New Taiwan Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.TWD</lx:path>
            <lx:label>New Taiwan Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TZS">
        <xs:annotation>
          <xs:documentation>Tanzanian Shilling</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.TZS</lx:path>
            <lx:label>Tanzanian Shilling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="UAH">
        <xs:annotation>
          <xs:documentation>Hryvnia</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.UAH</lx:path>
            <lx:label>Hryvnia</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="UGX">
        <xs:annotation>
          <xs:documentation>Uganda Shilling</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.UGX</lx:path>
            <lx:label>Uganda Shilling</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="USD">
        <xs:annotation>
          <xs:documentation>US Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.USD</lx:path>
            <lx:label>US Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="USN">
        <xs:annotation>
          <xs:documentation>US Dollar (Next day)</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.USN</lx:path>
            <lx:label>US Dollar (Next day)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="UYI">
        <xs:annotation>
          <xs:documentation>Uruguay Peso en Unidades Indexadas (URUIURUI)</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.UYI</lx:path>
            <lx:label>Uruguay Peso en Unidades Indexadas (URUIURUI)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="UYU">
        <xs:annotation>
          <xs:documentation>Peso Uruguayo</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.UYU</lx:path>
            <lx:label>Peso Uruguayo</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="UZS">
        <xs:annotation>
          <xs:documentation>Uzbekistan Sum</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.UZS</lx:path>
            <lx:label>Uzbekistan Sum</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="VEF">
        <xs:annotation>
          <xs:documentation>Bolivar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.VEF</lx:path>
            <lx:label>Bolivar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="VND">
        <xs:annotation>
          <xs:documentation>Dong</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.VND</lx:path>
            <lx:label>Dong</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="VUV">
        <xs:annotation>
          <xs:documentation>Vatu</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.VUV</lx:path>
            <lx:label>Vatu</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="WST">
        <xs:annotation>
          <xs:documentation>Tala</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.WST</lx:path>
            <lx:label>Tala</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="XAF">
        <xs:annotation>
          <xs:documentation>CFA Franc BEAC</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.XAF</lx:path>
            <lx:label>CFA Franc BEAC</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="XCD">
        <xs:annotation>
          <xs:documentation>East Caribbean Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.XCD</lx:path>
            <lx:label>East Caribbean Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="XDR">
        <xs:annotation>
          <xs:documentation>SDR (Special Drawing Right)</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.XDR</lx:path>
            <lx:label>SDR (Special Drawing Right)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="XOF">
        <xs:annotation>
          <xs:documentation>CFA Franc BCEAO</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.XOF</lx:path>
            <lx:label>CFA Franc BCEAO</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="XPF">
        <xs:annotation>
          <xs:documentation>CFP Franc</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.XPF</lx:path>
            <lx:label>CFP Franc</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="XSU">
        <xs:annotation>
          <xs:documentation>Sucre</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.XSU</lx:path>
            <lx:label>Sucre</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="XUA">
        <xs:annotation>
          <xs:documentation>ADB Unit of Account</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.XUA</lx:path>
            <lx:label>ADB Unit of Account</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="YER">
        <xs:annotation>
          <xs:documentation>Yemeni Rial</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.YER</lx:path>
            <lx:label>Yemeni Rial</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ZAR">
        <xs:annotation>
          <xs:documentation>Rand</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.ZAR</lx:path>
            <lx:label>Rand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ZMW">
        <xs:annotation>
          <xs:documentation>Zambian Kwacha</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.ZMW</lx:path>
            <lx:label>Zambian Kwacha</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ZWL">
        <xs:annotation>
          <xs:documentation>Zimbabwe Dollar</xs:documentation>
          <xs:appinfo>
            <lx:path>currencyCodeList.ZWL</lx:path>
            <lx:label>Zimbabwe Dollar</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="dataSourceList">
    <xs:annotation>
      <xs:documentation>dataSourceList</xs:documentation>
      <xs:appinfo>
        <lx:path>dataSourceList</lx:path>
        <lx:label>dataSourceList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="GNAF">
        <xs:annotation>
          <xs:documentation>GNAF</xs:documentation>
          <xs:appinfo>
            <lx:path>dataSourceList.GNAF</lx:path>
            <lx:label>GNAF</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RP Data">
        <xs:annotation>
          <xs:documentation>RP Data</xs:documentation>
          <xs:appinfo>
            <lx:path>dataSourceList.RP Data</lx:path>
            <lx:label>RP Data</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valocity">
        <xs:annotation>
          <xs:documentation>The source of data used in searching information about the real estate property is Valocity property search service.</xs:documentation>
          <xs:appinfo>
            <lx:path>dataSourceList.Valocity</lx:path>
            <lx:label>Valocity</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="dayOfWeekList">
    <xs:annotation>
      <xs:documentation>A list of the days of each week.</xs:documentation>
      <xs:appinfo>
        <lx:path>dayOfWeekList</lx:path>
        <lx:label>dayOfWeekList</lx:label>
        <li:transactions>CAL,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Friday">
        <xs:annotation>
          <xs:documentation>Friday</xs:documentation>
          <xs:appinfo>
            <lx:path>dayOfWeekList.Friday</lx:path>
            <lx:label>Friday</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Monday">
        <xs:annotation>
          <xs:documentation>Monday</xs:documentation>
          <xs:appinfo>
            <lx:path>dayOfWeekList.Monday</lx:path>
            <lx:label>Monday</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Saturday">
        <xs:annotation>
          <xs:documentation>Saturday</xs:documentation>
          <xs:appinfo>
            <lx:path>dayOfWeekList.Saturday</lx:path>
            <lx:label>Saturday</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sunday">
        <xs:annotation>
          <xs:documentation>Sunday</xs:documentation>
          <xs:appinfo>
            <lx:path>dayOfWeekList.Sunday</lx:path>
            <lx:label>Sunday</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Thursday">
        <xs:annotation>
          <xs:documentation>Thursday</xs:documentation>
          <xs:appinfo>
            <lx:path>dayOfWeekList.Thursday</lx:path>
            <lx:label>Thursday</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tuesday">
        <xs:annotation>
          <xs:documentation>Tuesday</xs:documentation>
          <xs:appinfo>
            <lx:path>dayOfWeekList.Tuesday</lx:path>
            <lx:label>Tuesday</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wednesday">
        <xs:annotation>
          <xs:documentation>Wednesday</xs:documentation>
          <xs:appinfo>
            <lx:path>dayOfWeekList.Wednesday</lx:path>
            <lx:label>Wednesday</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="declinedReasonApplicationInstructionsList">
    <xs:annotation>
      <xs:documentation>declinedReasonApplicationInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>declinedReasonApplicationInstructionsList</lx:path>
        <lx:label>declinedReasonApplicationInstructionsList</lx:label>
        <li:transactions>CAL,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Affordability">
        <xs:annotation>
          <xs:documentation>The loan has been declined because it does not confirm to the credit provider's affordability calculations</xs:documentation>
          <xs:appinfo>
            <lx:path>declinedReasonApplicationInstructionsList.Affordability</lx:path>
            <lx:label>The loan has been declined because it does not confirm to the credit provider's affordability calculations</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit History">
        <xs:annotation>
          <xs:documentation>The loan has been declined because the borrower has adverse credit history</xs:documentation>
          <xs:appinfo>
            <lx:path>declinedReasonApplicationInstructionsList.Credit History</lx:path>
            <lx:label>The loan has been declined because the borrower has adverse credit history</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Employment Security">
        <xs:annotation>
          <xs:documentation>The loan has been declined because the borrower does not have acceptable employment history</xs:documentation>
          <xs:appinfo>
            <lx:path>declinedReasonApplicationInstructionsList.Employment Security</lx:path>
            <lx:label>The loan has been declined because the borrower does not have acceptable employment history</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Level of Debt">
        <xs:annotation>
          <xs:documentation>The credit provider has determined that the borrower has an unacceptable level of debt</xs:documentation>
          <xs:appinfo>
            <lx:path>declinedReasonApplicationInstructionsList.Level of Debt</lx:path>
            <lx:label>The credit provider has determined that the borrower has an unacceptable level of debt</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security">
        <xs:annotation>
          <xs:documentation>The loan has been declined because the security property has come in too low or has some adverse comments against it</xs:documentation>
          <xs:appinfo>
            <lx:path>declinedReasonApplicationInstructionsList.Security</lx:path>
            <lx:label>The loan has been declined because the security property has come in too low or has some adverse comments against it</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="depositAccountTypeList">
    <xs:annotation>
      <xs:documentation>depositAccountTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>depositAccountTypeList</lx:path>
        <lx:label>depositAccountTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Standard">
        <xs:annotation>
          <xs:documentation>Standard deposit accounts that may or may not pay interest and funds can be withdrawn on demand.</xs:documentation>
          <xs:appinfo>
            <lx:path>depositAccountTypeList.Standard</lx:path>
            <lx:label>Standard</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Term Deposit">
        <xs:annotation>
          <xs:documentation>An investment of cash placed with a financial institution for a fixed term with a fixed interest rate.</xs:documentation>
          <xs:appinfo>
            <lx:path>depositAccountTypeList.Term Deposit</lx:path>
            <lx:label>Term Deposit</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Transactional">
        <xs:annotation>
          <xs:documentation>Funds can be immediately withdrawn, transferred or used by the depositor without restrictions or limits.</xs:documentation>
          <xs:appinfo>
            <lx:path>depositAccountTypeList.Transactional</lx:path>
            <lx:label>Transactional</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="discountReasonList">
    <xs:annotation>
      <xs:documentation>discountReasonList</xs:documentation>
      <xs:appinfo>
        <lx:path>discountReasonList</lx:path>
        <lx:label>discountReasonList</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Discount Extension">
        <xs:annotation>
          <xs:documentation>Discount Extension</xs:documentation>
          <xs:appinfo>
            <lx:path>discountReasonList.Discount Extension</lx:path>
            <lx:label>Discount Extension</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>discountReasonList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other Financial Institution Discount">
        <xs:annotation>
          <xs:documentation>Other Financial Institution Discount</xs:documentation>
          <xs:appinfo>
            <lx:path>discountReasonList.Other Financial Institution Discount</lx:path>
            <lx:label>Other Financial Institution Discount</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Retention Pricing">
        <xs:annotation>
          <xs:documentation>Retention Pricing</xs:documentation>
          <xs:appinfo>
            <lx:path>discountReasonList.Retention Pricing</lx:path>
            <lx:label>Retention Pricing</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="documentationDeliveryMethodList">
    <xs:annotation>
      <xs:documentation>documentationDeliveryMethodList</xs:documentation>
      <xs:appinfo>
        <lx:path>documentationDeliveryMethodList</lx:path>
        <lx:label>documentationDeliveryMethodList</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Collect From Nearest Branch">
        <xs:annotation>
          <xs:documentation>Collect From Nearest Branch</xs:documentation>
          <xs:appinfo>
            <lx:path>documentationDeliveryMethodList.Collect From Nearest Branch</lx:path>
            <lx:label>Collect From Nearest Branch</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Email">
        <xs:annotation>
          <xs:documentation>Email</xs:documentation>
          <xs:appinfo>
            <lx:path>documentationDeliveryMethodList.Email</lx:path>
            <lx:label>Email</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Post">
        <xs:annotation>
          <xs:documentation>Post</xs:documentation>
          <xs:appinfo>
            <lx:path>documentationDeliveryMethodList.Post</lx:path>
            <lx:label>Post</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="documentationInstructionsMethodList">
    <xs:annotation>
      <xs:documentation>documentationInstructionsMethodList</xs:documentation>
      <xs:appinfo>
        <lx:path>documentationInstructionsMethodList</lx:path>
        <lx:label>documentationInstructionsMethodList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Email">
        <xs:annotation>
          <xs:documentation>Email</xs:documentation>
          <xs:appinfo>
            <lx:path>documentationInstructionsMethodList.Email</lx:path>
            <lx:label>Email</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Post">
        <xs:annotation>
          <xs:documentation>Post</xs:documentation>
          <xs:appinfo>
            <lx:path>documentationInstructionsMethodList.Post</lx:path>
            <lx:label>Post</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="documentCategoryList">
    <xs:annotation>
      <xs:documentation>documentCategoryList</xs:documentation>
      <xs:appinfo>
        <lx:path>documentCategoryList</lx:path>
        <lx:label>documentCategoryList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Non Photographic">
        <xs:annotation>
          <xs:documentation>Non Photographic</xs:documentation>
          <xs:appinfo>
            <lx:path>documentCategoryList.Non Photographic</lx:path>
            <lx:label>Non Photographic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Photographic">
        <xs:annotation>
          <xs:documentation>Photographic</xs:documentation>
          <xs:appinfo>
            <lx:path>documentCategoryList.Photographic</lx:path>
            <lx:label>Photographic</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Secondary">
        <xs:annotation>
          <xs:documentation>Secondary</xs:documentation>
          <xs:appinfo>
            <lx:path>documentCategoryList.Secondary</lx:path>
            <lx:label>Secondary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="documentTypeList">
    <xs:annotation>
      <xs:documentation>A list of valid document types that can be used to support the verification of a person's identity.</xs:documentation>
      <xs:appinfo>
        <lx:path>documentTypeList</lx:path>
        <lx:label>documentTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="18+ Card">
        <xs:annotation>
          <xs:documentation>18+ Card</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.18+ Card</lx:path>
            <lx:label>18+ Card</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian Birth Certificate">
        <xs:annotation>
          <xs:documentation>Australian Birth Certificate</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Australian Birth Certificate</lx:path>
            <lx:label>Australian Birth Certificate</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian Citizenship Certificate">
        <xs:annotation>
          <xs:documentation>Australian Citizenship Certificate</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Australian Citizenship Certificate</lx:path>
            <lx:label>Australian Citizenship Certificate</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian Drivers Licence">
        <xs:annotation>
          <xs:documentation>Australian Drivers Licence</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Australian Drivers Licence</lx:path>
            <lx:label>Australian Drivers Licence</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian Passport">
        <xs:annotation>
          <xs:documentation>Australian Passport</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Australian Passport</lx:path>
            <lx:label>Australian Passport</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian Photo Firearms Licence">
        <xs:annotation>
          <xs:documentation>Australian Photo Firearms Licence</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Australian Photo Firearms Licence</lx:path>
            <lx:label>Australian Photo Firearms Licence</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian Taxation Office Correspondence (with TFN)">
        <xs:annotation>
          <xs:documentation>Australian Taxation Office Correspondence (with TFN)</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Australian Taxation Office Correspondence (with TFN)</lx:path>
            <lx:label>Australian Taxation Office Correspondence (with TFN)</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bank Statement">
        <xs:annotation>
          <xs:documentation>Bank Statement</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Bank Statement</lx:path>
            <lx:label>Bank Statement</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Centrelink or Social Security Card">
        <xs:annotation>
          <xs:documentation>Centrelink or Social Security Card</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Centrelink or Social Security Card</lx:path>
            <lx:label>Centrelink or Social Security Card</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Card">
        <xs:annotation>
          <xs:documentation>Credit Card</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Credit Card</lx:path>
            <lx:label>Credit Card</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Debit Card">
        <xs:annotation>
          <xs:documentation>Debit Card</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Debit Card</lx:path>
            <lx:label>Debit Card</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Deed Poll">
        <xs:annotation>
          <xs:documentation>Deed Poll</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Deed Poll</lx:path>
            <lx:label>Deed Poll</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Department of Veterans Affairs Card">
        <xs:annotation>
          <xs:documentation>Department of Veterans Affairs Card</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Department of Veterans Affairs Card</lx:path>
            <lx:label>Department of Veterans Affairs Card</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Descent Certificate">
        <xs:annotation>
          <xs:documentation>A Certificate of Registration by Descent is issued by the Government of Australia or New Zealand to a person who was born outside the country and at least one of their parents was an Australian or New Zealand citizen at the time of their birth.</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Descent Certificate</lx:path>
            <lx:label>Descent Certificate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Electoral Roll Acknowledgement">
        <xs:annotation>
          <xs:documentation>Electoral Roll acknowledgement is usually provided in the form of an email or a letter from a government body showing confirmation of registration on the electoral roll.</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Electoral Roll Acknowledgement</lx:path>
            <lx:label>Electoral Roll Acknowledgement</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Electricity Bill">
        <xs:annotation>
          <xs:documentation>Electricity Bill</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Electricity Bill</lx:path>
            <lx:label>Electricity Bill</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Emergency Travel Document">
        <xs:annotation>
          <xs:documentation>Emergency Travel Document</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Emergency Travel Document</lx:path>
            <lx:label>Emergency Travel Document</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Foreign Birth Certificate">
        <xs:annotation>
          <xs:documentation>Foreign Birth Certificate</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Foreign Birth Certificate</lx:path>
            <lx:label>Foreign Birth Certificate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Foreign Citizenship Certificate">
        <xs:annotation>
          <xs:documentation>Foreign Citizenship Certificate</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Foreign Citizenship Certificate</lx:path>
            <lx:label>Foreign Citizenship Certificate</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gas Bill">
        <xs:annotation>
          <xs:documentation>Gas Bill</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Gas Bill</lx:path>
            <lx:label>Gas Bill</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits Notice">
        <xs:annotation>
          <xs:documentation>Government Benefits Notice</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Government Benefits Notice</lx:path>
            <lx:label>Government Benefits Notice</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Health Care Card">
        <xs:annotation>
          <xs:documentation>Health Care Card</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Health Care Card</lx:path>
            <lx:label>Health Care Card</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="International Driver License">
        <xs:annotation>
          <xs:documentation>International Driver License</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.International Driver License</lx:path>
            <lx:label>International Driver License</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="International Passport">
        <xs:annotation>
          <xs:documentation>International Passport</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.International Passport</lx:path>
            <lx:label>International Passport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Internet Service Provider Bill">
        <xs:annotation>
          <xs:documentation>Internet Service Provider Bill</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Internet Service Provider Bill</lx:path>
            <lx:label>Internet Service Provider Bill</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Landline Phone Bill">
        <xs:annotation>
          <xs:documentation>Landline Phone Bill</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Landline Phone Bill</lx:path>
            <lx:label>Landline Phone Bill</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Marriage Certificate">
        <xs:annotation>
          <xs:documentation>Marriage Certificate</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Marriage Certificate</lx:path>
            <lx:label>Marriage Certificate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Medicare Card">
        <xs:annotation>
          <xs:documentation>Medicare Card</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Medicare Card</lx:path>
            <lx:label>Medicare Card</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Motor Vehicle Registration or Insurance Papers">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Registration or Insurance Papers</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Motor Vehicle Registration or Insurance Papers</lx:path>
            <lx:label>Motor Vehicle Registration or Insurance Papers</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="National Identity Card">
        <xs:annotation>
          <xs:documentation>National Identity Card</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.National Identity Card</lx:path>
            <lx:label>National Identity Card</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="New Zealand Birth Certificate">
        <xs:annotation>
          <xs:documentation>New Zealand Birth Certificate</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.New Zealand Birth Certificate</lx:path>
            <lx:label>New Zealand Birth Certificate</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="New Zealand Certificate of Identity">
        <xs:annotation>
          <xs:documentation>New Zealand Certificate of Identity</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.New Zealand Certificate of Identity</lx:path>
            <lx:label>New Zealand Certificate of Identity</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="New Zealand Citizenship Certificate">
        <xs:annotation>
          <xs:documentation>New Zealand Citizenship Certificate</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.New Zealand Citizenship Certificate</lx:path>
            <lx:label>New Zealand Citizenship Certificate</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="New Zealand Driver Licence">
        <xs:annotation>
          <xs:documentation>New Zealand Driver Licence</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.New Zealand Driver Licence</lx:path>
            <lx:label>New Zealand Driver Licence</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="New Zealand Firearms Licence">
        <xs:annotation>
          <xs:documentation>New Zealand Firearms Licence</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.New Zealand Firearms Licence</lx:path>
            <lx:label>New Zealand Firearms Licence</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="New Zealand Passport">
        <xs:annotation>
          <xs:documentation>New Zealand Passport</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.New Zealand Passport</lx:path>
            <lx:label>New Zealand Passport</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="New Zealand Refugee Travel Document">
        <xs:annotation>
          <xs:documentation>New Zealand Refugee Travel Document</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.New Zealand Refugee Travel Document</lx:path>
            <lx:label>New Zealand Refugee Travel Document</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pension Card">
        <xs:annotation>
          <xs:documentation>Pension Card</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Pension Card</lx:path>
            <lx:label>Pension Card</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Professional or Trade Association Card">
        <xs:annotation>
          <xs:documentation>Professional or Trade Association Card</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Professional or Trade Association Card</lx:path>
            <lx:label>Professional or Trade Association Card</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Proof of Age Card">
        <xs:annotation>
          <xs:documentation>Proof of Age Card</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Proof of Age Card</lx:path>
            <lx:label>Proof of Age Card</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Property Insurance Papers">
        <xs:annotation>
          <xs:documentation>Property Insurance Papers</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Property Insurance Papers</lx:path>
            <lx:label>Property Insurance Papers</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Property Lease or Rental Agreement">
        <xs:annotation>
          <xs:documentation>Property Lease or Rental Agreement</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Property Lease or Rental Agreement</lx:path>
            <lx:label>Property Lease or Rental Agreement</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rates Notice">
        <xs:annotation>
          <xs:documentation>Rates Notice</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Rates Notice</lx:path>
            <lx:label>Rates Notice</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="State or Federal Government Employee Photo Identity Card">
        <xs:annotation>
          <xs:documentation>State or Federal Government Employee Photo Identity Card</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.State or Federal Government Employee Photo Identity Card</lx:path>
            <lx:label>State or Federal Government Employee Photo Identity Card</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tax Assessment">
        <xs:annotation>
          <xs:documentation>Tax Assessment</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Tax Assessment</lx:path>
            <lx:label>Tax Assessment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tertiary Education Institution Photo Identity Card">
        <xs:annotation>
          <xs:documentation>Tertiary Education Institution Photo Identity Card</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Tertiary Education Institution Photo Identity Card</lx:path>
            <lx:label>Tertiary Education Institution Photo Identity Card</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Verification of Identity Document">
        <xs:annotation>
          <xs:documentation>A Verification of Identity Document can take the form of an electronic confirmation or documented proof of each identity check.</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Verification of Identity Document</lx:path>
            <lx:label>Verification of Identity Document</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Water Utility Bill">
        <xs:annotation>
          <xs:documentation>Water Utility Bill</xs:documentation>
          <xs:appinfo>
            <lx:path>documentTypeList.Water Utility Bill</lx:path>
            <lx:label>Water Utility Bill</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="durationUnitsList">
    <xs:annotation>
      <xs:documentation>durationUnitsList</xs:documentation>
      <xs:appinfo>
        <lx:path>durationUnitsList</lx:path>
        <lx:label>durationUnitsList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Days">
        <xs:annotation>
          <xs:documentation>Days</xs:documentation>
          <xs:appinfo>
            <lx:path>durationUnitsList.Days</lx:path>
            <lx:label>Days</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Months">
        <xs:annotation>
          <xs:documentation>Months</xs:documentation>
          <xs:appinfo>
            <lx:path>durationUnitsList.Months</lx:path>
            <lx:label>Months</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Weeks">
        <xs:annotation>
          <xs:documentation>Weeks</xs:documentation>
          <xs:appinfo>
            <lx:path>durationUnitsList.Weeks</lx:path>
            <lx:label>Weeks</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Years">
        <xs:annotation>
          <xs:documentation>Years</xs:documentation>
          <xs:appinfo>
            <lx:path>durationUnitsList.Years</lx:path>
            <lx:label>Years</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="earthMovingMiningAndConstructionSubtypeList">
    <xs:annotation>
      <xs:documentation>earthMovingMiningAndConstructionSubtypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>earthMovingMiningAndConstructionSubtypeList</lx:path>
        <lx:label>earthMovingMiningAndConstructionSubtypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Crusher">
        <xs:annotation>
          <xs:documentation>Crusher (subtype of Crushing / Screening).</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionSubtypeList.Crusher</lx:path>
            <lx:label>Crusher</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Directional Drill">
        <xs:annotation>
          <xs:documentation>Directional Drill (subtype of Drill Rig).</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionSubtypeList.Directional Drill</lx:path>
            <lx:label>Directional Drill</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Drill and Blast Rig">
        <xs:annotation>
          <xs:documentation>Drill and Blast Rig (subtype of Drill Rig).</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionSubtypeList.Drill and Blast Rig</lx:path>
            <lx:label>Drill and Blast Rig</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Screen">
        <xs:annotation>
          <xs:documentation>Screen (subtype of Crushing / Screening).</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionSubtypeList.Screen</lx:path>
            <lx:label>Screen</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Skid-Steer Loader">
        <xs:annotation>
          <xs:documentation>Skid-Steer Loader (subtype of Loader).</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionSubtypeList.Skid-Steer Loader</lx:path>
            <lx:label>Skid-Steer Loader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tracked Loader">
        <xs:annotation>
          <xs:documentation>Tracked Loader (subtype of Loader).</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionSubtypeList.Tracked Loader</lx:path>
            <lx:label>Tracked Loader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wheeled Loader">
        <xs:annotation>
          <xs:documentation>Wheeled Loader (subtype of Loader).</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionSubtypeList.Wheeled Loader</lx:path>
            <lx:label>Wheeled Loader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="earthMovingMiningAndConstructionTypeList">
    <xs:annotation>
      <xs:documentation>earthMovingMiningAndConstructionTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>earthMovingMiningAndConstructionTypeList</lx:path>
        <lx:label>earthMovingMiningAndConstructionTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Concrete Equipment">
        <xs:annotation>
          <xs:documentation>Concrete Equipment</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Concrete Equipment</lx:path>
            <lx:label>Concrete Equipment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Conveyor">
        <xs:annotation>
          <xs:documentation>Conveyor</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Conveyor</lx:path>
            <lx:label>Conveyor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Crushing / Screening">
        <xs:annotation>
          <xs:documentation>Crushing / Screening</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Crushing / Screening</lx:path>
            <lx:label>Crushing / Screening</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dozer">
        <xs:annotation>
          <xs:documentation>Dozer</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Dozer</lx:path>
            <lx:label>Dozer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Drill Rig">
        <xs:annotation>
          <xs:documentation>Drill Rig</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Drill Rig</lx:path>
            <lx:label>Drill Rig</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Excavator / Face Shovel">
        <xs:annotation>
          <xs:documentation>Excavator / Face Shovel</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Excavator / Face Shovel</lx:path>
            <lx:label>Excavator / Face Shovel</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Forestry">
        <xs:annotation>
          <xs:documentation>Forestry</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Forestry</lx:path>
            <lx:label>Forestry</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Grader">
        <xs:annotation>
          <xs:documentation>Grader</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Grader</lx:path>
            <lx:label>Grader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loader">
        <xs:annotation>
          <xs:documentation>Loader</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Loader</lx:path>
            <lx:label>Loader</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Off Highway Truck">
        <xs:annotation>
          <xs:documentation>Off Highway Truck</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Off Highway Truck</lx:path>
            <lx:label>Off Highway Truck</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Road Maintenance">
        <xs:annotation>
          <xs:documentation>Road Maintenance</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Road Maintenance</lx:path>
            <lx:label>Road Maintenance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Roller / Compacting">
        <xs:annotation>
          <xs:documentation>Roller / Compacting</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Roller / Compacting</lx:path>
            <lx:label>Roller / Compacting</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Scraper">
        <xs:annotation>
          <xs:documentation>Scraper</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Scraper</lx:path>
            <lx:label>Scraper</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sweeping / Cleaning">
        <xs:annotation>
          <xs:documentation>Sweeping / Cleaning</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Sweeping / Cleaning</lx:path>
            <lx:label>Sweeping / Cleaning</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trenching">
        <xs:annotation>
          <xs:documentation>Trenching</xs:documentation>
          <xs:appinfo>
            <lx:path>earthMovingMiningAndConstructionTypeList.Trenching</lx:path>
            <lx:label>Trenching</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="employerTypeList">
    <xs:annotation>
      <xs:documentation>employerTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>employerTypeList</lx:path>
        <lx:label>employerTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Private">
        <xs:annotation>
          <xs:documentation>The employer is a private enterprise.</xs:documentation>
          <xs:appinfo>
            <lx:path>employerTypeList.Private</lx:path>
            <lx:label>The employer is a private enterprise.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Public">
        <xs:annotation>
          <xs:documentation>The employer is a government department, agency or enterprise.</xs:documentation>
          <xs:appinfo>
            <lx:path>employerTypeList.Public</lx:path>
            <lx:label>The employer is a government department, agency or enterprise.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="employmentStatusList">
    <xs:annotation>
      <xs:documentation>employmentStatusList</xs:documentation>
      <xs:appinfo>
        <lx:path>employmentStatusList</lx:path>
        <lx:label>employmentStatusList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Previous">
        <xs:annotation>
          <xs:documentation>Previous</xs:documentation>
          <xs:appinfo>
            <lx:path>employmentStatusList.Previous</lx:path>
            <lx:label>Previous</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Primary">
        <xs:annotation>
          <xs:documentation>Primary</xs:documentation>
          <xs:appinfo>
            <lx:path>employmentStatusList.Primary</lx:path>
            <lx:label>Primary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Secondary">
        <xs:annotation>
          <xs:documentation>Secondary</xs:documentation>
          <xs:appinfo>
            <lx:path>employmentStatusList.Secondary</lx:path>
            <lx:label>Secondary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="encumbranceInFavourOfCapacityList">
    <xs:annotation>
      <xs:documentation>A list of capacities of the party to whom the encumbrance is in favour.</xs:documentation>
      <xs:appinfo>
        <lx:path>encumbranceInFavourOfCapacityList</lx:path>
        <lx:label>encumbranceInFavourOfCapacityList</lx:label>
        <li:transactions>CAL,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Asset Held in Own Right and on Behalf of Trust">
        <xs:annotation>
          <xs:documentation>Asset Held in Own Right and on Behalf of Trust</xs:documentation>
          <xs:appinfo>
            <lx:path>encumbranceInFavourOfCapacityList.Asset Held in Own Right and on Behalf of Trust</lx:path>
            <lx:label>Asset Held in Own Right and on Behalf of Trust</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Limited to Particular Property or Asset">
        <xs:annotation>
          <xs:documentation>Limited to Particular Property or Asset</xs:documentation>
          <xs:appinfo>
            <lx:path>encumbranceInFavourOfCapacityList.Limited to Particular Property or Asset</lx:path>
            <lx:label>Limited to Particular Property or Asset</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Own Right Only">
        <xs:annotation>
          <xs:documentation>Own Right Only</xs:documentation>
          <xs:appinfo>
            <lx:path>encumbranceInFavourOfCapacityList.Own Right Only</lx:path>
            <lx:label>Own Right Only</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Partnership">
        <xs:annotation>
          <xs:documentation>Partnership</xs:documentation>
          <xs:appinfo>
            <lx:path>encumbranceInFavourOfCapacityList.Partnership</lx:path>
            <lx:label>Partnership</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trust Assets Only">
        <xs:annotation>
          <xs:documentation>Trust Assets Only</xs:documentation>
          <xs:appinfo>
            <lx:path>encumbranceInFavourOfCapacityList.Trust Assets Only</lx:path>
            <lx:label>Trust Assets Only</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="errorInstructionsTypeList">
    <xs:annotation>
      <xs:documentation>errorInstructionsTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>errorInstructionsTypeList</lx:path>
        <lx:label>errorInstructionsTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Data Error">
        <xs:annotation>
          <xs:documentation>Data Error</xs:documentation>
          <xs:appinfo>
            <lx:path>errorInstructionsTypeList.Data Error</lx:path>
            <lx:label>Data Error</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Information">
        <xs:annotation>
          <xs:documentation>Information</xs:documentation>
          <xs:appinfo>
            <lx:path>errorInstructionsTypeList.Information</lx:path>
            <lx:label>Information</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="System Error">
        <xs:annotation>
          <xs:documentation>System Error</xs:documentation>
          <xs:appinfo>
            <lx:path>errorInstructionsTypeList.System Error</lx:path>
            <lx:label>System Error</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="errorSourceList">
    <xs:annotation>
      <xs:documentation>errorSourceList</xs:documentation>
      <xs:appinfo>
        <lx:path>errorSourceList</lx:path>
        <lx:label>errorSourceList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="DTD">
        <xs:annotation>
          <xs:documentation>DTD</xs:documentation>
          <xs:appinfo>
            <lx:path>errorSourceList.DTD</lx:path>
            <lx:label>DTD</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Parser">
        <xs:annotation>
          <xs:documentation>Parser</xs:documentation>
          <xs:appinfo>
            <lx:path>errorSourceList.Parser</lx:path>
            <lx:label>Parser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Proprietary">
        <xs:annotation>
          <xs:documentation>Proprietary</xs:documentation>
          <xs:appinfo>
            <lx:path>errorSourceList.Proprietary</lx:path>
            <lx:label>Proprietary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Schematron">
        <xs:annotation>
          <xs:documentation>Schematron</xs:documentation>
          <xs:appinfo>
            <lx:path>errorSourceList.Schematron</lx:path>
            <lx:label>Schematron</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="XSD">
        <xs:annotation>
          <xs:documentation>XSD</xs:documentation>
          <xs:appinfo>
            <lx:path>errorSourceList.XSD</lx:path>
            <lx:label>XSD</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="XSLT">
        <xs:annotation>
          <xs:documentation>XSLT</xs:documentation>
          <xs:appinfo>
            <lx:path>errorSourceList.XSLT</lx:path>
            <lx:label>XSLT</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="estimateBasisNonRealEstateAssetList">
    <xs:annotation>
      <xs:documentation>estimateBasisNonRealEstateAssetList</xs:documentation>
      <xs:appinfo>
        <lx:path>estimateBasisNonRealEstateAssetList</lx:path>
        <lx:label>estimateBasisNonRealEstateAssetList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Applicant Estimate">
        <xs:annotation>
          <xs:documentation>The value specified as the estimated value is the applicant's estimate of the value.</xs:documentation>
          <xs:appinfo>
            <lx:path>estimateBasisNonRealEstateAssetList.Applicant Estimate</lx:path>
            <lx:label>The value specified as the estimated value is the applicant's estimate of the value.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Certified Valuation">
        <xs:annotation>
          <xs:documentation>The value specified as the estimated value is a certified valuation provided by a valuer.</xs:documentation>
          <xs:appinfo>
            <lx:path>estimateBasisNonRealEstateAssetList.Certified Valuation</lx:path>
            <lx:label>The value specified as the estimated value is a certified valuation provided by a valuer.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="estimateBasisRealEstateAssetList">
    <xs:annotation>
      <xs:documentation>estimateBasisRealEstateAssetList</xs:documentation>
      <xs:appinfo>
        <lx:path>estimateBasisRealEstateAssetList</lx:path>
        <lx:label>estimateBasisRealEstateAssetList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Applicant Estimate">
        <xs:annotation>
          <xs:documentation>The value specified as the estimated value is the applicant's estimate of the value.</xs:documentation>
          <xs:appinfo>
            <lx:path>estimateBasisRealEstateAssetList.Applicant Estimate</lx:path>
            <lx:label>The value specified as the estimated value is the applicant's estimate of the value.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bank Assessment">
        <xs:annotation>
          <xs:documentation>The value specified as the estimated value is the bank's assessment of the value.</xs:documentation>
          <xs:appinfo>
            <lx:path>estimateBasisRealEstateAssetList.Bank Assessment</lx:path>
            <lx:label>Bank Assessment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Certified Valuation">
        <xs:annotation>
          <xs:documentation>The value specified as the estimated value is a certified valuation provided by a valuer.</xs:documentation>
          <xs:appinfo>
            <lx:path>estimateBasisRealEstateAssetList.Certified Valuation</lx:path>
            <lx:label>The value specified as the estimated value is a certified valuation provided by a valuer.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Contract of Sale">
        <xs:annotation>
          <xs:documentation>The value specified as the estimated value is provided by Contract of Sale.</xs:documentation>
          <xs:appinfo>
            <lx:path>estimateBasisRealEstateAssetList.Contract of Sale</lx:path>
            <lx:label>The value specified as the estimated value is provided by Contract of Sale.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Electronic Valuation">
        <xs:annotation>
          <xs:documentation>The value specified as the estimated value was calculated electronically.</xs:documentation>
          <xs:appinfo>
            <lx:path>estimateBasisRealEstateAssetList.Electronic Valuation</lx:path>
            <lx:label>The value specified as the estimated value was calculated electronically.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rates Notice">
        <xs:annotation>
          <xs:documentation>The value specified as the estimated value was taken from the rates notice.</xs:documentation>
          <xs:appinfo>
            <lx:path>estimateBasisRealEstateAssetList.Rates Notice</lx:path>
            <lx:label>The value specified as the estimated value was taken from the rates notice.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RP Data">
        <xs:annotation>
          <xs:documentation>The value specified as the estimated value is provided by RP Data property search service.</xs:documentation>
          <xs:appinfo>
            <lx:path>estimateBasisRealEstateAssetList.RP Data</lx:path>
            <lx:label>The value specified as the estimated value is provided by RP Data property search service.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valocity">
        <xs:annotation>
          <xs:documentation>The value specified as the estimated value is provided by Valocity property search service.</xs:documentation>
          <xs:appinfo>
            <lx:path>estimateBasisRealEstateAssetList.Valocity</lx:path>
            <lx:label>Estimated value provided by Valocity.</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="estimateMethodologyList">
    <xs:annotation>
      <xs:documentation>The methodology used for estimating the value.</xs:documentation>
      <xs:appinfo>
        <lx:path>estimateMethodologyList</lx:path>
        <lx:label>estimateMethodologyList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Completed Construction Value">
        <xs:annotation>
          <xs:documentation>The value specified as the estimated value is based on the expected value of the finalised construction. Sometimes called 'As If Complete'.</xs:documentation>
          <xs:appinfo>
            <lx:path>estimateMethodologyList.Completed Construction Value</lx:path>
            <lx:label>Completed Construction Value</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Current Market Value">
        <xs:annotation>
          <xs:documentation>The value specified as the estimated value is based on the current market value. Sometimes called 'Existing Property'.</xs:documentation>
          <xs:appinfo>
            <lx:path>estimateMethodologyList.Current Market Value</lx:path>
            <lx:label>Current Market Value</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="eventNameAccountCreationInstructionsList">
    <xs:annotation>
      <xs:documentation>eventNameAccountCreationInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>eventNameAccountCreationInstructionsList</lx:path>
        <lx:label>eventNameAccountCreationInstructionsList</lx:label>
        <li:transactions>ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Cancelled">
        <xs:annotation>
          <xs:documentation>The loan account creation application has been cancelled.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameAccountCreationInstructionsList.Cancelled</lx:path>
            <lx:label>Cancelled</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Discharged">
        <xs:annotation>
          <xs:documentation>The loan account creation aggreement has been discharged.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameAccountCreationInstructionsList.Discharged</lx:path>
            <lx:label>Discharged</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settled">
        <xs:annotation>
          <xs:documentation>The loan account creation application has been settled.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameAccountCreationInstructionsList.Settled</lx:path>
            <lx:label>Settled</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Withdrawn">
        <xs:annotation>
          <xs:documentation>The loan account creation application has been withdrawn.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameAccountCreationInstructionsList.Withdrawn</lx:path>
            <lx:label>Withdrawn</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="eventNameApplicationInstructionsList">
    <xs:annotation>
      <xs:documentation>eventNameApplicationInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>eventNameApplicationInstructionsList</lx:path>
        <lx:label>eventNameApplicationInstructionsList</lx:label>
        <li:transactions>CAL,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Additional Information Has Been Requested">
        <xs:annotation>
          <xs:documentation>Application has already been registered but additional explanations and/or comments are required from the applicants or loan writer in order to complete the assessment.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Additional Information Has Been Requested</lx:path>
            <lx:label>Additional Information Has Been Requested</lx:label>
            <li:transactions>CAL,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Application Received">
        <xs:annotation>
          <xs:documentation>Application successfully received by Lender system and can be further processed i.e. well formed.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Application Received</lx:path>
            <lx:label>Application successfully received by Lender system and can be further processed i.e. well formed.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Application Registered">
        <xs:annotation>
          <xs:documentation>Lender system ID allocated (Pre Settlement Application Identifier) and, if electronically lodged, application is ready for credit assessment ie. validates against Lender's schema and other business rules.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Application Registered</lx:path>
            <lx:label>Lender system ID allocated (Pre Settlement Application Identifier) and, if electronically lodged, application is ready for credit assessment ie. validates against Lender's schema and other business rules.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Application Sent">
        <xs:annotation>
          <xs:documentation>Application successfully transmitted by Broker system.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Application Sent</lx:path>
            <lx:label>Application successfully transmitted by Broker system.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Awaiting Supporting Documents">
        <xs:annotation>
          <xs:documentation>The customer is now required to to provide supporting documents.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Awaiting Supporting Documents</lx:path>
            <lx:label>Awaiting Supporting Documents</lx:label>
            <li:transactions>CAL,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cancelled">
        <xs:annotation>
          <xs:documentation>Application has been stopped at Lender's initiative. Cannot be resurrected.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Cancelled</lx:path>
            <lx:label>Application has been stopped at Lender's initiative. Cannot be resurrected.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Comment">
        <xs:annotation>
          <xs:documentation>The Details attribute contains free text comments.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Comment</lx:path>
            <lx:label>The Details attribute contains free text comments.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Commission Paid">
        <xs:annotation>
          <xs:documentation>The Commission Paid to loan originator event has been triggered.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Commission Paid</lx:path>
            <lx:label>The Commission Paid to loan originator event has been triggered.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Conditionally Approved">
        <xs:annotation>
          <xs:documentation>Application has been approved subject to conditions.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Conditionally Approved</lx:path>
            <lx:label>Application has been approved subject to conditions.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Check Ordered">
        <xs:annotation>
          <xs:documentation>A credit check with a credit bureau has been ordered.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Credit Check Ordered</lx:path>
            <lx:label>A credit check with a credit bureau has been ordered.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Check Received">
        <xs:annotation>
          <xs:documentation>A credit check has been received from a credit bureau.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Credit Check Received</lx:path>
            <lx:label>A credit check has been received from a credit bureau.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Declined">
        <xs:annotation>
          <xs:documentation>Application has been Declined.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Declined</lx:path>
            <lx:label>Application has been Declined.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Discharged">
        <xs:annotation>
          <xs:documentation>The borrowers current loan has been paid off in full.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Discharged</lx:path>
            <lx:label>The borrowers current loan has been paid off in full.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Duplicate Submission">
        <xs:annotation>
          <xs:documentation>The submission received already exists in the recipients system.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Duplicate Submission</lx:path>
            <lx:label>The submission received already exists in the recipients system.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Employment Check Commenced">
        <xs:annotation>
          <xs:documentation>An employment check on the borrower has commenced.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Employment Check Commenced</lx:path>
            <lx:label>An employment check on the borrower has commenced.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Employment Check Completed">
        <xs:annotation>
          <xs:documentation>An employment check on the borrower has been completed.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Employment Check Completed</lx:path>
            <lx:label>An employment check on the borrower has been completed.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Employment Check Delayed">
        <xs:annotation>
          <xs:documentation>An employment check on the borrower has been delayed.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Employment Check Delayed</lx:path>
            <lx:label>An employment check on the borrower has been delayed.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Error on Submission">
        <xs:annotation>
          <xs:documentation>The application either has missing data or has invalid combinations of data. You will need to re-submit the application after correcting any data issues.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Error on Submission</lx:path>
            <lx:label>The application either has missing data or has invalid combinations of data. You will need to re-submit the application after correcting any data issues.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LMI Ordered">
        <xs:annotation>
          <xs:documentation>Lenders Mortgage Insurance has been ordered by the credit provider with a Mortgage Insurer.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.LMI Ordered</lx:path>
            <lx:label>Lenders Mortgage Insurance has been ordered by the credit provider with a Mortgage Insurer.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LMI Response Received">
        <xs:annotation>
          <xs:documentation>A response has been received from the Mortgage Insurer, usually containing a Premium amount.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.LMI Response Received</lx:path>
            <lx:label>A response has been received from the Mortgage Insurer, usually containing a Premium amount.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents and Requirements Complete">
        <xs:annotation>
          <xs:documentation>The Loan Documents and requirements received from the Borrower are in order.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents and Requirements Complete</lx:path>
            <lx:label>The Loan Documents and requirements received from the Borrower are in order.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents and Requirements Incomplete">
        <xs:annotation>
          <xs:documentation>The Loan Documents and/or requirements received from the Borrower are not complete.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents and Requirements Incomplete</lx:path>
            <lx:label>The Loan Documents and/or requirements received from the Borrower are not complete.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Certified for Settlement">
        <xs:annotation>
          <xs:documentation>The Loan Documents have been certified.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents Certified for Settlement</lx:path>
            <lx:label>The Loan Documents have been certified.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Error Message Returned">
        <xs:annotation>
          <xs:documentation>Error message returned to the Lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents Error Message Returned</lx:path>
            <lx:label>Error message returned to the Lender.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Follow Up">
        <xs:annotation>
          <xs:documentation>A request to follow up the Loan Document production has been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents Follow Up</lx:path>
            <lx:label>A request to follow up the Loan Document production has been received.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Issued">
        <xs:annotation>
          <xs:documentation>Loan Documents have been issued as directed by Lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents Issued</lx:path>
            <lx:label>Loan Documents have been issued as directed by Lender.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Preparation Commenced">
        <xs:annotation>
          <xs:documentation>A request to start loan documents preparation has been issued by the Lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents Preparation Commenced</lx:path>
            <lx:label>Loan Documents Preparation Commenced</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Produced">
        <xs:annotation>
          <xs:documentation>Loan documents have been produced.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents Produced</lx:path>
            <lx:label>Loan documents have been produced.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Produced and Returned to Lender">
        <xs:annotation>
          <xs:documentation>The Loan Documents have been produced and returned to the Lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents Produced and Returned to Lender</lx:path>
            <lx:label>The Loan Documents have been produced and returned to the Lender.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Re-issue Request Sent">
        <xs:annotation>
          <xs:documentation>The request to reissue the Loan Documents has been sent.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents Re-issue Request Sent</lx:path>
            <lx:label>The request to reissue the Loan Documents has been sent.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Re-Issued">
        <xs:annotation>
          <xs:documentation>The Loan Documents have been re-issued.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents Re-Issued</lx:path>
            <lx:label>The Loan Documents have been re-issued</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Received">
        <xs:annotation>
          <xs:documentation>Signed Loan Documents have been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents Received</lx:path>
            <lx:label>Signed Loan Documents have been received.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Request Received">
        <xs:annotation>
          <xs:documentation>A request to produce Loan Documents has been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents Request Received</lx:path>
            <lx:label>A request to produce Loan Documents has been received.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Request Sent">
        <xs:annotation>
          <xs:documentation>A request to produce Loan Documents has been sent.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents Request Sent</lx:path>
            <lx:label>A request to produce Loan Documents has been sent.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Returned Enquiry">
        <xs:annotation>
          <xs:documentation>An enquiry about whether signed Loan Documents have been returned has been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Documents Returned Enquiry</lx:path>
            <lx:label>An enquiry about whether signed Loan Documents have been returned has been received.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Funded by Lender">
        <xs:annotation>
          <xs:documentation>The funds have been debited to the Loan Account.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Loan Funded by Lender</lx:path>
            <lx:label>The funds have been debited to the Loan Account.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Payout Figure Received">
        <xs:annotation>
          <xs:documentation>The total amount required to pay out the loan has been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Payout Figure Received</lx:path>
            <lx:label>The total amount required to pay out the loan has been received.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Payout Figure Rejected">
        <xs:annotation>
          <xs:documentation>The total amount required to pay out the loan has been rejected.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Payout Figure Rejected</lx:path>
            <lx:label>The total amount required to pay out the loan has been rejected.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PreApproved">
        <xs:annotation>
          <xs:documentation>Application conditionally approved subject to final identification of security property.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.PreApproved</lx:path>
            <lx:label>Application conditionally approved subject to final identification of security property.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Progress Payment Complete">
        <xs:annotation>
          <xs:documentation>A progress payment has been completed.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Progress Payment Complete</lx:path>
            <lx:label>A progress payment has been completed.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Progress Payment Message">
        <xs:annotation>
          <xs:documentation>A message about the Progress Payment process.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Progress Payment Message</lx:path>
            <lx:label>A message about the Progress Payment process.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ready for Progress Payment">
        <xs:annotation>
          <xs:documentation>The application is in a state in which it is ready to begin the Progress Payment process.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Ready for Progress Payment</lx:path>
            <lx:label>The application is in a state in which it is ready to begin the Progress Payment process.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ready for Settlement">
        <xs:annotation>
          <xs:documentation>The application is in a state in which it is ready to begin the Settlement process.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Ready for Settlement</lx:path>
            <lx:label>The application is in a state in which it is ready to begin the Settlement process.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Referred for Credit Assessment">
        <xs:annotation>
          <xs:documentation>Application has been referred for manual credit assessment.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Referred for Credit Assessment</lx:path>
            <lx:label>Application has been referred for manual credit assessment.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Referred to Broker">
        <xs:annotation>
          <xs:documentation>Application has been referred to the Broker.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Referred to Broker</lx:path>
            <lx:label>Application has been referred to the Broker.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Referred to Mortgage Insurer">
        <xs:annotation>
          <xs:documentation>Application has been referred to the Mortgage Insurer.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Referred to Mortgage Insurer</lx:path>
            <lx:label>Application has been referred to the Mortgage Insurer.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Documents Sent to Custodian">
        <xs:annotation>
          <xs:documentation>The security documents have been sent to the custodian.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Security Documents Sent to Custodian</lx:path>
            <lx:label>The security documents have been sent to the custodian.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Documents Sent to Lender">
        <xs:annotation>
          <xs:documentation>The security documents have been sent to the lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Security Documents Sent to Lender</lx:path>
            <lx:label>The security documents have been sent to the lender.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Packet Received">
        <xs:annotation>
          <xs:documentation>The security packet has been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Security Packet Received</lx:path>
            <lx:label>The security packet has been received.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Registered">
        <xs:annotation>
          <xs:documentation>The security has been registered with the Land Titles Office</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Security Registered</lx:path>
            <lx:label>The security has been registered with the Land Titles Office</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Service Coordinator Assigned">
        <xs:annotation>
          <xs:documentation>A service coordinator has been assigned to process the application.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Service Coordinator Assigned</lx:path>
            <lx:label>A service coordinator has been assigned to process the application.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Cancelled">
        <xs:annotation>
          <xs:documentation>The settlement process has been cancelled.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Settlement Cancelled</lx:path>
            <lx:label>The settlement process has been cancelled.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Completed">
        <xs:annotation>
          <xs:documentation>Application has gone to settlement; may or may not have been fully drawn down; accounts may have had a new identifier allocated to them by the Lender (known as Post Settlement Application Number or Facility identifier).</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Settlement Completed</lx:path>
            <lx:label>Application has gone to settlement; may or may not have been fully drawn down; accounts may have had a new identifier allocated to them by the Lender (known as Post Settlement Application Number or Facility identifier).</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Date Booked">
        <xs:annotation>
          <xs:documentation>The settlement date has been booked.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Settlement Date Booked</lx:path>
            <lx:label>The settlement date has been booked.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Date Confirmed">
        <xs:annotation>
          <xs:documentation>The settlement date has been confirmed by all parties.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Settlement Date Confirmed</lx:path>
            <lx:label>The settlement date has been confirmed by all parties.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Date Estimated">
        <xs:annotation>
          <xs:documentation>The estimated settlement date, which is 21 days or as long as six weeks post the exchange of contracts.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Settlement Date Estimated</lx:path>
            <lx:label>The estimated settlement date, which is 21 days or as long as six weeks post the exchange of contracts.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Delayed">
        <xs:annotation>
          <xs:documentation>The settlement process has been delayed.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Settlement Delayed</lx:path>
            <lx:label>The settlement process has been delayed.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Funds Received">
        <xs:annotation>
          <xs:documentation>The funds required at settlement have been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Settlement Funds Received</lx:path>
            <lx:label>The funds required at settlement have been received.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Rebooked">
        <xs:annotation>
          <xs:documentation>The settlement date has been rebooked.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Settlement Rebooked</lx:path>
            <lx:label>The settlement date has been rebooked.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Solicitor Instructed">
        <xs:annotation>
          <xs:documentation>The solicitors conducting the property transaction have been instructed.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Solicitor Instructed</lx:path>
            <lx:label>The solicitors conducting the property transaction have been instructed.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Solicitor Instruction Received">
        <xs:annotation>
          <xs:documentation>The instruction for the solicitors conducting the property transaction have been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Solicitor Instruction Received</lx:path>
            <lx:label>The instruction for the solicitors conducting the property transaction have been received.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Supporting Documents Determined to Be Incomplete">
        <xs:annotation>
          <xs:documentation>The supporting documents provided by the customer do not cover the checklist requirements, additional supporting documents are expected.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Supporting Documents Determined to Be Incomplete</lx:path>
            <lx:label>Supporting Documents Determined to Be Incomplete</lx:label>
            <li:transactions>CAL,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Supporting Documents Have Been Verified">
        <xs:annotation>
          <xs:documentation>The customer provided supporting documents have been verified.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Supporting Documents Have Been Verified</lx:path>
            <lx:label>Supporting Documents Have Been Verified</lx:label>
            <li:transactions>CAL,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Supporting Documents Received">
        <xs:annotation>
          <xs:documentation>The customer provided supporting documentation have been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Supporting Documents Received</lx:path>
            <lx:label>The customer provided supporting documentation have been received.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Title Search Ordered">
        <xs:annotation>
          <xs:documentation>A property title search has been ordered.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Title Search Ordered</lx:path>
            <lx:label>Title Search Ordered</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Title Search Received">
        <xs:annotation>
          <xs:documentation>The results of the property title search have been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Title Search Received</lx:path>
            <lx:label>Title Search Received</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unconditionally Approved">
        <xs:annotation>
          <xs:documentation>All conditions to conditional approval have been met and Application has received formal/final approval.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Unconditionally Approved</lx:path>
            <lx:label>All conditions to conditional approval have been met and Application has received formal/final approval.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valuation Approved">
        <xs:annotation>
          <xs:documentation>The property valuation has been approved.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Valuation Approved</lx:path>
            <lx:label>The property valuation has been approved.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valuation Cancelled">
        <xs:annotation>
          <xs:documentation>The property valuation ordered has been cancelled.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Valuation Cancelled</lx:path>
            <lx:label>The property valuation ordered has been cancelled.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valuation Not Required">
        <xs:annotation>
          <xs:documentation>A property valuation is not required.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Valuation Not Required</lx:path>
            <lx:label>A property valuation is not required.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valuation Ordered">
        <xs:annotation>
          <xs:documentation>The property valuation has been ordered.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Valuation Ordered</lx:path>
            <lx:label>The property valuation has been ordered.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valuation Received">
        <xs:annotation>
          <xs:documentation>A completed property valuation has been received from the Valuation Firm.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Valuation Received</lx:path>
            <lx:label>A completed property valuation has been received from the Valuation Firm.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Variation Initialised">
        <xs:annotation>
          <xs:documentation>A Variation to the borrower's facility has been initiated.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Variation Initialised</lx:path>
            <lx:label>A Variation to the borrower's facility has been initiated.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Withdrawn">
        <xs:annotation>
          <xs:documentation>Application has been stopped at client's initiative.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameApplicationInstructionsList.Withdrawn</lx:path>
            <lx:label>Application has been stopped at client's initiative.</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="eventNameCreditDecisionInstructionsList">
    <xs:annotation>
      <xs:documentation>eventNameCreditDecisionInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>eventNameCreditDecisionInstructionsList</lx:path>
        <lx:label>eventNameCreditDecisionInstructionsList</lx:label>
        <li:transactions>CDA</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Application Received">
        <xs:annotation>
          <xs:documentation>Application successfully received by credit decision system and can be further processed i.e. well formed.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Application Received</lx:path>
            <lx:label>Application Received</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Application Registered">
        <xs:annotation>
          <xs:documentation>Lender system ID allocated (Pre Settlement Application Identifier) and, if electronically lodged, application is ready for credit assessment ie. validates against Lender's schema and other business rules.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Application Registered</lx:path>
            <lx:label>Application Registered</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Application Sent">
        <xs:annotation>
          <xs:documentation>Application successfully transmitted by credit origination system.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Application Sent</lx:path>
            <lx:label>Application Sent</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cancelled">
        <xs:annotation>
          <xs:documentation>Application has been stopped at Lender's initiative. Cannot be resurrected.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Cancelled</lx:path>
            <lx:label>Cancelled</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Comment">
        <xs:annotation>
          <xs:documentation>The Details attribute contains free text comments.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Comment</lx:path>
            <lx:label>Comment</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Conditionally Approved">
        <xs:annotation>
          <xs:documentation>Application has been approved subject to conditions.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Conditionally Approved</lx:path>
            <lx:label>Conditionally Approved</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Check Ordered">
        <xs:annotation>
          <xs:documentation>A credit check with a credit bureau has been ordered.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Credit Check Ordered</lx:path>
            <lx:label>Credit Check Ordered</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Check Received">
        <xs:annotation>
          <xs:documentation>A credit check has been received from a credit bureau.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Credit Check Received</lx:path>
            <lx:label>Credit Check Received</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Declined">
        <xs:annotation>
          <xs:documentation>Application has been Declined.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Declined</lx:path>
            <lx:label>Declined</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Error on Submission">
        <xs:annotation>
          <xs:documentation>The application either has missing data or has invalid combinations of data. You will need to re-submit the application after correcting any data issues.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Error on Submission</lx:path>
            <lx:label>Error on Submission</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Identity Check Ordered">
        <xs:annotation>
          <xs:documentation>An identity check with an identity verification bureau has been ordered.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Identity Check Ordered</lx:path>
            <lx:label>Identity Check Ordered</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Identity Check Received">
        <xs:annotation>
          <xs:documentation>An identity check has been received from an identity verification bureau.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Identity Check Received</lx:path>
            <lx:label>Identity Check Received</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents and Requirements Complete">
        <xs:annotation>
          <xs:documentation>The Loan Documents and requirements received from the Borrower are in order.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Documents and Requirements Complete</lx:path>
            <lx:label>Loan Documents and Requirements Complete</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents and Requirements Incomplete">
        <xs:annotation>
          <xs:documentation>The Loan Documents and/or requirements received from the Borrower are not complete.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Documents and Requirements Incomplete</lx:path>
            <lx:label>Loan Documents and Requirements Incomplete</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Certified for Settlement">
        <xs:annotation>
          <xs:documentation>The Loan Documents have been certified.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Documents Certified for Settlement</lx:path>
            <lx:label>Loan Documents Certified for Settlement</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Error Message Returned">
        <xs:annotation>
          <xs:documentation>Error message returned to the Lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Documents Error Message Returned</lx:path>
            <lx:label>Loan Documents Error Message Returned</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Follow Up">
        <xs:annotation>
          <xs:documentation>A request to follow up the Loan Document production has been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Documents Follow Up</lx:path>
            <lx:label>Loan Documents Follow Up</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Issued">
        <xs:annotation>
          <xs:documentation>Loan Documents have been issued as directed by Lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Documents Issued</lx:path>
            <lx:label>Loan Documents Issued</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Produced">
        <xs:annotation>
          <xs:documentation>Loan documents have been produced.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Documents Produced</lx:path>
            <lx:label>Loan Documents Produced</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Produced and Returned to Lender">
        <xs:annotation>
          <xs:documentation>The Loan Documents have been produced and returned to the Lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Documents Produced and Returned to Lender</lx:path>
            <lx:label>Loan Documents Produced and Returned to Lender</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Re-issue Request Sent">
        <xs:annotation>
          <xs:documentation>The request to reissue the Loan Documents has been sent.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Documents Re-issue Request Sent</lx:path>
            <lx:label>Loan Documents Re-issue Request Sent</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Re-Issued">
        <xs:annotation>
          <xs:documentation>The Loan Documents have been re-issued.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Documents Re-Issued</lx:path>
            <lx:label>Loan Documents Re-Issued</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Received">
        <xs:annotation>
          <xs:documentation>Signed Loan Documents have been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Documents Received</lx:path>
            <lx:label>Loan Documents Received</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Request Received">
        <xs:annotation>
          <xs:documentation>A request to produce Loan Documents has been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Documents Request Received</lx:path>
            <lx:label>Loan Documents Request Received</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Request Sent">
        <xs:annotation>
          <xs:documentation>A request to produce Loan Documents has been sent.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Documents Request Sent</lx:path>
            <lx:label>Loan Documents Request Sent</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Returned Enquiry">
        <xs:annotation>
          <xs:documentation>An enquiry about whether signed Loan Documents have been returned has been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Documents Returned Enquiry</lx:path>
            <lx:label>Loan Documents Returned Enquiry</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Funded by Lender">
        <xs:annotation>
          <xs:documentation>The funds have been debited to the Loan Account.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Loan Funded by Lender</lx:path>
            <lx:label>Loan Funded by Lender</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Payout Figure Received">
        <xs:annotation>
          <xs:documentation>The total amount required to pay out the loan has been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Payout Figure Received</lx:path>
            <lx:label>Payout Figure Received</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Payout Figure Rejected">
        <xs:annotation>
          <xs:documentation>The total amount required to pay out the loan has been rejected.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Payout Figure Rejected</lx:path>
            <lx:label>Payout Figure Rejected</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Policy Exception Encountered">
        <xs:annotation>
          <xs:documentation>A credit policy has failed based on the application data provided.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Policy Exception Encountered</lx:path>
            <lx:label>Policy Exception Encountered</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PreApproved">
        <xs:annotation>
          <xs:documentation>Application conditionally approved subject to final identification of security property.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.PreApproved</lx:path>
            <lx:label>Pre-Approved</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Progress Payment Complete">
        <xs:annotation>
          <xs:documentation>A progress payment has been completed.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Progress Payment Complete</lx:path>
            <lx:label>Progress Payment Complete</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Progress Payment Message">
        <xs:annotation>
          <xs:documentation>A message about the Progress Payment process.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Progress Payment Message</lx:path>
            <lx:label>Progress Payment Message</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ready for Progress Payment">
        <xs:annotation>
          <xs:documentation>The application is in a state in which it is ready to begin the Progress Payment process.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Ready for Progress Payment</lx:path>
            <lx:label>Ready for Progress Payment</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ready for Settlement">
        <xs:annotation>
          <xs:documentation>The application is in a state in which it is ready to begin the Settlement process.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Ready for Settlement</lx:path>
            <lx:label>Ready for Settlement</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Referred for Credit Assessment">
        <xs:annotation>
          <xs:documentation>Application has been referred for manual credit assessment.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Referred for Credit Assessment</lx:path>
            <lx:label>Referred for Credit Assessment</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Referred to Broker">
        <xs:annotation>
          <xs:documentation>Application has been referred to the Broker.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Referred to Broker</lx:path>
            <lx:label>Referred to Broker</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Referred to Mortgage Insurer">
        <xs:annotation>
          <xs:documentation>Application has been referred to the Mortgage Insurer.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Referred to Mortgage Insurer</lx:path>
            <lx:label>Referred to Mortgage Insurer</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Reverted to Capture">
        <xs:annotation>
          <xs:documentation>Application has been reverted to the data capture stage to update or correct data.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Reverted to Capture</lx:path>
            <lx:label>Reverted to Capture</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Documents Sent to Custodian">
        <xs:annotation>
          <xs:documentation>The security documents have been sent to the custodian.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Security Documents Sent to Custodian</lx:path>
            <lx:label>Security Documents Sent to Custodian</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Documents Sent to Lender">
        <xs:annotation>
          <xs:documentation>The security documents have been sent to the lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Security Documents Sent to Lender</lx:path>
            <lx:label>Security Documents Sent to Lender</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Packet Received">
        <xs:annotation>
          <xs:documentation>The security packet has been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Security Packet Received</lx:path>
            <lx:label>Security Packet Received</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Registered">
        <xs:annotation>
          <xs:documentation>The security has been registered with the Land Titles Office</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Security Registered</lx:path>
            <lx:label>Security Registered</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Service Coordinator Assigned">
        <xs:annotation>
          <xs:documentation>A service coordinator has been assigned to process the application.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Service Coordinator Assigned</lx:path>
            <lx:label>Service Coordinator Assigned</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Cancelled">
        <xs:annotation>
          <xs:documentation>The settlement process has been cancelled.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Settlement Cancelled</lx:path>
            <lx:label>Settlement Cancelled</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Completed">
        <xs:annotation>
          <xs:documentation>Application has gone to settlement; may or may not have been fully drawn down; accounts may have had a new identifier allocated to them by the Lender (known as Post Settlement Application Number or Facility identifier).</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Settlement Completed</lx:path>
            <lx:label>Settlement Completed</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Date Booked">
        <xs:annotation>
          <xs:documentation>The settlement date has been booked.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Settlement Date Booked</lx:path>
            <lx:label>Settlement Date Booked</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Date Confirmed">
        <xs:annotation>
          <xs:documentation>The settlement date has been confirmed by all parties.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Settlement Date Confirmed</lx:path>
            <lx:label>Settlement Date Confirmed</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Date Estimated">
        <xs:annotation>
          <xs:documentation>The estimated settlement date, which is 21 days or as long as six weeks post the exchange of contracts.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Settlement Date Estimated</lx:path>
            <lx:label>Settlement Date Estimated</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Delayed">
        <xs:annotation>
          <xs:documentation>The settlement process has been delayed.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Settlement Delayed</lx:path>
            <lx:label>Settlement Delayed</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Funds Received">
        <xs:annotation>
          <xs:documentation>The funds required at settlement have been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Settlement Funds Received</lx:path>
            <lx:label>Settlement Funds Received</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Rebooked">
        <xs:annotation>
          <xs:documentation>The settlement date has been rebooked.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Settlement Rebooked</lx:path>
            <lx:label>Settlement Rebooked</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Solicitor Instructed">
        <xs:annotation>
          <xs:documentation>The solicitors conducting the property transaction have been instructed.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Solicitor Instructed</lx:path>
            <lx:label>Solicitor Instructed</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Solicitor Instruction Received">
        <xs:annotation>
          <xs:documentation>The instruction for the solicitors conducting the property transaction have been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Solicitor Instruction Received</lx:path>
            <lx:label>Solicitor Instruction Received</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Supporting Documents Received">
        <xs:annotation>
          <xs:documentation>The customer provided supporting documentation have been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Supporting Documents Received</lx:path>
            <lx:label>Supporting Documents Received</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unconditionally Approved">
        <xs:annotation>
          <xs:documentation>All conditions to conditional approval have been met and Application has received formal/final approval.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Unconditionally Approved</lx:path>
            <lx:label>Unconditionally Approved</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valuation Approved">
        <xs:annotation>
          <xs:documentation>The property valuation has been approved.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Valuation Approved</lx:path>
            <lx:label>Valuation Approved</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valuation Cancelled">
        <xs:annotation>
          <xs:documentation>The property valuation ordered has been cancelled.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Valuation Cancelled</lx:path>
            <lx:label>Valuation Cancelled</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valuation Not Required">
        <xs:annotation>
          <xs:documentation>A property valuation is not required.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Valuation Not Required</lx:path>
            <lx:label>Valuation Not Required</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valuation Ordered">
        <xs:annotation>
          <xs:documentation>The property valuation has been ordered.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Valuation Ordered</lx:path>
            <lx:label>Valuation Ordered</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valuation Received">
        <xs:annotation>
          <xs:documentation>A completed property valuation has been received from the Valuation Firm.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Valuation Received</lx:path>
            <lx:label>Valuation Received</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Variation Initialised">
        <xs:annotation>
          <xs:documentation>A Variation to the borrower's facility has been initiated.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Variation Initialised</lx:path>
            <lx:label>Variation Initialised</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Withdrawn">
        <xs:annotation>
          <xs:documentation>Application has been stopped at client's initiative.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameCreditDecisionInstructionsList.Withdrawn</lx:path>
            <lx:label>Withdrawn</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="eventNameLendersMortgageInsuranceInstructionsList">
    <xs:annotation>
      <xs:documentation>eventNameLendersMortgageInsuranceInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>eventNameLendersMortgageInsuranceInstructionsList</lx:path>
        <lx:label>eventNameLendersMortgageInsuranceInstructionsList</lx:label>
        <li:transactions>LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="More Information Received">
        <xs:annotation>
          <xs:documentation>More information has been received by the insurance provider from the instructor.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLendersMortgageInsuranceInstructionsList.More Information Received</lx:path>
            <lx:label>More Information Received</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="More Information Requested">
        <xs:annotation>
          <xs:documentation>More information has been requested from the instructor by the insurance provider.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLendersMortgageInsuranceInstructionsList.More Information Requested</lx:path>
            <lx:label>More Information Requested</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Supporting Documents Determined to Be Incomplete">
        <xs:annotation>
          <xs:documentation>The supporting documents provided by the customer do not cover the checklist requirements, additional supporting documents are expected.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLendersMortgageInsuranceInstructionsList.Supporting Documents Determined to Be Incomplete</lx:path>
            <lx:label>Supporting Documents Determined to Be Incomplete</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Supporting Documents Have Been Verified">
        <xs:annotation>
          <xs:documentation>The customer provided supporting documents have been verified.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLendersMortgageInsuranceInstructionsList.Supporting Documents Have Been Verified</lx:path>
            <lx:label>Supporting Documents Have Been Verified</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Supporting Documents Received">
        <xs:annotation>
          <xs:documentation>The customer provided supporting documentation have been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLendersMortgageInsuranceInstructionsList.Supporting Documents Received</lx:path>
            <lx:label>The customer provided supporting documentation have been received.</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="eventNameLoanDocumentsInstructionsList">
    <xs:annotation>
      <xs:documentation>eventNameLoanDocumentsInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>eventNameLoanDocumentsInstructionsList</lx:path>
        <lx:label>eventNameLoanDocumentsInstructionsList</lx:label>
        <li:transactions>DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Certified For Settlement">
        <xs:annotation>
          <xs:documentation>The loan application has been certified for settlement.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Certified For Settlement</lx:path>
            <lx:label>The loan application has been certified for settlement.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents and Requirements Complete">
        <xs:annotation>
          <xs:documentation>The Loan Documents and requirements received from the Borrower are in order.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Loan Documents and Requirements Complete</lx:path>
            <lx:label>The Loan Documents and requirements received from the Borrower are in order.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents and Requirements Incomplete">
        <xs:annotation>
          <xs:documentation>The Loan Documents and/or requirements received from the Borrower are not complete.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Loan Documents and Requirements Incomplete</lx:path>
            <lx:label>The Loan Documents and/or requirements received from the Borrower are not complete.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Error Message Returned">
        <xs:annotation>
          <xs:documentation>Error message returned to the Lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Loan Documents Error Message Returned</lx:path>
            <lx:label>Error message returned to the Lender.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Produced">
        <xs:annotation>
          <xs:documentation>Loan documents have been produced.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Loan Documents Produced</lx:path>
            <lx:label>Loan documents have been produced.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Produced and Returned to Lender">
        <xs:annotation>
          <xs:documentation>The Loan Documents have been produced and returned to the Lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Loan Documents Produced and Returned to Lender</lx:path>
            <lx:label>The Loan Documents have been produced and returned to the Lender.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Re-issue Request Sent">
        <xs:annotation>
          <xs:documentation>The request to reissue the Loan Documents has been sent.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Loan Documents Re-issue Request Sent</lx:path>
            <lx:label>The request to reissue the Loan Documents has been sent.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Received">
        <xs:annotation>
          <xs:documentation>Signed Loan Documents have been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Loan Documents Received</lx:path>
            <lx:label>Signed Loan Documents have been received.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Request Sent">
        <xs:annotation>
          <xs:documentation>A request to produce Loan Documents has been sent.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Loan Documents Request Sent</lx:path>
            <lx:label>A request to produce Loan Documents has been sent.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Returned Enquiry">
        <xs:annotation>
          <xs:documentation>An enquiry about whether signed Loan Documents have been returned has been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Loan Documents Returned Enquiry</lx:path>
            <lx:label>An enquiry about whether signed Loan Documents have been returned has been received.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Documents Served">
        <xs:annotation>
          <xs:documentation>Loan Documents have been served as directed by Lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Loan Documents Served</lx:path>
            <lx:label>Loan Documents have been served as directed by Lender.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Funded by Lender">
        <xs:annotation>
          <xs:documentation>The funds have been debited to the Loan Account.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Loan Funded by Lender</lx:path>
            <lx:label>The funds have been debited to the Loan Account.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Documents Sent to Custodian">
        <xs:annotation>
          <xs:documentation>The security documents have been sent to the custodian.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Security Documents Sent to Custodian</lx:path>
            <lx:label>The security documents have been sent to the custodian.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Documents Sent to Lender">
        <xs:annotation>
          <xs:documentation>The security documents have been sent to the lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Security Documents Sent to Lender</lx:path>
            <lx:label>The security documents have been sent to the lender.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Packet Received">
        <xs:annotation>
          <xs:documentation>The security packet has been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Security Packet Received</lx:path>
            <lx:label>The security packet has been received.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Registered">
        <xs:annotation>
          <xs:documentation>The security has been registered with the Land Titles Office</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Security Registered</lx:path>
            <lx:label>The security has been registered with the Land Titles Office</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Cancelled">
        <xs:annotation>
          <xs:documentation>The settlement process has been cancelled.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Settlement Cancelled</lx:path>
            <lx:label>The settlement process has been cancelled.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Completed">
        <xs:annotation>
          <xs:documentation>Application has gone to settlement; may or may not have been fully drawn down; accounts may have had a new identifier allocated to them by the Lender (known as Post Settlement Application Number or Facility identifier).</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Settlement Completed</lx:path>
            <lx:label>Application has gone to settlement; may or may not have been fully drawn down; accounts may have had a new identifier allocated to them by the Lender (known as Post Settlement Application Number or Facility identifier).</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Date Booked">
        <xs:annotation>
          <xs:documentation>The settlement date has been booked.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Settlement Date Booked</lx:path>
            <lx:label>The settlement date has been booked.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Date Confirmed">
        <xs:annotation>
          <xs:documentation>The settlement date has been confirmed by all parties.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Settlement Date Confirmed</lx:path>
            <lx:label>The settlement date has been confirmed by all parties.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Date Estimated">
        <xs:annotation>
          <xs:documentation>The estimated settlement date, which is 21 days or as long as six weeks post the exchange of contracts.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Settlement Date Estimated</lx:path>
            <lx:label>The estimated settlement date, which is 21 days or as long as six weeks post the exchange of contracts.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Delayed">
        <xs:annotation>
          <xs:documentation>The settlement process has been delayed.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Settlement Delayed</lx:path>
            <lx:label>The settlement process has been delayed.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Funds Received">
        <xs:annotation>
          <xs:documentation>The funds required at settlement have been received.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Settlement Funds Received</lx:path>
            <lx:label>The funds required at settlement have been received.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Rebooked">
        <xs:annotation>
          <xs:documentation>The settlement date has been rebooked.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameLoanDocumentsInstructionsList.Settlement Rebooked</lx:path>
            <lx:label>The settlement date has been rebooked.</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="eventNameValuationInstructionsList">
    <xs:annotation>
      <xs:documentation>eventNameValuationInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>eventNameValuationInstructionsList</lx:path>
        <lx:label>eventNameValuationInstructionsList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Comment">
        <xs:annotation>
          <xs:documentation>The subject attribute and details attribute contain free text comments.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameValuationInstructionsList.Comment</lx:path>
            <lx:label>Comment</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fee Change Accepted">
        <xs:annotation>
          <xs:documentation>The valuation fee change was accepted.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameValuationInstructionsList.Fee Change Accepted</lx:path>
            <lx:label>Fee Change Accepted</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fee Change Cancelled">
        <xs:annotation>
          <xs:documentation>The valuation fee change request was cancelled.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameValuationInstructionsList.Fee Change Cancelled</lx:path>
            <lx:label>Fee Change Cancelled</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fee Change Rejected">
        <xs:annotation>
          <xs:documentation>The valuation fee change was rejected.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameValuationInstructionsList.Fee Change Rejected</lx:path>
            <lx:label>Fee Change Rejected</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fee Change Requested">
        <xs:annotation>
          <xs:documentation>A valuation fee change was requested.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameValuationInstructionsList.Fee Change Requested</lx:path>
            <lx:label>Fee Change Requested</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Order Assigned to Valuer">
        <xs:annotation>
          <xs:documentation>The valuation order was assigned a valuer.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameValuationInstructionsList.Order Assigned to Valuer</lx:path>
            <lx:label>Order Assigned to Valuer</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Order Charged">
        <xs:annotation>
          <xs:documentation>The valuation order was charged to the lender's cost centre.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameValuationInstructionsList.Order Charged</lx:path>
            <lx:label>Order Charged</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Order Created">
        <xs:annotation>
          <xs:documentation>The valuation order was created.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameValuationInstructionsList.Order Created</lx:path>
            <lx:label>Order Created</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Order Reassigned to New Valuer">
        <xs:annotation>
          <xs:documentation>The valuation order was reassigned a different valuer.</xs:documentation>
          <xs:appinfo>
            <lx:path>eventNameValuationInstructionsList.Order Reassigned to New Valuer</lx:path>
            <lx:label>Order Reassigned to New Valuer</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="excessUsageCalculationMethodList">
    <xs:annotation>
      <xs:documentation>excessUsageCalculationMethodList</xs:documentation>
      <xs:appinfo>
        <lx:path>excessUsageCalculationMethodList</lx:path>
        <lx:label>excessUsageCalculationMethodList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Per Annum">
        <xs:annotation>
          <xs:documentation>Per Annum</xs:documentation>
          <xs:appinfo>
            <lx:path>excessUsageCalculationMethodList.Per Annum</lx:path>
            <lx:label>Per Annum</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Per Loan">
        <xs:annotation>
          <xs:documentation>Per Loan</xs:documentation>
          <xs:appinfo>
            <lx:path>excessUsageCalculationMethodList.Per Loan</lx:path>
            <lx:label>Per Loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Per Month">
        <xs:annotation>
          <xs:documentation>Per Month</xs:documentation>
          <xs:appinfo>
            <lx:path>excessUsageCalculationMethodList.Per Month</lx:path>
            <lx:label>Per Month</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="exemptStatusList">
    <xs:annotation>
      <xs:documentation>exemptStatusList</xs:documentation>
      <xs:appinfo>
        <lx:path>exemptStatusList</lx:path>
        <lx:label>exemptStatusList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Exempt">
        <xs:annotation>
          <xs:documentation>The company is an exempt company as defined in the Corporations Law.</xs:documentation>
          <xs:appinfo>
            <lx:path>exemptStatusList.Exempt</lx:path>
            <lx:label>The company is an exempt company as defined in the Corporations Law.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Non Exempt">
        <xs:annotation>
          <xs:documentation>The company is not an exempt company as defined in the Corporations Law.</xs:documentation>
          <xs:appinfo>
            <lx:path>exemptStatusList.Non Exempt</lx:path>
            <lx:label>The company is not an exempt company as defined in the Corporations Law.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="feeCategoryList">
    <xs:annotation>
      <xs:documentation>feeCategoryList</xs:documentation>
      <xs:appinfo>
        <lx:path>feeCategoryList</lx:path>
        <lx:label>feeCategoryList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Bank">
        <xs:annotation>
          <xs:documentation>Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>feeCategoryList.Bank</lx:path>
            <lx:label>Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government">
        <xs:annotation>
          <xs:documentation>Government</xs:documentation>
          <xs:appinfo>
            <lx:path>feeCategoryList.Government</lx:path>
            <lx:label>Government</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mortgage Insurer">
        <xs:annotation>
          <xs:documentation>Mortgage Insurer</xs:documentation>
          <xs:appinfo>
            <lx:path>feeCategoryList.Mortgage Insurer</lx:path>
            <lx:label>Mortgage Insurer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>feeCategoryList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valuer">
        <xs:annotation>
          <xs:documentation>Valuer</xs:documentation>
          <xs:appinfo>
            <lx:path>feeCategoryList.Valuer</lx:path>
            <lx:label>Valuer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="feePaymentTimingList">
    <xs:annotation>
      <xs:documentation>A list of timings that describe when fees are payable.</xs:documentation>
      <xs:appinfo>
        <lx:path>feePaymentTimingList</lx:path>
        <lx:label>Fee Payment Timing List</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Before Credit Drawdown">
        <xs:annotation>
          <xs:documentation>The fee is payable before credit drawdown occurs.</xs:documentation>
          <xs:appinfo>
            <lx:path>feePaymentTimingList.Before Credit Drawdown</lx:path>
            <lx:label>Before Credit Drawdown</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="On Credit Drawdown">
        <xs:annotation>
          <xs:documentation>The fee is payable upon credit drawdown.</xs:documentation>
          <xs:appinfo>
            <lx:path>feePaymentTimingList.On Credit Drawdown</lx:path>
            <lx:label>On Credit Drawdown</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="On Event Occurrence">
        <xs:annotation>
          <xs:documentation>The fee is payable upon the occurrence of a specific event.</xs:documentation>
          <xs:appinfo>
            <lx:path>feePaymentTimingList.On Event Occurrence</lx:path>
            <lx:label>On Event Occurrence</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Regular Recurring">
        <xs:annotation>
          <xs:documentation>The fee is payable on a regular schedule.</xs:documentation>
          <xs:appinfo>
            <lx:path>feePaymentTimingList.Regular Recurring</lx:path>
            <lx:label>Regular Recurring</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="feeTypeList">
    <xs:annotation>
      <xs:documentation>feeTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>feeTypeList</lx:path>
        <lx:label>feeTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Acceptance Fee">
        <xs:annotation>
          <xs:documentation>A fee charged by the bank for taking legal obligation of paying a commercial bill on its maturity date. Generally used when taking over a commercial bill.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Acceptance Fee</lx:path>
            <lx:label>Acceptance Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Additional Security Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged for adding one or more additional securities after loan settlement.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Additional Security Fee</lx:path>
            <lx:label>Additional Security Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Annual Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged for administration and maintenance of credit facilities, usually for revolving credit loans or credit cards.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Annual Fee</lx:path>
            <lx:label>Annual Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Application Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged for processing and assessment of the application for the loan.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Application Fee</lx:path>
            <lx:label>Fee is charged for processing and assessment of the application for the loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Brokerage Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged when a broker assisted borrowers in finding and obtaining approval for a loan. This fee may be charged to the buyer, seller or both.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Brokerage Fee</lx:path>
            <lx:label>Brokerage Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cash Benefit Insurance">
        <xs:annotation>
          <xs:documentation>Premium charged for taking out Cash Benefit Insurance. Cash benefit insurance is related to personal loan products (in particular car loans), and covers the owner if their car is written off/destroyed for any costs of replacing the vehicle (above and beyond the payout they would receive from full comprehensive coverage).</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Cash Benefit Insurance</lx:path>
            <lx:label>Cash Benefit Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Construction Progress Inspection Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged each time a property valuer inspects building works during the course of a property being built in order to ensure that building is in accordance with the plans and expectations.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Construction Progress Inspection Fee</lx:path>
            <lx:label>Construction Progress Inspection Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Consumer Credit Insurance">
        <xs:annotation>
          <xs:documentation>Premium charged for taking out Consumer Credit Insurance (CCI). CCI is an optional insurance policy that protects borrowers against changes in circumstances that negatively affect their ability to pay back the loan such as death, injury or illness and involuntary unemployment. The premium can generally be added to the loan as a lump sum or can be paid monthly.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Consumer Credit Insurance</lx:path>
            <lx:label>Consumer Credit Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Documentation Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged for preparation of loan documents. The amount varies based on the complexity of the loan application.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Documentation Fee</lx:path>
            <lx:label>Documentation Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Early Repayment Admin Fee">
        <xs:annotation>
          <xs:documentation>Fee charged when customer repays the loan earlier than the end of the loan term.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Early Repayment Admin Fee</lx:path>
            <lx:label>Early Repayment Admin Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Establishment Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged for the establishment of the loan facility. For some financial institutions, this is a general term used for fees that one may incur for the bank to set up a loan which may include Application Fee, Settlement Fee, etc. For some financial institutions, this is the same as an Application Fee.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Establishment Fee</lx:path>
            <lx:label>Fee is charged for the establishment of the loan facility</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Extended Warranty Insurance">
        <xs:annotation>
          <xs:documentation>Premium charged for taking out Extended Warranty Insurance. This insurance (also called mechanical breakdown insurance or motor breakdown insurance (MBI)) commences after your new car warranty ends and provides some cover for the repair or replacement of specific parts.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Extended Warranty Insurance</lx:path>
            <lx:label>Extended Warranty Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Guarantee Asset Protection Insurance">
        <xs:annotation>
          <xs:documentation>Premium charged for taking out Guarantee Asset Protection (GAP) Insurance. GAP insurance (also called motor equity insurance or shortfall insurance) covers the lender for the difference between what a borrower owes on the car loan, and what the car is insured for under comprehensive car insurance, if car is written off.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Guarantee Asset Protection Insurance</lx:path>
            <lx:label>Guarantee Asset Protection Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Inspection Fee">
        <xs:annotation>
          <xs:documentation>A fee charged for the inspection of an asset. An inspection usually occurs to confirm the location of the asset, determine the condition of the asset, and confirm the asset documents match with the physical asset.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Inspection Fee</lx:path>
            <lx:label>Inspection Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Land Titles Office Final Search Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged when a final search of the title is ordered to ensure that the property is clear from any interest or restrictions that may have been recorded prior to settlement.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Land Titles Office Final Search Fee</lx:path>
            <lx:label>Land Titles Office Final Search Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Land Titles Office Search Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged when title search is ordered to ascertain the ownership of a property and determine whether there are existing encumbrances, covenants or easements that may be recorded on the title.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Land Titles Office Search Fee</lx:path>
            <lx:label>Land Titles Office Search Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Legal Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged to cover legal fees incurred by the lender in arranging the loan</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Legal Fee</lx:path>
            <lx:label>Fee is charged to cover legal fees incurred by the lender in arranging the loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LMI">
        <xs:annotation>
          <xs:documentation>LMI</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.LMI</lx:path>
            <lx:label>LMI</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Fee">
        <xs:annotation>
          <xs:documentation>Loan Fee</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Loan Fee</lx:path>
            <lx:label>Loan Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mortgage Registration">
        <xs:annotation>
          <xs:documentation>Fee is charged by the Title Office to register the lender's mortgage on the title for a security</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Mortgage Registration</lx:path>
            <lx:label>Fee is charged by the Title Office to register the lender's mortgage on the title for a security</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Motor Vehicle Insurance">
        <xs:annotation>
          <xs:documentation>Motor Vehicle Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Motor Vehicle Insurance</lx:path>
            <lx:label>Motor Vehicle Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Fee charged is not described by one of the enumerated values in the list</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Other</lx:path>
            <lx:label>Fee charged is not described by one of the enumerated values in the list</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Package Fee">
        <xs:annotation>
          <xs:documentation>Package Fee</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Package Fee</lx:path>
            <lx:label>Package Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PPSR Fee">
        <xs:annotation>
          <xs:documentation>A fee associated with registering a non-real estate asset on the Personal Property Securities Register (PPSR) when it is used as security on a loan.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.PPSR Fee</lx:path>
            <lx:label>PPSR Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rate Lock Fee">
        <xs:annotation>
          <xs:documentation>Rate Lock Fee</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Rate Lock Fee</lx:path>
            <lx:label>Rate Lock Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Re-documentation Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged when documents such as Loan Agreement or Variation Contract are re-issued prior to settlement due to change requested by either customer or introducer.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Re-documentation Fee</lx:path>
            <lx:label>Re-documentation Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Reduced Equity Fee">
        <xs:annotation>
          <xs:documentation>Reduced Equity Fee</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Reduced Equity Fee</lx:path>
            <lx:label>Reduced Equity Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Risk Fee">
        <xs:annotation>
          <xs:documentation>Risk Fee</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Risk Fee</lx:path>
            <lx:label>Risk Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rollover Fee">
        <xs:annotation>
          <xs:documentation>A fee charged when rolling over a commerical bill.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Rollover Fee</lx:path>
            <lx:label>Rollover Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Guarantee Fee">
        <xs:annotation>
          <xs:documentation>Security Guarantee Fee</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Security Guarantee Fee</lx:path>
            <lx:label>Security Guarantee Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Release Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged when a lender removes their legal charge from the security held against a mortgage or loan. This is sometimes called Discharge Fee.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Security Release Fee</lx:path>
            <lx:label>Security Release Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Service Plan">
        <xs:annotation>
          <xs:documentation>Service Plan</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Service Plan</lx:path>
            <lx:label>Service Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement Fee">
        <xs:annotation>
          <xs:documentation>Settlement Fee</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Settlement Fee</lx:path>
            <lx:label>Settlement Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Stamp Duty on Mortgage">
        <xs:annotation>
          <xs:documentation>Fee paid to the State Revenue Office for stamping of mortgage or other security document</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Stamp Duty on Mortgage</lx:path>
            <lx:label>Fee paid to the State Revenue Office for stamping of mortgage or other security document</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Stamp Duty on Transfer of Property">
        <xs:annotation>
          <xs:documentation>Fee paid to the State Revenue Office for stamping of transfer or other assignment of ownership document</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Stamp Duty on Transfer of Property</lx:path>
            <lx:label>Fee paid to the State Revenue Office for stamping of transfer or other assignment of ownership document</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Substitute Security Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged when the current property security of a loan is substituted for another property. This usually occurs when the borrower intends to sell the existing property security.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Substitute Security Fee</lx:path>
            <lx:label>Substitute Security Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Switch Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged when all or part of an existing loan balance will be converted to another product, another type of interest rate (change between fixed and variable) or repayment type (change between interest only payments and principal and interest payments).</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Switch Fee</lx:path>
            <lx:label>Switch Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Third Party Document Production Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged when producing a Certificate of Title or other security documents.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Third Party Document Production Fee</lx:path>
            <lx:label>Third Party Document Production Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Title Transfer Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged when there is a transfer of ownership; for example property title transfer or vehicle registration transfer.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Title Transfer Fee</lx:path>
            <lx:label>Title Transfer Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tyre and Rim Insurance">
        <xs:annotation>
          <xs:documentation>Tyre and Rim Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Tyre and Rim Insurance</lx:path>
            <lx:label>Tyre and Rim Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unused Facility Fee">
        <xs:annotation>
          <xs:documentation>A fee charged when a customer has unused limit on their facility. Applies to commerical bills only.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Unused Facility Fee</lx:path>
            <lx:label>Unused Facility Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valuation Fee">
        <xs:annotation>
          <xs:documentation>Valuation Fee</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Valuation Fee</lx:path>
            <lx:label>Valuation Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Withdrawal of Caveat Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged when caveat will be withdrawn/removed and lodged to the corresponding land or property information office.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Withdrawal of Caveat Fee</lx:path>
            <lx:label>Withdrawal of Caveat Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Withdrawal of Writ Fee">
        <xs:annotation>
          <xs:documentation>Fee is charged for the removal of a writ recorded against a title.</xs:documentation>
          <xs:appinfo>
            <lx:path>feeTypeList.Withdrawal of Writ Fee</lx:path>
            <lx:label>Withdrawal of Writ Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="financialAccountStatusList">
    <xs:annotation>
      <xs:documentation>financialAccountStatusList</xs:documentation>
      <xs:appinfo>
        <lx:path>financialAccountStatusList</lx:path>
        <lx:label>financialAccountStatusList</lx:label>
        <li:transactions>CDA,SVC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Active">
        <xs:annotation>
          <xs:documentation>Indicates that the account is active (open).</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAccountStatusList.Active</lx:path>
            <lx:label>Active</lx:label>
            <li:transactions>CDA,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Arrears">
        <xs:annotation>
          <xs:documentation>Indicates that the account is in arrears.</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAccountStatusList.Arrears</lx:path>
            <lx:label>Arrears</lx:label>
            <li:transactions>CDA,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bad Debt">
        <xs:annotation>
          <xs:documentation>Indicates that the account has bad debt.</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAccountStatusList.Bad Debt</lx:path>
            <lx:label>Bad Debt</lx:label>
            <li:transactions>CDA,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Closed">
        <xs:annotation>
          <xs:documentation>Indicates that the account has been closed.</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAccountStatusList.Closed</lx:path>
            <lx:label>Closed</lx:label>
            <li:transactions>CDA,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pending">
        <xs:annotation>
          <xs:documentation>Indicates that the account is in a pending (not active) state.</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAccountStatusList.Pending</lx:path>
            <lx:label>Pending</lx:label>
            <li:transactions>CDA,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="financialAssetTypeList">
    <xs:annotation>
      <xs:documentation>financialAssetTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>financialAssetTypeList</lx:path>
        <lx:label>financialAssetTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Cash">
        <xs:annotation>
          <xs:documentation>Asset is cash that is not on deposit with a financial institution</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAssetTypeList.Cash</lx:path>
            <lx:label>Asset is cash that is not on deposit with a financial institution</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Investment">
        <xs:annotation>
          <xs:documentation>A financial asset that is purchased with the hope that it will generate income or appreciate in the future.</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAssetTypeList.Investment</lx:path>
            <lx:label>A financial asset that is purchased with the hope that it will generate income or appreciate in the future.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="KiwiSaver">
        <xs:annotation>
          <xs:documentation>Money available through KiwiSaver a New Zealand voluntary, work-based savings initiative.</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAssetTypeList.KiwiSaver</lx:path>
            <lx:label>KiwiSaver</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Managed Funds">
        <xs:annotation>
          <xs:documentation>Shares or units in a managed fund</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAssetTypeList.Managed Funds</lx:path>
            <lx:label>Shares or units in a managed fund</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Asset is not described by any of the other enumerated values in the list</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAssetTypeList.Other</lx:path>
            <lx:label>Asset is not described by any of the other enumerated values in the list</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pension Account">
        <xs:annotation>
          <xs:documentation>Pension Account</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAssetTypeList.Pension Account</lx:path>
            <lx:label>Pension Account</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Receivables">
        <xs:annotation>
          <xs:documentation>Trade receivables owed by debtors of a business</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAssetTypeList.Receivables</lx:path>
            <lx:label>Trade receivables owed by debtors of a business</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Savings Account">
        <xs:annotation>
          <xs:documentation>Savings Account</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAssetTypeList.Savings Account</lx:path>
            <lx:label>Savings Account</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Shares and Debentures">
        <xs:annotation>
          <xs:documentation>Listed and unlisted shares, business equity, bills, bonds and debentures</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAssetTypeList.Shares and Debentures</lx:path>
            <lx:label>Listed and unlisted shares, business equity, bills, bonds and debentures</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Superannuation">
        <xs:annotation>
          <xs:documentation>Value held in a superannuation fund</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAssetTypeList.Superannuation</lx:path>
            <lx:label>Value held in a superannuation fund</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Term Deposit">
        <xs:annotation>
          <xs:documentation>Term deposit accounts, including annuities</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAssetTypeList.Term Deposit</lx:path>
            <lx:label>Term deposit accounts, including annuities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Transaction Account">
        <xs:annotation>
          <xs:documentation>Includes deposit account, cash management account, and chequing account</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAssetTypeList.Transaction Account</lx:path>
            <lx:label>Includes deposit account, cash management account, and chequing account</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trust Beneficial Interest">
        <xs:annotation>
          <xs:documentation>Beneficial ownership of equity within a trust.</xs:documentation>
          <xs:appinfo>
            <lx:path>financialAssetTypeList.Trust Beneficial Interest</lx:path>
            <lx:label>Trust Beneficial Interest</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="financialInstitutionList">
    <xs:annotation>
      <xs:documentation>financialInstitutionList</xs:documentation>
      <xs:appinfo>
        <lx:path>financialInstitutionList</lx:path>
        <lx:label>financialInstitutionList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="86 400">
        <xs:annotation>
          <xs:documentation>86 400</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.86 400</lx:path>
            <lx:label>86 400</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AAMI">
        <xs:annotation>
          <xs:documentation>AAMI</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.AAMI</lx:path>
            <lx:label>AAMI</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AAPBS Settlements Services">
        <xs:annotation>
          <xs:documentation>AAPBS Settlements Services</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.AAPBS Settlements Services</lx:path>
            <lx:label>AAPBS Settlements Services</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ABN AMRO">
        <xs:annotation>
          <xs:documentation>ABN AMRO</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.ABN AMRO</lx:path>
            <lx:label>ABN AMRO</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Adelaide Bank">
        <xs:annotation>
          <xs:documentation>Adelaide Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Adelaide Bank</lx:path>
            <lx:label>Adelaide Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Advantedge">
        <xs:annotation>
          <xs:documentation>Advantedge</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Advantedge</lx:path>
            <lx:label>Advantedge</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AFG Home Loans">
        <xs:annotation>
          <xs:documentation>AFG Home Loans</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.AFG Home Loans</lx:path>
            <lx:label>AFG Home Loans</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Afterpay">
        <xs:annotation>
          <xs:documentation>Afterpay</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Afterpay</lx:path>
            <lx:label>Afterpay</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AGC">
        <xs:annotation>
          <xs:documentation>AGC</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.AGC</lx:path>
            <lx:label>AGC</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Allianz">
        <xs:annotation>
          <xs:documentation>Allianz</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Allianz</lx:path>
            <lx:label>Allianz</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Allstate Home Loans">
        <xs:annotation>
          <xs:documentation>Allstate Home Loans</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Allstate Home Loans</lx:path>
            <lx:label>Allstate Home Loans</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="American Express">
        <xs:annotation>
          <xs:documentation>American Express</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.American Express</lx:path>
            <lx:label>American Express</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AMG">
        <xs:annotation>
          <xs:documentation>AMG</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.AMG</lx:path>
            <lx:label>AMG</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AMIC">
        <xs:annotation>
          <xs:documentation>AMIC</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.AMIC</lx:path>
            <lx:label>AMIC</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AMIST">
        <xs:annotation>
          <xs:documentation>AMIST</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.AMIST</lx:path>
            <lx:label>AMIST</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AMP">
        <xs:annotation>
          <xs:documentation>AMP</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.AMP</lx:path>
            <lx:label>AMP</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AMS">
        <xs:annotation>
          <xs:documentation>AMS</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.AMS</lx:path>
            <lx:label>AMS</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ANZ Bank">
        <xs:annotation>
          <xs:documentation>ANZ Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.ANZ Bank</lx:path>
            <lx:label>ANZ Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ANZ Wealth">
        <xs:annotation>
          <xs:documentation>ANZ Wealth</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.ANZ Wealth</lx:path>
            <lx:label>ANZ Wealth</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Aon Master Trust">
        <xs:annotation>
          <xs:documentation>Aon Master Trust</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Aon Master Trust</lx:path>
            <lx:label>Aon Master Trust</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="API">
        <xs:annotation>
          <xs:documentation>API</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.API</lx:path>
            <lx:label>API</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Arab Bank of Australia">
        <xs:annotation>
          <xs:documentation>Arab Bank of Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Arab Bank of Australia</lx:path>
            <lx:label>Arab Bank of Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Armidale Building Society">
        <xs:annotation>
          <xs:documentation>Armidale Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Armidale Building Society</lx:path>
            <lx:label>Armidale Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Asahi Bank">
        <xs:annotation>
          <xs:documentation>Asahi Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Asahi Bank</lx:path>
            <lx:label>Asahi Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ASB Bank">
        <xs:annotation>
          <xs:documentation>ASB Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.ASB Bank</lx:path>
            <lx:label>ASB Bank</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Athena Home Loans">
        <xs:annotation>
          <xs:documentation>Athena Home Loans</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Athena Home Loans</lx:path>
            <lx:label>Athena Home Loans</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Aussie Home Loans">
        <xs:annotation>
          <xs:documentation>Aussie Home Loans</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Aussie Home Loans</lx:path>
            <lx:label>Aussie Home Loans</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Aussie Mastercard">
        <xs:annotation>
          <xs:documentation>Aussie Mastercard</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Aussie Mastercard</lx:path>
            <lx:label>Aussie Mastercard</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Aussie Mortgage Ltd">
        <xs:annotation>
          <xs:documentation>Aussie Mortgage Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Aussie Mortgage Ltd</lx:path>
            <lx:label>Aussie Mortgage Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Aust Alliance">
        <xs:annotation>
          <xs:documentation>Aust Alliance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Aust Alliance</lx:path>
            <lx:label>Aust Alliance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australia Central Credit Union">
        <xs:annotation>
          <xs:documentation>Australia Central Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Australia Central Credit Union</lx:path>
            <lx:label>Australia Central Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian Catholic">
        <xs:annotation>
          <xs:documentation>Australian Catholic</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Australian Catholic</lx:path>
            <lx:label>Australian Catholic</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian Defence Credit Union">
        <xs:annotation>
          <xs:documentation>Australian Defence Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Australian Defence Credit Union</lx:path>
            <lx:label>Australian Defence Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian Eagle">
        <xs:annotation>
          <xs:documentation>Australian Eagle</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Australian Eagle</lx:path>
            <lx:label>Australian Eagle</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian Ethical">
        <xs:annotation>
          <xs:documentation>Australian Ethical</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Australian Ethical</lx:path>
            <lx:label>Australian Ethical</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian First Mortgage">
        <xs:annotation>
          <xs:documentation>Australian First Mortgage</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Australian First Mortgage</lx:path>
            <lx:label>Australian First Mortgage</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian Pensioners Insurance">
        <xs:annotation>
          <xs:documentation>Australian Pensioners Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Australian Pensioners Insurance</lx:path>
            <lx:label>Australian Pensioners Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian Unity Building Society">
        <xs:annotation>
          <xs:documentation>Australian Unity Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Australian Unity Building Society</lx:path>
            <lx:label>Australian Unity Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AustralianSuper">
        <xs:annotation>
          <xs:documentation>AustralianSuper</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.AustralianSuper</lx:path>
            <lx:label>AustralianSuper</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AustSafe">
        <xs:annotation>
          <xs:documentation>AustSafe</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.AustSafe</lx:path>
            <lx:label>AustSafe</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Auswide Bank">
        <xs:annotation>
          <xs:documentation>Auswide Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Auswide Bank</lx:path>
            <lx:label>Auswide Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AV Super">
        <xs:annotation>
          <xs:documentation>AV Super</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.AV Super</lx:path>
            <lx:label>AV Super</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AVCO">
        <xs:annotation>
          <xs:documentation>AVCO</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.AVCO</lx:path>
            <lx:label>AVCO</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AXA">
        <xs:annotation>
          <xs:documentation>AXA</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.AXA</lx:path>
            <lx:label>AXA</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bank Australia">
        <xs:annotation>
          <xs:documentation>Bank Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bank Australia</lx:path>
            <lx:label>Bank Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bank of America">
        <xs:annotation>
          <xs:documentation>Bank of America</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bank of America</lx:path>
            <lx:label>Bank of America</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bank of Baroda">
        <xs:annotation>
          <xs:documentation>Bank of Baroda</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bank of Baroda</lx:path>
            <lx:label>Bank of Baroda</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bank of China">
        <xs:annotation>
          <xs:documentation>Bank of China</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bank of China</lx:path>
            <lx:label>Bank of China</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bank of India">
        <xs:annotation>
          <xs:documentation>Bank of India</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bank of India</lx:path>
            <lx:label>Bank of India</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bank of Melbourne">
        <xs:annotation>
          <xs:documentation>Bank of Melbourne</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bank of Melbourne</lx:path>
            <lx:label>Bank of Melbourne</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bank of New Zealand">
        <xs:annotation>
          <xs:documentation>Bank of New Zealand</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bank of New Zealand</lx:path>
            <lx:label>Bank of New Zealand</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bank of Queensland">
        <xs:annotation>
          <xs:documentation>Bank of Queensland</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bank of Queensland</lx:path>
            <lx:label>Bank of Queensland</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bank of Tokyo-Mitsubishi">
        <xs:annotation>
          <xs:documentation>Bank of Tokyo-Mitsubishi</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bank of Tokyo-Mitsubishi</lx:path>
            <lx:label>Bank of Tokyo-Mitsubishi</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bank SA">
        <xs:annotation>
          <xs:documentation>Bank SA</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bank SA</lx:path>
            <lx:label>Bank SA</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bankers Trust Australia">
        <xs:annotation>
          <xs:documentation>Bankers Trust Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bankers Trust Australia</lx:path>
            <lx:label>Bankers Trust Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bankwest">
        <xs:annotation>
          <xs:documentation>Bankwest</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bankwest</lx:path>
            <lx:label>Bankwest</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Banque Nationale De Paris">
        <xs:annotation>
          <xs:documentation>Banque Nationale De Paris</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Banque Nationale De Paris</lx:path>
            <lx:label>Banque Nationale De Paris</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Barclays">
        <xs:annotation>
          <xs:documentation>Barclays</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Barclays</lx:path>
            <lx:label>Barclays</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Barnes Home Loans">
        <xs:annotation>
          <xs:documentation>Barnes Home Loans</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Barnes Home Loans</lx:path>
            <lx:label>Barnes Home Loans</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bass and Equitable Building Society">
        <xs:annotation>
          <xs:documentation>Bass and Equitable Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bass and Equitable Building Society</lx:path>
            <lx:label>Bass and Equitable Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bendigo Bank">
        <xs:annotation>
          <xs:documentation>Bendigo Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bendigo Bank</lx:path>
            <lx:label>Bendigo Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Better Choice">
        <xs:annotation>
          <xs:documentation>Better Choice</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Better Choice</lx:path>
            <lx:label>Better Choice</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Better Mortgage Corporation">
        <xs:annotation>
          <xs:documentation>Better Mortgage Corporation</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Better Mortgage Corporation</lx:path>
            <lx:label>Better Mortgage Corporation</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Better Mortgage Mgmt">
        <xs:annotation>
          <xs:documentation>Better Mortgage Mgmt</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Better Mortgage Mgmt</lx:path>
            <lx:label>Better Mortgage Mgmt</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Beyond Bank Australia">
        <xs:annotation>
          <xs:documentation>Beyond Bank Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Beyond Bank Australia</lx:path>
            <lx:label>Beyond Bank Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bluebay Home Loans">
        <xs:annotation>
          <xs:documentation>Bluebay Home Loans</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bluebay Home Loans</lx:path>
            <lx:label>Bluebay Home Loans</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bluestone">
        <xs:annotation>
          <xs:documentation>Bluestone</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Bluestone</lx:path>
            <lx:label>Bluestone</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BMW Australia">
        <xs:annotation>
          <xs:documentation>BMW Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.BMW Australia</lx:path>
            <lx:label>BMW Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BOQ">
        <xs:annotation>
          <xs:documentation>BOQ</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.BOQ</lx:path>
            <lx:label>BOQ</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Brighten Home Loans">
        <xs:annotation>
          <xs:documentation>Brighten Home Loans</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Brighten Home Loans</lx:path>
            <lx:label>Brighten Home Loans</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Broken Hill Community Credit Union">
        <xs:annotation>
          <xs:documentation>Broken Hill Community Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Broken Hill Community Credit Union</lx:path>
            <lx:label>Broken Hill Community Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Building Society MP">
        <xs:annotation>
          <xs:documentation>Building Society MP</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Building Society MP</lx:path>
            <lx:label>Building Society MP</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BUSSQ">
        <xs:annotation>
          <xs:documentation>BUSSQ</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.BUSSQ</lx:path>
            <lx:label>BUSSQ</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Buyers Edge">
        <xs:annotation>
          <xs:documentation>Buyers Edge</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Buyers Edge</lx:path>
            <lx:label>Buyers Edge</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Capital Finance">
        <xs:annotation>
          <xs:documentation>Capital Finance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Capital Finance</lx:path>
            <lx:label>Capital Finance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Capricornia Credit Union">
        <xs:annotation>
          <xs:documentation>Capricornia Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Capricornia Credit Union</lx:path>
            <lx:label>Capricornia Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CareSuper">
        <xs:annotation>
          <xs:documentation>CareSuper</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.CareSuper</lx:path>
            <lx:label>CareSuper</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cash Doctors">
        <xs:annotation>
          <xs:documentation>Cash Doctors</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Cash Doctors</lx:path>
            <lx:label>Cash Doctors</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Catholic Church">
        <xs:annotation>
          <xs:documentation>Catholic Church</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Catholic Church</lx:path>
            <lx:label>Catholic Church</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Catholic Super">
        <xs:annotation>
          <xs:documentation>Catholic Super</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Catholic Super</lx:path>
            <lx:label>Catholic Super</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CBA">
        <xs:annotation>
          <xs:documentation>CBA</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.CBA</lx:path>
            <lx:label>CBA</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CBFC">
        <xs:annotation>
          <xs:documentation>CBFC</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.CBFC</lx:path>
            <lx:label>CBFC</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CBUS">
        <xs:annotation>
          <xs:documentation>CBUS</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.CBUS</lx:path>
            <lx:label>CBUS</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Certegy">
        <xs:annotation>
          <xs:documentation>Certegy</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Certegy</lx:path>
            <lx:label>Certegy</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CGU">
        <xs:annotation>
          <xs:documentation>CGU</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.CGU</lx:path>
            <lx:label>CGU</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Challenge Bank">
        <xs:annotation>
          <xs:documentation>Challenge Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Challenge Bank</lx:path>
            <lx:label>Challenge Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Chase Manhattan Bank">
        <xs:annotation>
          <xs:documentation>Chase Manhattan Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Chase Manhattan Bank</lx:path>
            <lx:label>Chase Manhattan Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Chequecash Pty Ltd">
        <xs:annotation>
          <xs:documentation>Chequecash Pty Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Chequecash Pty Ltd</lx:path>
            <lx:label>Chequecash Pty Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="China Construction Bank">
        <xs:annotation>
          <xs:documentation>China Construction Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.China Construction Bank</lx:path>
            <lx:label>China Construction Bank</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ChoiceLend">
        <xs:annotation>
          <xs:documentation>ChoiceLend</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.ChoiceLend</lx:path>
            <lx:label>ChoiceLend</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Christian Super">
        <xs:annotation>
          <xs:documentation>Christian Super</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Christian Super</lx:path>
            <lx:label>Christian Super</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CHU">
        <xs:annotation>
          <xs:documentation>CHU</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.CHU</lx:path>
            <lx:label>CHU</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CIC">
        <xs:annotation>
          <xs:documentation>CIC</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.CIC</lx:path>
            <lx:label>CIC</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CIL">
        <xs:annotation>
          <xs:documentation>CIL</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.CIL</lx:path>
            <lx:label>CIL</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Circle Credit Union">
        <xs:annotation>
          <xs:documentation>Circle Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Circle Credit Union</lx:path>
            <lx:label>Circle Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CIT Credit Pty">
        <xs:annotation>
          <xs:documentation>CIT Credit Pty</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.CIT Credit Pty</lx:path>
            <lx:label>CIT Credit Pty</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Citibank">
        <xs:annotation>
          <xs:documentation>Citibank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Citibank</lx:path>
            <lx:label>Citibank</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="City Coast Credit Union">
        <xs:annotation>
          <xs:documentation>City Coast Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.City Coast Credit Union</lx:path>
            <lx:label>City Coast Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Citywide Home Loans">
        <xs:annotation>
          <xs:documentation>Citywide Home Loans</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Citywide Home Loans</lx:path>
            <lx:label>Citywide Home Loans</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Club Plus Super">
        <xs:annotation>
          <xs:documentation>Club Plus Super</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Club Plus Super</lx:path>
            <lx:label>Club Plus Super</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Co Operative Ins Co">
        <xs:annotation>
          <xs:documentation>Co Operative Ins Co</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Co Operative Ins Co</lx:path>
            <lx:label>Co Operative Ins Co</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Coastline Credit Union">
        <xs:annotation>
          <xs:documentation>Coastline Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Coastline Credit Union</lx:path>
            <lx:label>Coastline Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Colonial">
        <xs:annotation>
          <xs:documentation>Colonial</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Colonial</lx:path>
            <lx:label>Colonial</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Colonial First State">
        <xs:annotation>
          <xs:documentation>Colonial First State</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Colonial First State</lx:path>
            <lx:label>Colonial First State</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Colonial Mutual">
        <xs:annotation>
          <xs:documentation>Colonial Mutual</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Colonial Mutual</lx:path>
            <lx:label>Colonial Mutual</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Columbus Capital Pty Ltd">
        <xs:annotation>
          <xs:documentation>Columbus Capital Pty Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Columbus Capital Pty Ltd</lx:path>
            <lx:label>Columbus Capital Pty Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CommBank Financial Planner">
        <xs:annotation>
          <xs:documentation>CommBank Financial Planner</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.CommBank Financial Planner</lx:path>
            <lx:label>CommBank Financial Planner</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Commercial Union">
        <xs:annotation>
          <xs:documentation>Commercial Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Commercial Union</lx:path>
            <lx:label>Commercial Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CommInsure">
        <xs:annotation>
          <xs:documentation>CommInsure</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.CommInsure</lx:path>
            <lx:label>CommInsure</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Commonwealth Bank">
        <xs:annotation>
          <xs:documentation>Commonwealth Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Commonwealth Bank</lx:path>
            <lx:label>Commonwealth Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Connective Home Loans">
        <xs:annotation>
          <xs:documentation>Connective Home Loans</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Connective Home Loans</lx:path>
            <lx:label>Connective Home Loans</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CPS Credit Union">
        <xs:annotation>
          <xs:documentation>CPS Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.CPS Credit Union</lx:path>
            <lx:label>CPS Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Corp">
        <xs:annotation>
          <xs:documentation>Credit Corp</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Credit Corp</lx:path>
            <lx:label>Credit Corp</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Union Australia">
        <xs:annotation>
          <xs:documentation>Credit Union Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Credit Union Australia</lx:path>
            <lx:label>Credit Union Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Union Financial Services">
        <xs:annotation>
          <xs:documentation>Credit Union Financial Services</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Credit Union Financial Services</lx:path>
            <lx:label>Credit Union Financial Services</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Union SA">
        <xs:annotation>
          <xs:documentation>Credit Union SA</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Credit Union SA</lx:path>
            <lx:label>Credit Union SA</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Union Settlement Services">
        <xs:annotation>
          <xs:documentation>Credit Union Settlement Services</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Credit Union Settlement Services</lx:path>
            <lx:label>Credit Union Settlement Services</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CUA">
        <xs:annotation>
          <xs:documentation>CUA</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.CUA</lx:path>
            <lx:label>CUA</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cumis">
        <xs:annotation>
          <xs:documentation>Cumis</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Cumis</lx:path>
            <lx:label>Cumis</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cuscal">
        <xs:annotation>
          <xs:documentation>Cuscal</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Cuscal</lx:path>
            <lx:label>Cuscal</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Custom Credit">
        <xs:annotation>
          <xs:documentation>Custom Credit</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Custom Credit</lx:path>
            <lx:label>Custom Credit</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dai-Ichi Kango Bank">
        <xs:annotation>
          <xs:documentation>Dai-Ichi Kango Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Dai-Ichi Kango Bank</lx:path>
            <lx:label>Dai-Ichi Kango Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Daimlerchrysler Finance">
        <xs:annotation>
          <xs:documentation>Daimlerchrysler Finance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Daimlerchrysler Finance</lx:path>
            <lx:label>Daimlerchrysler Finance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Debt Purchase Australia">
        <xs:annotation>
          <xs:documentation>Debt Purchase Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Debt Purchase Australia</lx:path>
            <lx:label>Debt Purchase Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Defence Bank">
        <xs:annotation>
          <xs:documentation>Defence Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Defence Bank</lx:path>
            <lx:label>Defence Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Defence Credit">
        <xs:annotation>
          <xs:documentation>Defence Credit</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Defence Credit</lx:path>
            <lx:label>Defence Credit</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Defence Force Credit Union Limited">
        <xs:annotation>
          <xs:documentation>Item to be deprecated since Defence Force Credit Union Limited changed name to Defence Bank in 2012.</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Defence Force Credit Union Limited</lx:path>
            <lx:label>Defence Force Credit Union Limited</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Defence Services Homes">
        <xs:annotation>
          <xs:documentation>Defence Services Homes</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Defence Services Homes</lx:path>
            <lx:label>Defence Services Homes</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Deutsche Bank">
        <xs:annotation>
          <xs:documentation>Deutsche Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Deutsche Bank</lx:path>
            <lx:label>Deutsche Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Diners Club">
        <xs:annotation>
          <xs:documentation>Diners Club</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Diners Club</lx:path>
            <lx:label>Diners Club</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dnister Ukrainian Credit Co-operative">
        <xs:annotation>
          <xs:documentation>Dnister Ukrainian Credit Co-operative</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Dnister Ukrainian Credit Co-operative</lx:path>
            <lx:label>Dnister Ukrainian Credit Co-operative</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="EECU">
        <xs:annotation>
          <xs:documentation>EECU</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.EECU</lx:path>
            <lx:label>EECU</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="EISS">
        <xs:annotation>
          <xs:documentation>EISS</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.EISS</lx:path>
            <lx:label>EISS</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ELCOM Credit Union">
        <xs:annotation>
          <xs:documentation>ELCOM Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.ELCOM Credit Union</lx:path>
            <lx:label>ELCOM Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Elders">
        <xs:annotation>
          <xs:documentation>Elders</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Elders</lx:path>
            <lx:label>Elders</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="EMoney Pty Ltd">
        <xs:annotation>
          <xs:documentation>EMoney Pty Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.EMoney Pty Ltd</lx:path>
            <lx:label>EMoney Pty Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Energy Super">
        <xs:annotation>
          <xs:documentation>Energy Super</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Energy Super</lx:path>
            <lx:label>Energy Super</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Equity Super">
        <xs:annotation>
          <xs:documentation>Equity Super</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Equity Super</lx:path>
            <lx:label>Equity Super</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Esanda">
        <xs:annotation>
          <xs:documentation>Esanda</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Esanda</lx:path>
            <lx:label>Esanda</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Essential Super">
        <xs:annotation>
          <xs:documentation>Essential Super</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Essential Super</lx:path>
            <lx:label>Essential Super</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ezy Bank">
        <xs:annotation>
          <xs:documentation>Ezy Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Ezy Bank</lx:path>
            <lx:label>Ezy Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FAI">
        <xs:annotation>
          <xs:documentation>FAI</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.FAI</lx:path>
            <lx:label>FAI</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Faraday West and Co Pty Ltd">
        <xs:annotation>
          <xs:documentation>Faraday West and Co Pty Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Faraday West and Co Pty Ltd</lx:path>
            <lx:label>Faraday West and Co Pty Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FASTLend">
        <xs:annotation>
          <xs:documentation>FASTLend</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.FASTLend</lx:path>
            <lx:label>FASTLend</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Federation">
        <xs:annotation>
          <xs:documentation>Federation</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Federation</lx:path>
            <lx:label>Federation</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FinancePath">
        <xs:annotation>
          <xs:documentation>FinancePath</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.FinancePath</lx:path>
            <lx:label>FinancePath</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fire Brigades Credit Union">
        <xs:annotation>
          <xs:documentation>Fire Brigades Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Fire Brigades Credit Union</lx:path>
            <lx:label>Fire Brigades Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fire Service Credit Union">
        <xs:annotation>
          <xs:documentation>Fire Service Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Fire Service Credit Union</lx:path>
            <lx:label>Fire Service Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="First Choice Credit Union">
        <xs:annotation>
          <xs:documentation>First Choice Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.First Choice Credit Union</lx:path>
            <lx:label>First Choice Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="First National Bank of Chicago">
        <xs:annotation>
          <xs:documentation>First National Bank of Chicago</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.First National Bank of Chicago</lx:path>
            <lx:label>First National Bank of Chicago</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="First Permanent Home Loans">
        <xs:annotation>
          <xs:documentation>First Permanent Home Loans</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.First Permanent Home Loans</lx:path>
            <lx:label>First Permanent Home Loans</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="First Provincial Building Society">
        <xs:annotation>
          <xs:documentation>First Provincial Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.First Provincial Building Society</lx:path>
            <lx:label>First Provincial Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="First State Super">
        <xs:annotation>
          <xs:documentation>First State Super</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.First State Super</lx:path>
            <lx:label>First State Super</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="First Super">
        <xs:annotation>
          <xs:documentation>First Super</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.First Super</lx:path>
            <lx:label>First Super</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Firstmac">
        <xs:annotation>
          <xs:documentation>Firstmac</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Firstmac</lx:path>
            <lx:label>Firstmac</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Flexirent">
        <xs:annotation>
          <xs:documentation>Flexirent</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Flexirent</lx:path>
            <lx:label>Flexirent</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ford Credit">
        <xs:annotation>
          <xs:documentation>Ford Credit</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Ford Credit</lx:path>
            <lx:label>Ford Credit</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fortis">
        <xs:annotation>
          <xs:documentation>Fortis</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Fortis</lx:path>
            <lx:label>Fortis</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fundcorp">
        <xs:annotation>
          <xs:documentation>Fundcorp</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Fundcorp</lx:path>
            <lx:label>Fundcorp</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Future Financial">
        <xs:annotation>
          <xs:documentation>Future Financial</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Future Financial</lx:path>
            <lx:label>Future Financial</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gateway Credit Union">
        <xs:annotation>
          <xs:documentation>Gateway Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Gateway Credit Union</lx:path>
            <lx:label>Gateway Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GE Automotive">
        <xs:annotation>
          <xs:documentation>GE Automotive</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.GE Automotive</lx:path>
            <lx:label>GE Automotive</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GE Capital Finance">
        <xs:annotation>
          <xs:documentation>GE Capital Finance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.GE Capital Finance</lx:path>
            <lx:label>GE Capital Finance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GE Capital Finance (Go Mastercard)">
        <xs:annotation>
          <xs:documentation>GE Capital Finance (Go Mastercard)</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.GE Capital Finance (Go Mastercard)</lx:path>
            <lx:label>GE Capital Finance (Go Mastercard)</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GE Credit Line">
        <xs:annotation>
          <xs:documentation>GE Credit Line</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.GE Credit Line</lx:path>
            <lx:label>GE Credit Line</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ge Finance">
        <xs:annotation>
          <xs:documentation>Ge Finance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Ge Finance</lx:path>
            <lx:label>Ge Finance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GE Money Aust Pty Ltd">
        <xs:annotation>
          <xs:documentation>GE Money Aust Pty Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.GE Money Aust Pty Ltd</lx:path>
            <lx:label>GE Money Aust Pty Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GE Mortgages">
        <xs:annotation>
          <xs:documentation>GE Mortgages</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.GE Mortgages</lx:path>
            <lx:label>GE Mortgages</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Genworth">
        <xs:annotation>
          <xs:documentation>Genworth</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Genworth</lx:path>
            <lx:label>Genworth</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GIO">
        <xs:annotation>
          <xs:documentation>GIO</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.GIO</lx:path>
            <lx:label>GIO</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GIO Building Society">
        <xs:annotation>
          <xs:documentation>GIO Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.GIO Building Society</lx:path>
            <lx:label>GIO Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GMAC Finance">
        <xs:annotation>
          <xs:documentation>GMAC Finance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.GMAC Finance</lx:path>
            <lx:label>GMAC Finance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Grace Bros">
        <xs:annotation>
          <xs:documentation>Grace Bros</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Grace Bros</lx:path>
            <lx:label>Grace Bros</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GRE">
        <xs:annotation>
          <xs:documentation>GRE</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.GRE</lx:path>
            <lx:label>GRE</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Great Pacific Financial">
        <xs:annotation>
          <xs:documentation>Great Pacific Financial</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Great Pacific Financial</lx:path>
            <lx:label>Great Pacific Financial</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Greater Building Society">
        <xs:annotation>
          <xs:documentation>Greater Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Greater Building Society</lx:path>
            <lx:label>Greater Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Guild">
        <xs:annotation>
          <xs:documentation>Guild</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Guild</lx:path>
            <lx:label>Guild</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hanover Consumer Finance">
        <xs:annotation>
          <xs:documentation>Hanover Consumer Finance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Hanover Consumer Finance</lx:path>
            <lx:label>Hanover Consumer Finance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HBF">
        <xs:annotation>
          <xs:documentation>HBF</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.HBF</lx:path>
            <lx:label>HBF</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Heartland Bank">
        <xs:annotation>
          <xs:documentation>Heartland Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Heartland Bank</lx:path>
            <lx:label>Heartland Bank</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Heritage">
        <xs:annotation>
          <xs:documentation>Heritage</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Heritage</lx:path>
            <lx:label>Heritage</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HESTA">
        <xs:annotation>
          <xs:documentation>HESTA</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.HESTA</lx:path>
            <lx:label>HESTA</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Home Building Society">
        <xs:annotation>
          <xs:documentation>Home Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Home Building Society</lx:path>
            <lx:label>Home Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Homeloans Ltd">
        <xs:annotation>
          <xs:documentation>Homeloans Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Homeloans Ltd</lx:path>
            <lx:label>Homeloans Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Homeside">
        <xs:annotation>
          <xs:documentation>Homeside</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Homeside</lx:path>
            <lx:label>Homeside</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Homestart Finance">
        <xs:annotation>
          <xs:documentation>Homestart Finance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Homestart Finance</lx:path>
            <lx:label>Homestart Finance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Homeswest">
        <xs:annotation>
          <xs:documentation>Homeswest</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Homeswest</lx:path>
            <lx:label>Homeswest</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HSBC">
        <xs:annotation>
          <xs:documentation>HSBC</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.HSBC</lx:path>
            <lx:label>HSBC</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hume Building Society">
        <xs:annotation>
          <xs:documentation>Hume Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Hume Building Society</lx:path>
            <lx:label>Hume Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IBJ Australia Bank">
        <xs:annotation>
          <xs:documentation>IBJ Australia Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.IBJ Australia Bank</lx:path>
            <lx:label>IBJ Australia Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Iden">
        <xs:annotation>
          <xs:documentation>Iden</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Iden</lx:path>
            <lx:label>Iden</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ignite">
        <xs:annotation>
          <xs:documentation>Ignite</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Ignite</lx:path>
            <lx:label>Ignite</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Illawarra Building Society">
        <xs:annotation>
          <xs:documentation>Illawarra Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Illawarra Building Society</lx:path>
            <lx:label>Illawarra Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IMB Ltd">
        <xs:annotation>
          <xs:documentation>IMB Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.IMB Ltd</lx:path>
            <lx:label>IMB Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="INC Invest">
        <xs:annotation>
          <xs:documentation>INC Invest</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.INC Invest</lx:path>
            <lx:label>INC Invest</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ING">
        <xs:annotation>
          <xs:documentation>ING</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.ING</lx:path>
            <lx:label>ING</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ING Mercantile Mutual Bank">
        <xs:annotation>
          <xs:documentation>ING Mercantile Mutual Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.ING Mercantile Mutual Bank</lx:path>
            <lx:label>ING Mercantile Mutual Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Intech Credit Union">
        <xs:annotation>
          <xs:documentation>Intech Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Intech Credit Union</lx:path>
            <lx:label>Intech Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="International Acceptance">
        <xs:annotation>
          <xs:documentation>International Acceptance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.International Acceptance</lx:path>
            <lx:label>International Acceptance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="International Commercial Bank of China">
        <xs:annotation>
          <xs:documentation>International Commercial Bank of China</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.International Commercial Bank of China</lx:path>
            <lx:label>International Commercial Bank of China</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="InTrust">
        <xs:annotation>
          <xs:documentation>InTrust</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.InTrust</lx:path>
            <lx:label>InTrust</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Investec">
        <xs:annotation>
          <xs:documentation>Investec</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Investec</lx:path>
            <lx:label>Investec</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IOOF Building Society">
        <xs:annotation>
          <xs:documentation>IOOF Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.IOOF Building Society</lx:path>
            <lx:label>IOOF Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="JPMorgan Chase Bank">
        <xs:annotation>
          <xs:documentation>JPMorgan Chase Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.JPMorgan Chase Bank</lx:path>
            <lx:label>JPMorgan Chase Bank</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Keystart Home Loans">
        <xs:annotation>
          <xs:documentation>Keystart Home Loans</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Keystart Home Loans</lx:path>
            <lx:label>Keystart Home Loans</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Kinetic">
        <xs:annotation>
          <xs:documentation>Kinetic</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Kinetic</lx:path>
            <lx:label>Kinetic</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Kiwibank">
        <xs:annotation>
          <xs:documentation>Kiwibank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Kiwibank</lx:path>
            <lx:label>Kiwibank</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Kookmin Bank">
        <xs:annotation>
          <xs:documentation>Kookmin Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Kookmin Bank</lx:path>
            <lx:label>Kookmin Bank</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Kwik Kash International">
        <xs:annotation>
          <xs:documentation>Kwik Kash International</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Kwik Kash International</lx:path>
            <lx:label>Kwik Kash International</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LaTrobe">
        <xs:annotation>
          <xs:documentation>LaTrobe</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.LaTrobe</lx:path>
            <lx:label>LaTrobe</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Legal Super">
        <xs:annotation>
          <xs:documentation>legalsuper</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Legal Super</lx:path>
            <lx:label>Legal Super</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Liberty">
        <xs:annotation>
          <xs:documentation>Liberty</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Liberty</lx:path>
            <lx:label>Liberty</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lion Finance">
        <xs:annotation>
          <xs:documentation>Lion Finance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Lion Finance</lx:path>
            <lx:label>Lion Finance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loanave">
        <xs:annotation>
          <xs:documentation>Loanave</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Loanave</lx:path>
            <lx:label>Loanave</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Local Government">
        <xs:annotation>
          <xs:documentation>Local Government</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Local Government</lx:path>
            <lx:label>Local Government</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lombard Finance">
        <xs:annotation>
          <xs:documentation>Lombard Finance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Lombard Finance</lx:path>
            <lx:label>Lombard Finance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LUCRF">
        <xs:annotation>
          <xs:documentation>LUCRF</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.LUCRF</lx:path>
            <lx:label>LUCRF</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="M Capitol">
        <xs:annotation>
          <xs:documentation>M Capitol</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.M Capitol</lx:path>
            <lx:label>M Capitol</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Macarthur Credit Union">
        <xs:annotation>
          <xs:documentation>Macarthur Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Macarthur Credit Union</lx:path>
            <lx:label>Macarthur Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mackay Permanent Building Society">
        <xs:annotation>
          <xs:documentation>Mackay Permanent Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Mackay Permanent Building Society</lx:path>
            <lx:label>Mackay Permanent Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Macquarie">
        <xs:annotation>
          <xs:documentation>Macquarie</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Macquarie</lx:path>
            <lx:label>Macquarie</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Maitland Mutual Building Society">
        <xs:annotation>
          <xs:documentation>Maitland Mutual Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Maitland Mutual Building Society</lx:path>
            <lx:label>Maitland Mutual Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Manufacture Mutual">
        <xs:annotation>
          <xs:documentation>Manufacture Mutual</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Manufacture Mutual</lx:path>
            <lx:label>Manufacture Mutual</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MAP">
        <xs:annotation>
          <xs:documentation>MAP</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.MAP</lx:path>
            <lx:label>MAP</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ME Bank">
        <xs:annotation>
          <xs:documentation>ME Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.ME Bank</lx:path>
            <lx:label>ME Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Media Super">
        <xs:annotation>
          <xs:documentation>Media Super</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Media Super</lx:path>
            <lx:label>Media Super</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mediplan Australia">
        <xs:annotation>
          <xs:documentation>Mediplan Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Mediplan Australia</lx:path>
            <lx:label>Mediplan Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Members Equity Bank">
        <xs:annotation>
          <xs:documentation>Members Equity Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Members Equity Bank</lx:path>
            <lx:label>Members Equity Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mercantile Mutual">
        <xs:annotation>
          <xs:documentation>Mercantile Mutual</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Mercantile Mutual</lx:path>
            <lx:label>Mercantile Mutual</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mercer Wealth">
        <xs:annotation>
          <xs:documentation>Mercer Wealth</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Mercer Wealth</lx:path>
            <lx:label>Mercer Wealth</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Midland Bank PLC">
        <xs:annotation>
          <xs:documentation>Midland Bank PLC</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Midland Bank PLC</lx:path>
            <lx:label>Midland Bank PLC</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MIIP">
        <xs:annotation>
          <xs:documentation>MIIP</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.MIIP</lx:path>
            <lx:label>MIIP</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mine Wealth and Wellbeing">
        <xs:annotation>
          <xs:documentation>Mine Wealth and Wellbeing</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Mine Wealth and Wellbeing</lx:path>
            <lx:label>Mine Wealth and Wellbeing</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MIS">
        <xs:annotation>
          <xs:documentation>MIS</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.MIS</lx:path>
            <lx:label>MIS</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mitchell Building Society">
        <xs:annotation>
          <xs:documentation>Mitchell Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Mitchell Building Society</lx:path>
            <lx:label>Mitchell Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MKM Capital">
        <xs:annotation>
          <xs:documentation>MKM Capital</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.MKM Capital</lx:path>
            <lx:label>MKM Capital</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MLC">
        <xs:annotation>
          <xs:documentation>MLC</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.MLC</lx:path>
            <lx:label>MLC</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MMI">
        <xs:annotation>
          <xs:documentation>MMI</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.MMI</lx:path>
            <lx:label>MMI</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Money Quest Australia Pty Ltd">
        <xs:annotation>
          <xs:documentation>Money Quest Australia Pty Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Money Quest Australia Pty Ltd</lx:path>
            <lx:label>Money Quest Australia Pty Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Morgan Guaranty Trust Bank of New York">
        <xs:annotation>
          <xs:documentation>Morgan Guaranty Trust Bank of New York</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Morgan Guaranty Trust Bank of New York</lx:path>
            <lx:label>Morgan Guaranty Trust Bank of New York</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mortgage Ezy">
        <xs:annotation>
          <xs:documentation>Mortgage Ezy</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Mortgage Ezy</lx:path>
            <lx:label>Mortgage Ezy</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mortgage House">
        <xs:annotation>
          <xs:documentation>Mortgage House</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Mortgage House</lx:path>
            <lx:label>Mortgage House</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mortgage Mart">
        <xs:annotation>
          <xs:documentation>Mortgage Mart</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Mortgage Mart</lx:path>
            <lx:label>Mortgage Mart</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MOVE Bank">
        <xs:annotation>
          <xs:documentation>Move Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.MOVE Bank</lx:path>
            <lx:label>MOVE Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MTAA">
        <xs:annotation>
          <xs:documentation>MTAA</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.MTAA</lx:path>
            <lx:label>MTAA</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mutual Community">
        <xs:annotation>
          <xs:documentation>Mutual Community</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Mutual Community</lx:path>
            <lx:label>Mutual Community</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NAB">
        <xs:annotation>
          <xs:documentation>NAB</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.NAB</lx:path>
            <lx:label>NAB</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="National Mortgage Corporation">
        <xs:annotation>
          <xs:documentation>National Mortgage Corporation</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.National Mortgage Corporation</lx:path>
            <lx:label>National Mortgage Corporation</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Natwest Markets Australia">
        <xs:annotation>
          <xs:documentation>Natwest Markets Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Natwest Markets Australia</lx:path>
            <lx:label>Natwest Markets Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NCUIS">
        <xs:annotation>
          <xs:documentation>NCUIS</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.NCUIS</lx:path>
            <lx:label>NCUIS</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Newcastle Permanent Building Society">
        <xs:annotation>
          <xs:documentation>Newcastle Permanent Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Newcastle Permanent Building Society</lx:path>
            <lx:label>Newcastle Permanent Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NGS">
        <xs:annotation>
          <xs:documentation>NGS</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.NGS</lx:path>
            <lx:label>NGS</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Northern Building Society">
        <xs:annotation>
          <xs:documentation>Northern Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Northern Building Society</lx:path>
            <lx:label>Northern Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NRMA">
        <xs:annotation>
          <xs:documentation>NRMA</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.NRMA</lx:path>
            <lx:label>NRMA</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NSF Super">
        <xs:annotation>
          <xs:documentation>NSF Super</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.NSF Super</lx:path>
            <lx:label>NSF Super</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NSW Home Loans">
        <xs:annotation>
          <xs:documentation>NSW Home Loans</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.NSW Home Loans</lx:path>
            <lx:label>NSW Home Loans</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NSW Teachers Credit Union">
        <xs:annotation>
          <xs:documentation>NSW Teachers Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.NSW Teachers Credit Union</lx:path>
            <lx:label>NSW Teachers Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NZI">
        <xs:annotation>
          <xs:documentation>NZI</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.NZI</lx:path>
            <lx:label>NZI</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="OneAnswer">
        <xs:annotation>
          <xs:documentation>OneAnswer</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.OneAnswer</lx:path>
            <lx:label>OneAnswer</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ORIX">
        <xs:annotation>
          <xs:documentation>ORIX</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.ORIX</lx:path>
            <lx:label>ORIX</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Overseas-Chinese Banking Corporation">
        <xs:annotation>
          <xs:documentation>Overseas-Chinese Banking Corporation</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Overseas-Chinese Banking Corporation</lx:path>
            <lx:label>Overseas-Chinese Banking Corporation</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Paramount">
        <xs:annotation>
          <xs:documentation>Paramount</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Paramount</lx:path>
            <lx:label>Paramount</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Payday Central">
        <xs:annotation>
          <xs:documentation>Payday Central</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Payday Central</lx:path>
            <lx:label>Payday Central</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pensioners Insurance">
        <xs:annotation>
          <xs:documentation>Pensioners Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Pensioners Insurance</lx:path>
            <lx:label>Pensioners Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="People's Choice Credit Union">
        <xs:annotation>
          <xs:documentation>People's Choice Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.People's Choice Credit Union</lx:path>
            <lx:label>People's Choice Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pepper">
        <xs:annotation>
          <xs:documentation>Pepper</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Pepper</lx:path>
            <lx:label>Pepper</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pepper New Zealand">
        <xs:annotation>
          <xs:documentation>Pepper New Zealand</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Pepper New Zealand</lx:path>
            <lx:label>Pepper New Zealand</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Perpetual Limited">
        <xs:annotation>
          <xs:documentation>Perpetual Limited</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Perpetual Limited</lx:path>
            <lx:label>Perpetual Limited</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Perpetual Trustee">
        <xs:annotation>
          <xs:documentation>Perpetual Trustee</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Perpetual Trustee</lx:path>
            <lx:label>Perpetual Trustee</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pioneer Permanent Building Society">
        <xs:annotation>
          <xs:documentation>Pioneer Permanent Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Pioneer Permanent Building Society</lx:path>
            <lx:label>Pioneer Permanent Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PLANLend">
        <xs:annotation>
          <xs:documentation>PLANLend</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.PLANLend</lx:path>
            <lx:label>PLANLend</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Plum">
        <xs:annotation>
          <xs:documentation>Plum</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Plum</lx:path>
            <lx:label>Plum</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Police and Nurses">
        <xs:annotation>
          <xs:documentation>Police and Nurses</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Police and Nurses</lx:path>
            <lx:label>Police and Nurses</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Police Credit Union">
        <xs:annotation>
          <xs:documentation>Police Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Police Credit Union</lx:path>
            <lx:label>Police Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Police Credit Union (NSW)">
        <xs:annotation>
          <xs:documentation>Police Credit Union (NSW)</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Police Credit Union (NSW)</lx:path>
            <lx:label>Police Credit Union (NSW)</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Powerstate Credit Union">
        <xs:annotation>
          <xs:documentation>Powerstate Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Powerstate Credit Union</lx:path>
            <lx:label>Powerstate Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Primary Industry Bank of Australia">
        <xs:annotation>
          <xs:documentation>Primary Industry Bank of Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Primary Industry Bank of Australia</lx:path>
            <lx:label>Primary Industry Bank of Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Prime">
        <xs:annotation>
          <xs:documentation>Prime</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Prime</lx:path>
            <lx:label>Prime</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Prospect Credit Union">
        <xs:annotation>
          <xs:documentation>Prospect Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Prospect Credit Union</lx:path>
            <lx:label>Prospect Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Public Trustee">
        <xs:annotation>
          <xs:documentation>Public Trustee</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Public Trustee</lx:path>
            <lx:label>Public Trustee</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pulse Credit Union">
        <xs:annotation>
          <xs:documentation>Pulse Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Pulse Credit Union</lx:path>
            <lx:label>Pulse Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="QANTAS Credit Union">
        <xs:annotation>
          <xs:documentation>QANTAS Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.QANTAS Credit Union</lx:path>
            <lx:label>QANTAS Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="QBE">
        <xs:annotation>
          <xs:documentation>QBE</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.QBE</lx:path>
            <lx:label>QBE</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Qudos Bank">
        <xs:annotation>
          <xs:documentation>Qudos Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Qudos Bank</lx:path>
            <lx:label>Qudos Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Queensland Country Credit Union">
        <xs:annotation>
          <xs:documentation>Queensland Country Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Queensland Country Credit Union</lx:path>
            <lx:label>Queensland Country Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Queensland Credit Union">
        <xs:annotation>
          <xs:documentation>Queensland Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Queensland Credit Union</lx:path>
            <lx:label>Queensland Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Queensland Industry Development Corp">
        <xs:annotation>
          <xs:documentation>Queensland Industry Development Corp</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Queensland Industry Development Corp</lx:path>
            <lx:label>Queensland Industry Development Corp</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Queensland Professional Credit Union">
        <xs:annotation>
          <xs:documentation>Queensland Professional Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Queensland Professional Credit Union</lx:path>
            <lx:label>Queensland Professional Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Queensland Teachers Credit Union">
        <xs:annotation>
          <xs:documentation>Queensland Teachers Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Queensland Teachers Credit Union</lx:path>
            <lx:label>Queensland Teachers Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rabobank Nederland">
        <xs:annotation>
          <xs:documentation>Rabobank Nederland</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Rabobank Nederland</lx:path>
            <lx:label>Rabobank Nederland</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rabobank New Zealand">
        <xs:annotation>
          <xs:documentation>Rabobank New Zealand</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Rabobank New Zealand</lx:path>
            <lx:label>Rabobank New Zealand</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RAC">
        <xs:annotation>
          <xs:documentation>RAC</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.RAC</lx:path>
            <lx:label>RAC</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RAC WA">
        <xs:annotation>
          <xs:documentation>RAC WA</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.RAC WA</lx:path>
            <lx:label>RAC WA</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RACV">
        <xs:annotation>
          <xs:documentation>RACV</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.RACV</lx:path>
            <lx:label>RACV</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Radio Rentals">
        <xs:annotation>
          <xs:documentation>Radio Rentals</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Radio Rentals</lx:path>
            <lx:label>Radio Rentals</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Railways Credit Union">
        <xs:annotation>
          <xs:documentation>Railways Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Railways Credit Union</lx:path>
            <lx:label>Railways Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RAMS">
        <xs:annotation>
          <xs:documentation>RAMS</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.RAMS</lx:path>
            <lx:label>RAMS</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rate Money Pty Ltd">
        <xs:annotation>
          <xs:documentation>Rate Money Pty Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Rate Money Pty Ltd</lx:path>
            <lx:label>Rate Money Pty Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="REA Home Loans">
        <xs:annotation>
          <xs:documentation>REA Home Loans</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.REA Home Loans</lx:path>
            <lx:label>REA Home Loans</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RedZed">
        <xs:annotation>
          <xs:documentation>RedZed</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.RedZed</lx:path>
            <lx:label>RedZed</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Repcol">
        <xs:annotation>
          <xs:documentation>Repcol</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Repcol</lx:path>
            <lx:label>Repcol</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Reserve Bank of Australia">
        <xs:annotation>
          <xs:documentation>Reserve Bank of Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Reserve Bank of Australia</lx:path>
            <lx:label>Reserve Bank of Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Resi">
        <xs:annotation>
          <xs:documentation>Resi</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Resi</lx:path>
            <lx:label>Resi</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RESI Mortgage Corporation">
        <xs:annotation>
          <xs:documentation>RESI Mortgage Corporation</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.RESI Mortgage Corporation</lx:path>
            <lx:label>RESI Mortgage Corporation</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RESIMAC">
        <xs:annotation>
          <xs:documentation>RESIMAC</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.RESIMAC</lx:path>
            <lx:label>RESIMAC</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Royal Life">
        <xs:annotation>
          <xs:documentation>Royal Life</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Royal Life</lx:path>
            <lx:label>Royal Life</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Russell Investments">
        <xs:annotation>
          <xs:documentation>Russell Investments</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Russell Investments</lx:path>
            <lx:label>Russell Investments</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Safrock Finance">
        <xs:annotation>
          <xs:documentation>Safrock Finance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Safrock Finance</lx:path>
            <lx:label>Safrock Finance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sanwa Australia Finance Limited">
        <xs:annotation>
          <xs:documentation>Sanwa Australia Finance Limited</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Sanwa Australia Finance Limited</lx:path>
            <lx:label>Sanwa Australia Finance Limited</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Savings And Loans Credit Union">
        <xs:annotation>
          <xs:documentation>Savings And Loans Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Savings And Loans Credit Union</lx:path>
            <lx:label>Savings And Loans Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SGIC">
        <xs:annotation>
          <xs:documentation>SGIC</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.SGIC</lx:path>
            <lx:label>SGIC</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SGIO">
        <xs:annotation>
          <xs:documentation>SGIO</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.SGIO</lx:path>
            <lx:label>SGIO</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Shannons">
        <xs:annotation>
          <xs:documentation>Shannons</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Shannons</lx:path>
            <lx:label>Shannons</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SmartSave">
        <xs:annotation>
          <xs:documentation>SmartSave</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.SmartSave</lx:path>
            <lx:label>SmartSave</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Southern Cross Credit Union">
        <xs:annotation>
          <xs:documentation>Southern Cross Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Southern Cross Credit Union</lx:path>
            <lx:label>Southern Cross Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Southland Building Society">
        <xs:annotation>
          <xs:documentation>Southland Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Southland Building Society</lx:path>
            <lx:label>Southland Building Society</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sovereign Credit">
        <xs:annotation>
          <xs:documentation>Sovereign Credit</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Sovereign Credit</lx:path>
            <lx:label>Sovereign Credit</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="St Andrew">
        <xs:annotation>
          <xs:documentation>St Andrew</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.St Andrew</lx:path>
            <lx:label>St Andrew</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="St George">
        <xs:annotation>
          <xs:documentation>St George</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.St George</lx:path>
            <lx:label>St George</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Standard Chartered Bank Aust">
        <xs:annotation>
          <xs:documentation>Standard Chartered Bank Aust</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Standard Chartered Bank Aust</lx:path>
            <lx:label>Standard Chartered Bank Aust</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="State Street Bank And Trust Company">
        <xs:annotation>
          <xs:documentation>State Street Bank And Trust Company</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.State Street Bank And Trust Company</lx:path>
            <lx:label>State Street Bank And Trust Company</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Statewide">
        <xs:annotation>
          <xs:documentation>Statewide</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Statewide</lx:path>
            <lx:label>Statewide</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Stronghold">
        <xs:annotation>
          <xs:documentation>Stronghold</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Stronghold</lx:path>
            <lx:label>Stronghold</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Summerland Credit Union">
        <xs:annotation>
          <xs:documentation>Summerland Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Summerland Credit Union</lx:path>
            <lx:label>Summerland Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sun Alliance">
        <xs:annotation>
          <xs:documentation>Sun Alliance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Sun Alliance</lx:path>
            <lx:label>Sun Alliance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Suncorp">
        <xs:annotation>
          <xs:documentation>Suncorp</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Suncorp</lx:path>
            <lx:label>Suncorp</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Suncorp Metway">
        <xs:annotation>
          <xs:documentation>Suncorp Metway</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Suncorp Metway</lx:path>
            <lx:label>Suncorp Metway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sunsuper">
        <xs:annotation>
          <xs:documentation>Sunsuper</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Sunsuper</lx:path>
            <lx:label>Sunsuper</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Swann">
        <xs:annotation>
          <xs:documentation>Swann</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Swann</lx:path>
            <lx:label>Swann</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TasPlan">
        <xs:annotation>
          <xs:documentation>TasPlan</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.TasPlan</lx:path>
            <lx:label>TasPlan</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Teachers Credit Union">
        <xs:annotation>
          <xs:documentation>Teachers Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Teachers Credit Union</lx:path>
            <lx:label>Teachers Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Teachers Mutual">
        <xs:annotation>
          <xs:documentation>Teachers Mutual</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Teachers Mutual</lx:path>
            <lx:label>Teachers Mutual</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Territory Mutual Building Society">
        <xs:annotation>
          <xs:documentation>Territory Mutual Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Territory Mutual Building Society</lx:path>
            <lx:label>Territory Mutual Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="The Co-operative Bank">
        <xs:annotation>
          <xs:documentation>The Co-operative Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.The Co-operative Bank</lx:path>
            <lx:label>The Co-operative Bank</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="The Hongkong and Shanghai Banking Corporation">
        <xs:annotation>
          <xs:documentation>The Hongkong and Shanghai Banking Corporation</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.The Hongkong and Shanghai Banking Corporation</lx:path>
            <lx:label>The Hongkong and Shanghai Banking Corporation</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="The Rock">
        <xs:annotation>
          <xs:documentation>The Rock</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.The Rock</lx:path>
            <lx:label>The Rock</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Think Tank Group Pty Ltd">
        <xs:annotation>
          <xs:documentation>Think Tank Group Pty Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Think Tank Group Pty Ltd</lx:path>
            <lx:label>Think Tank Group Pty Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tictoc Online">
        <xs:annotation>
          <xs:documentation>Tictoc Online</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Tictoc Online</lx:path>
            <lx:label>Tictoc Online</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tonto Home Loans">
        <xs:annotation>
          <xs:documentation>Tonto Home Loans</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Tonto Home Loans</lx:path>
            <lx:label>Tonto Home Loans</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Town And Country Bank">
        <xs:annotation>
          <xs:documentation>Town And Country Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Town And Country Bank</lx:path>
            <lx:label>Town And Country Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Toyota Finance Australia">
        <xs:annotation>
          <xs:documentation>Toyota Finance Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Toyota Finance Australia</lx:path>
            <lx:label>Toyota Finance Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trust Bank of Tasmania">
        <xs:annotation>
          <xs:documentation>Trust Bank of Tasmania</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Trust Bank of Tasmania</lx:path>
            <lx:label>Trust Bank of Tasmania</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TSB Bank">
        <xs:annotation>
          <xs:documentation>TSB Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.TSB Bank</lx:path>
            <lx:label>TSB Bank</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TWUSUPER">
        <xs:annotation>
          <xs:documentation>TWUSUPER</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.TWUSUPER</lx:path>
            <lx:label>TWUSUPER</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="UBank">
        <xs:annotation>
          <xs:documentation>UBank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.UBank</lx:path>
            <lx:label>UBank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="United Credit Union">
        <xs:annotation>
          <xs:documentation>United Credit Union</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.United Credit Union</lx:path>
            <lx:label>United Credit Union</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="United Overseas Bank">
        <xs:annotation>
          <xs:documentation>United Overseas Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.United Overseas Bank</lx:path>
            <lx:label>United Overseas Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="University Credit Society">
        <xs:annotation>
          <xs:documentation>University Credit Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.University Credit Society</lx:path>
            <lx:label>University Credit Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="VACC">
        <xs:annotation>
          <xs:documentation>VACC</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.VACC</lx:path>
            <lx:label>VACC</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vero">
        <xs:annotation>
          <xs:documentation>Vero</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Vero</lx:path>
            <lx:label>Vero</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="VicSuper">
        <xs:annotation>
          <xs:documentation>VicSuper</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.VicSuper</lx:path>
            <lx:label>VicSuper</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Victoria Teachers Credit Union Ltd">
        <xs:annotation>
          <xs:documentation>Victoria Teachers Credit Union Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Victoria Teachers Credit Union Ltd</lx:path>
            <lx:label>Victoria Teachers Credit Union Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Virgin Super">
        <xs:annotation>
          <xs:documentation>Virgin Super</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Virgin Super</lx:path>
            <lx:label>Virgin Super</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Visions">
        <xs:annotation>
          <xs:documentation>Visions</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Visions</lx:path>
            <lx:label>Visions</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Western QBE">
        <xs:annotation>
          <xs:documentation>Western QBE</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Western QBE</lx:path>
            <lx:label>Western QBE</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Westfarmers">
        <xs:annotation>
          <xs:documentation>Westfarmers</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Westfarmers</lx:path>
            <lx:label>Westfarmers</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Westpac Bank">
        <xs:annotation>
          <xs:documentation>Westpac Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Westpac Bank</lx:path>
            <lx:label>Westpac Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Westpack Insurance Services">
        <xs:annotation>
          <xs:documentation>Westpack Insurance Services</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Westpack Insurance Services</lx:path>
            <lx:label>Westpack Insurance Services</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wide Bay Capricorn Building Society">
        <xs:annotation>
          <xs:documentation>Wide Bay Capricorn Building Society</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Wide Bay Capricorn Building Society</lx:path>
            <lx:label>Wide Bay Capricorn Building Society</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wizard">
        <xs:annotation>
          <xs:documentation>Wizard</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Wizard</lx:path>
            <lx:label>Wizard</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="World Marine and General">
        <xs:annotation>
          <xs:documentation>World Marine and General</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.World Marine and General</lx:path>
            <lx:label>World Marine and General</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Your Manager">
        <xs:annotation>
          <xs:documentation>Your Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Your Manager</lx:path>
            <lx:label>Your Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Zip Co">
        <xs:annotation>
          <xs:documentation>Zip Co</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Zip Co</lx:path>
            <lx:label>Zip Co</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Zurich">
        <xs:annotation>
          <xs:documentation>Zurich</xs:documentation>
          <xs:appinfo>
            <lx:path>financialInstitutionList.Zurich</lx:path>
            <lx:label>Zurich</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="financialTransactionTypeList">
    <xs:annotation>
      <xs:documentation>financialTransactionTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>financialTransactionTypeList</lx:path>
        <lx:label>financialTransactionTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Leaseback">
        <xs:annotation>
          <xs:documentation>Leaseback</xs:documentation>
          <xs:appinfo>
            <lx:path>financialTransactionTypeList.Leaseback</lx:path>
            <lx:label>Leaseback</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Refinance">
        <xs:annotation>
          <xs:documentation>Refinance</xs:documentation>
          <xs:appinfo>
            <lx:path>financialTransactionTypeList.Refinance</lx:path>
            <lx:label>Refinance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rollover">
        <xs:annotation>
          <xs:documentation>Rollover</xs:documentation>
          <xs:appinfo>
            <lx:path>financialTransactionTypeList.Rollover</lx:path>
            <lx:label>Rollover</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sale and Hire">
        <xs:annotation>
          <xs:documentation>Sale and Hire</xs:documentation>
          <xs:appinfo>
            <lx:path>financialTransactionTypeList.Sale and Hire</lx:path>
            <lx:label>Sale and Hire</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="fixtureTypeList">
    <xs:annotation>
      <xs:documentation>A list of the types of fixtures that can be attached to a property.</xs:documentation>
      <xs:appinfo>
        <lx:path>fixtureTypeList</lx:path>
        <lx:label>fixtureTypeList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="All Standard Fixtures">
        <xs:annotation>
          <xs:documentation>All Standard Fixtures</xs:documentation>
          <xs:appinfo>
            <lx:path>fixtureTypeList.All Standard Fixtures</lx:path>
            <lx:label>All Standard Fixtures</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bathtub">
        <xs:annotation>
          <xs:documentation>Bathtub</xs:documentation>
          <xs:appinfo>
            <lx:path>fixtureTypeList.Bathtub</lx:path>
            <lx:label>Bathtub</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Built in Wardrobes">
        <xs:annotation>
          <xs:documentation>Built in Wardrobes</xs:documentation>
          <xs:appinfo>
            <lx:path>fixtureTypeList.Built in Wardrobes</lx:path>
            <lx:label>Built in Wardrobes</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fire Sprinkler System">
        <xs:annotation>
          <xs:documentation>Fire Sprinkler System</xs:documentation>
          <xs:appinfo>
            <lx:path>fixtureTypeList.Fire Sprinkler System</lx:path>
            <lx:label>Fire Sprinkler System</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fireplace">
        <xs:annotation>
          <xs:documentation>Fireplace</xs:documentation>
          <xs:appinfo>
            <lx:path>fixtureTypeList.Fireplace</lx:path>
            <lx:label>Fireplace</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hot Water System - Electric">
        <xs:annotation>
          <xs:documentation>Hot Water System - Electric</xs:documentation>
          <xs:appinfo>
            <lx:path>fixtureTypeList.Hot Water System - Electric</lx:path>
            <lx:label>Hot Water System - Electric</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hot Water System - Gas">
        <xs:annotation>
          <xs:documentation>Hot Water System - Gas</xs:documentation>
          <xs:appinfo>
            <lx:path>fixtureTypeList.Hot Water System - Gas</lx:path>
            <lx:label>Hot Water System - Gas</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>fixtureTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Alarm System">
        <xs:annotation>
          <xs:documentation>Security Alarm System</xs:documentation>
          <xs:appinfo>
            <lx:path>fixtureTypeList.Security Alarm System</lx:path>
            <lx:label>Security Alarm System</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="floorMaterialList">
    <xs:annotation>
      <xs:documentation>A list of the floor material types.</xs:documentation>
      <xs:appinfo>
        <lx:path>floorMaterialList</lx:path>
        <lx:label>floorMaterialList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Brick">
        <xs:annotation>
          <xs:documentation>Brick</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Brick</lx:path>
            <lx:label>Brick</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Carpet">
        <xs:annotation>
          <xs:documentation>Carpet</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Carpet</lx:path>
            <lx:label>Carpet</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Concrete">
        <xs:annotation>
          <xs:documentation>Concrete</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Concrete</lx:path>
            <lx:label>Concrete</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cork">
        <xs:annotation>
          <xs:documentation>Cork</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Cork</lx:path>
            <lx:label>Cork</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Glass">
        <xs:annotation>
          <xs:documentation>Glass</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Glass</lx:path>
            <lx:label>Glass</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Laminate">
        <xs:annotation>
          <xs:documentation>Laminate</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Laminate</lx:path>
            <lx:label>Laminate</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Linoleum">
        <xs:annotation>
          <xs:documentation>Linoleum</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Linoleum</lx:path>
            <lx:label>Linoleum</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Marble">
        <xs:annotation>
          <xs:documentation>Marble</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Marble</lx:path>
            <lx:label>Marble</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mosaic">
        <xs:annotation>
          <xs:documentation>Mosaic</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Mosaic</lx:path>
            <lx:label>Mosaic</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mud">
        <xs:annotation>
          <xs:documentation>Mud</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Mud</lx:path>
            <lx:label>Mud</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PVC">
        <xs:annotation>
          <xs:documentation>PVC</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.PVC</lx:path>
            <lx:label>PVC</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rubber">
        <xs:annotation>
          <xs:documentation>Rubber</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Rubber</lx:path>
            <lx:label>Rubber</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Stone">
        <xs:annotation>
          <xs:documentation>Stone</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Stone</lx:path>
            <lx:label>Stone</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Terrazzo">
        <xs:annotation>
          <xs:documentation>Terrazzo</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Terrazzo</lx:path>
            <lx:label>Terrazzo</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tile">
        <xs:annotation>
          <xs:documentation>Tile</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Tile</lx:path>
            <lx:label>Tile</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Timber">
        <xs:annotation>
          <xs:documentation>Timber</xs:documentation>
          <xs:appinfo>
            <lx:path>floorMaterialList.Timber</lx:path>
            <lx:label>Timber</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="foreignEmployedBasisList">
    <xs:annotation>
      <xs:documentation>foreignEmployedBasisList</xs:documentation>
      <xs:appinfo>
        <lx:path>foreignEmployedBasisList</lx:path>
        <lx:label>foreignEmployedBasisList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Casual">
        <xs:annotation>
          <xs:documentation>The employment described is under a casual contract of employment.</xs:documentation>
          <xs:appinfo>
            <lx:path>foreignEmployedBasisList.Casual</lx:path>
            <lx:label>Casual</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Commission Only">
        <xs:annotation>
          <xs:documentation>The employment described is remunerated by commission only, there is no base or retainer paid.</xs:documentation>
          <xs:appinfo>
            <lx:path>foreignEmployedBasisList.Commission Only</lx:path>
            <lx:label>Commission Only</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Contract">
        <xs:annotation>
          <xs:documentation>The employment described is under a contract for services.</xs:documentation>
          <xs:appinfo>
            <lx:path>foreignEmployedBasisList.Contract</lx:path>
            <lx:label>Contract</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Full Time">
        <xs:annotation>
          <xs:documentation>The employment described is under full time employment contract.</xs:documentation>
          <xs:appinfo>
            <lx:path>foreignEmployedBasisList.Full Time</lx:path>
            <lx:label>Full Time</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Part Time">
        <xs:annotation>
          <xs:documentation>The employment described is under part time employment contract.</xs:documentation>
          <xs:appinfo>
            <lx:path>foreignEmployedBasisList.Part Time</lx:path>
            <lx:label>Part Time</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Seasonal">
        <xs:annotation>
          <xs:documentation>The employment described is seasonal in nature, for example fruit picking.</xs:documentation>
          <xs:appinfo>
            <lx:path>foreignEmployedBasisList.Seasonal</lx:path>
            <lx:label>Seasonal</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Temporary">
        <xs:annotation>
          <xs:documentation>The employment described is under contract with an employment agency, also known as temping.</xs:documentation>
          <xs:appinfo>
            <lx:path>foreignEmployedBasisList.Temporary</lx:path>
            <lx:label>Temporary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="foreignTaxAssociationStatusList">
    <xs:annotation>
      <xs:documentation>foreignTaxAssociationStatusList</xs:documentation>
      <xs:appinfo>
        <lx:path>foreignTaxAssociationStatusList</lx:path>
        <lx:label>foreignTaxAssociationStatusList</lx:label>
        <li:transactions>CAL,CDA,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Fully Provided">
        <xs:annotation>
          <xs:documentation>The foreign tax associations have been fully provided by the entity or their representative.</xs:documentation>
          <xs:appinfo>
            <lx:path>foreignTaxAssociationStatusList.Fully Provided</lx:path>
            <lx:label>Fully Provided</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="None">
        <xs:annotation>
          <xs:documentation>The entity or their representative has disclosed that there are no foreign tax associations.</xs:documentation>
          <xs:appinfo>
            <lx:path>foreignTaxAssociationStatusList.None</lx:path>
            <lx:label>None</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Not Provided">
        <xs:annotation>
          <xs:documentation>The entity or their representative has not disclosed full details of whether there are any foreign tax associations and/or the details of those foreign tax associations.</xs:documentation>
          <xs:appinfo>
            <lx:path>foreignTaxAssociationStatusList.Not Provided</lx:path>
            <lx:label>Not Provided</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="frequencyFullList">
    <xs:annotation>
      <xs:documentation>frequencyFullList</xs:documentation>
      <xs:appinfo>
        <lx:path>frequencyFullList</lx:path>
        <lx:label>frequencyFullList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Daily">
        <xs:annotation>
          <xs:documentation>Daily</xs:documentation>
          <xs:appinfo>
            <lx:path>frequencyFullList.Daily</lx:path>
            <lx:label>Daily</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fortnightly">
        <xs:annotation>
          <xs:documentation>Fortnightly</xs:documentation>
          <xs:appinfo>
            <lx:path>frequencyFullList.Fortnightly</lx:path>
            <lx:label>Fortnightly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Half Yearly">
        <xs:annotation>
          <xs:documentation>Half Yearly</xs:documentation>
          <xs:appinfo>
            <lx:path>frequencyFullList.Half Yearly</lx:path>
            <lx:label>Half Yearly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Monthly">
        <xs:annotation>
          <xs:documentation>Monthly</xs:documentation>
          <xs:appinfo>
            <lx:path>frequencyFullList.Monthly</lx:path>
            <lx:label>Monthly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Quarterly">
        <xs:annotation>
          <xs:documentation>Quarterly</xs:documentation>
          <xs:appinfo>
            <lx:path>frequencyFullList.Quarterly</lx:path>
            <lx:label>Quarterly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Seasonal">
        <xs:annotation>
          <xs:documentation>Seasonal</xs:documentation>
          <xs:appinfo>
            <lx:path>frequencyFullList.Seasonal</lx:path>
            <lx:label>Seasonal</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Weekly">
        <xs:annotation>
          <xs:documentation>Weekly</xs:documentation>
          <xs:appinfo>
            <lx:path>frequencyFullList.Weekly</lx:path>
            <lx:label>Weekly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Yearly">
        <xs:annotation>
          <xs:documentation>Yearly</xs:documentation>
          <xs:appinfo>
            <lx:path>frequencyFullList.Yearly</lx:path>
            <lx:label>Yearly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="frequencyShortList">
    <xs:annotation>
      <xs:documentation>frequencyShortList</xs:documentation>
      <xs:appinfo>
        <lx:path>frequencyShortList</lx:path>
        <lx:label>frequencyShortList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Fortnightly">
        <xs:annotation>
          <xs:documentation>Fortnightly</xs:documentation>
          <xs:appinfo>
            <lx:path>frequencyShortList.Fortnightly</lx:path>
            <lx:label>Fortnightly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Monthly">
        <xs:annotation>
          <xs:documentation>Monthly</xs:documentation>
          <xs:appinfo>
            <lx:path>frequencyShortList.Monthly</lx:path>
            <lx:label>Monthly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Weekly">
        <xs:annotation>
          <xs:documentation>Weekly</xs:documentation>
          <xs:appinfo>
            <lx:path>frequencyShortList.Weekly</lx:path>
            <lx:label>Weekly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Yearly">
        <xs:annotation>
          <xs:documentation>Yearly</xs:documentation>
          <xs:appinfo>
            <lx:path>frequencyShortList.Yearly</lx:path>
            <lx:label>Yearly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="funderList">
    <xs:annotation>
      <xs:documentation>funderList</xs:documentation>
      <xs:appinfo>
        <lx:path>funderList</lx:path>
        <lx:label>funderList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Adelaide Bank">
        <xs:annotation>
          <xs:documentation>Adelaide Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.Adelaide Bank</lx:path>
            <lx:label>Adelaide Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Advantedge">
        <xs:annotation>
          <xs:documentation>Advantedge</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.Advantedge</lx:path>
            <lx:label>Advantedge</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Arch Finance">
        <xs:annotation>
          <xs:documentation>Arch Finance</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.Arch Finance</lx:path>
            <lx:label>Arch Finance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BC Securities">
        <xs:annotation>
          <xs:documentation>BC Securities</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.BC Securities</lx:path>
            <lx:label>BC Securities</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bendigo Bank">
        <xs:annotation>
          <xs:documentation>Bendigo Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.Bendigo Bank</lx:path>
            <lx:label>Bendigo Bank</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Firstmac">
        <xs:annotation>
          <xs:documentation>Firstmac</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.Firstmac</lx:path>
            <lx:label>Firstmac</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LaTrobe">
        <xs:annotation>
          <xs:documentation>LaTrobe</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.LaTrobe</lx:path>
            <lx:label>LaTrobe</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Macquarie">
        <xs:annotation>
          <xs:documentation>Macquarie</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.Macquarie</lx:path>
            <lx:label>Macquarie</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MKM Group">
        <xs:annotation>
          <xs:documentation>MKM Group</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.MKM Group</lx:path>
            <lx:label>MKM Group</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Origin">
        <xs:annotation>
          <xs:documentation>Origin</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.Origin</lx:path>
            <lx:label>Origin</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pepper">
        <xs:annotation>
          <xs:documentation>Pepper</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.Pepper</lx:path>
            <lx:label>Pepper</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RedZed">
        <xs:annotation>
          <xs:documentation>RedZed</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.RedZed</lx:path>
            <lx:label>RedZed</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Resimac">
        <xs:annotation>
          <xs:documentation>Resimac</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.Resimac</lx:path>
            <lx:label>Resimac</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sintex">
        <xs:annotation>
          <xs:documentation>Sintex</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.Sintex</lx:path>
            <lx:label>Sintex</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Think Tank">
        <xs:annotation>
          <xs:documentation>Think Tank</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.Think Tank</lx:path>
            <lx:label>Think Tank</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Well Nigh">
        <xs:annotation>
          <xs:documentation>Well Nigh</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.Well Nigh</lx:path>
            <lx:label>Well Nigh</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wilson and Copley">
        <xs:annotation>
          <xs:documentation>Wilson and Copley</xs:documentation>
          <xs:appinfo>
            <lx:path>funderList.Wilson and Copley</lx:path>
            <lx:label>Wilson and Copley</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="fundsAccessTypeList">
    <xs:annotation>
      <xs:documentation>fundsAccessTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>fundsAccessTypeList</lx:path>
        <lx:label>Funds Access Type List</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Line of Credit">
        <xs:annotation>
          <xs:documentation>Line of Credit</xs:documentation>
          <xs:appinfo>
            <lx:path>fundsAccessTypeList.Line of Credit</lx:path>
            <lx:label>Line of Credit</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="None">
        <xs:annotation>
          <xs:documentation>None</xs:documentation>
          <xs:appinfo>
            <lx:path>fundsAccessTypeList.None</lx:path>
            <lx:label>None</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Offset">
        <xs:annotation>
          <xs:documentation>Offset</xs:documentation>
          <xs:appinfo>
            <lx:path>fundsAccessTypeList.Offset</lx:path>
            <lx:label>Offset</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Redraw">
        <xs:annotation>
          <xs:documentation>Redraw</xs:documentation>
          <xs:appinfo>
            <lx:path>fundsAccessTypeList.Redraw</lx:path>
            <lx:label>Redraw</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="genderList">
    <xs:annotation>
      <xs:documentation>genderList</xs:documentation>
      <xs:appinfo>
        <lx:path>genderList</lx:path>
        <lx:label>genderList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Female">
        <xs:annotation>
          <xs:documentation>The person is a female.</xs:documentation>
          <xs:appinfo>
            <lx:path>genderList.Female</lx:path>
            <lx:label>The person is a female.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Male">
        <xs:annotation>
          <xs:documentation>The person is a male.</xs:documentation>
          <xs:appinfo>
            <lx:path>genderList.Male</lx:path>
            <lx:label>The person is a male.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Undisclosed">
        <xs:annotation>
          <xs:documentation>The person has elected not to disclose their gender.</xs:documentation>
          <xs:appinfo>
            <lx:path>genderList.Undisclosed</lx:path>
            <lx:label>The person has elected not to disclose their gender.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="generalSecurityAgreementTypeList">
    <xs:annotation>
      <xs:documentation>generalSecurityAgreementTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>generalSecurityAgreementTypeList</lx:path>
        <lx:label>generalSecurityAgreementTypeList</lx:label>
        <li:transactions>ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="All Present and Future Acquired Properties With Exceptions">
        <xs:annotation>
          <xs:documentation>All Present and Future Acquired Properties With Exceptions</xs:documentation>
          <xs:appinfo>
            <lx:path>generalSecurityAgreementTypeList.All Present and Future Acquired Properties With Exceptions</lx:path>
            <lx:label>All Present and Future Acquired Properties With Exceptions</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="All Present and Future Acquired Properties Without Exceptions">
        <xs:annotation>
          <xs:documentation>All Present and Future Acquired Properties Without Exceptions</xs:documentation>
          <xs:appinfo>
            <lx:path>generalSecurityAgreementTypeList.All Present and Future Acquired Properties Without Exceptions</lx:path>
            <lx:label>All Present and Future Acquired Properties Without Exceptions</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="goodsDeliveryList">
    <xs:annotation>
      <xs:documentation>goodsDeliveryList</xs:documentation>
      <xs:appinfo>
        <lx:path>goodsDeliveryList</lx:path>
        <lx:label>goodsDeliveryList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Already Delivered">
        <xs:annotation>
          <xs:documentation>Already Delivered</xs:documentation>
          <xs:appinfo>
            <lx:path>goodsDeliveryList.Already Delivered</lx:path>
            <lx:label>Already Delivered</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="To Be Delivered on Payment">
        <xs:annotation>
          <xs:documentation>To Be Delivered on Payment</xs:documentation>
          <xs:appinfo>
            <lx:path>goodsDeliveryList.To Be Delivered on Payment</lx:path>
            <lx:label>To Be Delivered on Payment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Will Be Delivered on Date">
        <xs:annotation>
          <xs:documentation>Will Be Delivered on Date</xs:documentation>
          <xs:appinfo>
            <lx:path>goodsDeliveryList.Will Be Delivered on Date</lx:path>
            <lx:label>Will Be Delivered on Date</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="goodToBeUsedLocationList">
    <xs:annotation>
      <xs:documentation>goodToBeUsedLocationList</xs:documentation>
      <xs:appinfo>
        <lx:path>goodToBeUsedLocationList</lx:path>
        <lx:label>goodToBeUsedLocationList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="On Business Premises">
        <xs:annotation>
          <xs:documentation>On Business Premises</xs:documentation>
          <xs:appinfo>
            <lx:path>goodToBeUsedLocationList.On Business Premises</lx:path>
            <lx:label>On Business Premises</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="On Site">
        <xs:annotation>
          <xs:documentation>On Site</xs:documentation>
          <xs:appinfo>
            <lx:path>goodToBeUsedLocationList.On Site</lx:path>
            <lx:label>On Site</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="On Site Daily Return">
        <xs:annotation>
          <xs:documentation>On Site Daily Return</xs:documentation>
          <xs:appinfo>
            <lx:path>goodToBeUsedLocationList.On Site Daily Return</lx:path>
            <lx:label>On Site Daily Return</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="governmentBenefitsTypeList">
    <xs:annotation>
      <xs:documentation>governmentBenefitsTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>governmentBenefitsTypeList</lx:path>
        <lx:label>governmentBenefitsTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Abstudy">
        <xs:annotation>
          <xs:documentation>Government support income for Aboriginal or Torres Strait Islander students or apprentices.</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Abstudy</lx:path>
            <lx:label>Abstudy</lx:label>
            <li:transactions>CAL,CDA,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Accident Compensation">
        <xs:annotation>
          <xs:documentation>Government support income for people who were injured in an accident.</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Accident Compensation</lx:path>
            <lx:label>Accident Compensation</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Age Pension">
        <xs:annotation>
          <xs:documentation>The Age Pension is a taxpayer-funded basic retirement income stream for life, for those Australians who meet the eligibility requirements for the Age Pension.</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Age Pension</lx:path>
            <lx:label>Age Pension</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Austudy">
        <xs:annotation>
          <xs:documentation>Government support income for people 25 years of age or older, and studying or an Australian apprentice.</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Austudy</lx:path>
            <lx:label>Austudy</lx:label>
            <li:transactions>CAL,CDA,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Carer Payment">
        <xs:annotation>
          <xs:documentation>Carer Payment</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Carer Payment</lx:path>
            <lx:label>Carer Payment</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Child Support">
        <xs:annotation>
          <xs:documentation>Child Support</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Child Support</lx:path>
            <lx:label>Child Support</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Crisis Payment">
        <xs:annotation>
          <xs:documentation>Crisis Payment</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Crisis Payment</lx:path>
            <lx:label>Crisis Payment</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Disability Support Pension">
        <xs:annotation>
          <xs:documentation>Disability Support Pension</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Disability Support Pension</lx:path>
            <lx:label>Disability Support Pension</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Family Allowance">
        <xs:annotation>
          <xs:documentation>Family Allowance</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Family Allowance</lx:path>
            <lx:label>Family Allowance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Invalid Benefits">
        <xs:annotation>
          <xs:documentation>Government income support for people who are permanently and severely restricted in their capacity for work because of sickness, injury, or disability, or be totally blind.</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Invalid Benefits</lx:path>
            <lx:label>Invalid Benefits</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mobility Allowance">
        <xs:annotation>
          <xs:documentation>Mobility Allowance</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Mobility Allowance</lx:path>
            <lx:label>Mobility Allowance</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Parenting Payment">
        <xs:annotation>
          <xs:documentation>Parenting Payment is government provided income support for parents or guardians to help with the cost of raising children.</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Parenting Payment</lx:path>
            <lx:label>Parenting Payment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sickness Benefits">
        <xs:annotation>
          <xs:documentation>Government support income for people who are temporarily off work or working at a reduced level because of sickness, an injury, pregnancy, or a disability.</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Sickness Benefits</lx:path>
            <lx:label>Sickness Benefits</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sole Parent Support">
        <xs:annotation>
          <xs:documentation>Government support income for single parents and caregivers of dependent children.</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Sole Parent Support</lx:path>
            <lx:label>Sole Parent Support</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Widows Allowance">
        <xs:annotation>
          <xs:documentation>Widows Allowance</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Widows Allowance</lx:path>
            <lx:label>Widows Allowance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Working for Families">
        <xs:annotation>
          <xs:documentation>Government support income for families with low income or large families.</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Working for Families</lx:path>
            <lx:label>Working for Families</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Youth Allowance">
        <xs:annotation>
          <xs:documentation>Youth Allowance.</xs:documentation>
          <xs:appinfo>
            <lx:path>governmentBenefitsTypeList.Youth Allowance</lx:path>
            <lx:label>Youth Allowance</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="gradeList">
    <xs:annotation>
      <xs:documentation>gradeList</xs:documentation>
      <xs:appinfo>
        <lx:path>gradeList</lx:path>
        <lx:label>gradeList</lx:label>
        <li:transactions>ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Near Prime">
        <xs:annotation>
          <xs:documentation>Near Prime</xs:documentation>
          <xs:appinfo>
            <lx:path>gradeList.Near Prime</lx:path>
            <lx:label>Near Prime</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Prime">
        <xs:annotation>
          <xs:documentation>Prime</xs:documentation>
          <xs:appinfo>
            <lx:path>gradeList.Prime</lx:path>
            <lx:label>Prime</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="guarantorAccessList">
    <xs:annotation>
      <xs:documentation>guarantorAccessList</xs:documentation>
      <xs:appinfo>
        <lx:path>guarantorAccessList</lx:path>
        <lx:label>guarantorAccessList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Full Access">
        <xs:annotation>
          <xs:documentation>Full Access</xs:documentation>
          <xs:appinfo>
            <lx:path>guarantorAccessList.Full Access</lx:path>
            <lx:label>Full Access</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Information Only Access">
        <xs:annotation>
          <xs:documentation>Information Only Access</xs:documentation>
          <xs:appinfo>
            <lx:path>guarantorAccessList.Information Only Access</lx:path>
            <lx:label>Information Only Access</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="No Access">
        <xs:annotation>
          <xs:documentation>No Access</xs:documentation>
          <xs:appinfo>
            <lx:path>guarantorAccessList.No Access</lx:path>
            <lx:label>No Access</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Transacting Access">
        <xs:annotation>
          <xs:documentation>Transacting Access</xs:documentation>
          <xs:appinfo>
            <lx:path>guarantorAccessList.Transacting Access</lx:path>
            <lx:label>Transacting Access</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="holdingList">
    <xs:annotation>
      <xs:documentation>holdingList</xs:documentation>
      <xs:appinfo>
        <lx:path>holdingList</lx:path>
        <lx:label>holdingList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Company Share">
        <xs:annotation>
          <xs:documentation>The parties specified as the owners of the property hold their interest as owners of shares in a company that owns the property.</xs:documentation>
          <xs:appinfo>
            <lx:path>holdingList.Company Share</lx:path>
            <lx:label>The parties specified as the owners of the property hold their interest as owners of shares in a company that owns the property.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Joint Tenants">
        <xs:annotation>
          <xs:documentation>The parties specified as the owners of the property hold their interest as joint tenants.</xs:documentation>
          <xs:appinfo>
            <lx:path>holdingList.Joint Tenants</lx:path>
            <lx:label>The parties specified as the owners of the property hold their interest as joint tenants.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sole">
        <xs:annotation>
          <xs:documentation>The party specified as the owner of the property holds their interest as sole proprietor.</xs:documentation>
          <xs:appinfo>
            <lx:path>holdingList.Sole</lx:path>
            <lx:label>The party specified as the owner of the property holds their interest as sole proprietor.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tennant Common in Equal Shares">
        <xs:annotation>
          <xs:documentation>The parties specified as the owners of the property hold their interest as tenants in common in equal shares.</xs:documentation>
          <xs:appinfo>
            <lx:path>holdingList.Tennant Common in Equal Shares</lx:path>
            <lx:label>The parties specified as the owners of the property hold their interest as tenants in common in equal shares.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tennant Common in Shares of">
        <xs:annotation>
          <xs:documentation>The parties specified as the owners of the property hold their interest as tenants in common in unspecified shares.</xs:documentation>
          <xs:appinfo>
            <lx:path>holdingList.Tennant Common in Shares of</lx:path>
            <lx:label>The parties specified as the owners of the property hold their interest as tenants in common in unspecified shares.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Time Share">
        <xs:annotation>
          <xs:documentation>The parties specified as the owners of the property hold their interest under a time share arrangement.</xs:documentation>
          <xs:appinfo>
            <lx:path>holdingList.Time Share</lx:path>
            <lx:label>The parties specified as the owners of the property hold their interest under a time share arrangement.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="hospitalityAndLeisureTypeList">
    <xs:annotation>
      <xs:documentation>hospitalityAndLeisureTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>hospitalityAndLeisureTypeList</lx:path>
        <lx:label>hospitalityAndLeisureTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Air Conditioning">
        <xs:annotation>
          <xs:documentation>Air Conditioning</xs:documentation>
          <xs:appinfo>
            <lx:path>hospitalityAndLeisureTypeList.Air Conditioning</lx:path>
            <lx:label>Air Conditioning</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cooking">
        <xs:annotation>
          <xs:documentation>Cooking</xs:documentation>
          <xs:appinfo>
            <lx:path>hospitalityAndLeisureTypeList.Cooking</lx:path>
            <lx:label>Cooking</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gaming">
        <xs:annotation>
          <xs:documentation>Gaming</xs:documentation>
          <xs:appinfo>
            <lx:path>hospitalityAndLeisureTypeList.Gaming</lx:path>
            <lx:label>Gaming</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Garden">
        <xs:annotation>
          <xs:documentation>Garden</xs:documentation>
          <xs:appinfo>
            <lx:path>hospitalityAndLeisureTypeList.Garden</lx:path>
            <lx:label>Garden</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Kitchen">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a future release. Kitchen</xs:documentation>
          <xs:appinfo>
            <lx:path>hospitalityAndLeisureTypeList.Kitchen</lx:path>
            <lx:label>Kitchen</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>hospitalityAndLeisureTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Refrigeration">
        <xs:annotation>
          <xs:documentation>Refrigeration</xs:documentation>
          <xs:appinfo>
            <lx:path>hospitalityAndLeisureTypeList.Refrigeration</lx:path>
            <lx:label>Refrigeration</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Salon">
        <xs:annotation>
          <xs:documentation>Salon</xs:documentation>
          <xs:appinfo>
            <lx:path>hospitalityAndLeisureTypeList.Salon</lx:path>
            <lx:label>Salon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Spa">
        <xs:annotation>
          <xs:documentation>Spa</xs:documentation>
          <xs:appinfo>
            <lx:path>hospitalityAndLeisureTypeList.Spa</lx:path>
            <lx:label>Spa</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="housingStatusList">
    <xs:annotation>
      <xs:documentation>housingStatusList</xs:documentation>
      <xs:appinfo>
        <lx:path>housingStatusList</lx:path>
        <lx:label>housingStatusList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Boarding">
        <xs:annotation>
          <xs:documentation>Person pays board to the person who has the right to occupy the property at the address they reside at.</xs:documentation>
          <xs:appinfo>
            <lx:path>housingStatusList.Boarding</lx:path>
            <lx:label>Person pays board to the person who has the right to occupy the property at the address they reside at.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Caravan">
        <xs:annotation>
          <xs:documentation>Person lives in a caravan or similar demountable dwelling at the address they reside at.</xs:documentation>
          <xs:appinfo>
            <lx:path>housingStatusList.Caravan</lx:path>
            <lx:label>Person lives in a caravan or similar demountable dwelling at the address they reside at.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Person's basis for living at the address they reside at is not found in the other enumerated values.</xs:documentation>
          <xs:appinfo>
            <lx:path>housingStatusList.Other</lx:path>
            <lx:label>Person's basis for living at the address they reside at is not found in the other enumerated values.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Own Home">
        <xs:annotation>
          <xs:documentation>Person owns outright the property at the address they reside at, property is not mortgaged.</xs:documentation>
          <xs:appinfo>
            <lx:path>housingStatusList.Own Home</lx:path>
            <lx:label>Person owns outright the property at the address they reside at, property is not mortgaged.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Own Home Mortgage">
        <xs:annotation>
          <xs:documentation>Person owns the property at the address they reside at, property is mortgaged.</xs:documentation>
          <xs:appinfo>
            <lx:path>housingStatusList.Own Home Mortgage</lx:path>
            <lx:label>Person owns the property at the address they reside at, property is mortgaged.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Renting">
        <xs:annotation>
          <xs:documentation>Person rents the property at the address they reside at.</xs:documentation>
          <xs:appinfo>
            <lx:path>housingStatusList.Renting</lx:path>
            <lx:label>Person rents the property at the address they reside at.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Supplied by Employer">
        <xs:annotation>
          <xs:documentation>Person lives at accommodation provided by their employer.</xs:documentation>
          <xs:appinfo>
            <lx:path>housingStatusList.Supplied by Employer</lx:path>
            <lx:label>Person lives at accommodation provided by their employer.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="With Parents">
        <xs:annotation>
          <xs:documentation>Person living with parents or other relatives in property at the address they reside at.</xs:documentation>
          <xs:appinfo>
            <lx:path>housingStatusList.With Parents</lx:path>
            <lx:label>Person living with parents or other relatives in property at the address they reside at.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="importanceList">
    <xs:annotation>
      <xs:documentation>importanceList</xs:documentation>
      <xs:appinfo>
        <lx:path>importanceList</lx:path>
        <lx:label>importanceList</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Don't Want">
        <xs:annotation>
          <xs:documentation>The applicant expressed they do not want the feature.</xs:documentation>
          <xs:appinfo>
            <lx:path>importanceList.Don't Want</lx:path>
            <lx:label>Don't Want</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Important">
        <xs:annotation>
          <xs:documentation>The feature is a high priority for the applicant.</xs:documentation>
          <xs:appinfo>
            <lx:path>importanceList.Important</lx:path>
            <lx:label>Important</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Not Important">
        <xs:annotation>
          <xs:documentation>The feature is not of high priority to the applicant.</xs:documentation>
          <xs:appinfo>
            <lx:path>importanceList.Not Important</lx:path>
            <lx:label>Not Important</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="incomeStatusOnOrBeforeSettlementList">
    <xs:annotation>
      <xs:documentation>A list of options for income status on or before settlement.</xs:documentation>
      <xs:appinfo>
        <lx:path>incomeStatusOnOrBeforeSettlementList</lx:path>
        <lx:label>incomeStatusOnOrBeforeSettlementList</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Ends">
        <xs:annotation>
          <xs:documentation>The income will end on, or before, the settlement date.</xs:documentation>
          <xs:appinfo>
            <lx:path>incomeStatusOnOrBeforeSettlementList.Ends</lx:path>
            <lx:label>Ends</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ongoing">
        <xs:annotation>
          <xs:documentation>The income is ongoing at the settlement date.</xs:documentation>
          <xs:appinfo>
            <lx:path>incomeStatusOnOrBeforeSettlementList.Ongoing</lx:path>
            <lx:label>Ongoing</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Starts">
        <xs:annotation>
          <xs:documentation>The income will start on, or before, the settlement date.</xs:documentation>
          <xs:appinfo>
            <lx:path>incomeStatusOnOrBeforeSettlementList.Starts</lx:path>
            <lx:label>Starts</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="incomeVerificationList">
    <xs:annotation>
      <xs:documentation>incomeVerificationList</xs:documentation>
      <xs:appinfo>
        <lx:path>incomeVerificationList</lx:path>
        <lx:label>incomeVerificationList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Fully Verified">
        <xs:annotation>
          <xs:documentation>Indicates that the income claimed will be verified by provision of evidence of the income, for example using payslips which evidence PAYG income.</xs:documentation>
          <xs:appinfo>
            <lx:path>incomeVerificationList.Fully Verified</lx:path>
            <lx:label>Indicates that the income claimed will be verified by provision of evidence of the income, for example using payslips which evidence PAYG income.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Self Certified">
        <xs:annotation>
          <xs:documentation>Indicates that the applicant will provide a written statement in which they certify that they earn the income claimed.</xs:documentation>
          <xs:appinfo>
            <lx:path>incomeVerificationList.Self Certified</lx:path>
            <lx:label>Indicates that the applicant will provide a written statement in which they certify that they earn the income claimed.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="independentAdviceTypeRequiredList">
    <xs:annotation>
      <xs:documentation>A list of the types of independent advice that a guarantor may be advised to seek.</xs:documentation>
      <xs:appinfo>
        <lx:path>independentAdviceTypeRequiredList</lx:path>
        <lx:label>Advised To Seek Independent Advice Type List</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Financial Advice">
        <xs:annotation>
          <xs:documentation>Financial Advice</xs:documentation>
          <xs:appinfo>
            <lx:path>independentAdviceTypeRequiredList.Financial Advice</lx:path>
            <lx:label>Financial Advice</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Financial and Legal Advice">
        <xs:annotation>
          <xs:documentation>Financial and Legal Advice</xs:documentation>
          <xs:appinfo>
            <lx:path>independentAdviceTypeRequiredList.Financial and Legal Advice</lx:path>
            <lx:label>Financial and Legal Advice</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Legal Advice">
        <xs:annotation>
          <xs:documentation>Legal Advice</xs:documentation>
          <xs:appinfo>
            <lx:path>independentAdviceTypeRequiredList.Legal Advice</lx:path>
            <lx:label>Legal Advice</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="industrialTypeList">
    <xs:annotation>
      <xs:documentation>industrialTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>industrialTypeList</lx:path>
        <lx:label>industrialTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Light Industrial">
        <xs:annotation>
          <xs:documentation>The industrial property is classified as light industrial.</xs:documentation>
          <xs:appinfo>
            <lx:path>industrialTypeList.Light Industrial</lx:path>
            <lx:label>The industrial property is classified as light industrial.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Non Specialised Industrial">
        <xs:annotation>
          <xs:documentation>The industrial property is classified as general industrial, not for specialised use.</xs:documentation>
          <xs:appinfo>
            <lx:path>industrialTypeList.Non Specialised Industrial</lx:path>
            <lx:label>The industrial property is classified as general industrial, not for specialised use.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>The industrial property cannot be classified using one of the other enumeration types.</xs:documentation>
          <xs:appinfo>
            <lx:path>industrialTypeList.Other</lx:path>
            <lx:label>The industrial property cannot be classified using one of the other enumeration types.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="instalmentsFrequencyList">
    <xs:annotation>
      <xs:documentation>instalmentsFrequencyList</xs:documentation>
      <xs:appinfo>
        <lx:path>instalmentsFrequencyList</lx:path>
        <lx:label>instalmentsFrequencyList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Fortnightly">
        <xs:annotation>
          <xs:documentation>Fortnightly</xs:documentation>
          <xs:appinfo>
            <lx:path>instalmentsFrequencyList.Fortnightly</lx:path>
            <lx:label>Fortnightly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Monthly">
        <xs:annotation>
          <xs:documentation>Monthly</xs:documentation>
          <xs:appinfo>
            <lx:path>instalmentsFrequencyList.Monthly</lx:path>
            <lx:label>Monthly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Weekly">
        <xs:annotation>
          <xs:documentation>Weekly</xs:documentation>
          <xs:appinfo>
            <lx:path>instalmentsFrequencyList.Weekly</lx:path>
            <lx:label>Weekly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="insuranceTypeList">
    <xs:annotation>
      <xs:documentation>insuranceTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>insuranceTypeList</lx:path>
        <lx:label>insuranceTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Body Corporate">
        <xs:annotation>
          <xs:documentation>Body Corporate</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Body Corporate</lx:path>
            <lx:label>Body Corporate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Building Insurance">
        <xs:annotation>
          <xs:documentation>Building Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Building Insurance</lx:path>
            <lx:label>Building Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Building Replacement">
        <xs:annotation>
          <xs:documentation>Building Replacement</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Building Replacement</lx:path>
            <lx:label>Building Replacement</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Consumer Credit Insurance">
        <xs:annotation>
          <xs:documentation>Consumer Credit Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Consumer Credit Insurance</lx:path>
            <lx:label>Consumer Credit Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Disablement">
        <xs:annotation>
          <xs:documentation>Disablement</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Disablement</lx:path>
            <lx:label>Disablement</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fire">
        <xs:annotation>
          <xs:documentation>Fire</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Fire</lx:path>
            <lx:label>Fire</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Home Contents Insurance">
        <xs:annotation>
          <xs:documentation>Home Contents Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Home Contents Insurance</lx:path>
            <lx:label>Home Contents Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Income Protection">
        <xs:annotation>
          <xs:documentation>Income Protection</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Income Protection</lx:path>
            <lx:label>Income Protection</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Involuntary Unemployment">
        <xs:annotation>
          <xs:documentation>Involuntary Unemployment</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Involuntary Unemployment</lx:path>
            <lx:label>Involuntary Unemployment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Land Only">
        <xs:annotation>
          <xs:documentation>Land Only</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Land Only</lx:path>
            <lx:label>Land Only</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Life Insurance">
        <xs:annotation>
          <xs:documentation>Life Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Life Insurance</lx:path>
            <lx:label>Life Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LMI">
        <xs:annotation>
          <xs:documentation>LMI</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.LMI</lx:path>
            <lx:label>LMI</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Professional Indemnity">
        <xs:annotation>
          <xs:documentation>Professional Indemnity</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Professional Indemnity</lx:path>
            <lx:label>Professional Indemnity</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Public Liability">
        <xs:annotation>
          <xs:documentation>Public Liability</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Public Liability</lx:path>
            <lx:label>Public Liability</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Theft">
        <xs:annotation>
          <xs:documentation>Theft</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Theft</lx:path>
            <lx:label>Theft</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Title Insurance">
        <xs:annotation>
          <xs:documentation>Title Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Title Insurance</lx:path>
            <lx:label>Title Insurance</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trauma">
        <xs:annotation>
          <xs:documentation>Trauma</xs:documentation>
          <xs:appinfo>
            <lx:path>insuranceTypeList.Trauma</lx:path>
            <lx:label>Trauma</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="insurerList">
    <xs:annotation>
      <xs:documentation>insurerList</xs:documentation>
      <xs:appinfo>
        <lx:path>insurerList</lx:path>
        <lx:label>insurerList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="AA Insurance">
        <xs:annotation>
          <xs:documentation>AA Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.AA Insurance</lx:path>
            <lx:label>AA Insurance</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AAMI">
        <xs:annotation>
          <xs:documentation>AAMI</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.AAMI</lx:path>
            <lx:label>AAMI</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AIA Australia">
        <xs:annotation>
          <xs:documentation>AIA Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.AIA Australia</lx:path>
            <lx:label>AIA Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Allianz Australia">
        <xs:annotation>
          <xs:documentation>Allianz Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Allianz Australia</lx:path>
            <lx:label>Allianz Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AMI Insurance">
        <xs:annotation>
          <xs:documentation>AMI Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.AMI Insurance</lx:path>
            <lx:label>AMI Insurance</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AMP Limited">
        <xs:annotation>
          <xs:documentation>AMP Limited</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.AMP Limited</lx:path>
            <lx:label>AMP Limited</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="API">
        <xs:annotation>
          <xs:documentation>API</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.API</lx:path>
            <lx:label>API</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="APIA">
        <xs:annotation>
          <xs:documentation>APIA</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.APIA</lx:path>
            <lx:label>APIA</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Asteron Life (Suncorp)">
        <xs:annotation>
          <xs:documentation>Asteron Life (Suncorp)</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Asteron Life (Suncorp)</lx:path>
            <lx:label>Asteron Life (Suncorp)</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Atradius">
        <xs:annotation>
          <xs:documentation>Atradius</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Atradius</lx:path>
            <lx:label>Atradius</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian Pensioners Insurance">
        <xs:annotation>
          <xs:documentation>Australian Pensioners Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Australian Pensioners Insurance</lx:path>
            <lx:label>Australian Pensioners Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Australian Unity">
        <xs:annotation>
          <xs:documentation>Australian Unity</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Australian Unity</lx:path>
            <lx:label>Australian Unity</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Auto and General Insurance">
        <xs:annotation>
          <xs:documentation>Auto and General Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Auto and General Insurance</lx:path>
            <lx:label>Auto and General Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="AXA Australia">
        <xs:annotation>
          <xs:documentation>AXA Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.AXA Australia</lx:path>
            <lx:label>AXA Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bingle">
        <xs:annotation>
          <xs:documentation>Bingle</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Bingle</lx:path>
            <lx:label>Bingle</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="BUPA Australia">
        <xs:annotation>
          <xs:documentation>BUPA Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.BUPA Australia</lx:path>
            <lx:label>BUPA Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Buzz">
        <xs:annotation>
          <xs:documentation>Buzz</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Buzz</lx:path>
            <lx:label>Buzz</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Calliden Insurance">
        <xs:annotation>
          <xs:documentation>Calliden Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Calliden Insurance</lx:path>
            <lx:label>Calliden Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CGU Insurance">
        <xs:annotation>
          <xs:documentation>CGU Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.CGU Insurance</lx:path>
            <lx:label>CGU Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CHU">
        <xs:annotation>
          <xs:documentation>CHU</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.CHU</lx:path>
            <lx:label>CHU</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Chubb">
        <xs:annotation>
          <xs:documentation>Chubb</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Chubb</lx:path>
            <lx:label>Chubb</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CIC">
        <xs:annotation>
          <xs:documentation>CIC</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.CIC</lx:path>
            <lx:label>CIC</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CIL">
        <xs:annotation>
          <xs:documentation>CIL</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.CIL</lx:path>
            <lx:label>CIL</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Colonial">
        <xs:annotation>
          <xs:documentation>Colonial</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Colonial</lx:path>
            <lx:label>Colonial</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Combined Insurance">
        <xs:annotation>
          <xs:documentation>Combined Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Combined Insurance</lx:path>
            <lx:label>Combined Insurance</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CommInsure">
        <xs:annotation>
          <xs:documentation>CommInsure</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.CommInsure</lx:path>
            <lx:label>CommInsure</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Commonwealth Bank">
        <xs:annotation>
          <xs:documentation>Commonwealth Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Commonwealth Bank</lx:path>
            <lx:label>Commonwealth Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Defence Services Homes">
        <xs:annotation>
          <xs:documentation>Defence Services Homes</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Defence Services Homes</lx:path>
            <lx:label>Defence Services Homes</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Elders">
        <xs:annotation>
          <xs:documentation>Elders</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Elders</lx:path>
            <lx:label>Elders</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Esanda">
        <xs:annotation>
          <xs:documentation>Esanda</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Esanda</lx:path>
            <lx:label>Esanda</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FAI">
        <xs:annotation>
          <xs:documentation>FAI</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.FAI</lx:path>
            <lx:label>FAI</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FMG Insurance">
        <xs:annotation>
          <xs:documentation>FMG Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.FMG Insurance</lx:path>
            <lx:label>FMG Insurance</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fortis">
        <xs:annotation>
          <xs:documentation>Fortis</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Fortis</lx:path>
            <lx:label>Fortis</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Genworth Financial">
        <xs:annotation>
          <xs:documentation>Genworth Financial</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Genworth Financial</lx:path>
            <lx:label>Genworth Financial</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GIO">
        <xs:annotation>
          <xs:documentation>GIO</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.GIO</lx:path>
            <lx:label>GIO</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Guardian Insurance">
        <xs:annotation>
          <xs:documentation>Guardian Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Guardian Insurance</lx:path>
            <lx:label>Guardian Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Guild">
        <xs:annotation>
          <xs:documentation>Guild</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Guild</lx:path>
            <lx:label>Guild</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HBF">
        <xs:annotation>
          <xs:documentation>HBF</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.HBF</lx:path>
            <lx:label>HBF</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IAG New Zealand">
        <xs:annotation>
          <xs:documentation>IAG New Zealand</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.IAG New Zealand</lx:path>
            <lx:label>IAG New Zealand</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="ING">
        <xs:annotation>
          <xs:documentation>ING</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.ING</lx:path>
            <lx:label>ING</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Insurance Australia Group">
        <xs:annotation>
          <xs:documentation>Insurance Australia Group</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Insurance Australia Group</lx:path>
            <lx:label>Insurance Australia Group</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="InsureMyRide">
        <xs:annotation>
          <xs:documentation>InsureMyRide</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.InsureMyRide</lx:path>
            <lx:label>InsureMyRide</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Just Car">
        <xs:annotation>
          <xs:documentation>Just Car</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Just Car</lx:path>
            <lx:label>Just Car</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LFI Group Pty Ltd">
        <xs:annotation>
          <xs:documentation>LFI Group Pty Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.LFI Group Pty Ltd</lx:path>
            <lx:label>LFI Group Pty Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lifebroker">
        <xs:annotation>
          <xs:documentation>Lifebroker</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Lifebroker</lx:path>
            <lx:label>Lifebroker</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Macquarie Life">
        <xs:annotation>
          <xs:documentation>Macquarie Life</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Macquarie Life</lx:path>
            <lx:label>Macquarie Life</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Manchester Unity Australia">
        <xs:annotation>
          <xs:documentation>Manchester Unity Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Manchester Unity Australia</lx:path>
            <lx:label>Manchester Unity Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Medibank Private">
        <xs:annotation>
          <xs:documentation>Medibank Private</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Medibank Private</lx:path>
            <lx:label>Medibank Private</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mercantile Mutual">
        <xs:annotation>
          <xs:documentation>Mercantile Mutual</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Mercantile Mutual</lx:path>
            <lx:label>Mercantile Mutual</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MLC">
        <xs:annotation>
          <xs:documentation>MLC</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.MLC</lx:path>
            <lx:label>MLC</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MMI">
        <xs:annotation>
          <xs:documentation>MMI</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.MMI</lx:path>
            <lx:label>MMI</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mutual Community">
        <xs:annotation>
          <xs:documentation>Mutual Community</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Mutual Community</lx:path>
            <lx:label>Mutual Community</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NCUIS">
        <xs:annotation>
          <xs:documentation>NCUIS</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.NCUIS</lx:path>
            <lx:label>NCUIS</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NIB Health Funds">
        <xs:annotation>
          <xs:documentation>NIB Health Funds</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.NIB Health Funds</lx:path>
            <lx:label>NIB Health Funds</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NRMA Insurance">
        <xs:annotation>
          <xs:documentation>NRMA Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.NRMA Insurance</lx:path>
            <lx:label>NRMA Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NZI">
        <xs:annotation>
          <xs:documentation>NZI</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.NZI</lx:path>
            <lx:label>NZI</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="OnePath (ANZ)">
        <xs:annotation>
          <xs:documentation>OnePath (ANZ)</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.OnePath (ANZ)</lx:path>
            <lx:label>OnePath (ANZ)</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Q Report Jewellery Insurance">
        <xs:annotation>
          <xs:documentation>Q Report Jewellery Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Q Report Jewellery Insurance</lx:path>
            <lx:label>Q Report Jewellery Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="QBE">
        <xs:annotation>
          <xs:documentation>QBE</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.QBE</lx:path>
            <lx:label>QBE</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="QBE Insurance - MemberCare">
        <xs:annotation>
          <xs:documentation>QBE Insurance - MemberCare</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.QBE Insurance - MemberCare</lx:path>
            <lx:label>QBE Insurance - MemberCare</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RAC">
        <xs:annotation>
          <xs:documentation>RAC</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.RAC</lx:path>
            <lx:label>RAC</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RAC WA">
        <xs:annotation>
          <xs:documentation>RAC WA</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.RAC WA</lx:path>
            <lx:label>RAC WA</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RACQ Insurance">
        <xs:annotation>
          <xs:documentation>RACQ Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.RACQ Insurance</lx:path>
            <lx:label>RACQ Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RACV">
        <xs:annotation>
          <xs:documentation>RACV</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.RACV</lx:path>
            <lx:label>RACV</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Real Insurance">
        <xs:annotation>
          <xs:documentation>Real Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Real Insurance</lx:path>
            <lx:label>Real Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Redwood Advisory">
        <xs:annotation>
          <xs:documentation>Redwood Advisory</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Redwood Advisory</lx:path>
            <lx:label>Redwood Advisory</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SGIC">
        <xs:annotation>
          <xs:documentation>SGIC</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.SGIC</lx:path>
            <lx:label>SGIC</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SGIO">
        <xs:annotation>
          <xs:documentation>SGIO</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.SGIO</lx:path>
            <lx:label>SGIO</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Shannons">
        <xs:annotation>
          <xs:documentation>Shannons</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Shannons</lx:path>
            <lx:label>Shannons</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Smart Business Insurance Pty Ltd.">
        <xs:annotation>
          <xs:documentation>Smart Business Insurance Pty Ltd.</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Smart Business Insurance Pty Ltd[-]</lx:path>
            <lx:label>Smart Business Insurance Pty Ltd.</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Southern Cross Healthcare Group">
        <xs:annotation>
          <xs:documentation>Southern Cross Healthcare Group</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Southern Cross Healthcare Group</lx:path>
            <lx:label>Southern Cross Healthcare Group</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sovereign">
        <xs:annotation>
          <xs:documentation>Sovereign</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Sovereign</lx:path>
            <lx:label>Sovereign</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sun Alliance">
        <xs:annotation>
          <xs:documentation>Sun Alliance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Sun Alliance</lx:path>
            <lx:label>Sun Alliance</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Suncorp">
        <xs:annotation>
          <xs:documentation>Suncorp</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Suncorp</lx:path>
            <lx:label>Suncorp</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Super Cheap Auto Group">
        <xs:annotation>
          <xs:documentation>Super Cheap Auto Group</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Super Cheap Auto Group</lx:path>
            <lx:label>Super Cheap Auto Group</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Superannuation Warehouse">
        <xs:annotation>
          <xs:documentation>Superannuation Warehouse</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Superannuation Warehouse</lx:path>
            <lx:label>Superannuation Warehouse</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Swann">
        <xs:annotation>
          <xs:documentation>Swann</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Swann</lx:path>
            <lx:label>Swann</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TAL (Tower Australia)">
        <xs:annotation>
          <xs:documentation>TAL (Tower Australia)</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.TAL (Tower Australia)</lx:path>
            <lx:label>TAL (Tower Australia)</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Terri Scheer">
        <xs:annotation>
          <xs:documentation>Terri Scheer</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Terri Scheer</lx:path>
            <lx:label>Terri Scheer</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TIO">
        <xs:annotation>
          <xs:documentation>TIO</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.TIO</lx:path>
            <lx:label>TIO</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tower Insurance">
        <xs:annotation>
          <xs:documentation>Tower Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Tower Insurance</lx:path>
            <lx:label>Tower Insurance</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="TradeCover">
        <xs:annotation>
          <xs:documentation>TradeCover</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.TradeCover</lx:path>
            <lx:label>TradeCover</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="VACC">
        <xs:annotation>
          <xs:documentation>VACC</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.VACC</lx:path>
            <lx:label>VACC</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vero">
        <xs:annotation>
          <xs:documentation>Vero</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Vero</lx:path>
            <lx:label>Vero</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Western QBE">
        <xs:annotation>
          <xs:documentation>Western QBE</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Western QBE</lx:path>
            <lx:label>Western QBE</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Westfarmers">
        <xs:annotation>
          <xs:documentation>Westfarmers</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Westfarmers</lx:path>
            <lx:label>Westfarmers</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Westpac Insurance Services">
        <xs:annotation>
          <xs:documentation>Westpac Insurance Services</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Westpac Insurance Services</lx:path>
            <lx:label>Westpac Insurance Services</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="World Marine and General">
        <xs:annotation>
          <xs:documentation>World Marine and General</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.World Marine and General</lx:path>
            <lx:label>World Marine and General</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="You.insured">
        <xs:annotation>
          <xs:documentation>You.insured</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.You[-]insured</lx:path>
            <lx:label>You.insured</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Zurich Australia">
        <xs:annotation>
          <xs:documentation>Zurich Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>insurerList.Zurich Australia</lx:path>
            <lx:label>Zurich Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="interestChargeFrequencyList">
    <xs:annotation>
      <xs:documentation>interestChargeFrequencyList</xs:documentation>
      <xs:appinfo>
        <lx:path>interestChargeFrequencyList</lx:path>
        <lx:label>interestChargeFrequencyList</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Daily">
        <xs:annotation>
          <xs:documentation>Daily</xs:documentation>
          <xs:appinfo>
            <lx:path>interestChargeFrequencyList.Daily</lx:path>
            <lx:label>Daily</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Every Two Months">
        <xs:annotation>
          <xs:documentation>Every Two Months</xs:documentation>
          <xs:appinfo>
            <lx:path>interestChargeFrequencyList.Every Two Months</lx:path>
            <lx:label>Every Two Months</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fortnightly">
        <xs:annotation>
          <xs:documentation>Fortnightly</xs:documentation>
          <xs:appinfo>
            <lx:path>interestChargeFrequencyList.Fortnightly</lx:path>
            <lx:label>Fortnightly</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Half Yearly">
        <xs:annotation>
          <xs:documentation>Half Yearly</xs:documentation>
          <xs:appinfo>
            <lx:path>interestChargeFrequencyList.Half Yearly</lx:path>
            <lx:label>Half Yearly</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Monthly">
        <xs:annotation>
          <xs:documentation>Monthly</xs:documentation>
          <xs:appinfo>
            <lx:path>interestChargeFrequencyList.Monthly</lx:path>
            <lx:label>Monthly</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Quarterly">
        <xs:annotation>
          <xs:documentation>Quarterly</xs:documentation>
          <xs:appinfo>
            <lx:path>interestChargeFrequencyList.Quarterly</lx:path>
            <lx:label>Quarterly</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Weekly">
        <xs:annotation>
          <xs:documentation>Weekly</xs:documentation>
          <xs:appinfo>
            <lx:path>interestChargeFrequencyList.Weekly</lx:path>
            <lx:label>Weekly</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Yearly">
        <xs:annotation>
          <xs:documentation>Yearly</xs:documentation>
          <xs:appinfo>
            <lx:path>interestChargeFrequencyList.Yearly</lx:path>
            <lx:label>Yearly</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="interestOnlyReasonList">
    <xs:annotation>
      <xs:documentation>Where the credit application includes an interest only period, this list contains options that can be used to prompt the applicant for reasons (with an option of 'other').</xs:documentation>
      <xs:appinfo>
        <lx:path>interestOnlyReasonList</lx:path>
        <lx:label>Interest Only Reason List</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Create funds for investment purposes">
        <xs:annotation>
          <xs:documentation>This application contains an interest only period to assist in creating funds for investment purposes.</xs:documentation>
          <xs:appinfo>
            <lx:path>interestOnlyReasonList.Create funds for investment purposes</lx:path>
            <lx:label>To create funds for investment purposes.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Large non-recurring expenses">
        <xs:annotation>
          <xs:documentation>This application contains an interest only period to accommodate an anticipated large non-recurring expenses.</xs:documentation>
          <xs:appinfo>
            <lx:path>interestOnlyReasonList.Large non-recurring expenses</lx:path>
            <lx:label>To accommodate anticipated large non-recurring expenses.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Maximise cash flow">
        <xs:annotation>
          <xs:documentation>This application contains an interest only period to assist in maximising cash flow.</xs:documentation>
          <xs:appinfo>
            <lx:path>interestOnlyReasonList.Maximise cash flow</lx:path>
            <lx:label>To assist in maximising cash flow.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Minimise Cash Flow Impact During Construction">
        <xs:annotation>
          <xs:documentation>This application contains an interest only period to minimise cash flow impact during the construction period on the security property.</xs:documentation>
          <xs:appinfo>
            <lx:path>interestOnlyReasonList.Minimise Cash Flow Impact During Construction</lx:path>
            <lx:label>Minimise Cash Flow Impact During Construction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Minimise Cash Flow Impact During the Bridging Period">
        <xs:annotation>
          <xs:documentation>This application contains an interest only period to minimise cash flow impact during the bridging finance period.</xs:documentation>
          <xs:appinfo>
            <lx:path>interestOnlyReasonList.Minimise Cash Flow Impact During the Bridging Period</lx:path>
            <lx:label>Minimise Cash Flow Impact During the Bridging Period</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other reason">
        <xs:annotation>
          <xs:documentation>This application contains an interest only period for reasons not otherwise listed above.</xs:documentation>
          <xs:appinfo>
            <lx:path>interestOnlyReasonList.Other reason</lx:path>
            <lx:label>For other reasons than those described above</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Plan to convert to investment property">
        <xs:annotation>
          <xs:documentation>This application contains an interest only period because it is intended to convert the property to an investment property in the future.</xs:documentation>
          <xs:appinfo>
            <lx:path>interestOnlyReasonList.Plan to convert to investment property</lx:path>
            <lx:label>It is intended that the property will be converted to an investment property in the future.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Principal reductions in an offset facility">
        <xs:annotation>
          <xs:documentation>This application contains an interest only period as it is intended that principal reductions will be made into an offset facility.</xs:documentation>
          <xs:appinfo>
            <lx:path>interestOnlyReasonList.Principal reductions in an offset facility</lx:path>
            <lx:label>It is intended that principal reductions will be made into an offset facility.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Recommendation From Independent Financial Adviser">
        <xs:annotation>
          <xs:documentation>This application contains an interest only period because of a recommendation from an independent financial adviser.</xs:documentation>
          <xs:appinfo>
            <lx:path>interestOnlyReasonList.Recommendation From Independent Financial Adviser</lx:path>
            <lx:label>Recommendation From Independent Financial Adviser</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Taxation, financial or accounting reasons">
        <xs:annotation>
          <xs:documentation>This application contains an interest only period for taxation, financial or accounting reasons.</xs:documentation>
          <xs:appinfo>
            <lx:path>interestOnlyReasonList.Taxation, financial or accounting reasons</lx:path>
            <lx:label>For Taxation, Financial or Accounting reasons.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Temporary reduction in income">
        <xs:annotation>
          <xs:documentation>This application contains an interest only period to accommodate a temporary reduction in income.</xs:documentation>
          <xs:appinfo>
            <lx:path>interestOnlyReasonList.Temporary reduction in income</lx:path>
            <lx:label>To accommodate a temporary reduction in income.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Variable income">
        <xs:annotation>
          <xs:documentation>This application contains an interest only period as income is of a variable nature.</xs:documentation>
          <xs:appinfo>
            <lx:path>interestOnlyReasonList.Variable income</lx:path>
            <lx:label>To accommodate income of a variable nature.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="interestRateTypeList">
    <xs:annotation>
      <xs:documentation>interestRateTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>interestRateTypeList</lx:path>
        <lx:label>Interest Rate Type List</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Basic Variable">
        <xs:annotation>
          <xs:documentation>Basic Variable</xs:documentation>
          <xs:appinfo>
            <lx:path>interestRateTypeList.Basic Variable</lx:path>
            <lx:label>Basic Variable</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fixed">
        <xs:annotation>
          <xs:documentation>Fixed</xs:documentation>
          <xs:appinfo>
            <lx:path>interestRateTypeList.Fixed</lx:path>
            <lx:label>Fixed</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Split">
        <xs:annotation>
          <xs:documentation>Split</xs:documentation>
          <xs:appinfo>
            <lx:path>interestRateTypeList.Split</lx:path>
            <lx:label>Split</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Standard Variable">
        <xs:annotation>
          <xs:documentation>Standard Variable</xs:documentation>
          <xs:appinfo>
            <lx:path>interestRateTypeList.Standard Variable</lx:path>
            <lx:label>Standard Variable</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="interestTypeList">
    <xs:annotation>
      <xs:documentation>interestTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>interestTypeList</lx:path>
        <lx:label>interestTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Fixed Rate">
        <xs:annotation>
          <xs:documentation>The term specified is the fixed rate term of the loan, being the period during which a fixed interest rate applies.</xs:documentation>
          <xs:appinfo>
            <lx:path>interestTypeList.Fixed Rate</lx:path>
            <lx:label>The term specified is the fixed rate term of the loan, being the period during which a fixed interest rate applies.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Variable">
        <xs:annotation>
          <xs:documentation>The term specified is the variable rate term of the loan, being the period during which a variable interest rate applies.</xs:documentation>
          <xs:appinfo>
            <lx:path>interestTypeList.Variable</lx:path>
            <lx:label>The term specified is the variable rate term of the loan, being the period during which a variable interest rate applies.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="interiorWallLiningMaterialList">
    <xs:annotation>
      <xs:documentation>A list of interior wall lining material types.</xs:documentation>
      <xs:appinfo>
        <lx:path>interiorWallLiningMaterialList</lx:path>
        <lx:label>interiorWallLiningMaterialList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Brick">
        <xs:annotation>
          <xs:documentation>Brick</xs:documentation>
          <xs:appinfo>
            <lx:path>interiorWallLiningMaterialList.Brick</lx:path>
            <lx:label>Brick</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Concrete">
        <xs:annotation>
          <xs:documentation>Concrete</xs:documentation>
          <xs:appinfo>
            <lx:path>interiorWallLiningMaterialList.Concrete</lx:path>
            <lx:label>Concrete</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>interiorWallLiningMaterialList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Plaster">
        <xs:annotation>
          <xs:documentation>Plaster</xs:documentation>
          <xs:appinfo>
            <lx:path>interiorWallLiningMaterialList.Plaster</lx:path>
            <lx:label>Plaster</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Plasterboard">
        <xs:annotation>
          <xs:documentation>Plasterboard</xs:documentation>
          <xs:appinfo>
            <lx:path>interiorWallLiningMaterialList.Plasterboard</lx:path>
            <lx:label>Plasterboard</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Render">
        <xs:annotation>
          <xs:documentation>Render</xs:documentation>
          <xs:appinfo>
            <lx:path>interiorWallLiningMaterialList.Render</lx:path>
            <lx:label>Render</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tile">
        <xs:annotation>
          <xs:documentation>Tile</xs:documentation>
          <xs:appinfo>
            <lx:path>interiorWallLiningMaterialList.Tile</lx:path>
            <lx:label>Tile</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wood Cladding">
        <xs:annotation>
          <xs:documentation>Wood Cladding</xs:documentation>
          <xs:appinfo>
            <lx:path>interiorWallLiningMaterialList.Wood Cladding</lx:path>
            <lx:label>Wood Cladding</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="itAndAVEquipmentTypeList">
    <xs:annotation>
      <xs:documentation>itAndAVEquipmentTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>itAndAVEquipmentTypeList</lx:path>
        <lx:label>itAndAVEquipmentTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Audio Visual">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a future release. Audio Visual</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Audio Visual</lx:path>
            <lx:label>Audio Visual</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Client Computing">
        <xs:annotation>
          <xs:documentation>Client Computing</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Client Computing</lx:path>
            <lx:label>Client Computing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gaming">
        <xs:annotation>
          <xs:documentation>Gaming</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Gaming</lx:path>
            <lx:label>Gaming</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HiFi">
        <xs:annotation>
          <xs:documentation>HiFi</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.HiFi</lx:path>
            <lx:label>HiFi</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Network Infrastructure">
        <xs:annotation>
          <xs:documentation>Network Infrastructure</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Network Infrastructure</lx:path>
            <lx:label>Network Infrastructure</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Peripherals">
        <xs:annotation>
          <xs:documentation>Peripherals</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Peripherals</lx:path>
            <lx:label>Peripherals</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Photography">
        <xs:annotation>
          <xs:documentation>Photography</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Photography</lx:path>
            <lx:label>Photography</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Point of Sale System">
        <xs:annotation>
          <xs:documentation>Point of Sale System</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Point of Sale System</lx:path>
            <lx:label>Point of Sale System</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Portable Audio">
        <xs:annotation>
          <xs:documentation>Portable Audio</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Portable Audio</lx:path>
            <lx:label>Portable Audio</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Hardware and Appliances">
        <xs:annotation>
          <xs:documentation>Security Hardware and Appliances</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Security Hardware and Appliances</lx:path>
            <lx:label>Security Hardware and Appliances</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Server Computing">
        <xs:annotation>
          <xs:documentation>Server Computing</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Server Computing</lx:path>
            <lx:label>Server Computing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Software">
        <xs:annotation>
          <xs:documentation>Software</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Software</lx:path>
            <lx:label>Software</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Storage Infrastructure">
        <xs:annotation>
          <xs:documentation>Storage Infrastructure</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Storage Infrastructure</lx:path>
            <lx:label>Storage Infrastructure</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Surveillance Equipment">
        <xs:annotation>
          <xs:documentation>Surveillance Equipment</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Surveillance Equipment</lx:path>
            <lx:label>Surveillance Equipment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Television">
        <xs:annotation>
          <xs:documentation>Television</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Television</lx:path>
            <lx:label>Television</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Video">
        <xs:annotation>
          <xs:documentation>Video</xs:documentation>
          <xs:appinfo>
            <lx:path>itAndAVEquipmentTypeList.Video</lx:path>
            <lx:label>Video</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="kinRelationshipList">
    <xs:annotation>
      <xs:documentation>kinRelationshipList</xs:documentation>
      <xs:appinfo>
        <lx:path>kinRelationshipList</lx:path>
        <lx:label>kinRelationshipList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Child">
        <xs:annotation>
          <xs:documentation>Next of kin is a child of the applicant.</xs:documentation>
          <xs:appinfo>
            <lx:path>kinRelationshipList.Child</lx:path>
            <lx:label>Next of kin is a child of the applicant.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Friend">
        <xs:annotation>
          <xs:documentation>Friend</xs:documentation>
          <xs:appinfo>
            <lx:path>kinRelationshipList.Friend</lx:path>
            <lx:label>Friend</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Grandparent">
        <xs:annotation>
          <xs:documentation>Grandparent</xs:documentation>
          <xs:appinfo>
            <lx:path>kinRelationshipList.Grandparent</lx:path>
            <lx:label>Next of kin is a grandparent of the applicant.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other Relative">
        <xs:annotation>
          <xs:documentation>Next of kin is a relative other than a parent, child or spouse of the applicant.</xs:documentation>
          <xs:appinfo>
            <lx:path>kinRelationshipList.Other Relative</lx:path>
            <lx:label>Next of kin is a relative other than a parent, child or spouse of the applicant.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Parent">
        <xs:annotation>
          <xs:documentation>Next of kin is a parent of the applicant.</xs:documentation>
          <xs:appinfo>
            <lx:path>kinRelationshipList.Parent</lx:path>
            <lx:label>Next of kin is a parent of the applicant.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sibling">
        <xs:annotation>
          <xs:documentation>Next of kin is a sibling of the applicant.</xs:documentation>
          <xs:appinfo>
            <lx:path>kinRelationshipList.Sibling</lx:path>
            <lx:label>Next of kin is a sibling of the applicant.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Spouse">
        <xs:annotation>
          <xs:documentation>Next of kin is a spouse.</xs:documentation>
          <xs:appinfo>
            <lx:path>kinRelationshipList.Spouse</lx:path>
            <lx:label>Next of kin is a spouse.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="landAreaUnitsList">
    <xs:annotation>
      <xs:documentation>landAreaUnitsList</xs:documentation>
      <xs:appinfo>
        <lx:path>landAreaUnitsList</lx:path>
        <lx:label>landAreaUnitsList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Acres">
        <xs:annotation>
          <xs:documentation>Acres</xs:documentation>
          <xs:appinfo>
            <lx:path>landAreaUnitsList.Acres</lx:path>
            <lx:label>Acres</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hectares">
        <xs:annotation>
          <xs:documentation>Hectares</xs:documentation>
          <xs:appinfo>
            <lx:path>landAreaUnitsList.Hectares</lx:path>
            <lx:label>Hectares</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Square Metres">
        <xs:annotation>
          <xs:documentation>Square Metres</xs:documentation>
          <xs:appinfo>
            <lx:path>landAreaUnitsList.Square Metres</lx:path>
            <lx:label>Square Metres</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="landValueEstimateBasisList">
    <xs:annotation>
      <xs:documentation>landValueEstimateBasisList</xs:documentation>
      <xs:appinfo>
        <lx:path>landValueEstimateBasisList</lx:path>
        <lx:label>landValueEstimateBasisList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Estimated Market Value">
        <xs:annotation>
          <xs:documentation>Estimated Market Value</xs:documentation>
          <xs:appinfo>
            <lx:path>landValueEstimateBasisList.Estimated Market Value</lx:path>
            <lx:label>Estimated Market Value</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase Price">
        <xs:annotation>
          <xs:documentation>Purchase Price</xs:documentation>
          <xs:appinfo>
            <lx:path>landValueEstimateBasisList.Purchase Price</lx:path>
            <lx:label>Purchase Price</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="lendersMortgageInsuranceInstructionList">
    <xs:annotation>
      <xs:documentation>lendersMortgageInsuranceInstructionList</xs:documentation>
      <xs:appinfo>
        <lx:path>lendersMortgageInsuranceInstructionList</lx:path>
        <lx:label>lendersMortgageInsuranceInstructionList</lx:label>
        <li:transactions>LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Additional Loan Application">
        <xs:annotation>
          <xs:documentation>A lenders mortgage insurance application for an increase to a loan that is associated to an LMI that is currently active.</xs:documentation>
          <xs:appinfo>
            <lx:path>lendersMortgageInsuranceInstructionList.Additional Loan Application</lx:path>
            <lx:label>Additional Loan Application</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Change Application">
        <xs:annotation>
          <xs:documentation>A revised lenders mortgage insurance application to an existing LMI application that has already been submitted.</xs:documentation>
          <xs:appinfo>
            <lx:path>lendersMortgageInsuranceInstructionList.Change Application</lx:path>
            <lx:label>Change Application</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="New Application">
        <xs:annotation>
          <xs:documentation>A new lenders mortgage insurance application.</xs:documentation>
          <xs:appinfo>
            <lx:path>lendersMortgageInsuranceInstructionList.New Application</lx:path>
            <lx:label>New Application</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Real Estate Valuation">
        <xs:annotation>
          <xs:documentation>The valuation details for a LMI application which has been conditionally approved, subject to valuation.</xs:documentation>
          <xs:appinfo>
            <lx:path>lendersMortgageInsuranceInstructionList.Real Estate Valuation</lx:path>
            <lx:label>Real Estate Valuation</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="lendingGuaranteeTypeList">
    <xs:annotation>
      <xs:documentation>lendingGuaranteeTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>lendingGuaranteeTypeList</lx:path>
        <lx:label>lendingGuaranteeTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Income and Security">
        <xs:annotation>
          <xs:documentation>Income and Security</xs:documentation>
          <xs:appinfo>
            <lx:path>lendingGuaranteeTypeList.Income and Security</lx:path>
            <lx:label>Income and Security</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Income Only">
        <xs:annotation>
          <xs:documentation>Income Only</xs:documentation>
          <xs:appinfo>
            <lx:path>lendingGuaranteeTypeList.Income Only</lx:path>
            <lx:label>Income Only</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Only">
        <xs:annotation>
          <xs:documentation>Security Only</xs:documentation>
          <xs:appinfo>
            <lx:path>lendingGuaranteeTypeList.Security Only</lx:path>
            <lx:label>Security Only</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="levelTypeList">
    <xs:annotation>
      <xs:documentation>levelTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>levelTypeList</lx:path>
        <lx:label>levelTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Basement">
        <xs:annotation>
          <xs:documentation>Basement</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Basement</lx:path>
            <lx:label>Basement</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Floor">
        <xs:annotation>
          <xs:documentation>Floor</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Floor</lx:path>
            <lx:label>Floor</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ground">
        <xs:annotation>
          <xs:documentation>Ground</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Ground</lx:path>
            <lx:label>Ground</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Level">
        <xs:annotation>
          <xs:documentation>Level</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Level</lx:path>
            <lx:label>Level</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lower Ground Floor">
        <xs:annotation>
          <xs:documentation>Lower Ground Floor</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Lower Ground Floor</lx:path>
            <lx:label>Lower Ground Floor</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lower Level">
        <xs:annotation>
          <xs:documentation>Lower Level</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Lower Level</lx:path>
            <lx:label>Lower Level</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mezzanine">
        <xs:annotation>
          <xs:documentation>Mezzanine</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Mezzanine</lx:path>
            <lx:label>Mezzanine</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Observation Deck">
        <xs:annotation>
          <xs:documentation>Observation Deck</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Observation Deck</lx:path>
            <lx:label>Observation Deck</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Parking">
        <xs:annotation>
          <xs:documentation>Parking</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Parking</lx:path>
            <lx:label>Parking</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Penthouse">
        <xs:annotation>
          <xs:documentation>Penthouse</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Penthouse</lx:path>
            <lx:label>Penthouse</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Platform">
        <xs:annotation>
          <xs:documentation>Platform</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Platform</lx:path>
            <lx:label>Platform</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Podium">
        <xs:annotation>
          <xs:documentation>Podium</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Podium</lx:path>
            <lx:label>Podium</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rooftop">
        <xs:annotation>
          <xs:documentation>Rooftop</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Rooftop</lx:path>
            <lx:label>Rooftop</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sub-basement">
        <xs:annotation>
          <xs:documentation>Sub-basement</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Sub-basement</lx:path>
            <lx:label>Sub-basement</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Upper Ground Floor">
        <xs:annotation>
          <xs:documentation>Upper Ground Floor</xs:documentation>
          <xs:appinfo>
            <lx:path>levelTypeList.Upper Ground Floor</lx:path>
            <lx:label>Upper Ground Floor</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="liabilityTypeList">
    <xs:annotation>
      <xs:documentation>liabilityTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>liabilityTypeList</lx:path>
        <lx:label>liabilityTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Bridging Finance">
        <xs:annotation>
          <xs:documentation>Liability is a bridging loan. A short term loan designed to allow a borrower to finance the purchase of a new property before selling an existing property. An assessment for a bridging loan involves calculation of peak debt and end debt.</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Bridging Finance</lx:path>
            <lx:label>Bridging Finance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Car Loan">
        <xs:annotation>
          <xs:documentation>Liability is a car loan</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Car Loan</lx:path>
            <lx:label>Liability is a car loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Commercial Bill">
        <xs:annotation>
          <xs:documentation>Liability is a commercial bill, typically associated with business lending or high-end investment lending</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Commercial Bill</lx:path>
            <lx:label>Liability is a commercial bill, typically associated with business lending or high-end investment lending</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Contingent Liability">
        <xs:annotation>
          <xs:documentation>Liability is has not yet crystallised but may arise in future on the happening of certain events or contingencies</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Contingent Liability</lx:path>
            <lx:label>Liability is has not yet crystallised but may arise in future on the happening of certain events or contingencies</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Card">
        <xs:annotation>
          <xs:documentation>Liability is a credit card</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Credit Card</lx:path>
            <lx:label>Liability is a credit card</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="HECS-HELP">
        <xs:annotation>
          <xs:documentation>Liability is a debt under the Commonwealth Higher Education Contribution Scheme</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.HECS-HELP</lx:path>
            <lx:label>Liability is a debt under the Commonwealth Higher Education Contribution Scheme</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hire Purchase">
        <xs:annotation>
          <xs:documentation>Liability is owed under a hire purchase agreement</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Hire Purchase</lx:path>
            <lx:label>Liability is owed under a hire purchase agreement</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Invoice Financing Loan">
        <xs:annotation>
          <xs:documentation>An invoice financing loan secured by one or more outstanding sales invoices owed to the borrower.</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Invoice Financing Loan</lx:path>
            <lx:label>Invoice Financing Loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lease">
        <xs:annotation>
          <xs:documentation>Liability is owed under a finance lease, including for a motor vehicle or equipment</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Lease</lx:path>
            <lx:label>Liability is owed under a finance lease, including for a motor vehicle or equipment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Line of Credit">
        <xs:annotation>
          <xs:documentation>Liability has a limit up to which the borrower may draw, but which has no regular principal repayment obligations; as opposed to a Term Loan</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Line of Credit</lx:path>
            <lx:label>Liability has a limit up to which the borrower may draw, but which has no regular principal repayment obligations; as opposed to a Term Loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan as Guarantor">
        <xs:annotation>
          <xs:documentation>Liability is the contingent liability of a guarantor for guaranteed obligations of a third party to a lender</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Loan as Guarantor</lx:path>
            <lx:label>Liability is the contingent liability of a guarantor for guaranteed obligations of a third party to a lender</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Margin Loan">
        <xs:annotation>
          <xs:documentation>Liability which lets you borrow money to invest and uses your shares or managed funds as security.</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Margin Loan</lx:path>
            <lx:label>Liability which lets you borrow money to invest and uses your shares or managed funds as security.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mortgage Loan">
        <xs:annotation>
          <xs:documentation>DEPRECATED and to be deleted in a breaking release on or after May 1 2017. Liability is a loan secured by real property</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Mortgage Loan</lx:path>
            <lx:label>DEPRECATED: Liability is a loan secured by real property</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Liability is not described by one of the other enumerated values in the list</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Other</lx:path>
            <lx:label>Liability is not described by one of the other enumerated values in the list</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other Loan">
        <xs:annotation>
          <xs:documentation>Liability is loan that is not described by the loan types Mortgage Loan, Personal Loan, Term Loan</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Other Loan</lx:path>
            <lx:label>Liability is loan that is not described by the loan types Mortgage Loan, Personal Loan, Term Loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Outstanding Taxation">
        <xs:annotation>
          <xs:documentation>Liability is a debt owed to a public authority for unpaid taxes, including BAS and PAYGW</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Outstanding Taxation</lx:path>
            <lx:label>Liability is a debt owed to a public authority for unpaid taxes, including BAS and PAYGW</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Overdraft">
        <xs:annotation>
          <xs:documentation>Liability is a credit buffer on a deposit account, unsecured, or an extra buffer above the approved limit of a credit facility</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Overdraft</lx:path>
            <lx:label>Liability is a credit buffer on a deposit account, unsecured, or an extra buffer above the approved limit of a credit facility</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Personal Loan">
        <xs:annotation>
          <xs:documentation>Liability is a personal loan, typically unsecured and for personal purposes such as funding travel or the purchase of a car</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Personal Loan</lx:path>
            <lx:label>Liability is a personal loan, typically unsecured and for personal purposes such as funding travel or the purchase of a car</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Store Card">
        <xs:annotation>
          <xs:documentation>Liability is a store card</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Store Card</lx:path>
            <lx:label>Liability is a store card</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Student Loan">
        <xs:annotation>
          <xs:documentation>Liability is a student loan</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Student Loan</lx:path>
            <lx:label>Student Loan</lx:label>
            <li:transactions>CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Term Loan">
        <xs:annotation>
          <xs:documentation>Liability is a loan that has a term, as opposed to a Line Of Credit</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Term Loan</lx:path>
            <lx:label>Liability is a loan that has a term, as opposed to a Line Of Credit</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trade Finance Loan">
        <xs:annotation>
          <xs:documentation>A trade finance loan for financing domestic and international trade.</xs:documentation>
          <xs:appinfo>
            <lx:path>liabilityTypeList.Trade Finance Loan</lx:path>
            <lx:label>Trade Finance Loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="licenceTypeList">
    <xs:annotation>
      <xs:documentation>licenceTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>licenceTypeList</lx:path>
        <lx:label>licenceTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,SVC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="ACL">
        <xs:annotation>
          <xs:documentation>ACL</xs:documentation>
          <xs:appinfo>
            <lx:path>licenceTypeList.ACL</lx:path>
            <lx:label>ACL</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="CRN">
        <xs:annotation>
          <xs:documentation>CRN</xs:documentation>
          <xs:appinfo>
            <lx:path>licenceTypeList.CRN</lx:path>
            <lx:label>CRN</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="livingExpenseCategoryList">
    <xs:annotation>
      <xs:documentation>List of living expense categories, as agreed by the Living Expenses Working Group.</xs:documentation>
      <xs:appinfo>
        <lx:path>livingExpenseCategoryList</lx:path>
        <lx:label>livingExpenseCategoryList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Body Corporate Fees, Strata Fees and Land Tax on Owner Occupied Principal Place of Residence">
        <xs:annotation>
          <xs:documentation>Body corporate fees, strata fees and land tax on owner occupied principal place of residence. Excludes investment properties, secondary residences, and properties maintained for other purposes such as holiday properties or residences maintained for parents or children.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Body Corporate Fees, Strata Fees and Land Tax on Owner Occupied Principal Place of Residence</lx:path>
            <lx:label>Body Corporate Fees, Strata Fees and Land Tax on Owner Occupied Principal Place of Residence</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Childcare">
        <xs:annotation>
          <xs:documentation>Childcare including nannies and non-compulsory pre-school.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Childcare</lx:path>
            <lx:label>Childcare</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Clothing and personal care">
        <xs:annotation>
          <xs:documentation>Clothing, footwear, personal care products, cosmetics, hair services and accessories (including travel goods, handbags, umbrellas, wallets, and baby-goods) for men, women, children and infants, including purchases, repairs and alterations.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Clothing and personal care</lx:path>
            <lx:label>Clothing and personal care</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Education">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a future release. Public and private education fees and associated costs (preschool, primary, secondary and tertiary) including books and uniforms etc.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Education</lx:path>
            <lx:label>Public and private education fees and associated costs (preschool, primary, secondary and tertiary) including books and uniforms etc.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="General Basic Insurances">
        <xs:annotation>
          <xs:documentation>General basic insurances include ambulance, car (not recreation vehicles), travel, and personal belongings insurance. Excludes property (house and contents), health, sickness and personal accident, life, income protection and pet insurance.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.General Basic Insurances</lx:path>
            <lx:label>General Basic Insurances</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Groceries">
        <xs:annotation>
          <xs:documentation>Groceries (food and non-alcoholic beverages), toiletries and cleaning products.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Groceries</lx:path>
            <lx:label>Groceries</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Health Insurance">
        <xs:annotation>
          <xs:documentation>Health insurance including hospital, medical and dental insurance (excludes separate ambulance insurance).</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Health Insurance</lx:path>
            <lx:label>Health Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Higher Education, Vocational Training and Professional Fees">
        <xs:annotation>
          <xs:documentation>Higher education and vocational training fees (including TAFE, business college, drama, music, dance) excluding HECS, and professional fees (union dues, professional association subscriptions, legal, accountant and tax agent fees).</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Higher Education, Vocational Training and Professional Fees</lx:path>
            <lx:label>Higher Education, Vocational Training and Professional Fees</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Insurance">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a future release. All insurance including health, home and contents, motor vehicle, life, income protection.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Insurance</lx:path>
            <lx:label>Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Investment Property Running Costs">
        <xs:annotation>
          <xs:documentation>Costs associated with any property owned for investment purposes. Includes body corporate fees, strata fees, land tax, property management fees, rates, water, sewage, repairs and maintenance, electricity, gas, heating oil and wood for fuel, purchase, hire and repair of household and outdoor furnishings and equipment, furnishings, floor coverings, linen, tableware, gardening tools, power tools, electrical, lights, trees, shrubs, plants, pool chemicals, landlords insurance, home insurance, contents insurance and appliance insurance.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Investment Property Running Costs</lx:path>
            <lx:label>Investment Property Running Costs</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Investment property utilities, rates and related costs">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a future release. Housing and property expenses on investment property including rates, taxes, levies, body corporate and strata fees, repairs and maintenance, other household items and utilities (excluding insurance, telephone, internet and pay TV as they are categorised separately).</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Investment property utilities, rates and related costs</lx:path>
            <lx:label>Investment property utilities, rates and related costs (excluding insurance)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Medical and health">
        <xs:annotation>
          <xs:documentation>Medical and health care services including GPs, specialists, physiotherapists, chiropractors, and opticians. Medicines, pharmaceuticals, prescriptions, glasses, purchase or hire of therapeutic appliances and equipment, hospital charges, nursing home charges. Excludes health insurance and ambulance insurance.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Medical and health</lx:path>
            <lx:label>Medical and health</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other ongoing or recurring items not included within the other categories including, but not limited to pet insurance, insurance not elsewhere classified, registration, insurance and operation of recreational vehicles (such as caravans, trailers, motorcycles, boats and aircraft), overseas holidays, gifts, jewellery, household services including cleaning, gardening, housekeeping, security and pest control services.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Owner occupied property utilities, rates and related costs">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a future release. Housing and property expenses on owner occupied property including rates, taxes, levies, body corporate and strata fees, repairs and maintenance, other household items and utilities (excluding insurance, telephone, internet and pay TV as they are categorised separately).</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Owner occupied property utilities, rates and related costs</lx:path>
            <lx:label>Owner occupied property utilities, rates and related costs (excluding insurance)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pet Care">
        <xs:annotation>
          <xs:documentation>Pet food, grooming, minding services, care products, health products, veterinarian fees.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Pet Care</lx:path>
            <lx:label>Pet Care</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Primary Residence Running Costs">
        <xs:annotation>
          <xs:documentation>Costs associated with the primary residence, either owned or rented. Includes rates, water, sewage, repairs and maintenance, electricity, gas, heating oil and wood for fuel, purchase, hire and repair of household and outdoor furnishings and equipment, furnishings, floor coverings, linen, tableware, gardening tools, power tools, electrical, lights, trees, shrubs, plants, pool chemicals, home insurance, contents insurance and appliance insurance. Excludes body corporate fees, strata fees and land tax.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Primary Residence Running Costs</lx:path>
            <lx:label>Primary Residence Running Costs</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Private Schooling and Tuition">
        <xs:annotation>
          <xs:documentation>Tuition fees, school fees and sports fees for private schooling, including independent schools (Catholic or non-Catholic) and private tuition including compulsory age kindergarten/reception/pre-primary/prep.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Private Schooling and Tuition</lx:path>
            <lx:label>Private Schooling and Tuition</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Public or Government Primary and Secondary Education">
        <xs:annotation>
          <xs:documentation>Public or government, primary and secondary school tuition, school and sports fees including compulsory kindergarten/reception/pre-primary/prep.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Public or Government Primary and Secondary Education</lx:path>
            <lx:label>Public or Government Primary and Secondary Education</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Recreation and entertainment">
        <xs:annotation>
          <xs:documentation>Meals in restaurants, hotels and clubs, fast food and takeaway including coffee. Alcoholic beverages and tobacco products. Purchase, hire, repair of recreational personal belongings including electronics, computers, games consoles, AV equipment, cameras. Toys, hobbies, stationery, arts, crafts, sports, camping, fishing and recreation equipment including musical instruments, accessories and lessons. Non-subscription TV, video, games, and audio content, books, newspapers, magazines (including physical material and digital content). Recreation, gym, and health and fitness studios fees, court hire, sports lessons, admission fees and tickets (sports, theatre, gallery, museum, music events). Domestic holidays (fares, fuel, accommodation, theme parks, zoos tours). Recreational gambling, including lottery-type games, scratch cards, poker machines, club and casino gaming, and sports betting. Excludes travel insurance.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Recreation and entertainment</lx:path>
            <lx:label>Recreation and entertainment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Secondary Residence Running Costs">
        <xs:annotation>
          <xs:documentation>Costs associated with any secondary residence, either rented or owned for non-investment purposes, such as a holiday property that is not rented to generate income, or a property that family members (parents or children) are allowed to live in rent-free. Includes body corporate fees, strata fees, land tax, property management fees, rates, water, sewage, repairs and maintenance, electricity, gas, heating oil and wood for fuel, purchase, hire and repair of household and outdoor furnishings and equipment, furnishings, floor coverings, linen, tableware, gardening tools, power tools, electrical, lights, trees, shrubs, plants, pool chemicals, landlords insurance, home insurance, contents insurance and appliance insurance.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Secondary Residence Running Costs</lx:path>
            <lx:label>Secondary Residence Running Costs</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sickness and Personal Accident Insurance, Life Insurance">
        <xs:annotation>
          <xs:documentation>Sickness and personal accident insurance, life insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Sickness and Personal Accident Insurance, Life Insurance</lx:path>
            <lx:label>Sickness and Personal Accident Insurance, Life Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Telephone, internet, pay TV and media streaming subscriptions">
        <xs:annotation>
          <xs:documentation>Telephone accounts (home and mobile), internet, pay TV and media streaming subscriptions (such as Netflix and Spotify).</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Telephone, internet, pay TV and media streaming subscriptions</lx:path>
            <lx:label>Telephone, internet, pay TV and media streaming subscriptions</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Transport">
        <xs:annotation>
          <xs:documentation>Public transport, taxis, ride-sharing and non-holiday domestic airfares (excludes overseas airfares). Running costs for essential vehicles including fuel, servicing, repairs, parking, tolls (excludes insurance) and registration of motor vehicles (excludes recreational vehicles such as caravans, trailers, motorcycles, boats and aircraft).</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Transport</lx:path>
            <lx:label>Transport</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Voluntary Superannuation Contributions">
        <xs:annotation>
          <xs:documentation>Superannuation contributions that the applicants choose to make into their own Superannuation schemes.</xs:documentation>
          <xs:appinfo>
            <lx:path>livingExpenseCategoryList.Voluntary Superannuation Contributions</lx:path>
            <lx:label>Voluntary Superannuation Contributions</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="loanPaymentScheduleTypeList">
    <xs:annotation>
      <xs:documentation>A list of types of loan payment schedules.</xs:documentation>
      <xs:appinfo>
        <lx:path>loanPaymentScheduleTypeList</lx:path>
        <lx:label>loanPaymentScheduleTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Even Principal Payments">
        <xs:annotation>
          <xs:documentation>The size of the principal payment is the same for every payment.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanPaymentScheduleTypeList.Even Principal Payments</lx:path>
            <lx:label>Even Principal Payments</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Even Total Payments">
        <xs:annotation>
          <xs:documentation>The size of the total payment (principal plus interest) is the same for every payment.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanPaymentScheduleTypeList.Even Total Payments</lx:path>
            <lx:label>Even Total Payments</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Half the Monthly Amount Paid in Fortnightly Payments">
        <xs:annotation>
          <xs:documentation>The repayment will be half of the monthly repayment amount and will be paid every two weeks.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanPaymentScheduleTypeList.Half the Monthly Amount Paid in Fortnightly Payments</lx:path>
            <lx:label>Half the Monthly Amount Paid in Fortnightly Payments</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Minimum Payments">
        <xs:annotation>
          <xs:documentation>The repayment will be the minimum amount that must be paid.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanPaymentScheduleTypeList.Minimum Payments</lx:path>
            <lx:label>Minimum Payments</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Outstanding Balance Based Payments">
        <xs:annotation>
          <xs:documentation>The repayment amount is based on the outstanding account balance amount.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanPaymentScheduleTypeList.Outstanding Balance Based Payments</lx:path>
            <lx:label>Outstanding Balance Based Payments</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="loanTermUnitsList">
    <xs:annotation>
      <xs:documentation>loanTermUnitsList</xs:documentation>
      <xs:appinfo>
        <lx:path>loanTermUnitsList</lx:path>
        <lx:label>loanTermUnitsList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Days">
        <xs:annotation>
          <xs:documentation>The loan term specified is in days.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTermUnitsList.Days</lx:path>
            <lx:label>The loan term specified is in days.</lx:label>
            <li:transactions>CAL,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Months">
        <xs:annotation>
          <xs:documentation>The loan term specified is in months.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTermUnitsList.Months</lx:path>
            <lx:label>The loan term specified is in months.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Weeks">
        <xs:annotation>
          <xs:documentation>The loan term specified is in weeks.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTermUnitsList.Weeks</lx:path>
            <lx:label>The loan term specified is in weeks.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Years">
        <xs:annotation>
          <xs:documentation>The loan term specified is in years.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTermUnitsList.Years</lx:path>
            <lx:label>The loan term specified is in years.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="loanTypeList">
    <xs:annotation>
      <xs:documentation>A list of credit facility types that are valid for those facilities described in the 'LoanDetails' element.</xs:documentation>
      <xs:appinfo>
        <lx:path>loanTypeList</lx:path>
        <lx:label>List of Credit Facility Types</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Bridging Finance">
        <xs:annotation>
          <xs:documentation>A short term loan designed to allow a borrower to finance the purchase of a new property before selling an existing property. An assessment for a bridging loan involves calculation of peak debt and end debt.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Bridging Finance</lx:path>
            <lx:label>Bridging Finance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Chattel Mortgage">
        <xs:annotation>
          <xs:documentation>A finance company lends money to the customer to purchase movable equipment (the 'chattel'), in return for regular repayments. The customer takes ownership upon purchase, and an ASIC-registered Charge provides loan security. At term, the customer can either pay any residual value (gaining clear title via removal of the Charge), trade in the asset, or refinance the residual value.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Chattel Mortgage</lx:path>
            <lx:label>Chattel Mortgage</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Commercial Bill">
        <xs:annotation>
          <xs:documentation>Commercial Bill</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Commercial Bill</lx:path>
            <lx:label>Commercial Bill</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Commercial Hire Purchase">
        <xs:annotation>
          <xs:documentation>The lessor secures and pays for an asset on the lessee's behalf with an offer to hire it back to the lessee for an agreed term. The lessor retains ownership of the asset until the lessee pays it off on a principal plus interest basis.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Commercial Hire Purchase</lx:path>
            <lx:label>Commercial Hire Purchase</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Card">
        <xs:annotation>
          <xs:documentation>A type of unsecured revolving credit facility.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Credit Card</lx:path>
            <lx:label>A type of unsecured revolving credit facility.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Finance Lease">
        <xs:annotation>
          <xs:documentation>An arrangement where a finance company purchases an asset, and use of the asset is granted to the customer through a series of payments until the end of term, when the customer can acquire ownership by paying the residual value.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Finance Lease</lx:path>
            <lx:label>Finance Lease</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Invoice Financing Loan">
        <xs:annotation>
          <xs:documentation>An invoice financing loan secured by one or more outstanding sales invoices owed to the borrower.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Invoice Financing Loan</lx:path>
            <lx:label>Invoice Financing Loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lease">
        <xs:annotation>
          <xs:documentation>A type of rental arrangement other than Operating Lease or Finance Lease (which are each are listed separately).</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Lease</lx:path>
            <lx:label>A type of rental arrangement.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Line of Credit">
        <xs:annotation>
          <xs:documentation>A revolving credit facility.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Line of Credit</lx:path>
            <lx:label>A revolving credit facility.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Margin Loan">
        <xs:annotation>
          <xs:documentation>Margin Loan</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Margin Loan</lx:path>
            <lx:label>Margin Loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mortgage Loan">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a future release. Mortgage Loan</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Mortgage Loan</lx:path>
            <lx:label>Mortgage Loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Operating Lease">
        <xs:annotation>
          <xs:documentation>An agreement where an asset's owner permits it's use by the customer for a term that is less than the useful or economic life of the asset in return for regular payments. Ownership rights are not transferred and the customer can return the equipment to the owner at the end of term without further obligation.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Operating Lease</lx:path>
            <lx:label>Operating Lease</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other Loan">
        <xs:annotation>
          <xs:documentation>Other Loan</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Other Loan</lx:path>
            <lx:label>Other Loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Overdraft">
        <xs:annotation>
          <xs:documentation>Overdraft</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Overdraft</lx:path>
            <lx:label>Overdraft</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Personal Loan">
        <xs:annotation>
          <xs:documentation>Personal Loan</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Personal Loan</lx:path>
            <lx:label>Personal Loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Reverse Mortgage">
        <xs:annotation>
          <xs:documentation>An equity release mortgage with accruing interest, repaid on the sale of the property.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Reverse Mortgage</lx:path>
            <lx:label>An equity release mortgage with accruing interest, repaid on the sale of the property.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Term Loan">
        <xs:annotation>
          <xs:documentation>A loan with a specified repayment period.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Term Loan</lx:path>
            <lx:label>A loan with a specified repayment period.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trade Finance Loan">
        <xs:annotation>
          <xs:documentation>A trade finance loan for financing domestic and international trade.</xs:documentation>
          <xs:appinfo>
            <lx:path>loanTypeList.Trade Finance Loan</lx:path>
            <lx:label>Trade Finance Loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="maritalStatusList">
    <xs:annotation>
      <xs:documentation>maritalStatusList</xs:documentation>
      <xs:appinfo>
        <lx:path>maritalStatusList</lx:path>
        <lx:label>maritalStatusList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="De Facto">
        <xs:annotation>
          <xs:documentation>The person is in a de facto relationship.</xs:documentation>
          <xs:appinfo>
            <lx:path>maritalStatusList.De Facto</lx:path>
            <lx:label>The person is in a de facto relationship.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Divorced">
        <xs:annotation>
          <xs:documentation>The person is a divorcee.</xs:documentation>
          <xs:appinfo>
            <lx:path>maritalStatusList.Divorced</lx:path>
            <lx:label>The person is a divorcee.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Married">
        <xs:annotation>
          <xs:documentation>The person is married.</xs:documentation>
          <xs:appinfo>
            <lx:path>maritalStatusList.Married</lx:path>
            <lx:label>The person is married.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Separated">
        <xs:annotation>
          <xs:documentation>The person is separated from their husband/wife.</xs:documentation>
          <xs:appinfo>
            <lx:path>maritalStatusList.Separated</lx:path>
            <lx:label>The person is separated from their husband/wife.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Single">
        <xs:annotation>
          <xs:documentation>The person is single with no financial dependency for or from a previous spouse.</xs:documentation>
          <xs:appinfo>
            <lx:path>maritalStatusList.Single</lx:path>
            <lx:label>The person is single with no financial dependency for or from a previous spouse.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unknown">
        <xs:annotation>
          <xs:documentation>The person's marital status is unknown.</xs:documentation>
          <xs:appinfo>
            <lx:path>maritalStatusList.Unknown</lx:path>
            <lx:label>Unknown</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Widowed">
        <xs:annotation>
          <xs:documentation>The person's husband/wife is deceased.</xs:documentation>
          <xs:appinfo>
            <lx:path>maritalStatusList.Widowed</lx:path>
            <lx:label>The person's husband/wife is deceased.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="masterAgreementStatusList">
    <xs:annotation>
      <xs:documentation>masterAgreementStatusList</xs:documentation>
      <xs:appinfo>
        <lx:path>masterAgreementStatusList</lx:path>
        <lx:label>masterAgreementStatusList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Approved">
        <xs:annotation>
          <xs:documentation>Approved</xs:documentation>
          <xs:appinfo>
            <lx:path>masterAgreementStatusList.Approved</lx:path>
            <lx:label>Approved</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Declined">
        <xs:annotation>
          <xs:documentation>Declined</xs:documentation>
          <xs:appinfo>
            <lx:path>masterAgreementStatusList.Declined</lx:path>
            <lx:label>Declined</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Existing">
        <xs:annotation>
          <xs:documentation>Existing</xs:documentation>
          <xs:appinfo>
            <lx:path>masterAgreementStatusList.Existing</lx:path>
            <lx:label>Existing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="New">
        <xs:annotation>
          <xs:documentation>A request for a new master agreement is being submitted.</xs:documentation>
          <xs:appinfo>
            <lx:path>masterAgreementStatusList.New</lx:path>
            <lx:label>New</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Withdrawn">
        <xs:annotation>
          <xs:documentation>Withdrawn</xs:documentation>
          <xs:appinfo>
            <lx:path>masterAgreementStatusList.Withdrawn</lx:path>
            <lx:label>Withdrawn</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="masterAgreementTypeList">
    <xs:annotation>
      <xs:documentation>masterAgreementTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>masterAgreementTypeList</lx:path>
        <lx:label>masterAgreementTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Reducing">
        <xs:annotation>
          <xs:documentation>The master agreement is a reducing credit facility.</xs:documentation>
          <xs:appinfo>
            <lx:path>masterAgreementTypeList.Reducing</lx:path>
            <lx:label>Reducing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Revolving">
        <xs:annotation>
          <xs:documentation>The master agreement is a revolving line of credit.</xs:documentation>
          <xs:appinfo>
            <lx:path>masterAgreementTypeList.Revolving</lx:path>
            <lx:label>Revolving</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="materialsHandlingAndLiftingSubtypeList">
    <xs:annotation>
      <xs:documentation>materialsHandlingAndLiftingSubtypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>materialsHandlingAndLiftingSubtypeList</lx:path>
        <lx:label>materialsHandlingAndLiftingSubtypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Diesel Forklift">
        <xs:annotation>
          <xs:documentation>Diesel Forklift (subtype of Forklift).</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingSubtypeList.Diesel Forklift</lx:path>
            <lx:label>Diesel Forklift</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Electric Forklift">
        <xs:annotation>
          <xs:documentation>Electric Forklift (subtype of Forklift).</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingSubtypeList.Electric Forklift</lx:path>
            <lx:label>Electric Forklift</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loader Crane">
        <xs:annotation>
          <xs:documentation>Loader Crane (subtype of Crane).</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingSubtypeList.Loader Crane</lx:path>
            <lx:label>Loader Crane</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mobile Crane">
        <xs:annotation>
          <xs:documentation>Mobile Crane (subtype of Crane).</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingSubtypeList.Mobile Crane</lx:path>
            <lx:label>Mobile Crane</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Overhead Crane">
        <xs:annotation>
          <xs:documentation>Overhead Crane (subtype of Crane).</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingSubtypeList.Overhead Crane</lx:path>
            <lx:label>Overhead Crane</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rough Terrain Crane">
        <xs:annotation>
          <xs:documentation>Rough Terrain Crane (subtype of Crane).</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingSubtypeList.Rough Terrain Crane</lx:path>
            <lx:label>Rough Terrain Crane</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Telescopic Crane">
        <xs:annotation>
          <xs:documentation>Telescopic Crane (subtype of Crane).</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingSubtypeList.Telescopic Crane</lx:path>
            <lx:label>Telescopic Crane</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tower Crane">
        <xs:annotation>
          <xs:documentation>Tower Crane (subtype of Crane).</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingSubtypeList.Tower Crane</lx:path>
            <lx:label>Tower Crane</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="materialsHandlingAndLiftingTypeList">
    <xs:annotation>
      <xs:documentation>materialsHandlingAndLiftingTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>materialsHandlingAndLiftingTypeList</lx:path>
        <lx:label>materialsHandlingAndLiftingTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Crane">
        <xs:annotation>
          <xs:documentation>Crane</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingTypeList.Crane</lx:path>
            <lx:label>Crane</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Forklift">
        <xs:annotation>
          <xs:documentation>Forklift</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingTypeList.Forklift</lx:path>
            <lx:label>Forklift</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Jack / Lifting">
        <xs:annotation>
          <xs:documentation>Jack / Lifting</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingTypeList.Jack / Lifting</lx:path>
            <lx:label>Jack / Lifting</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lift / Elevator">
        <xs:annotation>
          <xs:documentation>Lift / Elevator</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingTypeList.Lift / Elevator</lx:path>
            <lx:label>Lift / Elevator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Scissor Lift">
        <xs:annotation>
          <xs:documentation>Scissor Lift</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingTypeList.Scissor Lift</lx:path>
            <lx:label>Scissor Lift</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Storage">
        <xs:annotation>
          <xs:documentation>Storage</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingTypeList.Storage</lx:path>
            <lx:label>Storage</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Telescopic Handler">
        <xs:annotation>
          <xs:documentation>Telescopic Handler</xs:documentation>
          <xs:appinfo>
            <lx:path>materialsHandlingAndLiftingTypeList.Telescopic Handler</lx:path>
            <lx:label>Telescopic Handler</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="medicalEquipmentTypeList">
    <xs:annotation>
      <xs:documentation>medicalEquipmentTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>medicalEquipmentTypeList</lx:path>
        <lx:label>medicalEquipmentTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Diagnostic Equipment">
        <xs:annotation>
          <xs:documentation>Diagnostic Equipment</xs:documentation>
          <xs:appinfo>
            <lx:path>medicalEquipmentTypeList.Diagnostic Equipment</lx:path>
            <lx:label>Diagnostic Equipment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Imaging Equipment">
        <xs:annotation>
          <xs:documentation>X-ray, ultrasound, CT scanner, MRI etc</xs:documentation>
          <xs:appinfo>
            <lx:path>medicalEquipmentTypeList.Imaging Equipment</lx:path>
            <lx:label>X-ray, ultrasound, CT scanner, MRI etc</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Laboratory Equipment">
        <xs:annotation>
          <xs:documentation>Laboratory Equipment</xs:documentation>
          <xs:appinfo>
            <lx:path>medicalEquipmentTypeList.Laboratory Equipment</lx:path>
            <lx:label>Laboratory Equipment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>medicalEquipmentTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Therapy Systems">
        <xs:annotation>
          <xs:documentation>Therapy Systems</xs:documentation>
          <xs:appinfo>
            <lx:path>medicalEquipmentTypeList.Therapy Systems</lx:path>
            <lx:label>Therapy Systems</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="medicareCardColourList">
    <xs:annotation>
      <xs:documentation>medicareCardColourList</xs:documentation>
      <xs:appinfo>
        <lx:path>medicareCardColourList</lx:path>
        <lx:label>medicareCardColourList</lx:label>
        <li:transactions>CAL,CDA,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Blue">
        <xs:annotation>
          <xs:documentation>The Medicare Card is an Interim Medicare Card with a blue background.</xs:documentation>
          <xs:appinfo>
            <lx:path>medicareCardColourList.Blue</lx:path>
            <lx:label>Interim Medicare Card - Blue</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Green">
        <xs:annotation>
          <xs:documentation>The Medicare Card is an Australian Citizen Medicare Card with a green background.</xs:documentation>
          <xs:appinfo>
            <lx:path>medicareCardColourList.Green</lx:path>
            <lx:label>Australian Citizen Medicare Card - Green</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Yellow">
        <xs:annotation>
          <xs:documentation>The Medicare Card is a Reciprocal Health Care Card with a yellow background.</xs:documentation>
          <xs:appinfo>
            <lx:path>medicareCardColourList.Yellow</lx:path>
            <lx:label>Reciprocal Health Care Card - Yellow</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="mitigantFactorList">
    <xs:annotation>
      <xs:documentation>mitigantFactorList</xs:documentation>
      <xs:appinfo>
        <lx:path>mitigantFactorList</lx:path>
        <lx:label>mitigantFactorList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Application Reflects Changes">
        <xs:annotation>
          <xs:documentation>Application Reflects Changes</xs:documentation>
          <xs:appinfo>
            <lx:path>mitigantFactorList.Application Reflects Changes</lx:path>
            <lx:label>Application Reflects Changes</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Co-Applicant's Income">
        <xs:annotation>
          <xs:documentation>Co-Applicant's income.</xs:documentation>
          <xs:appinfo>
            <lx:path>mitigantFactorList.Co-Applicant's Income</lx:path>
            <lx:label>Co-Applicant's Income</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Downsize of Principal Residence">
        <xs:annotation>
          <xs:documentation>Downsize of Principal Residence</xs:documentation>
          <xs:appinfo>
            <lx:path>mitigantFactorList.Downsize of Principal Residence</lx:path>
            <lx:label>Downsize of Principal Residence</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>mitigantFactorList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Reducing Expenditure">
        <xs:annotation>
          <xs:documentation>Reducing Expenditure</xs:documentation>
          <xs:appinfo>
            <lx:path>mitigantFactorList.Reducing Expenditure</lx:path>
            <lx:label>Reducing Expenditure</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sale of Assets">
        <xs:annotation>
          <xs:documentation>Sale of Assets</xs:documentation>
          <xs:appinfo>
            <lx:path>mitigantFactorList.Sale of Assets</lx:path>
            <lx:label>Sale of Assets</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Savings or Superannuation">
        <xs:annotation>
          <xs:documentation>Savings and/or superannuation.</xs:documentation>
          <xs:appinfo>
            <lx:path>mitigantFactorList.Savings or Superannuation</lx:path>
            <lx:label>Savings or Superannuation</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Securing Additional Income">
        <xs:annotation>
          <xs:documentation>Securing Additional Income</xs:documentation>
          <xs:appinfo>
            <lx:path>mitigantFactorList.Securing Additional Income</lx:path>
            <lx:label>Securing Additional Income</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Using Savings">
        <xs:annotation>
          <xs:documentation>Using Savings</xs:documentation>
          <xs:appinfo>
            <lx:path>mitigantFactorList.Using Savings</lx:path>
            <lx:label>Using Savings</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="mmMonthList">
    <xs:annotation>
      <xs:documentation>mmMonthList</xs:documentation>
      <xs:appinfo>
        <lx:path>mmMonthList</lx:path>
        <lx:label>mmMonthList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="01">
        <xs:annotation>
          <xs:documentation>01</xs:documentation>
          <xs:appinfo>
            <lx:path>mmMonthList.01</lx:path>
            <lx:label>01</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="02">
        <xs:annotation>
          <xs:documentation>02</xs:documentation>
          <xs:appinfo>
            <lx:path>mmMonthList.02</lx:path>
            <lx:label>02</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="03">
        <xs:annotation>
          <xs:documentation>03</xs:documentation>
          <xs:appinfo>
            <lx:path>mmMonthList.03</lx:path>
            <lx:label>03</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="04">
        <xs:annotation>
          <xs:documentation>04</xs:documentation>
          <xs:appinfo>
            <lx:path>mmMonthList.04</lx:path>
            <lx:label>04</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="05">
        <xs:annotation>
          <xs:documentation>05</xs:documentation>
          <xs:appinfo>
            <lx:path>mmMonthList.05</lx:path>
            <lx:label>05</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="06">
        <xs:annotation>
          <xs:documentation>06</xs:documentation>
          <xs:appinfo>
            <lx:path>mmMonthList.06</lx:path>
            <lx:label>06</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="07">
        <xs:annotation>
          <xs:documentation>07</xs:documentation>
          <xs:appinfo>
            <lx:path>mmMonthList.07</lx:path>
            <lx:label>07</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="08">
        <xs:annotation>
          <xs:documentation>08</xs:documentation>
          <xs:appinfo>
            <lx:path>mmMonthList.08</lx:path>
            <lx:label>08</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="09">
        <xs:annotation>
          <xs:documentation>09</xs:documentation>
          <xs:appinfo>
            <lx:path>mmMonthList.09</lx:path>
            <lx:label>09</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="10">
        <xs:annotation>
          <xs:documentation>10</xs:documentation>
          <xs:appinfo>
            <lx:path>mmMonthList.10</lx:path>
            <lx:label>10</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="11">
        <xs:annotation>
          <xs:documentation>11</xs:documentation>
          <xs:appinfo>
            <lx:path>mmMonthList.11</lx:path>
            <lx:label>11</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="12">
        <xs:annotation>
          <xs:documentation>12</xs:documentation>
          <xs:appinfo>
            <lx:path>mmMonthList.12</lx:path>
            <lx:label>12</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="mobileComputingTypeList">
    <xs:annotation>
      <xs:documentation>mobileComputingTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>mobileComputingTypeList</lx:path>
        <lx:label>mobileComputingTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Laptop">
        <xs:annotation>
          <xs:documentation>Laptop</xs:documentation>
          <xs:appinfo>
            <lx:path>mobileComputingTypeList.Laptop</lx:path>
            <lx:label>Laptop</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>mobileComputingTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Smart Phone">
        <xs:annotation>
          <xs:documentation>Smart Phone</xs:documentation>
          <xs:appinfo>
            <lx:path>mobileComputingTypeList.Smart Phone</lx:path>
            <lx:label>Smart Phone</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tablet">
        <xs:annotation>
          <xs:documentation>Tablet</xs:documentation>
          <xs:appinfo>
            <lx:path>mobileComputingTypeList.Tablet</lx:path>
            <lx:label>Tablet</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wearable Device">
        <xs:annotation>
          <xs:documentation>Wearable Device</xs:documentation>
          <xs:appinfo>
            <lx:path>mobileComputingTypeList.Wearable Device</lx:path>
            <lx:label>Wearable Device</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="nameTitleList">
    <xs:annotation>
      <xs:documentation>A list of commonly used formal name titles.</xs:documentation>
      <xs:appinfo>
        <lx:path>nameTitleList</lx:path>
        <lx:label>nameTitleList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Dr">
        <xs:annotation>
          <xs:documentation>Dr</xs:documentation>
          <xs:appinfo>
            <lx:path>nameTitleList.Dr</lx:path>
            <lx:label>Dr</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hon">
        <xs:annotation>
          <xs:documentation>Hon</xs:documentation>
          <xs:appinfo>
            <lx:path>nameTitleList.Hon</lx:path>
            <lx:label>Hon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lady">
        <xs:annotation>
          <xs:documentation>Lady</xs:documentation>
          <xs:appinfo>
            <lx:path>nameTitleList.Lady</lx:path>
            <lx:label>Lady</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Miss">
        <xs:annotation>
          <xs:documentation>Miss</xs:documentation>
          <xs:appinfo>
            <lx:path>nameTitleList.Miss</lx:path>
            <lx:label>Miss</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mr">
        <xs:annotation>
          <xs:documentation>Mr</xs:documentation>
          <xs:appinfo>
            <lx:path>nameTitleList.Mr</lx:path>
            <lx:label>Mr</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mrs">
        <xs:annotation>
          <xs:documentation>Mrs</xs:documentation>
          <xs:appinfo>
            <lx:path>nameTitleList.Mrs</lx:path>
            <lx:label>Mrs</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ms">
        <xs:annotation>
          <xs:documentation>Ms</xs:documentation>
          <xs:appinfo>
            <lx:path>nameTitleList.Ms</lx:path>
            <lx:label>Ms</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>The name title is not included in this list and may be provided in the corresponding OtherNameTitle field if available.</xs:documentation>
          <xs:appinfo>
            <lx:path>nameTitleList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Prof">
        <xs:annotation>
          <xs:documentation>Prof</xs:documentation>
          <xs:appinfo>
            <lx:path>nameTitleList.Prof</lx:path>
            <lx:label>Prof</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="R Hon">
        <xs:annotation>
          <xs:documentation>R Hon</xs:documentation>
          <xs:appinfo>
            <lx:path>nameTitleList.R Hon</lx:path>
            <lx:label>R Hon</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rev">
        <xs:annotation>
          <xs:documentation>Rev</xs:documentation>
          <xs:appinfo>
            <lx:path>nameTitleList.Rev</lx:path>
            <lx:label>Rev</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sir">
        <xs:annotation>
          <xs:documentation>Sir</xs:documentation>
          <xs:appinfo>
            <lx:path>nameTitleList.Sir</lx:path>
            <lx:label>Sir</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="nccpStatusList">
    <xs:annotation>
      <xs:documentation>nccpStatusList</xs:documentation>
      <xs:appinfo>
        <lx:path>nccpStatusList</lx:path>
        <lx:label>nccpStatusList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Regulated">
        <xs:annotation>
          <xs:documentation>The loan is regulated by the NCCP as it is predominantly for private or domestic purposes.</xs:documentation>
          <xs:appinfo>
            <lx:path>nccpStatusList.Regulated</lx:path>
            <lx:label>The loan is regulated by the NCCP as it is predominantly for private or domestic purposes.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unregulated">
        <xs:annotation>
          <xs:documentation>The loan is not regulated by the NCCP as it is not predominantly for private or domestic purposes.</xs:documentation>
          <xs:appinfo>
            <lx:path>nccpStatusList.Unregulated</lx:path>
            <lx:label>The loan is not regulated by the NCCP as it is not predominantly for private or domestic purposes.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="nonRealEstateAssetEncumbranceList">
    <xs:annotation>
      <xs:documentation>nonRealEstateAssetEncumbranceList</xs:documentation>
      <xs:appinfo>
        <lx:path>nonRealEstateAssetEncumbranceList</lx:path>
        <lx:label>nonRealEstateAssetEncumbranceList</lx:label>
        <li:transactions>CAL,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Charge">
        <xs:annotation>
          <xs:documentation>Charge</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetEncumbranceList.Charge</lx:path>
            <lx:label>Charge</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lease">
        <xs:annotation>
          <xs:documentation>Lease</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetEncumbranceList.Lease</lx:path>
            <lx:label>Lease</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="None Disclosed">
        <xs:annotation>
          <xs:documentation>None Disclosed</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetEncumbranceList.None Disclosed</lx:path>
            <lx:label>None Disclosed</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Not Known">
        <xs:annotation>
          <xs:documentation>Not Known</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetEncumbranceList.Not Known</lx:path>
            <lx:label>Not Known</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetEncumbranceList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Warrant">
        <xs:annotation>
          <xs:documentation>Warrant</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetEncumbranceList.Warrant</lx:path>
            <lx:label>Warrant</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Writ">
        <xs:annotation>
          <xs:documentation>Writ</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetEncumbranceList.Writ</lx:path>
            <lx:label>Writ</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="nonRealEstateAssetTypeList">
    <xs:annotation>
      <xs:documentation>nonRealEstateAssetTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>nonRealEstateAssetTypeList</lx:path>
        <lx:label>nonRealEstateAssetTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Agricultural">
        <xs:annotation>
          <xs:documentation>Agricultural equipment, crops and livestock</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Agricultural</lx:path>
            <lx:label>Agricultural equipment, crops and livestock</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Aircraft">
        <xs:annotation>
          <xs:documentation>Fixed and rotary wing aircraft and aircraft engines</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Aircraft</lx:path>
            <lx:label>Fixed and rotary wing aircraft and aircraft engines</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Business">
        <xs:annotation>
          <xs:documentation>Business</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Business</lx:path>
            <lx:label>Business</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cleaning">
        <xs:annotation>
          <xs:documentation>Cleaning</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Cleaning</lx:path>
            <lx:label>Cleaning</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Earth Moving, Mining and Construction">
        <xs:annotation>
          <xs:documentation>Earth Moving, Mining and Construction</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Earth Moving, Mining and Construction</lx:path>
            <lx:label>Earth Moving, Mining and Construction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Financial Asset">
        <xs:annotation>
          <xs:documentation>Cash and accounts at financial institutions</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Financial Asset</lx:path>
            <lx:label>Cash and accounts at financial institutions</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hospitality and Leisure">
        <xs:annotation>
          <xs:documentation>Hospitality and Leisure</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Hospitality and Leisure</lx:path>
            <lx:label>Hospitality and Leisure</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="IT and AV Equipment">
        <xs:annotation>
          <xs:documentation>IT and AV Equipment</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.IT and AV Equipment</lx:path>
            <lx:label>IT and AV Equipment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Materials Handling and Lifting">
        <xs:annotation>
          <xs:documentation>Materials Handling and Lifting</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Materials Handling and Lifting</lx:path>
            <lx:label>Materials Handling and Lifting</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Medical">
        <xs:annotation>
          <xs:documentation>Medical</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Medical</lx:path>
            <lx:label>Medical</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mobile Computing">
        <xs:annotation>
          <xs:documentation>Mobile Computing</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Mobile Computing</lx:path>
            <lx:label>Mobile Computing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Motor Vehicle and Transport">
        <xs:annotation>
          <xs:documentation>Including cars, commercial vehicles and watercraft</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Motor Vehicle and Transport</lx:path>
            <lx:label>Including cars, commercial vehicles and watercraft</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Office Equipment">
        <xs:annotation>
          <xs:documentation>Office Equipment</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Office Equipment</lx:path>
            <lx:label>Office Equipment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Asset is not described by any of the other enumerated values in the list</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Other</lx:path>
            <lx:label>Asset is not described by any of the other enumerated values in the list</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Plant, Equipment and Industrial">
        <xs:annotation>
          <xs:documentation>Plant, Equipment and Industrial</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Plant, Equipment and Industrial</lx:path>
            <lx:label>Plant, Equipment and Industrial</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tools of Trade">
        <xs:annotation>
          <xs:documentation>Portable Tradesman tools</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Tools of Trade</lx:path>
            <lx:label>Portable Tradesman tools</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Watercraft">
        <xs:annotation>
          <xs:documentation>A boat or other vessel that travels on water.</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateAssetTypeList.Watercraft</lx:path>
            <lx:label>Watercraft</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="nonRealEstateSecurityPriorityList">
    <xs:annotation>
      <xs:documentation>A list of the order of priority of a security interest on a non real estate asset.</xs:documentation>
      <xs:appinfo>
        <lx:path>nonRealEstateSecurityPriorityList</lx:path>
        <lx:label>nonRealEstateSecurityPriorityList</lx:label>
        <li:transactions>CAL,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="First Security Interest">
        <xs:annotation>
          <xs:documentation>First Security Interest</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateSecurityPriorityList.First Security Interest</lx:path>
            <lx:label>First Security Interest</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Second Security Interest">
        <xs:annotation>
          <xs:documentation>Second Security Interest</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateSecurityPriorityList.Second Security Interest</lx:path>
            <lx:label>Second Security Interest</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Third Security Interest">
        <xs:annotation>
          <xs:documentation>Third Security Interest</xs:documentation>
          <xs:appinfo>
            <lx:path>nonRealEstateSecurityPriorityList.Third Security Interest</lx:path>
            <lx:label>Third Security Interest</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="nrasConsortiumList">
    <xs:annotation>
      <xs:documentation>nrasConsortiumList</xs:documentation>
      <xs:appinfo>
        <lx:path>nrasConsortiumList</lx:path>
        <lx:label>nrasConsortiumList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="4 Walls Ltd">
        <xs:annotation>
          <xs:documentation>4 Walls Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>nrasConsortiumList.4 Walls Ltd</lx:path>
            <lx:label>4 Walls Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Accelerated Wealth Systems (Quantum)">
        <xs:annotation>
          <xs:documentation>Accelerated Wealth Systems (Quantum)</xs:documentation>
          <xs:appinfo>
            <lx:path>nrasConsortiumList.Accelerated Wealth Systems (Quantum)</lx:path>
            <lx:label>Accelerated Wealth Systems (Quantum)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Affordable Housing Consulting Pty Ltd">
        <xs:annotation>
          <xs:documentation>Affordable Housing Consulting Pty Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>nrasConsortiumList.Affordable Housing Consulting Pty Ltd</lx:path>
            <lx:label>Affordable Housing Consulting Pty Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Affordable Management Corporation Pty Ltd">
        <xs:annotation>
          <xs:documentation>Affordable Management Corporation Pty Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>nrasConsortiumList.Affordable Management Corporation Pty Ltd</lx:path>
            <lx:label>Affordable Management Corporation Pty Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Aspire Housing Group">
        <xs:annotation>
          <xs:documentation>Aspire Housing Group</xs:documentation>
          <xs:appinfo>
            <lx:path>nrasConsortiumList.Aspire Housing Group</lx:path>
            <lx:label>Aspire Housing Group</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Brisbane Housing Company Ltd">
        <xs:annotation>
          <xs:documentation>Brisbane Housing Company Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>nrasConsortiumList.Brisbane Housing Company Ltd</lx:path>
            <lx:label>Brisbane Housing Company Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Community Housing Canberra Ltd">
        <xs:annotation>
          <xs:documentation>Community Housing Canberra Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>nrasConsortiumList.Community Housing Canberra Ltd</lx:path>
            <lx:label>Community Housing Canberra Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ethan Affordable Housing">
        <xs:annotation>
          <xs:documentation>Ethan Affordable Housing</xs:documentation>
          <xs:appinfo>
            <lx:path>nrasConsortiumList.Ethan Affordable Housing</lx:path>
            <lx:label>Ethan Affordable Housing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Providence Housing Pty Ltd">
        <xs:annotation>
          <xs:documentation>Providence Housing Pty Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>nrasConsortiumList.Providence Housing Pty Ltd</lx:path>
            <lx:label>Providence Housing Pty Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Quantum Housing Group">
        <xs:annotation>
          <xs:documentation>Quantum Housing Group</xs:documentation>
          <xs:appinfo>
            <lx:path>nrasConsortiumList.Quantum Housing Group</lx:path>
            <lx:label>Quantum Housing Group</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Queensland Affordable Housing Consortium">
        <xs:annotation>
          <xs:documentation>Queensland Affordable Housing Consortium</xs:documentation>
          <xs:appinfo>
            <lx:path>nrasConsortiumList.Queensland Affordable Housing Consortium</lx:path>
            <lx:label>Queensland Affordable Housing Consortium</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Questus Funds Management Pty Ltd">
        <xs:annotation>
          <xs:documentation>Questus Funds Management Pty Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>nrasConsortiumList.Questus Funds Management Pty Ltd</lx:path>
            <lx:label>Questus Funds Management Pty Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Urban Affordable Housing Corporation">
        <xs:annotation>
          <xs:documentation>Urban Affordable Housing Corporation</xs:documentation>
          <xs:appinfo>
            <lx:path>nrasConsortiumList.Urban Affordable Housing Corporation</lx:path>
            <lx:label>Urban Affordable Housing Corporation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Yaran Residential Investments Pty Ltd">
        <xs:annotation>
          <xs:documentation>Yaran Residential Investments Pty Ltd</xs:documentation>
          <xs:appinfo>
            <lx:path>nrasConsortiumList.Yaran Residential Investments Pty Ltd</lx:path>
            <lx:label>Yaran Residential Investments Pty Ltd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="nzDialingCodeList">
    <xs:annotation>
      <xs:documentation>A list of valid Area Codes for New Zealand telephone numbers.</xs:documentation>
      <xs:appinfo>
        <lx:path>nzDialingCodeList</lx:path>
        <lx:label>nzDialingCodeList</lx:label>
        <li:transactions>CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="01">
        <xs:annotation>
          <xs:documentation>Special Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.01</lx:path>
            <lx:label>01</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="011">
        <xs:annotation>
          <xs:documentation>Hand-Off Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.011</lx:path>
            <lx:label>011</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="020">
        <xs:annotation>
          <xs:documentation>Non-Geographic Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.020</lx:path>
            <lx:label>020</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="021">
        <xs:annotation>
          <xs:documentation>Non-Geographic Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.021</lx:path>
            <lx:label>021</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="022">
        <xs:annotation>
          <xs:documentation>Non-Geographic Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.022</lx:path>
            <lx:label>022</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="023">
        <xs:annotation>
          <xs:documentation>Non-Geographic Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.023</lx:path>
            <lx:label>023</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="024">
        <xs:annotation>
          <xs:documentation>Non-Geographic Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.024</lx:path>
            <lx:label>024</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="025">
        <xs:annotation>
          <xs:documentation>Non-Geographic Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.025</lx:path>
            <lx:label>025</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="026">
        <xs:annotation>
          <xs:documentation>Non-Geographic Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.026</lx:path>
            <lx:label>026</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="027">
        <xs:annotation>
          <xs:documentation>Non-Geographic Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.027</lx:path>
            <lx:label>027</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="028">
        <xs:annotation>
          <xs:documentation>Non-Geographic Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.028</lx:path>
            <lx:label>028</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="029">
        <xs:annotation>
          <xs:documentation>Non-Geographic Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.029</lx:path>
            <lx:label>029</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="03">
        <xs:annotation>
          <xs:documentation>Geographic Service Code for Nelson, West Coast, Southland, Marlborough, Christchurch, Oamaru, and Otago.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.03</lx:path>
            <lx:label>03</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="04">
        <xs:annotation>
          <xs:documentation>Geographic Service Code for the Wellington Region.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.04</lx:path>
            <lx:label>04</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="05">
        <xs:annotation>
          <xs:documentation>Service Provider Prefixes.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.05</lx:path>
            <lx:label>05</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="050">
        <xs:annotation>
          <xs:documentation>Free-Phone Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.050</lx:path>
            <lx:label>050</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0508">
        <xs:annotation>
          <xs:documentation>Free-Phone Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.0508</lx:path>
            <lx:label>0508</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="06">
        <xs:annotation>
          <xs:documentation>Geographic Service Code for Taranaki, Wanganui, Manawatu, Gisborne, Hawke's Bay, and Wairarapa.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.06</lx:path>
            <lx:label>06</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="07">
        <xs:annotation>
          <xs:documentation>Geographic Service Code for Waikato and Bay of Plenty.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.07</lx:path>
            <lx:label>07</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="070">
        <xs:annotation>
          <xs:documentation>Personal Number Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.070</lx:path>
            <lx:label>070</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="08">
        <xs:annotation>
          <xs:documentation>Value-Added Service Codes.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.08</lx:path>
            <lx:label>08</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="080">
        <xs:annotation>
          <xs:documentation>Free-Phone Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.080</lx:path>
            <lx:label>080</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="0800">
        <xs:annotation>
          <xs:documentation>Free-Phone Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.0800</lx:path>
            <lx:label>0800</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="09">
        <xs:annotation>
          <xs:documentation>Geographic Service Code for Northland and Auckland.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.09</lx:path>
            <lx:label>09</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="090">
        <xs:annotation>
          <xs:documentation>Premium-Rate Services Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.090</lx:path>
            <lx:label>090</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1">
        <xs:annotation>
          <xs:documentation>Special Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.1</lx:path>
            <lx:label>(1)</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="50">
        <xs:annotation>
          <xs:documentation>Nation-Wide Number Service Code.</xs:documentation>
          <xs:appinfo>
            <lx:path>nzDialingCodeList.50</lx:path>
            <lx:label>50</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="occupancyList">
    <xs:annotation>
      <xs:documentation>occupancyList</xs:documentation>
      <xs:appinfo>
        <lx:path>occupancyList</lx:path>
        <lx:label>occupancyList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Non Owner">
        <xs:annotation>
          <xs:documentation>Non Owner</xs:documentation>
          <xs:appinfo>
            <lx:path>occupancyList.Non Owner</lx:path>
            <lx:label>Non Owner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Owner Primary">
        <xs:annotation>
          <xs:documentation>Owner Primary</xs:documentation>
          <xs:appinfo>
            <lx:path>occupancyList.Owner Primary</lx:path>
            <lx:label>Owner Primary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Owner Secondary">
        <xs:annotation>
          <xs:documentation>Owner Secondary</xs:documentation>
          <xs:appinfo>
            <lx:path>occupancyList.Owner Secondary</lx:path>
            <lx:label>Owner Secondary</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="oecdCRSStatusList">
    <xs:annotation>
      <xs:documentation>oecdCRSStatusList</xs:documentation>
      <xs:appinfo>
        <lx:path>oecdCRSStatusList</lx:path>
        <lx:label>oecdCRSStatusList</lx:label>
        <li:transactions>CAL,CDA,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Active">
        <xs:annotation>
          <xs:documentation>The entity is considered to be Active as per the definition provided by the OECD with respect to the Common Reporting Standards.</xs:documentation>
          <xs:appinfo>
            <lx:path>oecdCRSStatusList.Active</lx:path>
            <lx:label>Active</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Passive">
        <xs:annotation>
          <xs:documentation>The entity is considered to be Passive as per the definition provided by the OECD with respect to the Common Reporting Standards.</xs:documentation>
          <xs:appinfo>
            <lx:path>oecdCRSStatusList.Passive</lx:path>
            <lx:label>Passive</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="officeEquipmentTypeList">
    <xs:annotation>
      <xs:documentation>officeEquipmentTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>officeEquipmentTypeList</lx:path>
        <lx:label>officeEquipmentTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Fixtures and Fittings">
        <xs:annotation>
          <xs:documentation>Shop and office furniture, fixtures and fittings</xs:documentation>
          <xs:appinfo>
            <lx:path>officeEquipmentTypeList.Fixtures and Fittings</lx:path>
            <lx:label>Shop and office furniture, fixtures and fittings</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>officeEquipmentTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Photocopier/Printer">
        <xs:annotation>
          <xs:documentation>Photocopier/Printer</xs:documentation>
          <xs:appinfo>
            <lx:path>officeEquipmentTypeList.Photocopier/Printer</lx:path>
            <lx:label>Photocopier/Printer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Telephone Systems">
        <xs:annotation>
          <xs:documentation>Telephone Systems</xs:documentation>
          <xs:appinfo>
            <lx:path>officeEquipmentTypeList.Telephone Systems</lx:path>
            <lx:label>Telephone Systems</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="otherAssetTypeList">
    <xs:annotation>
      <xs:documentation>otherAssetTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>otherAssetTypeList</lx:path>
        <lx:label>otherAssetTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Art Works">
        <xs:annotation>
          <xs:documentation>Art Works</xs:documentation>
          <xs:appinfo>
            <lx:path>otherAssetTypeList.Art Works</lx:path>
            <lx:label>Art Works</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Good Will of Business">
        <xs:annotation>
          <xs:documentation>Goodwill is value in a business over and above the value of its tangible and other intangible assets</xs:documentation>
          <xs:appinfo>
            <lx:path>otherAssetTypeList.Good Will of Business</lx:path>
            <lx:label>Goodwill is value in a business over and above the value of its tangible and other intangible assets</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Home Contents">
        <xs:annotation>
          <xs:documentation>Home Contents</xs:documentation>
          <xs:appinfo>
            <lx:path>otherAssetTypeList.Home Contents</lx:path>
            <lx:label>Home Contents</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Intellectual Property">
        <xs:annotation>
          <xs:documentation>Intellectual Property</xs:documentation>
          <xs:appinfo>
            <lx:path>otherAssetTypeList.Intellectual Property</lx:path>
            <lx:label>Intellectual Property</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>otherAssetTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Stock and Inventory">
        <xs:annotation>
          <xs:documentation>Stock and Inventory</xs:documentation>
          <xs:appinfo>
            <lx:path>otherAssetTypeList.Stock and Inventory</lx:path>
            <lx:label>Stock and Inventory</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="otherCommitmentCategoryList">
    <xs:annotation>
      <xs:documentation>Category of the other commitment.</xs:documentation>
      <xs:appinfo>
        <lx:path>otherCommitmentCategoryList</lx:path>
        <lx:label>otherCommitmentCategoryList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Board">
        <xs:annotation>
          <xs:documentation>Board paid.</xs:documentation>
          <xs:appinfo>
            <lx:path>otherCommitmentCategoryList.Board</lx:path>
            <lx:label>Board</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Child and Spousal Maintenance">
        <xs:annotation>
          <xs:documentation>Maintenance for either dependent or non-dependent children, and spousal or de facto partner maintenance.</xs:documentation>
          <xs:appinfo>
            <lx:path>otherCommitmentCategoryList.Child and Spousal Maintenance</lx:path>
            <lx:label>Child and Spousal Maintenance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Child Maintenance">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a future release. Maintenance expenses for nondependent children.</xs:documentation>
          <xs:appinfo>
            <lx:path>otherCommitmentCategoryList.Child Maintenance</lx:path>
            <lx:label>Child Maintenance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other Commitment.</xs:documentation>
          <xs:appinfo>
            <lx:path>otherCommitmentCategoryList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rent">
        <xs:annotation>
          <xs:documentation>Rent on primary residence.</xs:documentation>
          <xs:appinfo>
            <lx:path>otherCommitmentCategoryList.Rent</lx:path>
            <lx:label>Rent</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="otherExpenseCategoryList">
    <xs:annotation>
      <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017.</xs:documentation>
      <xs:appinfo>
        <lx:path>otherExpenseCategoryList</lx:path>
        <lx:label>otherExpenseCategoryList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Basic Living Expenses">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Most food, children's clothing, utilities, general dental, optical and pharmaceuticals, and most transport expenses.</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseCategoryList.Basic Living Expenses</lx:path>
            <lx:label>Most food, children's clothing, utilities, general dental, optical and pharmaceuticals, and most transport expenses.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Discretionary Living Expenses">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Childcare, entertainment, luxury food, and domestic holidays.</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseCategoryList.Discretionary Living Expenses</lx:path>
            <lx:label>Childcare, entertainment, luxury food, and domestic holidays.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Any other recurring expense category</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseCategoryList.Other</lx:path>
            <lx:label>Any other recurring expense category</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="otherExpenseTypeList">
    <xs:annotation>
      <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017.</xs:documentation>
      <xs:appinfo>
        <lx:path>otherExpenseTypeList</lx:path>
        <lx:label>otherExpenseTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Additional Car(s) Maintenance">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Additional Car(s) Maintenance</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Additional Car(s) Maintenance</lx:path>
            <lx:label>Additional Car(s) Maintenance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Additional Car(s) Petrol">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Additional Car(s) Petrol</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Additional Car(s) Petrol</lx:path>
            <lx:label>Additional Car(s) Petrol</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Additional Car(s) Registration">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Additional Car(s) Registration</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Additional Car(s) Registration</lx:path>
            <lx:label>Additional Car(s) Registration</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Board">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Board</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Board</lx:path>
            <lx:label>Board</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Body Corporate Levies">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Body Corporate Levies</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Body Corporate Levies</lx:path>
            <lx:label>Body Corporate Levies</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Building or Contents Insurance">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Building or Contents Insurance</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Building or Contents Insurance</lx:path>
            <lx:label>Building or Contents Insurance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Child Care">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Child Care</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Child Care</lx:path>
            <lx:label>Child Care</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Child Maintenance">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Maintenance expenses for nondependent children.</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Child Maintenance</lx:path>
            <lx:label>Maintenance expenses for nondependent children.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Education">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Private school and tertiary education.</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Education</lx:path>
            <lx:label>Private school and tertiary education.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gym Membership">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Gym Membership</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Gym Membership</lx:path>
            <lx:label>Gym Membership</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Maintenance Fee">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Maintenance Fee</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Maintenance Fee</lx:path>
            <lx:label>Maintenance Fee</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Medical">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. All medical expenses including GP, but excluding dental, optical and pharmaceutical.</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Medical</lx:path>
            <lx:label>All medical expenses including GP, but excluding dental, optical and pharmaceutical.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mobile Phone(s)">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Mobile Phone(s)</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Mobile Phone(s)</lx:path>
            <lx:label>Mobile Phone(s)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Non Government Education">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Non Government Education</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Non Government Education</lx:path>
            <lx:label>Non Government Education</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Any other recurring expense.</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Other</lx:path>
            <lx:label>Any other recurring expense.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Private Health / Medical">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Private Health / Medical</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Private Health / Medical</lx:path>
            <lx:label>Private Health / Medical</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rent">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Rent on primary residence.</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Rent</lx:path>
            <lx:label>Rent on primary residence.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Subscriptions">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a release in Q1 2017. Subscriptions</xs:documentation>
          <xs:appinfo>
            <lx:path>otherExpenseTypeList.Subscriptions</lx:path>
            <lx:label>Subscriptions</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="otherIncomeTypeList">
    <xs:annotation>
      <xs:documentation>otherIncomeTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>otherIncomeTypeList</lx:path>
        <lx:label>otherIncomeTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Annuities">
        <xs:annotation>
          <xs:documentation>Annuities</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Annuities</lx:path>
            <lx:label>Annuities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Board">
        <xs:annotation>
          <xs:documentation>Income is received from a paying boarder.</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Board</lx:path>
            <lx:label>Board</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Child Maintenance">
        <xs:annotation>
          <xs:documentation>Child Maintenance</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Child Maintenance</lx:path>
            <lx:label>Child Maintenance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dividends">
        <xs:annotation>
          <xs:documentation>Dividends</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Dividends</lx:path>
            <lx:label>Dividends</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Foreign Sourced">
        <xs:annotation>
          <xs:documentation>Foreign Sourced</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Foreign Sourced</lx:path>
            <lx:label>Foreign Sourced</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefits">
        <xs:annotation>
          <xs:documentation>Government Benefits</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Government Benefits</lx:path>
            <lx:label>Government Benefits</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Interest Income">
        <xs:annotation>
          <xs:documentation>Interest Income</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Interest Income</lx:path>
            <lx:label>Interest Income</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Managed Investment">
        <xs:annotation>
          <xs:documentation>Managed Investment</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Managed Investment</lx:path>
            <lx:label>Managed Investment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other Income">
        <xs:annotation>
          <xs:documentation>Other Income</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Other Income</lx:path>
            <lx:label>Other Income</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Private Pension">
        <xs:annotation>
          <xs:documentation>Private Pension</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Private Pension</lx:path>
            <lx:label>Private Pension</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Salary Sacrifice Contribution">
        <xs:annotation>
          <xs:documentation>Salary Sacrifice Contribution</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Salary Sacrifice Contribution</lx:path>
            <lx:label>Salary Sacrifice Contribution</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Super Concessional Contribution">
        <xs:annotation>
          <xs:documentation>Super Concessional Contribution</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Super Concessional Contribution</lx:path>
            <lx:label>Super Concessional Contribution</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Super Excess Concessional Contribution">
        <xs:annotation>
          <xs:documentation>Super Excess Concessional Contribution</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Super Excess Concessional Contribution</lx:path>
            <lx:label>Super Excess Concessional Contribution</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Super Non Concessional Contribution">
        <xs:annotation>
          <xs:documentation>Super Non Concessional Contribution</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Super Non Concessional Contribution</lx:path>
            <lx:label>Super Non Concessional Contribution</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Superannuation">
        <xs:annotation>
          <xs:documentation>Superannuation</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Superannuation</lx:path>
            <lx:label>Superannuation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Workers Compensation">
        <xs:annotation>
          <xs:documentation>Workers Compensation</xs:documentation>
          <xs:appinfo>
            <lx:path>otherIncomeTypeList.Workers Compensation</lx:path>
            <lx:label>Workers Compensation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="parkingFacilityTypeList">
    <xs:annotation>
      <xs:documentation>A list of parking facility types.</xs:documentation>
      <xs:appinfo>
        <lx:path>parkingFacilityTypeList</lx:path>
        <lx:label>parkingFacilityTypeList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Carport">
        <xs:annotation>
          <xs:documentation>Carport</xs:documentation>
          <xs:appinfo>
            <lx:path>parkingFacilityTypeList.Carport</lx:path>
            <lx:label>Carport</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Garage">
        <xs:annotation>
          <xs:documentation>Garage</xs:documentation>
          <xs:appinfo>
            <lx:path>parkingFacilityTypeList.Garage</lx:path>
            <lx:label>Garage</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>parkingFacilityTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Uncovered">
        <xs:annotation>
          <xs:documentation>Uncovered</xs:documentation>
          <xs:appinfo>
            <lx:path>parkingFacilityTypeList.Uncovered</lx:path>
            <lx:label>Uncovered</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="partnerTypeList">
    <xs:annotation>
      <xs:documentation>List of the types of partners.</xs:documentation>
      <xs:appinfo>
        <lx:path>partnerTypeList</lx:path>
        <lx:label>Partner Type List</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="General Partner">
        <xs:annotation>
          <xs:documentation>A General Partner is a partner in a partnership that normally has unlimited liability.</xs:documentation>
          <xs:appinfo>
            <lx:path>partnerTypeList.General Partner</lx:path>
            <lx:label>General Partner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Limited Partner">
        <xs:annotation>
          <xs:documentation>A Limited Partner is a partner in a partnership that has their liability limited to the capital amount they have invested</xs:documentation>
          <xs:appinfo>
            <lx:path>partnerTypeList.Limited Partner</lx:path>
            <lx:label>Limited Partner</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="payeBasisList">
    <xs:annotation>
      <xs:documentation>A list of employment basis types that are valid for New Zealand.</xs:documentation>
      <xs:appinfo>
        <lx:path>payeBasisList</lx:path>
        <lx:label>payeBasisList</lx:label>
        <li:transactions>CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Casual">
        <xs:annotation>
          <xs:documentation>The employment described is under a casual contract of employment.</xs:documentation>
          <xs:appinfo>
            <lx:path>payeBasisList.Casual</lx:path>
            <lx:label>The employment described is under a casual contract of employment.</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Commission Only">
        <xs:annotation>
          <xs:documentation>The employment described is remunerated by commission only, there is no base or retainer paid.</xs:documentation>
          <xs:appinfo>
            <lx:path>payeBasisList.Commission Only</lx:path>
            <lx:label>The employment described is remunerated by commission only, there is no base or retainer paid.</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Contract">
        <xs:annotation>
          <xs:documentation>The employment described is under a contract for services.</xs:documentation>
          <xs:appinfo>
            <lx:path>payeBasisList.Contract</lx:path>
            <lx:label>The employment described is under a contract for services.</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Full Time">
        <xs:annotation>
          <xs:documentation>The employment described is under full time employment contract.</xs:documentation>
          <xs:appinfo>
            <lx:path>payeBasisList.Full Time</lx:path>
            <lx:label>The employment described is under full time employment contract.</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Part Time">
        <xs:annotation>
          <xs:documentation>The employment described is under part time employment contract.</xs:documentation>
          <xs:appinfo>
            <lx:path>payeBasisList.Part Time</lx:path>
            <lx:label>The employment described is under part time employment contract.</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Seasonal">
        <xs:annotation>
          <xs:documentation>The employment described is seasonal in nature, for example fruit picking.</xs:documentation>
          <xs:appinfo>
            <lx:path>payeBasisList.Seasonal</lx:path>
            <lx:label>The employment described is seasonal in nature, for example fruit picking.</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Temporary">
        <xs:annotation>
          <xs:documentation>The employment described is under contract with an employment agency, also known as temping.</xs:documentation>
          <xs:appinfo>
            <lx:path>payeBasisList.Temporary</lx:path>
            <lx:label>The employment described is under contract with an employment agency, also known as temping.</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="payFeesFromList">
    <xs:annotation>
      <xs:documentation>payFeesFromList</xs:documentation>
      <xs:appinfo>
        <lx:path>payFeesFromList</lx:path>
        <lx:label>payFeesFromList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Account">
        <xs:annotation>
          <xs:documentation>Account</xs:documentation>
          <xs:appinfo>
            <lx:path>payFeesFromList.Account</lx:path>
            <lx:label>Account</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Add to Loan Amount">
        <xs:annotation>
          <xs:documentation>Add to Loan Amount</xs:documentation>
          <xs:appinfo>
            <lx:path>payFeesFromList.Add to Loan Amount</lx:path>
            <lx:label>Add to Loan Amount</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cash">
        <xs:annotation>
          <xs:documentation>Cash</xs:documentation>
          <xs:appinfo>
            <lx:path>payFeesFromList.Cash</lx:path>
            <lx:label>Cash</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cheque">
        <xs:annotation>
          <xs:documentation>Cheque</xs:documentation>
          <xs:appinfo>
            <lx:path>payFeesFromList.Cheque</lx:path>
            <lx:label>Cheque</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Card">
        <xs:annotation>
          <xs:documentation>Credit Card</xs:documentation>
          <xs:appinfo>
            <lx:path>payFeesFromList.Credit Card</lx:path>
            <lx:label>Credit Card</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Debit Card">
        <xs:annotation>
          <xs:documentation>Debit Card</xs:documentation>
          <xs:appinfo>
            <lx:path>payFeesFromList.Debit Card</lx:path>
            <lx:label>Debit Card</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Direct Credit">
        <xs:annotation>
          <xs:documentation>Direct Credit</xs:documentation>
          <xs:appinfo>
            <lx:path>payFeesFromList.Direct Credit</lx:path>
            <lx:label>Direct Credit</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Direct Deposit">
        <xs:annotation>
          <xs:documentation>Direct Deposit</xs:documentation>
          <xs:appinfo>
            <lx:path>payFeesFromList.Direct Deposit</lx:path>
            <lx:label>Direct Deposit</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="From Settlement Proceeds">
        <xs:annotation>
          <xs:documentation>From Settlement Proceeds</xs:documentation>
          <xs:appinfo>
            <lx:path>payFeesFromList.From Settlement Proceeds</lx:path>
            <lx:label>From Settlement Proceeds</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Included in Loan Amount">
        <xs:annotation>
          <xs:documentation>Included in Loan Amount</xs:documentation>
          <xs:appinfo>
            <lx:path>payFeesFromList.Included in Loan Amount</lx:path>
            <lx:label>Included in Loan Amount</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="paygBasisList">
    <xs:annotation>
      <xs:documentation>paygBasisList</xs:documentation>
      <xs:appinfo>
        <lx:path>paygBasisList</lx:path>
        <lx:label>paygBasisList</lx:label>
        <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Casual">
        <xs:annotation>
          <xs:documentation>The employment described is under a casual contract of employment.</xs:documentation>
          <xs:appinfo>
            <lx:path>paygBasisList.Casual</lx:path>
            <lx:label>The employment described is under a casual contract of employment.</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Commission Only">
        <xs:annotation>
          <xs:documentation>The employment described is remunerated by commission only, there is no base or retainer paid.</xs:documentation>
          <xs:appinfo>
            <lx:path>paygBasisList.Commission Only</lx:path>
            <lx:label>The employment described is remunerated by commission only, there is no base or retainer paid.</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Contract">
        <xs:annotation>
          <xs:documentation>The employment described is under a contract for services.</xs:documentation>
          <xs:appinfo>
            <lx:path>paygBasisList.Contract</lx:path>
            <lx:label>The employment described is under a contract for services.</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Full Time">
        <xs:annotation>
          <xs:documentation>The employment described is under full time employment contract.</xs:documentation>
          <xs:appinfo>
            <lx:path>paygBasisList.Full Time</lx:path>
            <lx:label>The employment described is under full time employment contract.</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Part Time">
        <xs:annotation>
          <xs:documentation>The employment described is under part time employment contract.</xs:documentation>
          <xs:appinfo>
            <lx:path>paygBasisList.Part Time</lx:path>
            <lx:label>The employment described is under part time employment contract.</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Seasonal">
        <xs:annotation>
          <xs:documentation>The employment described is seasonal in nature, for example fruit picking.</xs:documentation>
          <xs:appinfo>
            <lx:path>paygBasisList.Seasonal</lx:path>
            <lx:label>The employment described is seasonal in nature, for example fruit picking.</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Temporary">
        <xs:annotation>
          <xs:documentation>The employment described is under contract with an employment agency, also known as temping.</xs:documentation>
          <xs:appinfo>
            <lx:path>paygBasisList.Temporary</lx:path>
            <lx:label>The employment described is under contract with an employment agency, also known as temping.</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="paymentTimingList">
    <xs:annotation>
      <xs:documentation>List of items that describe the timing of the payment with respect to the payment period itself.</xs:documentation>
      <xs:appinfo>
        <lx:path>paymentTimingList</lx:path>
        <lx:label>Payment Timing List</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="In Advance">
        <xs:annotation>
          <xs:documentation>The payments are made at the start of the payment period (in advance).</xs:documentation>
          <xs:appinfo>
            <lx:path>paymentTimingList.In Advance</lx:path>
            <lx:label>In Advance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="In Arrears">
        <xs:annotation>
          <xs:documentation>The payment is made at the end of the payment period (in arrears).</xs:documentation>
          <xs:appinfo>
            <lx:path>paymentTimingList.In Arrears</lx:path>
            <lx:label>In Arrears</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="paymentTypeList">
    <xs:annotation>
      <xs:documentation>paymentTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>paymentTypeList</lx:path>
        <lx:label>paymentTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Interest Capitalised">
        <xs:annotation>
          <xs:documentation>There is no repayment during the specified term, interest is capitalised to the loan balance.</xs:documentation>
          <xs:appinfo>
            <lx:path>paymentTypeList.Interest Capitalised</lx:path>
            <lx:label>There is no repayment during the specified term, interest is capitalised to the loan balance.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Interest Only">
        <xs:annotation>
          <xs:documentation>The repayment during the specified term consists only of interest, calculated in arrears.</xs:documentation>
          <xs:appinfo>
            <lx:path>paymentTypeList.Interest Only</lx:path>
            <lx:label>The repayment during the specified term consists only of interest, calculated in arrears.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Prepaid Interest">
        <xs:annotation>
          <xs:documentation>The repayment during the specified term consists of an interest only payment made in advance.</xs:documentation>
          <xs:appinfo>
            <lx:path>paymentTypeList.Prepaid Interest</lx:path>
            <lx:label>The repayment during the specified term consists of an interest only payment made in advance.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Principal and Interest">
        <xs:annotation>
          <xs:documentation>The repayment during the specified term consists of both principal and interest.</xs:documentation>
          <xs:appinfo>
            <lx:path>paymentTypeList.Principal and Interest</lx:path>
            <lx:label>The repayment during the specified term consists of both principal and interest.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Principal Interest Fees">
        <xs:annotation>
          <xs:documentation>The repayment during the specified term consists of both principal and interest plus a servicing fee.</xs:documentation>
          <xs:appinfo>
            <lx:path>paymentTypeList.Principal Interest Fees</lx:path>
            <lx:label>The repayment during the specified term consists of both principal and interest plus a servicing fee.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="plantEquipmentAndIndustrialTypeList">
    <xs:annotation>
      <xs:documentation>plantEquipmentAndIndustrialTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>plantEquipmentAndIndustrialTypeList</lx:path>
        <lx:label>plantEquipmentAndIndustrialTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Arborist and Landscaping Equipment">
        <xs:annotation>
          <xs:documentation>Arborist and Landscaping Equipment</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Arborist and Landscaping Equipment</lx:path>
            <lx:label>Arborist and Landscaping Equipment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Butchers Plant">
        <xs:annotation>
          <xs:documentation>Butchers Plant</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Butchers Plant</lx:path>
            <lx:label>Butchers Plant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cleaning">
        <xs:annotation>
          <xs:documentation>Cleaning</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Cleaning</lx:path>
            <lx:label>Cleaning</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Compressors">
        <xs:annotation>
          <xs:documentation>Compressors</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Compressors</lx:path>
            <lx:label>Compressors</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Engine">
        <xs:annotation>
          <xs:documentation>Engine</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Engine</lx:path>
            <lx:label>Engine</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Engineering Machine">
        <xs:annotation>
          <xs:documentation>Engineering Machine</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Engineering Machine</lx:path>
            <lx:label>Engineering Machine</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fitout">
        <xs:annotation>
          <xs:documentation>Racks, shelving, benches, workstations, cool room.</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Fitout</lx:path>
            <lx:label>Racks, shelving, benches, workstations, cool room.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Generator">
        <xs:annotation>
          <xs:documentation>Generator</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Generator</lx:path>
            <lx:label>Generator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Heating and Cooling">
        <xs:annotation>
          <xs:documentation>Air conditioning, heaters, chillers, refrigeration plant, cold storage</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Heating and Cooling</lx:path>
            <lx:label>Air conditioning, heaters, chillers, refrigeration plant, cold storage</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hot Water Plant">
        <xs:annotation>
          <xs:documentation>Hot Water Plant</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Hot Water Plant</lx:path>
            <lx:label>Hot Water Plant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lighting Plant">
        <xs:annotation>
          <xs:documentation>Lighting Plant</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Lighting Plant</lx:path>
            <lx:label>Lighting Plant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Manufacturing and Machining">
        <xs:annotation>
          <xs:documentation>CNC machine, Lathe, Press, Drill Press, Borer, Welding, Dies</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Manufacturing and Machining</lx:path>
            <lx:label>CNC machine, Lathe, Press, Drill Press, Borer, Welding, Dies</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Photo Process Plant">
        <xs:annotation>
          <xs:documentation>Photo Process Plant</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Photo Process Plant</lx:path>
            <lx:label>Photo Process Plant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Power Metering Equipment">
        <xs:annotation>
          <xs:documentation>Power Metering Equipment</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Power Metering Equipment</lx:path>
            <lx:label>Power Metering Equipment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Printing Press">
        <xs:annotation>
          <xs:documentation>Printing Press</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Printing Press</lx:path>
            <lx:label>Printing Press</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pump">
        <xs:annotation>
          <xs:documentation>Pump</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Pump</lx:path>
            <lx:label>Pump</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sewing Machine">
        <xs:annotation>
          <xs:documentation>Sewing Machine</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Sewing Machine</lx:path>
            <lx:label>Sewing Machine</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Solar DC Inverters">
        <xs:annotation>
          <xs:documentation>Solar DC Inverters</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Solar DC Inverters</lx:path>
            <lx:label>Solar DC Inverters</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Solar Rooftop PV">
        <xs:annotation>
          <xs:documentation>Solar Rooftop PV</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Solar Rooftop PV</lx:path>
            <lx:label>Solar Rooftop PV</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wood Working Plant">
        <xs:annotation>
          <xs:documentation>Wood Working Plant</xs:documentation>
          <xs:appinfo>
            <lx:path>plantEquipmentAndIndustrialTypeList.Wood Working Plant</lx:path>
            <lx:label>Wood Working Plant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="planTypeList">
    <xs:annotation>
      <xs:documentation>planTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>planTypeList</lx:path>
        <lx:label>planTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Building Unit Plan">
        <xs:annotation>
          <xs:documentation>Building Unit Plan (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Building Unit Plan</lx:path>
            <lx:label>Building Unit Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Building Unit Plan of Amalgamation">
        <xs:annotation>
          <xs:documentation>Building Unit Plan of Amalgamation (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Building Unit Plan of Amalgamation</lx:path>
            <lx:label>Building Unit Plan of Amalgamation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Building Unit Plan of Resubdivision">
        <xs:annotation>
          <xs:documentation>Building Unit Plan of Resubdivision (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Building Unit Plan of Resubdivision</lx:path>
            <lx:label>Building Unit Plan of Resubdivision</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Community Plan">
        <xs:annotation>
          <xs:documentation>Community Plan (NSW, SA)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Community Plan</lx:path>
            <lx:label>Community Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Crown Plan">
        <xs:annotation>
          <xs:documentation>Crown Plan (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Crown Plan</lx:path>
            <lx:label>Crown Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Deposited Plan">
        <xs:annotation>
          <xs:documentation>Deposited Plan (NSW,SA)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Deposited Plan</lx:path>
            <lx:label>Deposited Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Diagram">
        <xs:annotation>
          <xs:documentation>Diagram (WA)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Diagram</lx:path>
            <lx:label>Diagram</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Filed Plan">
        <xs:annotation>
          <xs:documentation>Filed Plan (SA)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Filed Plan</lx:path>
            <lx:label>Filed Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Group Title Plan">
        <xs:annotation>
          <xs:documentation>Group Title Plan (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Group Title Plan</lx:path>
            <lx:label>Group Title Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Group Title Plan of Amalgamation">
        <xs:annotation>
          <xs:documentation>Group Title Plan of Amalgamation (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Group Title Plan of Amalgamation</lx:path>
            <lx:label>Group Title Plan of Amalgamation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Group Title Plan of Resubdivision">
        <xs:annotation>
          <xs:documentation>Group Title Plan of Resubdivision (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Group Title Plan of Resubdivision</lx:path>
            <lx:label>Group Title Plan of Resubdivision</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Leasehold Building Unit Plan (South Bank)">
        <xs:annotation>
          <xs:documentation>Leasehold Building Unit Plan (South Bank) (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Leasehold Building Unit Plan (South Bank)</lx:path>
            <lx:label>Leasehold Building Unit Plan (South Bank)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lower Building Unit Plan">
        <xs:annotation>
          <xs:documentation>Lower Building Unit Plan (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Lower Building Unit Plan</lx:path>
            <lx:label>Lower Building Unit Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mixed Community Plan">
        <xs:annotation>
          <xs:documentation>Mixed Community Plan (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Mixed Community Plan</lx:path>
            <lx:label>Mixed Community Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mixed Precinct Plan">
        <xs:annotation>
          <xs:documentation>Mixed Precinct Plan (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Mixed Precinct Plan</lx:path>
            <lx:label>Mixed Precinct Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mixed Stratum Plan">
        <xs:annotation>
          <xs:documentation>Mixed Stratum Plan (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Mixed Stratum Plan</lx:path>
            <lx:label>Mixed Stratum Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Plan">
        <xs:annotation>
          <xs:documentation>Plan (WA)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Plan</lx:path>
            <lx:label>Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Plan of Subdivision">
        <xs:annotation>
          <xs:documentation>Plan of Subdivision (VIC)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Plan of Subdivision</lx:path>
            <lx:label>Plan of Subdivision</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Registered Plan">
        <xs:annotation>
          <xs:documentation>Registered Plan (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Registered Plan</lx:path>
            <lx:label>Registered Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Strata Plan">
        <xs:annotation>
          <xs:documentation>Strata Plan (NSW, SA, WA)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Strata Plan</lx:path>
            <lx:label>Strata Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Stratum Plan (South Bank)">
        <xs:annotation>
          <xs:documentation>Stratum Plan (South Bank) (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Stratum Plan (South Bank)</lx:path>
            <lx:label>Stratum Plan (South Bank)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Stratum Plan of Amalgamation (South Bank)">
        <xs:annotation>
          <xs:documentation>Stratum Plan of Amalgamation (South Bank) (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Stratum Plan of Amalgamation (South Bank)</lx:path>
            <lx:label>Stratum Plan of Amalgamation (South Bank)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Stratum Plan of Subdivision (South Bank)">
        <xs:annotation>
          <xs:documentation>Stratum Plan of Subdivision (South Bank) (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Stratum Plan of Subdivision (South Bank)</lx:path>
            <lx:label>Stratum Plan of Subdivision (South Bank)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Survey Plan">
        <xs:annotation>
          <xs:documentation>Survey Plan (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Survey Plan</lx:path>
            <lx:label>Survey Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Survey Strata Plan">
        <xs:annotation>
          <xs:documentation>Survey Strata Plan (WA)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Survey Strata Plan</lx:path>
            <lx:label>Survey Strata Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Upper Building Unit Plan">
        <xs:annotation>
          <xs:documentation>Upper Building Unit Plan (QLD)</xs:documentation>
          <xs:appinfo>
            <lx:path>planTypeList.Upper Building Unit Plan</lx:path>
            <lx:label>Upper Building Unit Plan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="poBoxTypeList">
    <xs:annotation>
      <xs:documentation>poBoxTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>poBoxTypeList</lx:path>
        <lx:label>poBoxTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Care of Post Office">
        <xs:annotation>
          <xs:documentation>Care of Post Office</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Care of Post Office</lx:path>
            <lx:label>Care of Post Office</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Community Mail Agent">
        <xs:annotation>
          <xs:documentation>Community Mail Agent</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Community Mail Agent</lx:path>
            <lx:label>Community Mail Agent</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Community Mail Bag">
        <xs:annotation>
          <xs:documentation>Community Mail Bag</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Community Mail Bag</lx:path>
            <lx:label>Community Mail Bag (CMB)</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Community Postal Agent">
        <xs:annotation>
          <xs:documentation>Community Postal Agent</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Community Postal Agent</lx:path>
            <lx:label>Community Postal Agent</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Counter Delivery">
        <xs:annotation>
          <xs:documentation>Counter Delivery</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Counter Delivery</lx:path>
            <lx:label>Counter Delivery</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Freepost">
        <xs:annotation>
          <xs:documentation>Freepost</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Freepost</lx:path>
            <lx:label>Freepost</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="GPO Box">
        <xs:annotation>
          <xs:documentation>GPO Box</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.GPO Box</lx:path>
            <lx:label>GPO Box</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Locked Bag">
        <xs:annotation>
          <xs:documentation>Locked Bag</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Locked Bag</lx:path>
            <lx:label>Locked Bag</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mail Bag">
        <xs:annotation>
          <xs:documentation>Mail Bag</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Mail Bag</lx:path>
            <lx:label>Mail Bag</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mail Service">
        <xs:annotation>
          <xs:documentation>Mail Service</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Mail Service</lx:path>
            <lx:label>Mail Service</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PO Box">
        <xs:annotation>
          <xs:documentation>PO Box</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.PO Box</lx:path>
            <lx:label>PO Box</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Postal Mail Bag">
        <xs:annotation>
          <xs:documentation>Postal Mail Bag</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Postal Mail Bag</lx:path>
            <lx:label>Postal Mail Bag</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Poste Restante">
        <xs:annotation>
          <xs:documentation>Poste Restante</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Poste Restante</lx:path>
            <lx:label>Poste Restante</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Private Bag">
        <xs:annotation>
          <xs:documentation>Private Bag</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Private Bag</lx:path>
            <lx:label>Private Bag</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Response Bag">
        <xs:annotation>
          <xs:documentation>Response Bag</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Response Bag</lx:path>
            <lx:label>Response Bag</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Roadside Delivery">
        <xs:annotation>
          <xs:documentation>Roadside Delivery</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Roadside Delivery</lx:path>
            <lx:label>Roadside Delivery</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Roadside Mail Bag">
        <xs:annotation>
          <xs:documentation>Roadside Mail Bag</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Roadside Mail Bag</lx:path>
            <lx:label>Roadside Mail Bag</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Roadside Mail Box">
        <xs:annotation>
          <xs:documentation>Roadside Mail Box</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Roadside Mail Box</lx:path>
            <lx:label>Roadside Mail Box</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Roadside Mail Service">
        <xs:annotation>
          <xs:documentation>Roadside Mail Service</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Roadside Mail Service</lx:path>
            <lx:label>Roadside Mail Service</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rural Mail Bag">
        <xs:annotation>
          <xs:documentation>Rural Mail Bag</xs:documentation>
          <xs:appinfo>
            <lx:path>poBoxTypeList.Rural Mail Bag</lx:path>
            <lx:label>Rural Mail Bag</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="poolTypeList">
    <xs:annotation>
      <xs:documentation>poolTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>poolTypeList</lx:path>
        <lx:label>poolTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Above Ground">
        <xs:annotation>
          <xs:documentation>Above Ground</xs:documentation>
          <xs:appinfo>
            <lx:path>poolTypeList.Above Ground</lx:path>
            <lx:label>Above Ground</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Inground">
        <xs:annotation>
          <xs:documentation>Inground</xs:documentation>
          <xs:appinfo>
            <lx:path>poolTypeList.Inground</lx:path>
            <lx:label>Inground</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="powerOfAttorneyTypeList">
    <xs:annotation>
      <xs:documentation>powerOfAttorneyTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>powerOfAttorneyTypeList</lx:path>
        <lx:label>powerOfAttorneyTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Company">
        <xs:annotation>
          <xs:documentation>Company</xs:documentation>
          <xs:appinfo>
            <lx:path>powerOfAttorneyTypeList.Company</lx:path>
            <lx:label>Company</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Enduring">
        <xs:annotation>
          <xs:documentation>Enduring</xs:documentation>
          <xs:appinfo>
            <lx:path>powerOfAttorneyTypeList.Enduring</lx:path>
            <lx:label>Enduring</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="General">
        <xs:annotation>
          <xs:documentation>General</xs:documentation>
          <xs:appinfo>
            <lx:path>powerOfAttorneyTypeList.General</lx:path>
            <lx:label>General</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>powerOfAttorneyTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="preConditionToStageApplicationInstructionsList">
    <xs:annotation>
      <xs:documentation>preConditionToStageApplicationInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>preConditionToStageApplicationInstructionsList</lx:path>
        <lx:label>preConditionToStageApplicationInstructionsList</lx:label>
        <li:transactions>CAL,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Conditional Approval">
        <xs:annotation>
          <xs:documentation>The Condition must be satisfied or waived in order to progress to Conditional Approval status</xs:documentation>
          <xs:appinfo>
            <lx:path>preConditionToStageApplicationInstructionsList.Conditional Approval</lx:path>
            <lx:label>The Condition must be satisfied or waived in order to progress to Conditional Approval status</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PreApproval">
        <xs:annotation>
          <xs:documentation>The Condition must be satisfied or waived in order to progress to Pre Approval status</xs:documentation>
          <xs:appinfo>
            <lx:path>preConditionToStageApplicationInstructionsList.PreApproval</lx:path>
            <lx:label>The Condition must be satisfied or waived in order to progress to Pre Approval status</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Referred to Assessor">
        <xs:annotation>
          <xs:documentation>The Condition must be satisfied or waived in order to progress to Referred status</xs:documentation>
          <xs:appinfo>
            <lx:path>preConditionToStageApplicationInstructionsList.Referred to Assessor</lx:path>
            <lx:label>The Condition must be satisfied or waived in order to progress to Referred status</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement">
        <xs:annotation>
          <xs:documentation>The Condition must be satisfied or waived in order to progress to Settlement status</xs:documentation>
          <xs:appinfo>
            <lx:path>preConditionToStageApplicationInstructionsList.Settlement</lx:path>
            <lx:label>The Condition must be satisfied or waived in order to progress to Settlement status</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unconditional Approval">
        <xs:annotation>
          <xs:documentation>The Condition must be satisfied or waived in order to progress to Unconditional Approval status</xs:documentation>
          <xs:appinfo>
            <lx:path>preConditionToStageApplicationInstructionsList.Unconditional Approval</lx:path>
            <lx:label>The Condition must be satisfied or waived in order to progress to Unconditional Approval status</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="preConditionToStageLoanDocumentsInstructionsList">
    <xs:annotation>
      <xs:documentation>preConditionToStageLoanDocumentsInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>preConditionToStageLoanDocumentsInstructionsList</lx:path>
        <lx:label>preConditionToStageLoanDocumentsInstructionsList</lx:label>
        <li:transactions>DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Conditional Approval">
        <xs:annotation>
          <xs:documentation>The Condition must be satisfied or waived in order to progress to Conditional Approval status</xs:documentation>
          <xs:appinfo>
            <lx:path>preConditionToStageLoanDocumentsInstructionsList.Conditional Approval</lx:path>
            <lx:label>The Condition must be satisfied or waived in order to progress to Conditional Approval status</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PreApproval">
        <xs:annotation>
          <xs:documentation>The Condition must be satisfied or waived in order to progress to Pre Approval status</xs:documentation>
          <xs:appinfo>
            <lx:path>preConditionToStageLoanDocumentsInstructionsList.PreApproval</lx:path>
            <lx:label>The Condition must be satisfied or waived in order to progress to Pre Approval status</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Referred">
        <xs:annotation>
          <xs:documentation>The Condition must be satisfied or waived in order to progress to Referred status</xs:documentation>
          <xs:appinfo>
            <lx:path>preConditionToStageLoanDocumentsInstructionsList.Referred</lx:path>
            <lx:label>The Condition must be satisfied or waived in order to progress to Referred status</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Settlement">
        <xs:annotation>
          <xs:documentation>The Condition must be satisfied or waived in order to progress to Settlement status</xs:documentation>
          <xs:appinfo>
            <lx:path>preConditionToStageLoanDocumentsInstructionsList.Settlement</lx:path>
            <lx:label>The Condition must be satisfied or waived in order to progress to Settlement status</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unconditional Approval">
        <xs:annotation>
          <xs:documentation>The Condition must be satisfied or waived in order to progress to Unconditional Approval status</xs:documentation>
          <xs:appinfo>
            <lx:path>preConditionToStageLoanDocumentsInstructionsList.Unconditional Approval</lx:path>
            <lx:label>The Condition must be satisfied or waived in order to progress to Unconditional Approval status</lx:label>
            <li:transactions>DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="preferredContactCompanyList">
    <xs:annotation>
      <xs:documentation>A list of methods to contact a company.</xs:documentation>
      <xs:appinfo>
        <lx:path>preferredContactCompanyList</lx:path>
        <lx:label>preferredContactCompanyList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Email">
        <xs:annotation>
          <xs:documentation>Preferred contact method is email.</xs:documentation>
          <xs:appinfo>
            <lx:path>preferredContactCompanyList.Email</lx:path>
            <lx:label>Preferred contact method is email.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Office Phone">
        <xs:annotation>
          <xs:documentation>Preferred contact method is work phone number.</xs:documentation>
          <xs:appinfo>
            <lx:path>preferredContactCompanyList.Office Phone</lx:path>
            <lx:label>Preferred contact method is work phone number.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="preferredContactPersonList">
    <xs:annotation>
      <xs:documentation>A list of methods to contact a person.</xs:documentation>
      <xs:appinfo>
        <lx:path>preferredContactPersonList</lx:path>
        <lx:label>preferredContactPersonList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Email">
        <xs:annotation>
          <xs:documentation>Preferred contact method is email.</xs:documentation>
          <xs:appinfo>
            <lx:path>preferredContactPersonList.Email</lx:path>
            <lx:label>Preferred contact method is email.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Home Phone">
        <xs:annotation>
          <xs:documentation>Preferred contact method is home phone number.</xs:documentation>
          <xs:appinfo>
            <lx:path>preferredContactPersonList.Home Phone</lx:path>
            <lx:label>Preferred contact method is home phone number.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mobile">
        <xs:annotation>
          <xs:documentation>Preferred contact method is mobile phone number.</xs:documentation>
          <xs:appinfo>
            <lx:path>preferredContactPersonList.Mobile</lx:path>
            <lx:label>Preferred contact method is mobile phone number.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Work Phone">
        <xs:annotation>
          <xs:documentation>Preferred contact method is work phone number.</xs:documentation>
          <xs:appinfo>
            <lx:path>preferredContactPersonList.Work Phone</lx:path>
            <lx:label>Preferred contact method is work phone number.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="preferredCorrespondenceList">
    <xs:annotation>
      <xs:documentation>preferredCorrespondenceList</xs:documentation>
      <xs:appinfo>
        <lx:path>preferredCorrespondenceList</lx:path>
        <lx:label>preferredCorrespondenceList</lx:label>
        <li:transactions>ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Email">
        <xs:annotation>
          <xs:documentation>Email</xs:documentation>
          <xs:appinfo>
            <lx:path>preferredCorrespondenceList.Email</lx:path>
            <lx:label>Email</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="None">
        <xs:annotation>
          <xs:documentation>None</xs:documentation>
          <xs:appinfo>
            <lx:path>preferredCorrespondenceList.None</lx:path>
            <lx:label>None</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Post">
        <xs:annotation>
          <xs:documentation>Post</xs:documentation>
          <xs:appinfo>
            <lx:path>preferredCorrespondenceList.Post</lx:path>
            <lx:label>Post</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="primaryPurposeLoanPurposeList">
    <xs:annotation>
      <xs:documentation>primaryPurposeLoanPurposeList</xs:documentation>
      <xs:appinfo>
        <lx:path>primaryPurposeLoanPurposeList</lx:path>
        <lx:label>primaryPurposeLoanPurposeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Business">
        <xs:annotation>
          <xs:documentation>Primarily for business or other commercial purposes.</xs:documentation>
          <xs:appinfo>
            <lx:path>primaryPurposeLoanPurposeList.Business</lx:path>
            <lx:label>Primarily for business or other commercial purposes.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Investment Non Residential">
        <xs:annotation>
          <xs:documentation>Primarily for investment purposes other than investment in residential property.</xs:documentation>
          <xs:appinfo>
            <lx:path>primaryPurposeLoanPurposeList.Investment Non Residential</lx:path>
            <lx:label>Primarily for investment purposes other than investment in residential property.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Investment Residential">
        <xs:annotation>
          <xs:documentation>Primarily for the purchase, refinance, renovation or improvement of residential property used for investment purposes.</xs:documentation>
          <xs:appinfo>
            <lx:path>primaryPurposeLoanPurposeList.Investment Residential</lx:path>
            <lx:label>Primarily for the purchase, refinance, renovation or improvement of residential property used for investment purposes.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Owner Occupied">
        <xs:annotation>
          <xs:documentation>Primarily for the purchase, refinance, renovation or improvement of owner occupied residential property.</xs:documentation>
          <xs:appinfo>
            <lx:path>primaryPurposeLoanPurposeList.Owner Occupied</lx:path>
            <lx:label>Primarily for the purchase, refinance, renovation or improvement of owner occupied residential property.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Personal">
        <xs:annotation>
          <xs:documentation>Primarily for personal, domestic or household purposes.</xs:documentation>
          <xs:appinfo>
            <lx:path>primaryPurposeLoanPurposeList.Personal</lx:path>
            <lx:label>Primarily for personal, domestic or household purposes.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="primaryPurposeRealEstateAssetList">
    <xs:annotation>
      <xs:documentation>primaryPurposeRealEstateAssetList</xs:documentation>
      <xs:appinfo>
        <lx:path>primaryPurposeRealEstateAssetList</lx:path>
        <lx:label>primaryPurposeRealEstateAssetList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Business">
        <xs:annotation>
          <xs:documentation>The property is owned for the primary purpose as a business premises for the owners.</xs:documentation>
          <xs:appinfo>
            <lx:path>primaryPurposeRealEstateAssetList.Business</lx:path>
            <lx:label>The property is owned for the primary purpose as a business premises for the owners.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Investment">
        <xs:annotation>
          <xs:documentation>The property is owned for the primary purpose of an investment for the owners.</xs:documentation>
          <xs:appinfo>
            <lx:path>primaryPurposeRealEstateAssetList.Investment</lx:path>
            <lx:label>The property is owned for the primary purpose of an investment for the owners.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Owner Occupied">
        <xs:annotation>
          <xs:documentation>The property is owned for the primary purpose as a residence for the owners.</xs:documentation>
          <xs:appinfo>
            <lx:path>primaryPurposeRealEstateAssetList.Owner Occupied</lx:path>
            <lx:label>The property is owned for the primary purpose as a residence for the owners.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="primaryUsageList">
    <xs:annotation>
      <xs:documentation>primaryUsageList</xs:documentation>
      <xs:appinfo>
        <lx:path>primaryUsageList</lx:path>
        <lx:label>primaryUsageList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Commercial">
        <xs:annotation>
          <xs:documentation>Property is used primarily for commercial purposes.</xs:documentation>
          <xs:appinfo>
            <lx:path>primaryUsageList.Commercial</lx:path>
            <lx:label>Property is used primarily for commercial purposes.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Industrial">
        <xs:annotation>
          <xs:documentation>Property used for industrial purposes such as a factory or warehouse.</xs:documentation>
          <xs:appinfo>
            <lx:path>primaryUsageList.Industrial</lx:path>
            <lx:label>Property used for industrial purposes such as a factory or warehouse.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Residential">
        <xs:annotation>
          <xs:documentation>Indicates that a property is used primarily for residential purposes.</xs:documentation>
          <xs:appinfo>
            <lx:path>primaryUsageList.Residential</lx:path>
            <lx:label>Indicates that a property is used primarily for residential purposes.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rural">
        <xs:annotation>
          <xs:documentation>Property is located in a rural area and may be used for carrying on business of primary production.</xs:documentation>
          <xs:appinfo>
            <lx:path>primaryUsageList.Rural</lx:path>
            <lx:label>Property is located in a rural area and may be used for carrying on business of primary production.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="principalRefinancingReasonList">
    <xs:annotation>
      <xs:documentation>principalRefinancingReasonList</xs:documentation>
      <xs:appinfo>
        <lx:path>principalRefinancingReasonList</lx:path>
        <lx:label>principalRefinancingReasonList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Consolidating or Restructuring Finances">
        <xs:annotation>
          <xs:documentation>Consolidating or Restructuring Finances</xs:documentation>
          <xs:appinfo>
            <lx:path>principalRefinancingReasonList.Consolidating or Restructuring Finances</lx:path>
            <lx:label>Consolidating or Restructuring Finances</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Convenience / Flexibility">
        <xs:annotation>
          <xs:documentation>Convenience / Flexibility</xs:documentation>
          <xs:appinfo>
            <lx:path>principalRefinancingReasonList.Convenience / Flexibility</lx:path>
            <lx:label>Convenience / Flexibility</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dissatisfaction with Service at Current Lender">
        <xs:annotation>
          <xs:documentation>Dissatisfaction with Service at Current Lender</xs:documentation>
          <xs:appinfo>
            <lx:path>principalRefinancingReasonList.Dissatisfaction with Service at Current Lender</lx:path>
            <lx:label>Dissatisfaction with Service at Current Lender</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="More Competitive Pricing">
        <xs:annotation>
          <xs:documentation>More Competitive Pricing</xs:documentation>
          <xs:appinfo>
            <lx:path>principalRefinancingReasonList.More Competitive Pricing</lx:path>
            <lx:label>More Competitive Pricing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Specific Product or Feature">
        <xs:annotation>
          <xs:documentation>Specific Product or Feature</xs:documentation>
          <xs:appinfo>
            <lx:path>principalRefinancingReasonList.Specific Product or Feature</lx:path>
            <lx:label>Specific Product or Feature</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="proofCodeCompanyList">
    <xs:annotation>
      <xs:documentation>proofCodeCompanyList</xs:documentation>
      <xs:appinfo>
        <lx:path>proofCodeCompanyList</lx:path>
        <lx:label>proofCodeCompanyList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Financial Statement">
        <xs:annotation>
          <xs:documentation>Financial Statement</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeCompanyList.Financial Statement</lx:path>
            <lx:label>Financial Statement</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tax Assessment">
        <xs:annotation>
          <xs:documentation>Tax Assessment</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeCompanyList.Tax Assessment</lx:path>
            <lx:label>Tax Assessment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tax Return">
        <xs:annotation>
          <xs:documentation>Tax Return</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeCompanyList.Tax Return</lx:path>
            <lx:label>Tax Return</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="proofCodeNotEmployedList">
    <xs:annotation>
      <xs:documentation>proofCodeNotEmployedList</xs:documentation>
      <xs:appinfo>
        <lx:path>proofCodeNotEmployedList</lx:path>
        <lx:label>proofCodeNotEmployedList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Letter from Appropriate Body">
        <xs:annotation>
          <xs:documentation>Letter from Appropriate Body</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeNotEmployedList.Letter from Appropriate Body</lx:path>
            <lx:label>Letter from Appropriate Body</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Letter from Government">
        <xs:annotation>
          <xs:documentation>Letter from Government</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeNotEmployedList.Letter from Government</lx:path>
            <lx:label>Letter from Government</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Letter from Super Fund">
        <xs:annotation>
          <xs:documentation>Letter from Super Fund</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeNotEmployedList.Letter from Super Fund</lx:path>
            <lx:label>Letter from Super Fund</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tax Return">
        <xs:annotation>
          <xs:documentation>Tax Return</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeNotEmployedList.Tax Return</lx:path>
            <lx:label>Tax Return</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="proofCodeOtherList">
    <xs:annotation>
      <xs:documentation>proofCodeOtherList</xs:documentation>
      <xs:appinfo>
        <lx:path>proofCodeOtherList</lx:path>
        <lx:label>proofCodeOtherList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Letter from Appropriate Body">
        <xs:annotation>
          <xs:documentation>Letter from Appropriate Body</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeOtherList.Letter from Appropriate Body</lx:path>
            <lx:label>Letter from Appropriate Body</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Letter from Government">
        <xs:annotation>
          <xs:documentation>Letter from Government</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeOtherList.Letter from Government</lx:path>
            <lx:label>Letter from Government</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tax Return">
        <xs:annotation>
          <xs:documentation>Tax Return</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeOtherList.Tax Return</lx:path>
            <lx:label>Tax Return</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="proofCodePAYEList">
    <xs:annotation>
      <xs:documentation>A list of codes that indicate the type of evidence provided as proof of PAYE income.</xs:documentation>
      <xs:appinfo>
        <lx:path>proofCodePAYEList</lx:path>
        <lx:label>proofCodePAYEList</lx:label>
        <li:transactions>CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Employment Contract">
        <xs:annotation>
          <xs:documentation>Employment Contract</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodePAYEList.Employment Contract</lx:path>
            <lx:label>Employment Contract</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Group Certificate">
        <xs:annotation>
          <xs:documentation>Group Certificate</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodePAYEList.Group Certificate</lx:path>
            <lx:label>Group Certificate</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Letter from Employer">
        <xs:annotation>
          <xs:documentation>Letter from Employer</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodePAYEList.Letter from Employer</lx:path>
            <lx:label>Letter from Employer</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Payslip">
        <xs:annotation>
          <xs:documentation>Payslip</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodePAYEList.Payslip</lx:path>
            <lx:label>Payslip</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tax Return">
        <xs:annotation>
          <xs:documentation>Tax Return</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodePAYEList.Tax Return</lx:path>
            <lx:label>Tax Return</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="proofCodePAYGList">
    <xs:annotation>
      <xs:documentation>proofCodePAYGList</xs:documentation>
      <xs:appinfo>
        <lx:path>proofCodePAYGList</lx:path>
        <lx:label>proofCodePAYGList</lx:label>
        <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Employment Contract">
        <xs:annotation>
          <xs:documentation>Employment Contract</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodePAYGList.Employment Contract</lx:path>
            <lx:label>Employment Contract</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Group Certificate">
        <xs:annotation>
          <xs:documentation>Group Certificate</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodePAYGList.Group Certificate</lx:path>
            <lx:label>Group Certificate</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Letter from Employer">
        <xs:annotation>
          <xs:documentation>Letter from Employer</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodePAYGList.Letter from Employer</lx:path>
            <lx:label>Letter from Employer</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Payslip">
        <xs:annotation>
          <xs:documentation>Payslip</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodePAYGList.Payslip</lx:path>
            <lx:label>Payslip</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tax Return">
        <xs:annotation>
          <xs:documentation>Tax Return</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodePAYGList.Tax Return</lx:path>
            <lx:label>Tax Return</lx:label>
            <li:transactions>CAL,CDA,DAS,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="proofCodeSelfEmployedList">
    <xs:annotation>
      <xs:documentation>proofCodeSelfEmployedList</xs:documentation>
      <xs:appinfo>
        <lx:path>proofCodeSelfEmployedList</lx:path>
        <lx:label>proofCodeSelfEmployedList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Financial Statement">
        <xs:annotation>
          <xs:documentation>Financial Statement</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeSelfEmployedList.Financial Statement</lx:path>
            <lx:label>Financial Statement</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tax Return">
        <xs:annotation>
          <xs:documentation>Tax Return</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeSelfEmployedList.Tax Return</lx:path>
            <lx:label>Tax Return</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="proofCodeTrustList">
    <xs:annotation>
      <xs:documentation>proofCodeTrustList</xs:documentation>
      <xs:appinfo>
        <lx:path>proofCodeTrustList</lx:path>
        <lx:label>proofCodeTrustList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Financial Statement">
        <xs:annotation>
          <xs:documentation>Financial Statement</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeTrustList.Financial Statement</lx:path>
            <lx:label>Financial Statement</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tax Assessment">
        <xs:annotation>
          <xs:documentation>Tax Assessment</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeTrustList.Tax Assessment</lx:path>
            <lx:label>Tax Assessment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tax Return">
        <xs:annotation>
          <xs:documentation>Tax Return</xs:documentation>
          <xs:appinfo>
            <lx:path>proofCodeTrustList.Tax Return</lx:path>
            <lx:label>Tax Return</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="propertyPartTypeList">
    <xs:annotation>
      <xs:documentation>The types of property parts from which rental income can be earned.</xs:documentation>
      <xs:appinfo>
        <lx:path>propertyPartTypeList</lx:path>
        <lx:label>propertyPartTypeList</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Car Space">
        <xs:annotation>
          <xs:documentation>Car Space</xs:documentation>
          <xs:appinfo>
            <lx:path>propertyPartTypeList.Car Space</lx:path>
            <lx:label>Car Space</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Garage">
        <xs:annotation>
          <xs:documentation>Garage</xs:documentation>
          <xs:appinfo>
            <lx:path>propertyPartTypeList.Garage</lx:path>
            <lx:label>Garage</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Granny Flat">
        <xs:annotation>
          <xs:documentation>Granny Flat</xs:documentation>
          <xs:appinfo>
            <lx:path>propertyPartTypeList.Granny Flat</lx:path>
            <lx:label>Granny Flat</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Room Within Main Residence">
        <xs:annotation>
          <xs:documentation>Room Within Main Residence</xs:documentation>
          <xs:appinfo>
            <lx:path>propertyPartTypeList.Room Within Main Residence</lx:path>
            <lx:label>Room Within Main Residence</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="propertySearchValidationResultList">
    <xs:annotation>
      <xs:documentation>The valid list of results available from a Property Search Validation.</xs:documentation>
      <xs:appinfo>
        <lx:path>propertySearchValidationResultList</lx:path>
        <lx:label>Property Search Validation Result List</lx:label>
        <li:transactions>CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Caution">
        <xs:annotation>
          <xs:documentation>Indicates that the Purchase Price and/or LVRs are at the limit or slightly outside acceptable intervals for the current property.</xs:documentation>
          <xs:appinfo>
            <lx:path>propertySearchValidationResultList.Caution</lx:path>
            <lx:label>Caution</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pass">
        <xs:annotation>
          <xs:documentation>Indicates that the Purchase Price and LVRs are within acceptable intervals for the current property.</xs:documentation>
          <xs:appinfo>
            <lx:path>propertySearchValidationResultList.Pass</lx:path>
            <lx:label>Pass</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Refer">
        <xs:annotation>
          <xs:documentation>Indicates that the Purchase Price and/or LVRs are either outside acceptable intervals or the data is not available and a property valuation should be ordered.</xs:documentation>
          <xs:appinfo>
            <lx:path>propertySearchValidationResultList.Refer</lx:path>
            <lx:label>Refer</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="proportionsList">
    <xs:annotation>
      <xs:documentation>proportionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>proportionsList</lx:path>
        <lx:label>proportionsList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Equal">
        <xs:annotation>
          <xs:documentation>Equal</xs:documentation>
          <xs:appinfo>
            <lx:path>proportionsList.Equal</lx:path>
            <lx:label>Equal</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,SVC,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Not Specified">
        <xs:annotation>
          <xs:documentation>Not Specified</xs:documentation>
          <xs:appinfo>
            <lx:path>proportionsList.Not Specified</lx:path>
            <lx:label>Not Specified</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,SVC,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Specified">
        <xs:annotation>
          <xs:documentation>Specified</xs:documentation>
          <xs:appinfo>
            <lx:path>proportionsList.Specified</lx:path>
            <lx:label>Specified</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,SVC,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="proposedRepaymentMethodList">
    <xs:annotation>
      <xs:documentation>proposedRepaymentMethodList</xs:documentation>
      <xs:appinfo>
        <lx:path>proposedRepaymentMethodList</lx:path>
        <lx:label>proposedRepaymentMethodList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Australia Post">
        <xs:annotation>
          <xs:documentation>The loan will be repaid by deposit at Australia Post.</xs:documentation>
          <xs:appinfo>
            <lx:path>proposedRepaymentMethodList.Australia Post</lx:path>
            <lx:label>The loan will be repaid by deposit at Australia Post.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bpay">
        <xs:annotation>
          <xs:documentation>The loan will be repaid by Bpay.</xs:documentation>
          <xs:appinfo>
            <lx:path>proposedRepaymentMethodList.Bpay</lx:path>
            <lx:label>The loan will be repaid by Bpay.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Card">
        <xs:annotation>
          <xs:documentation>The loan payments will be drawn from a credit card.</xs:documentation>
          <xs:appinfo>
            <lx:path>proposedRepaymentMethodList.Credit Card</lx:path>
            <lx:label>The loan payments will be drawn from a credit card.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Direct Debit Existing Account">
        <xs:annotation>
          <xs:documentation>The loan will be repaid by direct debit from an bank account that is already open.</xs:documentation>
          <xs:appinfo>
            <lx:path>proposedRepaymentMethodList.Direct Debit Existing Account</lx:path>
            <lx:label>The loan will be repaid by direct debit from an bank account that is already open.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Direct Debit New Account">
        <xs:annotation>
          <xs:documentation>The loan will be repaid by direct debit from an bank account that will be opened in future.</xs:documentation>
          <xs:appinfo>
            <lx:path>proposedRepaymentMethodList.Direct Debit New Account</lx:path>
            <lx:label>The loan will be repaid by direct debit from an bank account that will be opened in future.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Direct Salary Credit">
        <xs:annotation>
          <xs:documentation>The loan will be repaid by direct payment from the person's employer under a salary credit arrangement.</xs:documentation>
          <xs:appinfo>
            <lx:path>proposedRepaymentMethodList.Direct Salary Credit</lx:path>
            <lx:label>The loan will be repaid by direct payment from the person's employer under a salary credit arrangement.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Funds Transfer">
        <xs:annotation>
          <xs:documentation>Payments will be made by direct funds transfer</xs:documentation>
          <xs:appinfo>
            <lx:path>proposedRepaymentMethodList.Funds Transfer</lx:path>
            <lx:label>Payments will be made by direct funds transfer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lender Branch">
        <xs:annotation>
          <xs:documentation>Payments will be made by deposit at a branch of the lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>proposedRepaymentMethodList.Lender Branch</lx:path>
            <lx:label>Payments will be made by deposit at a branch of the lender.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>The loan repayment method is not described in one of the other enumerated values.</xs:documentation>
          <xs:appinfo>
            <lx:path>proposedRepaymentMethodList.Other</lx:path>
            <lx:label>The loan repayment method is not described in one of the other enumerated values.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Staff Pay">
        <xs:annotation>
          <xs:documentation>The loan will be repaid by deduction from the staff member's pay; only applies when the borrower works for the lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>proposedRepaymentMethodList.Staff Pay</lx:path>
            <lx:label>The loan will be repaid by deduction from the staff member's pay; only applies when the borrower works for the lender.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="purposeOfApplicationList">
    <xs:annotation>
      <xs:documentation>purposeOfApplicationList</xs:documentation>
      <xs:appinfo>
        <lx:path>purposeOfApplicationList</lx:path>
        <lx:label>purposeOfApplicationList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Bridging Interest">
        <xs:annotation>
          <xs:documentation>Funds are intended to cover interest incurred on a bridging loan.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Bridging Interest</lx:path>
            <lx:label>Bridging Interest</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Business">
        <xs:annotation>
          <xs:documentation>Funds are intended for purchasing a business or any business related expenses.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Business</lx:path>
            <lx:label>Business</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Commercial Development">
        <xs:annotation>
          <xs:documentation>Funds are intended for the development of commercial property.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Commercial Development</lx:path>
            <lx:label>Commercial Development</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Computer">
        <xs:annotation>
          <xs:documentation>Funds are intended for the purchase of a computer or computer devices.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Computer</lx:path>
            <lx:label>Computer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Construction">
        <xs:annotation>
          <xs:documentation>Funds are intended for construction of a property asset.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Construction</lx:path>
            <lx:label>Construction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Consumer Goods">
        <xs:annotation>
          <xs:documentation>Funds are intended for purchase of consumer goods such as whitegoods.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Consumer Goods</lx:path>
            <lx:label>Consumer Goods</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Credit Card">
        <xs:annotation>
          <xs:documentation>Applying for a credit card product.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Credit Card</lx:path>
            <lx:label>Credit Card</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Debt Consolidation">
        <xs:annotation>
          <xs:documentation>Funds are intended for the consolidation of existing debt.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Debt Consolidation</lx:path>
            <lx:label>Debt Consolidation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dental Expenses">
        <xs:annotation>
          <xs:documentation>Funds are intended for dental expenses.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Dental Expenses</lx:path>
            <lx:label>Dental Expenses</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Deposit Account">
        <xs:annotation>
          <xs:documentation>Applying to open a deposit account.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Deposit Account</lx:path>
            <lx:label>Deposit Account</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Education Expenses">
        <xs:annotation>
          <xs:documentation>Funds are intended for expenses related to education.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Education Expenses</lx:path>
            <lx:label>Education Expenses</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Holiday Expenses">
        <xs:annotation>
          <xs:documentation>Funds are intended for holiday expenses.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Holiday Expenses</lx:path>
            <lx:label>Holiday Expenses</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Home Deposit">
        <xs:annotation>
          <xs:documentation>Funds are intended for a deposit taken when purchasing a home.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Home Deposit</lx:path>
            <lx:label>Home Deposit</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Home Equity">
        <xs:annotation>
          <xs:documentation>Funds are intended to be taken from the equity in a home.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Home Equity</lx:path>
            <lx:label>Home Equity</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Home Improvements">
        <xs:annotation>
          <xs:documentation>Funds are intended for home improvements such as renovations.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Home Improvements</lx:path>
            <lx:label>Home Improvements</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Household Items">
        <xs:annotation>
          <xs:documentation>Funds are intended for purchase of household items.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Household Items</lx:path>
            <lx:label>Household Items</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Legal Fees">
        <xs:annotation>
          <xs:documentation>Funds are intended for any legal fees incurred.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Legal Fees</lx:path>
            <lx:label>Legal Fees</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Marine">
        <xs:annotation>
          <xs:documentation>Funds are intended for the purchase of marine equipment.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Marine</lx:path>
            <lx:label>Marine</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Marital Settlement">
        <xs:annotation>
          <xs:documentation>Funds are intended for marital settlement.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Marital Settlement</lx:path>
            <lx:label>Marital Settlement</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Medical Expenses">
        <xs:annotation>
          <xs:documentation>Funds are intended for medical expenses.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Medical Expenses</lx:path>
            <lx:label>Medical Expenses</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Funds are intended for a purpose not described in any of the other enumerated values.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Overdraft">
        <xs:annotation>
          <xs:documentation>Applying for an overdraft or adding an overdraft facility to an existing account.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Overdraft</lx:path>
            <lx:label>Overdraft</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pay Taxes">
        <xs:annotation>
          <xs:documentation>Funds are intended for the payment of tax debt.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Pay Taxes</lx:path>
            <lx:label>Pay Taxes</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Personal Use">
        <xs:annotation>
          <xs:documentation>Funds are intended for potentially unspecified personal use.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Personal Use</lx:path>
            <lx:label>Personal Use</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Plant and Equipment">
        <xs:annotation>
          <xs:documentation>Funds are intended for the purchase of plant and equipment.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Plant and Equipment</lx:path>
            <lx:label>Plant and Equipment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Property Settlement">
        <xs:annotation>
          <xs:documentation>Funds are intended for settling the purchase of a property.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Property Settlement</lx:path>
            <lx:label>Property Settlement</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase Caravan">
        <xs:annotation>
          <xs:documentation>Funds are intended for the purchase of a caravan.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Purchase Caravan</lx:path>
            <lx:label>Purchase Caravan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase Land and Build">
        <xs:annotation>
          <xs:documentation>Funds are intended for the purchase of land and subsequent construction of a real estate asset on the land.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Purchase Land and Build</lx:path>
            <lx:label>Purchase Land and Build</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase Real Estate">
        <xs:annotation>
          <xs:documentation>Funds are intended for the purchase of real estate.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Purchase Real Estate</lx:path>
            <lx:label>Purchase Real Estate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase Shares">
        <xs:annotation>
          <xs:documentation>Funds are intended for the purchase of shares.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Purchase Shares</lx:path>
            <lx:label>Purchase Shares</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase Vehicle">
        <xs:annotation>
          <xs:documentation>Funds are intended for the purchase of a vehicle.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Purchase Vehicle</lx:path>
            <lx:label>Purchase Vehicle</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Recreation Expenses">
        <xs:annotation>
          <xs:documentation>Funds are intended for recreation expenses.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Recreation Expenses</lx:path>
            <lx:label>Recreation Expenses</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Refinance Home Loan">
        <xs:annotation>
          <xs:documentation>Funds are intended for refinancing an existing home loan.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Refinance Home Loan</lx:path>
            <lx:label>Refinance Home Loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Refinance Other">
        <xs:annotation>
          <xs:documentation>Funds are intended for refinancing unspecified existing debt.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Refinance Other</lx:path>
            <lx:label>Refinance Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Refinance Vehicle Loan">
        <xs:annotation>
          <xs:documentation>Funds are intended for refinancing an existing vehicle loan.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Refinance Vehicle Loan</lx:path>
            <lx:label>Refinance Vehicle Loan</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Travel Requirements">
        <xs:annotation>
          <xs:documentation>Funds are intended for travel or travel expenses.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Travel Requirements</lx:path>
            <lx:label>Travel Requirements</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vehicle Expenses">
        <xs:annotation>
          <xs:documentation>Funds are intended for vehicle expenses.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Vehicle Expenses</lx:path>
            <lx:label>Vehicle Expenses</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wedding Expenses">
        <xs:annotation>
          <xs:documentation>Funds are intended for wedding expenses.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfApplicationList.Wedding Expenses</lx:path>
            <lx:label>Wedding Expenses</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="purposeOfFundsList">
    <xs:annotation>
      <xs:documentation>purposeOfFundsList</xs:documentation>
      <xs:appinfo>
        <lx:path>purposeOfFundsList</lx:path>
        <lx:label>purposeOfFundsList</lx:label>
        <li:transactions>CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Construction of Dwelling">
        <xs:annotation>
          <xs:documentation>Funds are to be used for the construction of a dwelling.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Construction of Dwelling</lx:path>
            <lx:label>Construction of Dwelling</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Debt Consolidation">
        <xs:annotation>
          <xs:documentation>Funds are to be used to consolidate short term debts into a home loan.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Debt Consolidation</lx:path>
            <lx:label>Debt Consolidation</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="EECA Energy Efficiency Funding">
        <xs:annotation>
          <xs:documentation>Funds are to be used for purchase a dwelling through the EECA Energywise programme.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.EECA Energy Efficiency Funding</lx:path>
            <lx:label>EECA Energy Efficiency Funding</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fix Leaky Home">
        <xs:annotation>
          <xs:documentation>Funds are to be used for undertaking repairs on a property that has been identified as 'Leaky'.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Fix Leaky Home</lx:path>
            <lx:label>Fix Leaky Home</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fix Natural Disaster Affected Home">
        <xs:annotation>
          <xs:documentation>Funds are to be used for undertaking repairs on a property that has been affected by a natural disaster.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Fix Natural Disaster Affected Home</lx:path>
            <lx:label>Fix Natural Disaster Affected Home</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Holiday /Travel">
        <xs:annotation>
          <xs:documentation>Funds are to be used to pay for holiday and/or travel expenses.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Holiday /Travel</lx:path>
            <lx:label>Holiday /Travel</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Investment">
        <xs:annotation>
          <xs:documentation>Funds are to be used for investment.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Investment</lx:path>
            <lx:label>Investment</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other Purpose">
        <xs:annotation>
          <xs:documentation>Funds are to be used for another purpose.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Other Purpose</lx:path>
            <lx:label>Other Purpose</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase of Business or Shares">
        <xs:annotation>
          <xs:documentation>Funds are to be used to purchase business or shares.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Purchase of Business or Shares</lx:path>
            <lx:label>Purchase of Business or Shares</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase of Dwelling">
        <xs:annotation>
          <xs:documentation>Funds are to be used for the purchase of a newly erected dwelling.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Purchase of Dwelling</lx:path>
            <lx:label>Purchase of Dwelling</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase of Equipment or Assets">
        <xs:annotation>
          <xs:documentation>Funds are to be used to purchase equipment or assets.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Purchase of Equipment or Assets</lx:path>
            <lx:label>Purchase of Equipment or Assets</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase of Property">
        <xs:annotation>
          <xs:documentation>Funds are to be used to purchase property.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Purchase of Property</lx:path>
            <lx:label>Purchase of Property</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase Vehicle">
        <xs:annotation>
          <xs:documentation>Funds are to be used to purchase a vehicle.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Purchase Vehicle</lx:path>
            <lx:label>Purchase Vehicle</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Refinance of Liabilities with Current Lender">
        <xs:annotation>
          <xs:documentation>Funds are to be used for the refinance of a mortgage or other liabilities with the current lender.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Refinance of Liabilities with Current Lender</lx:path>
            <lx:label>Refinance of Liabilities with Current Lender</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Refinance of Liabilities with Other Lenders">
        <xs:annotation>
          <xs:documentation>Funds are to be used for the refinance of a mortgage or other liabilities with other lenders.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Refinance of Liabilities with Other Lenders</lx:path>
            <lx:label>Refinance of Liabilities with Other Lenders</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Renovations and Extensions">
        <xs:annotation>
          <xs:documentation>Funds are to be used for completing renovations or alterations on an existing dwelling.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Renovations and Extensions</lx:path>
            <lx:label>Renovations and Extensions</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Working Capital">
        <xs:annotation>
          <xs:documentation>Funds are to be used for working capital.</xs:documentation>
          <xs:appinfo>
            <lx:path>purposeOfFundsList.Working Capital</lx:path>
            <lx:label>Working Capital</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="ratingList">
    <xs:annotation>
      <xs:documentation>ratingList</xs:documentation>
      <xs:appinfo>
        <lx:path>ratingList</lx:path>
        <lx:label>ratingList</lx:label>
        <li:transactions>LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="0-NotKnown">
        <xs:annotation>
          <xs:documentation>Not Known</xs:documentation>
          <xs:appinfo>
            <lx:path>ratingList.0-NotKnown</lx:path>
            <lx:label>Not Known</lx:label>
            <li:transactions>LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1-Poor">
        <xs:annotation>
          <xs:documentation>Poor</xs:documentation>
          <xs:appinfo>
            <lx:path>ratingList.1-Poor</lx:path>
            <lx:label>Poor</lx:label>
            <li:transactions>LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2-Fair">
        <xs:annotation>
          <xs:documentation>Fair</xs:documentation>
          <xs:appinfo>
            <lx:path>ratingList.2-Fair</lx:path>
            <lx:label>Fair</lx:label>
            <li:transactions>LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3-Good">
        <xs:annotation>
          <xs:documentation>Good</xs:documentation>
          <xs:appinfo>
            <lx:path>ratingList.3-Good</lx:path>
            <lx:label>Good</lx:label>
            <li:transactions>LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4-VeryGood">
        <xs:annotation>
          <xs:documentation>Very Good</xs:documentation>
          <xs:appinfo>
            <lx:path>ratingList.4-VeryGood</lx:path>
            <lx:label>Very Good</lx:label>
            <li:transactions>LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5-Excellent">
        <xs:annotation>
          <xs:documentation>Excellent</xs:documentation>
          <xs:appinfo>
            <lx:path>ratingList.5-Excellent</lx:path>
            <lx:label>Excellent</lx:label>
            <li:transactions>LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="rbaLendingPurposeList">
    <xs:annotation>
      <xs:documentation>Enforces that a value entered for lending purposes is RBA specific.</xs:documentation>
      <xs:appinfo>
        <lx:path>rbaLendingPurposeList</lx:path>
        <lx:label>rbaLendingPurposeList</lx:label>
        <li:transactions>CAL,CDA,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Construction">
        <xs:annotation>
          <xs:documentation>Construction</xs:documentation>
          <xs:appinfo>
            <lx:path>rbaLendingPurposeList.Construction</lx:path>
            <lx:label>Construction</lx:label>
            <li:transactions>CAL,CDA,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Further Advance">
        <xs:annotation>
          <xs:documentation>Further Advance</xs:documentation>
          <xs:appinfo>
            <lx:path>rbaLendingPurposeList.Further Advance</lx:path>
            <lx:label>Further Advance</lx:label>
            <li:transactions>CAL,CDA,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>rbaLendingPurposeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase - Existing Dwelling">
        <xs:annotation>
          <xs:documentation>Purchase - Existing Dwelling</xs:documentation>
          <xs:appinfo>
            <lx:path>rbaLendingPurposeList.Purchase - Existing Dwelling</lx:path>
            <lx:label>Purchase - Existing Dwelling</lx:label>
            <li:transactions>CAL,CDA,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase - New Dwelling">
        <xs:annotation>
          <xs:documentation>Purchase - New dwelling</xs:documentation>
          <xs:appinfo>
            <lx:path>rbaLendingPurposeList.Purchase - New Dwelling</lx:path>
            <lx:label>Purchase - New Dwelling</lx:label>
            <li:transactions>CAL,CDA,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purchase - Other">
        <xs:annotation>
          <xs:documentation>Purchase - Other</xs:documentation>
          <xs:appinfo>
            <lx:path>rbaLendingPurposeList.Purchase - Other</lx:path>
            <lx:label>Purchase - Other</lx:label>
            <li:transactions>CAL,CDA,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Refinance">
        <xs:annotation>
          <xs:documentation>Refinance</xs:documentation>
          <xs:appinfo>
            <lx:path>rbaLendingPurposeList.Refinance</lx:path>
            <lx:label>Refinance</lx:label>
            <li:transactions>CAL,CDA,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="realEstateAssetEncumbranceList">
    <xs:annotation>
      <xs:documentation>realEstateAssetEncumbranceList</xs:documentation>
      <xs:appinfo>
        <lx:path>realEstateAssetEncumbranceList</lx:path>
        <lx:label>realEstateAssetEncumbranceList</lx:label>
        <li:transactions>CAL,DAS,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Caveat">
        <xs:annotation>
          <xs:documentation>Caveat</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetEncumbranceList.Caveat</lx:path>
            <lx:label>Caveat</lx:label>
            <li:transactions>CAL,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Charge">
        <xs:annotation>
          <xs:documentation>Charge</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetEncumbranceList.Charge</lx:path>
            <lx:label>Charge</lx:label>
            <li:transactions>CAL,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Covenant">
        <xs:annotation>
          <xs:documentation>Covenant</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetEncumbranceList.Covenant</lx:path>
            <lx:label>Covenant</lx:label>
            <li:transactions>CAL,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Easement">
        <xs:annotation>
          <xs:documentation>Easement</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetEncumbranceList.Easement</lx:path>
            <lx:label>Easement</lx:label>
            <li:transactions>CAL,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="FiFa">
        <xs:annotation>
          <xs:documentation>FiFa</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetEncumbranceList.FiFa</lx:path>
            <lx:label>FiFa</lx:label>
            <li:transactions>CAL,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Land Tax Charge">
        <xs:annotation>
          <xs:documentation>Land Tax Charge</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetEncumbranceList.Land Tax Charge</lx:path>
            <lx:label>Land Tax Charge</lx:label>
            <li:transactions>CAL,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lease">
        <xs:annotation>
          <xs:documentation>Lease</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetEncumbranceList.Lease</lx:path>
            <lx:label>Lease</lx:label>
            <li:transactions>CAL,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Memorial">
        <xs:annotation>
          <xs:documentation>Memorial</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetEncumbranceList.Memorial</lx:path>
            <lx:label>Memorial</lx:label>
            <li:transactions>CAL,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mortgage">
        <xs:annotation>
          <xs:documentation>Mortgage</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetEncumbranceList.Mortgage</lx:path>
            <lx:label>Mortgage</lx:label>
            <li:transactions>CAL,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="None Disclosed">
        <xs:annotation>
          <xs:documentation>None Disclosed</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetEncumbranceList.None Disclosed</lx:path>
            <lx:label>None Disclosed</lx:label>
            <li:transactions>CAL,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Not Known">
        <xs:annotation>
          <xs:documentation>Not Known</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetEncumbranceList.Not Known</lx:path>
            <lx:label>Not Known</lx:label>
            <li:transactions>CAL,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetEncumbranceList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Warrant">
        <xs:annotation>
          <xs:documentation>Warrant</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetEncumbranceList.Warrant</lx:path>
            <lx:label>Warrant</lx:label>
            <li:transactions>CAL,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Writ">
        <xs:annotation>
          <xs:documentation>Writ</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetEncumbranceList.Writ</lx:path>
            <lx:label>Writ</lx:label>
            <li:transactions>CAL,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="realEstateAssetStatusList">
    <xs:annotation>
      <xs:documentation>realEstateAssetStatusList</xs:documentation>
      <xs:appinfo>
        <lx:path>realEstateAssetStatusList</lx:path>
        <lx:label>realEstateAssetStatusList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Established">
        <xs:annotation>
          <xs:documentation>The dwelling on the property is an established home i.e. not newly constructed.</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetStatusList.Established</lx:path>
            <lx:label>The dwelling on the property is an established home i.e. not newly constructed.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="New Building">
        <xs:annotation>
          <xs:documentation>The dwelling on the property is newly constructed.</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetStatusList.New Building</lx:path>
            <lx:label>The dwelling on the property is newly constructed.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Off the Plan">
        <xs:annotation>
          <xs:documentation>A new construction that will be completed by the time of settlement.</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetStatusList.Off the Plan</lx:path>
            <lx:label>A new construction that will be completed by the time of settlement.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vacant Land">
        <xs:annotation>
          <xs:documentation>There is no dwelling on the property.</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateAssetStatusList.Vacant Land</lx:path>
            <lx:label>There is no dwelling on the property.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="realEstateRequestedPriorityList">
    <xs:annotation>
      <xs:documentation>realEstateRequestedPriorityList</xs:documentation>
      <xs:appinfo>
        <lx:path>realEstateRequestedPriorityList</lx:path>
        <lx:label>realEstateRequestedPriorityList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="High">
        <xs:annotation>
          <xs:documentation>The priority for the requested valuation is high.</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateRequestedPriorityList.High</lx:path>
            <lx:label>High</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Low">
        <xs:annotation>
          <xs:documentation>The priority for the requested valuation is low.</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateRequestedPriorityList.Low</lx:path>
            <lx:label>Low</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Normal">
        <xs:annotation>
          <xs:documentation>The priority for the requested valuation is normal.</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateRequestedPriorityList.Normal</lx:path>
            <lx:label>Normal</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="realEstateValuationFeeStatusList">
    <xs:annotation>
      <xs:documentation>realEstateValuationFeeStatusList</xs:documentation>
      <xs:appinfo>
        <lx:path>realEstateValuationFeeStatusList</lx:path>
        <lx:label>realEstateValuationFeeStatusList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Accepted">
        <xs:annotation>
          <xs:documentation>Accepted</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationFeeStatusList.Accepted</lx:path>
            <lx:label>Accepted</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cancelled">
        <xs:annotation>
          <xs:documentation>Cancelled</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationFeeStatusList.Cancelled</lx:path>
            <lx:label>Cancelled</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Declined">
        <xs:annotation>
          <xs:documentation>Declined</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationFeeStatusList.Declined</lx:path>
            <lx:label>Declined</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Requested">
        <xs:annotation>
          <xs:documentation>Requested</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationFeeStatusList.Requested</lx:path>
            <lx:label>Requested</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="realEstateValuationReasonList">
    <xs:annotation>
      <xs:documentation>realEstateValuationReasonList</xs:documentation>
      <xs:appinfo>
        <lx:path>realEstateValuationReasonList</lx:path>
        <lx:label>realEstateValuationReasonList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Audit">
        <xs:annotation>
          <xs:documentation>Audit</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationReasonList.Audit</lx:path>
            <lx:label>Audit</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Client Discharge">
        <xs:annotation>
          <xs:documentation>Client Discharge</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationReasonList.Client Discharge</lx:path>
            <lx:label>Client Discharge</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fair Market Value">
        <xs:annotation>
          <xs:documentation>Fair Market Value</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationReasonList.Fair Market Value</lx:path>
            <lx:label>Fair Market Value</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Maximum Borrowing Capacity">
        <xs:annotation>
          <xs:documentation>Maximum Borrowing Capacity</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationReasonList.Maximum Borrowing Capacity</lx:path>
            <lx:label>Maximum Borrowing Capacity</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mortgagee Sale">
        <xs:annotation>
          <xs:documentation>Mortgagee Sale</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationReasonList.Mortgagee Sale</lx:path>
            <lx:label>Mortgagee Sale</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationReasonList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Progress Drawdown">
        <xs:annotation>
          <xs:documentation>Progress Drawdown</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationReasonList.Progress Drawdown</lx:path>
            <lx:label>Progress Drawdown</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Revaluation">
        <xs:annotation>
          <xs:documentation>Revaluation</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationReasonList.Revaluation</lx:path>
            <lx:label>Revaluation</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="realEstateValuationTypeList">
    <xs:annotation>
      <xs:documentation>realEstateValuationTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>realEstateValuationTypeList</lx:path>
        <lx:label>realEstateValuationTypeList</lx:label>
        <li:transactions>CDA,LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="AVM">
        <xs:annotation>
          <xs:documentation>An AVM (Automated Valuation Model) requires no input from a qualified valuer, but instead use computer models based on attribute data and recent comparable sales data to estimate a value.</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationTypeList.AVM</lx:path>
            <lx:label>AVM</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Costing">
        <xs:annotation>
          <xs:documentation>A Costing is an evaluation of a construction contract to determine if the contract price is appropriate for the building being constructed, and to estimate the market value as if complete. Used for off the plan or To Be Erected (TBE) property.</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationTypeList.Costing</lx:path>
            <lx:label>Costing</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Desktop">
        <xs:annotation>
          <xs:documentation>A Desktop Valuation requires no internal or external inspection, instead qualified valuers use data about the property (attributes, photos, and comparable sales) to value a property.</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationTypeList.Desktop</lx:path>
            <lx:label>Desktop</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Full Registered">
        <xs:annotation>
          <xs:documentation>A Full Registered Valuation requires an internal and external inspection of the property by a qualified valuer.</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationTypeList.Full Registered</lx:path>
            <lx:label>Full Registered</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Kerb Side">
        <xs:annotation>
          <xs:documentation>A Kerb Side Valuation requires an external sighting of the property from the street by a qualified valuer.</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationTypeList.Kerb Side</lx:path>
            <lx:label>Kerb Side</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="None">
        <xs:annotation>
          <xs:documentation>None</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationTypeList.None</lx:path>
            <lx:label>None</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Progress Inspection">
        <xs:annotation>
          <xs:documentation>A Progress Inspection is an inspection and valuation of a building under construction.</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationTypeList.Progress Inspection</lx:path>
            <lx:label>Progress Inspection</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rates Notice">
        <xs:annotation>
          <xs:documentation>An official rates notice provided by the surveyor general.</xs:documentation>
          <xs:appinfo>
            <lx:path>realEstateValuationTypeList.Rates Notice</lx:path>
            <lx:label>Rates Notice</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="reasonForDifferentNameList">
    <xs:annotation>
      <xs:documentation>reasonForDifferentNameList</xs:documentation>
      <xs:appinfo>
        <lx:path>reasonForDifferentNameList</lx:path>
        <lx:label>reasonForDifferentNameList</lx:label>
        <li:transactions>CAL,CDA,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Abbreviation">
        <xs:annotation>
          <xs:documentation>Approved abbreviation.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Abbreviation</lx:path>
            <lx:label>Abbreviation</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Amalgamation">
        <xs:annotation>
          <xs:documentation>Amalgamation - Hospitals - Registered proprietor(s) may still exist.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Amalgamation</lx:path>
            <lx:label>Amalgamation</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Amalgamation Evidence">
        <xs:annotation>
          <xs:documentation>Amalgamation - Hospitals - Registered proprietor(s) may still exist - Registrar holds satisfactory evidence.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Amalgamation Evidence</lx:path>
            <lx:label>Amalgamation Evidence</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Change Incorporated Name">
        <xs:annotation>
          <xs:documentation>Change of incorporated name.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Change Incorporated Name</lx:path>
            <lx:label>Change Incorporated Name</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Deed Poll">
        <xs:annotation>
          <xs:documentation>Deed Poll.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Deed Poll</lx:path>
            <lx:label>Deed Poll</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Error in Register">
        <xs:annotation>
          <xs:documentation>Error in Register.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Error in Register</lx:path>
            <lx:label>Error in Register</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Legislative Change Company">
        <xs:annotation>
          <xs:documentation>Legislative change - Company - Registered proprietor(s) may still exist.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Legislative Change Company</lx:path>
            <lx:label>Legislative Change Company</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Legislative Change Company Evidence">
        <xs:annotation>
          <xs:documentation>Legislative change - Company - Registered proprietor(s) may still exist - Registrar holds satisfactory evidence.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Legislative Change Company Evidence</lx:path>
            <lx:label>Legislative Change Company Evidence</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Legislative Change Non Company">
        <xs:annotation>
          <xs:documentation>Legislative change - Non Company - Registered proprietor(s) may still exist.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Legislative Change Non Company</lx:path>
            <lx:label>Legislative Change Non Company</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Legislative Change Non Company Evidence">
        <xs:annotation>
          <xs:documentation>Legislative change - Non Company - Registered proprietor(s) may still exist - Registrar holds satisfactory evidence.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Legislative Change Non Company Evidence</lx:path>
            <lx:label>Legislative Change Non Company Evidence</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Marriage">
        <xs:annotation>
          <xs:documentation>Marriage.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Marriage</lx:path>
            <lx:label>Marriage</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Merger or Takeover">
        <xs:annotation>
          <xs:documentation>Merger or takeover - Registered proprietor(s) may still exist.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Merger or Takeover</lx:path>
            <lx:label>Merger or Takeover</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Merger or Takeover Evidence">
        <xs:annotation>
          <xs:documentation>Merger or takeover - Registered proprietor(s) may still exist - Registrar holds satisfactory evidence.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Merger or Takeover Evidence</lx:path>
            <lx:label>Merger or Takeover Evidence</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="New Name">
        <xs:annotation>
          <xs:documentation>Adoption of new name.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.New Name</lx:path>
            <lx:label>New Name</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Resumption of Maiden Name">
        <xs:annotation>
          <xs:documentation>Adoption of new name.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Resumption of Maiden Name</lx:path>
            <lx:label>Resumption of Maiden Name</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sale or Transfer">
        <xs:annotation>
          <xs:documentation>Sale or transfer of enterprise - Registered Proprietor(s) may still exist.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Sale or Transfer</lx:path>
            <lx:label>Sale or Transfer</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sale or Transfer Evidence">
        <xs:annotation>
          <xs:documentation>Sale or transfer of enterprise - Registered Proprietor(s) may still exist - Registrar holds satisfactory evidence.</xs:documentation>
          <xs:appinfo>
            <lx:path>reasonForDifferentNameList.Sale or Transfer Evidence</lx:path>
            <lx:label>Sale or Transfer Evidence</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="registrationEventTypeList">
    <xs:annotation>
      <xs:documentation>The types of PPSR registration event.</xs:documentation>
      <xs:appinfo>
        <lx:path>registrationEventTypeList</lx:path>
        <lx:label>registrationEventTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Amendment">
        <xs:annotation>
          <xs:documentation>Amendment</xs:documentation>
          <xs:appinfo>
            <lx:path>registrationEventTypeList.Amendment</lx:path>
            <lx:label>Amendment</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Creation">
        <xs:annotation>
          <xs:documentation>Creation</xs:documentation>
          <xs:appinfo>
            <lx:path>registrationEventTypeList.Creation</lx:path>
            <lx:label>Creation</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Discharge">
        <xs:annotation>
          <xs:documentation>Discharge</xs:documentation>
          <xs:appinfo>
            <lx:path>registrationEventTypeList.Discharge</lx:path>
            <lx:label>Discharge</lx:label>
            <li:transactions>CAL,CDA,DAS,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="relatedSearchReferenceTypeList">
    <xs:annotation>
      <xs:documentation>relatedSearchReferenceTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>relatedSearchReferenceTypeList</lx:path>
        <lx:label>relatedSearchReferenceTypeList</lx:label>
        <li:transactions>CDA</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Name Cross Reference">
        <xs:annotation>
          <xs:documentation>Name Cross Reference</xs:documentation>
          <xs:appinfo>
            <lx:path>relatedSearchReferenceTypeList.Name Cross Reference</lx:path>
            <lx:label>Name Cross Reference</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Previous Name">
        <xs:annotation>
          <xs:documentation>Previous Name</xs:documentation>
          <xs:appinfo>
            <lx:path>relatedSearchReferenceTypeList.Previous Name</lx:path>
            <lx:label>Previous Name</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Previous Name Cross Reference">
        <xs:annotation>
          <xs:documentation>Previous Name Cross Reference</xs:documentation>
          <xs:appinfo>
            <lx:path>relatedSearchReferenceTypeList.Previous Name Cross Reference</lx:path>
            <lx:label>Previous Name Cross Reference</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="releaseSecurityReasonList">
    <xs:annotation>
      <xs:documentation>releaseSecurityReasonList</xs:documentation>
      <xs:appinfo>
        <lx:path>releaseSecurityReasonList</lx:path>
        <lx:label>releaseSecurityReasonList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Debt Reduction">
        <xs:annotation>
          <xs:documentation>Debt Reduction</xs:documentation>
          <xs:appinfo>
            <lx:path>releaseSecurityReasonList.Debt Reduction</lx:path>
            <lx:label>Debt Reduction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dissolving Partnership">
        <xs:annotation>
          <xs:documentation>Dissolving Partnership</xs:documentation>
          <xs:appinfo>
            <lx:path>releaseSecurityReasonList.Dissolving Partnership</lx:path>
            <lx:label>Dissolving Partnership</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="No Longer Required">
        <xs:annotation>
          <xs:documentation>No Longer Required</xs:documentation>
          <xs:appinfo>
            <lx:path>releaseSecurityReasonList.No Longer Required</lx:path>
            <lx:label>No Longer Required</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Refinancing with Another Lender">
        <xs:annotation>
          <xs:documentation>Refinancing with Another Lender</xs:documentation>
          <xs:appinfo>
            <lx:path>releaseSecurityReasonList.Refinancing with Another Lender</lx:path>
            <lx:label>Refinancing with Another Lender</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sale of Property">
        <xs:annotation>
          <xs:documentation>Sale of Property</xs:documentation>
          <xs:appinfo>
            <lx:path>releaseSecurityReasonList.Sale of Property</lx:path>
            <lx:label>Sale of Property</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="remotenessList">
    <xs:annotation>
      <xs:documentation>The Australian Statistical Geography Standard (ASGS) defines remoteness areas into 5 classes of relative remoteness across Australia.</xs:documentation>
      <xs:appinfo>
        <lx:path>remotenessList</lx:path>
        <lx:label>remotenessList</lx:label>
        <li:transactions>ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Inner Regional">
        <xs:annotation>
          <xs:documentation>Inner Regional</xs:documentation>
          <xs:appinfo>
            <lx:path>remotenessList.Inner Regional</lx:path>
            <lx:label>Inner Regional</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Major Cities">
        <xs:annotation>
          <xs:documentation>Major Cities</xs:documentation>
          <xs:appinfo>
            <lx:path>remotenessList.Major Cities</lx:path>
            <lx:label>Major Cities</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Outer Regional">
        <xs:annotation>
          <xs:documentation>Outer Regional</xs:documentation>
          <xs:appinfo>
            <lx:path>remotenessList.Outer Regional</lx:path>
            <lx:label>Outer Regional</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Remote">
        <xs:annotation>
          <xs:documentation>Remote</xs:documentation>
          <xs:appinfo>
            <lx:path>remotenessList.Remote</lx:path>
            <lx:label>Remote</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Very Remote">
        <xs:annotation>
          <xs:documentation>Very Remote</xs:documentation>
          <xs:appinfo>
            <lx:path>remotenessList.Very Remote</lx:path>
            <lx:label>Very Remote</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="rentalIncomeBasisList">
    <xs:annotation>
      <xs:documentation>rentalIncomeBasisList</xs:documentation>
      <xs:appinfo>
        <lx:path>rentalIncomeBasisList</lx:path>
        <lx:label>rentalIncomeBasisList</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Applicant Estimate">
        <xs:annotation>
          <xs:documentation>Applicant Estimate</xs:documentation>
          <xs:appinfo>
            <lx:path>rentalIncomeBasisList.Applicant Estimate</lx:path>
            <lx:label>Applicant Estimate</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RP Data">
        <xs:annotation>
          <xs:documentation>RP Data</xs:documentation>
          <xs:appinfo>
            <lx:path>rentalIncomeBasisList.RP Data</lx:path>
            <lx:label>RP Data</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tenancy Agreement">
        <xs:annotation>
          <xs:documentation>Tenancy Agreement</xs:documentation>
          <xs:appinfo>
            <lx:path>rentalIncomeBasisList.Tenancy Agreement</lx:path>
            <lx:label>Tenancy Agreement</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="repaymentFrequencyList">
    <xs:annotation>
      <xs:documentation>repaymentFrequencyList</xs:documentation>
      <xs:appinfo>
        <lx:path>repaymentFrequencyList</lx:path>
        <lx:label>repaymentFrequencyList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Daily">
        <xs:annotation>
          <xs:documentation>Daily</xs:documentation>
          <xs:appinfo>
            <lx:path>repaymentFrequencyList.Daily</lx:path>
            <lx:label>Daily</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fortnightly">
        <xs:annotation>
          <xs:documentation>Fortnightly</xs:documentation>
          <xs:appinfo>
            <lx:path>repaymentFrequencyList.Fortnightly</lx:path>
            <lx:label>Fortnightly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Half Yearly">
        <xs:annotation>
          <xs:documentation>Half Yearly</xs:documentation>
          <xs:appinfo>
            <lx:path>repaymentFrequencyList.Half Yearly</lx:path>
            <lx:label>Half Yearly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Monthly">
        <xs:annotation>
          <xs:documentation>Monthly</xs:documentation>
          <xs:appinfo>
            <lx:path>repaymentFrequencyList.Monthly</lx:path>
            <lx:label>Monthly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="One Off">
        <xs:annotation>
          <xs:documentation>One Off</xs:documentation>
          <xs:appinfo>
            <lx:path>repaymentFrequencyList.One Off</lx:path>
            <lx:label>One Off</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Quarterly">
        <xs:annotation>
          <xs:documentation>Quarterly</xs:documentation>
          <xs:appinfo>
            <lx:path>repaymentFrequencyList.Quarterly</lx:path>
            <lx:label>Quarterly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Seasonal">
        <xs:annotation>
          <xs:documentation>Seasonal</xs:documentation>
          <xs:appinfo>
            <lx:path>repaymentFrequencyList.Seasonal</lx:path>
            <lx:label>Seasonal</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Weekly">
        <xs:annotation>
          <xs:documentation>Weekly</xs:documentation>
          <xs:appinfo>
            <lx:path>repaymentFrequencyList.Weekly</lx:path>
            <lx:label>Weekly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Yearly">
        <xs:annotation>
          <xs:documentation>Yearly</xs:documentation>
          <xs:appinfo>
            <lx:path>repaymentFrequencyList.Yearly</lx:path>
            <lx:label>Yearly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="reportContentTypeList">
    <xs:annotation>
      <xs:documentation>reportContentTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>reportContentTypeList</lx:path>
        <lx:label>reportContentTypeList</lx:label>
        <li:transactions>CDA,SVC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="HTML">
        <xs:annotation>
          <xs:documentation>HTML formatted string.</xs:documentation>
          <xs:appinfo>
            <lx:path>reportContentTypeList.HTML</lx:path>
            <lx:label>HTML</lx:label>
            <li:transactions>CDA,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Text">
        <xs:annotation>
          <xs:documentation>Plain text string.</xs:documentation>
          <xs:appinfo>
            <lx:path>reportContentTypeList.Text</lx:path>
            <lx:label>Text</lx:label>
            <li:transactions>CDA,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="XML">
        <xs:annotation>
          <xs:documentation>XML formatted string.</xs:documentation>
          <xs:appinfo>
            <lx:path>reportContentTypeList.XML</lx:path>
            <lx:label>XML</lx:label>
            <li:transactions>CDA,SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="residencyStatusList">
    <xs:annotation>
      <xs:documentation>residencyStatusList</xs:documentation>
      <xs:appinfo>
        <lx:path>residencyStatusList</lx:path>
        <lx:label>residencyStatusList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Non Resident">
        <xs:annotation>
          <xs:documentation>Non Resident</xs:documentation>
          <xs:appinfo>
            <lx:path>residencyStatusList.Non Resident</lx:path>
            <lx:label>Non Resident</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Permanently in Australia">
        <xs:annotation>
          <xs:documentation>Legally a Permanent Resident, holding either an Australian Permanent Residency Visa or Australian Citizenship.</xs:documentation>
          <xs:appinfo>
            <lx:path>residencyStatusList.Permanently in Australia</lx:path>
            <lx:label>Permanently in Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Permanently in New Zealand">
        <xs:annotation>
          <xs:documentation>Permanently in New Zealand</xs:documentation>
          <xs:appinfo>
            <lx:path>residencyStatusList.Permanently in New Zealand</lx:path>
            <lx:label>Permanently in New Zealand</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Temporarily in Australia">
        <xs:annotation>
          <xs:documentation>Temporarily in Australia</xs:documentation>
          <xs:appinfo>
            <lx:path>residencyStatusList.Temporarily in Australia</lx:path>
            <lx:label>Temporarily in Australia</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Temporarily in New Zealand">
        <xs:annotation>
          <xs:documentation>Temporarily in New Zealand</xs:documentation>
          <xs:appinfo>
            <lx:path>residencyStatusList.Temporarily in New Zealand</lx:path>
            <lx:label>Temporarily in New Zealand</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="residentialTypeList">
    <xs:annotation>
      <xs:documentation>residentialTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>residentialTypeList</lx:path>
        <lx:label>residentialTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Aged Care Unit">
        <xs:annotation>
          <xs:documentation>The dwelling is an aged care unit.</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Aged Care Unit</lx:path>
            <lx:label>Aged Care Unit</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Apartment Unit Flat">
        <xs:annotation>
          <xs:documentation>The dwelling is an apartment, unit or flat</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Apartment Unit Flat</lx:path>
            <lx:label>The dwelling is an apartment, unit or flat</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bedsitter Bachelor">
        <xs:annotation>
          <xs:documentation>The dwelling is a bedsit in which the bedroom and main living room are one</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Bedsitter Bachelor</lx:path>
            <lx:label>The dwelling is a bedsit in which the bedroom and main living room are one</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Company Title Unit">
        <xs:annotation>
          <xs:documentation>The dwelling is a unit in a block that is under company title</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Company Title Unit</lx:path>
            <lx:label>The dwelling is a unit in a block that is under company title</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Converted Commercial Property">
        <xs:annotation>
          <xs:documentation>The dwelling has been converted from a commercial property, such as a floor in a converted office building</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Converted Commercial Property</lx:path>
            <lx:label>The dwelling has been converted from a commercial property, such as a floor in a converted office building</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Converted Industrial Property">
        <xs:annotation>
          <xs:documentation>The dwelling has been converted from a commercial property, such as a converted warehouse</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Converted Industrial Property</lx:path>
            <lx:label>The dwelling has been converted from a commercial property, such as a converted warehouse</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Converted Motel Units">
        <xs:annotation>
          <xs:documentation>The dwelling is a unit in a building that has been converted from what were originally units in a motel</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Converted Motel Units</lx:path>
            <lx:label>The dwelling is a unit in a building that has been converted from what were originally units in a motel</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Display House">
        <xs:annotation>
          <xs:documentation>The dwelling is a display house.</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Display House</lx:path>
            <lx:label>Display House</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Duplex">
        <xs:annotation>
          <xs:documentation>The dwelling is a duplex, being one of two houses on neighbouring blocks of land and a common wall</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Duplex</lx:path>
            <lx:label>The dwelling is a duplex, being one of two houses on neighbouring blocks of land and a common wall</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fully Detached House">
        <xs:annotation>
          <xs:documentation>The dwelling is a house on its own block of land with no common walls</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Fully Detached House</lx:path>
            <lx:label>The dwelling is a house on its own block of land with no common walls</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="High Density Apartment">
        <xs:annotation>
          <xs:documentation>The dwelling meets the financial institution's definition of a high density apartment.</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.High Density Apartment</lx:path>
            <lx:label>High Density Apartment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Holiday Home">
        <xs:annotation>
          <xs:documentation>The dwelling is used by the owners as a holiday home</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Holiday Home</lx:path>
            <lx:label>The dwelling is used by the owners as a holiday home</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Holiday Rental">
        <xs:annotation>
          <xs:documentation>The dwelling is short term rented to holiday goers</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Holiday Rental</lx:path>
            <lx:label>The dwelling is short term rented to holiday goers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Improved Land">
        <xs:annotation>
          <xs:documentation>The property consists of improved land.</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Improved Land</lx:path>
            <lx:label>Improved Land</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Kit Home">
        <xs:annotation>
          <xs:documentation>The dwelling is a kit home.</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Kit Home</lx:path>
            <lx:label>Kit Home</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Low Density Apartment">
        <xs:annotation>
          <xs:documentation>The dwelling meets the financial institution's definition of a low density apartment.</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Low Density Apartment</lx:path>
            <lx:label>Low Density Apartment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Luxury House">
        <xs:annotation>
          <xs:documentation>The dwelling is a luxury home</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Luxury House</lx:path>
            <lx:label>The dwelling is a luxury home</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Luxury Other">
        <xs:annotation>
          <xs:documentation>The dwelling is a luxury accommodation other than a house</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Luxury Other</lx:path>
            <lx:label>The dwelling is a luxury accommodation other than a house</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Medium Density Apartment">
        <xs:annotation>
          <xs:documentation>The dwelling meets the financial institution's definition of a medium density apartment.</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Medium Density Apartment</lx:path>
            <lx:label>Medium Density Apartment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Property Development">
        <xs:annotation>
          <xs:documentation>Property Development</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Property Development</lx:path>
            <lx:label>Property Development</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Relocatable Home">
        <xs:annotation>
          <xs:documentation>The dwelling is a relocatable home.</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Relocatable Home</lx:path>
            <lx:label>Relocatable Home</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Resort Unit">
        <xs:annotation>
          <xs:documentation>The dwelling is a unit in a holiday resort</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Resort Unit</lx:path>
            <lx:label>The dwelling is a unit in a holiday resort</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Semi Detached House">
        <xs:annotation>
          <xs:documentation>The dwelling is a house on its own block of land with one or more common walls</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Semi Detached House</lx:path>
            <lx:label>The dwelling is a house on its own block of land with one or more common walls</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Serviced Apartment">
        <xs:annotation>
          <xs:documentation>The dwelling is a serviced apartment, being an apartment in a block that is serviced by a third party on contract attached to the property</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Serviced Apartment</lx:path>
            <lx:label>The dwelling is a serviced apartment, being an apartment in a block that is serviced by a third party on contract attached to the property</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Strata Hotel Unit">
        <xs:annotation>
          <xs:documentation>The dwelling is a strata hotel unit.</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Strata Hotel Unit</lx:path>
            <lx:label>Strata Hotel Unit</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Strata Title Unit">
        <xs:annotation>
          <xs:documentation>The dwelling is a unit in a block that is under strata title</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Strata Title Unit</lx:path>
            <lx:label>The dwelling is a unit in a block that is under strata title</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Studio Warehouse Apartment">
        <xs:annotation>
          <xs:documentation>The dwelling is a studio apartment in a converted warehouse</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Studio Warehouse Apartment</lx:path>
            <lx:label>The dwelling is a studio apartment in a converted warehouse</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Terrace">
        <xs:annotation>
          <xs:documentation>The dwelling is a terrace house, being a house on its own block of land in a row of such houses with common side walls</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Terrace</lx:path>
            <lx:label>The dwelling is a terrace house, being a house on its own block of land in a row of such houses with common side walls</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Townhouse">
        <xs:annotation>
          <xs:documentation>The dwelling is a town house, being a dwelling that is not a unit, apartment or flat but having common walls</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Townhouse</lx:path>
            <lx:label>The dwelling is a town house, being a dwelling that is not a unit, apartment or flat but having common walls</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unit Student Accommodation">
        <xs:annotation>
          <xs:documentation>The dwelling is a unit that is rented as student accommodation</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Unit Student Accommodation</lx:path>
            <lx:label>The dwelling is a unit that is rented as student accommodation</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vacant Land">
        <xs:annotation>
          <xs:documentation>No dwelling is currently scheduled for construction, the land is intended to remain unimproved</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Vacant Land</lx:path>
            <lx:label>No dwelling is currently scheduled for construction, the land is intended to remain unimproved</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Villa">
        <xs:annotation>
          <xs:documentation>The dwelling is a villa, or unit</xs:documentation>
          <xs:appinfo>
            <lx:path>residentialTypeList.Villa</lx:path>
            <lx:label>The dwelling is a villa, or unit</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="roadElevationGradeList">
    <xs:annotation>
      <xs:documentation>A list of property elevation grades.</xs:documentation>
      <xs:appinfo>
        <lx:path>roadElevationGradeList</lx:path>
        <lx:label>roadElevationGradeList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Above">
        <xs:annotation>
          <xs:documentation>The property is above the road level.</xs:documentation>
          <xs:appinfo>
            <lx:path>roadElevationGradeList.Above</lx:path>
            <lx:label>Above</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Below">
        <xs:annotation>
          <xs:documentation>The property is below the road level.</xs:documentation>
          <xs:appinfo>
            <lx:path>roadElevationGradeList.Below</lx:path>
            <lx:label>Below</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Level">
        <xs:annotation>
          <xs:documentation>The property is level with the road.</xs:documentation>
          <xs:appinfo>
            <lx:path>roadElevationGradeList.Level</lx:path>
            <lx:label>Level</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="roofMaterialList">
    <xs:annotation>
      <xs:documentation>A list of roof material types.</xs:documentation>
      <xs:appinfo>
        <lx:path>roofMaterialList</lx:path>
        <lx:label>roofMaterialList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Concrete Tiles">
        <xs:annotation>
          <xs:documentation>Concrete Tiles</xs:documentation>
          <xs:appinfo>
            <lx:path>roofMaterialList.Concrete Tiles</lx:path>
            <lx:label>Concrete Tiles</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Copper">
        <xs:annotation>
          <xs:documentation>Copper</xs:documentation>
          <xs:appinfo>
            <lx:path>roofMaterialList.Copper</lx:path>
            <lx:label>Copper</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Corrugated Metal">
        <xs:annotation>
          <xs:documentation>Corrugated Metal</xs:documentation>
          <xs:appinfo>
            <lx:path>roofMaterialList.Corrugated Metal</lx:path>
            <lx:label>Corrugated Metal</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Glass">
        <xs:annotation>
          <xs:documentation>Glass</xs:documentation>
          <xs:appinfo>
            <lx:path>roofMaterialList.Glass</lx:path>
            <lx:label>Glass</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Membrane">
        <xs:annotation>
          <xs:documentation>Membrane</xs:documentation>
          <xs:appinfo>
            <lx:path>roofMaterialList.Membrane</lx:path>
            <lx:label>Membrane</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>roofMaterialList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Shingles">
        <xs:annotation>
          <xs:documentation>Shingles</xs:documentation>
          <xs:appinfo>
            <lx:path>roofMaterialList.Shingles</lx:path>
            <lx:label>Shingles</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Slate Tiles">
        <xs:annotation>
          <xs:documentation>Slate Tiles</xs:documentation>
          <xs:appinfo>
            <lx:path>roofMaterialList.Slate Tiles</lx:path>
            <lx:label>Slate Tiles</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Solar Tiles">
        <xs:annotation>
          <xs:documentation>Solar Tiles</xs:documentation>
          <xs:appinfo>
            <lx:path>roofMaterialList.Solar Tiles</lx:path>
            <lx:label>Solar Tiles</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Steel">
        <xs:annotation>
          <xs:documentation>Steel</xs:documentation>
          <xs:appinfo>
            <lx:path>roofMaterialList.Steel</lx:path>
            <lx:label>Steel</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Terracotta Tiles">
        <xs:annotation>
          <xs:documentation>Terracotta Tiles</xs:documentation>
          <xs:appinfo>
            <lx:path>roofMaterialList.Terracotta Tiles</lx:path>
            <lx:label>Terracotta Tiles</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Thatched">
        <xs:annotation>
          <xs:documentation>Thatched</xs:documentation>
          <xs:appinfo>
            <lx:path>roofMaterialList.Thatched</lx:path>
            <lx:label>Thatched</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Timber Cladding">
        <xs:annotation>
          <xs:documentation>Timber Cladding</xs:documentation>
          <xs:appinfo>
            <lx:path>roofMaterialList.Timber Cladding</lx:path>
            <lx:label>Timber Cladding</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="ruralTypeList">
    <xs:annotation>
      <xs:documentation>ruralTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>ruralTypeList</lx:path>
        <lx:label>ruralTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="8 Hectares or Less">
        <xs:annotation>
          <xs:documentation>The rural property is 8 hectares or less in area.</xs:documentation>
          <xs:appinfo>
            <lx:path>ruralTypeList.8 Hectares or Less</lx:path>
            <lx:label>The rural property is 8 hectares or less in area.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Over 40 Hectares">
        <xs:annotation>
          <xs:documentation>The rural property is greater than 40 hectares in area.</xs:documentation>
          <xs:appinfo>
            <lx:path>ruralTypeList.Over 40 Hectares</lx:path>
            <lx:label>The rural property is greater than 40 hectares in area.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Over 8 Less than 40 Hectares">
        <xs:annotation>
          <xs:documentation>The rural property is greater than 8 hectares but not greater than 40 hectares in area.</xs:documentation>
          <xs:appinfo>
            <lx:path>ruralTypeList.Over 8 Less than 40 Hectares</lx:path>
            <lx:label>The rural property is greater than 8 hectares but not greater than 40 hectares in area.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="ruralUsageList">
    <xs:annotation>
      <xs:documentation>ruralUsageList</xs:documentation>
      <xs:appinfo>
        <lx:path>ruralUsageList</lx:path>
        <lx:label>ruralUsageList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Income Producing">
        <xs:annotation>
          <xs:documentation>The rural property is used for income producing purposes.</xs:documentation>
          <xs:appinfo>
            <lx:path>ruralUsageList.Income Producing</lx:path>
            <lx:label>The rural property is used for income producing purposes.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lifestyle Block">
        <xs:annotation>
          <xs:documentation>The rural property is a Lifestyle Block.</xs:documentation>
          <xs:appinfo>
            <lx:path>ruralUsageList.Lifestyle Block</lx:path>
            <lx:label>Lifestyle Block</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rural Life">
        <xs:annotation>
          <xs:documentation>The rural property is used as a rural residence.</xs:documentation>
          <xs:appinfo>
            <lx:path>ruralUsageList.Rural Life</lx:path>
            <lx:label>The rural property is used as a rural residence.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rural Residential">
        <xs:annotation>
          <xs:documentation>The rural property is used as a residence.</xs:documentation>
          <xs:appinfo>
            <lx:path>ruralUsageList.Rural Residential</lx:path>
            <lx:label>The rural property is used as a residence.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="salesChannelTypeList">
    <xs:annotation>
      <xs:documentation>salesChannelTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>salesChannelTypeList</lx:path>
        <lx:label>salesChannelTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Branch Network">
        <xs:annotation>
          <xs:documentation>Branch Network</xs:documentation>
          <xs:appinfo>
            <lx:path>salesChannelTypeList.Branch Network</lx:path>
            <lx:label>Branch Network</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Call Centre">
        <xs:annotation>
          <xs:documentation>Call Centre</xs:documentation>
          <xs:appinfo>
            <lx:path>salesChannelTypeList.Call Centre</lx:path>
            <lx:label>Call Centre</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mortgage Broker">
        <xs:annotation>
          <xs:documentation>Mortgage Broker</xs:documentation>
          <xs:appinfo>
            <lx:path>salesChannelTypeList.Mortgage Broker</lx:path>
            <lx:label>Mortgage Broker</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mortgage Manager">
        <xs:annotation>
          <xs:documentation>Mortgage Manager</xs:documentation>
          <xs:appinfo>
            <lx:path>salesChannelTypeList.Mortgage Manager</lx:path>
            <lx:label>Mortgage Manager</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mortgage Originator">
        <xs:annotation>
          <xs:documentation>Mortgage Originator</xs:documentation>
          <xs:appinfo>
            <lx:path>salesChannelTypeList.Mortgage Originator</lx:path>
            <lx:label>Mortgage Originator</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Online Channel">
        <xs:annotation>
          <xs:documentation>Online Channel</xs:documentation>
          <xs:appinfo>
            <lx:path>salesChannelTypeList.Online Channel</lx:path>
            <lx:label>Online Channel</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="securityPriorityList">
    <xs:annotation>
      <xs:documentation>securityPriorityList</xs:documentation>
      <xs:appinfo>
        <lx:path>securityPriorityList</lx:path>
        <lx:label>securityPriorityList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="First Mortgage">
        <xs:annotation>
          <xs:documentation>The mortgage is a first ranking mortgage.</xs:documentation>
          <xs:appinfo>
            <lx:path>securityPriorityList.First Mortgage</lx:path>
            <lx:label>The mortgage is a first ranking mortgage.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Registered Security">
        <xs:annotation>
          <xs:documentation>The security is registered on a property security register</xs:documentation>
          <xs:appinfo>
            <lx:path>securityPriorityList.Registered Security</lx:path>
            <lx:label>The security is registered on a property security register</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Second after DSH">
        <xs:annotation>
          <xs:documentation>The mortgage is a second ranking mortgage, ranking after a first ranking mortgage to Defence Services Housing.</xs:documentation>
          <xs:appinfo>
            <lx:path>securityPriorityList.Second after DSH</lx:path>
            <lx:label>The mortgage is a second ranking mortgage, ranking after a first ranking mortgage to Defence Services Housing.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Second Mortgage">
        <xs:annotation>
          <xs:documentation>The mortgage is a second ranking mortgage.</xs:documentation>
          <xs:appinfo>
            <lx:path>securityPriorityList.Second Mortgage</lx:path>
            <lx:label>The mortgage is a second ranking mortgage.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Third Mortgage">
        <xs:annotation>
          <xs:documentation>The mortgage is a third ranking mortgage.</xs:documentation>
          <xs:appinfo>
            <lx:path>securityPriorityList.Third Mortgage</lx:path>
            <lx:label>The mortgage is a third ranking mortgage.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unregistered Mortgage">
        <xs:annotation>
          <xs:documentation>The mortgage is not registered on the Land Register at the Land Titles Office.</xs:documentation>
          <xs:appinfo>
            <lx:path>securityPriorityList.Unregistered Mortgage</lx:path>
            <lx:label>The mortgage is not registered on the Land Register at the Land Titles Office.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="selfEmployedBasisList">
    <xs:annotation>
      <xs:documentation>selfEmployedBasisList</xs:documentation>
      <xs:appinfo>
        <lx:path>selfEmployedBasisList</lx:path>
        <lx:label>selfEmployedBasisList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Casual">
        <xs:annotation>
          <xs:documentation>The self employment described is under a casual contract of employment.</xs:documentation>
          <xs:appinfo>
            <lx:path>selfEmployedBasisList.Casual</lx:path>
            <lx:label>The self employment described is under a casual contract of employment.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Contract">
        <xs:annotation>
          <xs:documentation>The self employment described is under a contract for services.</xs:documentation>
          <xs:appinfo>
            <lx:path>selfEmployedBasisList.Contract</lx:path>
            <lx:label>The self employment described is under a contract for services.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Full Time">
        <xs:annotation>
          <xs:documentation>The self employment described is under full time employment contract.</xs:documentation>
          <xs:appinfo>
            <lx:path>selfEmployedBasisList.Full Time</lx:path>
            <lx:label>The self employment described is under full time employment contract.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Part Time">
        <xs:annotation>
          <xs:documentation>The self employment described is under part time employment contract.</xs:documentation>
          <xs:appinfo>
            <lx:path>selfEmployedBasisList.Part Time</lx:path>
            <lx:label>The self employment described is under part time employment contract.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Temporary">
        <xs:annotation>
          <xs:documentation>The self employment described is under contract with an employment agency, also known as temping.</xs:documentation>
          <xs:appinfo>
            <lx:path>selfEmployedBasisList.Temporary</lx:path>
            <lx:label>The self employment described is under contract with an employment agency, also known as temping.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="sendDocumentsToList">
    <xs:annotation>
      <xs:documentation>sendDocumentsToList</xs:documentation>
      <xs:appinfo>
        <lx:path>sendDocumentsToList</lx:path>
        <lx:label>sendDocumentsToList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Accountant">
        <xs:annotation>
          <xs:documentation>Accountant</xs:documentation>
          <xs:appinfo>
            <lx:path>sendDocumentsToList.Accountant</lx:path>
            <lx:label>Accountant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="All">
        <xs:annotation>
          <xs:documentation>All</xs:documentation>
          <xs:appinfo>
            <lx:path>sendDocumentsToList.All</lx:path>
            <lx:label>All</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Applicant">
        <xs:annotation>
          <xs:documentation>Applicant</xs:documentation>
          <xs:appinfo>
            <lx:path>sendDocumentsToList.Applicant</lx:path>
            <lx:label>Applicant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Branch">
        <xs:annotation>
          <xs:documentation>Branch</xs:documentation>
          <xs:appinfo>
            <lx:path>sendDocumentsToList.Branch</lx:path>
            <lx:label>Branch</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Writer">
        <xs:annotation>
          <xs:documentation>Loan Writer</xs:documentation>
          <xs:appinfo>
            <lx:path>sendDocumentsToList.Loan Writer</lx:path>
            <lx:label>Loan Writer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>sendDocumentsToList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Solicitor">
        <xs:annotation>
          <xs:documentation>Solicitor</xs:documentation>
          <xs:appinfo>
            <lx:path>sendDocumentsToList.Solicitor</lx:path>
            <lx:label>Solicitor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="sendDocumentsToPersonList">
    <xs:annotation>
      <xs:documentation>sendDocumentsToPersonList</xs:documentation>
      <xs:appinfo>
        <lx:path>sendDocumentsToPersonList</lx:path>
        <lx:label>sendDocumentsToPersonList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Accountant">
        <xs:annotation>
          <xs:documentation>Accountant</xs:documentation>
          <xs:appinfo>
            <lx:path>sendDocumentsToPersonList.Accountant</lx:path>
            <lx:label>Accountant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Branch">
        <xs:annotation>
          <xs:documentation>Branch</xs:documentation>
          <xs:appinfo>
            <lx:path>sendDocumentsToPersonList.Branch</lx:path>
            <lx:label>Branch</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loan Writer">
        <xs:annotation>
          <xs:documentation>Loan Writer</xs:documentation>
          <xs:appinfo>
            <lx:path>sendDocumentsToPersonList.Loan Writer</lx:path>
            <lx:label>Loan Writer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Myself">
        <xs:annotation>
          <xs:documentation>Myself</xs:documentation>
          <xs:appinfo>
            <lx:path>sendDocumentsToPersonList.Myself</lx:path>
            <lx:label>Myself</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>sendDocumentsToPersonList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Solicitor">
        <xs:annotation>
          <xs:documentation>Solicitor</xs:documentation>
          <xs:appinfo>
            <lx:path>sendDocumentsToPersonList.Solicitor</lx:path>
            <lx:label>Solicitor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="serviceabilityAssessmentTypeList">
    <xs:annotation>
      <xs:documentation>serviceabilityAssessmentTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>serviceabilityAssessmentTypeList</lx:path>
        <lx:label>serviceabilityAssessmentTypeList</lx:label>
        <li:transactions>SVC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Serviceability Request">
        <xs:annotation>
          <xs:documentation>Serviceability Request</xs:documentation>
          <xs:appinfo>
            <lx:path>serviceabilityAssessmentTypeList.Serviceability Request</lx:path>
            <lx:label>Serviceability Request</lx:label>
            <li:transactions>SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Serviceability Response">
        <xs:annotation>
          <xs:documentation>Serviceability Response</xs:documentation>
          <xs:appinfo>
            <lx:path>serviceabilityAssessmentTypeList.Serviceability Response</lx:path>
            <lx:label>Serviceability Response</lx:label>
            <li:transactions>SVC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="serviceTypeList">
    <xs:annotation>
      <xs:documentation>servicesList</xs:documentation>
      <xs:appinfo>
        <lx:path>serviceTypeList</lx:path>
        <lx:label>serviceTypeList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Cable TV">
        <xs:annotation>
          <xs:documentation>Cable TV</xs:documentation>
          <xs:appinfo>
            <lx:path>serviceTypeList.Cable TV</lx:path>
            <lx:label>Cable TV</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Electricity">
        <xs:annotation>
          <xs:documentation>Electricity</xs:documentation>
          <xs:appinfo>
            <lx:path>serviceTypeList.Electricity</lx:path>
            <lx:label>Electricity</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gas">
        <xs:annotation>
          <xs:documentation>Gas</xs:documentation>
          <xs:appinfo>
            <lx:path>serviceTypeList.Gas</lx:path>
            <lx:label>Gas</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NBN">
        <xs:annotation>
          <xs:documentation>NBN</xs:documentation>
          <xs:appinfo>
            <lx:path>serviceTypeList.NBN</lx:path>
            <lx:label>NBN</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sewerage">
        <xs:annotation>
          <xs:documentation>Sewerage</xs:documentation>
          <xs:appinfo>
            <lx:path>serviceTypeList.Sewerage</lx:path>
            <lx:label>Sewerage</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Telephone">
        <xs:annotation>
          <xs:documentation>Telephone</xs:documentation>
          <xs:appinfo>
            <lx:path>serviceTypeList.Telephone</lx:path>
            <lx:label>Telephone</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Town Water">
        <xs:annotation>
          <xs:documentation>Town Water</xs:documentation>
          <xs:appinfo>
            <lx:path>serviceTypeList.Town Water</lx:path>
            <lx:label>Town Water</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Waste Treatment System">
        <xs:annotation>
          <xs:documentation>Waste Treatment System</xs:documentation>
          <xs:appinfo>
            <lx:path>serviceTypeList.Waste Treatment System</lx:path>
            <lx:label>Waste Treatment System</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="shareTypeList">
    <xs:annotation>
      <xs:documentation>shareTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>shareTypeList</lx:path>
        <lx:label>shareTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Bearer Securities">
        <xs:annotation>
          <xs:documentation>Bearer Securities</xs:documentation>
          <xs:appinfo>
            <lx:path>shareTypeList.Bearer Securities</lx:path>
            <lx:label>Bearer Securities</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Debentures">
        <xs:annotation>
          <xs:documentation>Debentures</xs:documentation>
          <xs:appinfo>
            <lx:path>shareTypeList.Debentures</lx:path>
            <lx:label>Debentures</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Inscribed Stock Cth">
        <xs:annotation>
          <xs:documentation>Inscribed Stock Cth</xs:documentation>
          <xs:appinfo>
            <lx:path>shareTypeList.Inscribed Stock Cth</lx:path>
            <lx:label>Inscribed Stock Cth</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Inscribed Stock State">
        <xs:annotation>
          <xs:documentation>Inscribed Stock State</xs:documentation>
          <xs:appinfo>
            <lx:path>shareTypeList.Inscribed Stock State</lx:path>
            <lx:label>Inscribed Stock State</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Listed Shares">
        <xs:annotation>
          <xs:documentation>Listed Shares</xs:documentation>
          <xs:appinfo>
            <lx:path>shareTypeList.Listed Shares</lx:path>
            <lx:label>Listed Shares</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Options">
        <xs:annotation>
          <xs:documentation>Options</xs:documentation>
          <xs:appinfo>
            <lx:path>shareTypeList.Options</lx:path>
            <lx:label>Options</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unlisted Shares">
        <xs:annotation>
          <xs:documentation>Unlisted Shares</xs:documentation>
          <xs:appinfo>
            <lx:path>shareTypeList.Unlisted Shares</lx:path>
            <lx:label>Unlisted Shares</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="significantChangeList">
    <xs:annotation>
      <xs:documentation>significantChangeList</xs:documentation>
      <xs:appinfo>
        <lx:path>significantChangeList</lx:path>
        <lx:label>significantChangeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Anticipated Large Expenditure">
        <xs:annotation>
          <xs:documentation>Anticipated Large Expenditure</xs:documentation>
          <xs:appinfo>
            <lx:path>significantChangeList.Anticipated Large Expenditure</lx:path>
            <lx:label>Anticipated Large Expenditure</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="End of Contract or Loss of Employment">
        <xs:annotation>
          <xs:documentation>End of contract or loss of employment.</xs:documentation>
          <xs:appinfo>
            <lx:path>significantChangeList.End of Contract or Loss of Employment</lx:path>
            <lx:label>End of Contract or Loss of Employment</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Extended Unpaid Leave">
        <xs:annotation>
          <xs:documentation>Extended unpaid leave (e.g. parental leave).</xs:documentation>
          <xs:appinfo>
            <lx:path>significantChangeList.Extended Unpaid Leave</lx:path>
            <lx:label>Extended Unpaid Leave</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Increased Debt Repayments for an Existing Loan">
        <xs:annotation>
          <xs:documentation>Increased debt repayments for an existing loan.</xs:documentation>
          <xs:appinfo>
            <lx:path>significantChangeList.Increased Debt Repayments for an Existing Loan</lx:path>
            <lx:label>Increased Debt Repayments for an Existing Loan</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Leaving Employment">
        <xs:annotation>
          <xs:documentation>End of contract or loss of employment.</xs:documentation>
          <xs:appinfo>
            <lx:path>significantChangeList.Leaving Employment</lx:path>
            <lx:label>Leaving Employment</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Medical Treatment or Illness">
        <xs:annotation>
          <xs:documentation>Medical treatment or illness.</xs:documentation>
          <xs:appinfo>
            <lx:path>significantChangeList.Medical Treatment or Illness</lx:path>
            <lx:label>Medical Treatment or Illness</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>significantChangeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Permanent Decrease in Income">
        <xs:annotation>
          <xs:documentation>Permanent Decrease in Income</xs:documentation>
          <xs:appinfo>
            <lx:path>significantChangeList.Permanent Decrease in Income</lx:path>
            <lx:label>Permanent Decrease in Income</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Permanent Increase in Expense">
        <xs:annotation>
          <xs:documentation>Permanent Increase in Expense</xs:documentation>
          <xs:appinfo>
            <lx:path>significantChangeList.Permanent Increase in Expense</lx:path>
            <lx:label>Permanent Increase in Expense</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Retirement">
        <xs:annotation>
          <xs:documentation>Retirement</xs:documentation>
          <xs:appinfo>
            <lx:path>significantChangeList.Retirement</lx:path>
            <lx:label>Retirement</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Temporary Decrease in Income">
        <xs:annotation>
          <xs:documentation>Temporary Decrease in Income</xs:documentation>
          <xs:appinfo>
            <lx:path>significantChangeList.Temporary Decrease in Income</lx:path>
            <lx:label>Temporary Decrease in Income</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Temporary Increase in Expense">
        <xs:annotation>
          <xs:documentation>Temporary Increase in Expense</xs:documentation>
          <xs:appinfo>
            <lx:path>significantChangeList.Temporary Increase in Expense</lx:path>
            <lx:label>Temporary Increase in Expense</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="sourceOfFundsTypeList">
    <xs:annotation>
      <xs:documentation>sourceOfFundsTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>sourceOfFundsTypeList</lx:path>
        <lx:label>sourceOfFundsTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Business Ownership">
        <xs:annotation>
          <xs:documentation>Business Ownership</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Business Ownership</lx:path>
            <lx:label>Business Ownership</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Business Sourced Income">
        <xs:annotation>
          <xs:documentation>Business Sourced Income</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Business Sourced Income</lx:path>
            <lx:label>Business Sourced Income</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Company Sale or Earnings">
        <xs:annotation>
          <xs:documentation>Company Sale or Earnings</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Company Sale or Earnings</lx:path>
            <lx:label>Company Sale or Earnings</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Directorship">
        <xs:annotation>
          <xs:documentation>Directorship</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Directorship</lx:path>
            <lx:label>Directorship</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Divorce Settlement">
        <xs:annotation>
          <xs:documentation>Divorce Settlement</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Divorce Settlement</lx:path>
            <lx:label>Divorce Settlement</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Employment Income">
        <xs:annotation>
          <xs:documentation>Employment Income</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Employment Income</lx:path>
            <lx:label>Employment Income</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gift">
        <xs:annotation>
          <xs:documentation>Gift</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Gift</lx:path>
            <lx:label>Gift</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Government Benefit">
        <xs:annotation>
          <xs:documentation>Government Benefit</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Government Benefit</lx:path>
            <lx:label>Government Benefit</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Grant">
        <xs:annotation>
          <xs:documentation>Grant</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Grant</lx:path>
            <lx:label>Grant</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Inheritance">
        <xs:annotation>
          <xs:documentation>Inheritance</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Inheritance</lx:path>
            <lx:label>Inheritance</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Insurance Settlement Proceeds">
        <xs:annotation>
          <xs:documentation>Insurance Settlement Proceeds</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Insurance Settlement Proceeds</lx:path>
            <lx:label>Insurance Settlement Proceeds</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Investment Income">
        <xs:annotation>
          <xs:documentation>Investment Income</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Investment Income</lx:path>
            <lx:label>Investment Income</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lottery">
        <xs:annotation>
          <xs:documentation>Lottery</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Lottery</lx:path>
            <lx:label>Lottery</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Parental Support">
        <xs:annotation>
          <xs:documentation>Parental Support</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Parental Support</lx:path>
            <lx:label>Parental Support</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Payroll">
        <xs:annotation>
          <xs:documentation>Payroll</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Payroll</lx:path>
            <lx:label>Payroll</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pension">
        <xs:annotation>
          <xs:documentation>Pension</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Pension</lx:path>
            <lx:label>Pension</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Personal Savings">
        <xs:annotation>
          <xs:documentation>Personal Savings</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Personal Savings</lx:path>
            <lx:label>Personal Savings</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rental Income">
        <xs:annotation>
          <xs:documentation>Rental Income</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Rental Income</lx:path>
            <lx:label>Rental Income</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Royalties">
        <xs:annotation>
          <xs:documentation>Royalties</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Royalties</lx:path>
            <lx:label>Royalties</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sale of Asset">
        <xs:annotation>
          <xs:documentation>Sale of Non Real Estate Asset</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Sale of Asset</lx:path>
            <lx:label>Sale of Asset</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sale of Real Estate Asset">
        <xs:annotation>
          <xs:documentation>Sale of Real Estate Asset</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Sale of Real Estate Asset</lx:path>
            <lx:label>Sale of Real Estate Asset</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Scholarship">
        <xs:annotation>
          <xs:documentation>Scholarship</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Scholarship</lx:path>
            <lx:label>Scholarship</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Significant Other">
        <xs:annotation>
          <xs:documentation>Significant Other</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Significant Other</lx:path>
            <lx:label>Significant Other</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Top Executive">
        <xs:annotation>
          <xs:documentation>Top Executive</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Top Executive</lx:path>
            <lx:label>Top Executive</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trust Fund">
        <xs:annotation>
          <xs:documentation>Trust Fund</xs:documentation>
          <xs:appinfo>
            <lx:path>sourceOfFundsTypeList.Trust Fund</lx:path>
            <lx:label>Trust Fund</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="statementCycleList">
    <xs:annotation>
      <xs:documentation>statementCycleList</xs:documentation>
      <xs:appinfo>
        <lx:path>statementCycleList</lx:path>
        <lx:label>statementCycleList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Fortnightly">
        <xs:annotation>
          <xs:documentation>Fortnightly</xs:documentation>
          <xs:appinfo>
            <lx:path>statementCycleList.Fortnightly</lx:path>
            <lx:label>Fortnightly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Half Yearly">
        <xs:annotation>
          <xs:documentation>Half Yearly</xs:documentation>
          <xs:appinfo>
            <lx:path>statementCycleList.Half Yearly</lx:path>
            <lx:label>Half Yearly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Monthly">
        <xs:annotation>
          <xs:documentation>Monthly</xs:documentation>
          <xs:appinfo>
            <lx:path>statementCycleList.Monthly</lx:path>
            <lx:label>Monthly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Quarterly">
        <xs:annotation>
          <xs:documentation>Quarterly</xs:documentation>
          <xs:appinfo>
            <lx:path>statementCycleList.Quarterly</lx:path>
            <lx:label>Quarterly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Weekly">
        <xs:annotation>
          <xs:documentation>Weekly</xs:documentation>
          <xs:appinfo>
            <lx:path>statementCycleList.Weekly</lx:path>
            <lx:label>Weekly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Yearly">
        <xs:annotation>
          <xs:documentation>Yearly</xs:documentation>
          <xs:appinfo>
            <lx:path>statementCycleList.Yearly</lx:path>
            <lx:label>Yearly</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="statusNameAccountCreationInstructionsList">
    <xs:annotation>
      <xs:documentation>statusNameAccountCreationInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>statusNameAccountCreationInstructionsList</lx:path>
        <lx:label>statusNameAccountCreationInstructionsList</lx:label>
        <li:transactions>ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Current">
        <xs:annotation>
          <xs:documentation>The loan is current.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameAccountCreationInstructionsList.Current</lx:path>
            <lx:label>Current</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Discharged">
        <xs:annotation>
          <xs:documentation>The loan has been discharged.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameAccountCreationInstructionsList.Discharged</lx:path>
            <lx:label>Discharged</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pending Settlement">
        <xs:annotation>
          <xs:documentation>The application is pending settlement.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameAccountCreationInstructionsList.Pending Settlement</lx:path>
            <lx:label>Pending Settlement</lx:label>
            <li:transactions>ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="statusNameApplicationInstructionsList">
    <xs:annotation>
      <xs:documentation>statusNameApplicationInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>statusNameApplicationInstructionsList</lx:path>
        <lx:label>statusNameApplicationInstructionsList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Application Received">
        <xs:annotation>
          <xs:documentation>The intended recipient has received the application but has not yet registered it in the origination system.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameApplicationInstructionsList.Application Received</lx:path>
            <lx:label>Application Received</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Application Registered">
        <xs:annotation>
          <xs:documentation>Lender system ID allocated (Pre Settlement Application Identifier) and, if electronically lodged, application is ready for credit assessment ie. validates against Lender's schema and other business rules.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameApplicationInstructionsList.Application Registered</lx:path>
            <lx:label>Lender system ID allocated (Pre Settlement Application Identifier) and, if electronically lodged, application is ready for credit assessment ie. validates against Lender's schema and other business rules.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Application Sent">
        <xs:annotation>
          <xs:documentation>Application successfully transmitted by Broker system.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameApplicationInstructionsList.Application Sent</lx:path>
            <lx:label>Application successfully transmitted by Broker system.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Application Settled">
        <xs:annotation>
          <xs:documentation>Application has gone to settlement; may or may not have been fully drawn down; accounts may have had a new identifier allocated to them by the Lender (known as Post Settlement Application Number or Facility identifier).</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameApplicationInstructionsList.Application Settled</lx:path>
            <lx:label>Application has gone to settlement; may or may not have been fully drawn down; accounts may have had a new identifier allocated to them by the Lender (known as Post Settlement Application Number or Facility identifier).</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cancelled">
        <xs:annotation>
          <xs:documentation>Application has been stopped at Lender's initiative. Cannot be resurrected.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameApplicationInstructionsList.Cancelled</lx:path>
            <lx:label>Application has been stopped at Lender's initiative. Cannot be resurrected.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Conditionally Approved">
        <xs:annotation>
          <xs:documentation>Application has been approved subject to conditions.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameApplicationInstructionsList.Conditionally Approved</lx:path>
            <lx:label>Application has been approved subject to conditions.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Declined">
        <xs:annotation>
          <xs:documentation>Application has been Declined.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameApplicationInstructionsList.Declined</lx:path>
            <lx:label>Application has been Declined.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Indicatively Approved">
        <xs:annotation>
          <xs:documentation>The application has been indicatively approved without having done a credit check on the applicant.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameApplicationInstructionsList.Indicatively Approved</lx:path>
            <lx:label>Indicative Approval</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="PreApproved">
        <xs:annotation>
          <xs:documentation>Application conditionally approved subject to final identification of security property.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameApplicationInstructionsList.PreApproved</lx:path>
            <lx:label>Application conditionally approved subject to final identification of security property.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ready for Settlement">
        <xs:annotation>
          <xs:documentation>All documents have been received, confirmed as being signed and in order, and the loan is ready for settlement.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameApplicationInstructionsList.Ready for Settlement</lx:path>
            <lx:label>Ready for Settlement</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Referred to Assessor">
        <xs:annotation>
          <xs:documentation>Application has been referred for manual assessment.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameApplicationInstructionsList.Referred to Assessor</lx:path>
            <lx:label>Application has been referred for manual assessment.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Reverted to Capture">
        <xs:annotation>
          <xs:documentation>Application has been reverted to the data capture stage to update or correct data.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameApplicationInstructionsList.Reverted to Capture</lx:path>
            <lx:label>Reverted to Capture</lx:label>
            <li:transactions>CAL,CDA,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unconditionally Approved">
        <xs:annotation>
          <xs:documentation>All conditions to conditional approval have been met and Application has received formal/final approval.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameApplicationInstructionsList.Unconditionally Approved</lx:path>
            <lx:label>All conditions to conditional approval have been met and Application has received formal/final approval.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Withdrawn">
        <xs:annotation>
          <xs:documentation>Application has been stopped at client's initiative.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameApplicationInstructionsList.Withdrawn</lx:path>
            <lx:label>Application has been stopped at client's initiative.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="statusNameCreditDecisionAssessmentList">
    <xs:annotation>
      <xs:documentation>statusNameCreditDecisionAssessmentList</xs:documentation>
      <xs:appinfo>
        <lx:path>statusNameCreditDecisionAssessmentList</lx:path>
        <lx:label>Credit Decision Assessment Status Name List</lx:label>
        <li:transactions>CDA</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Complete">
        <xs:annotation>
          <xs:documentation>The credit decision assessment is complete.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameCreditDecisionAssessmentList.Complete</lx:path>
            <lx:label>Complete</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="statusNameLendersMortgageInsuranceInstructionsList">
    <xs:annotation>
      <xs:documentation>statusNameLendersMortgageInsuranceInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>statusNameLendersMortgageInsuranceInstructionsList</lx:path>
        <lx:label>statusNameLendersMortgageInsuranceInstructionsList</lx:label>
        <li:transactions>LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Accepted Under eDUA">
        <xs:annotation>
          <xs:documentation>Accepted Under eDUA</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameLendersMortgageInsuranceInstructionsList.Accepted Under eDUA</lx:path>
            <lx:label>Accepted Under eDUA</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Application Received">
        <xs:annotation>
          <xs:documentation>Application Received</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameLendersMortgageInsuranceInstructionsList.Application Received</lx:path>
            <lx:label>Application Received</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Approved">
        <xs:annotation>
          <xs:documentation>Approved</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameLendersMortgageInsuranceInstructionsList.Approved</lx:path>
            <lx:label>Approved</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Conditionally Approved">
        <xs:annotation>
          <xs:documentation>Conditionally Approved</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameLendersMortgageInsuranceInstructionsList.Conditionally Approved</lx:path>
            <lx:label>Conditionally Approved</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Declined">
        <xs:annotation>
          <xs:documentation>Declined</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameLendersMortgageInsuranceInstructionsList.Declined</lx:path>
            <lx:label>Declined</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Referred to Underwriter">
        <xs:annotation>
          <xs:documentation>Referred to Underwriter</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameLendersMortgageInsuranceInstructionsList.Referred to Underwriter</lx:path>
            <lx:label>Referred to Underwriter</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="statusNameValuationInstructionsList">
    <xs:annotation>
      <xs:documentation>statusNameValuationInstructionsList</xs:documentation>
      <xs:appinfo>
        <lx:path>statusNameValuationInstructionsList</lx:path>
        <lx:label>statusNameValuationInstructionsList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Accepted">
        <xs:annotation>
          <xs:documentation>The valuation job has been accepted.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameValuationInstructionsList.Accepted</lx:path>
            <lx:label>Accepted</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Amendment Required">
        <xs:annotation>
          <xs:documentation>An amendment is required to the valuation report.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameValuationInstructionsList.Amendment Required</lx:path>
            <lx:label>Amendment Required</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Assigned">
        <xs:annotation>
          <xs:documentation>The valuation job has been assigned to a valuer.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameValuationInstructionsList.Assigned</lx:path>
            <lx:label>Assigned</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Awaiting Authorisation">
        <xs:annotation>
          <xs:documentation>The valuation is awaiting authorisation.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameValuationInstructionsList.Awaiting Authorisation</lx:path>
            <lx:label>Awaiting Authorisation</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cancelled">
        <xs:annotation>
          <xs:documentation>The valuation job has been cancelled.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameValuationInstructionsList.Cancelled</lx:path>
            <lx:label>Cancelled</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Declined">
        <xs:annotation>
          <xs:documentation>The valuation job has been declined by the valuer.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameValuationInstructionsList.Declined</lx:path>
            <lx:label>Declined</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Delayed">
        <xs:annotation>
          <xs:documentation>The valuation has been delayed.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameValuationInstructionsList.Delayed</lx:path>
            <lx:label>Delayed</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Finalised">
        <xs:annotation>
          <xs:documentation>The valuation has been finalised.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameValuationInstructionsList.Finalised</lx:path>
            <lx:label>Finalised</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="In Progress">
        <xs:annotation>
          <xs:documentation>The delay has ended or expired.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameValuationInstructionsList.In Progress</lx:path>
            <lx:label>In Progress</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Initial">
        <xs:annotation>
          <xs:documentation>The valuation request is created for a standard request.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameValuationInstructionsList.Initial</lx:path>
            <lx:label>Initial</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Inquiring">
        <xs:annotation>
          <xs:documentation>The valuation request is created for a quote request.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameValuationInstructionsList.Inquiring</lx:path>
            <lx:label>Inquiring</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Inspected">
        <xs:annotation>
          <xs:documentation>The property has been inspected.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameValuationInstructionsList.Inspected</lx:path>
            <lx:label>Inspected</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Instructed">
        <xs:annotation>
          <xs:documentation>The valuation has been instructed.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameValuationInstructionsList.Instructed</lx:path>
            <lx:label>Instructed</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valuation Completed">
        <xs:annotation>
          <xs:documentation>The valuation has been completed.</xs:documentation>
          <xs:appinfo>
            <lx:path>statusNameValuationInstructionsList.Valuation Completed</lx:path>
            <lx:label>Valuation Completed</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="streetSuffixList">
    <xs:annotation>
      <xs:documentation>A list of valid Street Suffixes (accommodating the data item referred to as 'Street Suffix' in the Australia Post Data Guide and 'Street Direction' in New Zealand Post Address Standards).</xs:documentation>
      <xs:appinfo>
        <lx:path>streetSuffixList</lx:path>
        <lx:label>streetSuffixList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Central">
        <xs:annotation>
          <xs:documentation>Central</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.Central</lx:path>
            <lx:label>Central</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Deviation">
        <xs:annotation>
          <xs:documentation>Deviation</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.Deviation</lx:path>
            <lx:label>Deviation</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="East">
        <xs:annotation>
          <xs:documentation>East</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.East</lx:path>
            <lx:label>East</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Extension">
        <xs:annotation>
          <xs:documentation>Extension</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.Extension</lx:path>
            <lx:label>Extension</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Inner">
        <xs:annotation>
          <xs:documentation>Inner</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.Inner</lx:path>
            <lx:label>Inner</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lower">
        <xs:annotation>
          <xs:documentation>Lower</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.Lower</lx:path>
            <lx:label>Lower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mall">
        <xs:annotation>
          <xs:documentation>Mall</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.Mall</lx:path>
            <lx:label>Mall</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="North">
        <xs:annotation>
          <xs:documentation>North</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.North</lx:path>
            <lx:label>North</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="North East">
        <xs:annotation>
          <xs:documentation>North East</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.North East</lx:path>
            <lx:label>North East</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="North West">
        <xs:annotation>
          <xs:documentation>North West</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.North West</lx:path>
            <lx:label>North West</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Off">
        <xs:annotation>
          <xs:documentation>Off</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.Off</lx:path>
            <lx:label>Off</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="On">
        <xs:annotation>
          <xs:documentation>On</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.On</lx:path>
            <lx:label>On</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Outer">
        <xs:annotation>
          <xs:documentation>Outer</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.Outer</lx:path>
            <lx:label>Outer</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Overpass">
        <xs:annotation>
          <xs:documentation>Overpass</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.Overpass</lx:path>
            <lx:label>Overpass</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="South">
        <xs:annotation>
          <xs:documentation>South</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.South</lx:path>
            <lx:label>South</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="South East">
        <xs:annotation>
          <xs:documentation>South East</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.South East</lx:path>
            <lx:label>South East</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="South West">
        <xs:annotation>
          <xs:documentation>South West</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.South West</lx:path>
            <lx:label>South West</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Upper">
        <xs:annotation>
          <xs:documentation>Upper</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.Upper</lx:path>
            <lx:label>Upper</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="West">
        <xs:annotation>
          <xs:documentation>West</xs:documentation>
          <xs:appinfo>
            <lx:path>streetSuffixList.West</lx:path>
            <lx:label>West</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="streetTypeList">
    <xs:annotation>
      <xs:documentation>streetTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>streetTypeList</lx:path>
        <lx:label>streetTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Access">
        <xs:annotation>
          <xs:documentation>Access</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Access</lx:path>
            <lx:label>Access</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Accessway">
        <xs:annotation>
          <xs:documentation>Accessway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Accessway</lx:path>
            <lx:label>Accessway</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Alley">
        <xs:annotation>
          <xs:documentation>Alley</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Alley</lx:path>
            <lx:label>Alley</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Alleyway">
        <xs:annotation>
          <xs:documentation>Alleyway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Alleyway</lx:path>
            <lx:label>Alleyway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Amble">
        <xs:annotation>
          <xs:documentation>Amble</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Amble</lx:path>
            <lx:label>Amble</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Anchorage">
        <xs:annotation>
          <xs:documentation>Anchorage</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Anchorage</lx:path>
            <lx:label>Anchorage</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Approach">
        <xs:annotation>
          <xs:documentation>Approach</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Approach</lx:path>
            <lx:label>Approach</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Arcade">
        <xs:annotation>
          <xs:documentation>Arcade</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Arcade</lx:path>
            <lx:label>Arcade</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Arch">
        <xs:annotation>
          <xs:documentation>Arch</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Arch</lx:path>
            <lx:label>Arch</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Arterial">
        <xs:annotation>
          <xs:documentation>Arterial</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Arterial</lx:path>
            <lx:label>Arterial</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Artery">
        <xs:annotation>
          <xs:documentation>Artery</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Artery</lx:path>
            <lx:label>Artery</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Avenue">
        <xs:annotation>
          <xs:documentation>Avenue</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Avenue</lx:path>
            <lx:label>Avenue</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Banan">
        <xs:annotation>
          <xs:documentation>Banan</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Banan</lx:path>
            <lx:label>Banan</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bank">
        <xs:annotation>
          <xs:documentation>Bank</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Bank</lx:path>
            <lx:label>Bank</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Basin">
        <xs:annotation>
          <xs:documentation>Basin</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Basin</lx:path>
            <lx:label>Basin</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bay">
        <xs:annotation>
          <xs:documentation>Bay</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Bay</lx:path>
            <lx:label>Bay</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Beach">
        <xs:annotation>
          <xs:documentation>Beach</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Beach</lx:path>
            <lx:label>Beach</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Belt">
        <xs:annotation>
          <xs:documentation>Belt</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Belt</lx:path>
            <lx:label>Belt</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bend">
        <xs:annotation>
          <xs:documentation>Bend</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Bend</lx:path>
            <lx:label>Bend</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Block">
        <xs:annotation>
          <xs:documentation>Block</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Block</lx:path>
            <lx:label>Block</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bluff">
        <xs:annotation>
          <xs:documentation>Bluff</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Bluff</lx:path>
            <lx:label>Bluff</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Boardwalk">
        <xs:annotation>
          <xs:documentation>Boardwalk</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Boardwalk</lx:path>
            <lx:label>Boardwalk</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Boulevard">
        <xs:annotation>
          <xs:documentation>Boulevard</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Boulevard</lx:path>
            <lx:label>Boulevard</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Boulevarde">
        <xs:annotation>
          <xs:documentation>Boulevarde</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Boulevarde</lx:path>
            <lx:label>Boulevarde</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bowl">
        <xs:annotation>
          <xs:documentation>Bowl</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Bowl</lx:path>
            <lx:label>Bowl</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Brace">
        <xs:annotation>
          <xs:documentation>Brace</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Brace</lx:path>
            <lx:label>Brace</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Brae">
        <xs:annotation>
          <xs:documentation>Brae</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Brae</lx:path>
            <lx:label>Brae</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Break">
        <xs:annotation>
          <xs:documentation>Break</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Break</lx:path>
            <lx:label>Break</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Briars">
        <xs:annotation>
          <xs:documentation>Briars</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Briars</lx:path>
            <lx:label>Briars</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bridge">
        <xs:annotation>
          <xs:documentation>Bridge</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Bridge</lx:path>
            <lx:label>Bridge</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Broadway">
        <xs:annotation>
          <xs:documentation>Broadway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Broadway</lx:path>
            <lx:label>Broadway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Brow">
        <xs:annotation>
          <xs:documentation>Brow</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Brow</lx:path>
            <lx:label>Brow</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bypass">
        <xs:annotation>
          <xs:documentation>Bypass</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Bypass</lx:path>
            <lx:label>Bypass</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Byway">
        <xs:annotation>
          <xs:documentation>Byway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Byway</lx:path>
            <lx:label>Byway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Canal">
        <xs:annotation>
          <xs:documentation>Canal</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Canal</lx:path>
            <lx:label>Canal</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Causeway">
        <xs:annotation>
          <xs:documentation>Causeway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Causeway</lx:path>
            <lx:label>Causeway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Centre">
        <xs:annotation>
          <xs:documentation>Centre</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Centre</lx:path>
            <lx:label>Centre</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Centreway">
        <xs:annotation>
          <xs:documentation>Centreway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Centreway</lx:path>
            <lx:label>Centreway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Chase">
        <xs:annotation>
          <xs:documentation>Chase</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Chase</lx:path>
            <lx:label>Chase</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Circle">
        <xs:annotation>
          <xs:documentation>Circle</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Circle</lx:path>
            <lx:label>Circle</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Circlet">
        <xs:annotation>
          <xs:documentation>Circlet</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Circlet</lx:path>
            <lx:label>Circlet</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Circuit">
        <xs:annotation>
          <xs:documentation>Circuit</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Circuit</lx:path>
            <lx:label>Circuit</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Circus">
        <xs:annotation>
          <xs:documentation>Circus</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Circus</lx:path>
            <lx:label>Circus</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Claim">
        <xs:annotation>
          <xs:documentation>Claim</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Claim</lx:path>
            <lx:label>Claim</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Close">
        <xs:annotation>
          <xs:documentation>Close</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Close</lx:path>
            <lx:label>Close</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cluster">
        <xs:annotation>
          <xs:documentation>Cluster</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Cluster</lx:path>
            <lx:label>Cluster</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Colonnade">
        <xs:annotation>
          <xs:documentation>Colonnade</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Colonnade</lx:path>
            <lx:label>Colonnade</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Common">
        <xs:annotation>
          <xs:documentation>Common</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Common</lx:path>
            <lx:label>Common</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Concourse">
        <xs:annotation>
          <xs:documentation>Concourse</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Concourse</lx:path>
            <lx:label>Concourse</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Connection">
        <xs:annotation>
          <xs:documentation>Connection</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Connection</lx:path>
            <lx:label>Connection</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Copse">
        <xs:annotation>
          <xs:documentation>Copse</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Copse</lx:path>
            <lx:label>Copse</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Corner">
        <xs:annotation>
          <xs:documentation>Corner</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Corner</lx:path>
            <lx:label>Corner</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Corso">
        <xs:annotation>
          <xs:documentation>Corso</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Corso</lx:path>
            <lx:label>Corso</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Course">
        <xs:annotation>
          <xs:documentation>Course</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Course</lx:path>
            <lx:label>Course</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Court">
        <xs:annotation>
          <xs:documentation>Court</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Court</lx:path>
            <lx:label>Court</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Courts">
        <xs:annotation>
          <xs:documentation>Courts</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Courts</lx:path>
            <lx:label>Courts</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Courtyard">
        <xs:annotation>
          <xs:documentation>Courtyard</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Courtyard</lx:path>
            <lx:label>Courtyard</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cove">
        <xs:annotation>
          <xs:documentation>Cove</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Cove</lx:path>
            <lx:label>Cove</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Creek">
        <xs:annotation>
          <xs:documentation>Creek</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Creek</lx:path>
            <lx:label>Creek</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Crescent">
        <xs:annotation>
          <xs:documentation>Crescent</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Crescent</lx:path>
            <lx:label>Crescent</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Crest">
        <xs:annotation>
          <xs:documentation>Crest</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Crest</lx:path>
            <lx:label>Crest</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cross">
        <xs:annotation>
          <xs:documentation>Cross</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Cross</lx:path>
            <lx:label>Cross</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Crossing">
        <xs:annotation>
          <xs:documentation>Crossing</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Crossing</lx:path>
            <lx:label>Crossing</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Crossroad">
        <xs:annotation>
          <xs:documentation>Crossroad</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Crossroad</lx:path>
            <lx:label>Crossroad</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Crossway">
        <xs:annotation>
          <xs:documentation>Crossway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Crossway</lx:path>
            <lx:label>Crossway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cruiseway">
        <xs:annotation>
          <xs:documentation>Cruiseway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Cruiseway</lx:path>
            <lx:label>Cruiseway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cul">
        <xs:annotation>
          <xs:documentation>Cul</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Cul</lx:path>
            <lx:label>Cul</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cul-De-Sac">
        <xs:annotation>
          <xs:documentation>Cul-De-Sac</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Cul-De-Sac</lx:path>
            <lx:label>Cul-De-Sac</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cutting">
        <xs:annotation>
          <xs:documentation>Cutting</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Cutting</lx:path>
            <lx:label>Cutting</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dale">
        <xs:annotation>
          <xs:documentation>Dale</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Dale</lx:path>
            <lx:label>Dale</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dell">
        <xs:annotation>
          <xs:documentation>Dell</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Dell</lx:path>
            <lx:label>Dell</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dene">
        <xs:annotation>
          <xs:documentation>Dene</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Dene</lx:path>
            <lx:label>Dene</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Deviation">
        <xs:annotation>
          <xs:documentation>Deviation</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Deviation</lx:path>
            <lx:label>Deviation</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dip">
        <xs:annotation>
          <xs:documentation>Dip</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Dip</lx:path>
            <lx:label>Dip</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Distributor">
        <xs:annotation>
          <xs:documentation>Distributor</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Distributor</lx:path>
            <lx:label>Distributor</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Divide">
        <xs:annotation>
          <xs:documentation>Divide</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Divide</lx:path>
            <lx:label>Divide</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dock">
        <xs:annotation>
          <xs:documentation>Dock</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Dock</lx:path>
            <lx:label>Dock</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Domain">
        <xs:annotation>
          <xs:documentation>Domain</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Domain</lx:path>
            <lx:label>Domain</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Downs">
        <xs:annotation>
          <xs:documentation>Downs</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Downs</lx:path>
            <lx:label>Downs</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Drive">
        <xs:annotation>
          <xs:documentation>Drive</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Drive</lx:path>
            <lx:label>Drive</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Driveway">
        <xs:annotation>
          <xs:documentation>Driveway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Driveway</lx:path>
            <lx:label>Driveway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dune">
        <xs:annotation>
          <xs:documentation>Dune</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Dune</lx:path>
            <lx:label>Dune</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Edge">
        <xs:annotation>
          <xs:documentation>Edge</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Edge</lx:path>
            <lx:label>Edge</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Elbow">
        <xs:annotation>
          <xs:documentation>Elbow</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Elbow</lx:path>
            <lx:label>Elbow</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Elm">
        <xs:annotation>
          <xs:documentation>Elm</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Elm</lx:path>
            <lx:label>Elm</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="End">
        <xs:annotation>
          <xs:documentation>End</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.End</lx:path>
            <lx:label>End</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Entrance">
        <xs:annotation>
          <xs:documentation>Entrance</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Entrance</lx:path>
            <lx:label>Entrance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Esplanade">
        <xs:annotation>
          <xs:documentation>Esplanade</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Esplanade</lx:path>
            <lx:label>Esplanade</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Estate">
        <xs:annotation>
          <xs:documentation>Estate</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Estate</lx:path>
            <lx:label>Estate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Estuary">
        <xs:annotation>
          <xs:documentation>Estuary</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Estuary</lx:path>
            <lx:label>Estuary</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Expressway">
        <xs:annotation>
          <xs:documentation>Expressway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Expressway</lx:path>
            <lx:label>Expressway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Extension">
        <xs:annotation>
          <xs:documentation>Extension</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Extension</lx:path>
            <lx:label>Extension</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fairway">
        <xs:annotation>
          <xs:documentation>Fairway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Fairway</lx:path>
            <lx:label>Fairway</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fall">
        <xs:annotation>
          <xs:documentation>Fall</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Fall</lx:path>
            <lx:label>Fall</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fare">
        <xs:annotation>
          <xs:documentation>Fare</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Fare</lx:path>
            <lx:label>Fare</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Farms">
        <xs:annotation>
          <xs:documentation>Farms</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Farms</lx:path>
            <lx:label>Farms</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fen">
        <xs:annotation>
          <xs:documentation>Fen</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Fen</lx:path>
            <lx:label>Fen</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fern">
        <xs:annotation>
          <xs:documentation>Fern</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Fern</lx:path>
            <lx:label>Fern</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fire Track">
        <xs:annotation>
          <xs:documentation>Fire Track</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Fire Track</lx:path>
            <lx:label>Fire Track</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Firebreak">
        <xs:annotation>
          <xs:documentation>Firebreak</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Firebreak</lx:path>
            <lx:label>Firebreak</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fireline">
        <xs:annotation>
          <xs:documentation>Fireline</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Fireline</lx:path>
            <lx:label>Fireline</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Firetrack">
        <xs:annotation>
          <xs:documentation>Firetrack</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Firetrack</lx:path>
            <lx:label>Firetrack</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Firetrail">
        <xs:annotation>
          <xs:documentation>Firetrail</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Firetrail</lx:path>
            <lx:label>Firetrail</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Flat">
        <xs:annotation>
          <xs:documentation>Flat</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Flat</lx:path>
            <lx:label>Flat</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Flats">
        <xs:annotation>
          <xs:documentation>Flats</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Flats</lx:path>
            <lx:label>Flats</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Follow">
        <xs:annotation>
          <xs:documentation>Follow</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Follow</lx:path>
            <lx:label>Follow</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Footway">
        <xs:annotation>
          <xs:documentation>Footway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Footway</lx:path>
            <lx:label>Footway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ford">
        <xs:annotation>
          <xs:documentation>Ford</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Ford</lx:path>
            <lx:label>Ford</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Foreshore">
        <xs:annotation>
          <xs:documentation>Foreshore</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Foreshore</lx:path>
            <lx:label>Foreshore</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Formation">
        <xs:annotation>
          <xs:documentation>Formation</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Formation</lx:path>
            <lx:label>Formation</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Freeway">
        <xs:annotation>
          <xs:documentation>Freeway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Freeway</lx:path>
            <lx:label>Freeway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Front">
        <xs:annotation>
          <xs:documentation>Front</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Front</lx:path>
            <lx:label>Front</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Frontage">
        <xs:annotation>
          <xs:documentation>Frontage</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Frontage</lx:path>
            <lx:label>Frontage</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gap">
        <xs:annotation>
          <xs:documentation>Gap</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Gap</lx:path>
            <lx:label>Gap</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Garden">
        <xs:annotation>
          <xs:documentation>Garden</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Garden</lx:path>
            <lx:label>Garden</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gardens">
        <xs:annotation>
          <xs:documentation>Gardens</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Gardens</lx:path>
            <lx:label>Gardens</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gate">
        <xs:annotation>
          <xs:documentation>Gate</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Gate</lx:path>
            <lx:label>Gate</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gates">
        <xs:annotation>
          <xs:documentation>Gates</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Gates</lx:path>
            <lx:label>Gates</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gateway">
        <xs:annotation>
          <xs:documentation>Gateway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Gateway</lx:path>
            <lx:label>Gateway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Glade">
        <xs:annotation>
          <xs:documentation>Glade</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Glade</lx:path>
            <lx:label>Glade</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Glen">
        <xs:annotation>
          <xs:documentation>Glen</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Glen</lx:path>
            <lx:label>Glen</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Grange">
        <xs:annotation>
          <xs:documentation>Grange</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Grange</lx:path>
            <lx:label>Grange</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Green">
        <xs:annotation>
          <xs:documentation>Green</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Green</lx:path>
            <lx:label>Green</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ground">
        <xs:annotation>
          <xs:documentation>Ground</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Ground</lx:path>
            <lx:label>Ground</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Grove">
        <xs:annotation>
          <xs:documentation>Grove</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Grove</lx:path>
            <lx:label>Grove</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Gully">
        <xs:annotation>
          <xs:documentation>Gully</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Gully</lx:path>
            <lx:label>Gully</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Harbour">
        <xs:annotation>
          <xs:documentation>Harbour</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Harbour</lx:path>
            <lx:label>Harbour</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Haven">
        <xs:annotation>
          <xs:documentation>Haven</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Haven</lx:path>
            <lx:label>Haven</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Head">
        <xs:annotation>
          <xs:documentation>Head</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Head</lx:path>
            <lx:label>Head</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Heath">
        <xs:annotation>
          <xs:documentation>Heath</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Heath</lx:path>
            <lx:label>Heath</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Heights">
        <xs:annotation>
          <xs:documentation>Heights</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Heights</lx:path>
            <lx:label>Heights</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Highroad">
        <xs:annotation>
          <xs:documentation>Highroad</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Highroad</lx:path>
            <lx:label>Highroad</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Highway">
        <xs:annotation>
          <xs:documentation>Highway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Highway</lx:path>
            <lx:label>Highway</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hill">
        <xs:annotation>
          <xs:documentation>Hill</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Hill</lx:path>
            <lx:label>Hill</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hollow">
        <xs:annotation>
          <xs:documentation>Hollow</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Hollow</lx:path>
            <lx:label>Hollow</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hub">
        <xs:annotation>
          <xs:documentation>Hub</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Hub</lx:path>
            <lx:label>Hub</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Interchange">
        <xs:annotation>
          <xs:documentation>Interchange</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Interchange</lx:path>
            <lx:label>Interchange</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Intersection">
        <xs:annotation>
          <xs:documentation>Intersection</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Intersection</lx:path>
            <lx:label>Intersection</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Island">
        <xs:annotation>
          <xs:documentation>Island</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Island</lx:path>
            <lx:label>Island</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Junction">
        <xs:annotation>
          <xs:documentation>Junction</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Junction</lx:path>
            <lx:label>Junction</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Key">
        <xs:annotation>
          <xs:documentation>Key</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Key</lx:path>
            <lx:label>Key</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Keys">
        <xs:annotation>
          <xs:documentation>Keys</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Keys</lx:path>
            <lx:label>Keys</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Knob">
        <xs:annotation>
          <xs:documentation>Knob</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Knob</lx:path>
            <lx:label>Knob</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ladder">
        <xs:annotation>
          <xs:documentation>Ladder</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Ladder</lx:path>
            <lx:label>Ladder</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Landing">
        <xs:annotation>
          <xs:documentation>Landing</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Landing</lx:path>
            <lx:label>Landing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lane">
        <xs:annotation>
          <xs:documentation>Lane</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Lane</lx:path>
            <lx:label>Lane</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Laneway">
        <xs:annotation>
          <xs:documentation>Laneway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Laneway</lx:path>
            <lx:label>Laneway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lees">
        <xs:annotation>
          <xs:documentation>Lees</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Lees</lx:path>
            <lx:label>Lees</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Line">
        <xs:annotation>
          <xs:documentation>Line</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Line</lx:path>
            <lx:label>Line</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Link">
        <xs:annotation>
          <xs:documentation>Link</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Link</lx:path>
            <lx:label>Link</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Little">
        <xs:annotation>
          <xs:documentation>Little</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Little</lx:path>
            <lx:label>Little</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lookout">
        <xs:annotation>
          <xs:documentation>Lookout</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Lookout</lx:path>
            <lx:label>Lookout</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loop">
        <xs:annotation>
          <xs:documentation>Loop</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Loop</lx:path>
            <lx:label>Loop</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lower">
        <xs:annotation>
          <xs:documentation>Lower</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Lower</lx:path>
            <lx:label>Lower</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mall">
        <xs:annotation>
          <xs:documentation>Mall</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Mall</lx:path>
            <lx:label>Mall</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Manor">
        <xs:annotation>
          <xs:documentation>Manor</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Manor</lx:path>
            <lx:label>Manor</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mead">
        <xs:annotation>
          <xs:documentation>Mead</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Mead</lx:path>
            <lx:label>Mead</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Meadows">
        <xs:annotation>
          <xs:documentation>Meadows</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Meadows</lx:path>
            <lx:label>Meadows</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Meander">
        <xs:annotation>
          <xs:documentation>Meander</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Meander</lx:path>
            <lx:label>Meander</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mears">
        <xs:annotation>
          <xs:documentation>Mears</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Mears</lx:path>
            <lx:label>Mears</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mew">
        <xs:annotation>
          <xs:documentation>Mew</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Mew</lx:path>
            <lx:label>Mew</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mews">
        <xs:annotation>
          <xs:documentation>Mews</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Mews</lx:path>
            <lx:label>Mews</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mile">
        <xs:annotation>
          <xs:documentation>Mile</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Mile</lx:path>
            <lx:label>Mile</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Motorway">
        <xs:annotation>
          <xs:documentation>Motorway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Motorway</lx:path>
            <lx:label>Motorway</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Motu">
        <xs:annotation>
          <xs:documentation>Motu</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Motu</lx:path>
            <lx:label>Motu</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mount">
        <xs:annotation>
          <xs:documentation>Mount</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Mount</lx:path>
            <lx:label>Mount</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Neaves">
        <xs:annotation>
          <xs:documentation>Neaves</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Neaves</lx:path>
            <lx:label>Neaves</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Nook">
        <xs:annotation>
          <xs:documentation>Nook</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Nook</lx:path>
            <lx:label>Nook</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Oaks">
        <xs:annotation>
          <xs:documentation>Oaks</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Oaks</lx:path>
            <lx:label>Oaks</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Outlet">
        <xs:annotation>
          <xs:documentation>Outlet</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Outlet</lx:path>
            <lx:label>Outlet</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Outlook">
        <xs:annotation>
          <xs:documentation>Outlook</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Outlook</lx:path>
            <lx:label>Outlook</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Paddock">
        <xs:annotation>
          <xs:documentation>Paddock</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Paddock</lx:path>
            <lx:label>Paddock</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Paku">
        <xs:annotation>
          <xs:documentation>Paku</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Paku</lx:path>
            <lx:label>Paku</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Parade">
        <xs:annotation>
          <xs:documentation>Parade</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Parade</lx:path>
            <lx:label>Parade</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Park">
        <xs:annotation>
          <xs:documentation>Park</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Park</lx:path>
            <lx:label>Park</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Parklands">
        <xs:annotation>
          <xs:documentation>Parklands</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Parklands</lx:path>
            <lx:label>Parklands</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Parkway">
        <xs:annotation>
          <xs:documentation>Parkway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Parkway</lx:path>
            <lx:label>Parkway</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Part">
        <xs:annotation>
          <xs:documentation>Part</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Part</lx:path>
            <lx:label>Part</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pass">
        <xs:annotation>
          <xs:documentation>Pass</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Pass</lx:path>
            <lx:label>Pass</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Passage">
        <xs:annotation>
          <xs:documentation>Passage</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Passage</lx:path>
            <lx:label>Passage</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Path">
        <xs:annotation>
          <xs:documentation>Path</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Path</lx:path>
            <lx:label>Path</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pathway">
        <xs:annotation>
          <xs:documentation>Pathway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Pathway</lx:path>
            <lx:label>Pathway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Piazza">
        <xs:annotation>
          <xs:documentation>Piazza</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Piazza</lx:path>
            <lx:label>Piazza</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pier">
        <xs:annotation>
          <xs:documentation>Pier</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Pier</lx:path>
            <lx:label>Pier</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Place">
        <xs:annotation>
          <xs:documentation>Place</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Place</lx:path>
            <lx:label>Place</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Plateau">
        <xs:annotation>
          <xs:documentation>Plateau</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Plateau</lx:path>
            <lx:label>Plateau</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Plaza">
        <xs:annotation>
          <xs:documentation>Plaza</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Plaza</lx:path>
            <lx:label>Plaza</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pocket">
        <xs:annotation>
          <xs:documentation>Pocket</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Pocket</lx:path>
            <lx:label>Pocket</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Point">
        <xs:annotation>
          <xs:documentation>Point</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Point</lx:path>
            <lx:label>Point</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Port">
        <xs:annotation>
          <xs:documentation>Port</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Port</lx:path>
            <lx:label>Port</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Portal">
        <xs:annotation>
          <xs:documentation>Portal</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Portal</lx:path>
            <lx:label>Portal</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Priors">
        <xs:annotation>
          <xs:documentation>Priors</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Priors</lx:path>
            <lx:label>Priors</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Promenade">
        <xs:annotation>
          <xs:documentation>Promenade</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Promenade</lx:path>
            <lx:label>Promenade</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pursuit">
        <xs:annotation>
          <xs:documentation>Pursuit</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Pursuit</lx:path>
            <lx:label>Pursuit</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Quad">
        <xs:annotation>
          <xs:documentation>Quad</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Quad</lx:path>
            <lx:label>Quad</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Quadrangle">
        <xs:annotation>
          <xs:documentation>Quadrangle</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Quadrangle</lx:path>
            <lx:label>Quadrangle</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Quadrant">
        <xs:annotation>
          <xs:documentation>Quadrant</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Quadrant</lx:path>
            <lx:label>Quadrant</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Quay">
        <xs:annotation>
          <xs:documentation>Quay</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Quay</lx:path>
            <lx:label>Quay</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Quays">
        <xs:annotation>
          <xs:documentation>Quays</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Quays</lx:path>
            <lx:label>Quays</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ramble">
        <xs:annotation>
          <xs:documentation>Ramble</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Ramble</lx:path>
            <lx:label>Ramble</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ramp">
        <xs:annotation>
          <xs:documentation>Ramp</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Ramp</lx:path>
            <lx:label>Ramp</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Range">
        <xs:annotation>
          <xs:documentation>Range</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Range</lx:path>
            <lx:label>Range</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Reach">
        <xs:annotation>
          <xs:documentation>Reach</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Reach</lx:path>
            <lx:label>Reach</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Reef">
        <xs:annotation>
          <xs:documentation>Reef</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Reef</lx:path>
            <lx:label>Reef</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Reserve">
        <xs:annotation>
          <xs:documentation>Reserve</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Reserve</lx:path>
            <lx:label>Reserve</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rest">
        <xs:annotation>
          <xs:documentation>Rest</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Rest</lx:path>
            <lx:label>Rest</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Retreat">
        <xs:annotation>
          <xs:documentation>Retreat</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Retreat</lx:path>
            <lx:label>Retreat</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Return">
        <xs:annotation>
          <xs:documentation>Return</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Return</lx:path>
            <lx:label>Return</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ride">
        <xs:annotation>
          <xs:documentation>Ride</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Ride</lx:path>
            <lx:label>Ride</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ridge">
        <xs:annotation>
          <xs:documentation>Ridge</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Ridge</lx:path>
            <lx:label>Ridge</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ridgeway">
        <xs:annotation>
          <xs:documentation>Ridgeway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Ridgeway</lx:path>
            <lx:label>Ridgeway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Right Of Way">
        <xs:annotation>
          <xs:documentation>Right Of Way</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Right Of Way</lx:path>
            <lx:label>Right Of Way</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ring">
        <xs:annotation>
          <xs:documentation>Ring</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Ring</lx:path>
            <lx:label>Ring</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rise">
        <xs:annotation>
          <xs:documentation>Rise</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Rise</lx:path>
            <lx:label>Rise</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rising">
        <xs:annotation>
          <xs:documentation>Rising</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Rising</lx:path>
            <lx:label>Rising</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="River">
        <xs:annotation>
          <xs:documentation>River</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.River</lx:path>
            <lx:label>River</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Riverway">
        <xs:annotation>
          <xs:documentation>Riverway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Riverway</lx:path>
            <lx:label>Riverway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Riviera">
        <xs:annotation>
          <xs:documentation>Riviera</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Riviera</lx:path>
            <lx:label>Riviera</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Road">
        <xs:annotation>
          <xs:documentation>Road</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Road</lx:path>
            <lx:label>Road</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Roads">
        <xs:annotation>
          <xs:documentation>Roads</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Roads</lx:path>
            <lx:label>Roads</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Roadside">
        <xs:annotation>
          <xs:documentation>Roadside</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Roadside</lx:path>
            <lx:label>Roadside</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Roadway">
        <xs:annotation>
          <xs:documentation>Roadway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Roadway</lx:path>
            <lx:label>Roadway</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ronde">
        <xs:annotation>
          <xs:documentation>Ronde</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Ronde</lx:path>
            <lx:label>Ronde</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rosebowl">
        <xs:annotation>
          <xs:documentation>Rosebowl</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Rosebowl</lx:path>
            <lx:label>Rosebowl</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rotary">
        <xs:annotation>
          <xs:documentation>Rotary</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Rotary</lx:path>
            <lx:label>Rotary</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Round">
        <xs:annotation>
          <xs:documentation>Round</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Round</lx:path>
            <lx:label>Round</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Route">
        <xs:annotation>
          <xs:documentation>Route</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Route</lx:path>
            <lx:label>Route</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Row">
        <xs:annotation>
          <xs:documentation>Row</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Row</lx:path>
            <lx:label>Row</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rue">
        <xs:annotation>
          <xs:documentation>Rue</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Rue</lx:path>
            <lx:label>Rue</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Run">
        <xs:annotation>
          <xs:documentation>Run</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Run</lx:path>
            <lx:label>Run</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Service Lane">
        <xs:annotation>
          <xs:documentation>Service Lane</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Service Lane</lx:path>
            <lx:label>Service Lane</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Service Way">
        <xs:annotation>
          <xs:documentation>Service Way</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Service Way</lx:path>
            <lx:label>Service Way</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Shunt">
        <xs:annotation>
          <xs:documentation>Shunt</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Shunt</lx:path>
            <lx:label>Shunt</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Siding">
        <xs:annotation>
          <xs:documentation>Siding</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Siding</lx:path>
            <lx:label>Siding</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Slope">
        <xs:annotation>
          <xs:documentation>Slope</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Slope</lx:path>
            <lx:label>Slope</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sound">
        <xs:annotation>
          <xs:documentation>Sound</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Sound</lx:path>
            <lx:label>Sound</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Spa">
        <xs:annotation>
          <xs:documentation>Spa</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Spa</lx:path>
            <lx:label>Spa</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Spur">
        <xs:annotation>
          <xs:documentation>Spur</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Spur</lx:path>
            <lx:label>Spur</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Square">
        <xs:annotation>
          <xs:documentation>Square</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Square</lx:path>
            <lx:label>Square</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Stairs">
        <xs:annotation>
          <xs:documentation>Stairs</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Stairs</lx:path>
            <lx:label>Stairs</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="State Highway">
        <xs:annotation>
          <xs:documentation>State Highway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.State Highway</lx:path>
            <lx:label>State Highway</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Steep">
        <xs:annotation>
          <xs:documentation>Steep</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Steep</lx:path>
            <lx:label>Steep</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Steps">
        <xs:annotation>
          <xs:documentation>Steps</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Steps</lx:path>
            <lx:label>Steps</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Straight">
        <xs:annotation>
          <xs:documentation>Straight</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Straight</lx:path>
            <lx:label>Straight</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Strand">
        <xs:annotation>
          <xs:documentation>Strand</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Strand</lx:path>
            <lx:label>Strand</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Street">
        <xs:annotation>
          <xs:documentation>Street</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Street</lx:path>
            <lx:label>Street</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Streets">
        <xs:annotation>
          <xs:documentation>Streets</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Streets</lx:path>
            <lx:label>Streets</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Strip">
        <xs:annotation>
          <xs:documentation>Strip</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Strip</lx:path>
            <lx:label>Strip</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Subway">
        <xs:annotation>
          <xs:documentation>Subway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Subway</lx:path>
            <lx:label>Subway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tarn">
        <xs:annotation>
          <xs:documentation>Tarn</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Tarn</lx:path>
            <lx:label>Tarn</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tarnice Way">
        <xs:annotation>
          <xs:documentation>Tarnice Way</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Tarnice Way</lx:path>
            <lx:label>Tarnice Way</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Terrace">
        <xs:annotation>
          <xs:documentation>Terrace</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Terrace</lx:path>
            <lx:label>Terrace</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Thoroughfare">
        <xs:annotation>
          <xs:documentation>Thoroughfare</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Thoroughfare</lx:path>
            <lx:label>Thoroughfare</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Throughway">
        <xs:annotation>
          <xs:documentation>Throughway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Throughway</lx:path>
            <lx:label>Throughway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tollway">
        <xs:annotation>
          <xs:documentation>Tollway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Tollway</lx:path>
            <lx:label>Tollway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Top">
        <xs:annotation>
          <xs:documentation>Top</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Top</lx:path>
            <lx:label>Top</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tor">
        <xs:annotation>
          <xs:documentation>Tor</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Tor</lx:path>
            <lx:label>Tor</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Towers">
        <xs:annotation>
          <xs:documentation>Towers</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Towers</lx:path>
            <lx:label>Towers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Track">
        <xs:annotation>
          <xs:documentation>Track</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Track</lx:path>
            <lx:label>Track</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trail">
        <xs:annotation>
          <xs:documentation>Trail</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Trail</lx:path>
            <lx:label>Trail</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trailer">
        <xs:annotation>
          <xs:documentation>Trailer</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Trailer</lx:path>
            <lx:label>Trailer</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tramway">
        <xs:annotation>
          <xs:documentation>Tramway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Tramway</lx:path>
            <lx:label>Tramway</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trees">
        <xs:annotation>
          <xs:documentation>Trees</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Trees</lx:path>
            <lx:label>Trees</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Triangle">
        <xs:annotation>
          <xs:documentation>Triangle</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Triangle</lx:path>
            <lx:label>Triangle</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trunkway">
        <xs:annotation>
          <xs:documentation>Trunkway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Trunkway</lx:path>
            <lx:label>Trunkway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Turn">
        <xs:annotation>
          <xs:documentation>Turn</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Turn</lx:path>
            <lx:label>Turn</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Twist">
        <xs:annotation>
          <xs:documentation>Twist</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Twist</lx:path>
            <lx:label>Twist</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Underpass">
        <xs:annotation>
          <xs:documentation>Underpass</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Underpass</lx:path>
            <lx:label>Underpass</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Upper">
        <xs:annotation>
          <xs:documentation>Upper</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Upper</lx:path>
            <lx:label>Upper</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vale">
        <xs:annotation>
          <xs:documentation>Vale</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Vale</lx:path>
            <lx:label>Vale</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Valley">
        <xs:annotation>
          <xs:documentation>Valley</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Valley</lx:path>
            <lx:label>Valley</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Venus">
        <xs:annotation>
          <xs:documentation>Venus</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Venus</lx:path>
            <lx:label>Venus</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Viaduct">
        <xs:annotation>
          <xs:documentation>Viaduct</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Viaduct</lx:path>
            <lx:label>Viaduct</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="View">
        <xs:annotation>
          <xs:documentation>View</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.View</lx:path>
            <lx:label>View</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Views">
        <xs:annotation>
          <xs:documentation>Views</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Views</lx:path>
            <lx:label>Views</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Village">
        <xs:annotation>
          <xs:documentation>Village</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Village</lx:path>
            <lx:label>Village</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Villas">
        <xs:annotation>
          <xs:documentation>Villas</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Villas</lx:path>
            <lx:label>Villas</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vista">
        <xs:annotation>
          <xs:documentation>Vista</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Vista</lx:path>
            <lx:label>Vista</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vue">
        <xs:annotation>
          <xs:documentation>Vue</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Vue</lx:path>
            <lx:label>Vue</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wade">
        <xs:annotation>
          <xs:documentation>Wade</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Wade</lx:path>
            <lx:label>Wade</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Walk">
        <xs:annotation>
          <xs:documentation>Walk</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Walk</lx:path>
            <lx:label>Walk</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Walkway">
        <xs:annotation>
          <xs:documentation>Walkway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Walkway</lx:path>
            <lx:label>Walkway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Waters">
        <xs:annotation>
          <xs:documentation>Waters</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Waters</lx:path>
            <lx:label>Waters</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Waterway">
        <xs:annotation>
          <xs:documentation>Waterway</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Waterway</lx:path>
            <lx:label>Waterway</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Way">
        <xs:annotation>
          <xs:documentation>Way</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Way</lx:path>
            <lx:label>Way</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wharf">
        <xs:annotation>
          <xs:documentation>Wharf</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Wharf</lx:path>
            <lx:label>Wharf</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Woods">
        <xs:annotation>
          <xs:documentation>Woods</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Woods</lx:path>
            <lx:label>Woods</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wynd">
        <xs:annotation>
          <xs:documentation>Wynd</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Wynd</lx:path>
            <lx:label>Wynd</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Yard">
        <xs:annotation>
          <xs:documentation>Yard</xs:documentation>
          <xs:appinfo>
            <lx:path>streetTypeList.Yard</lx:path>
            <lx:label>Yard</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="submissionTypeList">
    <xs:annotation>
      <xs:documentation>submissionTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>submissionTypeList</lx:path>
        <lx:label>submissionTypeList</lx:label>
        <li:transactions>CDA</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Broker Application">
        <xs:annotation>
          <xs:documentation>Broker Application</xs:documentation>
          <xs:appinfo>
            <lx:path>submissionTypeList.Broker Application</lx:path>
            <lx:label>Broker Application</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Online Application">
        <xs:annotation>
          <xs:documentation>Online Application</xs:documentation>
          <xs:appinfo>
            <lx:path>submissionTypeList.Online Application</lx:path>
            <lx:label>Online Application</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Override Credit Decision">
        <xs:annotation>
          <xs:documentation>Override Credit Decision</xs:documentation>
          <xs:appinfo>
            <lx:path>submissionTypeList.Override Credit Decision</lx:path>
            <lx:label>Override Credit Decision</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Post Bureau Credit Decision">
        <xs:annotation>
          <xs:documentation>Post Bureau Credit Decision</xs:documentation>
          <xs:appinfo>
            <lx:path>submissionTypeList.Post Bureau Credit Decision</lx:path>
            <lx:label>Post Bureau Credit Decision</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pre Bureau Credit Decision">
        <xs:annotation>
          <xs:documentation>Pre Bureau Credit Decision</xs:documentation>
          <xs:appinfo>
            <lx:path>submissionTypeList.Pre Bureau Credit Decision</lx:path>
            <lx:label>Pre Bureau Credit Decision</lx:label>
            <li:transactions>CDA</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="supplierTypeList">
    <xs:annotation>
      <xs:documentation>supplierTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>supplierTypeList</lx:path>
        <lx:label>supplierTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Licensed Dealer">
        <xs:annotation>
          <xs:documentation>Licensed Dealer</xs:documentation>
          <xs:appinfo>
            <lx:path>supplierTypeList.Licensed Dealer</lx:path>
            <lx:label>Licensed Dealer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Manufacturer">
        <xs:annotation>
          <xs:documentation>Manufacturer</xs:documentation>
          <xs:appinfo>
            <lx:path>supplierTypeList.Manufacturer</lx:path>
            <lx:label>Manufacturer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>supplierTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Private Sale">
        <xs:annotation>
          <xs:documentation>Private Sale</xs:documentation>
          <xs:appinfo>
            <lx:path>supplierTypeList.Private Sale</lx:path>
            <lx:label>Private Sale</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Recognised Supplier">
        <xs:annotation>
          <xs:documentation>Recognised Supplier</xs:documentation>
          <xs:appinfo>
            <lx:path>supplierTypeList.Recognised Supplier</lx:path>
            <lx:label>Recognised Supplier</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="surplusFundsDisbursementMethodList">
    <xs:annotation>
      <xs:documentation>surplusFundsDisbursementMethodList</xs:documentation>
      <xs:appinfo>
        <lx:path>surplusFundsDisbursementMethodList</lx:path>
        <lx:label>surplusFundsDisbursementMethodList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Bank Cheque">
        <xs:annotation>
          <xs:documentation>Bank Cheque</xs:documentation>
          <xs:appinfo>
            <lx:path>surplusFundsDisbursementMethodList.Bank Cheque</lx:path>
            <lx:label>Bank Cheque</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bpay">
        <xs:annotation>
          <xs:documentation>BPAY electronic bill payment.</xs:documentation>
          <xs:appinfo>
            <lx:path>surplusFundsDisbursementMethodList.Bpay</lx:path>
            <lx:label>BPAY</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Existing Direct Debit Account">
        <xs:annotation>
          <xs:documentation>Existing Direct Debit Account</xs:documentation>
          <xs:appinfo>
            <lx:path>surplusFundsDisbursementMethodList.Existing Direct Debit Account</lx:path>
            <lx:label>Existing Direct Debit Account</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Existing Other Account">
        <xs:annotation>
          <xs:documentation>Existing Other Account</xs:documentation>
          <xs:appinfo>
            <lx:path>surplusFundsDisbursementMethodList.Existing Other Account</lx:path>
            <lx:label>Existing Other Account</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="New Account">
        <xs:annotation>
          <xs:documentation>New Account</xs:documentation>
          <xs:appinfo>
            <lx:path>surplusFundsDisbursementMethodList.New Account</lx:path>
            <lx:label>New Account</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Redraw">
        <xs:annotation>
          <xs:documentation>Redraw</xs:documentation>
          <xs:appinfo>
            <lx:path>surplusFundsDisbursementMethodList.Redraw</lx:path>
            <lx:label>Redraw</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="taxDeclarationExemptionCategoryList">
    <xs:annotation>
      <xs:documentation>taxDeclarationExemptionCategoryList</xs:documentation>
      <xs:appinfo>
        <lx:path>taxDeclarationExemptionCategoryList</lx:path>
        <lx:label>taxDeclarationExemptionCategoryList</lx:label>
        <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Age Pension">
        <xs:annotation>
          <xs:documentation>Investment bodies are not required to withhold amounts for non-quotation of a TFN or ABN from investments held by recipients of an Age Pension.</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDeclarationExemptionCategoryList.Age Pension</lx:path>
            <lx:label>Age Pension</lx:label>
            <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Carer Payment">
        <xs:annotation>
          <xs:documentation>Investment bodies are not required to withhold amounts for non-quotation of a TFN or ABN from investments held by recipients of a Carer Payment.</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDeclarationExemptionCategoryList.Carer Payment</lx:path>
            <lx:label>Carer Payment</lx:label>
            <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Child Exemption">
        <xs:annotation>
          <xs:documentation>Investment bodies are not required to withhold for non-quotation of a TFN or ABN from investments (other than shares in a public company) held by children under 16 years of age under certain specific conditions.</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDeclarationExemptionCategoryList.Child Exemption</lx:path>
            <lx:label>Child Exemption</lx:label>
            <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Disability Support Pension">
        <xs:annotation>
          <xs:documentation>Investment bodies are not required to withhold amounts for non-quotation of a TFN or ABN from investments held by recipients of a Disability Support Pension.</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDeclarationExemptionCategoryList.Disability Support Pension</lx:path>
            <lx:label>Disability Support Pension</lx:label>
            <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Non-Resident Exemption">
        <xs:annotation>
          <xs:documentation>Investment bodies are not required to withhold amounts for non-quotation of a TFN or ABN from investments held by residents living outside of Australia.</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDeclarationExemptionCategoryList.Non-Resident Exemption</lx:path>
            <lx:label>Non-Resident Exemption</lx:label>
            <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Not Required for Interest Income">
        <xs:annotation>
          <xs:documentation>Investment bodies are not required to withhold for non-quotation of a TFN or ABN from interest bearing deposits and accounts with a financial institution, provided that the payment is less than $120 for the financial year (if the payment is not for the whole financial year, the $120 threshold is pro-rated).</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDeclarationExemptionCategoryList.Not Required for Interest Income</lx:path>
            <lx:label>Not Required for Interest Income</lx:label>
            <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Not Required to Lodge a Tax Return">
        <xs:annotation>
          <xs:documentation>An entity that isn't required to lodge an income tax return and doesn't have a TFN is taken to have quoted a TFN for an investment, which is reportable in an AIIR (Annual investment income report), to the investment body.</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDeclarationExemptionCategoryList.Not Required to Lodge a Tax Return</lx:path>
            <lx:label>Not Required to Lodge a Tax Return</lx:label>
            <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Parenting Payment Single">
        <xs:annotation>
          <xs:documentation>Investment bodies are not required to withhold amounts for non-quotation of a TFN or ABN from investments held by recipients of a Parenting Payment Single.</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDeclarationExemptionCategoryList.Parenting Payment Single</lx:path>
            <lx:label>Parenting Payment Single</lx:label>
            <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Special Benefit">
        <xs:annotation>
          <xs:documentation>Investment bodies are not required to withhold amounts for non-quotation of a TFN or ABN from investments held by recipients of a Special Benefit.</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDeclarationExemptionCategoryList.Special Benefit</lx:path>
            <lx:label>Special Benefit</lx:label>
            <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Special Needs Pension">
        <xs:annotation>
          <xs:documentation>Investment bodies are not required to withhold amounts for non-quotation of a TFN or ABN from investments held by recipients of a Special Needs Pension.</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDeclarationExemptionCategoryList.Special Needs Pension</lx:path>
            <lx:label>Special Needs Pension</lx:label>
            <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Veteran Allowance">
        <xs:annotation>
          <xs:documentation>Investment bodies are not required to withhold amounts for non-quotation of a TFN or ABN from investments held by recipients of a Defence Force income support allowance from Veterans' Affairs</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDeclarationExemptionCategoryList.Veteran Allowance</lx:path>
            <lx:label>Veteran Allowance</lx:label>
            <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Veteran Pension">
        <xs:annotation>
          <xs:documentation>Investment bodies are not required to withhold amounts for non-quotation of a TFN or ABN from investments held by recipients of an age, invalidity or partner service pension from Veterans' Affairs.</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDeclarationExemptionCategoryList.Veteran Pension</lx:path>
            <lx:label>Veteran Pension</lx:label>
            <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Widow B Pension">
        <xs:annotation>
          <xs:documentation>Investment bodies are not required to withhold amounts for non-quotation of a TFN or ABN from investments held by recipients of a Widow B Pension.</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDeclarationExemptionCategoryList.Widow B Pension</lx:path>
            <lx:label>Widow B Pension</lx:label>
            <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wife Pension">
        <xs:annotation>
          <xs:documentation>Investment bodies are not required to withhold amounts for non-quotation of a TFN or ABN from investments held by recipients of a Wife Pension.</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDeclarationExemptionCategoryList.Wife Pension</lx:path>
            <lx:label>Wife Pension</lx:label>
            <li:transactions>CAL,CDA,DAS,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="taxDepreciationMethodList">
    <xs:annotation>
      <xs:documentation>taxDepreciationMethodList</xs:documentation>
      <xs:appinfo>
        <lx:path>taxDepreciationMethodList</lx:path>
        <lx:label>taxDepreciationMethodList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Diminishing Value">
        <xs:annotation>
          <xs:documentation>Diminishing Value</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDepreciationMethodList.Diminishing Value</lx:path>
            <lx:label>Diminishing Value</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Straight Line">
        <xs:annotation>
          <xs:documentation>Straight Line</xs:documentation>
          <xs:appinfo>
            <lx:path>taxDepreciationMethodList.Straight Line</lx:path>
            <lx:label>Straight Line</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="tenureTypeList">
    <xs:annotation>
      <xs:documentation>tenureTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>tenureTypeList</lx:path>
        <lx:label>tenureTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Community Title">
        <xs:annotation>
          <xs:documentation>Tenure held is under a Community Scheme or a Community Strata Scheme. Both divide land to create lots and common property in a similar manner to strata titles. Unlike a strata title, a scheme may include a development lot, retained by the developer, for later division into further lots within the scheme.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Community Title</lx:path>
            <lx:label>Tenure held is under a Community Scheme or a Community Strata Scheme. Both divide land to create lots and common property in a similar manner to strata titles. Unlike a strata title, a scheme may include a development lot, retained by the developer, for later division into further lots within the scheme.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Company Title">
        <xs:annotation>
          <xs:documentation>Tenure is held by ownership of shares in a company that grant rights to exclusively occupy parts of land that the company owns.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Company Title</lx:path>
            <lx:label>Tenure is held by ownership of shares in a company that grant rights to exclusively occupy parts of land that the company owns.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cross Lease">
        <xs:annotation>
          <xs:documentation>Tenure held is under a Cross-Lease Title, where the property owner holds a registered lease over their property on the landlord, and an undivided ownership share in the land with other property owners.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Cross Lease</lx:path>
            <lx:label>Tenure held is Cross-Lease.</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Crown Land">
        <xs:annotation>
          <xs:documentation>Tenure held is in the 'right of the Crown' of either an individual State or the Commonwealth of Australia.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Crown Land</lx:path>
            <lx:label>Tenure held is in the 'right of the Crown' of either an individual State or the Commonwealth of Australia.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Crown Lease">
        <xs:annotation>
          <xs:documentation>Tenure held is under a Crown lease granted by the Crown over Crown land, commonly with a 99 year leasehold term.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Crown Lease</lx:path>
            <lx:label>Tenure held is under a Crown lease granted by the Crown over Crown land, commonly with a 99 year leasehold term.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Flat Owning Company Shares">
        <xs:annotation>
          <xs:documentation>Tenure is held by ownership of shares in a company that grant rights to exclusively occupy parts of land that the company owns.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Flat Owning Company Shares</lx:path>
            <lx:label>Flat Owning Company Shares</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Freehold">
        <xs:annotation>
          <xs:documentation>Tenure held is outright ownership, in perpetuity.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Freehold</lx:path>
            <lx:label>Tenure held is outright ownership, in perpetuity.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Freehold Retirement Complex">
        <xs:annotation>
          <xs:documentation>Tenure held is outright ownership, in perpetuity. This is specifically used for Retirement Complexes.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Freehold Retirement Complex</lx:path>
            <lx:label>Freehold Retirement Complex</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Freehold Stratum Estate">
        <xs:annotation>
          <xs:documentation>Tenure held is outright ownership, in perpetuity. This is specifically used for individual portions of freehold land, common with blocks of units, townhouses etc. were the building is collectively owned by the residents that live there.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Freehold Stratum Estate</lx:path>
            <lx:label>Freehold Stratum Estate</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Group Titles Plan">
        <xs:annotation>
          <xs:documentation>Tenure held is as a lot collectively administered under the Group Titles Act 1980 (QLD).</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Group Titles Plan</lx:path>
            <lx:label>Tenure held is as a lot collectively administered under the Group Titles Act 1980 (QLD).</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Leasehold">
        <xs:annotation>
          <xs:documentation>Tenure held is as a lessee under a lease from the freehold owner of the property (lessor).</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Leasehold</lx:path>
            <lx:label>Tenure held is as a lessee under a lease from the freehold owner of the property (lessor).</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Leasehold Retirement Complex">
        <xs:annotation>
          <xs:documentation>Tenure held is as a lessee under a lease from the freehold owner of the property (lessor). This is specifically used for Retirement Complexes.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Leasehold Retirement Complex</lx:path>
            <lx:label>Leasehold Retirement Complex</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Leasehold Stratum Estate">
        <xs:annotation>
          <xs:documentation>Tenure held is as a lessee under a lease from the freehold owner of the property (lessor). This is specifically used for individual portions of freehold land, common with blocks of units, townhouses etc. where the building is collectively owned by the residents that live there.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Leasehold Stratum Estate</lx:path>
            <lx:label>Tenure held is as a lessee under a lease from the freehold owner of the property (lessor).</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Licence to Occupy Retirement Complex">
        <xs:annotation>
          <xs:documentation>Tenure is held by a landlord who owns an apartment or building and rents the right to occupy the unit to a tenant. This is specifically used for Retirement Complexes.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Licence to Occupy Retirement Complex</lx:path>
            <lx:label>Licence to Occupy Retirement Complex</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Maori Land">
        <xs:annotation>
          <xs:documentation>Tenure held is for Maori Land.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Maori Land</lx:path>
            <lx:label>Maori Land</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Moiety Title">
        <xs:annotation>
          <xs:documentation>Tenure held is a share of the whole of the land and a lease over a defined portion of the land for from the other owner or owners, typically covering maisonettes or attached cottages.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Moiety Title</lx:path>
            <lx:label>Tenure held is a share of the whole of the land and a lease over a defined portion of the land for from the other owner or owners, typically covering maisonettes or attached cottages.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="None">
        <xs:annotation>
          <xs:documentation>Indicates that no tenure is specified.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.None</lx:path>
            <lx:label>Indicates that no tenure is specified.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other Title">
        <xs:annotation>
          <xs:documentation>Indicates that the type of tenure held is not described by any of the other enumerated values.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Other Title</lx:path>
            <lx:label>Indicates that the type of tenure held is not described by any of the other enumerated values.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Purple Title">
        <xs:annotation>
          <xs:documentation>Tenure held is ownership of an undivided share in a property, with the right to exclusively occupy certain areas in the complex determined by contract with each resident.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Purple Title</lx:path>
            <lx:label>Tenure held is ownership of an undivided share in a property, with the right to exclusively occupy certain areas in the complex determined by contract with each resident.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Strata Title">
        <xs:annotation>
          <xs:documentation>Tenure held is as a unit or lot on a plan of subdivision associated with townhouses, units and blocks of flats and based on the horizontal and vertical subdivision of air space. Owners have a certificate of title, are absolute owners of a freehold flat and have an undivided share of the common property.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Strata Title</lx:path>
            <lx:label>Tenure held is as a unit or lot on a plan of subdivision associated with townhouses, units and blocks of flats and based on the horizontal and vertical subdivision of air space. Owners have a certificate of title, are absolute owners of a freehold flat and have an undivided share of the common property.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Stratum Title">
        <xs:annotation>
          <xs:documentation>Tenure held is freehold over an apartment, unit or flat, with in addition the owner automatically becoming a member of a service company that administers, manages and maintains the property in which the owner's flat is registered.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Stratum Title</lx:path>
            <lx:label>Tenure held is freehold over an apartment, unit or flat, with in addition the owner automatically becoming a member of a service company that administers, manages and maintains the property in which the owner's flat is registered.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Survey Strata Title">
        <xs:annotation>
          <xs:documentation>Equivalent to a Strata Title with the exception that no buildings are shown on the Plan.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Survey Strata Title</lx:path>
            <lx:label>Equivalent to a Strata Title with the exception that no buildings are shown on the Plan.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unit Title">
        <xs:annotation>
          <xs:documentation>Tenure held is unit title under the Unit Titles Act (ACT) which is the ACT equivalent of Strata Title.</xs:documentation>
          <xs:appinfo>
            <lx:path>tenureTypeList.Unit Title</lx:path>
            <lx:label>Tenure held is unit title under the Unit Titles Act (ACT) which is the ACT equivalent of Strata Title.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="tinNotProvidedReasonList">
    <xs:annotation>
      <xs:documentation>List of valid reasons as to why a Tax Identification Number (TIN) is not provided by the entity or the representative of that entity.</xs:documentation>
      <xs:appinfo>
        <lx:path>tinNotProvidedReasonList</lx:path>
        <lx:label>TIN Not Provided Reason List</lx:label>
        <li:transactions>CAL,CDA,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Country of Tax Association Does Not Issue a TIN">
        <xs:annotation>
          <xs:documentation>The country/jurisdiction of tax association does not issue TINs to its residents.</xs:documentation>
          <xs:appinfo>
            <lx:path>tinNotProvidedReasonList.Country of Tax Association Does Not Issue a TIN</lx:path>
            <lx:label>Country of Tax Association Does Not Issue a TIN</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Country of Tax Association Does Not Require TIN to Be Disclosed">
        <xs:annotation>
          <xs:documentation>The country/jurisdiction of tax association does not require the disclosure of the TIN.</xs:documentation>
          <xs:appinfo>
            <lx:path>tinNotProvidedReasonList.Country of Tax Association Does Not Require TIN to Be Disclosed</lx:path>
            <lx:label>Country of Tax Association Does Not Require TIN to Be Disclosed</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Entity Is Otherwise Unable to Obtain a TIN or Equivalent Number">
        <xs:annotation>
          <xs:documentation>The entity is otherwise unable to obtain a TIN or equivalent number.</xs:documentation>
          <xs:appinfo>
            <lx:path>tinNotProvidedReasonList.Entity Is Otherwise Unable to Obtain a TIN or Equivalent Number</lx:path>
            <lx:label>Entity Is Otherwise Unable to Obtain a TIN or Equivalent Number</lx:label>
            <li:transactions>CAL,CDA,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="titleSystemList">
    <xs:annotation>
      <xs:documentation>titleSystemList</xs:documentation>
      <xs:appinfo>
        <lx:path>titleSystemList</lx:path>
        <lx:label>titleSystemList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Company">
        <xs:annotation>
          <xs:documentation>The title is granted by ownership of shares in a company that have rights to exclusively occupy parts of land that the company owns.</xs:documentation>
          <xs:appinfo>
            <lx:path>titleSystemList.Company</lx:path>
            <lx:label>The title is granted by ownership of shares in a company that have rights to exclusively occupy parts of land that the company owns.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Old Law">
        <xs:annotation>
          <xs:documentation>The title is granted by a chain of title documents or deeds, the system replaced by Torrens Title.</xs:documentation>
          <xs:appinfo>
            <lx:path>titleSystemList.Old Law</lx:path>
            <lx:label>The title is granted by a chain of title documents or deeds, the system replaced by Torrens Title.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Qualified">
        <xs:annotation>
          <xs:documentation>A folio in the Torrens Title Register in which is recorded a "caution" that has not been cancelled. A caution warns persons dealing with the registered proprietor that the land is held subject to any "subsisting interests" that may exist under Common Law or Old System.</xs:documentation>
          <xs:appinfo>
            <lx:path>titleSystemList.Qualified</lx:path>
            <lx:label>A folio in the Torrens Title Register in which is recorded a "caution" that has not been cancelled. A caution warns persons dealing with the registered proprietor that the land is held subject to any "subsisting interests" that may exist under Common Law or Old System.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Torrens">
        <xs:annotation>
          <xs:documentation>The title is registered in a public register that is maintained by a Land Registry under relevant legislation.</xs:documentation>
          <xs:appinfo>
            <lx:path>titleSystemList.Torrens</lx:path>
            <lx:label>The title is registered in a public register that is maintained by a Land Registry under relevant legislation.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unregistered">
        <xs:annotation>
          <xs:documentation>The title is pending registration in the Torrens Title Register.</xs:documentation>
          <xs:appinfo>
            <lx:path>titleSystemList.Unregistered</lx:path>
            <lx:label>The title is pending registration in the Torrens Title Register.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="toolsOfTradeTypeList">
    <xs:annotation>
      <xs:documentation>toolsOfTradeTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>toolsOfTradeTypeList</lx:path>
        <lx:label>toolsOfTradeTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Camera">
        <xs:annotation>
          <xs:documentation>Camera</xs:documentation>
          <xs:appinfo>
            <lx:path>toolsOfTradeTypeList.Camera</lx:path>
            <lx:label>Camera</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>toolsOfTradeTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Power Tools">
        <xs:annotation>
          <xs:documentation>Power Tools</xs:documentation>
          <xs:appinfo>
            <lx:path>toolsOfTradeTypeList.Power Tools</lx:path>
            <lx:label>Power Tools</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Surveying Equipment">
        <xs:annotation>
          <xs:documentation>Surveying Equipment</xs:documentation>
          <xs:appinfo>
            <lx:path>toolsOfTradeTypeList.Surveying Equipment</lx:path>
            <lx:label>Surveying Equipment</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tools - Laser Beam">
        <xs:annotation>
          <xs:documentation>Tools - Laser Beam</xs:documentation>
          <xs:appinfo>
            <lx:path>toolsOfTradeTypeList.Tools - Laser Beam</lx:path>
            <lx:label>Tools - Laser Beam</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tools General">
        <xs:annotation>
          <xs:documentation>Tools General</xs:documentation>
          <xs:appinfo>
            <lx:path>toolsOfTradeTypeList.Tools General</lx:path>
            <lx:label>Tools General</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="totalTermTypeList">
    <xs:annotation>
      <xs:documentation>totalTermTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>totalTermTypeList</lx:path>
        <lx:label>totalTermTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Amortisation Term">
        <xs:annotation>
          <xs:documentation>The term specified is the amortisation term of the loan, being the period over which the amortising repayment is calculated.</xs:documentation>
          <xs:appinfo>
            <lx:path>totalTermTypeList.Amortisation Term</lx:path>
            <lx:label>The term specified is the amortisation term of the loan, being the period over which the amortising repayment is calculated.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Total Term">
        <xs:annotation>
          <xs:documentation>The term specified is the total term of the loan, being the period at the end of which the loan principal must be repaid.</xs:documentation>
          <xs:appinfo>
            <lx:path>totalTermTypeList.Total Term</lx:path>
            <lx:label>The term specified is the total term of the loan, being the period at the end of which the loan principal must be repaid.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="transactionTypeList">
    <xs:annotation>
      <xs:documentation>transactionTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>transactionTypeList</lx:path>
        <lx:label>transactionTypeList</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Credit">
        <xs:annotation>
          <xs:documentation>Credit</xs:documentation>
          <xs:appinfo>
            <lx:path>transactionTypeList.Credit</lx:path>
            <lx:label>Credit</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Debit">
        <xs:annotation>
          <xs:documentation>Debit</xs:documentation>
          <xs:appinfo>
            <lx:path>transactionTypeList.Debit</lx:path>
            <lx:label>Debit</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="trustPurposeList">
    <xs:annotation>
      <xs:documentation>trustPurposeList</xs:documentation>
      <xs:appinfo>
        <lx:path>trustPurposeList</lx:path>
        <lx:label>trustPurposeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Family Trust">
        <xs:annotation>
          <xs:documentation>A discretionary trust for which a family trust election has been made.</xs:documentation>
          <xs:appinfo>
            <lx:path>trustPurposeList.Family Trust</lx:path>
            <lx:label>A discretionary trust for which a family trust election has been made.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Protective">
        <xs:annotation>
          <xs:documentation>The trust has been established for a person who is unable to manage his or her own affairs.</xs:documentation>
          <xs:appinfo>
            <lx:path>trustPurposeList.Protective</lx:path>
            <lx:label>The trust has been established for a person who is unable to manage his or her own affairs.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Security Trust">
        <xs:annotation>
          <xs:documentation>A trust, typically a simple (bare) trust, set up to hold property as security for a loan, for example for an SMSF. Also known as a property trust.</xs:documentation>
          <xs:appinfo>
            <lx:path>trustPurposeList.Security Trust</lx:path>
            <lx:label>A trust, typically a simple (bare) trust, set up to hold property as security for a loan, for example for an SMSF. Also known as a property trust.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SMSF">
        <xs:annotation>
          <xs:documentation>Self managed super fund trust.</xs:documentation>
          <xs:appinfo>
            <lx:path>trustPurposeList.SMSF</lx:path>
            <lx:label>Self managed super fund trust.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trading Trust">
        <xs:annotation>
          <xs:documentation>A trust that actively carries on business, and hence where the trust assets are used in the conduct of business.</xs:documentation>
          <xs:appinfo>
            <lx:path>trustPurposeList.Trading Trust</lx:path>
            <lx:label>A trust that actively carries on business, and hence where the trust assets are used in the conduct of business.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="trustStructureList">
    <xs:annotation>
      <xs:documentation>trustStructureList</xs:documentation>
      <xs:appinfo>
        <lx:path>trustStructureList</lx:path>
        <lx:label>trustStructureList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Discretionary">
        <xs:annotation>
          <xs:documentation>The trustee has discretion as to the distribution of trust property to the beneficiaries.</xs:documentation>
          <xs:appinfo>
            <lx:path>trustStructureList.Discretionary</lx:path>
            <lx:label>The trustee has discretion as to the distribution of trust property to the beneficiaries.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fixed">
        <xs:annotation>
          <xs:documentation>The trustee holds the trust assets for the benefit of the beneficiaries in specified fixed proportions.</xs:documentation>
          <xs:appinfo>
            <lx:path>trustStructureList.Fixed</lx:path>
            <lx:label>The trustee holds the trust assets for the benefit of the beneficiaries in specified fixed proportions.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hybrid">
        <xs:annotation>
          <xs:documentation>A trust that combines elements of both fixed and discretionary trusts, in that the trustee must pay a certain amount of the trust property to each beneficiary, but otherwise has discretion as to how any remaining trust property is distributed.</xs:documentation>
          <xs:appinfo>
            <lx:path>trustStructureList.Hybrid</lx:path>
            <lx:label>A trust that combines elements of both fixed and discretionary trusts, in that the trustee must pay a certain amount of the trust property to each beneficiary, but otherwise has discretion as to how any remaining trust property is distributed.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Simple">
        <xs:annotation>
          <xs:documentation>A trust with one trustee and one beneficiary, and where the trustee has no active duty beyond conveying the property to the beneficiary at some future time determined by the trust. Also known as a bare trust.</xs:documentation>
          <xs:appinfo>
            <lx:path>trustStructureList.Simple</lx:path>
            <lx:label>A trust with one trustee and one beneficiary, and where the trustee has no active duty beyond conveying the property to the beneficiary at some future time determined by the trust. Also known as a bare trust.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unit">
        <xs:annotation>
          <xs:documentation>The beneficiaries (called unit holders) each possess a certain share (called units) and can direct the trustee to pay money to them out of the trust property according to the number of units they possess.</xs:documentation>
          <xs:appinfo>
            <lx:path>trustStructureList.Unit</lx:path>
            <lx:label>The beneficiaries (called unit holders) each possess a certain share (called units) and can direct the trustee to pay money to them out of the trust property according to the number of units they possess.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="typeOfIncorporationList">
    <xs:annotation>
      <xs:documentation>A list of valid types of incorporation that describe the valid company structures.</xs:documentation>
      <xs:appinfo>
        <lx:path>typeOfIncorporationList</lx:path>
        <lx:label>typeOfIncorporationList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Inc">
        <xs:annotation>
          <xs:documentation>An incorporated association under the relevant state or territory legislation.</xs:documentation>
          <xs:appinfo>
            <lx:path>typeOfIncorporationList.Inc</lx:path>
            <lx:label>An incorporated association under the relevant state or territory legislation.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LTC">
        <xs:annotation>
          <xs:documentation>This item is now deprecated and will be deleted in a future release. A Look-Through Company (LTC) is a kind of tax structure for New Zealand companies with limited liability, which allows the company in question to transfer its income and expenditure to its shareholders directly.</xs:documentation>
          <xs:appinfo>
            <lx:path>typeOfIncorporationList.LTC</lx:path>
            <lx:label>Look-Through Company</lx:label>
            <li:transactions>CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ltd">
        <xs:annotation>
          <xs:documentation>A public company limited by shares.</xs:documentation>
          <xs:appinfo>
            <lx:path>typeOfIncorporationList.Ltd</lx:path>
            <lx:label>A public company limited by shares.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="NL">
        <xs:annotation>
          <xs:documentation>A "No Liability Company", meaning that it must be solely for the objective of mining and have no further call on any of its unpaid shares.</xs:documentation>
          <xs:appinfo>
            <lx:path>typeOfIncorporationList.NL</lx:path>
            <lx:label>A "No Liability Company", meaning that it must be solely for the objective of mining and have no further call on any of its unpaid shares.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Pty Ltd">
        <xs:annotation>
          <xs:documentation>A Proprietary Limited company, being a private company limited by shares.</xs:documentation>
          <xs:appinfo>
            <lx:path>typeOfIncorporationList.Pty Ltd</lx:path>
            <lx:label>A Proprietary Limited company, being a private company limited by shares.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="underwritingTypeList">
    <xs:annotation>
      <xs:documentation>underwritingTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>underwritingTypeList</lx:path>
        <lx:label>underwritingTypeList</lx:label>
        <li:transactions>LMI</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="eDUA">
        <xs:annotation>
          <xs:documentation>eDUA</xs:documentation>
          <xs:appinfo>
            <lx:path>underwritingTypeList.eDUA</lx:path>
            <lx:label>eDUA</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Standard">
        <xs:annotation>
          <xs:documentation>Standard</xs:documentation>
          <xs:appinfo>
            <lx:path>underwritingTypeList.Standard</lx:path>
            <lx:label>Standard</lx:label>
            <li:transactions>LMI</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="unitTypeList">
    <xs:annotation>
      <xs:documentation>unitTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>unitTypeList</lx:path>
        <lx:label>unitTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Antenna">
        <xs:annotation>
          <xs:documentation>Antenna</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Antenna</lx:path>
            <lx:label>Antenna</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Apartment">
        <xs:annotation>
          <xs:documentation>Apartment</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Apartment</lx:path>
            <lx:label>Apartment</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Automated Teller Machine">
        <xs:annotation>
          <xs:documentation>Automated Teller Machine</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Automated Teller Machine</lx:path>
            <lx:label>Automated Teller Machine</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Automatic Teller">
        <xs:annotation>
          <xs:documentation>Automatic Teller</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Automatic Teller</lx:path>
            <lx:label>Automatic Teller</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Barbecue">
        <xs:annotation>
          <xs:documentation>Barbecue</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Barbecue</lx:path>
            <lx:label>Barbecue</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Boatshed">
        <xs:annotation>
          <xs:documentation>Boatshed</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Boatshed</lx:path>
            <lx:label>Boatshed</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Building">
        <xs:annotation>
          <xs:documentation>Building</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Building</lx:path>
            <lx:label>Building</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bungalow">
        <xs:annotation>
          <xs:documentation>Bungalow</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Bungalow</lx:path>
            <lx:label>Bungalow</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cage">
        <xs:annotation>
          <xs:documentation>Cage</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Cage</lx:path>
            <lx:label>Cage</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Carpark">
        <xs:annotation>
          <xs:documentation>Carpark</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Carpark</lx:path>
            <lx:label>Carpark</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Carspace">
        <xs:annotation>
          <xs:documentation>Carspace</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Carspace</lx:path>
            <lx:label>Carspace</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Club">
        <xs:annotation>
          <xs:documentation>Club</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Club</lx:path>
            <lx:label>Club</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Condominium">
        <xs:annotation>
          <xs:documentation>Condominium</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Condominium</lx:path>
            <lx:label>Condominium</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Coolroom">
        <xs:annotation>
          <xs:documentation>Coolroom</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Coolroom</lx:path>
            <lx:label>Coolroom</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cottage">
        <xs:annotation>
          <xs:documentation>Cottage</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Cottage</lx:path>
            <lx:label>Cottage</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Duplex">
        <xs:annotation>
          <xs:documentation>Duplex</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Duplex</lx:path>
            <lx:label>Duplex</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Factory">
        <xs:annotation>
          <xs:documentation>Factory</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Factory</lx:path>
            <lx:label>Factory</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Flat">
        <xs:annotation>
          <xs:documentation>Flat</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Flat</lx:path>
            <lx:label>Flat</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Garage">
        <xs:annotation>
          <xs:documentation>Garage</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Garage</lx:path>
            <lx:label>Garage</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hall">
        <xs:annotation>
          <xs:documentation>Hall</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Hall</lx:path>
            <lx:label>Hall</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="House">
        <xs:annotation>
          <xs:documentation>House</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.House</lx:path>
            <lx:label>House</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Kiosk">
        <xs:annotation>
          <xs:documentation>Kiosk</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Kiosk</lx:path>
            <lx:label>Kiosk</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lease">
        <xs:annotation>
          <xs:documentation>Lease</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Lease</lx:path>
            <lx:label>Lease</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lobby">
        <xs:annotation>
          <xs:documentation>Lobby</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Lobby</lx:path>
            <lx:label>Lobby</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Loft">
        <xs:annotation>
          <xs:documentation>Loft</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Loft</lx:path>
            <lx:label>Loft</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lot">
        <xs:annotation>
          <xs:documentation>Lot</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Lot</lx:path>
            <lx:label>Lot</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Maisonette">
        <xs:annotation>
          <xs:documentation>Maisonette</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Maisonette</lx:path>
            <lx:label>Maisonette</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Marine Berth">
        <xs:annotation>
          <xs:documentation>Marine Berth</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Marine Berth</lx:path>
            <lx:label>Marine Berth</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Office">
        <xs:annotation>
          <xs:documentation>Office</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Office</lx:path>
            <lx:label>Office</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Penthouse">
        <xs:annotation>
          <xs:documentation>Penthouse</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Penthouse</lx:path>
            <lx:label>Penthouse</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Reserve">
        <xs:annotation>
          <xs:documentation>Reserve</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Reserve</lx:path>
            <lx:label>Reserve</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Room">
        <xs:annotation>
          <xs:documentation>Room</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Room</lx:path>
            <lx:label>Room</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Shed">
        <xs:annotation>
          <xs:documentation>Shed</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Shed</lx:path>
            <lx:label>Shed</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Shop">
        <xs:annotation>
          <xs:documentation>Shop</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Shop</lx:path>
            <lx:label>Shop</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Showroom">
        <xs:annotation>
          <xs:documentation>Showroom</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Showroom</lx:path>
            <lx:label>Showroom</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Sign">
        <xs:annotation>
          <xs:documentation>Sign</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Sign</lx:path>
            <lx:label>Sign</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Site">
        <xs:annotation>
          <xs:documentation>Site</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Site</lx:path>
            <lx:label>Site</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Stall">
        <xs:annotation>
          <xs:documentation>Stall</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Stall</lx:path>
            <lx:label>Stall</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Store">
        <xs:annotation>
          <xs:documentation>Store</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Store</lx:path>
            <lx:label>Store</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Strata Unit">
        <xs:annotation>
          <xs:documentation>Strata Unit</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Strata Unit</lx:path>
            <lx:label>Strata Unit</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Studio / Studio Apartment">
        <xs:annotation>
          <xs:documentation>Studio / Studio Apartment</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Studio / Studio Apartment</lx:path>
            <lx:label>Studio / Studio Apartment</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Substation">
        <xs:annotation>
          <xs:documentation>Substation</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Substation</lx:path>
            <lx:label>Substation</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Suite">
        <xs:annotation>
          <xs:documentation>Suite</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Suite</lx:path>
            <lx:label>Suite</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tenancy">
        <xs:annotation>
          <xs:documentation>Tenancy</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Tenancy</lx:path>
            <lx:label>Tenancy</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tower">
        <xs:annotation>
          <xs:documentation>Tower</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Tower</lx:path>
            <lx:label>Tower</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Townhouse">
        <xs:annotation>
          <xs:documentation>Townhouse</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Townhouse</lx:path>
            <lx:label>Townhouse</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Unit">
        <xs:annotation>
          <xs:documentation>Unit</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Unit</lx:path>
            <lx:label>Unit</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vault">
        <xs:annotation>
          <xs:documentation>Vault</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Vault</lx:path>
            <lx:label>Vault</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Villa">
        <xs:annotation>
          <xs:documentation>Villa</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Villa</lx:path>
            <lx:label>Villa</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Ward">
        <xs:annotation>
          <xs:documentation>Ward</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Ward</lx:path>
            <lx:label>Ward</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Warehouse">
        <xs:annotation>
          <xs:documentation>Warehouse</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Warehouse</lx:path>
            <lx:label>Warehouse</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Workshop">
        <xs:annotation>
          <xs:documentation>Workshop</xs:documentation>
          <xs:appinfo>
            <lx:path>unitTypeList.Workshop</lx:path>
            <lx:label>Workshop</lx:label>
            <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="usageUnitOfMeasurementList">
    <xs:annotation>
      <xs:documentation>usageUnitOfMeasurementList</xs:documentation>
      <xs:appinfo>
        <lx:path>usageUnitOfMeasurementList</lx:path>
        <lx:label>usageUnitOfMeasurementList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Days">
        <xs:annotation>
          <xs:documentation>Days</xs:documentation>
          <xs:appinfo>
            <lx:path>usageUnitOfMeasurementList.Days</lx:path>
            <lx:label>Days</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Hours">
        <xs:annotation>
          <xs:documentation>Hours</xs:documentation>
          <xs:appinfo>
            <lx:path>usageUnitOfMeasurementList.Hours</lx:path>
            <lx:label>Hours</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Km">
        <xs:annotation>
          <xs:documentation>Km</xs:documentation>
          <xs:appinfo>
            <lx:path>usageUnitOfMeasurementList.Km</lx:path>
            <lx:label>Km</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Months">
        <xs:annotation>
          <xs:documentation>Months</xs:documentation>
          <xs:appinfo>
            <lx:path>usageUnitOfMeasurementList.Months</lx:path>
            <lx:label>Months</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Years">
        <xs:annotation>
          <xs:documentation>Years</xs:documentation>
          <xs:appinfo>
            <lx:path>usageUnitOfMeasurementList.Years</lx:path>
            <lx:label>Years</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="valuationProgramList">
    <xs:annotation>
      <xs:documentation>valuationProgramList</xs:documentation>
      <xs:appinfo>
        <lx:path>valuationProgramList</lx:path>
        <lx:label>valuationProgramList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="AVM">
        <xs:annotation>
          <xs:documentation>An AVM (Automated Valuation Model) requires no input from a qualified valuer, but instead use computer models based on attribute data and recent comparable sales data to estimate a value.</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationProgramList.AVM</lx:path>
            <lx:label>AVM</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Costing">
        <xs:annotation>
          <xs:documentation>A Costing is an evaluation of a construction contract to determine if the contract price is appropriate for the building being constructed, and to estimate the market value as if complete. Used for off the plan or To Be Erected (TBE) property.</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationProgramList.Costing</lx:path>
            <lx:label>Costing</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Desktop">
        <xs:annotation>
          <xs:documentation>A Desktop Valuation requires no internal or external inspection, instead qualified valuers use data about the property (attributes, photos, and comparable sales) to value a property.</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationProgramList.Desktop</lx:path>
            <lx:label>Desktop</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Full Registered">
        <xs:annotation>
          <xs:documentation>A Full Registered Valuation requires an internal and external inspection of the property by a qualified valuer.</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationProgramList.Full Registered</lx:path>
            <lx:label>Full Registered</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Kerb Side">
        <xs:annotation>
          <xs:documentation>A Kerb Side Valuation requires an external sighting of the property from the street by a qualified valuer.</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationProgramList.Kerb Side</lx:path>
            <lx:label>Kerb Side</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="None">
        <xs:annotation>
          <xs:documentation>None</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationProgramList.None</lx:path>
            <lx:label>None</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Progress Inspection">
        <xs:annotation>
          <xs:documentation>A Progress Inspection is an inspection and valuation of a building under construction.</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationProgramList.Progress Inspection</lx:path>
            <lx:label>Progress Inspection</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="valuationRiskRatingList">
    <xs:annotation>
      <xs:documentation>valuationRiskRatingList</xs:documentation>
      <xs:appinfo>
        <lx:path>valuationRiskRatingList</lx:path>
        <lx:label>valuationRiskRatingList</lx:label>
        <li:transactions>CDA,LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="0-NotKnown">
        <xs:annotation>
          <xs:documentation>0-NotKnown</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingList.0-NotKnown</lx:path>
            <lx:label>0-NotKnown</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="1-Low">
        <xs:annotation>
          <xs:documentation>1-Low</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingList.1-Low</lx:path>
            <lx:label>1-Low</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="2-LowToMedium">
        <xs:annotation>
          <xs:documentation>2-LowToMedium</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingList.2-LowToMedium</lx:path>
            <lx:label>2-LowToMedium</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="3-Medium">
        <xs:annotation>
          <xs:documentation>3-Medium</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingList.3-Medium</lx:path>
            <lx:label>3-Medium</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="4-MediumToHigh">
        <xs:annotation>
          <xs:documentation>4-MediumToHigh</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingList.4-MediumToHigh</lx:path>
            <lx:label>4-MediumToHigh</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="5-High">
        <xs:annotation>
          <xs:documentation>5-High</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingList.5-High</lx:path>
            <lx:label>5-High</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="valuationRiskRatingTypeList">
    <xs:annotation>
      <xs:documentation>valuationRiskRatingTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>valuationRiskRatingTypeList</lx:path>
        <lx:label>valuationRiskRatingTypeList</lx:label>
        <li:transactions>CDA,LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Environmental">
        <xs:annotation>
          <xs:documentation>Environmental</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingTypeList.Environmental</lx:path>
            <lx:label>Environmental</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Improvements">
        <xs:annotation>
          <xs:documentation>Improvements</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingTypeList.Improvements</lx:path>
            <lx:label>Improvements</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Land">
        <xs:annotation>
          <xs:documentation>Land</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingTypeList.Land</lx:path>
            <lx:label>Land</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LocalEconomy">
        <xs:annotation>
          <xs:documentation>LocalEconomy</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingTypeList.LocalEconomy</lx:path>
            <lx:label>LocalEconomy</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="LocationNeighbourhood">
        <xs:annotation>
          <xs:documentation>LocationNeighbourhood</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingTypeList.LocationNeighbourhood</lx:path>
            <lx:label>LocationNeighbourhood</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MarketSegment">
        <xs:annotation>
          <xs:documentation>MarketSegment</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingTypeList.MarketSegment</lx:path>
            <lx:label>MarketSegment</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="MarketVolatility">
        <xs:annotation>
          <xs:documentation>MarketVolatility</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingTypeList.MarketVolatility</lx:path>
            <lx:label>MarketVolatility</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="RecentMarketDirection">
        <xs:annotation>
          <xs:documentation>RecentMarketDirection</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationRiskRatingTypeList.RecentMarketDirection</lx:path>
            <lx:label>RecentMarketDirection</lx:label>
            <li:transactions>CDA,LMI,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="valuationVariationTypeList">
    <xs:annotation>
      <xs:documentation>valuationVariationTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>valuationVariationTypeList</lx:path>
        <lx:label>valuationVariationTypeList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Change Fee Amount">
        <xs:annotation>
          <xs:documentation>Change Fee Amount</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationVariationTypeList.Change Fee Amount</lx:path>
            <lx:label>Change Fee Amount</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Change Valuer">
        <xs:annotation>
          <xs:documentation>Change Valuer</xs:documentation>
          <xs:appinfo>
            <lx:path>valuationVariationTypeList.Change Valuer</lx:path>
            <lx:label>Change Valuer</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="variationTypeList">
    <xs:annotation>
      <xs:documentation>variationTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>variationTypeList</lx:path>
        <lx:label>variationTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Add Borrower">
        <xs:annotation>
          <xs:documentation>Add Borrower</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Add Borrower</lx:path>
            <lx:label>Add Borrower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Add Guarantee">
        <xs:annotation>
          <xs:documentation>Add a guarantee to an existing loan account.</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Add Guarantee</lx:path>
            <lx:label>Add Guarantee</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Add Guarantee Security">
        <xs:annotation>
          <xs:documentation>Add a security to an existing guarantee.</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Add Guarantee Security</lx:path>
            <lx:label>Add Guarantee Security</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Add Guarantor">
        <xs:annotation>
          <xs:documentation>Add Guarantor</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Add Guarantor</lx:path>
            <lx:label>Add Guarantor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Add Security">
        <xs:annotation>
          <xs:documentation>Add Security</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Add Security</lx:path>
            <lx:label>Add Security</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Balance Increase">
        <xs:annotation>
          <xs:documentation>Balance Increase</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Balance Increase</lx:path>
            <lx:label>Balance Increase</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Change Loan Term">
        <xs:annotation>
          <xs:documentation>Change Loan Term</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Change Loan Term</lx:path>
            <lx:label>Change Loan Term</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Change Repayment Amount">
        <xs:annotation>
          <xs:documentation>Change Repayment Amount</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Change Repayment Amount</lx:path>
            <lx:label>Change Repayment Amount</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Change Repayment Frequency">
        <xs:annotation>
          <xs:documentation>Change Repayment Frequency</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Change Repayment Frequency</lx:path>
            <lx:label>Change Repayment Frequency</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Change Repayment Type">
        <xs:annotation>
          <xs:documentation>Change Repayment Type</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Change Repayment Type</lx:path>
            <lx:label>Change Repayment Type</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Close Account">
        <xs:annotation>
          <xs:documentation>Close Account</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Close Account</lx:path>
            <lx:label>Close Account</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Interest Rate Discount">
        <xs:annotation>
          <xs:documentation>Interest Rate Discount</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Interest Rate Discount</lx:path>
            <lx:label>Interest Rate Discount</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lending Guarantee Limit Decrease">
        <xs:annotation>
          <xs:documentation>Lending Guarantee Limit Decrease</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Lending Guarantee Limit Decrease</lx:path>
            <lx:label>Lending Guarantee Limit Decrease</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lending Guarantee Limit Increase">
        <xs:annotation>
          <xs:documentation>Lending Guarantee Limit Increase</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Lending Guarantee Limit Increase</lx:path>
            <lx:label>Lending Guarantee Limit Increase</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Limit Increase">
        <xs:annotation>
          <xs:documentation>Limit Increase</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Limit Increase</lx:path>
            <lx:label>Limit Increase</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Reduce Balance">
        <xs:annotation>
          <xs:documentation>Reduce Balance</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Reduce Balance</lx:path>
            <lx:label>Reduce Balance</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Reduce Limit">
        <xs:annotation>
          <xs:documentation>Reduce Limit</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Reduce Limit</lx:path>
            <lx:label>Reduce Limit</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Release Guarantee">
        <xs:annotation>
          <xs:documentation>Release a guarantee from an existing loan account.</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Release Guarantee</lx:path>
            <lx:label>Release Guarantee</lx:label>
            <li:transactions>CAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Release Guarantee Security">
        <xs:annotation>
          <xs:documentation>Release a security from an existing guarantee.</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Release Guarantee Security</lx:path>
            <lx:label>Release Guarantee Security</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Release Guarantor">
        <xs:annotation>
          <xs:documentation>Release Guarantor</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Release Guarantor</lx:path>
            <lx:label>Release Guarantor</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Release Security">
        <xs:annotation>
          <xs:documentation>Release Security</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Release Security</lx:path>
            <lx:label>Release Security</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Remove Borrower">
        <xs:annotation>
          <xs:documentation>Remove Borrower</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Remove Borrower</lx:path>
            <lx:label>Remove Borrower</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Split Account">
        <xs:annotation>
          <xs:documentation>Split Account</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Split Account</lx:path>
            <lx:label>Split Account</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Substitute Guarantee Security">
        <xs:annotation>
          <xs:documentation>Substitute one security for another under an existing guarantee.</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Substitute Guarantee Security</lx:path>
            <lx:label>Substitute Guarantee Security</lx:label>
            <li:transactions>CAL,DAS</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Substitute Security">
        <xs:annotation>
          <xs:documentation>Substitute Security</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Substitute Security</lx:path>
            <lx:label>Substitute Security</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Switch Product">
        <xs:annotation>
          <xs:documentation>Switch Product</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Switch Product</lx:path>
            <lx:label>Switch Product</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Update Party Details">
        <xs:annotation>
          <xs:documentation>Update Party Details</xs:documentation>
          <xs:appinfo>
            <lx:path>variationTypeList.Update Party Details</lx:path>
            <lx:label>Update Party Details</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="varyOnValuationList">
    <xs:annotation>
      <xs:documentation>varyOnValuationList</xs:documentation>
      <xs:appinfo>
        <lx:path>varyOnValuationList</lx:path>
        <lx:label>varyOnValuationList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Cash Reserve">
        <xs:annotation>
          <xs:documentation>Vary the cash reserve amount.</xs:documentation>
          <xs:appinfo>
            <lx:path>varyOnValuationList.Cash Reserve</lx:path>
            <lx:label>Vary the cash reserve amount.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Instalments">
        <xs:annotation>
          <xs:documentation>Vary the instalment amount.</xs:documentation>
          <xs:appinfo>
            <lx:path>varyOnValuationList.Instalments</lx:path>
            <lx:label>Vary the instalment amount.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Lump Sum">
        <xs:annotation>
          <xs:documentation>Vary the lump sum amount.</xs:documentation>
          <xs:appinfo>
            <lx:path>varyOnValuationList.Lump Sum</lx:path>
            <lx:label>Vary the lump sum amount.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="vehicleSubtypeList">
    <xs:annotation>
      <xs:documentation>vehicleSubtypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>vehicleSubtypeList</lx:path>
        <lx:label>vehicleSubtypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Beaver Tail Truck">
        <xs:annotation>
          <xs:documentation>Beaver Tail Truck (subtype of Truck).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Beaver Tail Truck</lx:path>
            <lx:label>Beaver Tail Truck</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cab Chassis Truck">
        <xs:annotation>
          <xs:documentation>Cab Chassis Truck (subtype of Truck).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Cab Chassis Truck</lx:path>
            <lx:label>Cab Chassis Truck</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Cable Trailer">
        <xs:annotation>
          <xs:documentation>Cable Trailer (subtype of Trailer).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Cable Trailer</lx:path>
            <lx:label>Cable Trailer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Dog Trailer">
        <xs:annotation>
          <xs:documentation>Dog Trailer (subtype of Trailer).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Dog Trailer</lx:path>
            <lx:label>Dog Trailer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Flat Top Trailer">
        <xs:annotation>
          <xs:documentation>Flat Top Trailer (subtype of Trailer).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Flat Top Trailer</lx:path>
            <lx:label>Flat Top Trailer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Light Truck Under 4 and a Half Tons GVM">
        <xs:annotation>
          <xs:documentation>Light Truck Under 4 and a Half Tons GVM (subtype of Truck).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Light Truck Under 4 and a Half Tons GVM</lx:path>
            <lx:label>Light Truck Under 4 and a Half Tons GVM</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Low Loader Trailer">
        <xs:annotation>
          <xs:documentation>Low Loader Trailer (subtype of Trailer).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Low Loader Trailer</lx:path>
            <lx:label>Low Loader Trailer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Passenger">
        <xs:annotation>
          <xs:documentation>Passenger (subtype of Car).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Passenger</lx:path>
            <lx:label>Passenger</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Plant Trailer">
        <xs:annotation>
          <xs:documentation>Plant Trailer (subtype of Trailer).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Plant Trailer</lx:path>
            <lx:label>Plant Trailer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Prime Mover">
        <xs:annotation>
          <xs:documentation>Prime Mover (subtype of Truck).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Prime Mover</lx:path>
            <lx:label>Prime Mover</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Refrigerated Trailer">
        <xs:annotation>
          <xs:documentation>Refrigerated Trailer (subtype of Trailer).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Refrigerated Trailer</lx:path>
            <lx:label>Refrigerated Trailer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Skel Trailer">
        <xs:annotation>
          <xs:documentation>Skel Trailer (subtype of Trailer).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Skel Trailer</lx:path>
            <lx:label>Skel Trailer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="SUV">
        <xs:annotation>
          <xs:documentation>SUV (subtype of Car).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.SUV</lx:path>
            <lx:label>SUV</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tautliner Trailer">
        <xs:annotation>
          <xs:documentation>Tautliner Trailer (subtype of Trailer).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Tautliner Trailer</lx:path>
            <lx:label>Tautliner Trailer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tipper Truck">
        <xs:annotation>
          <xs:documentation>Tipper Truck (subtype of Truck).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Tipper Truck</lx:path>
            <lx:label>Tipper Truck</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trailer-Mounted Crane">
        <xs:annotation>
          <xs:documentation>Trailer-Mounted Crane (subtype of Trailer).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Trailer-Mounted Crane</lx:path>
            <lx:label>Trailer-Mounted Crane</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trailer-Mounted Vacuum Unit">
        <xs:annotation>
          <xs:documentation>Trailer-Mounted Vacuum Unit (subtype of Trailer).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Trailer-Mounted Vacuum Unit</lx:path>
            <lx:label>Trailer-Mounted Vacuum Unit</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trayback Truck">
        <xs:annotation>
          <xs:documentation>Trayback Truck (subtype of Truck).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Trayback Truck</lx:path>
            <lx:label>Trayback Truck</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Truck-Mounted Crane">
        <xs:annotation>
          <xs:documentation>Truck-Mounted Crane (subtype of Truck).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Truck-Mounted Crane</lx:path>
            <lx:label>Truck-Mounted Crane</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Utility">
        <xs:annotation>
          <xs:documentation>Utility (subtype of Car).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Utility</lx:path>
            <lx:label>Utility</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vacuum Truck">
        <xs:annotation>
          <xs:documentation>Vacuum Truck (subtype of Truck).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Vacuum Truck</lx:path>
            <lx:label>Vacuum Truck</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Water Truck">
        <xs:annotation>
          <xs:documentation>Water Truck (subtype of Truck).</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleSubtypeList.Water Truck</lx:path>
            <lx:label>Water Truck</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="vehicleTypeList">
    <xs:annotation>
      <xs:documentation>vehicleTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>vehicleTypeList</lx:path>
        <lx:label>vehicleTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Bicycle">
        <xs:annotation>
          <xs:documentation>Bicycle</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleTypeList.Bicycle</lx:path>
            <lx:label>Bicycle</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bodies">
        <xs:annotation>
          <xs:documentation>Bodies</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleTypeList.Bodies</lx:path>
            <lx:label>Bodies</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Bus Coach or Special Purpose">
        <xs:annotation>
          <xs:documentation>Buses, coaches, tow trucks, hearses, motorised caravans and other non-freight carrying vehicles</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleTypeList.Bus Coach or Special Purpose</lx:path>
            <lx:label>Buses, coaches, tow trucks, hearses, motorised caravans and other non-freight carrying vehicles</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Car">
        <xs:annotation>
          <xs:documentation>Cars, station wagons, utilities, SUV, and other small passenger vehicles and vans</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleTypeList.Car</lx:path>
            <lx:label>Cars, station wagons, utilities, SUV, and other small passenger vehicles and vans</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Caravan">
        <xs:annotation>
          <xs:documentation>Wheeled vehicle, used to live in, typically towed by a car.</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleTypeList.Caravan</lx:path>
            <lx:label>Wheeled vehicle, used to live in, typically towed by a car.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Light Commercial">
        <xs:annotation>
          <xs:documentation>Vehicles constructed primarily for the carriage of goods and not exceeding 3.5 tonnes</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleTypeList.Light Commercial</lx:path>
            <lx:label>Vehicles constructed primarily for the carriage of goods and not exceeding 3.5 tonnes</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Locomotive">
        <xs:annotation>
          <xs:documentation>Locomotive</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleTypeList.Locomotive</lx:path>
            <lx:label>Locomotive</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Motorcycle">
        <xs:annotation>
          <xs:documentation>Motorcycle</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleTypeList.Motorcycle</lx:path>
            <lx:label>Motorcycle</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleTypeList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Rail Cars">
        <xs:annotation>
          <xs:documentation>Rail rolling stock</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleTypeList.Rail Cars</lx:path>
            <lx:label>Rail rolling stock</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Trailer">
        <xs:annotation>
          <xs:documentation>Trailer</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleTypeList.Trailer</lx:path>
            <lx:label>Trailer</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Truck">
        <xs:annotation>
          <xs:documentation>Freight carrying vehicles with a gross combination mass rating in excess of 3.5 tonnes including prime movers</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleTypeList.Truck</lx:path>
            <lx:label>Freight carrying vehicles with a gross combination mass rating in excess of 3.5 tonnes including prime movers</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Watercraft">
        <xs:annotation>
          <xs:documentation>Including personal watercraft, sail and power boats of any size</xs:documentation>
          <xs:appinfo>
            <lx:path>vehicleTypeList.Watercraft</lx:path>
            <lx:label>Including personal watercraft, sail and power boats of any size</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="visitContactTypeList">
    <xs:annotation>
      <xs:documentation>visitContactTypeList</xs:documentation>
      <xs:appinfo>
        <lx:path>visitContactTypeList</lx:path>
        <lx:label>visitContactTypeList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Applicant">
        <xs:annotation>
          <xs:documentation>The contact for arranging inspection of the property is the applicant.</xs:documentation>
          <xs:appinfo>
            <lx:path>visitContactTypeList.Applicant</lx:path>
            <lx:label>The contact for arranging inspection of the property is the applicant.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Builder">
        <xs:annotation>
          <xs:documentation>The contact for arranging the inspection of the property is the builder.</xs:documentation>
          <xs:appinfo>
            <lx:path>visitContactTypeList.Builder</lx:path>
            <lx:label>The contact for arranging the inspection of the property is the builder.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Property Agent">
        <xs:annotation>
          <xs:documentation>The contact for arranging inspection of the property is the real estate agent acting for the vendor of the property. Applicable when the property is being purchased by the applicant.</xs:documentation>
          <xs:appinfo>
            <lx:path>visitContactTypeList.Property Agent</lx:path>
            <lx:label>The contact for arranging inspection of the property is the real estate agent acting for the vendor of the property. Applicable when the property is being purchased by the applicant.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Tenant">
        <xs:annotation>
          <xs:documentation>The contact for arranging inspection of the property is the tenant that resides in the property. Applicable when the property is leased.</xs:documentation>
          <xs:appinfo>
            <lx:path>visitContactTypeList.Tenant</lx:path>
            <lx:label>The contact for arranging inspection of the property is the tenant that resides in the property. Applicable when the property is leased.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vendor">
        <xs:annotation>
          <xs:documentation>The contact for arranging inspection of the property is the seller.</xs:documentation>
          <xs:appinfo>
            <lx:path>visitContactTypeList.Vendor</lx:path>
            <lx:label>The contact for arranging inspection of the property is the seller.</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="wallMaterialList">
    <xs:annotation>
      <xs:documentation>A list of wall meterial types.</xs:documentation>
      <xs:appinfo>
        <lx:path>wallMaterialList</lx:path>
        <lx:label>wallMaterialList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Besser Block">
        <xs:annotation>
          <xs:documentation>Besser Block</xs:documentation>
          <xs:appinfo>
            <lx:path>wallMaterialList.Besser Block</lx:path>
            <lx:label>Besser Block</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Brick">
        <xs:annotation>
          <xs:documentation>Brick</xs:documentation>
          <xs:appinfo>
            <lx:path>wallMaterialList.Brick</lx:path>
            <lx:label>Brick</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Concrete">
        <xs:annotation>
          <xs:documentation>Concrete</xs:documentation>
          <xs:appinfo>
            <lx:path>wallMaterialList.Concrete</lx:path>
            <lx:label>Concrete</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Log">
        <xs:annotation>
          <xs:documentation>Log</xs:documentation>
          <xs:appinfo>
            <lx:path>wallMaterialList.Log</lx:path>
            <lx:label>Log</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Metal Cladding">
        <xs:annotation>
          <xs:documentation>Metal Cladding</xs:documentation>
          <xs:appinfo>
            <lx:path>wallMaterialList.Metal Cladding</lx:path>
            <lx:label>Metal Cladding</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Mud Brick">
        <xs:annotation>
          <xs:documentation>Mud Brick</xs:documentation>
          <xs:appinfo>
            <lx:path>wallMaterialList.Mud Brick</lx:path>
            <lx:label>Mud Brick</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>wallMaterialList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Stone">
        <xs:annotation>
          <xs:documentation>Stone</xs:documentation>
          <xs:appinfo>
            <lx:path>wallMaterialList.Stone</lx:path>
            <lx:label>Stone</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Stone Cladding">
        <xs:annotation>
          <xs:documentation>Stone Cladding</xs:documentation>
          <xs:appinfo>
            <lx:path>wallMaterialList.Stone Cladding</lx:path>
            <lx:label>Stone Cladding</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Straw Bale">
        <xs:annotation>
          <xs:documentation>Straw Bale</xs:documentation>
          <xs:appinfo>
            <lx:path>wallMaterialList.Straw Bale</lx:path>
            <lx:label>Straw Bale</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vinyl Cladding">
        <xs:annotation>
          <xs:documentation>Vinyl Cladding</xs:documentation>
          <xs:appinfo>
            <lx:path>wallMaterialList.Vinyl Cladding</lx:path>
            <lx:label>Vinyl Cladding</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wood Cladding">
        <xs:annotation>
          <xs:documentation>Wood Cladding</xs:documentation>
          <xs:appinfo>
            <lx:path>wallMaterialList.Wood Cladding</lx:path>
            <lx:label>Wood Cladding</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="weekList">
    <xs:annotation>
      <xs:documentation>A list of the weeks in each period.</xs:documentation>
      <xs:appinfo>
        <lx:path>weekList</lx:path>
        <lx:label>weekList</lx:label>
        <li:transactions>CAL,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="First">
        <xs:annotation>
          <xs:documentation>The first week of the period.</xs:documentation>
          <xs:appinfo>
            <lx:path>weekList.First</lx:path>
            <lx:label>First</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fourth">
        <xs:annotation>
          <xs:documentation>The fourth week of the period.</xs:documentation>
          <xs:appinfo>
            <lx:path>weekList.Fourth</lx:path>
            <lx:label>Fourth</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Last">
        <xs:annotation>
          <xs:documentation>The last week of the period.</xs:documentation>
          <xs:appinfo>
            <lx:path>weekList.Last</lx:path>
            <lx:label>Last</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Second">
        <xs:annotation>
          <xs:documentation>The second week of the period.</xs:documentation>
          <xs:appinfo>
            <lx:path>weekList.Second</lx:path>
            <lx:label>Second</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Third">
        <xs:annotation>
          <xs:documentation>The third week of the period.</xs:documentation>
          <xs:appinfo>
            <lx:path>weekList.Third</lx:path>
            <lx:label>Third</lx:label>
            <li:transactions>CAL,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="windowMaterialList">
    <xs:annotation>
      <xs:documentation>A list of window material types.</xs:documentation>
      <xs:appinfo>
        <lx:path>windowMaterialList</lx:path>
        <lx:label>windowMaterialList</lx:label>
        <li:transactions>VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Aluminium">
        <xs:annotation>
          <xs:documentation>Aluminium</xs:documentation>
          <xs:appinfo>
            <lx:path>windowMaterialList.Aluminium</lx:path>
            <lx:label>Aluminium</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Composite">
        <xs:annotation>
          <xs:documentation>Composite</xs:documentation>
          <xs:appinfo>
            <lx:path>windowMaterialList.Composite</lx:path>
            <lx:label>Composite</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Fibreglass">
        <xs:annotation>
          <xs:documentation>Fibreglass</xs:documentation>
          <xs:appinfo>
            <lx:path>windowMaterialList.Fibreglass</lx:path>
            <lx:label>Fibreglass</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Other">
        <xs:annotation>
          <xs:documentation>Other</xs:documentation>
          <xs:appinfo>
            <lx:path>windowMaterialList.Other</lx:path>
            <lx:label>Other</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Vinyl">
        <xs:annotation>
          <xs:documentation>Vinyl</xs:documentation>
          <xs:appinfo>
            <lx:path>windowMaterialList.Vinyl</lx:path>
            <lx:label>Vinyl</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Wood">
        <xs:annotation>
          <xs:documentation>Wood</xs:documentation>
          <xs:appinfo>
            <lx:path>windowMaterialList.Wood</lx:path>
            <lx:label>Wood</lx:label>
            <li:transactions>VAL</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="yesNoIntentList">
    <xs:annotation>
      <xs:documentation>A list of options to indicate answers of Yes, No, Future Intent.</xs:documentation>
      <xs:appinfo>
        <lx:path>yesNoIntentList</lx:path>
        <lx:label>yes No Intent List</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="Future Intent">
        <xs:annotation>
          <xs:documentation>In the future the response will be Yes, indicating future intent.</xs:documentation>
          <xs:appinfo>
            <lx:path>yesNoIntentList.Future Intent</lx:path>
            <lx:label>Future Intent</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="No">
        <xs:annotation>
          <xs:documentation>Currently No</xs:documentation>
          <xs:appinfo>
            <lx:path>yesNoIntentList.No</lx:path>
            <lx:label>No</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Yes">
        <xs:annotation>
          <xs:documentation>Currently Yes</xs:documentation>
          <xs:appinfo>
            <lx:path>yesNoIntentList.Yes</lx:path>
            <lx:label>Yes</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="yesNoList">
    <xs:annotation>
      <xs:documentation>yesNoList</xs:documentation>
      <xs:appinfo>
        <lx:path>yesNoList</lx:path>
        <lx:label>yesNoList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="No">
        <xs:annotation>
          <xs:documentation>No</xs:documentation>
          <xs:appinfo>
            <lx:path>yesNoList.No</lx:path>
            <lx:label>No</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="Yes">
        <xs:annotation>
          <xs:documentation>Yes</xs:documentation>
          <xs:appinfo>
            <lx:path>yesNoList.Yes</lx:path>
            <lx:label>Yes</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="yyYearList">
    <xs:annotation>
      <xs:documentation>yyYearList</xs:documentation>
      <xs:appinfo>
        <lx:path>yyYearList</lx:path>
        <lx:label>yyYearList</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:token">
      <xs:enumeration value="00">
        <xs:annotation>
          <xs:documentation>00</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.00</lx:path>
            <lx:label>00</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="01">
        <xs:annotation>
          <xs:documentation>01</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.01</lx:path>
            <lx:label>01</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="02">
        <xs:annotation>
          <xs:documentation>02</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.02</lx:path>
            <lx:label>02</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="03">
        <xs:annotation>
          <xs:documentation>03</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.03</lx:path>
            <lx:label>03</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="04">
        <xs:annotation>
          <xs:documentation>04</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.04</lx:path>
            <lx:label>04</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="05">
        <xs:annotation>
          <xs:documentation>05</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.05</lx:path>
            <lx:label>05</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="06">
        <xs:annotation>
          <xs:documentation>06</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.06</lx:path>
            <lx:label>06</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="07">
        <xs:annotation>
          <xs:documentation>07</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.07</lx:path>
            <lx:label>07</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="08">
        <xs:annotation>
          <xs:documentation>08</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.08</lx:path>
            <lx:label>08</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="09">
        <xs:annotation>
          <xs:documentation>09</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.09</lx:path>
            <lx:label>09</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="10">
        <xs:annotation>
          <xs:documentation>10</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.10</lx:path>
            <lx:label>10</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="11">
        <xs:annotation>
          <xs:documentation>11</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.11</lx:path>
            <lx:label>11</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="12">
        <xs:annotation>
          <xs:documentation>12</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.12</lx:path>
            <lx:label>12</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="13">
        <xs:annotation>
          <xs:documentation>13</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.13</lx:path>
            <lx:label>13</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="14">
        <xs:annotation>
          <xs:documentation>14</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.14</lx:path>
            <lx:label>14</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="15">
        <xs:annotation>
          <xs:documentation>15</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.15</lx:path>
            <lx:label>15</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="16">
        <xs:annotation>
          <xs:documentation>16</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.16</lx:path>
            <lx:label>16</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="17">
        <xs:annotation>
          <xs:documentation>17</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.17</lx:path>
            <lx:label>17</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="18">
        <xs:annotation>
          <xs:documentation>18</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.18</lx:path>
            <lx:label>18</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="19">
        <xs:annotation>
          <xs:documentation>19</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.19</lx:path>
            <lx:label>19</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="20">
        <xs:annotation>
          <xs:documentation>20</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.20</lx:path>
            <lx:label>20</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="21">
        <xs:annotation>
          <xs:documentation>21</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.21</lx:path>
            <lx:label>21</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="22">
        <xs:annotation>
          <xs:documentation>22</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.22</lx:path>
            <lx:label>22</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="23">
        <xs:annotation>
          <xs:documentation>23</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.23</lx:path>
            <lx:label>23</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="24">
        <xs:annotation>
          <xs:documentation>24</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.24</lx:path>
            <lx:label>24</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="25">
        <xs:annotation>
          <xs:documentation>25</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.25</lx:path>
            <lx:label>25</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="26">
        <xs:annotation>
          <xs:documentation>26</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.26</lx:path>
            <lx:label>26</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="27">
        <xs:annotation>
          <xs:documentation>27</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.27</lx:path>
            <lx:label>27</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="28">
        <xs:annotation>
          <xs:documentation>28</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.28</lx:path>
            <lx:label>28</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="29">
        <xs:annotation>
          <xs:documentation>29</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.29</lx:path>
            <lx:label>29</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="30">
        <xs:annotation>
          <xs:documentation>30</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.30</lx:path>
            <lx:label>30</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="31">
        <xs:annotation>
          <xs:documentation>31</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.31</lx:path>
            <lx:label>31</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="32">
        <xs:annotation>
          <xs:documentation>32</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.32</lx:path>
            <lx:label>32</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="33">
        <xs:annotation>
          <xs:documentation>33</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.33</lx:path>
            <lx:label>33</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="34">
        <xs:annotation>
          <xs:documentation>34</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.34</lx:path>
            <lx:label>34</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="35">
        <xs:annotation>
          <xs:documentation>35</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.35</lx:path>
            <lx:label>35</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="36">
        <xs:annotation>
          <xs:documentation>36</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.36</lx:path>
            <lx:label>36</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="37">
        <xs:annotation>
          <xs:documentation>37</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.37</lx:path>
            <lx:label>37</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="38">
        <xs:annotation>
          <xs:documentation>38</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.38</lx:path>
            <lx:label>38</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="39">
        <xs:annotation>
          <xs:documentation>39</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.39</lx:path>
            <lx:label>39</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="40">
        <xs:annotation>
          <xs:documentation>40</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.40</lx:path>
            <lx:label>40</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="41">
        <xs:annotation>
          <xs:documentation>41</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.41</lx:path>
            <lx:label>41</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="42">
        <xs:annotation>
          <xs:documentation>42</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.42</lx:path>
            <lx:label>42</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="43">
        <xs:annotation>
          <xs:documentation>43</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.43</lx:path>
            <lx:label>43</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="44">
        <xs:annotation>
          <xs:documentation>44</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.44</lx:path>
            <lx:label>44</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="45">
        <xs:annotation>
          <xs:documentation>45</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.45</lx:path>
            <lx:label>45</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="46">
        <xs:annotation>
          <xs:documentation>46</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.46</lx:path>
            <lx:label>46</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="47">
        <xs:annotation>
          <xs:documentation>47</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.47</lx:path>
            <lx:label>47</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="48">
        <xs:annotation>
          <xs:documentation>48</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.48</lx:path>
            <lx:label>48</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="49">
        <xs:annotation>
          <xs:documentation>49</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.49</lx:path>
            <lx:label>49</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="50">
        <xs:annotation>
          <xs:documentation>50</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.50</lx:path>
            <lx:label>50</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="51">
        <xs:annotation>
          <xs:documentation>51</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.51</lx:path>
            <lx:label>51</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="52">
        <xs:annotation>
          <xs:documentation>52</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.52</lx:path>
            <lx:label>52</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="53">
        <xs:annotation>
          <xs:documentation>53</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.53</lx:path>
            <lx:label>53</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="54">
        <xs:annotation>
          <xs:documentation>54</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.54</lx:path>
            <lx:label>54</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="55">
        <xs:annotation>
          <xs:documentation>55</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.55</lx:path>
            <lx:label>55</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="56">
        <xs:annotation>
          <xs:documentation>56</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.56</lx:path>
            <lx:label>56</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="57">
        <xs:annotation>
          <xs:documentation>57</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.57</lx:path>
            <lx:label>57</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="58">
        <xs:annotation>
          <xs:documentation>58</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.58</lx:path>
            <lx:label>58</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="59">
        <xs:annotation>
          <xs:documentation>59</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.59</lx:path>
            <lx:label>59</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="60">
        <xs:annotation>
          <xs:documentation>60</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.60</lx:path>
            <lx:label>60</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="61">
        <xs:annotation>
          <xs:documentation>61</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.61</lx:path>
            <lx:label>61</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="62">
        <xs:annotation>
          <xs:documentation>62</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.62</lx:path>
            <lx:label>62</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="63">
        <xs:annotation>
          <xs:documentation>63</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.63</lx:path>
            <lx:label>63</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="64">
        <xs:annotation>
          <xs:documentation>64</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.64</lx:path>
            <lx:label>64</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="65">
        <xs:annotation>
          <xs:documentation>65</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.65</lx:path>
            <lx:label>65</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="66">
        <xs:annotation>
          <xs:documentation>66</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.66</lx:path>
            <lx:label>66</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="67">
        <xs:annotation>
          <xs:documentation>67</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.67</lx:path>
            <lx:label>67</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="68">
        <xs:annotation>
          <xs:documentation>68</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.68</lx:path>
            <lx:label>68</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="69">
        <xs:annotation>
          <xs:documentation>69</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.69</lx:path>
            <lx:label>69</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="70">
        <xs:annotation>
          <xs:documentation>70</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.70</lx:path>
            <lx:label>70</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="71">
        <xs:annotation>
          <xs:documentation>71</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.71</lx:path>
            <lx:label>71</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="72">
        <xs:annotation>
          <xs:documentation>72</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.72</lx:path>
            <lx:label>72</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="73">
        <xs:annotation>
          <xs:documentation>73</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.73</lx:path>
            <lx:label>73</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="74">
        <xs:annotation>
          <xs:documentation>74</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.74</lx:path>
            <lx:label>74</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="75">
        <xs:annotation>
          <xs:documentation>75</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.75</lx:path>
            <lx:label>75</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="76">
        <xs:annotation>
          <xs:documentation>76</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.76</lx:path>
            <lx:label>76</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="77">
        <xs:annotation>
          <xs:documentation>77</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.77</lx:path>
            <lx:label>77</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="78">
        <xs:annotation>
          <xs:documentation>78</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.78</lx:path>
            <lx:label>78</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="79">
        <xs:annotation>
          <xs:documentation>79</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.79</lx:path>
            <lx:label>79</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="80">
        <xs:annotation>
          <xs:documentation>80</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.80</lx:path>
            <lx:label>80</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="81">
        <xs:annotation>
          <xs:documentation>81</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.81</lx:path>
            <lx:label>81</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="82">
        <xs:annotation>
          <xs:documentation>82</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.82</lx:path>
            <lx:label>82</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="83">
        <xs:annotation>
          <xs:documentation>83</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.83</lx:path>
            <lx:label>83</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="84">
        <xs:annotation>
          <xs:documentation>84</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.84</lx:path>
            <lx:label>84</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="85">
        <xs:annotation>
          <xs:documentation>85</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.85</lx:path>
            <lx:label>85</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="86">
        <xs:annotation>
          <xs:documentation>86</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.86</lx:path>
            <lx:label>86</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="87">
        <xs:annotation>
          <xs:documentation>87</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.87</lx:path>
            <lx:label>87</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="88">
        <xs:annotation>
          <xs:documentation>88</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.88</lx:path>
            <lx:label>88</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="89">
        <xs:annotation>
          <xs:documentation>89</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.89</lx:path>
            <lx:label>89</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="90">
        <xs:annotation>
          <xs:documentation>90</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.90</lx:path>
            <lx:label>90</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="91">
        <xs:annotation>
          <xs:documentation>91</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.91</lx:path>
            <lx:label>91</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="92">
        <xs:annotation>
          <xs:documentation>92</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.92</lx:path>
            <lx:label>92</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="93">
        <xs:annotation>
          <xs:documentation>93</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.93</lx:path>
            <lx:label>93</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="94">
        <xs:annotation>
          <xs:documentation>94</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.94</lx:path>
            <lx:label>94</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="95">
        <xs:annotation>
          <xs:documentation>95</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.95</lx:path>
            <lx:label>95</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="96">
        <xs:annotation>
          <xs:documentation>96</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.96</lx:path>
            <lx:label>96</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="97">
        <xs:annotation>
          <xs:documentation>97</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.97</lx:path>
            <lx:label>97</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="98">
        <xs:annotation>
          <xs:documentation>98</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.98</lx:path>
            <lx:label>98</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
      <xs:enumeration value="99">
        <xs:annotation>
          <xs:documentation>99</xs:documentation>
          <xs:appinfo>
            <lx:path>yyYearList.99</lx:path>
            <lx:label>99</lx:label>
            <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
          </xs:appinfo>
        </xs:annotation>
      </xs:enumeration>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="abnPattern">
    <xs:annotation>
      <xs:documentation>Enforces that a value entered for Australian Business Number (ABN) is an eleven-digit number (with no spaces).</xs:documentation>
      <xs:appinfo>
        <lx:path>abnPattern</lx:path>
        <lx:label>abnPattern</lx:label>
        <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="\d{11}"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="acnPattern">
    <xs:annotation>
      <xs:documentation>Enforces that a value entered for Australian Company Number (ACN) is a nine-digit number (with no spaces).</xs:documentation>
      <xs:appinfo>
        <lx:path>acnPattern</lx:path>
        <lx:label>acnPattern</lx:label>
        <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="\d{9}"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="arbnPattern">
    <xs:annotation>
      <xs:documentation>Enforces that a value entered for Australian Registered Body Number (ARBN) is a nine-digit number (with no spaces).</xs:documentation>
      <xs:appinfo>
        <lx:path>arbnPattern</lx:path>
        <lx:label>arbnPattern</lx:label>
        <li:transactions>CAL,CDA,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="\d{9}"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="arsnPattern">
    <xs:annotation>
      <xs:documentation>Enforces that a value entered for Australian Registered Scheme Number (ARSN) is a nine-digit number (with no spaces).</xs:documentation>
      <xs:appinfo>
        <lx:path>arsnPattern</lx:path>
        <lx:label>arsnPattern</lx:label>
        <li:transactions>CAL,CDA,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="\d{9}"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="aubicPattern">
    <xs:annotation>
      <xs:documentation>A pattern to restrict a Business Industry Code to be exactly 5 numerical digits.</xs:documentation>
      <xs:appinfo>
        <lx:path>aubicPattern</lx:path>
        <lx:label>aubicPattern</lx:label>
        <li:transactions>CAL,CDA,DAS</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="\d{5}"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="auPostCodePattern">
    <xs:annotation>
      <xs:documentation>A four digit numeric descriptor for a postal delivery area, aligned with placename, suburb or locality and in some circumstances a unique Postal Delivery Type.</xs:documentation>
      <xs:appinfo>
        <lx:path>auPostCodePattern</lx:path>
        <lx:label>auPostCodePattern</lx:label>
        <li:transactions>CAL,CDA,DAS,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="\d{4}"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="bsbPattern">
    <xs:annotation>
      <xs:documentation>A pattern to restrict BSB values to be exactly 6 numerical digits.</xs:documentation>
      <xs:appinfo>
        <lx:path>bsbPattern</lx:path>
        <lx:label>bsbPattern</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="\d{6}"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="dayOfMonthPattern">
    <xs:annotation>
      <xs:documentation>A pattern restricting the day of the month to a number between 1 and 31.</xs:documentation>
      <xs:appinfo>
        <lx:path>dayOfMonthPattern</lx:path>
        <lx:label>dayOfMonthPattern</lx:label>
        <li:transactions>CAL,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="0?[1-9]|[12][0-9]|3[01]"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="emailPattern">
    <xs:annotation>
      <xs:documentation>Messages distributed by electronic means from one computer user to one or more recipients via a network.</xs:documentation>
      <xs:appinfo>
        <lx:path>emailPattern</lx:path>
        <lx:label>emailPattern</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value=".+@.+\..+"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="gicsCodePattern">
    <xs:annotation>
      <xs:documentation>A pattern to restrict a Global Industry Classification Standard (GICS) code to be either 2, 4, 6, or 8 numerical digits.</xs:documentation>
      <xs:appinfo>
        <lx:path>gicsCodePattern</lx:path>
        <lx:label>gicsCodePattern</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="\d{2}(\d{2})?(\d{2})?(\d{2})?"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="irdNumberPattern">
    <xs:annotation>
      <xs:documentation>The Inland Revenue Department Number (or IRD Number) is a unique identifier (consisting of 8 or 9 numbers) issued by the New Zealand Inland Revenue Department.</xs:documentation>
      <xs:appinfo>
        <lx:path>irdNumberPattern</lx:path>
        <lx:label>IRD Number Pattern</lx:label>
        <li:transactions>CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="\d{8,9}"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="nzbicPattern">
    <xs:annotation>
      <xs:documentation>A pattern to restrict a Business Industry Classification Code to be exactly 1 alphabetic character followed by 6 numerical digits.</xs:documentation>
      <xs:appinfo>
        <lx:path>nzbicPattern</lx:path>
        <lx:label>nzbicPattern</lx:label>
        <li:transactions>CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="[A-z]\d{6}"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="nzbnPattern">
    <xs:annotation>
      <xs:documentation>The New Zealand Business Number (NZBN) is a universal identifier which enables businesses to easily update, share key information and interact with each other.</xs:documentation>
      <xs:appinfo>
        <lx:path>nzbnPattern</lx:path>
        <lx:label>nzbnPattern</lx:label>
        <li:transactions>CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="\d{13}"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="nzPostCodePattern">
    <xs:annotation>
      <xs:documentation>A four digit numeric descriptor that helps to identify delivery addresses in New Zealand when there are the same or similar street and town names used nationwide.</xs:documentation>
      <xs:appinfo>
        <lx:path>nzPostCodePattern</lx:path>
        <lx:label>nzPostCodePattern</lx:label>
        <li:transactions>CNZ</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="\d{4}"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="urlPattern">
    <xs:annotation>
      <xs:documentation>The uniform resource locator, is a specific character string that constitutes a reference to a resource. In most web browsers, the URL of a web page is displayed on top inside an address bar.</xs:documentation>
      <xs:appinfo>
        <lx:path>urlPattern</lx:path>
        <lx:label>urlPattern</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value=".+\..+"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="visaSubclassCodePattern">
    <xs:annotation>
      <xs:documentation>Enforces that a value entered for visa subclass code is three digit number.</xs:documentation>
      <xs:appinfo>
        <lx:path>visaSubclassCodePattern</lx:path>
        <lx:label>visaSubclassCodePattern</lx:label>
        <li:transactions>CAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string">
      <xs:pattern value="\d{3}"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="base64BinaryType">
    <xs:annotation>
      <xs:documentation>base64BinaryType is a restriction of the base64Binary xml data type.</xs:documentation>
      <xs:appinfo>
        <lx:path>base64BinaryType</lx:path>
        <lx:label>base64BinaryType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:base64Binary"/>
  </xs:simpleType>
  <xs:simpleType name="currencyType">
    <xs:annotation>
      <xs:documentation>Local type that is based on a restriction of the decimal xml data type.</xs:documentation>
      <xs:appinfo>
        <lx:path>currencyType</lx:path>
        <lx:label>currencyType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:decimal"/>
  </xs:simpleType>
  <xs:simpleType name="dateTimeType">
    <xs:annotation>
      <xs:documentation>Local type that is based on a restriction of the dateTime xml data type.</xs:documentation>
      <xs:appinfo>
        <lx:path>dateTimeType</lx:path>
        <lx:label>dateTimeType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:dateTime"/>
  </xs:simpleType>
  <xs:simpleType name="dateType">
    <xs:annotation>
      <xs:documentation>Local type that is based on a restriction of the date xml data type.</xs:documentation>
      <xs:appinfo>
        <lx:path>dateType</lx:path>
        <lx:label>dateType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:date"/>
  </xs:simpleType>
  <xs:simpleType name="decimalType">
    <xs:annotation>
      <xs:documentation>Local type that is based on a restriction of the decimal xml data type.</xs:documentation>
      <xs:appinfo>
        <lx:path>decimalType</lx:path>
        <lx:label>decimalType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:decimal"/>
  </xs:simpleType>
  <xs:simpleType name="gYearType">
    <xs:annotation>
      <xs:documentation>Local type that is based on a restriction of the gYear xml data type.</xs:documentation>
      <xs:appinfo>
        <lx:path>gYearType</lx:path>
        <lx:label>gYearType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:gYear"/>
  </xs:simpleType>
  <xs:simpleType name="integerType">
    <xs:annotation>
      <xs:documentation>An integer is a number that is written without a fractional component. For example, 21, 4, and -2048 are integers; 9.75 and 51/2 are not integers.</xs:documentation>
      <xs:appinfo>
        <lx:path>integerType</lx:path>
        <lx:label>integerType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:integer"/>
  </xs:simpleType>
  <xs:simpleType name="percentType">
    <xs:annotation>
      <xs:documentation>Local type that is a restriction of the decimal xml data type.</xs:documentation>
      <xs:appinfo>
        <lx:path>percentType</lx:path>
        <lx:label>percentType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:decimal"/>
  </xs:simpleType>
  <xs:simpleType name="referenceType">
    <xs:annotation>
      <xs:documentation>Must point to a Unique ID</xs:documentation>
      <xs:appinfo>
        <lx:path>referenceType</lx:path>
        <lx:label>referenceType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:IDREF"/>
  </xs:simpleType>
  <xs:simpleType name="stringType">
    <xs:annotation>
      <xs:documentation>Local type that is a restriction of the string xml data type.</xs:documentation>
      <xs:appinfo>
        <lx:path>stringType</lx:path>
        <lx:label>stringType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:string"/>
  </xs:simpleType>
  <xs:simpleType name="uniqueIDType">
    <xs:annotation>
      <xs:documentation>Value must be unique within the scope, e.g. Application, Valuation, LMI response, etc. Until namespaces are implemented, the uniqueness will span Package.</xs:documentation>
      <xs:appinfo>
        <lx:path>uniqueIDType</lx:path>
        <lx:label>uniqueIDType</lx:label>
        <li:transactions>CAL,CDA,DAS,CNZ,SVC,LMI,VAL,ACC</li:transactions>
      </xs:appinfo>
    </xs:annotation>
    <xs:restriction base="xs:ID"/>
  </xs:simpleType>
</xs:schema>
`